static VENDORS: phf::Map<u16, Vendor> = ::phf::Map {
    key: 12913932095322966823,
    disps: &[
        (0, 1),
        (0, 433),
        (0, 8),
        (0, 154),
        (0, 32),
        (0, 5),
        (0, 332),
        (0, 120),
        (0, 3),
        (0, 38),
        (0, 60),
        (0, 24),
        (0, 0),
        (0, 343),
        (0, 154),
        (0, 363),
        (0, 0),
        (0, 392),
        (0, 51),
        (0, 89),
        (0, 102),
        (0, 10),
        (0, 0),
        (0, 1032),
        (0, 127),
        (0, 130),
        (0, 8),
        (0, 474),
        (0, 107),
        (0, 447),
        (0, 95),
        (0, 169),
        (0, 36),
        (0, 9),
        (0, 53),
        (0, 0),
        (0, 0),
        (0, 35),
        (0, 31),
        (0, 1432),
        (0, 69),
        (0, 22),
        (0, 784),
        (0, 165),
        (0, 163),
        (0, 4),
        (0, 16),
        (0, 3),
        (0, 0),
        (0, 4),
        (0, 4),
        (1, 0),
        (0, 1136),
        (0, 0),
        (0, 890),
        (0, 244),
        (0, 94),
        (0, 244),
        (0, 65),
        (0, 12),
        (0, 1),
        (0, 849),
        (0, 315),
        (0, 1),
        (0, 12),
        (0, 3),
        (0, 14),
        (0, 0),
        (0, 1),
        (0, 18),
        (0, 896),
        (0, 629),
        (0, 780),
        (0, 2),
        (0, 20),
        (0, 241),
        (0, 1463),
        (0, 17),
        (0, 2),
        (0, 66),
        (0, 1086),
        (0, 28),
        (0, 0),
        (0, 112),
        (0, 23),
        (0, 0),
        (0, 1839),
        (1, 1553),
        (0, 3),
        (0, 73),
        (0, 4),
        (0, 73),
        (0, 98),
        (0, 49),
        (0, 729),
        (0, 0),
        (0, 0),
        (0, 18),
        (0, 2),
        (0, 149),
        (0, 304),
        (0, 696),
        (0, 19),
        (0, 22),
        (0, 738),
        (0, 67),
        (0, 106),
        (0, 1696),
        (0, 610),
        (0, 7),
        (0, 139),
        (0, 0),
        (0, 445),
        (0, 2321),
        (0, 617),
        (0, 19),
        (0, 1),
        (0, 0),
        (0, 330),
        (0, 129),
        (0, 1036),
        (0, 2),
        (0, 468),
        (0, 143),
        (0, 1),
        (0, 17),
        (0, 5),
        (0, 33),
        (0, 83),
        (0, 7),
        (0, 296),
        (0, 15),
        (0, 1002),
        (0, 0),
        (0, 4),
        (0, 7),
        (0, 3),
        (0, 0),
        (0, 21),
        (0, 25),
        (0, 2),
        (0, 311),
        (0, 463),
        (0, 487),
        (0, 311),
        (0, 2),
        (0, 47),
        (0, 14),
        (0, 1),
        (0, 110),
        (0, 3),
        (0, 4),
        (0, 0),
        (0, 41),
        (0, 1374),
        (0, 13),
        (0, 212),
        (0, 311),
        (0, 158),
        (0, 638),
        (0, 26),
        (2, 1007),
        (0, 197),
        (0, 143),
        (0, 568),
        (0, 13),
        (0, 5),
        (0, 26),
        (0, 33),
        (0, 1405),
        (0, 466),
        (0, 2490),
        (0, 130),
        (0, 12),
        (0, 6),
        (1, 687),
        (0, 174),
        (0, 0),
        (0, 2),
        (0, 251),
        (0, 220),
        (0, 20),
        (0, 82),
        (0, 63),
        (0, 8),
        (0, 1074),
        (0, 1494),
        (0, 5),
        (0, 177),
        (0, 241),
        (0, 298),
        (0, 5),
        (0, 297),
        (0, 2166),
        (0, 2),
        (0, 2),
        (0, 2),
        (0, 36),
        (0, 0),
        (0, 14),
        (0, 717),
        (0, 0),
        (0, 1),
        (0, 1392),
        (0, 22),
        (0, 92),
        (0, 525),
        (0, 3),
        (0, 93),
        (0, 2147),
        (0, 52),
        (0, 1600),
        (0, 81),
        (0, 74),
        (0, 55),
        (0, 50),
        (0, 8),
        (0, 0),
        (1, 1553),
        (0, 194),
        (0, 192),
        (0, 197),
        (0, 2),
        (0, 2),
        (0, 2),
        (0, 99),
        (0, 167),
        (0, 740),
        (0, 292),
        (0, 169),
        (0, 21),
        (0, 284),
        (0, 25),
        (0, 26),
        (0, 26),
        (1, 151),
        (0, 6),
        (0, 51),
        (0, 290),
        (0, 558),
        (0, 3),
        (0, 6),
        (1, 1094),
        (0, 19),
        (0, 228),
        (1, 1511),
        (0, 40),
        (3, 971),
        (0, 24),
        (0, 7),
        (0, 136),
        (0, 168),
        (0, 49),
        (0, 6),
        (0, 3),
        (0, 63),
        (0, 0),
        (0, 931),
        (0, 281),
        (1, 439),
        (0, 714),
        (0, 52),
        (0, 1),
        (0, 2411),
        (0, 3),
        (0, 4),
        (0, 1249),
        (1, 1943),
        (0, 148),
        (0, 1405),
        (0, 0),
        (0, 390),
        (0, 29),
        (0, 349),
        (0, 8),
        (0, 51),
        (0, 32),
        (0, 14),
        (0, 84),
        (0, 127),
        (0, 6),
        (0, 1),
        (0, 2052),
        (0, 84),
        (0, 989),
        (1, 1216),
        (0, 3),
        (0, 168),
        (0, 6),
        (0, 356),
        (0, 1797),
        (0, 568),
        (0, 1888),
        (0, 3),
        (0, 559),
        (0, 1710),
        (0, 93),
        (0, 52),
        (0, 280),
        (0, 44),
        (0, 147),
        (0, 45),
        (0, 394),
        (0, 6),
        (0, 1729),
        (1, 498),
        (0, 1),
        (0, 0),
        (0, 1),
        (0, 1),
        (0, 6),
        (0, 1219),
        (1, 379),
        (0, 7),
        (0, 1932),
        (0, 169),
        (0, 0),
        (1, 886),
        (0, 178),
        (0, 1149),
        (0, 8),
        (0, 994),
        (0, 1264),
        (0, 0),
        (0, 68),
        (0, 37),
        (0, 1),
        (1, 920),
        (0, 0),
        (6, 1549),
        (0, 1),
        (0, 1),
        (0, 102),
        (0, 1),
        (2, 393),
        (0, 0),
        (0, 174),
        (0, 1264),
        (0, 0),
        (1, 36),
        (0, 322),
        (0, 1410),
        (1, 569),
        (0, 901),
        (0, 67),
        (0, 4),
        (0, 0),
        (0, 244),
        (0, 622),
        (0, 4),
        (0, 122),
        (0, 31),
        (0, 12),
        (0, 50),
        (2, 650),
        (0, 1386),
        (0, 997),
        (0, 1149),
        (0, 1427),
        (0, 216),
        (0, 70),
        (0, 73),
        (0, 146),
        (0, 22),
        (0, 32),
        (0, 54),
        (0, 18),
        (0, 812),
        (0, 28),
        (0, 14),
        (0, 8),
        (0, 4),
        (0, 177),
        (0, 104),
        (0, 1139),
        (0, 225),
        (0, 6),
        (1, 1336),
        (0, 2255),
        (0, 50),
        (0, 978),
        (3, 850),
        (0, 7),
        (0, 9),
        (0, 308),
        (0, 40),
        (0, 497),
        (0, 2),
        (0, 68),
        (0, 185),
        (1, 1551),
        (0, 2071),
        (1, 1718),
        (0, 1014),
        (2, 1666),
        (1, 2258),
        (0, 229),
        (0, 2),
        (1, 443),
        (0, 1254),
        (1, 17),
        (0, 9),
        (0, 234),
        (0, 235),
        (0, 2155),
        (0, 746),
        (0, 167),
        (2, 1007),
        (0, 2090),
        (0, 110),
        (0, 244),
        (0, 683),
        (5, 2381),
        (0, 602),
        (0, 2374),
        (2, 689),
        (0, 7),
        (0, 19),
        (0, 251),
        (0, 0),
        (2, 626),
        (0, 8),
        (0, 19),
        (0, 283),
        (0, 0),
        (0, 4),
        (0, 1022),
        (0, 8),
        (0, 5),
        (0, 13),
        (0, 5),
        (0, 1253),
        (7, 863),
        (0, 1006),
        (4, 1346),
        (0, 29),
        (0, 6),
        (0, 380),
        (0, 1675),
        (0, 365),
        (0, 2431),
        (0, 192),
        (2, 1119),
        (10, 1330),
        (6, 2345),
        (0, 0),
        (0, 118),
        (1, 506),
        (0, 35),
        (0, 16),
        (0, 1816),
        (0, 6),
        (0, 1),
        (6, 1133),
        (0, 26),
        (0, 0),
        (0, 0),
        (1, 1302),
        (0, 481),
        (0, 1005),
        (7, 2210),
        (0, 2),
        (3, 988),
        (1, 658),
        (4, 1433),
        (0, 0),
        (1, 102),
        (6, 1700),
        (0, 22),
        (0, 703),
        (0, 568),
        (0, 26),
        (0, 1095),
        (0, 2),
        (0, 52),
        (0, 904),
        (0, 83),
        (3, 722),
        (0, 407),
        (9, 1774),
        (0, 9),
        (3, 2256),
        (0, 1),
        (0, 0),
        (4, 115),
        (0, 0),
        (0, 1705),
        (0, 24),
        (0, 22),
        (0, 7),
        (0, 1740),
        (0, 24),
        (0, 0),
        (1, 1029),
        (2, 2507),
        (0, 9),
        (0, 0),
        (11, 1465),
        (1, 497),
        (0, 0),
        (1, 2271),
        (4, 935),
        (0, 1),
    ],
    entries: &[
        (4781, Vendor { id : 4781u16 , name : "Multidata GmbH" , devices : & [] }),
        (4147, Vendor { id : 4147u16 , name : "NEC Corporation" , devices : & [Device { vendor_id : 4147u16 , id : 0u16 , name : "Vr4181A USB Host or Function Control Unit" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 1u16 , name : "PCI to 486-like bus Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 2u16 , name : "PCI to VL98 Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 3u16 , name : "ATM Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 4u16 , name : "R4000 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 5u16 , name : "PCI to 486-like bus Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 6u16 , name : "PC-9800 Graphic Accelerator" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 7u16 , name : "PCI to UX-Bus Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 8u16 , name : "PC-9800 Graphic Accelerator" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 9u16 , name : "PCI to PC9800 Core-Graph Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 22u16 , name : "PCI to VL Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 26u16 , name : "[Nile II]" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 33u16 , name : "Vrc4373 [Nile I]" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 41u16 , name : "PowerVR PCX1" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 42u16 , name : "PowerVR 3D" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 44u16 , name : "Star Alpha 2" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 45u16 , name : "PCI to C-bus Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 53u16 , name : "OHCI USB Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4147u16 , subdevice : 53u16 , name : "USB Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4755u16 , name : "USB add-in card" } , SubSystem { subvendor : 4156u16 , subdevice : 4756u16 , name : "USB 2.0 add-in card" } , SubSystem { subvendor : 4436u16 , subdevice : 519u16 , name : "Buffalo IFC-USB2P4" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "USB" } , SubSystem { subvendor : 4486u16 , subdevice : 53u16 , name : "DUB-C2 USB 2.0 2-port 32-bit cardbus controller" } , SubSystem { subvendor : 4846u16 , subdevice : 28672u16 , name : "Root Hub" } , SubSystem { subvendor : 5314u16 , subdevice : 261u16 , name : "PTI-205N USB 2.0 Host Controller" } , SubSystem { subvendor : 6041u16 , subdevice : 1u16 , name : "Root Hub" } , SubSystem { subvendor : 6449u16 , subdevice : 10u16 , name : "GlobeTrotter Fusion Quad Lite (PPP data)" } , SubSystem { subvendor : 6449u16 , subdevice : 11u16 , name : "GlobeTrotter Fusion Quad Lite (GSM data)" } , SubSystem { subvendor : 32893u16 , subdevice : 53u16 , name : "PCI-USB2 (OHCI subsystem)" } , SubSystem { subvendor : 32902u16 , subdevice : 19780u16 , name : "D850EMV2 motherboard" }] } , Device { vendor_id : 4147u16 , id : 59u16 , name : "PCI to C-bus Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 62u16 , name : "NAPCCARD Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 70u16 , name : "PowerVR PCX2 [midas]" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 90u16 , name : "Vrc5074 [Nile 4]" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 99u16 , name : "uPD72862 [Firewarden] IEEE1394 OHCI 1.0 Link Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 103u16 , name : "PowerVR Neon 250 Chipset" , subsystems : & [SubSystem { subvendor : 4112u16 , subdevice : 32u16 , name : "PowerVR Neon 250 AGP 32Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 128u16 , name : "PowerVR Neon 250 AGP 16Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 136u16 , name : "PowerVR Neon 250 16Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 144u16 , name : "PowerVR Neon 250 AGP 16Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 152u16 , name : "PowerVR Neon 250 16Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 160u16 , name : "PowerVR Neon 250 AGP 32Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 168u16 , name : "PowerVR Neon 250 32Mb" } , SubSystem { subvendor : 4112u16 , subdevice : 288u16 , name : "PowerVR Neon 250 AGP 32Mb" }] } , Device { vendor_id : 4147u16 , id : 114u16 , name : "uPD72874 IEEE1394 OHCI 1.1 3-port PHY-Link Ctrlr" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 116u16 , name : "56k Voice Modem" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32788u16 , name : "RCV56ACF 56k Voice Modem" }] } , Device { vendor_id : 4147u16 , id : 155u16 , name : "Vrc5476" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 165u16 , name : "VRC4173" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 166u16 , name : "VRC5477 AC97" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 205u16 , name : "uPD72870 [Firewarden] IEEE1394a OHCI 1.0 Link/3-port PHY Controller" , subsystems : & [SubSystem { subvendor : 4846u16 , subdevice : 32785u16 , name : "Root hub" }] } , Device { vendor_id : 4147u16 , id : 206u16 , name : "uPD72871 [Firewarden] IEEE1394a OHCI 1.0 Link/1-port PHY Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 223u16 , name : "Vr4131" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 224u16 , name : "uPD72010x USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4436u16 , subdevice : 520u16 , name : "Buffalo IFC-USB2P4" } , SubSystem { subvendor : 4486u16 , subdevice : 61696u16 , name : "DUB-C2 USB 2.0 2-port 32-bit cardbus controller" } , SubSystem { subvendor : 4846u16 , subdevice : 28673u16 , name : "Root hub" } , SubSystem { subvendor : 5314u16 , subdevice : 517u16 , name : "PTI-205N USB 2.0 Host Controller" } , SubSystem { subvendor : 6041u16 , subdevice : 2u16 , name : "Root Hub" } , SubSystem { subvendor : 32893u16 , subdevice : 4163u16 , name : "PCI-USB2 (EHCI subsystem)" } , SubSystem { subvendor : 32902u16 , subdevice : 19780u16 , name : "D850EMV2 motherboard" }] } , Device { vendor_id : 4147u16 , id : 231u16 , name : "uPD72873 [Firewarden] IEEE1394a OHCI 1.1 Link/2-port PHY Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 242u16 , name : "uPD72874 [Firewarden] IEEE1394a OHCI 1.1 Link/3-port PHY Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 243u16 , name : "uPD6113x Multimedia Decoder/Processor [EMMA2]" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 268u16 , name : "VR7701" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 293u16 , name : "uPD720400 PCI Express - PCI/PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 314u16 , name : "Dual Tuner/MPEG Encoder" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 357u16 , name : "uPD61253F1 MPEG Video Broadcast Decoder" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 404u16 , name : "uPD720200 USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33811u16 , name : "P8P67 Deluxe Motherboard" } , SubSystem { subvendor : 4173u16 , subdevice : 36986u16 , name : "Vaio VPCF1" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 7062u16 , subdevice : 1u16 , name : "USB 3.0 PCIe Card" }] } , Device { vendor_id : 4147u16 , id : 487u16 , name : "uPD72873 [Firewarden] IEEE1394a OHCI 1.1 Link/2-port PHY Controller" , subsystems : & [] } , Device { vendor_id : 4147u16 , id : 498u16 , name : "uPD72874 [Firewarden] IEEE1394a OHCI 1.1 Link/3-port PHY Controller" , subsystems : & [] }] }),
        (5398, Vendor { id : 5398u16 , name : "MYSON Technology Inc" , devices : & [Device { vendor_id : 5398u16 , id : 2048u16 , name : "MTD-8xx 100/10M Ethernet PCI Adapter" , subsystems : & [] } , Device { vendor_id : 5398u16 , id : 2051u16 , name : "SURECOM EP-320X-S 100/10M Ethernet PCI Adapter" , subsystems : & [SubSystem { subvendor : 4896u16 , subdevice : 4285u16 , name : "SURECOM EP-320X-S 100/10M Ethernet PCI Adapter" }] } , Device { vendor_id : 5398u16 , id : 2193u16 , name : "MTD-8xx 100/10M Ethernet PCI Adapter" , subsystems : & [] }] }),
        (65049, Vendor { id : 65049u16 , name : "TenaFe, Inc." , devices : & [Device { vendor_id : 65049u16 , id : 1u16 , name : "TC2200/TC2201 NVMe Controller (DRAM-less)" , subsystems : & [] }] }),
        (44288, Vendor { id : 44288u16 , name : "Alta Data Technologies LLC" , devices : & [] }),
        (6538, Vendor { id : 6538u16 , name : "Nallatech Ltd." , devices : & [] }),
        (4260, Vendor { id : 4260u16 , name : "Globe Manufacturing Sales" , devices : & [] }),
        (4997, Vendor { id : 4997u16 , name : "Netgear" , devices : & [Device { vendor_id : 4997u16 , id : 107u16 , name : "WA301 802.11b Wireless PCI Adapter" , subsystems : & [] } , Device { vendor_id : 4997u16 , id : 16640u16 , name : "MA301 802.11b Wireless PCI Adapter" , subsystems : & [] } , Device { vendor_id : 4997u16 , id : 17921u16 , name : "WAG511 802.11a/b/g Dual Band Wireless PC Card" , subsystems : & [] } , Device { vendor_id : 4997u16 , id : 25098u16 , name : "GA620 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4997u16 , id : 25354u16 , name : "GA630 Gigabit Ethernet" , subsystems : & [] }] }),
        (6840, Vendor { id : 6840u16 , name : "Parallels, Inc." , devices : & [Device { vendor_id : 6840u16 , id : 16384u16 , name : "Virtual Machine Communication Interface" , subsystems : & [] } , Device { vendor_id : 6840u16 , id : 16389u16 , name : "Accelerated Virtual Video Adapter" , subsystems : & [] } , Device { vendor_id : 6840u16 , id : 16390u16 , name : "Memory Ballooning Controller" , subsystems : & [] }] }),
        (5002, Vendor { id : 5002u16 , name : "Fusion Micromedia Corp" , devices : & [Device { vendor_id : 5002u16 , id : 61u16 , name : "VFS491 Validity Sensor" , subsystems : & [] }] }),
        (21581, Vendor { id : 21581u16 , name : "TBS Technologies" , devices : & [Device { vendor_id : 21581u16 , id : 24952u16 , name : "DVB Tuner PCIe Card" , subsystems : & [SubSystem { subvendor : 21581u16 , subdevice : 26884u16 , name : "TBS6904 DVB-S2 Quad Tuner PCIe Card" } , SubSystem { subvendor : 21581u16 , subdevice : 26885u16 , name : "TBS6905 DVB-S2 Quad Tuner PCIe Card" } , SubSystem { subvendor : 25093u16 , subdevice : 1u16 , name : "TBS6205 DVB-T2/T/C Quad TV Tuner PCIe Card" } , SubSystem { subvendor : 25097u16 , subdevice : 1u16 , name : "TBS6209 DVB-T2/C2/T/C/ISDB-T OctaTV Tuner" } , SubSystem { subvendor : 26000u16 , subdevice : 1u16 , name : "TBS6590 DVB-S/S2/S2X/T/T2/C/C2/ISDB-T + 2xCI" } , SubSystem { subvendor : 26000u16 , subdevice : 2u16 , name : "TBS6590SE DVB-S/S2/S2X/T/T2/C/C2 + 2xCI" } , SubSystem { subvendor : 26372u16 , subdevice : 65535u16 , name : "TBS6704 (Quad ATSC/QAMB)" } , SubSystem { subvendor : 26883u16 , subdevice : 32u16 , name : "TBS6903x (Dual DVB-S/S2/S2X)" } , SubSystem { subvendor : 26883u16 , subdevice : 33u16 , name : "TBS6903x (Dual DVB-S/S2/S2X)" } , SubSystem { subvendor : 26883u16 , subdevice : 34952u16 , name : "TBS6903x (Dual DVB-S/S2/S2X)" } , SubSystem { subvendor : 26889u16 , subdevice : 9u16 , name : "TBS6909x (Octa DVB-S/S2/S2X)" } , SubSystem { subvendor : 26889u16 , subdevice : 16u16 , name : "TBS6909x (Octa DVB-S/S2/S2X)" } , SubSystem { subvendor : 26889u16 , subdevice : 25u16 , name : "TBS6909x (Octa DVB-S/S2/S2X)" } , SubSystem { subvendor : 26896u16 , subdevice : 1u16 , name : "TBS6910 DVB-S/S2 + 2xCI" } , SubSystem { subvendor : 26896u16 , subdevice : 2u16 , name : "TBS6910SE DVB-S/S2/S2x + 2xCI" }] }] }),
        (5074, Vendor { id : 5074u16 , name : "Shark Multimedia Inc" , devices : & [] }),
        (43690, Vendor { id : 43690u16 , name : "Adnaco Technology Inc." , devices : & [Device { vendor_id : 43690u16 , id : 1u16 , name : "H1 PCIe over fiber optic host controller" , subsystems : & [] } , Device { vendor_id : 43690u16 , id : 2u16 , name : "R1BP1 PCIe over fiber optic expansion chassis" , subsystems : & [] }] }),
        (4957, Vendor { id : 4957u16 , name : "ABB Network Partner AB" , devices : & [] }),
        (5149, Vendor { id : 5149u16 , name : "Digitan Systems Inc" , devices : & [] }),
        (6388, Vendor { id : 6388u16 , name : "Napatech A/S" , devices : & [Device { vendor_id : 6388u16 , id : 49u16 , name : "NT20X Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 81u16 , name : "NT20X Capture Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 97u16 , name : "NT20E Capture Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 100u16 , name : "NT20E Inline Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 113u16 , name : "NT4E Capture Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 116u16 , name : "NT4E Inline Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 129u16 , name : "NT4E 4-port Expansion Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 145u16 , name : "NT20X Capture Card [New Rev]" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 161u16 , name : "NT4E-STD Capture Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 164u16 , name : "NT4E-STD Inline Card" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 177u16 , name : "NTBPE Optical Bypass Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 197u16 , name : "NT20E2 Network Adapter 2x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 213u16 , name : "NT40E2-4 Network Adapter 4x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 229u16 , name : "NT40E2-1 Network Adapter 1x40Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 245u16 , name : "NT4E2-4T-BP Network Adapter 4x1Gb with Electrical Bypass" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 261u16 , name : "NT4E2-4-PTP Network Adapter 4x1Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 277u16 , name : "NT20E2-PTP Network Adapter 2x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 293u16 , name : "NT4E2-4-PTP Network Adapter 4x1Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 309u16 , name : "NT20E2-PTP Network Adapter 2x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 325u16 , name : "NT40E3-4-PTP Network Adapter 4x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 341u16 , name : "NT100E3-1-PTP Network Adapter 1x100Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 357u16 , name : "NT80E3-2-PTP Network Adapter 2x40Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 373u16 , name : "NT20E3-2-PTP Network Adapter 2x10Gb" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 389u16 , name : "NT40A01 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 421u16 , name : "NT200A01 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 453u16 , name : "NT200A02 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 469u16 , name : "NT50B01 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 485u16 , name : "NT100A01 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 533u16 , name : "NT400D11 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 549u16 , name : "NT40A11 Network Adapter" , subsystems : & [] } , Device { vendor_id : 6388u16 , id : 661u16 , name : "NT400D13 Network Adapter" , subsystems : & [] }] }),
        (5741, Vendor { id : 5741u16 , name : "Broadcom Corporation" , devices : & [Device { vendor_id : 5741u16 , id : 1u16 , name : "SiByte BCM1125/1125H/1250 System-on-a-Chip PCI" , subsystems : & [] } , Device { vendor_id : 5741u16 , id : 2u16 , name : "SiByte BCM1125H/1250 System-on-a-Chip HyperTransport" , subsystems : & [] } , Device { vendor_id : 5741u16 , id : 18u16 , name : "SiByte BCM1280/BCM1480 System-on-a-Chip PCI-X" , subsystems : & [] } , Device { vendor_id : 5741u16 , id : 20u16 , name : "Sibyte BCM1280/BCM1480 System-on-a-Chip HyperTransport" , subsystems : & [] }] }),
        (4397, Vendor { id : 4397u16 , name : "Ravicad" , devices : & [] }),
        (4252, Vendor { id : 4252u16 , name : "Megachips Corporation" , devices : & [] }),
        (6239, Vendor { id : 6239u16 , name : "Wistron NeWeb Corp." , devices : & [] }),
        (49153, Vendor { id : 49153u16 , name : "TSI Telsys" , devices : & [] }),
        (5018, Vendor { id : 5018u16 , name : "Alacritech Inc" , devices : & [Device { vendor_id : 5018u16 , id : 1u16 , name : "Quad Port 10/100 Server Accelerator" , subsystems : & [] } , Device { vendor_id : 5018u16 , id : 3u16 , name : "Single Port 10/100 Server Accelerator" , subsystems : & [] } , Device { vendor_id : 5018u16 , id : 5u16 , name : "Single Port Gigabit Server Accelerator" , subsystems : & [] }] }),
        (8274, Vendor { id : 8274u16 , name : "Frontgrade Gaisler AB" , devices : & [] }),
        (4785, Vendor { id : 4785u16 , name : "GammaLink" , devices : & [] }),
        (7760, Vendor { id : 7760u16 , name : "IP3 Tech (HK) Limited" , devices : & [] }),
        (7890, Vendor { id : 7890u16 , name : "FuriosaAI, Inc." , devices : & [Device { vendor_id : 7890u16 , id : 0u16 , name : "Warboy" , subsystems : & [] } , Device { vendor_id : 7890u16 , id : 1u16 , name : "RNGD" , subsystems : & [] } , Device { vendor_id : 7890u16 , id : 2u16 , name : "RNGD-Plus" , subsystems : & [] } , Device { vendor_id : 7890u16 , id : 8738u16 , name : "RNGD-S" , subsystems : & [] }] }),
        (4505, Vendor { id : 4505u16 , name : "Attachmate Corporation" , devices : & [Device { vendor_id : 4505u16 , id : 257u16 , name : "Advanced ISCA/PCI Adapter" , subsystems : & [] }] }),
        (4741, Vendor { id : 4741u16 , name : "Platform Technologies, Inc." , devices : & [Device { vendor_id : 4741u16 , id : 256u16 , name : "AGOGO sound chip (aka ESS Maestro 1)" , subsystems : & [] }] }),
        (6354, Vendor { id : 6354u16 , name : "Sitecom Europe BV (Wrong ID)" , devices : & [Device { vendor_id : 6354u16 , id : 12393u16 , name : "DC-105v2 ISDN controller" , subsystems : & [] }] }),
        (4163, Vendor { id : 4163u16 , name : "ASUSTeK Computer Inc." , devices : & [Device { vendor_id : 4163u16 , id : 1653u16 , name : "ISDNLink P-IN100-ST-D" , subsystems : & [SubSystem { subvendor : 1653u16 , subdevice : 5892u16 , name : "ISDN Adapter (PCI Bus, D, C)" } , SubSystem { subvendor : 1653u16 , subdevice : 5895u16 , name : "ISDN Adapter (PCI Bus, DV, W)" } , SubSystem { subvendor : 4303u16 , subdevice : 4190u16 , name : "ISDN Adapter (PCI Bus, DV, W)" }] } , Device { vendor_id : 4163u16 , id : 38402u16 , name : "AMD RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33698u16 , name : "M4A785TD Motherboard" }] }] }),
        (5601, Vendor { id : 5601u16 , name : "Voice Technologies Group Inc" , devices : & [] }),
        (4368, Vendor { id : 4368u16 , name : "Powerhouse Systems" , devices : & [Device { vendor_id : 4368u16 , id : 24631u16 , name : "Firepower Powerized SMP I/O ASIC" , subsystems : & [] } , Device { vendor_id : 4368u16 , id : 24691u16 , name : "Firepower Powerized SMP I/O ASIC" , subsystems : & [] }] }),
        (5282, Vendor { id : 5282u16 , name : "Millennium Engineering Inc" , devices : & [] }),
        (5358, Vendor { id : 5358u16 , name : "MASPRO KENKOH Corp" , devices : & [] }),
        (6099, Vendor { id : 6099u16 , name : "Areca Technology Corp." , devices : & [Device { vendor_id : 6099u16 , id : 4368u16 , name : "ARC-1110 4-Port PCI-X to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4384u16 , name : "ARC-1120 8-Port PCI-X to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4400u16 , name : "ARC-1130 12-Port PCI-X to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4448u16 , name : "ARC-1160 16-Port PCI-X to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4464u16 , name : "ARC-1170 24-Port PCI-X to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4609u16 , name : "ARC-1200 2-Port PCI-Express to SATA II RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4611u16 , name : "ARC-1203 2/4/8 Port PCIe 2.0 to SATA 6Gb RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4624u16 , name : "ARC-1210 4-Port PCI-Express to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4628u16 , name : "ARC-12x4 PCIe 2.0 to SAS/SATA 6Gb RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 4628u16 , name : "ARC-1214 4-Port PCIe 2.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4644u16 , name : "ARC-1224 8-Port PCIe 2.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4708u16 , name : "ARC-1264 12/16 Port PCIe 2.0 to SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4740u16 , name : "ARC-1284 24 Port PCIe 2.0 to SATA 6Gb RAID Controller" }] } , Device { vendor_id : 6099u16 , id : 4640u16 , name : "ARC-1220 8-Port PCI-Express to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4642u16 , name : "ARC-1222 8-Port PCI-Express to SAS/SATA II RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4656u16 , name : "ARC-1230 12-Port PCI-Express to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4704u16 , name : "ARC-1260 16-Port PCI-Express to SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4736u16 , name : "ARC-1280/1280ML 24-Port PCI-Express to SATA II RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 4641u16 , name : "ARC-1221 8-Port PCI-Express to SATA RAID Controller" }] } , Device { vendor_id : 6099u16 , id : 4864u16 , name : "ARC-1300ix-16 16-Port PCI-Express to SAS Non-RAID Host Adapter" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4896u16 , name : "ARC-1320 8/16 Port PCIe 2.0 to SAS/SATA 6Gb Non-RAID Host Adapter" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 4912u16 , name : "ARC-1330 16 Port PCIe 3.0 to SAS/SATA 12Gb Non-RAID Host Adapter" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 5760u16 , name : "ARC-1680 series PCIe to SAS/SATA 3Gb RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 4626u16 , name : "ARC-1212 4-Port PCIe to SAS/SATA II RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4642u16 , name : "ARC-1222 8-Port PCIe to SAS/SATA 3Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 5760u16 , name : "ARC-1680 8/12/16/24 Port PCIe to SAS/SATA 3Gb RAID Controller" }] } , Device { vendor_id : 6099u16 , id : 6272u16 , name : "ARC-188x series PCIe 2.0/3.0 to SAS/SATA 6/12Gb RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 4627u16 , name : "ARC-1213 4-Port PCIe 2.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4629u16 , name : "ARC-1215 4-Port PCIe 3.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4630u16 , name : "ARC-1216 4-Port PCIe 3.0 to SAS/SATA 12Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4643u16 , name : "ARC-1223 8-Port PCIe 2.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4645u16 , name : "ARC-1225 8-Port PCIe 3.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4646u16 , name : "ARC-1226 8-Port PCIe 3.0 to SAS/SATA 12Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 6272u16 , name : "ARC-1880 8/12/16/24 Port PCIe 2.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 6274u16 , name : "ARC-1882 8/12/16/24 Port PCIe 3.0 to SAS/SATA 6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 6275u16 , name : "ARC-1883 8/12/16/24 Port PCIe 3.0 to SAS/SATA 12Gb RAID Controller" }] } , Device { vendor_id : 6099u16 , id : 6276u16 , name : "ARC-1884 series PCIe 3.0 to SAS/SATA 12/6Gb RAID Controller" , subsystems : & [] } , Device { vendor_id : 6099u16 , id : 6278u16 , name : "ARC-1886 series PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 5766u16 , name : "ARC-1686 PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 6278u16 , name : "ARC-1886 PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" }] } , Device { vendor_id : 6099u16 , id : 6282u16 , name : "ARC-1886 series PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" , subsystems : & [SubSystem { subvendor : 6099u16 , subdevice : 4631u16 , name : "ARC-1217 4-Port PCIe 4.0 to SAS/SATA 12/6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 4647u16 , name : "ARC-1227 8-Port PCIe 4.0 to SAS/SATA 12/6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 5766u16 , name : "ARC-1686 PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 5768u16 , name : "ARC-1688 PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" } , SubSystem { subvendor : 6099u16 , subdevice : 6278u16 , name : "ARC-1886 PCIe 4.0 to NVMe/SAS/SATA 16/12/6Gb RAID Controller" }] }] }),
        (4241, Vendor { id : 4241u16 , name : "Intergraph Corporation" , devices : & [Device { vendor_id : 4241u16 , id : 32u16 , name : "3D graphics processor" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 33u16 , name : "3D graphics processor w/Texturing" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 64u16 , name : "3D graphics frame buffer" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 65u16 , name : "3D graphics frame buffer" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 96u16 , name : "Proprietary bus bridge" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 228u16 , name : "Powerstorm 4D50T" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 1824u16 , name : "Motion JPEG codec" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 1920u16 , name : "Intense3D Wildcat 3410 (MSMT496)" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 1952u16 , name : "Sun Expert3D-Lite Graphics Accelerator" , subsystems : & [] } , Device { vendor_id : 4241u16 , id : 4241u16 , name : "Sun Expert3D Graphics Accelerator" , subsystems : & [] }] }),
        (7352, Vendor { id : 7352u16 , name : "Dawning Information Industry Co., Ltd." , devices : & [] }),
        (4810, Vendor { id : 4810u16 , name : "Integrated Computing Engines" , devices : & [] }),
        (7648, Vendor { id : 7648u16 , name : "Groq" , devices : & [Device { vendor_id : 7648u16 , id : 0u16 , name : "TSP [GroqChip]" , subsystems : & [SubSystem { subvendor : 7648u16 , subdevice : 0u16 , name : "GC1-010X-ES [GroqCard]" } , SubSystem { subvendor : 7648u16 , subdevice : 49157u16 , name : "GC1-010B [GroqCard]" } , SubSystem { subvendor : 7648u16 , subdevice : 49158u16 , name : "GC1-0109 [GroqCard]" } , SubSystem { subvendor : 7648u16 , subdevice : 49159u16 , name : "GC1-0100 [GroqCard]" }] }] }),
        (35914, Vendor { id : 35914u16 , name : "Winbond" , devices : & [Device { vendor_id : 35914u16 , id : 6528u16 , name : "W89C940 misprogrammed [ne2k]" , subsystems : & [] }] }),
        (4856, Vendor { id : 4856u16 , name : "Electronic Design GmbH" , devices : & [Device { vendor_id : 4856u16 , id : 2u16 , name : "VideoMaker" , subsystems : & [] }] }),
        (6663, Vendor { id : 6663u16 , name : "Kvaser AB" , devices : & [Device { vendor_id : 6663u16 , id : 6u16 , name : "CAN interface PC104+ HS/HS" , subsystems : & [] } , Device { vendor_id : 6663u16 , id : 7u16 , name : "CAN interface PCIcanx II HS or HS/HS" , subsystems : & [] } , Device { vendor_id : 6663u16 , id : 8u16 , name : "CAN interface PCIEcan HS or HS/HS" , subsystems : & [] } , Device { vendor_id : 6663u16 , id : 9u16 , name : "CAN interface PCI104 HS/HS" , subsystems : & [] }] }),
        (4131, Vendor { id : 4131u16 , name : "Trident Microsystems" , devices : & [Device { vendor_id : 4131u16 , id : 404u16 , name : "82C194" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 8192u16 , name : "4DWave DX" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 8193u16 , name : "4DWave NX" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 5120u16 , name : "Trident PCI288-Q3DII (NX)" }] } , Device { vendor_id : 4131u16 , id : 8448u16 , name : "CyberBlade XP4m32" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 8704u16 , name : "XGI Volari XP5" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 33792u16 , name : "CyberBlade/i7" , subsystems : & [SubSystem { subvendor : 4131u16 , subdevice : 33792u16 , name : "CyberBlade i7 AGP" }] } , Device { vendor_id : 4131u16 , id : 33824u16 , name : "CyberBlade/i7d" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45402u16 , name : "CyberBlade i7 AGP" }] } , Device { vendor_id : 4131u16 , id : 34048u16 , name : "CyberBlade/i1" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 34080u16 , name : "CyberBlade i1" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45422u16 , name : "CyberBlade i1 AGP" } , SubSystem { subvendor : 4131u16 , subdevice : 34080u16 , name : "CyberBlade i1 AGP" }] } , Device { vendor_id : 4131u16 , id : 34336u16 , name : "CyberBlade/i1" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1282u16 , name : "ThinkPad R30/T30" } , SubSystem { subvendor : 4116u16 , subdevice : 4133u16 , name : "Travelmate 352TE" }] } , Device { vendor_id : 4131u16 , id : 34848u16 , name : "CyberBlade XPAi1" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37664u16 , name : "TGUI 9320" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37712u16 , name : "GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37728u16 , name : "Flat panel GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37762u16 , name : "Cyber 9382 [Reference design]" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37763u16 , name : "Cyber 9383 [Reference design]" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37765u16 , name : "Cyber 9385 [Reference design]" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37766u16 , name : "Cyber 9386" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37768u16 , name : "Cyber 9388" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37783u16 , name : "Cyber 9397" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37786u16 , name : "Cyber 9397DVD" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37920u16 , name : "TGUI 9420" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37936u16 , name : "TGUI 9430" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37952u16 , name : "TGUI 9440" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 37984u16 , name : "TGUI 9460" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38000u16 , name : "TGUI 9470" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38176u16 , name : "Cyber 9520" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38181u16 , name : "Cyber 9525" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38208u16 , name : "Cyber 9540" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38496u16 , name : "TGUI 9660/938x/968x" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38528u16 , name : "TGUI 9680" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38530u16 , name : "TGUI 9682" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38531u16 , name : "TGUI 9683" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38533u16 , name : "ProVIDIA 9685" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38736u16 , name : "3DImage 9750" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 38736u16 , name : "3DImage 9750" } , SubSystem { subvendor : 4131u16 , subdevice : 38736u16 , name : "3DImage 9750" }] } , Device { vendor_id : 4131u16 , id : 38739u16 , name : "TGUI 9753" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38740u16 , name : "TGUI 9754" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38745u16 , name : "TGUI 975" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38787u16 , name : "TGUI 9783" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38789u16 , name : "TGUI 9785" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 38992u16 , name : "3DImage 9850" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 39040u16 , name : "Blade 3D PCI/AGP" , subsystems : & [SubSystem { subvendor : 4131u16 , subdevice : 39040u16 , name : "Blade 3D" }] } , Device { vendor_id : 4131u16 , id : 39184u16 , name : "CyberBlade/XP" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 39216u16 , name : "CyberBlade/XPm" , subsystems : & [] } , Device { vendor_id : 4131u16 , id : 39264u16 , name : "CyberBlade XP2" , subsystems : & [] }] }),
        (4410, Vendor { id : 4410u16 , name : "FWB Inc" , devices : & [] }),
        (56026, Vendor { id : 56026u16 , name : "Datapath Limited" , devices : & [Device { vendor_id : 56026u16 , id : 307u16 , name : "VisionRGB-X2" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 313u16 , name : "VisionRGB-E1" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 324u16 , name : "VisionSD8" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 336u16 , name : "VisionRGB-E2" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 337u16 , name : "VisionSD4+1" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 345u16 , name : "VisionAV" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 353u16 , name : "DGC161" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 357u16 , name : "DGC165" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 359u16 , name : "DGC167" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 360u16 , name : "DGC168" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 4409u16 , name : "VisionRGB-E1S" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 4432u16 , name : "VisionRGB-E2S" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 4433u16 , name : "VisionSD4+1S" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 4435u16 , name : "VisionDVI-DL" , subsystems : & [] } , Device { vendor_id : 56026u16 , id : 4436u16 , name : "VisionSDI2" , subsystems : & [] }] }),
        (26470, Vendor { id : 26470u16 , name : "Glenfly Tech Co., Ltd." , devices : & [Device { vendor_id : 26470u16 , id : 15616u16 , name : "Arise-GT-10C0" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15618u16 , name : "Arise1020" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15619u16 , name : "Arise-GT-1040" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15620u16 , name : "Arise1010" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15621u16 , name : "Arise2050" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15622u16 , name : "Arise-GT-10C0t" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15623u16 , name : "Arise2030" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15624u16 , name : "Arise2020" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15625u16 , name : "Arise2020C" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15630u16 , name : "Arise10D0" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15680u16 , name : "GLF HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15681u16 , name : "GLF HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15682u16 , name : "GLF HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 26470u16 , id : 15683u16 , name : "GLF HDMI/DP Audio" , subsystems : & [] }] }),
        (8280, Vendor { id : 8280u16 , name : "Lime Microsystems Ltd." , devices : & [Device { vendor_id : 8280u16 , id : 26u16 , name : "LimeSDR X3" , subsystems : & [] } , Device { vendor_id : 8280u16 , id : 27u16 , name : "LimeSDR XTRX" , subsystems : & [] } , Device { vendor_id : 8280u16 , id : 29u16 , name : "LA9310 [LimeSDR Micro]" , subsystems : & [] }] }),
        (5247, Vendor { id : 5247u16 , name : "NIHON UNISYS, Ltd." , devices : & [] }),
        (8361, Vendor { id : 8361u16 , name : "LDA Technologies Ltd." , devices : & [Device { vendor_id : 8361u16 , id : 4104u16 , name : "NEOTAPX FPGA Accelerator Card" , subsystems : & [] } , Device { vendor_id : 8361u16 , id : 4356u16 , name : "NEOTAPX FPGA Timing Synchronization Card" , subsystems : & [] } , Device { vendor_id : 8361u16 , id : 4608u16 , name : "MUX Ultimate FPGA Accelerator Card" , subsystems : & [] } , Device { vendor_id : 8361u16 , id : 4609u16 , name : "MUX Ultimate FPGA Accelerator Card" , subsystems : & [] }] }),
        (60000, Vendor { id : 60000u16 , name : "RME" , devices : & [Device { vendor_id : 60000u16 , id : 39062u16 , name : "Digi32" , subsystems : & [] } , Device { vendor_id : 60000u16 , id : 39063u16 , name : "Digi32 Pro" , subsystems : & [] } , Device { vendor_id : 60000u16 , id : 39064u16 , name : "Digi32/8" , subsystems : & [] }] }),
        (7290, Vendor { id : 7290u16 , name : "OWC" , devices : & [] }),
        (4779, Vendor { id : 4779u16 , name : "YUAN High-Tech Development Co., Ltd." , devices : & [Device { vendor_id : 4779u16 , id : 0u16 , name : "MPG160/Kuroutoshikou ITVC15-STVLP" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 2u16 , name : "AU8830 [Vortex2] Based Sound Card With A3D Support" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 3u16 , name : "T507 (DVB-T) TV tuner/capture device" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 8960u16 , name : "Club-3D Zap TV2100" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 12288u16 , name : "MPG-200C PCI DVD Decoder Card" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 18313u16 , name : "MPC788 MiniPCI Hybrid TV Tuner" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 65523u16 , name : "MPG600/Kuroutoshikou ITVC16-STVLP" , subsystems : & [] } , Device { vendor_id : 4779u16 , id : 65535u16 , name : "MPG600/Kuroutoshikou ITVC16-STVLP" , subsystems : & [] }] }),
        (5502, Vendor { id : 5502u16 , name : "Transition Networks" , devices : & [] }),
        (4191, Vendor { id : 4191u16 , name : "Infotronic America Inc" , devices : & [] }),
        (4442, Vendor { id : 4442u16 , name : "Harlequin Ltd" , devices : & [] }),
        (7787, Vendor { id : 7787u16 , name : "Axiado Corp." , devices : & [] }),
        (8359, Vendor { id : 8359u16 , name : "EEVengers Inc." , devices : & [] }),
        (56577, Vendor { id : 56577u16 , name : "Digital Devices GmbH" , devices : & [Device { vendor_id : 56577u16 , id : 3u16 , name : "Octopus DVB Adapter" , subsystems : & [SubSystem { subvendor : 56577u16 , subdevice : 1u16 , name : "Octopus DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 2u16 , name : "Octopus LE DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 3u16 , name : "Octopus OEM" } , SubSystem { subvendor : 56577u16 , subdevice : 4u16 , name : "Octopus V3 DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 16u16 , name : "Octopus Mini" } , SubSystem { subvendor : 56577u16 , subdevice : 32u16 , name : "Cine S2 V6 DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 33u16 , name : "Cine S2 V6.5 DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 48u16 , name : "Cine CT V6.1 DVB adapter" } , SubSystem { subvendor : 56577u16 , subdevice : 56067u16 , name : "Mystique SaTiX-S2 V3 DVB adapter" }] } , Device { vendor_id : 56577u16 , id : 6u16 , name : "Cine V7" , subsystems : & [] } , Device { vendor_id : 56577u16 , id : 7u16 , name : "Max" , subsystems : & [SubSystem { subvendor : 56577u16 , subdevice : 35u16 , name : "Max S8 4/8" }] } , Device { vendor_id : 56577u16 , id : 17u16 , name : "Octopus CI DVB Adapter" , subsystems : & [SubSystem { subvendor : 56577u16 , subdevice : 64u16 , name : "Octopus CI" } , SubSystem { subvendor : 56577u16 , subdevice : 65u16 , name : "Octopus CI Single" }] } , Device { vendor_id : 56577u16 , id : 18u16 , name : "Octopus Duo CI" , subsystems : & [] } , Device { vendor_id : 56577u16 , id : 513u16 , name : "Resi DVB-C Modulator" , subsystems : & [SubSystem { subvendor : 56577u16 , subdevice : 1u16 , name : "Resi DVB-C Modulator" }] }] }),
        (4904, Vendor { id : 4904u16 , name : "quadrant international" , devices : & [] }),
        (21785, Vendor { id : 21785u16 , name : "Cnet Technologies, Inc." , devices : & [] }),
        (5105, Vendor { id : 5105u16 , name : "Oce' - Technologies B.V." , devices : & [] }),
        (48401, Vendor { id : 48401u16 , name : "Pinnacle Systems, Inc. (Wrong ID)" , devices : & [] }),
        (4287, Vendor { id : 4287u16 , name : "Most Inc" , devices : & [] }),
        (4588, Vendor { id : 4588u16 , name : "Coreco Inc" , devices : & [Device { vendor_id : 4588u16 , id : 13u16 , name : "Oculus-F/64P" , subsystems : & [] } , Device { vendor_id : 4588u16 , id : 6144u16 , name : "Cobra/C6" , subsystems : & [] }] }),
        (6737, Vendor { id : 6737u16 , name : "Hectronic AB" , devices : & [] }),
        (25372, Vendor { id : 25372u16 , name : "SmartInfra Ltd" , devices : & [Device { vendor_id : 25372u16 , id : 5714u16 , name : "PXI-1652 Signal Generator" , subsystems : & [] } , Device { vendor_id : 25372u16 , id : 9476u16 , name : "PXI-2504 Signal Interrogator" , subsystems : & [] }] }),
        (21589, Vendor { id : 21589u16 , name : "Technische Universitaet Berlin" , devices : & [Device { vendor_id : 21589u16 , id : 17496u16 , name : "S5933" , subsystems : & [] }] }),
        (6887, Vendor { id : 6887u16 , name : "First Wise Media GmbH" , devices : & [Device { vendor_id : 6887u16 , id : 1312u16 , name : "HFC-S PCI A [X-TENSIONS XC-520]" , subsystems : & [] }] }),
        (5523, Vendor { id : 5523u16 , name : "Bops Inc" , devices : & [] }),
        (8301, Vendor { id : 8301u16 , name : "GigaIO Networks, Inc." , devices : & [] }),
        (5362, Vendor { id : 5362u16 , name : "MOBILITY Electronics" , devices : & [Device { vendor_id : 5362u16 , id : 288u16 , name : "EV1000 bridge" , subsystems : & [] } , Device { vendor_id : 5362u16 , id : 289u16 , name : "EV1000 Parallel port" , subsystems : & [] } , Device { vendor_id : 5362u16 , id : 290u16 , name : "EV1000 Serial port" , subsystems : & [] } , Device { vendor_id : 5362u16 , id : 291u16 , name : "EV1000 Keyboard controller" , subsystems : & [] } , Device { vendor_id : 5362u16 , id : 292u16 , name : "EV1000 Mouse controller" , subsystems : & [] }] }),
        (5287, Vendor { id : 5287u16 , name : "MYTHOS Systems Inc" , devices : & [] }),
        (7748, Vendor { id : 7748u16 , name : "Valve Software" , devices : & [] }),
        (58596, Vendor { id : 58596u16 , name : "Xorcom" , devices : & [] }),
        (6893, Vendor { id : 6893u16 , name : "SanDisk" , devices : & [Device { vendor_id : 6893u16 , id : 4099u16 , name : "ioDimm3 (v1.2)" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 4101u16 , name : "ioDimm3" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 963u16 , name : "High IOPS SSD PCIe Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5999u16 , name : "1.28TB MLC PCIe ioDrive Duo" } , SubSystem { subvendor : 4156u16 , subdevice : 6000u16 , name : "5.2TB MLC PCIe ioDrive Octal" } , SubSystem { subvendor : 4156u16 , subdevice : 6027u16 , name : "160GB SLC PCIe ioDrive" } , SubSystem { subvendor : 4156u16 , subdevice : 6028u16 , name : "320GB MLC PCIe ioDrive" } , SubSystem { subvendor : 4156u16 , subdevice : 6029u16 , name : "320GB SLC PCIe ioDrive Duo" } , SubSystem { subvendor : 4156u16 , subdevice : 6030u16 , name : "640GB MLC PCIe ioDrive Duo" }] } , Device { vendor_id : 6893u16 , id : 4102u16 , name : "ioXtreme" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 4103u16 , name : "ioXtreme Pro" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 4104u16 , name : "ioXtreme-2" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 8193u16 , name : "ioDrive2" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 12289u16 , name : "ioMemory FHHL" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 12290u16 , name : "ioMemory HHHL" , subsystems : & [] } , Device { vendor_id : 6893u16 , id : 12291u16 , name : "ioMemory Mezzanine" , subsystems : & [] }] }),
        (8009, Vendor { id : 8009u16 , name : "NeuReality LTD" , devices : & [] }),
        (5955, Vendor { id : 5955u16 , name : "Peppercon AG" , devices : & [Device { vendor_id : 5955u16 , id : 33081u16 , name : "ROL/F-100 Fast Ethernet Adapter with ROL" , subsystems : & [] }] }),
        (4487, Vendor { id : 4487u16 , name : "Advanced Technology Laboratories, Inc." , devices : & [] }),
        (4907, Vendor { id : 4907u16 , name : "Broadband Technologies" , devices : & [] }),
        (4906, Vendor { id : 4906u16 , name : "Microcom Inc." , devices : & [] }),
        (4608, Vendor { id : 4608u16 , name : "CSS Corporation" , devices : & [] }),
        (5303, Vendor { id : 5303u16 , name : "PROXIM Inc" , devices : & [Device { vendor_id : 5303u16 , id : 1u16 , name : "Symphony 4110" , subsystems : & [] }] }),
        (5608, Vendor { id : 5608u16 , name : "National Datacomm Corp" , devices : & [Device { vendor_id : 5608u16 , id : 304u16 , name : "Wireless PCI Card" , subsystems : & [] } , Device { vendor_id : 5608u16 , id : 305u16 , name : "NCP130A2 Wireless NIC" , subsystems : & [] }] }),
        (63561, Vendor { id : 63561u16 , name : "ASRock Incorporation (Wrong ID)" , devices : & [] }),
        (6715, Vendor { id : 6715u16 , name : "AzureWave" , devices : & [Device { vendor_id : 6715u16 , id : 4370u16 , name : "AR9285 Wireless Network Adapter (PCI-Express)" , subsystems : & [] }] }),
        (7448, Vendor { id : 7448u16 , name : "RME" , devices : & [Device { vendor_id : 7448u16 , id : 1u16 , name : "Fireface UFX+" , subsystems : & [] }] }),
        (5722, Vendor { id : 5722u16 , name : "Epix Inc" , devices : & [Device { vendor_id : 5722u16 , id : 49408u16 , name : "PIXCI(R) CL1 Camera Link Video Capture Board [custom QL5232]" , subsystems : & [] } , Device { vendor_id : 5722u16 , id : 53760u16 , name : "PIXCI(R) D2X Digital Video Capture Board [custom QL5232]" , subsystems : & [] } , Device { vendor_id : 5722u16 , id : 54016u16 , name : "PIXCI(R) D3X Digital Video Capture Board [custom QL5232]" , subsystems : & [] } , Device { vendor_id : 5722u16 , id : 60161u16 , name : "PIXCI(R) EB1 PCI Camera Link Video Capture Board" , subsystems : & [] }] }),
        (6365, Vendor { id : 6365u16 , name : "Artimi Inc" , devices : & [Device { vendor_id : 6365u16 , id : 19567u16 , name : "Artimi RTMI-100 UWB adapter" , subsystems : & [] }] }),
        (4101, Vendor { id : 4101u16 , name : "Avance Logic Inc. [ALI]" , devices : & [Device { vendor_id : 4101u16 , id : 8292u16 , name : "ALG2032/2064" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 8488u16 , name : "ALG2364A" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 8961u16 , name : "ALG2301" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 8962u16 , name : "ALG2302" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 9060u16 , name : "ALG2364" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 9316u16 , name : "ALG2364A" , subsystems : & [] } , Device { vendor_id : 4101u16 , id : 9473u16 , name : "ALG2564A/25128A" , subsystems : & [] }] }),
        (4941, Vendor { id : 4941u16 , name : "PCTel Inc" , devices : & [Device { vendor_id : 4941u16 , id : 8585u16 , name : "HSP56 MicroModem" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 9350u16 , name : "2304WT V.92 MDC Modem" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30864u16 , name : "HSP MicroModem 56" , subsystems : & [SubSystem { subvendor : 4941u16 , subdevice : 1u16 , name : "PCT789 adapter" }] } , Device { vendor_id : 4941u16 , id : 30865u16 , name : "HSP MicroModem 56" , subsystems : & [SubSystem { subvendor : 4941u16 , subdevice : 1u16 , name : "HSP MicroModem 56" }] } , Device { vendor_id : 4941u16 , id : 30866u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30867u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30868u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30869u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30870u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4941u16 , id : 30871u16 , name : "HSP MicroModem 56" , subsystems : & [] }] }),
        (8011, Vendor { id : 8011u16 , name : "Axera Semiconductor Co., Ltd" , devices : & [] }),
        (4513, Vendor { id : 4513u16 , name : "Hamamatsu Photonics K.K." , devices : & [] }),
        (8184, Vendor { id : 8184u16 , name : "Beijing Gengtu Technology Co.Ltd" , devices : & [Device { vendor_id : 8184u16 , id : 8192u16 , name : "GT6910" , subsystems : & [] } , Device { vendor_id : 8184u16 , id : 8208u16 , name : "GT6908" , subsystems : & [] }] }),
        (4649, Vendor { id : 4649u16 , name : "Data Kinesis Inc." , devices : & [] }),
        (8340, Vendor { id : 8340u16 , name : "Shenzhen Wodposit Electronics Co., Ltd." , devices : & [Device { vendor_id : 8340u16 , id : 4737u16 , name : "WPBSNM8-256GTP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4738u16 , name : "WPBSNM8-512GTP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4739u16 , name : "WPBSNM8-1TTP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4740u16 , name : "WPBSNM8-256GMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4741u16 , name : "WPBSN4M8-512GMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4742u16 , name : "WPBSN4M8-1TMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 4743u16 , name : "WPBSN4M8-2TMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5729u16 , name : "WPBSNM8-512GMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5730u16 , name : "WPBSNM8-1TMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5731u16 , name : "WPBSNM8-2TMP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5732u16 , name : "WPBSN4M8-1TGP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5733u16 , name : "WPBSN4M8-2TGP" , subsystems : & [] } , Device { vendor_id : 8340u16 , id : 5734u16 , name : "WPBSN4M8-4TGP" , subsystems : & [] }] }),
        (8185, Vendor { id : 8185u16 , name : "Inagile Electronic Technology Co., LTD" , devices : & [] }),
        (5480, Vendor { id : 5480u16 , name : "DDK Electronics Inc" , devices : & [] }),
        (6792, Vendor { id : 6792u16 , name : "Duagon AG" , devices : & [Device { vendor_id : 6792u16 , id : 19781u16 , name : "Multifunction IP core" , subsystems : & [] }] }),
        (7365, Vendor { id : 7365u16 , name : "Embedded Intelligence, Inc." , devices : & [Device { vendor_id : 7365u16 , id : 256u16 , name : "PCIe-CAN-02 Dual CAN bus (9-pin male). PCI Express x1." , subsystems : & [] } , Device { vendor_id : 7365u16 , id : 257u16 , name : "PCIe-CAN-01 Single CAN bus (9-pin male). PCI Express x1." , subsystems : & [] }] }),
        (4356, Vendor { id : 4356u16 , name : "RasterOps Corp." , devices : & [] }),
        (7964, Vendor { id : 7964u16 , name : "Sophgo Technologies Inc." , devices : & [Device { vendor_id : 7964u16 , id : 5766u16 , name : "BM1684X [Sophon Series Deep Learning Accelerator]" , subsystems : & [] }] }),
        (7271, Vendor { id : 7271u16 , name : "PreSonus Audio Electronics Inc." , devices : & [Device { vendor_id : 7271u16 , id : 257u16 , name : "Quantum" , subsystems : & [] } , Device { vendor_id : 7271u16 , id : 258u16 , name : "Quantum 2" , subsystems : & [] } , Device { vendor_id : 7271u16 , id : 259u16 , name : "Quantum 4848" , subsystems : & [] } , Device { vendor_id : 7271u16 , id : 260u16 , name : "Quantum 2626" , subsystems : & [] } , Device { vendor_id : 7271u16 , id : 261u16 , name : "Quantum Mobile" , subsystems : & [] }] }),
        (6380, Vendor { id : 6380u16 , name : "Cesnet, z.s.p.o." , devices : & [Device { vendor_id : 6380u16 , id : 27909u16 , name : "ML555" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 256u16 , name : "NIC (ethernet interfaces)" } , SubSystem { subvendor : 6380u16 , subdevice : 512u16 , name : "NIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 513u16 , name : "NIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 768u16 , name : "NIFIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 770u16 , name : "NIFIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 16896u16 , name : "Flexible FlowMon (szedata2) 1x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 65280u16 , name : "Testing design" } , SubSystem { subvendor : 6380u16 , subdevice : 65281u16 , name : "Boot design" }] } , Device { vendor_id : 6380u16 , id : 49158u16 , name : "COMBO6" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 53249u16 , name : "COMBO-4MTX" } , SubSystem { subvendor : 6380u16 , subdevice : 53250u16 , name : "COMBO-4SFP" } , SubSystem { subvendor : 6380u16 , subdevice : 53251u16 , name : "COMBO-4SFPRO" } , SubSystem { subvendor : 6380u16 , subdevice : 53252u16 , name : "COMBO-2XFP" }] } , Device { vendor_id : 6380u16 , id : 49202u16 , name : "COMBO-LXT110" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 256u16 , name : "NIC (ethernet interfaces)" } , SubSystem { subvendor : 6380u16 , subdevice : 512u16 , name : "NIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 513u16 , name : "NIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 768u16 , name : "NIFIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 770u16 , name : "NIFIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 16896u16 , name : "Flexible FlowMon (szedata2) 1x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 65280u16 , name : "Testing design" } , SubSystem { subvendor : 6380u16 , subdevice : 65281u16 , name : "Boot design" }] } , Device { vendor_id : 6380u16 , id : 49221u16 , name : "COMBO6E" , subsystems : & [] } , Device { vendor_id : 6380u16 , id : 49232u16 , name : "COMBO-PTM" , subsystems : & [] } , Device { vendor_id : 6380u16 , id : 49240u16 , name : "COMBO6X" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 53249u16 , name : "COMBO-4MTX" } , SubSystem { subvendor : 6380u16 , subdevice : 53250u16 , name : "COMBO-4SFP" } , SubSystem { subvendor : 6380u16 , subdevice : 53251u16 , name : "COMBO-4SFPRO" } , SubSystem { subvendor : 6380u16 , subdevice : 53252u16 , name : "COMBO-2XFP" }] } , Device { vendor_id : 6380u16 , id : 49458u16 , name : "COMBO-LXT155" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 256u16 , name : "NIC (ethernet interfaces)" } , SubSystem { subvendor : 6380u16 , subdevice : 512u16 , name : "NIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 513u16 , name : "NIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 768u16 , name : "NIFIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 770u16 , name : "NIFIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 16896u16 , name : "Flexible FlowMon (szedata2) 1x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 65280u16 , name : "Testing design" } , SubSystem { subvendor : 6380u16 , subdevice : 65281u16 , name : "Boot design" }] } , Device { vendor_id : 6380u16 , id : 49714u16 , name : "COMBO-FXT100" , subsystems : & [SubSystem { subvendor : 6380u16 , subdevice : 256u16 , name : "NIC (ethernet interfaces)" } , SubSystem { subvendor : 6380u16 , subdevice : 512u16 , name : "NIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 513u16 , name : "NIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 768u16 , name : "NIFIC (szedata2) 4x1G" } , SubSystem { subvendor : 6380u16 , subdevice : 770u16 , name : "NIFIC (szedata2) 2x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 16896u16 , name : "Flexible FlowMon (szedata2) 1x10G" } , SubSystem { subvendor : 6380u16 , subdevice : 65280u16 , name : "Testing design" } , SubSystem { subvendor : 6380u16 , subdevice : 65281u16 , name : "Boot design" }] } , Device { vendor_id : 6380u16 , id : 50176u16 , name : "COMBO-400G1" , subsystems : & [] }] }),
        (20549, Vendor { id : 20549u16 , name : "University of Toronto" , devices : & [Device { vendor_id : 20549u16 , id : 16963u16 , name : "BLASTbus PCI Interface Card v1" , subsystems : & [] }] }),
        (4277, Vendor { id : 4277u16 , name : "PLX Technology, Inc." , devices : & [Device { vendor_id : 4277u16 , id : 1u16 , name : "i960 PCI bus interface" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 1367u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "Digium Tormenta 2 T400P-SS7 or E400P-SS7 Quad T1 or E1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 4096u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "ATCOM AT400P Quad T1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 4132u16 , name : "Acromag, Inc. IndustryPack Carrier Card" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4162u16 , name : "Brandywine / jxi2, Inc. - PMC-SyncClock32, IRIG A & B, Nasa 36" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4202u16 , name : "Dual OX16C952 4 port serial adapter [Megawolf Romulus/4]" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4214u16 , name : "VScom 800 8 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4215u16 , name : "VScom 400 4 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4216u16 , name : "VScom 210 2 port serial and 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4355u16 , name : "VScom 200 2 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4422u16 , name : "VScom 010 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 4423u16 , name : "VScom 020 2 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 8192u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "ATCOM AE400P Quad E1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 8960u16 , name : "Euresys DOMINO Gamma" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 9076u16 , name : "Euresys DOMINO Alpha" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 9361u16 , name : "Euresys GRABLINK Value" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 9363u16 , name : "Euresys GRABLINK Expert" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 9536u16 , name : "IXXAT CAN-Interface PC-I 04/PCI" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 10020u16 , name : "Thales PCSM Security Card" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 13174u16 , name : "Cosateq 4 Port CAN Card" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 16384u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "Tormenta 3 Varion V400P/ATCOM TE400P Quad E1/T1/J1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 16385u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "ATCOM A400PE Quad E1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 16386u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "ATCOM A400PT Quad T1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 24896u16 , name : "PCI6140 32-bit 33MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 24912u16 , name : "PCI6150 32-bit 33MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 24914u16 , name : "PCI6152 32-bit 66MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 24916u16 , name : "PCI6154 64-bit 66MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 25172u16 , name : "PCI6254 64-bit 66MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 25702u16 , name : "PCI6466 64-bit 66MHz PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 25888u16 , name : "PCI6520 64-bit 133MHz PCI-X-to-PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 25920u16 , name : "PCI6540 64-bit 133MHz PCI-X-to-PCI-X Bridge" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 4277u16 , id : 25921u16 , name : "PCI6540/6466 PCI-PCI bridge (non-transparent mode, primary side)" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 4277u16 , id : 25922u16 , name : "PCI6540/6466 PCI-PCI bridge (non-transparent mode, secondary side)" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 4277u16 , id : 33041u16 , name : "PEX 8111 PCI Express-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 33042u16 , name : "PEX8112 x1 Lane PCI Express-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 33044u16 , name : "PEX 8114 PCI Express-to-PCI/PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 33553u16 , name : "PEX8311 x1 Lane PCI Express-to-Generic Local Bus Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34053u16 , name : "PEX 8505 5-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34056u16 , name : "PEX 8508 8-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34057u16 , name : "PEX 8509 8-lane, 8-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34066u16 , name : "PEX 8512 12-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34070u16 , name : "PEX 8516  Versatile PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34071u16 , name : "PEX 8517 16-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34072u16 , name : "PEX 8518 16-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34084u16 , name : "PEX 8524 24-lane, 6-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34085u16 , name : "PEX 8525 24-lane, 5-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34098u16 , name : "PEX 8532  Versatile PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34099u16 , name : "PEX 8533 32-lane, 6-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34119u16 , name : "PEX 8547 48-lane, 3-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34120u16 , name : "PEX 8548 48-lane, 9-port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34307u16 , name : "PEX 8603 3-lane, 3-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34308u16 , name : "PEX 8604 4-lane, 4-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34309u16 , name : "PEX 8605 PCI Express 4-port Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34310u16 , name : "PEX 8606 6 Lane, 6 Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34312u16 , name : "PEX 8608 8-lane, 8-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34313u16 , name : "PEX 8609 8-lane, 8-Port PCI Express Gen 2 (5.0 GT/s) Switch with DMA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34322u16 , name : "PEX 8612 12-lane, 4-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34323u16 , name : "PEX 8613 12-lane, 3-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34324u16 , name : "PEX 8614 12-lane, 12-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34325u16 , name : "PEX 8615 12-lane, 12-Port PCI Express Gen 2 (5.0 GT/s) Switch with DMA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34326u16 , name : "PEX 8616 16-lane, 4-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34327u16 , name : "PEX 8617 16-lane, 4-Port PCI Express Gen 2 (5.0 GT/s) Switch with P2P" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34328u16 , name : "PEX 8618 16-lane, 16-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34329u16 , name : "PEX 8619 16-lane, 16-Port PCI Express Gen 2 (5.0 GT/s) Switch with DMA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34340u16 , name : "PEX 8624 24-lane, 6-Port PCI Express Gen 2 (5.0 GT/s) Switch [ExpressLane]" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 6213u16 , name : "DX1845 Acceleration Card" }] } , Device { vendor_id : 4277u16 , id : 34341u16 , name : "PEX 8625 24-lane, 24-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34354u16 , name : "PEX 8632 32-lane, 12-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34358u16 , name : "PEX 8636 36-lane, 24-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34375u16 , name : "PEX 8647 48-Lane, 3-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34376u16 , name : "PEX 8648 48-lane, 12-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34377u16 , name : "PEX 8649 48-lane, 12-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34404u16 , name : "PEX 8664 64-lane, 16-Port PCI Express Gen 2 (5.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34432u16 , name : "PEX 8680 80-lane, 20-Port PCI Express Gen 2 (5.0 GT/s) Multi-Root Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34454u16 , name : "PEX 8696 96-lane, 24-Port PCI Express Gen 2 (5.0 GT/s) Multi-Root Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34580u16 , name : "PEX 8714 12-Lane, 5-Port PCI Express Gen 3 (8 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34583u16 , name : "PEX 8717 16-lane, 8-Port PCI Express Gen 3 (8.0 GT/s) Switch with DMA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34584u16 , name : "PEX 8718 16-Lane, 5-Port PCI Express Gen 3 (8.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34595u16 , name : "PEX 8723 24-Lane, 6-Port PCI Express Gen 3 (8 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34596u16 , name : "PEX 8724 24-Lane, 6-Port PCI Express Gen 3 (8 GT/s) Switch, 19 x 19mm FCBGA" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 37428u16 , name : "LRNV9324 2-port Built-in 8643 NVMe Exchange Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38180u16 , name : "LRNV9524 2-port M.2 NVMe SSD Exchange Adapter" }] } , Device { vendor_id : 4277u16 , id : 34597u16 , name : "PEX 8725 24-Lane, 10-Port PCI Express Gen 3 (8.0 GT/s) Multi-Root Switch with DMA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34610u16 , name : "PEX 8732 32-lane, 8-Port PCI Express Gen 3 (8.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34611u16 , name : "PEX 8733 32-Lane, 19-Port PCI Express Gen 3 (8 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34612u16 , name : "PEX 8734 32-lane, 8-Port PCI Express Gen 3 (8.0GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34631u16 , name : "PEX 8747 48-Lane, 5-Port PCI Express Gen 3 (8.0 GT/s) Switch" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 37703u16 , name : "LRNV9347L 2-port Built-in 8643 NVMe Switching Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38215u16 , name : "LRNV9547 4-port M.2 NVMe SSD Exchange Adapter" }] } , Device { vendor_id : 4277u16 , id : 34632u16 , name : "PEX 8748 48-Lane, 12-Port PCI Express Gen 3 (8 GT/s) Switch, 27 x 27mm FCBGA" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34633u16 , name : "PEX 8749 48-Lane, 18-Port PCI Express Gen 3 (8.0 GT/s) Multi-Root Switch with DMA" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 37705u16 , name : "LRNV9349 8-port SFF-8643 NVMe SSD Exchange Adapter" }] } , Device { vendor_id : 4277u16 , id : 34660u16 , name : "PEX 8764 64-Lane, 16-Port PCI Express Gen 3 (8 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34688u16 , name : "PEX 8780 80-Lane, 20-Port PCI Express Gen 3 (8 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34710u16 , name : "PEX8796 96-Lane, 24-Port PCIe Gen 3 (8.0 GT/s) Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34720u16 , name : "PEX PCI Express Switch NT0 Port Link Interface" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34721u16 , name : "PEX PCI Express Switch NT1 Port Link Interface" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34736u16 , name : "PEX PCI Express Switch NT0 Port Virtual Interface" , subsystems : & [SubSystem { subvendor : 4243u16 , subdevice : 30561u16 , name : "PXIe-8830mc" }] } , Device { vendor_id : 4277u16 , id : 34737u16 , name : "PEX PCI Express Switch NT1 Port Virtual Interface" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 34768u16 , name : "PEX PCI Express Switch DMA interface" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36886u16 , name : "PLX 9016 8-port serial controller" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36912u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 4613u16 , name : "Becker & Hickl MSA-1000" } , SubSystem { subvendor : 4277u16 , subdevice : 9877u16 , name : "Hilscher CIF50-PB/DPS Profibus" } , SubSystem { subvendor : 4277u16 , subdevice : 10338u16 , name : "Alpermann+Velte PCL PCI LV (3V/5V): Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 10502u16 , name : "Alpermann+Velte PCI TS (3V/5V): Time Synchronisation Board" } , SubSystem { subvendor : 4277u16 , subdevice : 10560u16 , name : "Alpermann+Velte PCL PCI D (3V/5V): Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 10615u16 , name : "IXXAT iPC-I XC16/PCI CAN Board" } , SubSystem { subvendor : 4277u16 , subdevice : 10616u16 , name : "SH ARC-PCIu/SH ARC-PCI104/SH ARC-PCIe SOHARD ARCNET card" } , SubSystem { subvendor : 4277u16 , subdevice : 12325u16 , name : "Alpermann+Velte PCL PCI L (3V/5V): Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 12392u16 , name : "Alpermann+Velte PCL PCI HD (3V/5V): Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13411u16 , name : "Alpermann+Velte PCL PCI D (v2) (3V/5V): Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13713u16 , name : "PLURA PCL PCI L (v2) (3.3V/5V): Time Code Reader Board" } , SubSystem { subvendor : 4862u16 , subdevice : 273u16 , name : "CPCI-ASIO4 (ESD 4-port Serial Interface Board)" } , SubSystem { subvendor : 4969u16 , subdevice : 39937u16 , name : "VX222v2" } , SubSystem { subvendor : 4969u16 , subdevice : 40193u16 , name : "VX222-Mic" } , SubSystem { subvendor : 4969u16 , subdevice : 40194u16 , name : "VX222-Mic" } , SubSystem { subvendor : 4969u16 , subdevice : 40449u16 , name : "PCX924v2" } , SubSystem { subvendor : 4969u16 , subdevice : 40705u16 , name : "PCX924-Mic" } , SubSystem { subvendor : 4969u16 , subdevice : 40706u16 , name : "PCX924-Mic" } , SubSystem { subvendor : 4969u16 , subdevice : 40961u16 , name : "PCX22v2" } , SubSystem { subvendor : 4969u16 , subdevice : 42753u16 , name : "LCM220v2" } , SubSystem { subvendor : 4969u16 , subdevice : 43009u16 , name : "LCM200" } , SubSystem { subvendor : 5015u16 , subdevice : 12598u16 , name : "4xS0-ISDN PCI Adapter" } , SubSystem { subvendor : 5015u16 , subdevice : 12599u16 , name : "S2M-E1-ISDN PCI Adapter" } , SubSystem { subvendor : 5400u16 , subdevice : 512u16 , name : "ThinkIO-C" } , SubSystem { subvendor : 5613u16 , subdevice : 4098u16 , name : "MCCS 8-port Serial Hot Swap" } , SubSystem { subvendor : 5613u16 , subdevice : 4099u16 , name : "MCCS 16-port Serial Hot Swap" } , SubSystem { subvendor : 5985u16 , subdevice : 4386u16 , name : "Pickering Interfaces 50-297-010" } , SubSystem { subvendor : 5985u16 , subdevice : 4387u16 , name : "Pickering Interfaces 50-297-011" } , SubSystem { subvendor : 5985u16 , subdevice : 4388u16 , name : "Pickering Interfaces 50-297-012" } , SubSystem { subvendor : 5985u16 , subdevice : 4389u16 , name : "Pickering Interfaces 50-297-013" } , SubSystem { subvendor : 5985u16 , subdevice : 4390u16 , name : "Pickering Interfaces 50-297-020" } , SubSystem { subvendor : 5985u16 , subdevice : 4393u16 , name : "Pickering Interfaces 50-297-023" } , SubSystem { subvendor : 5985u16 , subdevice : 4395u16 , name : "Pickering Interfaces 50-297-030" } , SubSystem { subvendor : 5985u16 , subdevice : 4396u16 , name : "Pickering Interfaces 50-297-031" } , SubSystem { subvendor : 5985u16 , subdevice : 4397u16 , name : "Pickering Interfaces 50-297-032" } , SubSystem { subvendor : 5985u16 , subdevice : 4398u16 , name : "Pickering Interfaces 50-297-033" } , SubSystem { subvendor : 5985u16 , subdevice : 4399u16 , name : "Pickering Interfaces 50-297-034" } , SubSystem { subvendor : 5985u16 , subdevice : 4400u16 , name : "Pickering Interfaces 50-297-040" } , SubSystem { subvendor : 5985u16 , subdevice : 4401u16 , name : "Pickering Interfaces 50-297-041" } , SubSystem { subvendor : 5985u16 , subdevice : 4402u16 , name : "Pickering Interfaces 50-297-042" } , SubSystem { subvendor : 5985u16 , subdevice : 4403u16 , name : "Pickering Interfaces 50-297-043" } , SubSystem { subvendor : 5985u16 , subdevice : 4404u16 , name : "Pickering Interfaces 50-297-044" } , SubSystem { subvendor : 5985u16 , subdevice : 4405u16 , name : "Pickering Interfaces 50-297-050" } , SubSystem { subvendor : 5985u16 , subdevice : 4409u16 , name : "Pickering Interfaces 50-297-113" } , SubSystem { subvendor : 57797u16 , subdevice : 1u16 , name : "TE1-PCI" } , SubSystem { subvendor : 57797u16 , subdevice : 5u16 , name : "TA1-PCI" } , SubSystem { subvendor : 57797u16 , subdevice : 6u16 , name : "TA1-PCI4" }] } , Device { vendor_id : 4277u16 , id : 36918u16 , name : "9036" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36944u16 , name : "PCI9050 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4272u16 , name : "82350 PCI GPIB" } , SubSystem { subvendor : 4277u16 , subdevice : 4199u16 , name : "IXXAT CAN i165" } , SubSystem { subvendor : 4277u16 , subdevice : 4430u16 , name : "Wasco WITIO PCI168extended" } , SubSystem { subvendor : 4277u16 , subdevice : 4457u16 , name : "Wasco OPTOIO32standard 32 digital in, 32 digital out" } , SubSystem { subvendor : 4277u16 , subdevice : 4465u16 , name : "Becker & Hickl PMS-400" } , SubSystem { subvendor : 4277u16 , subdevice : 4466u16 , name : "IK220 (Heidenhain)" } , SubSystem { subvendor : 4277u16 , subdevice : 4609u16 , name : "Becker & Hickl SPC-6x0" } , SubSystem { subvendor : 4277u16 , subdevice : 4610u16 , name : "Becker & Hickl SPC-7x0" } , SubSystem { subvendor : 4277u16 , subdevice : 4611u16 , name : "Becker & Hickl MSA-300" } , SubSystem { subvendor : 4277u16 , subdevice : 4614u16 , name : "Becker & Hickl DCC-100" } , SubSystem { subvendor : 4277u16 , subdevice : 4618u16 , name : "Becker & Hickl STP-340" } , SubSystem { subvendor : 4277u16 , subdevice : 8246u16 , name : "SatPak GPS" } , SubSystem { subvendor : 4277u16 , subdevice : 8737u16 , name : "Alpermann+Velte PCL PCI LV: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 8819u16 , name : "SH ARC-PCI SOHARD ARCNET card" } , SubSystem { subvendor : 4277u16 , subdevice : 9265u16 , name : "Alpermann+Velte PCL PCI D: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 10501u16 , name : "Alpermann+Velte PCI TS: Time Synchronisation Board" } , SubSystem { subvendor : 4277u16 , subdevice : 12694u16 , name : "Goramo PLX200SYN sync serial card" } , SubSystem { subvendor : 4277u16 , subdevice : 36944u16 , name : "PCI-I04 PCI Passive PC/CAN Interface" } , SubSystem { subvendor : 4521u16 , subdevice : 21300u16 , name : "PDS4" } , SubSystem { subvendor : 4862u16 , subdevice : 1u16 , name : "CAN-PCI/331 CAN bus controller" } , SubSystem { subvendor : 4969u16 , subdevice : 35073u16 , name : "PCX11+ PCI" } , SubSystem { subvendor : 4969u16 , subdevice : 36609u16 , name : "VX222" } , SubSystem { subvendor : 4969u16 , subdevice : 37889u16 , name : "PCX924" } , SubSystem { subvendor : 4969u16 , subdevice : 38145u16 , name : "PCX22" } , SubSystem { subvendor : 5272u16 , subdevice : 866u16 , name : "TPMC866 8 Channel Serial Card" } , SubSystem { subvendor : 5410u16 , subdevice : 1u16 , name : "RockForce 4 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 2u16 , name : "RockForce 2 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 3u16 , name : "RockForce 6 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 4u16 , name : "RockForce 8 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 16u16 , name : "RockForce2000 4 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 32u16 , name : "RockForce2000 2 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5613u16 , subdevice : 4096u16 , name : "Macrolink MCCS 8-port Serial" } , SubSystem { subvendor : 5613u16 , subdevice : 4097u16 , name : "Macrolink MCCS 16-port Serial" } , SubSystem { subvendor : 5613u16 , subdevice : 4098u16 , name : "Macrolink MCCS 8-port Serial Hot Swap" } , SubSystem { subvendor : 5613u16 , subdevice : 4099u16 , name : "Macrolink MCCS 16-port Serial Hot Swap" } , SubSystem { subvendor : 5985u16 , subdevice : 398u16 , name : "40-199-002" } , SubSystem { subvendor : 5985u16 , subdevice : 605u16 , name : "40-662-001" } , SubSystem { subvendor : 22100u16 , subdevice : 8246u16 , name : "OpenSwitch 6 Telephony card" } , SubSystem { subvendor : 22100u16 , subdevice : 12594u16 , name : "OpenSwitch 12 Telephony card" } , SubSystem { subvendor : 22100u16 , subdevice : 22068u16 , name : "OpenLine4 Telephony Card" } , SubSystem { subvendor : 54577u16 , subdevice : 49154u16 , name : "PCIntelliCAN 2xSJA1000 CAN bus" } , SubSystem { subvendor : 55373u16 , subdevice : 16390u16 , name : "EX-4006 1P" } , SubSystem { subvendor : 55373u16 , subdevice : 16392u16 , name : "EX-4008 1P EPP/ECP" } , SubSystem { subvendor : 55373u16 , subdevice : 16404u16 , name : "EX-4014 2P" } , SubSystem { subvendor : 55373u16 , subdevice : 16408u16 , name : "EX-4018 3P EPP/ECP" } , SubSystem { subvendor : 55373u16 , subdevice : 16421u16 , name : "EX-4025 1S(16C550) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16423u16 , name : "EX-4027 1S(16C650) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16424u16 , name : "EX-4028 1S(16C850) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16438u16 , name : "EX-4036 2S(16C650) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16439u16 , name : "EX-4037 2S(16C650) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16440u16 , name : "EX-4038 2S(16C850) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16466u16 , name : "EX-4052 1S(16C550) RS-422/485" } , SubSystem { subvendor : 55373u16 , subdevice : 16467u16 , name : "EX-4053 2S(16C550) RS-422/485" } , SubSystem { subvendor : 55373u16 , subdevice : 16469u16 , name : "EX-4055 4S(16C550) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16472u16 , name : "EX-4055 4S(16C650) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16485u16 , name : "EX-4065 8S(16C550) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16488u16 , name : "EX-4068 8S(16C650) RS-232" } , SubSystem { subvendor : 55373u16 , subdevice : 16504u16 , name : "EX-4078 2S(16C552) RS-232+1P" }] } , Device { vendor_id : 4277u16 , id : 36946u16 , name : "PCI9052 PCI <-> IOBus Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36948u16 , name : "PCI9054 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 4465u16 , name : "Becker & Hickl PMS-400A" } , SubSystem { subvendor : 4277u16 , subdevice : 4616u16 , name : "Becker & Hickl SPC-830" } , SubSystem { subvendor : 4277u16 , subdevice : 4622u16 , name : "Becker & Hickl SPC-930" } , SubSystem { subvendor : 4277u16 , subdevice : 4623u16 , name : "Becker & Hickl SPC-150" } , SubSystem { subvendor : 4277u16 , subdevice : 4624u16 , name : "Becker & Hickl DPC-230" } , SubSystem { subvendor : 4277u16 , subdevice : 9301u16 , name : "Wessex Techology PHIL-PCI" } , SubSystem { subvendor : 4277u16 , subdevice : 9878u16 , name : "Innes Corp AM Radcap card" } , SubSystem { subvendor : 4277u16 , subdevice : 10007u16 , name : "Innes Corp Auricon card" } , SubSystem { subvendor : 4277u16 , subdevice : 10308u16 , name : "Innes Corp TVS Encoder card" } , SubSystem { subvendor : 4807u16 , subdevice : 16385u16 , name : "Intel Dialogic DM/V960-4T1 PCI" } , SubSystem { subvendor : 4825u16 , subdevice : 2u16 , name : "PCI Prosody Card rev 1.5" } , SubSystem { subvendor : 5300u16 , subdevice : 53504u16 , name : "Dektec DTA-100" } , SubSystem { subvendor : 5300u16 , subdevice : 53524u16 , name : "Dektec DTA-120" } , SubSystem { subvendor : 5855u16 , subdevice : 17u16 , name : "PIKA PrimeNet MM PCI" } , SubSystem { subvendor : 5855u16 , subdevice : 18u16 , name : "PIKA PrimeNet MM cPCI 8" } , SubSystem { subvendor : 5855u16 , subdevice : 19u16 , name : "PIKA PrimeNet MM cPCI 8 (without CAS Signaling)" } , SubSystem { subvendor : 5855u16 , subdevice : 20u16 , name : "PIKA PrimeNet MM cPCI 4" } , SubSystem { subvendor : 5855u16 , subdevice : 21u16 , name : "PIKA Daytona MM" } , SubSystem { subvendor : 5855u16 , subdevice : 22u16 , name : "PIKA InLine MM" }] } , Device { vendor_id : 4277u16 , id : 36950u16 , name : "PCI9056 32-bit 66MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 10617u16 , name : "CellinkBlade 11 - CPCI board VoATM AAL1" } , SubSystem { subvendor : 4277u16 , subdevice : 12904u16 , name : "IXXAT iPC-I XC16/PCIe CAN Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13108u16 , name : "Cambridge Pixel HPx Radar Input Card" } , SubSystem { subvendor : 4277u16 , subdevice : 13138u16 , name : "Alpermann+Velte PCL PCIe HD: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13139u16 , name : "Alpermann+Velte PCL PCIe D: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13140u16 , name : "Plura PCL PCIe LV: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13141u16 , name : "Plura PCL PCIe L: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13333u16 , name : "Plura PCIe TS: Time Synchronisation Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13459u16 , name : "Plura PCL PCIe 3G: Timecode Reader Board" } , SubSystem { subvendor : 4277u16 , subdevice : 13669u16 , name : "Cambridge Pixel HPx Radar Output Card" } , SubSystem { subvendor : 4969u16 , subdevice : 49153u16 , name : "LX6464ES" } , SubSystem { subvendor : 4969u16 , subdevice : 49665u16 , name : "LX1616ES" } , SubSystem { subvendor : 5300u16 , subdevice : 53514u16 , name : "DekTec DTA-110T" } , SubSystem { subvendor : 5300u16 , subdevice : 53544u16 , name : "Dektec DTA-140" } , SubSystem { subvendor : 5300u16 , subdevice : 53568u16 , name : "Dektec DTA-140" } , SubSystem { subvendor : 6670u16 , subdevice : 111u16 , name : "Dektec DTA-111" }] } , Device { vendor_id : 4277u16 , id : 36960u16 , name : "PCI9060 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36973u16 , name : "9060SD" , subsystems : & [SubSystem { subvendor : 4700u16 , subdevice : 1600u16 , name : "Aries 16000P" }] } , Device { vendor_id : 4277u16 , id : 36974u16 , name : "9060ES" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 36992u16 , name : "PCI9080 32-bit; 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4331u16 , name : "(Agilent) E2777B 83K Series Optical Communication Interface" } , SubSystem { subvendor : 4156u16 , subdevice : 4332u16 , name : "(Agilent) E6978-66442 PCI CIC" } , SubSystem { subvendor : 4277u16 , subdevice : 4387u16 , name : "Sectra KK631 encryption board" } , SubSystem { subvendor : 4277u16 , subdevice : 36992u16 , name : "9080 [real subsystem ID not set]" } , SubSystem { subvendor : 4825u16 , subdevice : 2u16 , name : "PCI Prosody Card" } , SubSystem { subvendor : 4831u16 , subdevice : 17442u16 , name : "4422PCI [\"Do-All\" Telemetry Data Acquisition System]" } , SubSystem { subvendor : 4969u16 , subdevice : 38401u16 , name : "PCX822np" } , SubSystem { subvendor : 4969u16 , subdevice : 41218u16 , name : "PCX822v2" } , SubSystem { subvendor : 4969u16 , subdevice : 41473u16 , name : "PCX442" } , SubSystem { subvendor : 4969u16 , subdevice : 41729u16 , name : "LCM440v2" } , SubSystem { subvendor : 4969u16 , subdevice : 41985u16 , name : "VX822" } , SubSystem { subvendor : 4969u16 , subdevice : 41986u16 , name : "VX822v2" } , SubSystem { subvendor : 4969u16 , subdevice : 43265u16 , name : "LCM420" } , SubSystem { subvendor : 4969u16 , subdevice : 43521u16 , name : "VX820v2" } , SubSystem { subvendor : 5399u16 , subdevice : 11u16 , name : "ECSG-1R3ADC-PMC Clock synthesizer" }] } , Device { vendor_id : 4277u16 , id : 38486u16 , name : "PCI9656 PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 5399u16 , subdevice : 15u16 , name : "ECDR-GC314-PMC Receiver" } , SubSystem { subvendor : 6277u16 , subdevice : 1792u16 , name : "Tsunami FPGA PMC with Altera Stratix S40" } , SubSystem { subvendor : 6277u16 , subdevice : 1793u16 , name : "Tsunami FPGA PMC with Altera Stratix S30" }] } , Device { vendor_id : 4277u16 , id : 38674u16 , name : "PEX9712 12-Lane, 5-Port PCIe Gen 3 (8.0 GT/s) ExpressFabric Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 38707u16 , name : "PEX 9733 33-lane, 9-port PCI Express Gen 3 (8.0 GT/s) Switch" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8135u16 , name : "UCEA-200 0235NK NVMe PCIe Extender Adapter with Full-Height Bracket" } , SubSystem { subvendor : 4136u16 , subdevice : 8136u16 , name : "UCEA-200 0CDC7W NVMe PCIe Extender Adapter with Half-Height Bracket" } , SubSystem { subvendor : 7497u16 , subdevice : 1u16 , name : "ThinkSystem 1610-4P NVMe Switch Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 2u16 , name : "ThinkSystem 810-4P NVMe Switch Adapter" }] } , Device { vendor_id : 4277u16 , id : 38729u16 , name : "PEX 9749 49-lane, 13-port PCI Express Gen 3 (8.0 GT/s) Switch" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 4u16 , name : "ThinkSystem 1610-8P NVMe Switch Adapter" }] } , Device { vendor_id : 4277u16 , id : 38757u16 , name : "PEX9765 65-Lane, 17-Port PCIe Gen 3 (8.0 GT/s) ExpressFabric Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 38785u16 , name : "PEX9781 81-Lane, 21-Port PCIe Gen 3 (8.0 GT/s) ExpressFabric Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 38807u16 , name : "PEX9781 97-Lane, 25-Port PCIe Gen 3 (8.0 GT/s) ExpressFabric Switch" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 41216u16 , name : "Blackmagic Design DeckLink" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 47876u16 , name : "B&B 3PCIOSD1A Isolated PCI Serial" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 49153u16 , name : "CronyxOmega-PCI (8-port RS232)" , subsystems : & [] } , Device { vendor_id : 4277u16 , id : 53261u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "Digium Tormenta 2 T400P or E400P Quad T1 or E1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 54077u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "Tormenta 3 Varion V401PT Quad T1/J1 PCI card" }] } , Device { vendor_id : 4277u16 , id : 54349u16 , name : "PCI9030 32-bit 33MHz PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 4277u16 , subdevice : 6134u16 , name : "Allo CP100P/E 1-port E1/T1/J1 PCI/PCIe card" } , SubSystem { subvendor : 4277u16 , subdevice : 6135u16 , name : "Allo CP400P/E 4-port E1/T1/J1 PCI/PCIe card" } , SubSystem { subvendor : 4277u16 , subdevice : 6136u16 , name : "Allo CP200P/E 2-port E1/T1/J1 PCI/PCIe card" } , SubSystem { subvendor : 4277u16 , subdevice : 36912u16 , name : "Tormenta 3 Varion V401PE Quad E1 PCI card" }] }] }),
        (10773, Vendor { id : 10773u16 , name : "3D Vision(???)" , devices : & [] }),
        (5492, Vendor { id : 5492u16 , name : "Fairchild Semiconductor" , devices : & [] }),
        (7926, Vendor { id : 7926u16 , name : "GrAI Matter Labs" , devices : & [] }),
        (5332, Vendor { id : 5332u16 , name : "Panacom Technology Corp" , devices : & [] }),
        (5250, Vendor { id : 5250u16 , name : "ISYTEC - Integrierte Systemtechnik GmBH" , devices : & [Device { vendor_id : 5250u16 , id : 1u16 , name : "PCI-16 Host Interface for ITC-16" , subsystems : & [] }] }),
        (7492, Vendor { id : 7492u16 , name : "DPT" , devices : & [Device { vendor_id : 7492u16 , id : 41984u16 , name : "PM2x24/PM3224" , subsystems : & [] }] }),
        (1074, Vendor { id : 1074u16 , name : "SCM Microsystems, Inc." , devices : & [Device { vendor_id : 1074u16 , id : 1u16 , name : "Pluto2 DVB-T Receiver for PCMCIA [EasyWatch MobilSet]" , subsystems : & [] }] }),
        (5618, Vendor { id : 5618u16 , name : "Diagnostic Instruments Inc" , devices : & [] }),
        (4549, Vendor { id : 4549u16 , name : "Shiva Corporation" , devices : & [] }),
        (6382, Vendor { id : 6382u16 , name : "Chenming Mold Ind. Corp." , devices : & [] }),
        (6436, Vendor { id : 6436u16 , name : "AMD Solarflare" , devices : & [Device { vendor_id : 6436u16 , id : 1795u16 , name : "SFC4000 rev A net [Solarstorm]" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 258u16 , name : "SMC10GPCIe-10BT (A2) [TigerCard]" } , SubSystem { subvendor : 4280u16 , subdevice : 259u16 , name : "SMC10GPCIe-10BT (A3) [TigerCard]" } , SubSystem { subvendor : 4280u16 , subdevice : 513u16 , name : "SMC10GPCIe-XFP (A1) [TigerCard]" } , SubSystem { subvendor : 6436u16 , subdevice : 257u16 , name : "SFE4001-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 258u16 , name : "SFE4001-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 259u16 , name : "SFE4001-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 513u16 , name : "SFE4002-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 769u16 , name : "SFE4003-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 770u16 , name : "SFE4003-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 771u16 , name : "SFE4003-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 772u16 , name : "SFE4003-A4" } , SubSystem { subvendor : 6436u16 , subdevice : 1280u16 , name : "SFE4005-A0" }] } , Device { vendor_id : 6436u16 , id : 1808u16 , name : "SFC4000 rev B [Solarstorm]" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 259u16 , name : "SMC10GPCIe-10BT (A3) [TigerCard]" } , SubSystem { subvendor : 4280u16 , subdevice : 513u16 , name : "SMC10GPCIe-XFP (A1) [TigerCard]" } , SubSystem { subvendor : 6436u16 , subdevice : 258u16 , name : "SFE4001-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 259u16 , name : "SFE4001-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 513u16 , name : "SFE4002-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 770u16 , name : "SFE4003-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 771u16 , name : "SFE4003-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 772u16 , name : "SFE4003-A4" } , SubSystem { subvendor : 6436u16 , subdevice : 1280u16 , name : "SFE4005-A0" } , SubSystem { subvendor : 6436u16 , subdevice : 20738u16 , name : "SFN4111T-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 20739u16 , name : "SFN4111T-R3" } , SubSystem { subvendor : 6436u16 , subdevice : 20740u16 , name : "SFN4111T-R4" } , SubSystem { subvendor : 6436u16 , subdevice : 20741u16 , name : "SFN4111T-R5" } , SubSystem { subvendor : 6436u16 , subdevice : 20993u16 , name : "SFN4112F-R1" } , SubSystem { subvendor : 6436u16 , subdevice : 20994u16 , name : "SFN4112F-R2" }] } , Device { vendor_id : 6436u16 , id : 2051u16 , name : "SFC9020 10G Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1144u16 , name : "2-port 10GbE Low-Latency (R7)" } , SubSystem { subvendor : 4116u16 , subdevice : 1145u16 , name : "2-port 10GbE OpenOnload (R7)" } , SubSystem { subvendor : 4116u16 , subdevice : 1191u16 , name : "Solarflare 10Gb Low-latency Dual-port HBA (R7)" } , SubSystem { subvendor : 4116u16 , subdevice : 1192u16 , name : "Solarflare 10Gb Dual-port HBA (R7)" } , SubSystem { subvendor : 4156u16 , subdevice : 8498u16 , name : "Ethernet 10Gb 2-port 570FLR-SFP+ Adapter (R1)" } , SubSystem { subvendor : 4156u16 , subdevice : 8502u16 , name : "Ethernet 10Gb 2-port 570SFP+ Adapter (R7)" } , SubSystem { subvendor : 6436u16 , subdevice : 4609u16 , name : "SFA6902F-R1 SFP+ AOE Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25088u16 , name : "SFN5122F-R0 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25089u16 , name : "SFN5122F-R1 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25090u16 , name : "SFN5122F-R2 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25092u16 , name : "SFN5122F-R4 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25093u16 , name : "SFN5122F-R5 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25094u16 , name : "SFN5122F-R6 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25095u16 , name : "SFN5122F-R7 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25104u16 , name : "SFN5322F-R0 SFP+ Precision Time Synchronization Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25105u16 , name : "SFN5322F-R1 SFP+ Precision Time Synchronization Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25111u16 , name : "SFN5322F-R7 SFP+ Precision Time Synchronization Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25127u16 , name : "SFN6122F-R7 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25143u16 , name : "SFN6322F-R7 SFP+ Precision Time Synchronization Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25857u16 , name : "SFN5802K-R1 Mezzanine Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25873u16 , name : "SFN5814H-R1 Mezzanine Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25889u16 , name : "SFN5812H-R1 Mezzanine Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 25954u16 , name : "SFN6832F-R2 SFP+ Mezzanine Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 27141u16 , name : "SFN5112F-R5 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 27142u16 , name : "SFN5112F-R6 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 29190u16 , name : "SFN5162F-R6 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 29191u16 , name : "SFN5162F-R7 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 31238u16 , name : "SFN5152F-R6 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 31239u16 , name : "SFN5152F-R7 SFP+ Server Adapter" }] } , Device { vendor_id : 6436u16 , id : 2067u16 , name : "SFL9021 10GBASE-T Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 24832u16 , name : "SFN5121T-R0 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 24834u16 , name : "SFN5121T-R2 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 24835u16 , name : "SFN5121T-R3 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 24836u16 , name : "SFN5121T-R4 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 26882u16 , name : "SFN5111T-R2 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 26884u16 , name : "SFN5111T-R4 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 28932u16 , name : "SFN5161T-R4 10GBASE-T Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 30980u16 , name : "SFN5151T-R4 10GBASE-T Server Adapter" }] } , Device { vendor_id : 6436u16 , id : 2307u16 , name : "SFC9120 10G Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1228u16 , name : "SFN7122F-R2 2x10GbE SFP+ Flareon Ultra" } , SubSystem { subvendor : 6436u16 , subdevice : 32770u16 , name : "SFN7122F-R1 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32771u16 , name : "SFN7x41Q-R1 Flareon Ultra 7000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32774u16 , name : "SFN7022F-R1 SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32775u16 , name : "SFN7322F-R2 Precision Time SFP+ Server Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32777u16 , name : "SFN7x22F-R2 Flareon Ultra 7000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32778u16 , name : "SFN7x02F-R2 Flareon 7000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32780u16 , name : "SFN7x22F-R3 Flareon Ultra 7000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32781u16 , name : "SFN7x02F-R3 Flareon 7000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32784u16 , name : "SFA7942Q-R1 QSFP+ AOE Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32789u16 , name : "SFA7942Q-A5-0-R1 QSFP+ AOE Adapter" }] } , Device { vendor_id : 6436u16 , id : 2339u16 , name : "SFC9140 10/40G Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 32779u16 , name : "SFN7x42Q-R1 Flareon Ultra 7000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32782u16 , name : "SFN7x42Q-R2 Flareon Ultra 7000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32783u16 , name : "SFN7xx4F-R1 Flareon Ultra 7000 Series 10G Adapter" }] } , Device { vendor_id : 6436u16 , id : 2563u16 , name : "SFC9220 10/40G Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 32785u16 , name : "SFN8022-R1 8000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32786u16 , name : "SFN8522-R1 8000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32787u16 , name : "SFN8042-R1 8000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32788u16 , name : "SFN8542-R1 8000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32790u16 , name : "SFN8022-R2 8000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32791u16 , name : "SFN8522-R2 8000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32792u16 , name : "SFN8042-R2 8000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32793u16 , name : "SFN8542-R2 8000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32794u16 , name : "SFN8722-R1 8000 Series OCP 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32795u16 , name : "SFN8522-R3 8000 Series 10G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32796u16 , name : "SFN8042-R3 8000 Series 10/40G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32801u16 , name : "SFN8041-R1 8000 Series 10/40G Adapter" }] } , Device { vendor_id : 6436u16 , id : 2819u16 , name : "XtremeScale SFC9250 10/25/40/50/100G Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 32797u16 , name : "x2522-R1 2000 Series 10/25G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32798u16 , name : "x2542-R1 2000 Series 40/100G Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32802u16 , name : "XtremeScale X2522 10G Network Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32804u16 , name : "XtremeScale X2562 OCP 3.0 Dual Port SFP28" } , SubSystem { subvendor : 6436u16 , subdevice : 32807u16 , name : "XtremeScale X2541 PCIe Single Port QSFP28" } , SubSystem { subvendor : 6436u16 , subdevice : 32808u16 , name : "XtremeScale X2522-25G Network Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32810u16 , name : "XtremeScale X2542 PCIe Dual Port QSFP28" } , SubSystem { subvendor : 6436u16 , subdevice : 32811u16 , name : "XtremeScale X2552 OCP 2.0 Dual Port SFP28" } , SubSystem { subvendor : 6436u16 , subdevice : 32812u16 , name : "XtremeScale X2522-25G PCIe Dual Port SFP28" } , SubSystem { subvendor : 6436u16 , subdevice : 32813u16 , name : "XtremeScale X2562 OCP 3.0 Dual Port SFP28" }] } , Device { vendor_id : 6436u16 , id : 3075u16 , name : "NS9480 1/10/25/40/50/100Gb Ethernet Controller PLUS" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 32817u16 , name : "X4522 Express and Enterprise Ethernet Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32818u16 , name : "X4542 Express and Enterprise Ethernet Adapter" }] } , Device { vendor_id : 6436u16 , id : 6147u16 , name : "SFC9020 10G Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 6163u16 , name : "SFL9021 10GBASE-T Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 6403u16 , name : "SFC9120 10G Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 6435u16 , name : "SFC9140 10/40G Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 6659u16 , name : "SFC9220 10/40G Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 6915u16 , name : "XtremeScale SFC9250 10/25/40/50/100G Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 7171u16 , name : "NS9480 1/10/25/40/50/100Gb Ethernet Controller PLUS (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 11267u16 , name : "NS9480 1/10/25/40/50/100Gb Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6436u16 , subdevice : 32817u16 , name : "X4522 Enterprise Ethernet Adapter" } , SubSystem { subvendor : 6436u16 , subdevice : 32818u16 , name : "X4542 Enterprise Ethernet Adapter" }] } , Device { vendor_id : 6436u16 , id : 15363u16 , name : "NS9480 1/10/25/40/50/100Gb Ethernet Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6436u16 , id : 26371u16 , name : "SFC4000 rev A iSCSI/Onload [Solarstorm]" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 258u16 , name : "SMC10GPCIe-10BT (A2) [TigerCard]" } , SubSystem { subvendor : 4280u16 , subdevice : 259u16 , name : "SMC10GPCIe-10BT (A3) [TigerCard]" } , SubSystem { subvendor : 4280u16 , subdevice : 513u16 , name : "SMC10GPCIe-XFP (A1) [TigerCard]" } , SubSystem { subvendor : 6436u16 , subdevice : 257u16 , name : "SFE4001-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 258u16 , name : "SFE4001-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 259u16 , name : "SFE4001-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 513u16 , name : "SFE4002-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 769u16 , name : "SFE4003-A1" } , SubSystem { subvendor : 6436u16 , subdevice : 770u16 , name : "SFE4003-A2" } , SubSystem { subvendor : 6436u16 , subdevice : 771u16 , name : "SFE4003-A3" } , SubSystem { subvendor : 6436u16 , subdevice : 772u16 , name : "SFE4003-A4" } , SubSystem { subvendor : 6436u16 , subdevice : 1280u16 , name : "SFE4005-A0" }] } , Device { vendor_id : 6436u16 , id : 49409u16 , name : "EF1-21022T [EtherFabric]" , subsystems : & [] }] }),
        (4867, Vendor { id : 4867u16 , name : "Innovative Integration" , devices : & [Device { vendor_id : 4867u16 , id : 48u16 , name : "X3-SDF 4-channel XMC acquisition board" , subsystems : & [] }] }),
        (2990, Vendor { id : 2990u16 , name : "Bachmann electronic GmbH" , devices : & [] }),
        (4290, Vendor { id : 4290u16 , name : "Auspex Systems Inc." , devices : & [] }),
        (5319, Vendor { id : 5319u16 , name : "Modular Technology Holdings Ltd" , devices : & [] }),
        (4832, Vendor { id : 4832u16 , name : "Chase Research" , devices : & [Device { vendor_id : 4832u16 , id : 16u16 , name : "ST16C654 Quad UART" , subsystems : & [] } , Device { vendor_id : 4832u16 , id : 32u16 , name : "ST16C654 Quad UART" , subsystems : & [] } , Device { vendor_id : 4832u16 , id : 48u16 , name : "ST16C654 Quad UART" , subsystems : & [] }] }),
        (5401, Vendor { id : 5401u16 , name : "TELEFON AKTIEBOLAGET LM Ericsson" , devices : & [] }),
        (5549, Vendor { id : 5549u16 , name : "VMware" , devices : & [Device { vendor_id : 5549u16 , id : 1029u16 , name : "SVGA II Adapter" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1030u16 , name : "SVGA II Adapter (Fusion)" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1808u16 , name : "SVGA Adapter" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1824u16 , name : "VMXNET Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1856u16 , name : "Virtual Machine Communication Interface" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1904u16 , name : "USB2 EHCI Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1908u16 , name : "USB1.1 UHCI Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1912u16 , name : "USB3 xHCI 0.96 Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1913u16 , name : "USB3 xHCI 1.0 Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1936u16 , name : "PCI bridge" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1952u16 , name : "PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1968u16 , name : "VMXNET3 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 1984u16 , name : "PVSCSI SCSI Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 2016u16 , name : "SATA AHCI controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 2032u16 , name : "NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 2049u16 , name : "Virtual Machine Interface" , subsystems : & [SubSystem { subvendor : 5549u16 , subdevice : 2048u16 , name : "Hypervisor ROM Interface" }] } , Device { vendor_id : 5549u16 , id : 2080u16 , name : "Paravirtual RDMA controller" , subsystems : & [] } , Device { vendor_id : 5549u16 , id : 6519u16 , name : "HD Audio Controller" , subsystems : & [] }] }),
        (8193, Vendor { id : 8193u16 , name : "Temporal Research Ltd" , devices : & [] }),
        (4154, Vendor { id : 4154u16 , name : "Seiko Epson Corporation" , devices : & [] }),
        (4879, Vendor { id : 4879u16 , name : "Advanet Inc" , devices : & [] }),
        (41, Vendor { id : 41u16 , name : "MICROCHIP" , devices : & [] }),
        (624, Vendor { id : 624u16 , name : "Hauppauge computer works Inc. (Wrong ID)" , devices : & [] }),
        (8246, Vendor { id : 8246u16 , name : "Netforward Microelectronics Co., Ltd." , devices : & [Device { vendor_id : 8246u16 , id : 5656u16 , name : "NF1618 PCI Express Ethernet Controller" , subsystems : & [SubSystem { subvendor : 8246u16 , subdevice : 2144u16 , name : "NF1618 Family NX860 (2*25GE)" } , SubSystem { subvendor : 8246u16 , subdevice : 2145u16 , name : "NF1618 Family NX861 (4*25GE)" } , SubSystem { subvendor : 8246u16 , subdevice : 2146u16 , name : "NF1618 Family NX862 (2*50GE)" } , SubSystem { subvendor : 8246u16 , subdevice : 2147u16 , name : "NF1618 Family NX863 (2*100GE)" } , SubSystem { subvendor : 8246u16 , subdevice : 2148u16 , name : "NF1618 Family NX864 (1*200GE)" } , SubSystem { subvendor : 8246u16 , subdevice : 2149u16 , name : "NF1618 Family NX865 (2*200GE)" }] } , Device { vendor_id : 8246u16 , id : 5657u16 , name : "NF1618 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 8246u16 , id : 32768u16 , name : "NP36xxx PCIe Gen 6 Switch" , subsystems : & [SubSystem { subvendor : 8246u16 , subdevice : 260u16 , name : "NP36104 PCIe Gen 6 26 port/104 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8246u16 , subdevice : 324u16 , name : "NP36144 PCIe Gen 6 36 port/144 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8246u16 , subdevice : 4096u16 , name : "NP36000 PCIe Gen 6 Virtual Upstream/Downstream Port" } , SubSystem { subvendor : 8246u16 , subdevice : 4097u16 , name : "NP36000 Virtual PCIe MGR 1.x Endpoint" } , SubSystem { subvendor : 8246u16 , subdevice : 4098u16 , name : "NP36000 Virtual PCIe DMA 1.x Endpoint" } , SubSystem { subvendor : 8246u16 , subdevice : 4099u16 , name : "NP36000 Virtual PCIe NTB 1.x Endpoint" } , SubSystem { subvendor : 8246u16 , subdevice : 4100u16 , name : "NP36000 Virtual PCIe Placeholder 1.x Endpoint" } , SubSystem { subvendor : 8246u16 , subdevice : 4101u16 , name : "NP36000 Virtual PCIe C2PMem 1.x Endpoint" } , SubSystem { subvendor : 8246u16 , subdevice : 4102u16 , name : "NP36000 Virtual PCIe Pktgen 1.x Endpoint" }] }] }),
        (6654, Vendor { id : 6654u16 , name : "ESI Audiotechnik GmbH" , devices : & [Device { vendor_id : 6654u16 , id : 28672u16 , name : "MAYA44 family PCI Audio Controller" , subsystems : & [] }] }),
        (8236, Vendor { id : 8236u16 , name : "CAEN S.p.A." , devices : & [Device { vendor_id : 8236u16 , id : 22552u16 , name : "A5818" , subsystems : & [] }] }),
        (22086, Vendor { id : 22086u16 , name : "Vector Fabrics BV" , devices : & [] }),
        (4775, Vendor { id : 4775u16 , name : "AMO GmbH" , devices : & [] }),
        (6332, Vendor { id : 6332u16 , name : "GeCube Technologies, Inc." , devices : & [] }),
        (7567, Vendor { id : 7567u16 , name : "Exegy" , devices : & [] }),
        (7738, Vendor { id : 7738u16 , name : "Cactus Technologies Limited" , devices : & [Device { vendor_id : 7738u16 , id : 8803u16 , name : "270PM6, 270PM7 series NVMe SSD" , subsystems : & [] }] }),
        (53769, Vendor { id : 53769u16 , name : "Ultimarc" , devices : & [Device { vendor_id : 53769u16 , id : 5376u16 , name : "PAC Drive" , subsystems : & [] } , Device { vendor_id : 53769u16 , id : 5538u16 , name : "SpinTrak" , subsystems : & [] } , Device { vendor_id : 53769u16 , id : 5633u16 , name : "AimTrak" , subsystems : & [] }] }),
        (5519, Vendor { id : 5519u16 , name : "Ditect Coop" , devices : & [] }),
        (7807, Vendor { id : 7807u16 , name : "Jiangsu Huacun Elec. Tech. Co., Ltd." , devices : & [Device { vendor_id : 7807u16 , id : 24578u16 , name : "MMY MMSP350 PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7807u16 , id : 24579u16 , name : "MMY HC512GP3KH2T PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (51949, Vendor { id : 51949u16 , name : "Canny Edge" , devices : & [] }),
        (43586, Vendor { id : 43586u16 , name : "Scitex Digital Video" , devices : & [] }),
        (5861, Vendor { id : 5861u16 , name : "Intellon Corp." , devices : & [Device { vendor_id : 5861u16 , id : 24576u16 , name : "INT6000 Ethernet-to-Powerline Bridge [HomePlug AV]" , subsystems : & [] } , Device { vendor_id : 5861u16 , id : 25344u16 , name : "INT6300 Ethernet-to-Powerline Bridge [HomePlug AV]" , subsystems : & [] }] }),
        (5336, Vendor { id : 5336u16 , name : "HOPF Elektronik GmBH" , devices : & [] }),
        (5947, Vendor { id : 5947u16 , name : "Altima (nee Broadcom)" , devices : & [Device { vendor_id : 5947u16 , id : 1u16 , name : "AC1002 PCI Gigabit Ethernet controller" , subsystems : & [] } , Device { vendor_id : 5947u16 , id : 1000u16 , name : "AC1000 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5947u16 , id : 1001u16 , name : "AC1001 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5947u16 , id : 1002u16 , name : "AC9100 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 5947u16 , subdevice : 1u16 , name : "AC1002" }] } , Device { vendor_id : 5947u16 , id : 1003u16 , name : "AC1003 Gigabit Ethernet" , subsystems : & [] }] }),
        (4990, Vendor { id : 4990u16 , name : "Patriot Scientific Corporation" , devices : & [] }),
        (6794, Vendor { id : 6794u16 , name : "StarBridge, Inc." , devices : & [] }),
        (6391, Vendor { id : 6391u16 , name : "Commtech, Inc." , devices : & [Device { vendor_id : 6391u16 , id : 1u16 , name : "ESCC-PCI-335 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 2u16 , name : "422/4-PCI-335 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 3u16 , name : "232/4-1M-PCI Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 4u16 , name : "422/2-PCI-335 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 5u16 , name : "IGESCC-PCI-ISO/1 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 10u16 , name : "232/4-PCI-335 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 11u16 , name : "232/8-PCI-335 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 15u16 , name : "FSCC Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 16u16 , name : "GSCC Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 17u16 , name : "QSSB Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 20u16 , name : "SuperFSCC Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 21u16 , name : "SuperFSCC-104-LVDS Serial PC/104+ Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 22u16 , name : "FSCC-232 RS-232 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 23u16 , name : "SuperFSCC-104 Serial PC/104+ Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 24u16 , name : "SuperFSCC/4 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 25u16 , name : "SuperFSCC Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 26u16 , name : "SuperFSCC-LVDS Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 27u16 , name : "FSCC/4 Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 28u16 , name : "SuperFSCC/4-LVDS Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 29u16 , name : "FSCC Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 30u16 , name : "SuperFSCC/4 Serial PCIe Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 31u16 , name : "SuperFSCC/4 Serial cPCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 32u16 , name : "422/4-PCIe Serial PCIe Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 33u16 , name : "422/8-PCIe Serial PCIe Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 34u16 , name : "SuperFSCC/4-LVDS Serial PCIe Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 35u16 , name : "SuperFSCC/4 Serial cPCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 37u16 , name : "SuperFSCC/4-LVDS Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 38u16 , name : "SuperFSCC-LVDS Serial PCI Adapter [Fastcom]" , subsystems : & [] } , Device { vendor_id : 6391u16 , id : 39u16 , name : "FSCC/4 Serial PCIe Adapter [Fastcom]" , subsystems : & [] }] }),
        (5634, Vendor { id : 5634u16 , name : "Prosys-tec Inc" , devices : & [] }),
        (6294, Vendor { id : 6294u16 , name : "B&B Electronics Manufacturing Company, Inc." , devices : & [Device { vendor_id : 6294u16 , id : 16898u16 , name : "MIport 3PCIU2 2-port Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 16900u16 , name : "MIport 3PCIU4 4-port Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 16904u16 , name : "MIport 3PCIU8 8-port Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 16913u16 , name : "MIport 3PCIOU1 1-port Isolated Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 16914u16 , name : "MIport 3PCIOU2 2-port Isolated Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 16916u16 , name : "MIport 3PCIOU4 4-port Isolated Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 47888u16 , name : "3PCI2 2-Port Serial" , subsystems : & [] } , Device { vendor_id : 6294u16 , id : 47889u16 , name : "3PCIO1 1-Port Isolated Serial" , subsystems : & [] }] }),
        (5136, Vendor { id : 5136u16 , name : "Midas lab Inc" , devices : & [] }),
        (4822, Vendor { id : 4822u16 , name : "Analogic Corp" , devices : & [] }),
        (4416, Vendor { id : 4416u16 , name : "Intervoice Inc" , devices : & [] }),
        (4535, Vendor { id : 4535u16 , name : "Motorola" , devices : & [] }),
        (4927, Vendor { id : 4927u16 , name : "SCM Microsystems" , devices : & [] }),
        (4267, Vendor { id : 4267u16 , name : "Digicom" , devices : & [] }),
        (4693, Vendor { id : 4693u16 , name : "Optibase Ltd" , devices : & [Device { vendor_id : 4693u16 , id : 4368u16 , name : "MPEG Forge" , subsystems : & [] } , Device { vendor_id : 4693u16 , id : 4624u16 , name : "MPEG Fusion" , subsystems : & [] } , Device { vendor_id : 4693u16 , id : 8464u16 , name : "VideoPlex" , subsystems : & [] } , Device { vendor_id : 4693u16 , id : 8480u16 , name : "VideoPlex CC" , subsystems : & [] } , Device { vendor_id : 4693u16 , id : 8496u16 , name : "VideoQuest" , subsystems : & [] }] }),
        (5375, Vendor { id : 5375u16 , name : "TWINHEAD INTERNATIONAL Corp" , devices : & [] }),
        (4790, Vendor { id : 4790u16 , name : "Natural Microsystems" , devices : & [] }),
        (5414, Vendor { id : 5414u16 , name : "ISS, Inc" , devices : & [] }),
        (19528, Vendor { id : 19528u16 , name : "LUNG HWA Electronics" , devices : & [] }),
        (4831, Vendor { id : 4831u16 , name : "SBS Technologies Inc" , devices : & [] }),
        (4827, Vendor { id : 4827u16 , name : "Annapolis Micro Systems, Inc" , devices : & [] }),
        (64087, Vendor { id : 64087u16 , name : "Interagon AS" , devices : & [Device { vendor_id : 64087u16 , id : 1u16 , name : "PMC [Pattern Matching Chip]" , subsystems : & [] }] }),
        (4930, Vendor { id : 4930u16 , name : "Promax Systems Inc" , devices : & [] }),
        (6153, Vendor { id : 6153u16 , name : "Lumanate, Inc." , devices : & [] }),
        (1911, Vendor { id : 1911u16 , name : "Ubiquiti Networks, Inc." , devices : & [] }),
        (10581, Vendor { id : 10581u16 , name : "Connectix Virtual PC" , devices : & [Device { vendor_id : 10581u16 , id : 28257u16 , name : "OHCI USB 1.1 controller" , subsystems : & [] }] }),
        (6144, Vendor { id : 6144u16 , name : "Qualcore Logic Inc." , devices : & [Device { vendor_id : 6144u16 , id : 4352u16 , name : "Nanospeed Trading Gateway" , subsystems : & [] }] }),
        (4432, Vendor { id : 4432u16 , name : "Thinking Machines Corp" , devices : & [] }),
        (4902, Vendor { id : 4902u16 , name : "Seachange international" , devices : & [] }),
        (4898, Vendor { id : 4898u16 , name : "MTT Co., Ltd" , devices : & [] }),
        (7741, Vendor { id : 7741u16 , name : "Burlywood, Inc" , devices : & [] }),
        (7804, Vendor { id : 7804u16 , name : "Brainchip Inc" , devices : & [Device { vendor_id : 7804u16 , id : 48289u16 , name : "AKD1000 Neural Network Coprocessor [Akida]" , subsystems : & [] }] }),
        (7421, Vendor { id : 7421u16 , name : "Mangstor" , devices : & [Device { vendor_id : 7421u16 , id : 25344u16 , name : "MX6300 series PCIe x8 NVMe SSD" , subsystems : & [] }] }),
        (12610, Vendor { id : 12610u16 , name : "Post Impression Systems." , devices : & [] }),
        (6466, Vendor { id : 6466u16 , name : "ClearSpeed Technology plc" , devices : & [Device { vendor_id : 6466u16 , id : 58641u16 , name : "Advance X620 accelerator card" , subsystems : & [] } , Device { vendor_id : 6466u16 , id : 58657u16 , name : "Advance e620 accelerator card" , subsystems : & [] }] }),
        (57797, Vendor { id : 57797u16 , name : "Elcus" , devices : & [] }),
        (7120, Vendor { id : 7120u16 , name : "Astronics Corporation" , devices : & [Device { vendor_id : 7120u16 , id : 4097u16 , name : "Mx5 PMC/XMC Databus Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4098u16 , name : "PM1553-5 (PC/104+ MIL-STD-1553 Interface Card)" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4100u16 , name : "AB3000 Series Rugged Computer" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4101u16 , name : "PE1000 (Multi-Protocol PCIe/104 Interface Card)" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4102u16 , name : "webCS Wireless Aircraft Communications Server" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4103u16 , name : "AB3000 Series Rugged Computer (Series N)" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4104u16 , name : "ME1000 mPCIe Avionics Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4106u16 , name : "NG1 Series Avionics Converter" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4107u16 , name : "LE5 429 Databus Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4108u16 , name : "LE5 1553 Databus Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4353u16 , name : "OmniBus II PCIe Multi-Protocol Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4354u16 , name : "OmniBusBox II Multi-Protocol Interface Core" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4355u16 , name : "OmniBus II cPCIe/PXIe Multi-Protocol Interface Card" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4608u16 , name : "NG3 Series Mil-Std-1553 Interface" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4609u16 , name : "NG3 Series ARINC 429 Interface" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4610u16 , name : "NG3 Series Avionics Discrete & Serial Interface" , subsystems : & [] } , Device { vendor_id : 7120u16 , id : 4611u16 , name : "NG3 Series Avionics Discrete Interface" , subsystems : & [] }] }),
        (8353, Vendor { id : 8353u16 , name : "Etched, Inc." , devices : & [Device { vendor_id : 8353u16 , id : 1u16 , name : "Sohu" , subsystems : & [] }] }),
        (4844, Vendor { id : 4844u16 , name : "3A International, Inc." , devices : & [] }),
        (4763, Vendor { id : 4763u16 , name : "Image Access" , devices : & [] }),
        (6092, Vendor { id : 6092u16 , name : "NetChip Technology, Inc" , devices : & [Device { vendor_id : 6092u16 , id : 8832u16 , name : "NET2280 PCI to USB 2.0 Hi-Speed Peripheral Controller" , subsystems : & [] } , Device { vendor_id : 6092u16 , id : 8834u16 , name : "NET2282 PCI to USB 2.0 Hi-Speed Peripheral Controller" , subsystems : & [] }] }),
        (5154, Vendor { id : 5154u16 , name : "Ygrec Systems Co Ltd" , devices : & [] }),
        (51918, Vendor { id : 51918u16 , name : "CACE Technologies, Inc." , devices : & [Device { vendor_id : 51918u16 , id : 1u16 , name : "TurboCap Port A" , subsystems : & [] } , Device { vendor_id : 51918u16 , id : 2u16 , name : "TurboCap Port B" , subsystems : & [] } , Device { vendor_id : 51918u16 , id : 35u16 , name : "AirPcap N" , subsystems : & [] }] }),
        (5326, Vendor { id : 5326u16 , name : "Whistle Communications" , devices : & [] }),
        (5379, Vendor { id : 5379u16 , name : "KAWASAKI LSI USA Inc" , devices : & [] }),
        (4174, Vendor { id : 4174u16 , name : "Oak Technology, Inc" , devices : & [Device { vendor_id : 4174u16 , id : 23u16 , name : "OTI-64017 [Eon]" , subsystems : & [] } , Device { vendor_id : 4174u16 , id : 263u16 , name : "OTI-107 [Spitfire]" , subsystems : & [] } , Device { vendor_id : 4174u16 , id : 265u16 , name : "Video Adapter" , subsystems : & [] } , Device { vendor_id : 4174u16 , id : 273u16 , name : "OTI-64111 [Spitfire]" , subsystems : & [] } , Device { vendor_id : 4174u16 , id : 535u16 , name : "OTI-64217 [Eon]" , subsystems : & [] } , Device { vendor_id : 4174u16 , id : 791u16 , name : "OTI-64317 [Warp 5]" , subsystems : & [] }] }),
        (6744, Vendor { id : 6744u16 , name : "Razer USA Ltd." , devices : & [] }),
        (5827, Vendor { id : 5827u16 , name : "Synopsys, Inc." , devices : & [Device { vendor_id : 5827u16 , id : 43981u16 , name : "DWC_usb3 / PCIe bridge" , subsystems : & [] } , Device { vendor_id : 5827u16 , id : 43982u16 , name : "DWC_usb3" , subsystems : & [] } , Device { vendor_id : 5827u16 , id : 43983u16 , name : "DWC_usb31" , subsystems : & [] } , Device { vendor_id : 5827u16 , id : 60890u16 , name : "EPMockUp" , subsystems : & [] }] }),
        (5467, Vendor { id : 5467u16 , name : "PROTAC INTERNATIONAL Corp" , devices : & [] }),
        (7661, Vendor { id : 7661u16 , name : "Alibaba (China) Co., Ltd." , devices : & [Device { vendor_id : 7661u16 , id : 4223u16 , name : "Elastic RDMA Adapter" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 20487u16 , name : "Elastic RDMA Adapter" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 32768u16 , name : "M1 Root Port" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 32769u16 , name : "ACC-RCiEP" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 32770u16 , name : "RCiEP VF" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 32771u16 , name : "RCEC PF" , subsystems : & [] } , Device { vendor_id : 7661u16 , id : 32772u16 , name : "RCEC VF" , subsystems : & [] }] }),
        (4330, Vendor { id : 4330u16 , name : "Integraphics" , devices : & [Device { vendor_id : 4330u16 , id : 5760u16 , name : "IGA-1680" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 5762u16 , name : "IGA-1682" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 5763u16 , name : "IGA-1683" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 8192u16 , name : "CyberPro 2000" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 8208u16 , name : "CyberPro 2000A" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 20480u16 , name : "CyberPro 5000" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 20560u16 , name : "CyberPro 5050" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 20994u16 , name : "CyberPro 5202" , subsystems : & [] } , Device { vendor_id : 4330u16 , id : 21074u16 , name : "CyberPro5252" , subsystems : & [] }] }),
        (7737, Vendor { id : 7737u16 , name : "MEDION AG" , devices : & [] }),
        (5229, Vendor { id : 5229u16 , name : "Tachyon, INC." , devices : & [] }),
        (5590, Vendor { id : 5590u16 , name : "Entridia Corp" , devices : & [] }),
        (20790, Vendor { id : 20790u16 , name : "S S Technologies" , devices : & [] }),
        (4648, Vendor { id : 4648u16 , name : "Norsk Elektro Optikk A/S" , devices : & [] }),
        (6530, Vendor { id : 6530u16 , name : "Distant Early Warning Communications Inc" , devices : & [Device { vendor_id : 6530u16 , id : 5632u16 , name : "OX16C954 HOST-A" , subsystems : & [] } , Device { vendor_id : 6530u16 , id : 5887u16 , name : "OX16C954 HOST-B" , subsystems : & [] }] }),
        (51792, Vendor { id : 51792u16 , name : "Varian Australia Pty Ltd" , devices : & [] }),
        (7306, Vendor { id : 7306u16 , name : "TSF5 Corporation" , devices : & [Device { vendor_id : 7306u16 , id : 1u16 , name : "Hunter PCI Express" , subsystems : & [] }] }),
        (8436, Vendor { id : 8436u16 , name : "TRENDnet" , devices : & [] }),
        (4886, Vendor { id : 4886u16 , name : "Teradyne Inc" , devices : & [] }),
        (5123, Vendor { id : 5123u16 , name : "Ascor Inc" , devices : & [] }),
        (7607, Vendor { id : 7607u16 , name : "Phytium Technology Co., Ltd." , devices : & [Device { vendor_id : 7607u16 , id : 56352u16 , name : "[X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56353u16 , name : "VPU Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56354u16 , name : "DC Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56355u16 , name : "I2S/DMA Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56356u16 , name : "NPU Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56358u16 , name : "SATA Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56359u16 , name : "USB Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56361u16 , name : "NANDFLASH Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56363u16 , name : "I2S Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56364u16 , name : "SPIM Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56365u16 , name : "CAN Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56366u16 , name : "UART Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56367u16 , name : "PWM Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56368u16 , name : "MIO Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56369u16 , name : "GPIO Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56370u16 , name : "SMBUS Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56372u16 , name : "PS2 Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56373u16 , name : "LPC Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56374u16 , name : "LDMA Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56376u16 , name : "LSD_CFG Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56378u16 , name : "SWITCH Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56379u16 , name : "MAC Controller [D3000M Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56380u16 , name : "GPU_DMA Controller [X100 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56382u16 , name : "DCController [E2000 Series]" , subsystems : & [] } , Device { vendor_id : 7607u16 , id : 56383u16 , name : "SATA Controller [D3000M Series]" , subsystems : & [] }] }),
        (5190, Vendor { id : 5190u16 , name : "Graphin Co Ltd" , devices : & [] }),
        (38424, Vendor { id : 38424u16 , name : "JusonTech Corporation" , devices : & [Device { vendor_id : 38424u16 , id : 1u16 , name : "JusonTech Gigabit Ethernet Controller" , subsystems : & [] }] }),
        (5019, Vendor { id : 5019u16 , name : "Mediasonic Multimedia Systems Ltd" , devices : & [] }),
        (4149, Vendor { id : 4149u16 , name : "Comp. & Comm. Research Lab" , devices : & [] }),
        (34816, Vendor { id : 34816u16 , name : "Trigem Computer Inc." , devices : & [Device { vendor_id : 34816u16 , id : 8200u16 , name : "Video assistant component" , subsystems : & [] }] }),
        (4352, Vendor { id : 4352u16 , name : "Jazz Multimedia" , devices : & [] }),
        (5001, Vendor { id : 5001u16 , name : "Applicom International" , devices : & [Device { vendor_id : 5001u16 , id : 1u16 , name : "PCI1500PFB [Intelligent fieldbus adaptor]" , subsystems : & [] }] }),
        (4394, Vendor { id : 4394u16 , name : "Hermes Electronics Company, Ltd." , devices : & [] }),
        (4868, Vendor { id : 4868u16 , name : "Juniper Networks" , devices : & [Device { vendor_id : 4868u16 , id : 123u16 , name : "TMC Chassis Interface" , subsystems : & [] } , Device { vendor_id : 4868u16 , id : 124u16 , name : "TMC PFE Interface" , subsystems : & [] } , Device { vendor_id : 4868u16 , id : 197u16 , name : "Supercon FPGA" , subsystems : & [] }] }),
        (4457, Vendor { id : 4457u16 , name : "Centre for Development of Advanced Computing" , devices : & [] }),
        (4601, Vendor { id : 4601u16 , name : "I-Cube Inc" , devices : & [] }),
        (5429, Vendor { id : 5429u16 , name : "EVERGREEN Technologies Inc" , devices : & [] }),
        (4606, Vendor { id : 4606u16 , name : "Pepperl+Fuchs" , devices : & [Device { vendor_id : 4606u16 , id : 1u16 , name : "RocketPort PCI 32-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2u16 , name : "RocketPort PCI 8-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 3u16 , name : "RocketPort PCI 16-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 4u16 , name : "RocketPort PCI 4-port w/Quad Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 5u16 , name : "RocketPort PCI 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 6u16 , name : "RocketPort PCI 8-port w/RJ11 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 7u16 , name : "RocketPort PCI 4-port w/RJ45 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 8u16 , name : "RocketPort PCI 8-port w/DB78 SNI connector (Siemens)" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 9u16 , name : "RocketPort PCI 16-port w/DB78 SNI connector (Siemens)" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 10u16 , name : "RocketPort PCI Plus 4-port w/Quad Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 11u16 , name : "RocketPort PCI Plus 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 12u16 , name : "RocketModem II 6-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 13u16 , name : "RocketModem 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 14u16 , name : "RocketPort PCI Plus 2-port RS-232 w/DB9 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 15u16 , name : "RocketPort PCI Plus 2-port SMPTE w/DB9 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 64u16 , name : "RocketPort INFINITY 8-port w/Octa Cable RJ45" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 65u16 , name : "RocketPort INFINITY 32-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 66u16 , name : "RocketPort INFINITY 8-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 67u16 , name : "RocketPort INFINITY 16-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 68u16 , name : "RocketPort INFINITY 4-port w/Quad Cable DB" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 69u16 , name : "RocketPort INFINITY 8-port w/Octa Cable DB" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 70u16 , name : "RocketPort INFINITY 4-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 71u16 , name : "RocketPort INFINITY 4J (4-port) w/RJ45 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 72u16 , name : "RocketPort INFINITY 4J (4-port) w/RJ45 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 74u16 , name : "RocketPort INFINITY Plus 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 75u16 , name : "RocketPort INFINITY Plus 8-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 76u16 , name : "RocketModem INFINITY III 8-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 77u16 , name : "RocketModem INFINITY III 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 78u16 , name : "RocketPort INFINITY Plus 2-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 79u16 , name : "RocketPort INFINITY 2-port SMPTE w/DB9 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 80u16 , name : "RocketPort INFINITY Plus 4-port RJ45" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 81u16 , name : "RocketPort INFINITY Plus 8-port RJ11" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 82u16 , name : "RocketPort INFINITY 8-port SMPTE w/DB9 Connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 96u16 , name : "RocketPort EXPRESS 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 97u16 , name : "RocketPort EXPRESS 32-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 98u16 , name : "RocketPort EXPRESS 8-Port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 99u16 , name : "RocketPort EXPRESS 16-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 100u16 , name : "RocketPort EXPRESS 4-port w/Quad Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 101u16 , name : "RocketPort EXPRESS 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 102u16 , name : "RocketPort EXPRESS 4-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 103u16 , name : "RocketPort EXPRESS 4J (4-port) w/RJ45 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 104u16 , name : "RocketPort EXPRESS 8J (8-port) w/RJ11 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 111u16 , name : "RocketPort EXPRESS SMPTE 2-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 114u16 , name : "RocketPort EXPRESS SMPTE 8-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2049u16 , name : "RocketPort uPCI 32-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2050u16 , name : "RocketPort uPCI 8-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2051u16 , name : "RocketPort uPCI 16-port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2053u16 , name : "RocketPort uPCI 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2059u16 , name : "RocketPort Plus uPCI 8-port w/Octa Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2060u16 , name : "RocketModem III 8-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2061u16 , name : "RocketModem III 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2062u16 , name : "RocketPort uPCI 2-port RS232 w/DB9 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2063u16 , name : "RocketPort uPCI SMPTE 2-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2064u16 , name : "RocketPort Plus uPCI 4J (4-port) w/RJ45 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2065u16 , name : "RocketPort Plus uPCI 8J (8-port) w/RJ11 connectors" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2066u16 , name : "RocketPort Plus uPCI 422 8-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2067u16 , name : "RocketModem IV uPCI 8-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2068u16 , name : "RocketModem IV uPCI 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 2307u16 , name : "RocketPort Compact PCI 16 port w/external I/F" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 32789u16 , name : "RocketPort 550 4-port" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 34821u16 , name : "RocketPort uPCI 4-port w/Quad Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 34827u16 , name : "RocketPort Plus uPCI 4-port w/Quad Cable" , subsystems : & [] } , Device { vendor_id : 4606u16 , id : 34834u16 , name : "RocketPort Plus uPCI 4-port RS422 w/Quad Cable" , subsystems : & [] }] }),
        (5324, Vendor { id : 5324u16 , name : "NAKAYO Telecommunications Inc" , devices : & [] }),
        (4688, Vendor { id : 4688u16 , name : "Hitachi Microcomputer System Ltd" , devices : & [] }),
        (52460, Vendor { id : 52460u16 , name : "Curtiss-Wright Controls Embedded Computing" , devices : & [] }),
        (8018, Vendor { id : 8018u16 , name : "MangoBoost Inc." , devices : & [Device { vendor_id : 8018u16 , id : 4104u16 , name : "Mango BoostX - RoCE AI" , subsystems : & [] } , Device { vendor_id : 8018u16 , id : 4128u16 , name : "Mango NetworkBoost - TCP" , subsystems : & [] } , Device { vendor_id : 8018u16 , id : 4130u16 , name : "Mango StorageBoost - NTI" , subsystems : & [] } , Device { vendor_id : 8018u16 , id : 4131u16 , name : "Mango StorageBoost - NTT" , subsystems : & [] }] }),
        (5757, Vendor { id : 5757u16 , name : "Samsung Electro-Mechanics Co., Ltd." , devices : & [Device { vendor_id : 5757u16 , id : 40960u16 , name : "MagicLAN SWL-2210P 802.11b [Intersil ISL3874]" , subsystems : & [] }] }),
        (7156, Vendor { id : 7156u16 , name : "VTI Instruments Corporation" , devices : & [Device { vendor_id : 7156u16 , id : 1u16 , name : "SentinelEX" , subsystems : & [] } , Device { vendor_id : 7156u16 , id : 28689u16 , name : "RX0xxx" , subsystems : & [] }] }),
        (6572, Vendor { id : 6572u16 , name : "Kasten Chase Applied Research" , devices : & [Device { vendor_id : 6572u16 , id : 1u16 , name : "ACA2400 Crypto Accelerator" , subsystems : & [] }] }),
        (4326, Vendor { id : 4326u16 , name : "Gainbery Computer Products Inc." , devices : & [] }),
        (5430, Vendor { id : 5430u16 , name : "ACTIS Computer" , devices : & [] }),
        (4982, Vendor { id : 4982u16 , name : "LMC" , devices : & [] }),
        (29697, Vendor { id : 29697u16 , name : "EndRun Technologies" , devices : & [Device { vendor_id : 29697u16 , id : 57600u16 , name : "PTP3100 PCIe PTP Slave Clock" , subsystems : & [] }] }),
        (4851, Vendor { id : 4851u16 , name : "Radstone Technology" , devices : & [] }),
        (5526, Vendor { id : 5526u16 , name : "Folsom Research Inc" , devices : & [] }),
        (4591, Vendor { id : 4591u16 , name : "Nicolet Technologies B.V." , devices : & [] }),
        (4264, Vendor { id : 4264u16 , name : "Sierra Semiconductor" , devices : & [Device { vendor_id : 4264u16 , id : 0u16 , name : "STB Horizon 64" , subsystems : & [] }] }),
        (65533, Vendor { id : 65533u16 , name : "XenSource, Inc." , devices : & [Device { vendor_id : 65533u16 , id : 257u16 , name : "PCI Event Channel Controller" , subsystems : & [] }] }),
        (7344, Vendor { id : 7344u16 , name : "Shannon Systems" , devices : & [Device { vendor_id : 7344u16 , id : 33382u16 , name : "SP4 Series SSD" , subsystems : & [SubSystem { subvendor : 7344u16 , subdevice : 8225u16 , name : "SP4 Series OCS U.2 SSD" } , SubSystem { subvendor : 7344u16 , subdevice : 8481u16 , name : "SP4 Series ZNS U.2 SSD" } , SubSystem { subvendor : 7344u16 , subdevice : 12065u16 , name : "SP4E Series NVMe U.2 SSD(1920/3840/7680GB)" } , SubSystem { subvendor : 7344u16 , subdevice : 12066u16 , name : "SP4X Series NVMe U.2 SSD(1600/3200/6400GB)" }] } , Device { vendor_id : 7344u16 , id : 53248u16 , name : "Venice NVMe SSD" , subsystems : & [SubSystem { subvendor : 7344u16 , subdevice : 8208u16 , name : "Venice-E Series OCS U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 8209u16 , name : "Venice Series OCS U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 8210u16 , name : "Venice-X Series OCS U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 12048u16 , name : "Venice-E Series NVMe U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 12049u16 , name : "Venice Series NVMe U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 12050u16 , name : "Venice-X Series NVMe U.2" } , SubSystem { subvendor : 7344u16 , subdevice : 40976u16 , name : "Venice-E Series OCS AIC" } , SubSystem { subvendor : 7344u16 , subdevice : 40978u16 , name : "Venice-X Series OCS AIC" } , SubSystem { subvendor : 7344u16 , subdevice : 44816u16 , name : "Venice-E Series NVMe AIC" } , SubSystem { subvendor : 7344u16 , subdevice : 44818u16 , name : "Venice-X Series NVMe AIC" }] }] }),
        (5498, Vendor { id : 5498u16 , name : "Japan Elecronics Ind Inc" , devices : & [] }),
        (5021, Vendor { id : 5021u16 , name : "EPL limited" , devices : & [] }),
        (6730, Vendor { id : 6730u16 , name : "SLAC National Accelerator Lab TID-ID" , devices : & [Device { vendor_id : 6730u16 , id : 4096u16 , name : "MCOR Power Supply Controller" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 4112u16 , name : "AMC EVR - Stockholm Timing Board" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 4128u16 , name : "PGPCard - Gen3 Cameralink Interface" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 4144u16 , name : "PGPCard - Gen3 GIGe Interface" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8192u16 , name : "PGPCard - 4 Lane" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8193u16 , name : "PGPCard - 8 Lane Plus EVR" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8208u16 , name : "PCI-Express EVR" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8209u16 , name : "PCI-Express EVR - TPR Version" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8224u16 , name : "PGP-GEN3 PCIe - 8 Lane Plus EVR" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 8240u16 , name : "AXI Stream DAQ PCIe card" , subsystems : & [SubSystem { subvendor : 4130u16 , subdevice : 5u16 , name : "Xilinx AC701" } , SubSystem { subvendor : 4130u16 , subdevice : 6u16 , name : "Xilinx Alveo U50" } , SubSystem { subvendor : 4130u16 , subdevice : 7u16 , name : "Xilinx Alveo U200" } , SubSystem { subvendor : 4130u16 , subdevice : 8u16 , name : "Xilinx Alveo U250" } , SubSystem { subvendor : 4130u16 , subdevice : 9u16 , name : "Xilinx Alveo U280" } , SubSystem { subvendor : 4130u16 , subdevice : 10u16 , name : "Xilinx KC705" } , SubSystem { subvendor : 4130u16 , subdevice : 11u16 , name : "Xilinx KCU105" } , SubSystem { subvendor : 4130u16 , subdevice : 12u16 , name : "Xilinx KCU116" } , SubSystem { subvendor : 4130u16 , subdevice : 13u16 , name : "Xilinx KCU1500" } , SubSystem { subvendor : 4130u16 , subdevice : 14u16 , name : "Xilinx VCU128" } , SubSystem { subvendor : 4130u16 , subdevice : 15u16 , name : "Xilinx Alveo U55C" } , SubSystem { subvendor : 4130u16 , subdevice : 16u16 , name : "Xilinx Varium C1100" } , SubSystem { subvendor : 4130u16 , subdevice : 4109u16 , name : "Xilinx KCU1500 Extended" } , SubSystem { subvendor : 4130u16 , subdevice : 4111u16 , name : "Xilinx Alveo U55C Extended" } , SubSystem { subvendor : 4130u16 , subdevice : 4112u16 , name : "Xilinx Varium C1100 Extended" } , SubSystem { subvendor : 4794u16 , subdevice : 2u16 , name : "XUPVV8 VU13P" } , SubSystem { subvendor : 4794u16 , subdevice : 19u16 , name : "XUPVV8 VU9P" } , SubSystem { subvendor : 6730u16 , subdevice : 3u16 , name : "PGP Card GEN3" } , SubSystem { subvendor : 6730u16 , subdevice : 4u16 , name : "PGP Card GEN4" } , SubSystem { subvendor : 7570u16 , subdevice : 17u16 , name : "PC821 KU085" } , SubSystem { subvendor : 7570u16 , subdevice : 18u16 , name : "PC821 KU115" } , SubSystem { subvendor : 16708u16 , subdevice : 1u16 , name : "ADM-PCIE-KU3" }] } , Device { vendor_id : 6730u16 , id : 8256u16 , name : "EXO PCIe TEM" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 12288u16 , name : "COB DTM V1" , subsystems : & [] } , Device { vendor_id : 6730u16 , id : 12289u16 , name : "COB DTM V2" , subsystems : & [] }] }),
        (8066, Vendor { id : 8066u16 , name : "d-Matrix" , devices : & [Device { vendor_id : 8066u16 , id : 17u16 , name : "Corsair [DMX 1000 Series]" , subsystems : & [] } , Device { vendor_id : 8066u16 , id : 241u16 , name : "JetStream [DMX F1 Transparent NIC]" , subsystems : & [] }] }),
        (8379, Vendor { id : 8379u16 , name : "Quantum Machines" , devices : & [Device { vendor_id : 8379u16 , id : 273u16 , name : "OPNIC" , subsystems : & [] }] }),
        (40242, Vendor { id : 40242u16 , name : "Beijing Starblaze Technology Co. Ltd." , devices : & [Device { vendor_id : 40242u16 , id : 0u16 , name : "STAR1000 PCIe NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4096u16 , name : "STAR1000 PCIe NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4097u16 , name : "STAR1000P PCIe NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4609u16 , name : "STAR1200C NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4610u16 , name : "STAR1200I NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4611u16 , name : "STAR1200L NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 4612u16 , name : "STAR1200E NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5376u16 , name : "STAR1500 PCIe NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5377u16 , name : "STAR1500C NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5378u16 , name : "STAR1500E NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5380u16 , name : "STAR1500L NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5381u16 , name : "STAR1500K NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 5398u16 , name : "STAR1516 PCIe NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8192u16 , name : "STAR2000 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8193u16 , name : "STAR2000E NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8194u16 , name : "STAR2000C NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8195u16 , name : "STAR2000L NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8196u16 , name : "EAST 2000K SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8200u16 , name : "STAR2008 PCIE NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 8208u16 , name : "STAR2010 PCIE NVMe Secure SSD Controller" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 47963u16 , name : "Asgard AN3+ NVMe SSD" , subsystems : & [] } , Device { vendor_id : 40242u16 , id : 64546u16 , name : "Asgard AN3+ NVMe SSD" , subsystems : & [] }] }),
        (5486, Vendor { id : 5486u16 , name : "Alfa Inc" , devices : & [] }),
        (4473, Vendor { id : 4473u16 , name : "Toshiba Corporation" , devices : & [Device { vendor_id : 4473u16 , id : 258u16 , name : "Extended IDE Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 259u16 , name : "EX-IDE Type-B" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 270u16 , name : "PXP04 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 271u16 , name : "XG3 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 272u16 , name : "Cx5 NVMe SSD Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8187u16 , name : "Express Flash NVMe 960G (RI) U.2 (CD5)" } , SubSystem { subvendor : 4136u16 , subdevice : 8188u16 , name : "Express Flash NVMe 1.92T (RI) U.2 (CD5)" } , SubSystem { subvendor : 4136u16 , subdevice : 8189u16 , name : "Express Flash NVMe 3.84T (RI) U.2 (CD5)" } , SubSystem { subvendor : 4136u16 , subdevice : 8190u16 , name : "Express Flash NVMe 7.68T (RI) U.2 (CD5)" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "KIOXIA CM5-R series SSD" } , SubSystem { subvendor : 4473u16 , subdevice : 33u16 , name : "KIOXIA CD5 series SSD" } , SubSystem { subvendor : 7497u16 , subdevice : 16441u16 , name : "Thinksystem U.2 CM5 NVMe SSD" } , SubSystem { subvendor : 7497u16 , subdevice : 16442u16 , name : "Thinksystem AIC CM5 NVMe SSD" }] } , Device { vendor_id : 4473u16 , id : 275u16 , name : "BG3 x2 NVMe SSD Controller (DRAM-less)" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Toshiba KBG30ZMS128G 128GB NVMe SSD" }] } , Device { vendor_id : 4473u16 , id : 277u16 , name : "XG4 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 278u16 , name : "XG5 NVMe SSD Controller" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "XG5 NVMe SSD Controller" }] } , Device { vendor_id : 4473u16 , id : 281u16 , name : "XD5 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 282u16 , name : "XG6 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1028u16 , name : "DVD Decoder card" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1030u16 , name : "Tecra Video Capture device" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1031u16 , name : "DVD Decoder card (Version 2)" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1537u16 , name : "CPU to PCI bridge" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Satellite Pro" }] } , Device { vendor_id : 4473u16 , id : 1538u16 , name : "PCI to ISA bridge" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1539u16 , name : "ToPIC95 PCI to CardBus Bridge for Notebooks" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1540u16 , name : "PCI-Docking Host bridge" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1546u16 , name : "ToPIC95" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Satellite Pro" }] } , Device { vendor_id : 4473u16 , id : 1551u16 , name : "ToPIC97" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Satellite 4010" }] } , Device { vendor_id : 4473u16 , id : 1559u16 , name : "ToPIC100 PCI to Cardbus Bridge with ZV Support" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1560u16 , name : "CPU to PCI and PCI to ISA bridge" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 1793u16 , name : "FIR Port Type-O" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 2051u16 , name : "TC6371AF SD Host Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 2052u16 , name : "TC6371AF SmartMedia Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 2053u16 , name : "SD TypA Controller" , subsystems : & [] } , Device { vendor_id : 4473u16 , id : 3329u16 , name : "FIR Port Type-DO" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "FIR Port Type-DO" }] } , Device { vendor_id : 4473u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (6463, Vendor { id : 6463u16 , name : "AHA Products Group" , devices : & [Device { vendor_id : 6463u16 , id : 1u16 , name : "AHA36x-PCIX" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 864u16 , name : "AHA360-PCIe" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 867u16 , name : "AHA363-PCIe" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 868u16 , name : "AHA364-PCIe" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 871u16 , name : "AHA367-PCIe" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 880u16 , name : "AHA370-PCIe" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 1540u16 , name : "AHA604" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 1541u16 , name : "AHA605" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 13889u16 , name : "AHA3641" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 13890u16 , name : "AHA3642" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 24833u16 , name : "AHA6101" , subsystems : & [] } , Device { vendor_id : 6463u16 , id : 24834u16 , name : "AHA6102" , subsystems : & [] }] }),
        (5588, Vendor { id : 5588u16 , name : "Iwill Corp" , devices : & [] }),
        (7424, Vendor { id : 7424u16 , name : "Everpure, Inc." , devices : & [] }),
        (5081, Vendor { id : 5081u16 , name : "Apex PC Solutions Inc" , devices : & [] }),
        (4166, Vendor { id : 4166u16 , name : "IPC Corporation, Ltd." , devices : & [] }),
        (4838, Vendor { id : 4838u16 , name : "Cirel Systems" , devices : & [] }),
        (5083, Vendor { id : 5083u16 , name : "Janich & Klass Computertechnik GmbH" , devices : & [] }),
        (5565, Vendor { id : 5565u16 , name : "DFI Inc" , devices : & [] }),
        (65246, Vendor { id : 65246u16 , name : "Fedetec Inc." , devices : & [Device { vendor_id : 65246u16 , id : 3u16 , name : "TABIC PCI v3" , subsystems : & [] }] }),
        (5098, Vendor { id : 5098u16 , name : "Dallas Semiconductor" , devices : & [] }),
        (4303, Vendor { id : 4303u16 , name : "Fujitsu Limited." , devices : & [Device { vendor_id : 4303u16 , id : 495u16 , name : "PCEA4 PCI-Express Dual Port ESCON Adapter" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5140u16 , name : "On-board USB 1.1 companion controller" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5141u16 , name : "On-board USB 2.0 EHCI controller" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5154u16 , name : "E8410 nVidia graphics adapter" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5165u16 , name : "HD audio (Realtek ALC262)" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5168u16 , name : "82566MM Intel 1Gb copper LAN interface" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 5667u16 , name : "PCEA4 PCI-Express Dual Port ESCON Adapter" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8193u16 , name : "mb86605" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8204u16 , name : "MB86613L IEEE1394 OHCI 1.0 Controller" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8208u16 , name : "MB86613S IEEE1394 OHCI 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8217u16 , name : "MB86295S [CORAL P]" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8222u16 , name : "MB86296S [CORAL PA]" , subsystems : & [] } , Device { vendor_id : 4303u16 , id : 8235u16 , name : "MB86297A [Carmine Graphics Controller]" , subsystems : & [] }] }),
        (5433, Vendor { id : 5433u16 , name : "ATELIER INFORMATIQUES et ELECTRONIQUE ETUDES S.A." , devices : & [] }),
        (36366, Vendor { id : 36366u16 , name : "Computone Corporation" , devices : & [] }),
        (7742, Vendor { id : 7742u16 , name : "Shanghai Iluvatar CoreX Semiconductor Co., Ltd." , devices : & [] }),
        (8378, Vendor { id : 8378u16 , name : "Hangzhou Hikstorage Technology Co., Ltd." , devices : & [Device { vendor_id : 8378u16 , id : 4610u16 , name : "NAND memory controller" , subsystems : & [SubSystem { subvendor : 8378u16 , subdevice : 4610u16 , name : "NVMe SSD V2000 (DRAM-less)" }] }] }),
        (8223, Vendor { id : 8223u16 , name : "SpacemiT" , devices : & [Device { vendor_id : 8223u16 , id : 1u16 , name : "X60 PCIe 2.0 x2 Root Complex" , subsystems : & [] } , Device { vendor_id : 8223u16 , id : 2u16 , name : "X100 PCIe Root Complex" , subsystems : & [] }] }),
        (5222, Vendor { id : 5222u16 , name : "Designpro Inc." , devices : & [] }),
        (6178, Vendor { id : 6178u16 , name : "Twinhan Technology Co. Ltd" , devices : & [Device { vendor_id : 6178u16 , id : 20021u16 , name : "Mantis DTV PCI Bridge Controller [Ver 1.0]" , subsystems : & [] }] }),
        (4920, Vendor { id : 4920u16 , name : "BT Electronics" , devices : & [] }),
        (4676, Vendor { id : 4676u16 , name : "AVM GmbH" , devices : & [Device { vendor_id : 4676u16 , id : 1792u16 , name : "B1 ISDN" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 2048u16 , name : "C4 ISDN" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 2560u16 , name : "A1 ISDN [Fritz]" , subsystems : & [SubSystem { subvendor : 4676u16 , subdevice : 2560u16 , name : "FRITZ!Card ISDN Controller" }] } , Device { vendor_id : 4676u16 , id : 3584u16 , name : "Fritz!Card PCI v2.0 ISDN" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 3712u16 , name : "Fritz!Card PCI v2.1 ISDN" , subsystems : & [SubSystem { subvendor : 4676u16 , subdevice : 3584u16 , name : "PSB 3100F (AVM KAFKA) [Fritz!Card PCI v2.1]" }] } , Device { vendor_id : 4676u16 , id : 4352u16 , name : "C2 ISDN" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 4608u16 , name : "T1 ISDN" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 9984u16 , name : "Fritz!Card DSL SL" , subsystems : & [] } , Device { vendor_id : 4676u16 , id : 10496u16 , name : "Fritz!Card DSL v2.0" , subsystems : & [] }] }),
        (6788, Vendor { id : 6788u16 , name : "Commex Technologies" , devices : & [Device { vendor_id : 6788u16 , id : 1u16 , name : "Vulcan SP HT6210 10-Gigabit Ethernet (rev 02)" , subsystems : & [] }] }),
        (4272, Vendor { id : 4272u16 , name : "CardExpert Technology" , devices : & [] }),
        (22272, Vendor { id : 22272u16 , name : "Netpower" , devices : & [] }),
        (7117, Vendor { id : 7117u16 , name : "Apacer Technology" , devices : & [Device { vendor_id : 7117u16 , id : 288u16 , name : "NVMe SSD Drive 960GB" , subsystems : & [] } , Device { vendor_id : 7117u16 , id : 384u16 , name : "PB4480 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7117u16 , id : 784u16 , name : "NVMe SSD Drive 480GB" , subsystems : & [] } , Device { vendor_id : 7117u16 , id : 20519u16 , name : "AS2280Q4 NVMe SSD" , subsystems : & [] }] }),
        (6945, Vendor { id : 6945u16 , name : "ASMedia Technology Inc." , devices : & [Device { vendor_id : 6945u16 , id : 1553u16 , name : "ASM1061 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 1554u16 , name : "ASM1061/ASM1062 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 1554u16 , name : "Motherboard" }] } , Device { vendor_id : 6945u16 , id : 1570u16 , name : "ASM106x Serial ATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 38497u16 , name : "LRST9661 2-port M.2 SATA3(6Gb/s) Raid Adapter" }] } , Device { vendor_id : 6945u16 , id : 1572u16 , name : "ASM106x SATA/RAID Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 1573u16 , name : "106x SATA/RAID Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4160u16 , name : "ASM1040 SuperSpeed USB Host Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4161u16 , name : "ASM1041 SuperSpeed USB Host Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4162u16 , name : "ASM1042 SuperSpeed USB Host Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4185u16 , name : "K53SM motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33928u16 , name : "P8B WS Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 4162u16 , name : "Motherboard" }] } , Device { vendor_id : 6945u16 , id : 4196u16 , name : "ASM1064 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4224u16 , name : "ASM1083/1085 PCIe to PCI Bridge" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 4224u16 , name : "Motherboard" }] } , Device { vendor_id : 6945u16 , id : 4418u16 , name : "ASM1042A USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4452u16 , name : "ASM1164 Serial ATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4454u16 , name : "ASM1166 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4482u16 , name : "ASM1182e 2-Port PCIe x1 Gen2 Packet Switch" , subsystems : & [SubSystem { subvendor : 6945u16 , subdevice : 4495u16 , name : "ASM1182e 2-Port PCIe x1 Gen2 Packet Switch" }] } , Device { vendor_id : 6945u16 , id : 4484u16 , name : "ASM1184e 4-Port PCIe x1 Gen2 Packet Switch" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 4484u16 , name : "ASM1184e 4-Port PCIe x1 Gen2 Packet Switch" }] } , Device { vendor_id : 6945u16 , id : 4487u16 , name : "ASM1187e 7-Port PCIe x1 Gen2 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4495u16 , name : "ASM1187e 7-Port PCIe x1 Gen2 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 4674u16 , name : "ASM1142 USB 3.1 Host Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 39490u16 , name : "LRSU9A42 2-Port Type-A Exchange Adapter" }] } , Device { vendor_id : 6945u16 , id : 4931u16 , name : "ASM1143 USB 3.1 Host Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 6150u16 , name : "ASM1806 4-Port PCIe x2 Gen2 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 6162u16 , name : "ASM1812 6-Port PCIe x4 Gen2 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 6180u16 , name : "ASM1824 12-Port PCIe x8 Gen2 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 8514u16 , name : "ASM2142/ASM3142 USB 3.1 Host Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 6945u16 , id : 9249u16 , name : "ASM4242 PCIe Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 9251u16 , name : "ASM4242 PCIe Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 9253u16 , name : "ASM4242 USB 4 / Thunderbolt 3 Host Router" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 9254u16 , name : "ASM4242 USB 3.2 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 9315u16 , name : "ASM2464PD USB4 Device Controller 40G" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 10246u16 , name : "ASM2806 4-Port PCIe x2 Gen3 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 10258u16 , name : "ASM2812 6-Port PCIe x4 Gen3 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 10276u16 , name : "ASM2824 PCIe Gen3 Packet Switch" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 12354u16 , name : "ASM3042 USB 3.2 Gen 1 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 12610u16 , name : "ASM3142 USB 3.2 Gen 2x1 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 12865u16 , name : "ASM3241 USB 3.2 Gen 2 Host Controller" , subsystems : & [] } , Device { vendor_id : 6945u16 , id : 12866u16 , name : "ASM3242 USB 3.2 Host Controller" , subsystems : & [] }] }),
        (8148, Vendor { id : 8148u16 , name : "SUNIX Co., Ltd." , devices : & [Device { vendor_id : 8148u16 , id : 1u16 , name : "Matrix multiport serial adapter" , subsystems : & [] } , Device { vendor_id : 8148u16 , id : 6553u16 , name : "Multiport serial controller" , subsystems : & [] }] }),
        (7852, Vendor { id : 7852u16 , name : "Quectel Wireless Solutions Co., Ltd." , devices : & [Device { vendor_id : 7852u16 , id : 4097u16 , name : "EM120R-GL LTE Modem" , subsystems : & [] } , Device { vendor_id : 7852u16 , id : 4098u16 , name : "EM160R-GL LTE Modem" , subsystems : & [] } , Device { vendor_id : 7852u16 , id : 8193u16 , name : "EM120R-GL" , subsystems : & [] }] }),
        (8454, Vendor { id : 8454u16 , name : "ZCHL Technology Co., Ltd" , devices : & [Device { vendor_id : 8454u16 , id : 1u16 , name : "HL100 Accelerator Controller" , subsystems : & [SubSystem { subvendor : 8454u16 , subdevice : 1u16 , name : "HLC100 Accelerator Card" }] }] }),
        (4949, Vendor { id : 4949u16 , name : "Kratos Analytical Ltd" , devices : & [] }),
        (4625, Vendor { id : 4625u16 , name : "Braintech Inc" , devices : & [] }),
        (4244, Vendor { id : 4244u16 , name : "First International Computers [FIC]" , devices : & [] }),
        (5619, Vendor { id : 5619u16 , name : "Digitmedia Corp" , devices : & [] }),
        (8336, Vendor { id : 8336u16 , name : "Rivian Automotive" , devices : & [] }),
        (4973, Vendor { id : 4973u16 , name : "Gigalabs Inc" , devices : & [] }),
        (34888, Vendor { id : 34888u16 , name : "MUCSE" , devices : & [Device { vendor_id : 34888u16 , id : 4096u16 , name : "Ethernet Controller N10 Series for 10GbE or 40GbE (Dual-port)" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 12338u16 , name : "LRES3032PF Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 34888u16 , subdevice : 33808u16 , name : "Ethernet Network Adapter N10G-X2-DC for 10GbE SFP+ 2-port" } , SubSystem { subvendor : 34888u16 , subdevice : 50192u16 , name : "Ethernet Network Adapter OCP3.0 N10G-X2-DCP for 10GbE SFP+ 2-port" }] } , Device { vendor_id : 34888u16 , id : 4097u16 , name : "Ethernet Controller N400 Series for 1GbE (Dual-port)" , subsystems : & [SubSystem { subvendor : 34888u16 , subdevice : 32768u16 , name : "Ethernet Network Adapter N400L-X4-DD for 1GbE SFP 2-port" }] } , Device { vendor_id : 34888u16 , id : 4099u16 , name : "Ethernet Controller N400 Series for 10GbE (Single-port)" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4176u16 , name : "LRES1050PF Single-port 10Gb Ethernet Network Adapter" }] } , Device { vendor_id : 34888u16 , id : 4128u16 , name : "Ethernet Controller N10 Series for 10GbE (Quad-port)" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4144u16 , name : "LRES1030PF Quad-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4145u16 , name : "LRES1031PF Dual-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12337u16 , name : "LRES3031PF Quad-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 34888u16 , subdevice : 33873u16 , name : "Ethernet Network Adapter N10G-X4-QC for 10GbE SFP+ 4-port" } , SubSystem { subvendor : 34888u16 , subdevice : 50257u16 , name : "Ethernet Network Adapter OCP3.0 N10G-X4-QCP for 10GbE SFP+ 4-port" }] } , Device { vendor_id : 34888u16 , id : 4129u16 , name : "Ethernet Controller N400 Series for 1GbE (Quad-port)" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4146u16 , name : "LRES1032PF Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4153u16 , name : "LRES1039PT Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 34888u16 , subdevice : 32833u16 , name : "Ethernet Network Adapter N400L-X4-QD for 1GbE SFP 4-port" }] } , Device { vendor_id : 34888u16 , id : 4192u16 , name : "Ethernet Controller N10 Series for 1GbE or 10GbE (8-port)" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 4224u16 , name : "Ethernet Controller N10 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 4225u16 , name : "Ethernet Controller N400 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 4227u16 , name : "Ethernet Controller N400 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 33288u16 , name : "Ethernet Controller N210M for 1GbE 1-port RJ45" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 33289u16 , name : "Ethernet Controller N210 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 33290u16 , name : "Ethernet Controller N210L for 1GbE 1-port RJ45" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 33544u16 , name : "Ethernet Controller N500 Series for 1GbE (Quad-port, Copper RJ45)" , subsystems : & [SubSystem { subvendor : 6461u16 , subdevice : 4232u16 , name : "NIC-ETH3M0T-3S-4P" } , SubSystem { subvendor : 19538u16 , subdevice : 4168u16 , name : "LRES1048PT Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12356u16 , name : "LRES3044PT Quad-port 1Gb Ethernet Server Adapter for OCP" }] } , Device { vendor_id : 34888u16 , id : 33545u16 , name : "Ethernet Controller N500 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 33560u16 , name : "Ethernet Controller N500 Series for 1GbE (Dual-port, Copper RJ45)" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4169u16 , name : "LRES1049PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12355u16 , name : "LRES3043PT Dual-port 1Gb Ethernet Server Adapter for OCP" }] } , Device { vendor_id : 34888u16 , id : 34048u16 , name : "Ethernet Controller N20 Series for 25GbE" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 5232u16 , name : "LRES1470PF Dual-Port 25G Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 9474u16 , name : "GQM2502E Dual-Port 25G Ethernet Network Adapter" } , SubSystem { subvendor : 34888u16 , subdevice : 2048u16 , name : "Ethernet Network Adapter N20 for 25GbE SFP28" } , SubSystem { subvendor : 34888u16 , subdevice : 34816u16 , name : "Ethernet Network Adapter N20 for RDMA 25GbE SFP28 2-port" } , SubSystem { subvendor : 34888u16 , subdevice : 51200u16 , name : "Ethernet Network Adapter N20 for RDMA 25GbE SFP28 2-port OCP 3.0" }] } , Device { vendor_id : 34888u16 , id : 34049u16 , name : "Ethernet Controller N20 Series for 100GbE" , subsystems : & [SubSystem { subvendor : 34888u16 , subdevice : 34816u16 , name : "Ethernet Network Adapter N20 for RDMA 100GbE QSFP28 2-port" }] } , Device { vendor_id : 34888u16 , id : 34050u16 , name : "Ethernet Controller N20 Series for 40GbE" , subsystems : & [SubSystem { subvendor : 34888u16 , subdevice : 34816u16 , name : "Ethernet Network Adapter N20 for RDMA 40GbE QSFP+ 2-port" }] } , Device { vendor_id : 34888u16 , id : 34051u16 , name : "Ethernet Controller N20 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 34055u16 , name : "Ethernet Controller B-Series for 25GbE" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 34056u16 , name : "Ethernet Controller B-Series for 100GbE" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 34057u16 , name : "Ethernet Controller B-Series for 40GbE" , subsystems : & [] } , Device { vendor_id : 34888u16 , id : 34058u16 , name : "Ethernet Controller B-Series Virtual Function" , subsystems : & [] }] }),
        (36398, Vendor { id : 36398u16 , name : "KTI" , devices : & [Device { vendor_id : 36398u16 , id : 12288u16 , name : "ET32P2" , subsystems : & [] }] }),
        (6966, Vendor { id : 6966u16 , name : "Red Hat, Inc." , devices : & [Device { vendor_id : 6966u16 , id : 1u16 , name : "QEMU PCI-PCI bridge" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 2u16 , name : "QEMU PCI 16550A Adapter" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 6966u16 , id : 3u16 , name : "QEMU PCI Dual-port 16550A Adapter" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 6966u16 , id : 4u16 , name : "QEMU PCI Quad-port 16550A Adapter" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 6966u16 , id : 5u16 , name : "QEMU PCI Test Device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 6966u16 , id : 6u16 , name : "PCI Rocker Ethernet switch device" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 7u16 , name : "PCI SD Card Host Controller Interface" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 8u16 , name : "QEMU PCIe Host bridge" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 9u16 , name : "QEMU PCI Expander bridge" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 10u16 , name : "PCI-PCI bridge (multiseat)" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 11u16 , name : "QEMU PCIe Expander bridge" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 12u16 , name : "QEMU PCIe Root port" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 13u16 , name : "QEMU XHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 14u16 , name : "QEMU PCIe-to-PCI bridge" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 16u16 , name : "QEMU NVM Express Controller" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 17u16 , name : "QEMU PVPanic device" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 19u16 , name : "QEMU UFS Host Controller" , subsystems : & [] } , Device { vendor_id : 6966u16 , id : 256u16 , name : "QXL paravirtual graphic card" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] }] }),
        (6042, Vendor { id : 6042u16 , name : "id Quantique" , devices : & [Device { vendor_id : 6042u16 , id : 1u16 , name : "Quantis PCI 16Mbps" , subsystems : & [] }] }),
        (4536, Vendor { id : 4536u16 , name : "XPoint Technologies, Inc" , devices : & [Device { vendor_id : 4536u16 , id : 1u16 , name : "Quad PeerMaster" , subsystems : & [] }] }),
        (4839, Vendor { id : 4839u16 , name : "Sunsgroup Corporation" , devices : & [] }),
        (4380, Vendor { id : 4380u16 , name : "Tricord Systems Inc." , devices : & [Device { vendor_id : 4380u16 , id : 1u16 , name : "Powerbis Bridge" , subsystems : & [] }] }),
        (55373, Vendor { id : 55373u16 , name : "Exsys" , devices : & [] }),
        (5060, Vendor { id : 5060u16 , name : "Phase Metrics" , devices : & [] }),
        (17956, Vendor { id : 17956u16 , name : "Budker Institute of Nuclear Physics" , devices : & [Device { vendor_id : 17956u16 , id : 44481u16 , name : "ADC200ME High speed ADC" , subsystems : & [] } , Device { vendor_id : 17956u16 , id : 56833u16 , name : "DL200ME High resolution delay line PCI based card" , subsystems : & [] } , Device { vendor_id : 17956u16 , id : 56834u16 , name : "DL200ME Middle resolution delay line PCI based card" , subsystems : & [] }] }),
        (4542, Vendor { id : 4542u16 , name : "International Microcircuits Inc" , devices : & [] }),
        (5233, Vendor { id : 5233u16 , name : "Integrated Telecom Express Inc" , devices : & [] }),
        (61719, Vendor { id : 61719u16 , name : "Cerio" , devices : & [Device { vendor_id : 61719u16 , id : 255u16 , name : "Test Device" , subsystems : & [] } , Device { vendor_id : 61719u16 , id : 4096u16 , name : "Emulated PCIe Switch" , subsystems : & [] } , Device { vendor_id : 61719u16 , id : 4112u16 , name : "Placeholder Device" , subsystems : & [] } , Device { vendor_id : 61719u16 , id : 4128u16 , name : "Pseudo-Device" , subsystems : & [] } , Device { vendor_id : 61719u16 , id : 4144u16 , name : "Test Device" , subsystems : & [] }] }),
        (6776, Vendor { id : 6776u16 , name : "Virident Systems Inc." , devices : & [Device { vendor_id : 6776u16 , id : 49u16 , name : "FlashMAX Drive" , subsystems : & [SubSystem { subvendor : 6776u16 , subdevice : 52u16 , name : "FlashMAX PCIe SSD [rev 3]" } , SubSystem { subvendor : 6776u16 , subdevice : 55u16 , name : "FlashMAX PCIe SSD [rev 3D]" } , SubSystem { subvendor : 6776u16 , subdevice : 56u16 , name : "FlashMAX PCIe SSD [rev 4]" } , SubSystem { subvendor : 6776u16 , subdevice : 57u16 , name : "FlashMAX PCIe SSD [rev 4D]" }] } , Device { vendor_id : 6776u16 , id : 64u16 , name : "FlashMAX II" , subsystems : & [] } , Device { vendor_id : 6776u16 , id : 65u16 , name : "FlashMAX II" , subsystems : & [] } , Device { vendor_id : 6776u16 , id : 66u16 , name : "FlashMAX II" , subsystems : & [] } , Device { vendor_id : 6776u16 , id : 80u16 , name : "FlashMAX III" , subsystems : & [] }] }),
        (4119, Vendor { id : 4119u16 , name : "SPEA Software AG" , devices : & [Device { vendor_id : 4119u16 , id : 21315u16 , name : "SPEA 3D Accelerator" , subsystems : & [] }] }),
        (5451, Vendor { id : 5451u16 , name : "COMPUTEX Co Ltd" , devices : & [] }),
        (7573, Vendor { id : 7573u16 , name : "Graphcore Ltd" , devices : & [Device { vendor_id : 7573u16 , id : 1u16 , name : "Colossus GC2 [C2]" , subsystems : & [] } , Device { vendor_id : 7573u16 , id : 2u16 , name : "Colossus GC1 [S1]" , subsystems : & [] }] }),
        (7545, Vendor { id : 7545u16 , name : "Transcend Information, Inc." , devices : & [Device { vendor_id : 7545u16 , id : 8802u16 , name : "NVMe PCIe SSD 220S/MTE662T2" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 8803u16 , name : "NVMe PCIe SSD 110S/112S/120S/MTE300S/MTE400S/MTE652T2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 8804u16 , name : "NVMe PCIe SSD 250H" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 8807u16 , name : "NVMe PCIe SSD 220S/240S/MTE710T" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 8808u16 , name : "NVMe PCIe SSD 245S (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 8809u16 , name : "NVMe PCIe SSD 410S (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 9992u16 , name : "USD710S microSDXC EX SD Express 7.1 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7545u16 , id : 22374u16 , name : "NVMe PCIe SSD 110Q (DRAM-less)" , subsystems : & [] }] }),
        (4214, Vendor { id : 4214u16 , name : "Chaintech Computer Co. Ltd" , devices : & [] }),
        (4576, Vendor { id : 4576u16 , name : "Cray Communications A/S" , devices : & [] }),
        (5437, Vendor { id : 5437u16 , name : "FILANET Corp" , devices : & [] }),
        (4985, Vendor { id : 4985u16 , name : "Asahi Kasei Microsystems Co Ltd" , devices : & [] }),
        (4969, Vendor { id : 4969u16 , name : "Digigram" , devices : & [] }),
        (4872, Vendor { id : 4872u16 , name : "Jato Technologies Inc." , devices : & [Device { vendor_id : 4872u16 , id : 1u16 , name : "NetCelerator Adapter" , subsystems : & [SubSystem { subvendor : 4872u16 , subdevice : 1u16 , name : "NetCelerator Adapter" }] }] }),
        (4223, Vendor { id : 4223u16 , name : "Data Technology Corporation" , devices : & [Device { vendor_id : 4223u16 , id : 2050u16 , name : "SL82C105" , subsystems : & [] }] }),
        (7572, Vendor { id : 7572u16 , name : "Chengdu Haiguang IC Design Co., Ltd." , devices : & [Device { vendor_id : 7572u16 , id : 5200u16 , name : "Root Complex" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5201u16 , name : "I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5202u16 , name : "PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5203u16 , name : "PCIE GPP Bridge" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5204u16 , name : "Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5205u16 , name : "PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5206u16 , name : "PSPCCP Command DMA Processor" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5208u16 , name : "10 Gb Ethernet Controller Port 0/Port1" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5209u16 , name : "10 Gb Ethernet Controller Port 2/Port3" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5210u16 , name : "PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5211u16 , name : "PCIE Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5212u16 , name : "USB3 XHCI" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5213u16 , name : "Switch upstream in PCIe" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5214u16 , name : "Switch downstream in PCIe" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5215u16 , name : "USB 3.0 Host controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5216u16 , name : "Data Fabric: Device 18h; Function 0" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5217u16 , name : "Data Fabric: Device 18h; Function 1" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5218u16 , name : "Data Fabric: Device 18h; Function 2" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5219u16 , name : "Data Fabric: Device 18h; Function 3" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5220u16 , name : "Data Fabric: Device 18h; Function 4" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5221u16 , name : "Data Fabric: Device 18h; Function 5" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5222u16 , name : "Data Fabric: Device 18h; Function 6" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5223u16 , name : "Data Fabric: Device 18h; Function 7" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5224u16 , name : "NTBCCP" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 5291u16 , name : "HGE1000 1000/100/10 Mb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 8203u16 , name : "HGE2500 2500/1000/100/10 Mb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 8204u16 , name : "HGE2500 1000/100/10 Mb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 25105u16 , name : "K100_AI" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 30977u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 30980u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 30982u16 , name : "FCH SD Flash Controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 30987u16 , name : "FCH SMBus Controller" , subsystems : & [] } , Device { vendor_id : 7572u16 , id : 30990u16 , name : "FCH LPC Bridge" , subsystems : & [] }] }),
        (5051, Vendor { id : 5051u16 , name : "Sanyo Technosound Co Ltd" , devices : & [] }),
        (4546, Vendor { id : 4546u16 , name : "Sand Microelectronics" , devices : & [] }),
        (36868, Vendor { id : 36868u16 , name : "Adaptec" , devices : & [Device { vendor_id : 36868u16 , id : 120u16 , name : "AHA-2940U_CN" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 4216u16 , name : "AIC-7810" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 4448u16 , name : "AIC-1160 [Family Fibre Channel Adapter]" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 8568u16 , name : "AIC-7821" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 14432u16 , name : "AHA-2930CU" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 15224u16 , name : "AHA-4844W/4844UW" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 20597u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 20600u16 , name : "AIC-7850T/7856T [AVA-2902/4/6 / AHA-2910]" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30800u16 , name : "AIC-7850T/7856T [AVA-290x / AHA-2910]" }] } , Device { vendor_id : 36868u16 , id : 20853u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 20856u16 , name : "AIC-7851" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21109u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21112u16 , name : "AIC-7852" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21365u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21368u16 , name : "AIC-7850" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21621u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21624u16 , name : "AIC-7850" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21877u16 , name : "AVA-2930" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 21880u16 , name : "AIC-7855" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22087u16 , name : "ANA-7711 TCP Offload Engine" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30480u16 , name : "ANA-7711F TCP Offload Engine - Optical" } , SubSystem { subvendor : 36868u16 , subdevice : 30481u16 , name : "ANA-7711LP TCP Offload Engine - Copper" }] } , Device { vendor_id : 36868u16 , id : 22133u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22136u16 , name : "AIC-7856" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22389u16 , name : "AIC-755x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22392u16 , name : "AIC-7850" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22528u16 , name : "AIC-5800" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22784u16 , name : "ANA-5910/5930/5940 ATM155 & 25 LAN Adapter" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 22789u16 , name : "ANA-5910A/5930A/5940A ATM Adapter" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 24632u16 , name : "AIC-3860" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 24693u16 , name : "AIC-1480 / APA-1480" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30048u16 , name : "AIC-1480 / APA-1480 Cardbus" }] } , Device { vendor_id : 36868u16 , id : 24696u16 , name : "AIC-7860" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 24952u16 , name : "AIC-7861" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30817u16 , name : "AHA-2940AU Single" }] } , Device { vendor_id : 36868u16 , id : 25208u16 , name : "AIC-7860" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 25464u16 , name : "AIC-7860" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 25720u16 , name : "AIC-786x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 25976u16 , name : "AIC-786x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 26232u16 , name : "AIC-786x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 26488u16 , name : "AIC-786x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 26901u16 , name : "ANA620xx/ANA69011A" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 8u16 , name : "ANA69011A/TX 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 9u16 , name : "ANA69011A/TX 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 16u16 , name : "ANA62022 2-port 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 24u16 , name : "ANA62044 4-port 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 25u16 , name : "ANA62044 4-port 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32u16 , name : "ANA62022 2-port 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 40u16 , name : "ANA69011A/TX 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32776u16 , name : "ANA69011A/TX 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32777u16 , name : "ANA69011A/TX 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32784u16 , name : "ANA62022 2-port 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32792u16 , name : "ANA62044 4-port 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32793u16 , name : "ANA62044 4-port 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32800u16 , name : "ANA62022 2-port 64 bit 10/100" } , SubSystem { subvendor : 36868u16 , subdevice : 32808u16 , name : "ANA69011A/TX 64 bit 10/100" }] } , Device { vendor_id : 36868u16 , id : 28792u16 , name : "AHA-294x / AIC-7870" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 29048u16 , name : "AIC-7870P/7871 [AHA-2940/W/S76]" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 29304u16 , name : "AHA-3940/3940W / AIC-7872" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 29560u16 , name : "AHA-3985 / AIC-7873" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 29816u16 , name : "AHA-2944/2944W / AIC-7874" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30072u16 , name : "AHA-3944/3944W / AIC-7875" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30328u16 , name : "AHA-4944W/UW / AIC-7876" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30480u16 , name : "ANA-7711F Network Accelerator Card (NAC) - Optical" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30481u16 , name : "ANA-7711C Network Accelerator Card (NAC) - Copper" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30584u16 , name : "AIC-787x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30736u16 , name : "AIC-7810" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30741u16 , name : "AIC-7815 RAID+Memory Controller IC" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30741u16 , name : "ARO-1130U2 RAID Controller" } , SubSystem { subvendor : 36868u16 , subdevice : 30784u16 , name : "AIC-7815 RAID+Memory Controller IC" }] } , Device { vendor_id : 36868u16 , id : 30800u16 , name : "AIC-7850" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30805u16 , name : "AHA-2930" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30816u16 , name : "AIC-7860" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30832u16 , name : "AIC-7870" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30833u16 , name : "AHA-2940" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30834u16 , name : "AHA-3940" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30835u16 , name : "AHA-3980" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30836u16 , name : "AHA-2944" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30848u16 , name : "AIC-7880P" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30864u16 , name : "AIC-7890" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30865u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30866u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30867u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30868u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30869u16 , name : "AHA-2940U/UW / AHA-39xx / AIC-7895" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30864u16 , name : "AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B" } , SubSystem { subvendor : 36868u16 , subdevice : 30865u16 , name : "AHA-2940U/2940UW Dual" } , SubSystem { subvendor : 36868u16 , subdevice : 30866u16 , name : "AHA-3940AU/AUW/AUWD/UWD" } , SubSystem { subvendor : 36868u16 , subdevice : 30868u16 , name : "AHA-3944AUWD" } , SubSystem { subvendor : 36868u16 , subdevice : 30869u16 , name : "AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B" } , SubSystem { subvendor : 36868u16 , subdevice : 30870u16 , name : "AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B" } , SubSystem { subvendor : 36868u16 , subdevice : 30871u16 , name : "AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B" }] } , Device { vendor_id : 36868u16 , id : 30870u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 30871u16 , name : "AIC-789x" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 32888u16 , name : "AIC-7880U" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30848u16 , name : "AIC-7880P Ultra/Ultra Wide SCSI Chipset" }] } , Device { vendor_id : 36868u16 , id : 33144u16 , name : "AIC-7870P/7881U [AHA-2940U/UW/D/S76]" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30849u16 , name : "AHA-2940UW SCSI Host Adapter" }] } , Device { vendor_id : 36868u16 , id : 33400u16 , name : "AHA-3940U/UW/UWD / AIC-7882U" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 33656u16 , name : "AHA-3940U/UW / AIC-7883U" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 33912u16 , name : "AHA-2944UW / AIC-7884U" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 34168u16 , name : "AHA-3944U/UWD / AIC-7885" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 34424u16 , name : "AHA-4944UW / AIC-7886" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 34680u16 , name : "AHA-2940UW Pro / AIC-788x" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30855u16 , name : "2940UW Pro Ultra-Wide SCSI Controller" }] } , Device { vendor_id : 36868u16 , id : 34936u16 , name : "AHA-2930UW / AIC-7888" , subsystems : & [SubSystem { subvendor : 36868u16 , subdevice : 30856u16 , name : "AHA-2930UW SCSI Controller" }] } , Device { vendor_id : 36868u16 , id : 35704u16 , name : "ABA-1030" , subsystems : & [] } , Device { vendor_id : 36868u16 , id : 60536u16 , name : "AHA-4944W/UW" , subsystems : & [] }] }),
        (5309, Vendor { id : 5309u16 , name : "CARDIO Control N.V." , devices : & [] }),
        (6537, Vendor { id : 6537u16 , name : "Montilio Inc." , devices : & [Device { vendor_id : 6537u16 , id : 1u16 , name : "RapidFile Bridge" , subsystems : & [] } , Device { vendor_id : 6537u16 , id : 32769u16 , name : "RapidFile" , subsystems : & [] }] }),
        (8289, Vendor { id : 8289u16 , name : "Unis Flash Memory" , devices : & [Device { vendor_id : 8289u16 , id : 16384u16 , name : "E4000 controller" , subsystems : & [] } , Device { vendor_id : 8289u16 , id : 16640u16 , name : "E4100 controller" , subsystems : & [] } , Device { vendor_id : 8289u16 , id : 16896u16 , name : "E4200 NVMe Controller" , subsystems : & [SubSystem { subvendor : 8289u16 , subdevice : 8225u16 , name : "E4200 NVMe SSD 1.6TB PCIe 4.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8226u16 , name : "E4200 NVMe SSD 1.92TB PCIe 4.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8227u16 , name : "E4200 NVMe SSD 3.2TB PCIe 4.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8228u16 , name : "E4200 NVMe SSD 3.84TB PCIe 4.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8229u16 , name : "E4200 NVMe SSD 6.4TB PCIe 4.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8230u16 , name : "E4200 NVMe SSD 7.68TB PCIe 4.0 U.2" }] } , Device { vendor_id : 8289u16 , id : 20736u16 , name : "E5100 NVMe Controller" , subsystems : & [SubSystem { subvendor : 8289u16 , subdevice : 8241u16 , name : "E5100 NVMe SSD 3.2TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8242u16 , name : "E5100 NVMe SSD 3.84TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8243u16 , name : "E5100 NVMe SSD 6.4TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8244u16 , name : "E5100 NVMe SSD 7.68TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8245u16 , name : "E5100 NVMe SSD 12.8TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8246u16 , name : "E5100 NVMe SSD 15.36TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8273u16 , name : "E510Q NVMe SSD 30.72TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8274u16 , name : "E510Q NVMe SSD 61.44TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8275u16 , name : "E510Q NVMe SSD 122.88TB PCIe 5.0 U.2" }] } , Device { vendor_id : 8289u16 , id : 20992u16 , name : "E5200 NVMe Controller" , subsystems : & [SubSystem { subvendor : 8289u16 , subdevice : 8257u16 , name : "E5200 NVMe SSD 3.2TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8258u16 , name : "E5200 NVMe SSD 3.84TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8259u16 , name : "E5200 NVMe SSD 6.4TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8260u16 , name : "E5200 NVMe SSD 7.68TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8261u16 , name : "E5200 NVMe SSD 12.8TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8262u16 , name : "E5200 NVMe SSD 15.36TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8289u16 , name : "E520Q NVMe SSD 30.72TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8290u16 , name : "E520Q NVMe SSD 61.44TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8291u16 , name : "E520Q NVMe SSD 122.88TB PCIe 5.0 U.2" }] } , Device { vendor_id : 8289u16 , id : 21248u16 , name : "E5300 NVMe Controller" , subsystems : & [SubSystem { subvendor : 8289u16 , subdevice : 8325u16 , name : "E5300 NVMe SSD 3.2TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8326u16 , name : "E5300 NVMe SSD 3.84TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8327u16 , name : "E5300 NVMe SSD 6.4TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8328u16 , name : "E5300 NVMe SSD 7.68TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8329u16 , name : "E5300 NVMe SSD 12.8TB PCIe 5.0 U.2" } , SubSystem { subvendor : 8289u16 , subdevice : 8330u16 , name : "E5300 NVMe SSD 15.36TB PCIe 5.0 U.2" }] }] }),
        (5299, Vendor { id : 5299u16 , name : "XPEED Inc" , devices : & [Device { vendor_id : 5299u16 , id : 0u16 , name : "DSL NIC" , subsystems : & [] }] }),
        (6067, Vendor { id : 6067u16 , name : "Hawking Technologies" , devices : & [Device { vendor_id : 6067u16 , id : 43784u16 , name : "PN672TX 10/100 Ethernet" , subsystems : & [] }] }),
        (5599, Vendor { id : 5599u16 , name : "Infinilink Corp" , devices : & [] }),
        (6690, Vendor { id : 6690u16 , name : "Ambric Inc." , devices : & [] }),
        (4305, Vendor { id : 4305u16 , name : "FuturePlus Systems Corp." , devices : & [] }),
        (45323, Vendor { id : 45323u16 , name : "Uakron PCI Project" , devices : & [] }),
        (6095, Vendor { id : 6095u16 , name : "Z-Com, Inc." , devices : & [] }),
        (4744, Vendor { id : 4744u16 , name : "Timestep Corporation" , devices : & [] }),
        (4381, Vendor { id : 4381u16 , name : "Microsemi / PMC / IDT" , devices : & [Device { vendor_id : 4381u16 , id : 1u16 , name : "IDT77201/77211 155Mbps ATM SAR Controller [NICStAR]" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 3u16 , name : "IDT77222/77252 155Mbps ATM MICRO ABR SAR Controller" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 4u16 , name : "IDT77V252 155Mbps ATM MICRO ABR SAR Controller" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 5u16 , name : "IDT77V222 155Mbps ATM MICRO ABR SAR Controller" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32792u16 , name : "PES12N3A 12-lane 3-Port PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32796u16 , name : "PES24N3A PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32808u16 , name : "PES4T4 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32811u16 , name : "PES8T5A PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32812u16 , name : "PES16T4 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32813u16 , name : "PES16T7 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32814u16 , name : "PES24T6 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32815u16 , name : "PES32T8 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32818u16 , name : "PES48T12 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32820u16 , name : "PES16/22/34H16 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32821u16 , name : "PES32H8 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32822u16 , name : "PES48H12 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32823u16 , name : "PES64H16 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32825u16 , name : "PES3T3 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32826u16 , name : "PES4T4 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32828u16 , name : "PES5T5 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32829u16 , name : "PES6T5 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32840u16 , name : "PES8NT2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32841u16 , name : "PES8NT2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32842u16 , name : "PES8NT2 PCI Express Internal NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32843u16 , name : "PES8NT2 PCI Express External NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32844u16 , name : "PES16NT2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32845u16 , name : "PES16NT2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32846u16 , name : "PES16NT2 PCI Express Internal NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32847u16 , name : "PES16NT2 PCI Express External NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32856u16 , name : "PES12NT3 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32857u16 , name : "PES12NT3 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32858u16 , name : "PES12NT3 PCI Express Internal NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32859u16 , name : "PES12NT3 PCI Express External NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32860u16 , name : "PES24NT3 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32861u16 , name : "PES24NT3 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32862u16 , name : "PES24NT3 PCI Express Internal NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32863u16 , name : "PES24NT3 PCI Express External NTB" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32864u16 , name : "PES16T4G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32865u16 , name : "PES12T3G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32872u16 , name : "PES6T6G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32874u16 , name : "PES24T3G2 PCI Express Gen2 Switch" , subsystems : & [SubSystem { subvendor : 5313u16 , subdevice : 12u16 , name : "10G-PCIE2-8B2" }] } , Device { vendor_id : 4381u16 , id : 32876u16 , name : "PES16T4A/4T4G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32878u16 , name : "PES24T6G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32879u16 , name : "HIO524G2 PCI Express Gen2 Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32887u16 , name : "89HPES64H16G2 64-Lane 16-Port PCIe Gen2 System Interconnect Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32904u16 , name : "PES32NT8BG2 PCI Express Switch" , subsystems : & [SubSystem { subvendor : 4243u16 , subdevice : 29999u16 , name : "PXIe-8383mc Device" } , SubSystem { subvendor : 4243u16 , subdevice : 30019u16 , name : "PXIe-8383mc System Host" } , SubSystem { subvendor : 4243u16 , subdevice : 30044u16 , name : "PXIe-8364" } , SubSystem { subvendor : 4243u16 , subdevice : 30045u16 , name : "PXIe-8374" } , SubSystem { subvendor : 4243u16 , subdevice : 30207u16 , name : "PXIe-8383mc DMA" } , SubSystem { subvendor : 4243u16 , subdevice : 30208u16 , name : "PXIe-8383mc DMA" } , SubSystem { subvendor : 4243u16 , subdevice : 30210u16 , name : "PXIe-8384" }] } , Device { vendor_id : 4381u16 , id : 32906u16 , name : "89HPES32NT24BG2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32908u16 , name : "89HPES32NT24AG2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32910u16 , name : "PES24NT24G2 PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32911u16 , name : "89HPES32NT8AG2 32-Lane 8-Port PCIe Gen2 System Interconnect Switch with Non-Transparent Bridging" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32912u16 , name : "89HPES16NT16G2 16-Lane 16-Port PCIe Gen2 System Interconnect Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32914u16 , name : "89HPES12NT12G2 12-Lane 12-Port PCIe Gen2 System Interconnect Switch" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32975u16 , name : "F32P08xG3 [PCIe boot mode]" , subsystems : & [] } , Device { vendor_id : 4381u16 , id : 32978u16 , name : "F32P08xG3 NVMe controller" , subsystems : & [] }] }),
        (4613, Vendor { id : 4613u16 , name : "Array Corporation" , devices : & [] }),
        (4616, Vendor { id : 4616u16 , name : "Parsytec GmbH" , devices : & [Device { vendor_id : 4616u16 , id : 18515u16 , name : "HS-Link Device" , subsystems : & [] }] }),
        (4799, Vendor { id : 4799u16 , name : "Fujifilm Microdevices" , devices : & [] }),
        (4471, Vendor { id : 4471u16 , name : "Silicon Engineering" , devices : & [] }),
        (17185, Vendor { id : 17185u16 , name : "Tata Power Strategic Electronics Division" , devices : & [] }),
        (7939, Vendor { id : 7939u16 , name : "Shenzhen Shichuangyi Electronics Co., Ltd" , devices : & [Device { vendor_id : 7939u16 , id : 4610u16 , name : "MAP1202-Based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 5640u16 , name : "SCY C5000 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 8802u16 , name : "SM2262EN-based OEM SSD" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 8803u16 , name : "SM2263XT-Based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 8808u16 , name : "SM2268XT-Based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 8809u16 , name : "SM2269XT-Based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 21014u16 , name : "IG5216-based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 21024u16 , name : "IG5220-Based NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 21046u16 , name : "IG5236-Based NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7939u16 , id : 22070u16 , name : "IG5636-Based NVMe SSD" , subsystems : & [] }] }),
        (5214, Vendor { id : 5214u16 , name : "Aashima Technology B.V." , devices : & [] }),
        (6192, Vendor { id : 6192u16 , name : "Credence Systems Corporation" , devices : & [Device { vendor_id : 6192u16 , id : 32768u16 , name : "CPIn" , subsystems : & [] } , Device { vendor_id : 6192u16 , id : 32769u16 , name : "CPId" , subsystems : & [] } , Device { vendor_id : 6192u16 , id : 32770u16 , name : "CPIx" , subsystems : & [] } , Device { vendor_id : 6192u16 , id : 32771u16 , name : "CPIq" , subsystems : & [] }] }),
        (8360, Vendor { id : 8360u16 , name : "Rayson HI-TECH(SZ) Co., Ltd." , devices : & [Device { vendor_id : 8360u16 , id : 4610u16 , name : "RS512GSSD510 PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5077, Vendor { id : 5077u16 , name : "Media 100 Inc" , devices : & [] }),
        (4784, Vendor { id : 4784u16 , name : "Jorge Scientific Corp" , devices : & [] }),
        (4736, Vendor { id : 4736u16 , name : "Photoscript Group Ltd." , devices : & [] }),
        (5753, Vendor { id : 5753u16 , name : "Tokyo Electron Device Ltd." , devices : & [Device { vendor_id : 5753u16 , id : 12288u16 , name : "SD Standard host controller [Ellen]" , subsystems : & [] }] }),
        (6038, Vendor { id : 6038u16 , name : "Research Centre Juelich" , devices : & [Device { vendor_id : 6038u16 , id : 1u16 , name : "SIS1100 [Gigabit link]" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 2u16 , name : "HOTlink" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 3u16 , name : "Counter Timer" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 4u16 , name : "CAMAC Controller" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 5u16 , name : "PROFIBUS" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 6u16 , name : "AMCC HOTlink" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 7u16 , name : "LVD Cable Bus" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 8u16 , name : "100MHz, 64bit Sequence Generator based on VirtexII" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 9u16 , name : "double 14bit-ADC" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 10u16 , name : "SIS1100 with N110 TDC" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 11u16 , name : "double 14bit-ADC with memory" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 13u16 , name : "Synchronisation Slave" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 14u16 , name : "SIS1100-eCMC" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 15u16 , name : "TDC (GPX)" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 16u16 , name : "PCIe Counter Timer" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 17u16 , name : "SIS1100-e single link" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 18u16 , name : "SIS1100-e quad link" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 19u16 , name : "4x2.5GHz SFP to 4 lane PCIe bridge" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 20u16 , name : "SIS1100 with GPX piggy back" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 21u16 , name : "SIS8100 [Gigabit link, MicroTCA]" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 22u16 , name : "SIS1100e with 4 lanes" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 23u16 , name : "Quad 14bit, 50MHz ADC with 2.5GHz SFP" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 24u16 , name : "SIS8300 4-lane PCI Express, Micro TCA for Physics ADC" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 25u16 , name : "SIS SIS8300-Lx MTCA.4 Digitizer" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 26u16 , name : "100MHz, 64bit Sequence Generator based on VirtexII" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 28u16 , name : "Quad 16bit, 150MHz ADC with 2.5GHz SFP" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 48u16 , name : "100MHz, 64bit Sequence Generator based on Spartan6" , subsystems : & [] } , Device { vendor_id : 6038u16 , id : 49u16 , name : "200MHz 64bit Sequence Generator based on Spartan7" , subsystems : & [] }] }),
        (5008, Vendor { id : 5008u16 , name : "Concept Development Inc" , devices : & [] }),
        (8408, Vendor { id : 8408u16 , name : "EHTech (Beijing) Co. Ltd." , devices : & [Device { vendor_id : 8408u16 , id : 513u16 , name : "TUNAN-7 1x200GbE Controller" , subsystems : & [SubSystem { subvendor : 8408u16 , subdevice : 513u16 , name : "TUNAN-7S 1x200GbE Controller" } , SubSystem { subvendor : 8408u16 , subdevice : 529u16 , name : "TUNAN-7 1x200GbE Controller" }] } , Device { vendor_id : 8408u16 , id : 4609u16 , name : "TUNAN-7 2x100GbE Controller" , subsystems : & [SubSystem { subvendor : 8408u16 , subdevice : 4609u16 , name : "TUNAN-7S 2x100GbE Controller" } , SubSystem { subvendor : 8408u16 , subdevice : 4625u16 , name : "TUNAN-7 2x100GbE Controller" }] } , Device { vendor_id : 8408u16 , id : 8705u16 , name : "TUNAN-7 2x25GbE Controller" , subsystems : & [SubSystem { subvendor : 8408u16 , subdevice : 8705u16 , name : "TUNAN-7 2x25GbE Controller" } , SubSystem { subvendor : 8408u16 , subdevice : 8721u16 , name : "TUNAN-7V 2x25GbE Controller" }] } , Device { vendor_id : 8408u16 , id : 12801u16 , name : "TUNAN-7 2x200GbE Controller" , subsystems : & [SubSystem { subvendor : 8408u16 , subdevice : 12801u16 , name : "TUNAN-7 2x200GbE Controller" }] } , Device { vendor_id : 8408u16 , id : 16897u16 , name : "TUNAN-7 1x400GbE Controller" , subsystems : & [SubSystem { subvendor : 8408u16 , subdevice : 16897u16 , name : "TUNAN-7V 1x400GbE Controller" }] }] }),
        (5238, Vendor { id : 5238u16 , name : "MORTARA Instrument Inc" , devices : & [] }),
        (6661, Vendor { id : 6661u16 , name : "deltaww" , devices : & [] }),
        (37906, Vendor { id : 37906u16 , name : "Holtek" , devices : & [Device { vendor_id : 37906u16 , id : 25957u16 , name : "6565" , subsystems : & [] }] }),
        (5368, Vendor { id : 5368u16 , name : "AUDIOCODES Inc" , devices : & [Device { vendor_id : 5368u16 , id : 8311u16 , name : "TP-240 dual span E1 VoIP PCI card" , subsystems : & [] }] }),
        (5270, Vendor { id : 5270u16 , name : "JOYTECH Computer Co., Ltd." , devices : & [] }),
        (4204, Vendor { id : 4204u16 , name : "Hynix Semiconductor" , devices : & [Device { vendor_id : 4204u16 , id : 33081u16 , name : "8139c 100BaseTX Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4204u16 , id : 34817u16 , name : "Dual Pentium ISA/PCI Motherboard" , subsystems : & [] } , Device { vendor_id : 4204u16 , id : 34818u16 , name : "PowerPC ISA/PCI Motherboard" , subsystems : & [] } , Device { vendor_id : 4204u16 , id : 34819u16 , name : "Dual Window Graphics Accelerator" , subsystems : & [] } , Device { vendor_id : 4204u16 , id : 34820u16 , name : "LAN Controller" , subsystems : & [] } , Device { vendor_id : 4204u16 , id : 34821u16 , name : "100-BaseT LAN" , subsystems : & [] }] }),
        (5610, Vendor { id : 5610u16 , name : "Tokyo Denshi Sekei K.K." , devices : & [] }),
        (8888, Vendor { id : 8888u16 , name : "Flex-Logix Technologies" , devices : & [Device { vendor_id : 8888u16 , id : 8864u16 , name : "Flex Logix InferX X1 Inference Accelerator" , subsystems : & [] }] }),
        (5875, Vendor { id : 5875u16 , name : "Jetway Information Co., Ltd." , devices : & [] }),
        (5481, Vendor { id : 5481u16 , name : "Palit Microsystems Inc." , devices : & [] }),
        (5491, Vendor { id : 5491u16 , name : "Lattice - Vantis" , devices : & [] }),
        (5205, Vendor { id : 5205u16 , name : "Logic Plus Plus Inc" , devices : & [] }),
        (4607, Vendor { id : 4607u16 , name : "Scion Corporation" , devices : & [Device { vendor_id : 4607u16 , id : 3u16 , name : "AG-5" , subsystems : & [] }] }),
        (4103, Vendor { id : 4103u16 , name : "NetFrame Systems Inc" , devices : & [] }),
        (22100, Vendor { id : 22100u16 , name : "VoiceTronix Pty Ltd" , devices : & [] }),
        (6116, Vendor { id : 6116u16 , name : "Sectra AB" , devices : & [Device { vendor_id : 6116u16 , id : 1u16 , name : "KK671 Cardbus encryption board" , subsystems : & [] } , Device { vendor_id : 6116u16 , id : 2u16 , name : "KK672 Cardbus encryption board" , subsystems : & [] }] }),
        (4959, Vendor { id : 4959u16 , name : "I-Data International A-S" , devices : & [] }),
        (4875, Vendor { id : 4875u16 , name : "Colorgraphic Communications Corp" , devices : & [] }),
        (4526, Vendor { id : 4526u16 , name : "Aztech System Ltd" , devices : & [] }),
        (4789, Vendor { id : 4789u16 , name : "Granite Systems Inc." , devices : & [] }),
        (6686, Vendor { id : 6686u16 , name : "3Leaf Systems, Inc." , devices : & [] }),
        (4853, Vendor { id : 4853u16 , name : "Forks" , devices : & [] }),
        (5553, Vendor { id : 5553u16 , name : "Source Technology Inc" , devices : & [] }),
        (39441, Vendor { id : 39441u16 , name : "Tiger Lake-H Gaussian & Neural Accelerator" , devices : & [] }),
        (6390, Vendor { id : 6390u16 , name : "NextIO" , devices : & [Device { vendor_id : 6390u16 , id : 4096u16 , name : "[Nexsis] Switch Virtual P2P PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 6390u16 , id : 4097u16 , name : "[Texsis] Switch Virtual P2P PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 6390u16 , id : 4176u16 , name : "[Nexsis] Switch Virtual P2P PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6390u16 , id : 4177u16 , name : "[Texsis] Switch Virtual P2P PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6390u16 , id : 8192u16 , name : "[Nexsis] Switch Integrated Mgmt. Endpoint" , subsystems : & [] } , Device { vendor_id : 6390u16 , id : 8193u16 , name : "[Texsis] Switch Integrated Mgmt. Endpoint" , subsystems : & [] }] }),
        (7951, Vendor { id : 7951u16 , name : "NebulaMatrix Technology" , devices : & [Device { vendor_id : 7951u16 , id : 4161u16 , name : "D1055AS vDPA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1055AS vDPA Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4162u16 , name : "D1055AS vDPA Storage Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1055AS vDPA Storage Controller" }] } , Device { vendor_id : 7951u16 , id : 4165u16 , name : "D1205CQ vPDA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ vPDA Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4166u16 , name : "D1205CQ vPDA Storage Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ vPDA Storage Controller" }] } , Device { vendor_id : 7951u16 , id : 4167u16 , name : "D1205CQ RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ RDMA Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4640u16 , name : "D1055AS Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 4641u16 , name : "D1055AS Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 4642u16 , name : "D1055AS Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 4643u16 , name : "D1055AS Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 4644u16 , name : "D1205CQ Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4645u16 , name : "D1205CQ Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4646u16 , name : "D1205CQ Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4647u16 , name : "D1205CQ Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 4648u16 , name : "D1205CQ Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "D1205CQ Ethernet Controller" }] } , Device { vendor_id : 7951u16 , id : 5632u16 , name : "M16104 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "N1045XS, Quad-port 10GbE, PCIe 3.0 x8" }] } , Device { vendor_id : 7951u16 , id : 5633u16 , name : "M16104 Family Virtual Function" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "M16104 Family Virtual Function" }] } , Device { vendor_id : 7951u16 , id : 5648u16 , name : "M16140 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "S1405WQ, Single-port 400GbE, PCIe5.0 x16" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "S1406VQ, Dual-port 200GbE, PCIe5.0 x16" }] } , Device { vendor_id : 7951u16 , id : 6144u16 , name : "M16104 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "CXP-6, Quad-port, PCIe 3.0 x4, Fan" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "CXP-12, Quad-port, PCIe 3.0 x8, Fan" }] } , Device { vendor_id : 7951u16 , id : 6145u16 , name : "M16105 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "CXP-31, Dual-port, PCIe 4.0 x8, Fan" }] } , Device { vendor_id : 7951u16 , id : 6656u16 , name : "M16104 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "N1045XT, Quad-port 10GbE, PCIe 3.0 x8" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "N1045XT, Quad-port 10GbE, PCIe 3.0 x8, Fan" }] } , Device { vendor_id : 7951u16 , id : 6657u16 , name : "M16104 Family Virtual Function" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "M16104 Family Virtual Function" }] } , Device { vendor_id : 7951u16 , id : 6658u16 , name : "M16102 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "N1025XT, Dual-port 10GbE, PCIe 3.0 x4, Fan" }] } , Device { vendor_id : 7951u16 , id : 8226u16 , name : "D1055AS PCI Express Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 8227u16 , name : "D1205CQ PCI Express Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13315u16 , name : "M18110 Family" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13316u16 , name : "M18110 Lx Family" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13317u16 , name : "M18110 Family BASE-T" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13318u16 , name : "M18110 Lx Family BASE-T" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13319u16 , name : "M18110 Family OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13320u16 , name : "M18110 Lx Family OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13321u16 , name : "M18110 Family BASE-T OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13322u16 , name : "M18110 Lx Family BASE-T OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13323u16 , name : "M18000 Family" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13324u16 , name : "M18000 Lx Family" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13325u16 , name : "M18000 Family BASE-T" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13326u16 , name : "M18000 Lx Family BASE-T" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13327u16 , name : "M18000 Family OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13328u16 , name : "M18000 Lx Family OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13329u16 , name : "M18000 Family BASE-T OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13330u16 , name : "M18000 Lx Family BASE-T OCP" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13331u16 , name : "M18000 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 13570u16 , name : "M18305 Family" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "S2055AS, 2x 25GbE, SFP28, PCIe 4.0 x8" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "S2025XS, 2x 10GbE, SFP+, PCIe 3.0 x8" } , SubSystem { subvendor : 7951u16 , subdevice : 3u16 , name : "S2045XS,\u{a0} 4x 10GbE, SFP+, PCIe 4.0 x8" }] } , Device { vendor_id : 7951u16 , id : 13572u16 , name : "M18305 Family BASE-T" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "S2025XT, 2x 10GbE, Base-T, PCIe 3.0 x8" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "S2025XT, 2x 10GbE, BASE-T, PCIe 4.0 x4, Fan" } , SubSystem { subvendor : 7951u16 , subdevice : 3u16 , name : "S2045XT, 4x 10GbE, Base-T, PCIe 4.0 x8" } , SubSystem { subvendor : 7951u16 , subdevice : 4u16 , name : "S2045XT, 4x 10GbE, BASE-T, PCIe 4.0 x8, Fan" }] } , Device { vendor_id : 7951u16 , id : 13574u16 , name : "M18305 Family OCP" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "S2055AS, 2x 25GbE, SFP28, PCIe 4.0 x8, OCP 3.0\u{a0}" } , SubSystem { subvendor : 7951u16 , subdevice : 2u16 , name : "S2025AS, 2x 10GbE, SFP+, PCIe 3.0 x8, OCP 3.0\u{a0}" }] } , Device { vendor_id : 7951u16 , id : 13578u16 , name : "M18305 Family Virtual Function" , subsystems : & [SubSystem { subvendor : 7951u16 , subdevice : 1u16 , name : "M18305 Family Virtual Function" }] } , Device { vendor_id : 7951u16 , id : 37000u16 , name : "D1055AS PCI Express Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 7951u16 , id : 37001u16 , name : "D1205CQ PCI Express Switch Downstream Port" , subsystems : & [] }] }),
        (4194, Vendor { id : 4194u16 , name : "Maspar Computer Corp" , devices : & [] }),
        (6449, Vendor { id : 6449u16 , name : "Option N.V." , devices : & [Device { vendor_id : 6449u16 , id : 12u16 , name : "Qualcomm MSM6275 UMTS chip" , subsystems : & [] }] }),
        (4465, Vendor { id : 4465u16 , name : "Loughborough Sound Images Plc" , devices : & [] }),
        (4932, Vendor { id : 4932u16 , name : "Micron Technology Inc" , devices : & [Device { vendor_id : 4932u16 , id : 20816u16 , name : "RealSSD P320h" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20817u16 , name : "RealSSD P320m" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20818u16 , name : "RealSSD P320s" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20819u16 , name : "RealSSD P325m" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20832u16 , name : "RealSSD P420h" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20833u16 , name : "RealSSD P420m" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20835u16 , name : "RealSSD P425m" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20864u16 , name : "9100 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20865u16 , name : "9100 MAX NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20872u16 , name : "7100 ECO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20873u16 , name : "7100 MAX NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20880u16 , name : "9200 ECO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20881u16 , name : "9200 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20882u16 , name : "9200 MAX NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20886u16 , name : "9400 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20887u16 , name : "9400 MAX NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 20898u16 , name : "7300 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4932u16 , subdevice : 8192u16 , name : "960GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 12288u16 , name : "1920GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "3480GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "7680GB U.2" }] } , Device { vendor_id : 4932u16 , id : 20899u16 , name : "7300 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4932u16 , subdevice : 4352u16 , name : "400GB M.2" } , SubSystem { subvendor : 4932u16 , subdevice : 8192u16 , name : "800GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 8448u16 , name : "800GB M.2" } , SubSystem { subvendor : 4932u16 , subdevice : 12288u16 , name : "1600GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "3200GB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "6400GB U.2" }] } , Device { vendor_id : 4932u16 , id : 20913u16 , name : "9300 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "3.84TB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "7.68TB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 24576u16 , name : "15.36TB U.2" }] } , Device { vendor_id : 4932u16 , id : 20914u16 , name : "9300 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "3.2TB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "6.4 TB U.2" } , SubSystem { subvendor : 4932u16 , subdevice : 24576u16 , name : "12.8TB U.2" }] } , Device { vendor_id : 4932u16 , id : 20919u16 , name : "7500 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8935u16 , name : "DC NVMe 7500 U.2 SED RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8936u16 , name : "DC NVMe 7500 U.2 SED RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8937u16 , name : "DC NVMe 7500 U.2 SED RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8938u16 , name : "DC NVMe 7500 U.2 SED RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8939u16 , name : "DC NVMe 7500 U.2 SED RI 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8940u16 , name : "DC NVMe 7500 U.2 ISE RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8941u16 , name : "DC NVMe 7500 U.2 ISE RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8942u16 , name : "DC NVMe 7500 U.2 ISE RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8943u16 , name : "DC NVMe 7500 U.2 ISE RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8944u16 , name : "DC NVMe 7500 U.2 ISE RI 960GB" }] } , Device { vendor_id : 4932u16 , id : 20920u16 , name : "7500 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8945u16 , name : "DC NVMe 7500 U.2 ISE MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8946u16 , name : "DC NVMe 7500 U.2 ISE MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8947u16 , name : "DC NVMe 7500 U.2 ISE MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8948u16 , name : "DC NVMe 7500 U.2 ISE MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8949u16 , name : "DC NVMe 7500 U.2 ISE MU 800GB" }] } , Device { vendor_id : 4932u16 , id : 20921u16 , name : "6500 ION NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8934u16 , name : "Ent NVMe 6500 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8950u16 , name : "Ent NVMe 6500 RI FIPS 30.72TB" }] } , Device { vendor_id : 4932u16 , id : 20923u16 , name : "9550 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8984u16 , name : "MTFDLBQ30T7THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8985u16 , name : "MTFDLBQ15T3THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8986u16 , name : "MTFDLBQ7T6THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8987u16 , name : "MTFDLBQ3T8THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8992u16 , name : "MTFDLAL30T7THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8993u16 , name : "MTFDLAL15T3THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8994u16 , name : "MTFDLAL7T6THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8995u16 , name : "MTFDLAL3T8THA-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9000u16 , name : "MTFDLAL30T7THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9001u16 , name : "MTFDLAL15T3THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9002u16 , name : "MTFDLAL7T6THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9003u16 , name : "MTFDLAL3T8THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9052u16 , name : "MTFDLBQ30T7THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9053u16 , name : "MTFDLBQ15T3THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9054u16 , name : "MTFDLBQ7T6THA-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9055u16 , name : "MTFDLBQ3T8THA-1BK1DFCDA" }] } , Device { vendor_id : 4932u16 , id : 20925u16 , name : "9550 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8988u16 , name : "MTFDLBQ25T6THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8989u16 , name : "MTFDLBQ12T8THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8990u16 , name : "MTFDLBQ6T4THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8991u16 , name : "MTFDLBQ3T2THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8996u16 , name : "MTFDLAL25T6THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8997u16 , name : "MTFDLAL12T8THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8998u16 , name : "MTFDLAL6T4THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 8999u16 , name : "MTFDLAL3T2THB-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9004u16 , name : "MTFDLAL25T6THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9005u16 , name : "MTFDLAL12T8THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9006u16 , name : "MTFDLAL6T4THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9007u16 , name : "MTFDLAL3T2THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9056u16 , name : "MTFDLBQ25T6THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9057u16 , name : "MTFDLBQ12T8THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9058u16 , name : "MTFDLBQ6T4THB-1BK1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9059u16 , name : "MTFDLBQ3T2THB-1BK1DFCDA" }] } , Device { vendor_id : 4932u16 , id : 20928u16 , name : "7400 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8546u16 , name : "EC NVMe OPAL 7400 RI M.2 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8547u16 , name : "EC NVMe OPAL 7400 RI M.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8548u16 , name : "EC NVMe ISE 7400 RI M.2 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8549u16 , name : "EC NVMe ISE 7400 RI M.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8556u16 , name : "EC NVMe ISE 7400 RI M.2 110 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8557u16 , name : "EC NVMe ISE 7400 RI M.2 110 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8558u16 , name : "EC NVMe OPAL 7400 RI M.2 110 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8559u16 , name : "EC NVMe OPAL 7400 RI M.2 110 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8599u16 , name : "EC NVMe OPAL 7400 RI M.2 110 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8781u16 , name : "EC NVMe ISE 7400 RI M.2 110 960GB" } , SubSystem { subvendor : 4932u16 , subdevice : 4352u16 , name : "M.2 480GB" } , SubSystem { subvendor : 4932u16 , subdevice : 8192u16 , name : "U.3 960GB" } , SubSystem { subvendor : 4932u16 , subdevice : 8448u16 , name : "M.2 960GB" } , SubSystem { subvendor : 4932u16 , subdevice : 9728u16 , name : "E1.S 960GB" } , SubSystem { subvendor : 4932u16 , subdevice : 11008u16 , name : "M.2 1920GB" } , SubSystem { subvendor : 4932u16 , subdevice : 11520u16 , name : "E1.2 1920GB" } , SubSystem { subvendor : 4932u16 , subdevice : 12288u16 , name : "U.3 1920GB" } , SubSystem { subvendor : 4932u16 , subdevice : 15872u16 , name : "M.2 3840GB" } , SubSystem { subvendor : 4932u16 , subdevice : 16128u16 , name : "E1.S 3840GB" } , SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "U.3 3840GB" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "U.3 7680GB" }] } , Device { vendor_id : 4932u16 , id : 20929u16 , name : "7400 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8554u16 , name : "EC NVMe ISE 7400 MU M.2 80 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8555u16 , name : "EC NVMe OPAL 7400 MU M.2 80 800GB" } , SubSystem { subvendor : 4932u16 , subdevice : 4352u16 , name : "M.2 400GB" } , SubSystem { subvendor : 4932u16 , subdevice : 8192u16 , name : "U.3 800GB" } , SubSystem { subvendor : 4932u16 , subdevice : 8448u16 , name : "M.2 800GB" } , SubSystem { subvendor : 4932u16 , subdevice : 9728u16 , name : "E1.S 800GB" } , SubSystem { subvendor : 4932u16 , subdevice : 11008u16 , name : "M.2 1600GB" } , SubSystem { subvendor : 4932u16 , subdevice : 11520u16 , name : "E1.S 1600GB" } , SubSystem { subvendor : 4932u16 , subdevice : 12288u16 , name : "U.3 1600GB" } , SubSystem { subvendor : 4932u16 , subdevice : 15872u16 , name : "M.2 3200GB" } , SubSystem { subvendor : 4932u16 , subdevice : 16128u16 , name : "E1.S 3200GB" } , SubSystem { subvendor : 4932u16 , subdevice : 16384u16 , name : "U.3 3200GB" } , SubSystem { subvendor : 4932u16 , subdevice : 20480u16 , name : "U.3 6400GB" }] } , Device { vendor_id : 4932u16 , id : 20931u16 , name : "7450 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8811u16 , name : "EC NVMe FIPS 7450 RI M.2 110 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8812u16 , name : "EC NVMe ISE 7450 RI M.2 80 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8813u16 , name : "EC NVMe ISE 7450 RI M.2 80 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8814u16 , name : "EC NVMe SED 7450 RI M.2 80 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8815u16 , name : "EC NVMe SED 7450 RI M.2 80 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8816u16 , name : "EC NVMe FIPS 7450 RI M.2 80 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8817u16 , name : "EC NVMe FIPS 7450 RI M.2 80 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8819u16 , name : "EC NVMe ISE 7450 RI M.2 110 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8820u16 , name : "EC NVMe ISE 7450 RI M.2 110 1920GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8821u16 , name : "EC NVMe ISE 7450 RI M.2 110 3840GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8824u16 , name : "DC NVMe ISE 7450 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8825u16 , name : "DC NVMe ISE 7450 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8826u16 , name : "DC NVMe ISE 7450 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8827u16 , name : "DC NVMe ISE 7450 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8828u16 , name : "DC NVMe ISE 7450 RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8829u16 , name : "DC NVMe SED 7450 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8830u16 , name : "DC NVMe SED 7450 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8831u16 , name : "DC NVMe SED 7450 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8832u16 , name : "DC NVMe SED 7450 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8833u16 , name : "DC NVMe SED 7450 RI U.2 15.36TB" }] } , Device { vendor_id : 4932u16 , id : 20932u16 , name : "7450 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8818u16 , name : "EC NVMe ISE 7450 MU M.2 80 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8843u16 , name : "DC NVMe SED 7450 MU U.2 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8844u16 , name : "DC NVMe ISE 7450 MU U.2 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8845u16 , name : "DC NVMe SED 7450 MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8846u16 , name : "DC NVMe ISE 7450 MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8847u16 , name : "DC NVMe SED 7450 MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8848u16 , name : "DC NVMe ISE 7450 MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8849u16 , name : "DC NVMe SED 7450 MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8850u16 , name : "DC NVMe ISE 7450 MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8851u16 , name : "DC NVMe SED 7450 MU U.2 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8852u16 , name : "DC NVMe ISE 7450 MU U.2 12.8TB" } , SubSystem { subvendor : 4932u16 , subdevice : 12288u16 , name : "U.3 1600GB [MTFDKCB1T6TFS/MTFDKCC1T6TFS]" }] } , Device { vendor_id : 4932u16 , id : 20937u16 , name : "7600 PRO NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9249u16 , name : "MTFDLBQ1T9THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9250u16 , name : "MTFDLBQ3T8THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9251u16 , name : "MTFDLBQ7T6THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9252u16 , name : "MTFDLBQ15T3THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9257u16 , name : "MTFDLBQ1T9THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9258u16 , name : "MTFDLBQ3T8THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9259u16 , name : "MTFDLBQ7T6THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9260u16 , name : "MTFDLBQ15T3THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9265u16 , name : "MTFDLAL1T9THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9266u16 , name : "MTFDLAL3T8THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9267u16 , name : "MTFDLAL7T6THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9268u16 , name : "MTFDLAL15T3THG-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9273u16 , name : "MTFDLAL1T9THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9274u16 , name : "MTFDLAL3T8THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9275u16 , name : "MTFDLAL7T6THG-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9276u16 , name : "MTFDLAL15T3THG-1BP1JABDA" }] } , Device { vendor_id : 4932u16 , id : 20938u16 , name : "7600 MAX NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9253u16 , name : "MTFDLBQ1T6THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9254u16 , name : "MTFDLBQ3T2THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9255u16 , name : "MTFDLBQ6T4THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9256u16 , name : "MTFDLBQ12T8THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9261u16 , name : "MTFDLBQ1T6THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9262u16 , name : "MTFDLBQ3T2THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9263u16 , name : "MTFDLBQ6T4THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9264u16 , name : "MTFDLBQ12T8THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9269u16 , name : "MTFDLAL1T6THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9270u16 , name : "MTFDLAL3T2THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9271u16 , name : "MTFDLAL6T4THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9272u16 , name : "MTFDLAL12T8THS-1BP1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9277u16 , name : "MTFDLAL1T6THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9278u16 , name : "MTFDLAL3T2THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9279u16 , name : "MTFDLAL6T4THS-1BP1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9280u16 , name : "MTFDLAL12T8THS-1BP1JABDA" }] } , Device { vendor_id : 4932u16 , id : 20939u16 , name : "6550 ION NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9081u16 , name : "MTFDLBQ61T4THL-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9126u16 , name : "MTFDLBQ30T7THL-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9127u16 , name : "MTFDLAL61T4THL-1BK1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9128u16 , name : "MTFDLAL30T7THL-1BK1JABDA" }] } , Device { vendor_id : 4932u16 , id : 20940u16 , name : "6600 ION NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9299u16 , name : "MTFDLBQ122T8QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9347u16 , name : "MTFDLBQ61T4QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9348u16 , name : "MTFDLBQ30T7QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9349u16 , name : "MTFDLBQ122T8QHF-1BQ1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9350u16 , name : "MTFDLBQ61T4QHF-1BQ1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9351u16 , name : "MTFDLBQ30T7QHF-1BQ1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9353u16 , name : "MTFDLAL122T8QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9354u16 , name : "MTFDLAL61T4QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9355u16 , name : "MTFDLAL30T7QHF-1BQ1JABDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9357u16 , name : "MTFDLAL122T8QHF-1BQ1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9358u16 , name : "MTFDLAL61T4QHF-1BQ1DFCDA" } , SubSystem { subvendor : 4136u16 , subdevice : 9359u16 , name : "MTFDLAL30T7QHF-1BQ1DFCDA" }] } , Device { vendor_id : 4932u16 , id : 20941u16 , name : "9650 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21508u16 , name : "2210 NVMe SSD [Cobain]" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21509u16 , name : "2300 NVMe SSD [Santana]" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21511u16 , name : "3400 NVMe SSD [Hendrix]" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21520u16 , name : "2200S NVMe SSD [Cassandra]" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21521u16 , name : "2450 NVMe SSD [HendrixV] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21523u16 , name : "2400 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21524u16 , name : "3460 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21525u16 , name : "3500 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21526u16 , name : "2550 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21541u16 , name : "2500 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21543u16 , name : "2650 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21544u16 , name : "4600 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 21545u16 , name : "2600 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 24577u16 , name : "2100AI NVMe SSD [Nitro]" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 24579u16 , name : "4150AT NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4932u16 , id : 24580u16 , name : "4100AT NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (4971, Vendor { id : 4971u16 , name : "Kawasaki Steel Corporation" , devices : & [Device { vendor_id : 4971u16 , id : 65281u16 , name : "KL5A72002 Motion JPEG" , subsystems : & [] }] }),
        (51714, Vendor { id : 51714u16 , name : "I-TEK OptoElectronics Co., LTD." , devices : & [Device { vendor_id : 51714u16 , id : 531u16 , name : "Vulcan-CXP Frame Grabber" , subsystems : & [] }] }),
        (5052, Vendor { id : 5052u16 , name : "Bitran Corporation" , devices : & [] }),
        (5509, Vendor { id : 5509u16 , name : "Logitron" , devices : & [] }),
        (5313, Vendor { id : 5313u16 , name : "MYRICOM Inc." , devices : & [Device { vendor_id : 5313u16 , id : 8u16 , name : "Myri-10G Dual-Protocol NIC" , subsystems : & [SubSystem { subvendor : 5313u16 , subdevice : 8u16 , name : "10G-PCIE-8A" } , SubSystem { subvendor : 5313u16 , subdevice : 9u16 , name : "10G-PCIE-8A (MSI-X firmware)" } , SubSystem { subvendor : 5313u16 , subdevice : 10u16 , name : "10G-PCIE-8B" }] } , Device { vendor_id : 5313u16 , id : 32835u16 , name : "Myrinet 2000 Scalable Cluster Interconnect" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4672u16 , name : "Myrinet M2L-PCI64/2-3.0 LANai 7.4 (HP OEM)" }] }] }),
        (6148, Vendor { id : 6148u16 , name : "Ralink corp. (wrong ID)" , devices : & [Device { vendor_id : 6148u16 , id : 12384u16 , name : "RT3060 Wireless 802.11n 1T/1R" , subsystems : & [] }] }),
        (4280, Vendor { id : 4280u16 , name : "Standard Microsystems Corp [SMC]" , devices : & [Device { vendor_id : 4280u16 , id : 5u16 , name : "83c170 EPIC/100 Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4181u16 , subdevice : 57344u16 , name : "LANEPIC 10/100 [EVB171Q-PCI]" } , SubSystem { subvendor : 4181u16 , subdevice : 57346u16 , name : "LANEPIC 10/100 [EVB171G-PCI]" } , SubSystem { subvendor : 4280u16 , subdevice : 40977u16 , name : "EtherPower II 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 40980u16 , name : "EtherPower II 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 40981u16 , name : "EtherPower II 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 40982u16 , name : "EtherPower II 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 40983u16 , name : "EtherPower II 10/100" }] } , Device { vendor_id : 4280u16 , id : 6u16 , name : "83c175 EPIC/100 Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4181u16 , subdevice : 57600u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 4181u16 , subdevice : 57602u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 4181u16 , subdevice : 58112u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 4181u16 , subdevice : 58114u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 4280u16 , subdevice : 40978u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 5026u16 , subdevice : 32770u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 5026u16 , subdevice : 32774u16 , name : "LANEPIC Cardbus Fast Ethernet Adapter" }] } , Device { vendor_id : 4280u16 , id : 4096u16 , name : "FDC 37c665" , subsystems : & [] } , Device { vendor_id : 4280u16 , id : 4097u16 , name : "FDC 37C922" , subsystems : & [] } , Device { vendor_id : 4280u16 , id : 40977u16 , name : "83C170QF" , subsystems : & [] } , Device { vendor_id : 4280u16 , id : 45318u16 , name : "SMC34C90" , subsystems : & [] }] }),
        (4230, Vendor { id : 4230u16 , name : "J. Bond Computer Systems" , devices : & [] }),
        (7753, Vendor { id : 7753u16 , name : "Yangtze Memory Technologies Co.,Ltd" , devices : & [Device { vendor_id : 7753u16 , id : 1u16 , name : "ZHITAI PC005 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 33u16 , name : "ZHITAI TiPro5000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 65u16 , name : "ZHITAI TiPro7000" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 113u16 , name : "ZHITAI TiPlus7100" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 129u16 , name : "ZHITAI Ti600 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 144u16 , name : "ZHITAI TiPro9000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 161u16 , name : "ZHITAI TiPlus7100s NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4097u16 , name : "PC005 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4113u16 , name : "PC210 M.2 2280 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4115u16 , name : "PC210 M.2 2242 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4145u16 , name : "PC300 M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4147u16 , name : "PC300 M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4209u16 , name : "PC411 M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4211u16 , name : "PC411 M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4216u16 , name : "PE321 U.2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4225u16 , name : "PC41Q M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4227u16 , name : "PC41Q M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4353u16 , name : "PC450 M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4355u16 , name : "PC450 M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4369u16 , name : "PC42Q M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4371u16 , name : "PC42Q M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4385u16 , name : "PC550 M.2 2280 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 4387u16 , name : "PC550 M.2 2242 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7753u16 , id : 6696u16 , name : "PE511 U.2 NVMe SSD" , subsystems : & [] }] }),
        (8923, Vendor { id : 8923u16 , name : "Missing Link Electronics, Inc." , devices : & [Device { vendor_id : 8923u16 , id : 4608u16 , name : "NVMe Streamer EP ERD" , subsystems : & [] }] }),
        (4519, Vendor { id : 4519u16 , name : "Power Computing Corp." , devices : & [] }),
        (4273, Vendor { id : 4273u16 , name : "Cabletron Systems Inc" , devices : & [] }),
        (60091, Vendor { id : 60091u16 , name : "Aashima Technology B.V." , devices : & [] }),
        (5012, Vendor { id : 5012u16 , name : "Level One Communications" , devices : & [Device { vendor_id : 5012u16 , id : 1u16 , name : "LXT1001 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 18432u16 , name : "DGE-500SX" } , SubSystem { subvendor : 5012u16 , subdevice : 1u16 , name : "NetCelerator Adapter" }] }] }),
        (5240, Vendor { id : 5240u16 , name : "DIATREND Corporation" , devices : & [] }),
        (7860, Vendor { id : 7860u16 , name : "Quantum Nebula Microelectronics Technology Co.,Ltd." , devices : & [Device { vendor_id : 7860u16 , id : 13313u16 , name : "SSD Contoller" , subsystems : & [] }] }),
        (5374, Vendor { id : 5374u16 , name : "ARCHTEK TELECOM Corp" , devices : & [] }),
        (5890, Vendor { id : 5890u16 , name : "Internet Machines Corporation (IMC)" , devices : & [] }),
        (5613, Vendor { id : 5613u16 , name : "Macrolink Inc" , devices : & [] }),
        (5497, Vendor { id : 5497u16 , name : "Dual Technology Corp" , devices : & [] }),
        (4768, Vendor { id : 4768u16 , name : "Allen-Bradley Company" , devices : & [] }),
        (7033, Vendor { id : 7033u16 , name : "Absolute Analysis" , devices : & [] }),
        (4643, Vendor { id : 4643u16 , name : "Artesyn Communication Products" , devices : & [Device { vendor_id : 4643u16 , id : 3u16 , name : "PM/Link" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 4u16 , name : "PM/T1" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 5u16 , name : "PM/E1" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 8u16 , name : "PM/SLS" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 9u16 , name : "BajaSpan Resource Target" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 10u16 , name : "BajaSpan Section 0" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 11u16 , name : "BajaSpan Section 1" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 12u16 , name : "BajaSpan Section 2" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 13u16 , name : "BajaSpan Section 3" , subsystems : & [] } , Device { vendor_id : 4643u16 , id : 14u16 , name : "PM/PPC" , subsystems : & [] }] }),
        (7523, Vendor { id : 7523u16 , name : "StorArt Technology Co., Ltd" , devices : & [Device { vendor_id : 7523u16 , id : 14337u16 , name : "SA3801 PCIe 3.0 x2 NVMe controller" , subsystems : & [] }] }),
        (6025, Vendor { id : 6025u16 , name : "Ennyah Technologies Corp." , devices : & [] }),
        (4874, Vendor { id : 4874u16 , name : "Mitsubishi Electric Microcomputer" , devices : & [] }),
        (16962, Vendor { id : 16962u16 , name : "Universall Answer Generators" , devices : & [] }),
        (8044, Vendor { id : 8044u16 , name : "CIX Technology Group Co., Ltd." , devices : & [Device { vendor_id : 8044u16 , id : 1u16 , name : "CIX P1 CD8180 PCI Express Root Port" , subsystems : & [] }] }),
        (4673, Vendor { id : 4673u16 , name : "DSC Communications" , devices : & [] }),
        (4443, Vendor { id : 4443u16 , name : "Parallax Graphics" , devices : & [] }),
        (4828, Vendor { id : 4828u16 , name : "Symicron Computer Communication Ltd." , devices : & [] }),
        (4464, Vendor { id : 4464u16 , name : "Inventec Corporation" , devices : & [] }),
        (48573, Vendor { id : 48573u16 , name : "Blackmagic Design" , devices : & [Device { vendor_id : 48573u16 , id : 41222u16 , name : "Multibridge Extreme" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41239u16 , name : "Intensity Pro" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41242u16 , name : "DeckLink HD Extreme 2" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41243u16 , name : "DeckLink SDI/Duo/Quad" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41244u16 , name : "DeckLink HD Extreme 3" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41245u16 , name : "DeckLink Studio" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41246u16 , name : "DeckLink Optical Fibre" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41248u16 , name : "Decklink Studio 2" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41249u16 , name : "DeckLink HD Extreme 3D/3D+" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41252u16 , name : "Intensity Extreme" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41254u16 , name : "Intensity Shuttle" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41255u16 , name : "UltraStudio Express" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41257u16 , name : "UltraStudio Mini Recorder" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41258u16 , name : "UltraStudio Mini Monitor" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41261u16 , name : "UltraStudio 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41262u16 , name : "DeckLink 4K Extreme" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41263u16 , name : "DeckLink Mini Monitor" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41264u16 , name : "DeckLink Mini Recorder" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41266u16 , name : "UltraStudio 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41270u16 , name : "DeckLink 4K Extreme 12G" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41271u16 , name : "DeckLink Studio 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41272u16 , name : "Decklink SDI 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41273u16 , name : "Intensity Pro 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41275u16 , name : "DeckLink Micro Recorder" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41277u16 , name : "DeckLink 4K Pro" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41278u16 , name : "UltraStudio 4K Extreme" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41279u16 , name : "DeckLink Quad 2" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41280u16 , name : "DeckLink Duo 2" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41281u16 , name : "UltraStudio 4K Extreme 3" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41282u16 , name : "UltraStudio HD Mini" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41283u16 , name : "DeckLink Mini Recorder 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41284u16 , name : "DeckLink Mini Monitor 4K" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41288u16 , name : "DeckLink SDI Micro" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41291u16 , name : "DeckLink 8K Pro" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41294u16 , name : "DeckLink Quad HDMI Recorder" , subsystems : & [] } , Device { vendor_id : 48573u16 , id : 41471u16 , name : "eGPU RX580" , subsystems : & [] }] }),
        (5084, Vendor { id : 5084u16 , name : "Netboost Corporation" , devices : & [] }),
        (4627, Vendor { id : 4627u16 , name : "Applied Intelligent Systems, Inc." , devices : & [] }),
        (4283, Vendor { id : 4283u16 , name : "Dapha Electronics Corporation" , devices : & [] }),
        (5593, Vendor { id : 5593u16 , name : "Super Micro Computer Inc" , devices : & [] }),
        (57007, Vendor { id : 57007u16 , name : "Middle Digital Inc." , devices : & [Device { vendor_id : 57007u16 , id : 36944u16 , name : "PC Weasel Virtual VGA" , subsystems : & [] } , Device { vendor_id : 57007u16 , id : 36945u16 , name : "PC Weasel Serial Port" , subsystems : & [] } , Device { vendor_id : 57007u16 , id : 36946u16 , name : "PC Weasel Watchdog Timer" , subsystems : & [] }] }),
        (5755, Vendor { id : 5755u16 , name : "ZyDAS Technology Corp." , devices : & [Device { vendor_id : 5755u16 , id : 8450u16 , name : "ZyDAS ZD1202" , subsystems : & [SubSystem { subvendor : 6270u16 , subdevice : 13318u16 , name : "ZyAIR B-122 CardBus 11Mbs Wireless LAN Card" } , SubSystem { subvendor : 6270u16 , subdevice : 13319u16 , name : "ZyAIR B-320 802.11b Wireless PCI Adapter" }] } , Device { vendor_id : 5755u16 , id : 8470u16 , name : "ZD1212B Wireless Adapter" , subsystems : & [] }] }),
        (7662, Vendor { id : 7662u16 , name : "Biwin Storage Technology Co., Ltd." , devices : & [Device { vendor_id : 7662u16 , id : 4610u16 , name : "HP EX900 Plus NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 5634u16 , name : "HP FX700 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 8802u16 , name : "HP EX950 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 8803u16 , name : "HP EX900 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 9472u16 , name : "XCG 2500 Enterprise Gen 4 SSD" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 9858u16 , name : "AP923P Gen4 NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 16673u16 , name : "PCIe 4.0 SP406/416 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7662u16 , subdevice : 0u16 , name : "NVMe SSD SP416 800G 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 1u16 , name : "NVMe SSD SP416 1.6T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 2u16 , name : "NVMe SSD SP416 3.2T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 3u16 , name : "NVMe SSD SP416 6.4T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 16u16 , name : "NVMe SSD SP406 960G 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 17u16 , name : "NVMe SSD SP406 1.92T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 18u16 , name : "NVMe SSD SP406 3.84T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 19u16 , name : "NVMe SSD SP406 7.68T 2.5\" U.2" }] } , Device { vendor_id : 7662u16 , id : 20502u16 , name : "BIWIN NVMe SSD SP509/SP519" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 20833u16 , name : "BIWIN NVMe SSD SP506/SP516" , subsystems : & [SubSystem { subvendor : 7662u16 , subdevice : 1u16 , name : "NVMe SSD SP516 1.6T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 2u16 , name : "NVMe SSD SP516 3.2T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 3u16 , name : "NVMe SSD SP516 6.4T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 4u16 , name : "NVMe SSD SP516 12.8T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 17u16 , name : "NVMe SSD SP506 1.92T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 18u16 , name : "NVMe SSD SP506 3.84T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 19u16 , name : "NVMe SSD SP506 7.68T 2.5\" U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 20u16 , name : "NVMe SSD SP506 15.36T 2.5\" U.2" }] } , Device { vendor_id : 7662u16 , id : 21014u16 , name : "KingSpec NX series NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 21046u16 , name : "Acer Predator GM7000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7662u16 , id : 30464u16 , name : "BIWIN NVMe SSD SP5 series" , subsystems : & [SubSystem { subvendor : 7662u16 , subdevice : 1u16 , name : "NVMe SSD SP51Y 3.2TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 2u16 , name : "NVMe SSD SP51Y 6.4TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 3u16 , name : "NVMe SSD SP51Y 12.8TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 17u16 , name : "NVMe SSD SP50Y 3.84TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 18u16 , name : "NVMe SSD SP50Y 7.68TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 19u16 , name : "NVMe SSD SP50Y 15.36B 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 257u16 , name : "NVMe SSD SP518 3.2TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 258u16 , name : "NVMe SSD SP518 6.4TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 259u16 , name : "NVMe SSD SP518 12.8TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 273u16 , name : "NVMe SSD SP508 3.84TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 274u16 , name : "NVMe SSD SP508 7.68TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 275u16 , name : "NVMe SSD SP508 15.36B 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 513u16 , name : "NVMe SSD SP5000 30.72TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 514u16 , name : "NVMe SSD SP5000 61.44TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 4609u16 , name : "NVMe SSD SP5001 30.72TB 2.5\"U.2" } , SubSystem { subvendor : 7662u16 , subdevice : 4610u16 , name : "NVMe SSD SP5001 61.44TB 2.5\"U.2" }] }] }),
        (5457, Vendor { id : 5457u16 , name : "SMART Electronic DEVELOPMENT GmBH" , devices : & [] }),
        (5554, Vendor { id : 5554u16 , name : "Mosaid Technologies Inc" , devices : & [] }),
        (8502, Vendor { id : 8502u16 , name : "Dongguan Xincun Chengbang Technology Co., Ltd." , devices : & [Device { vendor_id : 8502u16 , id : 26112u16 , name : "NX6600U0 NVME SSD" , subsystems : & [] } , Device { vendor_id : 8502u16 , id : 29952u16 , name : "NX7500U0 NVME SSD" , subsystems : & [] } , Device { vendor_id : 8502u16 , id : 30208u16 , name : "NX7600U0 NVME SSD" , subsystems : & [] }] }),
        (4888, Vendor { id : 4888u16 , name : "Packet Engines Inc." , devices : & [Device { vendor_id : 4888u16 , id : 2321u16 , name : "GNIC-II PCI Gigabit Ethernet [Hamachi]" , subsystems : & [] }] }),
        (5263, Vendor { id : 5263u16 , name : "Plant Equipment, Inc." , devices : & [] }),
        (5325, Vendor { id : 5325u16 , name : "Universal Global Scientific Industrial Co.,Ltd" , devices : & [Device { vendor_id : 5325u16 , id : 1u16 , name : "USI-1514-1GbaseT [OCP1]" , subsystems : & [] } , Device { vendor_id : 5325u16 , id : 2u16 , name : "USI-4227-SFP [OCP2]" , subsystems : & [] } , Device { vendor_id : 5325u16 , id : 3u16 , name : "USI-X557-10GbaseT [OCP3]" , subsystems : & [] }] }),
        (7993, Vendor { id : 7993u16 , name : "Silicon Innovation Microelectronics Co. Ltd." , devices : & [Device { vendor_id : 7993u16 , id : 8704u16 , name : "XSAT2204 SATA 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 8705u16 , name : "XSAT2104 SATA 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 8707u16 , name : "XSAT2104A SATA 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 8708u16 , name : "XSAT2102A SATA 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 8960u16 , name : "XUSB2104 USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 8961u16 , name : "XUSB2102 USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10240u16 , name : "XPSW3106 6-Lane 5-Port PCIe 3.0 Packet Switch" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10241u16 , name : "XPSW3112 12-Lane 9-Port PCIe 3.0 Packet Switch" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10242u16 , name : "XPSW3116 16-Lane 9-Port PCIe 3.0 Packet Switch" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10243u16 , name : "XPSW2106 6-Lane 5-Port PCIe 2.0 Packet Switch" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10244u16 , name : "XPSW2112 12-Lane 9-Port PCIe 2.0 Packet Switch" , subsystems : & [] } , Device { vendor_id : 7993u16 , id : 10245u16 , name : "XPSW2116 16-Lane 9-Port PCIe 2.0 Packet Switch" , subsystems : & [] }] }),
        (4284, Vendor { id : 4284u16 , name : "Advanced Logic Research" , devices : & [] }),
        (20805, Vendor { id : 20805u16 , name : "Ensoniq (Old)" , devices : & [Device { vendor_id : 20805u16 , id : 12337u16 , name : "Concert AudioPCI" , subsystems : & [] }] }),
        (4760, Vendor { id : 4760u16 , name : "Spellcaster Telecommunications Inc." , devices : & [] }),
        (4678, Vendor { id : 4678u16 , name : "Dipix Technologies, Inc." , devices : & [] }),
        (4234, Vendor { id : 4234u16 , name : "SBS Technologies" , devices : & [Device { vendor_id : 4234u16 , id : 1u16 , name : "VME Bridge Model 617" , subsystems : & [] } , Device { vendor_id : 4234u16 , id : 16u16 , name : "VME Bridge Model 618" , subsystems : & [] } , Device { vendor_id : 4234u16 , id : 64u16 , name : "dataBLIZZARD" , subsystems : & [] } , Device { vendor_id : 4234u16 , id : 12288u16 , name : "VME Bridge Model 2706" , subsystems : & [] }] }),
        (4344, Vendor { id : 4344u16 , name : "Altos India Ltd" , devices : & [] }),
        (5366, Vendor { id : 5366u16 , name : "COYOTE Technologies LLC" , devices : & [] }),
        (5320, Vendor { id : 5320u16 , name : "Turbocomm Tech. Inc." , devices : & [] }),
        (7985, Vendor { id : 7985u16 , name : "Nextorage" , devices : & [Device { vendor_id : 7985u16 , id : 17682u16 , name : "NE1N NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7985u16 , id : 17691u16 , name : "NN4LE NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7985u16 , id : 17954u16 , name : "NEM-PAC NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5305, Vendor { id : 5305u16 , name : "Cisco Aironet Wireless Communications" , devices : & [Device { vendor_id : 5305u16 , id : 1u16 , name : "PC4800" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 832u16 , name : "PC4800" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 848u16 , name : "350 series 802.11b Wireless LAN Adapter" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 17664u16 , name : "PC4500" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 18432u16 , name : "Cisco Aironet 340 802.11b Wireless LAN Adapter/Aironet PC4800" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 42244u16 , name : "Cisco Aironet Wireless 802.11b" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 42245u16 , name : "Cisco Aironet CB20a 802.11a Wireless LAN Adapter" , subsystems : & [] } , Device { vendor_id : 5305u16 , id : 42246u16 , name : "Cisco Aironet Mini PCI b/g" , subsystems : & [] }] }),
        (4493, Vendor { id : 4493u16 , name : "BitFlow Inc" , devices : & [Device { vendor_id : 4493u16 , id : 1u16 , name : "Raptor-PCI framegrabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 18u16 , name : "Model 12 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 20u16 , name : "Model 14 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 36u16 , name : "Model 24 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 68u16 , name : "Model 44 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 274u16 , name : "Model 12 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 276u16 , name : "Model 14 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 292u16 , name : "Model 24 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 324u16 , name : "Model 44 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 530u16 , name : "Model 12 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 532u16 , name : "Model 14 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 548u16 , name : "Model 24 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 580u16 , name : "Model 44 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 786u16 , name : "Model 12 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 788u16 , name : "Model 14 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 804u16 , name : "Model 24 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 836u16 , name : "Model 44 Road Runner Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4493u16 , id : 45134u16 , name : "Claxon CXP4 CoaXPress frame grabber" , subsystems : & [] }] }),
        (5419, Vendor { id : 5419u16 , name : "FLYTECH Technology CO Ltd" , devices : & [] }),
        (5126, Vendor { id : 5126u16 , name : "Oce' Printing Systems GmbH" , devices : & [] }),
        (5086, Vendor { id : 5086u16 , name : "ABB Robotics Products AB" , devices : & [] }),
        (5339, Vendor { id : 5339u16 , name : "AFAVLAB Technology Inc" , devices : & [Device { vendor_id : 5339u16 , id : 8480u16 , name : "TK9902" , subsystems : & [] } , Device { vendor_id : 5339u16 , id : 8578u16 , name : "AFAVLAB Technology Inc. 8-port serial card" , subsystems : & [] }] }),
        (5615, Vendor { id : 5615u16 , name : "Intelligent Paradigm Inc" , devices : & [] }),
        (7587, Vendor { id : 7587u16 , name : "Habana Labs Ltd." , devices : & [Device { vendor_id : 7587u16 , id : 1u16 , name : "HL-1000 AI Inference Accelerator [Goya]" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 48u16 , name : "Greco AI Inference Accelerator" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 4096u16 , name : "HL-2000 AI Training Accelerator [Gaudi]" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 4112u16 , name : "HL-2000 AI Training Accelerator [Gaudi secured]" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 4128u16 , name : "Gaudi2 AI Training Accelerator" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 4192u16 , name : "Gaudi3 AI Training Accelerator" , subsystems : & [] } , Device { vendor_id : 7587u16 , id : 4195u16 , name : "Gaudi3 AI Training Accelerator Add-In Family" , subsystems : & [SubSystem { subvendor : 7587u16 , subdevice : 4195u16 , name : "HL-338 Gaudi3 AI Accelerator PCIe Add-In Card" }] }] }),
        (5559, Vendor { id : 5559u16 , name : "Sandisk Corp" , devices : & [Device { vendor_id : 5559u16 , id : 8193u16 , name : "Skyhawk Series NVME SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20481u16 , name : "WD Black NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20482u16 , name : "SanDisk Extreme Pro / WD Black 2018/SN750/PC SN720 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20483u16 , name : "WD Blue SN500 / PC SN520 x2 M.2 2280 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20484u16 , name : "PC SN520 x2 M.2 2230 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20485u16 , name : "PC SN520 x2 M.2 2242 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20486u16 , name : "SanDisk Extreme Pro / WD Black SN750 / PC SN730 / Red SN700 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20487u16 , name : "IX SN530 NVMe SSD / microSD Express Card (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20488u16 , name : "PC SN530 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20489u16 , name : "SanDisk Ultra 3D / WD PC SN530, IX SN530, Blue SN550 NVMe SSD (DRAM-less)" , subsystems : & [SubSystem { subvendor : 5559u16 , subdevice : 20489u16 , name : "WD Blue SN550 NVMe SSD" }] } , Device { vendor_id : 5559u16 , id : 20491u16 , name : "PC SN530 NVMe SSD" , subsystems : & [SubSystem { subvendor : 5140u16 , subdevice : 20491u16 , name : "Xbox Series X" }] } , Device { vendor_id : 5559u16 , id : 20493u16 , name : "WD Ultrastar DC SN340 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20497u16 , name : "WD PC SN810 / Black SN850 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20500u16 , name : "WD PC SN540 / Green SN350 NVMe SSD 1 TB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20501u16 , name : "PC SN740 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20502u16 , name : "WD PC SN740 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20503u16 , name : "WD SN560/SN740/SN770/SN5000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20505u16 , name : "WD Green SN350 240GB (DRAM-less) / SN560E NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20506u16 , name : "SanDisk Ultra 3D / WD Blue SN570 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20509u16 , name : "WD Blue SN550 NVMe SSD 2TB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20510u16 , name : "PC SN735 / WD_BLACK SN750 SE NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20511u16 , name : "WD PC SN735 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20517u16 , name : "WD Blue SN570 NVMe SSD 2TB" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20518u16 , name : "WD PC SN735 NVMe SSD 1TB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20520u16 , name : "WD CH SN560 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20528u16 , name : "WD Black SN850X NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20532u16 , name : "WD PC SN5000S M.2 2230 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20533u16 , name : "WD PC SN5000S M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20534u16 , name : "WD PC SN5000S M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20545u16 , name : "WD Blue SN580 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20546u16 , name : "WD Black SN770M NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20547u16 , name : "WD PC SN7100S M.2 2230 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20548u16 , name : "WD PC SN7100S M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20549u16 , name : "WD_BLACK SN7100/WD PC SN7100S M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20550u16 , name : "SanDisk Extreme NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20553u16 , name : "SN8000S NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20554u16 , name : "WD Blue SN5000 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20560u16 , name : "WD PC SN8050S / WD_BLACK SN8100 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20577u16 , name : "PC SN5100S M.2 2230 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20578u16 , name : "PC SN5100S M.2 2242 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5559u16 , id : 20579u16 , name : "PC SN5100S / WD Blue SN5100 M.2 2280 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5227, Vendor { id : 5227u16 , name : "Parascan Technologies Ltd" , devices : & [] }),
        (8160, Vendor { id : 8160u16 , name : "Allwinmeta Co., Ltd." , devices : & [Device { vendor_id : 8160u16 , id : 4112u16 , name : "AWM 1" , subsystems : & [] } , Device { vendor_id : 8160u16 , id : 8192u16 , name : "AWM 2" , subsystems : & [] } , Device { vendor_id : 8160u16 , id : 8208u16 , name : "AWM 2-M" , subsystems : & [] }] }),
        (5383, Vendor { id : 5383u16 , name : "Motorola ?? / HTEC" , devices : & [Device { vendor_id : 5383u16 , id : 1u16 , name : "MPC105 [Eagle]" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 2u16 , name : "MPC106 [Grackle]" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 3u16 , name : "MPC8240 [Kahlua]" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 256u16 , name : "MC145575 [HFC-PCI]" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 1073u16 , name : "KTI829c 100VG" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 18433u16 , name : "Raven" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 18434u16 , name : "Falcon" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 18435u16 , name : "Hawk" , subsystems : & [] } , Device { vendor_id : 5383u16 , id : 18438u16 , name : "CPX8216" , subsystems : & [] }] }),
        (5542, Vendor { id : 5542u16 , name : "Sunlight Ultrasound Technologies Ltd" , devices : & [] }),
        (5348, Vendor { id : 5348u16 , name : "Broadcom Inc. and subsidiaries" , devices : & [Device { vendor_id : 5348u16 , id : 1398u16 , name : "BCM43224 802.11a/b/g/n" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2048u16 , name : "Sentry5 Chipcommon I/O Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2052u16 , name : "Sentry5 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2053u16 , name : "Sentry5 MIPS32 CPU" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2054u16 , name : "Sentry5 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2059u16 , name : "Sentry5 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2063u16 , name : "Sentry5 DDR/SDR RAM Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2065u16 , name : "Sentry5 External Interface Core" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 2070u16 , name : "BCM3302 Sentry5 MIPS32 CPU" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5488u16 , name : "720p FaceTime HD Camera" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5632u16 , name : "NetXtreme BCM5752 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 449u16 , name : "Precision 490" } , SubSystem { subvendor : 4136u16 , subdevice : 450u16 , name : "Latitude D620" } , SubSystem { subvendor : 4156u16 , subdevice : 12309u16 , name : "PCIe LAN on Motherboard" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500 Onboard" } , SubSystem { subvendor : 4697u16 , subdevice : 9989u16 , name : "AT-2711FX" }] } , Device { vendor_id : 5348u16 , id : 5633u16 , name : "NetXtreme BCM5752M Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5636u16 , name : "BCM5745X NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5637u16 , name : "BCM5745X NetXtreme-E RDMA Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5638u16 , name : "BCM5745X NetXtreme-E RDMA Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5641u16 , name : "BCM5745X NetXtreme-E Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5650u16 , name : "BCM70012 Video Decoder [Crystal HD]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5652u16 , name : "BCM57454 NetXtreme-E 10Gb/25Gb/40Gb/50Gb/100Gb Ethernet" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 53568u16 , name : "NetXtreme-E Quad-port 10GBASE-T Ethernet PCIe Adapter (BCM957454-P410SDBT)" } , SubSystem { subvendor : 5348u16 , subdevice : 54080u16 , name : "NetXtreme-E Quad-port 10GBASE-T Ethernet OCP 3.0 Adapter (BCM957454-N410SDBT)" }] } , Device { vendor_id : 5348u16 , id : 5653u16 , name : "BCM70015 Video Decoder [Crystal HD]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5689u16 , name : "NetXtreme II BCM5709 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 BCM5709 Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 BCM5709 Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 BCM5709 Gigabit Ethernet" } , SubSystem { subvendor : 4156u16 , subdevice : 28757u16 , name : "NC382i Integrated Multi-port PCI Express Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28761u16 , name : "NC382T PCIe Dual Port Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4265u16 , subdevice : 32807u16 , name : "Quad port Gigabit Ethernet Controller" }] } , Device { vendor_id : 5348u16 , id : 5690u16 , name : "NetXtreme II BCM5709S Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 635u16 , name : "PowerEdge M805 Broadcom NetXtreme II BCM5709S" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 BCM5709S Gigabit Ethernet" } , SubSystem { subvendor : 4156u16 , subdevice : 5917u16 , name : "NC382m Dual Port 1GbE Multifunction BL-c Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28758u16 , name : "NC382i Integrated Quad Port PCI Express Gigabit Server Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10628u16 , name : "AT-2973SX" }] } , Device { vendor_id : 5348u16 , id : 5691u16 , name : "NetXtreme II BCM5716 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 BCM5716 Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 BCM5716 Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 753u16 , name : "PowerEdge R510 BCM5716 Gigabit Ethernet" }] } , Device { vendor_id : 5348u16 , id : 5692u16 , name : "NetXtreme II BCM5716S Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5693u16 , name : "NetXtreme II BCM57811 10-Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34186u16 , name : "PEB-10G/57811-1S" }] } , Device { vendor_id : 5348u16 , id : 5694u16 , name : "NetXtreme II BCM57811 10 Gigabit Ethernet Multi Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5695u16 , name : "NetXtreme II BCM57811 10-Gigabit Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5697u16 , name : "NetXtreme BCM57787 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5698u16 , name : "NetXtreme BCM57764 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5699u16 , name : "NetXtreme BCM5725 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5700u16 , name : "NetXtreme BCM5700 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 631u16 , name : "Broadcom Vigil B5700 1000Base-T" } , SubSystem { subvendor : 4136u16 , subdevice : 209u16 , name : "Broadcom BCM5700" } , SubSystem { subvendor : 4136u16 , subdevice : 262u16 , name : "Broadcom BCM5700" } , SubSystem { subvendor : 4136u16 , subdevice : 265u16 , name : "Broadcom BCM5700 1000Base-T" } , SubSystem { subvendor : 4136u16 , subdevice : 266u16 , name : "Broadcom BCM5700 1000BaseTX" } , SubSystem { subvendor : 4279u16 , subdevice : 4096u16 , name : "3C996-T 1000Base-T" } , SubSystem { subvendor : 4279u16 , subdevice : 4097u16 , name : "3C996B-T 1000Base-T" } , SubSystem { subvendor : 4279u16 , subdevice : 4098u16 , name : "3C996C-T 1000Base-T" } , SubSystem { subvendor : 4279u16 , subdevice : 4099u16 , name : "3C997-T 1000Base-T Dual Port" } , SubSystem { subvendor : 4279u16 , subdevice : 4100u16 , name : "3C996-SX 1000Base-SX" } , SubSystem { subvendor : 4279u16 , subdevice : 4101u16 , name : "3C997-SX 1000Base-SX Dual Port" } , SubSystem { subvendor : 4279u16 , subdevice : 4104u16 , name : "3C942 Gigabit LOM (31X31)" } , SubSystem { subvendor : 5348u16 , subdevice : 2u16 , name : "NetXtreme 1000Base-SX" } , SubSystem { subvendor : 5348u16 , subdevice : 3u16 , name : "NetXtreme 1000Base-SX" } , SubSystem { subvendor : 5348u16 , subdevice : 4u16 , name : "NetXtreme 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 4136u16 , name : "NetXtreme 1000BaseTX" } , SubSystem { subvendor : 5348u16 , subdevice : 5700u16 , name : "BCM5700 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5701u16 , name : "NetXtreme BCM5701 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 124u16 , name : "NC7770 Gigabit Server Adapter (PCI-X, 10/100/1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 125u16 , name : "NC6770 Gigabit Server Adapter (PCI-X, 1000-SX)" } , SubSystem { subvendor : 3601u16 , subdevice : 133u16 , name : "NC7780 Gigabit Server Adapter (embedded, WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 153u16 , name : "NC7780 Gigabit Server Adapter (embedded, WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 154u16 , name : "NC7770 Gigabit Server Adapter (PCI-X, 10/100/1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 193u16 , name : "NC6770 Gigabit Server Adapter (PCI-X, 1000-SX)" } , SubSystem { subvendor : 4136u16 , subdevice : 289u16 , name : "Broadcom BCM5701 1000Base-T" } , SubSystem { subvendor : 4156u16 , subdevice : 4746u16 , name : "BCM5701 1000Base-T (HP, OEM 3COM)" } , SubSystem { subvendor : 4156u16 , subdevice : 4747u16 , name : "1000Base-SX (PCI) [A7073A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4772u16 , name : "Core Lan 1000Base-T" } , SubSystem { subvendor : 4156u16 , subdevice : 4801u16 , name : "IOX Core Lan 1000Base-T [A7109AX]" } , SubSystem { subvendor : 4156u16 , subdevice : 4864u16 , name : "Core LAN/SCSI Combo [A6794A]" } , SubSystem { subvendor : 4265u16 , subdevice : 32784u16 , name : "IO9/IO10 Gigabit Ethernet (Copper)" } , SubSystem { subvendor : 4265u16 , subdevice : 32785u16 , name : "Gigabit Ethernet (Copper)" } , SubSystem { subvendor : 4265u16 , subdevice : 32786u16 , name : "Gigabit Ethernet (Fiber)" } , SubSystem { subvendor : 4279u16 , subdevice : 4100u16 , name : "3C996-SX 1000Base-SX" } , SubSystem { subvendor : 4279u16 , subdevice : 4102u16 , name : "3C996B-T 1000Base-T" } , SubSystem { subvendor : 4279u16 , subdevice : 4103u16 , name : "3C1000-T 1000Base-T" } , SubSystem { subvendor : 4279u16 , subdevice : 4104u16 , name : "3C940-BR01 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 1u16 , name : "BCM5701 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 5u16 , name : "BCM5701 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 6u16 , name : "BCM5701 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 7u16 , name : "BCM5701 1000Base-SX" } , SubSystem { subvendor : 5348u16 , subdevice : 8u16 , name : "BCM5701 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 5701u16 , name : "NetXtreme BCM5701 Gigabit Ethernet" } , SubSystem { subvendor : 5348u16 , subdevice : 32776u16 , name : "BCM5701 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5702u16 , name : "NetXtreme BCM5702 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 187u16 , name : "NC7760 1000BaseTX" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Broadcom BCM5702 1000BaseTX" } , SubSystem { subvendor : 5348u16 , subdevice : 32777u16 , name : "BCM5702 1000BaseTX" }] } , Device { vendor_id : 5348u16 , id : 5703u16 , name : "NetXtreme BCM5703 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 153u16 , name : "NC7780 1000BaseTX" } , SubSystem { subvendor : 3601u16 , subdevice : 154u16 , name : "NC7770 1000BaseTX" } , SubSystem { subvendor : 4265u16 , subdevice : 32784u16 , name : "IO9 Gigabit Ethernet (Copper)" } , SubSystem { subvendor : 5348u16 , subdevice : 9u16 , name : "BCM5703 1000BaseTX" } , SubSystem { subvendor : 5348u16 , subdevice : 10u16 , name : "BCM5703 1000BaseSX" } , SubSystem { subvendor : 5348u16 , subdevice : 11u16 , name : "BCM5703 1000BaseTX" } , SubSystem { subvendor : 5348u16 , subdevice : 32777u16 , name : "BCM5703 1000BaseTX" } , SubSystem { subvendor : 5348u16 , subdevice : 32778u16 , name : "BCM5703 1000BaseTX" }] } , Device { vendor_id : 5348u16 , id : 5704u16 , name : "NetXtreme BCM5704 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 207u16 , name : "NC7772 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 208u16 , name : "NC7782 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 209u16 , name : "NC7783 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 4136u16 , subdevice : 330u16 , name : "PowerEdge 1750" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Broadcom NetXtreme BCM5704" } , SubSystem { subvendor : 4156u16 , subdevice : 12559u16 , name : "NC7782 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 4265u16 , subdevice : 32787u16 , name : "Dual Port Gigabit Ethernet (PCI-X,Copper)" } , SubSystem { subvendor : 4265u16 , subdevice : 32792u16 , name : "Dual Port Gigabit Ethernet (A330)" } , SubSystem { subvendor : 4265u16 , subdevice : 32794u16 , name : "Dual Port Gigabit Ethernet (IA-blade)" } , SubSystem { subvendor : 4265u16 , subdevice : 32795u16 , name : "Quad Port Gigabit Ethernet (PCI-E,Copper)" } , SubSystem { subvendor : 4279u16 , subdevice : 8192u16 , name : "3C998-T Dual Port 10/100/1000 PCI-X" } , SubSystem { subvendor : 4279u16 , subdevice : 12288u16 , name : "3C999-T Quad Port 10/100/1000 PCI-X" } , SubSystem { subvendor : 4454u16 , subdevice : 5704u16 , name : "NetXtreme CIOB-E 1000Base-T" } , SubSystem { subvendor : 5940u16 , subdevice : 4107u16 , name : "PRIMERGY RX/TX series onboard LAN" }] } , Device { vendor_id : 5348u16 , id : 5705u16 , name : "NetXtreme BCM5704S_2 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5706u16 , name : "NetXtreme II BCM5706 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5897u16 , name : "NC371i Integrated PCI-X Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12400u16 , name : "NC380T PCI Express Dual Port Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12545u16 , name : "NC370T MultifuNCtion Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12550u16 , name : "NC370i Multifunction Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5708u16 , name : "NetXtreme II BCM5708 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "PowerEdge T605 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4136u16 , subdevice : 547u16 , name : "PowerEdge R905 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4136u16 , subdevice : 7954u16 , name : "PowerEdge R805/R905 Broadcom NetXtreme II BCM5708" } , SubSystem { subvendor : 4156u16 , subdevice : 28727u16 , name : "NC373T PCI Express Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28728u16 , name : "NC373i Integrated Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28741u16 , name : "NC374m PCI Express Dual Port Multifunction Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5709u16 , name : "NetXtreme BCM5702FE Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5710u16 , name : "NetXtreme II BCM57710 10-Gigabit PCIe [Everest]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5916u16 , name : "NC532m Dual Port 10GbE Multifunction BL-C Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28760u16 , name : "NC532i Dual Port 10GbE Multifunction BL-C Adapter" }] } , Device { vendor_id : 5348u16 , id : 5711u16 , name : "NetXtreme II BCM57711 10-Gigabit PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5712u16 , name : "NetXtreme II BCM57711E 10-Gigabit PCIe" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5916u16 , name : "NC532m Dual Port 10GbE Multifunction BL-C Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28760u16 , name : "NC532i Dual Port 10GbE Multifunction BL-C Adapter" }] } , Device { vendor_id : 5348u16 , id : 5715u16 , name : "NetXtreme BCM5705 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 227u16 , name : "NC7761 Gigabit Server Adapter" } , SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" }] } , Device { vendor_id : 5348u16 , id : 5716u16 , name : "NetXtreme BCM5705_2 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 227u16 , name : "NC7761 Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12544u16 , name : "NC1020 ProLiant Gigabit Server Adapter 32 PCI" } , SubSystem { subvendor : 4156u16 , subdevice : 12838u16 , name : "NC150T 4-port Gigabit Combo Switch & Adapter" }] } , Device { vendor_id : 5348u16 , id : 5717u16 , name : "NetXtreme BCM5717 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5718u16 , name : "NetXtreme BCM5718 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5719u16 , name : "NetXtreme BCM5719 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1056u16 , name : "FC 5260/5899 4-port 1 GbE Adapter for Power" } , SubSystem { subvendor : 4156u16 , subdevice : 5789u16 , name : "Ethernet 1Gb 4-port 331FLR Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8894u16 , name : "Ethernet 1Gb 4-port 331i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13187u16 , name : "Ethernet 1Gb 4-port 331T Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 6404u16 , name : "4-port 1Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 6409u16 , name : "Broadcom NetXtreme 5719 Quad Port Gigabit NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 53606u16 , name : "BCM95719-P41 4x1GBT Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 54118u16 , name : "BCM95719-N41 4x1GBT Ethernet NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4133u16 , name : "NIC-ETH330T-LP-4P" } , SubSystem { subvendor : 6461u16 , subdevice : 4230u16 , name : "NIC-ETH330T-3S-4P" }] } , Device { vendor_id : 5348u16 , id : 5721u16 , name : "NetXtreme BCM5721 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 710u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 Broadcom NetXtreme BCM5721" } , SubSystem { subvendor : 4156u16 , subdevice : 5899u16 , name : "NC320m PCI Express Dual Port Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28721u16 , name : "NC320T PCIe Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28722u16 , name : "NC320i PCIe Gigabit Server Adapter" } , SubSystem { subvendor : 5940u16 , subdevice : 4193u16 , name : "PRIMERGY RX/TX S2 series onboard LAN" }] } , Device { vendor_id : 5348u16 , id : 5722u16 , name : "NetXtreme BCM5722 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 888u16 , name : "System x3350 (Machine type 4192)" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 Broadcom NetXtreme 5722" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 Broadcom NetXtreme 5722" } , SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 Broadcom NetXtreme 5722" } , SubSystem { subvendor : 4156u16 , subdevice : 28753u16 , name : "NC105i PCIe Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28754u16 , name : "NC105T PCIe Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5723u16 , name : "NetXtreme BCM5723 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28765u16 , name : "NC107i Integrated PCI Express Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5724u16 , name : "NetXtreme BCM5724 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5725u16 , name : "NetXtreme BCM5705M Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 34397u16 , name : "Latitude D400" } , SubSystem { subvendor : 5348u16 , subdevice : 5725u16 , name : "Dell Latitude D600" }] } , Device { vendor_id : 5348u16 , id : 5726u16 , name : "NetXtreme BCM5705M_2 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4303u16 , subdevice : 4729u16 , name : "LifeBook E8010D" }] } , Device { vendor_id : 5348u16 , id : 5727u16 , name : "NetXtreme BCM5720 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "PowerEdge R320 server" } , SubSystem { subvendor : 4136u16 , subdevice : 2301u16 , name : "PowerEdge R6515/R7515 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 2303u16 , name : "PowerEdge Rx5xx LOM Board" } , SubSystem { subvendor : 4136u16 , subdevice : 2304u16 , name : "PowerEdge C6525 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 2327u16 , name : "PowerEdge C6520 LOM" } , SubSystem { subvendor : 4156u16 , subdevice : 6022u16 , name : "NC332T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6461u16 , name : "NC332i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8499u16 , name : "NC332i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8936u16 , name : "NC332i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8939u16 , name : "NC332i Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 5727u16 , name : "H12SSL-i" }] } , Device { vendor_id : 5348u16 , id : 5730u16 , name : "NetXtreme II BCM57712 10 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5731u16 , name : "NetXtreme II BCM57712 10 Gigabit Ethernet Multi Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5733u16 , name : "NetXtreme BCM5717 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5736u16 , name : "NetXtreme BCM5714 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28729u16 , name : "NC324i PCIe Dual Port Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5737u16 , name : "NetXtreme 5714S Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5738u16 , name : "NetXtreme BCM5780 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28725u16 , name : "NC325i Integrated Dual port PCIe Express Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5739u16 , name : "NetXtreme BCM5780S Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5742u16 , name : "570x 10/100 Integrated Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5743u16 , name : "NetXtreme II BCM57712 10 Gigabit Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5746u16 , name : "NetXtreme BCM5754M Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5747u16 , name : "NetXtreme BCM5755M Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5748u16 , name : "NetXtreme BCM5756ME Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5751u16 , name : "NetXtreme BCM5751 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 374u16 , name : "Dimension XPS Gen 4" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4136u16 , subdevice : 386u16 , name : "Latitude D610" } , SubSystem { subvendor : 4136u16 , subdevice : 391u16 , name : "Precision M70" } , SubSystem { subvendor : 4136u16 , subdevice : 419u16 , name : "Latitude X1" } , SubSystem { subvendor : 4136u16 , subdevice : 424u16 , name : "Precision 380" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4156u16 , subdevice : 12294u16 , name : "DC7100 SFF(DX878AV)" } , SubSystem { subvendor : 5218u16 , subdevice : 652u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4189u16 , name : "Scenic W620" }] } , Device { vendor_id : 5348u16 , id : 5752u16 , name : "NetXtreme BCM5715 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28734u16 , name : "NC326i PCIe Dual Port Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5753u16 , name : "NetXtreme BCM5715S Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5895u16 , name : "NC326m PCIe Dual Port Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5900u16 , name : "NC325m PCIe Quad Port Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28732u16 , name : "NC326i PCIe Dual Port Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5754u16 , name : "NetXtreme BCM5754 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 478u16 , name : "Precision 390" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 532u16 , name : "Precision T3400" } , SubSystem { subvendor : 4136u16 , subdevice : 542u16 , name : "Precision T5400" }] } , Device { vendor_id : 5348u16 , id : 5755u16 , name : "NetXtreme BCM5755 Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" }] } , Device { vendor_id : 5348u16 , id : 5757u16 , name : "NetXtreme BCM5751M Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1399u16 , name : "ThinkPad X41 / Z60t" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2368u16 , name : "Compaq nw8240 Mobile Workstation" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 6058u16 , subdevice : 8321u16 , name : "ThinkPad R60e" }] } , Device { vendor_id : 5348u16 , id : 5758u16 , name : "NetXtreme BCM5751F Fast Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5759u16 , name : "NetLink BCM5787F Fast Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5760u16 , name : "NetXtreme BCM5761e Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5761u16 , name : "NetXtreme BCM5761 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1075u16 , name : "OptiPlex 580" }] } , Device { vendor_id : 5348u16 , id : 5762u16 , name : "NetXtreme BCM57762 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 246u16 , name : "Thunderbolt to Gigabit Ethernet Adapter" }] } , Device { vendor_id : 5348u16 , id : 5763u16 , name : "NetXtreme BCM57767 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5764u16 , name : "NetXtreme BCM5764M Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5765u16 , name : "NetXtreme II BCM57500S Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5766u16 , name : "NetXtreme BCM57766 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5767u16 , name : "NetXtreme BCM5762 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5768u16 , name : "NetXtreme BCM5761 10/100/1000BASE-T Ethernet" , subsystems : & [SubSystem { subvendor : 4697u16 , subdevice : 9992u16 , name : "AT-2712 FX" }] } , Device { vendor_id : 5348u16 , id : 5770u16 , name : "NetXtreme II BCM57800 1/10 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1171u16 , name : "PCIe2 LP 4-Port (10Gb+1GbE) SR+RJ45 Adapter (FC EN0T; CCIN 2CC3)" } , SubSystem { subvendor : 4116u16 , subdevice : 1172u16 , name : "PCIe2 LP 4-Port (10Gb+1GbE) SR+RJ45 Adapter (FC EN0T; CCIN 2CC3)" } , SubSystem { subvendor : 4136u16 , subdevice : 8028u16 , name : "BCM57800 10-Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 8029u16 , name : "BCM57800 10-Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 8039u16 , name : "BCM57800 1-Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 8040u16 , name : "BCM57800 1-Gigabit Ethernet" }] } , Device { vendor_id : 5348u16 , id : 5773u16 , name : "NetXtreme II BCM57840 10/20 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5774u16 , name : "NetXtreme II BCM57810 10 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1170u16 , name : "PCIe2 2-port 10 GbE BaseT RJ45 Adapter (FC EN0W; CCIN 2CC4)" } , SubSystem { subvendor : 4156u16 , subdevice : 6040u16 , name : "Flex-10 10Gb 2-port 530FLB Adapter [Meru]" } , SubSystem { subvendor : 4156u16 , subdevice : 6053u16 , name : "Flex-10 10Gb 2-port 530M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6355u16 , name : "Ethernet 10Gb 2-port 530T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6448u16 , name : "FlexFabric 10Gb 2-port 534FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6449u16 , name : "StoreFabric CN1100R Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6450u16 , name : "FlexFabric 10Gb 2-port 534FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6451u16 , name : "FlexFabric 10Gb 2-port 534M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6458u16 , name : "FlexFabric 10Gb 2-port 533FLR-T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13186u16 , name : "Ethernet 10Gb 2-port 530FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13213u16 , name : "Ethernet 10Gb 2-port 530SFP+ Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4099u16 , name : "530F-B" } , SubSystem { subvendor : 6461u16 , subdevice : 4102u16 , name : "530F-L" } , SubSystem { subvendor : 6461u16 , subdevice : 4111u16 , name : "NIC-ETH522i-Mb-2x10G" } , SubSystem { subvendor : 19538u16 , subdevice : 38930u16 , name : "LREC9812AF Dual-port 10Gb Ethernet Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5776u16 , name : "NetXtreme BCM57760 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5777u16 , name : "NetLink BCM57788 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" }] } , Device { vendor_id : 5348u16 , id : 5778u16 , name : "NetLink BCM57780 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 829u16 , name : "Aspire 7740G" }] } , Device { vendor_id : 5348u16 , id : 5779u16 , name : "NetLink BCM5787M Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "6710b" } , SubSystem { subvendor : 6058u16 , subdevice : 8405u16 , name : "ThinkPad R61" }] } , Device { vendor_id : 5348u16 , id : 5780u16 , name : "NetLink BCM57790 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5782u16 , name : "NetXtreme BCM5782 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 5348u16 , subdevice : 13u16 , name : "NetXtreme BCM5782 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5784u16 , name : "NetLink BCM5784M Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5785u16 , name : "NetLink BCM5785 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5786u16 , name : "NetLink BCM5786 Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5787u16 , name : "NetLink BCM5787 Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5788u16 , name : "NetXtreme BCM5788 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 5197u16 , subdevice : 49176u16 , name : "X20" } , SubSystem { subvendor : 5218u16 , subdevice : 22796u16 , name : "KT6 Delta-FIS2R (MS-6590)" }] } , Device { vendor_id : 5348u16 , id : 5789u16 , name : "NetLink BCM5789 Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5792u16 , name : "NetLink BCM5785 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5793u16 , name : "BCM57840 NetXtreme II 10 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34414u16 , name : "PEB-10G/57840-2T 10GBase-T Network Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4107u16 , name : "NIC-ETH521i-Mb-4x10G" } , SubSystem { subvendor : 19538u16 , subdevice : 38932u16 , name : "LREC9814AF Quad-port 10Gb Ethernet Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5794u16 , name : "BCM57840 NetXtreme II 10/20-Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6422u16 , name : "FlexFabric 20Gb 2-port 630FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6423u16 , name : "FlexFabric 20Gb 2-port 630M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8753u16 , name : "3820C 10/20Gb Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8954u16 , name : "FlexFabric 10Gb 2-port 536FLB Adapter" }] } , Device { vendor_id : 5348u16 , id : 5795u16 , name : "NetXtreme BCM57786 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5796u16 , name : "BCM57840 NetXtreme II Ethernet Multi Function" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6422u16 , name : "NPAR 20Gb 2-port 630FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6423u16 , name : "NPAR 20Gb 2-port 630M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8753u16 , name : "3820C 10/20Gb Converged Network Adapter (NPAR 1.5)" } , SubSystem { subvendor : 4156u16 , subdevice : 8954u16 , name : "FlexFabric 10Gb 2-port 536FLB Adapter (NPAR 1.5)" }] } , Device { vendor_id : 5348u16 , id : 5797u16 , name : "NetXtreme II BCM57800 1/10 Gigabit Ethernet Multi Function" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8028u16 , name : "NetXtreme II BCM57800 10-Gigabit Ethernet Multi Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8029u16 , name : "NetXtreme II BCM57800 10-Gigabit Ethernet Multi Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8039u16 , name : "NetXtreme II BCM57800 1-Gigabit Ethernet Multi Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8040u16 , name : "NetXtreme II BCM57800 1-Gigabit Ethernet Multi Function" }] } , Device { vendor_id : 5348u16 , id : 5798u16 , name : "NetXtreme BCM5702X Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 187u16 , name : "NC7760 Gigabit Server Adapter (PCI-X, 10/100/1000-T)" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "BCM5702 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 12u16 , name : "BCM5702 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 32777u16 , name : "BCM5702 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5799u16 , name : "NetXtreme BCM5703X Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 202u16 , name : "NC7771 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 203u16 , name : "NC7781 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 4116u16 , subdevice : 623u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 5348u16 , subdevice : 9u16 , name : "NetXtreme BCM5703 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 10u16 , name : "NetXtreme BCM5703 1000Base-SX" } , SubSystem { subvendor : 5348u16 , subdevice : 11u16 , name : "NetXtreme BCM5703 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 32778u16 , name : "NetXtreme BCM5703 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5800u16 , name : "NetXtreme BCM5704S Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4907u16 , name : "PCI-X 1000Mbps Dual-port Built-in" } , SubSystem { subvendor : 4265u16 , subdevice : 32788u16 , name : "Dual Port Gigabit Ethernet (PCI-X,Fiber)" } , SubSystem { subvendor : 4265u16 , subdevice : 32796u16 , name : "Quad Port Gigabit Ethernet (PCI-E,Fiber)" } , SubSystem { subvendor : 4279u16 , subdevice : 8193u16 , name : "3C998-SX Dual Port 1000-SX PCI-X" }] } , Device { vendor_id : 5348u16 , id : 5801u16 , name : "NetXtreme II BCM57800 1/10 Gigabit Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8028u16 , name : "NetXtreme II BCM57800 10-Gigabit Ethernet Virtual Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8029u16 , name : "NetXtreme II BCM57800 10-Gigabit Ethernet Virtual Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8039u16 , name : "NetXtreme II BCM57800 1-Gigabit Ethernet Virtual Function" } , SubSystem { subvendor : 4136u16 , subdevice : 8040u16 , name : "NetXtreme II BCM57800 1-Gigabit Ethernet Virtual Function" }] } , Device { vendor_id : 5348u16 , id : 5802u16 , name : "NetXtreme II BCM5706S Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12546u16 , name : "NC370F MultifuNCtion Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12556u16 , name : "NC370i Multifunction Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5803u16 , name : "NetXtreme II BCM57840 10/20 Gigabit Ethernet Multi Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5804u16 , name : "NetXtreme II BCM5708S Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 772u16 , name : "NetXtreme II BCM5708S Gigabit Ethernet" } , SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 Broadcom NetXtreme II BCM5708S" } , SubSystem { subvendor : 4136u16 , subdevice : 524u16 , name : "PowerEdge M605 Broadcom NetXtreme II BCM5708S" } , SubSystem { subvendor : 4156u16 , subdevice : 5894u16 , name : "NC373m Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28728u16 , name : "NC373i PCI Express Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28731u16 , name : "NC373i Integrated Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28733u16 , name : "NC373F PCI Express Multifunction Gigabit Server Adapter" }] } , Device { vendor_id : 5348u16 , id : 5805u16 , name : "NetXtreme II BCM57840 10/20 Gigabit Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6422u16 , name : "FlexFabric 20Gb 2-port 630FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6423u16 , name : "FlexFabric 20Gb 2-port 630M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8753u16 , name : "3820C 10/20Gb Converged Network Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4156u16 , subdevice : 8954u16 , name : "FlexFabric 10Gb 2-port 536FLB Adapter (SR-IOV VF)" }] } , Device { vendor_id : 5348u16 , id : 5806u16 , name : "NetXtreme II BCM57810 10 Gigabit Ethernet Multi Function" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6040u16 , name : "NPAR 10Gb 2-port 530FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6053u16 , name : "NPAR 10Gb 2-port 530M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6355u16 , name : "NPAR 10Gb 2-port 530T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6448u16 , name : "NPAR 10Gb 2-port 534FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6449u16 , name : "NPAR CN1100R Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6450u16 , name : "NPAR 10Gb 2-port 534FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6451u16 , name : "NPAR 10Gb 2-port 534M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6458u16 , name : "NPAR 10Gb 2-port 533FLR-T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13186u16 , name : "NPAR 10Gb 2-port 530FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13213u16 , name : "NPAR 10Gb 2-port 530SFP+ Adapter" }] } , Device { vendor_id : 5348u16 , id : 5807u16 , name : "NetXtreme II BCM57810 10 Gigabit Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6040u16 , name : "Flex-10 10Gb 2-port 530FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6053u16 , name : "Flex-10 10Gb 2-port 530M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6355u16 , name : "Ethernet 10Gb 2-port 530T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6448u16 , name : "FlexFabric 10Gb 2-port 534FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6449u16 , name : "StoreFabric CN1100R Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6450u16 , name : "FlexFabric 10Gb 2-port 534FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6451u16 , name : "FlexFabric 10Gb 2-port 534M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6458u16 , name : "FlexFabric 10Gb 2-port 533FLR-T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13186u16 , name : "Ethernet 10Gb 2-port 530FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13213u16 , name : "Ethernet 10Gb 2-port 530SFP+ Adapter" }] } , Device { vendor_id : 5348u16 , id : 5808u16 , name : "NetXtreme BCM57761 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5809u16 , name : "NetLink BCM57781 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 38577u16 , name : "Z77 Extreme4 motherboard" }] } , Device { vendor_id : 5348u16 , id : 5810u16 , name : "NetLink BCM57791 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5811u16 , name : "NetXtreme BCM57786 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5812u16 , name : "NetXtreme BCM57765 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5813u16 , name : "NetLink BCM57785 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5814u16 , name : "NetLink BCM57795 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5815u16 , name : "NetXtreme BCM57782 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5820u16 , name : "BCM57765/57785 SDXC/MMC Card Reader" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5822u16 , name : "BCM57765/57785 MS Card Reader" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5823u16 , name : "BCM57765/57785 xD-Picture Card Reader" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5825u16 , name : "NetXtreme-E RDMA Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5830u16 , name : "NetXtreme BCM5702A3 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 4352u16 , name : "3C1000B-T 10/100/1000 PCI" } , SubSystem { subvendor : 5348u16 , subdevice : 12u16 , name : "BCM5702 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 32777u16 , name : "BCM5702 1000Base-T" }] } , Device { vendor_id : 5348u16 , id : 5831u16 , name : "NetXtreme BCM5703 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 202u16 , name : "NC7771 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 3601u16 , subdevice : 203u16 , name : "NC7781 Gigabit Server Adapter (PCI-X, 10,100,1000-T)" } , SubSystem { subvendor : 4156u16 , subdevice : 4803u16 , name : "Combo FC/GigE-SX [A9782A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4810u16 , name : "Combo FC/GigE-T [A9784A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4897u16 , name : "Core I/O LAN/SCSI Combo [AB314A]" } , SubSystem { subvendor : 5348u16 , subdevice : 9u16 , name : "NetXtreme BCM5703 1000Base-T" } , SubSystem { subvendor : 5348u16 , subdevice : 10u16 , name : "NetXtreme BCM5703 1000Base-SX" }] } , Device { vendor_id : 5348u16 , id : 5832u16 , name : "BCM57301 NetXtreme-C 10Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5833u16 , name : "BCM57302 NetXtreme-C 10Gb/25Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5834u16 , name : "BCM57304 NetXtreme-C 10Gb/25Gb/40Gb/50Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5835u16 , name : "NetXtreme-C Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5836u16 , name : "BCM57417 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5838u16 , name : "BCM57311 NetXtreme-C 10Gb RDMA Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5839u16 , name : "BCM57312 NetXtreme-C 10Gb/25Gb RDMA Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5840u16 , name : "BCM57402 NetXtreme-E 10Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5841u16 , name : "BCM57404 NetXtreme-E 10Gb/25Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5842u16 , name : "BCM57406 NetXtreme-E 10GBASE-T Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5843u16 , name : "NetXtreme-E Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5844u16 , name : "BCM57402 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5845u16 , name : "BCM57407 NetXtreme-E 10GBase-T Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5846u16 , name : "BCM57412 NetXtreme-E 10Gb RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 432u16 , name : "RD10057 10GbE interface" } , SubSystem { subvendor : 4476u16 , subdevice : 205u16 , name : "FastFrame N4S2 Dual-port 10Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 4610u16 , name : "BCM957412M4122C OCP 1x25G Type1 wRoCE" } , SubSystem { subvendor : 5348u16 , subdevice : 16672u16 , name : "NetXtreme E-Series Advanced Dual-port 10Gb SFP+ Ethernet Network Daughter Card" } , SubSystem { subvendor : 5348u16 , subdevice : 16678u16 , name : "NetXtreme-E Dual-port 10G SFP+ Ethernet OCP 3.0 Adapter (BCM957412N4120C)" } , SubSystem { subvendor : 5421u16 , subdevice : 35616u16 , name : "BCM57412 NetXtreme-E 10Gb RDMA Ethernet Controller" } , SubSystem { subvendor : 5421u16 , subdevice : 35618u16 , name : "BCM57412 NetXtreme-E 25Gb RDMA Ethernet Controller" } , SubSystem { subvendor : 6461u16 , subdevice : 4132u16 , name : "NIC-ETH531F-LP-2P" } , SubSystem { subvendor : 6461u16 , subdevice : 4231u16 , name : "NIC-ETH531F-3S-2P" }] } , Device { vendor_id : 5348u16 , id : 5847u16 , name : "BCM57414 NetXtreme-E 10Gb/25Gb RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 204u16 , name : "FastFrame N422 Dual-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16599u16 , name : "ThunderLink NS 5252 Dual-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 5122u16 , name : "BCM957414A4142CC 10Gb/25Gb Ethernet PCIe" } , SubSystem { subvendor : 5348u16 , subdevice : 5124u16 , name : "BCM957414M4142C OCP 2x25G Type1 wRoCE" } , SubSystem { subvendor : 5348u16 , subdevice : 16704u16 , name : "NetXtreme E-Series Advanced Dual-port 25Gb SFP28 Network Daughter Card" } , SubSystem { subvendor : 5348u16 , subdevice : 16707u16 , name : "NetXtreme-E Single-port 40Gb/50Gb Ethernet OCP 2.0 Adapter (BCM957414M4143C)" } , SubSystem { subvendor : 5348u16 , subdevice : 16710u16 , name : "NetXtreme-E Dual-port 25G SFP28 Ethernet OCP 3.0 Adapter (BCM957414N4140C)" } , SubSystem { subvendor : 5520u16 , subdevice : 526u16 , name : "Ethernet 25Gb 2-port 631SFP28 Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 529u16 , name : "Ethernet 25Gb 2-port 631FLR-SFP28 Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4147u16 , name : "NIC-ETH630F-LP-2P" } , SubSystem { subvendor : 7124u16 , subdevice : 142u16 , name : "ENFB2251-SP2" } , SubSystem { subvendor : 7124u16 , subdevice : 143u16 , name : "ENPB2251-SP2" } , SubSystem { subvendor : 7916u16 , subdevice : 257u16 , name : "VSE-225-31S Dual-port 10Gb/25Gb Ethernet PCIe" }] } , Device { vendor_id : 5348u16 , id : 5848u16 , name : "BCM57416 NetXtreme-E Dual-Media 10G RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8171u16 , name : "NetXtreme-E 10Gb SFP+ Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 206u16 , name : "FastFrame N4T2 Dual-port 10GBASE-T Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 16739u16 , name : "NetXtreme-E Dual-port 10GBASE-T Ethernet OCP 2.0 Adapter (BCM957416M4163C)" } , SubSystem { subvendor : 5348u16 , subdevice : 16742u16 , name : "NetXtreme-E Dual-port 10GBASE-T Ethernet OCP 3.0 Adapter (BCM957416N4160C)" } , SubSystem { subvendor : 5348u16 , subdevice : 53568u16 , name : "BCM957416-P410 4x10GBT Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 54080u16 , name : "BCM957416-N410 4x10GBT Ethernet NIC" } , SubSystem { subvendor : 5520u16 , subdevice : 524u16 , name : "Ethernet 10Gb 2-port 535T Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 530u16 , name : "Ethernet 10Gb 2-port 535FLR-T Adapter" }] } , Device { vendor_id : 5348u16 , id : 5849u16 , name : "BCM57417 NetXtreme-E 10GBASE-T RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 18534u16 , name : "Dual Port 10GBase-T Ethernet Controller" }] } , Device { vendor_id : 5348u16 , id : 5852u16 , name : "NetXtreme-E Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5853u16 , name : "NetLink BCM5781 Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5854u16 , name : "BCM57412 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5855u16 , name : "BCM57314 NetXtreme-C 10Gb/25Gb/40Gb/50Gb RDMA Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5857u16 , name : "NetXtreme-C Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5858u16 , name : "BCM57417 NetXtreme-E 10Gb/25Gb RDMA Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 18534u16 , name : "Dual Port 10Gb/25Gb SFP28 Ethernet Controller" }] } , Device { vendor_id : 5348u16 , id : 5859u16 , name : "BCM57416 NetXtreme-E 10Gb RDMA Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5861u16 , name : "NetXtreme-C RDMA Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5863u16 , name : "BCM57404 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5864u16 , name : "BCM57406 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5865u16 , name : "BCM57407 NetXtreme-E 25Gb Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5867u16 , name : "BCM57412 NetXtreme-E RDMA Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5868u16 , name : "BCM57414 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5869u16 , name : "BCM57414 NetXtreme-E RDMA Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5870u16 , name : "BCM57416 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5871u16 , name : "BCM57416 NetXtreme-E RDMA Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5873u16 , name : "BCM57452 NetXtreme-E 10Gb/25Gb/40Gb/50Gb Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5875u16 , name : "NetXtreme BCM5727 Gigabit Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5879u16 , name : "NetXtreme BCM5753 Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5885u16 , name : "NetXtreme BCM5753M Gigabit Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" }] } , Device { vendor_id : 5348u16 , id : 5886u16 , name : "NetXtreme BCM5753F Fast Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5900u16 , name : "BCM4401-B0 100Base-TX" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 392u16 , name : "Inspiron 6000 laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4136u16 , subdevice : 431u16 , name : "Inspiron 6400" } , SubSystem { subvendor : 4136u16 , subdevice : 461u16 , name : "Inspiron 9400 Laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 472u16 , name : "Inspiron E1405" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12450u16 , name : "NX7300 laptop" } , SubSystem { subvendor : 5348u16 , subdevice : 5900u16 , name : "HP Compaq 6720t Mobile Thin Client" }] } , Device { vendor_id : 5348u16 , id : 5901u16 , name : "NetXtreme BCM5901 100Base-TX" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1349u16 , name : "ThinkPad R40e" }] } , Device { vendor_id : 5348u16 , id : 5902u16 , name : "NetXtreme BCM5901 100Base-TX" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5906u16 , name : "NetLink BCM5906 Fast Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5907u16 , name : "NetLink BCM5906M Fast Ethernet PCI Express" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 521u16 , name : "XPS M1330" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 6058u16 , subdevice : 14883u16 , name : "IdeaPad S10e" }] } , Device { vendor_id : 5348u16 , id : 5968u16 , name : "BCM57508 NetXtreme-E 10Gb/25Gb/40Gb/50Gb/100Gb/200Gb Ethernet" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 207u16 , name : "FastFrame N412 Dual-port 100Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16598u16 , name : "ThunderLink NS 5102 Dual-port 100Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 8448u16 , name : "NetXtreme-E Dual-port 100G QSFP56 Ethernet PCIe4.0 x16 Adapter (BCM957508-P2100G)" } , SubSystem { subvendor : 5348u16 , subdevice : 21000u16 , name : "NetXtreme-E Dual-port 100G QSFP56 Ethernet OCP 3.0 Adapter (BCM957508-N2100G)" } , SubSystem { subvendor : 5348u16 , subdevice : 21002u16 , name : "NetXtreme-E Dual-port 100G DSFP Ethernet OCP 3.0 Adapter (BCM957508-N2100GD)" } , SubSystem { subvendor : 5348u16 , subdevice : 53540u16 , name : "NetXtreme-E P2100D BCM57508 2x100G QSFP PCIE" } , SubSystem { subvendor : 5348u16 , subdevice : 54052u16 , name : "NetXtreme-E N2100D BCM57508 2x100G QSFP OCP3.0 Ethernet" } , SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "NetXtreme-E NGM2100D BCM57508 2x100G KR Mezz Ethernet" }] } , Device { vendor_id : 5348u16 , id : 5969u16 , name : "BCM57504 NetXtreme-E 10Gb/25Gb/40Gb/50Gb/100Gb Ethernet" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2516u16 , name : "PowerEdge XR11/XR12 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 2843u16 , name : "PowerEdge XR5610 LOM" } , SubSystem { subvendor : 4476u16 , subdevice : 218u16 , name : "FastFrame N424 Quad-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16607u16 , name : "ThunderLink NS 5254 Quad-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 16976u16 , name : "NetXtreme-E Quad-port 25G SFP28 Ethernet PCIe4.0 x16 Adapter (BCM957504-P425G)" } , SubSystem { subvendor : 5348u16 , subdevice : 20549u16 , name : "NetXtreme-E BCM57504 4x25G OCP3.0" } , SubSystem { subvendor : 5348u16 , subdevice : 20736u16 , name : "NetXtreme-E Single-port 100G QSFP56 Ethernet OCP 3.0 Adapter (BCM957504-N1100G)" } , SubSystem { subvendor : 5348u16 , subdevice : 20741u16 , name : "NetXtreme-E Single-port 100G DSFP Ethernet OCP 3.0 Adapter (BCM957504-N1100GD)" } , SubSystem { subvendor : 5348u16 , subdevice : 21072u16 , name : "NetXtreme-E BCM57504 4x25G KR Mezz" } , SubSystem { subvendor : 5348u16 , subdevice : 21541u16 , name : "NetXtreme-E Quad-port 25G SFP28 Ethernet OCP 3.0 Adapter (BCM957504-N425G)" } , SubSystem { subvendor : 5348u16 , subdevice : 53570u16 , name : "NetXtreme-E P425D BCM57504 4x25G SFP28 PCIE" } , SubSystem { subvendor : 5520u16 , subdevice : 1056u16 , name : "HPE Ethernet 25/50Gb 2-port 6310C Adapter" }] } , Device { vendor_id : 5348u16 , id : 5970u16 , name : "BCM57502 NetXtreme-E 10Gb/25Gb/40Gb/50Gb Ethernet" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 5984u16 , name : "BCM57608 25Gb/50Gb/100Gb/200Gb/400Gb Ethernet" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 207u16 , name : "FastFrame N522 Dual-port 200Gb Ethernet Adapter" } , SubSystem { subvendor : 5348u16 , subdevice : 37136u16 , name : "BCM57608 1x400G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37152u16 , name : "BCM57608 2x200G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37153u16 , name : "BCM57608 2x100G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37157u16 , name : "BCM57608 2x200G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37158u16 , name : "BCM57608 2x100G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37184u16 , name : "BCM57608 1x400G QSFP-DD PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37648u16 , name : "BCM57608 1x400G QSFP-DD OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37649u16 , name : "BCM57608 1x400G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37650u16 , name : "BCM57608 1x200G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37664u16 , name : "BCM57608 2x200G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37669u16 , name : "BCM57608 2x200G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37670u16 , name : "BCM57608 2x100G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37696u16 , name : "BCM57608 4x100G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37701u16 , name : "BCM57608 4x25G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 53541u16 , name : "BCM57608 2x200G PCIe Ethernet NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4187u16 , name : "NIC-ETH2030F-LP-2P 2x200G PCIe Ethernet NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4188u16 , name : "NIC-ETH4030F-LP-1P 1x400G PCIe Ethernet NIC" }] } , Device { vendor_id : 5348u16 , id : 6016u16 , name : "BCM57708 50Gb/100Gb/200Gb/400Gb/800Gb Ethernet" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 4163u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 4419u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 5187u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 5443u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 6211u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 6467u16 , name : "BCM957708 1x800G OCP Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 32803u16 , name : "BCM957708 2x400G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 32835u16 , name : "BCM957708 1x800G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 33059u16 , name : "BCM957708 2x400G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 33091u16 , name : "BCM957708 1x800G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 36899u16 , name : "BCM957708 2x400G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 36931u16 , name : "BCM957708 1x800G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37155u16 , name : "BCM957708 2x400G PCIe Ethernet NIC" } , SubSystem { subvendor : 5348u16 , subdevice : 37187u16 , name : "BCM957708 1x800G PCIe Ethernet NIC" }] } , Device { vendor_id : 5348u16 , id : 6144u16 , name : "BCM57502 NetXtreme-E Ethernet Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 6145u16 , name : "BCM57504 NetXtreme-E Ethernet Partition" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 1056u16 , name : "Ethernet NPAR 6310C Adapter" }] } , Device { vendor_id : 5348u16 , id : 6146u16 , name : "BCM57508 NetXtreme-E Ethernet Partition" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "BCM57508 NetXtreme-E NGM2100D 2x100G KR Mezz Ethernet Partition" }] } , Device { vendor_id : 5348u16 , id : 6147u16 , name : "BCM57502 NetXtreme-E RDMA Partition" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 6148u16 , name : "BCM57504 NetXtreme-E RDMA Partition" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 1056u16 , name : "Ethernet NPAR RoCE 6310C Adapter" }] } , Device { vendor_id : 5348u16 , id : 6149u16 , name : "BCM57508 NetXtreme-E RDMA Partition" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "NetXtreme-E NGM2100D BCM57508 2x100G KR Mezz RDMA Partition" }] } , Device { vendor_id : 5348u16 , id : 6150u16 , name : "BCM5750X NetXtreme-E Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "BCM57508 NetXtreme-E NGM2100D 2x100G KR Mezz Ethernet Virtual Function" } , SubSystem { subvendor : 5520u16 , subdevice : 1056u16 , name : "Ethernet SRIOV 6310C Adapter" }] } , Device { vendor_id : 5348u16 , id : 6151u16 , name : "BCM5750X NetXtreme-E RDMA Virtual Function" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "BCM57508 NetXtreme-E NGM2100D 2x100G KR Mezz RDMA Virtual Function" } , SubSystem { subvendor : 5520u16 , subdevice : 1056u16 , name : "Ethernet SRIOV RoCE 6310C Adapter" }] } , Device { vendor_id : 5348u16 , id : 6152u16 , name : "BCM5750X NetXtreme-E Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "BCM57508 NetXtreme-E NGM2100D 2x100G KR Mezz Ethernet Virtual Function" }] } , Device { vendor_id : 5348u16 , id : 6153u16 , name : "BCM5750X NetXtreme-E RDMA Virtual Function" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 57124u16 , name : "BCM57508 NetXtreme-E NGM2100D 2x100G KR Mezz RDMA Virtual Function" }] } , Device { vendor_id : 5348u16 , id : 6169u16 , name : "BCM5760X Ethernet Virtual Function" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 10001u16 , name : "BCM2711 PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 10002u16 , name : "BCM2712 PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 13138u16 , name : "BCM3352" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 13152u16 , name : "BCM3360" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 16912u16 , name : "BCM4210 iLine10 HomePNA 2.0" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 16913u16 , name : "BCM4211 iLine10 HomePNA 2.0 + V.90 56k modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 16914u16 , name : "BCM4212 v.90 56k modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 16928u16 , name : "802-11b/g Wireless PCI controller, packaged as a Linksys WPC54G ver 1.2 PCMCIA card" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 16930u16 , name : "NetXtreme BCM5753M Gigabit Ethernet PCI Express" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17153u16 , name : "BCM4301 802.11b Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1031u16 , name : "TrueMobile 1180 Onboard WLAN" } , SubSystem { subvendor : 4163u16 , subdevice : 288u16 , name : "WL-103b Wireless LAN PC Card" } , SubSystem { subvendor : 5797u16 , subdevice : 5634u16 , name : "B-300 802.11b Wireless CardBus Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 17153u16 , name : "WMP11 v2.7 802.11b Wireless-B PCI Adapter" }] } , Device { vendor_id : 5348u16 , id : 17157u16 , name : "BCM4307 V.90 56k Modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17158u16 , name : "BCM4306 802.11bg Wireless LAN controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17159u16 , name : "BCM4306 802.11bg Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17168u16 , name : "BCM4310 Chipcommon I/OController" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17169u16 , name : "BCM4311 802.11b/g WLAN" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7u16 , name : "Wireless 1390 WLAN Mini-Card" } , SubSystem { subvendor : 4136u16 , subdevice : 8u16 , name : "Wireless 1390 WLAN ExpressCard" } , SubSystem { subvendor : 4156u16 , subdevice : 4963u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4964u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4965u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4980u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4981u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4982u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4983u16 , name : "BCM4311 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4991u16 , name : "BCM4322 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4992u16 , name : "BCM4322 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 5348u16 , subdevice : 17169u16 , name : "BCM94311MCG" }] } , Device { vendor_id : 5348u16 , id : 17170u16 , name : "BCM4311 802.11a/b/g" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7u16 , name : "Wireless 1490 Dual Band WLAN Mini-Card" } , SubSystem { subvendor : 4136u16 , subdevice : 8u16 , name : "Wireless 1490 Dual Band WLAN ExpressCard" } , SubSystem { subvendor : 4156u16 , subdevice : 4954u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4959u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4960u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4961u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4962u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4976u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4977u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4978u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4979u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4203u16 , subdevice : 137u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4977u16 , subdevice : 4156u16 , name : "Broadcom 802.11 Multiband-netwerkadapter(6715s)" }] } , Device { vendor_id : 5348u16 , id : 17171u16 , name : "BCM4311 802.11a" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17173u16 , name : "BCM4312 802.11b/g LP-PHY" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 11u16 , name : "Wireless 1395 WLAN Mini-Card" } , SubSystem { subvendor : 4136u16 , subdevice : 12u16 , name : "Wireless 1397 WLAN Mini-Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4988u16 , name : "BCM4312 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4989u16 , name : "BCM4312 802.11b/g Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 5383u16 , name : "U98Z049.00 Wireless Mini PCIe Card" } , SubSystem { subvendor : 4187u16 , subdevice : 57347u16 , name : "T77H030.00 Wireless Mini PCIe Card" } , SubSystem { subvendor : 4187u16 , subdevice : 57371u16 , name : "T77H106.00 Wireless Half-size Mini PCIe Card" }] } , Device { vendor_id : 5348u16 , id : 17176u16 , name : "BCM4318 [AirForce One 54g] 802.11g Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 5u16 , name : "Wireless 1370 WLAN Mini-PCI Card" } , SubSystem { subvendor : 4136u16 , subdevice : 6u16 , name : "Wireless 1370 WLAN PC Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4949u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4950u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4951u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4163u16 , subdevice : 4111u16 , name : "WL-138G v2 / WL-138gE / WL-100gE" } , SubSystem { subvendor : 4163u16 , subdevice : 4623u16 , name : "A6U notebook embedded card" } , SubSystem { subvendor : 4436u16 , subdevice : 853u16 , name : "Buffalo WLI2-PCI-G54S High Speed Mode Wireless Adapter" } , SubSystem { subvendor : 5224u16 , subdevice : 785u16 , name : "Aspire 3022WLMi, 5024WLMi, 5020" } , SubSystem { subvendor : 5224u16 , subdevice : 786u16 , name : "TravelMate 2410" } , SubSystem { subvendor : 5348u16 , subdevice : 1097u16 , name : "Gateway 7510GX" } , SubSystem { subvendor : 5868u16 , subdevice : 281u16 , name : "U.S.Robotics Wireless MAXg PC Card" } , SubSystem { subvendor : 5943u16 , subdevice : 66u16 , name : "WMP54GS v1.1 802.11g Wireless-G PCI Adapter with SpeedBooster" } , SubSystem { subvendor : 5943u16 , subdevice : 72u16 , name : "WPC54G v3 802.11g Wireless-G Notebook Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 73u16 , name : "WPC54GS v2 802.11g Wireless-G Notebook Adapter with SpeedBooster" } , SubSystem { subvendor : 6041u16 , subdevice : 28672u16 , name : "F5D7000 v4000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28673u16 , name : "F5D7001 v2000 Wireless G Plus Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28688u16 , name : "F5D7010 v4000 Wireless G Notebook Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28689u16 , name : "F5D7011 v2000 High-Speed Mode Wireless G Notebook Card" }] } , Device { vendor_id : 5348u16 , id : 17177u16 , name : "BCM4318 [AirForce 54g] 802.11a/b/g PCI Express Transceiver" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 5u16 , name : "Wireless 1470 Dual Band WLAN Mini-PCI Card" } , SubSystem { subvendor : 4136u16 , subdevice : 6u16 , name : "Wireless 1470 Dual Band WLAN PC Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4952u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4953u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4954u16 , name : "Broadcom 802.11a/b/g WLAN" }] } , Device { vendor_id : 5348u16 , id : 17184u16 , name : "BCM4306 802.11b/g Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1u16 , name : "TrueMobile 1300 WLAN Mini-PCI Card" } , SubSystem { subvendor : 4136u16 , subdevice : 2u16 , name : "TrueMobile 1300 WLAN PC Card" } , SubSystem { subvendor : 4136u16 , subdevice : 3u16 , name : "Wireless 1350 WLAN Mini-PCI Card" } , SubSystem { subvendor : 4136u16 , subdevice : 4u16 , name : "Wireless 1350 WLAN PC Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4852u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4856u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4858u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4859u16 , name : "Broadcom 802.11b/g WLAN" } , SubSystem { subvendor : 4163u16 , subdevice : 4111u16 , name : "WL-100G" } , SubSystem { subvendor : 4183u16 , subdevice : 28709u16 , name : "WN825G" } , SubSystem { subvendor : 4203u16 , subdevice : 78u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4436u16 , subdevice : 816u16 , name : "Buffalo WLI2-PCI-G54S High Speed Mode Wireless Desktop Adapter" } , SubSystem { subvendor : 5199u16 , subdevice : 28752u16 , name : "eMachines M6805 802.11g Built-in Wireless" } , SubSystem { subvendor : 5199u16 , subdevice : 28753u16 , name : "Sonnet Aria Extreme PCI" } , SubSystem { subvendor : 5797u16 , subdevice : 5636u16 , name : "WE602-B" } , SubSystem { subvendor : 5943u16 , subdevice : 19u16 , name : "WMP54G v1 802.11g PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 20u16 , name : "WMP54G v2 802.11g PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 21u16 , name : "WMP54GS v1.0 802.11g Wireless-G PCI Adapter with SpeedBooster" } , SubSystem { subvendor : 5943u16 , subdevice : 17184u16 , name : "WPC54G v1 / WPC54GS v1 802.11g Wireless-G Notebook Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 28672u16 , name : "F5D7000 v1000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28673u16 , name : "F5D7001 v1000 Wireless G Plus Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28688u16 , name : "F5D7010 v1000 Wireless G Notebook Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28689u16 , name : "F5D7011 v1000 High-Speed Mode Wireless G Notebook Card" } , SubSystem { subvendor : 6239u16 , subdevice : 4640u16 , name : "TravelMate 290E WLAN Mini-PCI Card" }] } , Device { vendor_id : 5348u16 , id : 17185u16 , name : "BCM4321 802.11a Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17186u16 , name : "BCM4322 802.11bgn Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17188u16 , name : "BCM4309 802.11abg Wireless Network Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1u16 , name : "Truemobile 1400" } , SubSystem { subvendor : 4136u16 , subdevice : 2u16 , name : "TrueMobile 1400 Dual Band WLAN PC Card" } , SubSystem { subvendor : 4136u16 , subdevice : 3u16 , name : "Truemobile 1450 MiniPCI" } , SubSystem { subvendor : 4136u16 , subdevice : 4u16 , name : "Wireless 1450 Dual Band WLAN PC Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4857u16 , name : "Broadcom 802.11a/b/g WLAN" } , SubSystem { subvendor : 4156u16 , subdevice : 4860u16 , name : "Broadcom 802.11a/b/g WLAN" }] } , Device { vendor_id : 5348u16 , id : 17189u16 , name : "BCM4306 802.11bg Wireless Network Controller" , subsystems : & [SubSystem { subvendor : 5140u16 , subdevice : 3u16 , name : "Wireless Notebook Adapter MN-720" } , SubSystem { subvendor : 5140u16 , subdevice : 4u16 , name : "Wireless PCI Adapter MN-730" }] } , Device { vendor_id : 5348u16 , id : 17190u16 , name : "BCM4307 Chipcommon I/O Controller?" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17192u16 , name : "BCM4321 802.11a/b/g/n" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9u16 , name : "Wireless 1500 Draft 802.11n WLAN Mini-Card" } , SubSystem { subvendor : 4136u16 , subdevice : 10u16 , name : "Wireless 1500 Draft 802.11n WLAN Mini-card" } , SubSystem { subvendor : 4156u16 , subdevice : 4966u16 , name : "BCM4321 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4967u16 , name : "BCM4321 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4968u16 , name : "BCM4321 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4969u16 , name : "BCM4321 802.11a/b/g/n Wireless LAN Controller" } , SubSystem { subvendor : 4203u16 , subdevice : 135u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 136u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 139u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 140u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 144u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 5348u16 , subdevice : 17192u16 , name : "BCM4328 802.11a/b/g/n" } , SubSystem { subvendor : 5943u16 , subdevice : 102u16 , name : "WPC600N v1 802.11a/b/g/n Wireless-N CardBus Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 104u16 , name : "WEC600N v1 802.11a/b/g/n Wireless-N ExpressCard" }] } , Device { vendor_id : 5348u16 , id : 17193u16 , name : "BCM4321 802.11b/g/n" , subsystems : & [SubSystem { subvendor : 4997u16 , subdevice : 31488u16 , name : "WN511B RangeMax NEXT Wireless Notebook Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 32000u16 , name : "WN311B RangeMax Next 270 Mbps Wireless PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 88u16 , name : "WPC300N v1 Wireless-N Notebook Adapter" }] } , Device { vendor_id : 5348u16 , id : 17194u16 , name : "BCM4321 802.11an Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17195u16 , name : "BCM4322 802.11a/b/g/n Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 13u16 , name : "Wireless 1510 Wireless-N WLAN Mini-Card" } , SubSystem { subvendor : 4203u16 , subdevice : 141u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 142u16 , name : "AirPort Extreme" }] } , Device { vendor_id : 5348u16 , id : 17196u16 , name : "BCM4322 802.11b/g/n" , subsystems : & [SubSystem { subvendor : 6041u16 , subdevice : 54033u16 , name : "Dynex DX-NNBX 802.11n WLAN Cardbus Card" }] } , Device { vendor_id : 5348u16 , id : 17197u16 , name : "BCM4322 802.11an Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17201u16 , name : "BCM4331 802.11a/b/g/n" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 214u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 228u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 239u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 244u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 245u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 270u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 271u16 , name : "AirPort Extreme" }] } , Device { vendor_id : 5348u16 , id : 17203u16 , name : "Serial (EDGE/GPRS modem part of Option GT Combo Edge)" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17220u16 , name : "EDGE/GPRS data and 802.11b/g combo cardbus [GC89]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17232u16 , name : "BCM43222 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17233u16 , name : "BCM43222 802.11abgn Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17235u16 , name : "BCM43224 802.11a/b/g/n" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 14u16 , name : "Wireless 1520 Half-size Mini PCIe Card" } , SubSystem { subvendor : 4156u16 , subdevice : 5385u16 , name : "WMIB-275N Half-size Mini PCIe Card" } , SubSystem { subvendor : 4203u16 , subdevice : 147u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 209u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4203u16 , subdevice : 233u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 5348u16 , subdevice : 1240u16 , name : "Pegatron UPWL6024" }] } , Device { vendor_id : 5348u16 , id : 17239u16 , name : "BCM43225 802.11b/g/n" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 57377u16 , name : "T77H103.00 Wireless Half-size Mini PCIe Card" }] } , Device { vendor_id : 5348u16 , id : 17240u16 , name : "BCM43227 802.11b/g/n" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17241u16 , name : "BCM43228 802.11a/b/g/n" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 17u16 , name : "Wireless 1530 Half-size Mini PCIe Card" } , SubSystem { subvendor : 4156u16 , subdevice : 6188u16 , name : "BCM943228HM4L 802.11a/b/g/n 2x2 Wi-Fi Adapter" }] } , Device { vendor_id : 5348u16 , id : 17248u16 , name : "BCM4360 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17253u16 , name : "BCM43142 802.11b/g/n" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 22u16 , name : "Wireless 1704 802.11n + BT 4.0" }] } , Device { vendor_id : 5348u16 , id : 17312u16 , name : "BCM4360 802.11ac Dual Band Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17313u16 , name : "BCM4360 802.11ac 2,4G Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17314u16 , name : "BCM4360 802.11ac 5G Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17315u16 , name : "BCM4350 802.11ac Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 33u16 , name : "Wireless 1820" } , SubSystem { subvendor : 6058u16 , subdevice : 1882u16 , name : "00JT494" }] } , Device { vendor_id : 5348u16 , id : 17321u16 , name : "BCM43217 802.11b/g/n" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17322u16 , name : "BCM43131 802.11b/g/n" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17326u16 , name : "BCM43162 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17329u16 , name : "BCM4352 802.11ac Dual Band Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34234u16 , name : "PCE-AC56 Dual-Band Wireless PCI-E Adapter" }] } , Device { vendor_id : 5348u16 , id : 17338u16 , name : "BCM43602 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17339u16 , name : "BCM43602 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17340u16 , name : "BCM43602 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17347u16 , name : "BCM4366/BCM43465 802.11ac Wave2 4x4 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34555u16 , name : "PCE-AC88" }] } , Device { vendor_id : 5348u16 , id : 17363u16 , name : "BCM43567 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17369u16 , name : "BCM43570 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17372u16 , name : "BCM4355 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17375u16 , name : "BCM4354 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17385u16 , name : "BCM4358 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17388u16 , name : "BCM4356 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17409u16 , name : "BCM4401 100Base-T" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 53u16 , name : "TravelMate 660" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 33063u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 32936u16 , name : "A7V8X motherboard" }] } , Device { vendor_id : 5348u16 , id : 17410u16 , name : "BCM4402 Integrated 10/100BaseT" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17411u16 , name : "BCM4402 V.90 56k Modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17424u16 , name : "BCM4413 iLine32 HomePNA 2.0" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17425u16 , name : "BCM4413 V.90 56k modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17426u16 , name : "BCM4412 10/100BaseT" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17429u16 , name : "BCM4359 802.11ac Dual-Band Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17439u16 , name : "BCM4361 802.11ac Dual-Band Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17440u16 , name : "BCM4361 802.11ac 2.4 GHz Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17441u16 , name : "BCM4361 802.11ac 5 GHz Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17445u16 , name : "BCM4378 802.11ax Dual Band Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17456u16 , name : "BCM44xx CardBus iLine32 HomePNA 2.0" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17458u16 , name : "BCM4432 CardBus 10/100BaseT" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17459u16 , name : "BCM4387 802.11ax Dual Band Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17460u16 , name : "BCM4388 802.11ax Dual Band Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17508u16 , name : "BCM4364 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17544u16 , name : "BCM4377b Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17565u16 , name : "BCM43752 802.11ax Dual Band Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17936u16 , name : "BCM4610 Sentry5 PCI to SB Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17937u16 , name : "BCM4610 Sentry5 iLine32 HomePNA 1.0" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17938u16 , name : "BCM4610 Sentry5 V.90 56k Modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17939u16 , name : "BCM4610 Sentry5 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17940u16 , name : "BCM4610 Sentry5 External Interface" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 17941u16 , name : "BCM4610 Sentry5 USB Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18180u16 , name : "BCM4704 PCI to SB Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18181u16 , name : "BCM4704 Sentry5 802.11b Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18182u16 , name : "BCM4704 Sentry5 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18183u16 , name : "BCM4704 Sentry5 USB Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18184u16 , name : "BCM4704 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18192u16 , name : "BCM4710 Sentry5 PCI to SB Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18193u16 , name : "BCM47xx Sentry5 iLine32 HomePNA 2.0" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18194u16 , name : "BCM47xx V.92 56k modem" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18195u16 , name : "Sentry5 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18196u16 , name : "BCM47xx Sentry5 External Interface" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18197u16 , name : "BCM47xx Sentry5 USB / Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18198u16 , name : "BCM47xx Sentry5 USB Host Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18199u16 , name : "BCM47xx Sentry5 USB Device Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18200u16 , name : "Sentry5 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18201u16 , name : "BCM47xx/53xx RoboSwitch Core" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18202u16 , name : "BCM47xx EHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18208u16 , name : "BCM4712 MIPS CPU" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 18215u16 , name : "BCM4313 802.11bgn Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 16u16 , name : "Inspiron M5010 / XPS 8300" }] } , Device { vendor_id : 5348u16 , id : 18218u16 , name : "BCM47xx xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 21349u16 , name : "BCM5365P Sentry5 Host Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22016u16 , name : "BCM5600 StrataSwitch 24+2 Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22021u16 , name : "BCM5605 StrataSwitch 24+2 Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22037u16 , name : "BCM5615 StrataSwitch 24+2 Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22053u16 , name : "BCM5625 StrataSwitch 24+2 Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22085u16 , name : "BCM5645 StrataSwitch 24+2 Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22128u16 , name : "BCM5670 8-Port 10GE Ethernet Switch Fabric" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22144u16 , name : "BCM5680 G-Switch 8 Port Gigabit Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22160u16 , name : "BCM5690 12-port Multi-Layer Gigabit Ethernet Switch" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22161u16 , name : "BCM5691 GE/10GE 8+2 Gigabit Ethernet Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22162u16 , name : "BCM5692 12-port Multi-Layer Gigabit Ethernet Switch" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22165u16 , name : "BCM5695 12-port + HiGig Multi-Layer Gigabit Ethernet Switch" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22168u16 , name : "BCM5698 12-port Multi-Layer Gigabit Ethernet Switch" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22560u16 , name : "BCM5820 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22561u16 , name : "BCM5821 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22562u16 , name : "BCM5822 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22563u16 , name : "BCM5823 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22564u16 , name : "BCM5824 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22592u16 , name : "BCM5840 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22593u16 , name : "BCM5841 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 22608u16 , name : "BCM5850 Crypto Accelerator" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24199u16 , name : "Valkyrie offload engine" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24200u16 , name : "Viper Offload Engine" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24425u16 , name : "BRCM4378 Bluetooth Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24433u16 , name : "BRCM4387 Bluetooth Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24434u16 , name : "BCM4388 Bluetooth Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 24480u16 , name : "BRCM4377 Bluetooth Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 26725u16 , name : "BCM68650 [Aspen] XGSPON OLT" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 33809u16 , name : "BCM47xx PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 34306u16 , name : "BCM7400/BCM7405 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 36902u16 , name : "CN99xx [ThunderX2] Integrated USB 3.0 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 36903u16 , name : "CN99xx [ThunderX2] Integrated AHCI/SATA 3 Host Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 43224u16 , name : "BCM43224/5 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 43602u16 , name : "BCM43602 802.11ac Wireless LAN SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45168u16 , name : "BCM56070 Switch ASIC [Firelight]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45184u16 , name : "BCM56080 Firelight2 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45426u16 , name : "BCM56172 Hurricane3-MG" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45826u16 , name : "BCM56302 StrataXGS 24x1GE 2x10GE Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45876u16 , name : "BCM56334 StrataXGS 24x1GE 4x10GE Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45936u16 , name : "BCM56370 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45937u16 , name : "BCM56371 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45938u16 , name : "BCM56372 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45941u16 , name : "BCM56375 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45942u16 , name : "BCM56376 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45943u16 , name : "BCM56377 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 45945u16 , name : "Broadcom BCM56379 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 46192u16 , name : "BCM56470 SWITCH ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 46193u16 , name : "BCM56471 SWITCH ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 46194u16 , name : "BCM56472 SWITCH ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47104u16 , name : "BCM56800 StrataXGS 10GE Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47170u16 , name : "BCM56842 Trident 10GE Switch Controller" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47184u16 , name : "BCM56850 Switch ASIC [Trident2]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47188u16 , name : "BCM56854 Trident2 switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47232u16 , name : "BCM56880 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47456u16 , name : "BCM56960 Switch ASIC [Tomahawk]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47504u16 , name : "BCM56990 Switch ASIC [Tomahawk4]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47505u16 , name : "Tomahawk4" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47506u16 , name : "BCM56992 [Tomahawk4]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47507u16 , name : "BCM56993 Switch ASIC [Tomahawk4]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47509u16 , name : "BCM56995 Switch ASIC [Tomahawk4]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47510u16 , name : "BCM56996 106G Switch ASIC [Tomahawk4G]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47512u16 , name : "BCM56998 106G Switch ASIC [Tomahawk4GT]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 47513u16 , name : "BCM56999 106G Switch ASIC [Tomahawk4D]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 51465u16 , name : "BCM78909 Switch ASIC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 55298u16 , name : "BCM58802 Stingray 50Gb Ethernet SoC" , subsystems : & [SubSystem { subvendor : 5348u16 , subdevice : 32801u16 , name : "Stingray Dual-Port 25Gb Ethernet PCIe SmartNIC w16GB DRAM (Part No BCM958802A8046C)" } , SubSystem { subvendor : 5348u16 , subdevice : 32803u16 , name : "PS410T-H04 NetXtreme-S 4x10G 10GBaseT PCIe SmartNIC" } , SubSystem { subvendor : 5348u16 , subdevice : 32804u16 , name : "Stingray Dual-Port 25Gb Ethernet PCIe SmartNIC w4GB DRAM (Part No BCM958802A8044C)" } , SubSystem { subvendor : 5348u16 , subdevice : 32808u16 , name : "Stingray Dual-Port 25Gb Ethernet PCIe SmartNIC w8GB DRAM (Part No BCM958802A8048C)" } , SubSystem { subvendor : 7088u16 , subdevice : 33u16 , name : "HPE SimpliVity Accelerator" }] } , Device { vendor_id : 5348u16 , id : 55300u16 , name : "BCM58804 Stingray 100Gb Ethernet SoC" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63488u16 , name : "BCM78800 Switch ASIC [Trident5-X12]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63744u16 , name : "BCM78900 Switch ASIC [Tomahawk5]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63746u16 , name : "BCM78902 Switch ASIC [Tomahawk5T]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63747u16 , name : "BCM78903 Switch ASIC [Tomahawk5]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63749u16 , name : "BCM78905 Switch ASIC [Tomahawk5+]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63760u16 , name : "BCM78910 Switch ASIC [Tomahawk6]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63764u16 , name : "BCM78914 Switch ASIC [Tomahawk6]" , subsystems : & [] } , Device { vendor_id : 5348u16 , id : 63767u16 , name : "BCM78917 Switch ASIC [Tomahawk6]" , subsystems : & [] }] }),
        (3669, Vendor { id : 3669u16 , name : "HaSoTec GmbH" , devices : & [] }),
        (4939, Vendor { id : 4939u16 , name : "ARK Research Corp." , devices : & [] }),
        (6395, Vendor { id : 6395u16 , name : "Resilience Corporation" , devices : & [] }),
        (4333, Vendor { id : 4333u16 , name : "Ascii Corporation" , devices : & [Device { vendor_id : 4333u16 , id : 29456u16 , name : "V7310" , subsystems : & [] }] }),
        (26214, Vendor { id : 26214u16 , name : "Decision Computer International Co." , devices : & [Device { vendor_id : 26214u16 , id : 1u16 , name : "PCCOM4" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 2u16 , name : "PCCOM8" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 4u16 , name : "PCCOM2" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 257u16 , name : "PCI 8255/8254 I/O Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 512u16 , name : "12-bit AD/DA Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 513u16 , name : "14-bit AD/DA Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 4113u16 , name : "Industrial Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 4129u16 , name : "8 photo couple 8 relay Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 4130u16 , name : "4 photo couple 4 relay Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 4133u16 , name : "16 photo couple 16 relay Card" , subsystems : & [] } , Device { vendor_id : 26214u16 , id : 16384u16 , name : "WatchDog Card" , subsystems : & [] }] }),
        (8080, Vendor { id : 8080u16 , name : "Quside Technologies" , devices : & [Device { vendor_id : 8080u16 , id : 28708u16 , name : "QRNG PCIe Device" , subsystems : & [] } , Device { vendor_id : 8080u16 , id : 28709u16 , name : "QRNG PCIe Device" , subsystems : & [] }] }),
        (5587, Vendor { id : 5587u16 , name : "NDS Technologies Israel Ltd" , devices : & [] }),
        (6931, Vendor { id : 6931u16 , name : "Jaton Corp" , devices : & [] }),
        (7415, Vendor { id : 7415u16 , name : "Subspace Dynamics" , devices : & [] }),
        (43533, Vendor { id : 43533u16 , name : "iTuner" , devices : & [] }),
        (4658, Vendor { id : 4658u16 , name : "GPT Limited" , devices : & [] }),
        (4834, Vendor { id : 4834u16 , name : "Datum Inc. Bancomm-Timing Division" , devices : & [] }),
        (4897, Vendor { id : 4897u16 , name : "Arcobel Graphics BV" , devices : & [] }),
        (5302, Vendor { id : 5302u16 , name : "Quantum Data Corp." , devices : & [] }),
        (4577, Vendor { id : 4577u16 , name : "GEC Plessey Semi Inc." , devices : & [] }),
        (4778, Vendor { id : 4778u16 , name : "SDL Communications, Inc." , devices : & [] }),
        (20686, Vendor { id : 20686u16 , name : "System-on-Chip Engineering S.L." , devices : & [Device { vendor_id : 20686u16 , id : 1u16 , name : "RELY-MIL-XMC-TSN-SWITCH" , subsystems : & [] } , Device { vendor_id : 20686u16 , id : 256u16 , name : "XMC_AV-Dual-ETH" , subsystems : & [] } , Device { vendor_id : 20686u16 , id : 257u16 , name : "XMC_AV-ETSN" , subsystems : & [] } , Device { vendor_id : 20686u16 , id : 258u16 , name : "XMC_AV-AFDX" , subsystems : & [] }] }),
        (4373, Vendor { id : 4373u16 , name : "3D Labs" , devices : & [] }),
        (6143, Vendor { id : 6143u16 , name : "Benq Corporation" , devices : & [] }),
        (5070, Vendor { id : 5070u16 , name : "Cocom A/S" , devices : & [] }),
        (4587, Vendor { id : 4587u16 , name : "Formation Inc." , devices : & [] }),
        (5596, Vendor { id : 5596u16 , name : "Litronic Inc" , devices : & [Device { vendor_id : 5596u16 , id : 1u16 , name : "Argus 300 PCI Cryptography Module" , subsystems : & [] }] }),
        (5187, Vendor { id : 5187u16 , name : "Unibrain S.A." , devices : & [] }),
        (7728, Vendor { id : 7728u16 , name : "Sophgo" , devices : & [Device { vendor_id : 7728u16 , id : 5764u16 , name : "BM1684 [Sophon Series Deep Learning Accelerator]" , subsystems : & [] } , Device { vendor_id : 7728u16 , id : 8258u16 , name : "SG2042 Root Complex" , subsystems : & [] }] }),
        (4349, Vendor { id : 4349u16 , name : "Soyo Computer, Inc" , devices : & [] }),
        (5449, Vendor { id : 5449u16 , name : "INTERCONNECT Systems Solutions" , devices : & [] }),
        (8441, Vendor { id : 8441u16 , name : "Shenzhen Silicon Dynamic Networks Co., Ltd." , devices : & [] }),
        (4197, Vendor { id : 4197u16 , name : "Texas Microsystems" , devices : & [] }),
        (4366, Vendor { id : 4366u16 , name : "CPU Technology" , devices : & [] }),
        (5397, Vendor { id : 5397u16 , name : "Advent design" , devices : & [] }),
        (4501, Vendor { id : 4501u16 , name : "Ratoc System Inc" , devices : & [] }),
        (4169, Vendor { id : 4169u16 , name : "Fountain Technologies, Inc." , devices : & [] }),
        (6627, Vendor { id : 6627u16 , name : "DDRdrive LLC" , devices : & [Device { vendor_id : 6627u16 , id : 22529u16 , name : "DDRdrive X1" , subsystems : & [] } , Device { vendor_id : 6627u16 , id : 22536u16 , name : "DDRdrive X8" , subsystems : & [] } , Device { vendor_id : 6627u16 , id : 56658u16 , name : "DDRdrive X1-30" , subsystems : & [] }] }),
        (6091, Vendor { id : 6091u16 , name : "Qualcomm Technologies, Inc" , devices : & [Device { vendor_id : 6091u16 , id : 1u16 , name : "AGN100 802.11 a/b/g True MIMO Wireless Card" , subsystems : & [SubSystem { subvendor : 4436u16 , subdevice : 823u16 , name : "WLI-CB-G108" } , SubSystem { subvendor : 4997u16 , subdevice : 23552u16 , name : "WGM511 Pre-N 802.11g Wireless CardBus Adapter" } , SubSystem { subvendor : 5354u16 , subdevice : 24833u16 , name : "CQW-NS108G" } , SubSystem { subvendor : 5943u16 , subdevice : 55u16 , name : "WPC54GX Wireless-G Notebook Adapter with SRX" } , SubSystem { subvendor : 5943u16 , subdevice : 69u16 , name : "WMP54GX v1 802.11g Wireless-G PCI Adapter with SRX" }] } , Device { vendor_id : 6091u16 , id : 2u16 , name : "AGN300 802.11 a/b/g True MIMO Wireless Card" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4207u16 , name : "WL-106gM 240 MIMO Wireless CardBus Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 27904u16 , name : "WPNT511 RangeMax 240 Mbps Wireless CardBus Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 84u16 , name : "WPC54GX4 v1 802.11g Wireless-G Notebook Adapter with SRX400" } , SubSystem { subvendor : 5943u16 , subdevice : 86u16 , name : "WMP54GX4 Wireless-G PCI Adapter with SRX400" }] } , Device { vendor_id : 6091u16 , id : 260u16 , name : "APQ8096 PCIe Root Complex [Snapdragon 820]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 261u16 , name : "MSM8998 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 262u16 , name : "SDM850 PCIe Root Complex [Snapdragon 850]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 263u16 , name : "SDM850 PCIe Root Port [Snapdragon 850]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 264u16 , name : "SM8150 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 265u16 , name : "SA8195P PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 267u16 , name : "SM8250 PCIe Root Complex [Snapdragon 865/870 5G]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 268u16 , name : "SM8350 PCIe Root Complex [Snapdragon 888]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 270u16 , name : "SC8280XP PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 272u16 , name : "SM8475 PCIe Root Complex [Snapdragon 8+ Gen 1]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 273u16 , name : "SC8380XP PCIe Root Complex [Snapdragon X / X Plus / X Elite]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 275u16 , name : "SM8550/SM8650 PCIe Root Complex [Snapdragon 8 Gen 2/Gen 3]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 768u16 , name : "MDM9x35 LTE Modem [Snapdragon X7]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 769u16 , name : "MDM9x45 LTE Modem [Snapdragon X12]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 770u16 , name : "MDM9x55 LTE Modem [Snapdragon X16]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 772u16 , name : "SDX24 [Snapdragon X24 4G]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 774u16 , name : "SDX55 [Snapdragon X55 5G]" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 776u16 , name : "SDX61/SDX62/SDX65 [Snapdragon 5G X6X-series]" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 57666u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57667u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57668u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57669u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57670u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57680u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57681u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57682u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57683u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57684u16 , name : "T99W696 5G Modem [Snapdragon X61]" } , SubSystem { subvendor : 4187u16 , subdevice : 57685u16 , name : "T99W696 5G Modem [Snapdragon X61]" }] } , Device { vendor_id : 6091u16 , id : 1024u16 , name : "Datacenter Technologies QDF2432 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 1025u16 , name : "Datacenter Technologies QDF2400 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4096u16 , name : "QCS405 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4099u16 , name : "IPQ6018 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4353u16 , name : "QCA6390 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4355u16 , name : "QCNFA765 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4356u16 , name : "QCN6024/9024/9074 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4359u16 , name : "WCN785x Wi-Fi 7(802.11be) 320MHz 2x2 [FastConnect 7800]" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 57591u16 , name : "High Band Simultaneous Wireless Network Adapter" }] } , Device { vendor_id : 6091u16 , id : 4360u16 , name : "IPQ95xx/97xx PCIe Root Port" , subsystems : & [] } , Device { vendor_id : 6091u16 , id : 4361u16 , name : "QCN62xx/92xx Wireless Network Adapter" , subsystems : & [] }] }),
        (6518, Vendor { id : 6518u16 , name : "TRENDnet" , devices : & [] }),
        (4837, Vendor { id : 4837u16 , name : "Apex Semiconductor Inc" , devices : & [] }),
        (4306, Vendor { id : 4306u16 , name : "Molex Incorporated" , devices : & [] }),
        (5129, Vendor { id : 5129u16 , name : "Timedia Technology Co Ltd" , devices : & [Device { vendor_id : 5129u16 , id : 29032u16 , name : "PCI2S550 (Dual 16550 UART)" , subsystems : & [SubSystem { subvendor : 5129u16 , subdevice : 2u16 , name : "SER4036A3V (2x RS232 port)" } , SubSystem { subvendor : 5129u16 , subdevice : 16423u16 , name : "SER4027A (1x RS232 port)" } , SubSystem { subvendor : 5129u16 , subdevice : 16439u16 , name : "SER4037A(L) [SUNIX SUN1889] (2x RS232 port)" } , SubSystem { subvendor : 5129u16 , subdevice : 16470u16 , name : "SER4056A (4x RS232)" } , SubSystem { subvendor : 5129u16 , subdevice : 20519u16 , name : "SER4027D" } , SubSystem { subvendor : 5129u16 , subdevice : 20535u16 , name : "SER4037D (2x RS232 port)" } , SubSystem { subvendor : 5129u16 , subdevice : 20582u16 , name : "SER4066R (8x RS232)" } , SubSystem { subvendor : 5129u16 , subdevice : 24662u16 , name : "SER4056D (4x RS232 port)" }] } , Device { vendor_id : 5129u16 , id : 29288u16 , name : "SUN1888 (Dual IEEE1284 parallel port)" , subsystems : & [SubSystem { subvendor : 5129u16 , subdevice : 259u16 , name : "PAR4008A" } , SubSystem { subvendor : 5129u16 , subdevice : 260u16 , name : "PAR4018A" }] }] }),
        (4682, Vendor { id : 4682u16 , name : "AEG Electrocom GmbH" , devices : & [] }),
        (28, Vendor { id : 28u16 , name : "PEAK-System Technik GmbH" , devices : & [Device { vendor_id : 28u16 , id : 1u16 , name : "PCAN-PCI CAN-Bus controller" , subsystems : & [SubSystem { subvendor : 28u16 , subdevice : 4u16 , name : "2 Channel CAN Bus SJC1000" } , SubSystem { subvendor : 28u16 , subdevice : 5u16 , name : "2 Channel CAN Bus SJC1000 (Optically Isolated)" }] }] }),
        (4484, Vendor { id : 4484u16 , name : "Forks Inc" , devices : & [] }),
        (4786, Vendor { id : 4786u16 , name : "General Signal Networks" , devices : & [] }),
        (5478, Vendor { id : 5478u16 , name : "Ardent Technologies Inc" , devices : & [] }),
        (8964, Vendor { id : 8964u16 , name : "Colorgraphic Communications Corp." , devices : & [] }),
        (5213, Vendor { id : 5213u16 , name : "Gallant Computer Inc" , devices : & [] }),
        (7538, Vendor { id : 7538u16 , name : "Xiaomi" , devices : & [] }),
        (21586, Vendor { id : 21586u16 , name : "SCANLAB AG" , devices : & [Device { vendor_id : 21586u16 , id : 13379u16 , name : "RTC4" , subsystems : & [] }] }),
        (26777, Vendor { id : 26777u16 , name : "ZT Systems" , devices : & [] }),
        (5300, Vendor { id : 5300u16 , name : "PHILIPS Business Electronics B.V." , devices : & [] }),
        (4692, Vendor { id : 4692u16 , name : "Linear Systems Ltd." , devices : & [Device { vendor_id : 4692u16 , id : 101u16 , name : "DVB Master FD" , subsystems : & [] } , Device { vendor_id : 4692u16 , id : 124u16 , name : "DVB Master Quad/o" , subsystems : & [] }] }),
        (34952, Vendor { id : 34952u16 , name : "Silicon Magic" , devices : & [Device { vendor_id : 34952u16 , id : 34052u16 , name : "AVMatrix VC42 4-port HDMI Capture" , subsystems : & [] } , Device { vendor_id : 34952u16 , id : 34177u16 , name : "AVMatrix VC12 4K HDMI Capture" , subsystems : & [] }] }),
        (4507, Vendor { id : 4507u16 , name : "Omega Micro Inc." , devices : & [Device { vendor_id : 4507u16 , id : 4641u16 , name : "82C092G" , subsystems : & [] }] }),
        (5513, Vendor { id : 5513u16 , name : "Atlantek Microsystems Pty Ltd" , devices : & [Device { vendor_id : 5513u16 , id : 8u16 , name : "Leutron Vision PicPortExpress CL" , subsystems : & [] } , Device { vendor_id : 5513u16 , id : 9u16 , name : "Leutron Vision PicPortExpress CL Stereo" , subsystems : & [] }] }),
        (4777, Vendor { id : 4777u16 , name : "Xiotech Corporation" , devices : & [] }),
        (16435, Vendor { id : 16435u16 , name : "Addtron Technology Co, Inc." , devices : & [Device { vendor_id : 16435u16 , id : 4960u16 , name : "RTL8139 Ethernet" , subsystems : & [] }] }),
        (5614, Vendor { id : 5614u16 , name : "In Win Development Inc" , devices : & [] }),
        (4328, Vendor { id : 4328u16 , name : "Applied Micro Circuits Corp." , devices : & [Device { vendor_id : 4328u16 , id : 4210u16 , name : "INES GPIB-PCI (AMCC5920 based)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 8209u16 , name : "Q-Motion Video Capture/Edit board" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 18256u16 , name : "S5930 [Matchmaker]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 22816u16 , name : "S5920" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32797u16 , name : "Roper Scientific PCI TAXI interface" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32835u16 , name : "LANai4.x [Myrinet LANai interface chip]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32866u16 , name : "S5933_PARASTATION" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32893u16 , name : "S5933 [Matchmaker]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32897u16 , name : "GPIB interface card [IOtech Inc. PCI488]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32904u16 , name : "Kongsberg Spacetec Format Synchronizer" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32905u16 , name : "Kongsberg Spacetec Serial Output Board" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32924u16 , name : "S5933_HEPC3" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32953u16 , name : "Harmonix Hi-Card P8 (4x active ISDN BRI)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32983u16 , name : "PCI-9112" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32984u16 , name : "PCI-7200" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32985u16 , name : "PCI-9118" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 32986u16 , name : "PCI-9812" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33020u16 , name : "APCI1500 Signal processing controller (16 dig. inputs + 16 dig. outputs)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33050u16 , name : "PCI-IEEE1355-DS-DE Interface" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33100u16 , name : "Fastcom ESCC-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33136u16 , name : "S5933 [Matchmaker] (Chipset Development Tool)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33254u16 , name : "Multimedia video controller" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33421u16 , name : "APCI3001 Signal processing controller (up to 16 analog inputs)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33425u16 , name : "Fastcom 232/8-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33476u16 , name : "Fastcom 422/4-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33477u16 , name : "Fastcom 422/2-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33478u16 , name : "Fastcom IG422/1-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33479u16 , name : "Fastcom IG232/2-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33482u16 , name : "Fastcom 232/4-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33499u16 , name : "AJA HDNTV HD SDI Framestore" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33506u16 , name : "Fastcom DIO24H-PCI (Commtech, Inc.)" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33798u16 , name : "PCIcanx/PCIcan CAN interface [Kvaser AB]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 33799u16 , name : "PCIcan II CAN interface (A1021, PCB-07, PCB-08) [Kvaser AB]" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 34897u16 , name : "S5933 on Innes Corp FM Radio Capture card" , subsystems : & [] } , Device { vendor_id : 4328u16 , id : 57348u16 , name : "X-Gene PCIe bridge" , subsystems : & [] }] }),
        (5194, Vendor { id : 5194u16 , name : "Adlink Technology" , devices : & [Device { vendor_id : 5194u16 , id : 25096u16 , name : "PCI-6208V" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 29264u16 , name : "PCI-7250" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 29334u16 , name : "PCI-7296" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 29746u16 , name : "PCI-7432" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 29747u16 , name : "PCI-7433" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 29748u16 , name : "PCI-7434" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 30785u16 , name : "PCI-7841" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 33075u16 , name : "PCI-8133" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 33124u16 , name : "PCI-8164" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 34132u16 , name : "PCI-8554" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 37137u16 , name : "PCI-9111" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 37139u16 , name : "PCI-9113" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 37140u16 , name : "PCI-9114" , subsystems : & [] } , Device { vendor_id : 5194u16 , id : 40961u16 , name : "ADi-BSEC" , subsystems : & [] }] }),
        (4892, Vendor { id : 4892u16 , name : "Nippon Electro-Sensory Devices Corp" , devices : & [] }),
        (4825, Vendor { id : 4825u16 , name : "Aculab PLC" , devices : & [Device { vendor_id : 4825u16 , id : 2u16 , name : "PCI Prosody" , subsystems : & [] } , Device { vendor_id : 4825u16 , id : 4u16 , name : "cPCI Prosody" , subsystems : & [] } , Device { vendor_id : 4825u16 , id : 5u16 , name : "Aculab E1/T1 PCI card" , subsystems : & [] } , Device { vendor_id : 4825u16 , id : 4216u16 , name : "Prosody X class e1000 device" , subsystems : & [SubSystem { subvendor : 4825u16 , subdevice : 13u16 , name : "Prosody X PCI" } , SubSystem { subvendor : 4825u16 , subdevice : 14u16 , name : "Prosody X cPCI" }] }] }),
        (4756, Vendor { id : 4756u16 , name : "Rhetorex, Inc." , devices : & [] }),
        (5278, Vendor { id : 5278u16 , name : "Mapletree Networks Inc." , devices : & [] }),
        (7829, Vendor { id : 7829u16 , name : "Solid State Storage Technology Corporation" , devices : & [Device { vendor_id : 7829u16 , id : 4096u16 , name : "XA1 Series NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4097u16 , name : "CA6-8D512 NVMe SSD M.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4098u16 , name : "NVMe SSD [3DNAND] 2.5\" U.2 (LJ1)" , subsystems : & [SubSystem { subvendor : 7829u16 , subdevice : 4353u16 , name : "NVMe SSD [3DNAND] 2.5\" U.2 (LJ1)" } , SubSystem { subvendor : 7840u16 , subdevice : 22070u16 , name : "TP1500 Series U.2 NVMe Datacenter SSD" }] } , Device { vendor_id : 7829u16 , id : 4099u16 , name : "CLR-8W512 NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4101u16 , name : "PLEXTOR M10P(GN) NVMe SSD M.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4102u16 , name : "CA8 Series NVMe SSD M.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4103u16 , name : "CL4 Series NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4104u16 , name : "CL5-8D512 NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4107u16 , name : "XB2-311024 NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4108u16 , name : "CL6 Series NVMe SSD M.2 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4109u16 , name : "PJ1 Series NVMe SSD" , subsystems : & [SubSystem { subvendor : 7829u16 , subdevice : 0u16 , name : "M.2 2280 480 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 1u16 , name : "M.2 2280 960 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 2u16 , name : "M.2 2280 1920 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 3u16 , name : "M.2 22110 3840 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 4u16 , name : "U.2 1920 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 5u16 , name : "U.2 3840 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 6u16 , name : "U.2 7680 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 7u16 , name : "U.2 1600 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 8u16 , name : "U.2 3200 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 9u16 , name : "U.2 6400 GB" }] } , Device { vendor_id : 7829u16 , id : 4111u16 , name : "EJ5-2W3840 NVMe SSD U.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 4112u16 , name : "CX3 Series NVMe SSD" , subsystems : & [SubSystem { subvendor : 7829u16 , subdevice : 0u16 , name : "M.2 2280 480 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 1u16 , name : "M.2 2280 960 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 2u16 , name : "M.2 2280 1,920 GB" } , SubSystem { subvendor : 7829u16 , subdevice : 3u16 , name : "M.2 2280 3,840 GB" }] } , Device { vendor_id : 7829u16 , id : 13568u16 , name : "CA5-8D256 NVMe SSD M.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 13809u16 , name : "PLEXTOR M9PGN Plus NVMe SSD M.2" , subsystems : & [] } , Device { vendor_id : 7829u16 , id : 37120u16 , name : "CL1-3D256-Q11 NVMe SSD M.2" , subsystems : & [] }] }),
        (7611, Vendor { id : 7611u16 , name : "NGD Systems, Inc." , devices : & [Device { vendor_id : 7611u16 , id : 4129u16 , name : "ICS-8100 NVMe SSD" , subsystems : & [] }] }),
        (5029, Vendor { id : 5029u16 , name : "Audio Digital Imaging Inc" , devices : & [] }),
        (4491, Vendor { id : 4491u16 , name : "Hypertec Pty Limited" , devices : & [] }),
        (7089, Vendor { id : 7089u16 , name : "Seagate Technology PLC" , devices : & [Device { vendor_id : 7089u16 , id : 93u16 , name : "Nytro PCIe Flash Storage" , subsystems : & [SubSystem { subvendor : 7089u16 , subdevice : 25857u16 , name : "Nytro XP6500-8A1536 1.5TB" } , SubSystem { subvendor : 7089u16 , subdevice : 25858u16 , name : "Nytro XP6500-8A2048" } , SubSystem { subvendor : 7089u16 , subdevice : 25859u16 , name : "Nytro XP6500-8A4096" } , SubSystem { subvendor : 7089u16 , subdevice : 25873u16 , name : "Nytro XH6550-2GB DRAM" } , SubSystem { subvendor : 7089u16 , subdevice : 25874u16 , name : "Nytro XH6550-8GB DRAM" } , SubSystem { subvendor : 7089u16 , subdevice : 25889u16 , name : "Nytro XP6500-8A1536 1.5TB" } , SubSystem { subvendor : 7089u16 , subdevice : 25890u16 , name : "Nytro XP6500-8A2048" } , SubSystem { subvendor : 7089u16 , subdevice : 25891u16 , name : "Nytro XP6500-8A4096" }] } , Device { vendor_id : 7089u16 , id : 256u16 , name : "Nytro Flash Storage" , subsystems : & [SubSystem { subvendor : 7089u16 , subdevice : 257u16 , name : "Nytro XF1440" } , SubSystem { subvendor : 7089u16 , subdevice : 259u16 , name : "Nytro 5000" } , SubSystem { subvendor : 7089u16 , subdevice : 261u16 , name : "Nytro 5020" } , SubSystem { subvendor : 7089u16 , subdevice : 262u16 , name : "Nytro 5020 TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 263u16 , name : "Nytro 5320" } , SubSystem { subvendor : 7089u16 , subdevice : 264u16 , name : "Nytro 5320 TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 289u16 , name : "Nytro XM1440" } , SubSystem { subvendor : 7089u16 , subdevice : 291u16 , name : "Nytro 5000" } , SubSystem { subvendor : 7089u16 , subdevice : 293u16 , name : "Nytro 5020" } , SubSystem { subvendor : 7089u16 , subdevice : 294u16 , name : "Nytro 5020" } , SubSystem { subvendor : 7089u16 , subdevice : 295u16 , name : "Nytro 5320 M.2" } , SubSystem { subvendor : 7089u16 , subdevice : 296u16 , name : "Nytro 5320 M.2 TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 305u16 , name : "Nytro 5320 M.2" } , SubSystem { subvendor : 7089u16 , subdevice : 306u16 , name : "Nytro 5320 M.2 TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 321u16 , name : "Nytro 5320 E1.S" } , SubSystem { subvendor : 7089u16 , subdevice : 322u16 , name : "Nytro 5320 E1.S TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 337u16 , name : "Nytro 5520" } , SubSystem { subvendor : 7089u16 , subdevice : 338u16 , name : "Nytro 5520 TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 339u16 , name : "Nytro 5050H" } , SubSystem { subvendor : 7089u16 , subdevice : 340u16 , name : "Nytro 5050H TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 341u16 , name : "Nytro 5050M" } , SubSystem { subvendor : 7089u16 , subdevice : 342u16 , name : "Nytro 5050M TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 343u16 , name : "Nytro 5050M 7mm" } , SubSystem { subvendor : 7089u16 , subdevice : 344u16 , name : "Nytro 5050M TCG 7mm" } , SubSystem { subvendor : 7089u16 , subdevice : 345u16 , name : "Nytro 5060M" } , SubSystem { subvendor : 7089u16 , subdevice : 352u16 , name : "Nytro 5060M TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 353u16 , name : "Nytro 5060M 7mm" } , SubSystem { subvendor : 7089u16 , subdevice : 354u16 , name : "Nytro 5060M TCG 7mm" } , SubSystem { subvendor : 7089u16 , subdevice : 355u16 , name : "Nytro 5060H" } , SubSystem { subvendor : 7089u16 , subdevice : 356u16 , name : "Nytro 5060H TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 357u16 , name : "Nytro 5060H E3.S 1T" } , SubSystem { subvendor : 7089u16 , subdevice : 358u16 , name : "Nytro 5060H E3.S 1T TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 359u16 , name : "Nytro 5060H E3.L 1T" } , SubSystem { subvendor : 7089u16 , subdevice : 360u16 , name : "Nytro 5060H E3.L 1T TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 361u16 , name : "Nytro 5060M E3.S 1T" } , SubSystem { subvendor : 7089u16 , subdevice : 368u16 , name : "Nytro 5060M E3.S 1T TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 369u16 , name : "Nytro 5060M E3.L 1T" } , SubSystem { subvendor : 7089u16 , subdevice : 370u16 , name : "Nytro 5060M E3.L 1T TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 371u16 , name : "Nytro 5060M E1.S" } , SubSystem { subvendor : 7089u16 , subdevice : 372u16 , name : "Nytro 5060M E1.S TCG" } , SubSystem { subvendor : 7089u16 , subdevice : 373u16 , name : "Nytro 5350S - 15mm" } , SubSystem { subvendor : 7089u16 , subdevice : 374u16 , name : "Nytro 5350S TCG - 15mm" } , SubSystem { subvendor : 7089u16 , subdevice : 375u16 , name : "Nytro 5360S - 15mm" } , SubSystem { subvendor : 7089u16 , subdevice : 376u16 , name : "Nytro 5360S TCG - 15mm" } , SubSystem { subvendor : 7089u16 , subdevice : 377u16 , name : "Nytro 5360S - E3.S" } , SubSystem { subvendor : 7089u16 , subdevice : 384u16 , name : "Nytro 5360S TCG - E3.S" } , SubSystem { subvendor : 7089u16 , subdevice : 385u16 , name : "Nytro 5060H" } , SubSystem { subvendor : 7089u16 , subdevice : 417u16 , name : "Nytro XP7102" }] } , Device { vendor_id : 7089u16 , id : 339u16 , name : "Nytro 5x50 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 341u16 , name : "Nytro 5x50 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 8192u16 , name : "PCIe Gen4 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20498u16 , name : "FireCuda/IronWolf 510 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20499u16 , name : "BarraCuda Q5 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20502u16 , name : "FireCuda 520/IronWolf 525 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20504u16 , name : "E18 PCIe SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20505u16 , name : "BarraCuda PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20513u16 , name : "PCIe Gen4 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20518u16 , name : "FireCuda 540 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20519u16 , name : "BarraCuda 530 SSD / LaCie Rugged SSD Pro5" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20736u16 , name : "PCIe Gen3 SSD" , subsystems : & [] } , Device { vendor_id : 7089u16 , id : 20737u16 , name : "PCIe Gen5 SSD" , subsystems : & [] }] }),
        (6593, Vendor { id : 6593u16 , name : "Exegy Inc." , devices : & [] }),
        (4551, Vendor { id : 4551u16 , name : "D.C.M. Data Systems" , devices : & [] }),
        (5142, Vendor { id : 5142u16 , name : "Multiwave Innovation pte Ltd" , devices : & [] }),
        (7769, Vendor { id : 7769u16 , name : "Oxford Nanopore Technologies" , devices : & [Device { vendor_id : 7769u16 , id : 1u16 , name : "MinION Mk1C" , subsystems : & [] }] }),
        (4220, Vendor { id : 4220u16 , name : "LG Electronics [Lucky Goldstar Co. Ltd]" , devices : & [] }),
        (5576, Vendor { id : 5576u16 , name : "Penta Media Co Ltd" , devices : & [] }),
        (20563, Vendor { id : 20563u16 , name : "Voyetra Technologies" , devices : & [Device { vendor_id : 20563u16 , id : 8208u16 , name : "Daytona Audio Adapter" , subsystems : & [] }] }),
        (4804, Vendor { id : 4804u16 , name : "Connect Tech Inc" , devices : & [Device { vendor_id : 4804u16 , id : 1u16 , name : "Blue HEAT/PCI 8 (RS232/CL/RJ11)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 2u16 , name : "Blue HEAT/PCI 4 (RS232)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 3u16 , name : "Blue HEAT/PCI 2 (RS232)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 4u16 , name : "Blue HEAT/PCI 8 (UNIV, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 5u16 , name : "Blue HEAT/PCI 4+4/6+2 (UNIV, RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 6u16 , name : "Blue HEAT/PCI 4 (OPTO, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 7u16 , name : "Blue HEAT/PCI 2+2 (RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 8u16 , name : "Blue HEAT/PCI 2 (OPTO, Tx, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 9u16 , name : "Blue HEAT/PCI 2+6 (RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 10u16 , name : "Blue HEAT/PCI 8 (Tx, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 11u16 , name : "Blue HEAT/PCI 4 (Tx, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 12u16 , name : "Blue HEAT/PCI 2 (20 MHz, RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 13u16 , name : "Blue HEAT/PCI 2 PTM" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 256u16 , name : "NT960/PCI" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 513u16 , name : "cPCI Titan - 2 Port" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 514u16 , name : "cPCI Titan - 4 Port" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 768u16 , name : "CTI PCI UART 2 (RS232)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 769u16 , name : "CTI PCI UART 4 (RS232)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 770u16 , name : "CTI PCI UART 8 (RS232)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 784u16 , name : "CTI PCI UART 1+1 (RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 785u16 , name : "CTI PCI UART 2+2 (RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 786u16 , name : "CTI PCI UART 4+4 (RS232/485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 800u16 , name : "CTI PCI UART 2" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 801u16 , name : "CTI PCI UART 4" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 802u16 , name : "CTI PCI UART 8" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 816u16 , name : "CTI PCI UART 2 (RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 817u16 , name : "CTI PCI UART 4 (RS485)" , subsystems : & [] } , Device { vendor_id : 4804u16 , id : 818u16 , name : "CTI PCI UART 8 (RS485)" , subsystems : & [] }] }),
        (13587, Vendor { id : 13587u16 , name : "ARCOM Control Systems Ltd" , devices : & [] }),
        (6147, Vendor { id : 6147u16 , name : "ProdaSafe GmbH" , devices : & [] }),
        (5013, Vendor { id : 5013u16 , name : "Ambicom Inc" , devices : & [] }),
        (6472, Vendor { id : 6472u16 , name : "Alpha Networks Inc." , devices : & [] }),
        (33006, Vendor { id : 33006u16 , name : "InnoTek Systemberatung GmbH" , devices : & [Device { vendor_id : 33006u16 , id : 48879u16 , name : "VirtualBox Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 33006u16 , id : 51966u16 , name : "VirtualBox Guest Service" , subsystems : & [] }] }),
        (5512, Vendor { id : 5512u16 , name : "Solidum Systems Corp" , devices : & [] }),
        (4433, Vendor { id : 4433u16 , name : "JAE Electronics Inc." , devices : & [] }),
        (6513, Vendor { id : 6513u16 , name : "AGEIA Technologies, Inc." , devices : & [Device { vendor_id : 6513u16 , id : 0u16 , name : "Physics Processing Unit [PhysX] 100 Series PCI Express Card" , subsystems : & [] } , Device { vendor_id : 6513u16 , id : 4113u16 , name : "Physics Processing Unit [PhysX] 100 Series PCI Card" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1u16 , name : "PhysX P1" }] } , Device { vendor_id : 6513u16 , id : 4129u16 , name : "Physics Processing Unit [PhysX] 200 Series PCI Express Card" , subsystems : & [] }] }),
        (4390, Vendor { id : 4390u16 , name : "Vigra" , devices : & [] }),
        (4622, Vendor { id : 4622u16 , name : "Cyclades Corporation" , devices : & [Device { vendor_id : 4622u16 , id : 256u16 , name : "Cyclom-Y below first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 257u16 , name : "Cyclom-Y above first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 258u16 , name : "Cyclom-4Y below first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 259u16 , name : "Cyclom-4Y above first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 260u16 , name : "Cyclom-8Y below first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 261u16 , name : "Cyclom-8Y above first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 512u16 , name : "Cyclades-Z below first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 513u16 , name : "Cyclades-Z above first megabyte" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 768u16 , name : "PC300/RSV or /X21 (2 ports)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 769u16 , name : "PC300/RSV or /X21 (1 port)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 784u16 , name : "PC300/TE (2 ports)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 785u16 , name : "PC300/TE (1 port)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 800u16 , name : "PC300/TE-M (2 ports)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 801u16 , name : "PC300/TE-M (1 port)" , subsystems : & [] } , Device { vendor_id : 4622u16 , id : 1024u16 , name : "PC400" , subsystems : & [] }] }),
        (5045, Vendor { id : 5045u16 , name : "ARM" , devices : & [Device { vendor_id : 5045u16 , id : 768u16 , name : "Integrated [AGI CPU] PCIe/CXL Root Port" , subsystems : & [] } , Device { vendor_id : 5045u16 , id : 769u16 , name : "Integrated [AGI CPU] Platform PCIe Root Port" , subsystems : & [] }] }),
        (4862, Vendor { id : 4862u16 , name : "ESD Electronic System Design GmbH" , devices : & [] }),
        (5546, Vendor { id : 5546u16 , name : "Moreton Bay" , devices : & [] }),
        (7983, Vendor { id : 7983u16 , name : "China Mobile (Hangzhou) Information Technology Co.Ltd." , devices : & [Device { vendor_id : 7983u16 , id : 5395u16 , name : "DERA MENG NVMe Controller" , subsystems : & [SubSystem { subvendor : 7983u16 , subdevice : 24851u16 , name : "KM660 U.2 1.6TB NVMe SSD" } , SubSystem { subvendor : 7983u16 , subdevice : 24852u16 , name : "KM560 U.2 1.92TB NVMe SSD" } , SubSystem { subvendor : 7983u16 , subdevice : 24853u16 , name : "KM660 U.2 3.2TB NVMe SSD" } , SubSystem { subvendor : 7983u16 , subdevice : 24854u16 , name : "KM560 U.2 3.84TB NVMe SSD" } , SubSystem { subvendor : 7983u16 , subdevice : 24856u16 , name : "KM560 U.2 7.68TB NVMe SSD" }] }] }),
        (5067, Vendor { id : 5067u16 , name : "Yano Electric Co Ltd" , devices : & [] }),
        (5392, Vendor { id : 5392u16 , name : "BEHAVIOR TECH Computer Corp" , devices : & [] }),
        (5323, Vendor { id : 5323u16 , name : "Billionton Systems Inc" , devices : & [] }),
        (4467, Vendor { id : 4467u16 , name : "Adobe Systems, Inc" , devices : & [] }),
        (6446, Vendor { id : 6446u16 , name : "TransDimension" , devices : & [] }),
        (4102, Vendor { id : 4102u16 , name : "Reply Group" , devices : & [] }),
        (8100, Vendor { id : 8100u16 , name : "Shandong SinoChip Semiconductors Co., Ltd" , devices : & [] }),
        (6836, Vendor { id : 6836u16 , name : "Distributed Management Task Force, Inc. (DMTF)" , devices : & [] }),
        (57344, Vendor { id : 57344u16 , name : "Winbond" , devices : & [Device { vendor_id : 57344u16 , id : 57344u16 , name : "W89C940" , subsystems : & [] }] }),
        (5160, Vendor { id : 5160u16 , name : "Edec Co Ltd" , devices : & [] }),
        (6083, Vendor { id : 6083u16 , name : "Protogate, Inc." , devices : & [] }),
        (5346, Vendor { id : 5346u16 , name : "INFOLIBRIA" , devices : & [] }),
        (4294, Vendor { id : 4294u16 , name : "Rambus Inc." , devices : & [] }),
        (6346, Vendor { id : 6346u16 , name : "XGI Technology Inc. (eXtreme Graphics Innovation)" , devices : & [Device { vendor_id : 6346u16 , id : 32u16 , name : "Z7/Z9 (XG20 core)" , subsystems : & [] } , Device { vendor_id : 6346u16 , id : 33u16 , name : "Z9s/Z9m (XG21 core)" , subsystems : & [] } , Device { vendor_id : 6346u16 , id : 39u16 , name : "Z11/Z11M" , subsystems : & [] } , Device { vendor_id : 6346u16 , id : 64u16 , name : "Volari V3XT/V5/V8" , subsystems : & [] } , Device { vendor_id : 6346u16 , id : 71u16 , name : "Volari 8300 (chip: XP10, codename: XG47)" , subsystems : & [] }] }),
        (4164, Vendor { id : 4164u16 , name : "Adaptec (formerly DPT)" , devices : & [Device { vendor_id : 4164u16 , id : 4114u16 , name : "Domino RAID Engine" , subsystems : & [] } , Device { vendor_id : 4164u16 , id : 41984u16 , name : "SmartCache/Raid I-IV Controller" , subsystems : & [] } , Device { vendor_id : 4164u16 , id : 42240u16 , name : "PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4164u16 , id : 42241u16 , name : "SmartRAID V Controller" , subsystems : & [SubSystem { subvendor : 4164u16 , subdevice : 49153u16 , name : "PM1554U2 Ultra2 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49154u16 , name : "PM1654U2 Ultra2 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49155u16 , name : "PM1564U3 Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49156u16 , name : "PM1564U3 Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49157u16 , name : "PM1554U2 Ultra2 Single Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49162u16 , name : "PM2554U2 Ultra2 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49163u16 , name : "PM2654U2 Ultra2 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49164u16 , name : "PM2664U3 Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49165u16 , name : "PM2664U3 Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49166u16 , name : "PM2554U2 Ultra2 Single Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49167u16 , name : "PM2654U2 Ultra2 Single Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49172u16 , name : "PM3754U2 Ultra2 Single Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49173u16 , name : "PM3755U2B Ultra2 Single Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49174u16 , name : "PM3755F Fibre Channel (NON ACPI)" } , SubSystem { subvendor : 4164u16 , subdevice : 49182u16 , name : "PM3757U2 Ultra2 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49183u16 , name : "PM3757U2 Ultra2 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49184u16 , name : "PM3767U3 Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49185u16 , name : "PM3767U3 Ultra3 Quad Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49192u16 , name : "PM2865U3 Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49193u16 , name : "PM2865U3 Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49194u16 , name : "PM2865F Fibre Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49212u16 , name : "2000S Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49213u16 , name : "2000S Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49214u16 , name : "2000F Fibre Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49222u16 , name : "3000S Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49223u16 , name : "3000S Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49224u16 , name : "3000F Fibre Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49232u16 , name : "5000S Ultra3 Single Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49233u16 , name : "5000S Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49234u16 , name : "5000F Fibre Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49242u16 , name : "2400A UDMA Four Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49243u16 , name : "2400A UDMA Four Channel DAC" } , SubSystem { subvendor : 4164u16 , subdevice : 49252u16 , name : "3010S Ultra3 Dual Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49253u16 , name : "3410S Ultra160 Four Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49254u16 , name : "3010S Fibre Channel" }] } , Device { vendor_id : 4164u16 , id : 42257u16 , name : "SmartRAID V Controller" , subsystems : & [SubSystem { subvendor : 4164u16 , subdevice : 49202u16 , name : "ASR-2005S I2O Zero Channel" } , SubSystem { subvendor : 4164u16 , subdevice : 49205u16 , name : "ASR-2010S I2O Zero Channel" }] } , Device { vendor_id : 4164u16 , id : 49254u16 , name : "3010S Ultra3 Dual Channel" , subsystems : & [] }] }),
        (5228, Vendor { id : 5228u16 , name : "Ruby Tech Corp." , devices : & [Device { vendor_id : 5228u16 , id : 5168u16 , name : "FE-1430TX Fast Ethernet PCI Adapter" , subsystems : & [] }] }),
        (6235, Vendor { id : 6235u16 , name : "Compro Technology, Inc." , devices : & [Device { vendor_id : 6235u16 , id : 5257u16 , name : "VideoMate Vista T100" , subsystems : & [] }] }),
        (4602, Vendor { id : 4602u16 , name : "Kasan Electronics Company, Ltd." , devices : & [] }),
        (7621, Vendor { id : 7621u16 , name : "FADU Inc." , devices : & [Device { vendor_id : 7621u16 , id : 16513u16 , name : "FC4121 PCIe 4.0 NVMe controller [DELTA]" , subsystems : & [] } , Device { vendor_id : 7621u16 , id : 20833u16 , name : "FC5161 PCIe 5.0 NVMe controller [ECHO]" , subsystems : & [] } , Device { vendor_id : 7621u16 , id : 24912u16 , name : "FC3081 PCIe 3.0 NVMe controller [BRAVO]" , subsystems : & [] }] }),
        (7212, Vendor { id : 7212u16 , name : "Silicom Denmark" , devices : & [Device { vendor_id : 7212u16 , id : 10u16 , name : "Capture" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 15u16 , name : "SmartNIC" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 160u16 , name : "FBC4G Capture 4x1Gb [Herculaneum]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 161u16 , name : "FBC4XG Capture 4x10Gb [Ancona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 162u16 , name : "FBC8XG Capture 8x10Gb [Livorno]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 163u16 , name : "FBC2XG Capture 2x10Gb [Genoa]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 164u16 , name : "FBC4XGG3 Capture 4x10Gb [Livigno]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 165u16 , name : "FBC2XLG Capture 2x40Gb [Livorno]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 166u16 , name : "FBC1CG Capture 1x100Gb" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 169u16 , name : "FBC2XGHH Capture 2x10Gb [Latina]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 173u16 , name : "FBC2CGG3HL Capture 2x100Gb [Padua]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 175u16 , name : "Capture slave device" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 224u16 , name : "PacketMover 2x100Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 225u16 , name : "PacketMover 2x100Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 226u16 , name : "PacketMover 2x100Gb [Mango_04]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 227u16 , name : "PacketMover 2x10Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 229u16 , name : "PacketMover 2x10Gb [Corfu]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 4096u16 , name : "SmartNIC N5010 4x100Gb" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 4097u16 , name : "SmartNIC N5011 w/2xE810 4x100Gb" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 4098u16 , name : "SmartNIC N5013 4x100Gb" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 4099u16 , name : "SmartNIC N5014 w/2xE810 4x100Gb" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40960u16 , name : "FBC2CGG3 Capture 2x40Gb [Mango_02]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40961u16 , name : "FBC2CGG3 Capture 2x100Gb [Mango_02]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40963u16 , name : "FBC2CGG3 Capture 16x10Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40966u16 , name : "FBC2CGG3 Capture 2x100Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40967u16 , name : "FBC2CGG3 Capture 2x40Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40968u16 , name : "FBC2CGG3 Capture 2x25Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40969u16 , name : "FBC2CGG3 Capture 16x10Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40970u16 , name : "FBC2CGG3 Capture 8x10Gb [Mango]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40974u16 , name : "FB2CG Capture 2x100Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40975u16 , name : "FB2CG Capture 2x40Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40976u16 , name : "FB2CGHH Capture 2x40Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40977u16 , name : "FB2CG Capture 2x25Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40978u16 , name : "FB2CG Capture 8x10Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40979u16 , name : "FB2CGHH Capture 2x25Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40980u16 , name : "FB2CGHH Capture 8x10Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40981u16 , name : "FB2CGHH Capture 2x100Gb [Tivoli]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40982u16 , name : "FB2CG Capture 8x25Gb [Savona]" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40983u16 , name : "FB2CGHH Capture 8x25Gb [Tivoli] a017" , subsystems : & [] } , Device { vendor_id : 7212u16 , id : 40987u16 , name : "FB2CDG1 Capture 2x100Gb [Thunderfjord]" , subsystems : & [] }] }),
        (6140, Vendor { id : 6140u16 , name : "IOGEAR, Inc." , devices : & [] }),
        (4979, Vendor { id : 4979u16 , name : "Silicon Vision Inc" , devices : & [] }),
        (8470, Vendor { id : 8470u16 , name : "ZyDAS Technology Corp." , devices : & [] }),
        (4780, Vendor { id : 4780u16 , name : "Measurex Corporation" , devices : & [] }),
        (5113, Vendor { id : 5113u16 , name : "NTT Advanced Technology Corp." , devices : & [] }),
        (6068, Vendor { id : 6068u16 , name : "Indra Networks, Inc." , devices : & [Device { vendor_id : 6068u16 , id : 17u16 , name : "WebEnhance 100 GZIP Compression Card" , subsystems : & [] } , Device { vendor_id : 6068u16 , id : 18u16 , name : "WebEnhance 200 GZIP Compression Card" , subsystems : & [] } , Device { vendor_id : 6068u16 , id : 21u16 , name : "WebEnhance 300 GZIP Compression Card" , subsystems : & [] } , Device { vendor_id : 6068u16 , id : 22u16 , name : "StorCompress 300 GZIP Compression Card" , subsystems : & [] } , Device { vendor_id : 6068u16 , id : 23u16 , name : "StorSecure 300 GZIP Compression and AES Encryption Card" , subsystems : & [] }] }),
        (16448, Vendor { id : 16448u16 , name : "NetXen Incorporated" , devices : & [Device { vendor_id : 16448u16 , id : 1u16 , name : "NXB-10GXSR 10-Gigabit Ethernet PCIe Adapter with SR-XFP optical interface" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28743u16 , name : "NC510F PCIe 10-Gigabit Server Adapter" }] } , Device { vendor_id : 16448u16 , id : 2u16 , name : "NXB-10GCX4 10-Gigabit Ethernet PCIe Adapter with CX4 copper interface" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28744u16 , name : "NC510c PCIe 10-Gigabit Server Adapter" }] } , Device { vendor_id : 16448u16 , id : 3u16 , name : "NXB-4GCU Quad Gigabit Ethernet PCIe Adapter with 1000-BASE-T interface" , subsystems : & [] } , Device { vendor_id : 16448u16 , id : 4u16 , name : "BladeCenter-H 10-Gigabit Ethernet High Speed Daughter Card" , subsystems : & [] } , Device { vendor_id : 16448u16 , id : 5u16 , name : "NetXen Dual Port 10GbE Multifunction Adapter for c-Class" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5902u16 , name : "NC512m Dual Port 10GbE Multifunction BL-C Adapter" }] } , Device { vendor_id : 16448u16 , id : 36u16 , name : "XG Mgmt" , subsystems : & [] } , Device { vendor_id : 16448u16 , id : 37u16 , name : "XG Mgmt" , subsystems : & [] } , Device { vendor_id : 16448u16 , id : 256u16 , name : "NX3031 Multifunction 1/10-Gigabit Server Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5915u16 , name : "NC522m Dual Port 10GbE Multifunction BL-c Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5952u16 , name : "NC375T PCI Express Quad Port Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12881u16 , name : "NC375i 1G w/NC524SFP 10G Module" } , SubSystem { subvendor : 4156u16 , subdevice : 28762u16 , name : "NC375i Integrated Quad Port Multifunction Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28763u16 , name : "NC522SFP Dual Port 10GbE Server Adapter" } , SubSystem { subvendor : 5421u16 , subdevice : 35179u16 , name : "TG20 Dual Port 10GbE Server/Storage Adapter" } , SubSystem { subvendor : 16448u16 , subdevice : 292u16 , name : "NX3031 Quad Port Gigabit Server Adapter" } , SubSystem { subvendor : 16448u16 , subdevice : 294u16 , name : "Dual Port SFP+ 10GbE Server Adapter" }] }] }),
        (4739, Vendor { id : 4739u16 , name : "Integrated Technology Express, Inc." , devices : & [Device { vendor_id : 4739u16 , id : 26426u16 , name : "IT8330G" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 33106u16 , name : "IT8152F/G Advanced RISC-to-PCI Companion Chip" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 33297u16 , name : "ITE 8211F Single Channel UDMA 133" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33080u16 , name : "P5GD1-VW Mainboard" }] } , Device { vendor_id : 4739u16 , id : 33298u16 , name : "IT8212 Dual channel ATA RAID controller" , subsystems : & [SubSystem { subvendor : 4739u16 , subdevice : 1u16 , name : "IT/ITE8212 Dual channel ATA RAID controller" }] } , Device { vendor_id : 4739u16 , id : 33299u16 , name : "IT8213 IDE Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 45056u16 , name : "GA-EG45M-DS2H Mainboard" }] } , Device { vendor_id : 4739u16 , id : 33584u16 , name : "IT8330G" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 34930u16 , name : "IT887xF PCI to ISA I/O chip with SMB, GPIO, Serial or Parallel Port" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 34952u16 , name : "IT8888F/G PCI to ISA Bridge with SMB [Golden Gate]" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 34953u16 , name : "IT8889F PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 34962u16 , name : "IT8892E PCIe to PCI Bridge" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" }] } , Device { vendor_id : 4739u16 , id : 34963u16 , name : "IT8893E PCIe to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4739u16 , id : 59526u16 , name : "IT8330G" , subsystems : & [] }] }),
        (5403, Vendor { id : 5403u16 , name : "COMBOX Ltd" , devices : & [] }),
        (4922, Vendor { id : 4922u16 , name : "Vtel Corp" , devices : & [] }),
        (7014, Vendor { id : 7014u16 , name : "DELTACAST" , devices : & [Device { vendor_id : 7014u16 , id : 7u16 , name : "DELTA-3G-elp-d" , subsystems : & [] }] }),
        (4509, Vendor { id : 4509u16 , name : "Bug, Inc. Sapporo Japan" , devices : & [] }),
        (6641, Vendor { id : 6641u16 , name : "BFG Tech" , devices : & [] }),
        (6976, Vendor { id : 6976u16 , name : "Schooner Information Technology, Inc." , devices : & [] }),
        (6612, Vendor { id : 6612u16 , name : "Quixant Limited" , devices : & [] }),
        (5638, Vendor { id : 5638u16 , name : "Europop AG" , devices : & [] }),
        (5170, Vendor { id : 5170u16 , name : "Edimax Computer Co." , devices : & [Device { vendor_id : 5170u16 , id : 37168u16 , name : "RTL81xx Fast Ethernet" , subsystems : & [] }] }),
        (6367, Vendor { id : 6367u16 , name : "LeWiz Communications" , devices : & [] }),
        (7652, Vendor { id : 7652u16 , name : "Raspberry Pi Ltd" , devices : & [Device { vendor_id : 7652u16 , id : 1u16 , name : "RP1 PCIe 2.0 South Bridge" , subsystems : & [] }] }),
        (7693, Vendor { id : 7693u16 , name : "SambaNova Systems, Inc" , devices : & [] }),
        (5501, Vendor { id : 5501u16 , name : "Gemflex Networks" , devices : & [] }),
        (7104, Vendor { id : 7104u16 , name : "Innodisk Corporation" , devices : & [Device { vendor_id : 7104u16 , id : 4097u16 , name : "PCIe 3TG6-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4098u16 , name : "PCIe 3TE6 Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4448u16 , name : "PCIe 3TE2 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4610u16 , name : "PCIe 3TE8 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4618u16 , name : "PCIe 3IE8 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4619u16 , name : "PCIe 3TO8 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4897u16 , name : "PCIe 4TG-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 4898u16 , name : "PCIe 4TE Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 5634u16 , name : "PCIe 4TE3 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 5642u16 , name : "PCIe 4IE3 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 8802u16 , name : "PCIe 3TG3-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21000u16 , name : "PCIe 3TE7 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21014u16 , name : "PCIe 3TE9 controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21018u16 , name : "PCIe 3IE9 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21024u16 , name : "PCIe 4TE2 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21034u16 , name : "PCIe 4IE2 Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21046u16 , name : "PCIe 4TG2-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21050u16 , name : "PCIe 4TS2-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 21051u16 , name : "PCIe 4IG2-P Controller" , subsystems : & [] } , Device { vendor_id : 7104u16 , id : 33638u16 , name : "PCIe 5TS-P Controller" , subsystems : & [] }] }),
        (5874, Vendor { id : 5874u16 , name : "ETAS GmbH" , devices : & [Device { vendor_id : 5874u16 , id : 512u16 , name : "I/O board" , subsystems : & [SubSystem { subvendor : 5874u16 , subdevice : 16u16 , name : "ES53xx I/O board" }] }] }),
        (8128, Vendor { id : 8128u16 , name : "Ascom (Finland) Oy" , devices : & [Device { vendor_id : 8128u16 , id : 768u16 , name : "E2200 Dual E1/Rawpipe Card" , subsystems : & [] } , Device { vendor_id : 8128u16 , id : 769u16 , name : "C5400 SHDSL/E1 Card" , subsystems : & [] }] }),
        (5510, Vendor { id : 5510u16 , name : "Lancast Inc" , devices : & [] }),
        (5629, Vendor { id : 5629u16 , name : "N-CUBED.NET" , devices : & [] }),
        (33793, Vendor { id : 33793u16 , name : "TRENDware International Inc." , devices : & [] }),
        (5941, Vendor { id : 5941u16 , name : "Aten International Co. Ltd." , devices : & [] }),
        (5415, Vendor { id : 5415u16 , name : "SOLECTRON" , devices : & [] }),
        (4374, Vendor { id : 4374u16 , name : "Data Translation" , devices : & [Device { vendor_id : 4374u16 , id : 34u16 , name : "DT3001" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 35u16 , name : "DT3002" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 36u16 , name : "DT3003" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 37u16 , name : "DT3004" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 38u16 , name : "DT3005" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 39u16 , name : "DT3001-PGL" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 40u16 , name : "DT3003-PGL" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 81u16 , name : "DT322" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 96u16 , name : "DT340" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 105u16 , name : "DT332" , subsystems : & [] } , Device { vendor_id : 4374u16 , id : 32962u16 , name : "DT3162" , subsystems : & [] }] }),
        (4502, Vendor { id : 4502u16 , name : "Hytec Electronics Ltd" , devices : & [] }),
        (4599, Vendor { id : 4599u16 , name : "Scientific Atlanta" , devices : & [] }),
        (7637, Vendor { id : 7637u16 , name : "Manli Technology Group Limited" , devices : & [] }),
        (4621, Vendor { id : 4621u16 , name : "Compression Labs, Inc." , devices : & [] }),
        (4966, Vendor { id : 4966u16 , name : "Teijin Seiki Co. Ltd" , devices : & [] }),
        (4246, Vendor { id : 4246u16 , name : "Alacron" , devices : & [] }),
        (5487, Vendor { id : 5487u16 , name : "M-Systems Flash Disk Pioneers Ltd" , devices : & [] }),
        (4242, Vendor { id : 4242u16 , name : "Diamond Multimedia Systems" , devices : & [Device { vendor_id : 4242u16 , id : 40u16 , name : "Viper V770" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 18944u16 , name : "Viper V770 32MB" }] } , Device { vendor_id : 4242u16 , id : 160u16 , name : "Speedstar Pro SE" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 168u16 , name : "Speedstar 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 1360u16 , name : "Viper V550" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 2260u16 , name : "Supra 2260 Modem" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 2380u16 , name : "SupraExpress 56i Pro" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 4097u16 , name : "Video Crunch It 1001 capture card" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 4242u16 , name : "Viper V330" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 24864u16 , name : "Maximum DVD" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34832u16 , name : "Stealth SE" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34833u16 , name : "Stealth 64/SE" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34944u16 , name : "Stealth" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34945u16 , name : "Stealth" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34992u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 34993u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35008u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35009u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35024u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35025u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35056u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 35057u16 , name : "Stealth 64" , subsystems : & [] } , Device { vendor_id : 4242u16 , id : 39321u16 , name : "DMD-I0928-1 \"Monster sound\" sound chip" , subsystems : & [] }] }),
        (6706, Vendor { id : 6706u16 , name : "Quanta Microsystems, Inc" , devices : & [] }),
        (5230, Vendor { id : 5230u16 , name : "Williams Electronics Games, Inc." , devices : & [] }),
        (6118, Vendor { id : 6118u16 , name : "MaxLinear" , devices : & [Device { vendor_id : 6118u16 , id : 16u16 , name : "EN2010 [c.Link] MoCA Network Controller (Coax, PCI interface)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 17u16 , name : "EN2010 [c.Link] MoCA Network Controller (Coax, MPEG interface)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 33u16 , name : "EN2210 [c.Link] MoCA Network Controller (Coax)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 37u16 , name : "EN2510 [c.Link] MoCA Network Controller (Coax, PCIe interface)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 39u16 , name : "EN2710 [c.Link] MoCA 2.0 Network Controller (Coax, PCIe interface)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 14080u16 , name : "MoCA 2.0 Network Controller (Coax, PCIe interface)" , subsystems : & [] } , Device { vendor_id : 6118u16 , id : 14096u16 , name : "MoCA 2.5 Network Controller (Coax, PCIe interface)" , subsystems : & [] }] }),
        (4826, Vendor { id : 4826u16 , name : "True Time Inc." , devices : & [] }),
        (5404, Vendor { id : 5404u16 , name : "DIGITAL AUDIO LABS Inc" , devices : & [Device { vendor_id : 5404u16 , id : 3u16 , name : "Prodif T 2496" , subsystems : & [] } , Device { vendor_id : 5404u16 , id : 16384u16 , name : "Prodif 88" , subsystems : & [] }] }),
        (4955, Vendor { id : 4955u16 , name : "Giganet Inc" , devices : & [] }),
        (4316, Vendor { id : 4316u16 , name : "CERN/ECP/EDU" , devices : & [Device { vendor_id : 4316u16 , id : 1u16 , name : "STAR/RD24 SCI-PCI (PMC)" , subsystems : & [] } , Device { vendor_id : 4316u16 , id : 2u16 , name : "TAR/RD24 SCI-PCI (PMC)" , subsystems : & [] } , Device { vendor_id : 4316u16 , id : 33u16 , name : "HIPPI destination" , subsystems : & [] } , Device { vendor_id : 4316u16 , id : 34u16 , name : "HIPPI source" , subsystems : & [] } , Device { vendor_id : 4316u16 , id : 4316u16 , name : "ATT2C15-3 FPGA" , subsystems : & [] }] }),
        (4285, Vendor { id : 4285u16 , name : "Surecom Technology" , devices : & [Device { vendor_id : 4285u16 , id : 3636u16 , name : "NE-34" , subsystems : & [] }] }),
        (4391, Vendor { id : 4391u16 , name : "FORE Systems Inc" , devices : & [Device { vendor_id : 4391u16 , id : 512u16 , name : "ForeRunner PCA-200 ATM" , subsystems : & [] } , Device { vendor_id : 4391u16 , id : 528u16 , name : "PCA-200PC" , subsystems : & [] } , Device { vendor_id : 4391u16 , id : 592u16 , name : "ATM" , subsystems : & [] } , Device { vendor_id : 4391u16 , id : 768u16 , name : "ForeRunner PCA-200EPC ATM" , subsystems : & [] } , Device { vendor_id : 4391u16 , id : 784u16 , name : "ATM" , subsystems : & [] } , Device { vendor_id : 4391u16 , id : 1024u16 , name : "ForeRunnerHE ATM Adapter" , subsystems : & [SubSystem { subvendor : 4391u16 , subdevice : 1024u16 , name : "ForeRunnerHE ATM" }] }] }),
        (5413, Vendor { id : 5413u16 , name : "IMPACT Technologies" , devices : & [] }),
        (6107, Vendor { id : 6107u16 , name : "Cray Inc" , devices : & [Device { vendor_id : 6107u16 , id : 257u16 , name : "XT Series [Seastar] 3D Toroidal Router" , subsystems : & [] } , Device { vendor_id : 6107u16 , id : 1281u16 , name : "Cassini 1 [Slingshot 200Gb]" , subsystems : & [] }] }),
        (4543, Vendor { id : 4543u16 , name : "Astrodesign, Inc." , devices : & [] }),
        (5314, Vendor { id : 5314u16 , name : "DTK Computer" , devices : & [] }),
        (4641, Vendor { id : 4641u16 , name : "Contec Co., Ltd" , devices : & [Device { vendor_id : 4641u16 , id : 37234u16 , name : "PO-64L(PCI)H [Isolated Digital Output Board for PCI]" , subsystems : & [] } , Device { vendor_id : 4641u16 , id : 37282u16 , name : "PO-32L(PCI)H [Isolated Digital Output Board for PCI]" , subsystems : & [] } , Device { vendor_id : 4641u16 , id : 37315u16 , name : "DA16-16(LPCI)L [Un-insulated highly precise analog output board for Low Profile PCI]" , subsystems : & [] } , Device { vendor_id : 4641u16 , id : 45394u16 , name : "DIO-96D2-LPCI" , subsystems : & [] } , Device { vendor_id : 4641u16 , id : 49411u16 , name : "ADA16-32/2(PCI)F [High-Speed Analog I/O Board for PCI]" , subsystems : & [] }] }),
        (4710, Vendor { id : 4710u16 , name : "Microdyne Corporation" , devices : & [Device { vendor_id : 4710u16 , id : 1u16 , name : "NE10/100 Adapter (i82557B)" , subsystems : & [] } , Device { vendor_id : 4710u16 , id : 6416u16 , name : "NE2000Plus (RT8029) Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4710u16 , subdevice : 6416u16 , name : "NE2000Plus Ethernet Adapter" }] }] }),
        (4436, Vendor { id : 4436u16 , name : "Melco Inc" , devices : & [] }),
        (21828, Vendor { id : 21828u16 , name : "Dunord Technologies" , devices : & [Device { vendor_id : 21828u16 , id : 1u16 , name : "I-30xx Scanner Interface" , subsystems : & [] }] }),
        (7295, Vendor { id : 7295u16 , name : "Elektrobit Austria GmbH" , devices : & [Device { vendor_id : 7295u16 , id : 768u16 , name : "EBX3" , subsystems : & [] } , Device { vendor_id : 7295u16 , id : 20736u16 , name : "EB5100" , subsystems : & [] } , Device { vendor_id : 7295u16 , id : 29184u16 , name : "EB 7200 Automotive logging and replay card" , subsystems : & [] } , Device { vendor_id : 7295u16 , id : 29185u16 , name : "EB7200 with EB PCIe Driver Interface" , subsystems : & [] } , Device { vendor_id : 7295u16 , id : 29201u16 , name : "EB7210" , subsystems : & [] }] }),
        (6894, Vendor { id : 6894u16 , name : "Caustic Graphics Inc." , devices : & [] }),
        (4642, Vendor { id : 4642u16 , name : "Ancor Communications, Inc." , devices : & [] }),
        (4960, Vendor { id : 4960u16 , name : "Meinberg Funkuhren" , devices : & [Device { vendor_id : 4960u16 , id : 257u16 , name : "PCI32 DCF77 Radio Clock" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 258u16 , name : "PCI509 DCF77 Radio Clock" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 259u16 , name : "PCI510 DCF77 Radio Clock" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 260u16 , name : "PCI511 DCF77 Radio Clock" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 261u16 , name : "PEX511 DCF77 Radio Clock (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 262u16 , name : "PZF180PEX High Precision DCF77 Radio Clock (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 513u16 , name : "GPS167PCI GPS Receiver" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 514u16 , name : "GPS168PCI GPS Receiver" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 515u16 , name : "GPS169PCI GPS Receiver" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 516u16 , name : "GPS170PCI GPS Receiver" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 517u16 , name : "GPS170PEX GPS Receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 518u16 , name : "GPS180PEX GPS Receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 519u16 , name : "GLN180PEX GPS/GLONASS receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 520u16 , name : "GPS180AMC GPS Receiver (PCI Express / MicroTCA / AdvancedMC)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 521u16 , name : "GNS181PEX GNSS receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 522u16 , name : "GPS183PEX GPS Receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 523u16 , name : "GNS183PEX GNSS receiver (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 769u16 , name : "TCR510PCI IRIG Timecode Reader" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 770u16 , name : "TCR167PCI IRIG Timecode Reader" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 771u16 , name : "TCR511PCI IRIG Timecode Reader" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 772u16 , name : "TCR511PEX IRIG Timecode Reader (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 773u16 , name : "TCR170PEX IRIG Timecode Reader (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 774u16 , name : "TCR180PEX IRIG Timecode Reader (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 1281u16 , name : "PTP270PEX PTP/IEEE1588 slave card (PCI Express)" , subsystems : & [] } , Device { vendor_id : 4960u16 , id : 1537u16 , name : "FRC511PEX Free Running Clock (PCI Express)" , subsystems : & [] }] }),
        (4965, Vendor { id : 4965u16 , name : "Hypercope GmbH" , devices : & [] }),
        (5183, Vendor { id : 5183u16 , name : "Lightwell Co Ltd - Zax Division" , devices : & [] }),
        (4288, Vendor { id : 4288u16 , name : "Boca Research Inc." , devices : & [Device { vendor_id : 4288u16 , id : 37173u16 , name : "iX3D Ultimate Rez" , subsystems : & [] }] }),
        (5621, Vendor { id : 5621u16 , name : "Power Micro Research" , devices : & [] }),
        (5092, Vendor { id : 5092u16 , name : "Calculex Inc" , devices : & [] }),
        (4289, Vendor { id : 4289u16 , name : "ICM Co., Ltd." , devices : & [] }),
        (5315, Vendor { id : 5315u16 , name : "MEDIATEK Corp." , devices : & [Device { vendor_id : 5315u16 , id : 1544u16 , name : "MT7921K (RZ608) Wi-Fi 6E 80MHz" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 1558u16 , name : "MT7922 802.11ax PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 1815u16 , name : "MT7925 (RZ717) Wi-Fi 7 160MHz" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 19829u16 , name : "T700 5G Modem [5G Solution 5000]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30211u16 , name : "MT7603E 802.11bgn PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30226u16 , name : "MT7612E 802.11acbgn PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30229u16 , name : "MT7615E 802.11ac PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30256u16 , name : "MT7630e 802.11bgn Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30288u16 , name : "MT7650 802.11ac" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30306u16 , name : "MT7662E 802.11ac PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30307u16 , name : "MT7663 802.11ac PCI Express Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30978u16 , name : "MT7902 802.11ax PCIe Wireless Network Adapter [Filogic 310]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30982u16 , name : "MT7916A/MT7916D normal link PCIe Wi-Fi 6(802.11ax) 160MHz 2x2 Wireless Network Adapter [Filogic 630]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30997u16 , name : "MT7915A/MT7915D normal link PCIe Wi-Fi 6(802.11ax) 80MHz 4x4/2x2 Wireless Network Adapter [Filogic 615]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 30998u16 , name : "MT7915A/MT7915D hif link PCIe Wi-Fi 6(802.11ax) 80MHz 4x4/2x2 Wireless Network Adapter [Filogic 615]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31010u16 , name : "MT7922 802.11ax PCI Express Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 6715u16 , subdevice : 21248u16 , name : "ASUS PCE-AXE59BT" }] } , Device { vendor_id : 5315u16 , id : 31013u16 , name : "MT7925 802.11be 160MHz 2x2 PCIe Wireless Network Adapter [Filogic 360]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31015u16 , name : "MT7927 802.11be 320MHz 2x2 PCIe Wireless Network Adapter [Filogic 380]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31073u16 , name : "MT7921 802.11ax PCIe Wireless Network Adapter [Filogic 330]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31112u16 , name : "MT7988 PCI Express Host Bridge [Filogic 880]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31120u16 , name : "MT7996 primary link PCIe Wi-Fi 7(802.11be) 320MHz Wireless Network Adapter [Filogic 680]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31121u16 , name : "MT7996 secondary link PCIe Wi-Fi 7(802.11be) 320MHz Wireless Network Adapter [Filogic 680]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31122u16 , name : "MT7992 primary link PCIe Wi-Fi 7(802.11be) 160MHz Wireless Network Adapter [Filogic 660]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 31130u16 , name : "MT7992 secondary link PCIe Wi-Fi 7(802.11be) 160MHz Wireless Network Adapter [Filogic 660]" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 33160u16 , name : "MT8188 [Kompanio 838] Root Complex" , subsystems : & [] } , Device { vendor_id : 5315u16 , id : 34384u16 , name : "MT7650 Bluetooth" , subsystems : & [] }] }),
        (5761, Vendor { id : 5761u16 , name : "Hercules" , devices : & [] }),
        (4769, Vendor { id : 4769u16 , name : "Simpact Associates, Inc." , devices : & [] }),
        (3725, Vendor { id : 3725u16 , name : "MediaTek Inc. (Wrong ID)" , devices : & [Device { vendor_id : 3725u16 , id : 2049u16 , name : "MT7621 PCIe Bridge" , subsystems : & [] }] }),
        (5768, Vendor { id : 5768u16 , name : "CastleNet Technology Inc." , devices : & [Device { vendor_id : 5768u16 , id : 4464u16 , name : "WLAN 802.11b card" , subsystems : & [] }] }),
        (5091, Vendor { id : 5091u16 , name : "Nest Inc" , devices : & [] }),
        (7458, Vendor { id : 7458u16 , name : "Baidu Technology" , devices : & [Device { vendor_id : 7458u16 , id : 4992u16 , name : "Cloud Storage Device" , subsystems : & [] } , Device { vendor_id : 7458u16 , id : 13956u16 , name : "Kunlun AI Accelerator" , subsystems : & [] } , Device { vendor_id : 7458u16 , id : 13957u16 , name : "Kunlun2 AI Accelerator [VF]" , subsystems : & [] }] }),
        (4475, Vendor { id : 4475u16 , name : "L G Electronics, Inc." , devices : & [] }),
        (5436, Vendor { id : 5436u16 , name : "ANTAL Electronic" , devices : & [] }),
        (4718, Vendor { id : 4718u16 , name : "Sumitomo Metal Industries, Ltd." , devices : & [] }),
        (4414, Vendor { id : 4414u16 , name : "Sanyo Electric Co - Computer Engineering Dept" , devices : & [] }),
        (6288, Vendor { id : 6288u16 , name : "Egenera, Inc." , devices : & [] }),
        (4586, Vendor { id : 4586u16 , name : "Elsag Bailey" , devices : & [] }),
        (4970, Vendor { id : 4970u16 , name : "High Soft Tech" , devices : & [Device { vendor_id : 4970u16 , id : 4u16 , name : "HST Saphir VII mini PCI" , subsystems : & [] } , Device { vendor_id : 4970u16 , id : 7u16 , name : "HST Saphir III E MultiLink 4" , subsystems : & [] } , Device { vendor_id : 4970u16 , id : 8u16 , name : "HST Saphir III E MultiLink 8" , subsystems : & [] } , Device { vendor_id : 4970u16 , id : 10u16 , name : "HST Saphir III E MultiLink 2" , subsystems : & [] }] }),
        (4671, Vendor { id : 4671u16 , name : "LSI Logic" , devices : & [Device { vendor_id : 4671u16 , id : 228u16 , name : "MPEG" , subsystems : & [] } , Device { vendor_id : 4671u16 , id : 33056u16 , name : "DVxplore Codec" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 481u16 , name : "NVTV PAL" } , SubSystem { subvendor : 4318u16 , subdevice : 482u16 , name : "NVTV NTSC" } , SubSystem { subvendor : 4318u16 , subdevice : 483u16 , name : "NVTV PAL" } , SubSystem { subvendor : 4318u16 , subdevice : 584u16 , name : "NVTV NTSC" } , SubSystem { subvendor : 4318u16 , subdevice : 585u16 , name : "NVTV PAL" } , SubSystem { subvendor : 4541u16 , subdevice : 6u16 , name : "DV500 E4" } , SubSystem { subvendor : 4541u16 , subdevice : 10u16 , name : "DV500 E4" } , SubSystem { subvendor : 4541u16 , subdevice : 15u16 , name : "DV500 E4" } , SubSystem { subvendor : 6153u16 , subdevice : 22u16 , name : "Emuzed MAUI-III PCI PVR FM TV" }] } , Device { vendor_id : 4671u16 , id : 34952u16 , name : "Cinemaster C 3.0 DVD Decoder" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 1u16 , name : "Cinemaster C 3.0 DVD Decoder" } , SubSystem { subvendor : 4098u16 , subdevice : 2u16 , name : "Cinemaster C 3.0 DVD Decoder" } , SubSystem { subvendor : 4904u16 , subdevice : 1u16 , name : "Cinemaster C 3.0 DVD Decoder" }] }] }),
        (5298, Vendor { id : 5298u16 , name : "ENNOVATE Networks Inc" , devices : & [] }),
        (4569, Vendor { id : 4569u16 , name : "TEC Corporation" , devices : & [] }),
        (7453, Vendor { id : 7453u16 , name : "CNEX Labs" , devices : & [Device { vendor_id : 7453u16 , id : 7967u16 , name : "QEMU NVM Express LightNVM Controller" , subsystems : & [] } , Device { vendor_id : 7453u16 , id : 10247u16 , name : "8800 series NVMe SSD" , subsystems : & [] }] }),
        (4854, Vendor { id : 4854u16 , name : "Dawson France" , devices : & [] }),
        (5038, Vendor { id : 5038u16 , name : "ALD Technology Ltd" , devices : & [] }),
        (56506, Vendor { id : 56506u16 , name : "Dynamic Engineering" , devices : & [Device { vendor_id : 56506u16 , id : 70u16 , name : "PCIe Altera Cyclone IV" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 71u16 , name : "VPX-RCB" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 72u16 , name : "PMC-Biserial-III-BAE9" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 78u16 , name : "PC104p-Biserial-III-NVY5" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 79u16 , name : "PC104p-Biserial-III-NVY6" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 82u16 , name : "PCIeBiSerialDb37 BA22 LVDS IO" , subsystems : & [] } , Device { vendor_id : 56506u16 , id : 102u16 , name : "PMC-OctalUART-232" , subsystems : & [] }] }),
        (4566, Vendor { id : 4566u16 , name : "Tekelec Telecom" , devices : & [] }),
        (5673, Vendor { id : 5673u16 , name : "Kongsberg Spacetec AS" , devices : & [Device { vendor_id : 5673u16 , id : 4099u16 , name : "Format synchronizer v3.0" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 4102u16 , name : "Format synchronizer, model 10500" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 4103u16 , name : "Format synchronizer, model 21000" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 8194u16 , name : "Fast Universal Data Output" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 12544u16 , name : "IO31000 Frame Synchronizer and I/O" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 12800u16 , name : "IO32000 Frame Synchronizer and I/O" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 16386u16 , name : "High Rate Demodulator" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 20481u16 , name : "High Rate FEC" , subsystems : & [] } , Device { vendor_id : 5673u16 , id : 24577u16 , name : "High Rate Demodulator and FEC" , subsystems : & [] }] }),
        (5035, Vendor { id : 5035u16 , name : "Arcom Control Systems Ltd" , devices : & [] }),
        (4835, Vendor { id : 4835u16 , name : "Imation Corp - Medical Imaging Systems" , devices : & [] }),
        (4558, Vendor { id : 4558u16 , name : "Netaccess" , devices : & [] }),
        (4524, Vendor { id : 4524u16 , name : "Canon Information Systems Research Aust." , devices : & [] }),
        (5100, Vendor { id : 5100u16 , name : "Zydacron Inc" , devices : & [Device { vendor_id : 5100u16 , id : 10u16 , name : "NPC-RC01 Remote control receiver" , subsystems : & [] }] }),
        (5567, Vendor { id : 5567u16 , name : "High Tech Computer Corp (HTC)" , devices : & [] }),
        (53126, Vendor { id : 53126u16 , name : "Spectrum-4TOR" , devices : & [Device { vendor_id : 53126u16 , id : 630u16 , name : "Spectrum-4TOR in Flash Recovery Mode" , subsystems : & [] }] }),
        (4345, Vendor { id : 4345u16 , name : "PC Direct" , devices : & [] }),
        (7811, Vendor { id : 7811u16 , name : "Huaqin Technology Co.Ltd" , devices : & [] }),
        (5042, Vendor { id : 5042u16 , name : "Techno Chips Co. Ltd" , devices : & [] }),
        (4343, Vendor { id : 4343u16 , name : "Matsushita Electric Industrial Co., Ltd." , devices : & [] }),
        (28519, Vendor { id : 28519u16 , name : "NOVAIUM Technology" , devices : & [] }),
        (5276, Vendor { id : 5276u16 , name : "OVISLINK Corp." , devices : & [] }),
        (5568, Vendor { id : 5568u16 , name : "BVM Ltd" , devices : & [] }),
        (6553, Vendor { id : 6553u16 , name : "A-Logics" , devices : & [Device { vendor_id : 6553u16 , id : 43264u16 , name : "AM-7209 Video Processor" , subsystems : & [] }] }),
        (8058, Vendor { id : 8058u16 , name : "Efinix, Inc." , devices : & [Device { vendor_id : 8058u16 , id : 256u16 , name : "Default ID for Titanium FPGA PCIe Interface (AXI)" , subsystems : & [] }] }),
        (4462, Vendor { id : 4462u16 , name : "Electronics for Imaging" , devices : & [] }),
        (4730, Vendor { id : 4730u16 , name : "Rockwell International" , devices : & [Device { vendor_id : 4730u16 , id : 4098u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 2380u16 , name : "SupraExpress 56i PRO [Diamond SUP2380]" } , SubSystem { subvendor : 4653u16 , subdevice : 16386u16 , name : "HPG / MDP3858-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16389u16 , name : "MDP3858-E" } , SubSystem { subvendor : 4653u16 , subdevice : 16391u16 , name : "MDP3858-A/-NZ" } , SubSystem { subvendor : 4653u16 , subdevice : 16402u16 , name : "MDP3858-SA" } , SubSystem { subvendor : 4653u16 , subdevice : 16407u16 , name : "MDP3858-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16408u16 , name : "MDP3858-W" } , SubSystem { subvendor : 4730u16 , subdevice : 4098u16 , name : "Rockwell 56K D/F HCF Modem" }] } , Device { vendor_id : 4730u16 , id : 4099u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45244u16 , name : "229-DF Zephyr" } , SubSystem { subvendor : 3601u16 , subdevice : 45332u16 , name : "229-DF Cheetah" } , SubSystem { subvendor : 4147u16 , subdevice : 32811u16 , name : "229-DF" } , SubSystem { subvendor : 5087u16 , subdevice : 4099u16 , name : "PCI56RX Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 279u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 327u16 , name : "IBM F-1156IV+/R3 Spain V.90 Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 407u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 455u16 , name : "IBM F-1156IV+/R3 WW V.90 Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 503u16 , name : "IBM" } , SubSystem { subvendor : 5174u16 , subdevice : 4099u16 , name : "IBM" } , SubSystem { subvendor : 5174u16 , subdevice : 4355u16 , name : "IBM 5614PM3G V.90 Modem" } , SubSystem { subvendor : 5174u16 , subdevice : 5634u16 , name : "Compaq 229-DF Ducati" }] } , Device { vendor_id : 4730u16 , id : 4100u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 5376u16 , name : "MicroLink 56k Modem" } , SubSystem { subvendor : 4303u16 , subdevice : 4185u16 , name : "Fujitsu 229-DFRT" }] } , Device { vendor_id : 4730u16 , id : 4101u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [SubSystem { subvendor : 4101u16 , subdevice : 4730u16 , name : "AOpen FM56-P" } , SubSystem { subvendor : 4147u16 , subdevice : 32809u16 , name : "229-DFSV" } , SubSystem { subvendor : 4147u16 , subdevice : 32852u16 , name : "Modem" } , SubSystem { subvendor : 4303u16 , subdevice : 4156u16 , name : "Fujitsu" } , SubSystem { subvendor : 4303u16 , subdevice : 4181u16 , name : "Fujitsu 229-DFSV" } , SubSystem { subvendor : 4303u16 , subdevice : 4182u16 , name : "Fujitsu 229-DFSV" } , SubSystem { subvendor : 4653u16 , subdevice : 16387u16 , name : "MDP3858SP-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16390u16 , name : "Packard Bell MDP3858V-E" } , SubSystem { subvendor : 4653u16 , subdevice : 16392u16 , name : "MDP3858SP-A/SP-NZ" } , SubSystem { subvendor : 4653u16 , subdevice : 16393u16 , name : "MDP3858SP-E" } , SubSystem { subvendor : 4653u16 , subdevice : 16400u16 , name : "MDP3858V-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16401u16 , name : "MDP3858SP-SA" } , SubSystem { subvendor : 4653u16 , subdevice : 16403u16 , name : "MDP3858V-A/V-NZ" } , SubSystem { subvendor : 4653u16 , subdevice : 16405u16 , name : "MDP3858SP-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16406u16 , name : "MDP3858V-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16409u16 , name : "MDP3858V-SA" } , SubSystem { subvendor : 5087u16 , subdevice : 4101u16 , name : "PCI56RVP Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 391u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 423u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 439u16 , name : "IBM DF-1156IV+/R3 Spain V.90 Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 471u16 , name : "IBM DF-1156IV+/R3 WW V.90 Modem" } , SubSystem { subvendor : 5174u16 , subdevice : 4101u16 , name : "IBM" } , SubSystem { subvendor : 5174u16 , subdevice : 4357u16 , name : "IBM" } , SubSystem { subvendor : 5175u16 , subdevice : 4357u16 , name : "IBM 5614PS3G V.90 Modem" }] } , Device { vendor_id : 4730u16 , id : 4130u16 , name : "HCF 56k Modem" , subsystems : & [SubSystem { subvendor : 5174u16 , subdevice : 4867u16 , name : "M3-5614PM3G V.90 Modem" }] } , Device { vendor_id : 4730u16 , id : 4131u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 16416u16 , name : "Packard Bell MDP3858-WE" } , SubSystem { subvendor : 4653u16 , subdevice : 16419u16 , name : "MDP3858-UE" } , SubSystem { subvendor : 5088u16 , subdevice : 583u16 , name : "IBM F-1156IV+/R6 Spain V.90 Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 663u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 711u16 , name : "IBM F-1156IV+/R6 WW V.90 Modem" } , SubSystem { subvendor : 5174u16 , subdevice : 4611u16 , name : "IBM" } , SubSystem { subvendor : 5174u16 , subdevice : 4867u16 , name : "IBM" }] } , Device { vendor_id : 4730u16 , id : 4132u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4133u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4202u16 , name : "Fujitsu 235-DFSV" } , SubSystem { subvendor : 4653u16 , subdevice : 16417u16 , name : "Packard Bell MDP3858V-WE" } , SubSystem { subvendor : 4653u16 , subdevice : 16418u16 , name : "MDP3858SP-WE" } , SubSystem { subvendor : 4653u16 , subdevice : 16420u16 , name : "MDP3858V-UE" } , SubSystem { subvendor : 4653u16 , subdevice : 16421u16 , name : "MDP3858SP-UE" }] } , Device { vendor_id : 4730u16 , id : 4134u16 , name : "HCF 56k PCI Speakerphone Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4146u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4147u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4148u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4149u16 , name : "HCF 56k PCI Speakerphone Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4150u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 4229u16 , name : "HCF 56k Volcano PCI Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 8196u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 4730u16 , id : 8197u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32836u16 , name : "229-DFSV" } , SubSystem { subvendor : 4173u16 , subdevice : 32837u16 , name : "229-DFSV" } , SubSystem { subvendor : 4173u16 , subdevice : 32853u16 , name : "PBE/Aztech 235W-DFSV" } , SubSystem { subvendor : 4173u16 , subdevice : 32854u16 , name : "235-DFSV" } , SubSystem { subvendor : 4173u16 , subdevice : 32858u16 , name : "Modem" } , SubSystem { subvendor : 4173u16 , subdevice : 32863u16 , name : "Modem" } , SubSystem { subvendor : 4173u16 , subdevice : 32884u16 , name : "Modem" }] } , Device { vendor_id : 4730u16 , id : 8211u16 , name : "HSF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Modem" } , SubSystem { subvendor : 4473u16 , subdevice : 65280u16 , name : "Modem" }] } , Device { vendor_id : 4730u16 , id : 8212u16 , name : "HSF 56k Data/Fax/Voice Modem" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4183u16 , name : "Fujitsu Citicorp III" } , SubSystem { subvendor : 4653u16 , subdevice : 16464u16 , name : "MSP3880-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16469u16 , name : "MSP3880-W" }] } , Device { vendor_id : 4730u16 , id : 8213u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4195u16 , name : "Fujitsu" } , SubSystem { subvendor : 4303u16 , subdevice : 4196u16 , name : "Fujitsu" } , SubSystem { subvendor : 5224u16 , subdevice : 8213u16 , name : "Fujitsu" }] } , Device { vendor_id : 4730u16 , id : 8214u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 16465u16 , name : "MSP3880V-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16466u16 , name : "MSP3880SP-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16468u16 , name : "MSP3880V-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16470u16 , name : "MSP3880SP-U" } , SubSystem { subvendor : 4653u16 , subdevice : 16471u16 , name : "MSP3880SP-A" }] } , Device { vendor_id : 4730u16 , id : 17169u16 , name : "Riptide HSF 56k PCI Modem" , subsystems : & [SubSystem { subvendor : 4730u16 , subdevice : 17169u16 , name : "Ring Modular? Riptide HSF RT HP Dom" } , SubSystem { subvendor : 5088u16 , subdevice : 528u16 , name : "HP-GVC" }] } , Device { vendor_id : 4730u16 , id : 17184u16 , name : "Riptide PCI Audio Controller" , subsystems : & [SubSystem { subvendor : 4661u16 , subdevice : 17184u16 , name : "Riptide PCI Audio Controller" }] } , Device { vendor_id : 4730u16 , id : 17185u16 , name : "Riptide HCF 56k PCI Modem" , subsystems : & [SubSystem { subvendor : 4661u16 , subdevice : 17185u16 , name : "Hewlett Packard DF" } , SubSystem { subvendor : 4661u16 , subdevice : 17188u16 , name : "Hewlett Packard DF" } , SubSystem { subvendor : 5088u16 , subdevice : 528u16 , name : "Hewlett Packard DF" } , SubSystem { subvendor : 5197u16 , subdevice : 8993u16 , name : "Riptide" }] } , Device { vendor_id : 4730u16 , id : 17186u16 , name : "Riptide PCI Game Controller" , subsystems : & [SubSystem { subvendor : 4661u16 , subdevice : 17186u16 , name : "Riptide PCI Game Controller" }] } , Device { vendor_id : 4730u16 , id : 33332u16 , name : "RapidFire 616X ATM155 Adapter" , subsystems : & [SubSystem { subvendor : 4237u16 , subdevice : 34u16 , name : "RapidFire 616X ATM155 Adapter" } , SubSystem { subvendor : 4237u16 , subdevice : 39u16 , name : "RapidFire 616X ATM155 Adapter" }] }] }),
        (52730, Vendor { id : 52730u16 , name : "NextSilicon Ltd" , devices : & [Device { vendor_id : 52730u16 , id : 7u16 , name : "Maverick" , subsystems : & [] } , Device { vendor_id : 52730u16 , id : 9u16 , name : "Maverick-2" , subsystems : & [SubSystem { subvendor : 52730u16 , subdevice : 2304u16 , name : "Maverick-2 Single-Die PCIe Card" } , SubSystem { subvendor : 52730u16 , subdevice : 2310u16 , name : "Maverick-2 Dual-Die OAM" } , SubSystem { subvendor : 52730u16 , subdevice : 3071u16 , name : "Maverick-2 Firmware Update" }] } , Device { vendor_id : 52730u16 , id : 128u16 , name : "Arbel Test Chip" , subsystems : & [] }] }),
        (5876, Vendor { id : 5876u16 , name : "Vweb Corp" , devices : & [Device { vendor_id : 5876u16 , id : 32768u16 , name : "VW2010" , subsystems : & [] }] }),
        (5000, Vendor { id : 5000u16 , name : "Hitachi Information Technology Co Ltd" , devices : & [] }),
        (4450, Vendor { id : 4450u16 , name : "OA Laboratory Co Ltd" , devices : & [] }),
        (5356, Vendor { id : 5356u16 , name : "Agilent Technologies" , devices : & [Device { vendor_id : 5356u16 , id : 0u16 , name : "Aciris Digitizer (malformed ID)" , subsystems : & [] }] }),
        (4894, Vendor { id : 4894u16 , name : "Xinex Networks Inc" , devices : & [] }),
        (7548, Vendor { id : 7548u16 , name : "Aerotech, Inc." , devices : & [Device { vendor_id : 7548u16 , id : 1u16 , name : "HyperWire Adapter" , subsystems : & [] }] }),
        (4201, Vendor { id : 4201u16 , name : "Mylex Corporation" , devices : & [Device { vendor_id : 4201u16 , id : 1u16 , name : "DAC960P" , subsystems : & [] } , Device { vendor_id : 4201u16 , id : 2u16 , name : "DAC960PD" , subsystems : & [] } , Device { vendor_id : 4201u16 , id : 16u16 , name : "DAC960PG" , subsystems : & [] } , Device { vendor_id : 4201u16 , id : 32u16 , name : "DAC960LA" , subsystems : & [] } , Device { vendor_id : 4201u16 , id : 80u16 , name : "AcceleRAID 352/170/160 support Device" , subsystems : & [SubSystem { subvendor : 4201u16 , subdevice : 80u16 , name : "AcceleRAID 352 support Device" } , SubSystem { subvendor : 4201u16 , subdevice : 82u16 , name : "AcceleRAID 170 support Device" } , SubSystem { subvendor : 4201u16 , subdevice : 84u16 , name : "AcceleRAID 160 support Device" }] } , Device { vendor_id : 4201u16 , id : 45414u16 , name : "AcceleRAID 600/500/400/Sapphire support Device" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 578u16 , name : "iSeries 2872 DASD IOA" } , SubSystem { subvendor : 4116u16 , subdevice : 614u16 , name : "Dual Channel PCI-X U320 SCSI Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 632u16 , name : "Dual Channel PCI-X U320 SCSI RAID Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 723u16 , name : "Dual Channel PCI-X U320 SCSI Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 724u16 , name : "Dual Channel PCI-X U320 SCSI RAID Adapter" } , SubSystem { subvendor : 4201u16 , subdevice : 512u16 , name : "AcceleRAID 400, Single Channel, PCI-X, U320, SCSI RAID" } , SubSystem { subvendor : 4201u16 , subdevice : 514u16 , name : "AcceleRAID Sapphire, Dual Channel, PCI-X, U320, SCSI RAID" } , SubSystem { subvendor : 4201u16 , subdevice : 516u16 , name : "AcceleRAID 500, Dual Channel, Low-Profile, PCI-X, U320, SCSI RAID" } , SubSystem { subvendor : 4201u16 , subdevice : 518u16 , name : "AcceleRAID 600, Dual Channel, PCI-X, U320, SCSI RAID" }] } , Device { vendor_id : 4201u16 , id : 47701u16 , name : "eXtremeRAID 1100 support Device" , subsystems : & [] } , Device { vendor_id : 4201u16 , id : 47702u16 , name : "eXtremeRAID 2000/3000 support Device" , subsystems : & [SubSystem { subvendor : 4201u16 , subdevice : 48u16 , name : "eXtremeRAID 3000 support Device" } , SubSystem { subvendor : 4201u16 , subdevice : 64u16 , name : "eXtremeRAID 2000 support Device" }] } , Device { vendor_id : 4201u16 , id : 47703u16 , name : "eXtremeRAID 4000/5000 support Device" , subsystems : & [SubSystem { subvendor : 4201u16 , subdevice : 114u16 , name : "eXtremeRAID 5000 support Device" }] }] }),
        (4377, Vendor { id : 4377u16 , name : "ICP Vortex Computersysteme GmbH" , devices : & [Device { vendor_id : 4377u16 , id : 0u16 , name : "GDT 6000/6020/6050" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 1u16 , name : "GDT 6000B/6010" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 2u16 , name : "GDT 6110/6510" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 3u16 , name : "GDT 6120/6520" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 4u16 , name : "GDT 6530" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 5u16 , name : "GDT 6550" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 6u16 , name : "GDT 6117/6517" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 7u16 , name : "GDT 6127/6527" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 8u16 , name : "GDT 6537" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 9u16 , name : "GDT 6557/6557-ECC" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 10u16 , name : "GDT 6115/6515" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 11u16 , name : "GDT 6125/6525" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 12u16 , name : "GDT 6535" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 13u16 , name : "GDT 6555/6555-ECC" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 256u16 , name : "GDT 6117RP/6517RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 257u16 , name : "GDT 6127RP/6527RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 258u16 , name : "GDT 6537RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 259u16 , name : "GDT 6557RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 260u16 , name : "GDT 6111RP/6511RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 261u16 , name : "GDT 6121RP/6521RP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 272u16 , name : "GDT 6117RD/6517RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 273u16 , name : "GDT 6127RD/6527RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 274u16 , name : "GDT 6537RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 275u16 , name : "GDT 6557RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 276u16 , name : "GDT 6111RD/6511RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 277u16 , name : "GDT 6121RD/6521RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 280u16 , name : "GDT 6118RD/6518RD/6618RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 281u16 , name : "GDT 6128RD/6528RD/6628RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 282u16 , name : "GDT 6538RD/6638RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 283u16 , name : "GDT 6558RD/6658RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 288u16 , name : "GDT 6117RP2/6517RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 289u16 , name : "GDT 6127RP2/6527RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 290u16 , name : "GDT 6537RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 291u16 , name : "GDT 6557RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 292u16 , name : "GDT 6111RP2/6511RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 293u16 , name : "GDT 6121RP2/6521RP2" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 310u16 , name : "GDT 6113RS/6513RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 311u16 , name : "GDT 6123RS/6523RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 312u16 , name : "GDT 6118RS/6518RS/6618RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 313u16 , name : "GDT 6128RS/6528RS/6628RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 314u16 , name : "GDT 6538RS/6638RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 315u16 , name : "GDT 6558RS/6658RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 316u16 , name : "GDT 6533RS/6633RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 317u16 , name : "GDT 6543RS/6643RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 318u16 , name : "GDT 6553RS/6653RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 319u16 , name : "GDT 6563RS/6663RS" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 358u16 , name : "GDT 7113RN/7513RN/7613RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 359u16 , name : "GDT 7123RN/7523RN/7623RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 360u16 , name : "GDT 7118RN/7518RN/7518RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 361u16 , name : "GDT 7128RN/7528RN/7628RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 362u16 , name : "GDT 7538RN/7638RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 363u16 , name : "GDT 7558RN/7658RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 364u16 , name : "GDT 7533RN/7633RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 365u16 , name : "GDT 7543RN/7643RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 366u16 , name : "GDT 7553RN/7653RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 367u16 , name : "GDT 7563RN/7663RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 470u16 , name : "GDT 4x13RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 471u16 , name : "GDT 4x23RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 502u16 , name : "GDT 8x13RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 503u16 , name : "GDT 8x23RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 508u16 , name : "GDT 8x33RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 509u16 , name : "GDT 8x43RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 510u16 , name : "GDT 8x53RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 511u16 , name : "GDT 8x63RZ" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 528u16 , name : "GDT 6519RD/6619RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 529u16 , name : "GDT 6529RD/6629RD" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 608u16 , name : "GDT 7519RN/7619RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 609u16 , name : "GDT 7529RN/7629RN" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 767u16 , name : "GDT MAXRP" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 768u16 , name : "GDT NEWRX" , subsystems : & [] } , Device { vendor_id : 4377u16 , id : 769u16 , name : "GDT NEWRX2" , subsystems : & [] }] }),
        (5466, Vendor { id : 5466u16 , name : "INNOMEDIA Inc" , devices : & [] }),
        (5294, Vendor { id : 5294u16 , name : "CTI, Inc" , devices : & [] }),
        (5014, Vendor { id : 5014u16 , name : "Cipher Systems Inc" , devices : & [] }),
        (34438, Vendor { id : 34438u16 , name : "SAP" , devices : & [Device { vendor_id : 34438u16 , id : 4112u16 , name : "vSMP Foundation controller [vSMP CTL]" , subsystems : & [] } , Device { vendor_id : 34438u16 , id : 4113u16 , name : "vSMP Foundation MEX/FLX controller [vSMP CTL]" , subsystems : & [] }] }),
        (4350, Vendor { id : 4350u16 , name : "Fast Multimedia AG" , devices : & [] }),
        (6856, Vendor { id : 6856u16 , name : "Aeroflex Gaisler" , devices : & [] }),
        (8345, Vendor { id : 8345u16 , name : "Rolling Wireless S.a.r.l." , devices : & [] }),
        (4248, Vendor { id : 4248u16 , name : "Quantum Designs (H.K.) Ltd" , devices : & [Device { vendor_id : 4248u16 , id : 1u16 , name : "QD-8500" , subsystems : & [] } , Device { vendor_id : 4248u16 , id : 2u16 , name : "QD-8580" , subsystems : & [] }] }),
        (7640, Vendor { id : 7640u16 , name : "AMD Pensando Systems" , devices : & [Device { vendor_id : 7640u16 , id : 2u16 , name : "DSC2 Elba Upstream Port" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" }] } , Device { vendor_id : 7640u16 , id : 8u16 , name : "DSC3 Salina Upstream Port" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4096u16 , name : "DSC Capri Upstream Port" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" }] } , Device { vendor_id : 7640u16 , id : 4097u16 , name : "DSC Virtual Downstream Port" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4098u16 , name : "DSC Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4099u16 , name : "DSC Ethernet Controller VF" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4100u16 , name : "DSC Management Controller" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4101u16 , name : "DSC NVMe Controller" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" }] } , Device { vendor_id : 7640u16 , id : 4102u16 , name : "DSC NVMe Controller VF" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" }] } , Device { vendor_id : 7640u16 , id : 4103u16 , name : "DSC Storage Accelerator" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16384u16 , name : "Naples 100Gb 2-port QSFP28 x16 8GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16385u16 , name : "Naples 100Gb 2-port QSFP28 x16 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16386u16 , name : "Naples 25Gb 2-port SFP28 x8 4GB" } , SubSystem { subvendor : 7640u16 , subdevice : 16391u16 , name : "DSP DSC-25 Ent 10/25G OCP3 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16392u16 , name : "DSP DSC-25 10/25G 2p SFP28 Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16394u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16396u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16397u16 , name : "DSP DSC-100 Ent 100Gb Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16398u16 , name : "DSC-25 10/25G 2-port 4G RAM 8G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 16404u16 , name : "DSC-100 40/100G 2-port 8G RAM 16G eMMC G1 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" }] } , Device { vendor_id : 7640u16 , id : 4105u16 , name : "DSC Ethernet Controller UPT" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" }] } , Device { vendor_id : 7640u16 , id : 4106u16 , name : "DSC Serial Port Controller" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" }] } , Device { vendor_id : 7640u16 , id : 4107u16 , name : "DSC vDPA Network Device VF" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" }] } , Device { vendor_id : 7640u16 , id : 4108u16 , name : "DSC PDS Core Management" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4109u16 , name : "DSC Virtio Network Device VF" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 4110u16 , name : "Distributed Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20481u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20483u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 Services Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20488u16 , name : "DSC25v2 10/25G 2p 32GB Spl Card" } , SubSystem { subvendor : 7640u16 , subdevice : 20489u16 , name : "DSC2-25 10/25G 2p SFP56 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20490u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R2" } , SubSystem { subvendor : 7640u16 , subdevice : 20491u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU" } , SubSystem { subvendor : 7640u16 , subdevice : 20492u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R3" } , SubSystem { subvendor : 7640u16 , subdevice : 20493u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4" } , SubSystem { subvendor : 7640u16 , subdevice : 20494u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R5" } , SubSystem { subvendor : 7640u16 , subdevice : 20495u16 , name : "DSC2-200 50/100/200G 2-port 32G RAM 64G eMMC G2 DPU R4-T" } , SubSystem { subvendor : 7640u16 , subdevice : 20992u16 , name : "DSC3-2Q400 100/200/400G 2-port DDR5 128G eMMC G3 Services Card" }] } , Device { vendor_id : 7640u16 , id : 4111u16 , name : "Register/Memory Resource Device" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] } , Device { vendor_id : 7640u16 , id : 4114u16 , name : "TAWK IPC Device" , subsystems : & [SubSystem { subvendor : 7640u16 , subdevice : 20993u16 , name : "POLLARA-1Q400 100/200/400G 1-port Card" }] }] }),
        (5473, Vendor { id : 5473u16 , name : "Viewgraphics Inc" , devices : & [] }),
        (4469, Vendor { id : 4469u16 , name : "Mitron Computer Inc." , devices : & [] }),
        (4654, Vendor { id : 4654u16 , name : "Xyratex" , devices : & [Device { vendor_id : 4654u16 , id : 30498u16 , name : "Napatech XL1" , subsystems : & [] } , Device { vendor_id : 4654u16 , id : 30500u16 , name : "Napatech XL2/XA" , subsystems : & [] } , Device { vendor_id : 4654u16 , id : 30505u16 , name : "Napatech XD" , subsystems : & [] }] }),
        (4776, Vendor { id : 4776u16 , name : "News Datacom" , devices : & [] }),
        (4686, Vendor { id : 4686u16 , name : "Cylink" , devices : & [] }),
        (7236, Vendor { id : 7236u16 , name : "Enmotus Inc" , devices : & [Device { vendor_id : 7236u16 , id : 4352u16 , name : "Fuzedrive NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7236u16 , id : 32768u16 , name : "8000 Storage IO Controller" , subsystems : & [] }] }),
        (7479, Vendor { id : 7479u16 , name : "NovaSparks" , devices : & [Device { vendor_id : 7479u16 , id : 19u16 , name : "PM3" , subsystems : & [] } , Device { vendor_id : 7479u16 , id : 20u16 , name : "PM4" , subsystems : & [] } , Device { vendor_id : 7479u16 , id : 21u16 , name : "PM4edge" , subsystems : & [] } , Device { vendor_id : 7479u16 , id : 22u16 , name : "PM4edge User Device" , subsystems : & [] }] }),
        (4794, Vendor { id : 4794u16 , name : "BittWare, Inc." , devices : & [Device { vendor_id : 4794u16 , id : 105u16 , name : "VectorPath S7t-VG6" , subsystems : & [SubSystem { subvendor : 4794u16 , subdevice : 46548u16 , name : "BW_BMC_IF" }] } , Device { vendor_id : 4794u16 , id : 121u16 , name : "VectorPath VP815" , subsystems : & [SubSystem { subvendor : 4794u16 , subdevice : 46548u16 , name : "BW_BMC_IF" }] } , Device { vendor_id : 4794u16 , id : 122u16 , name : "VectorPath VP708" , subsystems : & [] }] }),
        (8483, Vendor { id : 8483u16 , name : "Shanghai Warpdrive Technology Co., Ltd" , devices : & [] }),
        (5518, Vendor { id : 5518u16 , name : "Lara Technology Inc" , devices : & [] }),
        (5252, Vendor { id : 5252u16 , name : "Logic Corporation" , devices : & [] }),
        (8385, Vendor { id : 8385u16 , name : "Suzhou Yige Technology Co., Ltd." , devices : & [Device { vendor_id : 8385u16 , id : 4660u16 , name : "YG Testchip" , subsystems : & [] }] }),
        (7168, Vendor { id : 7168u16 , name : "WCH (Nanjing Qinheng Microelectronics Co., Ltd.)" , devices : & [Device { vendor_id : 7168u16 , id : 8560u16 , name : "CH351 PCIe Parallel Port Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 8819u16 , name : "CH351 PCIe Dual Port Serial Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 12368u16 , name : "CH382L PCIe Parallel Port Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 12880u16 , name : "CH384 Dual Port Serial and Parallel Port Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 12882u16 , name : "CH382 PCIe Dual Port Serial Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 12883u16 , name : "CH352/CH382 PCI/PCIe Dual Port Serial Adapter" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 13424u16 , name : "CH384 Serial Adapter, 4-port mode" , subsystems : & [] } , Device { vendor_id : 7168u16 , id : 14419u16 , name : "CH384 Serial Adapter, 8-port mode" , subsystems : & [] }] }),
        (6506, Vendor { id : 6506u16 , name : "Sensory Networks Inc." , devices : & [Device { vendor_id : 6506u16 , id : 257u16 , name : "NodalCore C-1000 Content Classification Accelerator" , subsystems : & [] } , Device { vendor_id : 6506u16 , id : 258u16 , name : "NodalCore C-2000 Content Classification Accelerator" , subsystems : & [] } , Device { vendor_id : 6506u16 , id : 261u16 , name : "NodalCore C-3000 Content Classification Accelerator" , subsystems : & [] }] }),
        (4670, Vendor { id : 4670u16 , name : "Simutech, Inc." , devices : & [] }),
        (4388, Vendor { id : 4388u16 , name : "Leutron Vision AG" , devices : & [Device { vendor_id : 4388u16 , id : 9601u16 , name : "Picport Monochrome" , subsystems : & [] }] }),
        (6262, Vendor { id : 6262u16 , name : "L-3 Communications" , devices : & [Device { vendor_id : 6262u16 , id : 41217u16 , name : "VigraWATCH PCI" , subsystems : & [] } , Device { vendor_id : 6262u16 , id : 41218u16 , name : "VigraWATCH PMC" , subsystems : & [] } , Device { vendor_id : 6262u16 , id : 41219u16 , name : "Vigra I/O" , subsystems : & [] }] }),
        (4893, Vendor { id : 4893u16 , name : "Sysmic, Inc." , devices : & [] }),
        (4361, Vendor { id : 4361u16 , name : "Cogent Data Technologies, Inc." , devices : & [Device { vendor_id : 4361u16 , id : 5120u16 , name : "EM110TX [EX110TX]" , subsystems : & [] }] }),
        (4700, Vendor { id : 4700u16 , name : "Aurora Technologies, Inc." , devices : & [Device { vendor_id : 4700u16 , id : 257u16 , name : "Saturn 4520P" , subsystems : & [] } , Device { vendor_id : 4700u16 , id : 1600u16 , name : "Aries 16000P" , subsystems : & [] }] }),
        (4506, Vendor { id : 4506u16 , name : "Mind Share, Inc." , devices : & [] }),
        (6156, Vendor { id : 6156u16 , name : "IEI Integration Corp" , devices : & [] }),
        (4840, Vendor { id : 4840u16 , name : "Crisc Corp" , devices : & [] }),
        (6425, Vendor { id : 6425u16 , name : "Soltek Computer Inc." , devices : & [] }),
        (51966, Vendor { id : 51966u16 , name : "Thales" , devices : & [Device { vendor_id : 51966u16 , id : 3u16 , name : "Luna K3 Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 51966u16 , id : 6u16 , name : "Luna PCI-e 3000 Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 51966u16 , id : 7u16 , name : "Luna K6 Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 51966u16 , id : 8u16 , name : "Luna K7 Hardware Security Module" , subsystems : & [] }] }),
        (5724, Vendor { id : 5724u16 , name : "Gidel Ltd." , devices : & [Device { vendor_id : 5724u16 , id : 21345u16 , name : "PROCStarII60-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 21346u16 , name : "PROCStarII60-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 21348u16 , name : "PROCStarII60-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 21557u16 , name : "ProcSparkII" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22113u16 , name : "ProcE60" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22241u16 , name : "ProcE180" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22801u16 , name : "ProcStarIII110-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22802u16 , name : "ProcStarIII110-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22803u16 , name : "ProcStarIII110-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22804u16 , name : "ProcStarIII110-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22817u16 , name : "ProcStarIII150-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22818u16 , name : "ProcStarIII150-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22819u16 , name : "ProcStarIII150-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22820u16 , name : "ProcStarIII150-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22833u16 , name : "ProcStarIII260-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22834u16 , name : "ProcStarIII260-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22835u16 , name : "ProcStarIII260-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22836u16 , name : "ProcStarIII260-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22849u16 , name : "ProcStarIII340-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22850u16 , name : "ProcStarIII340-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22851u16 , name : "ProcStarIII340-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 22852u16 , name : "ProcStarIII340-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23041u16 , name : "ProceIII80" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23057u16 , name : "ProceIII110" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23073u16 , name : "ProceIII150" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23089u16 , name : "ProceIII260" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23105u16 , name : "ProceIII340" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23377u16 , name : "ProceIV360" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23393u16 , name : "ProceIV530" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23409u16 , name : "ProceIV820" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23553u16 , name : "ProcStarIV80-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23554u16 , name : "ProcStarIV80-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23555u16 , name : "ProcStarIV80-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23556u16 , name : "ProcStarIV80-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23569u16 , name : "ProcStarIV110-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23570u16 , name : "ProcStarIV110-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23571u16 , name : "ProcStarIV110-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23572u16 , name : "ProcStarIV110-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23633u16 , name : "ProcStarIV360-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23634u16 , name : "ProcStarIV360-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23635u16 , name : "ProcStarIV360-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23636u16 , name : "ProcStarIV360-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23649u16 , name : "ProcStarIV530-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23650u16 , name : "ProcStarIV530-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23651u16 , name : "ProcStarIV530-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23652u16 , name : "ProcStarIV530-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23665u16 , name : "ProcStarIV820-1" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23666u16 , name : "ProcStarIV820-2" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23667u16 , name : "ProcStarIV820-3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23668u16 , name : "ProcStarIV820-4" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23809u16 , name : "Proc10480" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 23825u16 , name : "Proc104110" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 24321u16 , name : "ProceV_A3" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 24337u16 , name : "ProceV_A7" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 24353u16 , name : "ProceV_AB" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 24369u16 , name : "ProceV_D5" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 24385u16 , name : "ProceV_D8" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 26418u16 , name : "Proc6M" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 26674u16 , name : "Proc12M" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 28929u16 , name : "Proc10a_27" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 28945u16 , name : "Proc10a_48" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 28961u16 , name : "Proc10a_66" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 28993u16 , name : "Proc10a_115" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29057u16 , name : "Proc10a_27S" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29073u16 , name : "Proc10a_48S" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29089u16 , name : "Proc10a_66S" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29105u16 , name : "Proc10A" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29361u16 , name : "HawkEye" , subsystems : & [] } , Device { vendor_id : 5724u16 , id : 29617u16 , name : "Proc10s" , subsystems : & [] }] }),
        (7378, Vendor { id : 7378u16 , name : "Seskion GmbH" , devices : & [Device { vendor_id : 7378u16 , id : 769u16 , name : "Simulyzer-RT CompactPCI Serial DIO-1 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 770u16 , name : "Simulyzer-RT CompactPCI Serial PSI5-ECU-1 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 771u16 , name : "Simulyzer-RT CompactPCI Serial PSI5-SIM-1 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 772u16 , name : "Simulyzer-RT CompactPCI Serial PWR-ANA-1 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 773u16 , name : "Simulyzer-RT CompactPCI Serial CAN-1 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 774u16 , name : "Simulyzer-RT CompactPCI Serial CAN-2 card (CAN-FD)" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 775u16 , name : "Simulyzer-RT CompactPCI Serial DIO-2 card [Xilinx Zynq UltraScale+]" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 776u16 , name : "Simulyzer-RT CompactPCI Serial SENT-DIO-2 card" , subsystems : & [] } , Device { vendor_id : 7378u16 , id : 777u16 , name : "Simulyzer-RT CompactPCI Serial SN-ADC card" , subsystems : & [] }] }),
        (7663, Vendor { id : 7663u16 , name : "Ampere Computing, LLC" , devices : & [Device { vendor_id : 7663u16 , id : 57349u16 , name : "eMAG PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57350u16 , name : "eMAG PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57351u16 , name : "eMAG PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57352u16 , name : "eMAG PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57353u16 , name : "eMAG PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57354u16 , name : "eMAG PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57355u16 , name : "eMAG PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57356u16 , name : "eMAG PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57600u16 , name : "Altra PCI Express Root Complex A" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57601u16 , name : "Altra PCI Express Root Port a0" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57602u16 , name : "Altra PCI Express Root Port a1" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57603u16 , name : "Altra PCI Express Root Port a2" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57604u16 , name : "Altra PCI Express Root Port a3" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57605u16 , name : "Altra PCI Express Root Port a4" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57606u16 , name : "Altra PCI Express Root Port a5" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57607u16 , name : "Altra PCI Express Root Port a6" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57608u16 , name : "Altra PCI Express Root Port a7" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57616u16 , name : "Altra PCI Express Root Complex B" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57617u16 , name : "Altra PCI Express Root Port b0" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57618u16 , name : "Altra PCI Express Root Port b1" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57619u16 , name : "Altra PCI Express Root Port b2" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57620u16 , name : "Altra PCI Express Root Port b3" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57621u16 , name : "Altra PCI Express Root Port b4" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57622u16 , name : "Altra PCI Express Root Port b5" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57623u16 , name : "Altra PCI Express Root Port b6" , subsystems : & [] } , Device { vendor_id : 7663u16 , id : 57624u16 , name : "Altra PCI Express Root Port b7" , subsystems : & [] }] }),
        (3756, Vendor { id : 3756u16 , name : "SHF Communication Technologies AG" , devices : & [Device { vendor_id : 3756u16 , id : 8u16 , name : "Ethernet Powerlink Managing Node 01" , subsystems : & [] }] }),
        (4168, Vendor { id : 4168u16 , name : "Elsa AG" , devices : & [Device { vendor_id : 4168u16 , id : 3168u16 , name : "Gladiac MX" , subsystems : & [] } , Device { vendor_id : 4168u16 , id : 3362u16 , name : "Quadro4 900XGL [ELSA GLoria4 900XGL]" , subsystems : & [] } , Device { vendor_id : 4168u16 , id : 4096u16 , name : "QuickStep 1000" , subsystems : & [] } , Device { vendor_id : 4168u16 , id : 12288u16 , name : "QuickStep 3000" , subsystems : & [] } , Device { vendor_id : 4168u16 , id : 35073u16 , name : "Gloria XL" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 2357u16 , name : "GLoria XL (Virge)" }] }] }),
        (5072, Vendor { id : 5072u16 , name : "Techsan Electronics Co Ltd" , devices : & [Device { vendor_id : 5072u16 , id : 8451u16 , name : "B2C2 FlexCopII DVB chip / Technisat SkyStar2 DVB card" , subsystems : & [] } , Device { vendor_id : 5072u16 , id : 8452u16 , name : "B2C2 FlexCopIII DVB chip / Technisat SkyStar2 DVB card (rev 01)" , subsystems : & [] } , Device { vendor_id : 5072u16 , id : 8704u16 , name : "B2C2 FlexCopIII DVB chip / Technisat SkyStar2 DVB card" , subsystems : & [] }] }),
        (4841, Vendor { id : 4841u16 , name : "GE Spacenet" , devices : & [] }),
        (6244, Vendor { id : 6244u16 , name : "SilverBack" , devices : & [Device { vendor_id : 6244u16 , id : 8464u16 , name : "ISNAP 2110" , subsystems : & [] }] }),
        (776, Vendor { id : 776u16 , name : "ZyXEL Communications Corporation (Wrong ID)" , devices : & [] }),
        (18048, Vendor { id : 18048u16 , name : "Umax Computer Corp" , devices : & [] }),
        (7704, Vendor { id : 7704u16 , name : "Beijing GuangRunTong Technology Development Co.,Ltd" , devices : & [] }),
        (5246, Vendor { id : 5246u16 , name : "Matsushita Graphic Communication Systems, Inc." , devices : & [] }),
        (4419, Vendor { id : 4419u16 , name : "NetPower, Inc" , devices : & [] }),
        (5069, Vendor { id : 5069u16 , name : "Compatible Systems Corporation" , devices : & [] }),
        (7530, Vendor { id : 7530u16 , name : "Aquantia Corp." , devices : & [Device { vendor_id : 7530u16 , id : 1u16 , name : "AQC107 NBase-T/IEEE 802.3bz Ethernet Controller [AQtion]" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 26752u16 , name : "LREC6880BT Single-port 10Gb Ethernet Network Adapter" }] } , Device { vendor_id : 7530u16 , id : 177u16 , name : "AQtion AQC100 NBase-T/IEEE 802.3an Ethernet Controller [Atlantic 10G]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34634u16 , name : "XG-C100F 10GbE SFP+ Ethernet Adapter" }] } , Device { vendor_id : 7530u16 , id : 192u16 , name : "Antigua NBase-T/IEEE 802.3an Ethernet Controller - Engineering Sample" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 1216u16 , name : "AQtion AQC113 NBase-T/IEEE 802.3an Ethernet Controller [Antigua 10G]" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4179u16 , name : "LRES1053PT Quad-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8273u16 , name : "LRES2051PT Single-port 10Gb Ethernet Network Adapter" }] } , Device { vendor_id : 7530u16 , id : 1969u16 , name : "AQtion AQC107 NBase-T/IEEE 802.3an Ethernet Controller [Atlantic 10G]" , subsystems : & [SubSystem { subvendor : 7082u16 , subdevice : 1969u16 , name : "QM2-2P10G1TA [QXG 10GbE Network Adapter]" } , SubSystem { subvendor : 7082u16 , subdevice : 1970u16 , name : "QM2-2P10G1TA [QM2 Expansion Adapter]" } , SubSystem { subvendor : 19538u16 , subdevice : 26752u16 , name : "LREC6880BT Single-port 10Gb Ethernet Network Adapter" }] } , Device { vendor_id : 7530u16 , id : 2225u16 , name : "AQC108 NBase-T/IEEE 802.3bz Ethernet Controller [AQtion]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 4529u16 , name : "AQC111 NBase-T/IEEE 802.3bz Ethernet Controller [AQtion]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 4785u16 , name : "AQC112 NBase-T/IEEE 802.3bz Ethernet Controller [AQtion]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 4800u16 , name : "AQtion AQC115C NBase-T/IEEE 802.3bz Ethernet Controller [Antigua 2.5G]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 5312u16 , name : "AQC113C NBase-T/IEEE 802.3an Ethernet Controller [Marvell Scalable mGig]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 32945u16 , name : "AQtion AQC100S NBase-T/IEEE 802.3an Ethernet Controller [Atlantic 10G]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 34737u16 , name : "AQtion AQC107S NBase-T/IEEE 802.3an Ethernet Controller [Atlantic 10G]" , subsystems : & [SubSystem { subvendor : 5816u16 , subdevice : 29409u16 , name : "Solo 10G Thunderbolt 3 PCIe 10Gb Ethernet Adapter" } , SubSystem { subvendor : 7290u16 , subdevice : 56875u16 , name : "Thunderbolt 10G Ethernet Adapter" }] } , Device { vendor_id : 7530u16 , id : 37824u16 , name : "AQtion AQC114CS NBase-T/IEEE 802.3bz Ethernet Controller [Antigua 5G]" , subsystems : & [] } , Device { vendor_id : 7530u16 , id : 38080u16 , name : "AQtion AQC113CS NBase-T/IEEE 802.3an Ethernet Controller [Antigua 10G]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34805u16 , name : "ProArt X570-CREATOR WIFI" } , SubSystem { subvendor : 4163u16 , subdevice : 34834u16 , name : "ASUS XG-C100C 10G PCI-E Network Adapter" }] } , Device { vendor_id : 7530u16 , id : 53511u16 , name : "AQtion AQC107 NBase-T/IEEE 802.3an Ethernet Controller [Atlantic 10G]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34625u16 , name : "XG-C100C" }] } , Device { vendor_id : 7530u16 , id : 53512u16 , name : "AQC108 NBase-T/IEEE 802.3bz Ethernet Controller [AQtion]" , subsystems : & [] }] }),
        (5165, Vendor { id : 5165u16 , name : "Pix stream Inc" , devices : & [] }),
        (4782, Vendor { id : 4782u16 , name : "Alteon Networks Inc." , devices : & [Device { vendor_id : 4782u16 , id : 1u16 , name : "AceNIC Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 260u16 , name : "Gigabit Ethernet-SX PCI Adapter" } , SubSystem { subvendor : 4782u16 , subdevice : 1u16 , name : "Gigabit Ethernet-SX (Universal)" }] } , Device { vendor_id : 4782u16 , id : 2u16 , name : "AceNIC Gigabit Ethernet (Copper)" , subsystems : & [SubSystem { subvendor : 4265u16 , subdevice : 32770u16 , name : "Acenic Gigabit Ethernet" } , SubSystem { subvendor : 4782u16 , subdevice : 2u16 , name : "Gigabit Ethernet-T (3C986-T)" }] } , Device { vendor_id : 4782u16 , id : 250u16 , name : "Farallon PN9100-T Gigabit Ethernet" , subsystems : & [] }] }),
        (7603, Vendor { id : 7603u16 , name : "Unisoc (Shanghai) Technologies Co., Ltd." , devices : & [] }),
        (5725, Vendor { id : 5725u16 , name : "Hsing Tech. Enterprise Co., Ltd." , devices : & [] }),
        (21299, Vendor { id : 21299u16 , name : "S3 Graphics Ltd." , devices : & [Device { vendor_id : 21299u16 , id : 1361u16 , name : "Plato/PX (system)" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 22065u16 , name : "86c325 [ViRGE]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34816u16 , name : "86c866 [Vision 866]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34817u16 , name : "86c964 [Vision 964]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34832u16 , name : "86c764_0 [Trio 32 vers 0]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34833u16 , name : "86c764/765 [Trio32/64/64V+]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34834u16 , name : "86cM65 [Aurora64V+]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34835u16 , name : "86c764_3 [Trio 32/64 vers 3]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34836u16 , name : "86c767 [Trio 64UV+]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34837u16 , name : "86cM65 [Aurora 128]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34877u16 , name : "86c988 [ViRGE/VX]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34928u16 , name : "FireGL" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34944u16 , name : "86c868 [Vision 868 VRAM] vers 0" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34945u16 , name : "86c868 [Vision 868 VRAM] vers 1" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34946u16 , name : "86c868 [Vision 868 VRAM] vers 2" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34947u16 , name : "86c868 [Vision 868 VRAM] vers 3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34992u16 , name : "86c928 [Vision 928 VRAM] vers 0" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34993u16 , name : "86c928 [Vision 928 VRAM] vers 1" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34994u16 , name : "86c928 [Vision 928 VRAM] vers 2" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 34995u16 , name : "86c928 [Vision 928 VRAM] vers 3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35008u16 , name : "86c864 [Vision 864 DRAM] vers 0" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35009u16 , name : "86c864 [Vision 864 DRAM] vers 1" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35010u16 , name : "86c864 [Vision 864-P DRAM] vers 2" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35011u16 , name : "86c864 [Vision 864-P DRAM] vers 3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35024u16 , name : "86c964 [Vision 964 VRAM] vers 0" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35025u16 , name : "86c964 [Vision 964 VRAM] vers 1" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35026u16 , name : "86c964 [Vision 964-P VRAM] vers 2" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35027u16 , name : "86c964 [Vision 964-P VRAM] vers 3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35056u16 , name : "86c968 [Vision 968 VRAM] rev 0" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35057u16 , name : "86c968 [Vision 968 VRAM] rev 1" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35058u16 , name : "86c968 [Vision 968 VRAM] rev 2" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35059u16 , name : "86c968 [Vision 968 VRAM] rev 3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35072u16 , name : "86c755 [Trio 64V2/DX]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 35072u16 , name : "86C775 Trio64V2/DX" }] } , Device { vendor_id : 21299u16 , id : 35073u16 , name : "86c775/86c785 [Trio 64V2/DX or /GX]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 35073u16 , name : "86C775 Trio64V2/DX, 86C785 Trio64V2/GX" }] } , Device { vendor_id : 21299u16 , id : 35074u16 , name : "Plato/PX" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35075u16 , name : "Trio 3D business multimedia" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35076u16 , name : "86c365, 86c366 [Trio 3D]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 219u16 , name : "Integrated Trio3D" } , SubSystem { subvendor : 18499u16 , subdevice : 12618u16 , name : "Terminator 128/3D GLH" } , SubSystem { subvendor : 21299u16 , subdevice : 35076u16 , name : "86C365 Trio3D AGP" }] } , Device { vendor_id : 21299u16 , id : 35077u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35078u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35079u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35080u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35081u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35082u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35083u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35084u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35085u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35086u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35087u16 , name : "Trio 64V+ family" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35329u16 , name : "86c375 [ViRGE/DX] or 86c385 [ViRGE/GX]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45106u16 , name : "ViRGE/GX" } , SubSystem { subvendor : 4276u16 , subdevice : 5655u16 , name : "Nitro 3D" } , SubSystem { subvendor : 4276u16 , subdevice : 5911u16 , name : "Nitro 3D" } , SubSystem { subvendor : 21299u16 , subdevice : 35329u16 , name : "ViRGE/DX" }] } , Device { vendor_id : 21299u16 , id : 35344u16 , name : "ViRGE/GX2" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 35344u16 , name : "Stealth 3D 4000" }] } , Device { vendor_id : 21299u16 , id : 35347u16 , name : "86c360 [Trio 3D/1X], 86c362, 86c368 [Trio 3D/2X]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 35347u16 , name : "Trio3D/2X" }] } , Device { vendor_id : 21299u16 , id : 35360u16 , name : "86c794 [Savage 3D]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 35360u16 , name : "86C391 Savage3D" }] } , Device { vendor_id : 21299u16 , id : 35361u16 , name : "86c390 [Savage 3D/MV]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 35361u16 , name : "86C390 Savage3D/MV" }] } , Device { vendor_id : 21299u16 , id : 35362u16 , name : "Savage 4" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32872u16 , name : "Savage 4" } , SubSystem { subvendor : 4147u16 , subdevice : 32873u16 , name : "Savage 4" } , SubSystem { subvendor : 4147u16 , subdevice : 33040u16 , name : "Savage 4 LT" } , SubSystem { subvendor : 4189u16 , subdevice : 24u16 , name : "SR9 8Mb SDRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 42u16 , name : "SR9 Pro 16Mb SDRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 58u16 , name : "SR9 Pro 32Mb SDRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 2351u16 , name : "SR9 Pro+ 16Mb SGRAM" } , SubSystem { subvendor : 4242u16 , subdevice : 16903u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18432u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18439u16 , name : "SpeedStar A90" } , SubSystem { subvendor : 4242u16 , subdevice : 18440u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18441u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18446u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18692u16 , name : "Stealth III S520" } , SubSystem { subvendor : 4242u16 , subdevice : 18693u16 , name : "SpeedStar A200" } , SubSystem { subvendor : 4242u16 , subdevice : 18953u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 18955u16 , name : "Stealth III S540 Xtreme" } , SubSystem { subvendor : 4242u16 , subdevice : 18959u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4242u16 , subdevice : 19969u16 , name : "Stealth III S540" } , SubSystem { subvendor : 4354u16 , subdevice : 4125u16 , name : "3d Blaster Savage 4" } , SubSystem { subvendor : 4354u16 , subdevice : 4126u16 , name : "3d Blaster Savage 4" } , SubSystem { subvendor : 21299u16 , subdevice : 33024u16 , name : "86C394-397 Savage4 SDRAM 100" } , SubSystem { subvendor : 21299u16 , subdevice : 33040u16 , name : "86C394-397 Savage4 SDRAM 110" } , SubSystem { subvendor : 21299u16 , subdevice : 33061u16 , name : "86C394-397 Savage4 SDRAM 125" } , SubSystem { subvendor : 21299u16 , subdevice : 33091u16 , name : "86C394-397 Savage4 SDRAM 143" } , SubSystem { subvendor : 21299u16 , subdevice : 35362u16 , name : "86C394-397 Savage4" } , SubSystem { subvendor : 21299u16 , subdevice : 35374u16 , name : "86C394-397 Savage4 32bit" } , SubSystem { subvendor : 21299u16 , subdevice : 37157u16 , name : "86C394-397 Savage4 SGRAM 125" } , SubSystem { subvendor : 21299u16 , subdevice : 37187u16 , name : "86C394-397 Savage4 SGRAM 143" }] } , Device { vendor_id : 21299u16 , id : 35363u16 , name : "Savage 4" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35365u16 , name : "ProSavage PM133" , subsystems : & [SubSystem { subvendor : 771u16 , subdevice : 771u16 , name : "D9840-60001 [Brio BA410 Motherboard]" }] } , Device { vendor_id : 21299u16 , id : 35366u16 , name : "ProSavage KM133" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35840u16 , name : "ViRGE/M3" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35841u16 , name : "ViRGE/MX" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "ViRGE/MX" }] } , Device { vendor_id : 21299u16 , id : 35842u16 , name : "ViRGE/MX+" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35843u16 , name : "ViRGE/MX+MV" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35856u16 , name : "86C270-294 [SavageMX-MV]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35857u16 , name : "82C270-294 [SavageMX]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35858u16 , name : "86C270-294 [SavageIX-MV]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 383u16 , name : "ThinkPad T20/T22" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "86C584 SuperSavage/IXC Toshiba" }] } , Device { vendor_id : 21299u16 , id : 35859u16 , name : "86C270-294 [SavageIX]" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" }] } , Device { vendor_id : 21299u16 , id : 35874u16 , name : "SuperSavage MX/128" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35876u16 , name : "SuperSavage MX/64" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35878u16 , name : "SuperSavage MX/64C" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35882u16 , name : "SuperSavage IX/128 SDR" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35883u16 , name : "SuperSavage IX/128 DDR" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35884u16 , name : "SuperSavage IX/64 SDR" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35885u16 , name : "SuperSavage IX/64 DDR" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 35886u16 , name : "SuperSavage IX/C SDR" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 508u16 , name : "ThinkPad T23" }] } , Device { vendor_id : 21299u16 , id : 35887u16 , name : "SuperSavage IX/C DDR" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36097u16 , name : "86C380 [ProSavageDDR K4M266]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36098u16 , name : "VT8636A [ProSavage KN133] AGP4X VGA Controller (TwisterK)" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36099u16 , name : "VT8751 [ProSavageDDR P4M266]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36100u16 , name : "VT8375 [ProSavage8 KM266/KL266]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36352u16 , name : "DeltaChrome" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36390u16 , name : "ProSavage" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36416u16 , name : "2300E Graphics Processor" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36424u16 , name : "Matrix [Chrome S25 / S27]" , subsystems : & [SubSystem { subvendor : 21299u16 , subdevice : 304u16 , name : "Chrome S27 256M DDR2" }] } , Device { vendor_id : 21299u16 , id : 36896u16 , name : "86C920 [Chrome 460 ESV]" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36931u16 , name : "Chrome 430 GT" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36933u16 , name : "Chrome 430 ULP / 435 ULP / 440 GTX" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36960u16 , name : "Chrome 530 GT / 5400E" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 36976u16 , name : "Chrome 5400EW" , subsystems : & [] } , Device { vendor_id : 21299u16 , id : 37122u16 , name : "86C410 [Savage 2000]" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 22834u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 22836u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 22866u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 22868u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 23093u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 23095u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 23125u16 , name : "Viper II Z200" } , SubSystem { subvendor : 4242u16 , subdevice : 23127u16 , name : "Viper II Z200" }] } , Device { vendor_id : 21299u16 , id : 51712u16 , name : "SonicVibes" , subsystems : & [] }] }),
        (4704, Vendor { id : 4704u16 , name : "Intersil Corporation" , devices : & [Device { vendor_id : 4704u16 , id : 14450u16 , name : "ISL3872 [Prism 3]" , subsystems : & [SubSystem { subvendor : 5224u16 , subdevice : 514u16 , name : "LAN-Express IEEE 802.11b Wireless LAN" }] } , Device { vendor_id : 4704u16 , id : 14451u16 , name : "ISL3874 [Prism 2.5]/ISL3872 [Prism 3]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 1u16 , name : "3CRDW696 rev A Wireless LAN PCI Adapter [ISL3874]" } , SubSystem { subvendor : 4303u16 , subdevice : 4457u16 , name : "MBH7WM01-8734 802.11b Wireless Mini PCI Card [ISL3874]" } , SubSystem { subvendor : 4348u16 , subdevice : 53257u16 , name : "WN-B11/PCIH" } , SubSystem { subvendor : 4486u16 , subdevice : 13569u16 , name : "DWL-520 Wireless PCI Adapter (rev A or B) [ISL3874]" } , SubSystem { subvendor : 4486u16 , subdevice : 14080u16 , name : "DWL-520 Wireless PCI Adapter (rev E1) [ISL3872]" } , SubSystem { subvendor : 4997u16 , subdevice : 16645u16 , name : "MA311 802.11b wireless adapter [ISL3874]" } , SubSystem { subvendor : 5736u16 , subdevice : 1044u16 , name : "HWP01170-01 802.11b PCI Wireless Adapter" } , SubSystem { subvendor : 5736u16 , subdevice : 5126u16 , name : "802MIP 802.11b Mini PCI Adapter [ISL3874]" } , SubSystem { subvendor : 5797u16 , subdevice : 5633u16 , name : "AIR.mate PC-400 PCI Wireless LAN Adapter" } , SubSystem { subvendor : 5822u16 , subdevice : 8193u16 , name : "CTX712" } , SubSystem { subvendor : 5943u16 , subdevice : 14452u16 , name : "WMP11 v1 802.11b Wireless-B PCI Adapter [ISL3874]" } , SubSystem { subvendor : 16435u16 , subdevice : 28723u16 , name : "PCW200 802.11b Wireless PCI Adapter [ISL3874]" } , SubSystem { subvendor : 32902u16 , subdevice : 9488u16 , name : "M3AWEB Wireless 802.11b MiniPCI Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 9491u16 , name : "Wireless 802.11b MiniPCI Adapter" }] } , Device { vendor_id : 4704u16 , id : 14455u16 , name : "ISL3877 [Prism Indigo]" , subsystems : & [] } , Device { vendor_id : 4704u16 , id : 14470u16 , name : "ISL3886 [Prism Javelin/Prism Xbow]" , subsystems : & [SubSystem { subvendor : 6095u16 , subdevice : 55u16 , name : "XG-901 and clones Wireless Adapter" }] } , Device { vendor_id : 4704u16 , id : 14480u16 , name : "ISL3890 [Prism GT/Prism Duette]/ISL3886 [Prism Javelin/Prism Xbow]" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 10242u16 , name : "SMC2802W V1 Wireless PCI Adapter [ISL3890]" } , SubSystem { subvendor : 4280u16 , subdevice : 10293u16 , name : "SMC2835W Wireless Cardbus Adapter" } , SubSystem { subvendor : 4280u16 , subdevice : 43061u16 , name : "SMC2835W V2 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 16899u16 , name : "WN4201B" } , SubSystem { subvendor : 4371u16 , subdevice : 33281u16 , name : "T-Com T-Sinus 154pcicard Wireless PCI Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 45825u16 , name : "T-Sinus 154card Cardbus" } , SubSystem { subvendor : 4371u16 , subdevice : 60931u16 , name : "SMC2802W V2 Wireless PCI Adapter [ISL3886]" } , SubSystem { subvendor : 4371u16 , subdevice : 60936u16 , name : "SMC2835W V3 EU Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 12802u16 , name : "DWL-G650 A1 Wireless Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 49412u16 , name : "CG-WLCB54GT Wireless Adapter" } , SubSystem { subvendor : 4704u16 , subdevice : 0u16 , name : "WG511 v1 54 Mbps Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 18432u16 , name : "WG511 v2/v3 54 Mbps Wireless PC Card" } , SubSystem { subvendor : 5797u16 , subdevice : 5637u16 , name : "ALLNET ALL0271 Wireless PCI Adapter" } , SubSystem { subvendor : 6095u16 , subdevice : 20u16 , name : "XG-600 and clones Wireless Adapter" } , SubSystem { subvendor : 6095u16 , subdevice : 32u16 , name : "XG-900 and clones Wireless Adapter" } , SubSystem { subvendor : 6270u16 , subdevice : 13315u16 , name : "G-110 802.11g Wireless Cardbus Adapter" }] } , Device { vendor_id : 4704u16 , id : 33072u16 , name : "HMP8130 NTSC/PAL Video Decoder" , subsystems : & [] } , Device { vendor_id : 4704u16 , id : 33073u16 , name : "HMP8131 NTSC/PAL Video Decoder" , subsystems : & [] } , Device { vendor_id : 4704u16 , id : 65535u16 , name : "ISL3886IK" , subsystems : & [SubSystem { subvendor : 4704u16 , subdevice : 0u16 , name : "Senao 3054MP+ (J) mini-PCI WLAN 802.11g adapter" }] }] }),
        (4595, Vendor { id : 4595u16 , name : "Keithley Metrabyte" , devices : & [Device { vendor_id : 4595u16 , id : 17u16 , name : "KPCI-PIO24" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16896u16 , name : "42x0-SMU" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16901u16 , name : "4205-VPU" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16917u16 , name : "4200-VPU" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16918u16 , name : "422x-PxU" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16928u16 , name : "4200-CVU" , subsystems : & [] } , Device { vendor_id : 4595u16 , id : 16929u16 , name : "4210-CVU" , subsystems : & [] }] }),
        (7062, Vendor { id : 7062u16 , name : "Western Digital" , devices : & [Device { vendor_id : 7062u16 , id : 8704u16 , name : "Ultrastar DC SN630 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 8705u16 , name : "Ultrastar DC SN630 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 8960u16 , name : "Ultrastar DC SN840 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9216u16 , name : "Ultrastar DC SN640 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9217u16 , name : "Ultrastar DC SN640 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9218u16 , name : "Ultrastar DC SN640 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9220u16 , name : "Ultrastar DC SN640 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9472u16 , name : "Ultrastar DC SN840 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9728u16 , name : "Ultrastar DC ZN540 ZNS NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9984u16 , name : "Ultrastar DC SN650 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9985u16 , name : "Ultrastar DC SN650 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 9986u16 , name : "Ultrastar DC SN650 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 10016u16 , name : "Ultrastar DC SN650 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 10017u16 , name : "Ultrastar DC SN650 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 10018u16 , name : "Ultrastar DC SN655 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 10064u16 , name : "Ultrastar DC SN861 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 10065u16 , name : "Ultrastar DC SN861 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 12289u16 , name : "RapidFlex C2000 NVMe Initiator" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 14100u16 , name : "PC SN730 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7062u16 , id : 14132u16 , name : "PC SN730 NVMe SSD" , subsystems : & [] }] }),
        (7151, Vendor { id : 7151u16 , name : "Lantiq" , devices : & [Device { vendor_id : 7151u16 , id : 17u16 , name : "MIPS SoC PCI Express Port" , subsystems : & [] }] }),
        (4855, Vendor { id : 4855u16 , name : "Cognex" , devices : & [] }),
        (6405, Vendor { id : 6405u16 , name : "Micronas USA, Inc." , devices : & [] }),
        (7082, Vendor { id : 7082u16 , name : "QNAP Systems, Inc." , devices : & [] }),
        (5040, Vendor { id : 5040u16 , name : "Maxspeed Corp" , devices : & [] }),
        (5548, Vendor { id : 5548u16 , name : "North Atlantic Instruments" , devices : & [Device { vendor_id : 5548u16 , id : 26771u16 , name : "3U OpenVPX Multi-function I/O Board [Model 68C3]" , subsystems : & [] }] }),
        (6741, Vendor { id : 6741u16 , name : "Rohde & Schwarz DVS GmbH" , devices : & [Device { vendor_id : 6741u16 , id : 16u16 , name : "SDStationOEM" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 17u16 , name : "SDStationOEM II" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 32u16 , name : "Centaurus" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 33u16 , name : "Centaurus II" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 34u16 , name : "Centaurus II LT" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 48u16 , name : "CLIPSTER-VPU 1.x (Hugo)" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 64u16 , name : "Hydra Cinema (JPEG)" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 80u16 , name : "CLIPSTER-VPU 2.x (DigiLab)" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 96u16 , name : "CLIPSTER-DCI 2.x (HydraX)" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 97u16 , name : "Atomix" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 98u16 , name : "Atomix LT" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 99u16 , name : "Atomix HDMI" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 100u16 , name : "Atomix STAN" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 101u16 , name : "Atomix HDMI STAN" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 112u16 , name : "RED Rocket" , subsystems : & [] } , Device { vendor_id : 6741u16 , id : 144u16 , name : "CinePlay" , subsystems : & [] }] }),
        (7001, Vendor { id : 7001u16 , name : "Achronix Semiconductor Corporation" , devices : & [Device { vendor_id : 7001u16 , id : 16u16 , name : "VectorPath S7t-VG6" , subsystems : & [SubSystem { subvendor : 7001u16 , subdevice : 44048u16 , name : "SDK demonstration design" } , SubSystem { subvendor : 7001u16 , subdevice : 44050u16 , name : "Ethernet board link" } , SubSystem { subvendor : 7001u16 , subdevice : 44051u16 , name : "DDR4 PCIe Ethernet demonstration design" } , SubSystem { subvendor : 7001u16 , subdevice : 44053u16 , name : "Licensing reference design" } , SubSystem { subvendor : 7001u16 , subdevice : 44054u16 , name : "Partial reconfiguration reference design" }] } , Device { vendor_id : 7001u16 , id : 18u16 , name : "VectorPath VP815" , subsystems : & [SubSystem { subvendor : 7001u16 , subdevice : 44048u16 , name : "SDK demonstration design" } , SubSystem { subvendor : 7001u16 , subdevice : 44050u16 , name : "Ethernet board link" } , SubSystem { subvendor : 7001u16 , subdevice : 44051u16 , name : "DDR4 PCIe Ethernet demonstration design" } , SubSystem { subvendor : 7001u16 , subdevice : 44053u16 , name : "Licensing reference design" } , SubSystem { subvendor : 7001u16 , subdevice : 44054u16 , name : "Partial reconfiguration reference design" }] } , Device { vendor_id : 7001u16 , id : 19u16 , name : "VectorPath VP708" , subsystems : & [] }] }),
        (18499, Vendor { id : 18499u16 , name : "Hercules Computer Technology Inc" , devices : & [] }),
        (4871, Vendor { id : 4871u16 , name : "Measurement Computing" , devices : & [Device { vendor_id : 4871u16 , id : 1u16 , name : "PCI-DAS1602/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 11u16 , name : "PCI-DIO48H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 12u16 , name : "PCI-PDISO8" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 13u16 , name : "PCI-PDISO16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 15u16 , name : "PCI-DAS1200" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 16u16 , name : "PCI-DAS1602/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 20u16 , name : "PCI-DIO24H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 21u16 , name : "PCI-DIO24H/CTR3" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 22u16 , name : "PCI-DIO48H/CTR15" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 23u16 , name : "PCI-DIO96H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 24u16 , name : "PCI-CTR05" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 25u16 , name : "PCI-DAS1200/JR" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 26u16 , name : "PCI-DAS1001" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 27u16 , name : "PCI-DAS1002" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 28u16 , name : "PCI-DAS1602JR/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 29u16 , name : "PCI-DAS6402/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 30u16 , name : "PCI-DAS6402/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 31u16 , name : "PCI-DAS16/M1" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 32u16 , name : "PCI-DDA02/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 33u16 , name : "PCI-DDA04/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 34u16 , name : "PCI-DDA08/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 35u16 , name : "PCI-DDA02/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 36u16 , name : "PCI-DDA04/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 37u16 , name : "PCI-DDA08/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 38u16 , name : "PCI-DAC04/12-HS" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 39u16 , name : "PCI-DAC04/16-HS" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 40u16 , name : "PCI-DIO24" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 41u16 , name : "PCI-DAS08" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 44u16 , name : "PCI-INT32" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 51u16 , name : "PCI-DUAL-AC5" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 52u16 , name : "PCI-DAS-TC" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 53u16 , name : "PCI-DAS64/M1/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 54u16 , name : "PCI-DAS64/M2/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 55u16 , name : "PCI-DAS64/M3/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 75u16 , name : "PCI-MDB64" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 76u16 , name : "PCI-DAS1000" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 77u16 , name : "PCI-QUAD04" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 82u16 , name : "PCI-DAS4020/12" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 83u16 , name : "PCIM-DDA06/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 84u16 , name : "PCI-DIO96" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 85u16 , name : "CPCI-DIO24H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 86u16 , name : "PCIM-DAS1602/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 87u16 , name : "PCI-DAS3202/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 89u16 , name : "PCI-QUAD-AC5" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 90u16 , name : "CPCI-DIO96H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 91u16 , name : "CPCI-DIO48H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 93u16 , name : "PCI-DAS6023" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 94u16 , name : "PCI-DAS6025" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 95u16 , name : "PCI-DAS6030" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 96u16 , name : "PCI-DAS6031" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 97u16 , name : "PCI-DAS6032" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 98u16 , name : "PCI-DAS6033" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 99u16 , name : "PCI-DAS6034" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 100u16 , name : "PCI-DAS6035" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 101u16 , name : "PCI-DAS6040" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 102u16 , name : "PCI-DAS6052" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 103u16 , name : "PCI-DAS6070" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 104u16 , name : "PCI-DAS6071" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 110u16 , name : "PCI-CTR10" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 111u16 , name : "PCI-DAS6036" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 112u16 , name : "PCI-DAC6702" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 113u16 , name : "PCI-DAC6703" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 116u16 , name : "PCI-CTR20HD" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 119u16 , name : "PCI-DIO24/LP" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 120u16 , name : "PCI-DAS6013" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 121u16 , name : "PCI-DAS6014" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 123u16 , name : "PCIM-DAS16JR/16" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 126u16 , name : "PCI-DIO24/S" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 165u16 , name : "PCI-2511" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 166u16 , name : "PCI-2513" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 167u16 , name : "PCI-2515" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 168u16 , name : "PCI-2517" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 190u16 , name : "PCI-QUAD05" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 218u16 , name : "PCIe-DIO96H" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 219u16 , name : "PCIe-DIO24" , subsystems : & [] } , Device { vendor_id : 4871u16 , id : 277u16 , name : "PCIe-DAS1602/16" , subsystems : & [] }] }),
        (1463, Vendor { id : 1463u16 , name : "THEAD" , devices : & [] }),
        (5382, Vendor { id : 5382u16 , name : "CHAMELEON Systems Inc" , devices : & [] }),
        (4758, Vendor { id : 4758u16 , name : "Kofax Image Products" , devices : & [] }),
        (5285, Vendor { id : 5285u16 , name : "XIONICS Document Technologies Inc" , devices : & [] }),
        (5935, Vendor { id : 5935u16 , name : "Sparkle Computer Co., Ltd." , devices : & [] }),
        (5628, Vendor { id : 5628u16 , name : "Techsan Electronics Co Ltd" , devices : & [] }),
        (4173, Vendor { id : 4173u16 , name : "Sony Corporation" , devices : & [Device { vendor_id : 4173u16 , id : 32772u16 , name : "DTL-H2500 [Playstation development board]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32777u16 , name : "CXD1947Q i.LINK Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32780u16 , name : "DTL-H800 [PS1 sound development board]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32781u16 , name : "DVBK-2000(E) DV Still Image Capture Board" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32825u16 , name : "CXD3222 i.LINK Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32839u16 , name : "PS2 TOOL MRP" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32854u16 , name : "Rockwell HCF 56K modem" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 32906u16 , name : "Memory Stick Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33023u16 , name : "PS2 Performance Analyzer" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33098u16 , name : "PS2 Performance Analyzer" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33155u16 , name : "ATHENS [PS3 prototype developer interface card]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33200u16 , name : "BM-1 [PSP TOOL Board Management Device]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33219u16 , name : "VO-4 [PSP TOOL Video Output Device]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33230u16 , name : "SxS Pro memory card" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33279u16 , name : "PS3 TOOL MRP" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33280u16 , name : "PS3 TOOL RSX Tracing FPGA" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 33294u16 , name : "CXD9208GP [PS3 PS2 emulation subsystem adapter]" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 36956u16 , name : "SxS Pro memory card" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 36991u16 , name : "SxS Pro+ memory card" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37007u16 , name : "Aeolia ACPI" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37022u16 , name : "Aeolia Ethernet Controller (Marvell Yukon 2 Family)" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37023u16 , name : "Aeolia SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37024u16 , name : "Aeolia SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37025u16 , name : "Aeolia PCI Express Glue and Miscellaneous Devices" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37026u16 , name : "Aeolia DMA Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37027u16 , name : "Aeolia Memory (DDR3/SPM)" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37028u16 , name : "Aeolia USB 3.0 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37052u16 , name : "SxS Pro+ memory card" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37056u16 , name : "PCIe x8 XAVC Codec Board" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37064u16 , name : "Belize ACPI" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37065u16 , name : "Belize Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37066u16 , name : "Belize SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37067u16 , name : "Belize SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37068u16 , name : "Belize PCI Express Glue and Miscellaneous Devices" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37069u16 , name : "Belize DMA Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37070u16 , name : "Belize Memory (DDR3/SPM)" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37071u16 , name : "Belize USB 3.0 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37079u16 , name : "Baikal ACPI" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37080u16 , name : "Baikal Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37081u16 , name : "Baikal SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37082u16 , name : "Baikal SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37083u16 , name : "Baikal PCI Express Glue and Miscellaneous Devices" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37084u16 , name : "Baikal DMA Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37085u16 , name : "Baikal Memory (DDR3/SPM)" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37086u16 , name : "Baikal USB 3.0 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37099u16 , name : "CXD90062GG" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37100u16 , name : "Salina PCIe Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37124u16 , name : "Salina Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37125u16 , name : "Salina SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37126u16 , name : "Salina SATA AHCI Controller (secondary)" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37127u16 , name : "Salina PCIe Glue and Miscellaneous Devices" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37128u16 , name : "Salina USB xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 4173u16 , id : 37153u16 , name : "Nextorage NEM-PA NVMe SSD for PlayStation" , subsystems : & [] }] }),
        (5752, Vendor { id : 5752u16 , name : "NetEffect" , devices : & [Device { vendor_id : 5752u16 , id : 256u16 , name : "NE020 10Gb Accelerated Ethernet Adapter (iWARP RNIC)" , subsystems : & [] }] }),
        (4991, Vendor { id : 4991u16 , name : "Japan Satellite Systems Inc" , devices : & [] }),
        (4809, Vendor { id : 4809u16 , name : "Gigi Operations" , devices : & [] }),
        (4447, Vendor { id : 4447u16 , name : "Maxtor Corporation" , devices : & [] }),
        (5385, Vendor { id : 5385u16 , name : "FIRST INTERNATIONAL Computer Inc" , devices : & [] }),
        (89, Vendor { id : 89u16 , name : "Tiger Jet Network Inc. (Wrong ID)" , devices : & [] }),
        (4668, Vendor { id : 4668u16 , name : "Century Systems, Inc." , devices : & [] }),
        (8317, Vendor { id : 8317u16 , name : "HRDT" , devices : & [Device { vendor_id : 8317u16 , id : 1552u16 , name : "2-port SATA 6 Gb/s RAID Controller" , subsystems : & [SubSystem { subvendor : 8317u16 , subdevice : 2049u16 , name : "TrustRAID B260s 2-port SATA RAID Adapter" }] } , Device { vendor_id : 8317u16 , id : 1553u16 , name : "JZ9234 SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 8317u16 , subdevice : 2051u16 , name : "B261s SATA RAID Adapter" }] } , Device { vendor_id : 8317u16 , id : 1554u16 , name : "SATA BootRAID Controller" , subsystems : & [] }] }),
        (7085, Vendor { id : 7085u16 , name : "ReFLEX CES" , devices : & [Device { vendor_id : 7085u16 , id : 49153u16 , name : "XpressGXA10-LP1150" , subsystems : & [] } , Device { vendor_id : 7085u16 , id : 49154u16 , name : "XpressGXA10-LP1151" , subsystems : & [] }] }),
        (5438, Vendor { id : 5438u16 , name : "TECHWELL Inc" , devices : & [] }),
        (5602, Vendor { id : 5602u16 , name : "Quicknet Technologies Inc" , devices : & [Device { vendor_id : 5602u16 , id : 1280u16 , name : "PhoneJack-PCI" , subsystems : & [] }] }),
        (5094, Vendor { id : 5094u16 , name : "Argosy research Inc" , devices : & [] }),
        (6669, Vendor { id : 6669u16 , name : "SEAKR Engineering" , devices : & [] }),
        (4852, Vendor { id : 4852u16 , name : "Megatel" , devices : & [] }),
        (4477, Vendor { id : 4477u16 , name : "Becton & Dickinson" , devices : & [] }),
        (6461, Vendor { id : 6461u16 , name : "New H3C Technologies Co., Ltd." , devices : & [] }),
        (5312, Vendor { id : 5312u16 , name : "COMPAL Electronics Inc" , devices : & [Device { vendor_id : 5312u16 , id : 4609u16 , name : "X550 10Gb 2P RJ45 OCP Mezz" , subsystems : & [] }] }),
        (32904, Vendor { id : 32904u16 , name : "Beijing Wangxun Technology Co., Ltd." , devices : & [Device { vendor_id : 32904u16 , id : 256u16 , name : "WX1860AL-W Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 257u16 , name : "WX1860A2 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8228u16 , name : "LRES2024PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8229u16 , name : "LRES2025PT Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8231u16 , name : "LRES2027PF Dual-port 1Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12312u16 , name : "LRES3018PT Dual-port 1Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 32904u16 , subdevice : 513u16 , name : "Dual-Port Ethernet Network Adaptor SF200T" } , SubSystem { subvendor : 32904u16 , subdevice : 1281u16 , name : "Dual-Port Ethernet Network Adapter SF200T-C101" } , SubSystem { subvendor : 32904u16 , subdevice : 2305u16 , name : "Dual-Port Ethernet Network Adapter SF200T-B401" } , SubSystem { subvendor : 32904u16 , subdevice : 2817u16 , name : "Dual-Port Ethernet Network Adapter SF200T-B402" } , SubSystem { subvendor : 32904u16 , subdevice : 16897u16 , name : "Dual-Port Ethernet Network Adaptor SF200T (WOL)" } , SubSystem { subvendor : 32904u16 , subdevice : 33281u16 , name : "Dual-Port Ethernet Network Adaptor SF200T (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 49665u16 , name : "Dual-Port Ethernet Network Adaptor SF200T (WOL, NCSI)" }] } , Device { vendor_id : 32904u16 , id : 258u16 , name : "WX1860A2S Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 528u16 , name : "Dual-Port Ethernet Network Adaptor SF200T-S" }] } , Device { vendor_id : 32904u16 , id : 259u16 , name : "WX1860A4 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 158u16 , name : "ENPW2100-T4" } , SubSystem { subvendor : 19538u16 , subdevice : 8232u16 , name : "LRES2028PF Quad-port 1Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12313u16 , name : "LRES3019PT Quad-port 1Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 32904u16 , subdevice : 1025u16 , name : "Qual-Port Ethernet Network Adaptor SF400T" } , SubSystem { subvendor : 32904u16 , subdevice : 1088u16 , name : "Qual-Port Ethernet Network Adaptor SF400-OCP" } , SubSystem { subvendor : 32904u16 , subdevice : 2561u16 , name : "Quad-Port Ethernet Network Adapter SF400T-B401" } , SubSystem { subvendor : 32904u16 , subdevice : 3073u16 , name : "Quad-Port Ethernet Network Adapter SF400T-B402" } , SubSystem { subvendor : 32904u16 , subdevice : 17409u16 , name : "Quad-Port Ethernet Network Adapter SF400T (WOL)" } , SubSystem { subvendor : 32904u16 , subdevice : 33027u16 , name : "Quad-Port Ethernet Network Adaptor SF400T (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 33793u16 , name : "Quad-Port Ethernet Network Adapter SF400T (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 49411u16 , name : "Quad-Port Ethernet Network Adaptor SF400T (WOL, NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 50177u16 , name : "Quad-Port Ethernet Network Adapter SF400T (WOL, NCSI)" }] } , Device { vendor_id : 32904u16 , id : 260u16 , name : "WX1860A4S Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 1040u16 , name : "Qual-Port Ethernet Network Adaptor SF400T-S" }] } , Device { vendor_id : 32904u16 , id : 261u16 , name : "WX1860AL2 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 514u16 , name : "Dual-Port Ethernet Network Adaptor SF200HT" } , SubSystem { subvendor : 32904u16 , subdevice : 16898u16 , name : "Dual-Port Ethernet Network Adaptor SF200HT (WOL)" } , SubSystem { subvendor : 32904u16 , subdevice : 33282u16 , name : "Dual-Port Ethernet Network Adaptor SF200HT (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 49666u16 , name : "Dual-Port Ethernet Network Adaptor SF200HT (WOL, NCSI)" }] } , Device { vendor_id : 32904u16 , id : 262u16 , name : "WX1860AL2S Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 544u16 , name : "Dual-Port Ethernet Network Adaptor SF200HT-S" }] } , Device { vendor_id : 32904u16 , id : 263u16 , name : "WX1860AL4 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 1026u16 , name : "Qual-Port Ethernet Network Adaptor SF400HT" } , SubSystem { subvendor : 32904u16 , subdevice : 17410u16 , name : "Quad-Port Ethernet Network Adaptor SF400HT (WOL)" } , SubSystem { subvendor : 32904u16 , subdevice : 33794u16 , name : "Quad-Port Ethernet Network Adaptor SF400HT (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 50178u16 , name : "Quad-Port Ethernet Network Adaptor SF400HT (WOL, NCSI)" }] } , Device { vendor_id : 32904u16 , id : 264u16 , name : "WX1860AL4S Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 1056u16 , name : "Qual-Port Ethernet Network Adaptor SF400HT-S" }] } , Device { vendor_id : 32904u16 , id : 265u16 , name : "WX1860-LC Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 266u16 , name : "WX1860A1 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8230u16 , name : "LRES2026PF Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8244u16 , name : "LRES2034PT Single-port 1Gb Ethernet Network Adapter" }] } , Device { vendor_id : 32904u16 , id : 267u16 , name : "WX1860AL1 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8725u16 , name : "LRES2215PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32904u16 , subdevice : 258u16 , name : "Single-Port Ethernet Network Adaptor SF100HT" } , SubSystem { subvendor : 32904u16 , subdevice : 16642u16 , name : "Single-Port Ethernet Network Adaptor SF100HT (WOL)" } , SubSystem { subvendor : 32904u16 , subdevice : 33026u16 , name : "Single-Port Ethernet Network Adaptor SF100HT (NCSI)" } , SubSystem { subvendor : 32904u16 , subdevice : 49410u16 , name : "Single-Port Ethernet Network Adaptor SF100HT (WOL, NCSI)" }] } , Device { vendor_id : 32904u16 , id : 273u16 , name : "WX1860A2 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 275u16 , name : "WX1860A4 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 277u16 , name : "WX1860AL2 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 279u16 , name : "WX1860AL4 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 281u16 , name : "WX1860-LC Gigabit Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 282u16 , name : "WX1860A1 Gigabit Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 283u16 , name : "WX1860AL1 Gigabit Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 4096u16 , name : "Ethernet Controller SP1000A Virtual Function for 10GbE SFP+" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 4097u16 , name : "Ethernet Controller SP1000A for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 132u16 , name : "Ethernet Controller SP1000A for 10GbE SFP+(lldp)" } , SubSystem { subvendor : 7124u16 , subdevice : 133u16 , name : "Ethernet Controller SP1000A for 10GBASE-T" } , SubSystem { subvendor : 19538u16 , subdevice : 4098u16 , name : "LRES1002PF Dual-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4099u16 , name : "LRES1003PF Single-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12289u16 , name : "LRES3001PF Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 32773u16 , name : "LRES8005PF Dual-port 10Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 32904u16 , subdevice : 0u16 , name : "Ethernet Network Adaptor RP1000 for 10GbE SFP+" } , SubSystem { subvendor : 32904u16 , subdevice : 768u16 , name : "Ethernet Network Adaptor RP1000-A03 for 10GbE SFP+" } , SubSystem { subvendor : 32904u16 , subdevice : 1024u16 , name : "Ethernet Network Adaptor RP1000-A04 for 10GbE SFP+" }] } , Device { vendor_id : 32904u16 , id : 8192u16 , name : "Ethernet Controller WX1820AL Virtual Function for 10GbE SFP+" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 8193u16 , name : "Ethernet Controller WX1820AL for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 8192u16 , name : "Ethernet Network Adaptor RP2000 for 10GbE SFP+" } , SubSystem { subvendor : 32904u16 , subdevice : 8960u16 , name : "Ethernet Network Adaptor RP2000-A03 for 10GbE SFP+" } , SubSystem { subvendor : 32904u16 , subdevice : 9216u16 , name : "Ethernet Network Adaptor RP2000-A04 for 10GbE SFP+" }] } , Device { vendor_id : 32904u16 , id : 20516u16 , name : "Ethernet Controller WX5025 Virtual Function for 25GbE SFP28" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 20517u16 , name : "Ethernet Controller WX5025 for 25GbE SFP28" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 4096u16 , name : "Dual-Port Ethernet Network Adapter FF5025-DDATACXX" }] } , Device { vendor_id : 32904u16 , id : 20543u16 , name : "Ethernet Controller WX5040 Virtual Function for 40GbE QSFP+" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 20544u16 , name : "Ethernet Controller WX5040 for 40GbE QSFP+" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 4096u16 , name : "Dual-Port Ethernet Network Adapter FF5040-DDBTACXX" }] } , Device { vendor_id : 32904u16 , id : 20772u16 , name : "Ethernet Controller WX5025AL Virtual Function for 25GbE SFP28" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 20773u16 , name : "Ethernet Controller WX5025AL for 25GbE SFP28" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 12288u16 , name : "Dual-Port Ethernet Network Adapter FF5025-DDATAIXX" }] } , Device { vendor_id : 32904u16 , id : 20799u16 , name : "Ethernet Controller WX5040AL Virtual Function for 40GbE QSFP+" , subsystems : & [] } , Device { vendor_id : 32904u16 , id : 20800u16 , name : "Ethernet Controller WX5040AL for 40GbE QSFP+" , subsystems : & [SubSystem { subvendor : 32904u16 , subdevice : 12288u16 , name : "Dual-Port Ethernet Network Adapter FF5040-DDBTAIXX" }] }] }),
        (4100, Vendor { id : 4100u16 , name : "VLSI Technology Inc" , devices : & [Device { vendor_id : 4100u16 , id : 5u16 , name : "82C592-FC1" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 6u16 , name : "82C593-FC1" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 7u16 , name : "82C594-AFC2" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 8u16 , name : "82C596/7 [Wildcat]" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 9u16 , name : "82C597-AFC2" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 12u16 , name : "82C541 [Lynx]" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 13u16 , name : "82C543 [Lynx]" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 257u16 , name : "82C532" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 258u16 , name : "82C534 [Eagle]" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 259u16 , name : "82C538" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 260u16 , name : "82C535" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 261u16 , name : "82C147" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 512u16 , name : "82C975" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 640u16 , name : "82C925" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 772u16 , name : "QSound ThunderBird PCI Audio" , subsystems : & [SubSystem { subvendor : 4100u16 , subdevice : 772u16 , name : "QSound ThunderBird PCI Audio" } , SubSystem { subvendor : 4653u16 , subdevice : 4614u16 , name : "DSP368 Audio" } , SubSystem { subvendor : 5251u16 , subdevice : 20512u16 , name : "XWave Thunder 3D Audio" }] } , Device { vendor_id : 4100u16 , id : 773u16 , name : "QSound ThunderBird PCI Audio Gameport" , subsystems : & [SubSystem { subvendor : 4100u16 , subdevice : 773u16 , name : "QSound ThunderBird PCI Audio Gameport" } , SubSystem { subvendor : 4653u16 , subdevice : 4615u16 , name : "DSP368 Audio Gameport" } , SubSystem { subvendor : 5251u16 , subdevice : 20513u16 , name : "XWave Thunder 3D Audio Gameport" }] } , Device { vendor_id : 4100u16 , id : 774u16 , name : "QSound ThunderBird PCI Audio Support Registers" , subsystems : & [SubSystem { subvendor : 4100u16 , subdevice : 774u16 , name : "QSound ThunderBird PCI Audio Support Registers" } , SubSystem { subvendor : 4653u16 , subdevice : 4616u16 , name : "DSP368 Audio Support Registers" } , SubSystem { subvendor : 5251u16 , subdevice : 20514u16 , name : "XWave Thunder 3D Audio Support Registers" }] } , Device { vendor_id : 4100u16 , id : 775u16 , name : "SAA7785 ThunderBird PCI Audio" , subsystems : & [SubSystem { subvendor : 4100u16 , subdevice : 1795u16 , name : "Philips Rhythmic Edge PSC703" } , SubSystem { subvendor : 4100u16 , subdevice : 1797u16 , name : "Philips Seismic Edge PSC705" } , SubSystem { subvendor : 4100u16 , subdevice : 1798u16 , name : "Philips Acoustic Edge PSC706" }] } , Device { vendor_id : 4100u16 , id : 776u16 , name : "SAA7785 ThunderBird PCI Audio Gameport" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 1794u16 , name : "VAS96011 [Golden Gate II]" , subsystems : & [] } , Device { vendor_id : 4100u16 , id : 1795u16 , name : "Tollgate" , subsystems : & [] }] }),
        (5386, Vendor { id : 5386u16 , name : "FORVUS RESEARCH Inc" , devices : & [] }),
        (4861, Vendor { id : 4861u16 , name : "I2S" , devices : & [] }),
        (4106, Vendor { id : 4106u16 , name : "Phoenix Technologies" , devices : & [] }),
        (4764, Vendor { id : 4764u16 , name : "Jaycor" , devices : & [] }),
        (4914, Vendor { id : 4914u16 , name : "Micro Memory" , devices : & [Device { vendor_id : 4914u16 , id : 21525u16 , name : "MM-5415CN PCI Memory Module with Battery Backup" , subsystems : & [] } , Device { vendor_id : 4914u16 , id : 21541u16 , name : "MM-5425CN PCI 64/66 Memory Module with Battery Backup" , subsystems : & [] } , Device { vendor_id : 4914u16 , id : 24896u16 , name : "MM-6140D" , subsystems : & [] }] }),
        (8000, Vendor { id : 8000u16 , name : "Netac Technology Co.,Ltd" , devices : & [Device { vendor_id : 8000u16 , id : 1u16 , name : "PCIe 4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 4610u16 , name : "NV3000 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 5634u16 , name : "NV7000-t NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 8000u16 , name : "PCIe 4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 8803u16 , name : "PCIe 3 SM based NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 21014u16 , name : "PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 21046u16 , name : "NV7000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8000u16 , id : 22373u16 , name : "PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (4924, Vendor { id : 4924u16 , name : "Holontech Corp" , devices : & [] }),
        (5943, Vendor { id : 5943u16 , name : "Linksys" , devices : & [Device { vendor_id : 5943u16 , id : 41u16 , name : "WPG54G ver. 4 PCI Card" , subsystems : & [] } , Device { vendor_id : 5943u16 , id : 4146u16 , name : "Gigabit Network Adapter" , subsystems : & [SubSystem { subvendor : 5943u16 , subdevice : 21u16 , name : "EG1032 v2 Instant Gigabit Network Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 36u16 , name : "EG1032 v3 Instant Gigabit Network Adapter" }] } , Device { vendor_id : 5943u16 , id : 4196u16 , name : "Gigabit Network Adapter" , subsystems : & [SubSystem { subvendor : 5943u16 , subdevice : 22u16 , name : "EG1064 v2 Instant Gigabit Network Adapter" }] } , Device { vendor_id : 5943u16 , id : 43784u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 5943u16 , id : 43785u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] }] }),
        (6892, Vendor { id : 6892u16 , name : "Wolfson Microelectronics" , devices : & [] }),
        (5816, Vendor { id : 5816u16 , name : "Sonnet Technologies, Inc." , devices : & [] }),
        (6228, Vendor { id : 6228u16 , name : "LG Electronics, Inc." , devices : & [] }),
        (5435, Vendor { id : 5435u16 , name : "TERRATEC Electronic GmbH" , devices : & [] }),
        (5158, Vendor { id : 5158u16 , name : "Storage Technology Corp." , devices : & [] }),
        (5552, Vendor { id : 5552u16 , name : "Zoltrix International Ltd" , devices : & [] }),
        (7762, Vendor { id : 7762u16 , name : "Tenstorrent Inc" , devices : & [Device { vendor_id : 7762u16 , id : 16414u16 , name : "Wormhole" , subsystems : & [SubSystem { subvendor : 7762u16 , subdevice : 20u16 , name : "n300" } , SubSystem { subvendor : 7762u16 , subdevice : 24u16 , name : "n150" }] } , Device { vendor_id : 7762u16 , id : 45376u16 , name : "Blackhole" , subsystems : & [SubSystem { subvendor : 7762u16 , subdevice : 64u16 , name : "p150a" } , SubSystem { subvendor : 7762u16 , subdevice : 65u16 , name : "p150b" } , SubSystem { subvendor : 7762u16 , subdevice : 67u16 , name : "p100a" }] } , Device { vendor_id : 7762u16 , id : 64202u16 , name : "Grayskull" , subsystems : & [SubSystem { subvendor : 7762u16 , subdevice : 3u16 , name : "e150" } , SubSystem { subvendor : 7762u16 , subdevice : 7u16 , name : "e75" } , SubSystem { subvendor : 7762u16 , subdevice : 10u16 , name : "e300" }] }] }),
        (5424, Vendor { id : 5424u16 , name : "ACQIS Technology Inc" , devices : & [] }),
        (4680, Vendor { id : 4680u16 , name : "Central Data Corporation" , devices : & [] }),
        (4824, Vendor { id : 4824u16 , name : "Pericom Semiconductor" , devices : & [Device { vendor_id : 4824u16 , id : 423u16 , name : "7C21P100 2-port PCI-X to PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 771u16 , name : "PCI Express Switch 3-3" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 1288u16 , name : "PI7C9X20508GP PCI Express Switch 5Port-8Lane" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 8964u16 , name : "PI7C9X2G304 EL/SL PCIe2 3-Port/4-Lane Packet Switch" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 8968u16 , name : "PI7C9X2G308GP 3-Ports/8-lane PCIe 2.0 Switch" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 9220u16 , name : "PI7C9X2G404 EL/SL PCIe2 4-Port/4-Lane Packet Switch" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 9736u16 , name : "PI7C9X2G608GP PCIe2 6-Port/8-Lane Packet Switch" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52240u16 , name : "RXi2-BP" }] } , Device { vendor_id : 4824u16 , id : 16394u16 , name : "PI7C9X442SL PCIe Bridge Port" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 16396u16 , name : "PI7C9X440SL PCIe Bridge Port" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 16398u16 , name : "PI7C9X440SL/PI7C9X442SL USB OHCI Controller" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 16399u16 , name : "PI7C9X440SL/PI7C9X442SL USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 29154u16 , name : "PI7C7300A/PI7C7300D PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 29155u16 , name : "PI7C7300A/PI7C7300D PCI-to-PCI Bridge (Secondary Bus 2)" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 33088u16 , name : "PI7C8140A PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 33096u16 , name : "PI7C8148A/PI7C8148B PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 33104u16 , name : "PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 33106u16 , name : "PI7C8152A/PI7C8152B/PI7C8152BI PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 33108u16 , name : "PI7C8154A/PI7C8154B/PI7C8154BI PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 34329u16 , name : "PI7C9X2G1616PR PCIe2 16-Port/16-Lane Packet Switch" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 46084u16 , name : "PI7C9X2G404 EV/SV PCIe2 4-Port/4-Lane Packet Switch" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 57616u16 , name : "PI7C9X110 PCIe- to-PCI bridge" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11 CompactPCI Bridge" }] } , Device { vendor_id : 4824u16 , id : 57617u16 , name : "PI7C9X111SL PCIe-to-PCI Reversible Bridge" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 15291u16 , name : "OCuLink to RX3i backplane bridge" }] } , Device { vendor_id : 4824u16 , id : 57618u16 , name : "PI7C9X112SL PCIe-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 57619u16 , name : "PI7C9X113SL/PI7C9X118SL PCIe-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4824u16 , id : 57648u16 , name : "PCI Express to PCI-XPI7C9X130 PCI-X Bridge" , subsystems : & [] }] }),
        (4995, Vendor { id : 4995u16 , name : "Controlnet Inc" , devices : & [] }),
        (4307, Vendor { id : 4307u16 , name : "Jabil Circuit Inc" , devices : & [] }),
        (4408, Vendor { id : 4408u16 , name : "Ziatech Corporation" , devices : & [Device { vendor_id : 4408u16 , id : 35077u16 , name : "8905 [STD 32 Bridge]" , subsystems : & [] }] }),
        (5469, Vendor { id : 5469u16 , name : "Mac System Co Ltd" , devices : & [] }),
        (4849, Vendor { id : 4849u16 , name : "Sorenson Vision Inc" , devices : & [] }),
        (5195, Vendor { id : 5195u16 , name : "Verint Systems Inc." , devices : & [] }),
        (6149, Vendor { id : 6149u16 , name : "Euresys S.A." , devices : & [Device { vendor_id : 6149u16 , id : 513u16 , name : "PICOLO Alert PCI" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 514u16 , name : "PICOLO Diligent" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 516u16 , name : "PICOLO Alert-RC" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 517u16 , name : "PICOLO Alert PCIe" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 518u16 , name : "PICOLO Diligent Plus PCIe" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 519u16 , name : "PICOLO Alert-RC PCIe" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 768u16 , name : "GRABLINK Expert 2" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 769u16 , name : "GRABLINK Quickpack ColorScan" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 770u16 , name : "GRABLINK Value cPCI" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 771u16 , name : "GRABLINK Expert 2 cPCI" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 773u16 , name : "GRABLINK Avenue" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 774u16 , name : "GRABLINK Quickpack CFA" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 775u16 , name : "GRABLINK Express" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 776u16 , name : "GRABLINK Quickpack CFA PCIe" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 777u16 , name : "GRABLINK Quickpack CFA PCIe (Recovery)" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 778u16 , name : "GRABLINK Full" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 779u16 , name : "GRABLINK Full (Recovery)" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 780u16 , name : "GRABLINK DualBase" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 781u16 , name : "GRABLINK DualBase (Recovery)" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 782u16 , name : "GRABLINK Base" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 783u16 , name : "GRABLINK Base (Recovery)" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 784u16 , name : "GRABLINK Full XR" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 785u16 , name : "GRABLINK Full XR (Recovery)" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1025u16 , name : "DOMINO Iota" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1026u16 , name : "DOMINO Alpha 2" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1027u16 , name : "DOMINO Harmony" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1028u16 , name : "DOMINO Melody" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1031u16 , name : "DOMINO Symphony" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 1032u16 , name : "DOMINO Symphony PCIe" , subsystems : & [] } , Device { vendor_id : 6149u16 , id : 2068u16 , name : "Coaxlink Quad CXP-12" , subsystems : & [] }] }),
        (6519, Vendor { id : 6519u16 , name : "Parsec" , devices : & [] }),
        (7618, Vendor { id : 7618u16 , name : "Alco Digital Devices Limited" , devices : & [] }),
        (4565, Vendor { id : 4565u16 , name : "Ikon Corporation" , devices : & [Device { vendor_id : 4565u16 , id : 277u16 , name : "10115" , subsystems : & [] } , Device { vendor_id : 4565u16 , id : 279u16 , name : "10117" , subsystems : & [] }] }),
        (6743, Vendor { id : 6743u16 , name : "Highly Reliable Systems" , devices : & [] }),
        (295, Vendor { id : 295u16 , name : "MIPS" , devices : & [] }),
        (43531, Vendor { id : 43531u16 , name : "iTuner" , devices : & [] }),
        (4399, Vendor { id : 4399u16 , name : "Dalsa Inc." , devices : & [Device { vendor_id : 4399u16 , id : 0u16 , name : "MVC IC-PCI" , subsystems : & [] } , Device { vendor_id : 4399u16 , id : 1u16 , name : "MVC IM-PCI Video frame grabber/processor" , subsystems : & [] } , Device { vendor_id : 4399u16 , id : 4u16 , name : "PCDig Digital Image Capture" , subsystems : & [] } , Device { vendor_id : 4399u16 , id : 8u16 , name : "PC-CamLink PCI framegrabber" , subsystems : & [] }] }),
        (4860, Vendor { id : 4860u16 , name : "Capital Equipment Corp" , devices : & [] }),
        (4181, Vendor { id : 4181u16 , name : "Microchip Technology / SMSC" , devices : & [Device { vendor_id : 4181u16 , id : 29744u16 , name : "LAN7430" , subsystems : & [] } , Device { vendor_id : 4181u16 , id : 29745u16 , name : "LAN7431" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 416u16 , name : "RD10055 M.2 Dual 1GbE-T1 40m" }] } , Device { vendor_id : 4181u16 , id : 37168u16 , name : "SLC90E66 [Victory66] IDE" , subsystems : & [] } , Device { vendor_id : 4181u16 , id : 37984u16 , name : "SLC90E66 [Victory66] ISA" , subsystems : & [] } , Device { vendor_id : 4181u16 , id : 37986u16 , name : "SLC90E66 [Victory66] USB" , subsystems : & [] } , Device { vendor_id : 4181u16 , id : 37987u16 , name : "SLC90E66 [Victory66] ACPI" , subsystems : & [] } , Device { vendor_id : 4181u16 , id : 58400u16 , name : "LAN9420/LAN9420i" , subsystems : & [] }] }),
        (5632, Vendor { id : 5632u16 , name : "Northrop Grumman - Canada Ltd" , devices : & [] }),
        (61531, Vendor { id : 61531u16 , name : "Foxconn International, Inc. (Wrong ID)" , devices : & [] }),
        (5275, Vendor { id : 5275u16 , name : "SEIKO Instruments Inc" , devices : & [] }),
        (6110, Vendor { id : 6110u16 , name : "KWorld Computer Co. Ltd." , devices : & [] }),
        (6489, Vendor { id : 6489u16 , name : "PA Semi, Inc" , devices : & [Device { vendor_id : 6489u16 , id : 40960u16 , name : "PA6T Core" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40961u16 , name : "PWRficient Host Bridge" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40962u16 , name : "PWRficient PCI-Express Port" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40963u16 , name : "PWRficient SMBus Controller" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40964u16 , name : "PWRficient 16550 UART" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40965u16 , name : "PWRficient Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40966u16 , name : "PWRficient 10-Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40967u16 , name : "PWRficient DMA Controller" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40968u16 , name : "PWRficient LPC/Localbus Interface" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40969u16 , name : "PWRficient L2 Cache" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40970u16 , name : "PWRficient DDR2 Memory Controller" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40971u16 , name : "PWRficient SERDES" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40972u16 , name : "PWRficient System/Debug Controller" , subsystems : & [] } , Device { vendor_id : 6489u16 , id : 40973u16 , name : "PWRficient PCI-Express Internal Endpoint" , subsystems : & [] }] }),
        (8039, Vendor { id : 8039u16 , name : "Yunsilicon Technology" , devices : & [Device { vendor_id : 8039u16 , id : 4113u16 , name : "metaConnect SmartNIC Physical Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4114u16 , name : "metaConnect SmartNIC Virtual Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4177u16 , name : "metaFusion DPU Physical Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4178u16 , name : "metaFusion DPU Virtual Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4185u16 , name : "metaFusion DPU SoC Network Interface" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4369u16 , name : "YSA-100 RDMA Ethernet Network Controller" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 178u16 , name : "O1E2YYSA100" } , SubSystem { subvendor : 7124u16 , subdevice : 185u16 , name : "O2E1YYSA100" } , SubSystem { subvendor : 8185u16 , subdevice : 177u16 , name : "ENFY1102-QP2" } , SubSystem { subvendor : 8185u16 , subdevice : 184u16 , name : "ENFY1202-QP1" } , SubSystem { subvendor : 8185u16 , subdevice : 186u16 , name : "ENFY1251-SP4" } , SubSystem { subvendor : 8185u16 , subdevice : 187u16 , name : "ENFY1251-SP4 (Roce)" }] } , Device { vendor_id : 8039u16 , id : 4370u16 , name : "metaScale SmartNIC Virtual Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4433u16 , name : "metaVisor DPU Physical Function" , subsystems : & [] } , Device { vendor_id : 8039u16 , id : 4434u16 , name : "metaVisor DPU Virtual Function" , subsystems : & [] }] }),
        (5531, Vendor { id : 5531u16 , name : "Faraday Technology Corp" , devices : & [Device { vendor_id : 5531u16 , id : 17185u16 , name : "StorLink SL3516 (Gemini) Host Bridge" , subsystems : & [] }] }),
        (5837, Vendor { id : 5837u16 , name : "Advantech Co. Ltd" , devices : & [Device { vendor_id : 5837u16 , id : 257u16 , name : "DirectPCI SRAM for DPX-11x series" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 258u16 , name : "DirectPCI SRAM for DPX-S/C/E-series" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 259u16 , name : "DirectPCI ROM for DPX-11x series" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 260u16 , name : "DirectPCI ROM for DPX-S/C/E-series" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 261u16 , name : "DirectPCI I/O for DPX-114/DPX-115" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 262u16 , name : "DirectPCI I/O for DPX-116" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 263u16 , name : "DirectPCI I/O for DPX-116U" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 264u16 , name : "DirectPCI I/O for DPX-117" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 265u16 , name : "DirectPCI I/O for DPX-112" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 266u16 , name : "DirectPCI I/O for DPX-C/E-series" , subsystems : & [] } , Device { vendor_id : 5837u16 , id : 267u16 , name : "DirectPCI I/O for DPX-S series" , subsystems : & [] }] }),
        (4517, Vendor { id : 4517u16 , name : "Microunity Systems Eng. Inc" , devices : & [] }),
        (4958, Vendor { id : 4958u16 , name : "Sealevel Systems Inc" , devices : & [Device { vendor_id : 4958u16 , id : 20737u16 , name : "Route 56.PCI - Multi-Protocol Serial Interface (Zilog Z16C32)" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 28929u16 , name : "Single Port RS-232/422/485/530" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 29185u16 , name : "Dual Port RS-232/422/485 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 29186u16 , name : "Dual Port RS-232 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 29697u16 , name : "Four Port RS-232 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 29698u16 , name : "Four Port RS-422/485 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 30721u16 , name : "Eight Port RS-232 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 30724u16 , name : "Eight Port RS-232/422/485 Interface" , subsystems : & [] } , Device { vendor_id : 4958u16 , id : 32769u16 , name : "8001 Digital I/O Adapter" , subsystems : & [] }] }),
        (4320, Vendor { id : 4320u16 , name : "Integrated Micro Solutions Inc." , devices : & [Device { vendor_id : 4320u16 , id : 20518u16 , name : "IMS5026/27/28" , subsystems : & [] } , Device { vendor_id : 4320u16 , id : 20519u16 , name : "IMS5027" , subsystems : & [] } , Device { vendor_id : 4320u16 , id : 20520u16 , name : "IMS5028" , subsystems : & [] } , Device { vendor_id : 4320u16 , id : 34889u16 , name : "IMS8849" , subsystems : & [] } , Device { vendor_id : 4320u16 , id : 34899u16 , name : "IMS8853" , subsystems : & [] } , Device { vendor_id : 4320u16 , id : 37160u16 , name : "IMS9128 [Twin turbo 128]" , subsystems : & [] }] }),
        (35090, Vendor { id : 35090u16 , name : "TRX" , devices : & [] }),
        (4584, Vendor { id : 4584u16 , name : "Digital Processing Systems Inc." , devices : & [] }),
        (5444, Vendor { id : 5444u16 , name : "DCM DATA Systems" , devices : & [] }),
        (5127, Vendor { id : 5127u16 , name : "Lava Computer mfg Inc" , devices : & [Device { vendor_id : 5127u16 , id : 256u16 , name : "Lava Dual Serial" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 257u16 , name : "Lava Quatro A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 258u16 , name : "Lava Quatro B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 272u16 , name : "Lava DSerial-PCI Port A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 273u16 , name : "Lava DSerial-PCI Port B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 288u16 , name : "Quattro-PCI A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 289u16 , name : "Quattro-PCI B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 384u16 , name : "Lava Octo A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 385u16 , name : "Lava Octo B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 512u16 , name : "Lava Port Plus" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 513u16 , name : "Lava Quad A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 514u16 , name : "Lava Quad B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 544u16 , name : "Lava Quattro PCI Ports A/B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 545u16 , name : "Lava Quattro PCI Ports C/D" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 1024u16 , name : "Lava 8255-PIO-PCI" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 1280u16 , name : "Lava Single Serial" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 1312u16 , name : "Lava RS422-SS-PCI" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 1536u16 , name : "Lava Port 650" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 32768u16 , name : "Lava Parallel" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 32769u16 , name : "Dual parallel port controller A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 32770u16 , name : "Lava Dual Parallel port A" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 32771u16 , name : "Lava Dual Parallel port B" , subsystems : & [] } , Device { vendor_id : 5127u16 , id : 34816u16 , name : "BOCA Research IOPPAR" , subsystems : & [] }] }),
        (7830, Vendor { id : 7830u16 , name : "Drut Technologies Inc." , devices : & [] }),
        (4385, Vendor { id : 4385u16 , name : "Zilog" , devices : & [] }),
        (7739, Vendor { id : 7739u16 , name : "DapuStor Corporation" , devices : & [Device { vendor_id : 7739u16 , id : 1536u16 , name : "NVMe SSD Controller DP600" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 6u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 12u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5060)" } , SubSystem { subvendor : 7739u16 , subdevice : 13u16 , name : "Enterprise NVMe SSD U.2 61.44TB (J5060)" } , SubSystem { subvendor : 7739u16 , subdevice : 14u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5060D)" } , SubSystem { subvendor : 7739u16 , subdevice : 15u16 , name : "Enterprise NVMe SSD U.2 61.44TB (J5060D)" } , SubSystem { subvendor : 7739u16 , subdevice : 16u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R5102)" } , SubSystem { subvendor : 7739u16 , subdevice : 19u16 , name : "Enterprise NVMe SSD U.2 3.20TB (R5302)" } , SubSystem { subvendor : 7739u16 , subdevice : 39u16 , name : "Enterprise NVMe SSD U.2 61.44TB (J5060)" } , SubSystem { subvendor : 7739u16 , subdevice : 40u16 , name : "Enterprise NVMe SSD U.2 61.44TB (J5060D)" } , SubSystem { subvendor : 7739u16 , subdevice : 41u16 , name : "Enterprise NVMe SSD U.2 15.36TB (J5060)" } , SubSystem { subvendor : 7739u16 , subdevice : 42u16 , name : "Enterprise NVMe SSD U.2 15.36TB (J5060D)" } , SubSystem { subvendor : 7739u16 , subdevice : 48u16 , name : "Enterprise NVMe SSD U.2 3.84TB (J5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 49u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 50u16 , name : "Enterprise NVMe SSD U.2 15.36TB (J5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 51u16 , name : "Enterprise NVMe SSD U.2 3.20TB (J5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 52u16 , name : "Enterprise NVMe SSD U.2 6.40TB (J5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 53u16 , name : "Enterprise NVMe SSD U.2 12.80TB (J5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 54u16 , name : "Enterprise NVMe SSD AIC 7.68TB (J5110)" } , SubSystem { subvendor : 7739u16 , subdevice : 55u16 , name : "Enterprise NVMe SSD AIC 6.40TB (J5310)" } , SubSystem { subvendor : 7739u16 , subdevice : 56u16 , name : "Enterprise NVMe SSD U.2 3.84TB (J5100D)" } , SubSystem { subvendor : 7739u16 , subdevice : 57u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5100D)" } , SubSystem { subvendor : 7739u16 , subdevice : 59u16 , name : "Enterprise NVMe SSD U.2 3.20TB (J5300D)" } , SubSystem { subvendor : 7739u16 , subdevice : 60u16 , name : "Enterprise NVMe SSD U.2 6.40TB (J5300D)" } , SubSystem { subvendor : 7739u16 , subdevice : 62u16 , name : "Enterprise NVMe SSD AIC 3.84TB (J5110)" } , SubSystem { subvendor : 7739u16 , subdevice : 63u16 , name : "Enterprise NVMe SSD AIC 3.20TB (J5310)" } , SubSystem { subvendor : 7739u16 , subdevice : 76u16 , name : "Enterprise NVMe SSD U.2 QDP 1.92TB (J5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 77u16 , name : "Enterprise NVMe SSD U.2 QDP 1.60TB (J5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 80u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 81u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 82u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 83u16 , name : "Enterprise NVMe SSD U.2 3.20TB (R5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 84u16 , name : "Enterprise NVMe SSD U.2 6.40TB (R5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 85u16 , name : "Enterprise NVMe SSD U.2 12.80TB (R5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 86u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 89u16 , name : "Enterprise NVMe SSD U.2 3.20TB (R5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 96u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R5100D)" } , SubSystem { subvendor : 7739u16 , subdevice : 97u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R5100D)" } , SubSystem { subvendor : 7739u16 , subdevice : 99u16 , name : "Enterprise NVMe SSD U.2 3.20TB (R5300D)" } , SubSystem { subvendor : 7739u16 , subdevice : 100u16 , name : "Enterprise NVMe SSD U.2 6.40TB (R5300D)" } , SubSystem { subvendor : 7739u16 , subdevice : 102u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R5101D)" } , SubSystem { subvendor : 7739u16 , subdevice : 105u16 , name : "Enterprise NVMe SSD U.2 3.20TB (R5301D)" } , SubSystem { subvendor : 7739u16 , subdevice : 108u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 109u16 , name : "Enterprise NVMe SSD U.2 1.60TB (J5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 131u16 , name : "Enterprise NVMe SSD AIC 1.6TB (J5310)" } , SubSystem { subvendor : 7739u16 , subdevice : 150u16 , name : "Enterprise NVMe SSD AIC 1.92TB (J5110)" } , SubSystem { subvendor : 7739u16 , subdevice : 185u16 , name : "Enterprise NVMe SSD U.2 25.60TB" } , SubSystem { subvendor : 7739u16 , subdevice : 190u16 , name : "Enterprise NVMe SSD U.2 30.72TB" } , SubSystem { subvendor : 7739u16 , subdevice : 193u16 , name : "Enterprise NVMe SSD U.2 25.60TB" } , SubSystem { subvendor : 7739u16 , subdevice : 196u16 , name : "Enterprise NVMe SSD U.2 30.72TB" } , SubSystem { subvendor : 7739u16 , subdevice : 199u16 , name : "Enterprise NVMe SSD U.2 25.60TB (J5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 200u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 201u16 , name : "Enterprise NVMe SSD U.2 15.36TB (J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 202u16 , name : "Enterprise NVMe SSD U.2 3.84TB (J5102)" } , SubSystem { subvendor : 7739u16 , subdevice : 203u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5102)" } , SubSystem { subvendor : 7739u16 , subdevice : 204u16 , name : "Enterprise NVMe SSD U.2 3.84TB (J5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 205u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 206u16 , name : "Enterprise NVMe SSD U.2 3.84TB (J5101D)" } , SubSystem { subvendor : 7739u16 , subdevice : 207u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5101D)" } , SubSystem { subvendor : 7739u16 , subdevice : 220u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 221u16 , name : "Enterprise NVMe SSD U.2 30.72TB(J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 222u16 , name : "Enterprise NVMe SSD U.2 15.36TB (J5000D)" } , SubSystem { subvendor : 7739u16 , subdevice : 223u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5000)" } , SubSystem { subvendor : 7739u16 , subdevice : 231u16 , name : "Enterprise NVMe SSD U.2 30.72TB (J5000D)" } , SubSystem { subvendor : 7739u16 , subdevice : 232u16 , name : "Enterprise NVMe SSD U.2 3.20TB (J5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 233u16 , name : "Enterprise NVMe SSD U.2 6.40TB (J5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 234u16 , name : "Enterprise NVMe SSD U.2 3.20TB (J5301D)" } , SubSystem { subvendor : 7739u16 , subdevice : 235u16 , name : "Enterprise NVMe SSD U.2 6.40TB (J5301D)" } , SubSystem { subvendor : 7739u16 , subdevice : 236u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 237u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 238u16 , name : "Enterprise NVMe SSD U.2 15.36B (J5101)" } , SubSystem { subvendor : 7739u16 , subdevice : 239u16 , name : "Enterprise NVMe SSD U.2 12.80TB (J5301)" } , SubSystem { subvendor : 7739u16 , subdevice : 240u16 , name : "Enterprise NVMe SSD U.2 0.40TB (X2900)" } , SubSystem { subvendor : 7739u16 , subdevice : 241u16 , name : "Enterprise NVMe SSD U.2 0.80TB (X2900)" } , SubSystem { subvendor : 7739u16 , subdevice : 242u16 , name : "Enterprise NVMe SSD U.2 1.60TB (X2900)" } , SubSystem { subvendor : 7739u16 , subdevice : 243u16 , name : "Enterprise NVMe SSD U.2 3.20TB (X2900)" } , SubSystem { subvendor : 7739u16 , subdevice : 245u16 , name : "Enterprise NVMe SSD U.2 0.40TB (X2900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 246u16 , name : "Enterprise NVMe SSD U.2 0.80TB (X2900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 248u16 , name : "Enterprise NVMe SSD U.2 122.88TB" } , SubSystem { subvendor : 7739u16 , subdevice : 249u16 , name : "Enterprise NVMe SSD U.2 122.88TB (J5060D)" } , SubSystem { subvendor : 7739u16 , subdevice : 252u16 , name : "Enterprise NVMe SSD U.2 7.68TB (J5000D)" }] } , Device { vendor_id : 7739u16 , id : 2048u16 , name : "NVMe SSD Controller DP800" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 1u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 3u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 4u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 6u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 7u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 10u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 12u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 13u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 15u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 16u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 24u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101C)" } , SubSystem { subvendor : 7739u16 , subdevice : 25u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101C)" } , SubSystem { subvendor : 7739u16 , subdevice : 26u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6301C)" } , SubSystem { subvendor : 7739u16 , subdevice : 27u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301C)" } , SubSystem { subvendor : 7739u16 , subdevice : 29u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 30u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 31u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 36u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 39u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 40u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 42u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 43u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 44u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 45u16 , name : "Enterprise NVMe SSD U.2 25.6TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 46u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6000)" } , SubSystem { subvendor : 7739u16 , subdevice : 47u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6000)" } , SubSystem { subvendor : 7739u16 , subdevice : 52u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 53u16 , name : "Enterprise NVMe SSD U.2 61.44TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 54u16 , name : "Enterprise NVMe SSD U.2 122.88TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 59u16 , name : "Enterprise NVMe SSD U.2 1.6TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 63u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 64u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 65u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 66u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 67u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 73u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 81u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 82u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 84u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 85u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 88u16 , name : "Enterprise NVMe SSD U.2 0.8TB (X5900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 89u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 90u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 92u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 96u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6000)" } , SubSystem { subvendor : 7739u16 , subdevice : 97u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6000)" } , SubSystem { subvendor : 7739u16 , subdevice : 98u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 99u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 100u16 , name : "Enterprise NVMe SSD U.2 61.44TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 101u16 , name : "Enterprise NVMe SSD U.2 122.88TB (R6060)" } , SubSystem { subvendor : 7739u16 , subdevice : 103u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 104u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 105u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 106u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 112u16 , name : "Enterprise NVMe SSD U.2 7.68TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 113u16 , name : "Enterprise NVMe SSD U.2 15.36TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 114u16 , name : "Enterprise NVMe SSD U.2 6.4TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 115u16 , name : "Enterprise NVMe SSD U.2 12.8TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 119u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 120u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 121u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 122u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 123u16 , name : "Enterprise NVMe SSD U.2 30.72TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 124u16 , name : "Enterprise NVMe SSD U.2 25.6TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 125u16 , name : "Enterprise NVMe SSD U.2 1.6TB (X5900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 126u16 , name : "Enterprise NVMe SSD U.2 0.8TB (X5900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 127u16 , name : "Enterprise NVMe SSD U.2 1.6TB (X5900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 323u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 324u16 , name : "Enterprise NVMe SSD U.2 1.6TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 329u16 , name : "Enterprise NVMe SSD U.2 1.92TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 332u16 , name : "Enterprise NVMe SSD U.2 1.6TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 346u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 347u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 348u16 , name : "Enterprise NVMe SSD U.2 3.84TB (R6101)" } , SubSystem { subvendor : 7739u16 , subdevice : 349u16 , name : "Enterprise NVMe SSD U.2 3.2TB (R6301)" } , SubSystem { subvendor : 7739u16 , subdevice : 8192u16 , name : "Enterprise NVMe SSD E1.S 9.5 3.2TB (X5900P)" } , SubSystem { subvendor : 7739u16 , subdevice : 12289u16 , name : "Enterprise NVMe SSD E3.S 3.84TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 12290u16 , name : "Enterprise NVMe SSD E3.S 7.68TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 12291u16 , name : "Enterprise NVMe SSD E3.S 15.36TB (R6100)" } , SubSystem { subvendor : 7739u16 , subdevice : 12292u16 , name : "Enterprise NVMe SSD E3.S 3.2TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 12293u16 , name : "Enterprise NVMe SSD E3.S 6.4TB (R6300)" } , SubSystem { subvendor : 7739u16 , subdevice : 12294u16 , name : "Enterprise NVMe SSD E3.S 12.8TB (R6300)" }] } , Device { vendor_id : 7739u16 , id : 4248u16 , name : "Haishen3 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 1u16 , name : "Enterprise NVMe SSD U.2 0.8TB (H2100)" } , SubSystem { subvendor : 7739u16 , subdevice : 2u16 , name : "Enterprise NVMe SSD U.2 0.96TB (H2200)" } , SubSystem { subvendor : 7739u16 , subdevice : 4u16 , name : "Enterprise NVMe SSD U.2 1.6TB (H2100)" } , SubSystem { subvendor : 7739u16 , subdevice : 5u16 , name : "Enterprise NVMe SSD U.2 1.92TB (H2200)" } , SubSystem { subvendor : 7739u16 , subdevice : 9u16 , name : "Enterprise NVMe SSD U.2 0.8TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 10u16 , name : "Enterprise NVMe SSD U.2 0.96TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 12u16 , name : "Enterprise NVMe SSD U.2 1.6TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 13u16 , name : "Enterprise NVMe SSD U.2 1.92TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 20u16 , name : "Enterprise NVMe SSD U.2 3.2TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 21u16 , name : "Enterprise NVMe SSD U.2 3.84TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 33u16 , name : "Enterprise NVMe SSD U.2 6.4TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 34u16 , name : "Enterprise NVMe SSD U.2 7.68TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 82u16 , name : "Enterprise NVMe SSD U.2 0.8TB (H3900)" } , SubSystem { subvendor : 7739u16 , subdevice : 83u16 , name : "Enterprise NVMe SSD U.2 1.6TB (H3900)" } , SubSystem { subvendor : 7739u16 , subdevice : 89u16 , name : "Enterprise NVMe SSD U.2 0.75TB (H3900)" } , SubSystem { subvendor : 7739u16 , subdevice : 97u16 , name : "Enterprise NVMe SSD HHHL 0.8TB (H2100)" } , SubSystem { subvendor : 7739u16 , subdevice : 98u16 , name : "Enterprise NVMe SSD HHHL 0.96TB (H2200)" } , SubSystem { subvendor : 7739u16 , subdevice : 100u16 , name : "Enterprise NVMe SSD HHHL 1.6TB (H2100)" } , SubSystem { subvendor : 7739u16 , subdevice : 101u16 , name : "Enterprise NVMe SSD HHHL 1.92TB (H2200)" } , SubSystem { subvendor : 7739u16 , subdevice : 108u16 , name : "Enterprise NVMe SSD HHHL 0.8TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 109u16 , name : "Enterprise NVMe SSD HHHL 0.96TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 111u16 , name : "Enterprise NVMe SSD HHHL 1.6TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 112u16 , name : "Enterprise NVMe SSD HHHL 1.92TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 124u16 , name : "Enterprise NVMe SSD HHHL 3.2TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 125u16 , name : "Enterprise NVMe SSD HHHL 3.84TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 127u16 , name : "Enterprise NVMe SSD HHHL 6.4TB (H3100)" } , SubSystem { subvendor : 7739u16 , subdevice : 128u16 , name : "Enterprise NVMe SSD HHHL 7.68TB (H3200)" } , SubSystem { subvendor : 7739u16 , subdevice : 138u16 , name : "Enterprise NVMe SSD HHHL 0.8TB (H3900)" } , SubSystem { subvendor : 7739u16 , subdevice : 139u16 , name : "Enterprise NVMe SSD HHHL 1.6TB (H3900)" } , SubSystem { subvendor : 7739u16 , subdevice : 145u16 , name : "Enterprise NVMe SSD HHHL 0.75TB (H3900)" }] } , Device { vendor_id : 7739u16 , id : 4897u16 , name : "AdgeStor NVMe SSD" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 16384u16 , name : "AdgeStor NVMe SSD M.2 0.48TB" } , SubSystem { subvendor : 7739u16 , subdevice : 16385u16 , name : "AdgeStor NVMe SSD M.2 0.96TB" } , SubSystem { subvendor : 7739u16 , subdevice : 16386u16 , name : "AdgeStor NVMe SSD M.2 1.92TB" } , SubSystem { subvendor : 7739u16 , subdevice : 16640u16 , name : "AdgeStor NVMe SSD M.2 0.48TB" } , SubSystem { subvendor : 7739u16 , subdevice : 16641u16 , name : "AdgeStor NVMe SSD M.2 0.96TB" } , SubSystem { subvendor : 7739u16 , subdevice : 16642u16 , name : "AdgeStor NVMe SSD M.2 1.92TB" }] } , Device { vendor_id : 7739u16 , id : 4915u16 , name : "Haishen5 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 26u16 , name : "Enterprise NVMe SSD E3.S 1T 3.84TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 27u16 , name : "Enterprise NVMe SSD E3.S 1T 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 29u16 , name : "Enterprise NVMe SSD E3.S 1T 3.2TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 30u16 , name : "Enterprise NVMe SSD E3.S 1T 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 49u16 , name : "Enterprise NVMe SSD U2 3.84TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 50u16 , name : "Enterprise NVMe SSD U2 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 51u16 , name : "Enterprise NVMe SSD U2 15.36TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 52u16 , name : "Enterprise NVMe SSD U2 3.2TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 53u16 , name : "Enterprise NVMe SSD U2 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 54u16 , name : "Enterprise NVMe SSD U2 12.8TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 74u16 , name : "Enterprise NVMe SSD E3.S 1T 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 75u16 , name : "Enterprise NVMe SSD E3.S 1T 15.36TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 76u16 , name : "Enterprise NVMe SSD E3.S 1T 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 77u16 , name : "Enterprise NVMe SSD E3.S 1T 12.8TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 129u16 , name : "Enterprise NVMe SSD U.2 3.84TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 130u16 , name : "Enterprise NVMe SSD U.2 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 132u16 , name : "Enterprise NVMe SSD U.2 3.2TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 133u16 , name : "Enterprise NVMe SSD U.2 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 138u16 , name : "Enterprise NVMe SSD U.2 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 139u16 , name : "Enterprise NVMe SSD U.2 15.36TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 140u16 , name : "Enterprise NVMe SSD U.2 30.72TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 141u16 , name : "Enterprise NVMe SSD U.2 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 142u16 , name : "Enterprise NVMe SSD U.2 12.8TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 143u16 , name : "Enterprise NVMe SSD U.2 25.6TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 144u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 3.84TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 145u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 146u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 3.2TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 147u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 148u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 3.84TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 150u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 3.2TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 152u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 7.68TB (H5100)" } , SubSystem { subvendor : 7739u16 , subdevice : 153u16 , name : "Enterprise NVMe SSD E1.S 9.5mm 6.4TB (H5300)" } , SubSystem { subvendor : 7739u16 , subdevice : 240u16 , name : "Enterprise NVMe SSD U2 Mold" }] } , Device { vendor_id : 7739u16 , id : 12289u16 , name : "Ethernet Controller DN200 for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 12289u16 , name : "Ethernet Network Adapter DN200-X1V for 10GbE SFP+ 2-port" }] } , Device { vendor_id : 7739u16 , id : 12290u16 , name : "Ethernet Controller DN200 Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 12300u16 , name : "Ethernet RAID Combo Controller DN200C for 1GbE" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 12300u16 , name : "Ethernet RAID Combo Adapter DN200C-G2 for 1GbE 4-port" }] } , Device { vendor_id : 7739u16 , id : 12301u16 , name : "Ethernet RAID Combo Controller DN200C Series Virtual Function" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20481u16 , name : "RAID Controller DP808A" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20482u16 , name : "RAID Controller DP808AL" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20656u16 , name : "Flexraid6 BootRaid - Type I" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20657u16 , name : "Flexraid6 BootRaid - Type II" , subsystems : & [SubSystem { subvendor : 7739u16 , subdevice : 1u16 , name : "Flexraid6 BootRAID F6100" }] } , Device { vendor_id : 7739u16 , id : 20673u16 , name : "Combo Card of BootRaid and Ethernet - Type I" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20674u16 , name : "Combo Card of BootRaid and Ethernet - Type II" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20688u16 , name : "NVME RAID Card DP800" , subsystems : & [] } , Device { vendor_id : 7739u16 , id : 20689u16 , name : "Flexraid6 NVMeRAID" , subsystems : & [] }] }),
        (5159, Vendor { id : 5159u16 , name : "Better On-Line Solutions" , devices : & [] }),
        (7986, Vendor { id : 7986u16 , name : "Wuhan YuXin Semiconductor Co., Ltd." , devices : & [Device { vendor_id : 7986u16 , id : 60757u16 , name : "U800G NVMe SSD" , subsystems : & [] }] }),
        (4189, Vendor { id : 4189u16 , name : "Number 9 Computer Company" , devices : & [Device { vendor_id : 4189u16 , id : 8969u16 , name : "Imagine 128" , subsystems : & [] } , Device { vendor_id : 4189u16 , id : 9017u16 , name : "Imagine 128-II" , subsystems : & [SubSystem { subvendor : 4189u16 , subdevice : 0u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 1u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 2u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 3u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 4u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 5u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 6u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 7u16 , name : "Imagine 128 series 2 4Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 8u16 , name : "Imagine 128 series 2e 4Mb DRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 9u16 , name : "Imagine 128 series 2e 4Mb DRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 10u16 , name : "Imagine 128 series 2 8Mb VRAM" } , SubSystem { subvendor : 4189u16 , subdevice : 11u16 , name : "Imagine 128 series 2 8Mb H-VRAM" } , SubSystem { subvendor : 4516u16 , subdevice : 10u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 0u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 4u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 5u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 6u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 8u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 9u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 10u16 , name : "Metheus 5 Megapixel" } , SubSystem { subvendor : 5068u16 , subdevice : 12u16 , name : "Metheus 5 Megapixel" }] } , Device { vendor_id : 4189u16 , id : 18749u16 , name : "Imagine 128 T2R [Ticket to Ride]" , subsystems : & [SubSystem { subvendor : 4516u16 , subdevice : 10u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 4516u16 , subdevice : 11u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 2u16 , name : "Metheus 4 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 3u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 7u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 8u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 9u16 , name : "Metheus 5 Megapixel, Dual Head" } , SubSystem { subvendor : 5068u16 , subdevice : 10u16 , name : "Metheus 5 Megapixel, Dual Head" }] } , Device { vendor_id : 4189u16 , id : 21320u16 , name : "Revolution 4" , subsystems : & [SubSystem { subvendor : 4189u16 , subdevice : 55u16 , name : "Revolution IV-FP AGP (For SGI 1600SW)" } , SubSystem { subvendor : 4516u16 , subdevice : 40u16 , name : "PVS5600M" } , SubSystem { subvendor : 4516u16 , subdevice : 56u16 , name : "PVS5600D" }] }] }),
        (5096, Vendor { id : 5096u16 , name : "Chip Express Corporation" , devices : & [] }),
        (5656, Vendor { id : 5656u16 , name : "Stone Ridge Technology" , devices : & [Device { vendor_id : 5656u16 , id : 1u16 , name : "RDX 11" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 2u16 , name : "HFT-01" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 1024u16 , name : "FarSync T2P (2 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 1088u16 , name : "FarSync T4P (4 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 1552u16 , name : "FarSync T1U (1 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 1568u16 , name : "FarSync T2U (2 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 1600u16 , name : "FarSync T4U (4 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 5648u16 , name : "FarSync TE1 (T1,E1)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 9744u16 , name : "FarSync DSL-S1 (SHDSL)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 13888u16 , name : "FarSync T4E (4-port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 17952u16 , name : "FarSync T2Ue PCI Express (2-port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5656u16 , id : 17984u16 , name : "FarSync T4Ue PCI Express (4-port X.21/V.35/V.24)" , subsystems : & [] }] }),
        (5095, Vendor { id : 5095u16 , name : "NAC Incorporated" , devices : & [] }),
        (7691, Vendor { id : 7691u16 , name : "Shenzhen Decenta Technology Co.,LTD" , devices : & [] }),
        (4498, Vendor { id : 4498u16 , name : "Densan Company Ltd" , devices : & [] }),
        (44747, Vendor { id : 44747u16 , name : "Adrienne Electronics Corporation" , devices : & [Device { vendor_id : 44747u16 , id : 25168u16 , name : "VITC/LTC Timecode Reader card [PCI-VLTC/RDR]" , subsystems : & [] }] }),
        (6214, Vendor { id : 6214u16 , name : "Alcatel-Lucent" , devices : & [] }),
        (6586, Vendor { id : 6586u16 , name : "ZyXEL Communications Corp." , devices : & [Device { vendor_id : 6586u16 , id : 9008u16 , name : "ZyWALL Turbo Card" , subsystems : & [] }] }),
        (6137, Vendor { id : 6137u16 , name : "Gemtek Technology Co., Ltd" , devices : & [] }),
        (4547, Vendor { id : 4547u16 , name : "NEC Corporation" , devices : & [] }),
        (6685, Vendor { id : 6685u16 , name : "GFaI e.V." , devices : & [Device { vendor_id : 6685u16 , id : 6679u16 , name : "Meta Networks MTP-1G IDPS NIC" , subsystems : & [] }] }),
        (4801, Vendor { id : 4801u16 , name : "GMM Research Corp" , devices : & [] }),
        (8380, Vendor { id : 8380u16 , name : "Xinsheng Technology Co., Ltd." , devices : & [Device { vendor_id : 8380u16 , id : 4097u16 , name : "Smart Network Adapter" , subsystems : & [SubSystem { subvendor : 8380u16 , subdevice : 4226u16 , name : "Ethernet Network Adapter XS-N1082-2X for 10GbE SFP+ Dual-port" } , SubSystem { subvendor : 8380u16 , subdevice : 8323u16 , name : "Ethernet Network Adapter XS-N2083-2X for 25GbE SFP28 SFP+ Dual-port ROCEv2" } , SubSystem { subvendor : 8380u16 , subdevice : 8324u16 , name : "Ethernet Network Adapter XS-N2084-2X for 25GbE SFP28 SFP+ Dual-port ROCEv2" } , SubSystem { subvendor : 8380u16 , subdevice : 16516u16 , name : "Ethernet Network Adapter XS-N4084-2X for 100GbE QSFP28/56 Dual-port" }] } , Device { vendor_id : 8380u16 , id : 8193u16 , name : "Solid State Storage" , subsystems : & [] }] }),
        (4624, Vendor { id : 4624u16 , name : "Hyperparallel Technologies" , devices : & [] }),
        (6915, Vendor { id : 6915u16 , name : "Magnum Semiconductor, Inc," , devices : & [Device { vendor_id : 6915u16 , id : 24832u16 , name : "DXT/DXTPro Multiformat Broadcast HD/SD Encoder/Decoder/Transcoder" , subsystems : & [] } , Device { vendor_id : 6915u16 , id : 28672u16 , name : "D7 Multiformat Broadcast HD/SD Encoder/Decoder/Transcoder" , subsystems : & [] }] }),
        (5301, Vendor { id : 5301u16 , name : "Creamware GmBH" , devices : & [Device { vendor_id : 5301u16 , id : 512u16 , name : "Scope" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 768u16 , name : "Pulsar" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 1024u16 , name : "PulsarSRB" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 1536u16 , name : "Pulsar2" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 2048u16 , name : "DSP-Board" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 2304u16 , name : "DSP-Board" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 2560u16 , name : "DSP-Board" , subsystems : & [] } , Device { vendor_id : 5301u16 , id : 2816u16 , name : "DSP-Board" , subsystems : & [] }] }),
        (4338, Vendor { id : 4338u16 , name : "Achme Computer, Inc." , devices : & [] }),
        (5119, Vendor { id : 5119u16 , name : "Silicon Spice Inc" , devices : & [] }),
        (2321, Vendor { id : 2321u16 , name : "Hantick" , devices : & [] }),
        (4702, Vendor { id : 4702u16 , name : "Specialvideo Engineering SRL" , devices : & [] }),
        (8311, Vendor { id : 8311u16 , name : "OpenAI" , devices : & [] }),
        (7862, Vendor { id : 7862u16 , name : "Wuxi Stars Microsystem Technology Co., Ltd" , devices : & [] }),
        (5198, Vendor { id : 5198u16 , name : "OLITEC" , devices : & [] }),
        (6679, Vendor { id : 6679u16 , name : "Force10 Networks, Inc." , devices : & [Device { vendor_id : 6679u16 , id : 32770u16 , name : "PB-10GE-2P 10GbE Security Card" , subsystems : & [] }] }),
        (62965, Vendor { id : 62965u16 , name : "F5 Networks, Inc." , devices : & [] }),
        (6435, Vendor { id : 6435u16 , name : "Sangoma Technologies Corp." , devices : & [Device { vendor_id : 6435u16 , id : 64u16 , name : "A200/Remora FXO/FXS Analog AFT card" , subsystems : & [] } , Device { vendor_id : 6435u16 , id : 256u16 , name : "A104d QUAD T1/E1 AFT card" , subsystems : & [] } , Device { vendor_id : 6435u16 , id : 768u16 , name : "A101 single-port T1/E1" , subsystems : & [] } , Device { vendor_id : 6435u16 , id : 1024u16 , name : "A104u Quad T1/E1 AFT" , subsystems : & [] }] }),
        (43524, Vendor { id : 43524u16 , name : "iTuner" , devices : & [] }),
        (5177, Vendor { id : 5177u16 , name : "Outsource Engineering & Mfg. Inc" , devices : & [] }),
        (4406, Vendor { id : 4406u16 , name : "Momentum Data Systems" , devices : & [Device { vendor_id : 4406u16 , id : 2u16 , name : "PCI-JTAG" , subsystems : & [] }] }),
        (4445, Vendor { id : 4445u16 , name : "Xircom" , devices : & [Device { vendor_id : 4445u16 , id : 3u16 , name : "Cardbus Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 385u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 4481u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 33153u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 37249u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4445u16 , subdevice : 385u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4445u16 , subdevice : 386u16 , name : "RealPort2 CardBus Ethernet 10/100 (R2BE-100)" } , SubSystem { subvendor : 4445u16 , subdevice : 4481u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4473u16 , subdevice : 385u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 32902u16 , subdevice : 33153u16 , name : "EtherExpress PRO/100 Mobile CardBus 32 Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 37249u16 , name : "EtherExpress PRO/100 Mobile CardBus 32 Adapter" }] } , Device { vendor_id : 4445u16 , id : 5u16 , name : "Cardbus Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 386u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 4482u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4445u16 , subdevice : 386u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4445u16 , subdevice : 4482u16 , name : "Cardbus Ethernet 10/100" }] } , Device { vendor_id : 4445u16 , id : 7u16 , name : "Cardbus Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 386u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 4482u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4445u16 , subdevice : 386u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4445u16 , subdevice : 4482u16 , name : "Cardbus Ethernet 10/100" }] } , Device { vendor_id : 4445u16 , id : 11u16 , name : "Cardbus Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 387u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4445u16 , subdevice : 387u16 , name : "Cardbus Ethernet 10/100" }] } , Device { vendor_id : 4445u16 , id : 12u16 , name : "Mini-PCI V.90 56k Modem" , subsystems : & [] } , Device { vendor_id : 4445u16 , id : 15u16 , name : "Cardbus Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 387u16 , name : "10/100 EtherJet Cardbus Adapter" } , SubSystem { subvendor : 4445u16 , subdevice : 387u16 , name : "Cardbus Ethernet 10/100" }] } , Device { vendor_id : 4445u16 , id : 212u16 , name : "Mini-PCI K56Flex Modem" , subsystems : & [] } , Device { vendor_id : 4445u16 , id : 257u16 , name : "Cardbus 56k modem" , subsystems : & [SubSystem { subvendor : 4445u16 , subdevice : 4225u16 , name : "Cardbus 56k Modem" }] } , Device { vendor_id : 4445u16 , id : 259u16 , name : "Cardbus Ethernet + 56k Modem" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 37249u16 , name : "Cardbus 56k Modem" } , SubSystem { subvendor : 4373u16 , subdevice : 4481u16 , name : "Cardbus Ethernet 100 + 56k Modem" } , SubSystem { subvendor : 4445u16 , subdevice : 4481u16 , name : "CBEM56G-100 Ethernet + 56k Modem" } , SubSystem { subvendor : 32902u16 , subdevice : 37249u16 , name : "PRO/100 LAN + Modem56 CardBus" }] }] }),
        (7958, Vendor { id : 7958u16 , name : "XConn Technologies" , devices : & [Device { vendor_id : 7958u16 , id : 50432u16 , name : "XC50256 CXL2.0 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50448u16 , name : "XC51256 PCIe 5.0 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50689u16 , name : "C60064 CXL 3.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50690u16 , name : "XC60256 CXL 3.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50691u16 , name : "XC60064 CXL 3.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50692u16 , name : "XC60148 CXL 3.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50693u16 , name : "XC60260 CXL 3.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50705u16 , name : "XC61064 PCIe 6.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50706u16 , name : "XC61256 PCIe 6.1 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50707u16 , name : "XC61064 PCIe 6.2 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50708u16 , name : "XC61148 PCIe 6.2 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50709u16 , name : "XC61260 PCIe 6.2 Switch" , subsystems : & [] } , Device { vendor_id : 7958u16 , id : 50752u16 , name : "XC61064 PCIe 6.2 Switch" , subsystems : & [] }] }),
        (4099, Vendor { id : 4099u16 , name : "ULSI Systems" , devices : & [Device { vendor_id : 4099u16 , id : 513u16 , name : "US201" , subsystems : & [] }] }),
        (4617, Vendor { id : 4617u16 , name : "SCI Systems Inc" , devices : & [] }),
        (6471, Vendor { id : 6471u16 , name : "C-guys, Inc." , devices : & [Device { vendor_id : 6471u16 , id : 18243u16 , name : "CG200 Dual SD/SDIO Host controller device" , subsystems : & [] }] }),
        (7597, Vendor { id : 7597u16 , name : "Fungible" , devices : & [Device { vendor_id : 7597u16 , id : 264u16 , name : "FC50, FC100, FC200 DPU NVMeoF Adapters" , subsystems : & [] }] }),
        (39321, Vendor { id : 39321u16 , name : "MetaX Integrated Circuits (Shanghai) Co., Ltd." , devices : & [Device { vendor_id : 39321u16 , id : 2561u16 , name : "C500" , subsystems : & [] }] }),
        (4812, Vendor { id : 4812u16 , name : "Pluto Technologies International" , devices : & [] }),
        (4537, Vendor { id : 4537u16 , name : "Pathlight Technology Inc." , devices : & [Device { vendor_id : 4537u16 , id : 49389u16 , name : "SSA Controller" , subsystems : & [] }] }),
        (37907, Vendor { id : 37907u16 , name : "Softlogic Co., Ltd." , devices : & [Device { vendor_id : 37907u16 , id : 24592u16 , name : "SOLO6010 MPEG-4 Video encoder/decoder" , subsystems : & [] } , Device { vendor_id : 37907u16 , id : 24848u16 , name : "SOLO6110 H.264 Video encoder/decoder" , subsystems : & [] }] }),
        (4150, Vendor { id : 4150u16 , name : "Future Domain Corp." , devices : & [Device { vendor_id : 4150u16 , id : 0u16 , name : "TMC-18C30 [36C70]" , subsystems : & [] }] }),
        (4403, Vendor { id : 4403u16 , name : "Dialogic Corporation" , devices : & [Device { vendor_id : 4403u16 , id : 30465u16 , name : "Eiconcard C90" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 30481u16 , name : "Eiconcard C91" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 30977u16 , name : "EiconCard S90" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 30978u16 , name : "EiconCard S90" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 30993u16 , name : "EiconCard S91" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 30994u16 , name : "EiconCard S91" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31009u16 , name : "Eiconcard S92" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31041u16 , name : "EiconCard S94" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31042u16 , name : "EiconCard S94" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31043u16 , name : "EiconCard S94" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31044u16 , name : "EiconCard S94" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31045u16 , name : "Eiconcard S94" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 31048u16 , name : "Eiconcard S94 64bit/66MHz" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 38673u16 , name : "Eiconcard S91 V2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 39185u16 , name : "Eiconcard S91 V2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 39233u16 , name : "Eiconcard S94 V2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 39489u16 , name : "Eiconcard S94 PCIe" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 47393u16 , name : "EiconCard P92" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 47394u16 , name : "EiconCard P92" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 47395u16 , name : "EiconCard P92" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57345u16 , name : "Diva Pro 2.0 S/T" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57346u16 , name : "Diva 2.0 S/T PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57347u16 , name : "Diva Pro 2.0 U" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57348u16 , name : "Diva 2.0 U PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57349u16 , name : "Diva 2.01 S/T PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57350u16 , name : "Diva CT S/T PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57351u16 , name : "Diva CT U PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57352u16 , name : "Diva CT Lite S/T PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57353u16 , name : "Diva CT Lite U PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57354u16 , name : "Diva ISDN+V.90 PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57355u16 , name : "Diva ISDN PCI 2.02" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57356u16 , name : "Diva 2.02 PCI U" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57357u16 , name : "Diva Pro 3.0 PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57358u16 , name : "Diva ISDN+CT S/T PCI Rev 2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57360u16 , name : "Diva Server BRI-2M PCI" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 33u16 , name : "Fujitsu Siemens ISDN S0" }] } , Device { vendor_id : 4403u16 , id : 57361u16 , name : "Diva Server BRI S/T Rev 2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57362u16 , name : "Diva Server 4BRI-8M PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57363u16 , name : "4BRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 4864u16 , name : "Diva V-4BRI-8 PCI v2" } , SubSystem { subvendor : 4403u16 , subdevice : 57363u16 , name : "Diva 4BRI-8 PCI v2" }] } , Device { vendor_id : 4403u16 , id : 57364u16 , name : "Diva Server PRI-30M PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57365u16 , name : "Diva PRI PCI v2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57366u16 , name : "Diva Server Voice 4BRI PCI" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57367u16 , name : "Diva Server Voice 4BRI Rev 2" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 57367u16 , name : "Diva Server Voice 4BRI-8M 2.0 PCI" }] } , Device { vendor_id : 4403u16 , id : 57368u16 , name : "BRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 6144u16 , name : "Diva V-BRI-2 PCI v2" } , SubSystem { subvendor : 4403u16 , subdevice : 57368u16 , name : "Diva BRI-2 PCI v2" }] } , Device { vendor_id : 4403u16 , id : 57369u16 , name : "Diva Server Voice PRI Rev 2" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 57369u16 , name : "Diva Server Voice PRI 2.0 PCI" }] } , Device { vendor_id : 4403u16 , id : 57370u16 , name : "Diva BRI-2FX PCI v2" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57371u16 , name : "Diva Server Voice BRI-2M 2.0 PCI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 57371u16 , name : "Diva Server Voice BRI-2M 2.0 PCI" }] } , Device { vendor_id : 4403u16 , id : 57372u16 , name : "PRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 7169u16 , name : "Diva PRI/E1/T1-8 PCI v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7170u16 , name : "Diva PRI/T1-24 PCI(e) v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7171u16 , name : "Diva PRI/E1-30 PCI(e) v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7172u16 , name : "Diva PRI/E1/T1-CTI PCI(e) v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7173u16 , name : "Diva V-PRI/T1-24 PCI(e) v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7174u16 , name : "Diva V-PRI/E1-30 PCI(e) v3" } , SubSystem { subvendor : 4403u16 , subdevice : 7175u16 , name : "Diva Server PRI/E1/T1-8 Cornet NQ" } , SubSystem { subvendor : 4403u16 , subdevice : 7176u16 , name : "Diva Server PRI/T1-24 Cornet NQ" } , SubSystem { subvendor : 4403u16 , subdevice : 7177u16 , name : "Diva Server PRI/E1-30 Cornet NQ" } , SubSystem { subvendor : 4403u16 , subdevice : 7178u16 , name : "Diva Server PRI/E1/T1 Cornet NQ" } , SubSystem { subvendor : 4403u16 , subdevice : 7179u16 , name : "Diva Server V-PRI/T1-24 Cornet NQ" } , SubSystem { subvendor : 4403u16 , subdevice : 7180u16 , name : "Diva Server V-PRI/E1-30 Cornet NQ" }] } , Device { vendor_id : 4403u16 , id : 57374u16 , name : "2PRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 7681u16 , name : "Diva 2PRI/E1/T1-60 PCI v1" } , SubSystem { subvendor : 4403u16 , subdevice : 57374u16 , name : "Diva V-2PRI/E1/T1-60 PCI v1" }] } , Device { vendor_id : 4403u16 , id : 57376u16 , name : "4PRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 8193u16 , name : "Diva 4PRI/E1/T1-120 PCI v1" } , SubSystem { subvendor : 4403u16 , subdevice : 57376u16 , name : "Diva V-4PRI/E1/T1-120 PCI v1" }] } , Device { vendor_id : 4403u16 , id : 57378u16 , name : "Analog-2" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 8704u16 , name : "Diva V-Analog-2 PCI v1" } , SubSystem { subvendor : 4403u16 , subdevice : 57378u16 , name : "Diva Analog-2 PCI v1" }] } , Device { vendor_id : 4403u16 , id : 57380u16 , name : "Analog-4" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 9216u16 , name : "Diva V-Analog-4 PCI v1" } , SubSystem { subvendor : 4403u16 , subdevice : 57380u16 , name : "Diva Analog-4 PCI v1" }] } , Device { vendor_id : 4403u16 , id : 57384u16 , name : "Analog-8" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 10240u16 , name : "Diva V-Analog-8 PCI v1" } , SubSystem { subvendor : 4403u16 , subdevice : 57384u16 , name : "Diva Analog-8 PCI v1" }] } , Device { vendor_id : 4403u16 , id : 57386u16 , name : "Diva IPM-300 PCI v1" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57388u16 , name : "Diva IPM-600 PCI v1" , subsystems : & [] } , Device { vendor_id : 4403u16 , id : 57390u16 , name : "4BRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 11777u16 , name : "Diva V-4BRI-8 PCIe v2" } , SubSystem { subvendor : 4403u16 , subdevice : 57390u16 , name : "Diva 4BRI-8 PCIe v2" }] } , Device { vendor_id : 4403u16 , id : 57394u16 , name : "BRI" , subsystems : & [SubSystem { subvendor : 4403u16 , subdevice : 12801u16 , name : "Diva V-BRI-2 PCIe v2" } , SubSystem { subvendor : 4403u16 , subdevice : 57394u16 , name : "Diva BRI-2 PCIe v2" }] } , Device { vendor_id : 4403u16 , id : 57396u16 , name : "Diva BRI-CTI PCI v2" , subsystems : & [] }] }),
        (6751, Vendor { id : 6751u16 , name : "System TALKS Inc." , devices : & [] }),
        (6059, Vendor { id : 6059u16 , name : "Phillips Components" , devices : & [] }),
        (7752, Vendor { id : 7752u16 , name : "Aumovio" , devices : & [] }),
        (7452, Vendor { id : 7452u16 , name : "Barefoot Networks, Inc." , devices : & [Device { vendor_id : 7452u16 , id : 1u16 , name : "Tofino 1" , subsystems : & [] } , Device { vendor_id : 7452u16 , id : 16u16 , name : "Tofino 1" , subsystems : & [] } , Device { vendor_id : 7452u16 , id : 256u16 , name : "Tofino 2" , subsystems : & [] } , Device { vendor_id : 7452u16 , id : 272u16 , name : "Tofino 2" , subsystems : & [] }] }),
        (4664, Vendor { id : 4664u16 , name : "Adtran" , devices : & [] }),
        (4579, Vendor { id : 4579u16 , name : "Quicklogic Corporation" , devices : & [Device { vendor_id : 4579u16 , id : 1u16 , name : "COM-ON-AIR Dosch&Amand DECT" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 16u16 , name : "QL5032 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 17u16 , name : "QL5032 (PBGA256) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 18u16 , name : "QL5232 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [SubSystem { subvendor : 4579u16 , subdevice : 4612u16 , name : "Becker & Hickl SPC-130" } , SubSystem { subvendor : 4579u16 , subdevice : 4615u16 , name : "Becker & Hickl DDG-200" } , SubSystem { subvendor : 4579u16 , subdevice : 4617u16 , name : "Becker & Hickl SHM-180" } , SubSystem { subvendor : 4579u16 , subdevice : 4620u16 , name : "Becker & Hickl PMM-428" }] } , Device { vendor_id : 4579u16 , id : 19u16 , name : "QL5232 (PBGA456) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 20u16 , name : "QL5030 (TQFP144) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 21u16 , name : "QL5130 (TQFP144) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 22u16 , name : "QL5130 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [SubSystem { subvendor : 4579u16 , subdevice : 4619u16 , name : "Becker & Hickl DEL-350" }] } , Device { vendor_id : 4579u16 , id : 23u16 , name : "QL5130 (PBGA256) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 25u16 , name : "QL5332 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 26u16 , name : "QL5332 (PBGA256) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 27u16 , name : "QL5432 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [SubSystem { subvendor : 4579u16 , subdevice : 4621u16 , name : "Becker & Hickl SPC-140" } , SubSystem { subvendor : 4579u16 , subdevice : 4625u16 , name : "Becker & Hickl GVD-120" } , SubSystem { subvendor : 4579u16 , subdevice : 4626u16 , name : "Becker & Hickl DDG-210" }] } , Device { vendor_id : 4579u16 , id : 28u16 , name : "QL5432 (PBGA456) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 29u16 , name : "QL5632 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 30u16 , name : "QL5632 (PBGA280) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 31u16 , name : "QL5632 (PBGA484) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 32u16 , name : "QL5632 (PBGA516) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 33u16 , name : "QL5732 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 34u16 , name : "QL5732 (PBGA280) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 35u16 , name : "QL5732 (PBGA484) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 36u16 , name : "QL5732 (PBGA516) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 45u16 , name : "QL5022 (TQFP144) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 46u16 , name : "QL5022 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Master/Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 47u16 , name : "QL5020 (TQFP144) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 48u16 , name : "QL5020 (PQFP208) [QuickPCI] 33 MHz/32-bit PCI Target with Embedded Programmable Logic and Dual Port SRAM" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 1376u16 , name : "QL5064 Companion Design Demo Board" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 20528u16 , name : "PC Watchdog" , subsystems : & [] } , Device { vendor_id : 4579u16 , id : 33815u16 , name : "QL5064 [QuickPCI] PCI v2.2 bridge for SMT417 Dual TMS320C6416T PMC Module" , subsystems : & [] }] }),
        (4159, Vendor { id : 4159u16 , name : "Synopsys/Logic Modeling Group" , devices : & [] }),
        (5145, Vendor { id : 5145u16 , name : "Excel Switching Corp" , devices : & [] }),
        (5581, Vendor { id : 5581u16 , name : "Dreamtech Co Ltd" , devices : & [] }),
        (4437, Vendor { id : 4437u16 , name : "Pine Technology Ltd" , devices : & [] }),
        (7817, Vendor { id : 7817u16 , name : "ID Quantique SA" , devices : & [Device { vendor_id : 7817u16 , id : 2u16 , name : "Quantis-PCIe-40M" , subsystems : & [] } , Device { vendor_id : 7817u16 , id : 3u16 , name : "Quantis-PCIe-240M" , subsystems : & [] }] }),
        (4096, Vendor { id : 4096u16 , name : "Broadcom / LSI" , devices : & [Device { vendor_id : 4096u16 , id : 1u16 , name : "53c810" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C810AE PCI to SCSI I/O Processor" }] } , Device { vendor_id : 4096u16 , id : 2u16 , name : "53c820" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 3u16 , name : "53c825" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C825AE PCI to SCSI I/O Processor (Ultra Wide)" }] } , Device { vendor_id : 4096u16 , id : 4u16 , name : "53c815" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 5u16 , name : "53c810AP" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 6u16 , name : "53c860" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C860E PCI to Ultra SCSI I/O Processor" }] } , Device { vendor_id : 4096u16 , id : 10u16 , name : "53c1510" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45379u16 , name : "Integrated Dual Channel Wide Ultra2 SCSI Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C1510 PCI to Dual Channel Wide Ultra2 SCSI Controller (Nonintelligent mode)" }] } , Device { vendor_id : 4096u16 , id : 11u16 , name : "53C896/897" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 24580u16 , name : "EOB003 Series SCSI host adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C896/7 PCI to Dual Channel Ultra2 SCSI Multifunction Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "LSI22910 PCI to Dual Channel Ultra2 SCSI host adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 4128u16 , name : "LSI21002 PCI to Dual Channel Ultra2 SCSI host adapter" } , SubSystem { subvendor : 5097u16 , subdevice : 4096u16 , name : "6221L-4U (Dual U2W SCSI, dual 10/100TX, graphics)" }] } , Device { vendor_id : 4096u16 , id : 12u16 , name : "53c895" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "LSI8951U PCI to Ultra2 SCSI host adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 4128u16 , name : "LSI8952U PCI to Ultra2 SCSI host adapter" } , SubSystem { subvendor : 7649u16 , subdevice : 14598u16 , name : "DC-390U2B SCSI adapter" } , SubSystem { subvendor : 7649u16 , subdevice : 14599u16 , name : "DC-390U2W" }] } , Device { vendor_id : 4096u16 , id : 13u16 , name : "53c885" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 15u16 , name : "53c875" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 28676u16 , name : "Embedded Ultra Wide SCSI Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C876/E PCI to Dual Channel SCSI Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "LSI22801 PCI to Dual Channel Ultra SCSI host adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 4128u16 , name : "LSI22802 PCI to Dual Channel Ultra SCSI host adapter" } , SubSystem { subvendor : 4242u16 , subdevice : 34656u16 , name : "FirePort 40 Dual SCSI Controller" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer Wide Ultra SCSI" } , SubSystem { subvendor : 6005u16 , subdevice : 4305u16 , name : "V5D Single Board Computer Ultra SCSI" } , SubSystem { subvendor : 7649u16 , subdevice : 14596u16 , name : "DC390F/U Ultra Wide SCSI Adapter" } , SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" }] } , Device { vendor_id : 4096u16 , id : 16u16 , name : "53C1510" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 16448u16 , name : "Integrated Smart Array Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16456u16 , name : "RAID LC2 Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "53C1510 PCI to Dual Channel Wide Ultra2 SCSI Controller (Intelligent mode)" }] } , Device { vendor_id : 4096u16 , id : 18u16 , name : "53c895a" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C895A PCI to Ultra2 SCSI Controller" }] } , Device { vendor_id : 4096u16 , id : 19u16 , name : "53c875a" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C875A PCI to Ultra SCSI Controller" }] } , Device { vendor_id : 4096u16 , id : 20u16 , name : "MegaRAID Tri-Mode SAS3516" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37984u16 , name : "MegaRAID 9460-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 38016u16 , name : "MegaRAID 9480-8i8e" } , SubSystem { subvendor : 4096u16 , subdevice : 38017u16 , name : "MegaRAID 9480-8e" } , SubSystem { subvendor : 4136u16 , subdevice : 7994u16 , name : "PERC H745 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7995u16 , name : "PERC H745 Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8148u16 , name : "PERC H745P MX" } , SubSystem { subvendor : 4407u16 , subdevice : 526u16 , name : "UCSC-RAID-M5 12G Modular RAID Controller" } , SubSystem { subvendor : 7497u16 , subdevice : 1538u16 , name : "ThinkSystem RAID 930-16i 4GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1540u16 , name : "ThinkSystem RAID 930-8e 4GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1543u16 , name : "ThinkSystem RAID 930-16i 8GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13613u16 , name : "Integrated RAID Module RMSP3AD160F" } , SubSystem { subvendor : 32902u16 , subdevice : 37984u16 , name : "RAID Controller RSP3TD160F" } , SubSystem { subvendor : 32902u16 , subdevice : 38016u16 , name : "RAID Controller RSP3MD088F" }] } , Device { vendor_id : 4096u16 , id : 21u16 , name : "MegaRAID Tri-Mode SAS3416" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37953u16 , name : "MegaRAID 9440-16i" } , SubSystem { subvendor : 4136u16 , subdevice : 7996u16 , name : "PERC H345 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7997u16 , name : "PERC H345 Front" } , SubSystem { subvendor : 7497u16 , subdevice : 1283u16 , name : "ThinkSystem RAID 530-16i PCIe 12Gb Adapter" }] } , Device { vendor_id : 4096u16 , id : 22u16 , name : "MegaRAID Tri-Mode SAS3508" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37985u16 , name : "MegaRAID 9460-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37986u16 , name : "MegaRAID 9460-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37987u16 , name : "MegaRAID 9365-28i" } , SubSystem { subvendor : 4096u16 , subdevice : 37988u16 , name : "MegaRAID 9365-24i" } , SubSystem { subvendor : 4136u16 , subdevice : 8137u16 , name : "PERC H840 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8139u16 , name : "PERC H740P Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8141u16 , name : "PERC H740P Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8143u16 , name : "PERC H740P Mini" } , SubSystem { subvendor : 7497u16 , subdevice : 1537u16 , name : "ThinkSystem RAID 930-8i 2GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1539u16 , name : "ThinkSystem RAID 930-24i 4GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13614u16 , name : "Integrated RAID Module RMSP3CD080F" } , SubSystem { subvendor : 32902u16 , subdevice : 13615u16 , name : "Integrated RAID Module RMSP3HD080E" } , SubSystem { subvendor : 32902u16 , subdevice : 37985u16 , name : "RAID Controller RSP3DD080F" }] } , Device { vendor_id : 4096u16 , id : 23u16 , name : "MegaRAID Tri-Mode SAS3408" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37952u16 , name : "MegaRAID 9440-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37954u16 , name : "MegaRAID 9440-4i" } , SubSystem { subvendor : 7497u16 , subdevice : 1280u16 , name : "ThinkSystem RAID 530-8i PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1282u16 , name : "ThinkSystem RAID 530-8i Dense Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13608u16 , name : "Integrated RAID RMSP3LD060" } , SubSystem { subvendor : 32902u16 , subdevice : 13609u16 , name : "Integrated RAID RMSP3LD060" } , SubSystem { subvendor : 32902u16 , subdevice : 37953u16 , name : "RAID Controller RSP3WD080E" }] } , Device { vendor_id : 4096u16 , id : 27u16 , name : "MegaRAID Tri-Mode SAS3504" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 1541u16 , name : "ThinkSystem RAID 930-4i 2GB Flash Flex Adapter" }] } , Device { vendor_id : 4096u16 , id : 28u16 , name : "MegaRAID Tri-Mode SAS3404" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 1281u16 , name : "ThinkSystem RAID 530-4i Flex Adapter" }] } , Device { vendor_id : 4096u16 , id : 32u16 , name : "53c1010 Ultra3 SCSI Adapter" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C1010-33 PCI to Dual Channel Ultra160 SCSI Controller" } , SubSystem { subvendor : 4219u16 , subdevice : 4160u16 , name : "Server Onboard 53C1010-33" } , SubSystem { subvendor : 7649u16 , subdevice : 4128u16 , name : "DC-390U3W" }] } , Device { vendor_id : 4096u16 , id : 33u16 , name : "53c1010 66MHz  Ultra3 SCSI Adapter" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C1000/1000R/1010R/1010-66 PCI to Ultra160 SCSI Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "Asus TR-DLS onboard 53C1010-66" } , SubSystem { subvendor : 4156u16 , subdevice : 4864u16 , name : "Ultra160 SCSI [AB306A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4880u16 , name : "Ultra160 SCSI [A9918A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4912u16 , name : "Ultra160 SCSI [A7059A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4928u16 , name : "Ultra160 SCSI [A7060A]" } , SubSystem { subvendor : 4683u16 , subdevice : 4208u16 , name : "PMC-USCSI3" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4864u16 , name : "P017 mezzanine (32-bit PMC)" } , SubSystem { subvendor : 19539u16 , subdevice : 4880u16 , name : "P017 mezzanine (64-bit PMC)" }] } , Device { vendor_id : 4096u16 , id : 47u16 , name : "MegaRAID SAS 2208 IOV [Thunderbolt]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7993u16 , name : "SPERC8-e" } , SubSystem { subvendor : 4136u16 , subdevice : 7998u16 , name : "SPERC 8" }] } , Device { vendor_id : 4096u16 , id : 48u16 , name : "53c1030 PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 218u16 , name : "ProLiant ML 350" } , SubSystem { subvendor : 4136u16 , subdevice : 291u16 , name : "LSI Logic 1020/1030" } , SubSystem { subvendor : 4136u16 , subdevice : 330u16 , name : "LSI Logic 1020/1030" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 MPT Fusion SCSI/RAID (Perc 4)" } , SubSystem { subvendor : 4136u16 , subdevice : 387u16 , name : "LSI Logic 1020/1030" } , SubSystem { subvendor : 4136u16 , subdevice : 394u16 , name : "PERC 4/IM" } , SubSystem { subvendor : 4136u16 , subdevice : 4112u16 , name : "LSI U320 SCSI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4805u16 , name : "Ultra320 SCSI [A7173A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4899u16 , name : "Core I/O LAN/SCSI Combo [AB314A]" } , SubSystem { subvendor : 4156u16 , subdevice : 12552u16 , name : "Single Channel Ultra320 SCSI HBA G2" } , SubSystem { subvendor : 4156u16 , subdevice : 12842u16 , name : "SC11Xe Ultra320 Single Channel PCIe x4 SCSI Host Bus Adapter (412911-B21)" } , SubSystem { subvendor : 4683u16 , subdevice : 4464u16 , name : "PMC-USCSI320" } , SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "LSI Logic Parallel SCSI Controller (emulated)" } , SubSystem { subvendor : 5940u16 , subdevice : 4178u16 , name : "PRIMERGY BX/RX/TX S2 series onboard SCSI(IME)" }] } , Device { vendor_id : 4096u16 , id : 49u16 , name : "53c1030ZC PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 50u16 , name : "53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4096u16 , name : "LSI53C1020/1030 PCI-X to Ultra320 SCSI Controller" }] } , Device { vendor_id : 4096u16 , id : 51u16 , name : "1030ZC_53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 64u16 , name : "53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 51u16 , name : "MegaRAID SCSI 320-2XR" } , SubSystem { subvendor : 4096u16 , subdevice : 102u16 , name : "MegaRAID SCSI 320-2XRWS" }] } , Device { vendor_id : 4096u16 , id : 65u16 , name : "53C1035ZC PCI-X Fusion-MPT Dual Ultra320 SCSI" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 80u16 , name : "SAS1064 PCI-X Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7940u16 , name : "SAS 5/E" } , SubSystem { subvendor : 4136u16 , subdevice : 7945u16 , name : "SAS 5i/R" }] } , Device { vendor_id : 4096u16 , id : 82u16 , name : "MegaRAID SAS-3 3216/3224 [Cutlass]" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 83u16 , name : "MegaRAID SAS-3 3216/3224 [Cutlass]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37712u16 , name : "MegaRAID SAS 9341-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 37713u16 , name : "MegaRAID SAS 9341-24i" }] } , Device { vendor_id : 4096u16 , id : 84u16 , name : "SAS1068 PCI-X Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7940u16 , name : "SAS 5/E Adapter Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7941u16 , name : "SAS 5/i Adapter Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7942u16 , name : "SAS 5/i Integrated Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7943u16 , name : "SAS 5/iR Integrated RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7944u16 , name : "SAS 5/iR Integrated RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7945u16 , name : "SAS 5/iR Adapter RAID Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 12840u16 , name : "SAS3080X-HP 8-port PCI-X 133MHz Host Bus Adapter with 2xSFF-8484" } , SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "SAS Controller" }] } , Device { vendor_id : 4096u16 , id : 85u16 , name : "SAS1068 PCI-X Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 33590u16 , name : "SAS1068" }] } , Device { vendor_id : 4096u16 , id : 86u16 , name : "SAS1064ET PCI-Express Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 955u16 , name : "ServeRAID BR10il SAS/SATA Controller v2" } , SubSystem { subvendor : 32902u16 , subdevice : 13532u16 , name : "AXX4SASMOD RAID Controller" }] } , Device { vendor_id : 4096u16 , id : 87u16 , name : "M1064E MegaRAID SAS" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 13420u16 , name : "Embedded Software RAID Technology II (ESTRII)" }] } , Device { vendor_id : 4096u16 , id : 88u16 , name : "SAS1068E PCI-Express Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12608u16 , name : "SAS3081E-R 8-Port SAS/SATA Host Bus Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 541u16 , name : "SAS 6/iR Integrated Workstations RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7950u16 , name : "SAS 6/iR Adapter RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7951u16 , name : "SAS 6/iR Integrated Blades RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7952u16 , name : "SAS 6/iR Integrated RAID Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 12841u16 , name : "SC44Ge Host Bus Adapter" }] } , Device { vendor_id : 4096u16 , id : 89u16 , name : "MegaRAID SAS 8208ELP/8208ELP" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 90u16 , name : "SAS1066E PCI-Express Fusion-MPT SAS" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 91u16 , name : "MegaRAID SAS 2208 [Thunderbolt]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37477u16 , name : "MegaRAID SAS 9265-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37478u16 , name : "MegaRAID SAS 9266-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37479u16 , name : "MegaRAID SAS 9267-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37480u16 , name : "MegaRAID SAS 9265CV-8i / 9270CV-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37481u16 , name : "MegaRAID SAS 9266-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37488u16 , name : "MegaRAID SAS 9270-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37489u16 , name : "MegaRAID SAS 9271-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37490u16 , name : "MegaRAID SAS 9272-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37491u16 , name : "MegaRAID SAS 9270CV-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37492u16 , name : "MegaRAID SAS 9270-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37493u16 , name : "MegaRAID SAS 9271-8iCC" } , SubSystem { subvendor : 4096u16 , subdevice : 37494u16 , name : "MegaRAID SAS 9271-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37509u16 , name : "MegaRAID SAS 9285-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 37512u16 , name : "MegaRAID SAS 9285CV-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 37520u16 , name : "MegaRAID SAS 9286-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 37521u16 , name : "MegaRAID SAS 9286CV-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 37525u16 , name : "MegaRAID SAS 9286CV-8eCC" } , SubSystem { subvendor : 4116u16 , subdevice : 1035u16 , name : "ServeRAID M5110 SAS/SATA Controller" } , SubSystem { subvendor : 4116u16 , subdevice : 1036u16 , name : "ServeRAID M5120 SAS/SATA Controller" } , SubSystem { subvendor : 4116u16 , subdevice : 1042u16 , name : "ServeRAID M5110e SAS/SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7981u16 , name : "PERC H810 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7984u16 , name : "PERC H710 Embedded" } , SubSystem { subvendor : 4136u16 , subdevice : 7985u16 , name : "PERC H710P Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7987u16 , name : "PERC H710P Mini (for blades)" } , SubSystem { subvendor : 4136u16 , subdevice : 7988u16 , name : "PERC H710P Mini (for monolithics)" } , SubSystem { subvendor : 4136u16 , subdevice : 7989u16 , name : "PERC H710 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7991u16 , name : "PERC H710 Mini (for blades)" } , SubSystem { subvendor : 4136u16 , subdevice : 7992u16 , name : "PERC H710 Mini (for monolithics)" } , SubSystem { subvendor : 5593u16 , subdevice : 1680u16 , name : "AOC-S2208L-H8iR" } , SubSystem { subvendor : 32902u16 , subdevice : 13584u16 , name : "RMS25PB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13585u16 , name : "RMS25PB040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13586u16 , name : "RMT3PB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13587u16 , name : "Integrated RAID Module RMS25CB080" } , SubSystem { subvendor : 32902u16 , subdevice : 13588u16 , name : "RMS25CB040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13596u16 , name : "RMS25PB080N RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13597u16 , name : "RMS25CB080N RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37477u16 , name : "RS25DB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37480u16 , name : "RS25AB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37509u16 , name : "RS25NB008 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37512u16 , name : "RS25SB008 RAID Controller" }] } , Device { vendor_id : 4096u16 , id : 92u16 , name : "SAS1064A PCI-X Fusion-MPT SAS" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 93u16 , name : "MegaRAID SAS-3 3108 [Invader]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37729u16 , name : "MegaRAID SAS 9361-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37731u16 , name : "MegaRAID SAS 9361-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37732u16 , name : "MegaRAID SAS 9364-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37738u16 , name : "MegaRAID SAS 9364-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37760u16 , name : "MegaRAID SAS 9380-8e" } , SubSystem { subvendor : 4136u16 , subdevice : 8001u16 , name : "PERC H830 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8002u16 , name : "PERC H730P Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8003u16 , name : "PERC H730 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8007u16 , name : "PERC H730P Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8008u16 , name : "PERC H730P Mini (for blades)" } , SubSystem { subvendor : 4136u16 , subdevice : 8009u16 , name : "PERC H730 Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8010u16 , name : "PERC H730 Mini (for blades)" } , SubSystem { subvendor : 4136u16 , subdevice : 8013u16 , name : "PERC FD33xS" } , SubSystem { subvendor : 4136u16 , subdevice : 8015u16 , name : "PERC H730P Slim" } , SubSystem { subvendor : 4136u16 , subdevice : 8020u16 , name : "PERC FD33xD" } , SubSystem { subvendor : 4136u16 , subdevice : 8145u16 , name : "PERC H730P MX" } , SubSystem { subvendor : 5940u16 , subdevice : 4598u16 , name : "PRAID EP400i" } , SubSystem { subvendor : 6058u16 , subdevice : 4178u16 , name : "ThinkServer RAID 720i" } , SubSystem { subvendor : 6058u16 , subdevice : 4179u16 , name : "ThinkServer RAID 720ix" } , SubSystem { subvendor : 7124u16 , subdevice : 20u16 , name : "12G SAS3108 2G" } , SubSystem { subvendor : 7124u16 , subdevice : 21u16 , name : "12G SAS3108 4G" } , SubSystem { subvendor : 7497u16 , subdevice : 1536u16 , name : "ThinkSystem RAID 730-8i 1GB Cache PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1544u16 , name : "ThinkSystem RAID 730-8i 2GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1545u16 , name : "ThinkSystem RAID 730-8i 4GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13598u16 , name : "RMS3CC080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13599u16 , name : "RMS3CC040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37728u16 , name : "RS3DC080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37730u16 , name : "RS3DC040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37760u16 , name : "RS3SC008 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37761u16 , name : "RS3MC044 RAID Controller" }] } , Device { vendor_id : 4096u16 , id : 94u16 , name : "SAS1066 PCI-X Fusion-MPT SAS" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 95u16 , name : "MegaRAID SAS-3 3008 [Fury]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8004u16 , name : "PERC H330 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8011u16 , name : "PERC H330 Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8012u16 , name : "PERC H330 Mini (for blades)" } , SubSystem { subvendor : 4136u16 , subdevice : 8013u16 , name : "PERC H330 Embedded (for monolithic)" } , SubSystem { subvendor : 4180u16 , subdevice : 12394u16 , name : "SAS 3004 iMR ROMB" } , SubSystem { subvendor : 5940u16 , subdevice : 4625u16 , name : "PRAID CP400i [D3307-A12]" } , SubSystem { subvendor : 7497u16 , subdevice : 1243u16 , name : "ServeRAID M1210 SAS/SATA Controller" } , SubSystem { subvendor : 7497u16 , subdevice : 1284u16 , name : "ThinkSystem RAID 520-8i PCIe 12Gb Adapter" }] } , Device { vendor_id : 4096u16 , id : 96u16 , name : "MegaRAID SAS 1078" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4102u16 , name : "MegaRAID SAS 8888ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4106u16 , name : "MegaRAID SAS 8708ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4110u16 , name : "MegaRAID SAS 8884E" } , SubSystem { subvendor : 4096u16 , subdevice : 4111u16 , name : "MegaRAID SAS 8708E" } , SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "MegaRAID SATA 350-8ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4113u16 , name : "MegaRAID SATA 350-4ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4114u16 , name : "MegaRAID SAS 8704ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4118u16 , name : "MegaRAID SAS 8880EM2" } , SubSystem { subvendor : 4116u16 , subdevice : 867u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 4116u16 , subdevice : 868u16 , name : "SystemX MegaRAID SAS 8808E" } , SubSystem { subvendor : 4116u16 , subdevice : 869u16 , name : "SystemX MegaRAID SAS 8884E" } , SubSystem { subvendor : 4116u16 , subdevice : 889u16 , name : "SystemX MegaRAID SAS 8880EM2" } , SubSystem { subvendor : 4136u16 , subdevice : 7946u16 , name : "PERC 6/E Adapter RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7947u16 , name : "PERC 6/i Adapter RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7948u16 , name : "PERC 6/i Integrated RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7949u16 , name : "PERC 6/i Integrated RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7953u16 , name : "CERC 6/i Integrated RAID Controller" } , SubSystem { subvendor : 4147u16 , subdevice : 33626u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 4163u16 , subdevice : 33357u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 4464u16 , subdevice : 47u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 4464u16 , subdevice : 54u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 5593u16 , subdevice : 49280u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 6058u16 , subdevice : 27516u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 6305u16 , subdevice : 3u16 , name : "LSI MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 32902u16 , subdevice : 4102u16 , name : "RAID Controller SRCSAS28EP" } , SubSystem { subvendor : 32902u16 , subdevice : 4106u16 , name : "RAID Controller SRCSAS28EV" } , SubSystem { subvendor : 32902u16 , subdevice : 4112u16 , name : "RAID Controller SRCSATA28E" } , SubSystem { subvendor : 32902u16 , subdevice : 13516u16 , name : "Integrated RAID Controller SROMBSAS28E" } , SubSystem { subvendor : 32902u16 , subdevice : 13517u16 , name : "Integrated RAID Controller SROMBSAS28E" } , SubSystem { subvendor : 32902u16 , subdevice : 13573u16 , name : "Integrated RAID Controller SROMBSASMP2" }] } , Device { vendor_id : 4096u16 , id : 98u16 , name : "SAS1078 PCI-Express Fusion-MPT SAS" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 98u16 , name : "SAS1078 PCI-Express Fusion-MPT SAS" }] } , Device { vendor_id : 4096u16 , id : 100u16 , name : "SAS2116 PCI-Express Fusion-MPT SAS-2 [Meteor]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12336u16 , name : "9200-16e 6Gb/s SAS/SATA PCIe x8 External HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 12480u16 , name : "SAS 9201-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 12496u16 , name : "9201-16e 6Gb/s SAS/SATA PCIe x8 External HBA" }] } , Device { vendor_id : 4096u16 , id : 101u16 , name : "SAS2116 PCI-Express Fusion-MPT SAS-2 [Meteor]" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 110u16 , name : "SAS2308 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 112u16 , name : "SAS2004 PCI-Express Fusion-MPT SAS-2 [Spitfire]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12304u16 , name : "SAS9211-4i" } , SubSystem { subvendor : 4116u16 , subdevice : 1038u16 , name : "ServeRAID H1110" }] } , Device { vendor_id : 4096u16 , id : 113u16 , name : "MR SAS HBA 2004" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 114u16 , name : "SAS2008 PCI-Express Fusion-MPT SAS-2 [Falcon]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12320u16 , name : "9211-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12352u16 , name : "9210-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12384u16 , name : "9212-4i4e" } , SubSystem { subvendor : 4096u16 , subdevice : 12416u16 , name : "9200-8e [LSI SAS 6Gb/s SAS/SATA PCIe x8 External HBA]" } , SubSystem { subvendor : 4096u16 , subdevice : 12464u16 , name : "9200-8e [LSI SAS 6Gb/s SAS/SATA PCIe x8 External HBA]" } , SubSystem { subvendor : 4096u16 , subdevice : 12512u16 , name : "9202-16e" } , SubSystem { subvendor : 4116u16 , subdevice : 970u16 , name : "IBM 6Gb SAS HBA [9212-4i4e]" } , SubSystem { subvendor : 4136u16 , subdevice : 7964u16 , name : "6Gbps SAS HBA Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7965u16 , name : "PERC H200 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7966u16 , name : "PERC H200 Integrated" } , SubSystem { subvendor : 4136u16 , subdevice : 7967u16 , name : "PERC H200 Modular" } , SubSystem { subvendor : 4136u16 , subdevice : 7968u16 , name : "PERC H200 Embedded" } , SubSystem { subvendor : 4136u16 , subdevice : 7970u16 , name : "PERC H200 Internal Tape Adapter" } , SubSystem { subvendor : 5940u16 , subdevice : 4471u16 , name : "HBA Ctrl SAS 6G 0/1 [D2607]" } , SubSystem { subvendor : 7124u16 , subdevice : 13u16 , name : "6G SAS2008IT" } , SubSystem { subvendor : 7124u16 , subdevice : 14u16 , name : "6G SAS2008IR" } , SubSystem { subvendor : 7124u16 , subdevice : 15u16 , name : "6G SAS2008IT SA5248" } , SubSystem { subvendor : 7124u16 , subdevice : 16u16 , name : "6G SAS2008IR SA5248" } , SubSystem { subvendor : 19538u16 , subdevice : 38600u16 , name : "LRSA96C8 8-Port SATA3 (6Gb/s) Exchange Adapter (with RAID)" } , SubSystem { subvendor : 32902u16 , subdevice : 13583u16 , name : "RMS2LL040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 14080u16 , name : "SSD 910 Series" }] } , Device { vendor_id : 4096u16 , id : 115u16 , name : "MegaRAID SAS 2008 [Falcon]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37440u16 , name : "MegaRAID SAS 9240-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37441u16 , name : "MegaRAID SAS 9240-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37536u16 , name : "MegaRAID SAS 9220-8i" } , SubSystem { subvendor : 4116u16 , subdevice : 945u16 , name : "ServeRAID M1015 SAS/SATA Controller" } , SubSystem { subvendor : 4116u16 , subdevice : 1037u16 , name : "ServeRAID M1115 SAS/SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 8014u16 , name : "PERC H310 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8015u16 , name : "PERC H310 Integrated" } , SubSystem { subvendor : 4136u16 , subdevice : 8016u16 , name : "PERC H310 Mini Blades" } , SubSystem { subvendor : 4136u16 , subdevice : 8017u16 , name : "PERC H310 Mini Monolithics" } , SubSystem { subvendor : 4136u16 , subdevice : 8018u16 , name : "PERC H310 Embedded1" } , SubSystem { subvendor : 4136u16 , subdevice : 8019u16 , name : "PERC H310 Embedded2" } , SubSystem { subvendor : 4136u16 , subdevice : 8020u16 , name : "PERC H310 Reserved" } , SubSystem { subvendor : 4136u16 , subdevice : 8056u16 , name : "PERC H310" } , SubSystem { subvendor : 4180u16 , subdevice : 12341u16 , name : "LSI MegaRAID SAS 9240-8i" } , SubSystem { subvendor : 4407u16 , subdevice : 114u16 , name : "2004 iMR ROMB" } , SubSystem { subvendor : 4407u16 , subdevice : 115u16 , name : "2008 ROMB" } , SubSystem { subvendor : 4407u16 , subdevice : 176u16 , name : "UCSC RAID SAS 2008M-8i" } , SubSystem { subvendor : 4407u16 , subdevice : 177u16 , name : "UCSC RAID SAS 2008M-8i" } , SubSystem { subvendor : 4407u16 , subdevice : 194u16 , name : "UCS E-Series Double Wide" } , SubSystem { subvendor : 4407u16 , subdevice : 195u16 , name : "UCS E-Series Single Wide" } , SubSystem { subvendor : 5593u16 , subdevice : 1024u16 , name : "Supermicro SMC2008-iMR" } , SubSystem { subvendor : 5940u16 , subdevice : 4471u16 , name : "RAID Ctrl SAS 6G 0/1 (D2607)" } , SubSystem { subvendor : 6058u16 , subdevice : 4177u16 , name : "ThinkServer RAID 510i" } , SubSystem { subvendor : 32902u16 , subdevice : 13581u16 , name : "RMS2AF040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37440u16 , name : "RAID Controller RS2WC080" } , SubSystem { subvendor : 32902u16 , subdevice : 37441u16 , name : "RAID Controller RS2WC040" }] } , Device { vendor_id : 4096u16 , id : 116u16 , name : "SAS2108 PCI-Express Fusion-MPT SAS-2 [Liberator]" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 118u16 , name : "SAS2108 PCI-Express Fusion-MPT SAS-2 [Liberator]" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 119u16 , name : "SAS2108 PCI-Express Fusion-MPT SAS-2 [Liberator]" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 121u16 , name : "MegaRAID SAS 2108 [Liberator]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37457u16 , name : "MegaRAID SAS 9260-4ix" } , SubSystem { subvendor : 4096u16 , subdevice : 37462u16 , name : "MegaRAID SAS 9260-8ix" } , SubSystem { subvendor : 4096u16 , subdevice : 37472u16 , name : "MegaRAID SAS 9260-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37473u16 , name : "MegaRAID SAS 9260-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37474u16 , name : "MegaRAID SAS 9262-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37475u16 , name : "MegaRAID SAS 9261-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37476u16 , name : "MegaRAID SAS 9264-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37479u16 , name : "MegaRAID SAS 9260CV-4i" } , SubSystem { subvendor : 4096u16 , subdevice : 37480u16 , name : "MegaRAID SAS 9260CV-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 37493u16 , name : "MegaRAID SAS 9280-8ex" } , SubSystem { subvendor : 4096u16 , subdevice : 37494u16 , name : "MR9260-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 37504u16 , name : "MegaRAID SAS 9280-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 37505u16 , name : "MegaRAID SAS 9281-8E" } , SubSystem { subvendor : 4096u16 , subdevice : 37506u16 , name : "MegaRAID SAS 9280-4i4e" } , SubSystem { subvendor : 4096u16 , subdevice : 37520u16 , name : "MegaRAID SAS 9280DE-24i4e" } , SubSystem { subvendor : 4116u16 , subdevice : 946u16 , name : "ServeRAID M5015 SAS/SATA Controller" } , SubSystem { subvendor : 4116u16 , subdevice : 947u16 , name : "ServeRAID M5025 SAS/SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7957u16 , name : "PERC H800 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7958u16 , name : "PERC H700 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7959u16 , name : "PERC H700 Integrated" } , SubSystem { subvendor : 4136u16 , subdevice : 7960u16 , name : "PERC H700 Modular" } , SubSystem { subvendor : 4136u16 , subdevice : 7962u16 , name : "PERC H800 Proto Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 7963u16 , name : "PERC H700 Integrated" } , SubSystem { subvendor : 4163u16 , subdevice : 33920u16 , name : "PIKE-2108 16PD" } , SubSystem { subvendor : 5940u16 , subdevice : 4470u16 , name : "RAID Ctrl SAS 6G 5/6 512MB (D2616)" } , SubSystem { subvendor : 5940u16 , subdevice : 4471u16 , name : "RAID Ctrl SAS 6G 0/1 (D2607)" } , SubSystem { subvendor : 32902u16 , subdevice : 13579u16 , name : "RMS2MH080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 37462u16 , name : "MegaRAID SAS 9260DE-8i RS2BL080DE" } , SubSystem { subvendor : 32902u16 , subdevice : 37472u16 , name : "RAID Controller RS2BL040" } , SubSystem { subvendor : 32902u16 , subdevice : 37473u16 , name : "RAID Controller RS2BL080" } , SubSystem { subvendor : 32902u16 , subdevice : 37476u16 , name : "RAID Controller RT3WB080 Warm Beach (Caster Lite)" } , SubSystem { subvendor : 32902u16 , subdevice : 37479u16 , name : "RAID Controller RS2VB040" } , SubSystem { subvendor : 32902u16 , subdevice : 37480u16 , name : "RAID Controller RS2VB080" } , SubSystem { subvendor : 32902u16 , subdevice : 37493u16 , name : "RAID Controller RS2PI008DE" } , SubSystem { subvendor : 32902u16 , subdevice : 37494u16 , name : "RAID Controller RS2WG160" } , SubSystem { subvendor : 32902u16 , subdevice : 37504u16 , name : "RAID Controller RS2PI008" } , SubSystem { subvendor : 32902u16 , subdevice : 37506u16 , name : "RAID Controller RS2MB044" } , SubSystem { subvendor : 32902u16 , subdevice : 37520u16 , name : "RAID Controller RS2SG244" }] } , Device { vendor_id : 4096u16 , id : 124u16 , name : "MegaRAID SAS 1078DE" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 917u16 , name : "ServeRAID-AR10is SAS/SATA Controller" }] } , Device { vendor_id : 4096u16 , id : 126u16 , name : "SSS6200 PCI-Express Flash SSD" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 1284u16 , name : "Nytro NWD-BLP4-800" } , SubSystem { subvendor : 4096u16 , subdevice : 1287u16 , name : "Nytro NWD-BLP4-1600" } , SubSystem { subvendor : 4096u16 , subdevice : 1409u16 , name : "Nytro NWD-BLP4-400" } , SubSystem { subvendor : 4096u16 , subdevice : 4109u16 , name : "Nytro NWD-BFH6-1200" } , SubSystem { subvendor : 4096u16 , subdevice : 4110u16 , name : "Nytro NWD-BFH8-1600" } , SubSystem { subvendor : 4096u16 , subdevice : 4222u16 , name : "Nytro NWD-BFH8-3200" } , SubSystem { subvendor : 4096u16 , subdevice : 4880u16 , name : "Nytro XP6302-8B1536" } , SubSystem { subvendor : 4096u16 , subdevice : 4881u16 , name : "Nytro XP6302-8B2048" } , SubSystem { subvendor : 4096u16 , subdevice : 4884u16 , name : "Nytro XP6302-8B4096" } , SubSystem { subvendor : 4096u16 , subdevice : 5388u16 , name : "Nytro XP6210-4A2048" } , SubSystem { subvendor : 4096u16 , subdevice : 5391u16 , name : "Nytro XP6210-4B2048" } , SubSystem { subvendor : 4096u16 , subdevice : 5643u16 , name : "Nytro XP6209-4A1024" } , SubSystem { subvendor : 4096u16 , subdevice : 5651u16 , name : "Nytro XP6209-4B2048" } , SubSystem { subvendor : 4238u16 , subdevice : 1290u16 , name : "Nytro ELP4x200_4d_n" } , SubSystem { subvendor : 4238u16 , subdevice : 1409u16 , name : "Nytro ELP4x100_4d_n" }] } , Device { vendor_id : 4096u16 , id : 128u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 129u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 130u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 131u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 132u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 133u16 , name : "SAS2208 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 134u16 , name : "SAS2308 PCI-Express Fusion-MPT SAS-2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1680u16 , name : "Onboard MegaRAID SAS2208 [Thunderbolt]" } , SubSystem { subvendor : 5593u16 , subdevice : 1681u16 , name : "Onboard SAS2308 PCI-Express Fusion-MPT SAS-2" }] } , Device { vendor_id : 4096u16 , id : 135u16 , name : "SAS2308 PCI-Express Fusion-MPT SAS-2" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12320u16 , name : "9207-8i SAS2.1 HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 12336u16 , name : "SAS9207-4i4e" } , SubSystem { subvendor : 4096u16 , subdevice : 12352u16 , name : "9207-8e SAS2.1 HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 12368u16 , name : "SAS9217-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12384u16 , name : "SAS9217-4i4e" } , SubSystem { subvendor : 4096u16 , subdevice : 12400u16 , name : "SAS9206-16e" } , SubSystem { subvendor : 4116u16 , subdevice : 1138u16 , name : "N2125 External Host Bus Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 1146u16 , name : "N2115 Internal Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 65u16 , name : "H220i" } , SubSystem { subvendor : 5520u16 , subdevice : 66u16 , name : "H221 / 9207-8e" } , SubSystem { subvendor : 5520u16 , subdevice : 68u16 , name : "H220i" } , SubSystem { subvendor : 7124u16 , subdevice : 9u16 , name : "6G SAS2308IR" } , SubSystem { subvendor : 7124u16 , subdevice : 10u16 , name : "6G SAS2308IT" } , SubSystem { subvendor : 32902u16 , subdevice : 12288u16 , name : "RS25GB008 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 12384u16 , name : "RS25FB044 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13590u16 , name : "RMS25JB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13591u16 , name : "RMS25JB040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13592u16 , name : "RMS25KB080 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13593u16 , name : "RMS25KB040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13594u16 , name : "RMS25LB040 RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13595u16 , name : "RMS25LB080 RAID Controller" }] } , Device { vendor_id : 4096u16 , id : 143u16 , name : "53c875J" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 32768u16 , name : "FirePort 40 SCSI Controller" } , SubSystem { subvendor : 4242u16 , subdevice : 34656u16 , name : "FirePort 40 Dual SCSI Host Adapter" }] } , Device { vendor_id : 4096u16 , id : 144u16 , name : "SAS3108 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 145u16 , name : "SAS3108 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 148u16 , name : "SAS3108 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 149u16 , name : "SAS3108 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 150u16 , name : "SAS3004 PCI-Express Fusion-MPT SAS-3" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12560u16 , name : "SAS9300-4i" }] } , Device { vendor_id : 4096u16 , id : 151u16 , name : "SAS3008 PCI-Express Fusion-MPT SAS-3" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12432u16 , name : "SAS9311-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12448u16 , name : "SAS9300-8e" } , SubSystem { subvendor : 4096u16 , subdevice : 12512u16 , name : "SAS9300-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12592u16 , name : "SAS 9300-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 12656u16 , name : "SAS9302-16e" } , SubSystem { subvendor : 4116u16 , subdevice : 1111u16 , name : "9302-8i [N2215]" } , SubSystem { subvendor : 4136u16 , subdevice : 8005u16 , name : "HBA330 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8006u16 , name : "12Gbps HBA" } , SubSystem { subvendor : 4136u16 , subdevice : 8019u16 , name : "HBA330 Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8146u16 , name : "HBA330 MX" } , SubSystem { subvendor : 4136u16 , subdevice : 8147u16 , name : "HBA330 MMZ" } , SubSystem { subvendor : 5593u16 , subdevice : 2056u16 , name : "AOC-S3008L-L8e" } , SubSystem { subvendor : 7124u16 , subdevice : 8u16 , name : "12G SAS3008IMR Onboard" } , SubSystem { subvendor : 7124u16 , subdevice : 11u16 , name : "12G SAS3008IR" } , SubSystem { subvendor : 7124u16 , subdevice : 12u16 , name : "12G SAS3008IT" } , SubSystem { subvendor : 7124u16 , subdevice : 17u16 , name : "Inspur 12Gb 8i-3008 IT SAS HBA" } , SubSystem { subvendor : 7124u16 , subdevice : 18u16 , name : "12Gb SAS3008IR UDM" } , SubSystem { subvendor : 7124u16 , subdevice : 31u16 , name : "12G SAS3008IR Onboard" } , SubSystem { subvendor : 7124u16 , subdevice : 32u16 , name : "12G SAS3008IT Onboard" } , SubSystem { subvendor : 7124u16 , subdevice : 38u16 , name : "12G SAS3008IT RACK" } , SubSystem { subvendor : 7124u16 , subdevice : 39u16 , name : "12G SAS3008IMR RACK" } , SubSystem { subvendor : 7124u16 , subdevice : 40u16 , name : "12G SAS3008IR RACK" }] } , Device { vendor_id : 4096u16 , id : 165u16 , name : "Fusion-MPT 24GSAS/PCIe SAS40xx/41xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 17920u16 , name : "MegaRAID 9670W-16i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 17936u16 , name : "MegaRAID 9670-24i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 17952u16 , name : "MegaRAID 9660-16i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 17968u16 , name : "MegaRAID 9660-8i8e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 17984u16 , name : "eHBA 9600W-16i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18000u16 , name : "eHBA 9600W-16e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18016u16 , name : "eHBA 9600-24i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18032u16 , name : "eHBA 9600-16i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18048u16 , name : "eHBA 9600-16e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18064u16 , name : "MegaRAID 9620-16i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18080u16 , name : "MegaRAID 9660-24i Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18112u16 , name : "eHBA 9680W-16e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18128u16 , name : "eHBA 9600-8i8e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 18144u16 , name : "eHBA 9602W-16e Tri-Mode Storage Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8468u16 , name : "PERC H965i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8469u16 , name : "PERC H965i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8471u16 , name : "PERC H965i MX" } , SubSystem { subvendor : 4136u16 , subdevice : 8506u16 , name : "PERC H965e Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8507u16 , name : "PERC H765i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8508u16 , name : "PERC H765i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8509u16 , name : "PERC H765N Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8510u16 , name : "PERC H765i MX" } , SubSystem { subvendor : 4136u16 , subdevice : 8511u16 , name : "PERC H365i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8512u16 , name : "PERC H365i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8514u16 , name : "HBA465e Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8713u16 , name : "HBA465i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8714u16 , name : "HBA465i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8907u16 , name : "PERC H365i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8908u16 , name : "PERC H965i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8909u16 , name : "HBA465i Front" } , SubSystem { subvendor : 5593u16 , subdevice : 7427u16 , name : "AOC-S4116L-H16IR (16DD/96DD) RAID Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7431u16 , name : "AOC-S4016L-L16IT Storage Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7432u16 , name : "AOC-S4016L-L16IR Storage Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 30805u16 , name : "ThinkSystem RAID 950W-16i 8GB Flash PCIe Gen4 24Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 522u16 , name : "ThinkSystem 450W-16e SAS/SATA PCIe Gen4 24Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 524u16 , name : "ThinkSystem 450W-16e PCIe Gen4 HBA" }] } , Device { vendor_id : 4096u16 , id : 171u16 , name : "SAS3516 Fusion-MPT Tri-Mode RAID On Chip (ROC)" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12352u16 , name : "HBA 9400-8i8e" } , SubSystem { subvendor : 32902u16 , subdevice : 13616u16 , name : "Integrated RAID Module RMSP3JD160J" }] } , Device { vendor_id : 4096u16 , id : 172u16 , name : "SAS3416 Fusion-MPT Tri-Mode I/O Controller Chip (IOC)" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12288u16 , name : "HBA 9400-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 12320u16 , name : "HBA 9400-16e" } , SubSystem { subvendor : 4136u16 , subdevice : 8163u16 , name : "HBA345 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8164u16 , name : "HBA345 Front" } , SubSystem { subvendor : 7497u16 , subdevice : 513u16 , name : "ThinkSystem 430-16i SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 515u16 , name : "ThinkSystem 430-16e SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 32902u16 , subdevice : 12288u16 , name : "RAID Controller RSP3QD160J" } , SubSystem { subvendor : 32902u16 , subdevice : 12320u16 , name : "RAID Controller RSP3GD016J" }] } , Device { vendor_id : 4096u16 , id : 174u16 , name : "SAS3508 Fusion-MPT Tri-Mode RAID On Chip (ROC)" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 175u16 , name : "SAS3408 Fusion-MPT Tri-Mode I/O Controller Chip (IOC)" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12304u16 , name : "HBA 9400-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 12336u16 , name : "HBA 9400-8e" } , SubSystem { subvendor : 7497u16 , subdevice : 512u16 , name : "ThinkSystem 430-8i SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 514u16 , name : "ThinkSystem 430-8e SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 516u16 , name : "ThinkSystem 430-8i SAS/SATA 12Gb Dense HBA" }] } , Device { vendor_id : 4096u16 , id : 178u16 , name : "PCIe Switch management endpoint" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 7460u16 , name : "AOM-PCIE5-418P" } , SubSystem { subvendor : 7497u16 , subdevice : 3u16 , name : "ThinkSystem 1611-8P PCIe Gen4 NVMe Switch Adapter" }] } , Device { vendor_id : 4096u16 , id : 179u16 , name : "Fusion-MPT 24G SAS/PCIe SAS50xx/SAS51xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20480u16 , name : "MegaRAID 9760W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20481u16 , name : "MegaRAID 9760W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20496u16 , name : "MegaRAID 9760W-16i16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20512u16 , name : "eHBA 9700W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20513u16 , name : "eHBA 9700W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20528u16 , name : "eHBA 9700-16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8914u16 , name : "PERC H975i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8915u16 , name : "PERC H975i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 9163u16 , name : "PERC H975i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 9165u16 , name : "PERC H975i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 9286u16 , name : "PERC H976i Front" } , SubSystem { subvendor : 6058u16 , subdevice : 30796u16 , name : "ThinkSystem RAID 960W PCIe Gen5 Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 523u16 , name : "ThinkSystem 460-16e SAS/SATA PCIe Gen5 24Gb HBA" }] } , Device { vendor_id : 4096u16 , id : 180u16 , name : "Fusion-MPT 24G SAS/PCIe SAS50xx/SAS51xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 181u16 , name : "Fusion-MPT 24G SAS/PCIe SAS50xx/SAS51xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20480u16 , name : "MegaRAID 9760W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20481u16 , name : "MegaRAID 9760W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20496u16 , name : "MegaRAID 9760W-16i16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20512u16 , name : "eHBA 9700W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20513u16 , name : "eHBA 9700W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20528u16 , name : "eHBA 9700-16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8914u16 , name : "PERC H975i Front - Virtual" } , SubSystem { subvendor : 4136u16 , subdevice : 8915u16 , name : "PERC H975i Adapter - Virtual" } , SubSystem { subvendor : 4136u16 , subdevice : 9163u16 , name : "PERC H975i Front - Virtual" } , SubSystem { subvendor : 4136u16 , subdevice : 9165u16 , name : "PERC H975i Adapter - Virtual" } , SubSystem { subvendor : 6058u16 , subdevice : 30796u16 , name : "ThinkSystem RAID 960W PCIe Gen5 Adapter" }] } , Device { vendor_id : 4096u16 , id : 182u16 , name : "Fusion-MPT 24G SAS/PCIe SAS50xx/SAS51xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20480u16 , name : "MegaRAID 9760W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20481u16 , name : "MegaRAID 9760W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20496u16 , name : "MegaRAID 9760W-16i16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20512u16 , name : "eHBA 9700W-32i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20513u16 , name : "eHBA 9700W-16i 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20528u16 , name : "eHBA 9700-16e 24G SAS/PCIe Storage Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 9286u16 , name : "PERC H976i Front - Virtual" }] } , Device { vendor_id : 4096u16 , id : 184u16 , name : "Fusion-MPT Switch SAS50xx/SAS51xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 190u16 , name : "SAS3504 Fusion-MPT Tri-Mode RAID On Chip (ROC)" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 191u16 , name : "SAS3404 Fusion-MPT Tri-Mode I/O Controller Chip (IOC)" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 192u16 , name : "SAS3324 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 193u16 , name : "SAS3324 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 194u16 , name : "SAS3324 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 195u16 , name : "SAS3324 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 196u16 , name : "SAS3224 PCI-Express Fusion-MPT SAS-3" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12688u16 , name : "SAS9305-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 12704u16 , name : "SAS9305-24i" } , SubSystem { subvendor : 4096u16 , subdevice : 12720u16 , name : "SAS9306-24i" } , SubSystem { subvendor : 4464u16 , subdevice : 2u16 , name : "SAS3224 PCI Express to 12Gb HBA MEZZ CARD" }] } , Device { vendor_id : 4096u16 , id : 197u16 , name : "SAS3316 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 198u16 , name : "SAS3316 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 199u16 , name : "SAS3316 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 200u16 , name : "SAS3316 PCI-Express Fusion-MPT SAS-3" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 201u16 , name : "SAS3216 PCI-Express Fusion-MPT SAS-3" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12672u16 , name : "SAS9305-16e" }] } , Device { vendor_id : 4096u16 , id : 206u16 , name : "MegaRAID SAS-3 3316 [Intruder]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37745u16 , name : "MegaRAID SAS 9361-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 37776u16 , name : "MegaRAID SAS 9380-8i8e" }] } , Device { vendor_id : 4096u16 , id : 207u16 , name : "MegaRAID SAS-3 3324 [Intruder]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 37744u16 , name : "MegaRAID SAS 9361-24i" }] } , Device { vendor_id : 4096u16 , id : 208u16 , name : "SAS3716 Fusion-MPT Tri-Mode RAID Controller Chip (ROC)" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12368u16 , name : "HBA 9405W-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 12400u16 , name : "HBA 9405W-8i8e" }] } , Device { vendor_id : 4096u16 , id : 209u16 , name : "SAS3616 Fusion-MPT Tri-Mode I/O Controller Chip (IOC)" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12416u16 , name : "HBA 9405W-16e" } , SubSystem { subvendor : 4096u16 , subdevice : 12432u16 , name : "HBA 9405W-16i" }] } , Device { vendor_id : 4096u16 , id : 211u16 , name : "MegaRAID Tri-Mode SAS3716W" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 224u16 , name : "Fusion-MPT 12GSAS/PCIe Unsupported SAS39xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 225u16 , name : "Fusion-MPT 12GSAS/PCIe SAS39xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 226u16 , name : "Fusion-MPT 12GSAS/PCIe Secure SAS39xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 227u16 , name : "Fusion-MPT 12GSAS/PCIe Unsupported SAS39xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 228u16 , name : "Fusion-MPT 12GSAS/PCIe Unsupported SAS38xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8203u16 , name : "HBA355i Adapter Invalid" } , SubSystem { subvendor : 4136u16 , subdevice : 8204u16 , name : "HBA355i Front Invalid" } , SubSystem { subvendor : 4136u16 , subdevice : 8205u16 , name : "HBA355e Adapter Invalid" } , SubSystem { subvendor : 4136u16 , subdevice : 8206u16 , name : "HBA350i MX Invalid" }] } , Device { vendor_id : 4096u16 , id : 229u16 , name : "Fusion-MPT 12GSAS/PCIe SAS38xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8203u16 , name : "HBA355i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8204u16 , name : "HBA355i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8205u16 , name : "HBA355e Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8206u16 , name : "HBA350i MX" } , SubSystem { subvendor : 7497u16 , subdevice : 517u16 , name : "ThinkSystem 440-16i SAS/SATA PCIe Gen4 12Gb Internal HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 518u16 , name : "ThinkSystem 440-16e SAS/SATA PCIe Gen4 12Gb HBA" }] } , Device { vendor_id : 4096u16 , id : 230u16 , name : "Fusion-MPT 12GSAS/PCIe Secure SAS38xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 16464u16 , name : "9500-16i Tri-Mode HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 16480u16 , name : "9500-8i Tri-Mode HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 16496u16 , name : "9500-16e Tri-Mode HBA" } , SubSystem { subvendor : 4096u16 , subdevice : 16512u16 , name : "9500-8e Tri-Mode HBA" } , SubSystem { subvendor : 4136u16 , subdevice : 8203u16 , name : "HBA355i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8204u16 , name : "HBA355i Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8205u16 , name : "HBA355e Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8206u16 , name : "HBA350i MX" } , SubSystem { subvendor : 4136u16 , subdevice : 8560u16 , name : "HBA350i MM" } , SubSystem { subvendor : 4136u16 , subdevice : 8565u16 , name : "HBA350i Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8599u16 , name : "HBA350i MM LP" } , SubSystem { subvendor : 4136u16 , subdevice : 8722u16 , name : "HBA355e-s Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7012u16 , name : "SCC-B8SB80-B1 (NI22) Storage Controller Card" } , SubSystem { subvendor : 5593u16 , subdevice : 7013u16 , name : "AOC-S3816L-L16iT (NI22) Storage Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7065u16 , name : "AOC-S3808L-L8iT Storage Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 517u16 , name : "ThinkSystem 440-16i SAS/SATA PCIe Gen4 12Gb Internal HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 518u16 , name : "ThinkSystem 440-16e SAS/SATA PCIe Gen4 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 519u16 , name : "ThinkSystem 440-8i SAS/SATA PCIe Gen4 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 520u16 , name : "ThinkSystem 440-16i SAS/SATA PCIe Gen4 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 521u16 , name : "ThinkSystem 440-8e SAS/SATA PCIe Gen4 12Gb HBA" } , SubSystem { subvendor : 32902u16 , subdevice : 16464u16 , name : "Storage Controller RS3P4QF160J" } , SubSystem { subvendor : 32902u16 , subdevice : 16496u16 , name : "Storage Controller RS3P4GF016J" }] } , Device { vendor_id : 4096u16 , id : 231u16 , name : "Fusion-MPT 12GSAS/PCIe Unsupported SAS38xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8203u16 , name : "HBA355i Adapter Tampered" } , SubSystem { subvendor : 4136u16 , subdevice : 8204u16 , name : "HBA355i Front Tampered" } , SubSystem { subvendor : 4136u16 , subdevice : 8205u16 , name : "HBA355e Adapter Tampered" } , SubSystem { subvendor : 4136u16 , subdevice : 8206u16 , name : "HBA350i MX Tampered" }] } , Device { vendor_id : 4096u16 , id : 240u16 , name : "Fusion-MPT SAS/SATA SAS52xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20736u16 , name : "eHBA 9705-32i SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20737u16 , name : "eHBA 9705-16i SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20738u16 , name : "eHBA 9705-16e SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20739u16 , name : "eHBA 9705-48i SAS/SATA Storage Adapter" }] } , Device { vendor_id : 4096u16 , id : 241u16 , name : "Fusion-MPT SAS/SATA SAS52xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20736u16 , name : "eHBA 9705-32i SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20737u16 , name : "eHBA 9705-16i SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20738u16 , name : "eHBA 9705-16e SAS/SATA Storage Adapter" } , SubSystem { subvendor : 4096u16 , subdevice : 20739u16 , name : "eHBA 9705-48i SAS/SATA Storage Adapter" }] } , Device { vendor_id : 4096u16 , id : 242u16 , name : "Fusion-MPT Switch SAS52xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 688u16 , name : "Virtual Endpoint on PCIe Switch" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 1u16 , name : "ThinkSystem 1610-4P NVMe Switch Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 2u16 , name : "ThinkSystem 810-4P NVMe Switch Adapter" }] } , Device { vendor_id : 4096u16 , id : 689u16 , name : "Virtual Endpoint on PCIe Switch (9749)" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 4u16 , name : "ThinkSystem 1610-8P NVMe Switch Adapter" }] } , Device { vendor_id : 4096u16 , id : 690u16 , name : "Virtual PCIe Placeholder Endpoint" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1031u16 , name : "MegaRAID" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 1328u16 , name : "MegaRAID 530 SCSI 320-0X RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 1329u16 , name : "MegaRAID 531 SCSI 320-4X RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 1330u16 , name : "MegaRAID 532 SCSI 320-2X RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1329u16 , name : "PowerEdge Expandable RAID Controller 4/QC" } , SubSystem { subvendor : 4136u16 , subdevice : 1331u16 , name : "PowerEdge Expandable RAID Controller 4/QC" } , SubSystem { subvendor : 32902u16 , subdevice : 1328u16 , name : "MegaRAID Intel RAID Controller SRCZCRX" } , SubSystem { subvendor : 32902u16 , subdevice : 1330u16 , name : "MegaRAID Intel RAID Controller SRCU42X" }] } , Device { vendor_id : 4096u16 , id : 1032u16 , name : "MegaRAID" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 1u16 , name : "MegaRAID SCSI 320-1E RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 2u16 , name : "MegaRAID SCSI 320-2E RAID Controller" } , SubSystem { subvendor : 4133u16 , subdevice : 77u16 , name : "MegaRAID ACER ROMB-2E RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1u16 , name : "PowerEdge RAID Controller PERC4e/SC" } , SubSystem { subvendor : 4136u16 , subdevice : 2u16 , name : "PowerEdge RAID Controller PERC4e/DC" } , SubSystem { subvendor : 4136u16 , subdevice : 18u16 , name : "PowerEdge RAID Controller RAC4" } , SubSystem { subvendor : 4136u16 , subdevice : 21u16 , name : "PowerEdge RAID Controller PERC5" } , SubSystem { subvendor : 4136u16 , subdevice : 7939u16 , name : "PowerEdge RAID Controller PERC5" } , SubSystem { subvendor : 5940u16 , subdevice : 4197u16 , name : "FSC MegaRAID PCI Express ROMB" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "MegaRAID Intel RAID Controller SRCU42E" } , SubSystem { subvendor : 32902u16 , subdevice : 13385u16 , name : "MegaRAID Intel RAID Controller SROMBU" }] } , Device { vendor_id : 4096u16 , id : 1033u16 , name : "MegaRAID" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 12292u16 , name : "MegaRAID SATA 300-4X RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 12296u16 , name : "MegaRAID SATA 300-8X RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 12296u16 , name : "MegaRAID RAID Controller SRCS28X" } , SubSystem { subvendor : 32902u16 , subdevice : 13361u16 , name : "MegaRAID RAID Controller Alief SROMBU42E" } , SubSystem { subvendor : 32902u16 , subdevice : 13465u16 , name : "MegaRAID RAID Controller Harwich SROMBU42E" }] } , Device { vendor_id : 4096u16 , id : 1041u16 , name : "MegaRAID SAS 1068" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4097u16 , name : "MegaRAID SAS 8408E" } , SubSystem { subvendor : 4096u16 , subdevice : 4098u16 , name : "MegaRAID SAS 8480E" } , SubSystem { subvendor : 4096u16 , subdevice : 4099u16 , name : "MegaRAID SAS 8344ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4100u16 , name : "MegaRAID SAS 8308ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 4104u16 , name : "MegaRAID SAS 84016E" } , SubSystem { subvendor : 4096u16 , subdevice : 4108u16 , name : "MegaRAID SATA 300-12E" } , SubSystem { subvendor : 4096u16 , subdevice : 4109u16 , name : "MegaRAID SATA 300-16E" } , SubSystem { subvendor : 4096u16 , subdevice : 8196u16 , name : "MegaRAID SATA 300-8ELP" } , SubSystem { subvendor : 4096u16 , subdevice : 8197u16 , name : "MegaRAID SATA 300-4ELP" } , SubSystem { subvendor : 4147u16 , subdevice : 33415u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 4180u16 , subdevice : 12310u16 , name : "MegaRAID SAS RoMB Server" } , SubSystem { subvendor : 5940u16 , subdevice : 4225u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 5940u16 , subdevice : 4259u16 , name : "MegaRAID SAS PCI Express ROMB" } , SubSystem { subvendor : 32902u16 , subdevice : 4097u16 , name : "RAID Controller SRCSAS18E" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "RAID Controller SRCSAS144E" } , SubSystem { subvendor : 32902u16 , subdevice : 13568u16 , name : "SROMBSAS18E RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13569u16 , name : "SROMBSAS18E RAID Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 13572u16 , name : "SROMBSAS18E RAID Controller" }] } , Device { vendor_id : 4096u16 , id : 1043u16 , name : "MegaRAID SAS 1068 [Verde ZCR]" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4101u16 , name : "MegaRAID SAS 8300XLP" }] } , Device { vendor_id : 4096u16 , id : 1569u16 , name : "FC909 Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1570u16 , name : "FC929 Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4128u16 , name : "44929 O Dual Fibre Channel card" }] } , Device { vendor_id : 4096u16 , id : 1571u16 , name : "FC929 LAN" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1572u16 , name : "FC919 Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1573u16 , name : "FC919 LAN" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1574u16 , name : "FC929X Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4112u16 , name : "7202-XP-LC Dual Fibre Channel card" }] } , Device { vendor_id : 4096u16 , id : 1575u16 , name : "FC929X LAN" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1576u16 , name : "FC919X Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1577u16 , name : "FC919X LAN" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1600u16 , name : "FC949X Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1602u16 , name : "FC939X Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1606u16 , name : "FC949ES Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1793u16 , name : "83C885 NT50 DigitalScape Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 1794u16 , name : "Yellowfin G-NIC gigabit ethernet" , subsystems : & [SubSystem { subvendor : 4888u16 , subdevice : 0u16 , name : "PEI100X" }] } , Device { vendor_id : 4096u16 , id : 2052u16 , name : "SA2010" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 2053u16 , name : "SA2010ZC" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 2054u16 , name : "SA2020" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 2055u16 , name : "SA2020ZC" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 2305u16 , name : "61C102" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 4096u16 , name : "63C815" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 4320u16 , name : "MegaRAID 12GSAS/PCIe Unsupported SAS39xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 6880u16 , name : "PERC H755 Adapter - Invalid Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6881u16 , name : "PERC H755 Front - Invalid Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6882u16 , name : "PERC H755N Front - Invalid Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6883u16 , name : "PERC H755 MX - Invalid Device" }] } , Device { vendor_id : 4096u16 , id : 4321u16 , name : "MegaRAID 12GSAS/PCIe SAS39xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 6880u16 , name : "PERC H755 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 6881u16 , name : "PERC H755 Front" } , SubSystem { subvendor : 4136u16 , subdevice : 6882u16 , name : "PERC H755N Front" } , SubSystem { subvendor : 4136u16 , subdevice : 6883u16 , name : "PERC H755 MX" } , SubSystem { subvendor : 7497u16 , subdevice : 1546u16 , name : "ThinkSystem RAID 940-8i 4GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1547u16 , name : "ThinkSystem RAID 940-8i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1548u16 , name : "ThinkSystem RAID 940-16i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1549u16 , name : "ThinkSystem RAID 940-16i 8GB Flash PCIe Gen4 12Gb Internal Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1550u16 , name : "ThinkSystem RAID 940-32i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1551u16 , name : "ThinkSystem RAID 940-8e 4GB Flash PCIe Gen4 12Gb Adapter" }] } , Device { vendor_id : 4096u16 , id : 4322u16 , name : "MegaRAID 12GSAS/PCIe Secure SAS39xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 16384u16 , name : "MegaRAID 9560-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 16386u16 , name : "MegaRAID 9561-16i" } , SubSystem { subvendor : 4096u16 , subdevice : 16400u16 , name : "MegaRAID 9560-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 16405u16 , name : "MegaRAID 9550-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 16416u16 , name : "MegaRAID 9580-8i8e" } , SubSystem { subvendor : 4096u16 , subdevice : 16560u16 , name : "MegaRAID 9562-16i" } , SubSystem { subvendor : 4136u16 , subdevice : 6880u16 , name : "PERC H755 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 6881u16 , name : "PERC H755 Front" } , SubSystem { subvendor : 4136u16 , subdevice : 6882u16 , name : "PERC H755N Front" } , SubSystem { subvendor : 4136u16 , subdevice : 6883u16 , name : "PERC H755 MX" } , SubSystem { subvendor : 4136u16 , subdevice : 8561u16 , name : "PERC H750 Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8566u16 , name : "PERC H750 Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7014u16 , name : "AOC-S3908L-H8iR RAID Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7015u16 , name : "AOC-S3916L-H16iR RAID Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7174u16 , name : "AOC-S3916L-H16iR-32DD+ RAID Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1546u16 , name : "ThinkSystem RAID 940-8i 4GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1547u16 , name : "ThinkSystem RAID 940-8i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1548u16 , name : "ThinkSystem RAID 940-16i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1549u16 , name : "ThinkSystem RAID 940-16i 8GB Flash PCIe Gen4 12Gb Internal Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1550u16 , name : "ThinkSystem RAID 940-32i 8GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1551u16 , name : "ThinkSystem RAID 940-8e 4GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1552u16 , name : "ThinkSystem RAID 940-16i 4GB Flash PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 16384u16 , name : "RAID Controller RS3P4TF160F" } , SubSystem { subvendor : 32902u16 , subdevice : 16416u16 , name : "RAID Controller RS3P4MF088F" }] } , Device { vendor_id : 4096u16 , id : 4323u16 , name : "MegaRAID 12GSAS/PCIe Unsupported SAS39xx" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 6880u16 , name : "PERC H755 Adapter - Tampered Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6881u16 , name : "PERC H755 Front - Tampered Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6882u16 , name : "PERC H755N Front - Tampered Device" } , SubSystem { subvendor : 4136u16 , subdevice : 6883u16 , name : "PERC H755 MX - Tampered Device" }] } , Device { vendor_id : 4096u16 , id : 4324u16 , name : "MegaRAID 12GSAS/PCIe Unsupported SAS38xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 4325u16 , name : "MegaRAID 12GSAS/PCIe SAS38xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 4326u16 , name : "MegaRAID 12GSAS/PCIe Secure SAS38xx" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 1241u16 , name : "3808N iMR ROMB" } , SubSystem { subvendor : 4096u16 , subdevice : 1242u16 , name : "3808N iMR ROMB" } , SubSystem { subvendor : 4096u16 , subdevice : 1243u16 , name : "3808N iMR ROMB" } , SubSystem { subvendor : 4096u16 , subdevice : 1244u16 , name : "3808N iMR ROMB" } , SubSystem { subvendor : 4096u16 , subdevice : 1245u16 , name : "3808N iMR ROMB" } , SubSystem { subvendor : 4096u16 , subdevice : 16597u16 , name : "MegaRAID 9540-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 16598u16 , name : "MegaRAID 9520-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 16600u16 , name : "MegaRAID 9524-8i" } , SubSystem { subvendor : 4096u16 , subdevice : 16608u16 , name : "MegaRAID 9540-2M2" } , SubSystem { subvendor : 4136u16 , subdevice : 6717u16 , name : "BOSS-MR1" } , SubSystem { subvendor : 4136u16 , subdevice : 8562u16 , name : "PERC H355 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8563u16 , name : "PERC H355 Front" } , SubSystem { subvendor : 4136u16 , subdevice : 8564u16 , name : "PERC H350 Mini" } , SubSystem { subvendor : 4136u16 , subdevice : 8567u16 , name : "PERC H350 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8601u16 , name : "PERC H350 Mini LP" } , SubSystem { subvendor : 5593u16 , subdevice : 7069u16 , name : "AOC-S3816L-L16IR Storage Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7071u16 , name : "AOC-S3816L-L8IR Storage Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7277u16 , name : "AOC-S3808L-L8IR Storage Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 7278u16 , name : "AOC-SLG4-2H8M2 Storage Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1285u16 , name : "ThinkSystem RAID 540-8i PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1286u16 , name : "ThinkSystem RAID 540-16i PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1287u16 , name : "ThinkSystem RAID 545-8i PCIe Gen4 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1288u16 , name : "ThinkSystem RAID 545-8i PCIe Gen4 12Gb Internal Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1792u16 , name : "ThinkSystem M.2 RAID B540i-2i SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1793u16 , name : "ThinkSystem 7mm RAID B540p-2HS SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1794u16 , name : "ThinkSystem M.2 RAID B540p-2HS SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1795u16 , name : "ThinkSystem M.2 RAID B540d-2HS SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1796u16 , name : "ThinkSystem M.2 RAID B545i-2i SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1797u16 , name : "ThinkSystem M.2 RAID B550p-2HS SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1798u16 , name : "ThinkSystem M.2 RAID B550d-2HS SATA/NVMe Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 1799u16 , name : "ThinkSystem M.2 RAID B550i-2i SATA/NVMe Enablement Kit" }] } , Device { vendor_id : 4096u16 , id : 4327u16 , name : "MegaRAID 12GSAS/PCIe Unsupported SAS38xx" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 6496u16 , name : "MegaRAID" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 1304u16 , name : "MegaRAID 518 SCSI 320-2 Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 1312u16 , name : "MegaRAID 520 SCSI 320-1 Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 1314u16 , name : "MegaRAID 522 i4 133 RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 1315u16 , name : "MegaRAID SATA 150-6 RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 17699u16 , name : "MegaRAID SATA 150-4 RAID Controller" } , SubSystem { subvendor : 4096u16 , subdevice : 42272u16 , name : "MegaRAID ZCR SCSI 320-0 Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1304u16 , name : "MegaRAID 518 DELL PERC 4/DC RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1312u16 , name : "MegaRAID 520 DELL PERC 4/SC RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1329u16 , name : "PowerEdge Expandable RAID Controller 4/QC" } , SubSystem { subvendor : 4136u16 , subdevice : 1331u16 , name : "PowerEdge Expandable RAID Controller 4/QC" } , SubSystem { subvendor : 32902u16 , subdevice : 1312u16 , name : "MegaRAID RAID Controller SRCU41L" } , SubSystem { subvendor : 32902u16 , subdevice : 1315u16 , name : "MegaRAID RAID Controller SRCS16" }] } , Device { vendor_id : 4096u16 , id : 12368u16 , name : "SAS2008 PCI-Express Fusion-MPT SAS-2" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 12624u16 , name : "1068e" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 24577u16 , name : "DX1 Multiformat Broadcast HD/SD Encoder/Decoder" , subsystems : & [] } , Device { vendor_id : 4096u16 , id : 49168u16 , name : "PEX880xx PCIe Gen 4 Switch" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4107u16 , name : "PEX88000 PCIe Gen 4 Virtual Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 8196u16 , name : "PEX88000 Virtual PCIe TWC/NT2 Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 8197u16 , name : "PEX88000 Virtual PCIe gDMA Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 40996u16 , name : "PEX88024 24 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41010u16 , name : "PEX88032 32 lane/port PCIe Gen 4 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41032u16 , name : "PEX88048 50 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41060u16 , name : "PEX88064 64 lane/port PCIe Gen 4 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41088u16 , name : "PEX88080 80 lane/port PCIe Gen 4 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41110u16 , name : "PEX88096 98 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 19538u16 , subdevice : 40776u16 , name : "LRNV9F48 4-port Built-in 8654 NVMe Switching Adapter" }] } , Device { vendor_id : 4096u16 , id : 49170u16 , name : "PEX880xx PCIe Gen 4 Switch" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4107u16 , name : "PEX88000 PCIe Gen 4 Virtual Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 8196u16 , name : "PEX88000 Virtual PCIe TWC/NT2 Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 8197u16 , name : "PEX88000 Virtual PCIe gDMA Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 40996u16 , name : "PEX88024 24 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41010u16 , name : "PEX88032 32 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41032u16 , name : "PEX88048 48 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41060u16 , name : "PEX88064 64 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41088u16 , name : "PEX88080 80 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 4096u16 , subdevice : 41110u16 , name : "PEX88096 96 lane/port PCIe Gen 4.0 Switch" } , SubSystem { subvendor : 7497u16 , subdevice : 3u16 , name : "ThinkSystem 1611-8P PCIe Gen4 NVMe Switch Adapter" }] } , Device { vendor_id : 4096u16 , id : 49200u16 , name : "PEX890xx PCIe Gen 5 Switch" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 36u16 , name : "PEX89024 PCIe Gen 5 24 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 50u16 , name : "PEX89032 PCIe Gen 5 32 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 72u16 , name : "PEX89048 PCIe Gen 5 48 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 114u16 , name : "PEX89072 PCIe Gen 5 36 port/72 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 136u16 , name : "PEX89088 PCIe Gen 5 44 port/88 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 260u16 , name : "PEX89104 PCIe Gen 5 52 port/104 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 324u16 , name : "PEX89144 PCIe Gen 5 72 port/144 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 4107u16 , name : "PEX89000 PCIe Gen 5 Virtual Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 8196u16 , name : "PEX89000 Virtual PCIe TWC/NT 2.0 Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 8197u16 , name : "PEX89000 Virtual PCIe gDMA Endpoint" } , SubSystem { subvendor : 5593u16 , subdevice : 7460u16 , name : "AOM-PCIE5-418P" }] } , Device { vendor_id : 4096u16 , id : 49204u16 , name : "PEX890xx PCIe Gen 5 Switch" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 36u16 , name : "PEX89024 PCIe Gen 5 24 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 50u16 , name : "PEX89032 PCIe Gen 5 32 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 72u16 , name : "PEX89048 PCIe Gen 5 48 port/lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 4107u16 , name : "PEX89000 PCIe Gen 5 Virtual Upstream/Downstream Port" } , SubSystem { subvendor : 4096u16 , subdevice : 8196u16 , name : "PEX89000 Virtual PCIe TWC/NT 2.0 Endpoint" } , SubSystem { subvendor : 4096u16 , subdevice : 8197u16 , name : "PEX89000 Virtual PCIe gDMA Endpoint" }] } , Device { vendor_id : 4096u16 , id : 49216u16 , name : "PEX90xxx PCIe Gen 6 Switch" , subsystems : & [] }] }),
        (4684, Vendor { id : 4684u16 , name : "Solitron Technologies, Inc." , devices : & [] }),
        (7716, Vendor { id : 7716u16 , name : "Squirrels Research Labs" , devices : & [Device { vendor_id : 7716u16 , id : 257u16 , name : "Acorn CLE-101" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 533u16 , name : "Acorn CLE-215" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 543u16 , name : "Acorn CLE-215+" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 5413u16 , name : "Xilinx BCU-1525" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 5427u16 , name : "ForestKitten 33" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 5683u16 , name : "JCM33" , subsystems : & [] } , Device { vendor_id : 7716u16 , id : 5685u16 , name : "JCM35" , subsystems : & [] }] }),
        (6554, Vendor { id : 6554u16 , name : "Pulse-LINK, Inc." , devices : & [] }),
        (5400, Vendor { id : 5400u16 , name : "Kontron" , devices : & [] }),
        (6564, Vendor { id : 6564u16 , name : "Owl Cyber Defense Solutions" , devices : & [] }),
        (5108, Vendor { id : 5108u16 , name : "Troika Networks, Inc." , devices : & [Device { vendor_id : 5108u16 , id : 5121u16 , name : "Zentai Fibre Channel Adapter" , subsystems : & [] }] }),
        (5345, Vendor { id : 5345u16 , name : "INVERTEX" , devices : & [] }),
        (7361, Vendor { id : 7361u16 , name : "ADATA Technology Co., Ltd." , devices : & [Device { vendor_id : 7361u16 , id : 4610u16 , name : "IM2P32A8 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 5634u16 , name : "LEGEND 900 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 8803u16 , name : "XPG SX6000 Lite NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 9992u16 , name : "Premier Extreme SDXC SD 7.0 / microSDXC SD 7.1 Express Card (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 12968u16 , name : "SM2P32A8 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 13299u16 , name : "IM2P33F3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 13300u16 , name : "IM2P33F4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 13304u16 , name : "IM2P33F8 series NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 16701u16 , name : "SM2P41D3Q NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 16824u16 , name : "IM2P41B8P NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 16835u16 , name : "SM2P41C3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 16840u16 , name : "SM2P41C8 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 16851u16 , name : "SM2P41D3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 21046u16 , name : "XPG GAMMIX S70 BLADE NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 21328u16 , name : "XPG GAMMIX S50, S50 Lite NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 21360u16 , name : "GAMMIX S70 BLADE, PS5 PREMIUM NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 22370u16 , name : "FALCON, GAMMIX S41, SPECTRIX S40G NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 22371u16 , name : "XPG GAMMIX S5 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 22374u16 , name : "XPG GAMMIXS1 1L, XPG GAMMIX S5, LEGEND 710 / 740, SWORDFISH NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 22386u16 , name : "LEGEND 850 LITE NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 24874u16 , name : "LEGEND 750 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 24890u16 , name : "ATOM 50, LEGEND 840 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25114u16 , name : "LEGEND 850 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25130u16 , name : "LEGEND 960 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25146u16 , name : "LEGEND 710 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25162u16 , name : "LEGEND 700, XPG GAMMIX S20 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25178u16 , name : "LEGEND 850 LITE NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25210u16 , name : "LEGEND 800 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25226u16 , name : "LEGEND 800 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25370u16 , name : "LEGEND 970 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25402u16 , name : "LEGEND 900 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25420u16 , name : "LEGEND 820 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25434u16 , name : "XPG GAMMIX S60 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25450u16 , name : "XPG GAMMIX S55 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25626u16 , name : "LEGEND 970 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25642u16 , name : "XPG GAMMIX S50 CORE NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25706u16 , name : "XPG MARS 980 BLADE NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25722u16 , name : "XPG MARS 980 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25738u16 , name : "LEGEND 860 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 25786u16 , name : "LEGEND 900 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 33281u16 , name : "XPG SX8200 Pro PCIe Gen3x4 M.2 2280 Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 40976u16 , name : "TRUSTA T7P5 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7361u16 , id : 41248u16 , name : "TRUSTA T5P4B NVMe SSD" , subsystems : & [] }] }),
        (61713, Vendor { id : 61713u16 , name : "Framework Computer Inc." , devices : & [] }),
        (7571, Vendor { id : 7571u16 , name : "YADRO" , devices : & [] }),
        (19793, Vendor { id : 19793u16 , name : "MediaQ Inc." , devices : & [Device { vendor_id : 19793u16 , id : 512u16 , name : "MQ-200" , subsystems : & [] }] }),
        (6889, Vendor { id : 6889u16 , name : "Wilocity Ltd." , devices : & [Device { vendor_id : 6889u16 , id : 257u16 , name : "Wil6200 PCI Express Upstream Port" , subsystems : & [] } , Device { vendor_id : 6889u16 , id : 512u16 , name : "Wil6200 PCI Express Port" , subsystems : & [] } , Device { vendor_id : 6889u16 , id : 513u16 , name : "Wil6200 Wireless PCI Express Port" , subsystems : & [] } , Device { vendor_id : 6889u16 , id : 769u16 , name : "Wil6200 802.11ad Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6889u16 , id : 770u16 , name : "Wil6200 802.11ad Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6889u16 , id : 784u16 , name : "Wil6200 802.11ad Wireless Network Adapter" , subsystems : & [] }] }),
        (8169, Vendor { id : 8169u16 , name : "MemryX" , devices : & [Device { vendor_id : 8169u16 , id : 256u16 , name : "MX3" , subsystems : & [] }] }),
        (5167, Vendor { id : 5167u16 , name : "Radicom Research Inc" , devices : & [] }),
        (5144, Vendor { id : 5144u16 , name : "Kyushu electronics systems Inc" , devices : & [] }),
        (6176, Vendor { id : 6176u16 , name : "InfiniCon Systems Inc." , devices : & [] }),
        (4846, Vendor { id : 4846u16 , name : "Orange Micro" , devices : & [] }),
        (4909, Vendor { id : 4909u16 , name : "Integrated Silicon Solution, Inc." , devices : & [] }),
        (4205, Vendor { id : 4205u16 , name : "Sequent Computer Systems" , devices : & [] }),
        (5636, Vendor { id : 5636u16 , name : "Central System Research Co Ltd" , devices : & [] }),
        (6484, Vendor { id : 6484u16 , name : "One Stop Systems, Inc." , devices : & [] }),
        (4661, Vendor { id : 4661u16 , name : "SMART Modular Technologies" , devices : & [Device { vendor_id : 4661u16 , id : 22560u16 , name : "DC5820 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 4661u16 , id : 49729u16 , name : "CXA-4F1W" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9090u16 , name : "4-DIMM Add In Card" }] }] }),
        (6760, Vendor { id : 6760u16 , name : "VirtenSys Limited" , devices : & [] }),
        (4479, Vendor { id : 4479u16 , name : "Integrated Circuit Systems" , devices : & [] }),
        (4257, Vendor { id : 4257u16 , name : "Juko Electronics Ind. Co. Ltd" , devices : & [] }),
        (8258, Vendor { id : 8258u16 , name : "Xi'an UniIC Semiconductors Co., Ltd" , devices : & [Device { vendor_id : 8258u16 , id : 56064u16 , name : "UWSC256DDQYACC-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56065u16 , name : "UWSC512DDQLBCC-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56066u16 , name : "UWSC512DDQLACC-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56067u16 , name : "UWSC1T0DDQLACC-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56320u16 , name : "UWSD512DDQYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56321u16 , name : "UWSD1T0DDQYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56322u16 , name : "UWSD2T0DDQYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56323u16 , name : "UWSD1T0DDTYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56324u16 , name : "UWSD2T0DDTYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56325u16 , name : "UWSD4T0DDTYACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56326u16 , name : "UWSD512DDQYACB-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56327u16 , name : "UWSD1T0DDQYACB-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56328u16 , name : "UWSD512DDTMACA-N" , subsystems : & [] } , Device { vendor_id : 8258u16 , id : 56329u16 , name : "UWSD1T0DDTMACA-N" , subsystems : & [] }] }),
        (5566, Vendor { id : 5566u16 , name : "Sola Electronics" , devices : & [] }),
        (4468, Vendor { id : 4468u16 , name : "Bridgeport Machines" , devices : & [] }),
        (5470, Vendor { id : 5470u16 , name : "LP Elektronik GmbH" , devices : & [] }),
        (5625, Vendor { id : 5625u16 , name : "Zenith Electronics Corp" , devices : & [] }),
        (5409, Vendor { id : 5409u16 , name : "BELL Corp" , devices : & [] }),
        (4754, Vendor { id : 4754u16 , name : "Tritech Microelectronics Inc" , devices : & [Device { vendor_id : 4754u16 , id : 64514u16 , name : "Pyramid3D TR25202" , subsystems : & [] }] }),
        (4108, Vendor { id : 4108u16 , name : "Tseng Labs Inc" , devices : & [Device { vendor_id : 4108u16 , id : 12802u16 , name : "ET4000/W32p rev A" , subsystems : & [] } , Device { vendor_id : 4108u16 , id : 12805u16 , name : "ET4000/W32p rev B" , subsystems : & [] } , Device { vendor_id : 4108u16 , id : 12806u16 , name : "ET4000/W32p rev C" , subsystems : & [] } , Device { vendor_id : 4108u16 , id : 12807u16 , name : "ET4000/W32p rev D" , subsystems : & [] } , Device { vendor_id : 4108u16 , id : 12808u16 , name : "ET6000" , subsystems : & [] } , Device { vendor_id : 4108u16 , id : 18178u16 , name : "ET6300" , subsystems : & [] }] }),
        (7916, Vendor { id : 7916u16 , name : "Viscore Technologies Ltd" , devices : & [Device { vendor_id : 7916u16 , id : 258u16 , name : "VSE250231S Dual-port 10Gb/25Gb Ethernet PCIe" , subsystems : & [] } , Device { vendor_id : 7916u16 , id : 7916u16 , name : "VSE250231S Dual-port 10Gb/25Gb Ethernet PCIe" , subsystems : & [] }] }),
        (5338, Vendor { id : 5338u16 , name : "National Aerospace Laboratories" , devices : & [] }),
        (54484, Vendor { id : 54484u16 , name : "Dy4 Systems Inc" , devices : & [Device { vendor_id : 54484u16 , id : 1537u16 , name : "PCI Mezzanine Card" , subsystems : & [] }] }),
        (5426, Vendor { id : 5426u16 , name : "ECHELON Corp" , devices : & [Device { vendor_id : 5426u16 , id : 32u16 , name : "LonWorks PCLTA-20 PCI LonTalk Adapter" , subsystems : & [] }] }),
        (4193, Vendor { id : 4193u16 , name : "I.I.T." , devices : & [Device { vendor_id : 4193u16 , id : 1u16 , name : "AGX016" , subsystems : & [] } , Device { vendor_id : 4193u16 , id : 2u16 , name : "IIT3204/3501" , subsystems : & [] }] }),
        (4721, Vendor { id : 4721u16 , name : "GW Instruments" , devices : & [] }),
        (4759, Vendor { id : 4759u16 , name : "Holco Enterprise Co, Ltd/Shuttle Computer" , devices : & [Device { vendor_id : 4759u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (20803, Vendor { id : 20803u16 , name : "Qualcomm Inc" , devices : & [] }),
        (5015, Vendor { id : 5015u16 , name : "Cologne Chip Designs GmbH" , devices : & [Device { vendor_id : 5015u16 , id : 2228u16 , name : "ISDN network Controller [HFC-4S]" , subsystems : & [SubSystem { subvendor : 5015u16 , subdevice : 2228u16 , name : "HFC-4S [Cologne Chip HFC-4S Eval. Board]" } , SubSystem { subvendor : 5015u16 , subdevice : 46362u16 , name : "HFC-4S [Allo.com BRI card]" } , SubSystem { subvendor : 5015u16 , subdevice : 46368u16 , name : "HFC-4S [IOB4ST]" } , SubSystem { subvendor : 5015u16 , subdevice : 46400u16 , name : "HFC-4S [Swyx SX2 QuadBri]" } , SubSystem { subvendor : 5015u16 , subdevice : 46416u16 , name : "HFC-4S [Junghanns.NET quadBRI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46422u16 , name : "HFC-4S [Junghanns.NET duoBRI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46425u16 , name : "HFC-4S [Junghanns.NET duoBRI miniPCI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46432u16 , name : "HFC-4S [BeroNet BN4S0]" } , SubSystem { subvendor : 5015u16 , subdevice : 46438u16 , name : "HFC-4S [BeroNet BN2S0]" } , SubSystem { subvendor : 5015u16 , subdevice : 46439u16 , name : "HFC-4S [BeroNet BN1S0 miniPCI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46440u16 , name : "HFC-4S [BeroNet BN4S0 miniPCI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46441u16 , name : "HFC-4S [BeroNet BN2S0 miniPCI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46624u16 , name : "HFC-4S" } , SubSystem { subvendor : 5015u16 , subdevice : 46930u16 , name : "HFC-4S [Junghanns.NET quadBRI PCIe]" } , SubSystem { subvendor : 5015u16 , subdevice : 46945u16 , name : "HFC-4S [BeroNet BN2S0 PCIe]" } , SubSystem { subvendor : 5015u16 , subdevice : 46946u16 , name : "HFC-4S [BeroNet BN4S0 PCIe]" } , SubSystem { subvendor : 5015u16 , subdevice : 59524u16 , name : "HFC-4S [OpenVox B200P]" } , SubSystem { subvendor : 5015u16 , subdevice : 59528u16 , name : "HFC-4S [OpenVox B200P / B400P]" }] } , Device { vendor_id : 5015u16 , id : 5816u16 , name : "ISDN network Controller [HFC-8S]" , subsystems : & [SubSystem { subvendor : 5015u16 , subdevice : 5816u16 , name : "HFC-8S [Cologne Chip HFC-8S Eval. Board]" } , SubSystem { subvendor : 5015u16 , subdevice : 46369u16 , name : "HFC-8S [IOB4ST Recording]" } , SubSystem { subvendor : 5015u16 , subdevice : 46370u16 , name : "HFC-8S [IOB8ST]" } , SubSystem { subvendor : 5015u16 , subdevice : 46418u16 , name : "HFC-8S [Junghanns.NET octoBRI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46427u16 , name : "HFC-8S [Junghanns.NET octoBRI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46434u16 , name : "HFC-8S [BeroNet BN8S0]" } , SubSystem { subvendor : 5015u16 , subdevice : 46443u16 , name : "HFC-8S [BeroNet BN8S0+]" } , SubSystem { subvendor : 5015u16 , subdevice : 46626u16 , name : "HFC-8S" } , SubSystem { subvendor : 5015u16 , subdevice : 59800u16 , name : "HFC-8S [OpenVox B800P]" }] } , Device { vendor_id : 5015u16 , id : 11216u16 , name : "ISDN network controller [HFC-PCI]" , subsystems : & [SubSystem { subvendor : 1653u16 , subdevice : 5892u16 , name : "ISDN Adapter (PCI Bus, D, C)" } , SubSystem { subvendor : 1653u16 , subdevice : 5896u16 , name : "ISDN Adapter (PCI Bus, D, C, ACPI)" } , SubSystem { subvendor : 5015u16 , subdevice : 11216u16 , name : "ISDN Board" } , SubSystem { subvendor : 58559u16 , subdevice : 4096u16 , name : "CI1-1-Harp" }] } , Device { vendor_id : 5015u16 , id : 12465u16 , name : "ISDN network Controller [HFC-E1]" , subsystems : & [SubSystem { subvendor : 5015u16 , subdevice : 12465u16 , name : "HFC-E1 [Cologne Chip HFC-E1 Eval. Board]" } , SubSystem { subvendor : 5015u16 , subdevice : 46371u16 , name : "HFC-E1 [IOB1E1]" } , SubSystem { subvendor : 5015u16 , subdevice : 46403u16 , name : "HFC-E1 [Swyx SX2 SinglePRI V2]" } , SubSystem { subvendor : 5015u16 , subdevice : 46404u16 , name : "HFC-E1 [Swyx SX2 DualPRI V2]" } , SubSystem { subvendor : 5015u16 , subdevice : 46419u16 , name : "HFC-E1 [Junghanns.NET singleE1]" } , SubSystem { subvendor : 5015u16 , subdevice : 46420u16 , name : "HFC-E1 [Junghanns.NET doubleE1]" } , SubSystem { subvendor : 5015u16 , subdevice : 46421u16 , name : "HFC-E1 [Junghanns.NET doubleE1 2.0]" } , SubSystem { subvendor : 5015u16 , subdevice : 46426u16 , name : "HFC-E1 [Junghanns.NET singleE1 miniPCI]" } , SubSystem { subvendor : 5015u16 , subdevice : 46435u16 , name : "HFC-E1 [beroNet BN1E1]" } , SubSystem { subvendor : 5015u16 , subdevice : 46436u16 , name : "HFC-E1 [beroNet BN2E1]" } , SubSystem { subvendor : 5015u16 , subdevice : 46437u16 , name : "HFC-E1 [beroNet BN2E1+]" } , SubSystem { subvendor : 5015u16 , subdevice : 46442u16 , name : "HFC-E1 [beroNet BN1E1 miniPCI]" }] } , Device { vendor_id : 5015u16 , id : 46848u16 , name : "ISDN network controller PrimuX S0 [HFC-PCI]" , subsystems : & [] } , Device { vendor_id : 5015u16 , id : 61441u16 , name : "GSM Network Controller [HFC-4GSM]" , subsystems : & [] }] }),
        (7645, Vendor { id : 7645u16 , name : "Thorlabs" , devices : & [] }),
        (5162, Vendor { id : 5162u16 , name : "Kingmax Technology Inc" , devices : & [] }),
        (7196, Vendor { id : 7196u16 , name : "Symphony" , devices : & [Device { vendor_id : 7196u16 , id : 1u16 , name : "82C101" , subsystems : & [] }] }),
        (4222, Vendor { id : 4222u16 , name : "Interphase Corporation" , devices : & [Device { vendor_id : 4222u16 , id : 1u16 , name : "5515 ATM Adapter [Flipper]" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 2u16 , name : "100 VG AnyLan Controller" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 4u16 , name : "5526 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 5u16 , name : "x526 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 8u16 , name : "5525/5575 ATM Adapter (155 Mbit) [Atlantic]" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36867u16 , name : "5535-4P-BRI-ST" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36871u16 , name : "5535-4P-BRI-U" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36872u16 , name : "5535-1P-SR" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36876u16 , name : "5535-1P-SR-ST" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36878u16 , name : "5535-1P-SR-U" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36881u16 , name : "5535-1P-PRI" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36883u16 , name : "5535-2P-PRI" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36899u16 , name : "5536-4P-BRI-ST" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36903u16 , name : "5536-4P-BRI-U" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36913u16 , name : "5536-1P-PRI" , subsystems : & [] } , Device { vendor_id : 4222u16 , id : 36915u16 , name : "5536-2P-PRI" , subsystems : & [] }] }),
        (4774, Vendor { id : 4774u16 , name : "Scalable Networks, Inc." , devices : & [] }),
        (6912, Vendor { id : 6912u16 , name : "Montage Technology Co., Ltd." , devices : & [Device { vendor_id : 6912u16 , id : 49153u16 , name : "CXL Memory Expander Controller M88MX5891" , subsystems : & [SubSystem { subvendor : 8185u16 , subdevice : 162u16 , name : "CXL Memory Expander" } , SubSystem { subvendor : 8185u16 , subdevice : 164u16 , name : "CXL Memory Expander" }] }] }),
        (4545, Vendor { id : 4545u16 , name : "LSI Corporation" , devices : & [Device { vendor_id : 4545u16 , id : 1088u16 , name : "56k WinModem" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32789u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 4147u16 , subdevice : 32839u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 4147u16 , subdevice : 32847u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 4303u16 , subdevice : 4140u16 , name : "LB LT Modem V.90 56k" } , SubSystem { subvendor : 4303u16 , subdevice : 4170u16 , name : "BIBLO LT Modem 56k" } , SubSystem { subvendor : 4303u16 , subdevice : 4191u16 , name : "LB2 LT Modem V.90 56k" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Internal V.90 Modem" } , SubSystem { subvendor : 4545u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 4653u16 , subdevice : 16641u16 , name : "MDP7800-U Modem" } , SubSystem { subvendor : 4653u16 , subdevice : 16642u16 , name : "MDP7800SP-U Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 64u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 1089u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 1104u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 61696u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 61697u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5197u16 , subdevice : 8449u16 , name : "LT56PV Modem" } , SubSystem { subvendor : 5279u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" }] } , Device { vendor_id : 4545u16 , id : 1089u16 , name : "56k WinModem" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32845u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 4147u16 , subdevice : 32869u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 4242u16 , subdevice : 1088u16 , name : "Supra 56i" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Internal V.90 Modem" } , SubSystem { subvendor : 4545u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 4545u16 , subdevice : 1089u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 4653u16 , subdevice : 16640u16 , name : "MDP7800-U Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 64u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 256u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 1040u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 1056u16 , name : "TelePath Internet 56k WinModem" } , SubSystem { subvendor : 5088u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 1091u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 61698u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5142u16 , subdevice : 38916u16 , name : "CommWave 56k Modem" } , SubSystem { subvendor : 5149u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5199u16 , subdevice : 1089u16 , name : "Lucent 56k V.90 DF Modem" } , SubSystem { subvendor : 5199u16 , subdevice : 1097u16 , name : "Lucent 56k V.90 DF Modem" } , SubSystem { subvendor : 5199u16 , subdevice : 4365u16 , name : "Lucent Win Modem" } , SubSystem { subvendor : 5224u16 , subdevice : 1089u16 , name : "Presario 56k V.90 DF Modem" } , SubSystem { subvendor : 5736u16 , subdevice : 1088u16 , name : "Lucent Win Modem" }] } , Device { vendor_id : 4545u16 , id : 1090u16 , name : "56k WinModem" , subsystems : & [SubSystem { subvendor : 4545u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 4545u16 , subdevice : 1090u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 1042u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 1090u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5116u16 , subdevice : 9329u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5197u16 , subdevice : 8452u16 , name : "LT56PT Modem" } , SubSystem { subvendor : 5199u16 , subdevice : 4356u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5279u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5736u16 , subdevice : 1088u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" }] } , Device { vendor_id : 4545u16 , id : 1091u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1092u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1093u16 , name : "LT WinModem" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 8707u16 , name : "PRO/100+ MiniPCI (probably an Ambit U98.003.C.00 combo card)" } , SubSystem { subvendor : 32902u16 , subdevice : 8708u16 , name : "PRO/100+ MiniPCI on Armada E500" }] } , Device { vendor_id : 4545u16 , id : 1094u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1095u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1096u16 , name : "WinModem 56k" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 305u16 , name : "Lucent Win Modem" } , SubSystem { subvendor : 4147u16 , subdevice : 32870u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 48u16 , name : "56k Voice Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 64u16 , name : "LT WinModem 56k Data+Fax+Voice+Dsvd" } , SubSystem { subvendor : 5736u16 , subdevice : 9216u16 , name : "LT WinModem 56k (MiniPCI Ethernet+Modem)" }] } , Device { vendor_id : 4545u16 , id : 1097u16 , name : "L56xM+S [Mars-2] WinModem 56k" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45389u16 , name : "56k V.90 Modem" } , SubSystem { subvendor : 4116u16 , subdevice : 396u16 , name : "ThinkPad 600X" } , SubSystem { subvendor : 5088u16 , subdevice : 32u16 , name : "LT WinModem 56k Data+Fax" } , SubSystem { subvendor : 5088u16 , subdevice : 65u16 , name : "TelePath Internet 56k WinModem" } , SubSystem { subvendor : 5174u16 , subdevice : 1088u16 , name : "Lucent Win Modem" } , SubSystem { subvendor : 5199u16 , subdevice : 1097u16 , name : "Lucent 56k V.90 DFi Modem" } , SubSystem { subvendor : 5224u16 , subdevice : 1040u16 , name : "IBM ThinkPad T23" } , SubSystem { subvendor : 5224u16 , subdevice : 1088u16 , name : "Lucent Win Modem" } , SubSystem { subvendor : 5224u16 , subdevice : 1097u16 , name : "Presario 56k V.90 DFi Modem" }] } , Device { vendor_id : 4545u16 , id : 1098u16 , name : "F-1156IV WinModem (V90, 56KFlex)" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4210u16 , name : "LB Global LT Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 18u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5088u16 , subdevice : 66u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" } , SubSystem { subvendor : 5199u16 , subdevice : 4101u16 , name : "LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd" }] } , Device { vendor_id : 4545u16 , id : 1099u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1100u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1101u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1102u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1103u16 , name : "V90 WildWire Modem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1104u16 , name : "LT WinModem" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32936u16 , name : "Versa Note Vxi" } , SubSystem { subvendor : 5199u16 , subdevice : 16389u16 , name : "Magnia SG20" } , SubSystem { subvendor : 5224u16 , subdevice : 1104u16 , name : "Evo N600c" }] } , Device { vendor_id : 4545u16 , id : 1105u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1106u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1107u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1108u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1109u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1110u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1111u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1112u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1113u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1114u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1116u16 , name : "LT WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1121u16 , name : "V90 WildWire Modem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1122u16 , name : "V90 WildWire Modem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1152u16 , name : "Venus Modem (V90, 56KFlex)" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1164u16 , name : "V.92 56K WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1167u16 , name : "V.92 56k WinModem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 1568u16 , name : "Lucent V.92 Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 9728u16 , name : "StarPro26XX family (SP2601, SP2603, SP2612) DSP" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 21504u16 , name : "OR3TP12 FPSC" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 22102u16 , name : "Venus Modem" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 22529u16 , name : "USB" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 22530u16 , name : "USS-312 USB Controller" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 22531u16 , name : "USS-344S USB Controller" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 22545u16 , name : "FW322/323 [TrueFire] 1394a Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33428u16 , name : "LSI FW322/323 IEEE 1394a FireWire Controller" } , SubSystem { subvendor : 4541u16 , subdevice : 14u16 , name : "LSI FW323" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" } , SubSystem { subvendor : 36869u16 , subdevice : 51u16 , name : "Adaptec AFW-2100 (HP) 2102900-R" } , SubSystem { subvendor : 57005u16 , subdevice : 2048u16 , name : "FireWire Host Bus Adapter" }] } , Device { vendor_id : 4545u16 , id : 22785u16 , name : "FW643 [TrueFire] PCIe 1394b Controller" , subsystems : & [SubSystem { subvendor : 4545u16 , subdevice : 22784u16 , name : "FW643 [TrueFire] PCIe 1394b Controller" } , SubSystem { subvendor : 5187u16 , subdevice : 1603u16 , name : "FireBoard800-e V.2" } , SubSystem { subvendor : 5446u16 , subdevice : 1603u16 , name : "FWB-PCIE1X2x" }] } , Device { vendor_id : 4545u16 , id : 22787u16 , name : "FW533 [TrueFire] PCIe 1394a Controller" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 33040u16 , name : "T8110 H.100/H.110 TDM switch" , subsystems : & [SubSystem { subvendor : 4825u16 , subdevice : 12u16 , name : "E1/T1 PMXc cPCI carrier card" }] } , Device { vendor_id : 4545u16 , id : 43792u16 , name : "WL60010 Wireless LAN MAC" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 43793u16 , name : "WL60040 Multimode Wireles LAN MAC" , subsystems : & [SubSystem { subvendor : 4545u16 , subdevice : 43794u16 , name : "WaveLAN 11abg Cardbus card (Model 1102)" } , SubSystem { subvendor : 4545u16 , subdevice : 43795u16 , name : "WaveLAN 11abg MiniPCI card (Model 0512)" } , SubSystem { subvendor : 4545u16 , subdevice : 43797u16 , name : "WaveLAN 11abg Cardbus card (Model 1106)" } , SubSystem { subvendor : 4545u16 , subdevice : 43798u16 , name : "WaveLAN 11abg MiniPCI card (Model 0516)" }] } , Device { vendor_id : 4545u16 , id : 43808u16 , name : "ORiNOCO PCI Adapter" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 43809u16 , name : "Agere Wireless PCI Adapter" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 43824u16 , name : "Hermes2 Mini-PCI WaveLAN a/b/g" , subsystems : & [SubSystem { subvendor : 5325u16 , subdevice : 8210u16 , name : "Hermes2 Mini-PCI WaveLAN a/b/g" }] } , Device { vendor_id : 4545u16 , id : 60672u16 , name : "ET-131x PCI-E Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4545u16 , id : 60673u16 , name : "ET-131x PCI-E Ethernet Controller" , subsystems : & [] }] }),
        (12489, Vendor { id : 12489u16 , name : "Luxvisions Innovation Technology Ltd." , devices : & [] }),
        (18755, Vendor { id : 18755u16 , name : "Growth Networks" , devices : & [] }),
        (13498, Vendor { id : 13498u16 , name : "Ice Lake-LP PCI Express Root Port #3" , devices : & [] }),
        (4271, Vendor { id : 4271u16 , name : "Micro Computer Systems Inc" , devices : & [] }),
        (6557, Vendor { id : 6557u16 , name : "Xsigo Systems" , devices : & [Device { vendor_id : 6557u16 , id : 33289u16 , name : "Virtual NIC Device" , subsystems : & [] } , Device { vendor_id : 6557u16 , id : 35082u16 , name : "Virtual HBA Device" , subsystems : & [] }] }),
        (4376, Vendor { id : 4376u16 , name : "Berg Electronics" , devices : & [] }),
        (7756, Vendor { id : 7756u16 , name : "GSI Technology" , devices : & [Device { vendor_id : 7756u16 , id : 16u16 , name : "Gemini-I" , subsystems : & [SubSystem { subvendor : 7756u16 , subdevice : 16u16 , name : "Leda-E" } , SubSystem { subvendor : 7756u16 , subdevice : 288u16 , name : "SE120" }] } , Device { vendor_id : 7756u16 , id : 32u16 , name : "Gemini-II" , subsystems : & [SubSystem { subvendor : 7756u16 , subdevice : 17u16 , name : "Leda-E2 rev 1.1" } , SubSystem { subvendor : 7756u16 , subdevice : 32u16 , name : "G2-Edge" }] }] }),
        (4666, Vendor { id : 4666u16 , name : "Visicom Laboratories, Inc." , devices : & [] }),
        (4677, Vendor { id : 4677u16 , name : "A.P.D., S.A." , devices : & [] }),
        (5293, Vendor { id : 5293u16 , name : "Time Space Radio AB" , devices : & [] }),
        (7525, Vendor { id : 7525u16 , name : "Imagine Communications Corp." , devices : & [Device { vendor_id : 7525u16 , id : 1246u16 , name : "Taurus/McKinley" , subsystems : & [] }] }),
        (4889, Vendor { id : 4889u16 , name : "Fortemedia, Inc" , devices : & [Device { vendor_id : 4889u16 , id : 2049u16 , name : "Xwave QS3000A [FM801]" , subsystems : & [SubSystem { subvendor : 4889u16 , subdevice : 4889u16 , name : "FM801 PCI Audio" }] } , Device { vendor_id : 4889u16 , id : 2050u16 , name : "Xwave QS3000A [FM801 game port]" , subsystems : & [SubSystem { subvendor : 4889u16 , subdevice : 4889u16 , name : "FM801 PCI Joystick" }] } , Device { vendor_id : 4889u16 , id : 4096u16 , name : "FM801 PCI Audio" , subsystems : & [] } , Device { vendor_id : 4889u16 , id : 4097u16 , name : "FM801 PCI Joystick" , subsystems : & [] }] }),
        (4128, Vendor { id : 4128u16 , name : "Hitachi Computer Products" , devices : & [] }),
        (8157, Vendor { id : 8157u16 , name : "Wuqi Microelectronics Co., Ltd." , devices : & [Device { vendor_id : 8157u16 , id : 1u16 , name : "WQ9201 802.11ax PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 8157u16 , id : 4097u16 , name : "WQ9301 802.11ax PCIe Wireless Access Points" , subsystems : & [] }] }),
        (4726, Vendor { id : 4726u16 , name : "Switched Network Technologies, Inc." , devices : & [] }),
        (45968, Vendor { id : 45968u16 , name : "TD3-X3+" , devices : & [] }),
        (7263, Vendor { id : 7263u16 , name : "Beijing Memblaze Technology Co. Ltd." , devices : & [Device { vendor_id : 7263u16 , id : 13u16 , name : "PBlaze5 520/526" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 544u16 , name : "NVMe SSD PBlaze5 520 1920G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 545u16 , name : "NVMe SSD PBlaze5 520 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 560u16 , name : "NVMe SSD PBlaze5 520 3840G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 561u16 , name : "NVMe SSD PBlaze5 520 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 16928u16 , name : "NVMe SSD PBlaze5 526 1600G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 16929u16 , name : "NVMe SSD PBlaze5 526 1600G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 16944u16 , name : "NVMe SSD PBlaze5 526 3200G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 16945u16 , name : "NVMe SSD PBlaze5 526 3200G 2.5\" U.2" }] } , Device { vendor_id : 7263u16 , id : 14u16 , name : "PBlaze6 6530/6531/6541/6630" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 2848u16 , name : "NVMe SSD PBlaze6 6530 1920G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 2849u16 , name : "NVMe SSD PBlaze6 6530 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2853u16 , name : "NVMe SSD PBlaze6 6530 1920G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 2855u16 , name : "NVMe SSD PBlaze6 6630 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2864u16 , name : "NVMe SSD PBlaze6 6530 3840G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 2865u16 , name : "NVMe SSD PBlaze6 6530 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2869u16 , name : "NVMe SSD PBlaze6 6530 3840G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 2871u16 , name : "NVMe SSD PBlaze6 6630 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2880u16 , name : "NVMe SSD PBlaze6 6530 7680G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 2881u16 , name : "NVMe SSD PBlaze6 6530 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2887u16 , name : "NVMe SSD PBlaze6 6630 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 4896u16 , name : "NVMe SSD PBlaze6 6531 1920G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 4897u16 , name : "NVMe SSD PBlaze6 6531 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 4912u16 , name : "NVMe SSD PBlaze6 6531 3840G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 4913u16 , name : "NVMe SSD PBlaze6 6531 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 4928u16 , name : "NVMe SSD PBlaze6 6531 7680G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 4929u16 , name : "NVMe SSD PBlaze6 6531 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5153u16 , name : "NVMe SSD PBlaze6 6541 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5159u16 , name : "NVMe SSD PBlaze6 6641 1920G 2.5\" U.2(dual port)" } , SubSystem { subvendor : 7263u16 , subdevice : 5169u16 , name : "NVMe SSD PBlaze6 6541 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5175u16 , name : "NVMe SSD PBlaze6 6641 3840G 2.5\" U.2(dual port)" } , SubSystem { subvendor : 7263u16 , subdevice : 5185u16 , name : "NVMe SSD PBlaze6 6541 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5191u16 , name : "NVMe SSD PBlaze6 6641 7680G 2.5\" U.2(dual port)" } , SubSystem { subvendor : 7263u16 , subdevice : 19232u16 , name : "NVMe SSD PBlaze6 6536 1600G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 19233u16 , name : "NVMe SSD PBlaze6 6536 1600G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19237u16 , name : "NVMe SSD PBlaze6 6536 1600G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 19239u16 , name : "NVMe SSD PBlaze6 6636 1600G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19248u16 , name : "NVMe SSD PBlaze6 6536 3200G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 19249u16 , name : "NVMe SSD PBlaze6 6536 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19253u16 , name : "NVMe SSD PBlaze6 6536 3200G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 19255u16 , name : "NVMe SSD PBlaze6 6636 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19264u16 , name : "NVMe SSD PBlaze6 6536 6400G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 19265u16 , name : "NVMe SSD PBlaze6 6536 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19271u16 , name : "NVMe SSD PBlaze6 6636 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21280u16 , name : "NVMe SSD PBlaze6 6537 1600G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21281u16 , name : "NVMe SSD PBlaze6 6537 1600G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21296u16 , name : "NVMe SSD PBlaze6 6537 3200G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21297u16 , name : "NVMe SSD PBlaze6 6537 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21312u16 , name : "NVMe SSD PBlaze6 6537 6400G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21313u16 , name : "NVMe SSD PBlaze6 6537 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21537u16 , name : "NVMe SSD PBlaze6 6547 1600G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21543u16 , name : "NVMe SSD PBlaze6 6647 1600G 2.5\" U.2(dual port)" } , SubSystem { subvendor : 7263u16 , subdevice : 21553u16 , name : "NVMe SSD PBlaze6 6547 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21559u16 , name : "NVMe SSD PBlaze6 6647 3200G 2.5\" U.2(dual port)" } , SubSystem { subvendor : 7263u16 , subdevice : 21569u16 , name : "NVMe SSD PBlaze6 6547 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21575u16 , name : "NVMe SSD PBlaze6 6647 6400G 2.5\" U.2(dual port)" }] } , Device { vendor_id : 7263u16 , id : 39u16 , name : "PBlaze7 7A40/7A46 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 5153u16 , name : "NVMe SSD PBlaze7 7A40 1920G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5169u16 , name : "NVMe SSD PBlaze7 7A40 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5185u16 , name : "NVMe SSD PBlaze7 7A40 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5681u16 , name : "NVMe SSD PBlaze7 7A40 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5697u16 , name : "NVMe SSD PBlaze7 7A40 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5713u16 , name : "NVMe SSD PBlaze7 7A40 15360G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5729u16 , name : "NVMe SSD PBlaze7 7A40 30720G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5969u16 , name : "NVMe SSD PBlaze7 7A40 Ocean 15360G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5985u16 , name : "NVMe SSD PBlaze7 7A40 Ocean 30720G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 6001u16 , name : "NVMe SSD PBlaze7 7A40 Ocean 61440G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 6017u16 , name : "NVMe SSD PBlaze7 7A40 Ocean 122880G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21553u16 , name : "NVMe SSD PBlaze7 7A46 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21569u16 , name : "NVMe SSD PBlaze7 7A46 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 22065u16 , name : "NVMe SSD PBlaze7 7A46 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 22081u16 , name : "NVMe SSD PBlaze7 7A46 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 22097u16 , name : "NVMe SSD PBlaze7 7A46 12800G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 22113u16 , name : "NVMe SSD PBlaze7 7A46 25600G 2.5\" U.2" }] } , Device { vendor_id : 7263u16 , id : 61u16 , name : "PBlaze5 920/926" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 2608u16 , name : "NVMe SSD PBlaze5 920 3840G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 2609u16 , name : "NVMe SSD PBlaze5 920 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2624u16 , name : "NVMe SSD PBlaze5 920 7680G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 2625u16 , name : "NVMe SSD PBlaze5 920 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 18992u16 , name : "NVMe SSD PBlaze5 926 3200G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 18993u16 , name : "NVMe SSD PBlaze5 926 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19008u16 , name : "NVMe SSD PBlaze5 926 6400G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 19009u16 , name : "NVMe SSD PBlaze5 926 6400G 2.5\" U.2" }] } , Device { vendor_id : 7263u16 , id : 62u16 , name : "PBlaze6 6920/6930" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 2609u16 , name : "NVMe SSD PBlaze6 6920 3840GB 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2625u16 , name : "NVMe SSD PBlaze6 6920 7680GB 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 2865u16 , name : "NVMe SSD PBlaze6 6930 3840GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 2881u16 , name : "NVMe SSD PBlaze6 6930 7680GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 2897u16 , name : "NVMe SSD PBlaze6 6930 15360GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 2913u16 , name : "NVMe SSD PBlaze6 6930 30720GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 18993u16 , name : "NVMe SSD PBlaze6 6920 3200GB 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19009u16 , name : "NVMe SSD PBlaze6 6920 6400GB 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19265u16 , name : "NVMe SSD PBlaze6 6936 6400GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 19281u16 , name : "NVMe SSD PBlaze6 6936 12800GB 2.5\" U.3" } , SubSystem { subvendor : 7263u16 , subdevice : 19297u16 , name : "NVMe SSD PBlaze6 6936 25600GB 2.5\" U.3" }] } , Device { vendor_id : 7263u16 , id : 63u16 , name : "PBlaze7 7940/7946 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7263u16 , subdevice : 1073u16 , name : "NVMe SSD PBlaze7 7940 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 1089u16 , name : "NVMe SSD PBlaze7 7940 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 1105u16 , name : "NVMe SSD PBlaze7 7940 15360G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 3121u16 , name : "NVMe SSD PBlaze7 7940 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 3137u16 , name : "NVMe SSD PBlaze7 7940 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 3153u16 , name : "NVMe SSD PBlaze7 7940 15360G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 3169u16 , name : "NVMe SSD PBlaze7 7940 30720G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5168u16 , name : "NVMe SSD PBlaze7 7940 3840G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 5169u16 , name : "NVMe SSD PBlaze7 7940 3840G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5173u16 , name : "NVMe SSD PBlaze7 7940 3840G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 5184u16 , name : "NVMe SSD PBlaze7 7940 7680G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 5185u16 , name : "NVMe SSD PBlaze7 7940 7680G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 5189u16 , name : "NVMe SSD PBlaze7 7940 7680G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 5200u16 , name : "NVMe SSD PBlaze7 7940 15360G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 5201u16 , name : "NVMe SSD PBlaze7 7940 15360G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 17457u16 , name : "NVMe SSD PBlaze7 7946 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 17473u16 , name : "NVMe SSD PBlaze7 7946 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 17489u16 , name : "NVMe SSD PBlaze7 7946 12800G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19505u16 , name : "NVMe SSD PBlaze7 7946 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19521u16 , name : "NVMe SSD PBlaze7 7946 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 19537u16 , name : "NVMe SSD PBlaze7 7946 12800G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21552u16 , name : "NVMe SSD PBlaze7 7946 3200G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21553u16 , name : "NVMe SSD PBlaze7 7946 3200G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21557u16 , name : "NVMe SSD PBlaze7 7946 3200G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 21568u16 , name : "NVMe SSD PBlaze7 7946 6400G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21569u16 , name : "NVMe SSD PBlaze7 7946 6400G 2.5\" U.2" } , SubSystem { subvendor : 7263u16 , subdevice : 21573u16 , name : "NVMe SSD PBlaze7 7946 6400G E1.S" } , SubSystem { subvendor : 7263u16 , subdevice : 21584u16 , name : "NVMe SSD PBlaze7 7946 12800G AIC" } , SubSystem { subvendor : 7263u16 , subdevice : 21585u16 , name : "NVMe SSD PBlaze7 7946 12800G 2.5\" U.2" } , SubSystem { subvendor : 7840u16 , subdevice : 8448u16 , name : "TP2510 Series U.2 NVMe Datacenter SSD(3.84TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8449u16 , name : "TP2511 Series U.2 NVMe Datacenter SSD(3.84TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8464u16 , name : "TP2510 Series E3.S NVMe Datacenter SSD(3.84TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8465u16 , name : "TP2511 Series E3.S NVMe Datacenter SSD(3.84TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8704u16 , name : "TP2510 Series U.2 NVMe Datacenter SSD(7.68TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8705u16 , name : "TP2511 Series U.2 NVMe Datacenter SSD(7.68TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8706u16 , name : "TP2310 Series U.2 NVMe Datacenter SSD(7.68TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8720u16 , name : "TP2510 Series E3.S NVMe Datacenter SSD(7.68TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8721u16 , name : "TP2511 Series E3.S NVMe Datacenter SSD(7.68TB)" } , SubSystem { subvendor : 7840u16 , subdevice : 8961u16 , name : "TP2511 Series U.2 NVMe Datacenter SSD (15.36TB)" }] } , Device { vendor_id : 7263u16 , id : 1344u16 , name : "PBlaze4 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7263u16 , id : 1360u16 , name : "PBlaze5 700/900" , subsystems : & [] } , Device { vendor_id : 7263u16 , id : 1365u16 , name : "PBlaze5 510/516" , subsystems : & [] } , Device { vendor_id : 7263u16 , id : 1367u16 , name : "PBlaze5 910/916" , subsystems : & [] }] }),
        (5626, Vendor { id : 5626u16 , name : "J.P. Axzam Corp" , devices : & [] }),
        (7674, Vendor { id : 7674u16 , name : "Astera Labs, Inc." , devices : & [Device { vendor_id : 7674u16 , id : 482u16 , name : "CXL 2.0 Memory Controller CM5xxxx [Leo]" , subsystems : & [] } , Device { vendor_id : 7674u16 , id : 1472u16 , name : "PCIe 6 Fabric Switch PF6xxx [Scorpio]" , subsystems : & [] }] }),
        (5078, Vendor { id : 5078u16 , name : "K.I. Technology Co Ltd" , devices : & [] }),
        (5088, Vendor { id : 5088u16 , name : "GVC Corporation" , devices : & [] }),
        (9999, Vendor { id : 9999u16 , name : "Chaintech Computer Co. Ltd" , devices : & [] }),
        (5245, Vendor { id : 5245u16 , name : "Interworks Computer Products" , devices : & [] }),
        (4351, Vendor { id : 4351u16 , name : "NCube" , devices : & [] }),
        (4265, Vendor { id : 4265u16 , name : "Silicon Graphics Intl. Corp." , devices : & [Device { vendor_id : 4265u16 , id : 1u16 , name : "Crosstalk to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 2u16 , name : "Linc I/O controller" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 3u16 , name : "IOC3 I/O controller" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4u16 , name : "O2 MACE" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 5u16 , name : "RAD Audio" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 6u16 , name : "HPCEX" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 7u16 , name : "RPCEX" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 8u16 , name : "DiVO VIP" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 9u16 , name : "AceNIC Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4265u16 , subdevice : 32770u16 , name : "AceNIC Gigabit Ethernet" }] } , Device { vendor_id : 4265u16 , id : 16u16 , name : "AMP Video I/O" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 17u16 , name : "GRIP" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 18u16 , name : "SGH PSHAC GSN" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 520u16 , name : "SSIM1 SAS Adapter" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4097u16 , name : "Magic Carpet" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4098u16 , name : "Lithium" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4099u16 , name : "Dual JPEG 1" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4100u16 , name : "Dual JPEG 2" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4101u16 , name : "Dual JPEG 3" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4102u16 , name : "Dual JPEG 4" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4103u16 , name : "Dual JPEG 5" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4104u16 , name : "Cesium" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 4106u16 , name : "IOC4 I/O controller" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 5380u16 , name : "SSIM1 Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 8193u16 , name : "Fibre Channel" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 8194u16 , name : "ASDE" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 16385u16 , name : "TIO-CE PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 16386u16 , name : "TIO-CE PCI Express Port" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 32769u16 , name : "O2 1394" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 32770u16 , name : "G-net NT" , subsystems : & [] } , Device { vendor_id : 4265u16 , id : 32811u16 , name : "REACT external interrupt controller" , subsystems : & [] }] }),
        (5005, Vendor { id : 5005u16 , name : "Future Techno Designs pte Ltd" , devices : & [] }),
        (12592, Vendor { id : 12592u16 , name : "AUDIOTRAK" , devices : & [] }),
        (4314, Vendor { id : 4314u16 , name : "Compaq IPG-Austin" , devices : & [Device { vendor_id : 4314u16 , id : 1288u16 , name : "TC4048 Token Ring 4/16" , subsystems : & [] } , Device { vendor_id : 4314u16 , id : 13200u16 , name : "Tl3c3x9" , subsystems : & [] }] }),
        (4746, Vendor { id : 4746u16 , name : "Asante Technologies, Inc." , devices : & [] }),
        (19538, Vendor { id : 19538u16 , name : "LR-LINK" , devices : & [Device { vendor_id : 19538u16 , id : 4097u16 , name : "Smart Network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 40968u16 , name : "LREG1008PT Single-port 1Gb Smart Ethernet Network Adapter" }] } , Device { vendor_id : 19538u16 , id : 4098u16 , name : "Smart Network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 40966u16 , name : "LREG1006PT Single-port 1.2Gb Network Security Isolation Adapter" }] } , Device { vendor_id : 19538u16 , id : 4099u16 , name : "Smart Network Adapter" , subsystems : & [] } , Device { vendor_id : 19538u16 , id : 4100u16 , name : "Smart Network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 45072u16 , name : "LREG1010PF Single-port 10Gb FPGA Network Security Isolation Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 45073u16 , name : "LREG1011PF Dual-port 10Gb FPGA Network Security Isolation Adapter" }] } , Device { vendor_id : 19538u16 , id : 8193u16 , name : "100G Ethernet network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 6178u16 , name : "LRES1822PF Dual-Port 100G Ethernet network Adapter" }] } , Device { vendor_id : 19538u16 , id : 8194u16 , name : "100G Ethernet network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 6178u16 , name : "LRES1822PF Ethernet Network Virtual Function" }] } , Device { vendor_id : 19538u16 , id : 8195u16 , name : "25G Ethernet network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 6194u16 , name : "LRES1832PF Dual-Port 25G Ethernet Network Adapter" }] } , Device { vendor_id : 19538u16 , id : 8196u16 , name : "25G Ethernet Network Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 6194u16 , name : "LRES1832PF Ethernet Network Virtual Function" }] }] }),
        (5372, Vendor { id : 5372u16 , name : "Quadrics Ltd" , devices : & [Device { vendor_id : 5372u16 , id : 0u16 , name : "QsNet Elan3 Network Adapter" , subsystems : & [] } , Device { vendor_id : 5372u16 , id : 1u16 , name : "QsNetII Elan4 Network Adapter" , subsystems : & [] } , Device { vendor_id : 5372u16 , id : 2u16 , name : "QsNetIII Elan5 Network Adapter" , subsystems : & [] }] }),
        (5121, Vendor { id : 5121u16 , name : "CR-Systems A/S" , devices : & [] }),
        (4869, Vendor { id : 4869u16 , name : "Netphone, Inc" , devices : & [] }),
        (6509, Vendor { id : 6509u16 , name : "Club-3D BV" , devices : & [] }),
        (5539, Vendor { id : 5539u16 , name : "Italtel" , devices : & [] }),
        (4724, Vendor { id : 4724u16 , name : "Ensoniq" , devices : & [Device { vendor_id : 4724u16 , id : 4465u16 , name : "ES1373 / Creative Labs CT5803 [AudioPCI]" , subsystems : & [] } , Device { vendor_id : 4724u16 , id : 4977u16 , name : "ES1371/ES1373 / Creative Labs CT2518" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 36u16 , name : "AudioPCI on Motherboard Compaq Deskpro" } , SubSystem { subvendor : 3601u16 , subdevice : 45479u16 , name : "ES1371, ES1373 AudioPCI" } , SubSystem { subvendor : 4147u16 , subdevice : 32940u16 , name : "ES1371, ES1373 AudioPCI" } , SubSystem { subvendor : 4162u16 , subdevice : 6228u16 , name : "Tazer" } , SubSystem { subvendor : 4219u16 , subdevice : 32852u16 , name : "Tabor2" } , SubSystem { subvendor : 4724u16 , subdevice : 4977u16 , name : "Audio PCI 64V/128/5200 / Creative CT4810/CT5803/CT5806 [Sound Blaster PCI]" } , SubSystem { subvendor : 4724u16 , subdevice : 32769u16 , name : "CT4751 board" } , SubSystem { subvendor : 5218u16 , subdevice : 25712u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6147 1.1A" } , SubSystem { subvendor : 5218u16 , subdevice : 25952u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6156 1.10" } , SubSystem { subvendor : 5218u16 , subdevice : 26160u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6163BX 1.0A" } , SubSystem { subvendor : 5218u16 , subdevice : 26161u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6163VIA 1.0A" } , SubSystem { subvendor : 5218u16 , subdevice : 26162u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6163BX 2.0A" } , SubSystem { subvendor : 5218u16 , subdevice : 26163u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6163VIA 2.0A" } , SubSystem { subvendor : 5218u16 , subdevice : 26656u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6182 1.00" } , SubSystem { subvendor : 5218u16 , subdevice : 26658u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6182 1.00A" } , SubSystem { subvendor : 5218u16 , subdevice : 26672u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6183 1.00" } , SubSystem { subvendor : 5218u16 , subdevice : 26752u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6188 1.00" } , SubSystem { subvendor : 5218u16 , subdevice : 26880u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6190 1.00" } , SubSystem { subvendor : 5218u16 , subdevice : 26896u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6191" } , SubSystem { subvendor : 5218u16 , subdevice : 26928u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6193" } , SubSystem { subvendor : 5218u16 , subdevice : 27024u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6199BX 2.0A" } , SubSystem { subvendor : 5218u16 , subdevice : 27025u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MS-6199VIA 2.0A" } , SubSystem { subvendor : 5284u16 , subdevice : 8311u16 , name : "ES1371, ES1373 AudioPCI On Motherboard KR639" } , SubSystem { subvendor : 5284u16 , subdevice : 8453u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MR800" } , SubSystem { subvendor : 5284u16 , subdevice : 8455u16 , name : "ES1371, ES1373 AudioPCI On Motherboard MR801" } , SubSystem { subvendor : 5284u16 , subdevice : 8562u16 , name : "ES1371, ES1373 AudioPCI On Motherboard DR739" } , SubSystem { subvendor : 5385u16 , subdevice : 39170u16 , name : "ES1371, ES1373 AudioPCI On Motherboard KW11" } , SubSystem { subvendor : 5385u16 , subdevice : 39171u16 , name : "ES1371, ES1373 AudioPCI On Motherboard KW31" } , SubSystem { subvendor : 5385u16 , subdevice : 39172u16 , name : "ES1371, ES1373 AudioPCI On Motherboard KA11" } , SubSystem { subvendor : 5385u16 , subdevice : 39173u16 , name : "ES1371, ES1373 AudioPCI On Motherboard KC13" } , SubSystem { subvendor : 5421u16 , subdevice : 34817u16 , name : "ES1371, ES1373 AudioPCI On Motherboard CP810E" } , SubSystem { subvendor : 5421u16 , subdevice : 34818u16 , name : "ES1371, ES1373 AudioPCI On Motherboard CP810" } , SubSystem { subvendor : 5421u16 , subdevice : 34819u16 , name : "ES1371, ES1373 AudioPCI On Motherboard P3810E" } , SubSystem { subvendor : 5421u16 , subdevice : 34820u16 , name : "ES1371, ES1373 AudioPCI On Motherboard P3810-S" } , SubSystem { subvendor : 5421u16 , subdevice : 34821u16 , name : "ES1371, ES1373 AudioPCI On Motherboard P3820-S" } , SubSystem { subvendor : 9999u16 , subdevice : 8193u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6CTR" } , SubSystem { subvendor : 9999u16 , subdevice : 8704u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6WTX" } , SubSystem { subvendor : 9999u16 , subdevice : 12288u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6WSV" } , SubSystem { subvendor : 9999u16 , subdevice : 12544u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6WIV2" } , SubSystem { subvendor : 9999u16 , subdevice : 12546u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6WIV" } , SubSystem { subvendor : 9999u16 , subdevice : 28768u16 , name : "ES1371, ES1373 AudioPCI On Motherboard 6ASA2" } , SubSystem { subvendor : 32902u16 , subdevice : 16969u16 , name : "ES1371, ES1373 AudioPCI On Motherboard BI440ZX" } , SubSystem { subvendor : 32902u16 , subdevice : 16972u16 , name : "ES1371, ES1373 AudioPCI On Motherboard BL440ZX" } , SubSystem { subvendor : 32902u16 , subdevice : 16986u16 , name : "ES1371, ES1373 AudioPCI On Motherboard BZ440ZX" } , SubSystem { subvendor : 32902u16 , subdevice : 17217u16 , name : "ES1371, ES1373 AudioPCI On Motherboard Cayman" } , SubSystem { subvendor : 32902u16 , subdevice : 17219u16 , name : "ES1371, ES1373 AudioPCI On Motherboard Cape Cod" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "D815EEA Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17993u16 , name : "ES1371, ES1373 AudioPCI On Motherboard Fire Island" } , SubSystem { subvendor : 32902u16 , subdevice : 17994u16 , name : "ES1371, ES1373 AudioPCI On Motherboard FJ440ZX" } , SubSystem { subvendor : 32902u16 , subdevice : 19791u16 , name : "ES1371, ES1373 AudioPCI On Motherboard Montreal" } , SubSystem { subvendor : 32902u16 , subdevice : 20291u16 , name : "ES1371, ES1373 AudioPCI On Motherboard OC440LX" } , SubSystem { subvendor : 32902u16 , subdevice : 21059u16 , name : "ES1371, ES1373 AudioPCI On Motherboard RC440BX" } , SubSystem { subvendor : 32902u16 , subdevice : 21330u16 , name : "ES1371, ES1373 AudioPCI On Motherboard SunRiver" } , SubSystem { subvendor : 32902u16 , subdevice : 22083u16 , name : "ES1371, ES1373 AudioPCI On Motherboard Vancouver" } , SubSystem { subvendor : 32902u16 , subdevice : 22355u16 , name : "ES1371, ES1373 AudioPCI On Motherboard WS440BX" }] } , Device { vendor_id : 4724u16 , id : 20480u16 , name : "ES1370 [AudioPCI]" , subsystems : & [] } , Device { vendor_id : 4724u16 , id : 22656u16 , name : "5880B / Creative Labs CT5880" , subsystems : & [SubSystem { subvendor : 4724u16 , subdevice : 8192u16 , name : "Creative CT4810 [Sound Blaster AudioPCI 128]" } , SubSystem { subvendor : 4724u16 , subdevice : 8195u16 , name : "Creative SoundBlaster AudioPCI 128" } , SubSystem { subvendor : 4724u16 , subdevice : 22656u16 , name : "Creative CT4750 [Sound Blaster PCI 128]" } , SubSystem { subvendor : 4724u16 , subdevice : 32769u16 , name : "Creative CT4750 [Sound Blaster 16 PCI/PCI 128/4.1 Digital]" } , SubSystem { subvendor : 5208u16 , subdevice : 40960u16 , name : "5880 AudioPCI On Motherboard 6OXET" } , SubSystem { subvendor : 5218u16 , subdevice : 26752u16 , name : "5880 AudioPCI On Motherboard MS-6188 1.00" } , SubSystem { subvendor : 9999u16 , subdevice : 8193u16 , name : "5880 AudioPCI On Motherboard 6CTR" } , SubSystem { subvendor : 9999u16 , subdevice : 8704u16 , name : "5880 AudioPCI On Motherboard 6WTX" } , SubSystem { subvendor : 9999u16 , subdevice : 28736u16 , name : "5880 AudioPCI On Motherboard 6ATA4" }] } , Device { vendor_id : 4724u16 , id : 32769u16 , name : "CT5880 [AudioPCI]" , subsystems : & [] } , Device { vendor_id : 4724u16 , id : 32770u16 , name : "5880A [AudioPCI]" , subsystems : & [] }] }),
        (8191, Vendor { id : 8191u16 , name : "DENSO Corporation" , devices : & [Device { vendor_id : 8191u16 , id : 32768u16 , name : "MCST PCIe controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32776u16 , name : "MCST USB 1.1 OHCI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32778u16 , name : "MCST HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32779u16 , name : "MCST PATA IDE Host Controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32782u16 , name : "MCST System Power Managment Controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32784u16 , name : "MCST PCIe 2.0 x4 Root Port" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32785u16 , name : "MCST PCIe 2.0 x16 Root Port" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32788u16 , name : "MCST MPV Controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32791u16 , name : "MCST PCIe controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32792u16 , name : "MCST I2C/SPI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32793u16 , name : "MCST UART controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32794u16 , name : "MCST SATA 3 AHCI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32798u16 , name : "MCST USB 2.0 EHCI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32805u16 , name : "MCST MPV Controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32806u16 , name : "MCST 10 Gigabit Ethernet MAC controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32807u16 , name : "MCST Gigabit Ethernet MAC controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32808u16 , name : "MCST I2C/SPI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32809u16 , name : "MCST USB 3 XHCI controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32810u16 , name : "MCST IMG GX6650 3D GPU" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32811u16 , name : "MCST Video encoder E5810" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32812u16 , name : "MCST Video decoder D5520" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32813u16 , name : "MCST VP9 Video encoder" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32814u16 , name : "MCST VP9 Video decoder" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32815u16 , name : "MCST UART controller" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32817u16 , name : "MCST MGA 2.5 GPU" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32818u16 , name : "MCST PCIe 3.0 x16 Root Complex" , subsystems : & [] } , Device { vendor_id : 8191u16 , id : 32819u16 , name : "MCST PCIe 3.0 x8 Root Complex" , subsystems : & [] }] }),
        (43529, Vendor { id : 43529u16 , name : "iTuner" , devices : & [] }),
        (5016, Vendor { id : 5016u16 , name : "Clarion co. Ltd" , devices : & [] }),
        (4864, Vendor { id : 4864u16 , name : "Harman International Industries Inc" , devices : & [] }),
        (7938, Vendor { id : 7938u16 , name : "Beijing Dayu Technology" , devices : & [] }),
        (7908, Vendor { id : 7908u16 , name : "PETAIO INC" , devices : & [Device { vendor_id : 7908u16 , id : 4480u16 , name : "PETA8118 NVMe SSD Series" , subsystems : & [SubSystem { subvendor : 7908u16 , subdevice : 21u16 , name : "NVMe SSD U.2 1.92TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 22u16 , name : "NVMe SSD U.2 3.84TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 23u16 , name : "NVMe SSD U.2 7.68TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 37u16 , name : "NVMe SSD U.2 1.6TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 38u16 , name : "NVMe SSD U.2 3.2TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 39u16 , name : "NVMe SSD U.2 6.4TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 277u16 , name : "NVMe SSD U.2 1.92TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 278u16 , name : "NVMe SSD U.2 3.84TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 279u16 , name : "NVMe SSD U.2 7.68TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 293u16 , name : "NVMe SSD U.2 1.6TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 294u16 , name : "NVMe SSD U.2 3.2TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 295u16 , name : "NVMe SSD U.2 6.4TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 533u16 , name : "NVMe SSD U.2 1.92TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 534u16 , name : "NVMe SSD U.2 3.84TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 535u16 , name : "NVMe SSD U.2 7.68TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 549u16 , name : "NVMe SSD U.2 1.6TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 550u16 , name : "NVMe SSD U.2 3.2TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 551u16 , name : "NVMe SSD U.2 6.4TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 789u16 , name : "NVMe SSD U.2 1.92TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 790u16 , name : "NVMe SSD U.2 3.84TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 791u16 , name : "NVMe SSD U.2 7.68TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 805u16 , name : "NVMe SSD U.2 1.6TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 806u16 , name : "NVMe SSD U.2 3.2TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 807u16 , name : "NVMe SSD U.2 6.4TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1045u16 , name : "NVMe SSD U.2 1.92TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1046u16 , name : "NVMe SSD U.2 3.84TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1047u16 , name : "NVMe SSD U.2 7.68TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1061u16 , name : "NVMe SSD U.2 1.6TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1062u16 , name : "NVMe SSD U.2 3.2TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1063u16 , name : "NVMe SSD U.2 6.4TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1301u16 , name : "NVMe SSD U.2 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1302u16 , name : "NVMe SSD U.2 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1303u16 , name : "NVMe SSD U.2 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1317u16 , name : "NVMe SSD U.2 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1318u16 , name : "NVMe SSD U.2 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1319u16 , name : "NVMe SSD U.2 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1557u16 , name : "NVMe SSD U.2 1.92TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1558u16 , name : "NVMe SSD U.2 3.84TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1559u16 , name : "NVMe SSD U.2 7.68TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1573u16 , name : "NVMe SSD U.2 1.6TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1574u16 , name : "NVMe SSD U.2 3.2TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1575u16 , name : "NVMe SSD U.2 6.4TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 1813u16 , name : "NVMe SSD U.2 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1814u16 , name : "NVMe SSD U.2 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1815u16 , name : "NVMe SSD U.2 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1816u16 , name : "NVMe SSD U.2 15.36TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1829u16 , name : "NVMe SSD U.2 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1830u16 , name : "NVMe SSD U.2 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1831u16 , name : "NVMe SSD U.2 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 1832u16 , name : "NVMe SSD U.2 12.8TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2069u16 , name : "NVMe SSD U.2 1.92TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2070u16 , name : "NVMe SSD U.2 3.84TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2071u16 , name : "NVMe SSD U.2 7.68TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2085u16 , name : "NVMe SSD U.2 1.6TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2086u16 , name : "NVMe SSD U.2 3.2TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2087u16 , name : "NVMe SSD U.2 6.4TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 2325u16 , name : "NVMe SSD U.2 1.92TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2326u16 , name : "NVMe SSD U.2 3.84TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2327u16 , name : "NVMe SSD U.2 7.68TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2341u16 , name : "NVMe SSD U.2 1.6TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2342u16 , name : "NVMe SSD U.2 3.2TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2343u16 , name : "NVMe SSD U.2 6.4TB (P8118E2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2581u16 , name : "NVMe SSD U.2 1.92TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2582u16 , name : "NVMe SSD U.2 3.84TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2583u16 , name : "NVMe SSD U.2 7.68TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2597u16 , name : "NVMe SSD U.2 1.6TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2598u16 , name : "NVMe SSD U.2 3.2TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2599u16 , name : "NVMe SSD U.2 6.4TB (P8118H2)" } , SubSystem { subvendor : 7908u16 , subdevice : 2837u16 , name : "NVMe SSD U.2 1.92TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 2838u16 , name : "NVMe SSD U.2 3.84TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 2839u16 , name : "NVMe SSD U.2 7.68TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 2853u16 , name : "NVMe SSD U.2 1.6TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 2854u16 , name : "NVMe SSD U.2 3.2TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 2855u16 , name : "NVMe SSD U.2 6.4TB (P8118H3)" } , SubSystem { subvendor : 7908u16 , subdevice : 12307u16 , name : "NVMe SSD AIC 480GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12308u16 , name : "NVMe SSD AIC 960GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12309u16 , name : "NVMe SSD AIC 1.92TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12310u16 , name : "NVMe SSD AIC 3.84TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12311u16 , name : "NVMe SSD AIC 7.68TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12325u16 , name : "NVMe SSD AIC 1.6TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12326u16 , name : "NVMe SSD AIC 3.2TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12327u16 , name : "NVMe SSD AIC 6.4TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 12563u16 , name : "NVMe SSD AIC 480GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12564u16 , name : "NVMe SSD AIC 960GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12565u16 , name : "NVMe SSD AIC 1.92TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12566u16 , name : "NVMe SSD AIC 3.84TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12567u16 , name : "NVMe SSD AIC 7.68TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12581u16 , name : "NVMe SSD AIC 1.6TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12582u16 , name : "NVMe SSD AIC 3.2TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12583u16 , name : "NVMe SSD AIC 6.4TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 12819u16 , name : "NVMe SSD AIC 480GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12820u16 , name : "NVMe SSD AIC 960GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12821u16 , name : "NVMe SSD AIC 1.92TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12822u16 , name : "NVMe SSD AIC 3.84TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12823u16 , name : "NVMe SSD AIC 7.68TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12837u16 , name : "NVMe SSD AIC 1.6TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12838u16 , name : "NVMe SSD AIC 3.2TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 12839u16 , name : "NVMe SSD AIC 6.4TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 13077u16 , name : "NVMe SSD AIC 1.92TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13078u16 , name : "NVMe SSD AIC 3.84TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13079u16 , name : "NVMe SSD AIC 7.68TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13093u16 , name : "NVMe SSD AIC 1.6TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13094u16 , name : "NVMe SSD AIC 3.2TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13095u16 , name : "NVMe SSD AIC 6.4TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13333u16 , name : "NVMe SSD AIC 1.92TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13334u16 , name : "NVMe SSD AIC 3.84TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13335u16 , name : "NVMe SSD AIC 7.68TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13349u16 , name : "NVMe SSD AIC 1.6TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13350u16 , name : "NVMe SSD AIC 3.2TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13351u16 , name : "NVMe SSD AIC 6.4TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13589u16 , name : "NVMe SSD AIC 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13590u16 , name : "NVMe SSD AIC 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13591u16 , name : "NVMe SSD AIC 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13605u16 , name : "NVMe SSD AIC 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13606u16 , name : "NVMe SSD AIC 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13607u16 , name : "NVMe SSD AIC 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 13845u16 , name : "NVMe SSD AIC 1.92TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13846u16 , name : "NVMe SSD AIC 3.84TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13847u16 , name : "NVMe SSD AIC 7.68TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13861u16 , name : "NVMe SSD AIC 1.6TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13862u16 , name : "NVMe SSD AIC 3.2TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 13863u16 , name : "NVMe SSD AIC 6.4TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 14101u16 , name : "NVMe SSD AIC 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14102u16 , name : "NVMe SSD AIC 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14103u16 , name : "NVMe SSD AIC 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14104u16 , name : "NVMe SSD AIC 15.36TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14117u16 , name : "NVMe SSD AIC 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14118u16 , name : "NVMe SSD AIC 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14119u16 , name : "NVMe SSD AIC 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 14120u16 , name : "NVMe SSD AIC 12.8TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 43981u16 , name : "NVMe SSD U.2" }] } , Device { vendor_id : 7908u16 , id : 4481u16 , name : "PETA8118 NVMe E1S Series" , subsystems : & [SubSystem { subvendor : 7908u16 , subdevice : 8213u16 , name : "NVMe SSD E1.S 1.92TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8214u16 , name : "NVMe SSD E1.S 3.84TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8215u16 , name : "NVMe SSD E1.S 7.68TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8229u16 , name : "NVMe SSD E1.S 1.6TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8230u16 , name : "NVMe SSD E1.S 3.2TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8231u16 , name : "NVMe SSD E1.S 6.4TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 8469u16 , name : "NVMe SSD E1.S 1.92TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8470u16 , name : "NVMe SSD E1.S 3.84TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8471u16 , name : "NVMe SSD E1.S 7.68TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8485u16 , name : "NVMe SSD E1.S 1.6TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8486u16 , name : "NVMe SSD E1.S 3.2TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8487u16 , name : "NVMe SSD E1.S 6.4TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 8725u16 , name : "NVMe SSD E1.S 1.92TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8726u16 , name : "NVMe SSD E1.S 3.84TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8727u16 , name : "NVMe SSD E1.S 7.68TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8741u16 , name : "NVMe SSD E1.S 1.6TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8742u16 , name : "NVMe SSD E1.S 3.2TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8743u16 , name : "NVMe SSD E1.S 6.4TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 8981u16 , name : "NVMe SSD E1.S 1.92TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 8982u16 , name : "NVMe SSD E1.S 3.84TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 8983u16 , name : "NVMe SSD E1.S 7.68TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 8997u16 , name : "NVMe SSD E1.S 1.6TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 8998u16 , name : "NVMe SSD E1.S 3.2TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 8999u16 , name : "NVMe SSD E1.S 6.4TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9237u16 , name : "NVMe SSD E1.S 1.92TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9238u16 , name : "NVMe SSD E1.S 3.84TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9239u16 , name : "NVMe SSD E1.S 7.68TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9253u16 , name : "NVMe SSD E1.S 1.6TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9254u16 , name : "NVMe SSD E1.S 3.2TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9255u16 , name : "NVMe SSD E1.S 6.4TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9493u16 , name : "NVMe SSD E1.S 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9494u16 , name : "NVMe SSD E1.S 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9495u16 , name : "NVMe SSD E1.S 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9509u16 , name : "NVMe SSD E1.S 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9510u16 , name : "NVMe SSD E1.S 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9511u16 , name : "NVMe SSD E1.S 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 9749u16 , name : "NVMe SSD E1.S 1.92TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9750u16 , name : "NVMe SSD E1.S 3.84TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9751u16 , name : "NVMe SSD E1.S 7.68TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9765u16 , name : "NVMe SSD E1.S 1.6TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9766u16 , name : "NVMe SSD E1.S 3.2TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 9767u16 , name : "NVMe SSD E1.S 6.4TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 10005u16 , name : "NVMe SSD E1.S 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10006u16 , name : "NVMe SSD E1.S 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10007u16 , name : "NVMe SSD E1.S 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10008u16 , name : "NVMe SSD E1.S 15.36TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10021u16 , name : "NVMe SSD E1.S 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10022u16 , name : "NVMe SSD E1.S 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10023u16 , name : "NVMe SSD E1.S 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 10024u16 , name : "NVMe SSD E1.S 12.8TB (P8118Z4)" }] } , Device { vendor_id : 7908u16 , id : 4482u16 , name : "PETA8118 NVMe M2 Series" , subsystems : & [SubSystem { subvendor : 7908u16 , subdevice : 4115u16 , name : "NVMe SSD M.2 480GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4116u16 , name : "NVMe SSD M.2 960GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4117u16 , name : "NVMe SSD M.2 1.92TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4118u16 , name : "NVMe SSD M.2 3.84TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4131u16 , name : "NVMe SSD M.2 400GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4132u16 , name : "NVMe SSD M.2 800GB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4133u16 , name : "NVMe SSD M.2 1.6TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4134u16 , name : "NVMe SSD M.2 3.2TB (P8118E)" } , SubSystem { subvendor : 7908u16 , subdevice : 4371u16 , name : "NVMe SSD M.2 480GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4372u16 , name : "NVMe SSD M.2 960GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4373u16 , name : "NVMe SSD M.2 1.92TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4374u16 , name : "NVMe SSD M.2 3.84TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4387u16 , name : "NVMe SSD M.2 400GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4388u16 , name : "NVMe SSD M.2 800GB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4389u16 , name : "NVMe SSD M.2 1.6TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4390u16 , name : "NVMe SSD M.2 3.2TB (P8118Z)" } , SubSystem { subvendor : 7908u16 , subdevice : 4627u16 , name : "NVMe SSD M.2 480GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4628u16 , name : "NVMe SSD M.2 960GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4629u16 , name : "NVMe SSD M.2 1.92TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4630u16 , name : "NVMe SSD M.2 3.84TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4643u16 , name : "NVMe SSD M.2 400GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4644u16 , name : "NVMe SSD M.2 800GB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4645u16 , name : "NVMe SSD M.2 1.6TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4646u16 , name : "NVMe SSD M.2 3.2TB (P8118X)" } , SubSystem { subvendor : 7908u16 , subdevice : 4883u16 , name : "NVMe SSD M.2 480GB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4884u16 , name : "NVMe SSD M.2 960GB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4885u16 , name : "NVMe SSD M.2 1.92TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4886u16 , name : "NVMe SSD M.2 3.84TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4887u16 , name : "NVMe SSD M.2 7.68TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4899u16 , name : "NVMe SSD M.2 400GB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4900u16 , name : "NVMe SSD M.2 800GB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4901u16 , name : "NVMe SSD M.2 1.6TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4902u16 , name : "NVMe SSD M.2 3.2TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 4903u16 , name : "NVMe SSD M.2 6.4TB (P8118E4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5139u16 , name : "NVMe SSD M.2 480GB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5140u16 , name : "NVMe SSD M.2 960GB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5141u16 , name : "NVMe SSD M.2 1.92TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5142u16 , name : "NVMe SSD M.2 3.84TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5143u16 , name : "NVMe SSD M.2 7.68TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5155u16 , name : "NVMe SSD M.2 400GB(P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5156u16 , name : "NVMe SSD M.2 800GB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5157u16 , name : "NVMe SSD M.2 1.6TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5158u16 , name : "NVMe SSD M.2 3.2TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5159u16 , name : "NVMe SSD M.2 6.4TB (P8118Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5395u16 , name : "NVMe SSD M.2 480GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5396u16 , name : "NVMe SSD M.2 960GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5397u16 , name : "NVMe SSD M.2 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5398u16 , name : "NVMe SSD M.2 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5399u16 , name : "NVMe SSD M.2 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5411u16 , name : "NVMe SSD M.2 400GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5412u16 , name : "NVMe SSD M.2 800GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5413u16 , name : "NVMe SSD M.2 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5414u16 , name : "NVMe SSD M.2 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5415u16 , name : "NVMe SSD M.2 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5651u16 , name : "NVMe SSD M.2 480GB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5652u16 , name : "NVMe SSD M.2 960GB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5653u16 , name : "NVMe SSD M.2 1.92TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5654u16 , name : "NVMe SSD M.2 3.84TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5655u16 , name : "NVMe SSD M.2 7.68TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5667u16 , name : "NVMe SSD M.2 400GB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5668u16 , name : "NVMe SSD M.2 800GB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5669u16 , name : "NVMe SSD M.2 1.6TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5670u16 , name : "NVMe SSD M.2 3.2TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5671u16 , name : "NVMe SSD M.2 6.4TB (P8128Z3)" } , SubSystem { subvendor : 7908u16 , subdevice : 5908u16 , name : "NVMe SSD M.2 960GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5909u16 , name : "NVMe SSD M.2 1.92TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5910u16 , name : "NVMe SSD M.2 3.84TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5911u16 , name : "NVMe SSD M.2 7.68TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5924u16 , name : "NVMe SSD M.2 800GB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5925u16 , name : "NVMe SSD M.2 1.6TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5926u16 , name : "NVMe SSD M.2 3.2TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 5927u16 , name : "NVMe SSD M.2 6.4TB (P8118Z4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6164u16 , name : "NVMe SSD M.2 960GB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6165u16 , name : "NVMe SSD M.2 1.92TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6166u16 , name : "NVMe SSD M.2 3.84TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6167u16 , name : "NVMe SSD M.2 7.68TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6180u16 , name : "NVMe SSD M.2 800GB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6181u16 , name : "NVMe SSD M.2 1.6TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6182u16 , name : "NVMe SSD M.2 3.2TB (P8118H4)" } , SubSystem { subvendor : 7908u16 , subdevice : 6183u16 , name : "NVMe SSD M.2 6.4TB (P8118H4)" }] }] }),
        (6988, Vendor { id : 6988u16 , name : "GALAX" , devices : & [] }),
        (4630, Vendor { id : 4630u16 , name : "Purup Prepress A/S" , devices : & [] }),
        (6165, Vendor { id : 6165u16 , name : "Devolo AG" , devices : & [] }),
        (4873, Vendor { id : 4873u16 , name : "AB Semiconductor Ltd" , devices : & [] }),
        (6044, Vendor { id : 6044u16 , name : "Data Patterns" , devices : & [Device { vendor_id : 6044u16 , id : 1367u16 , name : "DP-PCI-557 [PCI 1553B]" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 1382u16 , name : "DP-PCI-566 [Intelligent PCI 1553B]" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 4434u16 , name : "DP-cPCI-1152 (8-channel Isolated ADC Module)" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 20529u16 , name : "DP-CPCI-5031-Synchro Module" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 20754u16 , name : "DP-cPCI-5112 [MM-Carrier]" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 20769u16 , name : "DP-CPCI-5121-IP Carrier" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 21009u16 , name : "DP-CPCI-5211-IP Carrier" , subsystems : & [] } , Device { vendor_id : 6044u16 , id : 22137u16 , name : "AGE Display Module" , subsystems : & [] }] }),
        (5464, Vendor { id : 5464u16 , name : "CLEVO/KAPOK Computer" , devices : & [Device { vendor_id : 5464u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (4667, Vendor { id : 4667u16 , name : "Seeq Technology, Inc." , devices : & [] }),
        (5226, Vendor { id : 5226u16 , name : "Aeroflex" , devices : & [Device { vendor_id : 5226u16 , id : 12304u16 , name : "3010 RF Synthesizer" , subsystems : & [] } , Device { vendor_id : 5226u16 , id : 14865u16 , name : "3011A PXI RF Synthesizer" , subsystems : & [] }] }),
        (45529, Vendor { id : 45529u16 , name : "ATCOM Technology co., LTD." , devices : & [] }),
        (5940, Vendor { id : 5940u16 , name : "Fujitsu Technology Solutions" , devices : & [Device { vendor_id : 5940u16 , id : 4648u16 , name : "iRMC-S5 HTI Device" , subsystems : & [] } , Device { vendor_id : 5940u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (6718, Vendor { id : 6718u16 , name : "Micro-Research Finland Oy" , devices : & [Device { vendor_id : 6718u16 , id : 4908u16 , name : "MTCA Event Receiver 300" , subsystems : & [] } , Device { vendor_id : 6718u16 , id : 5420u16 , name : "CompactPCI Event Receiver 300" , subsystems : & [] } , Device { vendor_id : 6718u16 , id : 5932u16 , name : "PCI Express Event Receiver 300" , subsystems : & [] } , Device { vendor_id : 6718u16 , id : 6444u16 , name : "CompactPCI Event Receiver TG 300" , subsystems : & [] } , Device { vendor_id : 6718u16 , id : 9004u16 , name : "MTCA Event Master 300" , subsystems : & [] } , Device { vendor_id : 6718u16 , id : 9516u16 , name : "CompactPCI Event Generator 300" , subsystems : & [] }] }),
        (5353, Vendor { id : 5353u16 , name : "GARNETS System CO Ltd" , devices : & [] }),
        (7960, Vendor { id : 7960u16 , name : "c-payne GmbH" , devices : & [Device { vendor_id : 7960u16 , id : 1u16 , name : "PCIe gen4 Switch Backplane 5x x16 - 4W" , subsystems : & [] } , Device { vendor_id : 7960u16 , id : 2u16 , name : "PCIe gen4 Switch Backplane 4x x16 - 5W" , subsystems : & [] } , Device { vendor_id : 7960u16 , id : 257u16 , name : "PCIe gen5 MCIO Switch 100-Lane" , subsystems : & [] } , Device { vendor_id : 7960u16 , id : 258u16 , name : "PCIe gen5 MCIO Switch 52-Lane" , subsystems : & [] }] }),
        (7164, Vendor { id : 7164u16 , name : "Duagon AG" , devices : & [] }),
        (4598, Vendor { id : 4598u16 , name : "Compex" , devices : & [Device { vendor_id : 4598u16 , id : 274u16 , name : "ENet100VG4" , subsystems : & [] } , Device { vendor_id : 4598u16 , id : 275u16 , name : "FreedomLine 100" , subsystems : & [] } , Device { vendor_id : 4598u16 , id : 5121u16 , name : "ReadyLink 2000" , subsystems : & [] } , Device { vendor_id : 4598u16 , id : 8209u16 , name : "RL100-ATX 10/100" , subsystems : & [SubSystem { subvendor : 4598u16 , subdevice : 8209u16 , name : "RL100-ATX" }] } , Device { vendor_id : 4598u16 , id : 8705u16 , name : "ReadyLink 100TX (Winbond W89C840)" , subsystems : & [SubSystem { subvendor : 4598u16 , subdevice : 8209u16 , name : "ReadyLink 100TX" }] } , Device { vendor_id : 4598u16 , id : 39041u16 , name : "RL100TX Fast Ethernet" , subsystems : & [] }] }),
        (5377, Vendor { id : 5377u16 , name : "BANKSOFT CANADA Ltd" , devices : & [] }),
        (4900, Vendor { id : 4900u16 , name : "Sphere Communications" , devices : & [] }),
        (4742, Vendor { id : 4742u16 , name : "Mazet GmbH" , devices : & [] }),
        (22609, Vendor { id : 22609u16 , name : "Exacq Technologies" , devices : & [Device { vendor_id : 22609u16 , id : 32776u16 , name : "tDVR8008 8-port video capture card" , subsystems : & [] } , Device { vendor_id : 22609u16 , id : 32790u16 , name : "tDVR8016 16-chan video capture card" , subsystems : & [] } , Device { vendor_id : 22609u16 , id : 32818u16 , name : "tDVR8032 32-chan video capture card" , subsystems : & [] }] }),
        (43526, Vendor { id : 43526u16 , name : "iTuner" , devices : & [] }),
        (4659, Vendor { id : 4659u16 , name : "Bus-Tech, Inc." , devices : & [] }),
        (4122, Vendor { id : 4122u16 , name : "AT&T GIS (NCR)" , devices : & [Device { vendor_id : 4122u16 , id : 5u16 , name : "100VG ethernet" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 7u16 , name : "BYNET BIC4G/2C/2G" , subsystems : & [SubSystem { subvendor : 4122u16 , subdevice : 25u16 , name : "BYNET BIC2C" } , SubSystem { subvendor : 4122u16 , subdevice : 28u16 , name : "BYNET BIC2G" } , SubSystem { subvendor : 4122u16 , subdevice : 31u16 , name : "BYNET BIC4G" }] } , Device { vendor_id : 4122u16 , id : 9u16 , name : "PQS Memory Controller" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 10u16 , name : "BYNET BPCI Adapter" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 11u16 , name : "BYNET 4 Port BYA Switch (BYA4P)" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 12u16 , name : "BYNET 4 Port BYA Switch (BYA4G)" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 16u16 , name : "NCR AMC Memory Controller" , subsystems : & [] } , Device { vendor_id : 4122u16 , id : 7617u16 , name : "BYNET BIC2M/BIC4M/BYA4M" , subsystems : & [SubSystem { subvendor : 4122u16 , subdevice : 25u16 , name : "BIC2M" } , SubSystem { subvendor : 4122u16 , subdevice : 31u16 , name : "BIC4M" } , SubSystem { subvendor : 4122u16 , subdevice : 3790u16 , name : "BYA4M" }] } , Device { vendor_id : 4122u16 , id : 8104u16 , name : "BYNET Multi-port BIC Adapter (XBIC Based)" , subsystems : & [SubSystem { subvendor : 4122u16 , subdevice : 195u16 , name : "BYNET BIC2SE" }] }] }),
        (5371, Vendor { id : 5371u16 , name : "TRANSAS MARINE (UK) Ltd" , devices : & [] }),
        (5484, Vendor { id : 5484u16 , name : "Vidac Electronics GmbH" , devices : & [] }),
        (4325, Vendor { id : 4325u16 , name : "Micro Industries Corporation" , devices : & [] }),
        (59905, Vendor { id : 59905u16 , name : "Eagle Technology" , devices : & [Device { vendor_id : 59905u16 , id : 10u16 , name : "PCI-773 Temperature Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 50u16 , name : "PCI-730 & PC104P-30 Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 62u16 , name : "PCI-762 Opto-Isolator Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 65u16 , name : "PCI-763 Reed Relay Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 67u16 , name : "PCI-769 Opto-Isolator Reed Relay Combo Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 70u16 , name : "PCI-766 Analog Output Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 82u16 , name : "PCI-703 Analog I/O Card" , subsystems : & [] } , Device { vendor_id : 59905u16 , id : 2048u16 , name : "PCI-800 Digital I/O Card" , subsystems : & [] }] }),
        (7747, Vendor { id : 7747u16 , name : "MaxLinear Inc" , devices : & [Device { vendor_id : 7747u16 , id : 35076u16 , name : "MxL8904" , subsystems : & [] } , Device { vendor_id : 7747u16 , id : 35078u16 , name : "MxL8906" , subsystems : & [] } , Device { vendor_id : 7747u16 , id : 35080u16 , name : "MxL8908" , subsystems : & [] }] }),
        (4489, Vendor { id : 4489u16 , name : "Matsushita Electronics Co Ltd" , devices : & [] }),
        (5370, Vendor { id : 5370u16 , name : "WANDEL & GOLTERMANN" , devices : & [] }),
        (7703, Vendor { id : 7703u16 , name : "Arnold & Richter Cine Technik GmbH & Co. Betriebs KG" , devices : & [] }),
        (5540, Vendor { id : 5540u16 , name : "X-Net OY" , devices : & [] }),
        (5493, Vendor { id : 5493u16 , name : "Voltaire Advanced Data Security Ltd" , devices : & [] }),
        (38553, Vendor { id : 38553u16 , name : "Omni Media Technology Inc" , devices : & [Device { vendor_id : 38553u16 , id : 25957u16 , name : "6565" , subsystems : & [] }] }),
        (4593, Vendor { id : 4593u16 , name : "Symbios Logic Inc" , devices : & [] }),
        (8129, Vendor { id : 8129u16 , name : "QLogic, Corp." , devices : & [Device { vendor_id : 8129u16 , id : 13u16 , name : "IBA6110 InfiniBand HCA" , subsystems : & [] } , Device { vendor_id : 8129u16 , id : 16u16 , name : "IBA6120 InfiniBand HCA" , subsystems : & [] }] }),
        (4570, Vendor { id : 4570u16 , name : "Novell" , devices : & [] }),
        (5477, Vendor { id : 5477u16 , name : "Biostar Microtech Int'l Corp" , devices : & [] }),
        (8342, Vendor { id : 8342u16 , name : "Kaitian Information Technology Co., Ltd." , devices : & [Device { vendor_id : 8342u16 , id : 21505u16 , name : "KCP54(01) 2280 PCIe G4 x4 TLCKCP54(01) 2280 PCIe G4 x4 TLC" , subsystems : & [] } , Device { vendor_id : 8342u16 , id : 21506u16 , name : "KCP54(02) 2280 PCIe G4 x4 TLC" , subsystems : & [] } , Device { vendor_id : 8342u16 , id : 21507u16 , name : "KCP54(03) 2280 PCIe G4 x4 TLC" , subsystems : & [] } , Device { vendor_id : 8342u16 , id : 21508u16 , name : "KCP54(04) 2280 PCIe G4 x4 TLC" , subsystems : & [] }] }),
        (5093, Vendor { id : 5093u16 , name : "Telesoft Design Ltd" , devices : & [] }),
        (7851, Vendor { id : 7851u16 , name : "Hefei DATANG Storage Technology Co.,LTD." , devices : & [Device { vendor_id : 7851u16 , id : 12298u16 , name : "NVMe SSD Controller 300A" , subsystems : & [] } , Device { vendor_id : 7851u16 , id : 12299u16 , name : "NVMe SSD Controller 300B (DRAM-less)" , subsystems : & [] }] }),
        (6969, Vendor { id : 6969u16 , name : "sTec, Inc." , devices : & [Device { vendor_id : 6969u16 , id : 1u16 , name : "S1120 PCIe Accelerator SSD" , subsystems : & [] }] }),
        (4212, Vendor { id : 4212u16 , name : "NexGen Microsystems" , devices : & [Device { vendor_id : 4212u16 , id : 20088u16 , name : "82c500/1" , subsystems : & [] }] }),
        (4409, Vendor { id : 4409u16 , name : "Dynamic Pictures, Inc" , devices : & [Device { vendor_id : 4409u16 , id : 1u16 , name : "VGA Compatible 3D Graphics" , subsystems : & [] }] }),
        (13378, Vendor { id : 13378u16 , name : "Bihl+Wiedemann GmbH" , devices : & [Device { vendor_id : 13378u16 , id : 6019u16 , name : "AS-i 3.0 cPCI Master" , subsystems : & [] } , Device { vendor_id : 13378u16 , id : 6434u16 , name : "AS-i 3.0 PCI Master" , subsystems : & [] }] }),
        (5402, Vendor { id : 5402u16 , name : "Globetek" , devices : & [Device { vendor_id : 5402u16 , id : 4098u16 , name : "PCI-1002" , subsystems : & [] } , Device { vendor_id : 5402u16 , id : 4100u16 , name : "PCI-1004" , subsystems : & [] } , Device { vendor_id : 5402u16 , id : 4104u16 , name : "PCI-1008" , subsystems : & [] }] }),
        (6474, Vendor { id : 6474u16 , name : "DapTechnology B.V." , devices : & [Device { vendor_id : 6474u16 , id : 4369u16 , name : "FireSpy3850" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4370u16 , name : "FireSpy450b" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4371u16 , name : "FireSpy450bT" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4372u16 , name : "FireSpy850" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4373u16 , name : "FireSpy850bT" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4608u16 , name : "FireTrac 3460bT" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4609u16 , name : "FireTrac 3460bT (fallback firmware)" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4610u16 , name : "FireTrac 3460bT" , subsystems : & [] } , Device { vendor_id : 6474u16 , id : 4611u16 , name : "FireTrac 3460bT (fallback firmware)" , subsystems : & [] }] }),
        (34848, Vendor { id : 34848u16 , name : "Stryker Corporation" , devices : & [Device { vendor_id : 34848u16 , id : 10020u16 , name : "Mako Front Side Motor Controller [cPCI]" , subsystems : & [] }] }),
        (6970, Vendor { id : 6970u16 , name : "Westar Display Technologies" , devices : & [Device { vendor_id : 6970u16 , id : 30089u16 , name : "HRED J2000 - JPEG 2000 Video Codec Device" , subsystems : & [] }] }),
        (26248, Vendor { id : 26248u16 , name : "GUANGZHOU MAXSUN INFORMATION TECHNOLOGY CO., LTD." , devices : & [Device { vendor_id : 26248u16 , id : 4608u16 , name : "CooVox TDM Analog Module" , subsystems : & [] } , Device { vendor_id : 26248u16 , id : 5120u16 , name : "CooVOX TDM GSM Module" , subsystems : & [] } , Device { vendor_id : 26248u16 , id : 5632u16 , name : "CooVOX TDM E1/T1 Module" , subsystems : & [] } , Device { vendor_id : 26248u16 , id : 6144u16 , name : "CooVOX TDM BRI Module" , subsystems : & [] }] }),
        (7803, Vendor { id : 7803u16 , name : "Dataland" , devices : & [] }),
        (5279, Vendor { id : 5279u16 , name : "LECTRON Co Ltd" , devices : & [] }),
        (5337, Vendor { id : 5337u16 , name : "Alliance Semiconductor Corporation" , devices : & [Device { vendor_id : 5337u16 , id : 16u16 , name : "AP1011/SP1011 HyperTransport-PCI Bridge [Sturgeon]" , subsystems : & [] } , Device { vendor_id : 5337u16 , id : 36864u16 , name : "AS90L10204/10208 HyperTransport to PCI-X Bridge" , subsystems : & [] }] }),
        (4717, Vendor { id : 4717u16 , name : "Splash Technology, Inc." , devices : & [] }),
        (7893, Vendor { id : 7893u16 , name : "Moore Threads Technology Co.,Ltd" , devices : & [Device { vendor_id : 7893u16 , id : 256u16 , name : "MTT S10" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 257u16 , name : "MTT S10" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 258u16 , name : "MTT S30" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 261u16 , name : "MTT S50" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 262u16 , name : "MTT S60" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 273u16 , name : "MTT S100" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 289u16 , name : "MTT S1000M" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 290u16 , name : "MTT S1000" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 291u16 , name : "MTT S2000" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 511u16 , name : "MTT HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 512u16 , name : "MTT S80 Engineering Sample" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 513u16 , name : "MTT S80" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 514u16 , name : "MTT S70" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 515u16 , name : "MTT S60" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 529u16 , name : "MTT X300" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 545u16 , name : "G2S80" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 546u16 , name : "MTT S3000" , subsystems : & [SubSystem { subvendor : 7893u16 , subdevice : 1u16 , name : "C3150" }] } , Device { vendor_id : 7893u16 , id : 547u16 , name : "G2S4" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 549u16 , name : "MTT S3000E" , subsystems : & [SubSystem { subvendor : 7893u16 , subdevice : 1u16 , name : "C3150" }] } , Device { vendor_id : 7893u16 , id : 593u16 , name : "G2N10" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 767u16 , name : "MTT HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 768u16 , name : "MTT S90 Engineering Sample" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 769u16 , name : "MTT S90" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 787u16 , name : "MTT X500" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 803u16 , name : "MTT S4000" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 807u16 , name : "MTT S4000" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 808u16 , name : "MTT S4000" , subsystems : & [] } , Device { vendor_id : 7893u16 , id : 1023u16 , name : "MTT HDMI/DP Audio" , subsystems : & [] }] }),
        (4916, Vendor { id : 4916u16 , name : "Redcreek Communications, Inc" , devices : & [] }),
        (4384, Vendor { id : 4384u16 , name : "Dell EMC" , devices : & [Device { vendor_id : 4384u16 , id : 8966u16 , name : "Unity Fibre Channel Controller" , subsystems : & [] } , Device { vendor_id : 4384u16 , id : 9473u16 , name : "Unity Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4384u16 , id : 9477u16 , name : "Unity Fibre Channel Controller" , subsystems : & [] }] }),
        (4988, Vendor { id : 4988u16 , name : "Iwatsu Electric Co Ltd" , devices : & [] }),
        (5289, Vendor { id : 5289u16 , name : "HIVERTEC Inc" , devices : & [] }),
        (5254, Vendor { id : 5254u16 , name : "L3 Communications Telemetry & Instrumentation" , devices : & [] }),
        (5024, Vendor { id : 5024u16 , name : "Crystal Group Inc" , devices : & [] }),
        (4942, Vendor { id : 4942u16 , name : "CSTI" , devices : & [] }),
        (4968, Vendor { id : 4968u16 , name : "Skyware Corporation" , devices : & [] }),
        (4358, Vendor { id : 4358u16 , name : "VIA Technologies, Inc." , devices : & [Device { vendor_id : 4358u16 , id : 258u16 , name : "Embedded VIA Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 304u16 , name : "VT6305 1394.A Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 408u16 , name : "P4X600 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 516u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 520u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 568u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 600u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 601u16 , name : "CN333/CN400/PM880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 617u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 642u16 , name : "K8T800Pro Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32931u16 , name : "A8V Deluxe" }] } , Device { vendor_id : 4358u16 , id : 656u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 659u16 , name : "PM896 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 662u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 773u16 , name : "VT8363/8365 [KT133/KM133]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2439u16 , name : "K7VZA Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32819u16 , name : "A7V Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32830u16 , name : "A7V-E Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32834u16 , name : "A7V133/A7V133-C Mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 41985u16 , name : "KT7/KT7-RAID/KT7A/KT7A-RAID Mainboard" }] } , Device { vendor_id : 4358u16 , id : 776u16 , name : "PT880 Ultra/PT894 Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33177u16 , name : "P4V800D-X Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 776u16 , name : "Motherboard" }] } , Device { vendor_id : 4358u16 , id : 788u16 , name : "CN700/VN800/P4M800CE/Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 804u16 , name : "CX700/VX700 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 807u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 822u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 832u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 849u16 , name : "K8T890CF Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 851u16 , name : "VX800/820-Series Chipset Host-Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 868u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33230u16 , name : "P5VD2-VM mothervoard" }] } , Device { vendor_id : 4358u16 , id : 913u16 , name : "VT8371 [KX133]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1033u16 , name : "VX855/VX875 Host Bridge: Host Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1040u16 , name : "VX900 Series Host Bridge: Host Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1045u16 , name : "VT6415 PATA IDE Host Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33679u16 , name : "Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1045u16 , name : "Motherboard" }] } , Device { vendor_id : 4358u16 , id : 1049u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1281u16 , name : "VT8501 [Apollo MVP4]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1285u16 , name : "VT82C505" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1377u16 , name : "VT82C576MV" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1393u16 , name : "VT82C586A/B/VT82C686/A/B/VT823x/A/C PIPC Bus Master IDE" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 2689u16 , name : "L7VTA v1.0 Motherboard (KT400-8235)" } , SubSystem { subvendor : 4163u16 , subdevice : 32850u16 , name : "VT8233A Bus Master ATA100/66/33 IDE" } , SubSystem { subvendor : 4163u16 , subdevice : 32908u16 , name : "A7V8X / A7V333 motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X motherboard rev. 1.01" } , SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600/K8V-X/A8V Deluxe motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 1393u16 , name : "VT82C586/B/VT82C686/A/B/VT8233/A/C/VT8235 PIPC Bus Master IDE" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" } , SubSystem { subvendor : 4759u16 , subdevice : 63041u16 , name : "FX41 motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20482u16 , name : "GA-7VAX Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22785u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28704u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28820u16 , name : "K8T Neo2-F V2.0" } , SubSystem { subvendor : 5218u16 , subdevice : 28960u16 , name : "KT4AV motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1393u16 , name : "K7VT series Motherboards" }] } , Device { vendor_id : 4358u16 , id : 1398u16 , name : "VT82C576 3V [Apollo Master]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1409u16 , name : "CX700/VX700/VX800/820-Series Serial ATA RAID-Controller" , subsystems : & [SubSystem { subvendor : 4358u16 , subdevice : 1409u16 , name : "Wrong IDE ID" }] } , Device { vendor_id : 4358u16 , id : 1413u16 , name : "VT82C585VP [Apollo VP1/VPX]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1414u16 , name : "VT82C586/A/B PCI-to-ISA [Apollo VP]" , subsystems : & [SubSystem { subvendor : 4358u16 , subdevice : 0u16 , name : "MVP3 ISA Bridge" }] } , Device { vendor_id : 4358u16 , id : 1425u16 , name : "VT8237A Integrated SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1429u16 , name : "VT82C595 [Apollo VP2]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1430u16 , name : "VT82C596 ISA [Mobile South]" , subsystems : & [SubSystem { subvendor : 4358u16 , subdevice : 0u16 , name : "VT82C596/A/B PCI to ISA Bridge" } , SubSystem { subvendor : 5208u16 , subdevice : 1430u16 , name : "VT82C596/A/B PCI to ISA Bridge" }] } , Device { vendor_id : 4358u16 , id : 1431u16 , name : "VT82C597 [Apollo VP3]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1432u16 , name : "VT82C598 [Apollo MVP3]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1537u16 , name : "VT8601 [Apollo ProMedia]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1541u16 , name : "VT8605 [ProSavage PM133]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4692u16 , name : "D9840-60001 [Brio BA410 Motherboard]" } , SubSystem { subvendor : 4163u16 , subdevice : 32812u16 , name : "CUV4X mainboard" }] } , Device { vendor_id : 4358u16 , id : 1664u16 , name : "VT82C680 [Apollo P6]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1670u16 , name : "VT82C686 [Apollo Super South]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 4694u16 , name : "D9840-60001 [Brio BA410 Motherboard]" } , SubSystem { subvendor : 4163u16 , subdevice : 32812u16 , name : "CUV4X mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32819u16 , name : "A7V Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32830u16 , name : "A7V-E Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32832u16 , name : "A7M266 Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32834u16 , name : "A7V133/A7V133-C Mainboard" } , SubSystem { subvendor : 4358u16 , subdevice : 0u16 , name : "VT82C686/A PCI to ISA Bridge" } , SubSystem { subvendor : 4358u16 , subdevice : 1670u16 , name : "VT82C686/A PCI to ISA Bridge" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" } , SubSystem { subvendor : 5243u16 , subdevice : 42754u16 , name : "KG7-Lite Mainboard" }] } , Device { vendor_id : 4358u16 , id : 1681u16 , name : "VT82C693A/694x [Apollo PRO133x]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" } , SubSystem { subvendor : 5208u16 , subdevice : 1681u16 , name : "VT82C691 Apollo Pro System Controller" }] } , Device { vendor_id : 4358u16 , id : 1683u16 , name : "VT82C693 [Apollo Pro Plus]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1688u16 , name : "VT82C693A [Apollo Pro133 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1801u16 , name : "VX11 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1802u16 , name : "VX11 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1803u16 , name : "VX11 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1804u16 , name : "VX11 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1805u16 , name : "VX11 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 1806u16 , name : "VX11 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 2342u16 , name : "VT82C926 [Amazon]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4096u16 , name : "VT82C570MV" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4358u16 , name : "VT82C570MV" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4386u16 , name : "VX800/VX820 Chrome 9 HC3 Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4612u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4616u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4664u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4696u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4697u16 , name : "CN333/CN400/PM880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4713u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4738u16 , name : "K8T800Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4752u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4755u16 , name : "PM896 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4758u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4872u16 , name : "PT894 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4884u16 , name : "CN700/VN800/P4M800CE/Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4900u16 , name : "CX700/VX700-Series Error Reporting" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4903u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4918u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4928u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4945u16 , name : "VT3351 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4947u16 , name : "VX800/VX820 Error Reporting" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 4964u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5129u16 , name : "VX855/VX875 Error Reporting" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5136u16 , name : "VX900 Series Error Reporting" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5145u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5489u16 , name : "VT82C576M/VT82C586" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5525u16 , name : "VT82C595/97 [Apollo VP2/97]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 5938u16 , name : "VT1732 [Envy24 II] PCI Multi-Channel Audio Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8454u16 , name : "VIA Rhine Family Fast Ethernet Adapter (VT6105)" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8708u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8712u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8760u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8792u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8793u16 , name : "CN333/CN400/PM880 CPU Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8809u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8834u16 , name : "K8T800Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8848u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8851u16 , name : "PM896 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8854u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8968u16 , name : "PT894 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8980u16 , name : "CN700/VN800/P4M800CE/Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8996u16 , name : "CX700/VX700-Series Host Interface Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 8999u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9014u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9024u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9041u16 , name : "VT3351 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9043u16 , name : "VX800/VX820 Host Bus Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9060u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9225u16 , name : "VX855/VX875 Host Bus Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9232u16 , name : "VX900 Series CPU Bus Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 9241u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 10362u16 , name : "VT8251 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 10363u16 , name : "VT8251 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 10364u16 , name : "VT8251 PCIE Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 10365u16 , name : "VT8251 PCIE Root Port" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 10366u16 , name : "VT8237/8251/8261 Ultra VLINK Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12322u16 , name : "CLE266" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12344u16 , name : "VT82xx/62xx/VX700/8x0/900 UHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 2341u16 , subdevice : 4660u16 , name : "onboard UHCI USB 1.1 Controller" } , SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 2689u16 , name : "L7VTA v1.0 Motherboard (KT400-8235)" } , SubSystem { subvendor : 4163u16 , subdevice : 32896u16 , name : "A7V333 motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32908u16 , name : "VT62xx USB1.1 4 port controller" } , SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600/K8V-X/A8V Deluxe motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 12344u16 , name : "USB 1.1 UHCI controller" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" } , SubSystem { subvendor : 4660u16 , subdevice : 2341u16 , name : "MVP3 USB Controller" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-7VAX Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22785u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28704u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28820u16 , name : "K8T Neo2-F V2.0" } , SubSystem { subvendor : 5218u16 , subdevice : 28960u16 , name : "KT4AV motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 6189u16 , subdevice : 8221u16 , name : "CN-029 USB2.0 4 port PCI Card" } , SubSystem { subvendor : 6217u16 , subdevice : 12344u16 , name : "K7VT series Motherboards" } , SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 4358u16 , id : 12352u16 , name : "VT82C586B ACPI" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12355u16 , name : "VT86C100A [Rhine]" , subsystems : & [SubSystem { subvendor : 4285u16 , subdevice : 0u16 , name : "VT86C100A Fast Ethernet Adapter" } , SubSystem { subvendor : 4358u16 , subdevice : 256u16 , name : "VT86C100A Fast Ethernet Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 5120u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. A)" }] } , Device { vendor_id : 4358u16 , id : 12356u16 , name : "VT6306/7/8 [Fire II(M)] IEEE 1394 OHCI Controller" , subsystems : & [SubSystem { subvendor : 16u16 , subdevice : 1u16 , name : "IEEE 1394 4port DCST 1394-3+1B" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Media Center PC m7590n" } , SubSystem { subvendor : 4163u16 , subdevice : 32906u16 , name : "A8V/A8N/P4P800/P5SD2 series motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33278u16 , name : "Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 4096u16 , name : "GA-7VT600-1394 Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 8317u16 , name : "K8NGM2 series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 8573u16 , name : "Aspire L250" } , SubSystem { subvendor : 5218u16 , subdevice : 22797u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28717u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 38685u16 , name : "MS-6917" } , SubSystem { subvendor : 5435u16 , subdevice : 4422u16 , name : "Cameo DV Firewire controller" }] } , Device { vendor_id : 4358u16 , id : 12368u16 , name : "VT82C596 Power Management" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12369u16 , name : "VT82C596 Power Management" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12371u16 , name : "VT6105M [Rhine-III]" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 5124u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. D)" }] } , Device { vendor_id : 4358u16 , id : 12375u16 , name : "VT82C686 [Apollo Super ACPI]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 2439u16 , name : "K7VZA Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32819u16 , name : "A7V Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32830u16 , name : "A7V-E Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32832u16 , name : "A7M266 Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32834u16 , name : "A7V133/A7V133-C Mainboard" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" }] } , Device { vendor_id : 4358u16 , id : 12376u16 , name : "VT82C686 AC97 Audio Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 151u16 , name : "SoundMax Digital Integrated Audio" } , SubSystem { subvendor : 3601u16 , subdevice : 45460u16 , name : "Soundmax integrated digital audio" } , SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 2439u16 , name : "K7VZA Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 4689u16 , name : "D9840-60001 [Brio BA410 Motherboard]" } , SubSystem { subvendor : 4163u16 , subdevice : 4358u16 , name : "A7V133/A7V133-C Mainboard" } , SubSystem { subvendor : 4358u16 , subdevice : 17681u16 , name : "Onboard Audio on EP7KXA" } , SubSystem { subvendor : 4358u16 , subdevice : 43523u16 , name : "VT1612A AC'97 Audio Controller" } , SubSystem { subvendor : 4564u16 , subdevice : 21320u16 , name : "AD1881A audio" } , SubSystem { subvendor : 5208u16 , subdevice : 30208u16 , name : "Onboard Audio" } , SubSystem { subvendor : 5218u16 , subdevice : 12433u16 , name : "MS-6309 Onboard Audio" } , SubSystem { subvendor : 5218u16 , subdevice : 12434u16 , name : "MS-6309 v2.x Mainboard (VIA VT1611A codec)" } , SubSystem { subvendor : 5218u16 , subdevice : 13056u16 , name : "MS-6330 Onboard Audio" } , SubSystem { subvendor : 5218u16 , subdevice : 13312u16 , name : "MS-6340 (VT8363) motherboard" } , SubSystem { subvendor : 5597u16 , subdevice : 30217u16 , name : "Onboard Audio" }] } , Device { vendor_id : 4358u16 , id : 12377u16 , name : "VT8233/A/8235/8237 AC97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2689u16 , name : "L7VTA v1.0 Motherboard (KT400-8235)" } , SubSystem { subvendor : 4121u16 , subdevice : 6209u16 , name : "M811 (VT8367/VT8235/VT6103) [KT333] motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 6263u16 , name : "K8M800-M2 (V2.0) onboard audio" } , SubSystem { subvendor : 4163u16 , subdevice : 32917u16 , name : "A7V8X Motherboard (Realtek ALC650 codec)" } , SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32944u16 , name : "A7V600/K8V-X/K8V Deluxe motherboard (ADI AD1980 codec [SoundMAX])" } , SubSystem { subvendor : 4163u16 , subdevice : 33011u16 , name : "SK8V motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33037u16 , name : "P5VD1-X (AD1888 codec [SoundMax])" } , SubSystem { subvendor : 4163u16 , subdevice : 33066u16 , name : "A8V Deluxe motherboard (Realtek ALC850 codec)" } , SubSystem { subvendor : 4332u16 , subdevice : 33128u16 , name : "High Definition Audio" } , SubSystem { subvendor : 4358u16 , subdevice : 12377u16 , name : "L7VMM2 Motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 16737u16 , name : "K7VT2 motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 16752u16 , name : "PCPartner P4M800-8237R Motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 17746u16 , name : "Soyo KT-600 Dragon Plus (Realtek ALC 650)" } , SubSystem { subvendor : 4759u16 , subdevice : 49504u16 , name : "FX41 motherboard (Realtek ALC650 codec)" } , SubSystem { subvendor : 5139u16 , subdevice : 5243u16 , name : "KV8 Pro motherboard onboard audio" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-7VAX Onboard Audio (Realtek ALC650)" } , SubSystem { subvendor : 5218u16 , subdevice : 128u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 14336u16 , name : "KT266 onboard audio" } , SubSystem { subvendor : 5218u16 , subdevice : 22785u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 5781u16 , subdevice : 12300u16 , name : "Realtek ALC655 audio on EP-8KRA series mainboard" } , SubSystem { subvendor : 5875u16 , subdevice : 16752u16 , name : "J7F2 motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4216u16 , name : "Amilo L7300T notebook" } , SubSystem { subvendor : 6217u16 , subdevice : 2128u16 , name : "ASRock 775Dual-880 Pro onboard audio (Realtek ALC850)" } , SubSystem { subvendor : 6217u16 , subdevice : 38713u16 , name : "P4VT8 Mainboard (C-Media CMI9739A codec)" } , SubSystem { subvendor : 6217u16 , subdevice : 38753u16 , name : "K7VT series Motherboards" } , SubSystem { subvendor : 16389u16 , subdevice : 18192u16 , name : "MSI K7T266 Pro2-RU (MSI-6380 v2) onboard audio (Realtek/ALC 200/200P)" } , SubSystem { subvendor : 41120u16 , subdevice : 438u16 , name : "AK77-8XN onboard audio" } , SubSystem { subvendor : 41120u16 , subdevice : 834u16 , name : "AK86-L motherboard" }] } , Device { vendor_id : 4358u16 , id : 12389u16 , name : "VT6102/VT6103 [Rhine-II]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600-X Motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 258u16 , name : "VT6102/6103 [Rhine II] Ethernet Controller" } , SubSystem { subvendor : 4486u16 , subdevice : 5120u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. A)" } , SubSystem { subvendor : 4486u16 , subdevice : 5121u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. B)" } , SubSystem { subvendor : 4486u16 , subdevice : 5122u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. B)" } , SubSystem { subvendor : 5049u16 , subdevice : 5153u16 , name : "LD-10/100AL PCI Fast Ethernet Adapter (rev.B)" } , SubSystem { subvendor : 5218u16 , subdevice : 28769u16 , name : "MS-7061" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" } , SubSystem { subvendor : 5781u16 , subdevice : 12293u16 , name : "VT6103" } , SubSystem { subvendor : 6217u16 , subdevice : 12389u16 , name : "K7VT series Motherboards" }] } , Device { vendor_id : 4358u16 , id : 12392u16 , name : "AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 12446u16 , name : "MS-6309 Saturn Motherboard" }] } , Device { vendor_id : 4358u16 , id : 12404u16 , name : "VT8233 PCI to ISA Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32850u16 , name : "VT8233A" }] } , Device { vendor_id : 4358u16 , id : 12433u16 , name : "VT8633 [Apollo Pro266]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12441u16 , name : "VT8366/A/7 [Apollo KT266/A/333]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 6209u16 , name : "M811 (VT8367/VT8235/VT6103) [KT333] motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32868u16 , name : "A7V266-E Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32895u16 , name : "A7V333 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 12441u16 , name : "K7VT2 motherboard" }] } , Device { vendor_id : 4358u16 , id : 12545u16 , name : "VT8653 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12546u16 , name : "VT8662 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12547u16 , name : "VT8615 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12548u16 , name : "USB 2.0 EHCI-Compliant Host-Controller" , subsystems : & [SubSystem { subvendor : 2341u16 , subdevice : 4660u16 , name : "onboard EHCI USB 2.0 Controller" } , SubSystem { subvendor : 4121u16 , subdevice : 2689u16 , name : "L7VTA v1.0 Motherboard (KT400-8235)" } , SubSystem { subvendor : 4163u16 , subdevice : 32908u16 , name : "A7V8X motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X motherboard rev 1.01" } , SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600/K8V-X/A8V Deluxe motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 12548u16 , name : "USB 2.0 EHCI controller" } , SubSystem { subvendor : 4759u16 , subdevice : 63041u16 , name : "FX41 motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-7VAX Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22785u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28704u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28820u16 , name : "K8T Neo2-F V2.0" } , SubSystem { subvendor : 5218u16 , subdevice : 28960u16 , name : "KT4AV motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 6189u16 , subdevice : 8221u16 , name : "CN-029 USB 2.0 4 port PCI Card" } , SubSystem { subvendor : 6217u16 , subdevice : 12548u16 , name : "K7VT series Motherboards" } , SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" }] } , Device { vendor_id : 4358u16 , id : 12550u16 , name : "VT6105/VT6106S [Rhine-III]" , subsystems : & [SubSystem { subvendor : 4358u16 , subdevice : 261u16 , name : "VT6106S [Rhine-III]" } , SubSystem { subvendor : 4486u16 , subdevice : 5123u16 , name : "DFE-530TX PCI Fast Ethernet Adapter (rev. C)" } , SubSystem { subvendor : 4486u16 , subdevice : 5125u16 , name : "DFE-520TX Fast Ethernet PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 5126u16 , name : "DFE-530TX+ rev F2" } , SubSystem { subvendor : 4486u16 , subdevice : 5127u16 , name : "DFE-538TX" }] } , Device { vendor_id : 4358u16 , id : 12552u16 , name : "K8M800/K8N800/K8N800A [S3 UniChrome Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12553u16 , name : "VT8233C PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12562u16 , name : "VT8361 [KLE133] Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12563u16 , name : "VPX/VPX2 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12566u16 , name : "VT8375 [KM266/KL266] Host Bridge" , subsystems : & [SubSystem { subvendor : 4759u16 , subdevice : 63041u16 , name : "FX41 motherboard" }] } , Device { vendor_id : 4358u16 , id : 12568u16 , name : "CN400/PM800/PM880/PN800/PN880 [S3 UniChrome Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12569u16 , name : "VT6120/VT6121/VT6122/VT6130 Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12578u16 , name : "VT8623 [Apollo CLE266] integrated CastleRock graphics" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12579u16 , name : "VT8623 [Apollo CLE266]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12584u16 , name : "VT8753 [P4X266 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12595u16 , name : "VT3133 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12610u16 , name : "VT6651 WiFi Adapter, 802.11b" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12615u16 , name : "VT8233A ISA Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32908u16 , name : "A7V333 motherboard" }] } , Device { vendor_id : 4358u16 , id : 12616u16 , name : "P4M266 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12617u16 , name : "VIA VT6420 SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600/K8V Deluxe/K8V-X/A8V Deluxe motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 45059u16 , name : "GA-7VM400AM(F) Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22785u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28704u16 , name : "K8T Neo 2 Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28820u16 , name : "K8T Neo2-F V2.0" } , SubSystem { subvendor : 5218u16 , subdevice : 29057u16 , name : "K8MM3-V mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5128u16 , name : "KV7" } , SubSystem { subvendor : 6217u16 , subdevice : 12617u16 , name : "K7VT6 motherboard" } , SubSystem { subvendor : 41120u16 , subdevice : 1197u16 , name : "AK86-L motherboard" }] } , Device { vendor_id : 4358u16 , id : 12630u16 , name : "P/KN266 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12631u16 , name : "CX700/VX700 [S3 UniChrome Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12644u16 , name : "VT6410 ATA133 RAID controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33012u16 , name : "P4P800 Mainboard Deluxe ATX" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" }] } , Device { vendor_id : 4358u16 , id : 12648u16 , name : "P4X333/P4X400/PT800 AGP Bridge" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 12648u16 , name : "P4VT8 Mainboard" }] } , Device { vendor_id : 4358u16 , id : 12663u16 , name : "VT8235 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2689u16 , name : "L7VTA v1.0 Motherboard (KT400-8235)" } , SubSystem { subvendor : 4121u16 , subdevice : 6209u16 , name : "M811 (VT8367/VT8235/VT6103) [KT333] motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32908u16 , name : "A7V8X motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32929u16 , name : "A7V8X-X motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 0u16 , name : "KT4AV motherboard" } , SubSystem { subvendor : 4759u16 , subdevice : 63041u16 , name : "FX41 motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-7VAX Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 12663u16 , name : "K7VT series Motherboards" }] } , Device { vendor_id : 4358u16 , id : 12664u16 , name : "ProSavageDDR P4N333 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12680u16 , name : "VT8385 [K8T800 AGP] Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32931u16 , name : "K8V Deluxe/K8V-X motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" }] } , Device { vendor_id : 4358u16 , id : 12681u16 , name : "VT8377 [KT400/KT600 AGP] Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32895u16 , name : "A7V8X motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 0u16 , name : "KT4AV motherboard (KT400A)" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-7VAX Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 12681u16 , name : "K7VT series Motherboards" }] } , Device { vendor_id : 4358u16 , id : 12720u16 , name : "VX11 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12721u16 , name : "VX11 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12722u16 , name : "VX11 DRAM Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12723u16 , name : "VX11 Power Management Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12724u16 , name : "VX11 I/O APIC" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12725u16 , name : "VX11 Scratch Device" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12727u16 , name : "VX11 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12728u16 , name : "VX11 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12804u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12805u16 , name : "VT8378 [KM400/A] Chipset Host Bridge" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-7VM400M Motherboard" }] } , Device { vendor_id : 4358u16 , id : 12808u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12819u16 , name : "VPX/VPX2 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12824u16 , name : "K8T800M Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12839u16 , name : "VT8237 ISA bridge [KT600/K8T800/K8T890/CN700 South]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33005u16 , name : "A7V600/K8V-X/A8V Deluxe motherboard" } , SubSystem { subvendor : 4358u16 , subdevice : 12839u16 , name : "VT8237 ISA bridge" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-7VT600 Motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 12839u16 , name : "K7VT4 motherboard" }] } , Device { vendor_id : 4358u16 , id : 12848u16 , name : "K8M890CE/K8N890CE [Chrome 9]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12856u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12873u16 , name : "VT6421 IDE/SATA Controller" , subsystems : & [SubSystem { subvendor : 4358u16 , subdevice : 12873u16 , name : "VT6421 IDE/SATA Controller" }] } , Device { vendor_id : 4358u16 , id : 12874u16 , name : "CX700/VX700 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12875u16 , name : "CX700/VX700 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12878u16 , name : "CX700/VX700 Internal Module Bus" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12883u16 , name : "VT6655 WiFi Adapter, 802.11a/b/g" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12888u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12889u16 , name : "CN333/CN400/PM880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12896u16 , name : "VIA Chrome9 HC IGP" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12905u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12930u16 , name : "K8T800Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12935u16 , name : "VT8251 PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12936u16 , name : "VX900/VT8xxx High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX VD01" }] } , Device { vendor_id : 4358u16 , id : 12944u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 12950u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13092u16 , name : "CX700/VX700-Series DRAM Bus Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13095u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13110u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13111u16 , name : "VT8237A PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13120u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13123u16 , name : "P4M890 [S3 UniChrome Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13124u16 , name : "CN700/P4M800 Pro/P4M800 CE/VN800 Graphics [S3 UniChrome Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13129u16 , name : "VT8251 AHCI/SATA 4-Port Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13137u16 , name : "VT3351 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13139u16 , name : "VX800/820 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13156u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13157u16 , name : "VT630x IEEE 1394 Host Controller [Fire II/M]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13169u16 , name : "CN896/VN896/P4M900 [Chrome 9 HC]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13170u16 , name : "VT8237S PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13178u16 , name : "VT8237A PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13179u16 , name : "VT8237A Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13313u16 , name : "VT6325 Firewire Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13314u16 , name : "VT8261 PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13315u16 , name : "VT6315 Series Firewire Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33652u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4163u16 , subdevice : 33668u16 , name : "P8P67 Deluxe Motherboard" }] } , Device { vendor_id : 4358u16 , id : 13321u16 , name : "VX855/VX875 DRAM Bus Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13328u16 , name : "VX900 Series DRAM Bus Control" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" }] } , Device { vendor_id : 4358u16 , id : 13337u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13362u16 , name : "VL800/801 xHCI USB 3.0 Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13398u16 , name : "VX11 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13403u16 , name : "VX11 Miscellaneous Bus" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13443u16 , name : "VL805/806 xHCI USB 3.0 Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13914u16 , name : "VT3365 Card Reader Host Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 13915u16 , name : "VT3365 SDIO Host Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 14849u16 , name : "VX11 Graphics [Chrome 645/640]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16410u16 , name : "VT6325 Card Reader Host Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16411u16 , name : "VT6325 SDIO PCI-E Host Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16713u16 , name : "VIA VT6420 (ATA133) Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16900u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16904u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16952u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16984u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 16985u16 , name : "CN333/CN400/PM880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17001u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17026u16 , name : "K8T800Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17040u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17043u16 , name : "PM896 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17046u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17160u16 , name : "PT894 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17172u16 , name : "CN700/VN800/P4M800CE/Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17188u16 , name : "CX700/VX700-Series Power Management and Testing Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17191u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17206u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17216u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17233u16 , name : "VT3351 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17235u16 , name : "VX800/VX820 Power Management Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17252u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17417u16 , name : "VX855/VX875 Power Management Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 17424u16 , name : "VX900 Series Power Management and Chip Testing Control" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" }] } , Device { vendor_id : 4358u16 , id : 17433u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 20528u16 , name : "VT82C596 ACPI [Apollo PRO]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 20770u16 , name : "VX855/VX875 Chrome 9 HCM Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21000u16 , name : "PT890 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21048u16 , name : "K8T890 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21127u16 , name : "VT8251 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21136u16 , name : "K8M890 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21256u16 , name : "PT894 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21284u16 , name : "CX700M2/VX700/VX800/820-Series Serial ATA & EIDE-Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21287u16 , name : "P4M890 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21302u16 , name : "K8M890CE I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21303u16 , name : "VT8237A Integrated SATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21312u16 , name : "PT900 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21329u16 , name : "VT3351 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21331u16 , name : "VX800/VX820 APIC and Central Traffic Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21348u16 , name : "CN896/VN896/P4M900 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21362u16 , name : "VT8237S SATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21513u16 , name : "VX855/VX875 APIC and Central Traffic Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21520u16 , name : "VX900 Series APIC and Central Traffic Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 21529u16 , name : "VN1000 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 24832u16 , name : "VT85C100A [Rhine II]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 24866u16 , name : "VN1000 Graphics [Chrome 520 IGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25223u16 , name : "VT8251 AHCI SATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25232u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25383u16 , name : "P4M890 Security Device" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25427u16 , name : "VX800/VX820 Scratch Registers" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25444u16 , name : "CN896/VN896/P4M900 Security Device" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25609u16 , name : "VX855/VX875 Scratch Registers" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 25616u16 , name : "VX900 Series Scratch Registers" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" }] } , Device { vendor_id : 4358u16 , id : 25625u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 28962u16 , name : "VX900 Graphics [Chrome9 HD]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29188u16 , name : "K8M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29189u16 , name : "KM400/KN400/P4M800 [S3 UniChrome]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "Gigabyte GA-7VM400(A)M(F) Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28769u16 , name : "MS-7061" }] } , Device { vendor_id : 4358u16 , id : 29192u16 , name : "PT890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29240u16 , name : "K8T890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29272u16 , name : "PT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29273u16 , name : "CN333/CN400/PM880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29289u16 , name : "KT880 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29314u16 , name : "K8T800Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29328u16 , name : "K8M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29331u16 , name : "PM896 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29334u16 , name : "P4M800 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29448u16 , name : "PT894 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29460u16 , name : "CN700/VN800/P4M800CE/Pro Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29476u16 , name : "CX700/VX700-Series North-South Module Interface Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29479u16 , name : "P4M890 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29494u16 , name : "K8M890CE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29504u16 , name : "PT900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29521u16 , name : "VT3351 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29523u16 , name : "VX800/VX820 North-South Module Interface Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29540u16 , name : "CN896/VN896/P4M900 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29554u16 , name : "VT8237S SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29705u16 , name : "VX855/VX875 North-South Module Interface Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 29712u16 , name : "VX900 Series North-South Module Interface Control" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX nano VD01" }] } , Device { vendor_id : 4358u16 , id : 29721u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33329u16 , name : "VT8231 [PCI-to-ISA Bridge]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33333u16 , name : "VT8235 ACPI" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33541u16 , name : "VT8363/8365 [KT133/KM133 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33572u16 , name : "CX700/VX700-Series Bus Control and Power Management" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33619u16 , name : "VX800/VX820 Bus Control and Power Management" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33681u16 , name : "VT8371 [KX133 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33792u16 , name : "MVP4" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33801u16 , name : "VX855/VX875 Bus Control and Power Management" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 33808u16 , name : "VX900 Series Bus Control and Power Management" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41337u16 , name : "ZBOX VD01" }] } , Device { vendor_id : 4358u16 , id : 34048u16 , name : "KLE133/PLE133/PLE133T" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34049u16 , name : "VT8501 [Apollo MVP4 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34198u16 , name : "VT82C596 [Apollo PRO AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34199u16 , name : "VT82C597 [Apollo VP3 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34200u16 , name : "VT82C598/694x [Apollo MVP3/Pro133x AGP]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2437u16 , name : "P6VXA Motherboard" }] } , Device { vendor_id : 4358u16 , id : 34305u16 , name : "VT8601 [Apollo ProMedia AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34309u16 , name : "VT8605 [PM133 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34449u16 , name : "VT82C691 [Apollo Pro]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 34451u16 , name : "VT82C693 [Apollo Pro Plus] PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 35365u16 , name : "PL133/PL133T [S3 ProSavage]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 35366u16 , name : "KL133/KL133A/KM133/KM133A [S3 ProSavage]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36097u16 , name : "PN133/PN133T [S3 Twister]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36100u16 , name : "KM266/P4M266/P4M266A/P4N266 [S3 ProSavageDDR]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36864u16 , name : "VT8261 IDE Controller [StorX IDE Controller - 9000]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36865u16 , name : "VX900 Series Serial-ATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36928u16 , name : "VT8261 SATA Controller [StorX RAID Controller - 9040]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36929u16 , name : "VX900/VX11 SATA Controller [StorX RAID Controller - 9041]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 36994u16 , name : "Standard AHCI 1.0 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 37184u16 , name : "VX11 Graphics [Chrome 645/640] HDMI Audio Device" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 37232u16 , name : "VX900 Graphics [Chrome9 HD] HDMI Audio Device" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 37377u16 , name : "USB3.0 Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 37760u16 , name : "Ncore Coprocessor for Centaur CNS" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 38192u16 , name : "VX800/820/900/VT8261 Series Secure Digital Memory Card Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 38352u16 , name : "VX800/820/900 Series SDIO Host Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41480u16 , name : "PT890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41528u16 , name : "K8T890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41767u16 , name : "P4M890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41811u16 , name : "VX8xx/900 Series South-North Module Interface Control" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41828u16 , name : "CN896/VN896/P4M900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 41993u16 , name : "VX855/VX875/VX900 Series USB Device Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 42000u16 , name : "VX900 Series PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 42009u16 , name : "VN1000 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45201u16 , name : "VT8633 [Apollo Pro266 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45209u16 , name : "VT8366/A/7 [Apollo KT266/A/333 AGP]" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45313u16 , name : "VT8653 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45314u16 , name : "VT8362 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45315u16 , name : "VT8615 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45330u16 , name : "VT8361 [KLE133] AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45331u16 , name : "VPX/VPX2 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45333u16 , name : "VT8363/8365 [KT133/KM133] PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45416u16 , name : "VT8235 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45448u16 , name : "VT8237/8251 PCI bridge [K8M890/K8T800/K8T890 South]" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" }] } , Device { vendor_id : 4358u16 , id : 45464u16 , name : "VT8237/CX700/VX700-Series PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45587u16 , name : "VPX/VPX2 I/O APIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 45907u16 , name : "VX855/VX875/VX900 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 46096u16 , name : "VX900 Series PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 46105u16 , name : "VN1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 47513u16 , name : "[K8T890 North / VT8237 South] PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 49672u16 , name : "PT890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 49720u16 , name : "K8T890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 49959u16 , name : "P4M890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 49984u16 , name : "PT900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 50003u16 , name : "VX800/820-Series PCI-Express Root (PCI-to-PCI Virtual Bridge)" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 50020u16 , name : "CN896/VN896/P4M900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 50185u16 , name : "VX855/VX875 EIDE Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 50192u16 , name : "VX900 Series PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 50201u16 , name : "VN1000 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 53508u16 , name : "VT8237R USB UDCI Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 53768u16 , name : "PT890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 53779u16 , name : "VPX/VPX2 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 53816u16 , name : "K8T890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 54080u16 , name : "PT900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 54288u16 , name : "VX900 Series PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 54297u16 , name : "VN1000 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 57864u16 , name : "PT890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 57912u16 , name : "K8T890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 58176u16 , name : "PT900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 58195u16 , name : "VX800/820-Series PCI-Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 58384u16 , name : "VX900 Series PCI Express Physical Layer Electrical Sub-block" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 58393u16 , name : "VN1000 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 61960u16 , name : "PT890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 62008u16 , name : "K8T890 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 62272u16 , name : "PT900 PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 62291u16 , name : "VX800/820-Series PCI-Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 62480u16 , name : "VX900 Series PCI UART Port 0-3" , subsystems : & [] } , Device { vendor_id : 4358u16 , id : 62489u16 , name : "VN1000 PCI to PCI Bridge" , subsystems : & [] }] }),
        (5733, Vendor { id : 5733u16 , name : "EDAX Inc" , devices : & [Device { vendor_id : 5733u16 , id : 6515u16 , name : "DPP-II FR2 Board" , subsystems : & [] } , Device { vendor_id : 5733u16 , id : 8192u16 , name : "SG-IIP Board" , subsystems : & [] }] }),
        (5220, Vendor { id : 5220u16 , name : "Interactive Circuits & Systems Ltd" , devices : & [] }),
        (5089, Vendor { id : 5089u16 , name : "Silicom Multimedia Systems Inc" , devices : & [] }),
        (5412, Vendor { id : 5412u16 , name : "ENE Technology Inc" , devices : & [Device { vendor_id : 5412u16 , id : 1296u16 , name : "CB710 Memory Card Reader Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" }] } , Device { vendor_id : 5412u16 , id : 1312u16 , name : "FLASH memory: ENE Technology Inc:" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1328u16 , name : "ENE PCI Memory Stick Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1360u16 , name : "ENE PCI Secure Digital Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1361u16 , name : "SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1552u16 , name : "PCI Smart Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1824u16 , name : "Memory Stick Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1840u16 , name : "ENE PCI Memory Stick Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1872u16 , name : "ENE PCI SmartMedia / xD Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 1873u16 , name : "ENE PCI Secure Digital / MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 4625u16 , name : "CB1211 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 4645u16 , name : "CB1225 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 5136u16 , name : "CB1410 Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 60u16 , name : "CL50 motherboard" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" }] } , Device { vendor_id : 5412u16 , id : 5137u16 , name : "CB-710/2/4 Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" }] } , Device { vendor_id : 5412u16 , id : 5138u16 , name : "CB-712/4 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 5152u16 , name : "CB1420 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 5153u16 , name : "CB-720/2/4 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 5412u16 , id : 5154u16 , name : "CB-722/4 Cardbus Controller" , subsystems : & [] }] }),
        (4999, Vendor { id : 4999u16 , name : "Systran Corp" , devices : & [] }),
        (28771, Vendor { id : 28771u16 , name : "pcHDTV" , devices : & [Device { vendor_id : 28771u16 , id : 8192u16 , name : "HD-2000" , subsystems : & [] } , Device { vendor_id : 28771u16 , id : 12288u16 , name : "HD-3000" , subsystems : & [] } , Device { vendor_id : 28771u16 , id : 21760u16 , name : "HD5500 HDTV" , subsystems : & [] }] }),
        (4807, Vendor { id : 4807u16 , name : "Dialogic Corp" , devices : & [Device { vendor_id : 4807u16 , id : 1350u16 , name : "Springware D/120JCT-LS" , subsystems : & [] } , Device { vendor_id : 4807u16 , id : 1607u16 , name : "Springware D/240JCT-T1" , subsystems : & [] } , Device { vendor_id : 4807u16 , id : 1654u16 , name : "Springware D/41JCT-LS" , subsystems : & [] } , Device { vendor_id : 4807u16 , id : 1669u16 , name : "Springware D/480JCT-2T1" , subsystems : & [] }] }),
        (4675, Vendor { id : 4675u16 , name : "Delphax" , devices : & [] }),
        (4200, Vendor { id : 4200u16 , name : "Diversified Technology" , devices : & [] }),
        (4488, Vendor { id : 4488u16 , name : "Shima Seiki Manufacturing Ltd." , devices : & [] }),
        (5893, Vendor { id : 5893u16 , name : "Digital First, Inc." , devices : & [] }),
        (4815, Vendor { id : 4815u16 , name : "Prophet Systems, Inc." , devices : & [] }),
        (12562, Vendor { id : 12562u16 , name : "Satelco Ingenieria S.A." , devices : & [] }),
        (5292, Vendor { id : 5292u16 , name : "Novaweb Technologies Inc" , devices : & [] }),
        (5231, Vendor { id : 5231u16 , name : "Multi Dimensional Consulting Inc" , devices : & [] }),
        (13875, Vendor { id : 13875u16 , name : "DeepCool" , devices : & [Device { vendor_id : 13875u16 , id : 1u16 , name : "AK400 DIGITAL CPU Air Cooler" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 2u16 , name : "AK620 DIGITAL CPU Air Cooler" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 3u16 , name : "AK500 DIGITAL CPU Air Cooler" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 4u16 , name : "AK500S DIGITAL CPU Air Cooler" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 5u16 , name : "CH560 DIGITAL PC Case" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 6u16 , name : "LS Series (LS520 / LS720 SE DIGITAL) CPU Liquid Cooler" , subsystems : & [] } , Device { vendor_id : 13875u16 , id : 7u16 , name : "MORPHEUS PC Case" , subsystems : & [] }] }),
        (5395, Vendor { id : 5395u16 , name : "Raychem" , devices : & [] }),
        (4138, Vendor { id : 4138u16 , name : "LSI Logic" , devices : & [Device { vendor_id : 4138u16 , id : 0u16 , name : "HYDRA" , subsystems : & [] } , Device { vendor_id : 4138u16 , id : 16u16 , name : "ASPEN" , subsystems : & [] } , Device { vendor_id : 4138u16 , id : 31u16 , name : "AHA-2940U2/U2W /7890/7891 SCSI Controllers" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 15u16 , name : "2940U2W SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 262u16 , name : "2940U2W SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 41344u16 , name : "2940U2W SCSI Controller" }] } , Device { vendor_id : 4138u16 , id : 197u16 , name : "AIC-7899 U160/m SCSI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 197u16 , name : "PowerEdge 2550/2650/4600" }] } , Device { vendor_id : 4138u16 , id : 207u16 , name : "AIC-7899P U160/m" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 262u16 , name : "PowerEdge 4600" } , SubSystem { subvendor : 4136u16 , subdevice : 289u16 , name : "PowerEdge 2650" }] }] }),
        (4881, Vendor { id : 4881u16 , name : "Videoserver, Inc" , devices : & [] }),
        (5265, Vendor { id : 5265u16 , name : "ZEAL Corporation" , devices : & [] }),
        (5797, Vendor { id : 5797u16 , name : "Tekram Technology Co.,Ltd." , devices : & [] }),
        (6841, Vendor { id : 6841u16 , name : "Espia Srl" , devices : & [] }),
        (4227, Vendor { id : 4227u16 , name : "Forex Computer Corporation" , devices : & [Device { vendor_id : 4227u16 , id : 1u16 , name : "FR710" , subsystems : & [] }] }),
        (4415, Vendor { id : 4415u16 , name : "Equinox Systems, Inc." , devices : & [Device { vendor_id : 4415u16 , id : 2056u16 , name : "SST-64P Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 4112u16 , name : "SST-128P Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 32960u16 , name : "SST-16P DB Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 32964u16 , name : "SST-16P RJ Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 32968u16 , name : "SST-16P Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 34952u16 , name : "SST-4P Adapter" , subsystems : & [] } , Device { vendor_id : 4415u16 , id : 37008u16 , name : "SST-8P Adapter" , subsystems : & [] }] }),
        (8260, Vendor { id : 8260u16 , name : "Shenzhen Jiahua Zhongli Technology Co., LTD." , devices : & [Device { vendor_id : 8260u16 , id : 33280u16 , name : "CeaCent CS211X 12G SAS RAID controller" , subsystems : & [SubSystem { subvendor : 8260u16 , subdevice : 8464u16 , name : "CeaCent CS2110-8i" } , SubSystem { subvendor : 8260u16 , subdevice : 8465u16 , name : "CeaCent CS2110-16i" }] } , Device { vendor_id : 8260u16 , id : 33281u16 , name : "CeaCent CS215X 12G SAS RAID controller" , subsystems : & [SubSystem { subvendor : 8260u16 , subdevice : 8528u16 , name : "CeaCent CS2150-8i" } , SubSystem { subvendor : 8260u16 , subdevice : 8529u16 , name : "CeaCent CS2150-16i" }] }] }),
        (51900, Vendor { id : 51900u16 , name : "Cambricon Technologies Corporation Limited" , devices : & [Device { vendor_id : 51900u16 , id : 624u16 , name : "MLU270-S4" , subsystems : & [] } , Device { vendor_id : 51900u16 , id : 880u16 , name : "MLU370-S4" , subsystems : & [] }] }),
        (29527, Vendor { id : 29527u16 , name : "IOxOS Technologies SA" , devices : & [Device { vendor_id : 29527u16 , id : 30992u16 , name : "7910 [Althea]" , subsystems : & [] }] }),
        (4980, Vendor { id : 4980u16 , name : "Silicom Ltd." , devices : & [Device { vendor_id : 4980u16 , id : 36u16 , name : "Silicom Dual port Giga Ethernet BGE Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 37u16 , name : "Silicom Quad port Giga Ethernet BGE Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 38u16 , name : "Silicom Dual port Fiber Giga Ethernet 546 Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 39u16 , name : "Silicom Dual port Fiber LX Giga Ethernet 546 Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 41u16 , name : "Silicom Dual port Copper Giga Ethernet 546GB Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 42u16 , name : "Silicom Dual port Fiber Giga Ethernet 546 TAP/Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 43u16 , name : "Silicom Dual port Copper Fast Ethernet 546 TAP/Bypass Server Adapter (PXE2TBI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 44u16 , name : "Silicom Quad port Copper Giga Ethernet 546GB Bypass Server Adapter (PXG4BPI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 45u16 , name : "Silicom Quad port Fiber-SX Giga Ethernet 546GB Bypass Server Adapter (PXG4BPFI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 46u16 , name : "Silicom Quad port Fiber-LX Giga Ethernet 546GB Bypass Server Adapter (PXG4BPFI-LX)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 47u16 , name : "Silicom Dual port Fiber-SX Giga Ethernet 546GB Low profile Bypass Server Adapter (PXG2BPFIL)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 48u16 , name : "Silicom Dual port Fiber-LX Giga Ethernet 546GB Low profile Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 49u16 , name : "Silicom Quad port Copper Giga Ethernet PCI-E Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 50u16 , name : "Silicom Dual port Copper Fast Ethernet 546 TAP/Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 52u16 , name : "Silicom Dual port Copper Giga Ethernet PCI-E BGE Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 53u16 , name : "Silicom Quad port Copper Giga Ethernet PCI-E BGE Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 54u16 , name : "Silicom Dual port Fiber Giga Ethernet PCI-E BGE Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 55u16 , name : "Silicom Dual port Copper Ethernet PCI-E Intel based Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 56u16 , name : "Silicom Quad port Copper Ethernet PCI-E Intel based Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 57u16 , name : "Silicom Dual port Fiber-SX Ethernet PCI-E Intel based Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 58u16 , name : "Silicom Dual port Fiber-LX Ethernet PCI-E Intel based Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 59u16 , name : "Silicom Dual port Fiber Ethernet PMC Intel based Bypass Server Adapter (PMCX2BPFI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 60u16 , name : "Silicom Dual port Copper Ethernet PCI-X BGE based Bypass Server Adapter (PXG2BPRB)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 61u16 , name : "2-port Copper GBE Bypass with Caviume 1010 PCI-X" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 62u16 , name : "Silicom Dual port Fiber Giga Ethernet PCI-E 571 TAP/Bypass Server Adapter (PEG2TBFI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 63u16 , name : "Silicom Dual port Copper Giga Ethernet PCI-X 546 TAP/Bypass Server Adapter (PXG2TBI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 64u16 , name : "Silicom Quad port Fiber-SX Giga Ethernet 571 Bypass Server Adapter (PEG4BPFI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 66u16 , name : "4-port Copper GBE PMC-X Bypass" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 67u16 , name : "Silicom Quad port Fiber-SX Giga Ethernet 546 Bypass Server Adapter (PXG4BPFID)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 69u16 , name : "Silicom 6 port Copper Giga Ethernet 546 Bypass Server Adapter (PXG6BPI)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 70u16 , name : "4-port bypass PCI-E w disconnect low profile" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 71u16 , name : "Silicom Dual port Fiber-SX Giga Ethernet 571 Bypass Disconnect Server Adapter (PEG2BPFID)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 74u16 , name : "Silicom Quad port Fiber-LX Giga Ethernet 571 Bypass Server Adapter (PEG4BPFI-LX)" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 77u16 , name : "Dual port Copper Giga Ethernet PCI-E Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1025u16 , name : "Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1056u16 , name : "Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1120u16 , name : "Gigabit Ethernet Express Module Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1121u16 , name : "Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1122u16 , name : "Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1136u16 , name : "Octal-port Copper Gigabit Ethernet Express Module Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1154u16 , name : "Dual-port Fiber (SR) 10 Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] } , Device { vendor_id : 4980u16 , id : 1155u16 , name : "Dual-port Fiber (LR) 10 Gigabit Ethernet ExpressModule Bypass Server Adapter" , subsystems : & [] }] }),
        (7481, Vendor { id : 7481u16 , name : "Baikal Electronics" , devices : & [Device { vendor_id : 7481u16 , id : 32864u16 , name : "Baikal-M PCIe Controller" , subsystems : & [] }] }),
        (6385, Vendor { id : 6385u16 , name : "Spectrum GmbH" , devices : & [] }),
        (5448, Vendor { id : 5448u16 , name : "JET PROPULSION Laboratory" , devices : & [] }),
        (5474, Vendor { id : 5474u16 , name : "Symbol Technologies" , devices : & [] }),
        (4282, Vendor { id : 4282u16 , name : "Mitsubishi Electric Corp." , devices : & [Device { vendor_id : 4282u16 , id : 769u16 , name : "AccelGraphics AccelECLIPSE" , subsystems : & [] } , Device { vendor_id : 4282u16 , id : 772u16 , name : "AccelGALAXY A2100 [OEM Evans & Sutherland]" , subsystems : & [] } , Device { vendor_id : 4282u16 , id : 776u16 , name : "Tornado 3000 [OEM Evans & Sutherland]" , subsystems : & [SubSystem { subvendor : 4317u16 , subdevice : 36u16 , name : "Tornado 3000" }] } , Device { vendor_id : 4282u16 , id : 4098u16 , name : "VG500 [VolumePro Volume Rendering Accelerator]" , subsystems : & [] }] }),
        (4141, Vendor { id : 4141u16 , name : "Wyse Technology Inc." , devices : & [Device { vendor_id : 4141u16 , id : 20700u16 , name : "3328 Audio" , subsystems : & [] }] }),
        (4752, Vendor { id : 4752u16 , name : "Sord Computer Corporation" , devices : & [] }),
        (7088, Vendor { id : 7088u16 , name : "SimpliVity Corporation" , devices : & [Device { vendor_id : 7088u16 , id : 2u16 , name : "OmniCube Accelerator OA-3000" , subsystems : & [] } , Device { vendor_id : 7088u16 , id : 16u16 , name : "OmniCube Accelerator OA-3000-2" , subsystems : & [] }] }),
        (4636, Vendor { id : 4636u16 , name : "Nippon Texaco., Ltd" , devices : & [] }),
        (18808, Vendor { id : 18808u16 , name : "Axil Computer Inc" , devices : & [] }),
        (4478, Vendor { id : 4478u16 , name : "T/R Systems" , devices : & [] }),
        (5157, Vendor { id : 5157u16 , name : "Chelsio Communications Inc" , devices : & [Device { vendor_id : 5157u16 , id : 11u16 , name : "T210 Protocol Engine" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 12u16 , name : "T204 Protocol Engine" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 34u16 , name : "10GbE Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 48u16 , name : "T310 10GbE Single Port Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28766u16 , name : "PCIe 10GBase-SR [AD386A]" }] } , Device { vendor_id : 5157u16 , id : 49u16 , name : "T320 10GbE Dual Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 50u16 , name : "T302 1GbE Dual Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 51u16 , name : "T304 1GbE Quad Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 52u16 , name : "B320 10GbE Dual Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 53u16 , name : "S310-CR 10GbE Single Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 54u16 , name : "S320-LP-CR 10GbE Dual Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 55u16 , name : "N320-G2-CR 10GbE Dual Port Adapter" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16385u16 , name : "T420-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16386u16 , name : "T422-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16387u16 , name : "T440-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16388u16 , name : "T420-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16389u16 , name : "T440-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16390u16 , name : "T440-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16391u16 , name : "T420-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16392u16 , name : "T420-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16393u16 , name : "T420-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16394u16 , name : "T404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16395u16 , name : "B420-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16396u16 , name : "B404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16397u16 , name : "T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16398u16 , name : "T440-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16399u16 , name : "T440 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16512u16 , name : "T480-4080 T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16513u16 , name : "T440F-4081 T440-FCoE Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16514u16 , name : "T420-4082  Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16515u16 , name : "T420X-4083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16516u16 , name : "T440-4084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16517u16 , name : "T420-4085 SFP+ Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16518u16 , name : "T440-4086 10Gbase-T Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16519u16 , name : "T440T-4087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 16520u16 , name : "T440-4088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17409u16 , name : "T420-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17410u16 , name : "T422-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17411u16 , name : "T440-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17412u16 , name : "T420-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17413u16 , name : "T440-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17414u16 , name : "T440-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17415u16 , name : "T420-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17416u16 , name : "T420-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17417u16 , name : "T420-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17418u16 , name : "T404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17419u16 , name : "B420-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17420u16 , name : "B404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17421u16 , name : "T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17422u16 , name : "T440-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17423u16 , name : "T440 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17536u16 , name : "T480-4080 T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17537u16 , name : "T440F-4081 T440-FCoE Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17538u16 , name : "T420-4082  Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17539u16 , name : "T420X-4083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17540u16 , name : "T440-4084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17541u16 , name : "T420-4085 SFP+ Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17542u16 , name : "T440-4086 10Gbase-T Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17543u16 , name : "T440T-4087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17544u16 , name : "T440-4088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17665u16 , name : "T420-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17666u16 , name : "T422-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17667u16 , name : "T440-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17668u16 , name : "T420-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17669u16 , name : "T440-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17670u16 , name : "T440-CH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17671u16 , name : "T420-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17672u16 , name : "T420-CX Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17673u16 , name : "T420-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17674u16 , name : "T404-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17675u16 , name : "B420-SR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17676u16 , name : "B404-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17677u16 , name : "T480 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17678u16 , name : "T440-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17679u16 , name : "T440 [Amsterdam] Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17792u16 , name : "T480-4080 T480 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17793u16 , name : "T440F-4081 T440-FCoE Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17794u16 , name : "T420-4082  Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17795u16 , name : "T420X-4083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17796u16 , name : "T440-4084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17797u16 , name : "T420-4085 SFP+ Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17798u16 , name : "T440-4086 10Gbase-T Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17799u16 , name : "T440T-4087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17800u16 , name : "T440-4088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17921u16 , name : "T420-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17922u16 , name : "T422-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17923u16 , name : "T440-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17924u16 , name : "T420-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17925u16 , name : "T440-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17926u16 , name : "T440-CH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17927u16 , name : "T420-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17928u16 , name : "T420-CX Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17929u16 , name : "T420-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17930u16 , name : "T404-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17931u16 , name : "B420-SR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17932u16 , name : "B404-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17933u16 , name : "T480 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17934u16 , name : "T440-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 17935u16 , name : "T440 [Amsterdam] Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18048u16 , name : "T480-4080 T480 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18049u16 , name : "T440F-4081 T440-FCoE Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18050u16 , name : "T420-4082  Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18051u16 , name : "T420X-4083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18052u16 , name : "T440-4084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18053u16 , name : "T420-4085 SFP+ Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18054u16 , name : "T440-4086 10Gbase-T Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18055u16 , name : "T440T-4087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18056u16 , name : "T440-4088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18177u16 , name : "T420-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18178u16 , name : "T422-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18179u16 , name : "T440-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18180u16 , name : "T420-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18181u16 , name : "T440-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18182u16 , name : "T440-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18183u16 , name : "T420-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18184u16 , name : "T420-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18185u16 , name : "T420-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18186u16 , name : "T404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18187u16 , name : "B420-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18188u16 , name : "B404-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18189u16 , name : "T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18190u16 , name : "T440-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18191u16 , name : "T440 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18304u16 , name : "T480-4080 T480 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18305u16 , name : "T440F-4081 T440-FCoE Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18306u16 , name : "T420-4082  Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18307u16 , name : "T420X-4083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18308u16 , name : "T440-4084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18309u16 , name : "T420-4085 SFP+ Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18310u16 , name : "T440-4086 10Gbase-T Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18311u16 , name : "T440T-4087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18312u16 , name : "T440-4088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18433u16 , name : "T420-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18434u16 , name : "T422-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18435u16 , name : "T440-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18436u16 , name : "T420-BCH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18437u16 , name : "T440-BCH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18438u16 , name : "T440-CH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18439u16 , name : "T420-SO Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18440u16 , name : "T420-CX Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18441u16 , name : "T420-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18442u16 , name : "T404-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18443u16 , name : "B420-SR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18444u16 , name : "B404-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18445u16 , name : "T480 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18446u16 , name : "T440-LP-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18447u16 , name : "T440 [Amsterdam] Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18560u16 , name : "T480-4080 T480 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18561u16 , name : "T440F-4081 T440-FCoE Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18562u16 , name : "T420-4082 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18563u16 , name : "T420X-4083 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18564u16 , name : "T440-4084 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18565u16 , name : "T420-4085 SFP+ Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18566u16 , name : "T440-4086 10Gbase-T Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18567u16 , name : "T440T-4087 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 18568u16 , name : "T440-4088 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20481u16 , name : "T520-CR Unified Wire Ethernet Controller" , subsystems : & [SubSystem { subvendor : 6461u16 , subdevice : 4097u16 , name : "510F-B" }] } , Device { vendor_id : 5157u16 , id : 20482u16 , name : "T522-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20483u16 , name : "T540-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20484u16 , name : "T520-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20485u16 , name : "T540-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20486u16 , name : "T540-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20487u16 , name : "T520-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20488u16 , name : "T520-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20489u16 , name : "T520-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20490u16 , name : "T504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20491u16 , name : "B520-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20492u16 , name : "B504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20493u16 , name : "T580-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20494u16 , name : "T540-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20495u16 , name : "T540 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20496u16 , name : "T580-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20497u16 , name : "T520-LL-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20498u16 , name : "T560-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20499u16 , name : "T580-CHR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20500u16 , name : "T580-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20501u16 , name : "T502-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20502u16 , name : "T580-OCP-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20503u16 , name : "T520-OCP-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20504u16 , name : "T540-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20505u16 , name : "T540-LP-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20506u16 , name : "T540-SO-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20507u16 , name : "T540-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20608u16 , name : "T540-5080 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20609u16 , name : "T540-5081 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20610u16 , name : "T504-5082 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20611u16 , name : "T540-5083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20612u16 , name : "T540-5084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20613u16 , name : "T580-5085 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20614u16 , name : "T580-5086 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20615u16 , name : "T580-5087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20616u16 , name : "T570-5088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20617u16 , name : "T520-5089 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20624u16 , name : "T540-5090 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20625u16 , name : "T522-5091 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20626u16 , name : "T520-5092 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20627u16 , name : "T580-5093 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20628u16 , name : "T540-5094 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20629u16 , name : "T540-5095 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20630u16 , name : "T580-5096 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20631u16 , name : "T520-5097 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20632u16 , name : "T580-5098 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20633u16 , name : "T580-5099 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20634u16 , name : "T520-509A Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20635u16 , name : "T540-509B Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20636u16 , name : "T520-509C Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20637u16 , name : "T540-509D Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20638u16 , name : "T520-509E Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20639u16 , name : "T540-509F Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20640u16 , name : "T540-50A0 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20641u16 , name : "T540-50A1 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20642u16 , name : "T580-50A2 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20643u16 , name : "T580-50A3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20644u16 , name : "T540-50A4 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20645u16 , name : "T522-50A5 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20646u16 , name : "T522-50A6 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20647u16 , name : "T580-50A7 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20648u16 , name : "T580-50A8 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20649u16 , name : "T580-50A9 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20650u16 , name : "T580-50AA Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20651u16 , name : "T520-50AB Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20652u16 , name : "T540-50AC Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20653u16 , name : "T520-50AD Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20654u16 , name : "T540-50AE Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20655u16 , name : "T580-50AF Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 20656u16 , name : "T520-50B0 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21505u16 , name : "T520-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21506u16 , name : "T522-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21507u16 , name : "T540-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21508u16 , name : "T520-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21509u16 , name : "T540-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21510u16 , name : "T540-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21511u16 , name : "T520-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21512u16 , name : "T520-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21513u16 , name : "T520-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21514u16 , name : "T504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21515u16 , name : "B520-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21516u16 , name : "B504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21517u16 , name : "T580-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21518u16 , name : "T540-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21519u16 , name : "T540 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21520u16 , name : "T580-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21521u16 , name : "T520-LL-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21522u16 , name : "T560-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21523u16 , name : "T580-CHR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21524u16 , name : "T580-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21525u16 , name : "T502-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21526u16 , name : "T580-OCP-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21527u16 , name : "T520-OCP-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21528u16 , name : "T540-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21529u16 , name : "T540-LP-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21530u16 , name : "T540-SO-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21531u16 , name : "T540-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21632u16 , name : "T540-5080 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21633u16 , name : "T540-5081 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21634u16 , name : "T504-5082 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21635u16 , name : "T540-5083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21636u16 , name : "T540-5084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21637u16 , name : "T580-5085 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21638u16 , name : "T580-5086 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21639u16 , name : "T580-5087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21640u16 , name : "T570-5088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21641u16 , name : "T520-5089 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21648u16 , name : "T540-5090 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21649u16 , name : "T522-5091 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21650u16 , name : "T520-5092 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21651u16 , name : "T580-5093 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21652u16 , name : "T540-5094 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21653u16 , name : "T540-5095 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21654u16 , name : "T580-5096 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21655u16 , name : "T520-5097 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21656u16 , name : "T580-5098 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21657u16 , name : "T580-5099 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21658u16 , name : "T520-509A Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21659u16 , name : "T540-509B Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21660u16 , name : "T520-509C Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21661u16 , name : "T540-509D Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21662u16 , name : "T520-509E Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21663u16 , name : "T540-509F Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21664u16 , name : "T540-50A0 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21665u16 , name : "T540-50A1 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21666u16 , name : "T580-50A2 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21667u16 , name : "T580-50A3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21668u16 , name : "T540-50A4 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21669u16 , name : "T522-50A5 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21670u16 , name : "T522-50A6 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21671u16 , name : "T580-50A7 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21672u16 , name : "T580-50A8 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21673u16 , name : "T580-50A9 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21674u16 , name : "T580-50AA Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21675u16 , name : "T520-50AB Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21676u16 , name : "T540-50AC Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21677u16 , name : "T520-50AD Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21678u16 , name : "T540-50AE Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21679u16 , name : "T580-50AF Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21680u16 , name : "T520-50B0 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21761u16 , name : "T520-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21762u16 , name : "T522-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21763u16 , name : "T540-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21764u16 , name : "T520-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21765u16 , name : "T540-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21766u16 , name : "T540-CH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21767u16 , name : "T520-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21768u16 , name : "T520-CX Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21769u16 , name : "T520-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21770u16 , name : "T504-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21771u16 , name : "B520-SR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21772u16 , name : "B504-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21773u16 , name : "T580-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21774u16 , name : "T540-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21775u16 , name : "T540 [Amsterdam] Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21776u16 , name : "T580-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21777u16 , name : "T520-LL-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21778u16 , name : "T560-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21779u16 , name : "T580-CHR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21780u16 , name : "T580-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21781u16 , name : "T502-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21782u16 , name : "T580-OCP-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21783u16 , name : "T520-OCP-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21784u16 , name : "T540-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21785u16 , name : "T540-LP-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21786u16 , name : "T540-SO-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21787u16 , name : "T540-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21888u16 , name : "T540-5080 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21889u16 , name : "T540-5081 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21890u16 , name : "T504-5082 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21891u16 , name : "T540-5083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21892u16 , name : "T540-5084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21893u16 , name : "T580-5085 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21894u16 , name : "T580-5086 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21895u16 , name : "T580-5087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21896u16 , name : "T570-5088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21897u16 , name : "T520-5089 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21904u16 , name : "T540-5090 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21905u16 , name : "T522-5091 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21906u16 , name : "T520-5092 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21907u16 , name : "T580-5093 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21908u16 , name : "T540-5094 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21909u16 , name : "T540-5095 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21910u16 , name : "T580-5096 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21911u16 , name : "T520-5097 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21912u16 , name : "T580-5098 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21913u16 , name : "T580-5099 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21914u16 , name : "T520-509A Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21915u16 , name : "T540-509B Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21916u16 , name : "T520-509C Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21917u16 , name : "T540-509D Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21918u16 , name : "T520-509E Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21919u16 , name : "T540-509F Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21920u16 , name : "T540-50A0 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21921u16 , name : "T540-50A1 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21922u16 , name : "T580-50A2 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21923u16 , name : "T580-50A3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21924u16 , name : "T540-50A4 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21925u16 , name : "T522-50A5 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21926u16 , name : "T522-50A6 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21927u16 , name : "T580-50A7 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21928u16 , name : "T580-50A8 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21929u16 , name : "T580-50A9 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21930u16 , name : "T580-50AA Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21931u16 , name : "T520-50AB Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21932u16 , name : "T540-50AC Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21933u16 , name : "T520-50AD Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21934u16 , name : "T540-50AE Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21935u16 , name : "T580-50AF Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 21936u16 , name : "T520-50B0 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22017u16 , name : "T520-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22018u16 , name : "T522-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22019u16 , name : "T540-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22020u16 , name : "T520-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22021u16 , name : "T540-BCH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22022u16 , name : "T540-CH Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22023u16 , name : "T520-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22024u16 , name : "T520-CX Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22025u16 , name : "T520-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22026u16 , name : "T504-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22027u16 , name : "B520-SR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22028u16 , name : "B504-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22029u16 , name : "T580-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22030u16 , name : "T540-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22031u16 , name : "T540 [Amsterdam] Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22032u16 , name : "T580-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22033u16 , name : "T520-LL-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22034u16 , name : "T560-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22035u16 , name : "T580-CHR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22036u16 , name : "T580-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22037u16 , name : "T502-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22038u16 , name : "T580-OCP-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22039u16 , name : "T520-OCP-SO Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22040u16 , name : "T540-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22041u16 , name : "T540-LP-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22042u16 , name : "T540-SO-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22043u16 , name : "T540-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22144u16 , name : "T540-5080 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22145u16 , name : "T540-5081 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22146u16 , name : "T504-5082 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22147u16 , name : "T540-5083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22148u16 , name : "T540-5084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22149u16 , name : "T580-5085 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22150u16 , name : "T580-5086 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22151u16 , name : "T580-5087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22152u16 , name : "T570-5088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22153u16 , name : "T520-5089 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22160u16 , name : "T540-5090 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22161u16 , name : "T522-5091 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22162u16 , name : "T520-5092 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22163u16 , name : "T580-5093 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22164u16 , name : "T540-5094 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22165u16 , name : "T540-5095 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22166u16 , name : "T580-5096 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22167u16 , name : "T520-5097 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22168u16 , name : "T580-5098 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22169u16 , name : "T580-5099 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22170u16 , name : "T520-509A Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22171u16 , name : "T540-509B Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22172u16 , name : "T520-509C Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22173u16 , name : "T540-509D Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22174u16 , name : "T520-509E Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22175u16 , name : "T540-509F Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22176u16 , name : "T540-50A0 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22177u16 , name : "T540-50A1 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22178u16 , name : "T580-50A2 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22179u16 , name : "T580-50A3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22180u16 , name : "T540-50A4 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22181u16 , name : "T522-50A5 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22182u16 , name : "T522-50A6 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22183u16 , name : "T580-50A7 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22184u16 , name : "T580-50A8 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22185u16 , name : "T580-50A9 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22186u16 , name : "T580-50AA Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22187u16 , name : "T520-50AB Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22188u16 , name : "T540-50AC Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22189u16 , name : "T520-50AD Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22190u16 , name : "T540-50AE Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22191u16 , name : "T580-50AF Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22192u16 , name : "T520-50B0 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22273u16 , name : "T520-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22274u16 , name : "T522-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22275u16 , name : "T540-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22276u16 , name : "T520-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22277u16 , name : "T540-BCH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22278u16 , name : "T540-CH Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22279u16 , name : "T520-SO Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22280u16 , name : "T520-CX Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22281u16 , name : "T520-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22282u16 , name : "T504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22283u16 , name : "B520-SR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22284u16 , name : "B504-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22285u16 , name : "T580-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22286u16 , name : "T540-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22287u16 , name : "T540 [Amsterdam] Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22288u16 , name : "T580-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22289u16 , name : "T520-LL-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22290u16 , name : "T560-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22291u16 , name : "T580-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22292u16 , name : "T580-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22293u16 , name : "T502-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22400u16 , name : "T540-5080 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22401u16 , name : "T540-5081 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22402u16 , name : "T504-5082 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22403u16 , name : "T540-5083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22404u16 , name : "T580-5084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22405u16 , name : "T580-5085 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22406u16 , name : "T580-5086 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22407u16 , name : "T580-5087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22408u16 , name : "T570-5088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22409u16 , name : "T520-5089 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22416u16 , name : "T540-5090 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22417u16 , name : "T522-5091 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22418u16 , name : "T520-5092 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22419u16 , name : "T580-5093 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22420u16 , name : "T540-5094 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22421u16 , name : "T540-5095 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22422u16 , name : "T580-5096 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22423u16 , name : "T520-5097 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22529u16 , name : "T520-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22530u16 , name : "T522-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22531u16 , name : "T540-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22532u16 , name : "T520-BCH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22533u16 , name : "T540-BCH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22534u16 , name : "T540-CH Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22535u16 , name : "T520-SO Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22536u16 , name : "T520-CX Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22537u16 , name : "T520-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22538u16 , name : "T504-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22539u16 , name : "B520-SR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22540u16 , name : "B504-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22541u16 , name : "T580-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22542u16 , name : "T540-LP-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22543u16 , name : "T540 [Amsterdam] Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22544u16 , name : "T580-LP-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22545u16 , name : "T520-LL-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22546u16 , name : "T560-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22547u16 , name : "T580-CHR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22548u16 , name : "T580-SO-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22549u16 , name : "T502-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22550u16 , name : "T580-OCP-SO Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22551u16 , name : "T520-OCP-SO Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22552u16 , name : "T540-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22553u16 , name : "T540-LP-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22554u16 , name : "T540-SO-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22555u16 , name : "T540-SO-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22656u16 , name : "T540-5080 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22657u16 , name : "T540-5081 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22658u16 , name : "T504-5082 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22659u16 , name : "T540-5083 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22660u16 , name : "T540-5084 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22661u16 , name : "T580-5085 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22662u16 , name : "T580-5086 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22663u16 , name : "T580-5087 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22664u16 , name : "T570-5088 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22665u16 , name : "T520-5089 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22672u16 , name : "T540-5090 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22673u16 , name : "T522-5091 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22674u16 , name : "T520-5092 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22675u16 , name : "T580-5093 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22676u16 , name : "T540-5094 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22677u16 , name : "T540-5095 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22678u16 , name : "T580-5096 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22679u16 , name : "T520-5097 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22680u16 , name : "T580-5098 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22681u16 , name : "T580-5099 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22682u16 , name : "T520-509A Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22683u16 , name : "T540-509B Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22684u16 , name : "T520-509C Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22685u16 , name : "T540-509D Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22686u16 , name : "T520-509E Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22687u16 , name : "T540-509F Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22688u16 , name : "T540-50A0 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22689u16 , name : "T540-50A1 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22690u16 , name : "T580-50A2 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22691u16 , name : "T580-50A3 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22692u16 , name : "T540-50A4 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22693u16 , name : "T522-50A5 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22694u16 , name : "T522-50A6 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22695u16 , name : "T580-50A7 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22696u16 , name : "T580-50A8 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22697u16 , name : "T580-50A9 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22698u16 , name : "T580-50AA Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22699u16 , name : "T520-50AB Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22700u16 , name : "T540-50AC Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22701u16 , name : "T520-50AD Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22702u16 , name : "T540-50AE Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22703u16 , name : "T580-50AF Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 22704u16 , name : "T520-50B0 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24577u16 , name : "T6225-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24578u16 , name : "T6225-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24579u16 , name : "T6425-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24580u16 , name : "T6425-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24581u16 , name : "T6225-SO-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24582u16 , name : "T6225-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24583u16 , name : "T62100-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24584u16 , name : "T62100-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24585u16 , name : "T6210-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24589u16 , name : "T62100-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24593u16 , name : "T6225-LL-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24596u16 , name : "T62100-SO-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24597u16 , name : "T6201-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24704u16 , name : "T6225-6080 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24705u16 , name : "T62100-6081 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24706u16 , name : "T6225-6082 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24707u16 , name : "T62100-6083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24708u16 , name : "T64100-6084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24709u16 , name : "T6240-6085 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24710u16 , name : "T6225-6086 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24711u16 , name : "T6225-6087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24712u16 , name : "T62100-6088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24713u16 , name : "T62100-6089 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24714u16 , name : "T62100-608a Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 24722u16 , name : "T62100-KR Unified Wire Ethernet Controller" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65528u16 , name : "Cosmo Baseboard" } , SubSystem { subvendor : 478u16 , subdevice : 65529u16 , name : "Gimlet Baseboard" }] } , Device { vendor_id : 5157u16 , id : 25601u16 , name : "T6225-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25602u16 , name : "T6225-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25603u16 , name : "T6425-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25604u16 , name : "T6425-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25605u16 , name : "T6225-SO-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25606u16 , name : "T6225-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25607u16 , name : "T62100-LP-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25608u16 , name : "T62100-SO-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25609u16 , name : "T6210-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25613u16 , name : "T62100-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25617u16 , name : "T6225-LL-CR Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25620u16 , name : "T62100-SO-OCP3 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25621u16 , name : "T6201-BT Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25728u16 , name : "T6225-6080 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25729u16 , name : "T62100-6081 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25730u16 , name : "T6225-6082 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25731u16 , name : "T62100-6083 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25732u16 , name : "T64100-6084 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25733u16 , name : "T6240-6085 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25734u16 , name : "T6225-6086 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25735u16 , name : "T6225-6087 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25736u16 , name : "T62100-6088 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25737u16 , name : "T62100-6089 Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25738u16 , name : "T62100-608a Unified Wire Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25746u16 , name : "T62100-KR Unified Wire Ethernet Controller" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65528u16 , name : "Cosmo Baseboard" } , SubSystem { subvendor : 478u16 , subdevice : 65529u16 , name : "Gimlet Baseboard" }] } , Device { vendor_id : 5157u16 , id : 25857u16 , name : "T6225-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25858u16 , name : "T6225-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25859u16 , name : "T6425-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25860u16 , name : "T6425-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25861u16 , name : "T6225-SO-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25862u16 , name : "T6225-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25863u16 , name : "T62100-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25864u16 , name : "T62100-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25865u16 , name : "T6210-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25869u16 , name : "T62100-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25873u16 , name : "T6225-LL-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25876u16 , name : "T62100-SO-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25877u16 , name : "T6201-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25984u16 , name : "T6225-6080 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25985u16 , name : "T62100-6081 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25986u16 , name : "T6225-6082 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25987u16 , name : "T62100-6083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25988u16 , name : "T64100-6084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25989u16 , name : "T6240-6085 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25990u16 , name : "T6225-6086 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25991u16 , name : "T6225-6087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25992u16 , name : "T62100-6088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25993u16 , name : "T62100-6089 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 25994u16 , name : "T62100-608a Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26002u16 , name : "T62100-KR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26113u16 , name : "T6225-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26114u16 , name : "T6225-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26115u16 , name : "T6425-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26116u16 , name : "T6425-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26117u16 , name : "T6225-SO-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26118u16 , name : "T6225-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26119u16 , name : "T62100-LP-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26120u16 , name : "T62100-SO-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26121u16 , name : "T6210-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26125u16 , name : "T62100-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26129u16 , name : "T6225-LL-CR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26132u16 , name : "T62100-SO-OCP3 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26133u16 , name : "T6201-BT Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26240u16 , name : "T6225-6080 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26241u16 , name : "T62100-6081 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26242u16 , name : "T6225-6082 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26243u16 , name : "T62100-6083 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26244u16 , name : "T64100-6084 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26245u16 , name : "T6240-6085 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26246u16 , name : "T6225-6086 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26247u16 , name : "T6225-6087 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26248u16 , name : "T62100-6088 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26249u16 , name : "T62100-6089 Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26250u16 , name : "T62100-608a Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26258u16 , name : "T62100-KR Unified Wire Storage Controller" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26625u16 , name : "T6225-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26626u16 , name : "T6225-SO-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26627u16 , name : "T6425-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26628u16 , name : "T6425-SO-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26629u16 , name : "T6225-SO-OCP3 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26630u16 , name : "T6225-OCP3 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26631u16 , name : "T62100-LP-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26632u16 , name : "T62100-SO-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26633u16 , name : "T6210-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26637u16 , name : "T62100-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26641u16 , name : "T6225-LL-CR Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26644u16 , name : "T62100-SO-OCP3 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26645u16 , name : "T6201-BT Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26752u16 , name : "T6225-6080 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26753u16 , name : "T62100-6081 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26754u16 , name : "T6225-6082 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26755u16 , name : "T62100-6083 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26756u16 , name : "T64100-6084 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26757u16 , name : "T6240-6085 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26758u16 , name : "T6225-6086 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26759u16 , name : "T6225-6087 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26760u16 , name : "T62100-6088 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26761u16 , name : "T62100-6089 Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 26762u16 , name : "T62100-608a Unified Wire Ethernet Controller [VF]" , subsystems : & [] } , Device { vendor_id : 5157u16 , id : 40960u16 , name : "PE10K Unified Wire Ethernet Controller" , subsystems : & [] }] }),
        (5329, Vendor { id : 5329u16 , name : "Computer Hi-Tech Co Ltd" , devices : & [] }),
        (5434, Vendor { id : 5434u16 , name : "ONO SOKKI" , devices : & [] }),
        (4448, Vendor { id : 4448u16 , name : "Megasoft Inc" , devices : & [] }),
        (5832, Vendor { id : 5832u16 , name : "Octasic Inc." , devices : & [] }),
        (4811, Vendor { id : 4811u16 , name : "Antex Electronics Corporation" , devices : & [Device { vendor_id : 4811u16 , id : 39u16 , name : "SC4 (StudioCard)" , subsystems : & [] } , Device { vendor_id : 4811u16 , id : 46u16 , name : "StudioCard 2000" , subsystems : & [] }] }),
        (5421, Vendor { id : 5421u16 , name : "QUANTA Computer Inc" , devices : & [] }),
        (4177, Vendor { id : 4177u16 , name : "Anigma, Inc." , devices : & [] }),
        (5351, Vendor { id : 5351u16 , name : "3CX" , devices : & [] }),
        (7575, Vendor { id : 7575u16 , name : "Shenzhen Longsys Electronics Co., Ltd." , devices : & [Device { vendor_id : 7575u16 , id : 772u16 , name : "HIKSEMI FUTURE Lite NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 4194u16 , name : "Lexar NM710 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 4448u16 , name : "FORESEE P900 BGA NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 4610u16 , name : "Lexar NM610 PRO NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 4836u16 , name : "ORCA 4836 Series eSSD" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 5634u16 , name : "Lexar NM790 / Patriot Viper VP4300 Lite NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 6150u16 , name : "Lexar NM990 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 7575u16 , name : "Lexar NM620 NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 8803u16 , name : "SM2263EN/SM2263XT-based OEM NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 8809u16 , name : "FORESEE XP2000, Lexar NM760 NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 9480u16 , name : "Lexar NM1090 PRO NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 9992u16 , name : "Lexar microSD Express card (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 21014u16 , name : "FORESEE XP1000 / Lexar Professional CFexpress Type B Gold series, NM620 PCIe NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 21024u16 , name : "FORESEE XP2100 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 21046u16 , name : "Lexar NM800 PRO NVME SSD" , subsystems : & [] } , Device { vendor_id : 7575u16 , id : 22386u16 , name : "Lexar NQ700 NVME SSD (DRAM-less)" , subsystems : & [] }] }),
        (4901, Vendor { id : 4901u16 , name : "Salix Technologies, Inc" , devices : & [] }),
        (5556, Vendor { id : 5556u16 , name : "CCI/TRIAD" , devices : & [] }),
        (1905, Vendor { id : 1905u16 , name : "Xi'an Microelectronics Technology Institute" , devices : & [] }),
        (5130, Vendor { id : 5130u16 , name : "DSP Research Inc" , devices : & [] }),
        (4259, Vendor { id : 4259u16 , name : "Everex Systems Inc" , devices : & [] }),
        (8262, Vendor { id : 8262u16 , name : "GXMICRO Technology (Shanghai) Co., Ltd." , devices : & [] }),
        (4943, Vendor { id : 4943u16 , name : "Algo System Co Ltd" , devices : & [] }),
        (4552, Vendor { id : 4552u16 , name : "Dolphin Interconnect Solutions AS" , devices : & [Device { vendor_id : 4552u16 , id : 1624u16 , name : "PSB32 SCI-Adapter D31x" , subsystems : & [] } , Device { vendor_id : 4552u16 , id : 54885u16 , name : "PSB64 SCI-Adapter D32x" , subsystems : & [] } , Device { vendor_id : 4552u16 , id : 54887u16 , name : "PSB66 SCI-Adapter D33x" , subsystems : & [] }] }),
        (4492, Vendor { id : 4492u16 , name : "Corollary, Inc" , devices : & [Device { vendor_id : 4492u16 , id : 20u16 , name : "PCIB [C-bus II to PCI bus host bridge chip]" , subsystems : & [] } , Device { vendor_id : 4492u16 , id : 4375u16 , name : "Intel 8-way XEON Profusion Chipset [Cache Coherency Filter]" , subsystems : & [] }] }),
        (4707, Vendor { id : 4707u16 , name : "Sonic Solutions" , devices : & [] }),
        (58559, Vendor { id : 58559u16 , name : "EKF Elektronik GmbH" , devices : & [Device { vendor_id : 58559u16 , id : 3277u16 , name : "CCD-CALYPSO" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 3281u16 , name : "CD1-OPERA" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 3282u16 , name : "CD2-BEBOP" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 3283u16 , name : "CD3-JIVE" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 20673u16 , name : "PC1-GROOVE" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 20674u16 , name : "PC2-LIMBO" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 21441u16 , name : "SC1-ALLEGRO" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 52295u16 , name : "CCG-RUMBA" , subsystems : & [] } , Device { vendor_id : 58559u16 , id : 52301u16 , name : "CCM-BOOGIE" , subsystems : & [] }] }),
        (4404, Vendor { id : 4404u16 , name : "Mercury Computer Systems" , devices : & [Device { vendor_id : 4404u16 , id : 1u16 , name : "Raceway Bridge" , subsystems : & [] } , Device { vendor_id : 4404u16 , id : 2u16 , name : "Dual PCI to RapidIO Bridge" , subsystems : & [] } , Device { vendor_id : 4404u16 , id : 11u16 , name : "POET Serial RapidIO Bridge" , subsystems : & [] } , Device { vendor_id : 4404u16 , id : 13u16 , name : "POET PSDMS Device" , subsystems : & [] }] }),
        (8195, Vendor { id : 8195u16 , name : "Smart Link Ltd." , devices : & [Device { vendor_id : 8195u16 , id : 34816u16 , name : "LM-I56N" , subsystems : & [] }] }),
        (4597, Vendor { id : 4597u16 , name : "Computing Devices International" , devices : & [] }),
        (6655, Vendor { id : 6655u16 , name : "Eclipse Electronic Systems, Inc." , devices : & [] }),
        (4687, Vendor { id : 4687u16 , name : "Infortrend Technology, Inc." , devices : & [Device { vendor_id : 4687u16 , id : 65u16 , name : "IFT-2000 Series RAID Controller" , subsystems : & [] }] }),
        (4637, Vendor { id : 4637u16 , name : "LiPPERT ADLINK Technology GmbH" , devices : & [] }),
        (4500, Vendor { id : 4500u16 , name : "Toucan Technology" , devices : & [] }),
        (7631, Vendor { id : 7631u16 , name : "Beijing Sinead Technology Co., Ltd." , devices : & [] }),
        (7959, Vendor { id : 7959u16 , name : "Zettastone Technology" , devices : & [] }),
        (5269, Vendor { id : 5269u16 , name : "TOKAI Communications Industry Co. Ltd" , devices : & [] }),
        (4560, Vendor { id : 4560u16 , name : "Lockheed Martin Federal Systems-Manassas" , devices : & [] }),
        (7028, Vendor { id : 7028u16 , name : "OpenVox Communication Co. Ltd." , devices : & [Device { vendor_id : 7028u16 , id : 277u16 , name : "D115P/D115E Single-port E1/T1 card" , subsystems : & [] } , Device { vendor_id : 7028u16 , id : 53552u16 , name : "D130P/D130E Single-port E1/T1 card (3rd GEN)" , subsystems : & [] } , Device { vendor_id : 7028u16 , id : 53776u16 , name : "D210P/D210E Dual-port E1/T1 card(2nd generation)" , subsystems : & [] } , Device { vendor_id : 7028u16 , id : 53808u16 , name : "D230 Dual-port E1/T1 card (2nd generation)" , subsystems : & [] } , Device { vendor_id : 7028u16 , id : 54288u16 , name : "D410/430 Quad-port E1/T1 card" , subsystems : & [] } , Device { vendor_id : 7028u16 , id : 54320u16 , name : "D410/430 Quad-port E1/T1 card" , subsystems : & [] }] }),
        (6226, Vendor { id : 6226u16 , name : "Anritsu Corp." , devices : & [] }),
        (4199, Vendor { id : 4199u16 , name : "Mitsubishi Electric" , devices : & [Device { vendor_id : 4199u16 , id : 769u16 , name : "AccelGraphics AccelECLIPSE" , subsystems : & [] } , Device { vendor_id : 4199u16 , id : 772u16 , name : "AccelGALAXY A2100 [OEM Evans & Sutherland]" , subsystems : & [] } , Device { vendor_id : 4199u16 , id : 776u16 , name : "Tornado 3000 [OEM Evans & Sutherland]" , subsystems : & [] } , Device { vendor_id : 4199u16 , id : 4098u16 , name : "VG500 [VolumePro Volume Rendering Accelerator]" , subsystems : & [] }] }),
        (657, Vendor { id : 657u16 , name : "Davicom Semiconductor, Inc. (Wrong ID)" , devices : & [] }),
        (2697, Vendor { id : 2697u16 , name : "BREA Technologies Inc" , devices : & [] }),
        (5117, Vendor { id : 5117u16 , name : "Micro Science Inc" , devices : & [] }),
        (5047, Vendor { id : 5047u16 , name : "Logic Devices Inc" , devices : & [] }),
        (4962, Vendor { id : 4962u16 , name : "Fujifacom Corporation" , devices : & [] }),
        (4975, Vendor { id : 4975u16 , name : "Applied Magic Inc" , devices : & [] }),
        (4821, Vendor { id : 4821u16 , name : "Equator Technologies Inc" , devices : & [Device { vendor_id : 4821u16 , id : 3u16 , name : "BSP16" , subsystems : & [] } , Device { vendor_id : 4821u16 , id : 4096u16 , name : "BSP15" , subsystems : & [] }] }),
        (4198, Vendor { id : 4198u16 , name : "PicoPower Technology" , devices : & [Device { vendor_id : 4198u16 , id : 0u16 , name : "PT80C826" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 1u16 , name : "PT86C521 [Vesuvius v1] Host Bridge" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 2u16 , name : "PT86C523 [Vesuvius v3] PCI-ISA Bridge Master" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 3u16 , name : "PT86C524 [Nile] PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 4u16 , name : "PT86C525 [Nile-II] PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 5u16 , name : "National PC87550 System Controller" , subsystems : & [] } , Device { vendor_id : 4198u16 , id : 32770u16 , name : "PT86C523 [Vesuvius v3] PCI-ISA Bridge Slave" , subsystems : & [] }] }),
        (4589, Vendor { id : 4589u16 , name : "Mediamatics" , devices : & [] }),
        (5727, Vendor { id : 5727u16 , name : "Linux Media Labs, LLC" , devices : & [Device { vendor_id : 5727u16 , id : 4128u16 , name : "LMLM4 MPEG-4 encoder" , subsystems : & [] }] }),
        (7060, Vendor { id : 7060u16 , name : "Signatec / Dynamic Signals Corp" , devices : & [Device { vendor_id : 7060u16 , id : 58368u16 , name : "PX14400 Dual Xilinx Virtex5 based Digitizer" , subsystems : & [] }] }),
        (4823, Vendor { id : 4823u16 , name : "Biotronic SRL" , devices : & [] }),
        (6023, Vendor { id : 6023u16 , name : "Hightech Information System Ltd." , devices : & [] }),
        (6306, Vendor { id : 6306u16 , name : "Stretch Inc." , devices : & [Device { vendor_id : 6306u16 , id : 2u16 , name : "VRC6016 16-Channel PCIe DVR Card" , subsystems : & [] }] }),
        (6460, Vendor { id : 6460u16 , name : "MAXIM Integrated Products" , devices : & [] }),
        (21590, Vendor { id : 21590u16 , name : "GoTView" , devices : & [] }),
        (5189, Vendor { id : 5189u16 , name : "Logical DO Ltd" , devices : & [] }),
        (4496, Vendor { id : 4496u16 , name : "Tripace" , devices : & [Device { vendor_id : 4496u16 , id : 50993u16 , name : "TP-910/920/940 PCI Ultra(Wide) SCSI Adapter" , subsystems : & [] }] }),
        (33810, Vendor { id : 33810u16 , name : "Sharetronic Data Technology Co., Ltd." , devices : & [] }),
        (4126, Vendor { id : 4126u16 , name : "American Megatrends Inc." , devices : & [Device { vendor_id : 4126u16 , id : 9u16 , name : "MegaRAID 428 Ultra RAID Controller (rev 03)" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 6496u16 , name : "MegaRAID" , subsystems : & [SubSystem { subvendor : 4126u16 , subdevice : 1137u16 , name : "MegaRAID 471 Enterprise 1600 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1141u16 , name : "MegaRAID 475 Express 500/500LC RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1143u16 , name : "MegaRAID 477 Elite 3100 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1171u16 , name : "MegaRAID 493 Elite 1600 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1172u16 , name : "MegaRAID 494 Elite 1650 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1283u16 , name : "MegaRAID 503 Enterprise 1650 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1297u16 , name : "MegaRAID 511 i4 IDE RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1314u16 , name : "MegaRAID 522 i4133 RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 1137u16 , name : "PowerEdge RAID Controller 3/QC" } , SubSystem { subvendor : 4136u16 , subdevice : 1141u16 , name : "PowerEdge RAID Controller 3/SC" } , SubSystem { subvendor : 4136u16 , subdevice : 1171u16 , name : "PowerEdge RAID Controller 3/DC" } , SubSystem { subvendor : 4136u16 , subdevice : 1297u16 , name : "PowerEdge Cost Effective RAID Controller ATA100/4Ch" } , SubSystem { subvendor : 4156u16 , subdevice : 24807u16 , name : "NetRAID-1M" } , SubSystem { subvendor : 4156u16 , subdevice : 24808u16 , name : "NetRaid 2M [AMI MegaRaid 493]" }] } , Device { vendor_id : 4126u16 , id : 36880u16 , name : "MegaRAID 428 Ultra RAID Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36912u16 , name : "EIDE Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36913u16 , name : "EIDE Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36914u16 , name : "EIDE & SCSI Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36915u16 , name : "SCSI Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36928u16 , name : "Multimedia card" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36960u16 , name : "MegaRAID 434 Ultra GT RAID Controller" , subsystems : & [] } , Device { vendor_id : 4126u16 , id : 36963u16 , name : "MegaRAC" , subsystems : & [SubSystem { subvendor : 4126u16 , subdevice : 1895u16 , name : "Dell Remote Assistant Card 2" }] }] }),
        (12943, Vendor { id : 12943u16 , name : "Shenzhen EMEET Technology Co., Ltd." , devices : & [Device { vendor_id : 12943u16 , id : 76u16 , name : "OfficeCore M1A" , subsystems : & [] } , Device { vendor_id : 12943u16 , id : 8217u16 , name : "REC 600 HD Webcam" , subsystems : & [] }] }),
        (4596, Vendor { id : 4596u16 , name : "Kinetic Systems Corporation" , devices : & [Device { vendor_id : 4596u16 , id : 10517u16 , name : "CAMAC controller" , subsystems : & [] }] }),
        (5453, Vendor { id : 5453u16 , name : "PAN INTERNATIONAL Industrial Corp" , devices : & [] }),
        (4120, Vendor { id : 4120u16 , name : "Unisys Systems" , devices : & [] }),
        (4313, Vendor { id : 4313u16 , name : "Macronix, Inc. [MXIC]" , devices : & [Device { vendor_id : 4313u16 , id : 1073u16 , name : "MX98715" , subsystems : & [] } , Device { vendor_id : 4313u16 , id : 1298u16 , name : "MX98713" , subsystems : & [] } , Device { vendor_id : 4313u16 , id : 1329u16 , name : "MX987x5" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 4608u16 , name : "DFE-540TX ProFAST 10/100 Adapter" }] } , Device { vendor_id : 4313u16 , id : 34341u16 , name : "MX86250" , subsystems : & [] } , Device { vendor_id : 4313u16 , id : 34342u16 , name : "Macronix MX86251 + 3Dfx Voodoo Rush" , subsystems : & [] } , Device { vendor_id : 4313u16 , id : 34952u16 , name : "MX86200" , subsystems : & [] }] }),
        (7357, Vendor { id : 7357u16 , name : "novachips" , devices : & [] }),
        (7227, Vendor { id : 7227u16 , name : "Accensus, LLC" , devices : & [Device { vendor_id : 7227u16 , id : 512u16 , name : "Telas2" , subsystems : & [] } , Device { vendor_id : 7227u16 , id : 768u16 , name : "Telas 2.V" , subsystems : & [] }] }),
        (7917, Vendor { id : 7917u16 , name : "XDX Computing Technology Ltd." , devices : & [Device { vendor_id : 7917u16 , id : 4256u16 , name : "XDX110 Graphic/VGA Controller" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4257u16 , name : "XDX110 Audio Controller" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4258u16 , name : "XDX110M" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4260u16 , name : "XDX E1100" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4352u16 , name : "XDX110" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4416u16 , name : "XDX120" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4418u16 , name : "XDX120M" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4420u16 , name : "XDX E1200" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4432u16 , name : "XDX120S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4448u16 , name : "XDX121" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4464u16 , name : "XDX121S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4576u16 , name : "XDX130" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4580u16 , name : "XDX E1300" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4896u16 , name : "XDX150" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4899u16 , name : "XDX R1500" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4900u16 , name : "XDX X1500" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4912u16 , name : "XDX150S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4915u16 , name : "XDX R1510" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4928u16 , name : "XDX151" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4944u16 , name : "XDX151S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 4960u16 , name : "XDX151T" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5056u16 , name : "XDX160T" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5072u16 , name : "XDX160S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5075u16 , name : "XDX R1610" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5376u16 , name : "XDX180" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5379u16 , name : "XDX R1800" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5380u16 , name : "XDX X1800" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5392u16 , name : "XDX180S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5536u16 , name : "XDX190" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5539u16 , name : "XDX R1900" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5540u16 , name : "XDX X1900" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5541u16 , name : "XDX X1900D" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 5552u16 , name : "XDX190S" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 6160u16 , name : "XDX TJ01 Audio" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 6176u16 , name : "XDX TJ02 Audio" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 6192u16 , name : "XDX TJ03 Audio" , subsystems : & [] } , Device { vendor_id : 7917u16 , id : 8448u16 , name : "XDX290" , subsystems : & [] }] }),
        (5364, Vendor { id : 5364u16 , name : "TOKYO Electronic Industry CO Ltd" , devices : & [] }),
        (36991, Vendor { id : 36991u16 , name : "Atronics" , devices : & [Device { vendor_id : 36991u16 , id : 8213u16 , name : "IDE-2015PL" , subsystems : & [] }] }),
        (7676, Vendor { id : 7676u16 , name : "JSC NT-COM" , devices : & [Device { vendor_id : 7676u16 , id : 4481u16 , name : "TDM 8 Port E1/T1/J1 Adapter" , subsystems : & [] }] }),
        (4300, Vendor { id : 4300u16 , name : "Mai Logic Incorporated" , devices : & [Device { vendor_id : 4300u16 , id : 1632u16 , name : "Articia S Host Bridge" , subsystems : & [] } , Device { vendor_id : 4300u16 , id : 1633u16 , name : "Articia S PCI Bridge" , subsystems : & [] }] }),
        (5605, Vendor { id : 5605u16 , name : "Valley technologies Inc" , devices : & [] }),
        (4132, Vendor { id : 4132u16 , name : "Beijing Dajia Internet Information Technology Co." , devices : & [Device { vendor_id : 4132u16 , id : 257u16 , name : "StreamLake 200 AI-VPU" , subsystems : & [SubSystem { subvendor : 4132u16 , subdevice : 513u16 , name : "SL200-NP" } , SubSystem { subvendor : 4132u16 , subdevice : 769u16 , name : "SL200-P" }] }] }),
        (4761, Vendor { id : 4761u16 , name : "Knowledge Technology Lab." , devices : & [] }),
        (113, Vendor { id : 113u16 , name : "Nebula Electronics Ltd." , devices : & [] }),
        (5242, Vendor { id : 5242u16 , name : "FORMOSA Industrial Computing" , devices : & [] }),
        (4238, Vendor { id : 4238u16 , name : "Oracle/SUN" , devices : & [Device { vendor_id : 4238u16 , id : 1u16 , name : "EBUS" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 4096u16 , name : "STP2003QFP [PCIO] EBUS" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 4097u16 , name : "Happy Meal 10/100 Ethernet [hme]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 4352u16 , name : "RIO EBUS" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 4352u16 , name : "RIO EBUS on Blade 100 motherboard" }] } , Device { vendor_id : 4238u16 , id : 4353u16 , name : "RIO 10/100 Ethernet [eri]" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 4353u16 , name : "RIO GEM on Blade 100 motherboard" }] } , Device { vendor_id : 4238u16 , id : 4354u16 , name : "RIO 1394" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 4354u16 , name : "RIO 1394 on Blade 100 motherboard" }] } , Device { vendor_id : 4238u16 , id : 4355u16 , name : "RIO USB" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 4355u16 , name : "RIO USB on Blade 100 motherboard" }] } , Device { vendor_id : 4238u16 , id : 5703u16 , name : "Broadcom 570x 10/100/1000 Ethernet [bge]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 5704u16 , name : "Broadcom 570x 10/100/1000 Ethernet [bge]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 5799u16 , name : "Broadcom 570x 10/100/1000 Ethernet [bge]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 5800u16 , name : "Broadcom 570x 10/100/1000 Ethernet [bge]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 11181u16 , name : "GEM 10/100/1000 Ethernet [ge]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 20480u16 , name : "Simba Advanced PCI Bridge" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 20480u16 , name : "Netra AX1105-500" }] } , Device { vendor_id : 4238u16 , id : 20547u16 , name : "SunPCI Co-processor" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 23712u16 , name : "Crypto Accelerator 6000 [mca]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25344u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25345u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25346u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25347u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25360u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25361u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25362u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25363u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25376u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25379u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25392u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25393u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25394u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25395u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25408u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25411u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25424u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 25427u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 26402u16 , name : "Intel 21554 PCI-PCI bus bridge [db21554]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 26478u16 , name : "SunPCiIII" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 28771u16 , name : "SunPCiII / SunPCiIIpro" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 32768u16 , name : "Psycho PCI Bus Module" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 32769u16 , name : "Schizo PCI Bus Module" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 32770u16 , name : "Schizo+ PCI Bus Module" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 33008u16 , name : "PCIe switch [px]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 33016u16 , name : "PCIe switch [px]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 36880u16 , name : "PCIe/PCI bridge switch [pxb_plx]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 36896u16 , name : "PCIe/PCI bridge switch [pxb_plx]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 37122u16 , name : "Davicom Fast Ethernet driver for Davicom DM9102A [dmfe]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 40960u16 , name : "Psycho UPA-PCI Bus Module [pcipsy]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 40961u16 , name : "Psycho UPA-PCI Bus Module [pcipsy]" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 40961u16 , name : "Ultra IIe on Blade 100 motherboard" }] } , Device { vendor_id : 4238u16 , id : 43009u16 , name : "Schizo Fireplane-PCI bus bridge module [pcisch]" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 43690u16 , name : "Multithreaded Shared 10GbE Ethernet Network Controller" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 43962u16 , name : "Cassini 10/100/1000" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 43981u16 , name : "Multithreaded 10-Gigabit Ethernet Network Controller" , subsystems : & [] } , Device { vendor_id : 4238u16 , id : 50198u16 , name : "Sun Fire System/System Controller Interface chip [sbbc]" , subsystems : & [] }] }),
        (52428, Vendor { id : 52428u16 , name : "Catapult Communications" , devices : & [] }),
        (6502, Vendor { id : 6502u16 , name : "Orad Hi-Tec Systems" , devices : & [Device { vendor_id : 6502u16 , id : 6517u16 , name : "DVG64 family" , subsystems : & [] } , Device { vendor_id : 6502u16 , id : 6519u16 , name : "DVG128 family" , subsystems : & [] } , Device { vendor_id : 6502u16 , id : 6521u16 , name : "3DVG/UHD3" , subsystems : & [] } , Device { vendor_id : 6502u16 , id : 6528u16 , name : "HDV2/UHD2" , subsystems : & [SubSystem { subvendor : 4660u16 , subdevice : 12640u16 , name : "UHD2LC" } , SubSystem { subvendor : 4660u16 , subdevice : 13056u16 , name : "Legacy UHD2" } , SubSystem { subvendor : 4660u16 , subdevice : 13328u16 , name : "UHD2" }] }] }),
        (4793, Vendor { id : 4793u16 , name : "3Com Corp, Modem Division" , devices : & [Device { vendor_id : 4793u16 , id : 4102u16 , name : "WinModem" , subsystems : & [SubSystem { subvendor : 4793u16 , subdevice : 92u16 , name : "USR 56k Internal Voice WinModem (Model 3472)" } , SubSystem { subvendor : 4793u16 , subdevice : 94u16 , name : "USR 56k Internal WinModem (Models 662975)" } , SubSystem { subvendor : 4793u16 , subdevice : 98u16 , name : "USR 56k Internal Voice WinModem (Model 662978)" } , SubSystem { subvendor : 4793u16 , subdevice : 104u16 , name : "USR 56k Internal Voice WinModem (Model 5690)" } , SubSystem { subvendor : 4793u16 , subdevice : 122u16 , name : "USR 56k Internal Voice WinModem (Model 662974)" } , SubSystem { subvendor : 4793u16 , subdevice : 127u16 , name : "USR 56k Internal WinModem (Models 5698, 5699)" } , SubSystem { subvendor : 4793u16 , subdevice : 128u16 , name : "USR 56k Internal WinModem (Models 2975, 3528)" } , SubSystem { subvendor : 4793u16 , subdevice : 129u16 , name : "USR 56k Internal Voice WinModem (Models 2974, 3529)" } , SubSystem { subvendor : 4793u16 , subdevice : 145u16 , name : "USR 56k Internal Voice WinModem (Model 2978)" }] } , Device { vendor_id : 4793u16 , id : 4103u16 , name : "USR 56k Internal WinModem" , subsystems : & [SubSystem { subvendor : 4793u16 , subdevice : 163u16 , name : "USR 56k Internal WinModem (Model 3595)" } , SubSystem { subvendor : 4793u16 , subdevice : 196u16 , name : "U.S. Robotics V.92 Voice Faxmodem (2884A/B/C)" }] } , Device { vendor_id : 4793u16 , id : 4104u16 , name : "56K FaxModem Model 5610" , subsystems : & [SubSystem { subvendor : 4793u16 , subdevice : 162u16 , name : "USR 56k Internal FAX Modem (Model 2977)" } , SubSystem { subvendor : 4793u16 , subdevice : 170u16 , name : "USR 56k Internal Voice Modem (Model 2976)" } , SubSystem { subvendor : 4793u16 , subdevice : 171u16 , name : "USR 56k Internal Voice Modem (Model 5609)" } , SubSystem { subvendor : 4793u16 , subdevice : 172u16 , name : "USR 56k Internal Voice Modem (Model 3298)" } , SubSystem { subvendor : 4793u16 , subdevice : 173u16 , name : "USR 56k Internal FAX Modem (Model 5610)" } , SubSystem { subvendor : 4793u16 , subdevice : 211u16 , name : "USR 56K Internal V92 FAX Modem (Model 5610)" } , SubSystem { subvendor : 4793u16 , subdevice : 47802u16 , name : "USR 56K Internal Voice Modem 3CP3298-DEL (Model 5601) [Hawk]" }] }] }),
        (49321, Vendor { id : 49321u16 , name : "Micron/Crucial Technology" , devices : & [Device { vendor_id : 49321u16 , id : 8803u16 , name : "P1 NVMe PCIe SSD[Frampton]" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21507u16 , name : "P1 NVMe PCIe SSD[Frampton2]" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21511u16 , name : "P5 Plus NVMe PCIe SSD" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21514u16 , name : "P2 [Nick P2] / P3 / P3 Plus NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21522u16 , name : "P5 NVMe PCIe SSD[SlashP5]" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21525u16 , name : "T500 NVMe PCIe SSD" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21529u16 , name : "T700 NVMe PCIe SSD" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21536u16 , name : "P3 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21537u16 , name : "P3 Plus NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21542u16 , name : "P310 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21543u16 , name : "P310 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21544u16 , name : "T710 NVMe PCIe SSD" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 21547u16 , name : "T705 NVMe PCIe SSD" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 22026u16 , name : "P510 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 49321u16 , id : 22027u16 , name : "E100 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] }] }),
        (4407, Vendor { id : 4407u16 , name : "Cisco Systems Inc" , devices : & [Device { vendor_id : 4407u16 , id : 35u16 , name : "VIC 81 PCIe Upstream Port" , subsystems : & [] } , Device { vendor_id : 4407u16 , id : 64u16 , name : "VIC PCIe Upstream Port" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 79u16 , name : "VIC 1280 Dual 40Gb Mezzanine" } , SubSystem { subvendor : 4407u16 , subdevice : 132u16 , name : "VIC 1240 Dual 40Gb MLOM" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 Dual 10Gb SFP+ PCIe" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 Dual 40Gb QSFP+ PCIe" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T Dual 10GBaseT PCIe" } , SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 Dual 40Gb" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 Dual 40Gb MLOM" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 Dual 10Gb SFP+ PCIe" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Dual 40Gb Mezzanine" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 Dual 40Gb PCIe" }] } , Device { vendor_id : 4407u16 , id : 65u16 , name : "VIC PCIe Downstream Port" , subsystems : & [] } , Device { vendor_id : 4407u16 , id : 66u16 , name : "VIC Management Controller" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 71u16 , name : "VIC P81E PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 535u16 , name : "VIC 1455 PCIe Management Controller" } , SubSystem { subvendor : 4407u16 , subdevice : 536u16 , name : "VIC 1457 PCIe Management Controller" }] } , Device { vendor_id : 4407u16 , id : 67u16 , name : "VIC Ethernet NIC" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 71u16 , name : "VIC P81E PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 72u16 , name : "VIC M81KR Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 79u16 , name : "VIC 1280 Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 132u16 , name : "VIC 1240 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 349u16 , name : "VIC 1387 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 533u16 , name : "VIC 1440 Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 534u16 , name : "VIC 1480 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 535u16 , name : "VIC 1455 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 536u16 , name : "VIC 1457 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 537u16 , name : "VIC 1485 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 538u16 , name : "VIC 1487 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 586u16 , name : "VIC 1495 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 587u16 , name : "VIC 1497 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 687u16 , name : "VIC 1467 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 688u16 , name : "VIC 1477 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 719u16 , name : "VIC 14425 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 720u16 , name : "VIC 14825 Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 731u16 , name : "VIC 15231 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 732u16 , name : "VIC 15411 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 733u16 , name : "VIC 15428 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 734u16 , name : "VIC 15420 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 735u16 , name : "VIC 15230 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 736u16 , name : "VIC 15427 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 737u16 , name : "VIC 15422 Mezzanine Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 740u16 , name : "VIC 15235 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 744u16 , name : "VIC 15238 MLOM Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 754u16 , name : "VIC 15425 PCIe Ethernet NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 755u16 , name : "VIC 15237 MLOM Ethernet NIC" }] } , Device { vendor_id : 4407u16 , id : 68u16 , name : "VIC Ethernet NIC Dynamic" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 71u16 , name : "VIC P81E PCIe Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 72u16 , name : "VIC M81KR Mezzanine Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 79u16 , name : "VIC 1280 Mezzanine Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 132u16 , name : "VIC 1240 MLOM Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 PCIe Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 PCIe Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T PCIe Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 MLOM Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 PCIe Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Mezzanine Ethernet NIC Dynamic" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 PCIe Ethernet NIC Dynamic" }] } , Device { vendor_id : 4407u16 , id : 69u16 , name : "VIC FCoE HBA" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 71u16 , name : "VIC P81E PCIe FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 72u16 , name : "VIC M81KR Mezzanine FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 79u16 , name : "VIC 1280 Mezzanine FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 132u16 , name : "VIC 1240 MLOM FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 PCIe FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 PCIe FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T PCIe FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 MLOM FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 PCIe FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Mezzanine FCoE HBA" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 PCIe FCoE HBA" }] } , Device { vendor_id : 4407u16 , id : 70u16 , name : "VIC SCSI Controller" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 SCSI Controller" }] } , Device { vendor_id : 4407u16 , id : 78u16 , name : "VIC 82 PCIe Upstream Port" , subsystems : & [] } , Device { vendor_id : 4407u16 , id : 113u16 , name : "VIC SR-IOV VF" , subsystems : & [] } , Device { vendor_id : 4407u16 , id : 122u16 , name : "VIC 1300 PCIe Upstream Port" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 Dual 40Gb" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 Dual 40Gb MLOM" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Dual 40Gb Mezzanine" } , SubSystem { subvendor : 4407u16 , subdevice : 333u16 , name : "VIC 1385 Dual 40Gb PCIe" }] } , Device { vendor_id : 4407u16 , id : 207u16 , name : "VIC Userspace NIC" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 79u16 , name : "VIC 1280 Mezzanine Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 132u16 , name : "VIC 1240 MLOM Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 133u16 , name : "VIC 1225 PCIe Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 205u16 , name : "VIC 1285 PCIe Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 206u16 , name : "VIC 1225T PCIe Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 298u16 , name : "VIC M4308 Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 300u16 , name : "VIC 1340 MLOM Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 302u16 , name : "VIC 1227 PCIe Userspace NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 311u16 , name : "VIC 1380 Mezzanine Userspace NIC" }] } , Device { vendor_id : 4407u16 , id : 574u16 , name : "1GigE I350 LOM" , subsystems : & [] } , Device { vendor_id : 4407u16 , id : 695u16 , name : "VIC SR-IOV Ethernet VF" , subsystems : & [] }] }),
        (23246, Vendor { id : 23246u16 , name : "Beholder International Ltd." , devices : & [] }),
        (4729, Vendor { id : 4729u16 , name : "Transmeta Corporation" , devices : & [Device { vendor_id : 4729u16 , id : 96u16 , name : "TM8000 Northbridge" , subsystems : & [] } , Device { vendor_id : 4729u16 , id : 97u16 , name : "TM8000 AGP bridge" , subsystems : & [] } , Device { vendor_id : 4729u16 , id : 661u16 , name : "Northbridge" , subsystems : & [] } , Device { vendor_id : 4729u16 , id : 917u16 , name : "LongRun Northbridge" , subsystems : & [] } , Device { vendor_id : 4729u16 , id : 918u16 , name : "SDRAM controller" , subsystems : & [] } , Device { vendor_id : 4729u16 , id : 919u16 , name : "BIOS scratchpad" , subsystems : & [] }] }),
        (5719, Vendor { id : 5719u16 , name : "Cavium QLogic" , devices : & [Device { vendor_id : 5719u16 , id : 19u16 , name : "425/825/42B/82B 4Gbps/8Gbps PCIe dual port FC HBA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5954u16 , name : "82B 8Gbps dual port FC HBA" } , SubSystem { subvendor : 4156u16 , subdevice : 5956u16 , name : "42B 4Gbps dual port FC HBA" } , SubSystem { subvendor : 5719u16 , subdevice : 20u16 , name : "425/825 4Gbps/8Gbps PCIe dual port FC HBA" }] } , Device { vendor_id : 5719u16 , id : 20u16 , name : "1010/1020/1007/1741 10Gbps CNA" , subsystems : & [SubSystem { subvendor : 5719u16 , subdevice : 20u16 , name : "1010/1020/1007/1741 10Gbps CNA - FCOE" } , SubSystem { subvendor : 5719u16 , subdevice : 21u16 , name : "1010/1020/1007/1741 10Gbps CNA - LL" }] } , Device { vendor_id : 5719u16 , id : 23u16 , name : "415/815/41B/81B 4Gbps/8Gbps PCIe single port FC HBA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5953u16 , name : "41B 4Gbps single port FC HBA" } , SubSystem { subvendor : 4156u16 , subdevice : 5955u16 , name : "81B 8Gbps single port FC HBA" } , SubSystem { subvendor : 5719u16 , subdevice : 20u16 , name : "415/815 4Gbps/8Gbps single port PCIe FC HBA" }] } , Device { vendor_id : 5719u16 , id : 33u16 , name : "804 8Gbps FC HBA for HP Bladesystem c-class" , subsystems : & [] } , Device { vendor_id : 5719u16 , id : 34u16 , name : "BR-1860 Fabric Adapter" , subsystems : & [SubSystem { subvendor : 5719u16 , subdevice : 34u16 , name : "10Gbps CNA - FCOE" } , SubSystem { subvendor : 5719u16 , subdevice : 35u16 , name : "10Gbps CNA - LL" } , SubSystem { subvendor : 5719u16 , subdevice : 36u16 , name : "16Gbps FC HBA" }] } , Device { vendor_id : 5719u16 , id : 35u16 , name : "1867/1869 16Gbps FC HBA" , subsystems : & [] } , Device { vendor_id : 5719u16 , id : 1606u16 , name : "400 4Gbps PCIe FC HBA" , subsystems : & [] }] }),
        (4411, Vendor { id : 4411u16 , name : "Network Computing Devices" , devices : & [] }),
        (7364, Vendor { id : 7364u16 , name : "Shenzhen Unionmemory Information System Ltd." , devices : & [Device { vendor_id : 7364u16 , id : 4611u16 , name : "NVMe SSD Controller UHXXXa series" , subsystems : & [SubSystem { subvendor : 7364u16 , subdevice : 41249u16 , name : "NVMe SSD UHXXXa series U.2 960GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41250u16 , name : "NVMe SSD UHXXXa series U.2 1920GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41251u16 , name : "NVMe SSD UHXXXa series U.2 3840GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41252u16 , name : "NVMe SSD UHXXXa series U.2 7680GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41253u16 , name : "NVMe SSD UHXXXa series U.2 15360GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41489u16 , name : "NVMe SSD UHXXXa series U.2 800GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41490u16 , name : "NVMe SSD UHXXXa series U.2 1600GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41491u16 , name : "NVMe SSD UHXXXa series U.2 3200GB" } , SubSystem { subvendor : 7364u16 , subdevice : 41492u16 , name : "NVMe SSD UHXXXa series U.2 6400GB" } , SubSystem { subvendor : 7364u16 , subdevice : 57634u16 , name : "NVMe SSD UH711a series U.2 1920GB" } , SubSystem { subvendor : 7364u16 , subdevice : 57635u16 , name : "NVMe SSD UH711a series U.2 3840GB" } , SubSystem { subvendor : 7364u16 , subdevice : 57636u16 , name : "NVMe SSD UH711a series U.2 7680GB" }] } , Device { vendor_id : 7364u16 , id : 6057u16 , name : "RPITJ1TBVME2HWD NVMe SSD 1024GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 6058u16 , name : "AH631 PCIe 3.0 NVMe SSD 512GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 6059u16 , name : "AH631 PCIe 3.0 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 8803u16 , name : "AM611 PCIe 3.0 x2 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 20488u16 , name : "AM610 PCIe 3.0 x2 NVMe SSD 128GB, 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 20498u16 , name : "AH530 PCIe 3.0 NVMe SSD 512GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 20993u16 , name : "AM520 PCIe 3.0 NVMe SSD 128GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 21010u16 , name : "AM521 PCIe 3.0 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 21524u16 , name : "AM541 PCIe 4.0 NVMe SSD 1024GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 24608u16 , name : "NVMe SSD Controller UM3X1X series" , subsystems : & [SubSystem { subvendor : 7364u16 , subdevice : 33568u16 , name : "NVMe SSD UM301a M.2 480GB" } , SubSystem { subvendor : 7364u16 , subdevice : 33569u16 , name : "NVMe SSD UM301a M.2 960GB" } , SubSystem { subvendor : 7364u16 , subdevice : 33570u16 , name : "NVMe SSD UM301a M.2 1.92TB" } , SubSystem { subvendor : 7840u16 , subdevice : 33568u16 , name : "NVMe SSD TM1300 M.2 480GB" } , SubSystem { subvendor : 7840u16 , subdevice : 33570u16 , name : "NVMe SSD TM1300 M.2 1.92TB" }] } , Device { vendor_id : 7364u16 , id : 25089u16 , name : "AM620 PCIe 3.0 NVMe SSD 128GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25090u16 , name : "AM620 PCIe 3.0 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25091u16 , name : "AM620 PCIe 3.0 NVMe SSD 512GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25092u16 , name : "AM620 PCIe 3.0 NVMe SSD 1024GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25346u16 , name : "AM630 PCIe 4.0 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25347u16 , name : "AM630 PCIe 4.0 x4 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25348u16 , name : "AM630 PCIe 4.0 NVMe SSD 1024GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25420u16 , name : "LEGEND 820 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 25434u16 , name : "GAMMIX S60 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 26123u16 , name : "AH660 PCIe 4.0 NVMe SSD 512GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 26124u16 , name : "AH660 PCIe 4.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27137u16 , name : "AM620 PCIe 3.0 NVMe SSD 128GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27138u16 , name : "AM6A0 PCIe 4.0 NVMe SSD 256GB" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27139u16 , name : "RPETJ512MKP1QDQ PCIe 4.0 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27140u16 , name : "AM6A0 PCIe 4.0 NVMe SSD 1024GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27155u16 , name : "RPJYJ512MKN1QWQ PCIe 4.0 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27156u16 , name : "AM6A1 PCIe 4.0 NVMe SSD 1024GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27394u16 , name : "AM6B0 PCIe 4.0 NVMe SSD 256GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27395u16 , name : "RPETJ512MMW1MDQ PCIe 4.0 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27396u16 , name : "AM6B0 PCIe 4.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27411u16 , name : "AM6B1 PCIe 4.0 NVMe SSD 512GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27412u16 , name : "RPJYJ1T24MLR1HWY PCIe 4.0 NVMe SSD 1024GB (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27667u16 , name : "AM6C1 PCIe 4.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 27907u16 , name : "AM6DX PCIe 5.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7364u16 , id : 28720u16 , name : "NVMe SSD Controller UH7X3X series" , subsystems : & [SubSystem { subvendor : 7364u16 , subdevice : 28946u16 , name : "NVMe SSD UH733a U.2 1.6TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28947u16 , name : "NVMe SSD UH733a U.2 3.2TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28948u16 , name : "NVMe SSD UH733a U.2 6.4TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28949u16 , name : "NVMe SSD UH733a U.2 12.8TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28962u16 , name : "NVMe SSD UH713a U.2 1.92TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28963u16 , name : "NVMe SSD UH713a U.2 3.84TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28964u16 , name : "NVMe SSD UH713a U.2 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 28965u16 , name : "NVMe SSD UH713a U.2 15.36TB" }] } , Device { vendor_id : 7364u16 , id : 32816u16 , name : "NVMe SSD Controller UH8X2X/UH7X2X series" , subsystems : & [SubSystem { subvendor : 7364u16 , subdevice : 4386u16 , name : "NVMe SSD UH812a U.2 1.92TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4387u16 , name : "NVMe SSD UH812a U.2 3.84TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4388u16 , name : "NVMe SSD UH812a U.2 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4389u16 , name : "NVMe SSD UH812a U.2 15.36TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4642u16 , name : "NVMe SSD UH812a E3.S 1.92TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4643u16 , name : "NVMe SSD UH812a E3.S 3.84TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4644u16 , name : "NVMe SSD UH812a E3.S 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 4645u16 , name : "NVMe SSD UH812a E3.S 15.36TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8466u16 , name : "NVMe SSD UH832a U.2 1.6TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8467u16 , name : "NVMe SSD UH832a U.2 3.2TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8468u16 , name : "NVMe SSD UH832a U.2 6.4TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8469u16 , name : "NVMe SSD UH832a U.2 12.8TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8722u16 , name : "NVMe SSD UH832a E3.S 1.6TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8723u16 , name : "NVMe SSD UH832a E3.S 3.2TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8724u16 , name : "NVMe SSD UH832a E3.S 6.4TB" } , SubSystem { subvendor : 7364u16 , subdevice : 8725u16 , name : "NVMe SSD UH832a E3.S 12.8TB" } , SubSystem { subvendor : 7364u16 , subdevice : 12578u16 , name : "NVMe SSD UH712a U.2 1.92TB" } , SubSystem { subvendor : 7364u16 , subdevice : 12579u16 , name : "NVMe SSD UH712a U.2 3.84TB" } , SubSystem { subvendor : 7364u16 , subdevice : 12580u16 , name : "NVMe SSD UH712a U.2 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 12581u16 , name : "NVMe SSD UH712a U.2 15.36TB" } , SubSystem { subvendor : 7364u16 , subdevice : 20773u16 , name : "NVMe SSD UH802a U.2 15.36TB" } , SubSystem { subvendor : 7364u16 , subdevice : 20774u16 , name : "NVMe SSD UH802a U.2 30.72TB" } , SubSystem { subvendor : 7364u16 , subdevice : 20775u16 , name : "NVMe SSD UH802a U.2 61.44TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24850u16 , name : "NVMe SSD UH832c U.2 1.6TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24851u16 , name : "NVMe SSD UH832c U.2 3.2TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24852u16 , name : "NVMe SSD UH832c U.2 6.4TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24853u16 , name : "NVMe SSD UH832c U.2 12.8TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24866u16 , name : "NVMe SSD UH812c U.2 1.92TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24867u16 , name : "NVMe SSD UH812c U.2 3.84TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24868u16 , name : "NVMe SSD UH812c U.2 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 24869u16 , name : "NVMe SSD UH812c U.2 15.36TB" } , SubSystem { subvendor : 7364u16 , subdevice : 25108u16 , name : "NVMe SSD UH832c E3.S 6.4TB" } , SubSystem { subvendor : 7364u16 , subdevice : 25109u16 , name : "NVMe SSD UH832c E3.S 12.8TB" } , SubSystem { subvendor : 7364u16 , subdevice : 25124u16 , name : "NVMe SSD UH812c E3.S 7.68TB" } , SubSystem { subvendor : 7364u16 , subdevice : 25125u16 , name : "NVMe SSD UH812c E3.S 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 16676u16 , name : "NVMe SSD TP3511 U.2 7.68TB" } , SubSystem { subvendor : 7840u16 , subdevice : 16677u16 , name : "NVMe SSD TP3511 U.2 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 16932u16 , name : "NVMe SSD TP3511 E3.S 7.68TB" } , SubSystem { subvendor : 7840u16 , subdevice : 16933u16 , name : "NVMe SSD TP3511 E3.S 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 20773u16 , name : "NVMe SSD TP3310 U.2 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 20774u16 , name : "NVMe SSD TP3310 U.2 30.72TB" } , SubSystem { subvendor : 7840u16 , subdevice : 20775u16 , name : "NVMe SSD TP3310 U.2 61.44TB" } , SubSystem { subvendor : 7840u16 , subdevice : 24868u16 , name : "NVMe SSD TP3510 U.2 7.68TB" } , SubSystem { subvendor : 7840u16 , subdevice : 24869u16 , name : "NVMe SSD TP3510 U.2 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 25124u16 , name : "NVMe SSD TP3510 E3.S 7.68TB" } , SubSystem { subvendor : 7840u16 , subdevice : 25125u16 , name : "NVMe SSD TP3510 E3.S 15.36TB" } , SubSystem { subvendor : 7840u16 , subdevice : 28964u16 , name : "NVMe SSD TP3512 U.2 7.68TB" } , SubSystem { subvendor : 7840u16 , subdevice : 28965u16 , name : "NVMe SSD TP3512 U.2 15.36TB" }] }] }),
        (4254, Vendor { id : 4254u16 , name : "Brooktree Corporation" , devices : & [Device { vendor_id : 4254u16 , id : 784u16 , name : "Bt848 Video Capture" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 814u16 , name : "Bt878 Video Capture" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 848u16 , name : "Bt848 Video Capture" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 849u16 , name : "Bt849A Video capture" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 873u16 , name : "Bt878 Video Capture" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 1u16 , name : "TV-Wonder" } , SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "TV-Wonder/VE" }] } , Device { vendor_id : 4254u16 , id : 876u16 , name : "Bt879(??) Video Capture" , subsystems : & [SubSystem { subvendor : 5097u16 , subdevice : 112u16 , name : "Win/TV (Video Section)" }] } , Device { vendor_id : 4254u16 , id : 878u16 , name : "Bt878 Video Capture" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 1u16 , name : "Euresys Picolo PCIe" } , SubSystem { subvendor : 0u16 , subdevice : 2u16 , name : "Euresys PICOLO Pro 2" } , SubSystem { subvendor : 0u16 , subdevice : 4u16 , name : "Euresys PICOLO Pro 3E" } , SubSystem { subvendor : 112u16 , subdevice : 5099u16 , name : "WinTV Series" } , SubSystem { subvendor : 112u16 , subdevice : 65281u16 , name : "Viewcast Osprey 200" } , SubSystem { subvendor : 113u16 , subdevice : 257u16 , name : "DigiTV PCI" } , SubSystem { subvendor : 4221u16 , subdevice : 26118u16 , name : "WinFast TV 2000" } , SubSystem { subvendor : 4541u16 , subdevice : 18u16 , name : "PCTV pro (TV + FM stereo receiver)" } , SubSystem { subvendor : 4541u16 , subdevice : 28u16 , name : "PCTV Sat (DBC receiver)" } , SubSystem { subvendor : 4730u16 , subdevice : 1u16 , name : "Bt878 Mediastream Controller NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 2u16 , name : "Bt878 Mediastream Controller PAL BG" } , SubSystem { subvendor : 4730u16 , subdevice : 3u16 , name : "Bt878a Mediastream Controller PAL BG" } , SubSystem { subvendor : 4730u16 , subdevice : 72u16 , name : "Bt878/832 Mediastream Controller" } , SubSystem { subvendor : 5199u16 , subdevice : 12288u16 , name : "MagicTView CPH060 - Video" } , SubSystem { subvendor : 5217u16 , subdevice : 2u16 , name : "TV98 Series (TV/No FM/Remote)" } , SubSystem { subvendor : 5217u16 , subdevice : 3u16 , name : "AverMedia UltraTV PCI 350" } , SubSystem { subvendor : 5217u16 , subdevice : 4u16 , name : "AVerTV WDM Video Capture" } , SubSystem { subvendor : 5217u16 , subdevice : 1889u16 , name : "AverTV DVB-T" } , SubSystem { subvendor : 5217u16 , subdevice : 1905u16 , name : "AverMedia AVerTV DVB-T 771" } , SubSystem { subvendor : 5361u16 , subdevice : 1u16 , name : "Bt878 Mediastream Controller NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 2u16 , name : "Bt878 Mediastream Controller PAL BG" } , SubSystem { subvendor : 5361u16 , subdevice : 3u16 , name : "Bt878a Mediastream Controller PAL BG" } , SubSystem { subvendor : 5361u16 , subdevice : 72u16 , name : "Bt878/832 Mediastream Controller" } , SubSystem { subvendor : 6149u16 , subdevice : 257u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 258u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 259u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 260u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 261u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 262u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 263u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 264u16 , name : "Euresys PICOLO Tetra" } , SubSystem { subvendor : 6149u16 , subdevice : 513u16 , name : "Euresys PICOLO Tetra-X" } , SubSystem { subvendor : 6149u16 , subdevice : 514u16 , name : "Euresys PICOLO Tetra-X" } , SubSystem { subvendor : 6149u16 , subdevice : 515u16 , name : "Euresys PICOLO Tetra-X" } , SubSystem { subvendor : 6149u16 , subdevice : 516u16 , name : "Euresys PICOLO Tetra-X" } , SubSystem { subvendor : 6149u16 , subdevice : 1025u16 , name : "Euresys PICOLO Tymo" } , SubSystem { subvendor : 6149u16 , subdevice : 1026u16 , name : "Euresys PICOLO Tymo" } , SubSystem { subvendor : 6149u16 , subdevice : 1027u16 , name : "Euresys PICOLO Tymo" } , SubSystem { subvendor : 6149u16 , subdevice : 1028u16 , name : "Euresys PICOLO Tymo" } , SubSystem { subvendor : 6149u16 , subdevice : 4097u16 , name : "Euresys PICOLO Junior 4" } , SubSystem { subvendor : 6178u16 , subdevice : 1u16 , name : "VisionPlus DVB card" } , SubSystem { subvendor : 6225u16 , subdevice : 6224u16 , name : "FlyVideo'98 - Video" } , SubSystem { subvendor : 6225u16 , subdevice : 6225u16 , name : "FlyVideo II" } , SubSystem { subvendor : 6226u16 , subdevice : 6226u16 , name : "FlyVideo'98 - Video (with FM Tuner)" } , SubSystem { subvendor : 6316u16 , subdevice : 54528u16 , name : "DViCO FusionHDTV5 Lite" } , SubSystem { subvendor : 9999u16 , subdevice : 64512u16 , name : "Digitop DTT-1000" } , SubSystem { subvendor : 43520u16 , subdevice : 5216u16 , name : "Spectra8 CardA Input0" } , SubSystem { subvendor : 43521u16 , subdevice : 5217u16 , name : "Spectra8 CardA Input1" } , SubSystem { subvendor : 43522u16 , subdevice : 5218u16 , name : "Spectra8 CardA Input2" } , SubSystem { subvendor : 43523u16 , subdevice : 5219u16 , name : "Spectra8 CardA Input3" } , SubSystem { subvendor : 43524u16 , subdevice : 5220u16 , name : "Spectra8 CardB Input0" } , SubSystem { subvendor : 43525u16 , subdevice : 5221u16 , name : "Spectra8 CardB Input1" } , SubSystem { subvendor : 43526u16 , subdevice : 5222u16 , name : "Spectra8 CardB Input2" } , SubSystem { subvendor : 43527u16 , subdevice : 5223u16 , name : "Spectra8 CardB Input3" } , SubSystem { subvendor : 43528u16 , subdevice : 5224u16 , name : "Spectra8 CardC Input0" } , SubSystem { subvendor : 43529u16 , subdevice : 5225u16 , name : "Spectra8 CardC Input1" } , SubSystem { subvendor : 43530u16 , subdevice : 5226u16 , name : "Spectra8 CardC Input2" } , SubSystem { subvendor : 43531u16 , subdevice : 5227u16 , name : "Spectra8 CardC Input3" } , SubSystem { subvendor : 43532u16 , subdevice : 5228u16 , name : "Spectra8 CardD Input0" } , SubSystem { subvendor : 43533u16 , subdevice : 5229u16 , name : "Spectra8 CardD Input1" } , SubSystem { subvendor : 43534u16 , subdevice : 5230u16 , name : "Spectra8 CardD Input2" } , SubSystem { subvendor : 43535u16 , subdevice : 5231u16 , name : "Spectra8 CardD Input3" } , SubSystem { subvendor : 48401u16 , subdevice : 4608u16 , name : "PCTV pro (TV + FM stereo receiver)" }] } , Device { vendor_id : 4254u16 , id : 879u16 , name : "Bt879 Video Capture" , subsystems : & [SubSystem { subvendor : 4730u16 , subdevice : 68u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 290u16 , name : "Bt879 Video Capture PAL I" } , SubSystem { subvendor : 4730u16 , subdevice : 324u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 546u16 , name : "Bt879 Video Capture PAL BG" } , SubSystem { subvendor : 4730u16 , subdevice : 580u16 , name : "Bt879a Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 802u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 1058u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 4386u16 , name : "Bt879 Video Capture PAL I" } , SubSystem { subvendor : 4730u16 , subdevice : 4642u16 , name : "Bt879 Video Capture PAL BG" } , SubSystem { subvendor : 4730u16 , subdevice : 4898u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 4730u16 , subdevice : 5410u16 , name : "Bt879a Video Capture PAL I" } , SubSystem { subvendor : 4730u16 , subdevice : 5666u16 , name : "Bt879a Video Capture PAL BG" } , SubSystem { subvendor : 4730u16 , subdevice : 5922u16 , name : "Bt879a Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 68u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 290u16 , name : "Bt879 Video Capture PAL I" } , SubSystem { subvendor : 5361u16 , subdevice : 324u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 546u16 , name : "Bt879 Video Capture PAL BG" } , SubSystem { subvendor : 5361u16 , subdevice : 580u16 , name : "Bt879a Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 802u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 1058u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 4386u16 , name : "Bt879 Video Capture PAL I" } , SubSystem { subvendor : 5361u16 , subdevice : 4642u16 , name : "Bt879 Video Capture PAL BG" } , SubSystem { subvendor : 5361u16 , subdevice : 4898u16 , name : "Bt879 Video Capture NTSC" } , SubSystem { subvendor : 5361u16 , subdevice : 5410u16 , name : "Bt879a Video Capture PAL I" } , SubSystem { subvendor : 5361u16 , subdevice : 5666u16 , name : "Bt879a Video Capture PAL BG" } , SubSystem { subvendor : 5361u16 , subdevice : 5922u16 , name : "Bt879a Video Capture NTSC" } , SubSystem { subvendor : 6225u16 , subdevice : 6224u16 , name : "FlyVideo'98 - Video" } , SubSystem { subvendor : 6225u16 , subdevice : 6225u16 , name : "FlyVideo II" } , SubSystem { subvendor : 6226u16 , subdevice : 6226u16 , name : "FlyVideo'98 - Video (with FM Tuner)" }] } , Device { vendor_id : 4254u16 , id : 880u16 , name : "Bt880 Video Capture" , subsystems : & [SubSystem { subvendor : 6225u16 , subdevice : 6224u16 , name : "FlyVideo'98" } , SubSystem { subvendor : 6225u16 , subdevice : 6225u16 , name : "FlyVideo'98 EZ - video" } , SubSystem { subvendor : 6226u16 , subdevice : 6226u16 , name : "FlyVideo'98 (with FM Tuner)" }] } , Device { vendor_id : 4254u16 , id : 2168u16 , name : "Bt878 Audio Capture" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 1u16 , name : "Euresys Picolo PCIe" } , SubSystem { subvendor : 0u16 , subdevice : 2u16 , name : "Euresys PICOLO Pro 2 (Audio Section)" } , SubSystem { subvendor : 0u16 , subdevice : 4u16 , name : "Euresys PICOLO Pro 3E (Audio Section)" } , SubSystem { subvendor : 112u16 , subdevice : 5099u16 , name : "WinTV Series" } , SubSystem { subvendor : 112u16 , subdevice : 65281u16 , name : "Viewcast Osprey 200" } , SubSystem { subvendor : 113u16 , subdevice : 257u16 , name : "DigiTV PCI" } , SubSystem { subvendor : 4098u16 , subdevice : 1u16 , name : "TV-Wonder" } , SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "TV-Wonder/VE" } , SubSystem { subvendor : 4541u16 , subdevice : 18u16 , name : "PCTV pro (TV + FM stereo receiver, audio section)" } , SubSystem { subvendor : 4541u16 , subdevice : 28u16 , name : "PCTV Sat (DBC receiver)" } , SubSystem { subvendor : 4730u16 , subdevice : 1u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 2u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 3u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 72u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 5097u16 , subdevice : 112u16 , name : "Win/TV (Audio Section)" } , SubSystem { subvendor : 5199u16 , subdevice : 12288u16 , name : "MagicTView CPH060 - Audio" } , SubSystem { subvendor : 5217u16 , subdevice : 2u16 , name : "Avermedia PCTV98 Audio Capture" } , SubSystem { subvendor : 5217u16 , subdevice : 3u16 , name : "UltraTV PCI 350" } , SubSystem { subvendor : 5217u16 , subdevice : 4u16 , name : "AVerTV WDM Audio Capture" } , SubSystem { subvendor : 5217u16 , subdevice : 1889u16 , name : "AVerTV DVB-T" } , SubSystem { subvendor : 5217u16 , subdevice : 1905u16 , name : "AverMedia AVerTV DVB-T 771" } , SubSystem { subvendor : 5361u16 , subdevice : 1u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 2u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 3u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 72u16 , name : "Bt878 Video Capture (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 257u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 258u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 259u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 260u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 261u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 262u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 263u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 264u16 , name : "Euresys PICOLO Tetra (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 513u16 , name : "Euresys PICOLO Tetra-X (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 514u16 , name : "Euresys PICOLO Tetra-X (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 515u16 , name : "Euresys PICOLO Tetra-X (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 516u16 , name : "Euresys PICOLO Tetra-X (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 1025u16 , name : "Euresys PICOLO Tymo (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 1026u16 , name : "Euresys PICOLO Tymo (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 1027u16 , name : "Euresys PICOLO Tymo (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 1028u16 , name : "Euresys PICOLO Tymo (Audio Section)" } , SubSystem { subvendor : 6149u16 , subdevice : 4097u16 , name : "Euresys PICOLO Junior 4 (Audio Section)" } , SubSystem { subvendor : 6178u16 , subdevice : 1u16 , name : "VisionPlus DVB Card" } , SubSystem { subvendor : 6316u16 , subdevice : 54528u16 , name : "DViCO FusionHDTV5 Lite" } , SubSystem { subvendor : 9999u16 , subdevice : 64512u16 , name : "Digitop DTT-1000" } , SubSystem { subvendor : 48401u16 , subdevice : 4608u16 , name : "PCTV pro (TV + FM stereo receiver, audio section)" }] } , Device { vendor_id : 4254u16 , id : 2169u16 , name : "Bt879 Audio Capture" , subsystems : & [SubSystem { subvendor : 4730u16 , subdevice : 68u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 290u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 324u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 546u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 580u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 802u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 1058u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 4386u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 4642u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 4898u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 5410u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 5666u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 4730u16 , subdevice : 5922u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 68u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 290u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 324u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 546u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 580u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 802u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 1058u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 4386u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 4642u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 4898u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 5410u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 5666u16 , name : "Bt879 Video Capture (Audio Section)" } , SubSystem { subvendor : 5361u16 , subdevice : 5922u16 , name : "Bt879 Video Capture (Audio Section)" }] } , Device { vendor_id : 4254u16 , id : 2176u16 , name : "Bt880 Audio Capture" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 8469u16 , name : "BtV 2115 Mediastream controller" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 8485u16 , name : "BtV 2125 Mediastream controller" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 8548u16 , name : "BtV 2164" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 8549u16 , name : "BtV 2165" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 33328u16 , name : "Bt8230 ATM Segment/Reassembly Ctrlr (SRC)" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 33906u16 , name : "Bt8472" , subsystems : & [] } , Device { vendor_id : 4254u16 , id : 33908u16 , name : "Bt8474" , subsystems : & [] }] }),
        (16980, Vendor { id : 16980u16 , name : "DVBSky" , devices : & [] }),
        (7345, Vendor { id : 7345u16 , name : "Collion UG & Co.KG" , devices : & [] }),
        (4662, Vendor { id : 4662u16 , name : "Sigma Designs Corporation" , devices : & [Device { vendor_id : 4662u16 , id : 0u16 , name : "RealMagic64/GX" , subsystems : & [] } , Device { vendor_id : 4662u16 , id : 25601u16 , name : "REALmagic 64/GX (SD 6425)" , subsystems : & [] }] }),
        (8372, Vendor { id : 8372u16 , name : "Genstoraige Technology Co., Ltd." , devices : & [Device { vendor_id : 8372u16 , id : 41484u16 , name : "PT200A PCIe 5.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8372u16 , id : 49756u16 , name : "MQ205 PCIe 5.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8372u16 , id : 53772u16 , name : "PT200 PCIe 5.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8372u16 , id : 53852u16 , name : "PT205 PCIe 5.0 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8372u16 , id : 53900u16 , name : "PT208 PCIe 5.0 NVMe SSD" , subsystems : & [] }] }),
        (5243, Vendor { id : 5243u16 , name : "ABIT Computer Corp." , devices : & [Device { vendor_id : 5243u16 , id : 4228u16 , name : "IP35 [Dark Raider]" , subsystems : & [] }] }),
        (4263, Vendor { id : 4263u16 , name : "Benchmarq Microelectronics" , devices : & [] }),
        (5963, Vendor { id : 5963u16 , name : "PC Partner Limited / Sapphire Technology" , devices : & [] }),
        (4573, Vendor { id : 4573u16 , name : "Crosfield Electronics Limited" , devices : & [] }),
        (5006, Vendor { id : 5006u16 , name : "Basler GmbH" , devices : & [] }),
        (5322, Vendor { id : 5322u16 , name : "PE Logic Corp." , devices : & [] }),
        (5845, Vendor { id : 5845u16 , name : "Acromag, Inc." , devices : & [Device { vendor_id : 5845u16 , id : 1284u16 , name : "PMC-DX504 Reconfigurable FPGA with LVDS I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 1312u16 , name : "PMC520 Serial Communication, 232 Octal" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 1313u16 , name : "PMC521 Serial Communication, 422/485 Octal" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 4128u16 , name : "PMC-AX1020 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 4197u16 , name : "PMC-AX1065 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 8196u16 , name : "PMC-DX2004 Reconfigurable FPGA with LVDS I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 8224u16 , name : "PMC-AX2020 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 8293u16 , name : "PMC-AX2065 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 12320u16 , name : "PMC-AX3020 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 12389u16 , name : "PMC-AX3065 Reconfigurable FPGA with A/D & D/A" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 16963u16 , name : "PMC424, APC424, AcPC424 Digital I/O and Counter Timer Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 16968u16 , name : "PMC464, APC464, AcPC464 Digital I/O and Counter Timer Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 16971u16 , name : "PMC-DX2002 Reconfigurable FPGA with Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 16979u16 , name : "PMC-DX503 Reconfigurable FPGA with TTL and Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17170u16 , name : "PMC-CX1002 Reconfigurable Conduction-Cooled FPGA Virtex-II with Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17171u16 , name : "PMC-CX1003 Reconfigurable Conduction-Cooled FPGA Virtex-II with CMOS and Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17186u16 , name : "PMC-CX2002 Reconfigurable Conduction-Cooled FPGA Virtex-II with Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17187u16 , name : "PMC-CX2003 Reconfigurable Conduction-Cooled FPGA Virtex-II with CMOS and Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17232u16 , name : "PMC-DX501 Reconfigurable Digital I/O Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17235u16 , name : "PMC-DX2003 Reconfigurable FPGA with TTL and Differential I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17239u16 , name : "PMC-DX502 Reconfigurable Differential I/O Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17495u16 , name : "PMC730, APC730, AcPC730 Multifunction Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17521u16 , name : "XMC730 Multi-function I/O module with front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17523u16 , name : "XMC730CC Multi-function I/O module with rear I/O Conduction-cooled" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 17997u16 , name : "PMC408 32-Channel Digital Input/Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 18512u16 , name : "PMC220-16 12-Bit Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19010u16 , name : "PMC483, APC483, AcPC483 Counter Timer Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19024u16 , name : "PMC484, APC484, AcPC484 Counter Timer Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19030u16 , name : "PMC230 16-Bit Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19271u16 , name : "PMC330, APC330, AcPC330 Analog Input Module, 16-bit A/D" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19520u16 , name : "PMC-LX40 Reconfigurable Virtex-4 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19552u16 , name : "PMC-LX60 Reconfigurable Virtex-4 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19789u16 , name : "PMC341, APC341, AcPC341 Analog Input Module, Simultaneous Sample & Hold" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 19790u16 , name : "PMC482, APC482, AcPC482 Counter Timer Board" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 21069u16 , name : "PMC-DX2001 Reconfigurable FPGA with TTL I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 21301u16 , name : "PMC-SX35 Reconfigurable Virtex-4 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 21590u16 , name : "PMC470 48-Channel Digital Input/Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22017u16 , name : "PMC-VLX85 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22018u16 , name : "PMC-VLX110 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22019u16 , name : "PMC-VSX95 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22020u16 , name : "PMC-VLX155 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22021u16 , name : "PMC-VFX70 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22022u16 , name : "PMC-VLX155-1M Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22273u16 , name : "PMC-SLX150: Reconfigurable Spartan-6 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22274u16 , name : "PMC-SLX150-1M: Reconfigurable Spartan-6 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22529u16 , name : "XMC-VLX85 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22530u16 , name : "XMC-VLX110 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22531u16 , name : "XMC-VSX95 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22532u16 , name : "XMC-VLX155 Reconfigurable Virtex-5 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22535u16 , name : "XMC-SLX150: Reconfigurable Spartan-6 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22536u16 , name : "XMC-SLX150-1M: Reconfigurable Spartan-6 FPGA with plug-in I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 22785u16 , name : "APCe8650 PCI Express IndustryPack Carrier Card" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 25345u16 , name : "XMC Module with user-configurable Virtex-6 FPGA, 240k logic cells, SFP front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 25346u16 , name : "XMC Module with user-configurable Virtex-6 FPGA, 365k logic cells, SFP front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 25347u16 , name : "XMC Module with user-configurable Virtex-6 FPGA, 240k logic cells, no front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 25348u16 , name : "XMC Module with user-configurable Virtex-6 FPGA, 365k logic cells, no front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28672u16 , name : "XMC-7K325F: User-configurable Kintex-7 FPGA, 325k logic cells plus SFP front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28673u16 , name : "XMC-7K410F: User-configurable Kintex-7 FPGA, 410k logic cells plus SFP front I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28674u16 , name : "XMC-7K325AX: User-Configurable Kintex-7 FPGA, 325k logic cells with AXM Plug-In I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28675u16 , name : "XMC-7K410AX: User-Configurable Kintex-7 FPGA, 410k logic cells with AXM Plug-In I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28676u16 , name : "XMC-7K325CC: User-Configurable Kintex-7 FPGA, 325k logic cells, conduction-cooled" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28677u16 , name : "XMC-7K410CC: User-Configurable Kintex-7 FPGA, 410k logic cells, conduction-cooled" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28678u16 , name : "XMC-7A200: User-Configurable Artix-7 FPGA, 200k logic cells with Plug-In I/O" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28679u16 , name : "XMC-7A200CC: User-Configurable Conduction-Cooled Artix-7 FPGA, with 200k logic cells" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28689u16 , name : "AP440-1: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28690u16 , name : "AP440-2: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28691u16 , name : "AP440-3: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28692u16 , name : "AP445: 32-Channel Isolated Digital Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28693u16 , name : "AP471 48-Channel TTL Level Digital Input/Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28694u16 , name : "AP470 48-Channel TTL Level Digital Input/Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28695u16 , name : "AP323 16-bit, 20 or 40 Channel Analog Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28696u16 , name : "AP408: 32-Channel Digital I/O Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28697u16 , name : "AP341 14-bit, 16-Channel Simultaneous Conversion Analog Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28698u16 , name : "AP220-16 12-Bit, 16-Channel Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28699u16 , name : "AP231-16 16-Bit, 16-Channel Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28700u16 , name : "AP225 12-Bit, 16-Channel Analog Output Module with Waveform Memory" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28701u16 , name : "AP235 16-Bit, 16-Channel Analog Output Module with Waveform Memory" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28705u16 , name : "APA7-201 Reconfigurable Artix-7 FPGA module 48 TTL channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28706u16 , name : "APA7-202 Reconfigurable Artix-7 FPGA module 24 RS485 channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28707u16 , name : "APA7-203 Reconfigurable Artix-7 FPGA module 24 TTL & 12 RS485 channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28708u16 , name : "APA7-204 Reconfigurable Artix-7 FPGA module 24 LVDS channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28711u16 , name : "AP418 16-Channel High Voltage Digital Input/Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28713u16 , name : "AP342 14-bit, 12-Channel Isolated Simultaneous Conversion Analog Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28714u16 , name : "AP226 12-Bit, 8-Channel Isolated Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28715u16 , name : "AP236 16-Bit, 8-Channel Isolated Analog Output Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28716u16 , name : "AP560A Module 4 Independent isolated CAN bus channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28721u16 , name : "AP441-1: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28722u16 , name : "AP441-2: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28723u16 , name : "AP441-3: 32-Channel Isolated Digital Input Module" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28738u16 , name : "AP482 Counter Timer Module with TTL Level Input/Output" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28739u16 , name : "AP483 Counter Timer Module with TTL Level and RS422 Input/Output" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28740u16 , name : "AP484 Counter Timer Module with RS422 Input/Output" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28753u16 , name : "APA7-501 Reconfigurable Artix-7 52,160 Cell FPGA module 48 TTL channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28754u16 , name : "APA7-502 Reconfigurable Artix-7 52,160 Cell FPGA module 24 RS485 channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28755u16 , name : "APA7-503 Reconfigurable Artix-7 52,160 Cell FPGA module 24 TTL & 12 RS485 channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28756u16 , name : "APA7-504 Reconfigurable Artix-7 52,160 Cell FPGA module 24 LVDS channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28786u16 , name : "AP731 Multi-function I/O Module with 12-bit DAC" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 28787u16 , name : "AP730 Multi-function I/O Module 16 Digital I/O 8 Differential Analog In 4 Analog Out" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 30513u16 , name : "APZU-301 Zynq Ultrascale+ Module 28 TTL channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 30515u16 , name : "APZU-303 Zynq Ultrascale+ Module 20 TTL & 3 RS485/422 channels" , subsystems : & [] } , Device { vendor_id : 5845u16 , id : 30516u16 , name : "APZU-304 Zynq Ultrascale+ Module 14 LVDS channels" , subsystems : & [] }] }),
        (7544, Vendor { id : 7544u16 , name : "DERA Storage" , devices : & [Device { vendor_id : 7544u16 , id : 5394u16 , name : "TAI NVMe Controller" , subsystems : & [SubSystem { subvendor : 7544u16 , subdevice : 8196u16 , name : "D5437 HHHL 2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 8198u16 , name : "D5437 HHHL 4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 8200u16 , name : "D5437 HHHL 8TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 8452u16 , name : "D5437 U.2 2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 8454u16 , name : "D5437 U.2 4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 8456u16 , name : "D5437 U.2 8TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12291u16 , name : "D5457 HHHL 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12293u16 , name : "D5457 HHHL 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12295u16 , name : "D5457 HHHL 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12547u16 , name : "D5457 U.2 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12549u16 , name : "D5457 U.2 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12551u16 , name : "D5457 U.2 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 12552u16 , name : "D5437 U.2 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20483u16 , name : "D5447 HHHL 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20484u16 , name : "D5427 HHHL 1.92TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20485u16 , name : "D5447 HHHL 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20486u16 , name : "D5427 HHHL 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20487u16 , name : "D5447 HHHL 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20488u16 , name : "D5427 HHHL 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20739u16 , name : "D5447 U.2 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20741u16 , name : "D5447 U.2 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20742u16 , name : "AF3UP34DTCUN3T8B SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20743u16 , name : "D5447 U.2 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20756u16 , name : "D5427 U.2 1.92TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20758u16 , name : "D5427 U.2 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 20760u16 , name : "D5427 U.2 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 33031u16 , name : "D5457 U.2 6.4TB NVMe SSD" }] } , Device { vendor_id : 7544u16 , id : 5395u16 , name : "DERA MENG NVMe Controller" , subsystems : & [SubSystem { subvendor : 7544u16 , subdevice : 24597u16 , name : "D6456 HHHL 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24598u16 , name : "D6436 HHHL 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24599u16 , name : "D6456 HHHL 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24600u16 , name : "D6436 HHHL 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24837u16 , name : "D6457 U.2 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24838u16 , name : "D6437 U.2 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24839u16 , name : "D6457 U.2 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24840u16 , name : "D6437 U.2 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24853u16 , name : "D6456 U.2 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24854u16 , name : "D6436 U.2 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24855u16 , name : "D6456 U.2 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 24856u16 , name : "D6436 U.2 7.68TB NVMe SSD" }] } , Device { vendor_id : 7544u16 , id : 5397u16 , name : "DERA EMEI NVMe Controller" , subsystems : & [SubSystem { subvendor : 7544u16 , subdevice : 28674u16 , name : "D7436 HHHL 1.92TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28675u16 , name : "D7456 HHHL 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28676u16 , name : "D7436 HHHL 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28677u16 , name : "D7456 HHHL 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28679u16 , name : "D7456 HHHL 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28680u16 , name : "D7436 HHHL 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28681u16 , name : "D7456 HHHL 12.8TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28682u16 , name : "D7436 HHHL 15.36TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28930u16 , name : "D7436 U.2 15mm 1.92TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28931u16 , name : "D7456 U.2 15mm 1.6TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28932u16 , name : "D7436 U.2 15mm 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28933u16 , name : "D7456 U.2 15mm 3.2TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28935u16 , name : "D7456 U.2 15mm 6.4TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28936u16 , name : "D7436 U.2 15mm 7.68TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28937u16 , name : "D7456 U.2 15mm 12.8TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28938u16 , name : "D7436 U.2 15mm 15.36TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28995u16 , name : "D7556 U.2 15mm 1.6TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28996u16 , name : "D7536 U.2 15mm 1.92TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28997u16 , name : "D7556 U.2 15mm 3.2TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28998u16 , name : "D7536 U.2 15mm 3.84TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 28999u16 , name : "D7556 U.2 15mm 6.4TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29000u16 , name : "D7536 U.2 15mm 7.68TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29001u16 , name : "D7556 U.2 15mm 12.8TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29002u16 , name : "D7536 U.2 15mm 15.36TB dual port NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29186u16 , name : "Aliflash V2 U.2 15mm 1.92TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29188u16 , name : "Aliflash V2 U.2 15mm 3.84TB NVMe SSD" } , SubSystem { subvendor : 7544u16 , subdevice : 29192u16 , name : "Aliflash V2 U.2 15mm 7.68TB NVMe SSD" }] } , Device { vendor_id : 7544u16 , id : 5398u16 , name : "DERA D8436/D8456 NVMe SSD" , subsystems : & [] }] }),
        (5580, Vendor { id : 5580u16 , name : "Hotrail Inc" , devices : & [] }),
        (32902, Vendor { id : 32902u16 , name : "Intel Corporation" , devices : & [Device { vendor_id : 32902u16 , id : 7u16 , name : "82379AB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8u16 , name : "Extended Express System Support Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57u16 , name : "21145 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 64u16 , name : "Core Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 65u16 , name : "Core Processor PCI Express x16 Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 66u16 , name : "Core Processor Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 67u16 , name : "Core Processor Secondary PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 68u16 , name : "Core Processor DRAM Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4133u16 , subdevice : 1159u16 , name : "TravelMate 5742" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 69u16 , name : "Core Processor PCI Express x16 Root Port" , subsystems : & [SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 70u16 , name : "Core Processor Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4313u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 71u16 , name : "Core Processor Secondary PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 72u16 , name : "Core Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 73u16 , name : "Core Processor PCI Express x16 Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 74u16 , name : "Core Processor Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 75u16 , name : "Core Processor Secondary PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 80u16 , name : "Core Processor Thermal Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 105u16 , name : "Core Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 130u16 , name : "Centrino Advanced-N 6205 [Taylor Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4865u16 , name : "Centrino Advanced-N 6205 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "Centrino Advanced-N 6205 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4871u16 , name : "Centrino Advanced-N 6205 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4897u16 , name : "Centrino Advanced-N 6205 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4902u16 , name : "Centrino Advanced-N 6205 ABG" }] } , Device { vendor_id : 32902u16 , id : 131u16 , name : "Centrino Wireless-N 1000 [Condor Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4613u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4614u16 , name : "Centrino Wireless-N 1000 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4645u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4646u16 , name : "Centrino Wireless-N 1000 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4869u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "Centrino Wireless-N 1000 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4901u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4902u16 , name : "Centrino Wireless-N 1000 BG" }] } , Device { vendor_id : 32902u16 , id : 132u16 , name : "Centrino Wireless-N 1000 [Condor Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4629u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4630u16 , name : "Centrino Wireless-N 1000 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4885u16 , name : "Centrino Wireless-N 1000 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "Centrino Wireless-N 1000 BG" }] } , Device { vendor_id : 32902u16 , id : 133u16 , name : "Centrino Advanced-N 6205 [Taylor Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4881u16 , name : "Centrino Advanced-N 6205 (802.11a/b/g/n)" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "Centrino Advanced-N 6205 ABG" }] } , Device { vendor_id : 32902u16 , id : 135u16 , name : "Centrino Advanced-N + WiMAX 6250 [Kilmer Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4865u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4897u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4902u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 ABG" }] } , Device { vendor_id : 32902u16 , id : 137u16 , name : "Centrino Advanced-N + WiMAX 6250 [Kilmer Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4881u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "Centrino Advanced-N + WiMAX 6250 2x2 ABG" }] } , Device { vendor_id : 32902u16 , id : 138u16 , name : "Centrino Wireless-N 1030 [Rainbow Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 21253u16 , name : "Centrino Wireless-N 1030 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21255u16 , name : "Centrino Wireless-N 1030 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 21285u16 , name : "Centrino Wireless-N 1030 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21287u16 , name : "Centrino Wireless-N 1030 BG" }] } , Device { vendor_id : 32902u16 , id : 139u16 , name : "Centrino Wireless-N 1030 [Rainbow Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 21269u16 , name : "Centrino Wireless-N 1030 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21271u16 , name : "Centrino Wireless-N 1030 BG" }] } , Device { vendor_id : 32902u16 , id : 144u16 , name : "Centrino Advanced-N 6230 [Rainbow Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 21009u16 , name : "Centrino Advanced-N 6230 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21013u16 , name : "Centrino Advanced-N 6230 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21014u16 , name : "Centrino Advanced-N 6230 ABG" }] } , Device { vendor_id : 32902u16 , id : 145u16 , name : "Centrino Advanced-N 6230 [Rainbow Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20993u16 , name : "Centrino Advanced-N 6230 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 20997u16 , name : "Centrino Advanced-N 6230 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 20998u16 , name : "Centrino Advanced-N 6230 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 20999u16 , name : "Centrino Advanced-N 6230 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 21025u16 , name : "Centrino Advanced-N 6230 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21029u16 , name : "Centrino Advanced-N 6230 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 21030u16 , name : "Centrino Advanced-N 6230 ABG" }] } , Device { vendor_id : 32902u16 , id : 256u16 , name : "2nd Generation Core Processor Family DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8P67/P8H67 Series Motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4537u16 , name : "Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" }] } , Device { vendor_id : 32902u16 , id : 257u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4203u16 , subdevice : 220u16 , name : "MacBookPro8,2 [Core i7, 15\", 2011]" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 258u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4163u16 , subdevice : 258u16 , name : "P8H67 Series Motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4537u16 , name : "G640 [Sandy Bridge, HD Graphics] on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" }] } , Device { vendor_id : 32902u16 , id : 260u16 , name : "2nd Generation Core Processor Family DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4203u16 , subdevice : 220u16 , name : "MacBookPro8,2 [Core i7, 15\", 2011]" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 32902u16 , id : 261u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 220u16 , name : "MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 262u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 264u16 , name : "Xeon E3-1200 Processor Family DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 265u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 266u16 , name : "Xeon E3-1200 Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 267u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 268u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 269u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 270u16 , name : "Xeon E3-1200/2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 274u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 278u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "integrated HD 3000 graphics controller on NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 290u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 294u16 , name : "2nd Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1228u16 , name : "Vostro 3350" } , SubSystem { subvendor : 6058u16 , subdevice : 8654u16 , name : "ThinkPad T420" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 32902u16 , id : 336u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor DRAM Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "Ivy Bridge GT1 [HD Graphics]" } , SubSystem { subvendor : 5593u16 , subdevice : 1572u16 , name : "X9SCM-F Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 336u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 337u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8H77-I Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 8208u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 338u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8H77-I Motherboard" }] } , Device { vendor_id : 32902u16 , id : 339u16 , name : "3rd Gen Core Processor Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" }] } , Device { vendor_id : 32902u16 , id : 340u16 , name : "3rd Gen Core processor DRAM Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 2054u16 , name : "Aspire E1-470G" } , SubSystem { subvendor : 4133u16 , subdevice : 2067u16 , name : "Aspire R7-571" } , SubSystem { subvendor : 4156u16 , subdevice : 6134u16 , name : "ProBook 4540s" } , SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4303u16 , subdevice : 5823u16 , name : "LIFEBOOK E752" }] } , Device { vendor_id : 32902u16 , id : 341u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 8208u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 342u16 , name : "Ivy Bridge mobile GT1 [HD Graphics]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" }] } , Device { vendor_id : 32902u16 , id : 344u16 , name : "Xeon E3-1200 v2/Ivy Bridge DRAM Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 8208u16 , name : "Server Board S1200BT Family" }] } , Device { vendor_id : 32902u16 , id : 345u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 346u16 , name : "Xeon E3-1200 v2/Ivy Bridge Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 348u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 349u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" }] } , Device { vendor_id : 32902u16 , id : 350u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 354u16 , name : "IvyBridge GT2 [HD Graphics 4000]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 354u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 358u16 , name : "Ivy Bridge mobile GT2 [HD Graphics 4000]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 8451u16 , name : "N56VZ" } , SubSystem { subvendor : 4303u16 , subdevice : 5825u16 , name : "LIFEBOOK E752" }] } , Device { vendor_id : 32902u16 , id : 362u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8B WS Motherboard" }] } , Device { vendor_id : 32902u16 , id : 370u16 , name : "Xeon E3-1200 v2/3rd Gen Core processor Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 374u16 , name : "3rd Gen Core processor Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 513u16 , name : "Arctic Sound" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 644u16 , name : "400 Series Chipset Family On-Package PCH-LP Prem-U LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 645u16 , name : "400 Series Chipset Family On-Package PCH-LP Mainstream/Base U LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 672u16 , name : "400 Series Chipset Family On-Package P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 673u16 , name : "400 Series Chipset Family On-Package PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 675u16 , name : "400 Series Chipset Family On-Package SMBus" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 676u16 , name : "400 Series Chipset Family On-Package SPI (flash) Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 678u16 , name : "400 Series Chipset Family On-Package Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 680u16 , name : "400 Series Chipset Family On-Package UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 681u16 , name : "400 Series Chipset Family On-Package UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 682u16 , name : "400 Series Chipset Family On-Package SPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 683u16 , name : "400 Series Chipset Family On-Package SPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 688u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 689u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 690u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 691u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 692u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 693u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 694u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 695u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 696u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 697u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 698u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 699u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 700u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 701u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 702u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 703u16 , name : "400 Series Chipset Family On-Package PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 708u16 , name : "400 Series Chipset Family On-Package eMMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 709u16 , name : "400 Series Chipset Family On-Package I2C #4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 710u16 , name : "400 Series Chipset Family On-Package I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 711u16 , name : "400 Series Chipset Family On-Package UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 712u16 , name : "400 Series Chipset Family On-Package HD Audio" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 723u16 , name : "400 Series Chipset Family On-Package SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 725u16 , name : "400 Series Chipset Family On-Package SATA Controller (RAID 0/1/5/10) no premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 727u16 , name : "400 Series Chipset Family On-Package SATA Controller (RAID 0/1/5/10) premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 736u16 , name : "400 Series Chipset Family On-Package MEI #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 737u16 , name : "400 Series Chipset Family On-Package MEI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 738u16 , name : "400 Series Chipset Family On-Package IDE Redirection (IDER-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 739u16 , name : "400 Series Chipset Family On-Package Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 740u16 , name : "400 Series Chipset Family On-Package MEI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 741u16 , name : "400 Series Chipset Family On-Package MEI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 744u16 , name : "400 Series Chipset Family On-Package I2C #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 745u16 , name : "400 Series Chipset Family On-Package I2C #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 746u16 , name : "400 Series Chipset Family On-Package I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 747u16 , name : "400 Series Chipset Family On-Package I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 749u16 , name : "400 Series Chipset Family On-Package USB 3.2 Gen 2x1 (10 Gbs) xHCI Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 750u16 , name : "400 Series Chipset Family On-Package USB 3.2 Gen 1x1 (5 Gbs) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 751u16 , name : "400 Series Chipset Family On-Package Shared SRAM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 752u16 , name : "400 Series Chipset Family On-Package CNVi WiFi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9560 160MHz 2x2 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 112u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 564u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9560 80MHz 2x2 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 612u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9461 80MHz 1x1 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 676u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9462 80MHz 1x1 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 16496u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" }] } , Device { vendor_id : 32902u16 , id : 757u16 , name : "400 Series Chipset Family On-Package SDXC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 761u16 , name : "400 Series Chipset Family On-Package Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 763u16 , name : "400 Series Chipset Family On-Package SPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 764u16 , name : "400 Series Chipset Family On-Package Integrated Sensor Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 777u16 , name : "80303 I/O Processor PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 781u16 , name : "80312 I/O Companion Chip PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 806u16 , name : "6700/6702PXH I/OxAPIC Interrupt Controller A" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 807u16 , name : "6700PXH I/OxAPIC Interrupt Controller B" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 809u16 , name : "6700PXH PCI Express-to-PCI Bridge A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 810u16 , name : "6700PXH PCI Express-to-PCI Bridge B" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 812u16 , name : "6702PXH PCI Express-to-PCI Bridge A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 816u16 , name : "80332 [Dobson] I/O processor (A-Segment Bridge)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 817u16 , name : "80332 [Dobson] I/O processor (A-Segment IOAPIC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 818u16 , name : "80332 [Dobson] I/O processor (B-Segment Bridge)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 819u16 , name : "80332 [Dobson] I/O processor (B-Segment IOAPIC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 820u16 , name : "80332 [Dobson] I/O processor (ATU)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 821u16 , name : "80331 [Lindsay] I/O processor (PCI-X Bridge)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 822u16 , name : "80331 [Lindsay] I/O processor (ATU)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 832u16 , name : "41210 [Lanai] Serial to Parallel PCI Bridge (A-Segment Bridge)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 833u16 , name : "41210 [Lanai] Serial to Parallel PCI Bridge (B-Segment Bridge)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 880u16 , name : "80333 Segment-A PCIe Express to PCI-X bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 881u16 , name : "80333 A-Bus IOAPIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 882u16 , name : "80333 Segment-B PCIe Express to PCI-X bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 883u16 , name : "80333 B-Bus IOAPIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 884u16 , name : "80333 Address Translation Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1026u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1030u16 , name : "Haswell Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1034u16 , name : "Xeon E3-1200 v3 Processor Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1042u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6375u16 , name : "ProDesk 600 G1" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 1046u16 , name : "4th Gen Core Processor Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" }] } , Device { vendor_id : 32902u16 , id : 1050u16 , name : "Xeon E3-1200 v3 Processor Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1054u16 , name : "4th Generation Core Processor Family Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1076u16 , name : "DH89XXCC Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1077u16 , name : "DH895XCC Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1078u16 , name : "DH8900CC Null Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1080u16 , name : "DH8900CC Series Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1082u16 , name : "DH8900CC Series Gigabit Fiber Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1084u16 , name : "DH8900CC Series Gigabit Backplane Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1088u16 , name : "DH8900CC Series Gigabit SFP Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1090u16 , name : "DH89XXCC Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1091u16 , name : "DH895XCC Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1154u16 , name : "82375EB/SB PCI to EISA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1155u16 , name : "82424TX/ZX [Saturn] CPU to PCI bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1156u16 , name : "82378ZB/IB, 82379AB (SIO, SIO.A) PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1158u16 , name : "82425EX/ZX [Aries] PCIset with ISA bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1187u16 , name : "82434LX/NX [Mercury/Neptune] Processor to PCI bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1232u16 , name : "82437FX [Triton FX]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1280u16 , name : "E8870 Processor bus control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1281u16 , name : "E8870 Memory controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1282u16 , name : "E8870 Scalability Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1283u16 , name : "E8870 Scalability Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1296u16 , name : "E8870IO Hub Interface Port 0 registers (8-bit compatibility port)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1297u16 , name : "E8870IO Hub Interface Port 1 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1298u16 , name : "E8870IO Hub Interface Port 2 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1299u16 , name : "E8870IO Hub Interface Port 3 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1300u16 , name : "E8870IO Hub Interface Port 4 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1301u16 , name : "E8870IO General SIOH registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1302u16 , name : "E8870IO RAS registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1328u16 , name : "E8870SP Scalability Port 0 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1329u16 , name : "E8870SP Scalability Port 1 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1330u16 , name : "E8870SP Scalability Port 2 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1331u16 , name : "E8870SP Scalability Port 3 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1332u16 , name : "E8870SP Scalability Port 4 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1333u16 , name : "E8870SP Scalability Port 5 registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1334u16 , name : "E8870SP Interleave registers 0 and 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1335u16 , name : "E8870SP Interleave registers 2 and 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1536u16 , name : "RAID Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 310u16 , name : "SRCU31L" } , SubSystem { subvendor : 32902u16 , subdevice : 431u16 , name : "SRCZCR" } , SubSystem { subvendor : 32902u16 , subdevice : 449u16 , name : "ICP Vortex GDT8546RZ" } , SubSystem { subvendor : 32902u16 , subdevice : 503u16 , name : "SCRU32" }] } , Device { vendor_id : 32902u16 , id : 1567u16 , name : "80303 I/O Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1668u16 , name : "H470 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1669u16 , name : "Z490 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1671u16 , name : "Q470 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1676u16 , name : "QM480 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1677u16 , name : "HM470 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1678u16 , name : "WM490 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1687u16 , name : "W480 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1696u16 , name : "400 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1697u16 , name : "400 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1699u16 , name : "400 Series Chipset Family SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1700u16 , name : "400 Series Chipset Family SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1702u16 , name : "400 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1704u16 , name : "400 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1705u16 , name : "400 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1706u16 , name : "400 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1707u16 , name : "400 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1708u16 , name : "400 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1709u16 , name : "400 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1710u16 , name : "400 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1711u16 , name : "400 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1712u16 , name : "400 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1713u16 , name : "400 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1714u16 , name : "400 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1715u16 , name : "400 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1716u16 , name : "400 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1717u16 , name : "400 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1718u16 , name : "400 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1719u16 , name : "400 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1720u16 , name : "400 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1721u16 , name : "400 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1722u16 , name : "400 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1723u16 , name : "400 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1724u16 , name : "400 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1725u16 , name : "400 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1726u16 , name : "400 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1727u16 , name : "400 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1728u16 , name : "400 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1729u16 , name : "400 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1730u16 , name : "400 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1731u16 , name : "400 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1735u16 , name : "400 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1736u16 , name : "400 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1746u16 , name : "400 Series Chipset Family SATA Controller (AHCI) (Desktop)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1747u16 , name : "400 Series Chipset Family SATA Controller (AHCI) (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1749u16 , name : "400 Series Chipset Family SATA Controller (RAID 0/1/5/10) no premium (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1750u16 , name : "Comet Lake PCH-H RAID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1751u16 , name : "400 Series Chipset Family SATA Controller (RAID 0/1/5/10) premium (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1758u16 , name : "400 Series Chipset Family SATA Controller (AHCI) Optane Caching" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1760u16 , name : "400 Series Chipset Family HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1761u16 , name : "400 Series Chipset Family HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1762u16 , name : "400 Series Chipset Family IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1763u16 , name : "400 Series Chipset Family Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1764u16 , name : "400 Series Chipset Family HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1765u16 , name : "400 Series Chipset Family HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1768u16 , name : "400 Series Chipset Family I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1769u16 , name : "400 Series Chipset Family I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1770u16 , name : "400 Series Chipset Family I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1771u16 , name : "400 Series Chipset Family I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1773u16 , name : "400 Series Chipset Family USB 3.2 Gen 2x1 (10 Gbs) xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1775u16 , name : "400 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1776u16 , name : "400 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5713u16 , name : "Dual Band Wi-Fi 6(802.11ax) Killer AX1650s 160MHz 2x2 [Cyclone Peak]" } , SubSystem { subvendor : 6742u16 , subdevice : 5714u16 , name : "Dual Band Wi-Fi 6(802.11ax) Killer AX1650i 160MHz 2x2 [Cyclone Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9560 160MHz 2x2 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 676u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9462 80MHz 1x1 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 17060u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9462 80MHz 1x1 [Jefferson Peak]" }] } , Device { vendor_id : 32902u16 , id : 1781u16 , name : "400 Series Chipset Family SCS3 SDXC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1785u16 , name : "400 Series Chipset Family Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1787u16 , name : "400 Series Chipset Family GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1788u16 , name : "400 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1792u16 , name : "CE Media Processor A/V Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1793u16 , name : "CE Media Processor NAND Flash Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1795u16 , name : "CE Media Processor Media Control Unit 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1796u16 , name : "CE Media Processor Video Capture Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1799u16 , name : "CE Media Processor SPI Slave" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1800u16 , name : "Atom Processor CE 4100" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 1801u16 , name : "Atom Processor CE 4200" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2048u16 , name : "Moorestown SPI Ctrl 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2049u16 , name : "Moorestown SPI Ctrl 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2050u16 , name : "Moorestown I2C 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2051u16 , name : "Moorestown I2C 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2052u16 , name : "Moorestown I2C 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2053u16 , name : "Moorestown Keyboard Ctrl" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2054u16 , name : "Moorestown USB Ctrl" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2055u16 , name : "Moorestown SD Host Ctrl 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2056u16 , name : "Moorestown SD Host Ctrl 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2057u16 , name : "Moorestown NAND Ctrl" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2058u16 , name : "Moorestown Audio Ctrl" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2059u16 , name : "Moorestown ISP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2060u16 , name : "Moorestown Security Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2061u16 , name : "Moorestown External Displays" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2062u16 , name : "Moorestown SCU IPC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2063u16 , name : "Moorestown GPIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2064u16 , name : "Moorestown Power Management Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2065u16 , name : "Moorestown OTG Ctrl" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2066u16 , name : "Moorestown SPI Ctrl 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2067u16 , name : "Moorestown SC DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2068u16 , name : "Moorestown LPE DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2069u16 , name : "Moorestown SSP0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2071u16 , name : "Medfield Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2072u16 , name : "Medfield Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2073u16 , name : "Medfield Serial IO I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2074u16 , name : "Medfield GPIO Controller [Core]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2075u16 , name : "Medfield Serial IO HSUART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2076u16 , name : "Medfield Serial IO HSUART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2077u16 , name : "Medfield Serial IO HSUART Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2078u16 , name : "Medfield Serial IO HSUART DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2079u16 , name : "Medfield GPIO Controller [AON]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2080u16 , name : "Medfield SD Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2081u16 , name : "Medfield SDIO Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2082u16 , name : "Medfield SDIO Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2083u16 , name : "Medfield eMMC Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2084u16 , name : "Medfield eMMC Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2087u16 , name : "Medfield Serial IO DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2088u16 , name : "Medfield Power Management Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2089u16 , name : "Medfield USB Device Controller (OTG)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2090u16 , name : "Medfield SCU IPC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2092u16 , name : "Medfield Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2093u16 , name : "Medfield Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2094u16 , name : "Medfield Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2181u16 , name : "Centrino Wireless-N + WiMAX 6150" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4869u16 , name : "Centrino Wireless-N + WiMAX 6150 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4871u16 , name : "Centrino Wireless-N + WiMAX 6150 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4901u16 , name : "Centrino Wireless-N + WiMAX 6150 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4903u16 , name : "Centrino Wireless-N + WiMAX 6150 BG" }] } , Device { vendor_id : 32902u16 , id : 2182u16 , name : "Centrino Wireless-N + WiMAX 6150" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4885u16 , name : "Centrino Wireless-N + WiMAX 6150 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4887u16 , name : "Centrino Wireless-N + WiMAX 6150 BG" }] } , Device { vendor_id : 32902u16 , id : 2183u16 , name : "Centrino Wireless-N 2230" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16482u16 , name : "Centrino Wireless-N 2230 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 17506u16 , name : "Centrino Wireless-N 2230 BGN" }] } , Device { vendor_id : 32902u16 , id : 2184u16 , name : "Centrino Wireless-N 2230" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16994u16 , name : "Centrino Wireless-N 2230 BGN" }] } , Device { vendor_id : 32902u16 , id : 2190u16 , name : "Centrino Advanced-N 6235" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16480u16 , name : "Centrino Advanced-N 6235 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 17504u16 , name : "Centrino Advanced-N 6235 AGN" }] } , Device { vendor_id : 32902u16 , id : 2191u16 , name : "Centrino Advanced-N 6235" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16992u16 , name : "Centrino Advanced-N 6235 AGN" }] } , Device { vendor_id : 32902u16 , id : 2192u16 , name : "Centrino Wireless-N 2200" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16418u16 , name : "Centrino Wireless-N 2200 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 17442u16 , name : "Centrino Wireless-N 2200 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 18466u16 , name : "Centrino Wireless-N 2200 BGN" }] } , Device { vendor_id : 32902u16 , id : 2193u16 , name : "Centrino Wireless-N 2200" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16930u16 , name : "Centrino Wireless-N 2200 BGN" }] } , Device { vendor_id : 32902u16 , id : 2194u16 , name : "Centrino Wireless-N 135" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 98u16 , name : "Centrino Wireless-N 135 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 1122u16 , name : "Centrino Wireless-N 135 BGN" }] } , Device { vendor_id : 32902u16 , id : 2195u16 , name : "Centrino Wireless-N 135" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 610u16 , name : "Centrino Wireless-N 135 BGN" }] } , Device { vendor_id : 32902u16 , id : 2196u16 , name : "Centrino Wireless-N 105" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 34u16 , name : "Centrino Wireless-N 105 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 1058u16 , name : "Centrino Wireless-N 105 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 2082u16 , name : "Centrino Wireless-N 105 BGN" }] } , Device { vendor_id : 32902u16 , id : 2197u16 , name : "Centrino Wireless-N 105" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 546u16 , name : "Centrino Wireless-N 105 BGN" }] } , Device { vendor_id : 32902u16 , id : 2198u16 , name : "Centrino Wireless-N 130" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20485u16 , name : "Centrino Wireless-N 130 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 20487u16 , name : "Centrino Wireless-N 130 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 20517u16 , name : "Centrino Wireless-N 130 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 20519u16 , name : "Centrino Wireless-N 130 BG" }] } , Device { vendor_id : 32902u16 , id : 2199u16 , name : "Centrino Wireless-N 130" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20501u16 , name : "Centrino Wireless-N 130 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 20503u16 , name : "Centrino Wireless-N 130 BG" }] } , Device { vendor_id : 32902u16 , id : 2215u16 , name : "Quark SoC X1000 SDIO / eMMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2222u16 , name : "Centrino Wireless-N 100" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4101u16 , name : "Centrino Wireless-N 100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4103u16 , name : "Centrino Wireless-N 100 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4133u16 , name : "Centrino Wireless-N 100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4135u16 , name : "Centrino Wireless-N 100 BG" }] } , Device { vendor_id : 32902u16 , id : 2223u16 , name : "Centrino Wireless-N 100" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4117u16 , name : "Centrino Wireless-N 100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4119u16 , name : "Centrino Wireless-N 100 BG" }] } , Device { vendor_id : 32902u16 , id : 2225u16 , name : "Wireless 7260" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16416u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16426u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16480u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16482u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16490u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16496u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16498u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16736u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16738u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16752u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17440u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17504u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17506u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17514u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17520u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17522u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17760u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17776u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 18542u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 18544u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 19052u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 19054u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 19056u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 19552u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 19568u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 20592u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 20594u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 20848u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 22384u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49184u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49194u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49248u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49250u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49258u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49264u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49266u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49504u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49506u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49520u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50016u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50208u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50272u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50274u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50288u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50290u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50528u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50544u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 51040u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 51056u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 52320u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 52336u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" }] } , Device { vendor_id : 32902u16 , id : 2226u16 , name : "Wireless 7260" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16928u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16992u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 16994u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17002u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17008u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17010u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17248u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 17264u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49696u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49760u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49762u16 , name : "Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49770u16 , name : "Dual Band Wireless-N 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49776u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 49778u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" } , SubSystem { subvendor : 32902u16 , subdevice : 50032u16 , name : "Dual Band Wireless-AC 7260 [Wilkins Peak 2]" }] } , Device { vendor_id : 32902u16 , id : 2227u16 , name : "Wireless 3160" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 96u16 , name : "Dual Band Wireless-N 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 98u16 , name : "Wireless-N 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 112u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 114u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 368u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 370u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 608u16 , name : "Dual Band Wireless-N 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 1136u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 1138u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 4208u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 4464u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 32864u16 , name : "Dual Band Wireless N-3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 32866u16 , name : "Wireless N-3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 32880u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 32882u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33136u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33138u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33904u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 34160u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" }] } , Device { vendor_id : 32902u16 , id : 2228u16 , name : "Wireless 3160" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 624u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 626u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 880u16 , name : "Dual Band Wireless-AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33376u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33392u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33394u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" } , SubSystem { subvendor : 32902u16 , subdevice : 33648u16 , name : "Dual Band Wireless AC 3160 [Wilkins Peak 1]" }] } , Device { vendor_id : 32902u16 , id : 2255u16 , name : "Atom Processor Z2760 Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2353u16 , name : "Atom Processor CE 2600 [Puma 6]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2356u16 , name : "Quark SoC X1000 I2C Controller and GPIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2357u16 , name : "Quark SoC X1000 SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2358u16 , name : "Quark SoC X1000 HS-UART" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2359u16 , name : "Quark SoC X1000 10/100 Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2361u16 , name : "Quark SoC X1000 USB EHCI Host Controller / USB 2.0 Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2362u16 , name : "Quark SoC X1000 USB OHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2364u16 , name : "WiGig(802.11ad) wireless network connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2387u16 , name : "PCIe Data Center SSD" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 14082u16 , name : "DC P3700 SSD" } , SubSystem { subvendor : 32902u16 , subdevice : 14083u16 , name : "DC P3700 SSD [2.5\" SFF]" } , SubSystem { subvendor : 32902u16 , subdevice : 14084u16 , name : "DC P3500 SSD [Add-in Card]" } , SubSystem { subvendor : 32902u16 , subdevice : 14085u16 , name : "DC P3500 SSD [2.5\" SFF]" } , SubSystem { subvendor : 32902u16 , subdevice : 14089u16 , name : "DC P3600 SSD [Add-in Card]" } , SubSystem { subvendor : 32902u16 , subdevice : 14090u16 , name : "DC P3600 SSD [2.5\" SFF]" } , SubSystem { subvendor : 32902u16 , subdevice : 14093u16 , name : "SSD 750 Series [Add-in Card]" } , SubSystem { subvendor : 32902u16 , subdevice : 14094u16 , name : "SSD 750 Series [2.5\" SFF]" }] } , Device { vendor_id : 32902u16 , id : 2392u16 , name : "Quark SoC X1000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2394u16 , name : "Wireless 7265" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4112u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20480u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20482u16 , name : "Wireless-N 7265 [Stone Peak 2 BGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20490u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20496u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20498u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20512u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20522u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20624u16 , name : "Dual Band Wireless-AC 7265 [Maple Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20736u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20738u16 , name : "Wireless-N 7265 [Stone Peak 2 BGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20746u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20752u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20754u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 20880u16 , name : "Dual Band Wireless-AC 7265 [Maple Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21504u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 21520u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21522u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21536u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 21648u16 , name : "Dual Band Wireless-AC 7265 [Maple Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21776u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21904u16 , name : "Dual Band Wireless-AC 7265 [Maple Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 36864u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 36874u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 36880u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 36882u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37136u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37138u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37392u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37648u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37888u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 37904u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 38160u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" }] } , Device { vendor_id : 32902u16 , id : 2395u16 , name : "Wireless 7265" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20992u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 20994u16 , name : "Wireless-N 7265 [Stone Peak 2 BGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 21002u16 , name : "Dual Band Wireless-N 7265 [Stone Peak 2 AGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 21008u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21010u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21136u16 , name : "Dual Band Wireless-AC 7265 [Maple Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 21250u16 , name : "Wireless-N 7265 [Stone Peak 2 BGN]" } , SubSystem { subvendor : 32902u16 , subdevice : 21264u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 37376u16 , name : "Dual Band Wireless-AC 7265 [Stone Peak 2 AGN]" }] } , Device { vendor_id : 32902u16 , id : 2398u16 , name : "Quark SoC X1000 Legacy Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2400u16 , name : "80960RP (i960RP) Microprocessor/Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2402u16 , name : "80960RM (i960RM) Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2404u16 , name : "80960RP (i960RP) Microprocessor/Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2421u16 , name : "Optane NVME SSD H10 with Solid State Storage [Teton Glacier]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2456u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors IEH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2466u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors VT-d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2467u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2468u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UFI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2470u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors MSM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2471u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PMON MSM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2475u16 , name : "RST VMD Managed Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2477u16 , name : "Optane NVME SSD H20 with Solid State Storage [Pyramid Glacier]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2500u16 , name : "PAC with Intel Arria 10 GX FPGA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2563u16 , name : "Haswell-ULT Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2564u16 , name : "Haswell-ULT DRAM Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 2566u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2572u16 , name : "Haswell-ULT HD Audio Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 2582u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 2594u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2598u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2602u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2606u16 , name : "Haswell-ULT Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2643u16 , name : "DC P3520 SSD" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2644u16 , name : "NVMe Datacenter SSD [3DNAND, Beta Rock Controller]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8161u16 , name : "Express Flash NVMe 1TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8162u16 , name : "Express Flash NVMe 2TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8163u16 , name : "Express Flash NVMe 4TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8164u16 , name : "Express Flash NVMe 4TB HHHL AIC (P4500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8174u16 , name : "Express Flash NVMe 1.6TB 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 4136u16 , subdevice : 8175u16 , name : "Express Flash NVMe 3.2TB 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 4136u16 , subdevice : 8176u16 , name : "Express Flash NVMe 6.4TB 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 4136u16 , subdevice : 8191u16 , name : "Express Flash NVMe 8.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4136u16 , subdevice : 8195u16 , name : "Express Flash NVMe 1.0 TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4136u16 , subdevice : 8196u16 , name : "Express Flash NVMe 2.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4136u16 , subdevice : 8197u16 , name : "Express Flash NVMe 4.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4238u16 , subdevice : 18544u16 , name : "NVMe PCIe 3.0 SSD 6.4TB AIC (P4608)" } , SubSystem { subvendor : 4238u16 , subdevice : 18545u16 , name : "NVMe PCIe 3.0 SSD 6.4TB 2.5-inch (P4600)" } , SubSystem { subvendor : 4238u16 , subdevice : 18553u16 , name : "NVMe PCIe 3.0 SSD v2 6.4TB AIC (P4618)" } , SubSystem { subvendor : 4238u16 , subdevice : 18554u16 , name : "NVMe PCIe 3.0 SSD v2 6.4TB 2.5-inch (P4610)" } , SubSystem { subvendor : 4407u16 , subdevice : 551u16 , name : "NVMe Datacenter SSD [3DNAND] 1.6TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4407u16 , subdevice : 552u16 , name : "NVMe Datacenter SSD [3DNAND] 2.0TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4407u16 , subdevice : 553u16 , name : "NVMe Datacenter SSD [3DNAND] 3.2TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4407u16 , subdevice : 555u16 , name : "NVMe Datacenter SSD [3DNAND] 1.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4407u16 , subdevice : 556u16 , name : "NVMe Datacenter SSD [3DNAND] 2.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4407u16 , subdevice : 557u16 , name : "NVMe Datacenter SSD [3DNAND] 4.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 4407u16 , subdevice : 561u16 , name : "NVMe Datacenter SSD [3DNAND] 0.5TB 2.5\" U.2 (P4501)" } , SubSystem { subvendor : 4407u16 , subdevice : 562u16 , name : "NVMe Datacenter SSD [3DNAND] 1.0TB 2.5\" U.2 (P4501)" } , SubSystem { subvendor : 4407u16 , subdevice : 563u16 , name : "NVMe Datacenter SSD [3DNAND] 2.0TB 2.5\" U.2 (P4501)" } , SubSystem { subvendor : 4407u16 , subdevice : 600u16 , name : "NVMe Datacenter SSD [3DNAND] 1.6TB 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 4407u16 , subdevice : 602u16 , name : "NVMe Datacenter SSD [3DNAND] 3.2TB 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 4407u16 , subdevice : 603u16 , name : "NVMe Datacenter SSD [3DNAND] 1.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4407u16 , subdevice : 604u16 , name : "NVMe Datacenter SSD [3DNAND] 2.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4407u16 , subdevice : 605u16 , name : "NVMe Datacenter SSD [3DNAND] 4.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 4407u16 , subdevice : 606u16 , name : "NVMe Datacenter SSD [3DNAND] 8.0TB 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 5520u16 , subdevice : 605u16 , name : "NVMe Datacenter SSD [3DNAND] 1.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 5520u16 , subdevice : 606u16 , name : "NVMe Datacenter SSD [3DNAND] 2.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 5520u16 , subdevice : 607u16 , name : "NVMe Datacenter SSD [3DNAND] 4.0TB 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 5520u16 , subdevice : 610u16 , name : "NVMe Datacenter SSD [3DNAND] 1.6TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 5520u16 , subdevice : 612u16 , name : "NVMe Datacenter SSD [3DNAND] 3.2TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 5520u16 , subdevice : 613u16 , name : "NVMe Datacenter SSD [3DNAND] 6.4TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 5520u16 , subdevice : 620u16 , name : "NVMe Datacenter SSD [3DNAND] 4.0TB AIC (P4500)" } , SubSystem { subvendor : 7497u16 , subdevice : 18178u16 , name : "Thinksystem Intel P4500 NVMe U.2" } , SubSystem { subvendor : 7497u16 , subdevice : 18180u16 , name : "Thinksystem Intel P4500 NVMe AIC" } , SubSystem { subvendor : 7497u16 , subdevice : 18194u16 , name : "Thinksystem Intel P4600 NVMe U.2" } , SubSystem { subvendor : 7497u16 , subdevice : 18196u16 , name : "Thinksystem Intel P4600 NVMe AIC" } , SubSystem { subvendor : 7497u16 , subdevice : 18434u16 , name : "Thinksystem U.2 P4510 NVMe SSD" } , SubSystem { subvendor : 7497u16 , subdevice : 18450u16 , name : "Thinksystem U.2 P4610 NVMe SSD" } , SubSystem { subvendor : 32902u16 , subdevice : 17160u16 , name : "SSD D5-P4320 and D5-P4326" } , SubSystem { subvendor : 32902u16 , subdevice : 18178u16 , name : "NVMe Datacenter SSD [3DNAND] SE 2.5\" U.2 (P4500)" } , SubSystem { subvendor : 32902u16 , subdevice : 18180u16 , name : "NVMe Datacenter SSD [3DNAND] SE AIC (P4500)" } , SubSystem { subvendor : 32902u16 , subdevice : 18194u16 , name : "NVMe Datacenter SSD [3DNAND] ME 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 32902u16 , subdevice : 18196u16 , name : "NVMe Datacenter SSD [3DNAND] ME AIC (P4600)" } , SubSystem { subvendor : 32902u16 , subdevice : 18434u16 , name : "NVMe Datacenter SSD [3DNAND] SE 2.5\" U.2 (P4510)" } , SubSystem { subvendor : 32902u16 , subdevice : 18436u16 , name : "NVMe Datacenter SSD [3DNAND] SE AIC (P4510)" } , SubSystem { subvendor : 32902u16 , subdevice : 18437u16 , name : "NVMe Datacenter SSD [3DNAND] SE M.2 (P4511)" } , SubSystem { subvendor : 32902u16 , subdevice : 18450u16 , name : "NVMe Datacenter SSD [3DNAND] ME 2.5\" U.2 (P4610)" } , SubSystem { subvendor : 32902u16 , subdevice : 18452u16 , name : "NVMe Datacenter SSD [3DNAND] ME AIC (P4610)" }] } , Device { vendor_id : 32902u16 , id : 2645u16 , name : "NVMe DC SSD [3DNAND, Beta Rock Controller]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8165u16 , name : "Express Flash NVMe 1.6TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8166u16 , name : "Express Flash NVMe 2TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8167u16 , name : "Express Flash NVMe 3.2TB 2.5\" U.2 (P4600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8168u16 , name : "Express Flash NVMe 2.0TB HHHL AIC (P4600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8169u16 , name : "Express Flash NVMe 4.0TB HHHL AIC (P4600)" }] } , Device { vendor_id : 32902u16 , id : 2816u16 , name : "Ice Lake CBDMA [QuickData Technology]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2851u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors IEH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2853u16 , name : "Data Streaming Accelerator (DSA)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2854u16 , name : "Thunderbolt 4 Bridge [Goshen Ridge 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2855u16 , name : "Thunderbolt 4 USB Controller [Goshen Ridge 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2859u16 , name : "PAC with Intel Stratix 10 SX FPGA [FPGA PAC D5005]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2912u16 , name : "NVMe DC SSD [3DNAND, Sentinel Rock Controller]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8288u16 , name : "NVMe SED MU U.2 1.6TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8289u16 , name : "NVMe SED MU U.2 3.2TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8290u16 , name : "NVMe SED MU U.2 6.4TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8292u16 , name : "NVMe SED RI U.2 1.92TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8293u16 , name : "NVMe SED RI U.2 3.84TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8294u16 , name : "NVMe SED RI U.2 7.68TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8350u16 , name : "NVMe MU U.2 1.6TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8351u16 , name : "NVMe MU U.2 3.2TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8448u16 , name : "NVMe MU U.2 6.4TB (P5600)" } , SubSystem { subvendor : 4136u16 , subdevice : 8450u16 , name : "NVMe RI U.2 1.92TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8451u16 , name : "NVMe RI U.2 3.84TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8452u16 , name : "NVMe RI U.2 7.68TB (P5500)" } , SubSystem { subvendor : 4136u16 , subdevice : 8602u16 , name : "NVMe P5316 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8603u16 , name : "NVMe P5316 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8604u16 , name : "NVMe SED P5316 RI 15.36" } , SubSystem { subvendor : 4136u16 , subdevice : 8605u16 , name : "NVMe SED P5316 RI 30.72" } , SubSystem { subvendor : 4136u16 , subdevice : 8606u16 , name : "NVMe FIPS P5316 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8607u16 , name : "NVMe FIPS P5316 RI 30.72" } , SubSystem { subvendor : 32902u16 , subdevice : 32776u16 , name : "NVMe Datacenter SSD [3DNAND] SE 2.5\" U.2 (P5510)" } , SubSystem { subvendor : 32902u16 , subdevice : 36104u16 , name : "NVMe Datacenter SSD [3DNAND] VE 2.5\" U.2 (P5316)" } , SubSystem { subvendor : 32902u16 , subdevice : 36125u16 , name : "NVMe Datacenter SSD [3DNAND] VE E1.L 9.5/18mm (P5316)" } , SubSystem { subvendor : 32902u16 , subdevice : 49160u16 , name : "NVMe Datacenter SSD [3DNAND] SE U.2 15mm (P5530)" }] } , Device { vendor_id : 32902u16 , id : 2921u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1450]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 2926u16 , name : "Ponte Vecchio XT (1 Tile) [Data Center GPU Max 1100C]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3024u16 , name : "Ponte Vecchio XL (2 Tile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3028u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1550VG]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3029u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1550]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3030u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1550]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3031u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1350]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3032u16 , name : "Ponte Vecchio XT (2 Tile) [Data Center GPU Max 1350]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3033u16 , name : "Ponte Vecchio XT (1 Tile) [Data Center GPU Max 1100]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3034u16 , name : "Ponte Vecchio XT (1 Tile) [Data Center GPU Max 1100]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3035u16 , name : "Ponte Vecchio XT (1 Tile) [Data Center GPU Max 1100]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3040u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3041u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" }] } , Device { vendor_id : 32902u16 , id : 3042u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3043u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3044u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3045u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3046u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3047u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3048u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3049u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3050u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3051u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3052u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3053u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3054u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3055u16 , name : "Atom Processor D2xxx/N2xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3056u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3057u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3058u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3059u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3060u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3061u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" }] } , Device { vendor_id : 32902u16 , id : 3062u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3063u16 , name : "Atom Processor D2xxx/N2xxx DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3072u16 , name : "4th Gen Core Processor DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 3073u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor PCI Express x16 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3076u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor DRAM Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" }] } , Device { vendor_id : 32902u16 , id : 3077u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor PCI Express x8 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3080u16 , name : "Xeon E3-1200 v3 Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3081u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor PCI Express x4 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3084u16 , name : "Xeon E3-1200 v3/4th Gen Core Processor HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 3136u16 , name : "Atom Processor CE 5300" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3142u16 , name : "Atom Processor S1200 PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3143u16 , name : "Atom Processor S1200 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3144u16 , name : "Atom Processor S1200 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3145u16 , name : "Atom Processor S1200 PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3150u16 , name : "Atom Processor S1200 NTB Primary" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3152u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3153u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3154u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3155u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3156u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3157u16 , name : "Atom Processor S1200 DFX 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3158u16 , name : "Atom Processor S1200 DFX 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3161u16 , name : "Atom Processor S1200 SMBus 2.0 Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3162u16 , name : "Atom Processor S1200 SMBus 2.0 Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3163u16 , name : "Atom Processor S1200 SMBus Controller 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3164u16 , name : "Atom Processor S1200 SMBus Controller 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3165u16 , name : "Atom Processor S1200 SMBus Controller 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3166u16 , name : "Atom Processor S1200 SMBus Controller 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3167u16 , name : "Atom Processor S1200 UART" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3168u16 , name : "Atom Processor S1200 Integrated Legacy Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3184u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3185u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3186u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3187u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3188u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3189u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3190u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3191u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3192u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3193u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3194u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3195u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3196u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3197u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3198u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3199u16 , name : "Atom Processor S1200 Internal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3320u16 , name : "Ethernet Controller X710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Controller X710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Controller X710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" }] } , Device { vendor_id : 32902u16 , id : 3326u16 , name : "In-Memory Analytics Accelerator (IAA)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3328u16 , name : "Crystal Well DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3329u16 , name : "Crystal Well PCI Express x16 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3331u16 , name : "Crystal Well Dynamic Platform and Thermal Framework Processor Participant" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3332u16 , name : "Crystal Well DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3333u16 , name : "Crystal Well PCI Express x8 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3337u16 , name : "Crystal Well PCI Express x4 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3340u16 , name : "Crystal Well HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3346u16 , name : "Crystal Well Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3350u16 , name : "Crystal Well Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3362u16 , name : "Crystal Well Integrated Iris Pro Graphics 5200" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3366u16 , name : "Crystal Well Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3382u16 , name : "Crystal Well Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3404u16 , name : "400 Series Chipset Family GbE Controller (Corporate/vPro)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3405u16 , name : "400 Series Chipset Family GbE Controller (Consumer)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 3405u16 , name : "Ethernet Connection (11) I219-V" }] } , Device { vendor_id : 32902u16 , id : 3406u16 , name : "400 Series Chipset Family On-Package GbE Controller (Corporate/vPro)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3407u16 , name : "400 Series Chipset Family On-Package GbE Controller (Consumer)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3411u16 , name : "Ethernet Connection (12) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3413u16 , name : "Ethernet Connection (12) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3416u16 , name : "Ethernet Controller XXV710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Controller XXV710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Controller XXV710 Intel(R) FPGA Programmable Acceleration Card N3000 for Networking" }] } , Device { vendor_id : 32902u16 , id : 3487u16 , name : "Ethernet Controller I225-IT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3504u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3505u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3506u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3507u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3508u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors NTB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3510u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3511u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3512u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3513u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors PCIe and CXL.io Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3525u16 , name : "Ethernet Connection (23) I219-LM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 3526u16 , name : "Ethernet Connection (23) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3527u16 , name : "Ethernet Connection I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3528u16 , name : "Ethernet Connection I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3533u16 , name : "Ethernet Connection C825-X" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3538u16 , name : "Ethernet Network Adapter I710" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "I710T4LG 4x1 GbE RJ45 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 739u16 , name : "I710T4LG 4x1 GbE RJ45 PCIe NIC" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Network Adapter I710-T4L" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet Network Adapter I710-T4L" } , SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Ethernet Network Adapter I710-T4L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16410u16 , name : "Ethernet Network Adapter I710-T4L" } , SubSystem { subvendor : 32902u16 , subdevice : 16411u16 , name : "Ethernet Network Adapter I710-T4L for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 3541u16 , name : "Ethernet Adaptive Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3546u16 , name : "Ethernet Connection X722 for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 118u16 , name : "Ethernet Connection F102IX722 for 10GbE SFP" } , SubSystem { subvendor : 7124u16 , subdevice : 130u16 , name : "Ethernet Connection F102IX722 for 10GbE SFP" }] } , Device { vendor_id : 32902u16 , id : 3584u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DMI2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3585u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port in DMI2 Mode" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3586u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 1a" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3587u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 1b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3588u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 2a" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3589u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 2b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3590u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 2c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3591u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 2d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3592u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 3a" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3593u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 3b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3594u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 3c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3595u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 PCI Express Root Port 3d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3600u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3603u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3607u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3608u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3612u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3613u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 R2PCIe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3614u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 UBOX Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3615u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 UBOX Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3616u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3617u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3618u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3619u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3620u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3621u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3622u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3623u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Crystal Beach DMA Channel 7" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3624u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 VTd/Memory Map/Misc" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3625u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Memory Hotplug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3626u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IIO RAS" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3628u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 IOAPIC" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 3630u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 CBDMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3631u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 CBDMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3632u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Home Agent 0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3634u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3635u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3636u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 R2PCIe" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3638u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Performance Ring Monitoring" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3639u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Performance Ring Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3640u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3642u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3646u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Performance Ring Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3647u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Performance Ring Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3648u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3649u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3651u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Reut 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3652u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Reut 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3653u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Agent Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3655u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Agent Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3680u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3688u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Target Address/Thermal Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3690u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3691u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3692u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3693u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3697u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3700u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 R2PCIe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3701u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 R2PCIe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3703u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3705u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3709u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 UBOX Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3711u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3712u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3713u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Ring Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3715u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Reut 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3716u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Reut 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3717u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Agent Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3719u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3728u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3731u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3732u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Reut 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3733u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 QPI Link Agent Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3744u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Home Agent 0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "Xeon E5 v2 on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 3752u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Target Address/Thermal Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3754u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3755u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3756u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3757u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3758u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3759u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3760u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 Thermal Control 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3761u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 Thermal Control 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3762u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 ERROR Registers 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3763u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 ERROR Registers 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3764u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 Thermal Control 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3765u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 Thermal Control 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3766u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 ERROR Registers 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3767u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 1 Channel 0-3 ERROR Registers 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3772u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3774u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3775u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3776u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Power Control Unit 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3777u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Power Control Unit 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3778u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Power Control Unit 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3779u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Power Control Unit 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3780u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Power Control Unit 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3784u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3785u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Broadcast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3786u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Broadcast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3800u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3801u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3804u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3805u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3806u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3807u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3808u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3809u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3810u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3811u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3812u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3813u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3814u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3815u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3816u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3817u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3818u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3819u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3820u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3821u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3822u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3824u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 Thermal Control 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3825u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 Thermal Control 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3826u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 ERROR Registers 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3827u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 ERROR Registers 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3828u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 Thermal Control 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3829u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 Thermal Control 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3830u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 ERROR Registers 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3831u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 Integrated Memory Controller 0 Channel 0-3 ERROR Registers 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3832u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3833u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3834u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3835u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3836u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3837u16 , name : "Xeon E7 v2/Xeon E5 v2/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3840u16 , name : "Atom Processor Z36xxx/Z37xxx Series SoC Transaction Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3844u16 , name : "Atom Processor Z36xxx/Z37xxx Series High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3846u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3848u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 PWM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3849u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 PWM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3850u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 HSUART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3852u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 HSUART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3854u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO1 SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3858u16 , name : "Atom Processor E3800/CE2700 Series SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3860u16 , name : "Atom Processor Z36xxx/Z37xxx Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3861u16 , name : "Atom Processor Z36xxx/Z37xxx Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3862u16 , name : "Atom Processor Z36xxx/Z37xxx Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3864u16 , name : "Atom Processor Z36xxx/Z37xxx Series Trusted Execution Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3868u16 , name : "Atom Processor Z36xxx/Z37xxx Series Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3872u16 , name : "Atom Processor E3800 Series SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3873u16 , name : "Atom Processor E3800 Series SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3874u16 , name : "Atom Processor E3800 Series SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3875u16 , name : "Atom Processor E3800 Series SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3880u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPE Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3889u16 , name : "Atom Processor Z36xxx/Z37xxx Series Graphics & Display" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3892u16 , name : "Atom Processor Z36xxx/Z37xxx Series USB EHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3893u16 , name : "Atom Processor Z36xxx/Z37xxx, Celeron N2000 Series USB xHCI" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 2358u16 , name : "Aspire ES1" }] } , Device { vendor_id : 32902u16 , id : 3895u16 , name : "Atom Processor Z36xxx/Z37xxx Series OTG USB Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3896u16 , name : "Atom Processor Z36xxx/Z37xxx Series Camera ISP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3904u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3905u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3906u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3907u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3908u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3909u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3910u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3911u16 , name : "Atom Processor Z36xxx/Z37xxx Series LPIO2 I2C Controller #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3912u16 , name : "Atom Processor E3800 Series PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3914u16 , name : "Atom Processor E3800 Series PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3916u16 , name : "Atom Processor E3800 Series PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3918u16 , name : "Atom Processor E3800 Series PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 3920u16 , name : "Atom Processor E3800 Series eMMC 4.5 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4096u16 , name : "82542 Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45279u16 , name : "NC6132 Gigabit Ethernet Adapter (1000-SX)" } , SubSystem { subvendor : 3601u16 , subdevice : 45280u16 , name : "NC6133 Gigabit Ethernet Adapter (1000-LX)" } , SubSystem { subvendor : 3601u16 , subdevice : 45347u16 , name : "NC6134 Gigabit Ethernet Adapter (1000-LX)" } , SubSystem { subvendor : 4116u16 , subdevice : 281u16 , name : "Netfinity Gigabit Ethernet SX Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4096u16 , name : "PRO/1000 Gigabit Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4097u16 , name : "82543GC Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 74u16 , name : "NC6136 Gigabit Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 490u16 , name : "Netfinity Gigabit Ethernet SX Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4098u16 , name : "PRO/1000 F Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "PRO/1000 F Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4098u16 , name : "Pro 100 LAN+Modem 56 Cardbus II" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 8206u16 , name : "Pro 100 LAN+Modem 56 Cardbus II" } , SubSystem { subvendor : 32902u16 , subdevice : 8211u16 , name : "Pro 100 SR Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8215u16 , name : "Pro 100 S Combo Mobile Adapter" }] } , Device { vendor_id : 32902u16 , id : 4100u16 , name : "82543GC Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 73u16 , name : "NC7132 Gigabit Upgrade Module" } , SubSystem { subvendor : 3601u16 , subdevice : 45476u16 , name : "NC7131 Gigabit Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 4338u16 , name : "Gigabit Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4100u16 , name : "PRO/1000 T Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8196u16 , name : "PRO/1000 T Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4104u16 , name : "82544EI Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 617u16 , name : "iSeries 1000/100/10 Ethernet Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 283u16 , name : "PowerEdge 1650/2550" } , SubSystem { subvendor : 4136u16 , subdevice : 284u16 , name : "PRO/1000 XT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4359u16 , name : "PRO/1000 XT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8455u16 , name : "PRO/1000 XT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8464u16 , name : "PRO/1000 XT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 12552u16 , name : "PRO/1000 XT Network Connection" }] } , Device { vendor_id : 32902u16 , id : 4105u16 , name : "82544EI Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 616u16 , name : "iSeries Gigabit Ethernet Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4361u16 , name : "PRO/1000 XF Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8457u16 , name : "PRO/1000 XF Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4106u16 , name : "82540EM Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4108u16 , name : "82544GC Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4370u16 , name : "PRO/1000 T Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8466u16 , name : "PRO/1000 T Desktop Adapter" }] } , Device { vendor_id : 32902u16 , id : 4109u16 , name : "82544GC Gigabit Ethernet Controller (LOM)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 291u16 , name : "PRO/1000 XT Network Connection" } , SubSystem { subvendor : 4217u16 , subdevice : 35103u16 , name : "82544GC Based Network Connection" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 4365u16 , name : "82544GC Based Network Connection" }] } , Device { vendor_id : 32902u16 , id : 4110u16 , name : "82540EM Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 613u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4116u16 , subdevice : 618u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4136u16 , subdevice : 46u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 308u16 , name : "PowerEdge 600SC" } , SubSystem { subvendor : 4136u16 , subdevice : 337u16 , name : "Optiplex GX270" } , SubSystem { subvendor : 4219u16 , subdevice : 35104u16 , name : "PRO/1000 MT Desktop Adapter" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 30u16 , name : "PRO/1000 MT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 46u16 , name : "PRO/1000 MT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4982u16 , name : "PRO/1000 GT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 5238u16 , name : "PRO/1000 GT Desktop Adapter" }] } , Device { vendor_id : 32902u16 , id : 4111u16 , name : "82545EM Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 617u16 , name : "iSeries 1000/100/10 Ethernet Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 654u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 5549u16 , subdevice : 1872u16 , name : "PRO/1000 MT Single Port Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4096u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4097u16 , name : "PRO/1000 MT Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4112u16 , name : "82546EB Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 219u16 , name : "NC7170 Gigabit Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 636u16 , name : "PRO/1000 MT Dual Port Network Adapter" } , SubSystem { subvendor : 5549u16 , subdevice : 1888u16 , name : "PRO/1000 MT Dual Port Adapter" } , SubSystem { subvendor : 6395u16 , subdevice : 30834u16 , name : "RESlink-X" } , SubSystem { subvendor : 8129u16 , subdevice : 38u16 , name : "Niagara 2260 Bypass Card" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4256u16 , name : "CA3/CR3 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 4113u16 , name : "PRO/1000 MT Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4114u16 , name : "PRO/1000 MT Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4122u16 , name : "PRO/1000 MT Dual Port Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 13348u16 , name : "SE7501HG2 Mainboard" }] } , Device { vendor_id : 32902u16 , id : 4113u16 , name : "82545EM Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 616u16 , name : "iSeries Gigabit Ethernet Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4098u16 , name : "PRO/1000 MF Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "PRO/1000 MF Server Adapter (LX)" }] } , Device { vendor_id : 32902u16 , id : 4114u16 , name : "82546EB Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 220u16 , name : "NC6170 Gigabit Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4114u16 , name : "PRO/1000 MF Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4115u16 , name : "82541EI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 19u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4115u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4371u16 , name : "PRO/1000 MT Desktop Adapter" }] } , Device { vendor_id : 32902u16 , id : 4116u16 , name : "82541ER Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20u16 , name : "PRO/1000 MT Desktop Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4116u16 , name : "PRO/1000 MT Network Connection" }] } , Device { vendor_id : 32902u16 , id : 4117u16 , name : "82540EM Gigabit Ethernet Controller (LOM)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4117u16 , name : "PRO/1000 MT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4118u16 , name : "82540EP Gigabit Ethernet Controller (Mobile)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1324u16 , name : "PRO/1000 MT Mobile Connection" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "PRO/1000 MT Mobile Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4118u16 , name : "PRO/1000 MT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4119u16 , name : "82540EP Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4119u16 , name : "PR0/1000 MT Desktop Connection" }] } , Device { vendor_id : 32902u16 , id : 4120u16 , name : "82541EI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4120u16 , name : "PRO/1000 MT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4121u16 , name : "82547EI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 4121u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "Intel Gigabit Ethernet (Kenai II)" } , SubSystem { subvendor : 32902u16 , subdevice : 4121u16 , name : "PRO/1000 CT Desktop Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12319u16 , name : "D865PERL mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 12325u16 , name : "D875PBZ motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 12332u16 , name : "D865GBF Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 4122u16 , name : "82547EI Gigabit Ethernet Controller (Mobile)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4122u16 , name : "PRO/1000 CT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4125u16 , name : "82546EB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4096u16 , name : "PRO/1000 MT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4126u16 , name : "82540EP Gigabit Ethernet Controller (Mobile)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1353u16 , name : "Thinkpad" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "PRO/1000 MT Mobile Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4126u16 , name : "PRO/1000 MT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4127u16 , name : "Ethernet Controller V710 for 5GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4134u16 , name : "82545GM Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 32902u16 , subdevice : 4096u16 , name : "PRO/1000 MT Server Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4097u16 , name : "PRO/1000 MT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4098u16 , name : "PRO/1000 MT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "PRO/1000 GT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4134u16 , name : "PRO/1000 MT Server Connection" }] } , Device { vendor_id : 32902u16 , id : 4135u16 , name : "82545GM Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12547u16 , name : "NC310F PCI-X Gigabit Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4097u16 , name : "PRO/1000 MF Server Adapter(LX)" } , SubSystem { subvendor : 32902u16 , subdevice : 4098u16 , name : "PRO/1000 MF Server Adapter(LX)" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "PRO/1000 MF Server Adapter(LX)" } , SubSystem { subvendor : 32902u16 , subdevice : 4135u16 , name : "PRO/1000 MF Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4136u16 , name : "82545GM Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4136u16 , name : "PRO/1000 MB Server Connection" }] } , Device { vendor_id : 32902u16 , id : 4137u16 , name : "82559 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4144u16 , name : "82559 InBusiness 10/100" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4145u16 , name : "82801CAM (ICH3) PRO/100 VE (LOM) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 521u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "Vaio PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 4173u16 , subdevice : 33084u16 , name : "Vaio PCG-GRV616G" } , SubSystem { subvendor : 4219u16 , subdevice : 21328u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 5197u16 , subdevice : 49152u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 5197u16 , subdevice : 49153u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 5197u16 , subdevice : 49155u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 5197u16 , subdevice : 49158u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 32902u16 , id : 4146u16 , name : "82801CAM (ICH3) PRO/100 VE Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4147u16 , name : "82801CAM (ICH3) PRO/100 VM (LOM) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4148u16 , name : "82801CAM (ICH3) PRO/100 VM Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4149u16 , name : "82801CAM (ICH3)/82562EH (LOM) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4150u16 , name : "82801CAM (ICH3) 82562EH Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4151u16 , name : "82801CAM (ICH3) Chipset Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4152u16 , name : "82801CAM (ICH3) PRO/100 VM (KM) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 152u16 , name : "Evo N600c" }] } , Device { vendor_id : 32902u16 , id : 4153u16 , name : "82801DB PRO/100 VE (LOM) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard ethernet ETH1" }] } , Device { vendor_id : 32902u16 , id : 4154u16 , name : "82801DB PRO/100 VE (CNR) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4155u16 , name : "82801DB PRO/100 VM (LOM) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4156u16 , name : "82801DB PRO/100 VM (CNR) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4157u16 , name : "82801DB PRO/100 VE (MOB) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1314u16 , name : "ThinkPad R40" } , SubSystem { subvendor : 4136u16 , subdevice : 8194u16 , name : "Latitude D500" } , SubSystem { subvendor : 32902u16 , subdevice : 4157u16 , name : "82562EZ 10/100 Ethernet Controller" }] } , Device { vendor_id : 32902u16 , id : 4158u16 , name : "82801DB PRO/100 VM (MOB) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4160u16 , name : "536EP Data Fax Modem" , subsystems : & [SubSystem { subvendor : 5822u16 , subdevice : 4160u16 , name : "V.9X DSP Data Fax Modem" }] } , Device { vendor_id : 32902u16 , id : 4163u16 , name : "PRO/Wireless LAN 2100 3B Mini PCI Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 32902u16 , subdevice : 9506u16 , name : "Samsung X10/P30 integrated WLAN" } , SubSystem { subvendor : 32902u16 , subdevice : 9511u16 , name : "MIM2000/Centrino" } , SubSystem { subvendor : 32902u16 , subdevice : 9569u16 , name : "Dell Latitude D800" } , SubSystem { subvendor : 32902u16 , subdevice : 9601u16 , name : "Toshiba Satellite M10" }] } , Device { vendor_id : 32902u16 , id : 4168u16 , name : "82597EX 10GbE Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "PRO/10GbE LR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41247u16 , name : "PRO/10GbE LR Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4169u16 , name : "82566MM Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 6058u16 , subdevice : 8377u16 , name : "ThinkPad T61/R61" }] } , Device { vendor_id : 32902u16 , id : 4170u16 , name : "82566DM Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4171u16 , name : "82566DC Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4172u16 , name : "82562V 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4173u16 , name : "82566MC Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4174u16 , name : "Ethernet Controller X710 for 10 Gigabit SFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4175u16 , name : "Ethernet Controller X710 for 10 Gigabit backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4176u16 , name : "82562EZ 10/100 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4136u16 , subdevice : 413u16 , name : "Dimension 3000" } , SubSystem { subvendor : 5218u16 , subdevice : 29324u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30092u16 , name : "MS-6758 (875P Neo)" } , SubSystem { subvendor : 32902u16 , subdevice : 12320u16 , name : "D865PERL mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 12335u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 4177u16 , name : "82801EB/ER (ICH5/ICH5R) integrated LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4178u16 , name : "PRO/100 VM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4179u16 , name : "PRO/100 VM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4180u16 , name : "PRO/100 VE Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4181u16 , name : "PRO/100 VM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4182u16 , name : "PRO/100 VE Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4183u16 , name : "PRO/100 VE Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4185u16 , name : "82551QM Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4187u16 , name : "82546GB Gigabit Ethernet Controller (Copper)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4190u16 , name : "82571EB/82571GB Gigabit Ethernet Controller D0/D1 (copper applications)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28740u16 , name : "NC360T PCI Express Dual Port Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 28750u16 , name : "Dual Port 1000Base-T (PCIe) [AD337A]" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 24579u16 , name : "Telum GE-QT" } , SubSystem { subvendor : 6367u16 , subdevice : 4628u16 , name : "2x 1GbE, PCIe x1, dual Intel 82571EB chips" } , SubSystem { subvendor : 32902u16 , subdevice : 94u16 , name : "PRO/1000 PT Dual Port Server Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4190u16 , name : "PRO/1000 PT Dual Port Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4309u16 , name : "82571PT Gigabit PT Quad Port Server ExpressModule" } , SubSystem { subvendor : 32902u16 , subdevice : 4446u16 , name : "PRO/1000 PT Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4702u16 , name : "PRO/1000 PT Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4958u16 , name : "PRO/1000 PT Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4191u16 , name : "82571EB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28751u16 , name : "Dual Port 1000Base-SX (PCIe) [AD338A]" } , SubSystem { subvendor : 32902u16 , subdevice : 90u16 , name : "PRO/1000 PF Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4447u16 , name : "PRO/1000 PF Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4703u16 , name : "PRO/1000 PF Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4959u16 , name : "PRO/1000 PF Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4192u16 , name : "82571EB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 96u16 , name : "PRO/1000 PB Dual Port Server Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4192u16 , name : "PRO/1000 PB Dual Port Server Connection" }] } , Device { vendor_id : 32902u16 , id : 4196u16 , name : "82562ET/EZ/GT/GZ - PRO/100 VE (LOM) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33016u16 , name : "P5GD1-VW Mainboard" }] } , Device { vendor_id : 32902u16 , id : 4197u16 , name : "82562ET/EZ/GT/GZ - PRO/100 VE Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4198u16 , name : "82562 EM/EX/GX - PRO/100 VM (LOM) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4199u16 , name : "82562 EM/EX/GX - PRO/100 VM Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4200u16 , name : "82562ET/EZ/GT/GZ - PRO/100 VE (LOM) Ethernet Controller Mobile" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" }] } , Device { vendor_id : 32902u16 , id : 4201u16 , name : "82562EM/EX/GX - PRO/100 VM (LOM) Ethernet Controller Mobile" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4202u16 , name : "82562G - PRO/100 VE (LOM) Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4203u16 , name : "82562G - PRO/100 VE Ethernet Controller Mobile" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4213u16 , name : "82547GI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 357u16 , name : "PowerEdge 750" } , SubSystem { subvendor : 32902u16 , subdevice : 117u16 , name : "PRO/1000 CT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4213u16 , name : "PRO/1000 CT Network Connection" }] } , Device { vendor_id : 32902u16 , id : 4214u16 , name : "82541GI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 357u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 4136u16 , subdevice : 4205u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 118u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4214u16 , name : "PRO/1000 MT Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4470u16 , name : "PRO/1000 MT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4726u16 , name : "PRO/1000 MT Network Adapter" }] } , Device { vendor_id : 32902u16 , id : 4215u16 , name : "82541GI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "PRO/1000 MT Mobile Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 119u16 , name : "PRO/1000 MT Mobile Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4215u16 , name : "PRO/1000 MT Mobile Connection" }] } , Device { vendor_id : 32902u16 , id : 4216u16 , name : "82541ER Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4216u16 , name : "82541ER-based Network Connection" }] } , Device { vendor_id : 32902u16 , id : 4217u16 , name : "82546GB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4774u16 , name : "Dual Port 1000Base-T [A9900A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4815u16 , name : "Core Dual Port 1000Base-T [AB352A]" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer Gigabit Ethernet" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 8129u16 , subdevice : 39u16 , name : "Niagara 2261 Failover NIC" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 121u16 , name : "PRO/1000 MT Dual Port Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4217u16 , name : "PRO/1000 MT Dual Port Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4473u16 , name : "PRO/1000 MT Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4474u16 , name : "PRO/1000 MT Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4218u16 , name : "82546GB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4776u16 , name : "Dual Port 1000base-SX [A9899A]" } , SubSystem { subvendor : 32902u16 , subdevice : 4218u16 , name : "PRO/1000 MF Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4730u16 , name : "PRO/1000 MF Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4219u16 , name : "82546GB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 123u16 , name : "PRO/1000 MB Dual Port Server Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4219u16 , name : "PRO/1000 MB Dual Port Server Connection" }] } , Device { vendor_id : 32902u16 , id : 4220u16 , name : "82541PI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4982u16 , name : "PRO/1000 GT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 5238u16 , name : "PRO/1000 GT Desktop Adapter" }] } , Device { vendor_id : 32902u16 , id : 4221u16 , name : "82572EI Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4226u16 , name : "PRO/1000 PT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4228u16 , name : "PRO/1000 PT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4242u16 , name : "PRO/1000 PT Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4222u16 , name : "82572EI Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4228u16 , name : "PRO/1000 PF Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4229u16 , name : "PRO/1000 PF Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4244u16 , name : "PRO/1000 PF Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4223u16 , name : "82572EI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4224u16 , name : "FA82537EP 56K V.92 Data/Fax Modem PCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4225u16 , name : "631xESB/632xESB LAN Controller Copper" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4226u16 , name : "631xESB/632xESB LAN Controller fiber" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4227u16 , name : "631xESB/632xESB LAN Controller SERDES" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4228u16 , name : "631xESB/632xESB IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4229u16 , name : "631xESB/632xESB Serial Port Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4230u16 , name : "631xESB/632xESB IPMI/KCS0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4231u16 , name : "631xESB/632xESB UHCI Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4233u16 , name : "631xESB/632xESB BT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4234u16 , name : "82546GB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4234u16 , name : "PRO/1000 P Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4490u16 , name : "PRO/1000 P Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4235u16 , name : "82573V Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 5996u16 , name : "on board on MSI 945P - NEO (MS-7176)" }] } , Device { vendor_id : 32902u16 , id : 4236u16 , name : "82573E Gigabit Ethernet Controller (Copper)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4238u16 , name : "82573E KCS (Active Management)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4239u16 , name : "Active Management Technology - SOL" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4241u16 , name : "PRO/100 VM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4242u16 , name : "PRO/100 VE Network Connection" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" }] } , Device { vendor_id : 32902u16 , id : 4243u16 , name : "PRO/100 VM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4244u16 , name : "PRO/100 VE Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4245u16 , name : "PRO/100 VE Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4246u16 , name : "80003ES2LAN Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 4246u16 , name : "Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 4247u16 , name : "631xESB/632xESB DPT LAN Controller (Fiber)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4248u16 , name : "80003ES2LAN Gigabit Ethernet Controller (Serdes)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4249u16 , name : "82546GB Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4249u16 , name : "PRO/1000 GT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4250u16 , name : "82573L Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 65296u16 , name : "PRO/1000 PL" } , SubSystem { subvendor : 6058u16 , subdevice : 8193u16 , name : "ThinkPad T60" } , SubSystem { subvendor : 6058u16 , subdevice : 8318u16 , name : "ThinkPad X60/X60s" } , SubSystem { subvendor : 32902u16 , subdevice : 4250u16 , name : "PRO/1000 PL Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12444u16 , name : "Desktop Board D945GTP" } , SubSystem { subvendor : 32902u16 , subdevice : 12453u16 , name : "Desktop Board D975XBX" }] } , Device { vendor_id : 32902u16 , id : 4251u16 , name : "82546GB PRO/1000 GF Quad Port Server Adapter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4254u16 , name : "82597EX 10GbE Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "PRO/10GbE CX4 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41247u16 , name : "PRO/10GbE CX4 Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4256u16 , name : "82571EB PRO/1000 AT Quad Port Bypass Adapter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4257u16 , name : "82571EB PRO/1000 AF Quad Port Bypass Adapter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4260u16 , name : "82571EB Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4260u16 , name : "PRO/1000 PT Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4516u16 , name : "PRO/1000 PT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4261u16 , name : "82571EB Gigabit Ethernet Controller (Fiber)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4261u16 , name : "PRO/1000 PF Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4262u16 , name : "PRO/1000 PF Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4262u16 , name : "82599EB 10-Gigabit Dummy Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4263u16 , name : "82575EB Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 4263u16 , name : "AOC-SG-i2" } , SubSystem { subvendor : 32902u16 , subdevice : 4264u16 , name : "82575EB Gigabit Riser Card" }] } , Device { vendor_id : 32902u16 , id : 4265u16 , name : "82575EB Gigabit Backplane Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4272u16 , name : "82573L PRO/1000 PL Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4274u16 , name : "82573V PRO/1000 PM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4275u16 , name : "82573E PRO/1000 PM Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4276u16 , name : "82573L PRO/1000 PL Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4277u16 , name : "82546GB Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12553u16 , name : "NC340T PCI-X Quad-port Gigabit Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4249u16 , name : "PRO/1000 GT Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4505u16 , name : "PRO/1000 GT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4278u16 , name : "82598 10GbE PCI-Express Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4281u16 , name : "82572EI Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28746u16 , name : "110T PCIe Gigabit Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4227u16 , name : "PRO/1000 PT Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4243u16 , name : "PRO/1000 PT Desktop Adapter" }] } , Device { vendor_id : 32902u16 , id : 4282u16 , name : "80003ES2LAN Gigabit Ethernet Controller (Copper)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4283u16 , name : "80003ES2LAN Gigabit Ethernet Controller (Serdes)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4284u16 , name : "82571EB/82571GB Gigabit Ethernet Controller (Copper)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 872u16 , name : "4-Port 10/100/1000 Base-TX PCI Express Adapter for POWER" } , SubSystem { subvendor : 4156u16 , subdevice : 28747u16 , name : "NC364T PCI Express Quad Port Gigabit Server Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 4540u16 , name : "Quad Port Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4284u16 , name : "PRO/1000 PT Quad Port LP Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4540u16 , name : "PRO/1000 PT Quad Port LP Server Adapter (Kirkwood Low Profile)" }] } , Device { vendor_id : 32902u16 , id : 4285u16 , name : "82566DM-2 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 4287u16 , name : "82567LF Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4288u16 , name : "82562V-2 10/100 Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" }] } , Device { vendor_id : 32902u16 , id : 4290u16 , name : "82562G-2 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4291u16 , name : "82562GT-2 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4292u16 , name : "82562GT 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4293u16 , name : "82562G 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4294u16 , name : "82598EB 10-Gigabit AF Dual Port Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41055u16 , name : "10-Gigabit XF SR Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41311u16 , name : "10-Gigabit XF SR Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4295u16 , name : "82598EB 10-Gigabit AF Network Connection" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 895u16 , name : "10-Gigabit XF SR Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 896u16 , name : "10-Gigabit XF LR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41055u16 , name : "10-Gigabit XF SR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41311u16 , name : "10-Gigabit XF SR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41327u16 , name : "10-Gigabit XF SR Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4296u16 , name : "82598EB 10-Gigabit AT Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41228u16 , name : "10-Gigabit AT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41244u16 , name : "10-Gigabit AT Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41260u16 , name : "10-Gigabit AT Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4297u16 , name : "82576 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12783u16 , name : "NC362i Integrated Dual port Gigabit Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12863u16 , name : "NC362i Integrated Dual port Gigabit Server Adapter" } , SubSystem { subvendor : 4265u16 , subdevice : 32808u16 , name : "UV-BaseIO dual-port GbE" } , SubSystem { subvendor : 5027u16 , subdevice : 55u16 , name : "DS4100 Secure Multi-Gigabit Server Adapter with Compression" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" } , SubSystem { subvendor : 19538u16 , subdevice : 37394u16 , name : "LREC9212PT Dual-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 40988u16 , name : "Gigabit ET Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41020u16 , name : "Gigabit ET Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41036u16 , name : "Gigabit ET Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4298u16 , name : "82576 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4299u16 , name : "82567V Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4300u16 , name : "82567LM-2 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4301u16 , name : "82567LF-2 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4302u16 , name : "82567V-2 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4307u16 , name : "82574L Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6021u16 , name : "NC112i 1-port Ethernet Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 12880u16 , name : "NC112T PCI Express single Port Gigabit Server Adapter" } , SubSystem { subvendor : 4163u16 , subdevice : 33641u16 , name : "Motherboard" } , SubSystem { subvendor : 4243u16 , subdevice : 30441u16 , name : "PCIe-8233 Ethernet Adapter" } , SubSystem { subvendor : 4265u16 , subdevice : 32809u16 , name : "Prism XL Single Port Gigabit Ethernet" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1546u16 , name : "X7SPA-H/X7SPA-HF Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 19538u16 , subdevice : 8705u16 , name : "LRES2201PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Gigabit CT2 Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13688u16 , name : "Server Board S1200BTLR" } , SubSystem { subvendor : 32902u16 , subdevice : 13690u16 , name : "Server Board S1200BTS" } , SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "Gigabit CT Desktop Adapter" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" } , SubSystem { subvendor : 58559u16 , subdevice : 20674u16 , name : "PC2-LIMBO" }] } , Device { vendor_id : 32902u16 , id : 4308u16 , name : "Matrox Concord GE (customized Intel 82574)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4309u16 , name : "82571PT Gigabit PT Quad Port Server ExpressModule" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4310u16 , name : "82575GB Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4310u16 , name : "Gigabit VT Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 5210u16 , name : "Gigabit VT Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 5242u16 , name : "Gigabit VT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4312u16 , name : "82599EB 10 Gigabit Unprogrammed" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4313u16 , name : "82571EB Dual Port Gigabit Mezzanine Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5910u16 , name : "NC360m Dual Port 1GbE BL-c Adapter" }] } , Device { vendor_id : 32902u16 , id : 4314u16 , name : "82571EB Quad Port Gigabit Mezzanine Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5911u16 , name : "NC364m Quad Port 1GbE BL-c Adapter" }] } , Device { vendor_id : 32902u16 , id : 4315u16 , name : "82598EB 10-Gigabit Dual Port Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4317u16 , name : "82598EB 10-Gigabit AT CX4 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4318u16 , name : "82567LM-3 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4319u16 , name : "82567LF-3 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4321u16 , name : "82598EB 10-Gigabit AF Dual Port Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41311u16 , name : "10-Gigabit SR Dual Port Express Module" }] } , Device { vendor_id : 32902u16 , id : 4322u16 , name : "82575GB Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4322u16 , name : "Gigabit VT Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4325u16 , name : "82567LM-4 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4326u16 , name : "82576 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "Gigabit EF Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41007u16 , name : "Gigabit EF Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4327u16 , name : "82576 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12799u16 , name : "NC362i Integrated Dual Port BL-c Gigabit Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38657u16 , name : "LREC9701EF Single-port 1Gb Ethernet Network Adapter" }] } , Device { vendor_id : 32902u16 , id : 4328u16 , name : "82576 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41003u16 , name : "Gigabit ET Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41004u16 , name : "Gigabit ET Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4330u16 , name : "82577LM Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 4331u16 , name : "82577LC Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4332u16 , name : "82598EB 10-Gigabit AT CX4 Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "10-Gigabit CX4 Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41247u16 , name : "10-Gigabit CX4 Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4333u16 , name : "82599 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4335u16 , name : "82578DM Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" }] } , Device { vendor_id : 32902u16 , id : 4336u16 , name : "82578DC Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4337u16 , name : "82598EB 10-Gigabit AF Dual Port Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41487u16 , name : "10-Gigabit AF DA Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41503u16 , name : "10-Gigabit AF DA Dual Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4340u16 , name : "82598EB 10-Gigabit AF Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4207u16 , name : "10-Gigabit XF LR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41071u16 , name : "10-Gigabit XF LR Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 4341u16 , name : "82567LM Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8430u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 4342u16 , name : "82574L Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4343u16 , name : "10 Gigabit BR KX4 Dual Port Network Connection" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 31506u16 , name : "Sun Dual 10GbE PCIe 2.0 FEM" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet Mezzanine Adapter X520-KX4-2" }] } , Device { vendor_id : 32902u16 , id : 4344u16 , name : "82599 10 Gigabit Dual Port Backplane Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8035u16 , name : "10GbE 2P X520k bNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 6098u16 , name : "Ethernet 10Gb 2-port 560M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6352u16 , name : "Ethernet 10Gb 2-port 560FLB Adapter" } , SubSystem { subvendor : 4185u16 , subdevice : 273u16 , name : "T4007 10GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 304u16 , name : "T4009 10GbE interface" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet X520 10GbE Dual Port KX4-KR Mezz" }] } , Device { vendor_id : 32902u16 , id : 4345u16 , name : "82599 10 Gigabit Dual Port Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4347u16 , name : "82599ES 10-Gigabit SFI/SFP+ Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8050u16 , name : "Ethernet 10G 4P X520/I350 rNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 6096u16 , name : "Ethernet 10Gb 2-port 560FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6098u16 , name : "Ethernet 10Gb 2-port 560M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6099u16 , name : "Ethernet 10Gb 2-port 560SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8475u16 , name : "Ethernet 10Gb 1-port P560FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8519u16 , name : "Ethernet 10Gb 1-port 561i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8537u16 , name : "Ethernet 10Gb 2-port 562i Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 31505u16 , name : "Ethernet Server Adapter X520-2" } , SubSystem { subvendor : 4464u16 , subdevice : 76u16 , name : "82599 DP 10G Mezzanine Adapter" } , SubSystem { subvendor : 4980u16 , subdevice : 6664u16 , name : "PE310G4SPI9/PE310G4SPI9L/PE310G4SPI9LA Quad Port Fiber 10 Gigabit Ethernet PCI Express Server Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 1553u16 , name : "AOC-STGN-i2S" } , SubSystem { subvendor : 5940u16 , subdevice : 4521u16 , name : "10 Gigabit Dual Port Network Connection" } , SubSystem { subvendor : 6058u16 , subdevice : 4209u16 , name : "ThinkServer X520-2 AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 16391u16 , name : "82599ES 10-Gigabit SFI/SFP+ Network Connection" } , SubSystem { subvendor : 6058u16 , subdevice : 16427u16 , name : "82599ES 10Gb 2-port Server Adapter X520-DA2" } , SubSystem { subvendor : 6058u16 , subdevice : 16431u16 , name : "FPGA Card XC7VX690T-3FFG1157E" } , SubSystem { subvendor : 6356u16 , subdevice : 3081u16 , name : "82599ES 10Gb 2-port SFP+ OCP Mezz Card MOP81-I-10GS2" } , SubSystem { subvendor : 6461u16 , subdevice : 4100u16 , name : "560F-B" } , SubSystem { subvendor : 7124u16 , subdevice : 27u16 , name : "10G SFP+ DP ER102Fi4 Rack Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 47u16 , name : "10G SFP+ DP EP102Fi4A Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 50u16 , name : "10G SFP+ DP EP102Fi4 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 103u16 , name : "F102I82599" } , SubSystem { subvendor : 7999u16 , subdevice : 2560u16 , name : "Dual-port 10-Gigabit SFI/SFP+ Network Connection" } , SubSystem { subvendor : 19538u16 , subdevice : 4130u16 , name : "LRES1022PF Dual-port 10Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4132u16 , name : "LR-LINK LRES9804BF Quad-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12290u16 , name : "LRES3002PF Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 12306u16 , name : "LRES3012PF Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 38913u16 , name : "LREC9801BF Single-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38914u16 , name : "LREC9802BF Dual-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Adapter X520-DA2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Server Adapter X520-2" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Server Adapter X520-1" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet OCP Server Adapter X520-2" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Server Adapter X520-1" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet Server Adapter X520-2" } , SubSystem { subvendor : 32902u16 , subdevice : 4262u16 , name : "82599ES 10Gb 2 port Server Adapter X520-DA2" } , SubSystem { subvendor : 32902u16 , subdevice : 31249u16 , name : "Ethernet Server Adapter X520-2" } , SubSystem { subvendor : 32902u16 , subdevice : 31250u16 , name : "Ethernet Server Adapter X520-2" }] } , Device { vendor_id : 32902u16 , id : 4348u16 , name : "82599 10 Gigabit Dual Port Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4350u16 , name : "82552 10/100 Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4359u16 , name : "PRO/1000 MF Server Adapter (LX)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4400u16 , name : "82815 815 Chipset Host Bridge and Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4163u16 , subdevice : 32807u16 , name : "CUSL2-C/TUSL2-C Mainboard" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" }] } , Device { vendor_id : 32902u16 , id : 4401u16 , name : "82815 815 Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4402u16 , name : "82815 Chipset Graphics Controller (CGC)" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4156u16 , subdevice : 8193u16 , name : "e-pc 40" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "D815EEA Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" }] } , Device { vendor_id : 32902u16 , id : 4403u16 , name : "Thunderbolt 4 Bridge [Maple Ridge 2C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4404u16 , name : "Thunderbolt 4 NHI [Maple Ridge 2C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4405u16 , name : "Thunderbolt 4 USB Controller [Maple Ridge 2C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4406u16 , name : "Thunderbolt 4 Bridge [Maple Ridge 4C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4407u16 , name : "Thunderbolt 4 NHI [Maple Ridge 4C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4408u16 , name : "Thunderbolt 4 USB Controller [Maple Ridge 4C 2020]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4449u16 , name : "82806AA PCI64 Hub Advanced Programmable Interrupt Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4449u16 , name : "82806AA PCI64 Hub APIC" }] } , Device { vendor_id : 32902u16 , id : 4450u16 , name : "Xscale 80200 Big Endian Companion Chip" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4496u16 , name : "Merrifield SD/SDIO/eMMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4497u16 , name : "Merrifield Serial IO HSUART Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4498u16 , name : "Merrifield Serial IO HSUART DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4500u16 , name : "Merrifield Serial IO SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4501u16 , name : "Merrifield Serial IO I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4502u16 , name : "Merrifield Serial IO I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4505u16 , name : "Merrifield GPIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4510u16 , name : "Merrifield USB Device Controller (OTG)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4512u16 , name : "Merrifield SCU IPC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4513u16 , name : "Merrifield Power Management Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4514u16 , name : "Merrifield Serial IO DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4517u16 , name : "Merrifield Serial IO PWM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4547u16 , name : "Quark SoC X1000 PCIe Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4548u16 , name : "Quark SoC X1000 PCIe Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4575u16 , name : "Infrastructure Data Path Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4587u16 , name : "Simics NVMe Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4608u16 , name : "IXP1200 Network Processor" , subsystems : & [SubSystem { subvendor : 5930u16 , subdevice : 0u16 , name : "AEP SSL Accelerator" }] } , Device { vendor_id : 32902u16 , id : 4617u16 , name : "8255xER/82551IT Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 5131u16 , subdevice : 1552u16 , name : "PMC610 quad Ethernet board" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4208u16 , name : "PC6 mainboard" }] } , Device { vendor_id : 32902u16 , id : 4641u16 , name : "82092AA PCI to PCMCIA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4642u16 , name : "82092AA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4643u16 , name : "SAA7116" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4645u16 , name : "82452KX/GX [Orion]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4646u16 , name : "82596 PRO/10 PCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4647u16 , name : "82865 EtherExpress PRO/100A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4648u16 , name : "82556 EtherExpress PRO/100 Smart" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4649u16 , name : "82557/8/9/0/1 Ethernet Pro 100" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 12289u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12290u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12291u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12292u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12293u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12294u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 12295u16 , name : "82559 Fast Ethernet LOM with Alert on LAN*" } , SubSystem { subvendor : 3601u16 , subdevice : 45086u16 , name : "NC3120 Fast Ethernet NIC" } , SubSystem { subvendor : 3601u16 , subdevice : 45087u16 , name : "NC3122 Fast Ethernet NIC (dual port)" } , SubSystem { subvendor : 3601u16 , subdevice : 45103u16 , name : "NC1120 Ethernet NIC" } , SubSystem { subvendor : 3601u16 , subdevice : 45130u16 , name : "Netelligent 10/100TX NIC with Wake on LAN" } , SubSystem { subvendor : 3601u16 , subdevice : 45254u16 , name : "NC3161 Fast Ethernet NIC (embedded, WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 45255u16 , name : "NC3160 Fast Ethernet NIC (embedded)" } , SubSystem { subvendor : 3601u16 , subdevice : 45271u16 , name : "NC3121 Fast Ethernet NIC (WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 45277u16 , name : "NC3131 Fast Ethernet NIC (dual port)" } , SubSystem { subvendor : 3601u16 , subdevice : 45278u16 , name : "NC3132 Fast Ethernet Module (dual port)" } , SubSystem { subvendor : 3601u16 , subdevice : 45281u16 , name : "NC3133 Fast Ethernet Module (100-FX)" } , SubSystem { subvendor : 3601u16 , subdevice : 45364u16 , name : "NC3163 Fast Ethernet NIC (embedded, WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 45372u16 , name : "NC3162 Fast Ethernet NIC (embedded)" } , SubSystem { subvendor : 3601u16 , subdevice : 45380u16 , name : "NC3123 Fast Ethernet NIC (WOL)" } , SubSystem { subvendor : 3601u16 , subdevice : 45411u16 , name : "NC3134 Fast Ethernet NIC (dual port)" } , SubSystem { subvendor : 3601u16 , subdevice : 45412u16 , name : "NC3135 Fast Ethernet Upgrade Module (dual port)" } , SubSystem { subvendor : 3601u16 , subdevice : 45476u16 , name : "NC7131 Gigabit Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 92u16 , name : "82558B Ethernet Pro 10/100" } , SubSystem { subvendor : 4116u16 , subdevice : 444u16 , name : "82559 Fast Ethernet LAN On Motherboard" } , SubSystem { subvendor : 4116u16 , subdevice : 497u16 , name : "10/100 Ethernet Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 498u16 , name : "10/100 Ethernet Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 519u16 , name : "Ethernet Pro/100 S" } , SubSystem { subvendor : 4116u16 , subdevice : 562u16 , name : "10/100 Dual Port Server Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 570u16 , name : "ThinkPad R30" } , SubSystem { subvendor : 4116u16 , subdevice : 4188u16 , name : "Netfinity 10/100" } , SubSystem { subvendor : 4116u16 , subdevice : 8709u16 , name : "ThinkPad A22p" } , SubSystem { subvendor : 4116u16 , subdevice : 12380u16 , name : "10/100 EtherJet Management Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 16476u16 , name : "10/100 EtherJet Adapter with Alert on LAN" } , SubSystem { subvendor : 4116u16 , subdevice : 20572u16 , name : "10/100 EtherJet Secure Management Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 24668u16 , name : "10/100 EtherJet Secure Management Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 28764u16 , name : "10/100 Netfinity 10/100 Ethernet Security Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 32860u16 , name : "10/100 Netfinity 10/100 Ethernet Security Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 155u16 , name : "10/100 Ethernet Server Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 206u16 , name : "10/100 Ethernet Server Adapter" } , SubSystem { subvendor : 4147u16 , subdevice : 32768u16 , name : "PC-9821X-B06" } , SubSystem { subvendor : 4147u16 , subdevice : 32790u16 , name : "PK-UG-X006" } , SubSystem { subvendor : 4147u16 , subdevice : 32799u16 , name : "PK-UG-X006" } , SubSystem { subvendor : 4147u16 , subdevice : 32806u16 , name : "PK-UG-X006" } , SubSystem { subvendor : 4147u16 , subdevice : 32867u16 , name : "82559-based Fast Ethernet Adapter" } , SubSystem { subvendor : 4147u16 , subdevice : 32868u16 , name : "82559-based Fast Ethernet Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4288u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4291u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4298u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4299u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4323u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4324u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4156u16 , subdevice : 4608u16 , name : "NetServer 10/100TX" } , SubSystem { subvendor : 4238u16 , subdevice : 4303u16 , name : "EtherExpress PRO/100(B)" } , SubSystem { subvendor : 4291u16 , subdevice : 4352u16 , name : "SmartEther100 SC1100" } , SubSystem { subvendor : 4303u16 , subdevice : 4373u16 , name : "8255x-based Ethernet Adapter (10/100)" } , SubSystem { subvendor : 4303u16 , subdevice : 4419u16 , name : "8255x-based Ethernet Adapter (10/100)" } , SubSystem { subvendor : 4362u16 , subdevice : 139u16 , name : "82551QM Fast Ethernet Multifuction PCI/CardBus Controller" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard ethernet ETH2" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "8255x-based Ethernet Adapter (10/100)" } , SubSystem { subvendor : 4473u16 , subdevice : 2u16 , name : "PCI FastEther LAN on Docker" } , SubSystem { subvendor : 4473u16 , subdevice : 3u16 , name : "8255x-based Fast Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9568u16 , name : "AT-2560 100" } , SubSystem { subvendor : 4697u16 , subdevice : 9569u16 , name : "AT-2560 100 FX Ethernet Adapter" } , SubSystem { subvendor : 4710u16 , subdevice : 1u16 , name : "NE10/100 Adapter" } , SubSystem { subvendor : 5097u16 , subdevice : 4096u16 , name : "6221L-4U" } , SubSystem { subvendor : 5197u16 , subdevice : 9473u16 , name : "SEM-2000 MiniPCI LAN Adapter" } , SubSystem { subvendor : 5197u16 , subdevice : 9474u16 , name : "SEM-2100IL MiniPCI LAN Adapter" } , SubSystem { subvendor : 5736u16 , subdevice : 4352u16 , name : "EtherExpress PRO/100B (TX) (MiniPCI Ethernet+Modem)" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "EtherExpress PRO/100B (TX)" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "EtherExpress PRO/100B (T4)" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "EtherExpress PRO/10+" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "EtherExpress PRO/100 WfM" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "82557 10/100" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "82557 10/100 with Wake on LAN" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "82558 10/100 Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "82558 10/100 with Wake on LAN" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "82558B PRO/100+ PCI (TP)" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "EtherExpress PRO/100+ Management Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "EtherExpress PRO/100+" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "EtherExpress PRO/100+ Management Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "EtherExpress PRO/100+ Alert On LAN II* Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 14u16 , name : "EtherExpress PRO/100+ Management Adapter with Alert On LAN*" } , SubSystem { subvendor : 32902u16 , subdevice : 15u16 , name : "EtherExpress PRO/100 Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "EtherExpress PRO/100 S Management Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 17u16 , name : "EtherExpress PRO/100 S Management Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 18u16 , name : "EtherExpress PRO/100 S Advanced Management Adapter (D)" } , SubSystem { subvendor : 32902u16 , subdevice : 19u16 , name : "EtherExpress PRO/100 S Advanced Management Adapter (E)" } , SubSystem { subvendor : 32902u16 , subdevice : 48u16 , name : "EtherExpress PRO/100  Management Adapter with Alert On LAN* GC" } , SubSystem { subvendor : 32902u16 , subdevice : 49u16 , name : "EtherExpress PRO/100 Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 64u16 , name : "EtherExpress PRO/100 S Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 65u16 , name : "EtherExpress PRO/100 S Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 66u16 , name : "EtherExpress PRO/100 Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 80u16 , name : "EtherExpress PRO/100 S Desktop Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4105u16 , name : "EtherExpress PRO/100+ Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4108u16 , name : "EtherExpress PRO/100+ Server Adapter (PILA8470B)" } , SubSystem { subvendor : 32902u16 , subdevice : 4114u16 , name : "EtherExpress PRO/100 S Server Adapter (D)" } , SubSystem { subvendor : 32902u16 , subdevice : 4115u16 , name : "EtherExpress PRO/100 S Server Adapter (E)" } , SubSystem { subvendor : 32902u16 , subdevice : 4117u16 , name : "EtherExpress PRO/100 S Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4119u16 , name : "EtherExpress PRO/100+ Dual Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4144u16 , name : "EtherExpress PRO/100+ Management Adapter with Alert On LAN* G Server" } , SubSystem { subvendor : 32902u16 , subdevice : 4160u16 , name : "EtherExpress PRO/100 S Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4161u16 , name : "EtherExpress PRO/100 S Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4162u16 , name : "EtherExpress PRO/100 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4176u16 , name : "EtherExpress PRO/100 S Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4177u16 , name : "EtherExpress PRO/100 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4178u16 , name : "EtherExpress PRO/100 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4336u16 , name : "EtherExpress PRO/100+ Dual Port Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4649u16 , name : "82557/8/9 [Ethernet Pro 100]" } , SubSystem { subvendor : 32902u16 , subdevice : 8201u16 , name : "EtherExpress PRO/100 S Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "EtherExpress PRO/100 Cardbus" } , SubSystem { subvendor : 32902u16 , subdevice : 8206u16 , name : "EtherExpress PRO/100 LAN+V90 Cardbus Modem" } , SubSystem { subvendor : 32902u16 , subdevice : 8207u16 , name : "EtherExpress PRO/100 SR Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8208u16 , name : "EtherExpress PRO/100 S Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8211u16 , name : "EtherExpress PRO/100 SR Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8214u16 , name : "EtherExpress PRO/100 S Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8215u16 , name : "EtherExpress PRO/100 S Combo Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8216u16 , name : "EtherExpress PRO/100 SR Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8217u16 , name : "EtherExpress PRO/100 SR Combo Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8449u16 , name : "EtherExpress PRO/100 P Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8450u16 , name : "EtherExpress PRO/100 SP Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8451u16 , name : "EtherExpress PRO/100 SP Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8452u16 , name : "EtherExpress PRO/100 SP Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8453u16 , name : "EtherExpress PRO/100 SP Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8454u16 , name : "EtherExpress PRO/100 P Mobile Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8455u16 , name : "EtherExpress PRO/100 Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 8456u16 , name : "EtherExpress PRO/100 Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 8704u16 , name : "EtherExpress PRO/100 P Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8705u16 , name : "EtherExpress PRO/100 P Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8706u16 , name : "EtherExpress PRO/100 SP Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8707u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 8708u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 8709u16 , name : "EtherExpress PRO/100 SP Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8710u16 , name : "EtherExpress PRO/100 SP Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8711u16 , name : "EtherExpress PRO/100 SP Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8712u16 , name : "EtherExpress PRO/100 P Mobile Combo Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 9218u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9223u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9224u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9225u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9231u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9232u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9233u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9234u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 9235u16 , name : "EtherExpress PRO/100+ MiniPCI" } , SubSystem { subvendor : 32902u16 , subdevice : 12288u16 , name : "82559 Fast Ethernet LAN on Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 12289u16 , name : "82559 Fast Ethernet LOM with Basic Alert on LAN*" } , SubSystem { subvendor : 32902u16 , subdevice : 12290u16 , name : "82559 Fast Ethernet LOM with Alert on LAN II*" } , SubSystem { subvendor : 32902u16 , subdevice : 12294u16 , name : "EtherExpress PRO/100 S Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12295u16 , name : "EtherExpress PRO/100 S Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12296u16 , name : "EtherExpress PRO/100 Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12304u16 , name : "EtherExpress PRO/100 S Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12305u16 , name : "EtherExpress PRO/100 S Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12306u16 , name : "EtherExpress PRO/100 Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 12314u16 , name : "S845WD1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13329u16 , name : "SDS2 Mainboard" }] } , Device { vendor_id : 32902u16 , id : 4653u16 , name : "430FX - 82437FX TSC [Triton I]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4654u16 , name : "82371FB PIIX ISA [Triton I]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4656u16 , name : "82371FB PIIX IDE [Triton I]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4657u16 , name : "DSVD Modem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4660u16 , name : "430MX - 82371MX Mobile PCI I/O IDE Xcelerator (MPIIX)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4661u16 , name : "430MX - 82437MX Mob. System Ctrlr (MTSC) & 82438MX Data Path (MTDP)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4663u16 , name : "440FX - 82441FX PMC [Natoma]" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65534u16 , name : "Propolis Virtual 440FX" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "Qemu virtual machine" }] } , Device { vendor_id : 32902u16 , id : 4665u16 , name : "82371FB PIIX IDE Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4667u16 , name : "82380PB PCI to PCI Docking Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4668u16 , name : "82380AB (MISA) Mobile PCI-to-ISA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4669u16 , name : "683053 Programmable Interrupt Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4670u16 , name : "82466GX (IHPC) Integrated Hot-Plug Controller (hidden mode)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4671u16 , name : "82466GX Integrated Hot-Plug Controller (IHPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4672u16 , name : "82752 (752) AGP Graphics Accelerator" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4680u16 , name : "Ethernet Controller E835-CC for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4681u16 , name : "Ethernet Controller E835-CC for QSFP" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E835-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E835-C-Q2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E835-CC-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E835-CC-Q1 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter E835-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Network Adapter E835-C-Q2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet Network Adapter E835-CC-Q1 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet Network Adapter E835-CC-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Network Adapter E835-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Network Adapter E835-C-Q2" }] } , Device { vendor_id : 32902u16 , id : 4682u16 , name : "Ethernet Controller E835-CC for SFP" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E835-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E835-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E835-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E835-XXV-4 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet Network Adapter E835-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Network Adapter E835-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Network Adapter E835-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet Network Adapter E835-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet Network Adapter E835-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet Network Adapter E835-XXV-4 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 4683u16 , name : "82380FB (MPCI2) Mobile Docking Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4684u16 , name : "Ethernet Connection E823-L for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4685u16 , name : "Ethernet Connection E823-L for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4686u16 , name : "Ethernet Connection E823-L/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4687u16 , name : "Ethernet Connection E823-L 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4688u16 , name : "430HX - 82439HX TXC [Triton II]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4699u16 , name : "Ethernet Controller I226-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4700u16 , name : "Ethernet Controller I226-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4701u16 , name : "Ethernet Controller I226-IT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4705u16 , name : "Ethernet Controller E835-C for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4706u16 , name : "Ethernet Controller E835-C for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4707u16 , name : "Ethernet Controller E835-C for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4709u16 , name : "Ethernet Controller E835-L for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4710u16 , name : "Ethernet Controller E835-L for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4711u16 , name : "Ethernet Controller E835-L for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4748u16 , name : "60Cx8980 Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4749u16 , name : "60Cx8980 Series QAT – Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4817u16 , name : "Ethernet Controller E830-CC for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4818u16 , name : "Ethernet Controller E830-CC for QSFP" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E830-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E830-C-Q2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E830-CC-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E830-CC-Q1 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 4819u16 , name : "Ethernet Controller E830-CC for SFP" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E830-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E830-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E830-XXV-4 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter E830-XXV-8F for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Network Adapter E830-XXV-8F" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet Network Adapter E830-XXV-4F" }] } , Device { vendor_id : 32902u16 , id : 4820u16 , name : "Ethernet Controller E830-CC for SFP-DD" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4821u16 , name : "Ethernet Controller E830-C for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4824u16 , name : "Ethernet Controller E830-C for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4826u16 , name : "Ethernet Controller E830-C for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4828u16 , name : "Ethernet Controller E830-L for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4829u16 , name : "Ethernet Controller E830-L for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4830u16 , name : "Ethernet Controller E830-L for SFP" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E830-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E830-XXV-2" }] } , Device { vendor_id : 32902u16 , id : 4960u16 , name : "82806AA PCI64 Hub PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 4961u16 , name : "82806AA PCI64 Hub Controller (HRes)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4961u16 , name : "82806AA PCI64 Hub Controller (HRes)" } , SubSystem { subvendor : 32902u16 , subdevice : 32768u16 , name : "82806AA PCI64 Hub Controller (HRes)" }] } , Device { vendor_id : 32902u16 , id : 5202u16 , name : "Infrastructure Data Path Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5207u16 , name : "NVMe Data Path Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5212u16 , name : "Infrastructure Data Path Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5216u16 , name : "82870P2 P64H2 Hub PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5217u16 , name : "82870P2 P64H2 I/OxAPIC" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9/Vx9 mainboard" }] } , Device { vendor_id : 32902u16 , id : 5218u16 , name : "82870P2 P64H2 Hot Plug Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5377u16 , name : "82567V-3 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5378u16 , name : "82579LM Gigabit Network Connection (Lewisville)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 6058u16 , subdevice : 8654u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 13688u16 , name : "Server Board S1200BTLR" } , SubSystem { subvendor : 32902u16 , subdevice : 13690u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 5379u16 , name : "82579V Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33948u16 , name : "P8P67 Deluxe Motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5660u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" }] } , Device { vendor_id : 32902u16 , id : 5383u16 , name : "Ethernet Express Module X520-P2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5384u16 , name : "82598EB Gigabit BX Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5385u16 , name : "82580 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5386u16 , name : "82576NS Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5387u16 , name : "82598EB 10-Gigabit AT2 Server Adapter" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41228u16 , name : "82598EB 10-Gigabit AT2 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41244u16 , name : "82598EB 10-Gigabit AT2 Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41260u16 , name : "82598EB 10-Gigabit AT2 Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 5388u16 , name : "82583V Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5389u16 , name : "82576 Gigabit Backplane Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41228u16 , name : "Gigabit ET Quad Port Mezzanine Card" }] } , Device { vendor_id : 32902u16 , id : 5390u16 , name : "82580 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6016u16 , name : "NC365T 4-port Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 4769u16 , name : "Ethernet Server Adapter I340-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 4770u16 , name : "Ethernet Server Adapter I340-T4" }] } , Device { vendor_id : 32902u16 , id : 5391u16 , name : "82580 Gigabit Fiber Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5392u16 , name : "82580 Gigabit Backplane Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5393u16 , name : "82580 Gigabit SFP Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5395u16 , name : "CV82524 Thunderbolt Controller [Light Ridge 4C 2010]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5396u16 , name : "Ethernet X520 10GbE Dual Port KX4 Mezz" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet X520 10GbE Dual Port KX4 Mezz" }] } , Device { vendor_id : 32902u16 , id : 5397u16 , name : "X540 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5398u16 , name : "82580 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4785u16 , name : "Ethernet Server Adapter I340-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 4786u16 , name : "Ethernet Server Adapter I340-T2" }] } , Device { vendor_id : 32902u16 , id : 5399u16 , name : "82599ES 10 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 106u16 , name : "UCS CNA M61KR-I Intel Converged Network Adapter" }] } , Device { vendor_id : 32902u16 , id : 5400u16 , name : "82576NS SerDes Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5402u16 , name : "DSL2310 Thunderbolt Controller [Eagle Ridge 2C 2011]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5403u16 , name : "CVL2510 Thunderbolt Controller [Light Peak 2C 2010]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5404u16 , name : "82599 10 Gigabit TN Network Connection" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 31507u16 , name : "Dual 10GBASE-T LP" }] } , Device { vendor_id : 32902u16 , id : 5405u16 , name : "Ethernet Connection E823-L for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5408u16 , name : "I350 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5409u16 , name : "I350 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1538u16 , name : "Gigabit 2P I350-t LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 1683u16 , name : "Gigabit 2P I350-t LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 1762u16 , name : "Gigabit 2P I350-t LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 1879u16 , name : "Gigabit I350-t LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 1882u16 , name : "Gigabit I350-t LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 8032u16 , name : "Gigabit 4P I350-t rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8034u16 , name : "Gigabit 4P X540/I350 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8104u16 , name : "Ethernet 10G 4P X550/I350 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8105u16 , name : "Ethernet 10G 4P X550 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8106u16 , name : "Gigabit 4P X550/I350 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 65434u16 , name : "Gigabit 4P X710/I350 rNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 6097u16 , name : "Ethernet 1Gb 4-port 366FLR Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8195u16 , name : "Ethernet 1Gb 2-port 367i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8742u16 , name : "Ethernet 1Gb 1-port 364i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13183u16 , name : "Ethernet 1Gb 2-port 361i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13184u16 , name : "Ethernet 1Gb 4-port 366i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13214u16 , name : "Ethernet 1Gb 2-port 361T Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 33111u16 , name : "Ethernet 1Gb 4-port 366T Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 31510u16 , name : "Quad Port GbE PCIe 2.0 ExpressModule, UTP" } , SubSystem { subvendor : 4238u16 , subdevice : 31512u16 , name : "Quad Port GbE PCIe 2.0 Low Profile Adapter, UTP" } , SubSystem { subvendor : 4243u16 , subdevice : 30280u16 , name : "PCIe-8237R Ethernet Adapter" } , SubSystem { subvendor : 4243u16 , subdevice : 30281u16 , name : "PCIe-8236 Ethernet Adapter" } , SubSystem { subvendor : 4243u16 , subdevice : 30385u16 , name : "PCIe-8237R-S Ethernet Adapter" } , SubSystem { subvendor : 4243u16 , subdevice : 30555u16 , name : "PCIe-8237 Ethernet Adapter" } , SubSystem { subvendor : 4265u16 , subdevice : 32810u16 , name : "UV2-BaseIO dual-port GbE" } , SubSystem { subvendor : 4407u16 , subdevice : 574u16 , name : "1GigE I350 LOM" } , SubSystem { subvendor : 5593u16 , subdevice : 0u16 , name : "AOC-SGP-i4" } , SubSystem { subvendor : 5593u16 , subdevice : 1618u16 , name : "Dual Port i350 GbE MicroLP [AOC-CGP-i2]" } , SubSystem { subvendor : 5593u16 , subdevice : 5409u16 , name : "X10DRW-i" } , SubSystem { subvendor : 6058u16 , subdevice : 4212u16 , name : "ThinkServer I350-T4 AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 16389u16 , name : "I350 Gigabit Network Connection" } , SubSystem { subvendor : 6356u16 , subdevice : 3079u16 , name : "I350 1Gb 2-port RJ45 OCP Mezz Card MOP41-I-1GT2" } , SubSystem { subvendor : 6461u16 , subdevice : 4101u16 , name : "360T-B" } , SubSystem { subvendor : 6461u16 , subdevice : 4103u16 , name : "360T-L" } , SubSystem { subvendor : 6461u16 , subdevice : 4224u16 , name : "NIC-ETH360T-3S-4P" } , SubSystem { subvendor : 7124u16 , subdevice : 29u16 , name : "1G base-T QP EP014Ti1 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 53u16 , name : "1G base-T QP EP014Ti1 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 102u16 , name : "F014I350" } , SubSystem { subvendor : 7124u16 , subdevice : 138u16 , name : "F012I350" } , SubSystem { subvendor : 7124u16 , subdevice : 141u16 , name : "ENFI1100-T4" } , SubSystem { subvendor : 19538u16 , subdevice : 848u16 , name : "I350 1Gb 2-port Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4944u16 , name : "LREC9222HT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8195u16 , name : "LRES2003PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8197u16 , name : "LRES2005PT Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8198u16 , name : "LRES2006PT Six-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8200u16 , name : "LRES2008PT Eight-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8216u16 , name : "LRES2018PT Twelve-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8706u16 , name : "LRES2202PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8727u16 , name : "LRES2217PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12304u16 , name : "LRES3010PF Dual-port 1Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 12323u16 , name : "LRES3023PT Quad-port 1Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 12353u16 , name : "LRES3041PT Dual-port 1Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 16390u16 , name : "LRES4006MT Quad-port 1Gb Ethernet Netwaork Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 32769u16 , name : "LRES8001PT Quad-port 1Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38674u16 , name : "LREC9712HT Dual-port 1Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Server Adapter I350-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Adapter I350-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter I350-T4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 161u16 , name : "Ethernet Server Adapter I350-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 162u16 , name : "Ethernet Server Adapter I350-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 163u16 , name : "Ethernet Network Adapter I350-T4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 170u16 , name : "Ethernet Network Adapter I350-T4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16407u16 , name : "Ethernet Network Adapter I350-T4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 20481u16 , name : "Ethernet Server Adapter I350-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 20482u16 , name : "Ethernet Server Adapter I350-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 20483u16 , name : "Ethernet 1G 4P I350-t OCP" }] } , Device { vendor_id : 32902u16 , id : 5410u16 , name : "I350 Gigabit Fiber Network Connection" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 31511u16 , name : "Quad Port GbE PCIe 2.0 ExpressModule, MMF" } , SubSystem { subvendor : 4238u16 , subdevice : 31513u16 , name : "Dual Port GbE PCIe 2.0 Low Profile Adapter, MMF" } , SubSystem { subvendor : 19538u16 , subdevice : 4102u16 , name : "LRES1006PF Six-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8707u16 , name : "LRES2203PF Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 32816u16 , name : "LRES8030PF Dual-port 1Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38672u16 , name : "LREC9710HF Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38674u16 , name : "LREC9712HF Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38676u16 , name : "LREC9714HF Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Adapter I350-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Server Adapter I350-F4" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Server Adapter I350-F2" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Server Adapter I350-F1" } , SubSystem { subvendor : 32902u16 , subdevice : 162u16 , name : "Ethernet Server Adapter I350-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 163u16 , name : "Ethernet Server Adapter I350-F4" } , SubSystem { subvendor : 32902u16 , subdevice : 164u16 , name : "Ethernet Server Adapter I350-F2" }] } , Device { vendor_id : 32902u16 , id : 5411u16 , name : "I350 Gigabit Backplane Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 96u16 , name : "Gigabit 2P I350 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 8091u16 , name : "Gigabit 4P I350-t bNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 6020u16 , name : "Ethernet 1Gb 2-port 361FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6353u16 , name : "Ethernet 1Gb 2-port 361FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6537u16 , name : "Ethernet 1Gb 2-port 363i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13215u16 , name : "Ethernet 1Gb 4-port 366M Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38676u16 , name : "LREC9714HT Quad-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8018u16 , name : "1GbE 4P I350 Mezz" }] } , Device { vendor_id : 32902u16 , id : 5412u16 , name : "I350 Gigabit Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5413u16 , name : "82567V-4 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5414u16 , name : "82576 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 41052u16 , name : "Gigabit ET2 Quad Port Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41068u16 , name : "Gigabit ET2 Quad Port Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 5415u16 , name : "82580 Gigabit Fiber Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Server Adapter I340-F4" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Adapter I340-F4" }] } , Device { vendor_id : 32902u16 , id : 5416u16 , name : "Ethernet Controller 10-Gigabit X540-AT2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8033u16 , name : "Ethernet 10G 4P X540/I350 rNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 6445u16 , name : "561FLR-T 2-port 10Gb Ethernet Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8196u16 , name : "Ethernet 10Gb 2-port 561i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8474u16 , name : "Ethernet 10Gb 2-port 561T Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 18515u16 , name : "Ethernet Controller 10-Gigabit X540-AT2" } , SubSystem { subvendor : 4238u16 , subdevice : 31508u16 , name : "Sun Dual Port 10 GbE PCIe 2.0 ExpressModule, Base-T" } , SubSystem { subvendor : 4238u16 , subdevice : 31509u16 , name : "Sun Dual Port 10 GbE PCIe 2.0 Low Profile Adapter, Base-T" } , SubSystem { subvendor : 4407u16 , subdevice : 191u16 , name : "Ethernet Converged Network Adapter X540-T2" } , SubSystem { subvendor : 4464u16 , subdevice : 82u16 , name : "Ethernet Controller 10-Gigabit X540-AT2" } , SubSystem { subvendor : 5593u16 , subdevice : 1844u16 , name : "AOC-STG-I2T" } , SubSystem { subvendor : 5816u16 , subdevice : 28946u16 , name : "Twin 10G Thunderbolt 3 PCIe 10Gb Ethernet Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 4211u16 , name : "ThinkServer X540-T2 AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 16390u16 , name : "Ethernet Controller 10-Gigabit X540-AT2" } , SubSystem { subvendor : 7124u16 , subdevice : 26u16 , name : "10G base-T DP ER102Ti3 Rack Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 51u16 , name : "10G base-T DP EP102Ti3 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 52u16 , name : "10G base-T DP EP102Ti3A Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38914u16 , name : "LREC9802BT Dual-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter X540-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter X540-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 26u16 , name : "Ethernet Converged Network Adapter X540-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 162u16 , name : "Ethernet Converged Network Adapter X540-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 8033u16 , name : "Ethernet 10G 4P X540/I350 rNDC" } , SubSystem { subvendor : 32902u16 , subdevice : 20483u16 , name : "Ethernet 10G 2P X540-t Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 20484u16 , name : "Ethernet 10G 2P X540-t Adapter" }] } , Device { vendor_id : 32902u16 , id : 5417u16 , name : "82599 10 Gigabit Dual Port Network Connection with FCoE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5418u16 , name : "82599 10 Gigabit Dual Port Backplane Connection with FCoE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5422u16 , name : "82599 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5423u16 , name : "I350 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5424u16 , name : "X540 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5425u16 , name : "I210 Gigabit Unprogrammed" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5427u16 , name : "I210 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1880u16 , name : "I210 PCIe 1Gb 1-Port RJ45 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 1881u16 , name : "I210 PCIe 1Gb 1-Port RJ45 LOM" } , SubSystem { subvendor : 4136u16 , subdevice : 2869u16 , name : "I210 Gigabit Network Connection" } , SubSystem { subvendor : 4156u16 , subdevice : 3u16 , name : "Ethernet I210-T1 GbE NIC" } , SubSystem { subvendor : 4185u16 , subdevice : 384u16 , name : "RD10019 1GbE interface" } , SubSystem { subvendor : 4243u16 , subdevice : 30470u16 , name : "Compact Vision System Ethernet Adapter" } , SubSystem { subvendor : 4265u16 , subdevice : 32812u16 , name : "UV300 BaseIO single-port GbE" } , SubSystem { subvendor : 4265u16 , subdevice : 32813u16 , name : "UV3000 BaseIO GbE Network" } , SubSystem { subvendor : 6058u16 , subdevice : 4185u16 , name : "I210 Gigabit Network Connection" } , SubSystem { subvendor : 6058u16 , subdevice : 4352u16 , name : "ThinkServer Ethernet Server Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 5385u16 , name : "I210 Gigabit Network Connection" } , SubSystem { subvendor : 6058u16 , subdevice : 16461u16 , name : "I210 PCIe 1Gb 1-Port RJ45 LOM" } , SubSystem { subvendor : 6058u16 , subdevice : 16506u16 , name : "I210 PCIe 1Gb 1-Port RJ45 LOM" } , SubSystem { subvendor : 6838u16 , subdevice : 532u16 , name : "TS3 Plus Thunderbolt 3 Dock PCIe NIC" } , SubSystem { subvendor : 19538u16 , subdevice : 4177u16 , name : "LRES1051PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4624u16 , name : "LREC9204CT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8279u16 , name : "LRES2057PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8710u16 , name : "LRES2206PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8720u16 , name : "LRES2210PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8721u16 , name : "LRES2211PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8724u16 , name : "LRES2214PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12290u16 , name : "LRES3002PT Dual-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12292u16 , name : "LRES3004PT Quad-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Server Adapter I210-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Adapter I210-T1" }] } , Device { vendor_id : 32902u16 , id : 5430u16 , name : "I210 Gigabit Fiber Network Connection" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8708u16 , name : "LRES2204PT Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8722u16 , name : "LRES2212PF Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 8723u16 , name : "LRES2213PF Single-port 1Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 25136u16 , name : "LREC6230PF Single-port 1Gb Ethernet Network Adapter" }] } , Device { vendor_id : 32902u16 , id : 5431u16 , name : "I210 Gigabit Backplane Connection" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 272u16 , name : "T4005 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 273u16 , name : "T4007 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 288u16 , name : "T4008 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 304u16 , name : "T4009 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 320u16 , name : "T2035 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 336u16 , name : "RD-01068 1GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 368u16 , name : "RD-01213 10GbE interface" }] } , Device { vendor_id : 32902u16 , id : 5432u16 , name : "I210 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 400u16 , name : "RD20057 M.2 Dual 1GbE-T1 15m" }] } , Device { vendor_id : 32902u16 , id : 5433u16 , name : "I211 Gigabit Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5434u16 , name : "Ethernet Connection I217-LM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 5435u16 , name : "Ethernet Connection I217-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5447u16 , name : "DSL3510 Thunderbolt Controller [Cactus Ridge 4C 2012]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5448u16 , name : "DSL3310 Thunderbolt Controller [Cactus Ridge 2C 2012]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5449u16 , name : "DSL2210 Thunderbolt Controller [Port Ridge 1C 2011]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5450u16 , name : "Ethernet Server Adapter X520-4" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 282u16 , name : "Ethernet Converged Network Adapter X520-4" } , SubSystem { subvendor : 32902u16 , subdevice : 283u16 , name : "Ethernet Converged Network Adapter X520-4" } , SubSystem { subvendor : 32902u16 , subdevice : 284u16 , name : "Ethernet Converged Network Adapter X520-4" }] } , Device { vendor_id : 32902u16 , id : 5452u16 , name : "Ethernet Virtual Function 700 Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5453u16 , name : "Ethernet 10G 2P X520 Adapter" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 31505u16 , name : "10GbE 2P X520 Adapter" }] } , Device { vendor_id : 32902u16 , id : 5463u16 , name : "82599 10 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 16392u16 , name : "82599EN 10 Gigabit Network Connection" } , SubSystem { subvendor : 7124u16 , subdevice : 28u16 , name : "10G SFP+ SP ER101Fi4 Rack Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 48u16 , name : "10G SFP+ SP EP101Fi4A Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet OCP Server Adapter X520-1" }] } , Device { vendor_id : 32902u16 , id : 5464u16 , name : "Ethernet Converged Network Adapter X520-Q1" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 282u16 , name : "Ethernet Converged Network Adapter X520-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 283u16 , name : "Ethernet Converged Network Adapter X520-Q1" }] } , Device { vendor_id : 32902u16 , id : 5465u16 , name : "Ethernet Connection I218-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5466u16 , name : "Ethernet Connection I218-LM" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 5468u16 , name : "Ethernet Server Bypass Adapter" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Server Bypass Adapter X540-T2" }] } , Device { vendor_id : 32902u16 , id : 5469u16 , name : "Ethernet Server Bypass Adapter" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Server Bypass Adapter X520-SR2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Server Bypass Adapter X520-LR2" }] } , Device { vendor_id : 32902u16 , id : 5472u16 , name : "Ethernet Controller X540" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 38913u16 , name : "LREC9801BT Single-port 10Gb Ethernet Network Adapter" }] } , Device { vendor_id : 32902u16 , id : 5475u16 , name : "Ethernet Controller X550" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8104u16 , name : "Ethernet 10G 4P X550/I350 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8105u16 , name : "Ethernet 10G 4P X550 rNDC" } , SubSystem { subvendor : 4407u16 , subdevice : 690u16 , name : "X550-TX 10 Gig LOM" } , SubSystem { subvendor : 4407u16 , subdevice : 691u16 , name : "X550-TX 10 Gig LOM" } , SubSystem { subvendor : 4464u16 , subdevice : 1u16 , name : "Intel Ethernet Controller X550-T2 OCP card" } , SubSystem { subvendor : 5312u16 , subdevice : 4609u16 , name : "X550 10Gb 2P RJ45 OCP Mezz" } , SubSystem { subvendor : 5520u16 , subdevice : 209u16 , name : "Ethernet 10Gb 2-port 562T Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 210u16 , name : "Ethernet 10Gb 2-port 562FLR-T Adapter" } , SubSystem { subvendor : 5816u16 , subdevice : 29207u16 , name : "Twin10G Thunderbolt 3 Edition" } , SubSystem { subvendor : 6356u16 , subdevice : 3080u16 , name : "X550 10Gb 2-port RJ45 OCP Mezz Card MOP81-I-10GT2" } , SubSystem { subvendor : 6461u16 , subdevice : 4104u16 , name : "560T-B" } , SubSystem { subvendor : 6461u16 , subdevice : 4105u16 , name : "560T-L" } , SubSystem { subvendor : 6461u16 , subdevice : 4113u16 , name : "UN-NIC-ETH563T-sL-2P" } , SubSystem { subvendor : 19538u16 , subdevice : 4133u16 , name : "LRES1025PT Dual-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38930u16 , name : "LREC9812BT Dual-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter X550-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 26u16 , name : "Ethernet Converged Network Adapter X550-T2" } , SubSystem { subvendor : 32902u16 , subdevice : 27u16 , name : "Ethernet Server Adapter X550-T2 for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 29u16 , name : "Ethernet 10G 2P X550-t Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 34u16 , name : "Ethernet Converged Network Adapter X550-T2" }] } , Device { vendor_id : 32902u16 , id : 5476u16 , name : "X550 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5477u16 , name : "X550 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5478u16 , name : "DSL4410 Thunderbolt NHI [Redwood Ridge 2C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5479u16 , name : "DSL4410 Thunderbolt Bridge [Redwood Ridge 2C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5480u16 , name : "DSL4510 Thunderbolt NHI [Redwood Ridge 4C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5481u16 , name : "DSL4510 Thunderbolt Bridge [Redwood Ridge 4C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5482u16 , name : "DSL5320 Thunderbolt 2 NHI [Falcon Ridge 2C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5483u16 , name : "DSL5320 Thunderbolt 2 Bridge [Falcon Ridge 2C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5484u16 , name : "DSL5520 Thunderbolt 2 NHI [Falcon Ridge 4C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5485u16 , name : "DSL5520 Thunderbolt 2 Bridge [Falcon Ridge 4C 2013]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5487u16 , name : "Ethernet Connection I219-LM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" }] } , Device { vendor_id : 32902u16 , id : 5488u16 , name : "Ethernet Connection I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5489u16 , name : "Ethernet Virtual Function 700 Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5490u16 , name : "Ethernet Controller X710 for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 0u16 , name : "Ethernet 10G X710 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8089u16 , name : "Ethernet 10G 4P X710/I350 rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8092u16 , name : "Ethernet 10G 4P X710 SFP+ rNDC" } , SubSystem { subvendor : 4156u16 , subdevice : 0u16 , name : "Ethernet 10Gb 562SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8956u16 , name : "Ethernet 10Gb 2-port 562FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8957u16 , name : "Ethernet 10Gb 2-port 562SFP+ Adapter" } , SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "Ethernet Converged NIC X710-DA" } , SubSystem { subvendor : 4407u16 , subdevice : 315u16 , name : "Ethernet Converged NIC X710-DA4" } , SubSystem { subvendor : 4407u16 , subdevice : 522u16 , name : "Ethernet Converged NIC X710-DA2" } , SubSystem { subvendor : 5520u16 , subdevice : 0u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 5520u16 , subdevice : 549u16 , name : "Ethernet 10GbE 4P 563SFP+ Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 559u16 , name : "Ethernet 10Gb 2-port 564i Communication Board" } , SubSystem { subvendor : 6058u16 , subdevice : 0u16 , name : "ThinkServer X710 AnyFabric for 10GbE SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16385u16 , name : "ThinkServer X710-4 AnyFabric for 10GbE SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16386u16 , name : "ThinkServer X710-2 AnyFabric for 10GbE SFP+" } , SubSystem { subvendor : 6461u16 , subdevice : 4128u16 , name : "NIC-ETH561F-sL-4x10G" } , SubSystem { subvendor : 6461u16 , subdevice : 4129u16 , name : "NIC-ETH561F-sL-2x10G" } , SubSystem { subvendor : 6461u16 , subdevice : 4225u16 , name : "NIC-ETH561F-3S-2P" } , SubSystem { subvendor : 6629u16 , subdevice : 53532u16 , name : "Ethernet 2-port X710 10Gb SFP+ Adapter SP330" } , SubSystem { subvendor : 7124u16 , subdevice : 66u16 , name : "10G SFP+ DP EP102Fi4 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 86u16 , name : "Ethernet Network Adapter X710-BM2 for OCP NIC 3.0" } , SubSystem { subvendor : 7124u16 , subdevice : 101u16 , name : "F102IX710" } , SubSystem { subvendor : 7124u16 , subdevice : 116u16 , name : "Ethernet Network Adapter X710-BM2 for lldp" } , SubSystem { subvendor : 7124u16 , subdevice : 139u16 , name : "F102IX710" } , SubSystem { subvendor : 19538u16 , subdevice : 12291u16 , name : "LRES3003PF Quad-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 12295u16 , name : "LRES3007PF Quad-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 12345u16 , name : "LRES3039PF Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 32771u16 , name : "LRES8003PT Quad-port 10Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38916u16 , name : "LREC9804BF Quad-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38930u16 , name : "LREC9812BF Dual-port 10Gb Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter X710" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter X710-4" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter X710-4" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Converged Network Adapter X710-4" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Converged Network Adapter X710" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Converged Network Adapter X710" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet Converged Network Adapter X710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet Converged Network Adapter X710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet Server Adapter X710-DA2 for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 14u16 , name : "Ethernet Server Adapter OCP X710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 15u16 , name : "Ethernet Server Adapter OCP X710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Ethernet Converged Network Adapter X710" } , SubSystem { subvendor : 32902u16 , subdevice : 17u16 , name : "Ethernet Network Adapter X710-2 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 18u16 , name : "Ethernet Network Adapter X710-4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 19u16 , name : "Ethernet 10G 2P X710 OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 20u16 , name : "Ethernet 10G 4P X710 OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 21u16 , name : "Ethernet Server Adapter X710-DA2 for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 161u16 , name : "Ethernet Network Adapter X710-2 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 162u16 , name : "Ethernet Network Adapter X710-4 for OCP NIC 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16389u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 16390u16 , name : "Ethernet Controller X710 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 16391u16 , name : "Ethernet Controller X710 for 10GbE SFP+" }] } , Device { vendor_id : 32902u16 , id : 5492u16 , name : "Ethernet Controller XL710 Emulation" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5493u16 , name : "DSL6340 Thunderbolt 3 NHI [Alpine Ridge 2C 2015]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5494u16 , name : "DSL6340 Thunderbolt 3 Bridge [Alpine Ridge 2C 2015]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5495u16 , name : "DSL6540 Thunderbolt 3 NHI [Alpine Ridge 4C 2015]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5496u16 , name : "DSL6540 Thunderbolt 3 Bridge [Alpine Ridge 4C 2015]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5499u16 , name : "I210 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52240u16 , name : "RXi2-BP" }] } , Device { vendor_id : 32902u16 , id : 5500u16 , name : "I210 Gigabit Backplane Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5501u16 , name : "DSL5110 Thunderbolt 2 NHI (Low Power) [Win Ridge 2C 2014]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5502u16 , name : "DSL5110 Thunderbolt 2 Bridge (Low Power) [Win Ridge 2C 2014]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5504u16 , name : "Ethernet Controller XL710 for 40GbE backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5505u16 , name : "Ethernet Controller X710 for 10GbE backplane" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 0u16 , name : "Ethernet 10G X710-k bNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8088u16 , name : "Ethernet 10G 4P X710-k bNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8094u16 , name : "Ethernet 10G 2P X710-k bNDC" } , SubSystem { subvendor : 4185u16 , subdevice : 336u16 , name : "RD-01068 10GbE-KR interface" } , SubSystem { subvendor : 4185u16 , subdevice : 368u16 , name : "RD-01213 10GbE interface" } , SubSystem { subvendor : 5520u16 , subdevice : 0u16 , name : "Ethernet 2-port 563i Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 248u16 , name : "Ethernet 2-port 563i Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4110u16 , name : "NIC-ETH561i-Mb-4x10G" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter XL710-Q2" }] } , Device { vendor_id : 32902u16 , id : 5507u16 , name : "Ethernet Controller XL710 for 40GbE QSFP+" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 0u16 , name : "Ethernet 40G 2P XL710 QSFP+ rNDC" } , SubSystem { subvendor : 4136u16 , subdevice : 8095u16 , name : "Ethernet 40G 2P XL710 QSFP+ rNDC" } , SubSystem { subvendor : 4238u16 , subdevice : 0u16 , name : "10 Gb/40 Gb Ethernet Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 31515u16 , name : "10 Gb/40 Gb Ethernet Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 31517u16 , name : "10Gb/40Gb Ethernet Adapter" } , SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "Ethernet Converged NIC XL710-QDA2" } , SubSystem { subvendor : 4407u16 , subdevice : 316u16 , name : "Ethernet Converged NIC XL710-QDA2" } , SubSystem { subvendor : 19538u16 , subdevice : 12354u16 , name : "LRES3042PF Dual-port 40Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 39170u16 , name : "LREC9902BF Dual-port 40Gb Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter XL710-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter XL710-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter XL710-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet I/O Module XL710-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Server Adapter XL710-Q2OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Converged Network Adapter XL710-Q2" }] } , Device { vendor_id : 32902u16 , id : 5508u16 , name : "Ethernet Controller XL710 for 40GbE QSFP+" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 39169u16 , name : "LREC9901BF Single-port 40Gb Ethernet Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter XL710-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter XL710-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter XL710-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet I/O Module XL710-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Server Adapter XL710-Q1OCP" }] } , Device { vendor_id : 32902u16 , id : 5509u16 , name : "Ethernet Controller X710 for 10GbE QSFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5510u16 , name : "Ethernet Controller X710 for 10GBASE-T" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 0u16 , name : "Ethernet Controller X710 for 10GBASE-T" } , SubSystem { subvendor : 4238u16 , subdevice : 18519u16 , name : "Ethernet Controller X710 for 10GBASE-T" }] } , Device { vendor_id : 32902u16 , id : 5511u16 , name : "Ethernet Controller XL710 for 20GbE backplane" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 0u16 , name : "Ethernet 10/20Gb 2-port 660FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8958u16 , name : "Ethernet 10/20Gb 2-port 660FLB Adapter" }] } , Device { vendor_id : 32902u16 , id : 5512u16 , name : "Ethernet Controller XL710 for 20GbE backplane" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 0u16 , name : "Ethernet 10/20Gb 2-port 660M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8959u16 , name : "Ethernet 10/20Gb 2-port 660M Adapter" } , SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "Ethernet Network Adapter XXV710" } , SubSystem { subvendor : 4407u16 , subdevice : 692u16 , name : "Ethernet Network Adapter XXV710 OCP 2.0" }] } , Device { vendor_id : 32902u16 , id : 5513u16 , name : "Ethernet Controller X710/X557-AT 10GBASE-T" , subsystems : & [SubSystem { subvendor : 4238u16 , subdevice : 0u16 , name : "Quad Port 10GBase-T Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 31516u16 , name : "Quad Port 10GBase-T Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 38916u16 , name : "LREC9804BT Quad-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter X710-T" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Converged Network Adapter X710-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter X710-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Converged Network Adapter X710-T" } , SubSystem { subvendor : 32902u16 , subdevice : 160u16 , name : "Ethernet Converged Network Adapter X710-T4" } , SubSystem { subvendor : 32902u16 , subdevice : 4099u16 , name : "Ethernet Converged Network Adapter X710-T" }] } , Device { vendor_id : 32902u16 , id : 5514u16 , name : "Ethernet Controller XXV710 for 25GbE backplane" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 0u16 , name : "10/25Gb Ethernet Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 646u16 , name : "Synergy 4610C 10/25Gb Ethernet Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Controller XXV710 for 25GbE backplane" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet 25G 2P XXV710 Mezz" }] } , Device { vendor_id : 32902u16 , id : 5515u16 , name : "Ethernet Controller XXV710 for 25GbE SFP28" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "Ethernet Network Adapter XXV710" } , SubSystem { subvendor : 4407u16 , subdevice : 549u16 , name : "Ethernet Network Adapter XXV710" } , SubSystem { subvendor : 4407u16 , subdevice : 692u16 , name : "Ethernet Network Adapter XXV710 OCP 2.0" } , SubSystem { subvendor : 4980u16 , subdevice : 560u16 , name : "Single Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G1I71)" } , SubSystem { subvendor : 4980u16 , subdevice : 561u16 , name : "Single Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G1I71EU)" } , SubSystem { subvendor : 4980u16 , subdevice : 564u16 , name : "Dual Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G2I71)" } , SubSystem { subvendor : 4980u16 , subdevice : 565u16 , name : "Dual Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G2I71EU)" } , SubSystem { subvendor : 4980u16 , subdevice : 568u16 , name : "Quad Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G4I71L)" } , SubSystem { subvendor : 4980u16 , subdevice : 569u16 , name : "Quad Port 25 Gigabit Ethernet PCI Express Server Adapter (PE325G4I71LEU)" } , SubSystem { subvendor : 4980u16 , subdevice : 570u16 , name : "Quad Port 25 Gigabit Ethernet PCI Express Server Adapter (PE31625G4I71L)" } , SubSystem { subvendor : 4980u16 , subdevice : 571u16 , name : "Quad Port 25 Gigabit Ethernet PCI Express Server Adapter (PE31625G4I71LEU)" } , SubSystem { subvendor : 5520u16 , subdevice : 0u16 , name : "Ethernet Network Adapter XXV710-2" } , SubSystem { subvendor : 5520u16 , subdevice : 595u16 , name : "Ethernet 10/25/Gb 2-port 661SFP28 Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12311u16 , name : "LRES3017PF Dual-port 25Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 32772u16 , name : "LRES8004PT Quad-port 25Gb Ethernet Bypass Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Network Adapter XXV710" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter XXV710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter XXV710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter XXV710-1" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter XXV710-1" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter OCP XXV710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Network Adapter OCP XXV710-2" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet Network Adapter OCP XXV710-1" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet Network Adapter OCP XXV710-1" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet 25G 2P XXV710 Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet 25G 2P XXV710 OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet Network Adapter XXV710-DA2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet 25G 2P XXV710 OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 16385u16 , name : "Ethernet Network Adapter XXV710-2" }] } , Device { vendor_id : 32902u16 , id : 5520u16 , name : "Ethernet Connection E810-C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5521u16 , name : "Ethernet Controller E810-C for backplane" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 48334u16 , name : "Ethernet Controller E810-C for Intel(R) Open FPGA Stack" }] } , Device { vendor_id : 32902u16 , id : 5522u16 , name : "Ethernet Controller E810-C for QSFP" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 703u16 , name : "E810CQDA2 2x100 GbE QSFP28 PCIe NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4176u16 , name : "NIC-ETH1060F-LP-2P 2x100GbE Ethernet PCIe Card" } , SubSystem { subvendor : 19538u16 , subdevice : 4116u16 , name : "LRES1014PF Dual-port 100Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12326u16 , name : "LRES3026PF Dual-port 100Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E810-C-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E810-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E810-C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter E810-C-Q1 for OCP3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Network Adapter E810-C-Q2 for OCP3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Network Adapter E810-C-Q1" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Network Adapter E810-C-Q1 for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet 100G 2P E810-C Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet 100G 2P E810-C OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet Network Adapter E810-L-Q2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 14u16 , name : "Ethernet Network Adapter E810-2C-Q2" } , SubSystem { subvendor : 32902u16 , subdevice : 15u16 , name : "Ethernet Network Adapter E810-C-Q2T" } , SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Ethernet 100G 2P E810-C-stg Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 17u16 , name : "Ethernet Network Adapter E810-C-Q1 for OCP3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 18u16 , name : "Ethernet 100G 2P E810-C-st Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 19u16 , name : "Ethernet Network Adapter E810-C-Q1 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 20u16 , name : "Ethernet 100G 2P E810-2C Adapter" }] } , Device { vendor_id : 32902u16 , id : 5523u16 , name : "Ethernet Controller E810-C for SFP" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 707u16 , name : "E810XXVDA4 4x25/10 GbE SFP28 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 745u16 , name : "E810XXVDA4TG 4x25/10 GbE SFP28 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 746u16 , name : "E810XXVDA4T 4x25/10 GbE SFP28 PCIe NIC" } , SubSystem { subvendor : 19538u16 , subdevice : 4131u16 , name : "LRES1023PF Quad-port 25Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12327u16 , name : "LRES3027PF Quad-port 25Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 19538u16 , subdevice : 24704u16 , name : "LRES6080PF TimeSync 25G/10GbE Quad-Port Time Synchronization Ethernet Card" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E810-L-2" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter E810-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet Network Adapter E810-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet Network Adapter E810-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet Network Adapter E810-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Network Adapter E810-XXV-2 for OCP 2.0" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet 25G 4P E810-XXV Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet Network Adapter E810-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet Network Adapter E810-XXV-4 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet 25G 4P E810-XXV OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 14u16 , name : "Ethernet Network Adapter E810-XXV-4T" } , SubSystem { subvendor : 32902u16 , subdevice : 15u16 , name : "Ethernet 25G 4P E810-XXV-stg Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Ethernet 25G 4P E810-XXV-st Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 18u16 , name : "Ethernet 25G 4P E810-XXV Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 16400u16 , name : "Ethernet Network Adapter E810-XXV-4" } , SubSystem { subvendor : 32902u16 , subdevice : 16403u16 , name : "Ethernet Network Adapter E810-XXV-4 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16412u16 , name : "Ethernet Network Adapter E810-XXV-4 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 5524u16 , name : "Ethernet Controller E810-C/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5525u16 , name : "Ethernet Controller E810-C 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5528u16 , name : "Ethernet Connection E810-XXV" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5529u16 , name : "Ethernet Controller E810-XXV for backplane" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet 25G 2P E810-XXV-k Mezz" }] } , Device { vendor_id : 32902u16 , id : 5530u16 , name : "Ethernet Controller E810-XXV for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5531u16 , name : "Ethernet Controller E810-XXV for SFP" , subsystems : & [SubSystem { subvendor : 4407u16 , subdevice : 702u16 , name : "E810XXVDA2 2x25/10 GbE SFP28 PCIe NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4229u16 , name : "NIC-ETH660F-3S-2P" } , SubSystem { subvendor : 7124u16 , subdevice : 87u16 , name : "Ethernet Network Adapter E810-XXVAM2" } , SubSystem { subvendor : 7124u16 , subdevice : 88u16 , name : "Ethernet Network Adapter E810-XXVAM2 for OCP 3.0" } , SubSystem { subvendor : 7124u16 , subdevice : 110u16 , name : "Ethernet Network Adapter E810-XXVAM2 for BD" } , SubSystem { subvendor : 7124u16 , subdevice : 131u16 , name : "Ethernet Network Adapter E810-XXVAM2 for lldp" } , SubSystem { subvendor : 7124u16 , subdevice : 160u16 , name : "S252IE810" } , SubSystem { subvendor : 7916u16 , subdevice : 258u16 , name : "VSE-225-41E Dual-port 10Gb/25Gb Etherent PCIe" } , SubSystem { subvendor : 19538u16 , subdevice : 3u16 , name : "LRES1021PF Dual-port 25Gb Ethernet Server Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12329u16 , name : "LRES3029PF Dual-port 25Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet 25G 2P E810-XXV OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet 25G 2P E810-XXV Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E810-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet Network Adapter E810-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16385u16 , name : "Ethernet Network Adapter E810-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 16386u16 , name : "Ethernet Network Adapter E810-XXV-2 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16387u16 , name : "Ethernet Network Adapter E810-XXV-2" } , SubSystem { subvendor : 32902u16 , subdevice : 16405u16 , name : "Ethernet Network Adapter E810-XXV-2 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 5532u16 , name : "Ethernet Controller E810-XXV/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5533u16 , name : "Ethernet Controller E810-XXV 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5536u16 , name : "Ethernet Connection (2) I218-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5537u16 , name : "Ethernet Connection (2) I218-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5538u16 , name : "Ethernet Connection (3) I218-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5539u16 , name : "Ethernet Connection (3) I218-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5540u16 , name : "Ethernet Switch FM10000 Host Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5541u16 , name : "Ethernet Switch FM10000 Host Virtual Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5544u16 , name : "Ethernet Connection X552 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5545u16 , name : "X552 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5546u16 , name : "Ethernet Connection X552 10 GbE Backplane" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 288u16 , name : "T4008 10GbE interface" }] } , Device { vendor_id : 32902u16 , id : 5547u16 , name : "Ethernet Connection X552 10 GbE Backplane" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 336u16 , name : "RD-01068 10GbE interface" } , SubSystem { subvendor : 4185u16 , subdevice : 368u16 , name : "RD-01213 10GbE interface" }] } , Device { vendor_id : 32902u16 , id : 5548u16 , name : "Ethernet Connection X552 10 GbE SFP+" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 352u16 , name : "RD-01167 10GbE interface" }] } , Device { vendor_id : 32902u16 , id : 5549u16 , name : "Ethernet Connection X552/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5550u16 , name : "Ethernet Connection X552 1000BASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5552u16 , name : "Ethernet Connection X552 Backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5556u16 , name : "X553 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5557u16 , name : "DSL6340 USB 3.1 Controller [Alpine Ridge]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5558u16 , name : "DSL6540 USB 3.1 Controller [Alpine Ridge]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5559u16 , name : "Ethernet Connection (2) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5560u16 , name : "Ethernet Connection (2) I219-V" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" } , SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 5561u16 , name : "Ethernet Connection (3) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5563u16 , name : "Ethernet Connection (7) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5564u16 , name : "Ethernet Connection (7) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5565u16 , name : "Ethernet Connection (6) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5566u16 , name : "Ethernet Connection (6) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5567u16 , name : "JHL6240 Thunderbolt 3 NHI (Low Power) [Alpine Ridge LP 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5568u16 , name : "JHL6240 Thunderbolt 3 Bridge (Low Power) [Alpine Ridge LP 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5569u16 , name : "JHL6240 Thunderbolt 3 USB 3.1 Controller (Low Power) [Alpine Ridge LP 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5570u16 , name : "Ethernet Connection X553 Backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5571u16 , name : "Ethernet Connection X553 Backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5572u16 , name : "Ethernet Connection X553 10 GbE SFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5573u16 , name : "X553 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5574u16 , name : "Ethernet Connection X553 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5575u16 , name : "Ethernet Connection X553 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5576u16 , name : "Ethernet Connection X553/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5582u16 , name : "Ethernet Connection X553 10 GbE SFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5584u16 , name : "Ethernet SDI Adapter" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet SDI Adapter FM10420-100GbE-QDA2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet SDI Adapter FM10840-MTP2" }] } , Device { vendor_id : 32902u16 , id : 5585u16 , name : "Ethernet Controller 10G X550T" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 38929u16 , name : "LREC9811BT Single-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Converged Network Adapter X550-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 27u16 , name : "Ethernet Server Adapter X550-T1 for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 33u16 , name : "Ethernet Converged Network Adapter X550-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 162u16 , name : "Ethernet Converged Network Adapter X550-T1" }] } , Device { vendor_id : 32902u16 , id : 5586u16 , name : "JHL6540 Thunderbolt 3 NHI (C step) [Alpine Ridge 4C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5587u16 , name : "JHL6540 Thunderbolt 3 Bridge (C step) [Alpine Ridge 4C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5588u16 , name : "JHL6540 Thunderbolt 3 USB Controller (C step) [Alpine Ridge 4C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5589u16 , name : "Ethernet SDI Adapter FM10420-25GbE-DA2" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet SDI Adapter FM10420-25GbE-DA2" }] } , Device { vendor_id : 32902u16 , id : 5590u16 , name : "Ethernet Connection (5) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5591u16 , name : "Ethernet Connection (4) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5592u16 , name : "Ethernet Connection (4) I219-V" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 5593u16 , name : "JHL6340 Thunderbolt 3 NHI (C step) [Alpine Ridge 2C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5594u16 , name : "JHL6340 Thunderbolt 3 Bridge (C step) [Alpine Ridge 2C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5595u16 , name : "JHL6340 Thunderbolt 3 USB 3.1 Controller (C step) [Alpine Ridge 2C 2016]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5599u16 , name : "Ethernet Connection (8) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5600u16 , name : "Ethernet Connection (8) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5601u16 , name : "Ethernet Connection (9) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5602u16 , name : "Ethernet Connection (9) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5603u16 , name : "Ethernet Connection (5) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5604u16 , name : "Ethernet Connection X553 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5605u16 , name : "Ethernet Connection X553 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5607u16 , name : "JHL7540 Thunderbolt 3 Bridge [Titan Ridge 2C 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5608u16 , name : "JHL7540 Thunderbolt 3 NHI [Titan Ridge 2C 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5609u16 , name : "JHL7540 Thunderbolt 3 USB Controller [Titan Ridge 2C 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5610u16 , name : "JHL7540 Thunderbolt 3 Bridge [Titan Ridge 4C 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5611u16 , name : "JHL7540 Thunderbolt 3 NHI [Titan Ridge 4C 2018]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 5612u16 , name : "JHL7540 Thunderbolt 3 USB Controller [Titan Ridge 4C 2018]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 5615u16 , name : "JHL7440 Thunderbolt 3 Bridge [Titan Ridge DD 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5616u16 , name : "JHL7440 Thunderbolt 3 USB Controller [Titan Ridge DD 2018]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5618u16 , name : "Ethernet Controller I225-LM" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8241u16 , name : "LRES2031PT Single-port 2.5Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter I225-T1" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter I225-T1" }] } , Device { vendor_id : 32902u16 , id : 5619u16 , name : "Ethernet Controller I225-V" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8241u16 , name : "LRES2031PT Single-port 2.5Gb Ethernet Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Intel(R) Ethernet Controller (3) I225-V" }] } , Device { vendor_id : 32902u16 , id : 5620u16 , name : "Ethernet Connection (15) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5621u16 , name : "Ethernet Connection (15) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5622u16 , name : "I210 Gigabit Ethernet Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5625u16 , name : "500 Series Chipset Family GbE Controller (Corporate/vPro)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5626u16 , name : "500 Series Chipset Family GbE Controller (Consumer)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5627u16 , name : "Ethernet Connection (13) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5628u16 , name : "Ethernet Connection (13) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5631u16 , name : "Ethernet Controller X710 for 10GBASE-T" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 0u16 , name : "PCIe3 4-port 10GbE Base-T Adapter" } , SubSystem { subvendor : 4238u16 , subdevice : 0u16 , name : "Quad Port 10GBase-T Adapter - CP" } , SubSystem { subvendor : 4238u16 , subdevice : 31519u16 , name : "Quad Port 10GBase-T Adapter - CP" } , SubSystem { subvendor : 4407u16 , subdevice : 0u16 , name : "X710TLG GbE RJ45 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 705u16 , name : "X710T2LG 2x10 GbE RJ45 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 706u16 , name : "X710T4LG 4x10 GbE RJ45 PCIe NIC" } , SubSystem { subvendor : 4407u16 , subdevice : 729u16 , name : "Ethernet Network Adapter X710-T2L OCP 3.0" } , SubSystem { subvendor : 4407u16 , subdevice : 730u16 , name : "Ethernet Network Adapter X710-T4L OCP 3.0" } , SubSystem { subvendor : 5593u16 , subdevice : 7286u16 , name : "AOM-PTG-I2T-P" } , SubSystem { subvendor : 6461u16 , subdevice : 4226u16 , name : "NIC-ETH565T-3S-2P" } , SubSystem { subvendor : 19538u16 , subdevice : 4114u16 , name : "LRES1012PT Dual-port 10Gb Ethernet Network Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 12321u16 , name : "LRES3021PT Dual-port 10Gb Ethernet Server Adapter for OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Network Adapter X710-TL" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter X710-T4L" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter X710-T4L" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter X710-T2L" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter X710-T2L" } , SubSystem { subvendor : 32902u16 , subdevice : 5u16 , name : "Ethernet 10G 2P X710-T2L-t Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 6u16 , name : "Ethernet 10G 4P X710-T4L-t Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 7u16 , name : "Ethernet 10G 2P X710-T2L-t OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 8u16 , name : "Ethernet 10G 4P X710-T4L-t OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 9u16 , name : "Ethernet Network Adapter X710-T4L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 10u16 , name : "Ethernet Network Adapter X710-T4L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 11u16 , name : "Ethernet Network Adapter X710-T2L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 12u16 , name : "Ethernet Network Adapter X710-T2L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 13u16 , name : "Ethernet 10G 2P X710-T2L-t OCP" } , SubSystem { subvendor : 32902u16 , subdevice : 15u16 , name : "Ethernet Network Adapter X710-T2L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16393u16 , name : "Ethernet Network Adapter X710-T2L" } , SubSystem { subvendor : 32902u16 , subdevice : 16402u16 , name : "Ethernet Network Adapter X710-T4L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16408u16 , name : "Ethernet Network Adapter X710-T2L for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 16409u16 , name : "Ethernet Network Adapter X710-T4L" }] } , Device { vendor_id : 32902u16 , id : 5632u16 , name : "Broadwell-U Host Bridge -OPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5633u16 , name : "Broadwell-U PCI Express x16 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5634u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5635u16 , name : "Broadwell-U Processor Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5636u16 , name : "Broadwell-U Host Bridge -OPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5637u16 , name : "Broadwell-U PCI Express x8 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5638u16 , name : "Broadwell-U GT1 [HD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5639u16 , name : "Broadwell-U CHAPS Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5640u16 , name : "Broadwell-U Host Bridge -OPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5641u16 , name : "Broadwell-U x4 PCIe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5642u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5643u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5644u16 , name : "Broadwell-U Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5645u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5646u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5647u16 , name : "Broadwell-U SoftSKU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5648u16 , name : "Broadwell-U Host Bridge - DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5650u16 , name : "Broadwell-H GT2 [HD Graphics 5600]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5652u16 , name : "Broadwell-U Host Bridge - DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5654u16 , name : "Broadwell-U GT2 [HD Graphics 5500]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 8726u16 , name : "ZBook 15u G2 Mobile Workstation" }] } , Device { vendor_id : 32902u16 , id : 5656u16 , name : "Broadwell-U Host Bridge - DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5658u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5659u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5661u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5662u16 , name : "Broadwell-Y GT2 [HD Graphics 5300]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5666u16 , name : "Broadwell-DT/H GT3 [Iris Pro Graphics 6200]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5670u16 , name : "Broadwell-U GT3 [HD Graphics 6000]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5674u16 , name : "Broadwell-DT GT3 [Iris Pro Graphics P6300]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5675u16 , name : "Broadwell-U GT3 [Iris Graphics 6100]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5677u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5678u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5682u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5686u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5690u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5691u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5693u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 5694u16 , name : "Broadwell-U Integrated Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6281u16 , name : "Ethernet Adaptive Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6282u16 , name : "Ethernet Connection E823-C for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6283u16 , name : "Ethernet Connection E823-C for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6284u16 , name : "Ethernet Connection E823-C for SFP" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2749u16 , name : "Ethernet Connection 25G 4P E823-C LOM" } , SubSystem { subvendor : 6058u16 , subdevice : 16478u16 , name : "E823 25G/10G Ethernet LOM Controller" }] } , Device { vendor_id : 32902u16 , id : 6285u16 , name : "Ethernet Connection E823-C/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6286u16 , name : "Ethernet Connection E823-C 1GbE" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 16479u16 , name : "E823 1G Ethernet LOM Controller" }] } , Device { vendor_id : 32902u16 , id : 6288u16 , name : "Ethernet Connection E822-C for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6289u16 , name : "Ethernet Connection E822-C for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6290u16 , name : "Ethernet Connection E822-C for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6291u16 , name : "Ethernet Connection E822-C/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6292u16 , name : "Ethernet Connection E822-C 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6295u16 , name : "Ethernet Connection E822-L for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6296u16 , name : "Ethernet Connection E822-L for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6297u16 , name : "Ethernet Connection E822-L/X557-AT 10GBASE-T" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6298u16 , name : "Ethernet Connection E822-L 1GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6304u16 , name : "Atom Processor P5xxx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6305u16 , name : "C4XXX Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6323u16 , name : "Atom Processor P5xxx Series SATA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6352u16 , name : "Atom Processor P5xxx Series USB xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6355u16 , name : "Atom Processor P5xxx Series MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6358u16 , name : "Atom Processor P5xxx Series MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6367u16 , name : "Atom Processor P5xxx Series SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6368u16 , name : "Atom Processor P5xxx Series SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6369u16 , name : "Atom Processor P5xxx Series Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6382u16 , name : "200xx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6383u16 , name : "200xx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6387u16 , name : "Atom Processor P5xxx Series SATA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6400u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6401u16 , name : "6th-10th Gen Core Processor PCIe Controller (x16)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6402u16 , name : "Skylake-S GT1 [HD Graphics 510]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6403u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 6404u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 6405u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor PCIe Controller (x8)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6406u16 , name : "Skylake-U GT1 [HD Graphics 510]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 6408u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6409u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor PCIe Controller (x4)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6411u16 , name : "HD Graphics 510" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6412u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 6415u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6416u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 32902u16 , id : 6417u16 , name : "Xeon E3-1200 v5/v6 / E3-1500 v5 / 6th/7th/8th Gen Core Processor Gaussian Mixture Model" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" } , SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" } , SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 6418u16 , name : "Skylake-S GT2 [HD Graphics 530]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6422u16 , name : "Skylake-U GT2 [HD Graphics 520]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" }] } , Device { vendor_id : 32902u16 , id : 6424u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6425u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Imaging Unit" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 6427u16 , name : "Skylake-H GT2 [HD Graphics 530]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 32902u16 , id : 6429u16 , name : "Skylake-DT/H GT2 [HD Graphics P530]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6430u16 , name : "Skylake-Y GT2 [HD Graphics 515]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 6431u16 , name : "Xeon E3-1200 v5/E3-1500 v5/6th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6433u16 , name : "HD Graphics 520" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6435u16 , name : "HD Graphics 535" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6438u16 , name : "Skylake-U GT3 [Iris Graphics 540]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" }] } , Device { vendor_id : 32902u16 , id : 6439u16 , name : "Skylake-U GT3 [Iris Graphics 550]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6443u16 , name : "Iris Graphics 555" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6445u16 , name : "Skylake-H GT3 [Iris Graphics P555]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6450u16 , name : "Iris Pro Graphics 580" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6458u16 , name : "Skylake-H GT4 [Iris Pro Graphics P580]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6459u16 , name : "Skylake-H GT4 [Iris Pro Graphics 580]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6461u16 , name : "Skylake-H GT4 [Iris Pro Graphics P580]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6496u16 , name : "80960RP (i960RP) Microprocessor" , subsystems : & [SubSystem { subvendor : 4126u16 , subdevice : 1073u16 , name : "MegaRAID 431 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1080u16 , name : "MegaRAID 438 Ultra2 LVD RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1126u16 , name : "MegaRAID 466 Express Plus RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1127u16 , name : "MegaRAID 467 Enterprise 1500 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1168u16 , name : "MegaRAID 490 Express 300 RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1890u16 , name : "MegaRAID 762 Express RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 2464u16 , name : "PowerEdge Expandable RAID Controller 2/SC" } , SubSystem { subvendor : 4136u16 , subdevice : 1127u16 , name : "PowerEdge Expandable RAID Controller 2/DC" } , SubSystem { subvendor : 4136u16 , subdevice : 4369u16 , name : "PowerEdge Expandable RAID Controller 2/SC" } , SubSystem { subvendor : 4156u16 , subdevice : 930u16 , name : "MegaRAID" } , SubSystem { subvendor : 4156u16 , subdevice : 4294u16 , name : "MegaRAID 438, NetRAID-3Si" } , SubSystem { subvendor : 4156u16 , subdevice : 4295u16 , name : "MegaRAID T5, Integrated NetRAID" } , SubSystem { subvendor : 4156u16 , subdevice : 4300u16 , name : "MegaRAID, Integrated NetRAID" } , SubSystem { subvendor : 4156u16 , subdevice : 4301u16 , name : "NetRAID-1Si" } , SubSystem { subvendor : 4186u16 , subdevice : 0u16 , name : "SuperTrak" } , SubSystem { subvendor : 4186u16 , subdevice : 8552u16 , name : "SuperTrak Pro" } , SubSystem { subvendor : 4186u16 , subdevice : 20840u16 , name : "SuperTrak66/100" } , SubSystem { subvendor : 4369u16 , subdevice : 4369u16 , name : "MegaRAID 466, PowerEdge Expandable RAID Controller 2/SC" } , SubSystem { subvendor : 4369u16 , subdevice : 4370u16 , name : "PowerEdge Expandable RAID Controller 2/SC" } , SubSystem { subvendor : 4412u16 , subdevice : 930u16 , name : "MegaRAID" } , SubSystem { subvendor : 58559u16 , subdevice : 4112u16 , name : "CG1-RADIO" } , SubSystem { subvendor : 58559u16 , subdevice : 4128u16 , name : "CU2-QUARTET" } , SubSystem { subvendor : 58559u16 , subdevice : 4160u16 , name : "CU1-CHORUS" } , SubSystem { subvendor : 58559u16 , subdevice : 12544u16 , name : "CX1-BAND" }] } , Device { vendor_id : 32902u16 , id : 6498u16 , name : "80960RM (i960RM) Microprocessor" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 0u16 , name : "SuperTrak SX6000 I2O CPU" }] } , Device { vendor_id : 32902u16 , id : 6500u16 , name : "80960RN (i960RN) Microprocessor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6528u16 , name : "Atom Processor C3000 Series System Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6561u16 , name : "Atom Processor C3000 Series Error Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6562u16 , name : "Atom Processor C3000 Series Root Complex Event Collector" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6563u16 , name : "Atom Processor C3000 Series Integrated QAT Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6564u16 , name : "Atom Processor C3000 Series PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6565u16 , name : "Atom Processor C3000 Series PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6566u16 , name : "Atom Processor C3000 Series PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6567u16 , name : "Atom Processor C3000 Series PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6568u16 , name : "Atom Processor C3000 Series PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6569u16 , name : "Atom Processor C3000 Series PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6570u16 , name : "Atom Processor C3000 Series PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6571u16 , name : "Atom Processor C3000 Series PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6572u16 , name : "Atom Processor C3000 Series SMBus Contoller - Host" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6576u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6577u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6578u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6579u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6580u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6581u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6582u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6583u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6590u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6591u16 , name : "Atom Processor C3000 Series SATA Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6592u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6593u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6594u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6595u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6596u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6597u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6598u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6599u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6606u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6607u16 , name : "Atom Processor C3000 Series SATA Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6608u16 , name : "Atom Processor C3000 Series USB 3.0 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6609u16 , name : "Atom Processor C3000 Series Integrated LAN Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6610u16 , name : "Atom Processor C3000 Series Integrated LAN Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6611u16 , name : "Atom Processor C3000 Series ME HECI 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6612u16 , name : "Atom Processor C3000 Series ME HECI 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6613u16 , name : "Atom Processor C3000 Series ME KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6614u16 , name : "Atom Processor C3000 Series ME HECI 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6616u16 , name : "Atom Processor C3000 Series HSUART Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6619u16 , name : "Atom Processor C3000 Series SD Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6620u16 , name : "Atom Processor C3000 Series LPC or eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6621u16 , name : "Atom Processor C3000 Series Primary to Side Band (P2SB) Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6622u16 , name : "Atom Processor C3000 Series Power Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6623u16 , name : "Atom Processor C3000 Series SMBus controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6624u16 , name : "Atom Processor C3000 Series SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6626u16 , name : "Atom Processor C3000 Series QuickAssist Technology" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6627u16 , name : "Atom Processor C3000 Series QuickAssist Technology Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6674u16 , name : "Timna CPU Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6684u16 , name : "Ethernet Connection (17) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6685u16 , name : "Ethernet Connection (17) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6686u16 , name : "Ethernet Connection (16) I219-LM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 6687u16 , name : "Ethernet Connection (16) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6689u16 , name : "82840 840 [Carmel] Chipset Host Bridge (Hub A)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6691u16 , name : "82840 840 [Carmel] Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6692u16 , name : "82840 840 [Carmel] Chipset PCI Bridge (Hub B)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6704u16 , name : "82845 845 [Brookdale] Chipset Host Bridge" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 12928u16 , name : "Supermicro P4SBE Mainboard" }] } , Device { vendor_id : 32902u16 , id : 6705u16 , name : "82845 845 [Brookdale] Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 6712u16 , name : "5000 Series Chipset DMA Engine" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 6728u16 , name : "82597EX 10GbE Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "PRO/10GbE SR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41247u16 , name : "PRO/10GbE SR Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 6984u16 , name : "82597EX 10GbE Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 40991u16 , name : "PRO/10GbE LR Server Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 41247u16 , name : "PRO/10GbE LR Server Adapter" }] } , Device { vendor_id : 32902u16 , id : 7041u16 , name : "C740 Series (Emmitsburg) Chipsets eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7074u16 , name : "C740 Series (Emmitsburg) Chipsets SATA0 Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7078u16 , name : "C740 Series (Emmitsburg) Chipsets SATA0 Controller (RAID)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7085u16 , name : "C740 Series (Emmitsburg) Chipsets UART0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7086u16 , name : "C740 Series (Emmitsburg) Chipsets UART1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7088u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7089u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7090u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7091u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7092u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7093u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7096u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7097u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7098u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7099u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7100u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7101u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7102u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7103u16 , name : "C740 Series (Emmitsburg) Chipsets PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7111u16 , name : "C740 Series (Emmitsburg) Chipsets Power Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7112u16 , name : "C740 Series (Emmitsburg) Chipsets Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7113u16 , name : "C740 Series (Emmitsburg) Chipsets SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7117u16 , name : "C740 Series (Emmitsburg) Chipsets USB 3.2 Gen 1 xHCI Controller" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 165u16 , name : "RS0800I5H16i" }] } , Device { vendor_id : 32902u16 , id : 7122u16 , name : "Sapphire Rapids SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7126u16 , name : "C740 Series (Emmitsburg) Chipsets SATA2 Controller (RAID)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7154u16 , name : "Sapphire Rapids SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7158u16 , name : "C740 Series (Emmitsburg) Chipsets SATA1 Controller (RAID)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7168u16 , name : "6 Series/C200 Series Chipset Family Desktop SATA Controller (IDE mode, ports 0-3)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7169u16 , name : "6 Series/C200 Series Chipset Family Mobile SATA Controller (IDE mode, ports 0-3)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7170u16 , name : "6 Series/C200 Series Chipset Family 6 port Desktop SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BT Family" }] } , Device { vendor_id : 32902u16 , id : 7171u16 , name : "6 Series/C200 Series Chipset Family 6 port Mobile SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7172u16 , name : "6 Series/C200 Series Desktop SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12568u16 , name : "Smart Array B110i SATA RAID Controller" }] } , Device { vendor_id : 32902u16 , id : 7173u16 , name : "6 Series/C200 Series Mobile SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7174u16 , name : "Z68 Express Chipset SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7176u16 , name : "6 Series/C200 Series Chipset Family Desktop SATA Controller (IDE mode, ports 4-5)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7177u16 , name : "6 Series/C200 Series Chipset Family Mobile SATA Controller (IDE mode, ports 4-5)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7184u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BTS / Apple MacBook Pro 8,1/8,2" }] } , Device { vendor_id : 32902u16 , id : 7186u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7188u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7190u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 32902u16 , id : 7192u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 7194u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" }] } , Device { vendor_id : 32902u16 , id : 7196u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7198u16 , name : "6 Series/C200 Series Chipset Family PCI Express Root Port 8" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" }] } , Device { vendor_id : 32902u16 , id : 7200u16 , name : "6 Series/C200 Series Chipset Family High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1168u16 , name : "Alienware M17x R3" } , SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4156u16 , subdevice : 10943u16 , name : "HP Pavilion p6-2100 Desktop PC Series" } , SubSystem { subvendor : 4163u16 , subdevice : 33816u16 , name : "P8P67 Deluxe Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33819u16 , name : "P8H67 Series Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8200u16 , name : "DQ67SW board" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7202u16 , name : "6 Series/C200 Series Chipset Family SMBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BT Family / Apple MacBook Pro 8,1/8,2" }] } , Device { vendor_id : 32902u16 , id : 7204u16 , name : "6 Series/C200 Series Chipset Family Thermal Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7205u16 , name : "6 Series/C200 Series Chipset Family DMI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7206u16 , name : "6 Series/C200 Series Chipset Family USB Enhanced Host Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BT Family / Apple MacBook Pro 8,1/8,2" }] } , Device { vendor_id : 32902u16 , id : 7207u16 , name : "6 Series/C200 Series Chipset Family USB Universal Host Controller #1" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7212u16 , name : "6 Series/C200 Series Chipset Family USB Universal Host Controller #5" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7213u16 , name : "6 Series/C200 Series Chipset Family USB Enhanced Host Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BT Family / Apple MacBook Pro 8,1/8,2" }] } , Device { vendor_id : 32902u16 , id : 7219u16 , name : "6 Series/C200 Series Chipset Family LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7221u16 , name : "6 Series/C200 Series Chipset Family VECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7226u16 , name : "6 Series/C200 Series Chipset Family MEI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1171u16 , name : "Latitude E6420" } , SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7227u16 , name : "6 Series/C200 Series Chipset Family MEI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7228u16 , name : "6 Series/C200 Series Chipset Family IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7229u16 , name : "6 Series/C200 Series Chipset Family KT Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" }] } , Device { vendor_id : 32902u16 , id : 7232u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7233u16 , name : "Mobile SFF 6 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7234u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7235u16 , name : "Mobile 6 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7236u16 , name : "Z68 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7237u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7238u16 , name : "P67 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8P67 Deluxe Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7239u16 , name : "UM67 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7240u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7241u16 , name : "HM65 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookPro8,2 [Core i7, 15\", 2011]" }] } , Device { vendor_id : 32902u16 , id : 7242u16 , name : "H67 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1194u16 , name : "XPS 8300" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8H67 Series Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7243u16 , name : "HM67 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" }] } , Device { vendor_id : 32902u16 , id : 7244u16 , name : "Q65 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7245u16 , name : "QS67 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7246u16 , name : "Q67 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" }] } , Device { vendor_id : 32902u16 , id : 7247u16 , name : "QM67 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 32902u16 , id : 7248u16 , name : "B65 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7249u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7250u16 , name : "C202 Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 7251u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7252u16 , name : "C204 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7253u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7254u16 , name : "C206 Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8B WS Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7255u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7256u16 , name : "Upgraded B65 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7257u16 , name : "Upgraded HM67 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7258u16 , name : "Upgraded Q67 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7259u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7260u16 , name : "H61 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 8205u16 , name : "DH61CR motherboard" }] } , Device { vendor_id : 32902u16 , id : 7261u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7262u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7263u16 , name : "6 Series/C200 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7424u16 , name : "C600/X79 series chipset 4-Port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7426u16 , name : "C600/X79 series chipset 6-Port SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 7428u16 , name : "C600/X79 series chipset SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7430u16 , name : "C600/X79 series chipset SATA Premium RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7432u16 , name : "C600/X79 series chipset 2-Port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7440u16 , name : "C600/X79 series chipset PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 7441u16 , name : "C600/X79 series chipset PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7442u16 , name : "C600/X79 series chipset PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7443u16 , name : "C600/X79 series chipset PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7444u16 , name : "C600/X79 series chipset PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7445u16 , name : "C600/X79 series chipset PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7446u16 , name : "C600/X79 series chipset PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7447u16 , name : "C600/X79 series chipset PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7448u16 , name : "C600/X79 series chipset PCI Express Root Port 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 7449u16 , name : "C600/X79 series chipset PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7450u16 , name : "C600/X79 series chipset PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7451u16 , name : "C600/X79 series chipset PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7452u16 , name : "C600/X79 series chipset PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7453u16 , name : "C600/X79 series chipset PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7454u16 , name : "C600/X79 series chipset PCI Express Root Port 8" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 7455u16 , name : "C600/X79 series chipset PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7456u16 , name : "C600/X79 series chipset High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7458u16 , name : "C600/X79 series chipset SMBus Host Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7460u16 , name : "C600/X79 series chipset Thermal Management Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7461u16 , name : "C600/X79 series chipset DMI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7462u16 , name : "C600/X79 series chipset USB2 Enhanced Host Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7469u16 , name : "C600/X79 series chipset USB2 Enhanced Host Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" } , SubSystem { subvendor : 4156u16 , subdevice : 6313u16 , name : "HP DL360e G8" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7475u16 , name : "C600/X79 series chipset LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7477u16 , name : "C600/X79 series chipset VECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7482u16 , name : "C600/X79 series chipset MEI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7483u16 , name : "C600/X79 series chipset MEI Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7484u16 , name : "C600/X79 series chipset IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7485u16 , name : "C600/X79 series chipset KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7486u16 , name : "C600/X79 series chipset PCI Express Virtual Root Port" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" }] } , Device { vendor_id : 32902u16 , id : 7487u16 , name : "C608/C606/X79 series chipset PCI Express Virtual Switch Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7488u16 , name : "C600/X79 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7489u16 , name : "C600/X79 series chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "C602J on PowerEdge R320 server" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" }] } , Device { vendor_id : 32902u16 , id : 7504u16 , name : "C608 chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7508u16 , name : "C600/X79 series chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7509u16 , name : "C600/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7512u16 , name : "C606 chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7513u16 , name : "C604/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7514u16 , name : "C600/X79 series chipset Dual 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7515u16 , name : "C602 chipset 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7516u16 , name : "C600/X79 series chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7517u16 , name : "C600/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7518u16 , name : "C600/X79 series chipset Dual 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7519u16 , name : "C600/X79 series chipset 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7520u16 , name : "C608 chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7524u16 , name : "C600/X79 series chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7525u16 , name : "C600/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7528u16 , name : "C606 chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7529u16 , name : "C604/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7530u16 , name : "C600/X79 series chipset Dual 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7531u16 , name : "C602 chipset 4-Port SATA Storage Control Unit" , subsystems : & [SubSystem { subvendor : 1175u16 , subdevice : 4136u16 , name : "Dell Precision T3600" }] } , Device { vendor_id : 32902u16 , id : 7532u16 , name : "C600/X79 series chipset Dual 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7533u16 , name : "C600/X79 series chipset 4-Port SATA/SAS Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7534u16 , name : "C600/X79 series chipset Dual 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7535u16 , name : "C600/X79 series chipset 4-Port SATA Storage Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7536u16 , name : "C600/X79 series chipset SMBus Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7537u16 , name : "C608/C606/X79 series chipset SMBus Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7538u16 , name : "C608 chipset SMBus Controller 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7540u16 , name : "C608/C606/X79 series chipset PCI Express Upstream Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7542u16 , name : "C600/X79 series chipset Multi-Function Glue" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7680u16 , name : "7 Series/C210 Series Chipset Family 4-port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" }] } , Device { vendor_id : 32902u16 , id : 7681u16 , name : "7 Series Chipset Family 4-port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 7682u16 , name : "7 Series/C210 Series Chipset Family 6-port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7682u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7683u16 , name : "7 Series Chipset Family 6-port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4303u16 , subdevice : 5858u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 7684u16 , name : "7 Series/C210 Series Chipset Family SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7685u16 , name : "7 Series Chipset SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7686u16 , name : "7 Series/C210 Series Chipset Family SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7687u16 , name : "7 Series Chipset Family SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7688u16 , name : "7 Series/C210 Series Chipset Family 2-port SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7689u16 , name : "7 Series Chipset Family 2-port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 7694u16 , name : "7 Series/C210 Series Chipset Family SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7696u16 , name : "7 Series/C216 Chipset Family PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8H77-I Motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5865u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7696u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7698u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 2" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" }] } , Device { vendor_id : 32902u16 , id : 7700u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 3" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 5865u16 , name : "LIFEBOOK E752" }] } , Device { vendor_id : 32902u16 , id : 7702u16 , name : "7 Series/C216 Chipset Family PCI Express Root Port 4" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 6217u16 , subdevice : 5656u16 , name : "Z77 Extreme4 motherboard" }] } , Device { vendor_id : 32902u16 , id : 7704u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 5" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8H77-I Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7704u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7706u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 6" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7706u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7708u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7710u16 , name : "7 Series/C210 Series Chipset Family PCI Express Root Port 8" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 5865u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 6217u16 , subdevice : 7710u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7712u16 , name : "7 Series/C216 Chipset Family High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1355u16 , name : "XPS One 2710" } , SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33813u16 , name : "P8H77-I Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33861u16 , name : "P8Z77-V LX Motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5975u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4568u16 , name : "B75 [Ivy Bridge]\u{a0}chipset CX20642 audio controller on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 6296u16 , name : "Z77 Extreme4 motherboard" }] } , Device { vendor_id : 32902u16 , id : 7714u16 , name : "7 Series/C216 Chipset Family SMBus Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5862u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7714u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7716u16 , name : "7 Series/C210 Series Chipset Family Thermal Management Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" }] } , Device { vendor_id : 32902u16 , id : 7717u16 , name : "7 Series/C210 Series Chipset Family DMI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7718u16 , name : "7 Series/C216 Chipset Family USB Enhanced Host Controller #1" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5864u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset USB 2.0 controller on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7718u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7725u16 , name : "7 Series/C216 Chipset Family USB Enhanced Host Controller #2" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5864u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset USB 2.0 controller on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7725u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7729u16 , name : "7 Series/C210 Series Chipset Family USB xHCI Host Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6043u16 , name : "Elitebook 8470p" } , SubSystem { subvendor : 4156u16 , subdevice : 6059u16 , name : "ProBook 6570b" } , SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5870u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset USB 3.0 controller on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8691u16 , name : "ThinkPad T430" } , SubSystem { subvendor : 6217u16 , subdevice : 7729u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7731u16 , name : "7 Series/C210 Series Chipset Family LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7738u16 , name : "7 Series/C216 Chipset Family MEI Controller #1" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" } , SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4303u16 , subdevice : 5866u16 , name : "LIFEBOOK E752" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7738u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7739u16 , name : "7 Series/C210 Series Chipset Family MEI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7740u16 , name : "7 Series/C210 Series Chipset Family IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7741u16 , name : "7 Series/C210 Series Chipset Family KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7745u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7746u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7747u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7748u16 , name : "Z77 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 7748u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7749u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7750u16 , name : "Z75 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7751u16 , name : "Q77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7752u16 , name : "Q75 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7753u16 , name : "B75 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "Esprimo P510 D3171 motherboard" }] } , Device { vendor_id : 32902u16 , id : 7754u16 , name : "H77 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33994u16 , name : "P8H77-I Motherboard" }] } , Device { vendor_id : 32902u16 , id : 7755u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7756u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7757u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7758u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7759u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7760u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7761u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7762u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7763u16 , name : "C216 Series Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7764u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7765u16 , name : "QM77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7766u16 , name : "QS77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7767u16 , name : "HM77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7768u16 , name : "UM77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7769u16 , name : "HM76 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 5399u16 , name : "Zenbook Prime UX31A" } , SubSystem { subvendor : 4303u16 , subdevice : 5856u16 , name : "LIFEBOOK E752" }] } , Device { vendor_id : 32902u16 , id : 7770u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7771u16 , name : "UM77 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7772u16 , name : "7 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7773u16 , name : "HM75 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "NP300E5C series laptop" }] } , Device { vendor_id : 32902u16 , id : 7774u16 , name : "HM70 Express Chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202EV" }] } , Device { vendor_id : 32902u16 , id : 7775u16 , name : "NM70 Express Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7936u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7937u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7938u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7939u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7940u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7941u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7942u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7943u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7944u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7945u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7946u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7947u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7948u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7949u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7950u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7951u16 , name : "Atom processor C2000 SoC Transaction Router" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7952u16 , name : "Atom processor C2000 PCIe Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7953u16 , name : "Atom processor C2000 PCIe Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7954u16 , name : "Atom processor C2000 PCIe Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7955u16 , name : "Atom processor C2000 PCIe Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7956u16 , name : "Atom processor C2000 RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7957u16 , name : "Atom processor C2000 SMBus 2.0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7958u16 , name : "Atom processor C2000 RCEC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7960u16 , name : "Atom processor C2000 QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7961u16 , name : "Atom processor C2000 QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7968u16 , name : "Atom processor C2000 4-Port IDE SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7969u16 , name : "Atom processor C2000 4-Port IDE SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7970u16 , name : "Atom processor C2000 AHCI SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7971u16 , name : "Atom processor C2000 AHCI SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7972u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7973u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7974u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7975u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7980u16 , name : "Atom processor C2000 USB Enhanced Host Controller" , subsystems : & [SubSystem { subvendor : 512u16 , subdevice : 4136u16 , name : "Atom C2338 on Dell 0K8Y0N motherboard" }] } , Device { vendor_id : 32902u16 , id : 7982u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7983u16 , name : "Atom processor C2000 RAID SATA2 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7984u16 , name : "Atom processor C2000 2-Port IDE SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7985u16 , name : "Atom processor C2000 2-Port IDE SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7986u16 , name : "Atom processor C2000 AHCI SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7987u16 , name : "Atom processor C2000 AHCI SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7988u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7989u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7990u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7991u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7992u16 , name : "Atom processor C2000 PCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7993u16 , name : "Atom processor C2000 PCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7994u16 , name : "Atom processor C2000 PCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7995u16 , name : "Atom processor C2000 PCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7996u16 , name : "Atom processor C2000 PCU SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7997u16 , name : "Atom Processor C2000 PECI SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7998u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 7999u16 , name : "Atom processor C2000 RAID SATA3 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8000u16 , name : "Ethernet Connection I354 1.0 GbE Backplane" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1521u16 , name : "Ethernet Connection I354 1.0 GbE Backplane" }] } , Device { vendor_id : 32902u16 , id : 8001u16 , name : "Ethernet Connection I354" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8002u16 , name : "Atom processor C2000 GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8004u16 , name : "Atom processor C2000 GbE Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8005u16 , name : "Ethernet Connection I354 2.5 GbE Backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8212u16 , name : "Sky Lake-E Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8213u16 , name : "Sky Lake-E Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8214u16 , name : "Sky Lake-E Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8216u16 , name : "Sky Lake-E M2PCI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8218u16 , name : "Sky Lake-E Non-Transparent Bridge Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8220u16 , name : "Sky Lake-E Non-Transparent Bridge Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8221u16 , name : "Volume Management Device NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8224u16 , name : "Sky Lake-E DMI3 Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2327u16 , name : "X11DPi-N(T)" } , SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 8225u16 , name : "Sky Lake-E CBDMA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8228u16 , name : "Sky Lake-E MM/Vt-d Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8229u16 , name : "Sky Lake-E RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8230u16 , name : "Sky Lake-E IOAPIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8240u16 , name : "Sky Lake-E PCI Express Root Port A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8241u16 , name : "Sky Lake-E PCI Express Root Port B" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8242u16 , name : "Sky Lake-E PCI Express Root Port C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8243u16 , name : "Sky Lake-E PCI Express Root Port D" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8244u16 , name : "Sky Lake-E VT-d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8245u16 , name : "Sky Lake-E RAS Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8246u16 , name : "Sky Lake-E IOxAPIC Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8256u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8257u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8258u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8259u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8260u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8261u16 , name : "Sky Lake-E LM Channel 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8262u16 , name : "Sky Lake-E LMS Channel 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8263u16 , name : "Sky Lake-E LMDP Channel 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8264u16 , name : "Sky Lake-E DECS Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8265u16 , name : "Sky Lake-E LM Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8266u16 , name : "Sky Lake-E LMS Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8267u16 , name : "Sky Lake-E LMDP Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8268u16 , name : "Sky Lake-E M3KTI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8269u16 , name : "Sky Lake-E M3KTI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8270u16 , name : "Sky Lake-E M3KTI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8276u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8277u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8278u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8279u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8280u16 , name : "Sky Lake-E KTI 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8281u16 , name : "Sky Lake-E UPI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8294u16 , name : "Sky Lake-E Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8296u16 , name : "Sky Lake-E DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8297u16 , name : "Sky Lake-E DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8298u16 , name : "Sky Lake-E IOxAPIC Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8302u16 , name : "Sky Lake-E DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8303u16 , name : "Sky Lake-E DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8312u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8314u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8320u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8321u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8322u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8323u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8324u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8325u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8326u16 , name : "Sky Lake-E PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8328u16 , name : "Sky Lake-E DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8333u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8334u16 , name : "Sky Lake-E CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8768u16 , name : "Larrabee" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8769u16 , name : "Larrabee" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8777u16 , name : "Knights Ferry coprocessor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8778u16 , name : "Knights Ferry coprocessor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8784u16 , name : "Xeon Phi coprocessor 5100 series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8796u16 , name : "Xeon Phi coprocessor SE10/7120 series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8797u16 , name : "Xeon Phi coprocessor 3120 series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8798u16 , name : "Xeon Phi coprocessor 31S1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8802u16 , name : "Xeon Phi coprocessor 7220" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8832u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series SoC Transaction Register" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8836u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8838u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO1 DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8842u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO1 HSUART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8844u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO1 HSUART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8850u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8852u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series MMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8853u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8854u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series SD Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8856u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series Trusted Execution Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8860u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series PCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8867u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series SATA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8868u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8872u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series Low Power Engine Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8880u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8881u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8885u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series USB xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8888u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series Imaging Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8896u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8897u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8898u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8899u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8900u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8901u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8902u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8903u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series LPIO2 I2C Controller #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8904u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series PCI Express Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8906u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series PCI Express Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8908u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series PCI Express Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8910u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series PCI Express Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8924u16 , name : "Atom/Celeron/Pentium Processor x5-E8000/J3xxx/N3xxx Series Power Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8976u16 , name : "DH89xxCC LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 8995u16 , name : "DH89xxCC 4 Port SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9008u16 , name : "DH89xxCC SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9009u16 , name : "DH89xxCC Chap Counter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9010u16 , name : "DH89xxCC Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9012u16 , name : "DH89xxCC USB2 Enhanced Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9013u16 , name : "DH89xxCC USB2 Enhanced Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9026u16 , name : "DH89xxCC PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9027u16 , name : "DH89xxCC PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9028u16 , name : "DH89xxCC PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9029u16 , name : "DH89xxCC PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9030u16 , name : "DH89xxCC PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9031u16 , name : "DH89xxCC PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9032u16 , name : "DH89xxCC PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9033u16 , name : "DH89xxCC PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9056u16 , name : "DH89xxCC Watchdog Timer" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9060u16 , name : "DH89xxCC MEI 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9061u16 , name : "DH89xxCC MEI 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9104u16 , name : "DH895XCC Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9121u16 , name : "DH895XCC Series 2-Port SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9123u16 , name : "DH895XCC Series 4-Port SATA Controller [AHCI Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9126u16 , name : "DH895XCC Series 2-Port SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9136u16 , name : "DH895XCC Series SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9137u16 , name : "DH895XCC Series CHAP Counter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9138u16 , name : "DH895XCC Series Thermal Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9140u16 , name : "DH895XCC Series USB2 Enhanced Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9141u16 , name : "DH895XCC Series USB2 Enhanced Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9154u16 , name : "DH895XCC Series PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9155u16 , name : "DH895XCC Series PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9156u16 , name : "DH895XCC Series PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9157u16 , name : "DH895XCC Series PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9158u16 , name : "CDH895XCC Series PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9159u16 , name : "DH895XCC Series PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9160u16 , name : "DH895XCC Series PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9161u16 , name : "DH895XCC Series PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9184u16 , name : "DH895XCC Series Watchdog Timer" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9188u16 , name : "DH895XCC Series MEI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9189u16 , name : "DH895XCC Series MEI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9232u16 , name : "82801AA ISA Bridge (LPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9233u16 , name : "82801AA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9234u16 , name : "82801AA USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9235u16 , name : "82801AA SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9237u16 , name : "82801AA AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 149u16 , name : "Precision Workstation 220 Integrated Digital Audio" } , SubSystem { subvendor : 4136u16 , subdevice : 180u16 , name : "OptiPlex GX110" } , SubSystem { subvendor : 4362u16 , subdevice : 81u16 , name : "Activy 2xx" } , SubSystem { subvendor : 4564u16 , subdevice : 64u16 , name : "SoundMAX Integrated Digital Audio" } , SubSystem { subvendor : 4564u16 , subdevice : 72u16 , name : "SoundMAX Integrated Digital Audio" } , SubSystem { subvendor : 4564u16 , subdevice : 21312u16 , name : "SoundMAX Integrated Digital Audio" } , SubSystem { subvendor : 5940u16 , subdevice : 4133u16 , name : "Activy 3xx" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 9238u16 , name : "82801AA AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9240u16 , name : "82801AA PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9248u16 , name : "82801AB ISA Bridge (LPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9249u16 , name : "82801AB IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9250u16 , name : "82801AB USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9251u16 , name : "82801AB SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9253u16 , name : "82801AB AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4564u16 , subdevice : 64u16 , name : "SoundMAX Integrated Digital Audio" } , SubSystem { subvendor : 4564u16 , subdevice : 72u16 , name : "SoundMAX Integrated Digital Audio" }] } , Device { vendor_id : 32902u16 , id : 9254u16 , name : "82801AB AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9256u16 , name : "82801AB PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9280u16 , name : "82801BA ISA Bridge (LPC)" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E" }] } , Device { vendor_id : 32902u16 , id : 9282u16 , name : "82801BA/BAM UHCI USB 1.1 Controller #1" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 454u16 , name : "Netvista A40/A40p" } , SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4136u16 , subdevice : 199u16 , name : "Dimension 8100" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 4156u16 , subdevice : 4719u16 , name : "e-pc 40" } , SubSystem { subvendor : 4163u16 , subdevice : 32807u16 , name : "CUSL2-C/TUSL2-C Mainboard" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 19780u16 , name : "D850EMV2 motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 9283u16 , name : "82801BA/BAM SMBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 454u16 , name : "Netvista A40/A40p" } , SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4136u16 , subdevice : 199u16 , name : "Dimension 8100" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 4156u16 , subdevice : 4719u16 , name : "e-pc 40" } , SubSystem { subvendor : 4163u16 , subdevice : 32807u16 , name : "CUSL2-C/TUSL2-C Mainboard" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" } , SubSystem { subvendor : 5593u16 , subdevice : 12928u16 , name : "Supermicro P4SBE Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 9284u16 , name : "82801BA/BAM UHCI USB 1.1 Controller #2" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4136u16 , subdevice : 199u16 , name : "Dimension 8100" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 4156u16 , subdevice : 4719u16 , name : "e-pc 40" } , SubSystem { subvendor : 4163u16 , subdevice : 32807u16 , name : "CUSL2-C/TUSL2-C Mainboard" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 9285u16 , name : "82801BA/BAM AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 11u16 , name : "Compaq Deskpro EN Audio" } , SubSystem { subvendor : 3601u16 , subdevice : 136u16 , name : "Evo D500" } , SubSystem { subvendor : 4116u16 , subdevice : 454u16 , name : "Netvista A40/A40p" } , SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4156u16 , subdevice : 4719u16 , name : "e-pc 40" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 4564u16 , subdevice : 21312u16 , name : "AD1885 AC97 codec" } , SubSystem { subvendor : 5218u16 , subdevice : 13168u16 , name : "STAC9721 AC" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 18006u16 , name : "Desktop Board D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 19780u16 , name : "D850EMV2 motherboard" }] } , Device { vendor_id : 32902u16 , id : 9286u16 , name : "82801BA/BAM AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" }] } , Device { vendor_id : 32902u16 , id : 9288u16 , name : "82801 Mobile PCI Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240 Mobile Workstation" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30 notebook" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 6058u16 , subdevice : 8211u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 6058u16 , subdevice : 8366u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6058u16 , subdevice : 8436u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 9289u16 , name : "82801BA/BAM/CA/CAM Ethernet Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 18u16 , name : "EtherExpress PRO/100 VM" } , SubSystem { subvendor : 3601u16 , subdevice : 145u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 462u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 476u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 491u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 492u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 514u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 517u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 535u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 564u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 573u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 580u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 581u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4116u16 , subdevice : 613u16 , name : "PRO/100 VE Desktop Connection" } , SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "PRO/100 VE Desktop Connection" } , SubSystem { subvendor : 4116u16 , subdevice : 618u16 , name : "PRO/100 VE Desktop Connection" } , SubSystem { subvendor : 4255u16 , subdevice : 12637u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4255u16 , subdevice : 12673u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 4473u16 , subdevice : 65281u16 , name : "PRO/100 VE Network Connection" } , SubSystem { subvendor : 4486u16 , subdevice : 30721u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 5197u16 , subdevice : 9730u16 , name : "HomePNA 1M CNR" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 12304u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 32902u16 , subdevice : 12305u16 , name : "EtherExpress PRO/100 VM" } , SubSystem { subvendor : 32902u16 , subdevice : 12306u16 , name : "82562EH based Phoneline" } , SubSystem { subvendor : 32902u16 , subdevice : 12307u16 , name : "EtherExpress PRO/100 VE" } , SubSystem { subvendor : 32902u16 , subdevice : 12308u16 , name : "EtherExpress PRO/100 VM" } , SubSystem { subvendor : 32902u16 , subdevice : 12309u16 , name : "82562EH based Phoneline" } , SubSystem { subvendor : 32902u16 , subdevice : 12310u16 , name : "EtherExpress PRO/100 P Mobile Combo" } , SubSystem { subvendor : 32902u16 , subdevice : 12311u16 , name : "EtherExpress PRO/100 P Mobile" } , SubSystem { subvendor : 32902u16 , subdevice : 12312u16 , name : "EtherExpress PRO/100" }] } , Device { vendor_id : 32902u16 , id : 9290u16 , name : "82801BAM IDE U100 Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612TX" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" }] } , Device { vendor_id : 32902u16 , id : 9291u16 , name : "82801BA IDE U100 Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 454u16 , name : "Netvista A40/A40p" } , SubSystem { subvendor : 4136u16 , subdevice : 199u16 , name : "Dimension 8100" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 4156u16 , subdevice : 4719u16 , name : "e-pc 40" } , SubSystem { subvendor : 4163u16 , subdevice : 32807u16 , name : "CUSL2-C/TUSL2-C Mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1285u16 , name : "BL7 motherboard" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" } , SubSystem { subvendor : 5593u16 , subdevice : 12928u16 , name : "Supermicro P4SBE Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17714u16 , name : "Desktop Board D815EEA2/D815EFV" } , SubSystem { subvendor : 32902u16 , subdevice : 17751u16 , name : "D815EGEW Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 19780u16 , name : "D850EMV2 motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 32902u16 , id : 9292u16 , name : "82801BAM ISA Bridge (LPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9294u16 , name : "82801 PCI Bridge" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "PowerEdge R320 server" } , SubSystem { subvendor : 4156u16 , subdevice : 4874u16 , name : "Z600 Workstation" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33869u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 34100u16 , name : "ASUS B85-PLUS" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4566u16 , name : "B75 [Ivy Bridge]\u{a0}chipset on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 12400u16 , name : "ThinkCentre M91p" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Server Board S1200BTS" }] } , Device { vendor_id : 32902u16 , id : 9296u16 , name : "82801E ISA Bridge (LPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9298u16 , name : "82801E USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9299u16 , name : "82801E SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9305u16 , name : "82801E Ethernet Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9307u16 , name : "82801E IDE U100 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9309u16 , name : "82801E Ethernet Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9310u16 , name : "82801E PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9344u16 , name : "82801CA LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9346u16 , name : "82801CA/CAM USB Controller #1" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 48u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 544u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 32902u16 , subdevice : 6488u16 , name : "vpr Matrix 170B4" } , SubSystem { subvendor : 32902u16 , subdevice : 13348u16 , name : "SE7501HG2 Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude C640" }] } , Device { vendor_id : 32902u16 , id : 9347u16 , name : "82801CA/CAM SMBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 544u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 32902u16 , subdevice : 6488u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 32902u16 , id : 9348u16 , name : "82801CA/CAM USB Controller #2" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 48u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 544u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 32902u16 , subdevice : 6488u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 32902u16 , id : 9349u16 , name : "82801CA/CAM AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4115u16 , subdevice : 22873u16 , name : "Crystal WMD Audio Codec" } , SubSystem { subvendor : 4116u16 , subdevice : 546u16 , name : "ThinkPad A30/A30p/T23" } , SubSystem { subvendor : 4116u16 , subdevice : 1288u16 , name : "ThinkPad T30" } , SubSystem { subvendor : 4116u16 , subdevice : 1308u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4163u16 , subdevice : 5507u16 , name : "L3C (SPDIF)" } , SubSystem { subvendor : 4163u16 , subdevice : 5667u16 , name : "L2B (no SPDIF)" } , SubSystem { subvendor : 4163u16 , subdevice : 5699u16 , name : "L3F" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 5197u16 , subdevice : 49158u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 32902u16 , id : 9350u16 , name : "82801CA/CAM AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 547u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4116u16 , subdevice : 1283u16 , name : "ThinkPad R31" } , SubSystem { subvendor : 4116u16 , subdevice : 1306u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4127u16 , subdevice : 4133u16 , name : "620 Series" } , SubSystem { subvendor : 4163u16 , subdevice : 5270u16 , name : "PCtel HSP56 MR" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 4941u16 , subdevice : 19489u16 , name : "Dell Inspiron 2100 internal modem" } , SubSystem { subvendor : 5197u16 , subdevice : 8469u16 , name : "vpr Matrix 170B4 internal modem" } , SubSystem { subvendor : 5361u16 , subdevice : 21537u16 , name : "MD56ORD V.92 MDC Modem" }] } , Device { vendor_id : 32902u16 , id : 9351u16 , name : "82801CA/CAM USB Controller #3" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 48u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 544u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 32902u16 , subdevice : 6488u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 32902u16 , id : 9354u16 , name : "82801CAM IDE U100 Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 48u16 , name : "Evo N600c" } , SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4116u16 , subdevice : 544u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 32902u16 , subdevice : 6488u16 , name : "vpr Matrix 170B4" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude C640" }] } , Device { vendor_id : 32902u16 , id : 9355u16 , name : "82801CA Ultra ATA Storage Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" }] } , Device { vendor_id : 32902u16 , id : 9356u16 , name : "82801CAM ISA Bridge (LPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9408u16 , name : "82801DB/DBL (ICH4/ICH4-L) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" }] } , Device { vendor_id : 32902u16 , id : 9409u16 , name : "82801DBL (ICH4-L) IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9410u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1325u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 32905u16 , name : "P4B533" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard USB 1.x" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5385u16 , subdevice : 10640u16 , name : "Averatec 5110H laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 9410u16 , name : "Latitude X300" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude D400/D500" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9411u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1325u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard SMbus" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5208u16 , subdevice : 9410u16 , name : "GA-8PE667 Ultra" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9412u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1325u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 32905u16 , name : "P4B533" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5385u16 , subdevice : 10640u16 , name : "Averatec 5110H" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 9410u16 , name : "Latitude X300" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude D400/D500" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9413u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 184u16 , name : "Analog Devices Inc. codec [SoundMAX]" } , SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1335u16 , name : "ThinkPad T4x Series" } , SubSystem { subvendor : 4116u16 , subdevice : 1375u16 , name : "Thinkpad R50e model 1634" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m [SigmaTel STAC9750,51]" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 5907u16 , name : "M2400N/M6800N laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 32944u16 , name : "P4B533" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4473u16 , subdevice : 513u16 , name : "Toshiba Tecra M1" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-8PE667 Ultra" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5940u16 , subdevice : 4101u16 , name : "D1451 (SCENIC N300, i845GV) Sigmatel STAC9750T" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" }] } , Device { vendor_id : 32902u16 , id : 9414u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1316u16 , name : "ThinkPad T4x Series" } , SubSystem { subvendor : 4116u16 , subdevice : 1317u16 , name : "ThinkPad" } , SubSystem { subvendor : 4116u16 , subdevice : 1369u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4133u16 , subdevice : 60u16 , name : "Aspire 2001WLCi (Compal CL50 motherboard) implementation" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 5910u16 , name : "M2400N laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 6182u16 , name : "M6800N" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4941u16 , subdevice : 19489u16 , name : "Latitude D500" } , SubSystem { subvendor : 5197u16 , subdevice : 8469u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5361u16 , subdevice : 21538u16 , name : "D480 MDC V.9x Modem" }] } , Device { vendor_id : 32902u16 , id : 9415u16 , name : "82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1325u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 32905u16 , name : "P4B533" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5385u16 , subdevice : 10640u16 , name : "Averatec 5110H" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 9410u16 , name : "Latitude X300" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude D400/D500" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9418u16 , name : "82801DBM (ICH4-M) IDE Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1325u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 32902u16 , subdevice : 17729u16 , name : "Latitude D400/D500" }] } , Device { vendor_id : 32902u16 , id : 9419u16 , name : "82801DB (ICH4) IDE Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4163u16 , subdevice : 32905u16 , name : "P4B533" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard IDE" } , SubSystem { subvendor : 5208u16 , subdevice : 9410u16 , name : "GA-8PE667 Ultra" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9420u16 , name : "82801DBM (ICH4-M) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30 notebook" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" }] } , Device { vendor_id : 32902u16 , id : 9421u16 , name : "82801DB/DBM (ICH4/ICH4-M) USB2 EHCI Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4116u16 , subdevice : 1326u16 , name : "ThinkPad" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: Intel 82801DBM (ICH4-M)" } , SubSystem { subvendor : 4136u16 , subdevice : 285u16 , name : "Latitude D600" } , SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 4163u16 , subdevice : 32905u16 , name : "P4B533" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 onboard USB 2.0" } , SubSystem { subvendor : 4473u16 , subdevice : 65280u16 , name : "Satellite 2430" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 14721u16 , name : "845PE Max (MS-6580)" } , SubSystem { subvendor : 5385u16 , subdevice : 6504u16 , name : "Averatec 5110H" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 9410u16 , name : "Latitude X300" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 9424u16 , name : "82801EB/ER (ICH5/ICH5R) LPC Interface Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9425u16 , name : "82801EB (ICH5) SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9425u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 20992u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9426u16 , name : "82801EB/ER (ICH5/ICH5R) USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 387u16 , name : "PowerEdge 1800" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000/8KNXP motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX series onboard UHCI" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9427u16 , name : "82801EB/ER (ICH5/ICH5R) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 342u16 , name : "Precision 360" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d330 uT" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX S2 series SMBus" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9428u16 , name : "82801EB/ER (ICH5/ICH5R) USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 387u16 , name : "PowerEdge 1800" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX S2 series onboard UHCI" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9429u16 , name : "82801EB/ER (ICH5/ICH5R) AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4106u16 , subdevice : 5243u16 , name : "Abit IS7-E motherboard" } , SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d330 uT" } , SubSystem { subvendor : 4163u16 , subdevice : 33011u16 , name : "P4P800 series motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33039u16 , name : "P5P800-MX Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-8IPE1000/8KNXP motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 128u16 , name : "865PE Neo2-V (MS-6788) Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 32902u16 , subdevice : 40960u16 , name : "D865PERL mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 57344u16 , name : "D865PERL mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 57345u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 57346u16 , name : "SoundMax Integrated Digital Audio" }] } , Device { vendor_id : 32902u16 , id : 9430u16 , name : "82801EB/ER (ICH5/ICH5R) AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" }] } , Device { vendor_id : 32902u16 , id : 9431u16 , name : "82801EB/ER (ICH5/ICH5R) USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 387u16 , name : "PowerEdge 1800" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX S2 series onboard UHCI" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9435u16 , name : "82801EB/ER (ICH5/ICH5R) IDE Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30080u16 , name : "MSI 875P" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX S2 series onboard IDE" } , SubSystem { subvendor : 32902u16 , subdevice : 9435u16 , name : "P4C800 Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9436u16 , name : "82801EB (ICH5) LPC Interface Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9437u16 , name : "82801EB/ER (ICH5/ICH5R) USB2 EHCI Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 387u16 , name : "PowerEdge 1800" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d530 CMT (DG746A)" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20486u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9438u16 , name : "82801EB/ER (ICH5/ICH5R) USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4116u16 , subdevice : 749u16 , name : "xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9426u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" } , SubSystem { subvendor : 5218u16 , subdevice : 29312u16 , name : "865PE Neo2 (MS-6728)" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5477u16 , subdevice : 12545u16 , name : "P4TSV Motherboard (865G)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4124u16 , name : "PRIMERGY RX/TX S2 series onboard UHCI" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" } , SubSystem { subvendor : 32902u16 , subdevice : 21068u16 , name : "D865PERL mainboard" }] } , Device { vendor_id : 32902u16 , id : 9439u16 , name : "82801ER (ICH5R) SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" }] } , Device { vendor_id : 32902u16 , id : 9456u16 , name : "Omni-Path HFI Silicon 100 Series [discrete]" , subsystems : & [SubSystem { subvendor : 4265u16 , subdevice : 32814u16 , name : "Omni-path HFI 100 Series, 1-port A-board" } , SubSystem { subvendor : 4265u16 , subdevice : 32815u16 , name : "Omni-path HFI 100 Series, 2-port A-board" } , SubSystem { subvendor : 4265u16 , subdevice : 32816u16 , name : "Omni-path HFI 100 Series, 1-port B-board" } , SubSystem { subvendor : 4265u16 , subdevice : 32817u16 , name : "Omni-path HFI 100 Series, 2-port B-board" } , SubSystem { subvendor : 5520u16 , subdevice : 231u16 , name : "100Gb 1-port OP101 QSFP28 x8 PCIe Gen3 with Intel Omni-Path Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 232u16 , name : "100Gb 1-port OP101 QSFP28 x16 PCIe Gen3 with Intel Omni-Path Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 540u16 , name : "Apollo 100Gb 1-port Intel Omni-Path Architecture 860z Mezzanine FIO Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 2356u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16, SIOM Module" } , SubSystem { subvendor : 5593u16 , subdevice : 2459u16 , name : "Omni-path HFI Mezz AOC, 1 Port, PCIe x16." } , SubSystem { subvendor : 7352u16 , subdevice : 1u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16, TC4600 QSFP28" } , SubSystem { subvendor : 7352u16 , subdevice : 2u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16, TC6600 Fixed Port" } , SubSystem { subvendor : 7352u16 , subdevice : 3u16 , name : "Omni-Path HFI Adapter 100 Series, 2 Port, 2 PCIe x16, Earth Simulation QSFP28" } , SubSystem { subvendor : 7352u16 , subdevice : 4u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16, TC4600E QSFP28" } , SubSystem { subvendor : 17230u16 , subdevice : 1u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, OCP 3.0" } , SubSystem { subvendor : 17230u16 , subdevice : 9768u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16" } , SubSystem { subvendor : 32902u16 , subdevice : 9768u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16" } , SubSystem { subvendor : 32902u16 , subdevice : 9769u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x8" } , SubSystem { subvendor : 32902u16 , subdevice : 9770u16 , name : "Omni-Path HFI Adapter 100 Series, 2 Ports, Split PCIe x16" } , SubSystem { subvendor : 32902u16 , subdevice : 9773u16 , name : "Omni-Path HFI Adapter 100 Series, 1 Port, PCIe x16, IO Module AHWKPTP100HF" }] } , Device { vendor_id : 32902u16 , id : 9457u16 , name : "Omni-Path HFI Silicon 100 Series [integrated]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9459u16 , name : "Wireless 8260" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Dual Band Wireless-AC 8260 [Snow Field Peak AC]" } , SubSystem { subvendor : 32902u16 , subdevice : 4112u16 , name : "Dual Band Wireless-AC 8260" }] } , Device { vendor_id : 32902u16 , id : 9460u16 , name : "Wireless 8260" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 48u16 , name : "Dual Band Wireless-AC 8260 [Snow Field Peak AC]" }] } , Device { vendor_id : 32902u16 , id : 9467u16 , name : "Dual Band Wireless-AC 3168NGW [Stone Peak]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9469u16 , name : "Wireless 8265 / 8275" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16u16 , name : "Dual Band Wireless-AC 8265 [Windstorm Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 336u16 , name : "Dual Band Wireless-AC 8265" } , SubSystem { subvendor : 32902u16 , subdevice : 4112u16 , name : "Dual Band Wireless-AC 8265" } , SubSystem { subvendor : 32902u16 , subdevice : 4400u16 , name : "Dual Band Wireless-AC 8265" }] } , Device { vendor_id : 32902u16 , id : 9472u16 , name : "82820 820 (Camino) Chipset Host Bridge (MCH)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 149u16 , name : "Precision Workstation 220 Chipset" } , SubSystem { subvendor : 4163u16 , subdevice : 32796u16 , name : "P3C-2000 system chipset" }] } , Device { vendor_id : 32902u16 , id : 9473u16 , name : "82820 820 (Camino) Chipset Host Bridge (MCH)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32796u16 , name : "P3C-2000 system chipset" }] } , Device { vendor_id : 32902u16 , id : 9483u16 , name : "82820 820 (Camino) Chipset Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9487u16 , name : "82820 820 (Camino) Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9504u16 , name : "82805AA MTH Memory Translator Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9505u16 , name : "82804AA MRH-S Memory Repeater Hub for SDRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9506u16 , name : "NVMe Optane Memory Series" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 14342u16 , name : "Optane Memory 16GB" } , SubSystem { subvendor : 32902u16 , subdevice : 14352u16 , name : "Optane Memory M10 16GB" }] } , Device { vendor_id : 32902u16 , id : 9509u16 , name : "Optane NVME SSD P1600X Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9510u16 , name : "Wi-Fi 5(802.11ac) Wireless-AC 9x6x [Thunder Peak]" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9260 160MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 528u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9260 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 532u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9260 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 560u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9560 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 564u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9560 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 568u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9560 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 572u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9560 80MHz 2x2" } , SubSystem { subvendor : 32902u16 , subdevice : 608u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9461 80MHz 1x1" } , SubSystem { subvendor : 32902u16 , subdevice : 612u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9461 80MHz 1x1" } , SubSystem { subvendor : 32902u16 , subdevice : 672u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9462 80MHz 1x1" } , SubSystem { subvendor : 32902u16 , subdevice : 676u16 , name : "Dual Band Wi-Fi 5 Wireless-AC 9462 80MHz 1x1" }] } , Device { vendor_id : 32902u16 , id : 9520u16 , name : "82850 850 (Tehama) Chipset Host Bridge (MCH)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 199u16 , name : "Dimension 8100" } , SubSystem { subvendor : 5243u16 , subdevice : 1287u16 , name : "TH7II-RAID" }] } , Device { vendor_id : 32902u16 , id : 9521u16 , name : "82860 860 (Wombat) Chipset Host Bridge (MCH)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" }] } , Device { vendor_id : 32902u16 , id : 9522u16 , name : "82850 850 (Tehama) Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9523u16 , name : "82860 860 (Wombat) Chipset AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9524u16 , name : "82860 860 (Wombat) Chipset PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9536u16 , name : "E7500 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" }] } , Device { vendor_id : 32902u16 , id : 9537u16 , name : "E7500/E7501 Host RASUM Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 13440u16 , name : "P4DP6" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13348u16 , name : "SE7501HG2 Mainboard" }] } , Device { vendor_id : 32902u16 , id : 9539u16 , name : "E7500/E7501 Hub Interface B PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9540u16 , name : "E7500/E7501 Hub Interface B RASUM Controller" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" }] } , Device { vendor_id : 32902u16 , id : 9541u16 , name : "E7500/E7501 Hub Interface C PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9542u16 , name : "E7500/E7501 Hub Interface C RASUM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9543u16 , name : "E7500/E7501 Hub Interface D PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9544u16 , name : "E7500/E7501 Hub Interface D RASUM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9548u16 , name : "E7501 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13348u16 , name : "SE7501HG2 Mainboard" }] } , Device { vendor_id : 32902u16 , id : 9552u16 , name : "E7505 Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9553u16 , name : "E7505/E7205 Series RAS Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9554u16 , name : "E7505/E7205 PCI-to-AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9555u16 , name : "E7505 Hub Interface B PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9556u16 , name : "E7505 Hub Interface B PCI-to-PCI Bridge RAS Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9565u16 , name : "E7205 Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9568u16 , name : "82845G/GL[Brookdale-G]/GE/PE DRAM Controller/Host-Hub Interface" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 294u16 , name : "Optiplex GX260" } , SubSystem { subvendor : 5208u16 , subdevice : 9568u16 , name : "GA-8PE667 Ultra" } , SubSystem { subvendor : 5218u16 , subdevice : 22528u16 , name : "845PE Max (MS-6580)" }] } , Device { vendor_id : 32902u16 , id : 9569u16 , name : "82845G/GL[Brookdale-G]/GE/PE Host-to-AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9570u16 , name : "82845G/GL[Brookdale-G]/GE Chipset Integrated Graphics Device" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 185u16 , name : "Evo D510 SFF" } , SubSystem { subvendor : 4116u16 , subdevice : 615u16 , name : "NetVista A30p" } , SubSystem { subvendor : 4136u16 , subdevice : 352u16 , name : "Dimension 2400" } , SubSystem { subvendor : 5940u16 , subdevice : 4099u16 , name : "D1521 Mainboard (Fujitsu-Siemens)" } , SubSystem { subvendor : 5940u16 , subdevice : 4100u16 , name : "D1451 Mainboard (SCENIC N300, i845GV)" }] } , Device { vendor_id : 32902u16 , id : 9584u16 , name : "82865G/PE/P DRAM Controller/Host-Hub Interface" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "d330 uT" } , SubSystem { subvendor : 4163u16 , subdevice : 33010u16 , name : "P4P800/P5P800 series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9584u16 , name : "GA-8IPE1000 Pro2 motherboard (865PE)" }] } , Device { vendor_id : 32902u16 , id : 9585u16 , name : "82865G/PE/P AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9586u16 , name : "82865G Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 647u16 , name : "ThinkCentre S50" } , SubSystem { subvendor : 4136u16 , subdevice : 413u16 , name : "Dimension 3000" } , SubSystem { subvendor : 4156u16 , subdevice : 4796u16 , name : "D530 sff(dc578av)" } , SubSystem { subvendor : 4163u16 , subdevice : 32933u16 , name : "P5P800-MX Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30288u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5940u16 , subdevice : 4123u16 , name : "Fujitsu-Siemens Scenic E300 i865GV" } , SubSystem { subvendor : 32902u16 , subdevice : 16966u16 , name : "Desktop Board D865GBF" } , SubSystem { subvendor : 32902u16 , subdevice : 19523u16 , name : "Desktop Board D865GLC" }] } , Device { vendor_id : 32902u16 , id : 9587u16 , name : "82865G/PE/P PCI to CSA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9590u16 , name : "82865G/PE/P Processor to I/O Memory Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9592u16 , name : "82875P/E7210 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 9592u16 , name : "GA-8KNXP motherboard (875P)" } , SubSystem { subvendor : 5218u16 , subdevice : 30080u16 , name : "MS-6758 (875P Neo)" } , SubSystem { subvendor : 5593u16 , subdevice : 17792u16 , name : "P4SCE Motherboard" }] } , Device { vendor_id : 32902u16 , id : 9593u16 , name : "82875P Processor to AGP Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9595u16 , name : "82875P/E7210 Processor to PCI to CSA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9598u16 , name : "82875P/E7210 Processor to I/O Memory Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9600u16 , name : "82915G/P/GV/GL/PL/910GL Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 9600u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4187u16 , name : "Scenic W620" }] } , Device { vendor_id : 32902u16 , id : 9601u16 , name : "82915G/P/GV/GL/PL/910GL PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9602u16 , name : "82915G/GV/910GL Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 4217u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 12294u16 , name : "DC7100 SFF(DX878AV)" } , SubSystem { subvendor : 4163u16 , subdevice : 9602u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9602u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4187u16 , name : "Scenic W620" } , SubSystem { subvendor : 6217u16 , subdevice : 9602u16 , name : "ASRock P4Dual-915GL" }] } , Device { vendor_id : 32902u16 , id : 9604u16 , name : "82925X/XE Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" }] } , Device { vendor_id : 32902u16 , id : 9605u16 , name : "82925X/XE PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9608u16 , name : "E7220/E7221 Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9609u16 , name : "E7220/E7221 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9610u16 , name : "E7221 Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9616u16 , name : "Mobile 915GM/PM/GMS/910GML Express Processor to DRAM Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1397u16 , name : "ThinkPad X41 / Z60t" } , SubSystem { subvendor : 4136u16 , subdevice : 386u16 , name : "Latitude C610" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 33497u16 , name : "Asus Eee PC 900" } , SubSystem { subvendor : 4173u16 , subdevice : 33207u16 , name : "Vaio VGN-S3XP" } , SubSystem { subvendor : 41732u16 , subdevice : 33207u16 , name : "Vaio VGN-S3XP" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9617u16 , name : "Mobile 915GM/PM Express PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240 Mobile Workstation" }] } , Device { vendor_id : 32902u16 , id : 9618u16 , name : "Mobile 915GM/GMS/910GML Express Graphics Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1410u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12426u16 , name : "NC6220" } , SubSystem { subvendor : 4163u16 , subdevice : 6273u16 , name : "GMA 900 915GM Integrated Graphics" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9633u16 , name : "6300ESB LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9634u16 , name : "6300ESB PATA Storage Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer IDE" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9635u16 , name : "6300ESB SATA Storage Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9636u16 , name : "6300ESB SMBus Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9638u16 , name : "6300ESB AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" }] } , Device { vendor_id : 32902u16 , id : 9639u16 , name : "6300ESB AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9641u16 , name : "6300ESB USB Universal Host Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer USB" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9642u16 , name : "6300ESB USB Universal Host Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9643u16 , name : "6300ESB Watchdog Timer" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9644u16 , name : "6300ESB I/O Advanced Programmable Interrupt Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9645u16 , name : "6300ESB USB2 Enhanced Host Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer USB 2.0" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9646u16 , name : "6300ESB 64-bit PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9648u16 , name : "6300ESB SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 9664u16 , name : "5000X Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9680u16 , name : "5000Z Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9684u16 , name : "5000V Chipset Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" }] } , Device { vendor_id : 32902u16 , id : 9688u16 , name : "5000P Chipset Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9698u16 , name : "5000 Series Chipset PCI Express x4 Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9699u16 , name : "5000 Series Chipset PCI Express x4 Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9700u16 , name : "5000 Series Chipset PCI Express x4 Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9701u16 , name : "5000 Series Chipset PCI Express x4 Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9702u16 , name : "5000 Series Chipset PCI Express x4 Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9703u16 , name : "5000 Series Chipset PCI Express x4 Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9712u16 , name : "5000 Series Chipset FSB Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 FSB Registers" } , SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9713u16 , name : "5000 Series Chipset Reserved Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9715u16 , name : "5000 Series Chipset Reserved Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9717u16 , name : "5000 Series Chipset FBD Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9718u16 , name : "5000 Series Chipset FBD Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12797u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9719u16 , name : "5000 Series Chipset PCI Express x8 Port 2-3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9720u16 , name : "5000 Series Chipset PCI Express x8 Port 4-5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9721u16 , name : "5000 Series Chipset PCI Express x8 Port 6-7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9722u16 , name : "5000X Chipset PCI Express x16 Port 4-7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9728u16 , name : "E8500/E8501 Hub Interface 1.5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Hub Interface" }] } , Device { vendor_id : 32902u16 , id : 9729u16 , name : "E8500/E8501 PCI Express x4 Port D" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9730u16 , name : "E8500/E8501 PCI Express x4 Port C0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9731u16 , name : "E8500/E8501 PCI Express x4 Port C1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9732u16 , name : "E8500/E8501 PCI Express x4 Port B0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9733u16 , name : "E8500/E8501 PCI Express x4 Port B1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9734u16 , name : "E8500/E8501 PCI Express x4 Port A0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9735u16 , name : "E8500/E8501 PCI Express x4 Port A1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9736u16 , name : "E8500/E8501 PCI Express x8 Port C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9737u16 , name : "E8500/E8501 PCI Express x8 Port B" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9738u16 , name : "E8500/E8501 PCI Express x8 Port A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9740u16 , name : "E8500/E8501 IMI Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9744u16 , name : "E8500/E8501 FSB Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9745u16 , name : "E8500/E8501 Address Mapping Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9746u16 , name : "E8500/E8501 RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9747u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9748u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9749u16 , name : "E8500/E8501 Miscellaneous Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9751u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9752u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9753u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9754u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9755u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9756u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9757u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9758u16 , name : "E8500/E8501 Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9760u16 , name : "E8500/E8501 eXternal Memory Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Memory Bridge" }] } , Device { vendor_id : 32902u16 , id : 9761u16 , name : "E8500/E8501 XMB Miscellaneous Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 XMB Registers" }] } , Device { vendor_id : 32902u16 , id : 9762u16 , name : "E8500/E8501 XMB Memory Interleaving Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Memory Interleaving Registers" }] } , Device { vendor_id : 32902u16 , id : 9763u16 , name : "E8500/E8501 XMB DDR Initialization and Calibration" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 DDR Initialization and Calibration" }] } , Device { vendor_id : 32902u16 , id : 9764u16 , name : "E8500/E8501 XMB Reserved Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Reserved Registers" }] } , Device { vendor_id : 32902u16 , id : 9765u16 , name : "E8500/E8501 XMB Reserved Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Reserved Registers" }] } , Device { vendor_id : 32902u16 , id : 9766u16 , name : "E8500/E8501 XMB Reserved Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Reserved Registers" }] } , Device { vendor_id : 32902u16 , id : 9767u16 , name : "E8500/E8501 XMB Reserved Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Reserved Registers" }] } , Device { vendor_id : 32902u16 , id : 9792u16 , name : "82801FB/FR (ICH6/ICH6R) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9793u16 , name : "82801FBM (ICH6M) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1384u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" }] } , Device { vendor_id : 32902u16 , id : 9794u16 , name : "82801FW/FRW (ICH6W/ICH6RW) LPC Interface Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9809u16 , name : "82801FB/FW (ICH6/ICH6W) SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4163u16 , subdevice : 9729u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 32902u16 , subdevice : 16711u16 , name : "D915GAG Motherboard" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9810u16 , name : "82801FR/FRW (ICH6R/ICH6RW) SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" }] } , Device { vendor_id : 32902u16 , id : 9811u16 , name : "82801FBM (ICH6M) SATA Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1386u16 , name : "ThinkPad X41" }] } , Device { vendor_id : 32902u16 , id : 9816u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #1" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1381u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9560u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9817u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #2" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1381u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9817u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9818u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #3" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1381u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9818u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9819u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #4" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1381u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9818u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9820u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) USB2 EHCI Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1382u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20486u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 32902u16 , subdevice : 9820u16 , name : "Dimension 3100" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9824u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240 Mobile Workstation" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9826u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240 Mobile Workstation" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9828u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 3" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9830u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 4" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9832u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1463u16 , name : "ThinkPad Z60t" } , SubSystem { subvendor : 4156u16 , subdevice : 10761u16 , name : "PufferM-UL8E" } , SubSystem { subvendor : 4163u16 , subdevice : 4467u16 , name : "A6VC" } , SubSystem { subvendor : 4163u16 , subdevice : 33102u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 9834u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1387u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9834u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9836u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9837u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 106u16 , name : "Conexant AC'97 CoDec (in Acer TravelMate 2410 serie laptop)" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" }] } , Device { vendor_id : 32902u16 , id : 9838u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1409u16 , name : "ThinkPad X41 (Analog Devices AD1981B codec)" } , SubSystem { subvendor : 4133u16 , subdevice : 106u16 , name : "Realtek ALC 655 codec (in Acer TravelMate 2410 serie laptop)" } , SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4136u16 , subdevice : 377u16 , name : "Optiplex GX280" } , SubSystem { subvendor : 4136u16 , subdevice : 386u16 , name : "Latitude D610 Laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 391u16 , name : "Precision M70 Laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 392u16 , name : "Inspiron 6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq NC6220" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12294u16 , name : "DC7100 SFF(DX878AV)" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5421u16 , subdevice : 1861u16 , name : "Packard Bell A8550 Laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4186u16 , name : "Scenic W620" }] } , Device { vendor_id : 32902u16 , id : 9839u16 , name : "82801FB/FBM/FR/FW/FRW (ICH6 Family) IDE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 375u16 , name : "Dimension 8400" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4163u16 , subdevice : 32934u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 9839u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28712u16 , name : "915P/G Neo2" } , SubSystem { subvendor : 5940u16 , subdevice : 4188u16 , name : "Scenic W620" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 9840u16 , name : "631xESB/632xESB/3100 Chipset LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9856u16 , name : "631xESB/632xESB/3100 Chipset SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9857u16 , name : "631xESB/632xESB SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9858u16 , name : "631xESB/632xESB SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "Adaptec Serial ATA HostRAID" }] } , Device { vendor_id : 32902u16 , id : 9859u16 , name : "631xESB/632xESB SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9864u16 , name : "631xESB/632xESB/3100 Chipset UHCI USB Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 onboard USB" } , SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 onboard USB" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9865u16 , name : "631xESB/632xESB/3100 Chipset UHCI USB Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 onboard USB" } , SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 onboard USB" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9866u16 , name : "631xESB/632xESB/3100 Chipset UHCI USB Controller #3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 onboard USB" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9867u16 , name : "631xESB/632xESB/3100 Chipset UHCI USB Controller #4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 onboard USB" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9868u16 , name : "631xESB/632xESB/3100 Chipset EHCI USB2 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 onboard USB" } , SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 onboard USB" } , SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9872u16 , name : "631xESB/632xESB/3100 Chipset PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 9874u16 , name : "631xESB/632xESB/3100 Chipset PCI Express Root Port 2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" }] } , Device { vendor_id : 32902u16 , id : 9876u16 , name : "631xESB/632xESB/3100 Chipset PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9878u16 , name : "631xESB/632xESB/3100 Chipset PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9880u16 , name : "631xESB/632xESB AC '97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9881u16 , name : "631xESB/632xESB AC '97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9882u16 , name : "631xESB/632xESB High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 9883u16 , name : "631xESB/632xESB/3100 Chipset SMBus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 32902u16 , id : 9886u16 , name : "631xESB/632xESB IDE Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3, DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 9984u16 , name : "Optane SSD 900P Series" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 14592u16 , name : "900P Series [Add-in Card]" } , SubSystem { subvendor : 32902u16 , subdevice : 14593u16 , name : "900P Series [2.5\" SFF]" }] } , Device { vendor_id : 32902u16 , id : 9985u16 , name : "NVMe Datacenter SSD [Optane]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8192u16 , name : "Express Flash NVMe [Optane] 375GB 2.5\" U.2 (P4800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8193u16 , name : "Express Flash NVMe [Optane] 750GB 2.5\" U.2 (P4800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8194u16 , name : "Express Flash NVMe [Optane] 750GB AIC (P4800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8202u16 , name : "Express Flash NVMe [Optane] 375GB AIC (P4800X)" } , SubSystem { subvendor : 32902u16 , subdevice : 14596u16 , name : "NVMe Datacenter SSD [Optane] x4 AIC (P4800X)" } , SubSystem { subvendor : 32902u16 , subdevice : 14597u16 , name : "NVMe Datacenter SSD [Optane] 15mm 2.5\" U.2 (P4800X)" }] } , Device { vendor_id : 32902u16 , id : 10000u16 , name : "Dynamic Load Balancer 2.0 (DLB)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10004u16 , name : "Dynamic Load Balancer 2.5 (DLB)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10005u16 , name : "Dynamic Load Balancer (DLB) Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10019u16 , name : "Wi-Fi 6 AX200" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5716u16 , name : "Killer Wi-Fi 6 AX1650x (AX200NGW)" } , SubSystem { subvendor : 32902u16 , subdevice : 132u16 , name : "Wi-Fi 6 AX200NGW" }] } , Device { vendor_id : 32902u16 , id : 10021u16 , name : "Wi-Fi 6E(802.11ax) AX210/AX1675* 2x2 [Typhoon Peak]" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5747u16 , name : "Killer AX1675w 160MHz" } , SubSystem { subvendor : 6742u16 , subdevice : 5748u16 , name : "Killer Wi-Fi 6E AX1675x 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 32u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 36u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 784u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 1296u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 2576u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 8224u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 16416u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 24608u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 24612u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 57376u16 , name : "Wi-Fi 6 AX210 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 57380u16 , name : "Wi-Fi 6 AX210 160MHz" }] } , Device { vendor_id : 32902u16 , id : 10027u16 , name : "Wi-Fi 7(802.11be) AX1775*/AX1790*/BE20*/BE401/BE1750* 2x2" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 240u16 , name : "BE200 320MHz [Gale Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 244u16 , name : "BE200 320MHz [Gale Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 756u16 , name : "BE202 160MHz [Misty Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 16624u16 , name : "BE200 320MHz [Gale Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 17140u16 , name : "BE202 160MHz [Misty Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 57588u16 , name : "BE200 320MHz [Gale Peak]" }] } , Device { vendor_id : 32902u16 , id : 10096u16 , name : "82945G/GZ/P/PL Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4163u16 , subdevice : 33146u16 , name : "P5LD2-VM Mainboard" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6217u16 , subdevice : 10096u16 , name : "ConRoe1333-D667" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10097u16 , name : "82945G/GZ/P/PL PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10098u16 , name : "82945G/GZ Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6217u16 , subdevice : 10098u16 , name : "ConRoe1333-D667" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" } , SubSystem { subvendor : 32902u16 , subdevice : 54789u16 , name : "Desktop Board D945GCCR" }] } , Device { vendor_id : 32902u16 , id : 10100u16 , name : "82955X Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10101u16 , name : "82955X PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10102u16 , name : "82945G/GZ Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10104u16 , name : "E7230/3000/3010 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" }] } , Device { vendor_id : 32902u16 , id : 10105u16 , name : "E7230/3000/3010 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10106u16 , name : "82975X/3010 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10108u16 , name : "82975X Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33144u16 , name : "P5WDG2 WS Professional motherboard" }] } , Device { vendor_id : 32902u16 , id : 10109u16 , name : "82975X PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10112u16 , name : "82915G/GV/GL/910GL [Grantsdale] Graphics Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10114u16 , name : "82915G Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 9602u16 , name : "P5GD1-VW Mainboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4187u16 , name : "Scenic W620" }] } , Device { vendor_id : 32902u16 , id : 10130u16 , name : "Mobile 915GM/GMS/910GML Express Graphics Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1410u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12426u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 6273u16 , name : "GMA 900 915GM Integrated Graphics" } , SubSystem { subvendor : 58559u16 , subdevice : 3277u16 , name : "CCD-CALYPSO" } , SubSystem { subvendor : 58559u16 , subdevice : 3283u16 , name : "CD3-JIVE" } , SubSystem { subvendor : 58559u16 , subdevice : 22705u16 , name : "XB1" }] } , Device { vendor_id : 32902u16 , id : 10144u16 , name : "Mobile 945GM/PM/GMS, 943/940GML and 945GT Express Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 6058u16 , subdevice : 8213u16 , name : "ThinkPad T60" } , SubSystem { subvendor : 6058u16 , subdevice : 8215u16 , name : "ThinkPad R60/T60/X60 series" }] } , Device { vendor_id : 32902u16 , id : 10145u16 , name : "Mobile 945GM/PM/GMS, 943/940GML and 945GT Express PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" }] } , Device { vendor_id : 32902u16 , id : 10146u16 , name : "Mobile 945GM/GMS, 943/940GML Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 6058u16 , subdevice : 8218u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 39170u16 , subdevice : 5508u16 , name : "CCE MPL-D10H120F" }] } , Device { vendor_id : 32902u16 , id : 10150u16 , name : "Mobile 945GM/GMS/GME, 943/940GML Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11 integrated graphics (secondary)" } , SubSystem { subvendor : 6058u16 , subdevice : 8218u16 , name : "ThinkPad R60/T60/X60 series" }] } , Device { vendor_id : 32902u16 , id : 10156u16 , name : "Mobile 945GSE Express Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" }] } , Device { vendor_id : 32902u16 , id : 10157u16 , name : "Mobile 945GSE Express PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10158u16 , name : "Mobile 945GSE Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11 integrated graphics (primary)" }] } , Device { vendor_id : 32902u16 , id : 10160u16 , name : "82801GH (ICH7DH) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10168u16 , name : "82801GB/GR (ICH7 Family) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM Motherboard" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10169u16 , name : "82801GBM (ICH7-M) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 6058u16 , subdevice : 8201u16 , name : "ThinkPad R60/T60/X60 series" }] } , Device { vendor_id : 32902u16 , id : 10172u16 , name : "NM10 Family LPC Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 10173u16 , name : "82801GHM (ICH7-M DH) LPC Interface Bridge" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" }] } , Device { vendor_id : 32902u16 , id : 10176u16 , name : "NM10/ICH7 Family SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM Motherboard" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5218u16 , subdevice : 8976u16 , name : "MSI Hetis 945" } , SubSystem { subvendor : 5218u16 , subdevice : 29238u16 , name : "945P Neo3-F Rev. 2.2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10177u16 , name : "NM10/ICH7 Family SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 45061u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 22594u16 , name : "DeskTop Board D975XBX" }] } , Device { vendor_id : 32902u16 , id : 10179u16 , name : "82801GR/GDH (ICH7R/ICH7DH) SATA Controller [RAID mode]" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10180u16 , name : "82801GBM/GHM (ICH7-M Family) SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 6058u16 , subdevice : 8206u16 , name : "ThinkPad T60" }] } , Device { vendor_id : 32902u16 , id : 10181u16 , name : "82801GBM/GHM (ICH7-M Family) SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 6058u16 , subdevice : 8205u16 , name : "ThinkPad R60/T60/X60 series" }] } , Device { vendor_id : 32902u16 , id : 10182u16 , name : "82801GHM (ICH7-M DH) SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10184u16 , name : "NM10/ICH7 Family USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM, P5LD2-VM Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8202u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10185u16 , name : "NM10/ICH7 Family USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM, P5LD2-VM Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8202u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10186u16 , name : "NM10/ICH7 Family USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM, P5LD2-VM Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8202u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10187u16 , name : "NM10/ICH7 Family USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM, P5LD2-VM Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8202u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10188u16 , name : "NM10/ICH7 Family USB2 EHCI Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM, P5LD2-VM Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33709u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20486u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8203u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10192u16 , name : "NM10/ICH7 Family PCI Express Port 1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8209u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 10194u16 , name : "NM10/ICH7 Family PCI Express Port 2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8209u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 10196u16 , name : "NM10/ICH7 Family PCI Express Port 3" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8209u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 10198u16 , name : "NM10/ICH7 Family PCI Express Port 4" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8209u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 10200u16 , name : "NM10/ICH7 Family High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4387u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 5060u16 , name : "G2P" } , SubSystem { subvendor : 4163u16 , subdevice : 33151u16 , name : "P5LD2-VM Mainboard (Realtek ALC 882 codec)" } , SubSystem { subvendor : 4163u16 , subdevice : 33353u16 , name : "P5B-MX/WiFi-AP" } , SubSystem { subvendor : 4163u16 , subdevice : 33424u16 , name : "P5KPL-VM Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P5KPL-CM Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33847u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33287u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 4473u16 , subdevice : 65296u16 , name : "Toshiba Satellite A100-796 audio (Realtek ALC861)" } , SubSystem { subvendor : 4473u16 , subdevice : 65329u16 , name : "AC97 Data Fax SoftModem with SmartCP" } , SubSystem { subvendor : 5191u16 , subdevice : 4163u16 , name : "Asus A8JP (Analog Devices AD1986A)" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-D525TUD (Realtek ALC887)" } , SubSystem { subvendor : 5208u16 , subdevice : 41218u16 , name : "GA-8I945PG-RH Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 5421u16 , subdevice : 1875u16 , name : "Softmodem" } , SubSystem { subvendor : 5940u16 , subdevice : 4269u16 , name : "Conexant softmodem SmartCP" } , SubSystem { subvendor : 6058u16 , subdevice : 8208u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 6058u16 , subdevice : 14338u16 , name : "3000 C200 audio [Realtek ALC861VD]" } , SubSystem { subvendor : 32902u16 , subdevice : 4370u16 , name : "DeskTop Board D945GTP" } , SubSystem { subvendor : 32902u16 , subdevice : 10200u16 , name : "DeskTop Board D945GTP" } , SubSystem { subvendor : 32902u16 , subdevice : 54808u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 33668u16 , subdevice : 30336u16 , name : "STAC9221 HD Audio Codec" }] } , Device { vendor_id : 32902u16 , id : 10202u16 , name : "NM10/ICH7 Family SMBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "9814 WKMI" } , SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM Motherboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-8I945PG-RH/GA-D525TUD Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8207u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" } , SubSystem { subvendor : 32902u16 , subdevice : 22594u16 , name : "DeskTop Board D975XBX" }] } , Device { vendor_id : 32902u16 , id : 10204u16 , name : "NM10/ICH7 Family LAN Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 32902u16 , subdevice : 12429u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10205u16 , name : "82801G (ICH7 Family) AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10206u16 , name : "82801G (ICH7 Family) AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 429u16 , name : "OptiPlex GX620" } , SubSystem { subvendor : 5218u16 , subdevice : 29287u16 , name : "Realtek ALC883 Audio Controller" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11 integrated audio (AD1981BL codec)" }] } , Device { vendor_id : 32902u16 , id : 10207u16 , name : "82801G (ICH7 Family) IDE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4156u16 , subdevice : 10811u16 , name : "Pavilion A1512X" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4156u16 , subdevice : 12501u16 , name : "530 Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 33145u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM Motherboard" } , SubSystem { subvendor : 4219u16 , subdevice : 20552u16 , name : "E4500" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 5218u16 , subdevice : 29720u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 8204u16 , name : "ThinkPad R60/T60/X60 series" } , SubSystem { subvendor : 32902u16 , subdevice : 21582u16 , name : "DeskTop Board D945GTP" }] } , Device { vendor_id : 32902u16 , id : 10208u16 , name : "82801GR/GH/GHM (ICH7 Family) PCI Express Port 5" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" }] } , Device { vendor_id : 32902u16 , id : 10210u16 , name : "82801GR/GH/GHM (ICH7 Family) PCI Express Port 6" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" }] } , Device { vendor_id : 32902u16 , id : 10256u16 , name : "82801HB/HR (ICH8/R) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" }] } , Device { vendor_id : 32902u16 , id : 10257u16 , name : "82801HEM (ICH8M-E) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 6058u16 , subdevice : 8374u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10258u16 , name : "82801HH (ICH8DH) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10260u16 , name : "82801HO (ICH8DO) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10261u16 , name : "82801HM (ICH8M) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8357u16 , name : "ThinkPad R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10272u16 , name : "82801H (ICH8 Family) 4 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" }] } , Device { vendor_id : 32902u16 , id : 10273u16 , name : "82801HR/HO/HH (ICH8R/DO/DH) 6 port SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10274u16 , name : "SATA Controller (RAID 0/1/5/10) In-box Compatible ID (Desktop RST)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" }] } , Device { vendor_id : 32902u16 , id : 10275u16 , name : "SSATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10276u16 , name : "82801HB (ICH8) 4 port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" }] } , Device { vendor_id : 32902u16 , id : 10277u16 , name : "82801HR/HO/HH (ICH8R/DO/DH) 2 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" }] } , Device { vendor_id : 32902u16 , id : 10278u16 , name : "SATA Controller (RAID 0/1/5/10) In-box Compatible ID (Server/Desktop RST)" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 256u16 , name : "Intel RSTe SATA Software RAID" } , SubSystem { subvendor : 7497u16 , subdevice : 257u16 , name : "Intel RSTe SATA Software RAID" } , SubSystem { subvendor : 7497u16 , subdevice : 258u16 , name : "Intel RSTe SATA Software RAID" } , SubSystem { subvendor : 7497u16 , subdevice : 259u16 , name : "Intel RSTe SATA Software RAID" } , SubSystem { subvendor : 7497u16 , subdevice : 260u16 , name : "Intel RSTe SATA Software RAID" } , SubSystem { subvendor : 7497u16 , subdevice : 261u16 , name : "Intel RSTe SATA Software RAID" }] } , Device { vendor_id : 32902u16 , id : 10279u16 , name : "SSATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10280u16 , name : "82801HM/HEM (ICH8M/ICH8M-E) SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 6058u16 , subdevice : 8360u16 , name : "ThinkPad R61" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10281u16 , name : "82801HM/HEM (ICH8M/ICH8M-E) SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8359u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10282u16 , name : "SATA Controller (RAID 0/1/5/10) In-box Compatible ID (Mobile RST)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 10283u16 , name : "C740 Series (Emmitsburg) Chipsets SATA2 Controller (RAID) Alternate ID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10287u16 , name : "tSATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10288u16 , name : "82801H (ICH8 Family) USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8362u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10289u16 , name : "82801H (ICH8 Family) USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8362u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10290u16 , name : "82801H (ICH8 Family) USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8362u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10291u16 , name : "82801H (ICH8 Family) USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" }] } , Device { vendor_id : 32902u16 , id : 10292u16 , name : "82801H (ICH8 Family) USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8362u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10293u16 , name : "82801H (ICH8 Family) USB UHCI Controller #5" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8362u16 , name : "Thinkpad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10294u16 , name : "82801H (ICH8 Family) USB2 EHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8363u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10298u16 , name : "82801H (ICH8 Family) USB2 EHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8363u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10302u16 , name : "82801H (ICH8 Family) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36872u16 , name : "Vaio VGN-SZ79SN_C" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 6058u16 , subdevice : 8361u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10303u16 , name : "82801H (ICH8 Family) PCI Express Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8365u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10305u16 , name : "82801H (ICH8 Family) PCI Express Port 2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8365u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10307u16 , name : "82801H (ICH8 Family) PCI Express Port 3" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8365u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10309u16 , name : "82801H (ICH8 Family) PCI Express Port 4" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 6058u16 , subdevice : 8365u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10311u16 , name : "82801H (ICH8 Family) PCI Express Port 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 6058u16 , subdevice : 8365u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 32902u16 , id : 10313u16 , name : "82801H (ICH8 Family) PCI Express Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10315u16 , name : "82801H (ICH8 Family) HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 287u16 , name : "Realtek ALC268 audio codec" } , SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4133u16 , subdevice : 325u16 , name : "Realtek ALC889 (Aspire 8920G w. Dolby Theater)" } , SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 505u16 , name : "Latitude D630" } , SubSystem { subvendor : 4136u16 , subdevice : 511u16 , name : "Precision M4300" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4136u16 , subdevice : 598u16 , name : "Studio 1735" } , SubSystem { subvendor : 4156u16 , subdevice : 10242u16 , name : "Compaq dc7700p" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4163u16 , subdevice : 4921u16 , name : "M51S series" } , SubSystem { subvendor : 4163u16 , subdevice : 6131u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 33260u16 , name : "P5B" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36872u16 , name : "Vaio VGN-SZ79SN_C" } , SubSystem { subvendor : 4173u16 , subdevice : 36886u16 , name : "Sony VAIO VGN-AR51M" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 5361u16 , subdevice : 20561u16 , name : "Presario C700" } , SubSystem { subvendor : 6058u16 , subdevice : 8364u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16520u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 33668u16 , subdevice : 30230u16 , name : "Dell Vostro 1400" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10319u16 , name : "82801H (ICH8 Family) Thermal Reporting Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10320u16 , name : "82801HM/HEM (ICH8M/ICH8M-E) IDE Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8358u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10368u16 , name : "Ice Lake DDRIO Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10432u16 , name : "Volume Management Device (VMD)" , subsystems : & [SubSystem { subvendor : 7497u16 , subdevice : 282u16 , name : "Intel VROC (VMD NVMe RAID) for ThinkSystem V4 PL" }] } , Device { vendor_id : 32902u16 , id : 10514u16 , name : "82801IH (ICH9DH) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10516u16 , name : "82801IO (ICH9DO) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" }] } , Device { vendor_id : 32902u16 , id : 10518u16 , name : "82801IR (ICH9R) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" }] } , Device { vendor_id : 32902u16 , id : 10519u16 , name : "ICH9M-E LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8437u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10520u16 , name : "82801IB (ICH9) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 82801IB (ICH9) LPC Interface Controller" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 10521u16 , name : "ICH9M LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10528u16 , name : "82801IR/IO/IH (ICH9R/DO/DH) 4 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 onboard SATA Controller" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801IR [ICH9R]" }] } , Device { vendor_id : 32902u16 , id : 10529u16 , name : "82801IB (ICH9) 2 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 SATA IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 SATA IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 SATA IDE Controller" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" }] } , Device { vendor_id : 32902u16 , id : 10530u16 , name : "82801IR/IO/IH (ICH9R/DO/DH) 6 port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801IR [ICH9R]" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" }] } , Device { vendor_id : 32902u16 , id : 10531u16 , name : "82801IB (ICH9) 4 port SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10533u16 , name : "82801IR/IO (ICH9R/DO) SATA Controller [RAID mode]" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4320u16 , name : "System Board D2542" } , SubSystem { subvendor : 32902u16 , subdevice : 10533u16 , name : "System Board D2542" }] } , Device { vendor_id : 32902u16 , id : 10534u16 , name : "82801I (ICH9 Family) 2 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard SATA Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" }] } , Device { vendor_id : 32902u16 , id : 10536u16 , name : "82801IBM/IEM (ICH9M/ICH9M-E) 2 port SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10537u16 , name : "82801IBM/IEM (ICH9M/ICH9M-E) 4 port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "dv6-1190en" } , SubSystem { subvendor : 6058u16 , subdevice : 8440u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10540u16 , name : "82801IEM (ICH9M-E) SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10541u16 , name : "82801IBM/IEM (ICH9M/ICH9M-E) 2 port SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10544u16 , name : "82801I (ICH9 Family) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "dv6-1190en" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8441u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10546u16 , name : "82801I (ICH9 Family) Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "dv6-1190en" }] } , Device { vendor_id : 32902u16 , id : 10548u16 , name : "82801I (ICH9 Family) USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 8209u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10549u16 , name : "82801I (ICH9 Family) USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10550u16 , name : "82801I (ICH9 Family) USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10551u16 , name : "82801I (ICH9 Family) USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 8209u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 10551u16 , name : "Optiplex 755" } , SubSystem { subvendor : 32902u16 , subdevice : 10562u16 , name : "828011 (ICH9 Family ) USB UHCI Controller" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10552u16 , name : "82801I (ICH9 Family) USB UHCI Controller #5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 10552u16 , name : "Optiplex 755" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10553u16 , name : "82801I (ICH9 Family) USB UHCI Controller #6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard UHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8432u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10554u16 , name : "82801I (ICH9 Family) USB2 EHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB EHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8433u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10556u16 , name : "82801I (ICH9 Family) USB2 EHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 onboard EHCI" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 USB EHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8433u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 10556u16 , name : "Optiplex 755" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10558u16 , name : "82801I (ICH9 Family) HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "dv6-1190en" } , SubSystem { subvendor : 4163u16 , subdevice : 33439u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29530u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6058u16 , subdevice : 8434u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 10558u16 , name : "Optiplex 755" } , SubSystem { subvendor : 32902u16 , subdevice : 10560u16 , name : "Optiplex 755" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10560u16 , name : "82801I (ICH9 Family) PCI Express Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 6058u16 , subdevice : 8435u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 32902u16 , subdevice : 10560u16 , name : "Optiplex 755" }] } , Device { vendor_id : 32902u16 , id : 10562u16 , name : "82801I (ICH9 Family) PCI Express Port 2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 6058u16 , subdevice : 8435u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 10564u16 , name : "82801I (ICH9 Family) PCI Express Port 3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" }] } , Device { vendor_id : 32902u16 , id : 10566u16 , name : "82801I (ICH9 Family) PCI Express Port 4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 6058u16 , subdevice : 8435u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 10568u16 , name : "82801I (ICH9 Family) PCI Express Port 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" } , SubSystem { subvendor : 6058u16 , subdevice : 8435u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 10570u16 , name : "82801I (ICH9 Family) PCI Express Port 6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4163u16 , subdevice : 33399u16 , name : "P5K PRO Motherboard: 82801IR [ICH9R]" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82801I/IR [ICH9/ICH9R]" }] } , Device { vendor_id : 32902u16 , id : 10572u16 , name : "82566DC-2 Gigabit Network Connection" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12334u16 , name : "82566DM-2 Gigabit Network Connection" }] } , Device { vendor_id : 32902u16 , id : 10608u16 , name : "82946GZ/PL/GL Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33339u16 , name : "P5B-MX/WiFi-AP" }] } , Device { vendor_id : 32902u16 , id : 10609u16 , name : "82946GZ/PL/GL PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10610u16 , name : "82946GZ/GL Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33339u16 , name : "P5B-MX/WiFi-AP" }] } , Device { vendor_id : 32902u16 , id : 10611u16 , name : "82946GZ/GL Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10612u16 , name : "82946GZ/GL HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10613u16 , name : "82946GZ/GL HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10614u16 , name : "82946GZ/GL PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10615u16 , name : "82946GZ/GL KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10624u16 , name : "82G35 Express DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10625u16 , name : "82G35 Express PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10626u16 , name : "82G35 Express Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10627u16 , name : "82G35 Express Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10628u16 , name : "82G35 Express HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10640u16 , name : "82Q963/Q965 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 474u16 , name : "OptiPlex 745" }] } , Device { vendor_id : 32902u16 , id : 10641u16 , name : "82Q963/Q965 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10642u16 , name : "82Q963/Q965 Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10643u16 , name : "82Q963/Q965 Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10644u16 , name : "82Q963/Q965 HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10645u16 , name : "82Q963/Q965 HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10646u16 , name : "82Q963/Q965 PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10647u16 , name : "82Q963/Q965 KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10656u16 , name : "82P965/G965 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33258u16 , name : "P5B" } , SubSystem { subvendor : 5218u16 , subdevice : 29302u16 , name : "MS-7276 [G965MDH]" }] } , Device { vendor_id : 32902u16 , id : 10657u16 , name : "82P965/G965 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10658u16 , name : "82G965 Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29302u16 , name : "MS-7276 [G965MDH]" }] } , Device { vendor_id : 32902u16 , id : 10659u16 , name : "82G965 Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10660u16 , name : "82P965/G965 HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10661u16 , name : "82P965/G965 HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10662u16 , name : "82P965/G965 PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10663u16 , name : "82P965/G965 KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10672u16 , name : "82Q35 Express DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10673u16 , name : "82Q35 Express PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10674u16 , name : "82Q35 Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10675u16 , name : "82Q35 Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10676u16 , name : "82Q35 Express MEI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10677u16 , name : "82Q35 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10678u16 , name : "82Q35 Express PT IDER Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10679u16 , name : "82Q35 Express Serial KT Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "OptiPlex 755" }] } , Device { vendor_id : 32902u16 , id : 10688u16 , name : "82G33/G31/P35/P31 Express DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33398u16 , name : "P5K PRO Motherboard: Intel 82P35 Northbridge" } , SubSystem { subvendor : 4163u16 , subdevice : 33456u16 , name : "P5KPL-VM Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29509u16 , name : "MS-7345 Motherboard: Intel 82G33/P35 Northbridge" } , SubSystem { subvendor : 5218u16 , subdevice : 29536u16 , name : "G33/P35 Neo" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" }] } , Device { vendor_id : 32902u16 , id : 10689u16 , name : "82G33/G31/P35/P31 Express PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4163u16 , subdevice : 33398u16 , name : "P5K PRO Motherboard: Intel 82P35 Northbridge" }] } , Device { vendor_id : 32902u16 , id : 10690u16 , name : "82G33/G31 Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4163u16 , subdevice : 33456u16 , name : "P5KPL-VM Motherboard" }] } , Device { vendor_id : 32902u16 , id : 10691u16 , name : "82G33/G31 Express Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 525u16 , name : "Inspiron 530" } , SubSystem { subvendor : 4163u16 , subdevice : 33456u16 , name : "P5KPL-VM Motherboard" }] } , Device { vendor_id : 32902u16 , id : 10692u16 , name : "82G33/G31/P35/P31 Express MEI Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" }] } , Device { vendor_id : 32902u16 , id : 10693u16 , name : "82G33/G31/P35/P31 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10694u16 , name : "82G33/G31/P35/P31 Express PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10695u16 , name : "82G33/G31/P35/P31 Express Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10703u16 , name : "Virtual HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10704u16 , name : "82Q33 Express DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10705u16 , name : "82Q33 Express PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10706u16 , name : "82Q33 Express Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10707u16 , name : "82Q33 Express Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10708u16 , name : "82Q33 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10709u16 , name : "82Q33 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10710u16 , name : "82Q33 Express PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10711u16 , name : "82Q33 Express Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10720u16 , name : "82X38/X48 Express DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10721u16 , name : "82X38/X48 Express Host-Primary PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10724u16 , name : "82X38/X48 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10725u16 , name : "82X38/X48 Express MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10726u16 , name : "82X38/X48 Express PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10727u16 , name : "82X38/X48 Express Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10729u16 , name : "82X38/X48 Express Host-Secondary PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10736u16 , name : "3200/3210 Chipset DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10737u16 , name : "3200/3210 Chipset Host-Primary PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10740u16 , name : "3200/3210 Chipset MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10741u16 , name : "3200/3210 Chipset MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10742u16 , name : "3200/3210 Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10743u16 , name : "3200/3210 Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10745u16 , name : "3210 Chipset Host-Secondary PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10752u16 , name : "Mobile PM965/GM965/GL960 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12485u16 , name : "Compaq 8510p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36869u16 , name : "Vaio VGN-FZ260E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8369u16 , name : "ThinkPad T61" } , SubSystem { subvendor : 6058u16 , subdevice : 8371u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16514u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10753u16 , name : "Mobile PM965/GM965/GL960 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10754u16 , name : "Mobile GM965/GL960 Integrated Graphics Controller (primary)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 505u16 , name : "Latitude D630" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 5346u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8373u16 , name : "GM965 [X3100] on ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16514u16 , name : "GM965 on Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10755u16 , name : "Mobile GM965/GL960 Integrated Graphics Controller (secondary)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 5346u16 , name : "X58LE" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" } , SubSystem { subvendor : 6058u16 , subdevice : 8373u16 , name : "GM965 [X3100] on ThinkPad T61/R61" } , SubSystem { subvendor : 6080u16 , subdevice : 16514u16 , name : "GM965 on Medion WIM 2210 Notebook PC [MD96850]" } , SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10756u16 , name : "Mobile PM965/GM965 MEI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" }] } , Device { vendor_id : 32902u16 , id : 10757u16 , name : "Mobile PM965/GM965 MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10758u16 , name : "Mobile PM965/GM965 PT IDER Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" }] } , Device { vendor_id : 32902u16 , id : 10759u16 , name : "Mobile PM965/GM965 KT Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" }] } , Device { vendor_id : 32902u16 , id : 10768u16 , name : "Mobile GME965/GLE960 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10769u16 , name : "Mobile GME965/GLE960 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10770u16 , name : "Mobile GME965/GLE960 Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10771u16 , name : "Mobile GME965/GLE960 Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 52295u16 , name : "CCG-RUMBA" }] } , Device { vendor_id : 32902u16 , id : 10772u16 , name : "Mobile GME965/GLE960 MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10773u16 , name : "Mobile GME965/GLE960 MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10774u16 , name : "Mobile GME965/GLE960 PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10775u16 , name : "Mobile GME965/GLE960 KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10816u16 , name : "Mobile 4 Series Chipset Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8416u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10817u16 , name : "Mobile 4 Series Chipset PCI Express Graphics Port" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10818u16 , name : "Mobile 4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 682u16 , name : "Dell Inspiron 1545" } , SubSystem { subvendor : 6058u16 , subdevice : 8420u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6058u16 , subdevice : 8466u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10819u16 , name : "Mobile 4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8420u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 6058u16 , subdevice : 8466u16 , name : "ThinkPad T400" } , SubSystem { subvendor : 58559u16 , subdevice : 52301u16 , name : "CCM-BOOGIE" }] } , Device { vendor_id : 32902u16 , id : 10820u16 , name : "Mobile 4 Series Chipset MEI Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8422u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 10821u16 , name : "Mobile 4 Series Chipset MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10822u16 , name : "Mobile 4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10823u16 , name : "Mobile 4 Series Chipset AMT SOL Redirection" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8428u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 32902u16 , id : 10832u16 , name : "Cantiga MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10833u16 , name : "Cantiga MEI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10834u16 , name : "Cantiga PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 10835u16 , name : "Cantiga AMT SOL Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11008u16 , name : "Xeon Processor E7 Product Family System Configuration Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11010u16 , name : "Xeon Processor E7 Product Family System Configuration Controller 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11012u16 , name : "Xeon Processor E7 Product Family Power Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11016u16 , name : "Xeon Processor E7 Product Family Caching Agent 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11020u16 , name : "Xeon Processor E7 Product Family Caching Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11024u16 , name : "Xeon Processor E7 Product Family QPI Home Agent 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11027u16 , name : "Xeon Processor E7 Product Family Memory Controller 0c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11028u16 , name : "Xeon Processor E7 Product Family Memory Controller 0a" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11030u16 , name : "Xeon Processor E7 Product Family Memory Controller 0b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11032u16 , name : "Xeon Processor E7 Product Family QPI Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11035u16 , name : "Xeon Processor E7 Product Family Memory Controller 1c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11036u16 , name : "Xeon Processor E7 Product Family Memory Controller 1a" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11038u16 , name : "Xeon Processor E7 Product Family Memory Controller 1b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11040u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11042u16 , name : "Xeon Processor E7 Product Family System Configuration Controller 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11044u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11048u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11050u16 , name : "Xeon Processor E7 Product Family System Configuration Controller 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11052u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11056u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11060u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11064u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11068u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11072u16 , name : "Xeon Processor E7 Product Family QPI Router Port 0-1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11074u16 , name : "Xeon Processor E7 Product Family QPI Router Port 2-3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11076u16 , name : "Xeon Processor E7 Product Family QPI Router Port 4-5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11078u16 , name : "Xeon Processor E7 Product Family QPI Router Port 6-7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11080u16 , name : "Xeon Processor E7 Product Family Test and Debug 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11084u16 , name : "Xeon Processor E7 Product Family Test and Debug 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11088u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 0: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11090u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 0: Misc. control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11092u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 1: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11094u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 1: Misc. control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11096u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 2: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11098u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 2: Misc. control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11100u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 3: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11102u16 , name : "Xeon Processor E7 Product Family QPI Physical Port 3: Misc. control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11104u16 , name : "Xeon Processor E7 Product Family SMI Physical Port 0: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11106u16 , name : "Xeon Processor E7 Product Family SMI Physical Port 0: Misc control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11108u16 , name : "Xeon Processor E7 Product Family SMI Physical Port 1: REUT control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11110u16 , name : "Xeon Processor E7 Product Family SMI Physical Port 1: Misc control/status" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11112u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11116u16 , name : "Xeon Processor E7 Product Family Last Level Cache Coherence Engine 9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11136u16 , name : "Atom CE2700 Series [Puma 7]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11160u16 , name : "Puma 7 Trusted Execution Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11189u16 , name : "Puma 7 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11191u16 , name : "Puma 7 USB Device Controller (OTG)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11228u16 , name : "Puma 7 Thermal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11234u16 , name : "Puma 7 Security Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11265u16 , name : "Xeon 5500/Core i7 QuickPath Architecture System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11280u16 , name : "Xeon 5500/Core i7 QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11281u16 , name : "Xeon 5500/Core i7 QPI Physical 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11284u16 , name : "Xeon 5500/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11285u16 , name : "Xeon 5500/Core i7 QPI Physical 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11288u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11289u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11290u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11292u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Test Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11296u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 0 Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11297u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 0 Address Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11298u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 0 Rank Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11299u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 0 Thermal Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11304u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 1 Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11305u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 1 Address Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11306u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 1 Rank Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11307u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 1 Thermal Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11312u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 2 Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11313u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 2 Address Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11314u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 2 Rank Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11315u16 , name : "Xeon 5500/Core i7 Integrated Memory Controller Channel 2 Thermal Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11328u16 , name : "Xeon 5500/Core i7 QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11329u16 , name : "Xeon 5500/Core i7 QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11344u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11345u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11346u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11347u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11348u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11349u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11350u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11351u16 , name : "Core Processor QuickPath Architecture Generic Non-Core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11352u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11353u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11354u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11355u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11356u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11357u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11358u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11359u16 , name : "Xeon C5500/C3500 QPI Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11361u16 , name : "Core Processor QuickPath Architecture Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11362u16 , name : "Core Processor QuickPath Architecture Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11376u16 , name : "Xeon 5600 Series QuickPath Architecture Generic Non-core Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11393u16 , name : "Core Processor QuickPath Architecture System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11408u16 , name : "Core Processor QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11409u16 , name : "Core Processor QPI Physical 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11416u16 , name : "Core Processor Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11417u16 , name : "Core Processor Integrated Memory Controller Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11418u16 , name : "Core Processor Integrated Memory Controller Test Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11420u16 , name : "Core Processor Integrated Memory Controller Test Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11424u16 , name : "Core Processor Integrated Memory Controller Channel 0 Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11425u16 , name : "Core Processor Integrated Memory Controller Channel 0 Address Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11426u16 , name : "Core Processor Integrated Memory Controller Channel 0 Rank Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11427u16 , name : "Core Processor Integrated Memory Controller Channel 0 Thermal Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11432u16 , name : "Core Processor Integrated Memory Controller Channel 1 Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11433u16 , name : "Core Processor Integrated Memory Controller Channel 1 Address Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11434u16 , name : "Core Processor Integrated Memory Controller Channel 1 Rank Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11435u16 , name : "Core Processor Integrated Memory Controller Channel 1 Thermal Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11457u16 , name : "Xeon C5500/C3500 QPI System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11472u16 , name : "Xeon C5500/C3500 QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11473u16 , name : "Xeon C5500/C3500 QPI Physical 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11476u16 , name : "Xeon C5500/C3500 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11477u16 , name : "Xeon C5500/C3500 QPI Physical 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11480u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11481u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11482u16 , name : "Xeon C5500/C3500 Integrated Memory Controller RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11484u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Test Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11488u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 0 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11489u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 0 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11490u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 0 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11491u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11496u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 1 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11497u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 1 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11498u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 1 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11499u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11504u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 2 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11505u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 2 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11506u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 2 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11507u16 , name : "Xeon C5500/C3500 Integrated Memory Controller Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11521u16 , name : "Core Processor QuickPath Architecture System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11536u16 , name : "Core Processor QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11537u16 , name : "1st Generation Core i3/5/7 Processor QPI Physical 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11538u16 , name : "1st Generation Core i3/5/7 Processor Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11539u16 , name : "1st Generation Core i3/5/7 Processor Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11649u16 , name : "Xeon 5600 Series QuickPath Architecture System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11664u16 , name : "Xeon 5600 Series QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11665u16 , name : "Xeon 5600 Series QPI Physical 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11666u16 , name : "Xeon 5600 Series Mirror Port Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11667u16 , name : "Xeon 5600 Series Mirror Port Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11668u16 , name : "Xeon 5600 Series QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11669u16 , name : "Xeon 5600 Series QPI Physical 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11672u16 , name : "Xeon 5600 Series Integrated Memory Controller Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11673u16 , name : "Xeon 5600 Series Integrated Memory Controller Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11674u16 , name : "Xeon 5600 Series Integrated Memory Controller RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11676u16 , name : "Xeon 5600 Series Integrated Memory Controller Test Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11680u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 0 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11681u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 0 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11682u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 0 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11683u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11688u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 1 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11689u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 1 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11690u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 1 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11691u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11696u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 2 Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11697u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 2 Address" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11698u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 2 Rank" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11699u16 , name : "Xeon 5600 Series Integrated Memory Controller Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11776u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11777u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11778u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11779u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11780u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11781u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11782u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11783u16 , name : "4 Series Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11792u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11793u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11794u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12360u16 , name : "ThinkCentre M6258" }] } , Device { vendor_id : 32902u16 , id : 11795u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11796u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11797u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11798u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11799u16 , name : "4 Series Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11808u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 643u16 , name : "Vostro 220" } , SubSystem { subvendor : 4163u16 , subdevice : 33491u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 11809u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33491u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 11810u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "GA-EG45M-DS2H Mainboard" }] } , Device { vendor_id : 32902u16 , id : 11811u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "GA-EG45M-DS2H Mainboard" }] } , Device { vendor_id : 32902u16 , id : 11812u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11813u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11814u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11815u16 , name : "4 Series Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11817u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11824u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" }] } , Device { vendor_id : 32902u16 , id : 11825u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11826u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" }] } , Device { vendor_id : 32902u16 , id : 11827u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11828u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11829u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11830u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11831u16 , name : "4 Series Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11840u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11841u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11842u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11843u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11844u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11845u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11846u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11847u16 , name : "4 Series Chipset Serial KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11856u16 , name : "CE Media Processor CE3100" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11858u16 , name : "CE Media Processor Clock and Reset Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11864u16 , name : "CE Media Processor Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11866u16 , name : "CE Media Processor CE3100 A/V Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11867u16 , name : "Graphics Media Accelerator 500 Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11868u16 , name : "CE Media Processor Video Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11869u16 , name : "CE Media Processor Transport Stream Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11870u16 , name : "CE Media Processor Transport Stream Processor 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11871u16 , name : "CE Media Processor Audio DSP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11872u16 , name : "CE Media Processor Audio Interfaces" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11873u16 , name : "CE Media Processor Video Display Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11874u16 , name : "CE Media Processor Video Processing Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11875u16 , name : "CE Media Processor HDMI Tx Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11876u16 , name : "Atom CE2600/3100/4100/4200/5300 Security Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11877u16 , name : "CE Media Processor Expansion Bus Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11878u16 , name : "CE Media Processor UART" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11879u16 , name : "CE Media Processor General Purpose I/Os" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11880u16 , name : "CE Media Processor I2C Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11881u16 , name : "CE Media Processor Smart Card Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11882u16 , name : "CE Media Processor SPI Master Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11886u16 , name : "CE Media Processor Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11887u16 , name : "CE Media Processor Media Timing Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11888u16 , name : "CE Media Processor USB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11889u16 , name : "CE Media Processor SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11891u16 , name : "CE Media Processor CE3100 PCI Express" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11920u16 , name : "4 Series Chipset DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11921u16 , name : "4 Series Chipset PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11922u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11923u16 , name : "4 Series Chipset Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11924u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11925u16 , name : "4 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 11926u16 , name : "4 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12032u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMI2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12033u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12034u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12035u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12036u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12037u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12038u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12039u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12040u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12041u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12042u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12043u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12045u16 , name : "Haswell Xeon Non-Transparent Bridge (Back-to-back)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12046u16 , name : "Haswell Xeon Non-Transparent Bridge (Primary Side)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12047u16 , name : "Haswell Xeon Non-Transparent Bridge (Secondary Side)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12048u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12049u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12050u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12051u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12052u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12053u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12054u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12055u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12056u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12057u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12058u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12059u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12060u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12061u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCIe Ring Interface" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12062u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Scratchpad & Semaphore Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12063u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Scratchpad & Semaphore Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12064u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12065u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 1" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12066u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12067u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 3" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12068u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 4" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12069u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 5" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12070u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 6" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12071u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DMA Channel 7" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12072u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Address Map, VTd_Misc, System Management" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12073u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Hot Plug" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12074u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 RAS, Control Status and Global Errors" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12076u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 I/O APIC" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12078u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 RAID 5/6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12079u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 RAID 5/6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12080u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12082u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12083u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12084u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 PCIe Ring Interface" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12086u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 0 & 1 Monitoring" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12087u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 0 & 1 Monitoring" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12088u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12089u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 I/O Performance Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12090u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12094u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 2 Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12095u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 2 Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12096u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12097u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 2 Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12099u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12101u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12102u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12103u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12128u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12136u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Target Address, Thermal & RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12138u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12139u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12140u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12141u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12142u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO Channel 2/3 Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12143u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO Global Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12144u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12145u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Target Address, Thermal & RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12150u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 E3 QPI Link Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12152u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12153u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Target Address, Thermal & RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12157u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Scratchpad & Semaphore Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12158u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 E3 QPI Link Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12160u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12161u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 R3 QPI Link 0 & 1 Monitoring" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12163u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12165u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12166u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12167u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12168u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 VCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12170u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 VCU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12176u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12179u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12181u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12182u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 QPI Link 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12184u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12185u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12186u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12188u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12192u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Home Agent 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12200u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Target Address, Thermal & RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12202u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12203u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12204u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12205u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel Target Address Decoder" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12206u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO Channel 0/1 Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12207u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO Global Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12208u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12209u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12210u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 0 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12211u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 1 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12212u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12213u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 3 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12214u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 2 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12215u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 0 Channel 3 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12216u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 2 & 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12217u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 2 & 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12218u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 2 & 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12219u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 2 & 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12220u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 0 & 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12221u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 0 & 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12222u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 0 & 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12223u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 DDRIO (VMSE) 0 & 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12224u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12225u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12226u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12227u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12228u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12229u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12240u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12241u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12242u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 0 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12243u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 1 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12244u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12245u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 3 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12246u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 2 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12247u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Integrated Memory Controller 1 Channel 3 ERROR Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12256u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12257u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12258u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12259u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12260u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12261u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12262u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12263u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12264u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12265u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12266u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12267u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12268u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12269u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12270u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12271u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12272u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12273u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12274u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12275u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12276u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12277u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12278u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12279u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Unicast Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12280u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Buffered Ring Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12281u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Buffered Ring Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12282u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Buffered Ring Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12283u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 Buffered Ring Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12284u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 System Address Decoder & Broadcast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12285u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 System Address Decoder & Broadcast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12286u16 , name : "Xeon E7 v3/Xeon E5 v3/Core i7 System Address Decoder & Broadcast Registers" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" }] } , Device { vendor_id : 32902u16 , id : 12545u16 , name : "Killer E3100X 2.5 Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12546u16 , name : "Ethernet Controller I226-K" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12608u16 , name : "Easel/Monette Hill Image Processor [Pixel Visual Core]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12645u16 , name : "Wireless 3165" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16400u16 , name : "Dual Band Wireless AC 3165 [Stone Peak 1x1]" } , SubSystem { subvendor : 32902u16 , subdevice : 16912u16 , name : "Dual Band Wireless AC 3165 [Stone Peak 1x1]" }] } , Device { vendor_id : 32902u16 , id : 12646u16 , name : "Dual Band Wireless-AC 3165 Plus Bluetooth" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16912u16 , name : "Dual Band Wireless-AC 3165" }] } , Device { vendor_id : 32902u16 , id : 12676u16 , name : "GeminiLake [UHD Graphics 605]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12677u16 , name : "GeminiLake [UHD Graphics 600]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12684u16 , name : "Celeron/Pentium Silver Processor Dynamic Platform and Thermal Framework Processor Participant" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12686u16 , name : "Celeron/Pentium Silver Processor NorthPeak" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12688u16 , name : "Celeron/Pentium Silver Processor Gaussian Mixture Model" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12690u16 , name : "Gemini Lake P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12694u16 , name : "Unknown P2SB serial controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12695u16 , name : "Celeron/Pentium Silver Processor PCI-default ISA-bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12696u16 , name : "Celeron/Pentium Silver Processor High Definition Audio" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14347u16 , name : "V130-15IGM Laptop (Lenovo) - Type 81HL" }] } , Device { vendor_id : 32902u16 , id : 12698u16 , name : "Celeron/Pentium Silver Processor Trusted Execution Engine Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12706u16 , name : "Celeron/Pentium Silver Processor Integrated Sensor Solution" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12712u16 , name : "Celeron/Pentium Silver Processor USB 3.0 xHCI Controller" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 12712u16 , name : "Celeron/Pentium Silver Processor USB 3.0 xHCI Controller" }] } , Device { vendor_id : 32902u16 , id : 12716u16 , name : "Celeron/Pentium Silver Processor I2C 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12718u16 , name : "Celeron/Pentium Silver Processor I2C 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12720u16 , name : "Celeron/Pentium Silver Processor I2C 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12722u16 , name : "Celeron/Pentium Silver Processor I2C 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12724u16 , name : "Celeron/Pentium Silver Processor I2C 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12726u16 , name : "Celeron/Pentium Silver Processor I2C 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12728u16 , name : "Celeron/Pentium Silver Processor I2C 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12730u16 , name : "Celeron/Pentium Silver Processor I2C 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12732u16 , name : "Celeron/Pentium Silver Processor Serial IO UART Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12734u16 , name : "Celeron/Pentium Silver Processor Serial IO UART Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12736u16 , name : "Celeron/Pentium Silver Processor Serial IO UART Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12738u16 , name : "Celeron/Pentium Silver Processor Serial IO SPI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12740u16 , name : "Celeron/Pentium Silver Processor Serial IO SPI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12742u16 , name : "Celeron/Pentium Silver Processor Serial IO SPI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12748u16 , name : "Celeron/Pentium Silver Processor SDA Standard Compliant SD Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12756u16 , name : "Celeron/Pentium Silver Processor Gaussian Mixture Model" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12758u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12759u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12760u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12761u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12762u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12763u16 , name : "Gemini Lake PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12764u16 , name : "Gemini Lake PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5458u16 , name : "Killer(R) Wireless-AC 1550i Wireless Network Adapter (9560NGW)" } , SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Wireless-AC 9560" }] } , Device { vendor_id : 32902u16 , id : 12771u16 , name : "Celeron/Pentium Silver Processor SATA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12776u16 , name : "Celeron/Pentium Silver Processor LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12782u16 , name : "Celeron/Pentium Silver Processor Serial IO UART Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12784u16 , name : "Gemini Lake Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12800u16 , name : "GD31244 PCI-X SATA HBA" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 49664u16 , name : "C2K onboard SATA host bus adapter" }] } , Device { vendor_id : 32902u16 , id : 12864u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UPI Misc" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12865u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UPI link/Phy0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12866u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UPI Phy0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12869u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12874u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors IMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12876u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors CHA Unicast Group 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12877u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors CHA Unicast Group 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12880u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors Ubox Event Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12881u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors Ubox Register Access Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12882u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors Ubox Decode" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12886u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12888u16 , name : "Power Control Unit (PCU) CR0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12889u16 , name : "Power Control Unit (PCU) CR1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12890u16 , name : "Power Control Unit (PCU) CR2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12891u16 , name : "Power Control Unit (PCU) CR3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12892u16 , name : "Power Control Unit (PCU) CR4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12893u16 , name : "Power Control Unit (PCU) CR5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12894u16 , name : "Power Control Unit (PCU) CR6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 12895u16 , name : "Power Control Unit (PCU) CR7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13072u16 , name : "IOP348 I/O Processor" , subsystems : & [SubSystem { subvendor : 4180u16 , subdevice : 12336u16 , name : "HRA380 Hitachi RAID Adapter to PCIe" } , SubSystem { subvendor : 4180u16 , subdevice : 12340u16 , name : "HRA381 Hitachi RAID Adapter to PCIe" }] } , Device { vendor_id : 32902u16 , id : 13075u16 , name : "IOP348 I/O Processor (SL8e) in IOC Mode SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13083u16 , name : "IOP348 I/O Processor (SL8x) in IOC Mode SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13105u16 , name : "IOC340 I/O Controller (VV8e) SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13113u16 , name : "IOC340 I/O Controller (VV8x) SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13120u16 , name : "82855PM Processor to I/O Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1321u16 , name : "Thinkpad T40 series" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" }] } , Device { vendor_id : 32902u16 , id : 13121u16 , name : "82855PM Processor to AGP Controller" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30 notebook" }] } , Device { vendor_id : 32902u16 , id : 13155u16 , name : "IOC340 I/O Controller in IOC Mode SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13186u16 , name : "81342 [Chevelon] I/O Processor (ATUe)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13251u16 , name : "IOP348 I/O Processor (SL8De) in IOC Mode SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13259u16 , name : "IOP348 I/O Processor (SL8Dx) in IOC Mode SAS/SATA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13312u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13313u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13314u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13315u16 , name : "5500 I/O Hub to ESI Port" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 I/O Hub to ESI Port" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 I/O Hub to ESI Port" } , SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 I/O Hub to ESI Port" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 I/O Hub to ESI Port" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" }] } , Device { vendor_id : 32902u16 , id : 13316u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13317u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13318u16 , name : "5520 I/O Hub to ESI Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" }] } , Device { vendor_id : 32902u16 , id : 13319u16 , name : "5520/5500/X58 I/O Hub to ESI Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13320u16 , name : "5520/5500/X58 I/O Hub PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" }] } , Device { vendor_id : 32902u16 , id : 13321u16 , name : "5520/5500/X58 I/O Hub PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13322u16 , name : "5520/5500/X58 I/O Hub PCI Express Root Port 3" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" }] } , Device { vendor_id : 32902u16 , id : 13323u16 , name : "5520/X58 I/O Hub PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13324u16 , name : "5520/X58 I/O Hub PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13325u16 , name : "5520/X58 I/O Hub PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13326u16 , name : "5520/5500/X58 I/O Hub PCI Express Root Port 7" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" }] } , Device { vendor_id : 32902u16 , id : 13327u16 , name : "5520/5500/X58 I/O Hub PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13328u16 , name : "7500/5520/5500/X58 I/O Hub PCI Express Root Port 9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13329u16 , name : "7500/5520/5500/X58 I/O Hub PCI Express Root Port 10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13336u16 , name : "7500/5520/5500/X58 Physical Layer Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13337u16 , name : "7500/5520/5500 Physical Layer Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13344u16 , name : "7500/5520/5500/X58 I/O Hub PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13345u16 , name : "7500/5520/5500/X58 I/O Hub PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13346u16 , name : "7500/5520/5500/X58 I/O Hub GPIO and Scratch Pad Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" }] } , Device { vendor_id : 32902u16 , id : 13347u16 , name : "7500/5520/5500/X58 I/O Hub Control Status and RAS Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" }] } , Device { vendor_id : 32902u16 , id : 13349u16 , name : "7500/5520/5500/X58 Physical and Link Layer Registers Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13350u16 , name : "7500/5520/5500/X58 Routing and Protocol Layer Registers Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13351u16 , name : "7500/5520/5500 Physical and Link Layer Registers Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13352u16 , name : "7500/5520/5500 Routing & Protocol Layer Register Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13353u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13354u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13355u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13356u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13357u16 , name : "7500/5520/5500/X58 I/O Hub I/OxAPIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13358u16 , name : "7500/5520/5500/X58 I/O Hub System Management Registers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" }] } , Device { vendor_id : 32902u16 , id : 13359u16 , name : "7500/5520/5500/X58 Trusted Execution Technology Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13360u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13361u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13362u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13363u16 , name : "5520/5500/X58 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13368u16 , name : "7500/5520/5500/X58 I/O Hub Throttle Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13376u16 , name : "Ice Lake UPI Misc" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13377u16 , name : "Ice Lake UPI Link/Phy0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13378u16 , name : "Ice Lake UPI Phy0 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13381u16 , name : "Ice Lake UPI Mesh Stop Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13382u16 , name : "Ice Lake UPI PMON0 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13383u16 , name : "Ice Lake UPI PMON1 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13384u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13386u16 , name : "Ice Lake Integrated Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13387u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13388u16 , name : "Ice Lake CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13389u16 , name : "Ice Lake CHA Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13391u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors CHA All 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13392u16 , name : "Ice Lake Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13393u16 , name : "Ice Lake Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13394u16 , name : "Ice Lake Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13397u16 , name : "Ice Lake Ubox Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13398u16 , name : "Ice Lake NorthPeak" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13399u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors CHA All 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13400u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13401u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13402u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13403u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13404u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13405u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13406u16 , name : "Ice Lake PCU Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13434u16 , name : "Ice Lake PCI Express Root Port A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13435u16 , name : "Ice Lake PCI Express Root Port B" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13436u16 , name : "Ice Lake PCI Express Root Port C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13437u16 , name : "Ice Lake PCI Express Root Port D" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13438u16 , name : "Ice Lake Xeon Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13442u16 , name : "Ice Lake-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13475u16 , name : "Ice Lake-LP SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13476u16 , name : "Ice Lake-LP SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13480u16 , name : "Ice Lake-LP Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13481u16 , name : "Ice Lake-LP Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13482u16 , name : "Ice Lake-LP Serial IO SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13483u16 , name : "Ice Lake-LP Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13488u16 , name : "Ice Lake-LP PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13489u16 , name : "Ice Lake-LP PCIe Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13492u16 , name : "Ice Lake-LP PCIe Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13493u16 , name : "Ice Lake-LP PCIe Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13495u16 , name : "Ice Lake-LP PCI Express Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13496u16 , name : "Ice Lake-LP PCIe Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13498u16 , name : "Ice Lake-LP PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13499u16 , name : "Ice Lake-LP PCIe Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13500u16 , name : "Ice Lake-LP PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13501u16 , name : "Ice Lake-LP PCIe Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13502u16 , name : "Ice Lake-LP PCIe Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13503u16 , name : "Ice Lake-LP PCIe Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13508u16 , name : "Ice Lake-LP SD Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13509u16 , name : "Ice Lake-LP Serial IO I2c Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13510u16 , name : "Ice Lake-LP Serial IO I2c Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13512u16 , name : "Ice Lake-LP Smart Sound Technology Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13523u16 , name : "Ice Lake-LP SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13536u16 , name : "Ice Lake-LP Management Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13544u16 , name : "Ice Lake-LP Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13545u16 , name : "Ice Lake-LP Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13546u16 , name : "Ice Lake-LP Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13547u16 , name : "Ice Lake-LP Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13549u16 , name : "Ice Lake-LP USB 3.1 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13551u16 , name : "Ice Lake-LP DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13552u16 , name : "Ice Lake-LP PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5458u16 , name : "Killer(R) Wireless-AC 1550i Wireless Network Adapter (9560NGW)" } , SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Wi-Fi 6 AX201" } , SubSystem { subvendor : 32902u16 , subdevice : 612u16 , name : "Wireless-AC 9461" }] } , Device { vendor_id : 32902u16 , id : 13560u16 , name : "Ice Lake-LP SD Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13564u16 , name : "Ice Lake-LP Integrated Sensor Solution" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13568u16 , name : "6311ESB/6321ESB PCI Express Upstream Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 13569u16 , name : "6310ESB PCI Express Upstream Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13572u16 , name : "6311ESB/6321ESB I/OxAPIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13573u16 , name : "6310ESB I/OxAPIC Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13580u16 , name : "6311ESB/6321ESB PCI Express to PCI-X Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 13581u16 , name : "6310ESB PCI Express to PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13584u16 , name : "6311ESB/6321ESB PCI Express Downstream Port E1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12798u16 , name : "ProLiant DL140 G3" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 13585u16 , name : "6310ESB PCI Express Downstream Port E1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13588u16 , name : "6311ESB/6321ESB PCI Express Downstream Port E2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13589u16 , name : "6310ESB PCI Express Downstream Port E2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13592u16 , name : "6311ESB/6321ESB PCI Express Downstream Port E3" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" }] } , Device { vendor_id : 32902u16 , id : 13593u16 , name : "6310ESB PCI Express Downstream Port E3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13610u16 , name : "PCI Express Gen5 Port A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13611u16 , name : "PCI Express Gen5 Port B" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13612u16 , name : "PCI Express Gen5 Port C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13613u16 , name : "PCI Express Gen5 Port D" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13685u16 , name : "82830M/MG/MP Host Bridge" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 48u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 541u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" }] } , Device { vendor_id : 32902u16 , id : 13686u16 , name : "82830M/MP AGP Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13687u16 , name : "82830M/MG Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1299u16 , name : "ThinkPad A/T/X Series" }] } , Device { vendor_id : 32902u16 , id : 13688u16 , name : "82830M/MG/MP Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13696u16 , name : "82852/82855 GM/GME/PM/GMV Processor to I/O Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1372u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 13697u16 , name : "82852/82855 GM/GME/PM/GMV Processor to AGP Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" }] } , Device { vendor_id : 32902u16 , id : 13698u16 , name : "82852/855GM Integrated Graphics Device" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1378u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8 integrated graphics" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer VGA" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" } , SubSystem { subvendor : 58559u16 , subdevice : 3273u16 , name : "CC9-SAMBA" } , SubSystem { subvendor : 58559u16 , subdevice : 3282u16 , name : "CD2-BEBOP" }] } , Device { vendor_id : 32902u16 , id : 13700u16 , name : "82852/82855 GM/GME/PM/GMV Processor to I/O Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1373u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 13701u16 , name : "82852/82855 GM/GME/PM/GMV Processor to I/O Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1374u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4426u16 , subdevice : 1410u16 , name : "PC8" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420" } , SubSystem { subvendor : 6005u16 , subdevice : 4304u16 , name : "V5D Single Board Computer" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4272u16 , name : "CL9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4320u16 , name : "PSL09 PrPMC" }] } , Device { vendor_id : 32902u16 , id : 13708u16 , name : "82854 GMCH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13710u16 , name : "82854 GMCH Integrated Graphics Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13712u16 , name : "E7520 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 Memory Controller Hub" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 Memory Controller Hub" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 5940u16 , subdevice : 4158u16 , name : "PRIMERGY RX/TX S2 series" } , SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" }] } , Device { vendor_id : 32902u16 , id : 13713u16 , name : "E7525/E7520 Error Reporting Registers" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" }] } , Device { vendor_id : 32902u16 , id : 13714u16 , name : "E7320 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" }] } , Device { vendor_id : 32902u16 , id : 13715u16 , name : "E7320 Error Reporting Registers" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" }] } , Device { vendor_id : 32902u16 , id : 13716u16 , name : "E7520 DMA Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" } , SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 Processor AMC" }] } , Device { vendor_id : 32902u16 , id : 13717u16 , name : "E7525/E7520/E7320 PCI Express Port A" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 13718u16 , name : "E7525/E7520/E7320 PCI Express Port A1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13719u16 , name : "E7525/E7520 PCI Express Port B" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 13720u16 , name : "E7520 PCI Express Port B1" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 13721u16 , name : "E7520 PCI Express Port C" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 32902u16 , id : 13722u16 , name : "E7520 PCI Express Port C1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13723u16 , name : "E7525/E7520/E7320 Extended Configuration Registers" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 733u16 , name : "eServer xSeries server mainboard" }] } , Device { vendor_id : 32902u16 , id : 13726u16 , name : "E7525 Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 361u16 , name : "Precision 470" }] } , Device { vendor_id : 32902u16 , id : 13744u16 , name : "3100 Chipset Memory I/O Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13745u16 , name : "3100 DRAM Controller Error Reporting Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13749u16 , name : "3100 Chipset Enhanced DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13750u16 , name : "3100 Chipset PCI Express Port A" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13751u16 , name : "3100 Chipset PCI Express Port A1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13768u16 , name : "3100 Extended Configuration Test Overflow Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13824u16 , name : "7300 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13828u16 , name : "7300 Chipset PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13829u16 , name : "7300 Chipset PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13830u16 , name : "7300 Chipset PCI Express Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13831u16 , name : "7300 Chipset PCI Express Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13832u16 , name : "7300 Chipset PCI Express Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13833u16 , name : "7300 Chipset PCI Express Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13834u16 , name : "7300 Chipset PCI Express Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13835u16 , name : "7300 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13836u16 , name : "7300 Chipset FSB Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 7300 Chipset FSB Registers" }] } , Device { vendor_id : 32902u16 , id : 13837u16 , name : "7300 Chipset Snoop Filter Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13838u16 , name : "7300 Chipset Debug and Miscellaneous Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13839u16 , name : "7300 Chipset FBD Branch 0 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 13840u16 , name : "7300 Chipset FBD Branch 1 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14080u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14081u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14082u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14083u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14084u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14085u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14086u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14087u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14088u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14089u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14090u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14091u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14092u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14093u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14094u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14095u16 , name : "Xeon C5500/C3500 DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14096u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14097u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14098u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14099u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14100u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14101u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14102u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14103u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14104u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14105u16 , name : "Xeon C5500/C3500 CB3 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14106u16 , name : "Xeon C5500/C3500 QPI Link" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14107u16 , name : "Xeon C5500/C3500 QPI Routing and Protocol" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14109u16 , name : "Xeon C5500/C3500 QPI Routing and Protocol" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14112u16 , name : "Xeon C5500/C3500 PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14113u16 , name : "Xeon C5500/C3500 PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14114u16 , name : "Xeon C5500/C3500 PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14115u16 , name : "Xeon C5500/C3500 PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14116u16 , name : "Xeon C5500/C3500 PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14117u16 , name : "Xeon C5500/C3500 NTB Primary" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14118u16 , name : "Xeon C5500/C3500 NTB Primary" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14119u16 , name : "Xeon C5500/C3500 NTB Secondary" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14120u16 , name : "Xeon C5500/C3500 Core" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14121u16 , name : "Xeon C5500/C3500 Core" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14122u16 , name : "Xeon C5500/C3500 Core" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14123u16 , name : "Xeon C5500/C3500 Core" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14124u16 , name : "Xeon C5500/C3500 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14143u16 , name : "Xeon C5500/C3500 IOxAPIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14257u16 , name : "C620 Series Chipset Family Thermal Sensor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14272u16 , name : "C620 Series Chipset Family PCIe Uplink (x16)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14273u16 , name : "C620 Series Chipset Family PCIe Uplink (x8)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14274u16 , name : "C620 Series Chipset Family Virtual Switch Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14275u16 , name : "C620 Series Chipset Family Virtual Switch Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14276u16 , name : "C620 Series Chipset Family Virtual Switch Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14277u16 , name : "C620 Series Chipset Family Virtual Switch Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14279u16 , name : "C620 Series Chipset Family Virtual Switch Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14280u16 , name : "C620 Series Chipset Family QAT" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "QuickAssist Adapter 8960" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "QuickAssist Adapter 8970" }] } , Device { vendor_id : 32902u16 , id : 14281u16 , name : "C620 Series Chipset Family QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14284u16 , name : "Ethernet Connection X722" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14285u16 , name : "Ethernet Virtual Function 700 Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14286u16 , name : "Ethernet Connection X722 for 10GbE backplane" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 533u16 , name : "Ethernet 10Gb 2-port 568i Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 16419u16 , name : "Intel Ethernet Connection X722 for 10GbE backplane" } , SubSystem { subvendor : 6058u16 , subdevice : 16421u16 , name : "Ethernet Connection X722 for 10GbE backplane" }] } , Device { vendor_id : 32902u16 , id : 14287u16 , name : "Ethernet Connection X722 for 10GbE QSFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14288u16 , name : "Ethernet Connection X722 for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 16416u16 , name : "Intel Ethernet Connection X722 for 10G SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16417u16 , name : "Intel Ethernet Connection X722 for 10G SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16418u16 , name : "Ethernet Connection X722 for 10GbE SFP+" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter X722-2" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter X722-2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter X722-4" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter X722-4" }] } , Device { vendor_id : 32902u16 , id : 14289u16 , name : "Ethernet Connection X722 for 1GbE" , subsystems : & [SubSystem { subvendor : 5325u16 , subdevice : 16u16 , name : "88E1514 Ethernet OCP 2x1G RJ45 Phy Card [USI-1514-1GbaseT]" } , SubSystem { subvendor : 5520u16 , subdevice : 534u16 , name : "Ethernet 1Gb 2-port 368i Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 535u16 , name : "Ethernet 1Gb 2-port 368FLR-MMT Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 583u16 , name : "Ethernet 1Gb 4-port 369i Adapter" } , SubSystem { subvendor : 5593u16 , subdevice : 14289u16 , name : "X11DPi-N" } , SubSystem { subvendor : 6058u16 , subdevice : 16416u16 , name : "Ethernet Connection X722 for 1GbE" } , SubSystem { subvendor : 6058u16 , subdevice : 16417u16 , name : "Ethernet Connection X722 for 1GbE" } , SubSystem { subvendor : 6058u16 , subdevice : 16418u16 , name : "Ethernet Connection X722 for 1GbE" } , SubSystem { subvendor : 6058u16 , subdevice : 16420u16 , name : "Ethernet Connection X722 for 1GbE" }] } , Device { vendor_id : 32902u16 , id : 14290u16 , name : "Ethernet Connection X722 for 10GBASE-T" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 384u16 , name : "RD10019 10GbE interface" } , SubSystem { subvendor : 4464u16 , subdevice : 14290u16 , name : "Ethernet Connection X722 for 10GBASE-T" } , SubSystem { subvendor : 5325u16 , subdevice : 48u16 , name : "Ethernet OCP 2x10G RJ45 Phy Card [USI-X557-10GbaseT]" } , SubSystem { subvendor : 5520u16 , subdevice : 536u16 , name : "Ethernet 10Gb 2-port 568FLR-MMT Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 16416u16 , name : "Ethernet Connection X722 for 10GBASE-T" } , SubSystem { subvendor : 6058u16 , subdevice : 16417u16 , name : "Ethernet Connection X722 for 10GBASE-T" } , SubSystem { subvendor : 6058u16 , subdevice : 16418u16 , name : "Ethernet Connection X722 for 10GBASE-T" } , SubSystem { subvendor : 6058u16 , subdevice : 16420u16 , name : "Ethernet Connection X722 for 10GBASE-T" } , SubSystem { subvendor : 6058u16 , subdevice : 16421u16 , name : "Ethernet Connection X722 for 10GBASE-T" }] } , Device { vendor_id : 32902u16 , id : 14291u16 , name : "Ethernet Connection X722 for 10GbE SFP+" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 537u16 , name : "Ethernet 10Gb 2-port 568FLR-MMSFP+ Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 16416u16 , name : "Ethernet Connection X722 for 10GbE SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16417u16 , name : "Ethernet Connection X722 for 10GbE SFP+" } , SubSystem { subvendor : 6058u16 , subdevice : 16421u16 , name : "Ethernet Connection X722 for 10GbE SFP+" }] } , Device { vendor_id : 32902u16 , id : 14292u16 , name : "Ethernet Connection X722 for 10GbE QSFP+" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14297u16 , name : "X722 Hyper-V Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14466u16 , name : "Ice Lake LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14500u16 , name : "Ice Lake SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14536u16 , name : "Ice Lake-LP Smart Sound Technology Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14560u16 , name : "Ice Lake Management Engine Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14848u16 , name : "82801JD/DO (ICH10 Family) 4-port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14850u16 , name : "82801JD/DO (ICH10 Family) SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14853u16 , name : "82801JD/DO (ICH10 Family) SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14854u16 , name : "82801JD/DO (ICH10 Family) 2-port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14868u16 , name : "82801JDO (ICH10DO) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14870u16 , name : "82801JIR (ICH10R) LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 LPC Interface Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 LPC Interface Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14872u16 , name : "82801JIB (ICH10) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14874u16 , name : "82801JD (ICH10D) LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14880u16 , name : "82801JI (ICH10 Family) 4 port SATA IDE Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 SATA IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 SATA IDE Controller" }] } , Device { vendor_id : 32902u16 , id : 14882u16 , name : "82801JI (ICH10 Family) SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 45061u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14885u16 , name : "82801JIR (ICH10R) SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PERC S100 Controller (PE R410)" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PERC S100 Controller (PE T410)" } , SubSystem { subvendor : 4136u16 , subdevice : 753u16 , name : "PERC S100 Controller (PE R510)" }] } , Device { vendor_id : 32902u16 , id : 14886u16 , name : "82801JI (ICH10 Family) 2 port SATA IDE Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 SATA IDE Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 SATA IDE Controller" }] } , Device { vendor_id : 32902u16 , id : 14896u16 , name : "82801JI (ICH10 Family) SMBus Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14898u16 , name : "82801JI (ICH10 Family) Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14900u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14901u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14902u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14903u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14904u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14905u16 , name : "82801JI (ICH10 Family) USB UHCI Controller #6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB UHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB UHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14906u16 , name : "82801JI (ICH10 Family) USB2 EHCI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB EHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20486u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14908u16 , name : "82801JI (ICH10 Family) USB2 EHCI Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 USB EHCI Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 USB EHCI Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant G6 series" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20486u16 , name : "Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14910u16 , name : "82801JI (ICH10 Family) HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33553u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 40962u16 , name : "GA-EP45-UD3R Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 41218u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14912u16 , name : "82801JI (ICH10 Family) PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 PCI Express Port 1" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 PCI Express Port 1" } , SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P6T DeLuxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14914u16 , name : "82801JI (ICH10 Family) PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14916u16 , name : "82801JI (ICH10 Family) PCI Express Root Port 3" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P6T DeLuxe Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14918u16 , name : "82801JI (ICH10 Family) PCI Express Root Port 4" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P6T DeLuxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14920u16 , name : "82801JI (ICH10 Family) PCI Express Root Port 5" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P6T Deluxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5 Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14922u16 , name : "82801JI (ICH10 Family) PCI Express Root Port 6" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13067u16 , name : "ProLiant ML150 G6 Server" } , SubSystem { subvendor : 4163u16 , subdevice : 33492u16 , name : "P5Q Deluxe Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33514u16 , name : "P6T DeLuxe Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-EP45-DS5/GA-EG45M-DS2H Motherboard" }] } , Device { vendor_id : 32902u16 , id : 14924u16 , name : "82801JI (ICH10 Family) Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14929u16 , name : "82801JDO (ICH10DO) VECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14933u16 , name : "82801JD/DO (ICH10 Family) Virtual SATA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14944u16 , name : "82801JD/DO (ICH10 Family) SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14946u16 , name : "82801JD/DO (ICH10 Family) Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14948u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14949u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14950u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14951u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14952u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14953u16 , name : "82801JD/DO (ICH10 Family) USB UHCI Controller #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14954u16 , name : "82801JD/DO (ICH10 Family) USB2 EHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14956u16 , name : "82801JD/DO (ICH10 Family) USB2 EHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14958u16 , name : "82801JD/DO (ICH10 Family) HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14960u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14962u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14964u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14966u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14968u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14970u16 , name : "82801JD/DO (ICH10 Family) PCI Express Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 14972u16 , name : "82801JD/DO (ICH10 Family) Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15104u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15105u16 , name : "Mobile 5 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15106u16 , name : "P55 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15107u16 , name : "PM55 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15108u16 , name : "5 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15109u16 , name : "Mobile 5 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15110u16 , name : "H55 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15111u16 , name : "QM57 Chipset LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15112u16 , name : "H57 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15113u16 , name : "HM55 Chipset LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15114u16 , name : "Q57 Chipset LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" }] } , Device { vendor_id : 32902u16 , id : 15115u16 , name : "HM57 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15116u16 , name : "5 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15117u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15118u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15119u16 , name : "QS57 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15120u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15121u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15122u16 , name : "3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15123u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15124u16 , name : "3420 Chipset LPC Interface Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" }] } , Device { vendor_id : 32902u16 , id : 15125u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15126u16 , name : "3450 Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15127u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15128u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15129u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15130u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15131u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15132u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15133u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15134u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15135u16 , name : "5 Series/3400 Series Chipset LPC Interface Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15136u16 , name : "5 Series/3400 Series Chipset 4 port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15137u16 , name : "5 Series/3400 Series Chipset 2 port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15138u16 , name : "5 Series/3400 Series Chipset 6 port SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" }] } , Device { vendor_id : 32902u16 , id : 15139u16 , name : "5 Series/3400 Series Chipset 4 port SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15141u16 , name : "5 Series/3400 Series Chipset SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12568u16 , name : "Smart Array B110i SATA RAID Controller" }] } , Device { vendor_id : 32902u16 , id : 15142u16 , name : "5 Series/3400 Series Chipset 2 port SATA IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15144u16 , name : "5 Series/3400 Series Chipset 4 port SATA IDE Controller" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15145u16 , name : "5 Series/3400 Series Chipset 4 port SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15148u16 , name : "5 Series/3400 Series Chipset SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15149u16 , name : "5 Series/3400 Series Chipset 2 port SATA IDE Controller" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15150u16 , name : "5 Series/3400 Series Chipset 4 port SATA IDE Controller" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15151u16 , name : "5 Series/3400 Series Chipset 6 port SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15152u16 , name : "5 Series/3400 Series Chipset SMBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4163u16 , subdevice : 14392u16 , name : "P7P55-M Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33667u16 , name : "P7P55-M Motherboard" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15154u16 , name : "5 Series/3400 Series Chipset Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15156u16 , name : "5 Series/3400 Series Chipset USB2 Enhanced Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15158u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15159u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15160u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15161u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15162u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15163u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15164u16 , name : "5 Series/3400 Series Chipset USB2 Enhanced Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15166u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15167u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15168u16 , name : "5 Series/3400 Series Chipset USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15169u16 , name : "5 Series/3400 Series Chipset LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15170u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4156u16 , subdevice : 5409u16 , name : "EliteBook 8540p" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15172u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15174u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 3" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15176u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" }] } , Device { vendor_id : 32902u16 , id : 15178u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 32902u16 , id : 15180u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15182u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15184u16 , name : "5 Series/3400 Series Chipset PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15187u16 , name : "5 Series/3400 Series Chipset VECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15190u16 , name : "5 Series/3400 Series Chipset High Definition Audio" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4163u16 , subdevice : 4979u16 , name : "G73-series gaming laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15191u16 , name : "5 Series/3400 Series Chipset High Definition Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15204u16 , name : "5 Series/3400 Series Chipset HECI Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" } , SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15205u16 , name : "5 Series/3400 Series Chipset HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15206u16 , name : "5 Series/3400 Series Chipset PT IDER Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15207u16 , name : "5 Series/3400 Series Chipset KT Controller" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 20673u16 , name : "PC1-GROOVE" }] } , Device { vendor_id : 32902u16 , id : 15360u16 , name : "Xeon E5/Core i7 DMI2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15361u16 , name : "Xeon E5/Core i7 DMI2 in PCI Express Mode" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15362u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 1a" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15363u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 1b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15364u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 2a" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15365u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 2b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15366u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 2c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15367u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 2d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15368u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 3a in PCI Express Mode" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15369u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 3b" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15370u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 3c" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15371u16 , name : "Xeon E5/Core i7 IIO PCI Express Root Port 3d" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15373u16 , name : "Xeon E5/Core i7 Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15374u16 , name : "Xeon E5/Core i7 Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15375u16 , name : "Xeon E5/Core i7 Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15392u16 , name : "Xeon E5/Core i7 DMA Channel 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15393u16 , name : "Xeon E5/Core i7 DMA Channel 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15394u16 , name : "Xeon E5/Core i7 DMA Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15395u16 , name : "Xeon E5/Core i7 DMA Channel 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15396u16 , name : "Xeon E5/Core i7 DMA Channel 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15397u16 , name : "Xeon E5/Core i7 DMA Channel 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15398u16 , name : "Xeon E5/Core i7 DMA Channel 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15399u16 , name : "Xeon E5/Core i7 DMA Channel 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15400u16 , name : "Xeon E5/Core i7 Address Map, VTd_Misc, System Management" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15402u16 , name : "Xeon E5/Core i7 Control Status and Global Errors" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15404u16 , name : "Xeon E5/Core i7 I/O APIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15406u16 , name : "Xeon E5/Core i7 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15407u16 , name : "Xeon E5/Core i7 DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15424u16 , name : "Xeon E5/Core i7 IIO Switch and IRP Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15425u16 , name : "Sandy Bridge QPI Port 0 Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15426u16 , name : "Sandy Bridge QPI Port 1 Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15427u16 , name : "Xeon E5/Core i7 Ring to PCI Express Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15428u16 , name : "Xeon E5/Core i7 Ring to QuickPath Interconnect Link 0 Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15429u16 , name : "Xeon E5/Core i7 Ring to QuickPath Interconnect Link 1 Performance Monitor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15430u16 , name : "Xeon E5/Core i7 Processor Home Agent Performance Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15473u16 , name : "Xeon E5/Core i7 Integrated Memory Controller RAS Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15488u16 , name : "Xeon E5/Core i7 QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15491u16 , name : "Xeon E5/Core i7 QPI Link Reut 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15492u16 , name : "Xeon E5/Core i7 QPI Link Reut 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15494u16 , name : "Sandy Bridge QPI Port 0 DFX Link" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15504u16 , name : "Xeon E5/Core i7 QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15507u16 , name : "Xeon E5/Core i7 QPI Link Reut 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15508u16 , name : "Xeon E5/Core i7 QPI Link Reut 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15510u16 , name : "Sandy Bridge QPI Port 1 DFX Link" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15520u16 , name : "Xeon E5/Core i7 Processor Home Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15528u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15530u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Target Address Decoder 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15531u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Target Address Decoder 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15532u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Target Address Decoder 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15533u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Target Address Decoder 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15534u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Target Address Decoder 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15536u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Channel 0-3 Thermal Control 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15537u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Channel 0-3 Thermal Control 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15538u16 , name : "Xeon E5/Core i7 Integrated Memory Controller ERROR Registers 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15539u16 , name : "Xeon E5/Core i7 Integrated Memory Controller ERROR Registers 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15540u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Channel 0-3 Thermal Control 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15541u16 , name : "Xeon E5/Core i7 Integrated Memory Controller Channel 0-3 Thermal Control 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15542u16 , name : "Xeon E5/Core i7 Integrated Memory Controller ERROR Registers 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15543u16 , name : "Xeon E5/Core i7 Integrated Memory Controller ERROR Registers 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15544u16 , name : "Xeon E5/Core i7 DDRIO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15552u16 , name : "Xeon E5/Core i7 Power Control Unit 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15553u16 , name : "Xeon E5/Core i7 Power Control Unit 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15554u16 , name : "Xeon E5/Core i7 Power Control Unit 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15568u16 , name : "Xeon E5/Core i7 Power Control Unit 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15584u16 , name : "Xeon E5/Core i7 Interrupt Control Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15587u16 , name : "Xeon E5/Core i7 Semaphore and Scratchpad Configuration Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15588u16 , name : "Xeon E5/Core i7 R2PCIe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15590u16 , name : "Xeon E5/Core i7 QuickPath Interconnect Agent Ring Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15592u16 , name : "Xeon E5/Core i7 Unicast Register 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15593u16 , name : "Xeon E5/Core i7 Unicast Register 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15594u16 , name : "Xeon E5/Core i7 Unicast Register 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15595u16 , name : "Xeon E5/Core i7 Unicast Register 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15596u16 , name : "Xeon E5/Core i7 Unicast Register 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15597u16 , name : "Xeon E5/Core i7 Unicast Register 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15598u16 , name : "Xeon E5/Core i7 Unicast Register 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15599u16 , name : "Xeon E5/Core i7 Unicast Register 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15604u16 , name : "Xeon E5/Core i7 Integrated Memory Controller System Address Decoder 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15605u16 , name : "Xeon E5/Core i7 Integrated Memory Controller System Address Decoder 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15606u16 , name : "Xeon E5/Core i7 System Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15888u16 , name : "8th Gen Core 4-core Processor Host Bridge/DRAM Registers [Coffee Lake H]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15896u16 , name : "8th Gen Core 4-core Workstation Processor Host Bridge/DRAM Registers [Coffee Lake S]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15903u16 , name : "8th Gen Core 4-core Desktop Processor Host Bridge/DRAM Registers [Coffee Lake S]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "Z370 AORUS Gaming K3-CF" }] } , Device { vendor_id : 32902u16 , id : 15904u16 , name : "i9 Core / Xeon E 2200M Coffee Lake Host Bridge / DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15920u16 , name : "8th/9th Gen Core 8-core Desktop Processor Host Bridge/DRAM Registers [Coffee Lake S]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15923u16 , name : "8th/9th Gen Core Processor Host Bridge/DRAM Registers [Coffee Lake]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15924u16 , name : "Coffee Lake HOST and DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 15925u16 , name : "Coffee Lake Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16001u16 , name : "8th Gen Core Processor PCIe Controller (x16)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16005u16 , name : "8th Gen Core Processor PCIe Controller (x8)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16009u16 , name : "8th Gen Core Processor PCIe Controller (x4)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16016u16 , name : "CoffeeLake-S GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16017u16 , name : "CoffeeLake-S GT2 [UHD Graphics 630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16018u16 , name : "CoffeeLake-S GT2 [UHD Graphics 630]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 16019u16 , name : "CoffeeLake-S GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16020u16 , name : "Coffee Lake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16022u16 , name : "CoffeeLake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16024u16 , name : "CoffeeLake-S GT2 [UHD Graphics 630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16026u16 , name : "Coffee Lake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16027u16 , name : "CoffeeLake-H GT2 [UHD Graphics 630]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 412u16 , name : "MacBookPro16,1 (16\", 2019)" }] } , Device { vendor_id : 32902u16 , id : 16028u16 , name : "Coffee Lake-S GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16032u16 , name : "WhiskeyLake-U GT2 [UHD Graphics 620]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 16033u16 , name : "Whiskey Lake-U GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16037u16 , name : "CoffeeLake-U GT3e [Iris Plus Graphics 655]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16038u16 , name : "Coffee Lake-U GT3 [Iris Plus Graphics 645]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16040u16 , name : "Coffee Lake-U GT3 [Iris Plus Graphics 655]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16041u16 , name : "Coffee Lake-U GT2 [UHD Graphics 620]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16066u16 , name : "8th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" } , SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "PRIME H310M-D" }] } , Device { vendor_id : 32902u16 , id : 16068u16 , name : "8th/9th Gen Core Processor Host Bridge / DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16070u16 , name : "8th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16074u16 , name : "8th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16080u16 , name : "8th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16384u16 , name : "5400 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16385u16 , name : "5400 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16387u16 , name : "5400 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16417u16 , name : "5400 Chipset PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16418u16 , name : "5400 Chipset PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16419u16 , name : "5400 Chipset PCI Express Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16420u16 , name : "5400 Chipset PCI Express Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16421u16 , name : "5400 Chipset PCI Express Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16422u16 , name : "5400 Chipset PCI Express Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16423u16 , name : "5400 Chipset PCI Express Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16424u16 , name : "5400 Chipset PCI Express Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16425u16 , name : "5400 Chipset PCI Express Port 9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16429u16 , name : "5400 Chipset IBIST Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16430u16 , name : "5400 Chipset IBIST Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16431u16 , name : "5400 Chipset QuickData Technology Device" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16432u16 , name : "5400 Chipset FSB Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16433u16 , name : "5400 Chipset CE/SF Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16434u16 , name : "5400 Chipset IOxAPIC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16437u16 , name : "5400 Chipset FBD Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16438u16 , name : "5400 Chipset FBD Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16449u16 , name : "NVMe Datacenter SSD [Optane]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16640u16 , name : "Moorestown Graphics and Video" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16648u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16649u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16650u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16651u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16652u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16653u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16654u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16655u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16660u16 , name : "Atom Processor E6xx PCI Host Bridge #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16661u16 , name : "Atom Processor E6xx PCI Host Bridge #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16662u16 , name : "Atom Processor E6xx PCI Host Bridge #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16663u16 , name : "Atom Processor E6xx PCI Host Bridge #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16704u16 , name : "NVMe Datacenter SSD [Optane]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8500u16 , name : "NVMe Datacenter SSD [Optane] SED 400GB 2.5\" U.2 (P5800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8501u16 , name : "NVMe Datacenter SSD [Optane] SED 800GB 2.5\" U.2 (P5800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8502u16 , name : "NVMe Datacenter SSD [Optane] SED 1.6TB 2.5\" U.2 (P5800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8503u16 , name : "NVMe Datacenter SSD [Optane] 400GB 2.5\" U.2 (P5800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8504u16 , name : "NVMe Datacenter SSD [Optane] 800GB 2.5\" U.2 (P5800X)" } , SubSystem { subvendor : 4136u16 , subdevice : 8505u16 , name : "NVMe Datacenter SSD [Optane] 1.6TB 2.5\" U.2 (P5800X)" }] } , Device { vendor_id : 32902u16 , id : 16928u16 , name : "PRO/Wireless 2200BG [Calexico2] Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 4854u16 , name : "nc6120/nc6220/nw8240/nx8220" } , SubSystem { subvendor : 32902u16 , subdevice : 9985u16 , name : "WM3B2200BG Mini-PCI Card" } , SubSystem { subvendor : 32902u16 , subdevice : 10002u16 , name : "IBM ThinkPad R50e" } , SubSystem { subvendor : 32902u16 , subdevice : 10017u16 , name : "Dell B130 laptop integrated WLAN" } , SubSystem { subvendor : 32902u16 , subdevice : 10018u16 , name : "Dell Latitude D600" } , SubSystem { subvendor : 32902u16 , subdevice : 10033u16 , name : "Samsung P35 integrated WLAN" }] } , Device { vendor_id : 32902u16 , id : 16930u16 , name : "PRO/Wireless 3945ABG [Golan] Network Connection" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4956u16 , name : "PRO/Wireless 3945ABG [Golan] Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4096u16 , name : "PRO/Wireless 3945ABG Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4097u16 , name : "WM3945ABG MOW2" } , SubSystem { subvendor : 32902u16 , subdevice : 4101u16 , name : "PRO/Wireless 3945BG Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4148u16 , name : "PRO/Wireless 3945BG Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 4164u16 , name : "PRO/Wireless 3945BG Network Connection" } , SubSystem { subvendor : 32902u16 , subdevice : 7168u16 , name : "PRO/Wireless 3945ABG Network Connection" }] } , Device { vendor_id : 32902u16 , id : 16931u16 , name : "PRO/Wireless 2915ABG [Calexico2] Network Connection" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 32902u16 , name : "mPCI 3B Americas/Europe ZZA" } , SubSystem { subvendor : 4097u16 , subdevice : 32902u16 , name : "mPCI 3B Europe ZZE" } , SubSystem { subvendor : 4098u16 , subdevice : 32902u16 , name : "mPCI 3B Japan ZZJ" } , SubSystem { subvendor : 4099u16 , subdevice : 32902u16 , name : "mPCI 3B High-Band ZZH" } , SubSystem { subvendor : 4156u16 , subdevice : 4945u16 , name : "Compaq nc6220" }] } , Device { vendor_id : 32902u16 , id : 16932u16 , name : "PRO/Wireless 2915ABG [Calexico2] Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16935u16 , name : "PRO/Wireless 3945ABG [Golan] Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4112u16 , name : "ThinkPad R60e" } , SubSystem { subvendor : 32902u16 , subdevice : 4113u16 , name : "ThinkPad T60/R60e/X60s/R61" } , SubSystem { subvendor : 32902u16 , subdevice : 4116u16 , name : "PRO/Wireless 3945BG Network Connection" }] } , Device { vendor_id : 32902u16 , id : 16937u16 , name : "PRO/Wireless 4965 AG or AGN [Kedron] Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4352u16 , name : "Vaio VGN-SZ79SN_C" } , SubSystem { subvendor : 32902u16 , subdevice : 4353u16 , name : "PRO/Wireless 4965 AG or AGN" }] } , Device { vendor_id : 32902u16 , id : 16939u16 , name : "Centrino Ultimate-N 6300" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4353u16 , name : "Centrino Ultimate-N 6300 3x3 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4385u16 , name : "Centrino Ultimate-N 6300 3x3 AGN" }] } , Device { vendor_id : 32902u16 , id : 16940u16 , name : "Centrino Advanced-N 6200" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4865u16 , name : "Centrino Advanced-N 6200 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "Centrino Advanced-N 6200 2x2 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4871u16 , name : "Centrino Advanced-N 6200 2x2 BG" } , SubSystem { subvendor : 32902u16 , subdevice : 4897u16 , name : "Centrino Advanced-N 6200 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4902u16 , name : "Centrino Advanced-N 6200 2x2 ABG" }] } , Device { vendor_id : 32902u16 , id : 16944u16 , name : "PRO/Wireless 4965 AG or AGN [Kedron] Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4368u16 , name : "Lenovo ThinkPad T51" } , SubSystem { subvendor : 32902u16 , subdevice : 4369u16 , name : "Lenovo ThinkPad T61" }] } , Device { vendor_id : 32902u16 , id : 16946u16 , name : "WiFi Link 5100" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4609u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4612u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4613u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4614u16 , name : "WiFi Link 5100 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4641u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4644u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4645u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4646u16 , name : "WiFi Link 5100 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4865u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4868u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4869u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "WiFi Link 5100 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4897u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4900u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4901u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4902u16 , name : "WiFi Link 5100 ABG" }] } , Device { vendor_id : 32902u16 , id : 16949u16 , name : "Ultimate N WiFi Link 5300" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16950u16 , name : "Ultimate N WiFi Link 5300" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16951u16 , name : "PRO/Wireless 5100 AGN [Shiloh] Network Connection" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4625u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4628u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4629u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4630u16 , name : "WiFi Link 5100 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4881u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4884u16 , name : "WiFi Link 5100 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4885u16 , name : "WiFi Link 5100 BGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "WiFi Link 5100 ABG" }] } , Device { vendor_id : 32902u16 , id : 16952u16 , name : "Centrino Ultimate-N 6300" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4369u16 , name : "Centrino Ultimate-N 6300 3x3 AGN" }] } , Device { vendor_id : 32902u16 , id : 16953u16 , name : "Centrino Advanced-N 6200" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4881u16 , name : "Centrino Advanced-N 6200 2x2 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "Centrino Advanced-N 6200 2x2 ABG" }] } , Device { vendor_id : 32902u16 , id : 16954u16 , name : "PRO/Wireless 5350 AGN [Echo Peak] Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16955u16 , name : "PRO/Wireless 5350 AGN [Echo Peak] Network Connection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 16956u16 , name : "WiMAX/WiFi Link 5150" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4609u16 , name : "WiMAX/WiFi Link 5150 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4614u16 , name : "WiMAX/WiFi Link 5150 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4641u16 , name : "WiMAX/WiFi Link 5150 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4865u16 , name : "WiMAX/WiFi Link 5150 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4870u16 , name : "WiMAX/WiFi Link 5150 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4897u16 , name : "WiMAX/WiFi Link 5150 AGN" }] } , Device { vendor_id : 32902u16 , id : 16957u16 , name : "WiMAX/WiFi Link 5150" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 4625u16 , name : "WiMAX/WiFi Link 5150 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4630u16 , name : "WiMAX/WiFi Link 5150 ABG" } , SubSystem { subvendor : 32902u16 , subdevice : 4881u16 , name : "WiMAX/WiFi Link 5150 AGN" } , SubSystem { subvendor : 32902u16 , subdevice : 4886u16 , name : "WiMAX/WiFi Link 5150 ABG" }] } , Device { vendor_id : 32902u16 , id : 17284u16 , name : "Q570 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17285u16 , name : "Z590 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17286u16 , name : "H570 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17287u16 , name : "B560 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17288u16 , name : "H510 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17289u16 , name : "WM590 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17290u16 , name : "QM580 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17291u16 , name : "HM570 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17292u16 , name : "C252 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17293u16 , name : "C256 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17294u16 , name : "H310D LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17295u16 , name : "W580 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17296u16 , name : "RM590E LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17297u16 , name : "R580E LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17312u16 , name : "500 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17313u16 , name : "500 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17315u16 , name : "500 Series Chipset Family SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17316u16 , name : "500 Series Chipset Family SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17318u16 , name : "500 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17319u16 , name : "500 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17320u16 , name : "500 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17321u16 , name : "500 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17322u16 , name : "500 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17323u16 , name : "500 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17325u16 , name : "500 Series Chipset Family I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17326u16 , name : "500 Series Chipset Family I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17328u16 , name : "500 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17329u16 , name : "500 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17330u16 , name : "500 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17331u16 , name : "500 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17332u16 , name : "500 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17333u16 , name : "500 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17334u16 , name : "500 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17335u16 , name : "500 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17336u16 , name : "500 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17337u16 , name : "500 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17338u16 , name : "500 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17339u16 , name : "500 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17340u16 , name : "500 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17341u16 , name : "500 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17342u16 , name : "500 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17343u16 , name : "500 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17344u16 , name : "500 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17345u16 , name : "500 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17346u16 , name : "500 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17347u16 , name : "500 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17348u16 , name : "500 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17349u16 , name : "500 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17350u16 , name : "500 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17351u16 , name : "500 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17352u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17353u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17354u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17355u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17356u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17357u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17358u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17359u16 , name : "500 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17360u16 , name : "500 Series Chipset Family Touch Host Controller (THC) #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17361u16 , name : "500 Series Chipset Family Touch Host Controller (THC) #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17362u16 , name : "500 Series Chipset Family SATA Controller (AHCI) (Server/Desktop)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17363u16 , name : "500 Series Chipset Family SATA Controller (AHCI) (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17364u16 , name : "500 Series Chipset Family SATA Controller (RAID 0/1/5/10) no premium (Desktop)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17365u16 , name : "500 Series Chipset Family SATA Controller (RAID 0/1/5/10) no premium (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17366u16 , name : "500 Series Chipset Family SATA Controller (RAID 0/1/5/10) premium (Server/Desktop)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17367u16 , name : "500 Series Chipset Family SATA Controller (RAID 0/1/5/10) premium (Mobile)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17368u16 , name : "500 Series Chipset Family I2C #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17370u16 , name : "500 Series Chipset Family UART #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17376u16 , name : "500 Series Chipset Family CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17377u16 , name : "500 Series Chipset Family CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17378u16 , name : "500 Series Chipset Family CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17379u16 , name : "500 Series Chipset Family CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17380u16 , name : "500 Series Chipset Family CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17381u16 , name : "500 Series Chipset Family CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17384u16 , name : "500 Series Chipset Family I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17385u16 , name : "500 Series Chipset Family I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17386u16 , name : "500 Series Chipset Family I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17387u16 , name : "500 Series Chipset Family I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17389u16 , name : "500 Series Chipset Family USB 3.2 Gen 2x2 (20 Gbs) xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17390u16 , name : "500 Series Chipset Family USB 3.2 Gen 1x1 (5 Gbs) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17391u16 , name : "500 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17392u16 , name : "500 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Wireless-AC 9560" } , SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Wi-Fi 6 AX201 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 612u16 , name : "Wireless-AC 9461" } , SubSystem { subvendor : 32902u16 , subdevice : 676u16 , name : "Wireless-AC 9462" }] } , Device { vendor_id : 32902u16 , id : 17393u16 , name : "500 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17394u16 , name : "500 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17395u16 , name : "500 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17397u16 , name : "500 Series Chipset Family CNVi Bluetooth" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17398u16 , name : "500 Series Chipset Family CNVi Bluetooth" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17399u16 , name : "500 Series Chipset Family CNVi Bluetooth" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17403u16 , name : "500 Series Chipset Family GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17404u16 , name : "500 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17405u16 , name : "500 Series Chipset Family GSPI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17486u16 , name : "Turbo Memory Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17681u16 , name : "Elkhart Lake Gaussian and Neural Accelerator" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17720u16 , name : "Elkhart Lake PCI-e Root Complex" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17749u16 , name : "Elkhart Lake [UHD Graphics Gen11 16EU]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17777u16 , name : "Elkhart Lake [UHD Graphics Gen11 32EU]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17921u16 , name : "Alder Lake-U15 Host and DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17922u16 , name : "Alder Lake Host and DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17933u16 , name : "12th Gen Core Processor PCI Express x16 Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17945u16 , name : "Core i9/i7/i5/i3 Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17947u16 , name : "Alder Lake-N Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17948u16 , name : "Alder Lake-N Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17949u16 , name : "Alder Lake Innovation Platform Framework Processor Participant" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 17950u16 , name : "Alder Lake-P Thunderbolt 4 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 17951u16 , name : "Alder Lake-P Thunderbolt 4 PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17958u16 , name : "Alder Lake-P Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17960u16 , name : "Alder Lake-UP3 GT2 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17961u16 , name : "12th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17962u16 , name : "AlderLake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17967u16 , name : "Alder Lake-P Thunderbolt 4 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17974u16 , name : "AlderLake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17976u16 , name : "AlderLake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17978u16 , name : "AlderLake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17981u16 , name : "12th Gen Core Processor PCI Express x4 Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17982u16 , name : "Alder Lake-P Thunderbolt 4 NHI #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 17983u16 , name : "Alder Lake-P Thunderbolt 4 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17985u16 , name : "12th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 17997u16 , name : "12th Gen Core Processor PCI Express x4 Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17998u16 , name : "Alder Lake-N Processor USB 3.2 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 17999u16 , name : "12th Gen Core Processor Gaussian & Neural Accelerator" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 18000u16 , name : "12th Gen Core Processor Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18013u16 , name : "Alder Lake Imaging Signal Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18016u16 , name : "12th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18024u16 , name : "12th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18029u16 , name : "Alder Lake-P Thunderbolt 4 NHI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18030u16 , name : "Alder Lake-P Thunderbolt 4 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18045u16 , name : "Platform Monitoring Technology" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18046u16 , name : "GNA Scoring Accelerator" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18047u16 , name : "Volume Management Device NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18048u16 , name : "Alder Lake-S GT1 [UHD Graphics 770]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18050u16 , name : "Alder Lake-S GT1 [UHD Graphics 730]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18056u16 , name : "Alder Lake-HX GT1 [UHD Graphics 770]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18058u16 , name : "Alder Lake-S [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18059u16 , name : "Alder Lake-S [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18064u16 , name : "Alder Lake-S GT1 [UHD Graphics 770]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18066u16 , name : "Alder Lake-S GT1 [UHD Graphics 730]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18067u16 , name : "Alder Lake-S GT1 [UHD Graphics 710]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18080u16 , name : "AlderLake-P GT2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18081u16 , name : "UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18083u16 , name : "Alder Lake-P GT1 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18086u16 , name : "Alder Lake-P GT2 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18088u16 , name : "Alder Lake-UP3 GT2 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18090u16 , name : "Alder Lake-UP4 GT2 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18096u16 , name : "AlderLake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18097u16 , name : "AlderLake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18099u16 , name : "Alder Lake-UP3 GT1 [UHD Graphics]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 5661u16 , name : "N22C6 [Extensa 15 EX215-55]" }] } , Device { vendor_id : 32902u16 , id : 18102u16 , name : "AlderLake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18104u16 , name : "AlderLake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18106u16 , name : "AlderLake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18112u16 , name : "AlderLake-M GT1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18113u16 , name : "AlderLake-M [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18115u16 , name : "Alder Lake-UP4 GT1 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18128u16 , name : "Alder Lake-N [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18129u16 , name : "Alder Lake-N [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18130u16 , name : "Alder Lake-N [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18131u16 , name : "Alder Lake-N [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18132u16 , name : "Alder Lake-N [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18693u16 , name : "DG1 [Iris Xe MAX Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18694u16 , name : "DG1 [Iris Xe Pod]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18695u16 , name : "SG1 [Server GPU SG-18M]" , subsystems : & [SubSystem { subvendor : 6461u16 , subdevice : 16384u16 , name : "UN-GPU-XG310-32GB-FHFL" }] } , Device { vendor_id : 32902u16 , id : 18696u16 , name : "DG1 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18697u16 , name : "DG1 [Iris Xe MAX 100]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18752u16 , name : "4xxx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18753u16 , name : "4xxx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18754u16 , name : "401xx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18755u16 , name : "401xx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18756u16 , name : "402xx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18757u16 , name : "402xx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18758u16 , name : "420xx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18759u16 , name : "420xx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18760u16 , name : "6xxx Series QAT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 18761u16 , name : "6xxx Series QAT Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19200u16 , name : "Elkhart Lake eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19235u16 , name : "Elkhart Lake SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19236u16 , name : "Elkhart Lake SPI (Flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19256u16 , name : "Elkhart Lake PCH PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19257u16 , name : "Elkhart Lake PCH PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19260u16 , name : "Elkhart Lake PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19262u16 , name : "Elkhart Lake PCH PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19275u16 , name : "Elkhart Lake Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19277u16 , name : "Elkhart Lake Serial IO UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19288u16 , name : "Elkhart Lake High Density Audio bus interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19299u16 , name : "Elkhart Lake SATA AHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19312u16 , name : "Elkhart Lake Management Engine Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19325u16 , name : "Elkhart Lake USB 3.10 XHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19327u16 , name : "Elkhart Lake PMC SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19393u16 , name : "Elkhart Lake Controller Area Network (CAN) interface #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19394u16 , name : "Elkhart Lake Controller Area Network (CAN) interface #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19517u16 , name : "Volume Management Device NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19594u16 , name : "RocketLake-S GT1 [UHD Graphics 750]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19595u16 , name : "RocketLake-S GT1 [UHD Graphics 730]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19600u16 , name : "RocketLake-S GT1 [UHD Graphics P750]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19610u16 , name : "RocketLake-S [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19847u16 , name : "Jasper Lake eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19875u16 , name : "Jasper Lake SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19876u16 , name : "Jasper Lake SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19880u16 , name : "Jasper Lake Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19883u16 , name : "Jasper Lake Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19896u16 , name : "Jasper Lake PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19897u16 , name : "Jasper Lake PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19900u16 , name : "Jasper Lake PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19902u16 , name : "Jasper Lake PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19908u16 , name : "Jasper Lake eMMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19909u16 , name : "Jasper Lake Serial IO I2C Host Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19910u16 , name : "Jasper Lake Serial IO I2C Host Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19912u16 , name : "Jasper Lake HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19923u16 , name : "Jasper Lake SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19936u16 , name : "Management Engine Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19944u16 , name : "Jasper Lake Serial IO I2C Host Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19945u16 , name : "Jasper Lake Serial IO I2C Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19946u16 , name : "Jasper Lake Serial IO I2C Host Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19947u16 , name : "Jasper Lake Serial IO I2C Host Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19949u16 , name : "Jasper Lake USB 3.1 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19951u16 , name : "Jasper Lake Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19952u16 , name : "Jasper Lake PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16432u16 , name : "Wireless-AC 9560 160MHz" }] } , Device { vendor_id : 32902u16 , id : 19960u16 , name : "Jasper Lake SD Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19971u16 , name : "Dynamic Tuning service" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19985u16 , name : "Gaussian Mixture Model and Neural Network Accelerator (GNA)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 19993u16 , name : "JasperLake IPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20002u16 , name : "Processor Transaction Router SKU 2 Core" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20053u16 , name : "JasperLake [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20065u16 , name : "JasperLake [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20081u16 , name : "JasperLake [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20352u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20353u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20354u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20355u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20356u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20357u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20358u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20359u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20360u16 , name : "DG2 [Intel Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20361u16 , name : "ACMP [Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20364u16 , name : "ACMP [Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20368u16 , name : "DG2 Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20369u16 , name : "DG2 Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20370u16 , name : "DG2 Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20481u16 , name : "LE80578" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20482u16 , name : "LE80578 Graphics Processor Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20489u16 , name : "LE80578 Video Display Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20493u16 , name : "LE80578 Expansion Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20494u16 , name : "LE80578 UART Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20495u16 , name : "LE80578 General Purpose IO" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20496u16 , name : "LE80578 I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20498u16 , name : "LE80578 Serial Peripheral Interface Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20512u16 , name : "EP80579 Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20513u16 , name : "EP80579 DRAM Error Reporting Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20515u16 , name : "EP80579 EDMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20516u16 , name : "EP80579 PCI Express Port PEA0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20517u16 , name : "EP80579 PCI Express Port PEA1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20520u16 , name : "EP80579 S-ATA IDE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20521u16 , name : "EP80579 S-ATA AHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20522u16 , name : "EP80579 S-ATA Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20523u16 , name : "EP80579 S-ATA Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20524u16 , name : "EP80579 Integrated Processor ASU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20525u16 , name : "EP80579 Integrated Processor with QuickAssist ASU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20526u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20527u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20528u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20529u16 , name : "EP80579 LPC Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20530u16 , name : "EP80579 SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20531u16 , name : "EP80579 USB 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20533u16 , name : "EP80579 USB 2.0 Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20535u16 , name : "EP80579 PCI-PCI Bridge (transparent mode)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20537u16 , name : "EP80579 Controller Area Network (CAN) interface #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20538u16 , name : "EP80579 Controller Area Network (CAN) interface #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20539u16 , name : "EP80579 Synchronous Serial Port (SPP)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20540u16 , name : "EP80579 IEEE 1588 Hardware Assist" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20541u16 , name : "EP80579 Local Expansion Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20542u16 , name : "EP80579 Global Control Unit (GCU)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20543u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20544u16 , name : "EP80579 Integrated Processor Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20545u16 , name : "EP80579 Integrated Processor with QuickAssist Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20546u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20547u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20548u16 , name : "EP80579 Integrated Processor Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20549u16 , name : "EP80579 Integrated Processor with QuickAssist Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20550u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20551u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20552u16 , name : "EP80579 Integrated Processor Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20553u16 , name : "EP80579 Integrated Processor with QuickAssist Gigabit Ethernet MAC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20554u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20555u16 , name : "EP80579 Reserved" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20556u16 , name : "EP80579 Integrated Processor with QuickAssist TDM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20865u16 , name : "Alder Lake PCH-P LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20866u16 , name : "Alder Lake PCH eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 20871u16 , name : "Alder Lake LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20893u16 , name : "Raptor Lake LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20899u16 , name : "Alder Lake PCH-P SMBus Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20900u16 , name : "Alder Lake-P PCH SPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20904u16 , name : "Alder Lake PCH UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20905u16 , name : "Alder Lake PCH UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20906u16 , name : "Alder Lake SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20907u16 , name : "Alder Lake SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20912u16 , name : "Alder Lake PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20913u16 , name : "Alder Lake PCI Express x1 Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20914u16 , name : "Alder Lake PCI Express x1 Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20915u16 , name : "Alder Lake PCI Express Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20923u16 , name : "Alder Lake-P PCH PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20924u16 , name : "Alder Lake PCI Express x4 Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20925u16 , name : "Alder Lake-P PCH PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20927u16 , name : "Alder Lake PCH-P PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20933u16 , name : "Alder Lake-P Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20934u16 , name : "Alder Lake-P Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20936u16 , name : "Alder Lake PCH-P High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 20938u16 , name : "Raptor Lake-P/U/H cAVS" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20940u16 , name : "Alder Lake Smart Sound Technology Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20947u16 , name : "Alder Lake-P SATA AHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 20952u16 , name : "Alder Lake-P Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20953u16 , name : "Alder Lake-P Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20960u16 , name : "Alder Lake PCH HECI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20963u16 , name : "Alder Lake AMT SOL Redirection" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 32902u16 , id : 20968u16 , name : "Alder Lake PCH Serial IO I2C Controller #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20969u16 , name : "Alder Lake PCH Serial IO I2C Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20970u16 , name : "Alder Lake PCH Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20971u16 , name : "Alder Lake PCH Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20973u16 , name : "Alder Lake PCH USB 3.2 xHCI Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20975u16 , name : "Alder Lake PCH Shared SRAM" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20976u16 , name : "Alder Lake-P PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5714u16 , name : "Dual Band Wi-Fi 6(802.11ax) Killer AX1650i 160MHz 2x2 [Cyclone Peak]" } , SubSystem { subvendor : 6742u16 , subdevice : 5745u16 , name : "Dual Band Wi-Fi 6E(802.11ax) AX1675s 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 6742u16 , subdevice : 5746u16 , name : "Dual Band Wi-Fi 6E(802.11ax) AX1675i 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 6742u16 , subdevice : 5778u16 , name : "Simultaneous Dual Band(Double Connect) Wi-Fi 6E AX1690i 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9560 160MHz 2x2 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 112u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 148u16 , name : "Dual Band Wi-Fi 6E(802.11ax) AX211 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 180u16 , name : "Simultaneous Dual Band(Double Connect) Wi-Fi 6E AX411 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 564u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9560 80MHz 2x2 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 580u16 , name : "Single Band Wi-Fi 6(802.11ax) AX101 80MHz 1x1 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 612u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9461 80MHz 1x1 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 628u16 , name : "Dual Band Wi-Fi E(802.11ax) AX203 80MHz 2x2 [Johnson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 676u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9462 80MHz 1x1 [Jefferson Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 16496u16 , name : "Dual Band Wi-Fi 6(802.11ax) AX201 160MHz 2x2 [Harrison Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 16528u16 , name : "Dual Band Wi-Fi 6E(802.11ax) AX211 160MHz 2x2 [Garfield Peak]" } , SubSystem { subvendor : 32902u16 , subdevice : 17060u16 , name : "Dual Band Wi-Fi 5(802.11ac) Wireless-AC 9462 80MHz 1x1 [Jefferson Peak]" }] } , Device { vendor_id : 32902u16 , id : 20977u16 , name : "Raptor Lake PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 16528u16 , name : "Wi-Fi 6E AX211 160MHz" }] } , Device { vendor_id : 32902u16 , id : 20988u16 , name : "Alder Lake-P Integrated Sensor Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 20992u16 , name : "EtherExpress PRO/100 Intelligent Server PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 20993u16 , name : "EtherExpress PRO/100 Intelligent Server Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "EtherExpress PRO/100 Server Ethernet Adapter" }] } , Device { vendor_id : 32902u16 , id : 21261u16 , name : "80310 (IOP) IO Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21633u16 , name : "Alder Lake-N PCH eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21664u16 , name : "Alder Lake-N P2SB Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21665u16 , name : "Alder Lake-N Power Management Controller (PMC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21667u16 , name : "Alder Lake-N SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21668u16 , name : "Alder Lake-N SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21670u16 , name : "Alder Lake-N Intel Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21672u16 , name : "Alder Lake-N Serial IO UART Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21673u16 , name : "Alder Lake-N UART Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21674u16 , name : "Alder Lake-N Generic SPI (GSPI) Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21675u16 , name : "Alder Lake-N Generic SPI (GSPI) Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21680u16 , name : "Alder Lake-N PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21681u16 , name : "Alder Lake-N PCI Express Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21682u16 , name : "Alder Lake-N PCI Express Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21683u16 , name : "Alder Lake-N PCI Express Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21688u16 , name : "Alder Lake-N PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21689u16 , name : "Alder Lake-N PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21690u16 , name : "Alder Lake-N PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21691u16 , name : "Alder Lake-N PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21694u16 , name : "Alder Lake-N PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21700u16 , name : "Alder Lake-N eMMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21701u16 , name : "Alder Lake-N I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21702u16 , name : "Alder Lake-N I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21703u16 , name : "Alder Lake-N UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21704u16 , name : "Alder Lake-N PCH High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21712u16 , name : "Alder Lake-N Touch Host Controller #0 (THC #0)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21713u16 , name : "Alder Lake-N Touch Host Controller #1 (THC #1)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21715u16 , name : "Alder Lake-N SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21722u16 , name : "Alder Lake-N UART Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21728u16 , name : "Alder Lake-N PCH HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21729u16 , name : "Alder Lake-N PCH HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21732u16 , name : "Alder Lake-N PCH HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21733u16 , name : "Alder Lake-N PCH HECI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21736u16 , name : "Alder Lake-N PCH I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21737u16 , name : "Alder Lake-N PCH I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21738u16 , name : "Alder Lake-N PCH I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21739u16 , name : "Alder Lake-N PCH I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21741u16 , name : "Alder Lake-N PCH USB 3.2 Gen 2x1 (10 Gb/s) xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21742u16 , name : "Alder Lake-N USB 3.2 Gen 1x1 (5 Gb/s) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21743u16 , name : "Alder Lake-N PCH Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21744u16 , name : "Alder Lake-N PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 580u16 , name : "Wi-Fi 6 AX101NGW" }] } , Device { vendor_id : 32902u16 , id : 21745u16 , name : "Alder Lake-N PCH CNVi WiFi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21746u16 , name : "Alder Lake-N PCH CNVi WiFi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21747u16 , name : "Alder Lake-N PCH CNVi WiFi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21755u16 , name : "Alder Lake-N Generic SPI (GSPI) Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21756u16 , name : "Alder Lake-N Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21759u16 , name : "Alder Lake-N UFS Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21762u16 , name : "Ethernet Controller (2) I225-LMvP" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34745u16 , name : "Thunderbolt Dock G4 PCIe NIC" } , SubSystem { subvendor : 6058u16 , subdevice : 8963u16 , name : "ThinkPad Universal Thunderbolt 4 Dock PCIe NIC" } , SubSystem { subvendor : 6838u16 , subdevice : 549u16 , name : "TS4 On-Board 2.5GbE Ethernet Adaptor" }] } , Device { vendor_id : 32902u16 , id : 21763u16 , name : "Ethernet Controller I226-LMvP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21770u16 , name : "Ethernet Connection (18) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21771u16 , name : "Ethernet Connection (18) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21772u16 , name : "800 Series Chipset Family GbE Controller (Corporate/vPro)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21773u16 , name : "800 Series Chipset Family GbE Controller (Consumer)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21774u16 , name : "Ethernet Connection (20) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21775u16 , name : "Ethernet Connection (20) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21776u16 , name : "Ethernet Connection (21) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 21777u16 , name : "Ethernet Connection (21) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22160u16 , name : "DG2 [Arc A770M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22161u16 , name : "DG2 [Arc A730M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22162u16 , name : "DG2 [Arc A550M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22163u16 , name : "DG2 [Arc A370M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22164u16 , name : "DG2 [Arc A350M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22165u16 , name : "DG2 [Iris Xe MAX A200M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22166u16 , name : "DG2 [Arc A570M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22167u16 , name : "DG2 [Arc A530M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22168u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22176u16 , name : "DG2 [Arc A770]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22177u16 , name : "DG2 [Arc A750]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22178u16 , name : "DG2 [Arc A580]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22179u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22180u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22181u16 , name : "DG2 [Arc A380]" , subsystems : & [SubSystem { subvendor : 5935u16 , subdevice : 14657u16 , name : "A380 ELF" } , SubSystem { subvendor : 5935u16 , subdevice : 14659u16 , name : "A380 ELF" } , SubSystem { subvendor : 5935u16 , subdevice : 16407u16 , name : "A380 Pioneer" }] } , Device { vendor_id : 32902u16 , id : 22182u16 , name : "DG2 [Arc A310]" , subsystems : & [SubSystem { subvendor : 5935u16 , subdevice : 16409u16 , name : "A380 ECO" }] } , Device { vendor_id : 32902u16 , id : 22183u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22184u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22185u16 , name : "DG2 [Arc Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22192u16 , name : "DG2 [Arc Pro A30M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22193u16 , name : "DG2 [Arc Pro A40/A50]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22194u16 , name : "DG2 [Arc Pro A60M]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22195u16 , name : "DG2 [Arc Pro A60]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22202u16 , name : "DG2 [Arc A380E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22203u16 , name : "DG2 [Arc A310E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22204u16 , name : "DG2 [Arc A370E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22205u16 , name : "DG2 [Arc A350E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22206u16 , name : "DG2 [Arc A750E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22207u16 , name : "DG2 [Arc A580E]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22208u16 , name : "ATS-M [Data Center GPU Flex 170]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22209u16 , name : "ATS-M [Data Center GPU Flex 140]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22210u16 , name : "ATS-M [Data Center GPU Flex 170V]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22400u16 , name : "JHL9580 Thunderbolt 5 80/120G Bridge [Barlow Ridge Host 80G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22401u16 , name : "JHL9580 Thunderbolt 5 80/120G NHI [Barlow Ridge Host 80G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22402u16 , name : "JHL9580 Thunderbolt 5 80/120G USB Controller [Barlow Ridge Host 80G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22403u16 , name : "JHL9540 Thunderbolt 4 Bridge [Barlow Ridge Host 40G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22404u16 , name : "JHL9540 Thunderbolt 4 NHI [Barlow Ridge Host 40G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22405u16 , name : "JHL9540 Thunderbolt 4 USB Controller [Barlow Ridge Host 40G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22406u16 , name : "JHL9480 Thunderbolt 5 80/120G Bridge [Barlow Ridge Hub 80G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22407u16 , name : "JHL9480 Thunderbolt 5 80/120G USB Controller [Barlow Ridge Hub 80G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22418u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors ILMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22419u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors ACPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22420u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors SPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22421u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22422u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22423u16 , name : "Xeon 6900 6700 6500 Series with P-Cores Processors UART" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22428u16 , name : "Ethernet Connection E825-C for backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22429u16 , name : "Ethernet Connection E825-C for QSFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22430u16 , name : "Ethernet Connection E825-C for SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22431u16 , name : "Ethernet Connection E825-C 10GbE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22432u16 , name : "Ethernet Connection (24) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22433u16 , name : "Ethernet Connection (24) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22436u16 , name : "JHL9440 Thunderbolt 4 Bridge [Barlow Ridge Hub 40G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22437u16 , name : "JHL9440 Thunderbolt 4 USB Controller [Barlow Ridge Hub 40G 2023]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22444u16 , name : "Ethernet Controller E610" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22445u16 , name : "Ethernet Controller E610 Virtual Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22446u16 , name : "Ethernet Controller E610 Backplane" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22447u16 , name : "Ethernet Controller E610 SFP" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22448u16 , name : "Ethernet Controller E610-XT/E610-XT2 10GBASE-T" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "Ethernet Network Adapter E610-XT4" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E610-XT2" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E610-XT4 for OCP 3.0" } , SubSystem { subvendor : 32902u16 , subdevice : 4u16 , name : "Ethernet Network Adapter E610-XT2 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 22449u16 , name : "Ethernet Controller E610-AT2 1000BASE-T" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Ethernet Converged Network Adapter E610" } , SubSystem { subvendor : 32902u16 , subdevice : 2u16 , name : "Ethernet Network Adapter E610-IT4" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Ethernet Network Adapter E610-IT4 for OCP 3.0" }] } , Device { vendor_id : 32902u16 , id : 22450u16 , name : "Ethernet Controller E610 SGMII" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22451u16 , name : "Ethernet Connection (25) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22452u16 , name : "Ethernet Connection (25) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22453u16 , name : "Ethernet Connection (26) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22454u16 , name : "Ethernet Connection (26) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22455u16 , name : "Ethernet Connection (27) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22456u16 , name : "Ethernet Connection (27) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22457u16 , name : "Ethernet Connection (29) I219-LM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22458u16 , name : "Ethernet Connection (29) I219-V" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22597u16 , name : "QEMU NVM Express Controller" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 22784u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22785u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor PCIe Controller (x16)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22786u16 , name : "Kaby Lake-S GT1 [HD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22788u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Aspire E5-575G" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" }] } , Device { vendor_id : 32902u16 , id : 22789u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor PCIe Controller (x8)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22790u16 , name : "Kaby Lake-U GT1 [HD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22793u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor PCIe Controller (x4)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22795u16 , name : "HD Graphics 610" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22796u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22799u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" } , SubSystem { subvendor : 5218u16 , subdevice : 31336u16 , name : "B250 KRAIT GAMING (MS-7A68)" } , SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 22800u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22801u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Gaussian Mixture Model" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22802u16 , name : "Kaby Lake-S GT2 [HD Graphics 630]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" } , SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 22804u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 22806u16 , name : "Kaby Lake-U GT2 [HD Graphics 620]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4244u16 , name : "Aspire E5-575G" } , SubSystem { subvendor : 6058u16 , subdevice : 8776u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" }] } , Device { vendor_id : 32902u16 , id : 22807u16 , name : "Kaby Lake-R GT2 [UHD Graphics 620]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480 (20L5)" } , SubSystem { subvendor : 6058u16 , subdevice : 8798u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 22808u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22811u16 , name : "Kaby Lake-H GT2 [HD Graphics 630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22812u16 , name : "UHD Graphics 615" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22813u16 , name : "Kaby Lake-DT GT2 [HD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22814u16 , name : "Kaby Lake-Y GT2 [HD Graphics 615]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22815u16 , name : "Xeon E3-1200 v6/7th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22817u16 , name : "HD Graphics 620" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22819u16 , name : "HD Graphics 635" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22822u16 , name : "Kaby Lake-U GT3 [Iris Plus Graphics 640]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 22823u16 , name : "Kaby Lake-U GT3 [Iris Plus Graphics 650]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23172u16 , name : "Apollo Lake [HD Graphics 505]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23173u16 , name : "Apollo Lake GT1 [HD Graphics 500]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23176u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Imaging Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23192u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Audio Cluster" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23194u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Trusted Execution Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23202u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23208u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series USB xHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23212u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23214u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23216u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23218u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23220u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23222u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23224u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23226u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series I2C Controller #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23228u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series HSUART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23230u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series HSUART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23232u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series HSUART Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23234u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23236u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SPI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23238u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SPI Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23240u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PWM Pin Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23242u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SDXC/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23244u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series eMMC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23248u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23252u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23254u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port B #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23255u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port B #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23256u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port A #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23257u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port A #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23258u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port A #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23259u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series PCI Express Port A #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23267u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23272u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Low Pin Count Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23278u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series HSUART Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 23280u16 , name : "Celeron N3350/Pentium N4200/Atom E3900 Series Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25600u16 , name : "Core Ultra 200V Series Processors with 4 P-Cores 4 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25629u16 , name : "Core Ultra 200V Series Processors Dynamic Tuning Technology (DTT)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25632u16 , name : "Lunar Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25662u16 , name : "Core Ultra 200V Series Processors NPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25693u16 , name : "Core Ultra 200V Series Processors IPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25725u16 , name : "Core Ultra 200V Series Processors Crash Log and Telemetry" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25760u16 , name : "Core Ultra 200V Series Processors Arc Graphics 130V/140V GPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 25776u16 , name : "Lunar Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26048u16 , name : "5100 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26082u16 , name : "5100 Chipset PCI Express x4 Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26083u16 , name : "5100 Chipset PCI Express x4 Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26084u16 , name : "5100 Chipset PCI Express x4 Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26085u16 , name : "5100 Chipset PCI Express x4 Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26086u16 , name : "5100 Chipset PCI Express x4 Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26087u16 , name : "5100 Chipset PCI Express x4 Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26096u16 , name : "5100 Chipset FSB Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300" }] } , Device { vendor_id : 32902u16 , id : 26097u16 , name : "5100 Chipset Reserved Registers" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300" }] } , Device { vendor_id : 32902u16 , id : 26099u16 , name : "5100 Chipset Reserved Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26101u16 , name : "5100 Chipset DDR Channel 0 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26102u16 , name : "5100 Chipset DDR Channel 1 Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26103u16 , name : "5100 Chipset PCI Express x8 Port 2-3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26104u16 , name : "5100 Chipset PCI Express x8 Port 4-5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26105u16 , name : "5100 Chipset PCI Express x8 Port 6-7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26106u16 , name : "5100 Chipset PCI Express x16 Port 4-7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26111u16 , name : "5100 Chipset DMA Engine" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26444u16 , name : "CRI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26445u16 , name : "CRI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26446u16 , name : "CRI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26447u16 , name : "CRI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 26448u16 , name : "CRI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28195u16 , name : "Nova Lake PCH-S SMbus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28196u16 , name : "Nova Lake PCH-S SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28200u16 , name : "Nova Lake PCH-S Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28201u16 , name : "Nova Lake PCH-S Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28202u16 , name : "Nova Lake PCH-S Serial IO SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28203u16 , name : "Nova Lake PCH-S Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28236u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28237u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28238u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28239u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28254u16 , name : "Nova Lake PCH-S Serial IO SPI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28282u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28283u16 , name : "Nova Lake PCH-S Serial IO I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28416u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DMI2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28417u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28418u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28419u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28420u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28421u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28422u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28423u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28424u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28425u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28426u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28427u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28432u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28433u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28434u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28435u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28436u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28437u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28438u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28439u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28440u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28441u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28442u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28443u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28444u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28445u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R2PCIe Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28446u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Ubox" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28447u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Ubox" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28448u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 0" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28449u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 1" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28450u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28451u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 3" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28452u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 4" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28453u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 5" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28454u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 6" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28455u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Crystal Beach DMA Channel 7" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28456u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Map/VTd_Misc/System Management" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28457u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO Hot Plug" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28458u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IIO RAS/Control Status/Global Errors" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28460u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D I/O APIC" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" }] } , Device { vendor_id : 32902u16 , id : 28464u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28466u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28467u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28468u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R2PCIe Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28470u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 0/1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28471u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 0/1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28472u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28473u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D IO Performance Monitoring" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28474u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28478u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28479u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28480u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28481u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28483u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28485u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28486u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28487u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 2 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28496u16 , name : "Xeon Processor D Family QuickData Technology Register DMA Channel 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28497u16 , name : "Xeon Processor D Family QuickData Technology Register DMA Channel 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28498u16 , name : "Xeon Processor D Family QuickData Technology Register DMA Channel 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28499u16 , name : "Xeon Processor D Family QuickData Technology Register DMA Channel 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28500u16 , name : "Xeon Processor D Family QuickAssist Technology" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28501u16 , name : "Xeon Processor D Family QuickAssist Technology Virtual Fuction" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28512u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28520u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Target Address/Thermal/RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28522u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28523u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28524u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28525u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28526u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 2/3 Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28527u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Global Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28528u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28529u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Target Address/Thermal/RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28534u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28536u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28537u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Target Address/Thermal/RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28541u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Ubox" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28542u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28544u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28545u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D R3 QPI Link 0/1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28547u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28549u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28550u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28551u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 0 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28552u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28554u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28560u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28563u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28565u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28566u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D QPI Link 1 Debug" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28568u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28569u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28570u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28572u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28576u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Home Agent 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28584u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Target Address/Thermal/RAS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28586u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28587u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28588u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28589u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel Target Address Decoder" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28590u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 0/1 Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28591u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Global Broadcast" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28592u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28593u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28594u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 0 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28595u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 1 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28596u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28597u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 3 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28598u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 2 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28599u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 0 - Channel 3 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28600u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 2/3 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28601u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 2/3 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28602u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 2/3 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28603u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 2/3 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28604u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 0/1 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28605u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 0/1 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28606u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 0/1 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28607u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D DDRIO Channel 0/1 Interface" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28608u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28609u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28610u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28611u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28612u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28613u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28614u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28615u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28616u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28617u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28618u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28619u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28620u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28621u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28622u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28623u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Power Control Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28624u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 0 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28625u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 1 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28626u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 0 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28627u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 1 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28628u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 2 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28629u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 3 Thermal Control" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28630u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 2 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28631u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Memory Controller 1 - Channel 3 Error" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28640u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28641u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28642u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28643u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28644u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28645u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28646u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28647u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28648u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28649u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28650u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28651u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28652u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28653u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28654u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28655u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28656u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28657u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28664u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28665u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28666u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28667u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28668u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28669u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28670u16 , name : "Xeon E7 v4/Xeon E5 v4/Xeon E3 v4/Xeon D Caching Agent" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28672u16 , name : "82371SB PIIX3 ISA [Natoma/Triton II]" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65533u16 , name : "Propolis Virtual PIIX3 ISA Controller" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "Qemu virtual machine" }] } , Device { vendor_id : 32902u16 , id : 28688u16 , name : "82371SB PIIX3 IDE [Natoma/Triton II]" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "Qemu virtual machine" }] } , Device { vendor_id : 32902u16 , id : 28704u16 , name : "82371SB PIIX3 USB [Natoma/Triton II]" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 28720u16 , name : "430VX - 82437VX TVX [Triton VX]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28752u16 , name : "Intercast Video Capture Card" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28753u16 , name : "PB 642365-003 (Business Video Conferencing Card)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28928u16 , name : "430TX - 82439TX MTXC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28944u16 , name : "82371AB/EB/MB PIIX4 ISA" , subsystems : & [SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "Virtual Machine Chipset" }] } , Device { vendor_id : 32902u16 , id : 28945u16 , name : "82371AB/EB/MB PIIX4 IDE" , subsystems : & [SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "Virtual Machine Chipset" }] } , Device { vendor_id : 32902u16 , id : 28946u16 , name : "82371AB/EB/MB PIIX4 USB" , subsystems : & [SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "Virtual Machine Chipset" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 32902u16 , id : 28947u16 , name : "82371AB/EB/MB PIIX4 ACPI" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65532u16 , name : "Propolis Virtual PIIX4 PM Controller" } , SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "Virtual Machine Chipset" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "Qemu virtual machine" }] } , Device { vendor_id : 32902u16 , id : 28960u16 , name : "82810 GMCH (Graphics Memory Controller Hub)" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4160u16 , name : "CL7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4192u16 , name : "PC7 mainboard" }] } , Device { vendor_id : 32902u16 , id : 28961u16 , name : "82810 (CGC) Chipset Graphics Controller" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4160u16 , name : "CL7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4192u16 , name : "PC7 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 17217u16 , name : "Cayman (CA810) Mainboard" }] } , Device { vendor_id : 32902u16 , id : 28962u16 , name : "82810 DC-100 (GMCH) Graphics Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28963u16 , name : "82810 DC-100 (CGC) Chipset Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28964u16 , name : "82810E DC-133 (GMCH) Graphics Memory Controller Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 180u16 , name : "OptiPlex GX110" }] } , Device { vendor_id : 32902u16 , id : 28965u16 , name : "82810E DC-133 (CGC) Chipset Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 180u16 , name : "OptiPlex GX110" }] } , Device { vendor_id : 32902u16 , id : 28966u16 , name : "82810 DC-133 System and Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28968u16 , name : "82810-M DC-100 System and Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 28970u16 , name : "82810-M DC-133 System and Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29056u16 , name : "440LX/EX - 82443LX/EX Host bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29057u16 , name : "440LX/EX - 82443LX/EX AGP bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29072u16 , name : "440BX/ZX/DX - 82443BX/ZX/DX Host bridge" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 1280u16 , name : "Armada 1750 Laptop System Chipset" } , SubSystem { subvendor : 3601u16 , subdevice : 45328u16 , name : "Armada M700/E500" } , SubSystem { subvendor : 4136u16 , subdevice : 142u16 , name : "PowerEdge 1300 mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32827u16 , name : "CUBX-L/E Mainboard" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Toshiba Tecra 8100 Laptop System Chipset" } , SubSystem { subvendor : 5549u16 , subdevice : 6518u16 , name : "Virtual Machine Chipset" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" }] } , Device { vendor_id : 32902u16 , id : 29073u16 , name : "440BX/ZX/DX - 82443BX/ZX/DX AGP bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 142u16 , name : "PowerEdge 1300 mainboard" }] } , Device { vendor_id : 32902u16 , id : 29074u16 , name : "440BX/ZX/DX - 82443BX/ZX/DX Host bridge (AGP disabled)" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 1120u16 , name : "Armada 1700 Laptop System Chipset" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Satellite 4010" } , SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 29072u16 , name : "Dell PowerEdge 350" }] } , Device { vendor_id : 32902u16 , id : 29076u16 , name : "82440MX Host Bridge" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 0u16 , name : "Versa Note Vxi" } , SubSystem { subvendor : 19539u16 , subdevice : 4256u16 , name : "CA3/CR3 mainboard" }] } , Device { vendor_id : 32902u16 , id : 29077u16 , name : "82440MX AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32972u16 , name : "Versa Note VXi" } , SubSystem { subvendor : 4303u16 , subdevice : 4249u16 , name : "QSound_SigmaTel Stac97 PCI Audio" } , SubSystem { subvendor : 4564u16 , subdevice : 64u16 , name : "SoundMAX Integrated Digital Audio" } , SubSystem { subvendor : 4564u16 , subdevice : 72u16 , name : "SoundMAX Integrated Digital Audio" }] } , Device { vendor_id : 32902u16 , id : 29078u16 , name : "82440MX AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29080u16 , name : "82440MX ISA Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29081u16 , name : "82440MX EIDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29082u16 , name : "82440MX USB Universal Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29083u16 , name : "82440MX Power Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29088u16 , name : "440GX - 82443GX Host bridge" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" }] } , Device { vendor_id : 32902u16 , id : 29089u16 , name : "440GX - 82443GX AGP bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29090u16 , name : "440GX - 82443GX Host bridge (AGP disabled)" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" }] } , Device { vendor_id : 32902u16 , id : 29186u16 , name : "Core Ultra 200H Series Processors eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29187u16 , name : "Core Ultra 200V Series Processors eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 29536u16 , name : "XMM7360 LTE Advanced Modem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30048u16 , name : "XMM7560 LTE Advanced Pro Modem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30208u16 , name : "82372FB PIIX5 ISA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30209u16 , name : "82372FB PIIX5 IDE" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30210u16 , name : "82372FB PIIX5 USB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30211u16 , name : "82372FB PIIX5 SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30466u16 , name : "Arrow Lake-H eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30496u16 , name : "Core Ultra 200H/200V Series Processors P2SB (SOC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30497u16 , name : "Core Ultra 200H/200V Series Processors PMC (SOC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30498u16 , name : "Core Ultra 200H/200V Series Processors SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30499u16 , name : "Core Ultra 200H/200V Series Processors SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30500u16 , name : "Core Ultra 200H/200V Series Processors Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30501u16 , name : "Core Ultra 200H/200V Series Processors UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30502u16 , name : "Core Ultra 200H/200V Series Processors UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30503u16 , name : "Core Ultra 200H/200V Series Processors GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30504u16 , name : "Core Ultra 200H/200V Series Processors HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30512u16 , name : "Arrow Lake-H [LPC/eSPI Controller]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30520u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30521u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30522u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30523u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30524u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30525u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30526u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30527u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30528u16 , name : "Arrow Lake CNVi WiFi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30533u16 , name : "Core Ultra 200H/200V Series Processors Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30534u16 , name : "Core Ultra 200H/200V Series Processors GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30536u16 , name : "Core Ultra 200H/200V Series Processors Touch Host Controller (THC) #0 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30537u16 , name : "Core Ultra 200H/200V Series Processors Touch Host Controller (THC) #0 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30538u16 , name : "Core Ultra 200H/200V Series Processors Touch Host Controller (THC) #1 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30539u16 , name : "Core Ultra 200H/200V Series Processors Touch Host Controller (THC) #1 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30540u16 , name : "Core Ultra 200H/200V Series Processors Gaussian & Neural-Network Accelerator (GNA)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30541u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #9 (PXPC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30544u16 , name : "Core Ultra 200H/200V Series Processors I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30545u16 , name : "Core Ultra 200H/200V Series Processors I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30546u16 , name : "Core Ultra 200H/200V Series Processors UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30552u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30553u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30554u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30563u16 , name : "Core Ultra 200H/200V Series Processors SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30567u16 , name : "Core Ultra 200H/200V Series Processors SATA Controller (RAID 0/1/5/10) premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30576u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30577u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30578u16 , name : "Core Ultra 200H/200V Series Processors CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30579u16 , name : "Core Ultra 200H/200V Series Processors CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30580u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30581u16 , name : "Core Ultra 200H/200V Series Processors CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30584u16 , name : "Core Ultra 200H/200V Series Processors I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30585u16 , name : "Core Ultra 200H/200V Series Processors I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30586u16 , name : "Core Ultra 200H/200V Series Processors I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30587u16 , name : "Core Ultra 200H/200V Series Processors I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30588u16 , name : "Core Ultra 200H/200V Series Processors I3C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30589u16 , name : "Core Ultra 200H/200V Series Processors Standalone xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30590u16 , name : "Core Ultra 200H/200V Series Processors Standalone USB Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30591u16 , name : "Core Ultra 200H/200V Series Processors Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 30720u16 , name : "82740 (i740) AGP Graphics Accelerator" , subsystems : & [SubSystem { subvendor : 61u16 , subdevice : 8u16 , name : "Starfighter AGP" } , SubSystem { subvendor : 61u16 , subdevice : 11u16 , name : "Starfighter AGP" } , SubSystem { subvendor : 4242u16 , subdevice : 256u16 , name : "Stealth II G460" } , SubSystem { subvendor : 4276u16 , subdevice : 8218u16 , name : "Lightspeed 740" } , SubSystem { subvendor : 4276u16 , subdevice : 8239u16 , name : "Lightspeed 740" } , SubSystem { subvendor : 32902u16 , subdevice : 0u16 , name : "Terminator 2x/i" } , SubSystem { subvendor : 32902u16 , subdevice : 256u16 , name : "Intel740 Graphics Accelerator" }] } , Device { vendor_id : 32902u16 , id : 31236u16 , name : "Z790 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31237u16 , name : "H770 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31238u16 , name : "B760 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31244u16 , name : "HM770 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31245u16 , name : "WM790 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31251u16 , name : "C266 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31252u16 , name : "C262 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31264u16 , name : "700 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31265u16 , name : "700 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31267u16 , name : "700 Series Chipset Family SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31268u16 , name : "700 Series Chipset Family SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31270u16 , name : "700 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31271u16 , name : "700 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31272u16 , name : "700 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31273u16 , name : "700 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31274u16 , name : "700 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31275u16 , name : "700 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31280u16 , name : "700 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31281u16 , name : "700 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31282u16 , name : "700 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31283u16 , name : "700 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31284u16 , name : "700 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31285u16 , name : "700 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31286u16 , name : "700 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31287u16 , name : "700 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31288u16 , name : "700 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31289u16 , name : "700 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31290u16 , name : "700 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31291u16 , name : "700 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31292u16 , name : "700 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31293u16 , name : "700 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31294u16 , name : "700 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31295u16 , name : "700 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31296u16 , name : "700 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31297u16 , name : "700 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31298u16 , name : "700 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31299u16 , name : "700 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31300u16 , name : "700 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31301u16 , name : "700 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31302u16 , name : "700 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31303u16 , name : "700 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31304u16 , name : "700 Series Chipset Family PCIe Root Port #25" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31305u16 , name : "700 Series Chipset Family PCIe Root Port #26" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31306u16 , name : "700 Series Chipset Family PCIe Root Port #27" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31307u16 , name : "700 Series Chipset Family PCIe Root Port #28" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31308u16 , name : "700 Series Chipset Family I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31309u16 , name : "700 Series Chipset Family I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31310u16 , name : "700 Series Chipset Family I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31311u16 , name : "700 Series Chipset Family I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31312u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31313u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31314u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31315u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31316u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31317u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31318u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31319u16 , name : "700 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31324u16 , name : "700 Series Chipset Family UART #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31328u16 , name : "700 Series Chipset Family USB 3.2 Gen 2x2 (20 Gbs) xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31329u16 , name : "700 Series Chipset Family USB 3.2 Gen 1x1 (5 Gbs) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31330u16 , name : "700 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31336u16 , name : "700 Series Chipset Family CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31337u16 , name : "700 Series Chipset Family CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31338u16 , name : "700 Series Chipset Family CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31339u16 , name : "700 Series Chipset Family CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31340u16 , name : "700 Series Chipset Family CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31341u16 , name : "700 Series Chipset Family CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31344u16 , name : "700 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 116u16 , name : "Wi-Fi 6 AX201 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 144u16 , name : "WiFi 6E AX211 160MHz" }] } , Device { vendor_id : 32902u16 , id : 31345u16 , name : "700 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31346u16 , name : "700 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31347u16 , name : "700 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31352u16 , name : "700 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31353u16 , name : "700 Series Chipset Family GSPI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31355u16 , name : "700 Series Chipset Family GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31356u16 , name : "700 Series Chipset Family I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31357u16 , name : "700 Series Chipset Family I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31358u16 , name : "700 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31363u16 , name : "Q670 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31364u16 , name : "Z690 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31365u16 , name : "H670 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31366u16 , name : "B660 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31367u16 , name : "H610 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31368u16 , name : "W680 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31370u16 , name : "W790 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31372u16 , name : "HM670 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31373u16 , name : "WM690 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31392u16 , name : "600 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31393u16 , name : "600 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31395u16 , name : "600 Series Chipset Family SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31396u16 , name : "600 Series Chipset Family SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31398u16 , name : "600 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31399u16 , name : "600 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31400u16 , name : "600 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31401u16 , name : "600 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31402u16 , name : "600 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31403u16 , name : "600 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31408u16 , name : "600 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31409u16 , name : "600 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31410u16 , name : "600 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31411u16 , name : "600 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31412u16 , name : "600 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31413u16 , name : "600 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31414u16 , name : "600 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31415u16 , name : "600 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31416u16 , name : "600 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31417u16 , name : "600 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31418u16 , name : "600 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31419u16 , name : "600 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31420u16 , name : "600 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31421u16 , name : "600 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31422u16 , name : "600 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31423u16 , name : "600 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31424u16 , name : "600 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31425u16 , name : "600 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31426u16 , name : "600 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31427u16 , name : "600 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31428u16 , name : "600 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31429u16 , name : "600 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31430u16 , name : "600 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31431u16 , name : "600 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31432u16 , name : "600 Series Chipset Family PCIe Root Port #25" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31433u16 , name : "600 Series Chipset Family PCIe Root Port #26" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31434u16 , name : "600 Series Chipset Family PCIe Root Port #27" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31435u16 , name : "600 Series Chipset Family PCIe Root Port #28" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31436u16 , name : "600 Series Chipset Family I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31437u16 , name : "600 Series Chipset Family I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31438u16 , name : "600 Series Chipset Family I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31439u16 , name : "600 Series Chipset Family 12C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31440u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31441u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31442u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31443u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31444u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31445u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31446u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31447u16 , name : "600 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31452u16 , name : "600 Series Chipset Family UART #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31456u16 , name : "600 Series Chipset Family USB 3.2 Gen 2x2 (20Gbs) XHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31457u16 , name : "600 Series Chipset Family USB 3.2 Gen 1x1 (5Gbs) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31458u16 , name : "600 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31464u16 , name : "600 Series Chipset Family CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31465u16 , name : "600 Series Chipset Family CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31466u16 , name : "600 Series Chipset Family CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31467u16 , name : "600 Series Chipset Family CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31468u16 , name : "600 Series Chipset Family CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31469u16 , name : "600 Series Chipset Family CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31472u16 , name : "600 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Wireless-AC 9560" } , SubSystem { subvendor : 32902u16 , subdevice : 112u16 , name : "Wi-Fi 6 AX201 160MHz" } , SubSystem { subvendor : 32902u16 , subdevice : 148u16 , name : "Wi-Fi 6 AX201 160MHz" }] } , Device { vendor_id : 32902u16 , id : 31473u16 , name : "600 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31474u16 , name : "600 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31475u16 , name : "600 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31480u16 , name : "600 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31481u16 , name : "600 Series Chipset Family GSPI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31483u16 , name : "600 Series Chipset Family GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31484u16 , name : "600 Series Chipset Family I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31485u16 , name : "600 Series Chipset Family I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 31486u16 , name : "600 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32001u16 , name : "Meteor Lake-H 6p+8e cores Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32003u16 , name : "Core Ultra 200H/200V Series Processors Dynamic Tuning Technology (DTT)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32006u16 , name : "Core Ultra 200H Series Processors with 6 P-Cores 8 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32011u16 , name : "Core Ultra 200H/200V Series Processors VMD" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32013u16 , name : "Core Ultra 200H/200V Series Processors Platform Monitoring Technology (PMT)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32025u16 , name : "Core Ultra 200H/200V Series Processors IPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32026u16 , name : "Core Ultra 200S/200S-Plus Series Processors with 8 P-Cores 16 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32027u16 , name : "Core Ultra 200S Series Processors with 8 P-Cores 12 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32028u16 , name : "Core Ultra 200HX/HX-Plus Series Processors with 8 P-Cores 16 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32029u16 , name : "Core Ultra 200H/200V Series Processors NPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32041u16 , name : "Core Ultra 200S-Plus Series Processors with 6 P-Cores 12 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32042u16 , name : "Core Ultra 200S Series Processors with 6 P-Cores 8 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32045u16 , name : "Core Ultra 200HX/HX-Plus Series Processors with 8 P-Cores 12 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32047u16 , name : "Core Ultra 200HX Series Processors with 6 P-Cores 8 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32048u16 , name : "Core Ultra 200U Series Processors with 2 P-Cores 8 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32053u16 , name : "Core Ultra 200S Series Processors with 6 P-Cores 4 E-Cores Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32064u16 , name : "Meteor Lake-M [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32065u16 , name : "Arrow Lake-U [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32069u16 , name : "Meteor Lake-P [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32081u16 , name : "Arrow Lake-P [Arc Pro 130T/140T]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32085u16 , name : "Meteor Lake-P [Intel Arc Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32096u16 , name : "Meteor Lake-M [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32103u16 , name : "Arrow Lake-S [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32209u16 , name : "Arrow Lake-P [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32213u16 , name : "Meteor Lake-P [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32257u16 , name : "Meteor Lake-P LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32258u16 , name : "Meteor Lake-H eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32290u16 , name : "Meteor Lake-P SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32291u16 , name : "Meteor Lake-P SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32292u16 , name : "Meteor Lake-P Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32293u16 , name : "Meteor Lake-P Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32294u16 , name : "Meteor Lake-P Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32295u16 , name : "Meteor Lake-P Serial IO SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32296u16 , name : "Meteor Lake-P HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32304u16 , name : "Meteor Lake-P Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32319u16 , name : "Meteor Lake-H/U PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32320u16 , name : "Meteor Lake PCH CNVi WiFi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 148u16 , name : "Wi-Fi 6E AX211 160MHz" }] } , Device { vendor_id : 32902u16 , id : 32325u16 , name : "Meteor Lake-P Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32326u16 , name : "Meteor Lake-P Serial IO SPI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32332u16 , name : "Meteor Lake-P Gaussian & Neural-Network Accelerator" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32333u16 , name : "Meteor Lake-H/U PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32336u16 , name : "Meteor Lake-P Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32337u16 , name : "Meteor Lake-P Serial IO I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32338u16 , name : "Meteor Lake-P Serial IO UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32368u16 , name : "Meteor Lake-P CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32371u16 , name : "Meteor Lake-P Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32376u16 , name : "Meteor Lake-P Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32377u16 , name : "Meteor Lake-P Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32378u16 , name : "Meteor Lake-P Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32379u16 , name : "Meteor Lake-P Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32381u16 , name : "Meteor Lake-P USB 3.2 Gen 2x1 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32382u16 , name : "Meteor Lake-P USB Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32383u16 , name : "Meteor Lake-H/U Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32448u16 , name : "Core Ultra 200 Series Processors USB xHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32449u16 , name : "Core Ultra 200 Series Processors USB xDCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32450u16 , name : "Core Ultra 200 Series Processors Thunderbolt DMA0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32451u16 , name : "Core Ultra 200 Series Processors Thunderbolt DMA1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32452u16 , name : "Core Ultra 200 Series Processors USB Type-C Subsystem PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32453u16 , name : "Core Ultra 200 Series Processors USB Type-C Subsystem PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32454u16 , name : "Core Ultra 200 Series Processors USB Type-C Subsystem PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32455u16 , name : "Core Ultra 200 Series Processors USB Type-C Subsystem PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32456u16 , name : "Core Ultra 200 Series Processors P2SB (IOE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32457u16 , name : "Core Ultra 200H/200V Series Processors IEH (IOE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32458u16 , name : "Core Ultra 200 Series Processors PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32459u16 , name : "Core Ultra 200H/200V Series Processors PCIe Root Port #11 (PXPE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32460u16 , name : "Core Ultra 200 Series Processors PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32462u16 , name : "Core Ultra 200 Series Processors PMC (IOE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32463u16 , name : "Core Ultra 200 Series Processors Shared SRAM (IOE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32515u16 , name : "Q870 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32516u16 , name : "Z890 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32518u16 , name : "B860 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32519u16 , name : "H810 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32520u16 , name : "W880 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32524u16 , name : "HM870 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32525u16 , name : "WM880 Chipset eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32544u16 , name : "800 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32545u16 , name : "800 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32547u16 , name : "800 Series Chipset Family SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32548u16 , name : "800 Series Chipset Family SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32550u16 , name : "800 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32551u16 , name : "800 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32552u16 , name : "800 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32553u16 , name : "800 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32554u16 , name : "800 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32555u16 , name : "800 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32560u16 , name : "800 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32561u16 , name : "800 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32562u16 , name : "800 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32563u16 , name : "800 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32564u16 , name : "800 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32565u16 , name : "800 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32566u16 , name : "800 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32567u16 , name : "800 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32568u16 , name : "800 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32569u16 , name : "800 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32570u16 , name : "800 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32571u16 , name : "800 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32572u16 , name : "800 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32573u16 , name : "800 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32574u16 , name : "800 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32575u16 , name : "800 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32576u16 , name : "800 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32577u16 , name : "800 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32578u16 , name : "800 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32579u16 , name : "800 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32580u16 , name : "800 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32581u16 , name : "800 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32582u16 , name : "800 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32583u16 , name : "800 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32588u16 , name : "800 Series Chipset Family I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32589u16 , name : "800 Series Chipset Family I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32590u16 , name : "800 Series Chipset Family I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32591u16 , name : "800 Series Chipset Family I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32592u16 , name : "800 Series Chipset Family Audio Context Engine (ACE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32600u16 , name : "800 Series Chipset Family Touch Host Controller (THC) #0 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32601u16 , name : "800 Series Chipset Family Touch Host Controller (THC) #0 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32602u16 , name : "800 Series Chipset Family Touch Host Controller (THC) #1 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32603u16 , name : "800 Series Chipset Family Touch Host Controller (THC) #1 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32604u16 , name : "800 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32605u16 , name : "800 Series Chipset Family UART #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32606u16 , name : "800 Series Chipset Family GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32607u16 , name : "800 Series Chipset Family GSPI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32610u16 , name : "800 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32614u16 , name : "800 Series Chipset Family SATA Controller (RAID 0/1/5/10) premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32616u16 , name : "800 Series Chipset Family CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32617u16 , name : "800 Series Chipset Family CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32618u16 , name : "800 Series Chipset Family IDE Redirection (IDER-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32619u16 , name : "800 Series Chipset Family Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32620u16 , name : "800 Series Chipset Family CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32621u16 , name : "800 Series Chipset Family CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32622u16 , name : "800 Series Chipset Family USB 3.1 xHCI HC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32623u16 , name : "800 Series Chipset Family USB Device Controller (OTG) (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32624u16 , name : "800 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 148u16 , name : "WiFi 6E AX211 160MHz" }] } , Device { vendor_id : 32902u16 , id : 32632u16 , name : "800 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32633u16 , name : "800 Series Chipset Family I3C" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32634u16 , name : "800 Series Chipset Family I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32635u16 , name : "800 Series Chipset Family I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32636u16 , name : "800 Series Chipset Family Silicon Security Engine HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32637u16 , name : "800 Series Chipset Family Silicon Security Engine HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32638u16 , name : "800 Series Chipset Family Silicon Security Engine HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32770u16 , name : "Trusted Execution Technology Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 32771u16 , name : "Trusted Execution Technology Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33024u16 , name : "US15W/US15X SCH [Poulsbo] Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33025u16 , name : "US15L/UL11L SCH [Poulsbo] Host Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33032u16 , name : "US15W/US15X SCH [Poulsbo] Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33033u16 , name : "US15L/UL11L SCH [Poulsbo] Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33040u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33042u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33044u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] USB UHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33045u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] USB UHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33046u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] USB UHCI Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33047u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33048u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] USB Client Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33049u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] LPC Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33050u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] IDE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33051u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33052u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] SDIO/MMC Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33053u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] SDIO/MMC Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33054u16 , name : "US15W/US15X/US15L/UL11L SCH [Poulsbo] SDIO/MMC Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33152u16 , name : "Atom Processor E6xx PCI Express Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33153u16 , name : "Atom Processor E6xx PCI Express Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33154u16 , name : "Atom Processor E6xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33155u16 , name : "Atom Processor E6xx Configuration Unit" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33156u16 , name : "Atom Processor E6xx PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33157u16 , name : "Atom Processor E6xx PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33158u16 , name : "Atom Processor E6xx LPC Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33988u16 , name : "450KX/GX [Orion] - 82454KX/GX PCI bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33989u16 , name : "450KX/GX [Orion] - 82453KX/GX Memory controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33994u16 , name : "450NX - 82451NX Memory & I/O Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 33995u16 , name : "450NX - 82454NX/84460GX PCI Expander Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34016u16 , name : "460GX - 84460GX System Address Controller (SAC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34017u16 , name : "460GX - 84460GX System Data Controller (SDC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34018u16 , name : "460GX - 84460GX AGP Bridge (GXB function 2)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34019u16 , name : "460GX - 84460GX Memory Address Controller (MAC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34020u16 , name : "460GX - 84460GX Memory Data Controller (MDC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34022u16 , name : "460GX - 82466GX Wide and fast PCI eXpander Bridge (WXB)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34026u16 , name : "460GX - 84460GX AGP Bridge (GXB function 1)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34048u16 , name : "IXP4XX Network Processor (IXP420/421/422/425/IXC1100)" , subsystems : & [SubSystem { subvendor : 6547u16 , subdevice : 3565u16 , name : "mGuard-PCI AV#2" } , SubSystem { subvendor : 6547u16 , subdevice : 3566u16 , name : "mGuard-PCI AV#1" } , SubSystem { subvendor : 6547u16 , subdevice : 3567u16 , name : "mGuard-PCI AV#0" }] } , Device { vendor_id : 32902u16 , id : 34307u16 , name : "Ice Lake-LP Dynamic Tuning Processor Participant" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34752u16 , name : "UHD Graphics 617" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34762u16 , name : "UHD Graphics 617" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34816u16 , name : "Platform Controller Hub EG20T PCI Express Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34817u16 , name : "Platform Controller Hub EG20T Packet Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34818u16 , name : "Platform Controller Hub EG20T Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34819u16 , name : "Platform Controller Hub EG20T General Purpose IO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34820u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34821u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34822u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34823u16 , name : "Platform Controller Hub EG20T USB2 EHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34824u16 , name : "Platform Controller Hub EG20T USB Client Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34825u16 , name : "Platform Controller Hub EG20T SDIO Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34826u16 , name : "Platform Controller Hub EG20T SDIO Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34827u16 , name : "Platform Controller Hub EG20T SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34828u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34829u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34830u16 , name : "Platform Controller Hub EG20T USB OHCI Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34831u16 , name : "Platform Controller Hub EG20T USB2 EHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34832u16 , name : "Platform Controller Hub EG20T DMA Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34833u16 , name : "Platform Controller Hub EG20T UART Controller 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34834u16 , name : "Platform Controller Hub EG20T UART Controller 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34835u16 , name : "Platform Controller Hub EG20T UART Controller 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34836u16 , name : "Platform Controller Hub EG20T UART Controller 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34837u16 , name : "Platform Controller Hub EG20T DMA Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34838u16 , name : "Platform Controller Hub EG20T Serial Peripheral Interface Bus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34839u16 , name : "Platform Controller Hub EG20T I2C Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34840u16 , name : "Platform Controller Hub EG20T Controller Area Network (CAN) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 34841u16 , name : "Platform Controller Hub EG20T IEEE 1588 Hardware Assist" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35331u16 , name : "Processor Power and Thermal Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35341u16 , name : "Ice Lake Thunderbolt 3 NHI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35345u16 , name : "GNA Scoring Accelerator Module" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35346u16 , name : "Ice Lake-LP Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35347u16 , name : "Ice Lake Thunderbolt 3 USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35348u16 , name : "Ice Lake Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35351u16 , name : "Ice Lake Thunderbolt 3 NHI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35353u16 , name : "Image Signal Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35357u16 , name : "Ice Lake Thunderbolt 3 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35359u16 , name : "Ice Lake Thunderbolt 3 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35361u16 , name : "Ice Lake Thunderbolt 3 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35363u16 , name : "Ice Lake Thunderbolt 3 PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35409u16 , name : "Iris Plus Graphics G7 (Ice Lake)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35410u16 , name : "Iris Plus Graphics G7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35411u16 , name : "Iris Plus Graphics G7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35414u16 , name : "Iris Plus Graphics G1 (Ice Lake)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35416u16 , name : "Ice Lake-Y GT1 [UHD Graphics G1]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35418u16 , name : "Iris Plus Graphics G4 (Ice Lake)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35420u16 , name : "Iris Plus Graphics G4 (Ice Lake)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35840u16 , name : "8 Series/C220 Series Chipset Family 4-port SATA Controller 1 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35841u16 , name : "8 Series Chipset Family 4-port SATA Controller 1 [IDE mode] - Mobile" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35842u16 , name : "8 Series/C220 Series Chipset Family 6-port SATA Controller 1 [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35843u16 , name : "8 Series/C220 Series Chipset Family 6-port SATA Controller 1 [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" }] } , Device { vendor_id : 32902u16 , id : 35844u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35845u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35846u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35847u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35848u16 , name : "8 Series/C220 Series Chipset Family 2-port SATA Controller 2 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35849u16 , name : "8 Series/C220 Series Chipset Family 2-port SATA Controller 2 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35854u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35855u16 , name : "8 Series/C220 Series Chipset Family SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35856u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 4163u16 , subdevice : 34100u16 , name : "ASUS H81I-PLUS" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" }] } , Device { vendor_id : 32902u16 , id : 35857u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35858u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" }] } , Device { vendor_id : 32902u16 , id : 35859u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35860u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35861u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35862u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35863u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35864u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35865u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35866u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #6" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" }] } , Device { vendor_id : 32902u16 , id : 35867u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35868u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35869u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35870u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35871u16 , name : "8 Series/C220 Series Chipset Family PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35872u16 , name : "8 Series/C220 Series Chipset High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35873u16 , name : "8 Series/C220 Series Chipset High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35874u16 , name : "8 Series/C220 Series Chipset Family SMBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35875u16 , name : "8 Series Chipset Family CHAP Counters" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35876u16 , name : "8 Series Chipset Family Thermal Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35878u16 , name : "8 Series/C220 Series Chipset Family USB EHCI #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 8720u16 , name : "ThinkPad T540p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" } , SubSystem { subvendor : 8720u16 , subdevice : 6058u16 , name : "ThinkPad T540p" }] } , Device { vendor_id : 32902u16 , id : 35885u16 , name : "8 Series/C220 Series Chipset Family USB EHCI #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35889u16 , name : "8 Series/C220 Series Chipset Family USB xHCI" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35891u16 , name : "8 Series/C220 Series Chipset Family LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35892u16 , name : "8 Series/C220 Series Chipset Family NAND Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35898u16 , name : "8 Series/C220 Series Chipset Family MEI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35899u16 , name : "8 Series/C220 Series Chipset Family MEI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35900u16 , name : "8 Series/C220 Series Chipset Family IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35901u16 , name : "8 Series/C220 Series Chipset Family KT Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" }] } , Device { vendor_id : 32902u16 , id : 35904u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35905u16 , name : "8 Series Chipset Family Mobile Super SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35906u16 , name : "8 Series/C220 Series Chipset Family Desktop Super SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35907u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35908u16 , name : "Z87 Express LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35909u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35910u16 , name : "Z85 Express LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35911u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35912u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35913u16 , name : "HM86 Express LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35914u16 , name : "H87 Express LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" }] } , Device { vendor_id : 32902u16 , id : 35915u16 , name : "HM87 Express LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35916u16 , name : "Q85 Express LPC Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12447u16 , name : "ThinkCentre M83" }] } , Device { vendor_id : 32902u16 , id : 35917u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35918u16 , name : "Q87 Express LPC Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6552u16 , name : "EliteDesk 800 G1" }] } , Device { vendor_id : 32902u16 , id : 35919u16 , name : "QM87 Express LPC Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" } , SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" }] } , Device { vendor_id : 32902u16 , id : 35920u16 , name : "B85 Express LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35921u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35922u16 , name : "C222 Series Chipset Family Server Essential SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35923u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35924u16 , name : "C224 Series Chipset Family Server Standard SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35925u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35926u16 , name : "C226 Series Chipset Family Server Advanced SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35927u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35928u16 , name : "8 Series/C220 Series Chipset Family WS SKU LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35929u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35930u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35931u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35932u16 , name : "H81 Express LPC Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" }] } , Device { vendor_id : 32902u16 , id : 35933u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35934u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35935u16 , name : "8 Series/C220 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35968u16 , name : "9 Series Chipset Family SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35969u16 , name : "9 Series Chipset Family SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35970u16 , name : "9 Series Chipset Family SATA Controller [AHCI Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35971u16 , name : "9 Series Chipset Family SATA Controller [AHCI Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35972u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35973u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35974u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35975u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35976u16 , name : "9 Series Chipset Family SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35977u16 , name : "9 Series Chipset Family SATA Controller [IDE Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35982u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35983u16 , name : "9 Series Chipset Family SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35984u16 , name : "9 Series Chipset Family PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35986u16 , name : "9 Series Chipset Family PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35988u16 , name : "9 Series Chipset Family PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35990u16 , name : "9 Series Chipset Family PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35992u16 , name : "9 Series Chipset Family PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35994u16 , name : "9 Series Chipset Family PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35996u16 , name : "9 Series Chipset Family PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 35998u16 , name : "9 Series Chipset Family PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36000u16 , name : "9 Series Chipset Family HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36002u16 , name : "9 Series Chipset Family SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36004u16 , name : "9 Series Chipset Family Thermal Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36006u16 , name : "9 Series Chipset Family USB EHCI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36013u16 , name : "9 Series Chipset Family USB EHCI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36017u16 , name : "9 Series Chipset Family USB xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36019u16 , name : "9 Series Chipset Family LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36026u16 , name : "9 Series Chipset Family ME Interface #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36027u16 , name : "9 Series Chipset Family ME Interface #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36028u16 , name : "9 Series Chipset Family IDE-R Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36029u16 , name : "9 Series Chipset Family KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36033u16 , name : "9 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36034u16 , name : "9 Series Chipset Family LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36035u16 , name : "HM97 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36036u16 , name : "Z97 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36037u16 , name : "QM97 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36038u16 , name : "H97 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36096u16 , name : "C610/X99 series chipset 4-port SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36098u16 , name : "C610/X99 series chipset 6-Port SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36100u16 , name : "C610/X99 series chipset SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36102u16 , name : "C610/X99 series chipset SATA Controller [RAID mode]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 4145u16 , name : "ThinkServer RAID 110i" }] } , Device { vendor_id : 32902u16 , id : 36104u16 , name : "C610/X99 series chipset 2-port SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36110u16 , name : "C610/X99 series chipset SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36112u16 , name : "C610/X99 series chipset PCI Express Root Port #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36113u16 , name : "C610/X99 series chipset PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36114u16 , name : "C610/X99 series chipset PCI Express Root Port #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" }] } , Device { vendor_id : 32902u16 , id : 36115u16 , name : "C610/X99 series chipset PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36116u16 , name : "C610/X99 series chipset PCI Express Root Port #3" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36117u16 , name : "C610/X99 series chipset PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36118u16 , name : "C610/X99 series chipset PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36119u16 , name : "C610/X99 series chipset PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36120u16 , name : "C610/X99 series chipset PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36121u16 , name : "C610/X99 series chipset PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36122u16 , name : "C610/X99 series chipset PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36123u16 , name : "C610/X99 series chipset PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36124u16 , name : "C610/X99 series chipset PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36125u16 , name : "C610/X99 series chipset PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36126u16 , name : "C610/X99 series chipset PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36127u16 , name : "C610/X99 series chipset PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36128u16 , name : "C610/X99 series chipset HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" }] } , Device { vendor_id : 32902u16 , id : 36129u16 , name : "C610/X99 series chipset HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36130u16 , name : "C610/X99 series chipset SMBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36132u16 , name : "C610/X99 series chipset Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36134u16 , name : "C610/X99 series chipset USB Enhanced Host Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36141u16 , name : "C610/X99 series chipset USB Enhanced Host Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36145u16 , name : "C610/X99 series chipset USB xHCI Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36147u16 , name : "C610/X99 series chipset LAN Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36148u16 , name : "C610/X99 series chipset NAND Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36154u16 , name : "C610/X99 series chipset MEI Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36155u16 , name : "C610/X99 series chipset MEI Controller #2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36156u16 , name : "C610/X99 series chipset IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36157u16 , name : "C610/X99 series chipset KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36160u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36161u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36162u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36163u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36164u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36165u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36166u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36167u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36168u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36169u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36170u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36171u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36172u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36173u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36174u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36175u16 , name : "C610/X99 series chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36192u16 , name : "C610/X99 series chipset sSATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36194u16 , name : "C610/X99 series chipset sSATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36196u16 , name : "C610/X99 series chipset sSATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36198u16 , name : "C610/X99 series chipset sSATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36200u16 , name : "C610/X99 series chipset sSATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36206u16 , name : "C610/X99 series chipset sSATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36220u16 , name : "C610/X99 series chipset SPSR" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1559u16 , name : "Precision T5810" } , SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36221u16 , name : "C610/X99 series chipset MS SMBus 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36222u16 , name : "C610/X99 series chipset MS SMBus 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36223u16 , name : "C610/X99 series chipset MS SMBus 2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" }] } , Device { vendor_id : 32902u16 , id : 36864u16 , name : "IXP2000 Family Network Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36865u16 , name : "IXP2400 Network Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36866u16 , name : "IXP2300 Network Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 36868u16 , name : "IXP2800 Network Processor" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38433u16 , name : "Integrated RAID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38434u16 , name : "Integrated RAID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38465u16 , name : "Integrated RAID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38561u16 , name : "Integrated RAID" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38976u16 , name : "Lakefield GT2 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 38977u16 , name : "Lakefield GT1.5 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39425u16 , name : "11th Gen Core Processor PCIe Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39426u16 , name : "Tiger Lake-UP4 2 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39427u16 , name : "TigerLake-LP Dynamic Tuning Processor Participant" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39428u16 , name : "Tiger Lake-UP3 2 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39429u16 , name : "11th Gen Core Processor PCIe Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39431u16 , name : "11th Gen Core Processor PCIe Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39433u16 , name : "11th Gen Core Processor PCIe Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39435u16 , name : "Volume Management Device NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39437u16 , name : "Tigerlake Telemetry Aggregator Driver" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39439u16 , name : "11th Gen Core Processor PCIe Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39441u16 , name : "GNA Scoring Accelerator module" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39442u16 , name : "Tiger Lake-UP4 4 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39443u16 , name : "Tiger Lake-LP Thunderbolt 4 USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39444u16 , name : "Tiger Lake-UP3/H35 4 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39445u16 , name : "Tiger Lake xDCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39447u16 , name : "Tiger Lake-H Thunderbolt 4 USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39449u16 , name : "Tiger Lake IPU6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39451u16 , name : "Tiger Lake-LP Thunderbolt 4 NHI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39453u16 , name : "Tiger Lake-LP Thunderbolt 4 NHI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39455u16 , name : "Tiger Lake-H Thunderbolt 4 NHI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39457u16 , name : "Tiger Lake-H Thunderbolt 4 NHI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39459u16 , name : "Tiger Lake-LP Thunderbolt 4 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39461u16 , name : "Tiger Lake-LP Thunderbolt 4 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39462u16 , name : "Tiger Lake-H 6 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39463u16 , name : "Tiger Lake-LP Thunderbolt 4 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39465u16 , name : "Tiger Lake-LP Thunderbolt 4 PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39467u16 , name : "Tiger Lake-H Thunderbolt 4 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39469u16 , name : "Tiger Lake-H Thunderbolt 4 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39471u16 , name : "Tiger Lake-H Thunderbolt 4 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39473u16 , name : "Tiger Lake-H Thunderbolt 4 PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39475u16 , name : "Tiger Lake Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39478u16 , name : "Tiger Lake-H 8 cores Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39481u16 , name : "Tiger Lake H IPU6-Slim" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39488u16 , name : "Tiger Lake-UP4 GT2 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39497u16 , name : "TigerLake-LP GT2 [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39520u16 , name : "TigerLake-H GT1 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39528u16 , name : "TigerLake-H GT1 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39536u16 , name : "Tiger Lake-H GT1 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39544u16 , name : "Tiger Lake-LP GT2 [UHD Graphics G4]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39713u16 , name : "Comet Lake-U GT2 [UHD Graphics 620]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39731u16 , name : "Comet Lake-S 6c Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39745u16 , name : "CometLake-U GT2 [UHD Graphics]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2493u16 , name : "Latitude 7310" } , SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 39747u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39748u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39761u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39763u16 , name : "Comet Lake-S 6c Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39764u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39777u16 , name : "Comet Lake-U v1 4c Host Bridge/DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" }] } , Device { vendor_id : 32902u16 , id : 39779u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39780u16 , name : "10th Gen Core Processor Host Bridge/DRAM Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39844u16 , name : "Comet Lake-H GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39848u16 , name : "CometLake-S GT1 [UHD Graphics 610]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39850u16 , name : "Comet Lake UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39852u16 , name : "Comet Lake UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39876u16 , name : "CometLake-H GT2 [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39877u16 , name : "CometLake-S GT2 [UHD Graphics 630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39878u16 , name : "Comet Lake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39880u16 , name : "CometLake-S GT2 [UHD Graphics 630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39882u16 , name : "Comet Lake UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39884u16 , name : "Comet Lake UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39910u16 , name : "Comet Lake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39926u16 , name : "Coffee Lake-S GT2 [UHD Graphics P630]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39936u16 , name : "8 Series SATA Controller 1 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39937u16 , name : "8 Series SATA Controller 1 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39938u16 , name : "8 Series SATA Controller 1 [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39939u16 , name : "8 Series SATA Controller 1 [AHCI mode]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 39940u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39941u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39942u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39943u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39944u16 , name : "8 Series SATA Controller 2 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39945u16 , name : "8 Series SATA Controller 2 [IDE mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39946u16 , name : "8 Series SATA Controller [Reserved]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39947u16 , name : "8 Series SATA Controller [Reserved]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39948u16 , name : "8 Series SATA Controller [Reserved]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39949u16 , name : "8 Series SATA Controller [Reserved]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39950u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39951u16 , name : "8 Series SATA Controller 1 [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39952u16 , name : "8 Series PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39953u16 , name : "8 Series PCI Express Root Port 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39954u16 , name : "8 Series PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39955u16 , name : "8 Series PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39956u16 , name : "8 Series PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39957u16 , name : "8 Series PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39958u16 , name : "8 Series PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39959u16 , name : "8 Series PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39960u16 , name : "8 Series PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39961u16 , name : "8 Series PCI Express Root Port 5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39962u16 , name : "8 Series PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39963u16 , name : "8 Series PCI Express Root Port 6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39964u16 , name : "8 Series PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39965u16 , name : "8 Series PCI Express Root Port 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39966u16 , name : "8 Series PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39967u16 , name : "8 Series PCI Express Root Port 8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39968u16 , name : "8 Series HD Audio Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 39969u16 , name : "8 Series HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39970u16 , name : "8 Series SMBus Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 39971u16 , name : "8 Series CHAP Counters" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39972u16 , name : "8 Series Thermal" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39974u16 , name : "8 Series USB EHCI #1" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8716u16 , name : "T440s" } , SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 39981u16 , name : "8 Series USB EHCI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39985u16 , name : "8 Series USB xHCI HC" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" } , SubSystem { subvendor : 32902u16 , subdevice : 29296u16 , name : "Apple MacBookAir6,2 / MacBookPro11,1" }] } , Device { vendor_id : 32902u16 , id : 39989u16 , name : "8 Series SDIO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39990u16 , name : "8 Series Audio DSP Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39994u16 , name : "8 Series HECI #0" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 39995u16 , name : "8 Series HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39996u16 , name : "8 Series HECI IDER" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 39997u16 , name : "8 Series HECI KT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40000u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40001u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40002u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40003u16 , name : "8 Series LPC Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 32902u16 , id : 40004u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40005u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40006u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40007u16 , name : "8 Series LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40032u16 , name : "8 Series Low Power Sub-System DMA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40033u16 , name : "8 Series I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40034u16 , name : "8 Series I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40035u16 , name : "8 Series UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40036u16 , name : "8 Series UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40037u16 , name : "8 Series SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40038u16 , name : "8 Series SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40067u16 , name : "Wildcat Point-LP SATA Controller [AHCI Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40069u16 , name : "Wildcat Point-LP SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40071u16 , name : "Wildcat Point-LP SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40079u16 , name : "Wildcat Point-LP SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40080u16 , name : "Wildcat Point-LP PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40082u16 , name : "Wildcat Point-LP PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40084u16 , name : "Wildcat Point-LP PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40086u16 , name : "Wildcat Point-LP PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40088u16 , name : "Wildcat Point-LP PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40090u16 , name : "Wildcat Point-LP PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40096u16 , name : "Wildcat Point-LP High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40098u16 , name : "Wildcat Point-LP SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40100u16 , name : "Wildcat Point-LP Thermal Management Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40102u16 , name : "Wildcat Point-LP USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40113u16 , name : "Wildcat Point-LP USB xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40117u16 , name : "Wildcat Point-LP Secure Digital IO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40118u16 , name : "Wildcat Point-LP Smart Sound Technology Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40122u16 , name : "Wildcat Point-LP MEI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40123u16 , name : "Wildcat Point-LP MEI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40124u16 , name : "Wildcat Point-LP IDE-r Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40125u16 , name : "Wildcat Point-LP KT Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40129u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40130u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40131u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40133u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40134u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40135u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40137u16 , name : "Wildcat Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40160u16 , name : "Wildcat Point-LP Serial IO DMA Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40161u16 , name : "Wildcat Point-LP Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40162u16 , name : "Wildcat Point-LP Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40163u16 , name : "Wildcat Point-LP Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40164u16 , name : "Wildcat Point-LP Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40165u16 , name : "Wildcat Point-LP Serial IO GSPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40166u16 , name : "Wildcat Point-LP Serial IO GSPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40195u16 , name : "Sunrise Point-LP SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40208u16 , name : "Sunrise Point-LP PCI Express Root Port #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" }] } , Device { vendor_id : 32902u16 , id : 40209u16 , name : "Sunrise Point-LP PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40210u16 , name : "Sunrise Point-LP PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40211u16 , name : "Sunrise Point-LP PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40212u16 , name : "Sunrise Point-LP PCI Express Root Port #5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40213u16 , name : "Sunrise Point-LP PCI Express Root Port #6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40214u16 , name : "Sunrise Point-LP PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40215u16 , name : "Sunrise Point-LP PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40216u16 , name : "Sunrise Point-LP PCI Express Root Port #9" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40217u16 , name : "Sunrise Point-LP PCI Express Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40218u16 , name : "Sunrise Point-LP PCI Express Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40225u16 , name : "Sunrise Point-LP PMC" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40227u16 , name : "Sunrise Point-LP SMBus" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40231u16 , name : "Sunrise Point-LP Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40232u16 , name : "Sunrise Point-LP Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40233u16 , name : "Sunrise Point-LP Serial IO SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40234u16 , name : "Sunrise Point-LP Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40237u16 , name : "Sunrise Point-LP Secure Digital IO Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40239u16 , name : "Sunrise Point-LP USB 3.0 xHCI Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40241u16 , name : "Sunrise Point-LP Thermal subsystem" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40242u16 , name : "CSI-2 Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 40245u16 , name : "Sunrise Point-LP Integrated Sensor Hub" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 40250u16 , name : "Sunrise Point-LP CSME HECI #1" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40253u16 , name : "Sunrise Point-LP Active Management Technology - SOL" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" }] } , Device { vendor_id : 32902u16 , id : 40254u16 , name : "iTouch Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40259u16 , name : "Sunrise Point-LP LPC Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40262u16 , name : "LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 40264u16 , name : "Sunrise Point-LP LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" }] } , Device { vendor_id : 32902u16 , id : 40270u16 , name : "Sunrise Point LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 40272u16 , name : "Sunrise Point LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40278u16 , name : "Sunrise Point-LP LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40280u16 , name : "Sunrise Point-LP LPC Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" }] } , Device { vendor_id : 32902u16 , id : 40288u16 , name : "Sunrise Point-LP Serial IO I2C Controller #0" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4447u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" } , SubSystem { subvendor : 32902u16 , subdevice : 40288u16 , name : "100 Series PCH/Sunrise Point PCH I2C0 [Skylake/Kaby Lake LPSS I2C]" }] } , Device { vendor_id : 32902u16 , id : 40289u16 , name : "Sunrise Point-LP Serial IO I2C Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" }] } , Device { vendor_id : 32902u16 , id : 40290u16 , name : "Sunrise Point-LP Serial IO I2C Controller #2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" }] } , Device { vendor_id : 32902u16 , id : 40291u16 , name : "Sunrise Point-LP Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40292u16 , name : "Sunrise Point-LP Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40293u16 , name : "Sunrise Point-LP Serial IO I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40294u16 , name : "Sunrise Point-LP Serial IO UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40304u16 , name : "Sunrise Point-LP HD Audio" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 6058u16 , subdevice : 8775u16 , name : "ThinkPad T570" } , SubSystem { subvendor : 6058u16 , subdevice : 14378u16 , name : "B51-80 Laptop" }] } , Device { vendor_id : 32902u16 , id : 40305u16 , name : "Sunrise Point-LP HD Audio" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4244u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" } , SubSystem { subvendor : 6058u16 , subdevice : 8797u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 32902u16 , id : 40324u16 , name : "Cannon Point-LP LPC Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 40355u16 , name : "Cannon Point-LP SMBus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40356u16 , name : "Cannon Point-LP SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40360u16 , name : "Cannon Point-LP Serial IO UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40362u16 , name : "Cannon Point-LP Serial IO SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40363u16 , name : "Cannon Point-LP Serial IO SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40368u16 , name : "Cannon Point-LP PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40369u16 , name : "Cannon Point-LP PCI Express Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40370u16 , name : "Cannon Point-LP PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40371u16 , name : "Cannon Point-LP PCI Express Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40372u16 , name : "Cannon Point-LP PCI Express Root Port #13" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 40373u16 , name : "Cannon Point-LP PCI Express Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40374u16 , name : "Cannon Point-LP PCI Express Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40376u16 , name : "Cannon Point-LP PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40380u16 , name : "Cannon Point-LP PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40381u16 , name : "Cannon Point-LP PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40382u16 , name : "Cannon Point-LP PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40383u16 , name : "Cannon Point PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40388u16 , name : "Cannon Point-LP SD Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40389u16 , name : "Cannon Point-LP Serial IO I2C Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40392u16 , name : "Cannon Point-LP High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 40403u16 , name : "Cannon Point-LP SATA Controller [AHCI Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40416u16 , name : "Cannon Point-LP MEI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40419u16 , name : "Cannon Point-LP Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40424u16 , name : "Cannon Point-LP Serial IO I2C Controller #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 40425u16 , name : "Cannon Point-LP Serial IO I2C Controller #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2206u16 , name : "Inspiron 5482" }] } , Device { vendor_id : 32902u16 , id : 40429u16 , name : "Cannon Point-LP USB 3.1 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40431u16 , name : "Cannon Point-LP Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40432u16 , name : "Cannon Point-LP CNVi [Wireless-AC]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40437u16 , name : "BayHubTech Integrated SD controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40441u16 , name : "Cannon Point-LP Thermal Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40444u16 , name : "Cannon Point-LP Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40960u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx DMI Bridge" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 40961u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "GA-D525TUD" } , SubSystem { subvendor : 32902u16 , subdevice : 20301u16 , name : "DeskTop Board D510MO" } , SubSystem { subvendor : 32902u16 , subdevice : 21579u16 , name : "Desktop Board D425KT" }] } , Device { vendor_id : 32902u16 , id : 40962u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx Integrated Graphics Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40963u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx CHAPS counter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 40976u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx DMI Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33708u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" }] } , Device { vendor_id : 32902u16 , id : 40977u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33708u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" }] } , Device { vendor_id : 32902u16 , id : 40978u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx Integrated Graphics Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33708u16 , name : "Eee PC 1015PX" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" }] } , Device { vendor_id : 32902u16 , id : 40979u16 , name : "Atom Processor D4xx/D5xx/N4xx/N5xx CHAPS counter" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41090u16 , name : "500 Series Chipset Family On-Package eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41120u16 , name : "500 Series Chipset Family On-Package Primary-to-Sideband Bridge (P2SB)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41121u16 , name : "500 Series Chipset Family On-Package Power Management Controller (PMC)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41123u16 , name : "500 Series Chipset Family On-Package System Management Bus (SMBus)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41124u16 , name : "500 Series Chipset Family On-Package SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41126u16 , name : "500 Series Chipset Family On-Package Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41128u16 , name : "500 Series Chipset Family On-Package UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41129u16 , name : "500 Series Chipset Family On-Package UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41130u16 , name : "500 Series Chipset Family On-Package Generic SPI (GSPI) #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41131u16 , name : "500 Series Chipset Family On-Package Generic SPI (GSPI) #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41136u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41137u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41138u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41139u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41144u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41145u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41146u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41147u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41148u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41149u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41150u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41151u16 , name : "500 Series Chipset Family On-Package PCI Express Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41157u16 , name : "500 Series Chipset Family On-Package I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41158u16 , name : "500 Series Chipset Family On-Package I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41159u16 , name : "500 Series Chipset Family On-Package UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41160u16 , name : "500 Series Chipset Family On-Package High Definition Audio (HD Audio)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41168u16 , name : "500 Series Chipset Family On-Package Touch Host Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41169u16 , name : "500 Series Chipset Family On-Package Touch Host Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41171u16 , name : "500 Series Chipset Family On-Package SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41173u16 , name : "500 Series Chipset Family On-Package SATA Controller (RAID 0/1/5/10) no premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41175u16 , name : "500 Series Chipset Family On-Package SATA Controller (RAID 0/1/5/10) premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41178u16 , name : "500 Series Chipset Family On-Package UART Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41184u16 , name : "500 Series Chipset Family On-Package CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41185u16 , name : "500 Series Chipset Family On-Package CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41186u16 , name : "500 Series Chipset Family On-Package CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41187u16 , name : "500 Series Chipset Family On-Package CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41188u16 , name : "500 Series Chipset Family On-Package CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41189u16 , name : "500 Series Chipset Family On-Package CSME HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41192u16 , name : "500 Series Chipset Family On-Package I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41193u16 , name : "500 Series Chipset Family On-Package I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41194u16 , name : "500 Series Chipset Family On-Package I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41195u16 , name : "500 Series Chipset Family On-Package I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41197u16 , name : "500 Series Chipset Family On-Package USB 3.2 Gen 2x1 (10 Gbs) xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41198u16 , name : "500 Series Chipset Family On-Package USB 3.2 Gen 1x1 (5 Gbs) Device Controller (xDCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41199u16 , name : "500 Series Chipset Family On-Package Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41200u16 , name : "Wi-Fi 6 AX201" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 580u16 , name : "Wi-Fi 6 AX101NGW" }] } , Device { vendor_id : 32902u16 , id : 41211u16 , name : "500 Series Chipset Family On-Package Generic SPI (GSPI) #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41212u16 , name : "500 Series Chipset Family On-Package Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41213u16 , name : "500 Series Chipset Family On-Package Generic SPI (GSPI) #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41218u16 , name : "Q170/Q150/B150/H170/H110/Z170/CM236 Chipset SATA Controller [AHCI Mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41219u16 , name : "HM170/QM170 Chipset SATA Controller [AHCI Mode]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 32902u16 , id : 41221u16 , name : "Sunrise Point-H SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41222u16 , name : "Q170/H170/Z170/CM236 Chipset SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41223u16 , name : "HM170/QM170 Chipset SATA Controller [RAID Mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41231u16 , name : "Sunrise Point-H SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41232u16 , name : "100/C230 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41233u16 , name : "100/C230 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41234u16 , name : "100/C230 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41235u16 , name : "100/C230 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41236u16 , name : "100/C230 Series Chipset Family PCIe Root Port #5" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" }] } , Device { vendor_id : 32902u16 , id : 41237u16 , name : "100/C230 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41238u16 , name : "100/C230 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41239u16 , name : "100/C230 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41240u16 , name : "100/C230 Series Chipset Family PCIe Root Port #9" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" }] } , Device { vendor_id : 32902u16 , id : 41241u16 , name : "100/C230 Series Chipset Family PCIe Root Port #10" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" }] } , Device { vendor_id : 32902u16 , id : 41242u16 , name : "100/C230 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41243u16 , name : "100/C230 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41244u16 , name : "100/C230 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41245u16 , name : "100/C230 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41246u16 , name : "100/C230 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41247u16 , name : "100/C230 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41248u16 , name : "100/C230 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41249u16 , name : "100/C230 Series Chipset Family PMC" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41250u16 , name : "Sunrise Point-H cAVS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41251u16 , name : "100/C230 Series Chipset Family SMBus" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41252u16 , name : "100/C230 Series Chipset Family SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41253u16 , name : "100/C230 Series Chipset Family GbE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41254u16 , name : "100/C230 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41255u16 , name : "100/C230 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41256u16 , name : "100/C230 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41257u16 , name : "100/C230 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41258u16 , name : "100/C230 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41263u16 , name : "100/C230 Series Chipset Family USB 3.0 xHCI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41264u16 , name : "100/C230 Series Chipset Family USB Device Controller (OTG)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41265u16 , name : "100/C230 Series Chipset Family Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41267u16 , name : "Sunrise Point-H Northpeak ACPI Function" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41269u16 , name : "100/C230 Series Chipset Family ISH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41274u16 , name : "100/C230 Series Chipset Family MEI #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41275u16 , name : "100/C230 Series Chipset Family MEI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41276u16 , name : "100/C230 Series Chipset Family IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41277u16 , name : "100/C230 Series Chipset Family Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41278u16 , name : "100/C230 Series Chipset Family MEI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41280u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41281u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41282u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41283u16 , name : "H110 Chipset LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34452u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31124u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41284u16 , name : "H170 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41285u16 , name : "Z170 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41286u16 , name : "Q170 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41287u16 , name : "Q150 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41288u16 , name : "B150 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41289u16 , name : "C236 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41290u16 , name : "C232 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41291u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41292u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41293u16 , name : "QM170 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41294u16 , name : "HM170 Chipset LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 32902u16 , id : 41295u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41296u16 , name : "CM236 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41297u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41298u16 , name : "HM175 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41299u16 , name : "QM175 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41300u16 , name : "CM238 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41301u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41302u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41303u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41304u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41305u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41306u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41307u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41308u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41309u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41310u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41311u16 , name : "Sunrise Point-H LPC Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41312u16 , name : "100/C230 Series Chipset Family I2C #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 32902u16 , id : 41313u16 , name : "100/C230 Series Chipset Family I2C #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" }] } , Device { vendor_id : 32902u16 , id : 41314u16 , name : "100/C230 Series Chipset Family I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41315u16 , name : "100/C230 Series Chipset Family I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41318u16 , name : "100/C230 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41319u16 , name : "100/C230 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41320u16 , name : "100/C230 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41321u16 , name : "100/C230 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41322u16 , name : "100/C230 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41328u16 , name : "100/C230 Series Chipset Family HD Audio" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4163u16 , subdevice : 34503u16 , name : "H110I-PLUS Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 63892u16 , name : "H110M ECO/GAMING" }] } , Device { vendor_id : 32902u16 , id : 41329u16 , name : "HM175/QM175/CM238 HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41346u16 , name : "C620 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41350u16 , name : "C620 Series Chipset Family SATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41360u16 , name : "C620 Series Chipset Family PCIe Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41361u16 , name : "C620 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41362u16 , name : "C620 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41363u16 , name : "C620 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41364u16 , name : "C620 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41365u16 , name : "C620 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41366u16 , name : "C620 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41367u16 , name : "C620 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41368u16 , name : "C620 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41369u16 , name : "C620 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41370u16 , name : "C620 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41371u16 , name : "C620 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41372u16 , name : "C620 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41373u16 , name : "C620 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41374u16 , name : "C620 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41375u16 , name : "C620 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41376u16 , name : "C620 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41377u16 , name : "C620 Series Chipset Family PMC" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41378u16 , name : "C620 Series Chipset Family cAVS" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41379u16 , name : "C620 Series Chipset Family SMBus" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41380u16 , name : "C620 Series Chipset Family SPI Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41382u16 , name : "C620 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41391u16 , name : "C620 Series Chipset Family USB 3.0 xHCI Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41393u16 , name : "C620 Series Chipset Family Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41402u16 , name : "C620 Series Chipset Family MEI HECI #1" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41403u16 , name : "C620 Series Chipset Family MEI HECI #2" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41404u16 , name : "C620 Series Chipset Family MEI IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41405u16 , name : "C620 Series Chipset Family MEI Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41406u16 , name : "C620 Series Chipset Family MEI HECI #3" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41409u16 , name : "C621 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41410u16 , name : "C622 Chipset LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2397u16 , name : "X11SPM-TF" }] } , Device { vendor_id : 32902u16 , id : 41411u16 , name : "C624 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41412u16 , name : "C625 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41413u16 , name : "C626 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41414u16 , name : "C627 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41415u16 , name : "C628 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41418u16 , name : "C629 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41426u16 , name : "C620 Series Chipset Family SSATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41430u16 , name : "C620 Series Chipset Family SSATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41447u16 , name : "C620 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41448u16 , name : "C620 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41449u16 , name : "C620 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41450u16 , name : "C620 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41452u16 , name : "C620 Series Chipset Family MROM 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41453u16 , name : "C620 Series Chipset Family MROM 1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41456u16 , name : "C620 Series Chipset Family HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41464u16 , name : "C620 Series Chipset Family IE HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41465u16 , name : "C620 Series Chipset Family IE HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41466u16 , name : "C620 Series Chipset Family IE IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41467u16 , name : "C620 Series Chipset Family IE Keyboard and Text Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41468u16 , name : "C620 Series Chipset Family IE HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41474u16 , name : "C620 Series Chipset Family (Super) SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41478u16 , name : "C620 Series Chipset Family (Super) SATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41488u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41489u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41490u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41491u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41492u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41493u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41494u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41495u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41496u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41497u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41498u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41499u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41500u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41501u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41502u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41503u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41504u16 , name : "C620 Series Chipset Family (Super) P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41505u16 , name : "C620 Series Chipset Family (Super) PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41507u16 , name : "C620 Series Chipset Family (Super) SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41508u16 , name : "C620 Series Chipset Family (Super) SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41510u16 , name : "C620 Series Chipset Family (Super) Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41519u16 , name : "C620 Series Chipset Family (Super) USB 3.0 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41521u16 , name : "C620 Series Chipset Family (Super) Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41530u16 , name : "C620 Series Chipset Family (Super) MEI HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41531u16 , name : "C620 Series Chipset Family (Super) MEI HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41532u16 , name : "C620 Series Chipset Family (Super) MEI IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41533u16 , name : "C620 Series Chipset Family (Super) MEI Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41534u16 , name : "C620 Series Chipset Family (Super) MEI HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41538u16 , name : "C624 Chipset (Super) LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41539u16 , name : "C627 Chipset (Super) LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41540u16 , name : "C621 Chipset (Super) LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41541u16 , name : "C627 Chipset (Super) LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41542u16 , name : "C628 Chipset (Super) LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41554u16 , name : "C620 Series Chipset Family (Super) SSATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41558u16 , name : "C620 Series Chipset Family (Super) SSATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41575u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41576u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41577u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41578u16 , name : "C620 Series Chipset Family (Super) PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41580u16 , name : "C620 Series Chipset Family (Super) MROM 0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41584u16 , name : "C620 Series Chipset Family (Super) Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41592u16 , name : "C620 Series Chipset Family (Super) IE HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41593u16 , name : "C620 Series Chipset Family (Super) IE HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41594u16 , name : "C620 Series Chipset Family (Super) IE IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41595u16 , name : "C620 Series Chipset Family (Super) IE Keyboard and Text Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41596u16 , name : "C620 Series Chipset Family (Super) IE HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41602u16 , name : "200 Series/Z370 Chipset Family SATA Controller (AHCI)" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41606u16 , name : "200 Series/Z370 Chipset Family SATA Controller (RAID) Premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41614u16 , name : "200 Series/Z370 Chipset Family SATA Controller (RST and Optane Technology)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41616u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41617u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41618u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41619u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41620u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #5" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41621u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41622u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #7" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41623u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41624u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #9" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41625u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41626u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41627u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41628u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41629u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41630u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41631u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41632u16 , name : "200 Series/Z370 Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41633u16 , name : "200 Series/Z370 Chipset Family PMC" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41635u16 , name : "200 Series/Z370 Chipset Family SMBus" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41636u16 , name : "200 Series/Z370 Chipset Family SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41637u16 , name : "200 Series/Z370 Chipset Family GbE Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41638u16 , name : "200 Series/Z370 Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41639u16 , name : "200 Series/Z370 Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41640u16 , name : "200 Series/Z370 Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41641u16 , name : "200 Series/Z370 Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41642u16 , name : "200 Series/Z370 Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41647u16 , name : "200 Series/Z370 Chipset Family USB 3.0 xHCI Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41648u16 , name : "200 Series/Z370 Chipset Family USB Device Controller (OTG)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41649u16 , name : "200 Series/Z370 Chipset Family Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41653u16 , name : "200 Series/Z370 Chipset Family ISH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41658u16 , name : "200 Series/Z370 Chipset Family MEI #1" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41659u16 , name : "200 Series/Z370 Chipset Family MEI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41660u16 , name : "200 Series/Z370 Chipset Family IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41661u16 , name : "200 Series/Z370 Chipset Family Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41662u16 , name : "200 Series/Z370 Chipset Family MEI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41668u16 , name : "H270 Chipset LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41669u16 , name : "Z270 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41670u16 , name : "Q270 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41671u16 , name : "Q250 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41672u16 , name : "B250 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41673u16 , name : "Z370 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41682u16 , name : "X299 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41683u16 , name : "C422 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41696u16 , name : "200 Series/Z370 Chipset Family I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41697u16 , name : "200 Series/Z370 Chipset Family I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41698u16 , name : "200 Series/Z370 Chipset Family I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41699u16 , name : "200 Series/Z370 Chipset Family I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41702u16 , name : "200 Series/Z370 Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41703u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41704u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41705u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41706u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41707u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41708u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41709u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41710u16 , name : "200 Series/Z370 Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41712u16 , name : "200 Series/Z370 Chipset Family HD Audio" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31346u16 , name : "H270 PC MATE" } , SubSystem { subvendor : 5218u16 , subdevice : 64114u16 , name : "H270 PC MATE" }] } , Device { vendor_id : 32902u16 , id : 41731u16 , name : "H310 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41732u16 , name : "H370 Chipset LPC/eSPI Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41733u16 , name : "Z390 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41734u16 , name : "Q370 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41735u16 , name : "Q360 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41736u16 , name : "B360 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41737u16 , name : "C246 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41738u16 , name : "C242 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41740u16 , name : "QM370 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41741u16 , name : "HM370 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41742u16 , name : "CM246 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41747u16 , name : "Cannon Lake LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41760u16 , name : "300/C240 Series Chipset Family P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41761u16 , name : "300/C240 Series Chipset Family PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41763u16 , name : "300/C240 Series Chipset Family SMBus" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41764u16 , name : "300/C240 Series Chipset Family SPI (Flash) Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41766u16 , name : "300/C240 Series Chipset Family Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41768u16 , name : "300/C240 Series Chipset Family UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41769u16 , name : "300/C240 Series Chipset Family UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41770u16 , name : "300/C240 Series Chipset Family GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41771u16 , name : "300/C240 Series Chipset Family GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41772u16 , name : "300/C240 Series Chipset Family PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41773u16 , name : "300/C240 Series Chipset Family PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41774u16 , name : "300/C240 Series Chipset Family PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41775u16 , name : "300/C240 Series Chipset Family PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41776u16 , name : "300/C240 Series Chipset Family PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41777u16 , name : "300/C240 Series Chipset Family PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41778u16 , name : "300/C240 Series Chipset Family PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41779u16 , name : "300/C240 Series Chipset Family PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41780u16 , name : "300/C240 Series Chipset Family PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41781u16 , name : "300/C240 Series Chipset Family PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41782u16 , name : "300/C240 Series Chipset Family PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41783u16 , name : "300/C240 Series Chipset Family PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41784u16 , name : "300/C240 Series Chipset Family PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41785u16 , name : "300/C240 Series Chipset Family PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41786u16 , name : "300/C240 Series Chipset Family PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41787u16 , name : "300/C240 Series Chipset Family PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41788u16 , name : "300/C240 Series Chipset Family PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41789u16 , name : "300/C240 Series Chipset Family PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41790u16 , name : "300/C240 Series Chipset Family PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41791u16 , name : "300/C240 Series Chipset Family PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41792u16 , name : "300/C240 Series Chipset Family PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41793u16 , name : "300/C240 Series Chipset Family PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41794u16 , name : "300/C240 Series Chipset Family PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41795u16 , name : "300/C240 Series Chipset Family PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41799u16 , name : "300/C240 Series Chipset Family UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41800u16 , name : "300/C240 Series Chipset Family HD Audio" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41810u16 , name : "300/C240 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41811u16 , name : "300/C240 Series Chipset Family SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41813u16 , name : "300/C240 Series Chipset Family SATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41814u16 , name : "300/C240 Series Chipset Family SATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41815u16 , name : "300/C240 Series Chipset Family SATA Controller (RAID 0/1/5/10)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41822u16 , name : "300/C240 Series Chipset Family SATA Controller Optane Memory" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41824u16 , name : "300/C240 Series Chipset Family HECI #1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41825u16 , name : "300/C240 Series Chipset Family HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41826u16 , name : "300/C240 Series Chipset Family IDE Redirection (IDER-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41827u16 , name : "300/C240 Series Chipset Family Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41828u16 , name : "300/C240 Series Chipset Family HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41829u16 , name : "300/C240 Series Chipset Family HECI #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41832u16 , name : "300/C240 Series Chipset Family I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41833u16 , name : "300/C240 Series Chipset Family I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41834u16 , name : "300/C240 Series Chipset Family I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41835u16 , name : "300/C240 Series Chipset Family I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41837u16 , name : "300/C240 Series Chipset Family USB 3.1 xHCI" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41838u16 , name : "300/C240 Series Chipset Family USB Device Controller (Dual Role)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41839u16 , name : "300/C240 Series Chipset Family Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41840u16 , name : "300/C240 Series Chipset Family CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5458u16 , name : "Killer(R) Wireless-AC 1550i Wireless Network Adapter (9560NGW)" } , SubSystem { subvendor : 32902u16 , subdevice : 52u16 , name : "Wireless-AC 9560" }] } , Device { vendor_id : 32902u16 , id : 41841u16 , name : "300/C240 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41842u16 , name : "300/C240 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41843u16 , name : "300/C240 Series Chipset Family CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41849u16 , name : "300/C240 Series Chipset Family Thermal Subsystem" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" }] } , Device { vendor_id : 32902u16 , id : 41851u16 , name : "300/C240 Series Chipset Family SPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41852u16 , name : "300/C240 Series Chipset Family Integrated Sensor Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41858u16 , name : "B460/H410 Chipset SATA Controller (AHCI)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41860u16 , name : "B460/H410 Chipset SATA Controller (RAID 0/1/5/10) Not Premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41862u16 , name : "B460/H410 Chipset SATA Controller (RAID 0/1/5/10) Premium" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41870u16 , name : "B460/H410 Chipset SATA Controller (RST Optane)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41872u16 , name : "B460/H410 Chipset PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41873u16 , name : "B460/H410 Chipset PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41874u16 , name : "B460/H410 Chipset PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41875u16 , name : "B460/H410 Chipset PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41876u16 , name : "B460/H410 Chipset PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41877u16 , name : "B460/H410 Chipset PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41878u16 , name : "B460/H410 Chipset PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41879u16 , name : "B460/H410 Chipset PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41880u16 , name : "B460/H410 Chipset PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41881u16 , name : "B460/H410 Chipset PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41882u16 , name : "B460/H410 Chipset PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41883u16 , name : "B460/H410 Chipset PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41884u16 , name : "B460/H410 Chipset PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41885u16 , name : "B460/H410 Chipset PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41886u16 , name : "B460/H410 Chipset PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41887u16 , name : "B460/H410 Chipset PCIe Root Port #16" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41888u16 , name : "B460/H410 Chipset P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41889u16 , name : "B460/H410 Chipset PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41891u16 , name : "B460/H410 Chipset SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41892u16 , name : "B460/H410 Chipset SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41894u16 , name : "B460/H410 Chipset Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41895u16 , name : "B460/H410 Chipset UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41896u16 , name : "B460/H410 Chipset UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41897u16 , name : "B460/H410 Chipset SPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41898u16 , name : "B460/H410 Chipset SPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41903u16 , name : "B460/H410 Chipset USB 3.2 Gen 1x1 (5 Gbs) xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41904u16 , name : "B460/H410 Chipset USB Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41905u16 , name : "B460/H410 Chipset Thermal Subsystem" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41909u16 , name : "B460/H410 Chipset ISH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41914u16 , name : "B460/H410 Chipset CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41915u16 , name : "B460/H410 Chipset CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41916u16 , name : "B460/H410 Chipset CSME IDE Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41917u16 , name : "B460/H410 Chipset CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41918u16 , name : "B460/H410 Chipset CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41928u16 , name : "B460 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41946u16 , name : "H410 Chipset LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41952u16 , name : "B460/H410 Chipset I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41953u16 , name : "B460/H410 Chipset I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41954u16 , name : "B460/H410 Chipset I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41955u16 , name : "B460/H410 Chipset I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41958u16 , name : "B460/H410 Chipset UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41959u16 , name : "B460/H410 Chipset PCIe Root Port #17" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41960u16 , name : "B460/H410 Chipset PCIe Root Port #18" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41961u16 , name : "B460/H410 Chipset PCIe Root Port #19" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41962u16 , name : "B460/H410 Chipset PCIe Root Port #20" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41963u16 , name : "B460/H410 Chipset PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41964u16 , name : "B460/H410 Chipset PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41965u16 , name : "B460/H410 Chipset PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41966u16 , name : "B460/H410 Chipset PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41968u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41969u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41970u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41971u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41972u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41973u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41974u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 41975u16 , name : "B460/H410 Chipset HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42528u16 , name : "6400/6402 Advanced Memory Buffer (AMB)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42752u16 , name : "Raptor Lake-S Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42755u16 , name : "Raptor Lake-S Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42758u16 , name : "Raptor Lake-P 6p+8e cores Host Bridge/DRAM Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42759u16 , name : "Raptor Lake-P/U 4p+8e cores Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42760u16 , name : "Raptor Lake-P/U 2p+8e cores Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42765u16 , name : "Raptor Lake PCI Express 5.0 Graphics Port (PEG010)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42781u16 , name : "Raptor Lake Dynamic Platform and Thermal Framework Processor Participant" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42782u16 , name : "Raptor Lake-P Thunderbolt 4 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42783u16 , name : "Raptor Lake-P Thunderbolt 4 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42784u16 , name : "Raptor Lake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42785u16 , name : "Raptor Lake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42799u16 , name : "Raptor Lake-P Thunderbolt 4 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42814u16 , name : "Raptor Lake-P Thunderbolt 4 NHI #0" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42816u16 , name : "Raptor Lake-S 8+12 - Host Bridge/DRAM Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42829u16 , name : "Raptor Lake PCI Express 4.0 Graphics Port" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42831u16 , name : "GNA Scoring Accelerator module" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42845u16 , name : "Raptor Lake IPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42861u16 , name : "Raptor Lake-P Thunderbolt 4 NHI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42862u16 , name : "Raptor Lake-P Thunderbolt 4 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42877u16 , name : "Raptor Lake Crashlog and Telemetry" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42879u16 , name : "RST Volume Management Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42880u16 , name : "Raptor Lake-S GT1 [UHD Graphics 770]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42881u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42882u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42883u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42888u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42889u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42890u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42891u16 , name : "Raptor Lake-S UHD Graphics" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42912u16 , name : "Raptor Lake-P [Iris Xe Graphics]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" }] } , Device { vendor_id : 32902u16 , id : 42913u16 , name : "Raptor Lake-P [Iris Xe Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42920u16 , name : "Raptor Lake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42921u16 , name : "Raptor Lake-P [UHD Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42922u16 , name : "Raptor Lake-P [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42923u16 , name : "Raptor Lake-P [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42924u16 , name : "Raptor Lake-U [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 42925u16 , name : "Raptor Lake-U [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43014u16 , name : "Lunar Lake-M LPC/eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43015u16 , name : "Core Ultra 200V Series Processors eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43040u16 , name : "Core Ultra 200V Series Processors P2SB (8 bit)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43041u16 , name : "Core Ultra 200V Series Processors PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43042u16 , name : "Lunar Lake-M SMbus Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43043u16 , name : "Core Ultra 200V Series Processors SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43044u16 , name : "Core Ultra 200V Series Processors Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43045u16 , name : "Core Ultra 200V Series Processors UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43046u16 , name : "Core Ultra 200V Series Processors UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43047u16 , name : "Core Ultra 200V Series Processors GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43048u16 , name : "Core Ultra 200V Series Processors HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43056u16 , name : "Core Ultra 200V Series Processors GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43057u16 , name : "Core Ultra 200V Series Processors Type-C Subsystem xHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43059u16 , name : "Core Ultra 200V Series Processors Thunderbolt DMA0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43060u16 , name : "Core Ultra 200V Series Processors Thunderbolt DMA1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43064u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43065u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43066u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43067u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43068u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43069u16 , name : "Core Ultra 200V Series Processors PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43072u16 , name : "BE200 Series Wi-Fi 7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43077u16 , name : "Core Ultra 200V Series Processors Integrated Sensor Hub (ISH)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43078u16 , name : "Core Ultra 200V Series Processors GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43079u16 , name : "Lunar Lake-M UFS Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43080u16 , name : "Core Ultra 200V Series Processors Touch Host Controller (THC) #0 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43081u16 , name : "Core Ultra 200V Series Processors Touch Host Controller (THC) #0 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43082u16 , name : "Core Ultra 200V Series Processors Touch Host Controller (THC) #1 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43083u16 , name : "Core Ultra 200V Series Processors Touch Host Controller (THC) #1 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43084u16 , name : "Core Ultra 200V Series Processors P2SB (16 bit)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43086u16 , name : "Core Ultra 200V Series Processors USB Type-C Subsystem PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43087u16 , name : "Core Ultra 200V Series Processors USB Type-C Subsystem PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43088u16 , name : "Core Ultra 200V Series Processors I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43089u16 , name : "Core Ultra 200V Series Processors I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43090u16 , name : "Core Ultra 200V Series Processors UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43101u16 , name : "Core Ultra 200V Series Processors CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43102u16 , name : "Core Ultra 200V Series Processors CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43103u16 , name : "Core Ultra 200V Series Processors CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43104u16 , name : "Core Ultra 200V Series Processors USB Type-C Subsystem PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43106u16 , name : "Core Ultra 200V Series Processors CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43107u16 , name : "Core Ultra 200V Series Processors CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43108u16 , name : "Core Ultra 200V Series Processors CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43120u16 , name : "Core Ultra 200V Series Processors CSME HECI #1 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43121u16 , name : "Core Ultra 200V Series Processors CSME HECI #2 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43122u16 , name : "Core Ultra 200V Series Processors CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43123u16 , name : "Core Ultra 200V Series Processors CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43124u16 , name : "Core Ultra 200V Series Processors CSME HECI #3 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43125u16 , name : "Core Ultra 200V Series Processors CSME HECI #4 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43127u16 , name : "Core Ultra 200V Series Processors I3C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43128u16 , name : "Core Ultra 200V Series Processors I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43129u16 , name : "Core Ultra 200V Series Processors I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43130u16 , name : "Core Ultra 200V Series Processors I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43131u16 , name : "Core Ultra 200V Series Processors I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43132u16 , name : "Core Ultra 200V Series Processors I3C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43133u16 , name : "Core Ultra 200V Series Processors Standalone xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43135u16 , name : "Core Ultra 200V Series Processors Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 43968u16 , name : "Omni-Path Fabric Switch Silicon 100 Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44291u16 , name : "Core Ultra 200 Series Processors Dynamic Tuning Technology (DTT)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44299u16 , name : "Core Ultra 200 Series Processors VMD" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44301u16 , name : "Core Ultra 200 Series Processors Crash Log and Telemetry" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44317u16 , name : "Core Ultra 200 Series Processors NPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44560u16 , name : "Arrow Lake-HX Direct eSPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44576u16 , name : "Core Ultra 200 Series Processors P2SB (SOC-S)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44577u16 , name : "Core Ultra 200 Series Processors PMC (SOC-S)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44578u16 , name : "Core Ultra 200 Series Processors SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44579u16 , name : "Core Ultra 200 Series Processors SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44580u16 , name : "Core Ultra 200 Series Processors Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44620u16 , name : "Core Ultra 200 Series Processors Gauss Newton Algorithm (GNA)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44621u16 , name : "Core Ultra 200 Series Processors PCIe Root Port #13" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44622u16 , name : "Core Ultra 200 Series Processors PCIe Root Port #14" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44623u16 , name : "Core Ultra 200 Series Processors PCIe Root Port #15" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44656u16 , name : "Core Ultra 200 Series Processors CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44657u16 , name : "Core Ultra 200 Series Processors CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44660u16 , name : "Core Ultra 200 Series Processors CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 44671u16 , name : "Core Ultra 200 Series Processors Shared SRAM (SOC-S)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45056u16 , name : "Core Ultra Processors (Series 3) PTL-404" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45057u16 , name : "Core Ultra Processors (Series 3) PTL-H12Xe" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45058u16 , name : "Core Ultra Processors (Series 3) PTL-H484" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45059u16 , name : "Core Ultra Processors (Series 3) PTL-204" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45060u16 , name : "Core Ultra Processors (Series 3) PTL-H444" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45061u16 , name : "Core Ultra Processors (Series 3) PTL-H444" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45085u16 , name : "Core Ultra Processors (Series 3) DTT" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45101u16 , name : "Core Ultra Processors (Series 3) IAA" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45118u16 , name : "Core Ultra Processors (Series 3) NPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45149u16 , name : "Core Ultra Processors (Series 3) IPU" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45181u16 , name : "Core Ultra Processors (Series 3) Crashlog and Telemetry" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45184u16 , name : "Panther Lake [Arc B390]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45185u16 , name : "Panther Lake [Arc B370]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45186u16 , name : "Panther Lake [Arc B390]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45187u16 , name : "Panther Lake [Arc B370]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45188u16 , name : "Panther Lake [Arc Pro B390]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45189u16 , name : "Panther Lake [Arc Pro B370]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45190u16 , name : "Panther Lake [Arc Pro B390]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45191u16 , name : "Panther Lake [Arc Pro B370]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45199u16 , name : "Panther Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45200u16 , name : "Panther Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45216u16 , name : "Panther Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45232u16 , name : "Panther Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45311u16 , name : "Panther Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 45394u16 , name : "21152 PCI-to-PCI Bridge" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 45394u16 , name : "21152 PCI-to-PCI Bridge" }] } , Device { vendor_id : 32902u16 , id : 45396u16 , name : "21154 PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 46421u16 , name : "21555 Non transparent PCI-to-PCI Bridge" , subsystems : & [SubSystem { subvendor : 4807u16 , subdevice : 20485u16 , name : "SS7HD PCI Adaptor Card" } , SubSystem { subvendor : 4807u16 , subdevice : 20486u16 , name : "SS7HDC cPCI Adaptor Card" } , SubSystem { subvendor : 4825u16 , subdevice : 10u16 , name : "PCI VoIP Gateway" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" } , SubSystem { subvendor : 58559u16 , subdevice : 4096u16 , name : "CC8-1-BLUES" }] } , Device { vendor_id : 32902u16 , id : 46656u16 , name : "Arrow Lake-H [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53552u16 , name : "Core Processor DMI" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" }] } , Device { vendor_id : 32902u16 , id : 53553u16 , name : "Core Processor DMI" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" }] } , Device { vendor_id : 32902u16 , id : 53554u16 , name : "Core Processor DMI" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" }] } , Device { vendor_id : 32902u16 , id : 53555u16 , name : "Core Processor DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53556u16 , name : "Core Processor DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53557u16 , name : "Core Processor DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53558u16 , name : "Core Processor DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53559u16 , name : "Core Processor DMI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53560u16 , name : "Core Processor PCI Express Root Port 1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 730u16 , name : "OptiPlex 980" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 5593u16 , subdevice : 1549u16 , name : "C7SIM-Q Motherboard" }] } , Device { vendor_id : 32902u16 , id : 53561u16 , name : "Core Processor PCI Express Root Port 2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53562u16 , name : "Core Processor PCI Express Root Port 3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53563u16 , name : "Core Processor PCI Express Root Port 4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53584u16 , name : "Core Processor QPI Link" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53585u16 , name : "Core Processor QPI Routing and Protocol Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53589u16 , name : "Core Processor System Management Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53590u16 , name : "Core Processor Semaphore and Scratchpad Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53591u16 , name : "Core Processor System Control and Status Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 53592u16 , name : "Core Processor Miscellaneous Registers" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54051u16 , name : "Nova Lake PCD-H SPI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54053u16 , name : "Nova Lake PCD-H Serial IO UART Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54054u16 , name : "Nova Lake PCD-H Serial IO UART Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54055u16 , name : "Nova Lake PCD-H Serial IO SPI Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54064u16 , name : "Nova Lake PCD-H Serial IO SPI Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54065u16 , name : "Nova Lake-H Thunderbolt 5 USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54067u16 , name : "Nova Lake-H Thunderbolt 5 NHI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54087u16 , name : "Nova Lake PCD-H Serial IO SPI Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54094u16 , name : "Nova Lake-H Thunderbolt 5 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54095u16 , name : "Nova Lake-H Thunderbolt 5 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54096u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54097u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54098u16 , name : "Nova Lake PCD-H Serial IO UART Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54112u16 , name : "Nova Lake-H Thunderbolt 5 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54136u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54137u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54138u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54139u16 , name : "Nova Lake PCD-H Serial IO I2C Controller #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54321u16 , name : "Nova Lake-S Thunderbolt 5 USB Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54323u16 , name : "Nova Lake-S Thunderbolt 5 NHI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54350u16 , name : "Nova Lake-S Thunderbolt 5 PCI Express Root Port #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54351u16 , name : "Nova Lake-S Thunderbolt 5 PCI Express Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 54368u16 , name : "Nova Lake-S Thunderbolt 5 PCI Express Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55104u16 , name : "NVL-S" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55105u16 , name : "NVL-U" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55106u16 , name : "NVL-H" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55107u16 , name : "NVL-HX" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55108u16 , name : "NVL-UL" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55109u16 , name : "NVL-HX" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55114u16 , name : "NVL-S" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55115u16 , name : "NVL-S" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55120u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55121u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55122u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55123u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55124u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55125u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55126u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55127u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 55135u16 , name : "NVL-P" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57858u16 , name : "Battlemage G21 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57867u16 , name : "Battlemage G21 [Arc B580]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57868u16 , name : "Battlemage G21 [Arc B570]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57869u16 , name : "Battlemage G21 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57872u16 , name : "Battlemage G21 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57873u16 , name : "Battlemage G21 [Arc Pro B60]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57874u16 , name : "Battlemage G21 [Arc Pro B50]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57877u16 , name : "Battlemage G21 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57878u16 , name : "Battlemage G21 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57888u16 , name : "Battlemage G31 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57889u16 , name : "Battlemage G31 [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57890u16 , name : "Battlemage G31 [Arc Pro B65]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 57891u16 , name : "Battlemage G31 [Arc Pro B70]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58112u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58114u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58143u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58144u16 , name : "Core Ultra Processors (Series 3) P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58145u16 , name : "Core Ultra Processors (Series 3) PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58146u16 , name : "Core Ultra Processors (Series 3) SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58147u16 , name : "Core Ultra Processors (Series 3) SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58148u16 , name : "Core Ultra Processors (Series 3) Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58149u16 , name : "Core Ultra Processors (Series 3) UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58150u16 , name : "Core Ultra Processors (Series 3) UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58151u16 , name : "Core Ultra Processors (Series 3) GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58152u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58153u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58154u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58155u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58156u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58157u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58158u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58159u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58160u16 , name : "Core Ultra Processors (Series 3) GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58161u16 , name : "Core Ultra Processors (Series 3) Type-C Subsystem xHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58162u16 , name : "Core Ultra Processors (Series 3) Type-C Subsystem xDCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58163u16 , name : "Core Ultra Processors (Series 3) Thunderbolt DMA0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58164u16 , name : "Core Ultra Processors (Series 3) Thunderbolt DMA1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58167u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58168u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58169u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58170u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58171u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58172u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58173u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58174u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58175u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58176u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 148u16 , name : "Wi-Fi 6E AX211 160MHz" }] } , Device { vendor_id : 32902u16 , id : 58177u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58178u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58179u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58180u16 , name : "Core Ultra Processors (Series 3) IEH #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58181u16 , name : "Core Ultra Processors (Series 3) ISH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58182u16 , name : "Core Ultra Processors (Series 3) GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58184u16 , name : "Core Ultra Processors (Series 3) THC #0 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58185u16 , name : "Core Ultra Processors (Series 3) THC #0 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58186u16 , name : "Core Ultra Processors (Series 3) THC #1 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58187u16 , name : "Core Ultra Processors (Series 3) THC #1 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58188u16 , name : "Core Ultra Processors (Series 3) P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58189u16 , name : "Core Ultra Processors (Series 3) IEH #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58190u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58191u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58192u16 , name : "Core Ultra Processors (Series 3) I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58193u16 , name : "Core Ultra Processors (Series 3) I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58194u16 , name : "Core Ultra Processors (Series 3) UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58204u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58205u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58206u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58207u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58208u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58209u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58210u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58211u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58212u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58213u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #11" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58214u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #12" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58215u16 , name : "Panther Lake PCI Express Root Port D3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58216u16 , name : "Panther Lake PCI Express Root Port D4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58217u16 , name : "Panther Lake PCI Express Root Port D5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58218u16 , name : "Panther Lake PCI Express Root Port D6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58219u16 , name : "Panther Lake PCI Express Root Port D7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58220u16 , name : "Panther Lake PCI Express Root Port D8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58223u16 , name : "Core Ultra Processors (Series 3) I3C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58224u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58225u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58226u16 , name : "Core Ultra Processors (Series 3) CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58227u16 , name : "Core Ultra Processors (Series 3) CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58228u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58229u16 , name : "Core Ultra Processors (Series 3) CSME HECI #4 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58230u16 , name : "Core Ultra Processors (Series 3) CNVi Bluetooth" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58232u16 , name : "Core Ultra Processors (Series 3) I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58233u16 , name : "Core Ultra Processors (Series 3) I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58234u16 , name : "Core Ultra Processors (Series 3) I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58235u16 , name : "Core Ultra Processors (Series 3) I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58236u16 , name : "Core Ultra Processors (Series 3) I3C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58237u16 , name : "Core Ultra Processors (Series 3) Standalone xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58238u16 , name : "Core Ultra Processors (Series 3) Standalone USB Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58239u16 , name : "Core Ultra Processors (Series 3) Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58368u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58370u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58399u16 , name : "Core Ultra Processors (Series 3) eSPI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58400u16 , name : "Core Ultra Processors (Series 3) P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58401u16 , name : "Core Ultra Processors (Series 3) PMC" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58402u16 , name : "Core Ultra Processors (Series 3) SMBus" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58403u16 , name : "Core Ultra Processors (Series 3) SPI (flash) Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58404u16 , name : "Core Ultra Processors (Series 3) Trace Hub" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58405u16 , name : "Core Ultra Processors (Series 3) UART #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58406u16 , name : "Core Ultra Processors (Series 3) UART #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58407u16 , name : "Core Ultra Processors (Series 3) GSPI #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58408u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58409u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58410u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58411u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58412u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58413u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58414u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58415u16 , name : "Core Ultra Processors (Series 3) HD Audio" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58416u16 , name : "Core Ultra Processors (Series 3) GSPI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58417u16 , name : "Core Ultra Processors (Series 3) Type-C Subsystem xHCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58418u16 , name : "Core Ultra Processors (Series 3) Type-C Subsystem xDCI" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58419u16 , name : "Core Ultra Processors (Series 3) Thunderbolt DMA0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58420u16 , name : "Core Ultra Processors (Series 3) Thunderbolt DMA1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58423u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #24" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58424u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58425u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58426u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58427u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58428u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58429u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #6" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58430u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #7" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58431u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #8" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58432u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 276u16 , name : "Wi-Fi 7 BE211 320MHz" }] } , Device { vendor_id : 32902u16 , id : 58433u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58434u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58435u16 , name : "Core Ultra Processors (Series 3) CNVi Wi-Fi" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58436u16 , name : "Core Ultra Processors (Series 3) IEH #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58437u16 , name : "Core Ultra Processors (Series 3) ISH" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58438u16 , name : "Core Ultra Processors (Series 3) GSPI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58440u16 , name : "Core Ultra Processors (Series 3) THC #0 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58441u16 , name : "Core Ultra Processors (Series 3) THC #0 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58442u16 , name : "Core Ultra Processors (Series 3) THC #1 ID1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58443u16 , name : "Core Ultra Processors (Series 3) THC #1 ID2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58444u16 , name : "Core Ultra Processors (Series 3) P2SB" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58445u16 , name : "Core Ultra Processors (Series 3) IEH #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58446u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #21" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58447u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #22" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58448u16 , name : "Core Ultra Processors (Series 3) I2C #4" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58449u16 , name : "Core Ultra Processors (Series 3) I2C #5" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58450u16 , name : "Core Ultra Processors (Series 3) UART #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58460u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #10" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58461u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58462u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58463u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58464u16 , name : "Core Ultra Processors (Series 3) USB Type-C Subsystem PCIe Root Port #23" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58465u16 , name : "Core Ultra Processors (Series 3) PCIe Root Port #9" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58466u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58467u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58468u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58479u16 , name : "Core Ultra Processors (Series 3) I3C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58480u16 , name : "Core Ultra Processors (Series 3) CSME HECI #1 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58481u16 , name : "Core Ultra Processors (Series 3) CSME HECI #2 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58482u16 , name : "Core Ultra Processors (Series 3) CSME IDE Redirection (IDE-R)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58483u16 , name : "Core Ultra Processors (Series 3) CSME Keyboard and Text (KT) Redirection" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58484u16 , name : "Core Ultra Processors (Series 3) CSME HECI #3 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58485u16 , name : "Core Ultra Processors (Series 3) CSME HECI #4 (CSE)" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58486u16 , name : "Core Ultra Processors (Series 3) CNVi Bluetooth" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58488u16 , name : "Core Ultra Processors (Series 3) I2C #0" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58489u16 , name : "Core Ultra Processors (Series 3) I2C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58490u16 , name : "Core Ultra Processors (Series 3) I2C #2" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58491u16 , name : "Core Ultra Processors (Series 3) I2C #3" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58492u16 , name : "Core Ultra Processors (Series 3) I3C #1" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58493u16 , name : "Core Ultra Processors (Series 3) Standalone xHCI Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58494u16 , name : "Core Ultra Processors (Series 3) Standalone USB Device Controller" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 58495u16 , name : "Core Ultra Processors (Series 3) Shared SRAM" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 61861u16 , name : "SSD 600P Series" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 14602u16 , name : "SSDPEKKW256G7 256GB" }] } , Device { vendor_id : 32902u16 , id : 61862u16 , name : "SSD DC P4101/Pro 7600p/760p/E 6100p Series" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 14603u16 , name : "SSD Pro 7600p/760p/E 6100p Series [NVM Express]" }] } , Device { vendor_id : 32902u16 , id : 61863u16 , name : "SSD 700p Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 61864u16 , name : "SSD 660P Series" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 61866u16 , name : "SSD 670p Series [Keystone Harbor]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 64240u16 , name : "SSD 665p Series [Neptune Harbor Refresh]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 64896u16 , name : "Wildcat Lake [Intel Graphics]" , subsystems : & [] } , Device { vendor_id : 32902u16 , id : 64897u16 , name : "Wildcat Lake [Intel Graphics]" , subsystems : & [] }] }),
        (4743, Vendor { id : 4743u16 , name : "M-Pact, Inc." , devices : & [Device { vendor_id : 4743u16 , id : 30u16 , name : "LS220D DVD Decoder" , subsystems : & [] } , Device { vendor_id : 4743u16 , id : 31u16 , name : "LS220C DVD Decoder" , subsystems : & [] }] }),
        (4695, Vendor { id : 4695u16 , name : "Vertex Networks, Inc." , devices : & [] }),
        (56467, Vendor { id : 56467u16 , name : "Dawicontrol GmbH" , devices : & [] }),
        (5031, Vendor { id : 5031u16 , name : "Teles AG" , devices : & [] }),
        (5295, Vendor { id : 5295u16 , name : "Guillemot Corporation" , devices : & [Device { vendor_id : 5295u16 , id : 28930u16 , name : "3D Prophet II MX" , subsystems : & [] }] }),
        (4645, Vendor { id : 4645u16 , name : "Power I/O, Inc." , devices : & [] }),
        (4572, Vendor { id : 4572u16 , name : "Questra Corporation" , devices : & [] }),
        (5330, Vendor { id : 5330u16 , name : "Titan Electronics Inc" , devices : & [Device { vendor_id : 5330u16 , id : 32769u16 , name : "VScom 010L 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32770u16 , name : "VScom 020L 2 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32784u16 , name : "VScom 100L 1 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32785u16 , name : "VScom 110L 1 port serial and 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32800u16 , name : "VScom 200L 1 or 2 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32801u16 , name : "VScom 210L 2 port serial and 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32808u16 , name : "VScom 200I/200I-SI 2-port serial adapter" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32832u16 , name : "VScom 400L 4 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32835u16 , name : "VScom 430L 4-port serial and 3-port parallel adapter" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32840u16 , name : "VScom 400I 4-port serial adapter" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32896u16 , name : "VScom 800L 8 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 32904u16 , name : "VScom 800I 8-port serial adapter" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40960u16 , name : "VScom 010H 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40961u16 , name : "VScom 100H 1 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40963u16 , name : "VScom 400H 4 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40964u16 , name : "VScom 400HF1 4 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40965u16 , name : "VScom 200H 2 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40967u16 , name : "VScom PCI800EH (PCIe) 8-port serial adapter Port 1-4" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40968u16 , name : "VScom PCI800EH (PCIe) 8-port serial adapter Port 5-8" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 40969u16 , name : "VScom PCI400EH (PCIe) 4-port serial adapter" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 57345u16 , name : "VScom 010HV2 1 port parallel adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 57360u16 , name : "VScom 100HV2 1 port serial adaptor" , subsystems : & [] } , Device { vendor_id : 5330u16 , id : 57376u16 , name : "VScom 200HV2 2 port serial adaptor" , subsystems : & [] }] }),
        (5461, Vendor { id : 5461u16 , name : "GESYTEC GmBH" , devices : & [] }),
        (19798, Vendor { id : 19798u16 , name : "Balluff MV GmbH" , devices : & [Device { vendor_id : 19798u16 , id : 0u16 , name : "[mvHYPERION-CLe/CLb] CameraLink PCI Express x1 Frame Grabber" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 1u16 , name : "[mvHYPERION-CLf/CLm] CameraLink PCI Express x4 Frame Grabber" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 16u16 , name : "[mvHYPERION-16R16/-32R16] 16 Video Channel PCI Express x4 Frame Grabber" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 32u16 , name : "[mvHYPERION-HD-SDI] HD-SDI PCI Express x4 Frame Grabber" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 48u16 , name : "[mvHYPERION-HD-SDI-Merger] HD-SDI PCI Express x4 Frame Grabber" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 28690u16 , name : "[mvBlueNAOS BVS CA-BN] PCIe Gen1 x2 Camera" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 28692u16 , name : "[mvBlueNAOS BVS CA-BN] PCIe Gen1 x4 Camera" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 28706u16 , name : "[mvBlueNAOS BVS CA-BN] PCIe Gen2 x2 Camera" , subsystems : & [] } , Device { vendor_id : 19798u16 , id : 28708u16 , name : "[mvBlueNAOS BVS CA-BN] PCIe Gen2 x4 Camera" , subsystems : & [] }] }),
        (4940, Vendor { id : 4940u16 , name : "Chori Joho System Co. Ltd" , devices : & [] }),
        (4771, Vendor { id : 4771u16 , name : "Lucent Technologies" , devices : & [Device { vendor_id : 4771u16 , id : 33029u16 , name : "T8105 H100 Digital Switch" , subsystems : & [] }] }),
        (4400, Vendor { id : 4400u16 , name : "Computervision" , devices : & [] }),
        (5574, Vendor { id : 5574u16 , name : "Technical University of Budapest" , devices : & [] }),
        (4387, Vendor { id : 4387u16 , name : "Excellent Design, Inc." , devices : & [] }),
        (15880, Vendor { id : 15880u16 , name : "Sinead Tech Co., Ltd" , devices : & [] }),
        (8491, Vendor { id : 8491u16 , name : "Flumeio" , devices : & [Device { vendor_id : 8491u16 , id : 22865u16 , name : "Flumeio F5950Q" , subsystems : & [] }] }),
        (4472, Vendor { id : 4472u16 , name : "Alfa, Inc." , devices : & [Device { vendor_id : 4472u16 , id : 44961u16 , name : "Fast Ethernet Adapter" , subsystems : & [] }] }),
        (4612, Vendor { id : 4612u16 , name : "Lattice Semiconductor Corporation" , devices : & [Device { vendor_id : 4612u16 , id : 6501u16 , name : "SB6501 802.11ad Wireless Network Adapter" , subsystems : & [] }] }),
        (5535, Vendor { id : 5535u16 , name : "Galea Network Security" , devices : & [] }),
        (5788, Vendor { id : 5788u16 , name : "Netcell Corporation" , devices : & [Device { vendor_id : 5788u16 , id : 68u16 , name : "Revolution Storage Processing Card" , subsystems : & [] }] }),
        (5417, Vendor { id : 5417u16 , name : "ON Semiconductor" , devices : & [] }),
        (4207, Vendor { id : 4207u16 , name : "City Gate Development Ltd" , devices : & [] }),
        (5225, Vendor { id : 5225u16 , name : "Cleveland Motion Controls" , devices : & [] }),
        (5650, Vendor { id : 5650u16 , name : "Telesynergy Research Inc." , devices : & [] }),
        (4322, Vendor { id : 4322u16 , name : "Aptix Corporation" , devices : & [] }),
        (5221, Vendor { id : 5221u16 , name : "GN NETTEST Telecom DIV." , devices : & [] }),
        (6063, Vendor { id : 6063u16 , name : "Hightech Information System Ltd." , devices : & [] }),
        (4948, Vendor { id : 4948u16 , name : "Dwave System Inc" , devices : & [] }),
        (5146, Vendor { id : 5146u16 , name : "Apache Micro Peripherals Inc" , devices : & [] }),
        (5578, Vendor { id : 5578u16 , name : "Bitboys OY" , devices : & [] }),
        (5391, Vendor { id : 5391u16 , name : "INTEC GmbH" , devices : & [] }),
        (4703, Vendor { id : 4703u16 , name : "Concurrent Technologies, Inc." , devices : & [Device { vendor_id : 4703u16 , id : 8305u16 , name : "CC PMC/232" , subsystems : & [] } , Device { vendor_id : 4703u16 , id : 8324u16 , name : "CC PMC/23P" , subsystems : & [] } , Device { vendor_id : 4703u16 , id : 8337u16 , name : "CC PMC/422" , subsystems : & [] }] }),
        (8993, Vendor { id : 8993u16 , name : "Bruker AXS Inc." , devices : & [Device { vendor_id : 8993u16 , id : 2u16 , name : "Hi-Star PCI Interface" , subsystems : & [] }] }),
        (5219, Vendor { id : 5219u16 , name : "Fast Corporation" , devices : & [] }),
        (5389, Vendor { id : 5389u16 , name : "WARPSPPED Inc" , devices : & [] }),
        (4929, Vendor { id : 4929u16 , name : "Kyoto Microcomputer Co" , devices : & [] }),
        (4453, Vendor { id : 4453u16 , name : "Imagraph Corporation" , devices : & [Device { vendor_id : 4453u16 , id : 1u16 , name : "Motion TPEG Recorder/Player with audio" , subsystems : & [] }] }),
        (4555, Vendor { id : 4555u16 , name : "Specialix Research Ltd." , devices : & [Device { vendor_id : 4555u16 , id : 8192u16 , name : "PCI_9050" , subsystems : & [SubSystem { subvendor : 4555u16 , subdevice : 512u16 , name : "SX" } , SubSystem { subvendor : 4555u16 , subdevice : 45064u16 , name : "I/O8+" }] } , Device { vendor_id : 4555u16 , id : 16384u16 , name : "SUPI_1" , subsystems : & [] } , Device { vendor_id : 4555u16 , id : 32768u16 , name : "T225" , subsystems : & [] }] }),
        (4183, Vendor { id : 4183u16 , name : "Motorola" , devices : & [Device { vendor_id : 4183u16 , id : 1u16 , name : "MPC105 [Eagle]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 2u16 , name : "MPC106 [Grackle]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 3u16 , name : "MPC8240 [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 4u16 , name : "MPC107" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 6u16 , name : "MPC8245 [Unity]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 8u16 , name : "MPC8540" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 9u16 , name : "MPC8560" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 18u16 , name : "MPC8548 [PowerQUICC III]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 256u16 , name : "MC145575 [HFC-PCI]" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 1073u16 , name : "KTI829c 100VG" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 4211u16 , name : "Nokia N770" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 4633u16 , name : "Nokia N800" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 6145u16 , name : "DSP56301 Digital Signal Processor" , subsystems : & [SubSystem { subvendor : 5371u16 , subdevice : 257u16 , name : "Transas Radar Imitator Board [RIM]" } , SubSystem { subvendor : 5371u16 , subdevice : 258u16 , name : "Transas Radar Imitator Board [RIM-2]" } , SubSystem { subvendor : 5371u16 , subdevice : 514u16 , name : "Transas Radar Integrator Board [RIB-2]" } , SubSystem { subvendor : 5371u16 , subdevice : 1553u16 , name : "1 channel CAN bus Controller [CanPci-1]" } , SubSystem { subvendor : 5371u16 , subdevice : 1554u16 , name : "2 channels CAN bus Controller [CanPci-2]" } , SubSystem { subvendor : 5371u16 , subdevice : 1555u16 , name : "3 channels CAN bus Controller [CanPci-3]" } , SubSystem { subvendor : 5371u16 , subdevice : 1556u16 , name : "4 channels CAN bus Controller [CanPci-4]" } , SubSystem { subvendor : 5371u16 , subdevice : 1569u16 , name : "1 channel CAN bus Controller [CanPci2-1]" } , SubSystem { subvendor : 5371u16 , subdevice : 1570u16 , name : "2 channels CAN bus Controller [CanPci2-2]" } , SubSystem { subvendor : 5371u16 , subdevice : 2064u16 , name : "Transas VTS Radar Integrator Board [RIB-4]" } , SubSystem { subvendor : 5980u16 , subdevice : 16896u16 , name : "ASI4215 Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 17152u16 , name : "ASI43xx Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 17408u16 , name : "ASI4401 Audio Adapter" } , SubSystem { subvendor : 60608u16 , subdevice : 16u16 , name : "Darla" } , SubSystem { subvendor : 60608u16 , subdevice : 32u16 , name : "Gina" } , SubSystem { subvendor : 60608u16 , subdevice : 48u16 , name : "Layla rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 49u16 , name : "Layla rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 64u16 , name : "Darla24 rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 65u16 , name : "Darla24 rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 80u16 , name : "Gina24 rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 81u16 , name : "Gina24 rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 112u16 , name : "Mona rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 113u16 , name : "Mona rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 114u16 , name : "Mona rev.2" }] } , Device { vendor_id : 4183u16 , id : 6336u16 , name : "MPC8265A/8266/8272" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 6337u16 , name : "MPC8271/MPC8272" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 12370u16 , name : "SM56 Data Fax Modem" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 13328u16 , name : "DSP56361 Digital Signal Processor" , subsystems : & [SubSystem { subvendor : 60608u16 , subdevice : 80u16 , name : "Gina24 rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 81u16 , name : "Gina24 rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 96u16 , name : "Layla24" } , SubSystem { subvendor : 60608u16 , subdevice : 112u16 , name : "Mona rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 113u16 , name : "Mona rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 114u16 , name : "Mona rev.2" } , SubSystem { subvendor : 60608u16 , subdevice : 128u16 , name : "Mia rev.0" } , SubSystem { subvendor : 60608u16 , subdevice : 129u16 , name : "Mia rev.1" } , SubSystem { subvendor : 60608u16 , subdevice : 144u16 , name : "Indigo" } , SubSystem { subvendor : 60608u16 , subdevice : 160u16 , name : "Indigo IO" } , SubSystem { subvendor : 60608u16 , subdevice : 176u16 , name : "Indigo DJ" } , SubSystem { subvendor : 60608u16 , subdevice : 256u16 , name : "3G" }] } , Device { vendor_id : 4183u16 , id : 18433u16 , name : "Raven" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 18434u16 , name : "Falcon" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 18435u16 , name : "Hawk" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 18438u16 , name : "CPX8216" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 18443u16 , name : "Harrier PCI Host Bridge" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 19816u16 , name : "20268" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 22016u16 , name : "SM56 PCI Modem" , subsystems : & [SubSystem { subvendor : 4183u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 4183u16 , subdevice : 769u16 , name : "SM56 PCI Voice Modem" } , SubSystem { subvendor : 4183u16 , subdevice : 770u16 , name : "SM56 PCI Fax Modem" } , SubSystem { subvendor : 4183u16 , subdevice : 22016u16 , name : "SM56 PCI Voice modem" } , SubSystem { subvendor : 5074u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 5074u16 , subdevice : 769u16 , name : "SM56 PCI Voice modem" } , SubSystem { subvendor : 5074u16 , subdevice : 770u16 , name : "SM56 PCI Fax Modem" } , SubSystem { subvendor : 5174u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 5174u16 , subdevice : 769u16 , name : "SM56 PCI Voice modem" } , SubSystem { subvendor : 5174u16 , subdevice : 770u16 , name : "SM56 PCI Fax Modem" } , SubSystem { subvendor : 5199u16 , subdevice : 4108u16 , name : "SM56 PCI Fax Modem" } , SubSystem { subvendor : 5268u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 5268u16 , subdevice : 769u16 , name : "SM56 PCI Voice modem" } , SubSystem { subvendor : 5320u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 5320u16 , subdevice : 770u16 , name : "SM56 PCI Fax Modem" } , SubSystem { subvendor : 5736u16 , subdevice : 768u16 , name : "SM56 PCI Speakerphone Modem" } , SubSystem { subvendor : 5736u16 , subdevice : 770u16 , name : "SM56 PCI Fax Modem" }] } , Device { vendor_id : 4183u16 , id : 22024u16 , name : "Wildcard X100P" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 22531u16 , name : "MPC5200" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 22534u16 , name : "MCF54 Coldfire" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 22536u16 , name : "MPC8220" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 22537u16 , name : "MPC5200B" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 25600u16 , name : "MPC190 Security Processor (S1 family, encryption)" , subsystems : & [] } , Device { vendor_id : 4183u16 , id : 25605u16 , name : "MPC184 Security Processor (S1 family)" , subsystems : & [] }] }),
        (19533, Vendor { id : 19533u16 , name : "Liquid-Markets GmbH" , devices : & [Device { vendor_id : 19533u16 , id : 39319u16 , name : "UberNIC" , subsystems : & [] } , Device { vendor_id : 19533u16 , id : 39320u16 , name : "TaSR" , subsystems : & [] } , Device { vendor_id : 19533u16 , id : 39321u16 , name : "UberNIC PoC/testing/dev" , subsystems : & [] }] }),
        (4848, Vendor { id : 4848u16 , name : "Pentek" , devices : & [] }),
        (7586, Vendor { id : 7586u16 , name : "Sapphire Technology Limited" , devices : & [Device { vendor_id : 7586u16 , id : 18269u16 , name : "Radeon RX 7800 XT [PULSE]" , subsystems : & [] } , Device { vendor_id : 7586u16 , id : 57962u16 , name : "Radeon R7 250" , subsystems : & [] } , Device { vendor_id : 7586u16 , id : 58437u16 , name : "Sapphire Radeon RX 6700" , subsystems : & [] }] }),
        (5085, Vendor { id : 5085u16 , name : "Multimedia Bundle Inc" , devices : & [] }),
        (5264, Vendor { id : 5264u16 , name : "Stone Microsystems PTY Ltd." , devices : & [] }),
        (4460, Vendor { id : 4460u16 , name : "Intelligent Resources Integrated Systems" , devices : & [] }),
        (34918, Vendor { id : 34918u16 , name : "T-Square Design Inc." , devices : & [] }),
        (8417, Vendor { id : 8417u16 , name : "CECloud Computing Technology Co., Ltd." , devices : & [Device { vendor_id : 8417u16 , id : 28929u16 , name : "LS X710-E" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 28931u16 , name : "LS X710-M" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 28932u16 , name : "LS X710-P" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 29056u16 , name : "LS X718" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 29201u16 , name : "LS X721-E" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 29219u16 , name : "LS X722-M" , subsystems : & [] } , Device { vendor_id : 8417u16 , id : 29220u16 , name : "LS X722-P" , subsystems : & [] }] }),
        (4296, Vendor { id : 4296u16 , name : "Neomagic Corporation" , devices : & [Device { vendor_id : 4296u16 , id : 1u16 , name : "NM2070 [MagicGraph 128]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 2u16 , name : "NM2090 [MagicGraph 128V]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 3u16 , name : "NM2093 [MagicGraph 128ZV]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 4u16 , name : "NM2160 [MagicGraph 128XD]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 186u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4133u16 , subdevice : 4103u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4136u16 , subdevice : 116u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4136u16 , subdevice : 117u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4136u16 , subdevice : 125u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4136u16 , subdevice : 126u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4147u16 , subdevice : 32815u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4173u16 , subdevice : 32795u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4173u16 , subdevice : 32815u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4173u16 , subdevice : 33547u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4282u16 , subdevice : 3584u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4296u16 , subdevice : 4u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4303u16 , subdevice : 4137u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4343u16 , subdevice : 33544u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4343u16 , subdevice : 33545u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4343u16 , subdevice : 33547u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4343u16 , subdevice : 33549u16 , name : "MagicGraph 128XD" } , SubSystem { subvendor : 4343u16 , subdevice : 33554u16 , name : "MagicGraph 128XD" }] } , Device { vendor_id : 4296u16 , id : 5u16 , name : "NM2200 [MagicGraph 256AV]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 221u16 , name : "ThinkPad 570" } , SubSystem { subvendor : 4136u16 , subdevice : 136u16 , name : "Latitude CPi A" }] } , Device { vendor_id : 4296u16 , id : 6u16 , name : "NM2360 [MagicMedia 256ZX]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 338u16 , name : "ThinkPad 600X" }] } , Device { vendor_id : 4296u16 , id : 22u16 , name : "NM2380 [MagicMedia 256XL+]" , subsystems : & [SubSystem { subvendor : 4296u16 , subdevice : 22u16 , name : "MagicMedia 256XL+" }] } , Device { vendor_id : 4296u16 , id : 37u16 , name : "NM2230 [MagicGraph 256AV+]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 131u16 , name : "NM2093 [MagicGraph 128ZV+]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 32773u16 , name : "NM2200 [MagicMedia 256AV Audio]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45265u16 , name : "MagicMedia 256AV Audio Device on Discovery" } , SubSystem { subvendor : 3601u16 , subdevice : 45350u16 , name : "MagicMedia 256AV Audio Device on Durango" } , SubSystem { subvendor : 4116u16 , subdevice : 221u16 , name : "ThinkPad 390/i1720/i1721" } , SubSystem { subvendor : 4133u16 , subdevice : 4099u16 , name : "MagicMedia 256AV Audio Device on TravelMate 720" } , SubSystem { subvendor : 4136u16 , subdevice : 136u16 , name : "Latitude CPi A" } , SubSystem { subvendor : 4136u16 , subdevice : 143u16 , name : "MagicMedia 256AV Audio Device on Colorado Inspiron" } , SubSystem { subvendor : 4156u16 , subdevice : 7u16 , name : "MagicMedia 256AV Audio Device on Voyager II" } , SubSystem { subvendor : 4156u16 , subdevice : 8u16 , name : "MagicMedia 256AV Audio Device on Voyager III" } , SubSystem { subvendor : 4156u16 , subdevice : 13u16 , name : "MagicMedia 256AV Audio Device on Omnibook 900" } , SubSystem { subvendor : 4296u16 , subdevice : 32773u16 , name : "MagicMedia 256AV Audio Device on FireAnt" } , SubSystem { subvendor : 4362u16 , subdevice : 32773u16 , name : "MagicMedia 256AV Audio Device" } , SubSystem { subvendor : 5312u16 , subdevice : 4u16 , name : "MagicMedia 256AV Audio Device" }] } , Device { vendor_id : 4296u16 , id : 32774u16 , name : "NM2360 [MagicMedia 256ZX Audio]" , subsystems : & [] } , Device { vendor_id : 4296u16 , id : 32790u16 , name : "NM2380 [MagicMedia 256XL+ Audio]" , subsystems : & [] }] }),
        (4611, Vendor { id : 4611u16 , name : "Bayer Corporation, Agfa Division" , devices : & [] }),
        (4275, Vendor { id : 4275u16 , name : "Databook Inc" , devices : & [Device { vendor_id : 4275u16 , id : 12550u16 , name : "DB87144" , subsystems : & [] } , Device { vendor_id : 4275u16 , id : 45318u16 , name : "DB87144" , subsystems : & [] }] }),
        (4355, Vendor { id : 4355u16 , name : "HighPoint Technologies, Inc." , devices : & [Device { vendor_id : 4355u16 , id : 3u16 , name : "HPT343/345/346/363" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 4u16 , name : "HPT366/368/370/370A/372/372N" , subsystems : & [SubSystem { subvendor : 4355u16 , subdevice : 1u16 , name : "HPT370A" } , SubSystem { subvendor : 4355u16 , subdevice : 4u16 , name : "HPT366 UDMA66 (r1) / HPT368 UDMA66 (r2) / HPT370 UDMA100 (r3) / HPT370 UDMA100 RAID (r4)" } , SubSystem { subvendor : 4355u16 , subdevice : 5u16 , name : "HPT370 UDMA100" } , SubSystem { subvendor : 4355u16 , subdevice : 6u16 , name : "HPT302/302N" }] } , Device { vendor_id : 4355u16 , id : 5u16 , name : "HPT372A/372N" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 6u16 , name : "HPT302/302N" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 7u16 , name : "HPT371/371N" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 8u16 , name : "HPT374" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 9u16 , name : "HPT372N" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1568u16 , name : "RocketRAID 620 2 Port SATA-III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1570u16 , name : "RocketRAID 622 2 Port SATA-III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1600u16 , name : "RocketRAID 640 4 Port SATA-III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1601u16 , name : "RocketRAID 640L 4 Port SATA-III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1602u16 , name : "RocketRAID 642L SATA-III Controller (2 eSATA ports + 2 internal SATA ports)" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1604u16 , name : "RocketRAID 644 4 Port SATA-III Controller (eSATA)" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1605u16 , name : "RocketRAID 644L 4 Port SATA-III Controller (eSATA)" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1606u16 , name : "RocketRAID 644LS SATA-III Controller (4 eSATA devices connected by 1 SAS cable)" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 1872u16 , name : "Rocket 750 PCIe Gen2 SATA III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 2112u16 , name : "RocketRAID 840 PCIe Gen3 SATA III Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 5920u16 , name : "RocketRAID 1720 (2x SATA II RAID Controller)" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 5952u16 , name : "RocketRAID 1740" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 5954u16 , name : "RocketRAID 1742" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 8720u16 , name : "RocketRAID 2210 SATA-II Controller" , subsystems : & [SubSystem { subvendor : 4523u16 , subdevice : 4523u16 , name : "88SX6042" }] } , Device { vendor_id : 4355u16 , id : 8960u16 , name : "RocketRAID 230x 4 Port SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 8976u16 , name : "RocketRAID 2310 4 Port SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 8992u16 , name : "RocketRAID 2320 SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 8994u16 , name : "RocketRAID 2322 SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 9024u16 , name : "RocketRAID 2340 16 Port SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 9792u16 , name : "RocketRAID 2640 SAS/SATA Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10016u16 , name : "RocketRAID 2720 PCIe Gen2 6Gb/s SAS/SATA Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10018u16 , name : "RocketRAID 2722" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10048u16 , name : "RocketRAID 2740" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10052u16 , name : "RocketRaid 2744" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10114u16 , name : "RocketRAID 2782" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 10304u16 , name : "RocketRAID 2840 PCIe Gen3 6Gb/s SAS/SATA Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 12576u16 , name : "RocketRAID 3120" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 12832u16 , name : "RocketRAID 3220" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 13088u16 , name : "RocketRAID 3320" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 13600u16 , name : "RocketRAID 3520 PCIe Gen1 8-Port SATA II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 13616u16 , name : "RocketRAID 3530 PCIe Gen1 12-Port SATA II Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 14144u16 , name : "RocketRAID 3740 PCIe Gen3 12Gb/s SAS/SATA Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 17168u16 , name : "RocketRaid 4310" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 17184u16 , name : "RocketRAID 4320 SAS Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 28931u16 , name : "SSD7103 PCIe Gen3 x16 4-Port M.2 NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 28933u16 , name : "SSD7105 PCIe Gen3 x16 4-Port M.2 NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 28944u16 , name : "SSD7110 PCIe Gen3 x16 NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 29957u16 , name : "SSD7505 PCIe Gen4 x16 4-Port M.2 NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 30016u16 , name : "SSD7540 PCIe Gen4 x16 8-Port M.2 NVMe RAID Controller" , subsystems : & [] } , Device { vendor_id : 4355u16 , id : 30080u16 , name : "SSD7580 PCIe Gen4 x16 8-Port M.2 NVMe RAID Controller" , subsystems : & [] }] }),
        (4870, Vendor { id : 4870u16 , name : "Duet Technologies" , devices : & [] }),
        (4699, Vendor { id : 4699u16 , name : "Asix Electronics Corporation" , devices : & [Device { vendor_id : 4699u16 , id : 5120u16 , name : "AX88141 Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 4352u16 , name : "AX8814X Based PCI Fast Ethernet Adapter" }] } , Device { vendor_id : 4699u16 , id : 37120u16 , name : "AX99100 PCIe to Multi I/O Controller" , subsystems : & [SubSystem { subvendor : 40960u16 , subdevice : 4096u16 , name : "Serial Port" } , SubSystem { subvendor : 40960u16 , subdevice : 8192u16 , name : "Parallel Port" } , SubSystem { subvendor : 40960u16 , subdevice : 24576u16 , name : "SPI" } , SubSystem { subvendor : 40960u16 , subdevice : 28672u16 , name : "Local Bus" } , SubSystem { subvendor : 59984u16 , subdevice : 7184u16 , name : "RXi2-BP Serial Port" }] } , Device { vendor_id : 4699u16 , id : 37125u16 , name : "AX99100 PCIe to I/O Bridge" , subsystems : & [] }] }),
        (9798, Vendor { id : 9798u16 , name : "Kingston Technology Company, Inc." , devices : & [Device { vendor_id : 9798u16 , id : 16u16 , name : "HyperX Predator PCIe AHCI SSD" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 8802u16 , name : "KC2000/KC2500 NVMe SSD [SM2262EN]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 8803u16 , name : "A2000 NVMe SSD [SM2263EN]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20488u16 , name : "A1000/U-SNS8154P3 x2 NVMe SSD [E8]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20490u16 , name : "DC1000B NVMe SSD [E12DC]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20491u16 , name : "DC1000M NVMe SSD [SM2270]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20492u16 , name : "OM8PCP3 PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20493u16 , name : "OM3PDP3 PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20494u16 , name : "NV1 NVMe SSD [E13T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20495u16 , name : "NV1 NVMe SSD [SM2263XT] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20496u16 , name : "OM8SBP NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20498u16 , name : "DC1500M NVMe SSD [SM2270]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20499u16 , name : "KC3000/FURY Renegade NVMe SSD [E18]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20500u16 , name : "OM8SEP4 PCIe 4 NVMe SSD (TLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20502u16 , name : "OM3PGP4 PCIe 4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20503u16 , name : "NV2 NVMe SSD [SM2267XT] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20504u16 , name : "OM8SFP4 PCIe 4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20505u16 , name : "NV2 NVMe SSD [E21T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20506u16 , name : "OM8PGP4 PCIe 4 NVMe SSD (TLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20507u16 , name : "OM8PGP4 NVMe PCIe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20508u16 , name : "NV2 NVMe SSD [E19T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20509u16 , name : "NV2 NVMe SSD [TC2200] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20510u16 , name : "OM3PGP4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20511u16 , name : "FURY Renegade NVMe SSD [E18] (Heatsink)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20513u16 , name : "OM8SEP4 PCIe 4 NVMe SSD (QLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20514u16 , name : "OM8PGP4 PCIe 4 NVMe SSD (QLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20515u16 , name : "NV2 NVMe SSD [SM2269XT] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20516u16 , name : "DC2000B NVMe SSD [E18DC]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20517u16 , name : "NV3 NVMe SSD [TC2201] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20518u16 , name : "NV3 NVMe SSD [E21T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20519u16 , name : "NV3 NVMe SSD [E27T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20520u16 , name : "NV3 NVMe SSD [SM2268XT2] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20521u16 , name : "OM8SGP4 PCIe 4 NVMe SSD (TLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20522u16 , name : "FURY Renegade G5 NVMe SSD [SM2508]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20523u16 , name : "NV3 NVMe SSD [E29T] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20524u16 , name : "DC3000ME NVMe SSD [SC5]" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20525u16 , name : "OM8TAP4 PCIe 4 NVMe SSD (QLC) (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20527u16 , name : "OM3SGP4 PCIe 4 NVMe SSD (TLC)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20528u16 , name : "NV3 2230 NVMe SSD [SM2268XT2] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 9798u16 , id : 20532u16 , name : "NV3 NVMe SSD [E33T] (DRAM-less)" , subsystems : & [] }] }),
        (5217, Vendor { id : 5217u16 , name : "Avermedia Technologies Inc" , devices : & [Device { vendor_id : 5217u16 , id : 41934u16 , name : "M179" , subsystems : & [] } , Device { vendor_id : 5217u16 , id : 41935u16 , name : "M179" , subsystems : & [] } , Device { vendor_id : 5217u16 , id : 43062u16 , name : "M115 DVB-T, PAL/SECAM/NTSC Tuner" , subsystems : & [] } , Device { vendor_id : 5217u16 , id : 59446u16 , name : "M115S Hybrid Analog/DVB PAL/SECAM/NTSC Tuner" , subsystems : & [] } , Device { vendor_id : 5217u16 , id : 62518u16 , name : "AVerTV Hybrid+FM" , subsystems : & [] }] }),
        (6013, Vendor { id : 6013u16 , name : "Cavium, Inc." , devices : & [Device { vendor_id : 6013u16 , id : 1u16 , name : "Nitrox XL N1" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 3u16 , name : "Nitrox XL N1 Lite" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 4u16 , name : "Octeon (and older) FIPS" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 5u16 , name : "Octeon CN38XX Network Processor Pass 3.x" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 6u16 , name : "RoHS" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 16u16 , name : "CN15XX/CN16XX [Nitrox PX]" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 17u16 , name : "CNN35XX [Nitrox III]" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 18u16 , name : "CNN55XX [Nitrox V]" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 32u16 , name : "Octeon CN31XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 48u16 , name : "Octeon CN30XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 64u16 , name : "Octeon CN58XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 80u16 , name : "Octeon CN57XX Network Processor (CN54XX/CN55XX/CN56XX)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 112u16 , name : "Octeon CN50XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 128u16 , name : "Octeon CN52XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 144u16 , name : "Octeon II CN63XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 145u16 , name : "Octeon II CN68XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 146u16 , name : "Octeon II CN65XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 147u16 , name : "Octeon II CN61XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 148u16 , name : "Octeon Fusion CNF71XX Cell processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 149u16 , name : "Octeon III CN78XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 150u16 , name : "Octeon III CN70XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 38656u16 , name : "Octeon III CN73XX Network Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 38658u16 , name : "CN23XX [LiquidIO II] Intelligent Adapter" , subsystems : & [SubSystem { subvendor : 6013u16 , subdevice : 3u16 , name : "CN2350 [LiquidIO II] 2-port 10GbE Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 4u16 , name : "CN2350 [LiquidIO II] 2-port 10GbE Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 5u16 , name : "CN2360 [LiquidIO II] 2-port 10GbE Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 6u16 , name : "CN2360 [LiquidIO II] 2-port 25GbE Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 7u16 , name : "CN2350 [LiquidIO II] 2-port 25GbE Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 8u16 , name : "CN2350 [LiquidIO II] 2-port 10GbE SFP+ Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 9u16 , name : "CN2360 [LiquidIO II] 2-port 10GbE SFP+ Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 10u16 , name : "CN2350 [LiquidIO II] 2-port 10GBASE-T Intelligent adapter" } , SubSystem { subvendor : 6013u16 , subdevice : 11u16 , name : "CN2360 [LiquidIO II] 2-port 10GBASE-T Intelligent adapter" }] } , Device { vendor_id : 6013u16 , id : 38659u16 , name : "CN23XX [LiquidIO II] NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 38674u16 , name : "CN23XX [LiquidIO II] SRIOV Virtual Function" , subsystems : & [SubSystem { subvendor : 6013u16 , subdevice : 3u16 , name : "CN2350 [LiquidIO II] 2-port 10GbE SRIOV Virtual Function" } , SubSystem { subvendor : 6013u16 , subdevice : 4u16 , name : "CN2350 [LiquidIO II] 2-port 10GbE SRIOV Virtual Function" } , SubSystem { subvendor : 6013u16 , subdevice : 5u16 , name : "CN2360 [LiquidIO II] 2-port 10GbE SRIOV Virtual Function" } , SubSystem { subvendor : 6013u16 , subdevice : 6u16 , name : "CN2360 [LiquidIO II] 2-port 25GbE SRIOV Virtual Function" } , SubSystem { subvendor : 6013u16 , subdevice : 7u16 , name : "CN2350 [LiquidIO II] 2-port 25GbE SRIOV Virtual Function" }] } , Device { vendor_id : 6013u16 , id : 38675u16 , name : "CN23XX [LiquidIO II] NVMe SRIOV Virtual Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 38912u16 , name : "Octeon Fusion CNF75XX Processor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40961u16 , name : "ThunderX MRML(Master RML Bridge to RSL devices)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40962u16 , name : "THUNDERX PCC Bridge" , subsystems : & [SubSystem { subvendor : 6013u16 , subdevice : 41218u16 , name : "CN88XX PCC Bridge" }] } , Device { vendor_id : 6013u16 , id : 40968u16 , name : "THUNDERX SMMU" , subsystems : & [SubSystem { subvendor : 6013u16 , subdevice : 41224u16 , name : "CN88XX SMMU" }] } , Device { vendor_id : 6013u16 , id : 40969u16 , name : "THUNDERX Generic Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40970u16 , name : "THUNDERX GPIO Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40971u16 , name : "THUNDERX MPI / SPI Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40972u16 , name : "THUNDERX MIO-PTP Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40973u16 , name : "THUNDERX MIX Network Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40974u16 , name : "THUNDERX Reset Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40975u16 , name : "THUNDERX UART Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40976u16 , name : "THUNDERX eMMC/SD Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40977u16 , name : "THUNDERX MIO-BOOT Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40978u16 , name : "THUNDERX TWSI / I2C Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40979u16 , name : "THUNDERX CCPI (Multi-node connect)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40980u16 , name : "THUNDERX Voltage Regulator Module" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40981u16 , name : "THUNDERX PCIe Switch Logic Interface" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40982u16 , name : "THUNDERX Key Memory" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40983u16 , name : "THUNDERX GTI (Global System Timers)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40984u16 , name : "THUNDERX Random Number Generator" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40985u16 , name : "THUNDERX DFA" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40986u16 , name : "THUNDERX Zip Coprocessor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40987u16 , name : "THUNDERX xHCI USB Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40988u16 , name : "THUNDERX AHCI SATA Controller" , subsystems : & [SubSystem { subvendor : 6013u16 , subdevice : 41244u16 , name : "CN88XX AHCI SATA Controller" }] } , Device { vendor_id : 6013u16 , id : 40989u16 , name : "THUNDERX RAID Coprocessor" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40990u16 , name : "THUNDERX Network Interface Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40991u16 , name : "THUNDERX Traffic Network Switch" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40992u16 , name : "THUNDERX PEM (PCI Express Interface)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40993u16 , name : "THUNDERX L2C (Level-2 Cache Controller)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40994u16 , name : "THUNDERX LMC (DRAM Controller)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40995u16 , name : "THUNDERX OCLA (On-Chip Logic Analyzer)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40996u16 , name : "THUNDERX OSM" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40997u16 , name : "THUNDERX GSER (General Serializer/Deserializer)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40998u16 , name : "THUNDERX BGX (Common Ethernet Interface)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 40999u16 , name : "THUNDERX IOBN" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41001u16 , name : "THUNDERX NCSI (Network Controller Sideband Interface)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41002u16 , name : "ThunderX SGPIO (Serial GPIO controller for SATA disk lights)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41003u16 , name : "THUNDERX SMI / MDIO Controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41004u16 , name : "THUNDERX DAP (Debug Access Port)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41005u16 , name : "THUNDERX PCIERC (PCIe Root Complex)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41006u16 , name : "ThunderX L2C-TAD (Level 2 cache tag and data)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41007u16 , name : "THUNDERX L2C-CBC" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41008u16 , name : "THUNDERX L2C-MCI" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41009u16 , name : "THUNDERX MIO-FUS (Fuse Access Controller)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41010u16 , name : "THUNDERX FUSF (Fuse Controller)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41011u16 , name : "THUNDERX Random Number Generator virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41012u16 , name : "THUNDERX Network Interface Controller virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41013u16 , name : "THUNDERX Parallel Bus" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41014u16 , name : "ThunderX RAD (RAID acceleration engine) virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41015u16 , name : "THUNDERX ZIP virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41024u16 , name : "THUNDERX CPT Cryptographic Accelerator" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41049u16 , name : "Octeon TX2 CGX (MAC)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41056u16 , name : "Octeon 10 RPM (MAC)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41057u16 , name : "Octeon Tx2 Loopback Interface (LBK)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41059u16 , name : "Octeon Tx2 RVU Physical Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41060u16 , name : "Octeon Tx2 RVU Virtual Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41061u16 , name : "Octeon Tx2 RVU Admin Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41118u16 , name : "Octeon 10 PTP controller" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41202u16 , name : "Octeon 10 CPT Cryptographic Accelerator, Physical function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41203u16 , name : "Octeon 10 CPT Cryptographic Accelerator, Virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41206u16 , name : "Octeon Tx2 SDP Physical Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41207u16 , name : "Octeon Tx2 SDP Virtual Function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41208u16 , name : "Octeon Tx2 Loopback Interface Virtual Function (LBKVF)" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41213u16 , name : "Octeon Tx2 CPT Cryptographic Accelerator, Physical function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41214u16 , name : "Octeon Tx2 CPT Cryptographic Accelerator, Virtual function" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41216u16 , name : "THUNDERX CN88XX 48 core SoC" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41472u16 , name : "OCTEON TX CN81XX/CN80XX" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 41728u16 , name : "OCTEON TX CN83XX" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 44800u16 , name : "CN99xx [ThunderX2] Integrated PCI Host bridge" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 44932u16 , name : "CN99xx [ThunderX2] Integrated PCI Express RP Bridge" , subsystems : & [] } , Device { vendor_id : 6013u16 , id : 61456u16 , name : "XPliant CNX880xx Network Processor" , subsystems : & [] }] }),
        (4788, Vendor { id : 4788u16 , name : "FutureTel Inc" , devices : & [] }),
        (8161, Vendor { id : 8161u16 , name : "Beijing ESWIN Computing Technology Co., Ltd." , devices : & [Device { vendor_id : 8161u16 , id : 8240u16 , name : "EIC7700 Root Complex" , subsystems : & [] }] }),
        (7999, Vendor { id : 7999u16 , name : "3SNIC Ltd" , devices : & [Device { vendor_id : 7999u16 , id : 8448u16 , name : "SSSHBA SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 288u16 , name : "HBA 32 Ports" } , SubSystem { subvendor : 7999u16 , subdevice : 293u16 , name : "HBA 40 Ports" } , SubSystem { subvendor : 7999u16 , subdevice : 384u16 , name : "HBA 16 Ports" } , SubSystem { subvendor : 7999u16 , subdevice : 389u16 , name : "HBA 8 Ports" }] } , Device { vendor_id : 7999u16 , id : 8704u16 , name : "SSSRAID RAID controller" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 389u16 , name : "RAID Controller 16 Ports with 2G Cache" } , SubSystem { subvendor : 7999u16 , subdevice : 417u16 , name : "RAID Controller 40 Ports with 2G Cache" } , SubSystem { subvendor : 7999u16 , subdevice : 420u16 , name : "RAID Controller 16 Ports with 4G Cache" } , SubSystem { subvendor : 7999u16 , subdevice : 424u16 , name : "RAID Controller 32 Ports with 4G Cache" } , SubSystem { subvendor : 7999u16 , subdevice : 429u16 , name : "RAID Controller 40 Ports with 4G Cache" }] } , Device { vendor_id : 7999u16 , id : 36865u16 , name : "SSSNIC Ethernet VF" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "Ethernet VF" }] } , Device { vendor_id : 7999u16 , id : 36866u16 , name : "SSSNIC Ethernet VF Hyper-V" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "Ethernet VF Hyper-V" }] } , Device { vendor_id : 7999u16 , id : 36867u16 , name : "SSSNIC Ethernet SDI VF" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "Ethernet SDI VF" }] } , Device { vendor_id : 7999u16 , id : 36868u16 , name : "SSSNIC Ethernet SDI VF Hyper-V" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "Ethernet SDI VF Hyper-V" }] } , Device { vendor_id : 7999u16 , id : 36872u16 , name : "SSSFC VF" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "3SFC VF" }] } , Device { vendor_id : 7999u16 , id : 36873u16 , name : "SSSFC VF Hyper-V" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 4096u16 , name : "3SFC VF Hyper-V" }] } , Device { vendor_id : 7999u16 , id : 36880u16 , name : "SSSFC FC Controller" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 65u16 , name : "Dual Port 16GE FC to PCIe Gen4 x16 Adapter" } , SubSystem { subvendor : 7999u16 , subdevice : 97u16 , name : "Dual Port 32GE FC to PCIe Gen4 x16 Adapter" }] } , Device { vendor_id : 7999u16 , id : 36896u16 , name : "SSSNIC Ethernet Network Adapter" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 81u16 , name : "Dual Port 25GE Ethernet Network Adapter" } , SubSystem { subvendor : 7999u16 , subdevice : 82u16 , name : "Quad Port 25GE Ethernet Network Adapter" } , SubSystem { subvendor : 7999u16 , subdevice : 161u16 , name : "Dual Port 100GE Ethernet Network Adapter" }] } , Device { vendor_id : 7999u16 , id : 36912u16 , name : "SSSNIC IPU" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 81u16 , name : "Dual Port 25GE IPU" }] } , Device { vendor_id : 7999u16 , id : 36913u16 , name : "SSSNIC SDI5.0" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 81u16 , name : "Dual Port 25GE SDI5.0lite" } , SubSystem { subvendor : 7999u16 , subdevice : 161u16 , name : "Dual Port 100GE SDI5.0" }] } , Device { vendor_id : 7999u16 , id : 36914u16 , name : "SSSNIC SDI5.1" , subsystems : & [SubSystem { subvendor : 7999u16 , subdevice : 161u16 , name : "Dual Port 100GE SDI5.1" }] }] }),
        (7927, Vendor { id : 7927u16 , name : "Shenzhen Gunnir Technology Development Co., Ltd" , devices : & [] }),
        (4628, Vendor { id : 4628u16 , name : "Performance Technologies, Inc." , devices : & [] }),
        (7497, Vendor { id : 7497u16 , name : "Lenovo" , devices : & [] }),
        (4795, Vendor { id : 4795u16 , name : "Nippon Unisoft Corporation" , devices : & [] }),
        (4817, Vendor { id : 4817u16 , name : "PSITech" , devices : & [] }),
        (8412, Vendor { id : 8412u16 , name : "Sharetronic Data Technology Co., Ltd." , devices : & [Device { vendor_id : 8412u16 , id : 4610u16 , name : "M.2 2280 PCIe Gen3 x 4 Series." , subsystems : & [SubSystem { subvendor : 4610u16 , subdevice : 4694u16 , name : "M.2 2280 256GB 1202+N38A" }] } , Device { vendor_id : 8412u16 , id : 20480u16 , name : "E5000 U.2 15mm 3.84TB NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8412u16 , id : 20737u16 , name : "E5000 U.2 15mm 7.68TB NVMe SSD" , subsystems : & [] }] }),
        (6769, Vendor { id : 6769u16 , name : "XenSource, Inc." , devices : & [] }),
        (4459, Vendor { id : 4459u16 , name : "Connectware Inc" , devices : & [] }),
        (4540, Vendor { id : 4540u16 , name : "Network Peripherals Inc" , devices : & [Device { vendor_id : 4540u16 , id : 1u16 , name : "NP-PCI" , subsystems : & [] }] }),
        (7488, Vendor { id : 7488u16 , name : "Techman Electronics (Changshu) Co., Ltd." , devices : & [Device { vendor_id : 7488u16 , id : 21761u16 , name : "XC100C55-xxxx NVME SSD" , subsystems : & [] } , Device { vendor_id : 7488u16 , id : 23553u16 , name : "XC100C5C-xxxx, XC100E5C-xxxx NVME SSD" , subsystems : & [] } , Device { vendor_id : 7488u16 , id : 45312u16 , name : "PV100C55-xxxx NVME SSD" , subsystems : & [] }] }),
        (4787, Vendor { id : 4787u16 , name : "Inter-Face Co Ltd" , devices : & [] }),
        (61904, Vendor { id : 61904u16 , name : "AJA Video" , devices : & [Device { vendor_id : 61904u16 , id : 49406u16 , name : "Xena HS/HD-R" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 49407u16 , name : "Kona/Xena 2" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 51966u16 , name : "Kona SD" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 53230u16 , name : "Xena LS/SD-22-DA/SD-DA" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56062u16 , name : "Corvid 1" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56063u16 , name : "KONA LHi" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56064u16 , name : "IoExpress" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56065u16 , name : "Corvid22" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56066u16 , name : "Kona 3G" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56067u16 , name : "Corvid 3G" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56068u16 , name : "Kona 3G QUAD" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56069u16 , name : "Kona LHe+" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56070u16 , name : "IoXT" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56071u16 , name : "Kona 3G P2P" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56072u16 , name : "Kona 3G QUAD P2P" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56073u16 , name : "Corvid 24" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56081u16 , name : "T-Tap" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 56495u16 , name : "Kona HD" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 57326u16 , name : "Xena HD-DA" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60167u16 , name : "Io4K" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60170u16 , name : "Io4K UFC" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60171u16 , name : "Kona 4" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60172u16 , name : "Kona 4 UFC" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60173u16 , name : "Corvid 88" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60174u16 , name : "Corvid 44" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60182u16 , name : "Corvid HEVC" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4169u16 , name : "Corvid HEVC M31" }] } , Device { vendor_id : 61904u16 , id : 60184u16 , name : "Corvid HB-R" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60186u16 , name : "Kona IP 1SFP" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60188u16 , name : "Kona IP 2SFP" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60189u16 , name : "Io4KPlus" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60190u16 , name : "IoIP" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60191u16 , name : "Kona 5" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60195u16 , name : "Kona 1" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60196u16 , name : "Kona HDMI" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60197u16 , name : "Corvid 44 12g" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60198u16 , name : "T-Tap Pro" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60199u16 , name : "IoX3" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60200u16 , name : "Kona XM" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60201u16 , name : "Kona X" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 60208u16 , name : "Kona IP 25G" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 61356u16 , name : "Xena SD-MM/SD-22-MM" , subsystems : & [] } , Device { vendor_id : 61904u16 , id : 64205u16 , name : "Xena HD-MM" , subsystems : & [] }] }),
        (43605, Vendor { id : 43605u16 , name : "Ncomputing X300 PCI-Engine" , devices : & [] }),
        (4685, Vendor { id : 4685u16 , name : "Stallion Technologies, Inc." , devices : & [Device { vendor_id : 4685u16 , id : 0u16 , name : "EasyConnection 8/32" , subsystems : & [] } , Device { vendor_id : 4685u16 , id : 2u16 , name : "EasyConnection 8/64" , subsystems : & [] } , Device { vendor_id : 4685u16 , id : 3u16 , name : "EasyIO" , subsystems : & [] } , Device { vendor_id : 4685u16 , id : 4u16 , name : "EasyConnection/RA" , subsystems : & [] }] }),
        (5249, Vendor { id : 5249u16 , name : "BIOPAC Systems Inc" , devices : & [] }),
        (53772, Vendor { id : 53772u16 , name : "Chengdu NorthLink Technology Co., Ltd." , devices : & [Device { vendor_id : 53772u16 , id : 20497u16 , name : "NE5025 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 53772u16 , subdevice : 57632u16 , name : "N5 Series 2-port 10GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57664u16 , name : "N5 Series 4-port 10GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57888u16 , name : "N5 Series 2-port 25GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57889u16 , name : "N5 Series 2-port 25GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57900u16 , name : "N5 Series 2-port 25GbE Network Adapter for OCP" }] } , Device { vendor_id : 53772u16 , id : 24593u16 , name : "NE6100 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 53772u16 , subdevice : 40961u16 , name : "N6 Series Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57902u16 , name : "N6 Series 2-port 25GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 57985u16 , name : "N6 Series 8-port 25GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 58401u16 , name : "N6 Series 2-port 40GbE Network Adapter" } , SubSystem { subvendor : 53772u16 , subdevice : 59937u16 , name : "N6 Series 2-port 100GbE Network Adapter" }] }] }),
        (5365, Vendor { id : 5365u16 , name : "SOPAC Ltd" , devices : & [] }),
        (4123, Vendor { id : 4123u16 , name : "Vitesse Semiconductor" , devices : & [Device { vendor_id : 4123u16 , id : 1106u16 , name : "VSC452 [SuperBMC]" , subsystems : & [] }] }),
        (5283, Vendor { id : 5283u16 , name : "Maverick Networks" , devices : & [] }),
        (5335, Vendor { id : 5335u16 , name : "Hirakawa Hewtech Corp" , devices : & [] }),
        (7529, Vendor { id : 7529u16 , name : "Renesas Electronics Corp." , devices : & [Device { vendor_id : 7529u16 , id : 9266u16 , name : "CL2432" , subsystems : & [] } , Device { vendor_id : 7529u16 , id : 9280u16 , name : "CL2440" , subsystems : & [] } , Device { vendor_id : 7529u16 , id : 32768u16 , name : "CL80x0 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 7529u16 , id : 32838u16 , name : "CL8046 Wireless Network Adapter" , subsystems : & [] }] }),
        (4331, Vendor { id : 4331u16 , name : "Artists Graphics" , devices : & [Device { vendor_id : 4331u16 , id : 257u16 , name : "3GA" , subsystems : & [] } , Device { vendor_id : 4331u16 , id : 33041u16 , name : "Twist3 Frame Grabber" , subsystems : & [] }] }),
        (4814, Vendor { id : 4814u16 , name : "Netspeed Inc." , devices : & [] }),
        (5496, Vendor { id : 5496u16 , name : "HITT" , devices : & [Device { vendor_id : 5496u16 , id : 19764u16 , name : "VPMK4 [Video Processor Mk IV]" , subsystems : & [] } , Device { vendor_id : 5496u16 , id : 22037u16 , name : "VPMK3 [Video Processor Mk III]" , subsystems : & [] }] }),
        (4554, Vendor { id : 4554u16 , name : "IBEX Technology Co., Ltd." , devices : & [Device { vendor_id : 4554u16 , id : 57u16 , name : "PAC with Altera Agilex 7 FPGA M-Series [IPAC-1000]" , subsystems : & [] }] }),
        (5200, Vendor { id : 5200u16 , name : "Octave Communications Ind." , devices : & [] }),
        (5178, Vendor { id : 5178u16 , name : "Stargate Solutions Inc" , devices : & [] }),
        (8140, Vendor { id : 8140u16 , name : "StreamLabs" , devices : & [Device { vendor_id : 8140u16 , id : 62486u16 , name : "MS416" , subsystems : & [] } , Device { vendor_id : 8140u16 , id : 64257u16 , name : "MH4LM" , subsystems : & [] }] }),
        (5511, Vendor { id : 5511u16 , name : "Konica Corp" , devices : & [] }),
        (7847, Vendor { id : 7847u16 , name : "Intelliprop, Inc" , devices : & [Device { vendor_id : 7847u16 , id : 8762u16 , name : "Typhon+ PCIe to Gen-Z Bridge" , subsystems : & [] } , Device { vendor_id : 7847u16 , id : 8778u16 , name : "IPA-PE224A CXL to Gen-Z Bridge [Sphinx]" , subsystems : & [] }] }),
        (4574, Vendor { id : 4574u16 , name : "Zoran Corporation" , devices : & [Device { vendor_id : 4574u16 , id : 24599u16 , name : "miroVIDEO DC30" , subsystems : & [] } , Device { vendor_id : 4574u16 , id : 24663u16 , name : "ZR36057PQC Video cutting chipset" , subsystems : & [SubSystem { subvendor : 4145u16 , subdevice : 32510u16 , name : "DC10 Plus" } , SubSystem { subvendor : 4145u16 , subdevice : 64512u16 , name : "MiroVIDEO DC50, Motion JPEG Capture/CODEC Board" } , SubSystem { subvendor : 4856u16 , subdevice : 35330u16 , name : "Tekram Video Kit" } , SubSystem { subvendor : 5066u16 , subdevice : 16945u16 , name : "JPEG/TV Card" }] } , Device { vendor_id : 4574u16 , id : 24864u16 , name : "ZR36120" , subsystems : & [SubSystem { subvendor : 4904u16 , subdevice : 61441u16 , name : "Cinemaster C DVD Decoder" } , SubSystem { subvendor : 5058u16 , subdevice : 0u16 , name : "MediaFocus Satellite TV Card" } , SubSystem { subvendor : 7649u16 , subdevice : 40959u16 , name : "Video Kit C210" }] }] }),
        (52307, Vendor { id : 52307u16 , name : "ScaleFlux Inc." , devices : & [Device { vendor_id : 52307u16 , id : 1u16 , name : "CSS 1000" , subsystems : & [] } , Device { vendor_id : 52307u16 , id : 16u16 , name : "CSD 3310" , subsystems : & [] }] }),
        (4751, Vendor { id : 4751u16 , name : "Tateno Dennou, Inc." , devices : & [] }),
        (5627, Vendor { id : 5627u16 , name : "Zilog Inc" , devices : & [] }),
        (4386, Vendor { id : 4386u16 , name : "Multi-tech Systems, Inc." , devices : & [] }),
        (4620, Vendor { id : 4620u16 , name : "Technical Corp." , devices : & [] }),
        (5751, Vendor { id : 5751u16 , name : "B&R Industrial Automation GmbH" , devices : & [Device { vendor_id : 5751u16 , id : 4174u16 , name : "5LS172.6 B&R Dual CAN Interface Card" , subsystems : & [] } , Device { vendor_id : 5751u16 , id : 4823u16 , name : "5LS172.61 B&R Dual CAN Interface Card" , subsystems : & [] } , Device { vendor_id : 5751u16 , id : 8365u16 , name : "5ACPCI.MFIO-K01 Profibus DP / K-Feldbus / COM" , subsystems : & [] }] }),
        (4145, Vendor { id : 4145u16 , name : "Miro Computer Products AG" , devices : & [Device { vendor_id : 4145u16 , id : 22017u16 , name : "DC20 ASIC" , subsystems : & [] } , Device { vendor_id : 4145u16 , id : 22023u16 , name : "Video I/O & motion JPEG compressor" , subsystems : & [] } , Device { vendor_id : 4145u16 , id : 22065u16 , name : "Media 3D" , subsystems : & [] } , Device { vendor_id : 4145u16 , id : 24663u16 , name : "MiroVideo DC10/DC30+" , subsystems : & [] }] }),
        (19617, Vendor { id : 19617u16 , name : "Seanix Technology Inc" , devices : & [] }),
        (8453, Vendor { id : 8453u16 , name : "Shanghai Timar Integrated Circuit Co., LTD" , devices : & [] }),
        (5058, Vendor { id : 5058u16 , name : "Technotrend Systemtechnik GmbH" , devices : & [Device { vendor_id : 5058u16 , id : 14u16 , name : "Technotrend/Hauppauge DVB card rev2.3" , subsystems : & [] } , Device { vendor_id : 5058u16 , id : 4121u16 , name : "TTechnoTrend-budget DVB S2-3200" , subsystems : & [] }] }),
        (4255, Vendor { id : 4255u16 , name : "Trigem Computer Inc." , devices : & [] }),
        (4354, Vendor { id : 4354u16 , name : "Creative Labs" , devices : & [Device { vendor_id : 4354u16 , id : 2u16 , name : "EMU10k1 [Sound Blaster Live! Series]" , subsystems : & [SubSystem { subvendor : 4106u16 , subdevice : 4354u16 , name : "SB Live! 5.1 Digital OEM SB0220 EMU10K1-JFF" } , SubSystem { subvendor : 4354u16 , subdevice : 32u16 , name : "CT4670/4850 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 33u16 , name : "CT4620 SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 47u16 , name : "M002/M003 Integrated SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 4106u16 , name : "SB0220/0229 SBLive! 5.1 Digital OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 16385u16 , name : "E-mu APS" } , SubSystem { subvendor : 4354u16 , subdevice : 32802u16 , name : "CT4780 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32803u16 , name : "CT4790 SoundBlaster PCI512" } , SubSystem { subvendor : 4354u16 , subdevice : 32804u16 , name : "CT4760 SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 32805u16 , name : "CT1140/SB0040 Integrated SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 32806u16 , name : "CT4830 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32807u16 , name : "CT4832 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32808u16 , name : "CT4870 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32809u16 , name : "CT4872 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32810u16 , name : "CT4890 SoundBlaster PCI256" } , SubSystem { subvendor : 4354u16 , subdevice : 32811u16 , name : "CT4891 SoundBlaster PCI256" } , SubSystem { subvendor : 4354u16 , subdevice : 32817u16 , name : "CT4831 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32818u16 , name : "CT4871 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32819u16 , name : "CT4893 SoundBlaster PCI256" } , SubSystem { subvendor : 4354u16 , subdevice : 32821u16 , name : "CT0060 SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 32832u16 , name : "CT4760 SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 32848u16 , name : "CT4750 SoundBlaster PCI512" } , SubSystem { subvendor : 4354u16 , subdevice : 32849u16 , name : "CT4850 SBLive! Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32865u16 , name : "SB060 SBLive! Player 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32866u16 , name : "SB0100 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32867u16 , name : "DXW Integrated SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32868u16 , name : "SB0100/SB0102 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32869u16 , name : "SB0220/0222 SBLive! 5.1 Digital" } , SubSystem { subvendor : 4354u16 , subdevice : 32870u16 , name : "SB0228 SBLive! 5.1 Digital" } , SubSystem { subvendor : 4354u16 , subdevice : 32871u16 , name : "SB0220 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32872u16 , name : "CT0061 SBLive!" } , SubSystem { subvendor : 4354u16 , subdevice : 32873u16 , name : "SB0101 SBLive! 5.1 Value" } , SubSystem { subvendor : 4354u16 , subdevice : 32874u16 , name : "SB0103 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32875u16 , name : "SB0105 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32876u16 , name : "SB0221 SBLive! 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 32881u16 , name : "SB0150 SoundBlaster PCI512" }] } , Device { vendor_id : 4354u16 , id : 3u16 , name : "SB AWE64(D)" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 16u16 , name : "CT4600 AWE64D" } , SubSystem { subvendor : 4354u16 , subdevice : 48u16 , name : "CT4650 AWE64D" } , SubSystem { subvendor : 4354u16 , subdevice : 49u16 , name : "CT4655 AWE64D" }] } , Device { vendor_id : 4354u16 , id : 4u16 , name : "EMU10k2/CA0100/CA0102/CA10200 [Sound Blaster Audigy Series]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 64u16 , name : "SB0090 Audigy Player" } , SubSystem { subvendor : 4354u16 , subdevice : 65u16 , name : "CT4820 SBLive!2" } , SubSystem { subvendor : 4354u16 , subdevice : 66u16 , name : "CT0070 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 67u16 , name : "CT0072 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 81u16 , name : "SB0090 Audigy Player/Platinum (EX)" } , SubSystem { subvendor : 4354u16 , subdevice : 82u16 , name : "SB0162 Audigy ES" } , SubSystem { subvendor : 4354u16 , subdevice : 83u16 , name : "CT0090/SB0092 Audigy Player/OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 84u16 , name : "SB0161 Audigy ES" } , SubSystem { subvendor : 4354u16 , subdevice : 85u16 , name : "SB0192 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 86u16 , name : "SB0191 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 87u16 , name : "SB0091 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 88u16 , name : "SB0095 Audigy Player/OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 89u16 , name : "SB0230 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 90u16 , name : "SB0231 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 91u16 , name : "SB0232 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 92u16 , name : "SB0238 Audigy" } , SubSystem { subvendor : 4354u16 , subdevice : 4098u16 , name : "SB0240 Audigy 2 Platinum 6.1" } , SubSystem { subvendor : 4354u16 , subdevice : 4099u16 , name : "SB0350 Audigy 2 / SB0243 Audigy 2 OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 4100u16 , name : "SB0242 Audigy 2" } , SubSystem { subvendor : 4354u16 , subdevice : 4101u16 , name : "SB0280 Audigy 2 Platinum Ex" } , SubSystem { subvendor : 4354u16 , subdevice : 4102u16 , name : "SB0245 Audigy 2 OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 4103u16 , name : "SB0240/SB0244 Audigy 2 Platinum" } , SubSystem { subvendor : 4354u16 , subdevice : 4104u16 , name : "SB0320 Audigy 2" } , SubSystem { subvendor : 4354u16 , subdevice : 4105u16 , name : "SB0249 Audigy 2 OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 4106u16 , name : "SB0246 Audigy 2" } , SubSystem { subvendor : 4354u16 , subdevice : 8193u16 , name : "SB0360 Audigy 2 ZS Platinum Pro" } , SubSystem { subvendor : 4354u16 , subdevice : 8194u16 , name : "SB0350 Audigy 2 ZS" } , SubSystem { subvendor : 4354u16 , subdevice : 8195u16 , name : "SB0352 Audigy 2 ZS" } , SubSystem { subvendor : 4354u16 , subdevice : 8196u16 , name : "SB0355 Audigy 2 ZS" } , SubSystem { subvendor : 4354u16 , subdevice : 8197u16 , name : "SB0359 Audigy 2 ZS" } , SubSystem { subvendor : 4354u16 , subdevice : 8198u16 , name : "SB035x Audigy 2 OEM" } , SubSystem { subvendor : 4354u16 , subdevice : 8199u16 , name : "SB0380 Audigy 4 Pro" } , SubSystem { subvendor : 4354u16 , subdevice : 16385u16 , name : "E-MU 1010 [MAEM8810]" } , SubSystem { subvendor : 4354u16 , subdevice : 16386u16 , name : "E-MU 0404" } , SubSystem { subvendor : 4354u16 , subdevice : 16387u16 , name : "E-MU 1010" }] } , Device { vendor_id : 4354u16 , id : 5u16 , name : "EMU20k1 [Sound Blaster X-Fi Series]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 33u16 , name : "X-Fi Platinum" } , SubSystem { subvendor : 4354u16 , subdevice : 44u16 , name : "X-Fi XtremeGamer FATAL1TY PRO" } , SubSystem { subvendor : 4354u16 , subdevice : 4099u16 , name : "X-Fi XtremeMusic" }] } , Device { vendor_id : 4354u16 , id : 6u16 , name : "EMU10k1X / CA0103 [SB Live! OEM / SB 5.1 / Ectiva 5.1]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 4097u16 , name : "SB0680 Sound Blaster 5.1" } , SubSystem { subvendor : 4354u16 , subdevice : 4099u16 , name : "SB0203 SB Live! 5.1 (Dell)" } , SubSystem { subvendor : 4354u16 , subdevice : 4100u16 , name : "TP0033 Ectiva Audio 5.1" }] } , Device { vendor_id : 4354u16 , id : 7u16 , name : "CA0106/CA0111 [SB Live!/Audigy/X-Fi Series]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 7u16 , name : "SBLive! 24bit" } , SubSystem { subvendor : 4354u16 , subdevice : 4097u16 , name : "SB0310 Audigy LS" } , SubSystem { subvendor : 4354u16 , subdevice : 4098u16 , name : "SB0312 Audigy LS" } , SubSystem { subvendor : 4354u16 , subdevice : 4102u16 , name : "SB0410 SBLive! 24-bit" } , SubSystem { subvendor : 4354u16 , subdevice : 4106u16 , name : "SB0570 [SB Audigy SE]" } , SubSystem { subvendor : 4354u16 , subdevice : 4114u16 , name : "SB0790 X-Fi XA" } , SubSystem { subvendor : 4354u16 , subdevice : 4115u16 , name : "Soundblaster X-Fi Xtreme Audio" } , SubSystem { subvendor : 5218u16 , subdevice : 4105u16 , name : "K8N Diamond" }] } , Device { vendor_id : 4354u16 , id : 8u16 , name : "CA0108/CA10300 [Sound Blaster Audigy Series]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 8u16 , name : "EMU0404 Digital Audio System" } , SubSystem { subvendor : 4354u16 , subdevice : 4097u16 , name : "SB0400 Audigy 2 Value" } , SubSystem { subvendor : 4354u16 , subdevice : 4129u16 , name : "SB0610 Audigy 4 Value" } , SubSystem { subvendor : 4354u16 , subdevice : 4130u16 , name : "SBxxx Audigy 2/4 Value" } , SubSystem { subvendor : 4354u16 , subdevice : 4131u16 , name : "SB0612 Audigy 2 LS" } , SubSystem { subvendor : 4354u16 , subdevice : 4132u16 , name : "SB1550 Audigy 5/Rx" } , SubSystem { subvendor : 4354u16 , subdevice : 4353u16 , name : "SBxxxx Audigy 2 SA" } , SubSystem { subvendor : 4354u16 , subdevice : 8193u16 , name : "SB0530 Audigy 2 ZS Notebook" } , SubSystem { subvendor : 4354u16 , subdevice : 8225u16 , name : "SBxxxx Audigy 4 Notebook" } , SubSystem { subvendor : 4354u16 , subdevice : 16386u16 , name : "E-MU 0404" } , SubSystem { subvendor : 4354u16 , subdevice : 16387u16 , name : "E-MU 1010" } , SubSystem { subvendor : 4354u16 , subdevice : 16388u16 , name : "EMU1010 Digital Audio System [MAEM8960]" } , SubSystem { subvendor : 4354u16 , subdevice : 16389u16 , name : "E-MU 0404 [MAEM8984]" } , SubSystem { subvendor : 4354u16 , subdevice : 16391u16 , name : "E-MU 1010 [MAEM8982]" } , SubSystem { subvendor : 4354u16 , subdevice : 16897u16 , name : "E-MU 0202 [MAEM8950]" }] } , Device { vendor_id : 4354u16 , id : 9u16 , name : "CA0110 [Sound Blaster X-Fi Xtreme Audio]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 16u16 , name : "MB0820 Integrated" } , SubSystem { subvendor : 4354u16 , subdevice : 24u16 , name : "SB1040 PCI Express" }] } , Device { vendor_id : 4354u16 , id : 11u16 , name : "EMU20k2 [Sound Blaster X-Fi Titanium Series]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 65u16 , name : "SB0880 [SoundBlaster X-Fi Titanium PCI-e]" } , SubSystem { subvendor : 4354u16 , subdevice : 98u16 , name : "SB1270 [SoundBlaster X-Fi Titanium HD]" }] } , Device { vendor_id : 4354u16 , id : 16u16 , name : "CA0132 Sound Core3D [Sound Blaster AE-7/AE-9]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 113u16 , name : "Sound Blaster AE-9" } , SubSystem { subvendor : 4354u16 , subdevice : 129u16 , name : "Sound Blaster AE-7" }] } , Device { vendor_id : 4354u16 , id : 18u16 , name : "CA0132 Sound Core3D [Sound Blaster Recon3D / Z-Series / Sound BlasterX AE-5 Plus]" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 16u16 , name : "SB1570 SB Audigy Fx" } , SubSystem { subvendor : 4354u16 , subdevice : 401u16 , name : "CA0132 Sound Core3D - CA0113 [ Sound BlasterX AE-5 Plus]" }] } , Device { vendor_id : 4354u16 , id : 16385u16 , name : "SB Audigy FireWire Port" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 16u16 , name : "SB Audigy FireWire Port" }] } , Device { vendor_id : 4354u16 , id : 28674u16 , name : "SB Live! Game Port" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 32u16 , name : "Gameport Joystick" }] } , Device { vendor_id : 4354u16 , id : 28675u16 , name : "SB Audigy Game Port" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 64u16 , name : "SB Audigy Game Port" } , SubSystem { subvendor : 4354u16 , subdevice : 96u16 , name : "SB Audigy2 MIDI/Game Port" }] } , Device { vendor_id : 4354u16 , id : 28676u16 , name : "[SB Live! Value] Input device controller" , subsystems : & [] } , Device { vendor_id : 4354u16 , id : 28677u16 , name : "SB Audigy LS Game Port" , subsystems : & [SubSystem { subvendor : 4354u16 , subdevice : 4097u16 , name : "SB0310 Audigy LS MIDI/Game port" } , SubSystem { subvendor : 4354u16 , subdevice : 4098u16 , name : "SB0312 Audigy LS MIDI/Game port" }] } , Device { vendor_id : 4354u16 , id : 28678u16 , name : "[SB X-Fi Xtreme Audio] CA0110-IBG PCIe to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4354u16 , id : 35128u16 , name : "Ectiva EV1938" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32997u16 , name : "SlimTower-Jim (NEC)" } , SubSystem { subvendor : 4209u16 , subdevice : 29008u16 , name : "Mitac 7150" } , SubSystem { subvendor : 4362u16 , subdevice : 22840u16 , name : "Siemens Scenic Mobile 510PIII" } , SubSystem { subvendor : 5053u16 , subdevice : 4108u16 , name : "Ceres-C (Sharp, Intel BX)" } , SubSystem { subvendor : 5053u16 , subdevice : 4109u16 , name : "Sharp, Intel Banister" } , SubSystem { subvendor : 5053u16 , subdevice : 4110u16 , name : "TwinHead P09S/P09S3 (Sharp)" } , SubSystem { subvendor : 5053u16 , subdevice : 63217u16 , name : "Marlin (Sharp)" } , SubSystem { subvendor : 5375u16 , subdevice : 3696u16 , name : "P88TE (TWINHEAD INTERNATIONAL Corp)" } , SubSystem { subvendor : 5375u16 , subdevice : 50177u16 , name : "Notebook 9100/9200/2000 (TWINHEAD INTERNATIONAL Corp)" } , SubSystem { subvendor : 5485u16 , subdevice : 46080u16 , name : "G400 - Geo (AlphaTop (Taiwan))" } , SubSystem { subvendor : 5485u16 , subdevice : 46416u16 , name : "G560  (AlphaTop (Taiwan))" } , SubSystem { subvendor : 5485u16 , subdevice : 46432u16 , name : "G560  (AlphaTop (Taiwan))" } , SubSystem { subvendor : 5485u16 , subdevice : 46848u16 , name : "G700/U700  (AlphaTop (Taiwan))" } , SubSystem { subvendor : 5485u16 , subdevice : 46997u16 , name : "G795  (AlphaTop (Taiwan))" } , SubSystem { subvendor : 5485u16 , subdevice : 46999u16 , name : "G797  (AlphaTop (Taiwan))" }] }] }),
        (7615, Vendor { id : 7615u16 , name : "Guizhou Huaxintong Semiconductor Technology Co., Ltd" , devices : & [Device { vendor_id : 7615u16 , id : 1025u16 , name : "StarDragon4800 PCI Express Root Port" , subsystems : & [] }] }),
        (7856, Vendor { id : 7856u16 , name : "Shenzhen Electrical Appliances CO." , devices : & [Device { vendor_id : 7856u16 , id : 6401u16 , name : "NVMe SSD Controller (DRAM-less)" , subsystems : & [] }] }),
        (12544, Vendor { id : 12544u16 , name : "Dynabook Inc." , devices : & [] }),
        (5155, Vendor { id : 5155u16 , name : "Custom Technology Corp." , devices : & [] }),
        (5112, Vendor { id : 5112u16 , name : "Ad Lib Multimedia Inc" , devices : & [] }),
        (7389, Vendor { id : 7389u16 , name : "secunet Security Networks AG" , devices : & [] }),
        (4734, Vendor { id : 4734u16 , name : "Winnov, L.P." , devices : & [Device { vendor_id : 4734u16 , id : 16u16 , name : "Videum 1000 Plus" , subsystems : & [] }] }),
        (4829, Vendor { id : 4829u16 , name : "Management Graphics" , devices : & [] }),
        (5204, Vendor { id : 5204u16 , name : "Altiga Networks" , devices : & [] }),
        (4701, Vendor { id : 4701u16 , name : "ESS Technology" , devices : & [Device { vendor_id : 4701u16 , id : 0u16 , name : "ES336H Fax Modem (Early Model)" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6472u16 , name : "ES1948 Maestro-1" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6504u16 , name : "ES1968 Maestro 2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 133u16 , name : "ES1968 Maestro-2 PCI" } , SubSystem { subvendor : 4147u16 , subdevice : 32849u16 , name : "ES1968 Maestro-2 Audiodrive" }] } , Device { vendor_id : 4701u16 , id : 6505u16 , name : "ES1938/ES1946/ES1969 Solo-1 Audiodrive" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 354u16 , name : "16 Bit PCI Audio Adapter (37L4457)" } , SubSystem { subvendor : 4116u16 , subdevice : 358u16 , name : "ES1969 SOLO-1 AudioDrive on IBM Aptiva Mainboard" } , SubSystem { subvendor : 4639u16 , subdevice : 34816u16 , name : "eDio Hi-Live SC1938" } , SubSystem { subvendor : 4701u16 , subdevice : 34952u16 , name : "Solo-1 Audio Adapter" } , SubSystem { subvendor : 4701u16 , subdevice : 34968u16 , name : "ES1938S TTSOLO1-SL [TerraTec 128i PCI]" } , SubSystem { subvendor : 5435u16 , subdevice : 4379u16 , name : "Terratec 128i PCI" }] } , Device { vendor_id : 4701u16 , id : 6520u16 , name : "ES1978 Maestro 2E" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45330u16 , name : "Armada M700/E500" } , SubSystem { subvendor : 4147u16 , subdevice : 32828u16 , name : "ES1978 Maestro-2E Audiodrive" } , SubSystem { subvendor : 4147u16 , subdevice : 32856u16 , name : "ES1978 Maestro-2E Audiodrive" } , SubSystem { subvendor : 4242u16 , subdevice : 16384u16 , name : "Monster Sound MX400" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "ES1978 Maestro-2E Audiodrive" }] } , Device { vendor_id : 4701u16 , id : 6536u16 , name : "ES1988 Allegro-1" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 152u16 , name : "Evo N600c" } , SubSystem { subvendor : 4242u16 , subdevice : 16640u16 , name : "Sonic Impact S100" } , SubSystem { subvendor : 4701u16 , subdevice : 1073u16 , name : "Allegro AudioDrive" } , SubSystem { subvendor : 4701u16 , subdevice : 6536u16 , name : "ESS Allegro-1 Audiodrive" } , SubSystem { subvendor : 4701u16 , subdevice : 6552u16 , name : "Allegro AudioDrive" } , SubSystem { subvendor : 4701u16 , subdevice : 6553u16 , name : "Allegro-1 AudioDrive" }] } , Device { vendor_id : 4701u16 , id : 6537u16 , name : "ESS Modem" , subsystems : & [SubSystem { subvendor : 4701u16 , subdevice : 6537u16 , name : "ESS Modem" }] } , Device { vendor_id : 4701u16 , id : 6544u16 , name : "ES1990S Canyon3D 2LE" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6546u16 , name : "ES1992S Canyon3D 2" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6552u16 , name : "ES1983S Maestro-3i PCI Audio Accelerator" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 177u16 , name : "Latitude C600" } , SubSystem { subvendor : 4136u16 , subdevice : 229u16 , name : "Latitude C810" } , SubSystem { subvendor : 4136u16 , subdevice : 230u16 , name : "ES1983S Maestro-3i (Dell Inspiron 8100)" }] } , Device { vendor_id : 4701u16 , id : 6553u16 , name : "ES1983S Maestro-3i PCI Modem Accelerator" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6554u16 , name : "ES1983S Maestro-3i PCI Audio Accelerator" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 6555u16 , name : "ES1983S Maestro-3i PCI Modem Accelerator" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 10248u16 , name : "ES336H Fax Modem (Later Model)" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 10296u16 , name : "ES2838/2839 SuperLink Modem" , subsystems : & [] } , Device { vendor_id : 4701u16 , id : 10392u16 , name : "ES2898 Modem" , subsystems : & [SubSystem { subvendor : 4701u16 , subdevice : 1060u16 , name : "ES56-PI Data Fax Modem" } , SubSystem { subvendor : 4701u16 , subdevice : 1061u16 , name : "ES56T-PI Data Fax Modem" } , SubSystem { subvendor : 4701u16 , subdevice : 1062u16 , name : "ES56V-PI Data Fax Modem" } , SubSystem { subvendor : 4701u16 , subdevice : 1063u16 , name : "VW-PI Data Fax Modem" } , SubSystem { subvendor : 4701u16 , subdevice : 1064u16 , name : "ES56ST-PI Data Fax Modem" } , SubSystem { subvendor : 4701u16 , subdevice : 1065u16 , name : "ES56SV-PI Data Fax Modem" } , SubSystem { subvendor : 5242u16 , subdevice : 49153u16 , name : "ES56-PI Data Fax Modem" } , SubSystem { subvendor : 5261u16 , subdevice : 4144u16 , name : "HCF WV-PI56 [ESS ES56-PI Data Fax Modem]" } , SubSystem { subvendor : 5374u16 , subdevice : 1064u16 , name : "ES56-PI Data Fax Modem" } , SubSystem { subvendor : 5374u16 , subdevice : 1065u16 , name : "ES56-PI Data Fax Modem" }] }] }),
        (4458, Vendor { id : 4458u16 , name : "Luminex Software, Inc." , devices : & [Device { vendor_id : 4458u16 , id : 24832u16 , name : "Bus/Tag Channel" , subsystems : & [] } , Device { vendor_id : 4458u16 , id : 26624u16 , name : "Escon Channel" , subsystems : & [] } , Device { vendor_id : 4458u16 , id : 28928u16 , name : "Bus/Tag Channel" , subsystems : & [] } , Device { vendor_id : 4458u16 , id : 30720u16 , name : "Escon Channel" , subsystems : & [] }] }),
        (5427, Vendor { id : 5427u16 , name : "BALTIMORE" , devices : & [] }),
        (4321, Vendor { id : 4321u16 , name : "Tekram Technology Co.,Ltd." , devices : & [Device { vendor_id : 4321u16 , id : 913u16 , name : "TRM-S1040" , subsystems : & [SubSystem { subvendor : 4321u16 , subdevice : 913u16 , name : "DC-315U SCSI-3 Host Adapter" }] } , Device { vendor_id : 4321u16 , id : 26892u16 , name : "DC-690c" , subsystems : & [] } , Device { vendor_id : 4321u16 , id : 56361u16 , name : "DC-290" , subsystems : & [] }] }),
        (4412, Vendor { id : 4412u16 , name : "Cyclone Microsystems, Inc." , devices : & [Device { vendor_id : 4412u16 , id : 0u16 , name : "PCI-9060 i960 Bridge" , subsystems : & [] } , Device { vendor_id : 4412u16 , id : 1u16 , name : "PCI-SDK [PCI i960 Evaluation Platform]" , subsystems : & [] } , Device { vendor_id : 4412u16 , id : 2321u16 , name : "PCI-911 [i960Jx-based Intelligent I/O Controller]" , subsystems : & [] } , Device { vendor_id : 4412u16 , id : 2322u16 , name : "PCI-912 [i960CF-based Intelligent I/O Controller]" , subsystems : & [] } , Device { vendor_id : 4412u16 , id : 2323u16 , name : "PCI-913" , subsystems : & [] } , Device { vendor_id : 4412u16 , id : 2324u16 , name : "PCI-914 [I/O Controller w/ secondary PCI bus]" , subsystems : & [] }] }),
        (5575, Vendor { id : 5575u16 , name : "Tateyama System Laboratory Co Ltd" , devices : & [Device { vendor_id : 5575u16 , id : 841u16 , name : "Tateyama C-PCI PLC/NC card Rev.01A" , subsystems : & [] }] }),
        (6880, Vendor { id : 6880u16 , name : "Google, Inc." , devices : & [Device { vendor_id : 6880u16 , id : 31u16 , name : "NVMe device" , subsystems : & [] } , Device { vendor_id : 6880u16 , id : 66u16 , name : "Compute Engine Virtual Ethernet [gVNIC]" , subsystems : & [] } , Device { vendor_id : 6880u16 , id : 43981u16 , name : "Airbrush Combined Paintbox IPU/Oscar Edge TPU [Pixel Neural Core]" , subsystems : & [] }] }),
        (4981, Vendor { id : 4981u16 , name : "Argosystems Inc" , devices : & [] }),
        (5468, Vendor { id : 5468u16 , name : "Cemax-Icon Inc" , devices : & [] }),
        (6796, Vendor { id : 6796u16 , name : "Verigy Pte. Ltd." , devices : & [Device { vendor_id : 6796u16 , id : 4352u16 , name : "E8001-66443 PCI Express CIC" , subsystems : & [] }] }),
        (5544, Vendor { id : 5544u16 , name : "Shanghai Communications Technologies Center" , devices : & [] }),
        (4887, Vendor { id : 4887u16 , name : "ADMtek" , devices : & [Device { vendor_id : 4887u16 , id : 2433u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 4887u16 , id : 2437u16 , name : "NC100 Network Everywhere Fast Ethernet 10/100" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4108u16 , name : "Scenic N300 ADMtek AN983 10/100 Mbps PCI Adapter" }] } , Device { vendor_id : 4887u16 , id : 6533u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [SubSystem { subvendor : 4997u16 , subdevice : 20762u16 , name : "FA511" } , SubSystem { subvendor : 5013u16 , subdevice : 8451u16 , name : "CB100-EZ (4-LED version)" }] } , Device { vendor_id : 4887u16 , id : 10320u16 , name : "HSP MicroModem 56" , subsystems : & [] } , Device { vendor_id : 4887u16 , id : 20768u16 , name : "ADM5120 OpenGate System-on-Chip" , subsystems : & [] } , Device { vendor_id : 4887u16 , id : 33281u16 , name : "ADM8211 802.11b Wireless Interface" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 9781u16 , name : "SMC2635W v1 802.11b Wireless Cardbus Adapter" } , SubSystem { subvendor : 4887u16 , subdevice : 33281u16 , name : "SMC2635W v2 802.11b Wireless Cardbus Adapter" }] } , Device { vendor_id : 4887u16 , id : 33297u16 , name : "ADM8211 802.11b Wireless Interface" , subsystems : & [] } , Device { vendor_id : 4887u16 , id : 38161u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] }] }),
        (6742, Vendor { id : 6742u16 , name : "Rivet Networks" , devices : & [] }),
        (4938, Vendor { id : 4938u16 , name : "DTC Technology Corp." , devices : & [Device { vendor_id : 4938u16 , id : 1u16 , name : "Domex 536" , subsystems : & [] } , Device { vendor_id : 4938u16 , id : 2u16 , name : "Domex DMX3194UP SCSI Adapter" , subsystems : & [] }] }),
        (684, Vendor { id : 684u16 , name : "SpeedStream" , devices : & [Device { vendor_id : 684u16 , id : 4114u16 , name : "1012 PCMCIA 10/100 Ethernet Card [RTL81xx]" , subsystems : & [] }] }),
        (4446, Vendor { id : 4446u16 , name : "Peer Protocols Inc" , devices : & [] }),
        (7840, Vendor { id : 7840u16 , name : "Tencent Technology (Shenzhen) Company Limited" , devices : & [Device { vendor_id : 7840u16 , id : 10774u16 , name : "Cloud Intelligent Inference Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10776u16 , name : "Video Transcode Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10784u16 , name : "Cloud Intelligent Inference and Training Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10786u16 , name : "Cloud Intelligent Inference Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10788u16 , name : "Cloud Intelligent Inference Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10790u16 , name : "Cloud Intelligent Inference Controller" , subsystems : & [] } , Device { vendor_id : 7840u16 , id : 10800u16 , name : "Cloud Video Transcode Controller" , subsystems : & [] }] }),
        (5500, Vendor { id : 5500u16 , name : "Eurosoft (UK)" , devices : & [Device { vendor_id : 5500u16 , id : 32769u16 , name : "Fix2000 PCI Y2K Compliance Card" , subsystems : & [] }] }),
        (4581, Vendor { id : 4581u16 , name : "IIX Consulting" , devices : & [] }),
        (5239, Vendor { id : 5239u16 , name : "Net Insight" , devices : & [] }),
        (10001, Vendor { id : 10001u16 , name : "AVID Technology Inc." , devices : & [] }),
        (8164, Vendor { id : 8164u16 , name : "HippStor Technology" , devices : & [Device { vendor_id : 8164u16 , id : 5632u16 , name : "HP600 Series NVMe SSD" , subsystems : & [SubSystem { subvendor : 8164u16 , subdevice : 1u16 , name : "Enterprise NVMe SSD U.2 15.36TB (HP600)" } , SubSystem { subvendor : 8164u16 , subdevice : 3u16 , name : "Enterprise NVMe SSD U.2 12.80TB (HP630)" } , SubSystem { subvendor : 8164u16 , subdevice : 17u16 , name : "Enterprise NVMe SSD U.2 30.72TB (HP620)" } , SubSystem { subvendor : 8164u16 , subdevice : 19u16 , name : "Enterprise NVMe SSD U.2 61.44TB (HP620)" } , SubSystem { subvendor : 8164u16 , subdevice : 117u16 , name : "Enterprise NVMe SSD U.2 3.84TB(HP610)" } , SubSystem { subvendor : 8164u16 , subdevice : 118u16 , name : "Enterprise NVMe SSD U.2 7.68TB(HP610)" } , SubSystem { subvendor : 8164u16 , subdevice : 120u16 , name : "Enterprise NVMe SSD U.2 3.20TB(HP630)" } , SubSystem { subvendor : 8164u16 , subdevice : 121u16 , name : "Enterprise NVMe SSD U.2 6.40TB (HP630)" }] }] }),
        (5076, Vendor { id : 5076u16 , name : "Graphics Microsystems Inc" , devices : & [] }),
        (4669, Vendor { id : 4669u16 , name : "Engineering Design Team, Inc." , devices : & [Device { vendor_id : 4669u16 , id : 0u16 , name : "PCI 11W" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 71u16 , name : "PCIe4 CDa" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 75u16 , name : "PCIe4 CDa 16" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 157u16 , name : "VisionLink F1" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 160u16 , name : "PCIe8g3 S5 (10G/40G)" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 163u16 , name : "VisionLink F4" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 169u16 , name : "VisionLink CLS" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 171u16 , name : "PCIe8g3 A5 10G" , subsystems : & [] } , Device { vendor_id : 4669u16 , id : 181u16 , name : "PCIe8 RFx SDR" , subsystems : & [] }] }),
        (5043, Vendor { id : 5043u16 , name : "Lanart Corporation" , devices : & [] }),
        (5150, Vendor { id : 5150u16 , name : "Fanuc Ltd" , devices : & [] }),
        (4644, Vendor { id : 4644u16 , name : "Interactive Images" , devices : & [] }),
        (6163, Vendor { id : 6163u16 , name : "Ambient Technologies Inc" , devices : & [Device { vendor_id : 6163u16 , id : 16384u16 , name : "HaM controllerless modem" , subsystems : & [SubSystem { subvendor : 5822u16 , subdevice : 1u16 , name : "V9x HAM Data Fax Modem" }] } , Device { vendor_id : 6163u16 , id : 16640u16 , name : "HaM plus Data Fax Modem" , subsystems : & [SubSystem { subvendor : 5822u16 , subdevice : 2u16 , name : "V9x HAM 1394" }] }] }),
        (4972, Vendor { id : 4972u16 , name : "Adtek System Science Co Ltd" , devices : & [] }),
        (43530, Vendor { id : 43530u16 , name : "iTuner" , devices : & [] }),
        (5758, Vendor { id : 5758u16 , name : "ONNTO Corp." , devices : & [] }),
        (5516, Vendor { id : 5516u16 , name : "Hitachi Semiconductor & Devices Sales Co Ltd" , devices : & [] }),
        (6609, Vendor { id : 6609u16 , name : "Motorola Expedience" , devices : & [] }),
        (6101, Vendor { id : 6101u16 , name : "Exar Corp." , devices : & [Device { vendor_id : 6101u16 , id : 22321u16 , name : "Xframe 10-Gigabit Ethernet PCI-X" , subsystems : & [] } , Device { vendor_id : 6101u16 , id : 22322u16 , name : "Xframe II 10-Gigabit Ethernet PCI-X 2.0" , subsystems : & [] } , Device { vendor_id : 6101u16 , id : 22577u16 , name : "Xframe 10-Gigabit Ethernet PCI-X" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4821u16 , name : "PCI-X 133MHz 10GbE SR Fiber" } , SubSystem { subvendor : 4265u16 , subdevice : 32800u16 , name : "Single Port 10-Gigabit Ethernet (PCI-X, Fiber)" } , SubSystem { subvendor : 4265u16 , subdevice : 32804u16 , name : "Single Port 10-Gigabit Ethernet (PCI-X, Fiber)" }] } , Device { vendor_id : 6101u16 , id : 22578u16 , name : "Xframe II 10-Gigabit Ethernet PCI-X 2.0" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4919u16 , name : "PCI-X 266MHz 10GigE SR [AD385A]" } , SubSystem { subvendor : 4265u16 , subdevice : 32801u16 , name : "Single Port 10-Gigabit Ethernet II (PCI-X, Fiber)" } , SubSystem { subvendor : 6101u16 , subdevice : 24608u16 , name : "Xframe II SR" } , SubSystem { subvendor : 6101u16 , subdevice : 24609u16 , name : "Xframe II SR, Low Profile" } , SubSystem { subvendor : 6101u16 , subdevice : 24610u16 , name : "Xframe E SR" } , SubSystem { subvendor : 6101u16 , subdevice : 25632u16 , name : "Xframe II LR" } , SubSystem { subvendor : 6101u16 , subdevice : 25633u16 , name : "Xframe II LR, Low Profile" } , SubSystem { subvendor : 6101u16 , subdevice : 25634u16 , name : "Xframe E LR" } , SubSystem { subvendor : 6101u16 , subdevice : 27680u16 , name : "Xframe II CX4" } , SubSystem { subvendor : 6101u16 , subdevice : 27681u16 , name : "Xframe II CX4, Low Profile" } , SubSystem { subvendor : 6101u16 , subdevice : 27682u16 , name : "Xframe E CX4" }] } , Device { vendor_id : 6101u16 , id : 22579u16 , name : "X3100 Series 10 Gigabit Ethernet PCIe" , subsystems : & [SubSystem { subvendor : 6101u16 , subdevice : 24624u16 , name : "X3110 Single Port SR" } , SubSystem { subvendor : 6101u16 , subdevice : 24625u16 , name : "X3120 Dual Port SR" } , SubSystem { subvendor : 6101u16 , subdevice : 25648u16 , name : "X3110 Single Port LR" } , SubSystem { subvendor : 6101u16 , subdevice : 25649u16 , name : "X3120 Dual Port LR" } , SubSystem { subvendor : 6101u16 , subdevice : 28720u16 , name : "X3110 Single Port LRM" } , SubSystem { subvendor : 6101u16 , subdevice : 28721u16 , name : "X3120 Dual Port LRM" } , SubSystem { subvendor : 6101u16 , subdevice : 29744u16 , name : "X3110 Single Port 10GBase-T" } , SubSystem { subvendor : 6101u16 , subdevice : 29745u16 , name : "X3120 Dual Port 10GBase-T" } , SubSystem { subvendor : 6101u16 , subdevice : 30768u16 , name : "X3110 Single Port 10GBase-CR" } , SubSystem { subvendor : 6101u16 , subdevice : 30769u16 , name : "X3120 Dual Port 10GBase-CR" }] }] }),
        (5984, Vendor { id : 5984u16 , name : "TEDIA spol. s r. o." , devices : & [Device { vendor_id : 5984u16 , id : 257u16 , name : "PCD-7004 Digital Bi-Directional Ports PCI Card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 258u16 , name : "PCD-7104 Digital Input & Output PCI Card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 289u16 , name : "PCT-7303A PC card with IRC counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 290u16 , name : "PCT-7408A PC card with counters and timers" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 291u16 , name : "PCT-7424 PCI card with standard counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 321u16 , name : "PCA7208AL - Analog Inputs/Outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 322u16 , name : "PCA7208AS - Analog inputs/Outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 323u16 , name : "PCA7408AL - Analog Inputs/Outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 324u16 , name : "PCA7408AS - Analog Inputs/Outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 325u16 , name : "PCA-7228AL Multifunction PCI IO card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 326u16 , name : "PCA-7228AS Multifunction PCI IO card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 327u16 , name : "PCA7428AL Multifunction PCI IO card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 328u16 , name : "PCA7428AS Multifunction PCI IO card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 329u16 , name : "PCA7228EL Multifunction PCI IO card with isolated analog inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 336u16 , name : "PCA7428EL Multifunction PCI IO card with isolated analog inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 337u16 , name : "PCA7628AL - PCI card with analog inputs, counters and DIO" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 338u16 , name : "PCA7628AS PCI card with analog inputs, outputs, counters and DIO" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 353u16 , name : "PCA7288A PCI card with analog outputs, counters and DIO" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 384u16 , name : "PCI1052 Communication card for MicroUnit network" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 532u16 , name : "PCT-7424C (F0) PC card with standard counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 533u16 , name : "PCT-7424C (F1) PC card with standard counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 534u16 , name : "PCT-7424E (F0) PC card with standard counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 535u16 , name : "PCT-7424E (F1) PC card with standard counters" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 576u16 , name : "PCA7428CL_F0 - analog Inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 577u16 , name : "PCA7428CL_F1 - analog Inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 578u16 , name : "PCA7428CS_F0 - Analog Inputs/Outputs non isolated" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 579u16 , name : "PCA7428CS_F1 - Analog Inputs/Outputs non isolated" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 580u16 , name : "PCA7428CE_F0 - Analog Inputs isolated" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 581u16 , name : "PCA7428CE_F1 - Analog Inputs isolated" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 771u16 , name : "PCD-7006C Digital Input & Output PCI Card" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 2048u16 , name : "PCD8006 - PCIe digital Inputs/Outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 2112u16 , name : "PCA-8428 General-purpose multifunctional PCIe card with 8 analog inputs and 2 analog outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 2113u16 , name : "PCA-8429 General-purpose multifunctional PCIe card with 8 analog inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 2114u16 , name : "PCA-8438 General-purpose multifunctional PCIe card with 16 analog inputs and 2 analog outputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 2115u16 , name : "PCA-8439 General-purpose multifunctional PCIe card with 16 analog inputs" , subsystems : & [] } , Device { vendor_id : 5984u16 , id : 65280u16 , name : "CTU CAN FD PCIe Card" , subsystems : & [] }] }),
        (6878, Vendor { id : 6878u16 , name : "Spin Master Ltd." , devices : & [Device { vendor_id : 6878u16 , id : 5377u16 , name : "Swipetech barcode scanner" , subsystems : & [] } , Device { vendor_id : 6878u16 , id : 12344u16 , name : "PCIe Video Bridge" , subsystems : & [SubSystem { subvendor : 5058u16 , subdevice : 12310u16 , name : "TT-budget S2-4200 Twin" } , SubSystem { subvendor : 16980u16 , subdevice : 1362u16 , name : "S952 v3" }] }] }),
        (8400, Vendor { id : 8400u16 , name : "Telin Semiconductor (Wuhan) Co., Ltd." , devices : & [Device { vendor_id : 8400u16 , id : 8193u16 , name : "PACIFIC-S1 NVMe SSD" , subsystems : & [] }] }),
        (4190, Vendor { id : 4190u16 , name : "Vtech Computers Ltd" , devices : & [] }),
        (6631, Vendor { id : 6631u16 , name : "NET (Network Equipment Technologies)" , devices : & [Device { vendor_id : 6631u16 , id : 4097u16 , name : "STIX DSP Card" , subsystems : & [] } , Device { vendor_id : 6631u16 , id : 4098u16 , name : "STIX - 1 Port T1/E1 Card" , subsystems : & [] } , Device { vendor_id : 6631u16 , id : 4099u16 , name : "STIX - 2 Port T1/E1 Card" , subsystems : & [] } , Device { vendor_id : 6631u16 , id : 4100u16 , name : "STIX - 4 Port T1/E1 Card" , subsystems : & [] } , Device { vendor_id : 6631u16 , id : 4101u16 , name : "STIX - 4 Port FXS Card" , subsystems : & [] }] }),
        (7439, Vendor { id : 7439u16 , name : "Amazon.com, Inc." , devices : & [Device { vendor_id : 7439u16 , id : 28772u16 , name : "NeuronDevice (Inferentia)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 29028u16 , name : "NeuronDevice (Trainium)" , subsystems : & [SubSystem { subvendor : 7439u16 , subdevice : 0u16 , name : "Trainium" }] } , Device { vendor_id : 7439u16 , id : 29284u16 , name : "NeuronDevice (Inferentia2)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 29540u16 , name : "NeuronDevice (Trainium2)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 30052u16 , name : "NeuronDevice (Trainium3)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 30053u16 , name : "NeuronDevice (Trainium3)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 32865u16 , name : "NVMe EBS Controller" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 52481u16 , name : "NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 60448u16 , name : "Elastic Network Adapter (ENA)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 61344u16 , name : "Elastic Fabric Adapter (EFA)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 61345u16 , name : "Elastic Fabric Adapter (EFA)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 61346u16 , name : "Elastic Fabric Adapter (EFA)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 61347u16 , name : "Elastic Fabric Adapter (EFA)" , subsystems : & [] } , Device { vendor_id : 7439u16 , id : 61348u16 , name : "Elastic Fabric Adapter (EFA)" , subsystems : & [] }] }),
        (6721, Vendor { id : 6721u16 , name : "Tilera Corp." , devices : & [Device { vendor_id : 6721u16 , id : 1u16 , name : "TILE64 processor" , subsystems : & [] } , Device { vendor_id : 6721u16 , id : 2u16 , name : "TILEPro processor" , subsystems : & [] } , Device { vendor_id : 6721u16 , id : 512u16 , name : "TILE-Gx processor" , subsystems : & [] } , Device { vendor_id : 6721u16 , id : 513u16 , name : "TILE-Gx Processor Virtual Function" , subsystems : & [] } , Device { vendor_id : 6721u16 , id : 8192u16 , name : "TILE-Gx PCI Express Root Port" , subsystems : & [] }] }),
        (478, Vendor { id : 478u16 , name : "Oxide Computer Company" , devices : & [Device { vendor_id : 478u16 , id : 0u16 , name : "Propolis NVMe Controller" , subsystems : & [] } , Device { vendor_id : 478u16 , id : 1u16 , name : "Propolis xHCI Controller" , subsystems : & [] } , Device { vendor_id : 478u16 , id : 2u16 , name : "Propolis PCI-PCI Bridge" , subsystems : & [] }] }),
        (5342, Vendor { id : 5342u16 , name : "Applied Integration Corporation" , devices : & [] }),
        (41120, Vendor { id : 41120u16 , name : "AOPEN Inc." , devices : & [] }),
        (5525, Vendor { id : 5525u16 , name : "Diva Systems Corp" , devices : & [] }),
        (13978, Vendor { id : 13978u16 , name : "HighSecLabs, Ltd." , devices : & [] }),
        (4926, Vendor { id : 4926u16 , name : "Virtual Computer Corp" , devices : & [] }),
        (6774, Vendor { id : 6774u16 , name : "Wavesat" , devices : & [] }),
        (6883, Vendor { id : 6883u16 , name : "SANBlaze Technology, Inc." , devices : & [] }),
        (5483, Vendor { id : 5483u16 , name : "2wire Inc" , devices : & [] }),
        (6450, Vendor { id : 6450u16 , name : "DiBcom" , devices : & [] }),
        (4770, Vendor { id : 4770u16 , name : "Newgen Systems Corporation" , devices : & [] }),
        (4311, Vendor { id : 4311u16 , name : "BCM Advanced Research" , devices : & [] }),
        (7150, Vendor { id : 7150u16 , name : "IXXAT Automation GmbH" , devices : & [Device { vendor_id : 7150u16 , id : 2u16 , name : "CAN-IB100/PCIe" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 3u16 , name : "CAN-IB200/PCIe" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 4u16 , name : "CAN-IB120/PCIe Mini" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 5u16 , name : "CAN-IB130/PCIe 104" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 6u16 , name : "CAN-IB230/PCIe 104" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 14u16 , name : "CAN-IB500/PCIe" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 15u16 , name : "CAN-IB600/PCIe" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 16u16 , name : "CAN-IB300/PCI" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 17u16 , name : "CAN-IB400/PCI" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 18u16 , name : "CAN-IB520/PCIe Mini" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 22u16 , name : "CAN-IB410/PMC" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 28u16 , name : "CAN-IB810/PMC" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 30u16 , name : "INpact PCIe" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 31u16 , name : "INpact PCIe Mini" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 41u16 , name : "INpact M.2" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 45u16 , name : "CAN-IB630/PCIe 104" , subsystems : & [] } , Device { vendor_id : 7150u16 , id : 46u16 , name : "CAN-IB640/PCIe" , subsystems : & [] }] }),
        (4762, Vendor { id : 4762u16 , name : "VMetro, inc." , devices : & [Device { vendor_id : 4762u16 , id : 1557u16 , name : "PBT-615 PCI-X Bus Analyzer" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 1813u16 , name : "cPCI Bus Analyzer" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 4352u16 , name : "PMC-FPGA05" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 4358u16 , name : "XMC-FPGA05F, PCI interface" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 4359u16 , name : "XMC-FPGA05F, PCIe interface" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 4360u16 , name : "XMC-FPGA05D, PCI interface" , subsystems : & [] } , Device { vendor_id : 4762u16 , id : 4361u16 , name : "XMC-FPGA05D, PCIe interface" , subsystems : & [] }] }),
        (5406, Vendor { id : 5406u16 , name : "MATRIX Corp" , devices : & [] }),
        (5387, Vendor { id : 5387u16 , name : "YAMASHITA Systems Corp" , devices : & [] }),
        (5361, Vendor { id : 5361u16 , name : "Conexant Systems, Inc." , devices : & [Device { vendor_id : 5361u16 , id : 4098u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4099u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4100u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4101u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4102u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4130u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4131u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4132u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4133u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4134u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4146u16 , name : "HCF 56k Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4147u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 32887u16 , name : "NEC" } , SubSystem { subvendor : 4653u16 , subdevice : 16423u16 , name : "Dell Zeus - MDP3880-W(B) Data Fax Modem" } , SubSystem { subvendor : 4653u16 , subdevice : 16432u16 , name : "Dell Mercury - MDP3880-U(B) Data Fax Modem" } , SubSystem { subvendor : 4653u16 , subdevice : 16436u16 , name : "Dell Thor - MDP3880-W(U) Data Fax Modem" } , SubSystem { subvendor : 5088u16 , subdevice : 525u16 , name : "Dell Copper" } , SubSystem { subvendor : 5088u16 , subdevice : 526u16 , name : "Dell Silver" } , SubSystem { subvendor : 5088u16 , subdevice : 609u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 656u16 , name : "Compaq Goldwing" } , SubSystem { subvendor : 5088u16 , subdevice : 672u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 688u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 704u16 , name : "Compaq Scooter" } , SubSystem { subvendor : 5088u16 , subdevice : 720u16 , name : "IBM" } , SubSystem { subvendor : 5199u16 , subdevice : 5376u16 , name : "IBM P85-DF (1)" } , SubSystem { subvendor : 5199u16 , subdevice : 5377u16 , name : "IBM P85-DF (2)" } , SubSystem { subvendor : 5199u16 , subdevice : 5386u16 , name : "IBM P85-DF (3)" } , SubSystem { subvendor : 5199u16 , subdevice : 5387u16 , name : "IBM P85-DF Low Profile (1)" } , SubSystem { subvendor : 5199u16 , subdevice : 5392u16 , name : "IBM P85-DF Low Profile (2)" }] } , Device { vendor_id : 5361u16 , id : 4148u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4149u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [SubSystem { subvendor : 4303u16 , subdevice : 4248u16 , name : "Fujitsu P85-DFSV" }] } , Device { vendor_id : 5361u16 , id : 4150u16 , name : "HCF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32871u16 , name : "HCF 56k Modem" } , SubSystem { subvendor : 4653u16 , subdevice : 16425u16 , name : "MDP3880SP-W" } , SubSystem { subvendor : 4653u16 , subdevice : 16433u16 , name : "MDP3880SP-U" } , SubSystem { subvendor : 5088u16 , subdevice : 521u16 , name : "Dell Titanium" } , SubSystem { subvendor : 5088u16 , subdevice : 522u16 , name : "Dell Graphite" } , SubSystem { subvendor : 5088u16 , subdevice : 608u16 , name : "Gateway Red Owl" } , SubSystem { subvendor : 5088u16 , subdevice : 624u16 , name : "Gateway White Horse" }] } , Device { vendor_id : 5361u16 , id : 4178u16 , name : "HCF 56k Data/Fax Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4179u16 , name : "HCF 56k Data/Fax Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4180u16 , name : "HCF 56k Data/Fax/Voice Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4181u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4182u16 , name : "HCF 56k Data/Fax/Voice/Spkp Modem (Worldwide)" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 16437u16 , name : "MDP3900V-W" }] } , Device { vendor_id : 5361u16 , id : 4183u16 , name : "HCF 56k Data/Fax/Voice/Spkp Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4185u16 , name : "HCF 56k Data/Fax/Voice Modem (Worldwide)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4195u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4196u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4197u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4198u16 , name : "HCF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 16435u16 , name : "Dell Athena - MDP3900V-U" }] } , Device { vendor_id : 5361u16 , id : 4229u16 , name : "HCF V90 56k Data/Fax/Voice/Spkp PCI Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 4278u16 , name : "CX06834-11 HCF V.92 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5171u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5172u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5173u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5174u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5203u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 5088u16 , subdevice : 576u16 , name : "IBM" } , SubSystem { subvendor : 5088u16 , subdevice : 592u16 , name : "IBM" } , SubSystem { subvendor : 5199u16 , subdevice : 5378u16 , name : "IBM P95-DF (1)" } , SubSystem { subvendor : 5199u16 , subdevice : 5379u16 , name : "IBM P95-DF (2)" }] } , Device { vendor_id : 5361u16 , id : 5204u16 , name : "HCF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5205u16 , name : "HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5206u16 , name : "HCF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 16437u16 , name : "Dell Europa - MDP3900V-W" } , SubSystem { subvendor : 4653u16 , subdevice : 17154u16 , name : "Dell MP3930V-W(C) MiniPCI" }] } , Device { vendor_id : 5361u16 , id : 5648u16 , name : "ADSL AccessRunner PCI Arbitration Device" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5649u16 , name : "AccessRunner PCI ADSL Interface Device" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5664u16 , name : "AccessRunner V2 PCI ADSL Arbitration Device" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5665u16 , name : "AccessRunner V2 PCI ADSL Interface Device" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 5666u16 , name : "AccessRunner V2 PCI ADSL Yukon WAN Adapter" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 6147u16 , name : "HCF 56k Modem" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 35u16 , name : "623-LAN Grizzly" } , SubSystem { subvendor : 3601u16 , subdevice : 67u16 , name : "623-LAN Yogi" }] } , Device { vendor_id : 5361u16 , id : 6161u16 , name : "MiniPCI Network Adapter" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 6165u16 , name : "HCF 56k Modem" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 34u16 , name : "Grizzly" } , SubSystem { subvendor : 3601u16 , subdevice : 66u16 , name : "Yogi" }] } , Device { vendor_id : 5361u16 , id : 6192u16 , name : "CX861xx Integrated Host Bridge" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8070u16 , name : "DBH CX11880 Codec" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8071u16 , name : "SMIC CX11880 Codec" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8195u16 , name : "HSF 56k Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8196u16 , name : "HSF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8197u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8198u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8211u16 , name : "HSF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45461u16 , name : "Bear" } , SubSystem { subvendor : 3601u16 , subdevice : 45462u16 , name : "Seminole 1" } , SubSystem { subvendor : 3601u16 , subdevice : 45502u16 , name : "Seminole 2" } , SubSystem { subvendor : 4133u16 , subdevice : 32787u16 , name : "Acer" } , SubSystem { subvendor : 4147u16 , subdevice : 32925u16 , name : "NEC" } , SubSystem { subvendor : 4147u16 , subdevice : 32956u16 , name : "NEC" } , SubSystem { subvendor : 5469u16 , subdevice : 26515u16 , name : "HP" } , SubSystem { subvendor : 5469u16 , subdevice : 34896u16 , name : "E Machines" }] } , Device { vendor_id : 5361u16 , id : 8212u16 , name : "HSF 56k Data/Fax/Voice Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8213u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8214u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8259u16 , name : "HSF 56k Data/Fax Modem (WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8260u16 , name : "HSF 56k Data/Fax/Voice Modem (WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8261u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (WorldW SmartDAA)" , subsystems : & [SubSystem { subvendor : 5361u16 , subdevice : 8261u16 , name : "Generic SoftK56" }] } , Device { vendor_id : 5361u16 , id : 8262u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem (WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8291u16 , name : "HSF 56k Data/Fax Modem (SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8292u16 , name : "HSF 56k Data/Fax/Voice Modem (SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8293u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8294u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem (SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8339u16 , name : "HSF 56k Modem" , subsystems : & [SubSystem { subvendor : 5469u16 , subdevice : 12039u16 , name : "Legend" }] } , Device { vendor_id : 5361u16 , id : 8515u16 , name : "HSF 56k Data/Fax/Cell Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8516u16 , name : "HSF 56k Data/Fax/Voice/Cell Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8517u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS)/Cell Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8518u16 , name : "HSF 56k Data/Fax/Voice/Spkp/Cell Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8547u16 , name : "HSF 56k Data/Fax/Cell Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8548u16 , name : "HSF 56k Data/Fax/Voice/Cell Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8549u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS)/Cell Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 8550u16 , name : "HSF 56k Data/Fax/Voice/Spkp/Cell Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9027u16 , name : "HSF 56k Data/Fax CardBus Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9028u16 , name : "HSF 56k Data/Fax/Voice CardBus Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9029u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS) CardBus Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9030u16 , name : "HSF 56k Data/Fax/Voice/Spkp CardBus Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9059u16 , name : "HSF 56k Data/Fax CardBus Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9060u16 , name : "HSF 56k Data/Fax/Voice CardBus Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9061u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS) CardBus Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9062u16 , name : "HSF 56k Data/Fax/Voice/Spkp CardBus Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9283u16 , name : "HSF 56k Data/Fax Modem (Mob WorldW SmartDAA)" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32885u16 , name : "Modem" } , SubSystem { subvendor : 4173u16 , subdevice : 32899u16 , name : "Modem" } , SubSystem { subvendor : 4173u16 , subdevice : 32919u16 , name : "Modem" }] } , Device { vendor_id : 5361u16 , id : 9284u16 , name : "HSF 56k Data/Fax/Voice Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9285u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS) Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9286u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem (Mob WorldW SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9315u16 , name : "HSF 56k Data/Fax Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9316u16 , name : "HSF 56k Data/Fax/Voice Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9317u16 , name : "HSF 56k Data/Fax/Voice/Spkp (w/HS) Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9318u16 , name : "HSF 56k Data/Fax/Voice/Spkp Modem (Mob SmartDAA)" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 9986u16 , name : "HSFi modem RD01-D270" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 36232u16 , name : "SmartHSFi V92 56K PCI Modem" }] } , Device { vendor_id : 5361u16 , id : 12032u16 , name : "HSF 56k HSFi Modem" , subsystems : & [SubSystem { subvendor : 5088u16 , subdevice : 36228u16 , name : "IBM HSFi V.90" } , SubSystem { subvendor : 5088u16 , subdevice : 36229u16 , name : "Compaq Stinger" } , SubSystem { subvendor : 5361u16 , subdevice : 8196u16 , name : "Dynalink 56PMi" }] } , Device { vendor_id : 5361u16 , id : 12034u16 , name : "HSF 56k HSFi Data/Fax" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 12049u16 , name : "HSF 56k HSFi Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 12064u16 , name : "HSF 56k Data/Fax Modem" , subsystems : & [SubSystem { subvendor : 5361u16 , subdevice : 8204u16 , name : "Soft Data Fax Modem with SmartCP" } , SubSystem { subvendor : 5361u16 , subdevice : 8207u16 , name : "Dimension 3000" }] } , Device { vendor_id : 5361u16 , id : 12080u16 , name : "SoftV92 SpeakerPhone SoftRing Modem with SmartSP" , subsystems : & [SubSystem { subvendor : 5361u16 , subdevice : 8212u16 , name : "Devolo MikroLink 56K Modem PCI" }] } , Device { vendor_id : 5361u16 , id : 12112u16 , name : "Conexant SoftK56 Data/Fax Modem" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 20751u16 , name : "Conexant CX 20751/20752" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 23418u16 , name : "CX23418 Single-Chip MPEG-2 Encoder with Integrated Analog Video/Broadcast Audio Decoder" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 29764u16 , name : "WinTV HVR-1600" } , SubSystem { subvendor : 4221u16 , subdevice : 28468u16 , name : "WinFast DVR3100 H" } , SubSystem { subvendor : 22612u16 , subdevice : 13123u16 , name : "GoTView PCI DVD3 Hybrid" }] } , Device { vendor_id : 5361u16 , id : 33280u16 , name : "CX25850" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 33332u16 , name : "RS8234 ATM SAR Controller [ServiceSAR Plus]" , subsystems : & [] } , Device { vendor_id : 5361u16 , id : 34816u16 , name : "CX23880/1/2/3 PCI Video and Audio Decoder" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 10241u16 , name : "Hauppauge WinTV 28xxx (Roslyn) models" } , SubSystem { subvendor : 112u16 , subdevice : 13312u16 , name : "WinTV 34604" } , SubSystem { subvendor : 112u16 , subdevice : 13313u16 , name : "Hauppauge WinTV 34xxx models" } , SubSystem { subvendor : 112u16 , subdevice : 26882u16 , name : "WinTV HVR-4000-HD" } , SubSystem { subvendor : 112u16 , subdevice : 30721u16 , name : "WinTV HVR-1800 MCE" } , SubSystem { subvendor : 112u16 , subdevice : 36865u16 , name : "Nova-T DVB-T" } , SubSystem { subvendor : 112u16 , subdevice : 36866u16 , name : "Nova-T DVB-T Model 909" } , SubSystem { subvendor : 112u16 , subdevice : 37376u16 , name : "Nova-SE2 DVB-S" } , SubSystem { subvendor : 112u16 , subdevice : 37378u16 , name : "Nova-S-Plus DVB-S" } , SubSystem { subvendor : 112u16 , subdevice : 37890u16 , name : "WinTV-HVR1100 DVB-T/Hybrid" } , SubSystem { subvendor : 112u16 , subdevice : 38400u16 , name : "WinTV 88x Video" } , SubSystem { subvendor : 112u16 , subdevice : 38914u16 , name : "WinTV-HVR1100 DVB-T/Hybrid (Low Profile)" } , SubSystem { subvendor : 4098u16 , subdevice : 248u16 , name : "ATI TV Wonder Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 249u16 , name : "ATI TV Wonder" } , SubSystem { subvendor : 4098u16 , subdevice : 41217u16 , name : "HDTV Wonder" } , SubSystem { subvendor : 4163u16 , subdevice : 18467u16 , name : "PVR-416" } , SubSystem { subvendor : 4221u16 , subdevice : 26129u16 , name : "Winfast TV 2000XP Expert" } , SubSystem { subvendor : 4221u16 , subdevice : 26131u16 , name : "Leadtek Winfast 2000XP Expert" } , SubSystem { subvendor : 4221u16 , subdevice : 26144u16 , name : "Leadtek Winfast DV2000" } , SubSystem { subvendor : 4221u16 , subdevice : 26172u16 , name : "Leadtek PVR 2000" } , SubSystem { subvendor : 4221u16 , subdevice : 26207u16 , name : "WinFast DTV1000-T" } , SubSystem { subvendor : 4348u16 , subdevice : 53251u16 , name : "IODATA GV-VCP3/PCI" } , SubSystem { subvendor : 4348u16 , subdevice : 53301u16 , name : "IODATA GV/BCTV7E" } , SubSystem { subvendor : 5153u16 , subdevice : 820u16 , name : "Instant TV DVB-T PCI" } , SubSystem { subvendor : 5217u16 , subdevice : 10u16 , name : "AVerTV 303 (M126)" } , SubSystem { subvendor : 5217u16 , subdevice : 11u16 , name : "AverTV Studio 303 (M126)" } , SubSystem { subvendor : 5217u16 , subdevice : 32785u16 , name : "UltraTV Media Center PCI 550" } , SubSystem { subvendor : 5218u16 , subdevice : 34310u16 , name : "MSI TV-@nywhere Master" } , SubSystem { subvendor : 5319u16 , subdevice : 263u16 , name : "GDI Black Gold" } , SubSystem { subvendor : 5361u16 , subdevice : 391u16 , name : "Conexant DVB-T reference design" } , SubSystem { subvendor : 5361u16 , subdevice : 834u16 , name : "Digital-Logic MICROSPACE Entertainment Center (MEC)" } , SubSystem { subvendor : 5435u16 , subdevice : 4454u16 , name : "Cinergy 1400 DVB-T" } , SubSystem { subvendor : 5440u16 , subdevice : 9600u16 , name : "Provideo PV259" } , SubSystem { subvendor : 5460u16 , subdevice : 18449u16 , name : "PixelView" } , SubSystem { subvendor : 5460u16 , subdevice : 18451u16 , name : "Club 3D  ZAP1000 MCE Edition" } , SubSystem { subvendor : 6110u16 , subdevice : 2209u16 , name : "KWorld/VStream XPert DVB-T with cx22702" } , SubSystem { subvendor : 6110u16 , subdevice : 2214u16 , name : "KWorld/VStream XPert DVB-T" } , SubSystem { subvendor : 6110u16 , subdevice : 2226u16 , name : "KWorld DVB-S 100" } , SubSystem { subvendor : 6110u16 , subdevice : 43174u16 , name : "digitalnow DNTV Live! DVB-T" } , SubSystem { subvendor : 6149u16 , subdevice : 273u16 , name : "PICOLO Jet-X Video" } , SubSystem { subvendor : 6149u16 , subdevice : 274u16 , name : "PICOLO Jet-X Video" } , SubSystem { subvendor : 6149u16 , subdevice : 275u16 , name : "PICOLO Jet-X Video" } , SubSystem { subvendor : 6149u16 , subdevice : 276u16 , name : "PICOLO Jet-X Video" } , SubSystem { subvendor : 6178u16 , subdevice : 37u16 , name : "digitalnow DNTV Live! DVB-T Pro" } , SubSystem { subvendor : 6235u16 , subdevice : 57344u16 , name : "VideoMate X500" } , SubSystem { subvendor : 6316u16 , subdevice : 54528u16 , name : "FusionHDTV 5 Gold" } , SubSystem { subvendor : 6316u16 , subdevice : 55312u16 , name : "FusionHDTV 3 Gold-Q" } , SubSystem { subvendor : 6316u16 , subdevice : 55328u16 , name : "FusionHDTV 3 Gold-T" } , SubSystem { subvendor : 6316u16 , subdevice : 56064u16 , name : "FusionHDTV DVB-T1" } , SubSystem { subvendor : 6316u16 , subdevice : 56081u16 , name : "FusionHDTV DVB-T Plus" } , SubSystem { subvendor : 6316u16 , subdevice : 56144u16 , name : "FusionHDTV DVB-T Dual Digital" } , SubSystem { subvendor : 22100u16 , subdevice : 9096u16 , name : "GoTView PCI Hybrid TV Tuner Card" } , SubSystem { subvendor : 28771u16 , subdevice : 12288u16 , name : "pcHDTV HD3000 HDTV" } , SubSystem { subvendor : 28771u16 , subdevice : 21760u16 , name : "pcHDTV HD-5500" }] } , Device { vendor_id : 5361u16 , id : 34817u16 , name : "CX23880/1/2/3 PCI Video and Audio Decoder [Audio Port]" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 10241u16 , name : "Hauppauge WinTV 28xxx (Roslyn) models" } , SubSystem { subvendor : 6235u16 , subdevice : 57344u16 , name : "VideoMate X500" } , SubSystem { subvendor : 22100u16 , subdevice : 9096u16 , name : "GoTView PCI Hybrid Audio AVStream Device" } , SubSystem { subvendor : 28771u16 , subdevice : 21760u16 , name : "pcHDTV HD-5500" }] } , Device { vendor_id : 5361u16 , id : 34818u16 , name : "CX23880/1/2/3 PCI Video and Audio Decoder [MPEG Port]" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 10241u16 , name : "Hauppauge WinTV 28xxx (Roslyn) models" } , SubSystem { subvendor : 112u16 , subdevice : 26882u16 , name : "WinTV HVR-4000-HD" } , SubSystem { subvendor : 112u16 , subdevice : 36866u16 , name : "Nova-T DVB-T Model 909" } , SubSystem { subvendor : 112u16 , subdevice : 37890u16 , name : "WinTV-HVR1100 DVB-T/Hybrid" } , SubSystem { subvendor : 112u16 , subdevice : 38400u16 , name : "WinTV 88x MPEG Encoder" } , SubSystem { subvendor : 4163u16 , subdevice : 18467u16 , name : "PVR-416" } , SubSystem { subvendor : 4221u16 , subdevice : 26172u16 , name : "Leadtek PVR 2000" } , SubSystem { subvendor : 4221u16 , subdevice : 26207u16 , name : "WinFast DTV1000-T" } , SubSystem { subvendor : 5361u16 , subdevice : 391u16 , name : "Conexant DVB-T reference design" } , SubSystem { subvendor : 6110u16 , subdevice : 2209u16 , name : "XPert DVB-T PCI BDA DVBT 23880 Transport Stream Capture" } , SubSystem { subvendor : 6110u16 , subdevice : 2214u16 , name : "KWorld/VStream XPert DVB-T" } , SubSystem { subvendor : 6149u16 , subdevice : 273u16 , name : "PICOLO Jet-X Jpeg" } , SubSystem { subvendor : 6149u16 , subdevice : 274u16 , name : "PICOLO Jet-X Jpeg" } , SubSystem { subvendor : 6149u16 , subdevice : 275u16 , name : "PICOLO Jet-X Jpeg" } , SubSystem { subvendor : 6149u16 , subdevice : 276u16 , name : "PICOLO Jet-X Jpeg" } , SubSystem { subvendor : 6316u16 , subdevice : 54528u16 , name : "DViCO FusionHDTV5 Gold" } , SubSystem { subvendor : 6316u16 , subdevice : 55312u16 , name : "DViCO FusionHDTV3 Gold-Q" } , SubSystem { subvendor : 6316u16 , subdevice : 55328u16 , name : "DViCO FusionHDTV3 Gold-T" } , SubSystem { subvendor : 6316u16 , subdevice : 56064u16 , name : "DVICO FusionHDTV DVB-T1" } , SubSystem { subvendor : 6316u16 , subdevice : 56080u16 , name : "DVICO FusionHDTV DVB-T Plus" } , SubSystem { subvendor : 22100u16 , subdevice : 9096u16 , name : "GoTView PCI Hybrid TS Capture Device" } , SubSystem { subvendor : 28771u16 , subdevice : 12288u16 , name : "pcHDTV HD3000 HDTV" } , SubSystem { subvendor : 28771u16 , subdevice : 21760u16 , name : "pcHDTV HD-5500" }] } , Device { vendor_id : 5361u16 , id : 34820u16 , name : "CX23880/1/2/3 PCI Video and Audio Decoder [IR Port]" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 26882u16 , name : "WinTV HVR-4000-HD" } , SubSystem { subvendor : 112u16 , subdevice : 36866u16 , name : "Nova-T DVB-T Model 909" } , SubSystem { subvendor : 112u16 , subdevice : 37890u16 , name : "WinTV-HVR1100 DVB-T/Hybrid" } , SubSystem { subvendor : 6149u16 , subdevice : 273u16 , name : "PICOLO Jet-X Control" } , SubSystem { subvendor : 6149u16 , subdevice : 274u16 , name : "PICOLO Jet-X Control" } , SubSystem { subvendor : 6149u16 , subdevice : 275u16 , name : "PICOLO Jet-X Control" } , SubSystem { subvendor : 6149u16 , subdevice : 276u16 , name : "PICOLO Jet-X Control" } , SubSystem { subvendor : 28771u16 , subdevice : 21760u16 , name : "pcHDTV HD-5500" }] } , Device { vendor_id : 5361u16 , id : 34833u16 , name : "CX23880/1/2/3 PCI Video and Audio Decoder [Audio Port]" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 13312u16 , name : "WinTV 34604" } , SubSystem { subvendor : 112u16 , subdevice : 13313u16 , name : "Hauppauge WinTV 34xxx models" } , SubSystem { subvendor : 112u16 , subdevice : 26882u16 , name : "WinTV HVR-4000-HD" } , SubSystem { subvendor : 112u16 , subdevice : 37890u16 , name : "WinTV-HVR1100 DVB-T/Hybrid" } , SubSystem { subvendor : 112u16 , subdevice : 38400u16 , name : "WinTV 88x Audio" } , SubSystem { subvendor : 5218u16 , subdevice : 34310u16 , name : "MSI TV-@nywhere Master" } , SubSystem { subvendor : 6316u16 , subdevice : 54528u16 , name : "DViCO FusionHDTV5 Gold" } , SubSystem { subvendor : 6316u16 , subdevice : 55312u16 , name : "DViCO FusionHDTV3 Gold-Q" } , SubSystem { subvendor : 6316u16 , subdevice : 55328u16 , name : "DViCO FusionHDTV3 Gold-T" } , SubSystem { subvendor : 6316u16 , subdevice : 56064u16 , name : "DVICO FusionHDTV DVB-T1" } , SubSystem { subvendor : 22100u16 , subdevice : 9096u16 , name : "GoTView PCI Hybrid Audio Capture Device" }] } , Device { vendor_id : 5361u16 , id : 34898u16 , name : "CX23885 PCI Video and Audio Decoder" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 32784u16 , name : "WinTV HVR-1400 ExpressCard" } , SubSystem { subvendor : 112u16 , subdevice : 61496u16 , name : "WinTV HVR-5525" } , SubSystem { subvendor : 4221u16 , subdevice : 28450u16 , name : "WinFast PxTV1200" } , SubSystem { subvendor : 4779u16 , subdevice : 54661u16 , name : "PE988J Hybrid ATSC/QAM PCI-E AVS Video Capture (SoftEncoder)" } , SubSystem { subvendor : 5058u16 , subdevice : 12307u16 , name : "TT-budget CT2-4500 CI" } , SubSystem { subvendor : 5217u16 , subdevice : 49209u16 , name : "AVerTV Hybrid Express (A577)" } , SubSystem { subvendor : 5435u16 , subdevice : 4478u16 , name : "Cinergy T PCIe Dual" } , SubSystem { subvendor : 6316u16 , subdevice : 56184u16 , name : "FusionHDTV DVB-T Dual Express" } , SubSystem { subvendor : 16980u16 , subdevice : 2384u16 , name : "S950" } , SubSystem { subvendor : 16980u16 , subdevice : 2386u16 , name : "S952" } , SubSystem { subvendor : 16980u16 , subdevice : 2434u16 , name : "T982" } , SubSystem { subvendor : 16980u16 , subdevice : 38272u16 , name : "T9580" } , SubSystem { subvendor : 16980u16 , subdevice : 38924u16 , name : "T980C" }] } , Device { vendor_id : 5361u16 , id : 34944u16 , name : "CX23887/8 PCIe Broadcast Audio and Video Decoder with 3D Comb" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 8793u16 , name : "WinTV HVR-1250" } , SubSystem { subvendor : 112u16 , subdevice : 27160u16 , name : "WinTV-quadHD" } , SubSystem { subvendor : 112u16 , subdevice : 49416u16 , name : "WinTV-HVR-4400-HD model 1278" } , SubSystem { subvendor : 5217u16 , subdevice : 12544u16 , name : "CE310B SD PCIe Video Capture Card" } , SubSystem { subvendor : 22100u16 , subdevice : 9097u16 , name : "GoTView X5 DVD Hybrid PCI-E" } , SubSystem { subvendor : 22100u16 , subdevice : 9104u16 , name : "GoTView X5 3D HYBRID PCI-E" }] }] }),
        (4866, Vendor { id : 4866u16 , name : "Computer Sciences Corp" , devices : & [] }),
        (5459, Vendor { id : 5459u16 , name : "CHICONY Electronics Co Ltd" , devices : & [] }),
        (4772, Vendor { id : 4772u16 , name : "NTT Innovative Devices Corporation" , devices : & [] }),
        (4833, Vendor { id : 4833u16 , name : "Nintendo Co, Ltd" , devices : & [] }),
        (4716, Vendor { id : 4716u16 , name : "Northern Telecom" , devices : & [Device { vendor_id : 4716u16 , id : 4625u16 , name : "10/100BaseTX [RTL81xx]" , subsystems : & [] } , Device { vendor_id : 4716u16 , id : 4716u16 , name : "802.11b Wireless Ethernet Adapter" , subsystems : & [] }] }),
        (6227, Vendor { id : 6227u16 , name : "SMSC Automotive Infotainment System Group" , devices : & [] }),
        (5803, Vendor { id : 5803u16 , name : "Global Sun Technology Inc" , devices : & [Device { vendor_id : 5803u16 , id : 4352u16 , name : "GL24110P" , subsystems : & [] } , Device { vendor_id : 5803u16 , id : 4353u16 , name : "PLX9052 PCMCIA-to-PCI Wireless LAN" , subsystems : & [] } , Device { vendor_id : 5803u16 , id : 4354u16 , name : "PCMCIA-to-PCI Wireless Network Bridge" , subsystems : & [] } , Device { vendor_id : 5803u16 , id : 34049u16 , name : "WL-8305 Wireless LAN PCI Adapter" , subsystems : & [] }] }),
        (5310, Vendor { id : 5310u16 , name : "L3 Communications" , devices : & [] }),
        (5071, Vendor { id : 5071u16 , name : "Studio Audio & Video Ltd" , devices : & [] }),
        (4364, Vendor { id : 4364u16 , name : "Mini-Max Technology, Inc." , devices : & [] }),
        (4418, Vendor { id : 4418u16 , name : "Alliance Semiconductor Corporation" , devices : & [Device { vendor_id : 4418u16 , id : 12816u16 , name : "AP6410" , subsystems : & [] } , Device { vendor_id : 4418u16 , id : 25634u16 , name : "ProVideo 6422" , subsystems : & [] } , Device { vendor_id : 4418u16 , id : 25636u16 , name : "ProVideo 6424" , subsystems : & [] } , Device { vendor_id : 4418u16 , id : 25637u16 , name : "ProMotion AT25" , subsystems : & [] } , Device { vendor_id : 4418u16 , id : 25661u16 , name : "ProMotion AT3D" , subsystems : & [] }] }),
        (18964, Vendor { id : 18964u16 , name : "NetVin" , devices : & [Device { vendor_id : 18964u16 , id : 20480u16 , name : "NV5000SC" , subsystems : & [SubSystem { subvendor : 18964u16 , subdevice : 20480u16 , name : "RT8029-Based Ethernet Adapter" }] }] }),
        (4329, Vendor { id : 4329u16 , name : "Alps Electric Co., Ltd." , devices : & [] }),
        (2827, Vendor { id : 2827u16 , name : "Rhino Equipment Corp." , devices : & [Device { vendor_id : 2827u16 , id : 261u16 , name : "R1T1" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 517u16 , name : "R4FXO" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 518u16 , name : "RCB4FXO 4-channel FXO analog telephony card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 773u16 , name : "R4T1" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1029u16 , name : "R8FXX" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1030u16 , name : "RCB8FXX 8-channel modular analog telephony card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1285u16 , name : "R24FXX" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1286u16 , name : "RCB24FXS 24-Channel FXS analog telephony card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1541u16 , name : "R2T1" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1797u16 , name : "R24FXS" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 1798u16 , name : "RCB24FXO 24-Channel FXO analog telephony card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 2309u16 , name : "R1T3 Single T3 Digital Telephony Card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 2310u16 , name : "RCB24FXX 24-channel modular analog telephony card" , subsystems : & [] } , Device { vendor_id : 2827u16 , id : 2566u16 , name : "RCB672FXX 672-channel modular analog telephony card" , subsystems : & [] }] }),
        (5027, Vendor { id : 5027u16 , name : "Hifn Inc." , devices : & [Device { vendor_id : 5027u16 , id : 5u16 , name : "7751 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 6u16 , name : "6500 Public Key Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 7u16 , name : "7811 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 18u16 , name : "7951 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 20u16 , name : "78XX Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 22u16 , name : "8065 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 23u16 , name : "8165 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 24u16 , name : "8154 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 29u16 , name : "7956 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 31u16 , name : "7855 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 32u16 , name : "7955 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 38u16 , name : "8155 Security Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 46u16 , name : "9630 Compression Processor" , subsystems : & [] } , Device { vendor_id : 5027u16 , id : 47u16 , name : "9725 Compression and Security Processor" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 5632u16 , name : "DR1600 Acceleration Card" } , SubSystem { subvendor : 5027u16 , subdevice : 5637u16 , name : "DR1605 Acceleration Card" } , SubSystem { subvendor : 5027u16 , subdevice : 5648u16 , name : "DR1610 Acceleration Card" } , SubSystem { subvendor : 5027u16 , subdevice : 5653u16 , name : "DR1615 Acceleration Card" } , SubSystem { subvendor : 5027u16 , subdevice : 5664u16 , name : "DR1620 Acceleration Card" } , SubSystem { subvendor : 5027u16 , subdevice : 5669u16 , name : "DR1625 Acceleration Card" }] } , Device { vendor_id : 5027u16 , id : 51u16 , name : "8201 Acceleration Processor" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 54u16 , name : "DX1710 Acceleration Card" }] } , Device { vendor_id : 5027u16 , id : 52u16 , name : "8202 Acceleration Processor" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 54u16 , name : "DX1720 Acceleration Card" }] } , Device { vendor_id : 5027u16 , id : 53u16 , name : "8203 Acceleration Processor" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 54u16 , name : "DX1730 Acceleration Card" }] } , Device { vendor_id : 5027u16 , id : 55u16 , name : "8204 Acceleration Processor" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 54u16 , name : "DX1740 Acceleration Card" }] } , Device { vendor_id : 5027u16 , id : 37440u16 , name : "XR9240 Compression and Security Coprocessor [Panther II]" , subsystems : & [SubSystem { subvendor : 5027u16 , subdevice : 37376u16 , name : "DX2040 Compression and Security Acceleration Card [Panther II]" }] }] }),
        (8402, Vendor { id : 8402u16 , name : "Awide Labs LTD." , devices : & [Device { vendor_id : 8402u16 , id : 512u16 , name : "XRaid [Extreme Performance Compression RAID Accelerator]" , subsystems : & [] }] }),
        (5450, Vendor { id : 5450u16 , name : "MAX Technologies Inc" , devices : & [] }),
        (4383, Vendor { id : 4383u16 , name : "Precision Digital Images" , devices : & [Device { vendor_id : 4383u16 , id : 19015u16 , name : "Precision MX Video engine interface" , subsystems : & [] } , Device { vendor_id : 4383u16 , id : 21059u16 , name : "Frame capture bus interface" , subsystems : & [] }] }),
        (5530, Vendor { id : 5530u16 , name : "General Instrument" , devices : & [] }),
        (4413, Vendor { id : 4413u16 , name : "Leading Edge Products Inc" , devices : & [] }),
        (4393, Vendor { id : 4393u16 , name : "Firmworks" , devices : & [] }),
        (5090, Vendor { id : 5090u16 , name : "Dynamics Research Corporation" , devices : & [] }),
        (4186, Vendor { id : 4186u16 , name : "Promise Technology, Inc." , devices : & [Device { vendor_id : 4186u16 , id : 3376u16 , name : "PDC20265 (FastTrak100 Lite/Ultra100)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32834u16 , name : "AV7266-E South Bridge Promise RAID" } , SubSystem { subvendor : 4186u16 , subdevice : 19763u16 , name : "Ultra100" }] } , Device { vendor_id : 4186u16 , id : 3384u16 , name : "20263" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19769u16 , name : "Fasttrak66" }] } , Device { vendor_id : 4186u16 , id : 4725u16 , name : "20275" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13080u16 , name : "PDC20318 (SATA150 TX4)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13081u16 , name : "PDC20319 (FastTrak S150 TX4)" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 13081u16 , name : "FastTrak S150 TX4 4 port SATA PCI board" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" }] } , Device { vendor_id : 4186u16 , id : 13169u16 , name : "PDC20371 (FastTrak S150 TX2plus)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13171u16 , name : "PDC20378 (FastTrak 378/SATA 378)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33013u16 , name : "K8V Deluxe/PC-DL Deluxe motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22797u16 , name : "KT6 Delta-FIS2R (MS-6590)" } , SubSystem { subvendor : 5218u16 , subdevice : 28718u16 , name : "K8T NEO FIS2R motherboard" }] } , Device { vendor_id : 4186u16 , id : 13173u16 , name : "PDC20375 (SATA150 TX2plus)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13174u16 , name : "PDC20376 (FastTrak 376)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32926u16 , name : "A7V8X motherboard" }] } , Device { vendor_id : 4186u16 , id : 13589u16 , name : "PDC40719 [FastTrak TX4300/TX4310]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13593u16 , name : "PDC40519 (FastTrak TX4200)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13680u16 , name : "PDC20771 [FastTrak TX2300]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13681u16 , name : "PDC20571 (FastTrak TX2200)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13684u16 , name : "PDC20579 SATAII 150 IDE Controller" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 13687u16 , name : "PDC40779 (SATA 300 779)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 15639u16 , name : "PDC40718 (SATA 300 TX4)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 15640u16 , name : "PDC20518/PDC40518 (SATAII 150 TX4)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 15731u16 , name : "PDC40775 (SATA 300 TX2plus)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 15733u16 , name : "PDC20575 (SATAII150 TX2plus)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 16160u16 , name : "PDC42819 [FastTrak TX2650/TX4650]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 17154u16 , name : "80333 [SuperTrak EX4350]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 19760u16 , name : "PDC20267 (FastTrak100/Ultra100)" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19763u16 , name : "Ultra100" } , SubSystem { subvendor : 4186u16 , subdevice : 19769u16 , name : "FastTrak100" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 4186u16 , id : 19763u16 , name : "20246" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19763u16 , name : "20246 IDE Controller" }] } , Device { vendor_id : 4186u16 , id : 19768u16 , name : "PDC20262 (FastTrak66/Ultra66)" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19760u16 , name : "Ultra Device on SuperTrak" } , SubSystem { subvendor : 4186u16 , subdevice : 19763u16 , name : "Ultra66" } , SubSystem { subvendor : 4186u16 , subdevice : 19769u16 , name : "FastTrak66" }] } , Device { vendor_id : 4186u16 , id : 19816u16 , name : "PDC20268 [Ultra100 TX2]" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19816u16 , name : "Ultra100 TX2" }] } , Device { vendor_id : 4186u16 , id : 19817u16 , name : "20269" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19816u16 , name : "Ultra133TX2" }] } , Device { vendor_id : 4186u16 , id : 21109u16 , name : "PDC20276 (MBFastTrak133 Lite)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32894u16 , name : "A7V333 motherboard." } , SubSystem { subvendor : 4186u16 , subdevice : 629u16 , name : "SuperTrak SX6000 IDE" } , SubSystem { subvendor : 4186u16 , subdevice : 4725u16 , name : "MBFastTrak133 Lite (tm) Controller (RAID mode)" } , SubSystem { subvendor : 5208u16 , subdevice : 45057u16 , name : "MBUltra 133" }] } , Device { vendor_id : 4186u16 , id : 21248u16 , name : "DC5300" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 25192u16 , name : "PDC20270 (FastTrak100 LP/TX2/TX4)" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 19816u16 , name : "FastTrak100 TX2" }] } , Device { vendor_id : 4186u16 , id : 25193u16 , name : "PDC20271 (FastTrak TX2000)" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 25193u16 , name : "FastTrak TX2/TX2000" }] } , Device { vendor_id : 4186u16 , id : 25344u16 , name : "PDC81731 [FastTrak SX8300]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 26145u16 , name : "PDC20621 (FastTrak S150 SX4/FastTrak SX4000 lite)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 26146u16 , name : "PDC20621 [SATA150 SX4] 4 Channel IDE RAID Controller" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 26148u16 , name : "PDC20621 [FastTrak SX4100]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 26150u16 , name : "PDC20618 (Ultra 618)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 26153u16 , name : "PDC20619 (FastTrak TX4000)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 29301u16 , name : "PDC20277 (SBFastTrak133 Lite)" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 32770u16 , name : "SATAII150 SX8" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 33616u16 , name : "80333 [SuperTrak EX8350/EX16350], 80331 [SuperTrak EX8300/EX16300]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 34384u16 , name : "81384 [SuperTrak EX SAS and SATA RAID Controller]" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 17920u16 , name : "SuperTrak EX4650A" } , SubSystem { subvendor : 4186u16 , subdevice : 17921u16 , name : "SuperTrak EX4650" } , SubSystem { subvendor : 4186u16 , subdevice : 17936u16 , name : "SuperTrak EX4650EL" } , SubSystem { subvendor : 4186u16 , subdevice : 34304u16 , name : "SuperTrak EX8650EL" } , SubSystem { subvendor : 4186u16 , subdevice : 34305u16 , name : "SuperTrak EX8650A" } , SubSystem { subvendor : 4186u16 , subdevice : 34306u16 , name : "SuperTrak EX8654" } , SubSystem { subvendor : 4186u16 , subdevice : 34307u16 , name : "SuperTrak EX8658" } , SubSystem { subvendor : 4186u16 , subdevice : 34308u16 , name : "SuperTrak EX8650" } , SubSystem { subvendor : 4186u16 , subdevice : 34320u16 , name : "SuperTrak EX8650M" } , SubSystem { subvendor : 4186u16 , subdevice : 42496u16 , name : "SuperTrak EX12650" } , SubSystem { subvendor : 4186u16 , subdevice : 46592u16 , name : "SuperTrak EX16650" } , SubSystem { subvendor : 4186u16 , subdevice : 46593u16 , name : "SuperTrak EX16654" } , SubSystem { subvendor : 4186u16 , subdevice : 46594u16 , name : "SuperTrak EX16658" }] } , Device { vendor_id : 4186u16 , id : 34656u16 , name : "PM8010 [SuperTrak EX SAS and SATA 6G RAID Controller]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 50000u16 , name : "80333 [SuperTrak EX12350]" , subsystems : & [] } , Device { vendor_id : 4186u16 , id : 58192u16 , name : "80333 [SuperTrak EX24350]" , subsystems : & [] }] }),
        (4146, Vendor { id : 4146u16 , name : "Compaq" , devices : & [] }),
        (5120, Vendor { id : 5120u16 , name : "Artx Inc" , devices : & [Device { vendor_id : 5120u16 , id : 5121u16 , name : "9432 TX" , subsystems : & [] }] }),
        (6277, Vendor { id : 6277u16 , name : "Avvida Systems Inc." , devices : & [] }),
        (4933, Vendor { id : 4933u16 , name : "Arescom Inc" , devices : & [] }),
        (5558, Vendor { id : 5558u16 , name : "Texas Memory Systems Inc" , devices : & [Device { vendor_id : 5558u16 , id : 1u16 , name : "XP15 DSP Accelerator" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 2u16 , name : "XP30 DSP Accelerator" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 3u16 , name : "XP00 Data Acquisition Device" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 4u16 , name : "XP35 DSP Accelerator" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 7u16 , name : "XP100 DSP Accelerator [XP100-T0]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 8u16 , name : "XP100 DSP Accelerator [XP100-T1]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 9u16 , name : "XP100 DSP Accelerator [XP100-E0]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 10u16 , name : "XP100 DSP Accelerator [XP100-E1]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 14u16 , name : "XP100 DSP Accelerator [XP100-0]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 15u16 , name : "XP100 DSP Accelerator [XP100-1]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 16u16 , name : "XP100 DSP Accelerator [XP100-P0]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 17u16 , name : "XP100 DSP Accelerator [XP100-P1]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 18u16 , name : "XP100 DSP Accelerator [XP100-P2]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 19u16 , name : "XP100 DSP Accelerator [XP100-P3]" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 20u16 , name : "RamSan Flash SSD" , subsystems : & [] } , Device { vendor_id : 5558u16 , id : 21u16 , name : "ZBox" , subsystems : & [] }] }),
        (6005, Vendor { id : 6005u16 , name : "General Electric" , devices : & [] }),
        (7256, Vendor { id : 7256u16 , name : "HGST, Inc." , devices : & [Device { vendor_id : 7256u16 , id : 3u16 , name : "Ultrastar SN100 Series NVMe SSD" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1269u16 , name : "PCIe3 1.6TB NVMe Flash Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 1270u16 , name : "PCIe3 3.2TB NVMe Flash Adapter" } , SubSystem { subvendor : 7256u16 , subdevice : 3u16 , name : "Ultrastar SN100/SN150 NVMe SSD" }] } , Device { vendor_id : 7256u16 , id : 35u16 , name : "Ultrastar SN200 Series NVMe SSD" , subsystems : & [SubSystem { subvendor : 7256u16 , subdevice : 34851u16 , name : "Ultrastar Memory (ME200)" }] }] }),
        (7653, Vendor { id : 7653u16 , name : "Eideticom, Inc" , devices : & [Device { vendor_id : 7653u16 , id : 4096u16 , name : "IO Memory Controller" , subsystems : & [] } , Device { vendor_id : 7653u16 , id : 8192u16 , name : "NoLoad Hardware Development Kit" , subsystems : & [] } , Device { vendor_id : 7653u16 , id : 8448u16 , name : "NoLoad Accelerator Platform" , subsystems : & [] } , Device { vendor_id : 7653u16 , id : 8704u16 , name : "NoLoad Cryptographic Accelerator" , subsystems : & [] } , Device { vendor_id : 7653u16 , id : 8960u16 , name : "NoLoad Data Services" , subsystems : & [] } , Device { vendor_id : 7653u16 , id : 12288u16 , name : "eBPF-based PCIe Accelerator" , subsystems : & [] }] }),
        (4219, Vendor { id : 4219u16 , name : "Gateway, Inc." , devices : & [] }),
        (4363, Vendor { id : 4363u16 , name : "Chromatic Research Inc." , devices : & [Device { vendor_id : 4363u16 , id : 1u16 , name : "Mpact Media Processor" , subsystems : & [] } , Device { vendor_id : 4363u16 , id : 4u16 , name : "Mpact 2" , subsystems : & [] }] }),
        (3601, Vendor { id : 3601u16 , name : "Compaq Computer Corporation" , devices : & [Device { vendor_id : 3601u16 , id : 1u16 , name : "PCI to EISA Bridge" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 2u16 , name : "PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 70u16 , name : "Smart Array 64xx" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 16529u16 , name : "Smart Array 6i" } , SubSystem { subvendor : 3601u16 , subdevice : 16538u16 , name : "Smart Array 641" } , SubSystem { subvendor : 3601u16 , subdevice : 16539u16 , name : "Smart Array 642" } , SubSystem { subvendor : 3601u16 , subdevice : 16540u16 , name : "Smart Array 6400" } , SubSystem { subvendor : 3601u16 , subdevice : 16541u16 , name : "Smart Array 6400 EM" }] } , Device { vendor_id : 3601u16 , id : 90u16 , name : "Remote Insight II board - Lights-Out" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 177u16 , name : "Remote Insight II board - PCI device" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 1288u16 , name : "Netelligent 4/16 Token Ring" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 4096u16 , name : "Triflex/Pentium Bridge, Model 1000" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 8192u16 , name : "Triflex/Pentium Bridge, Model 2000" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 12338u16 , name : "QVision 1280/p" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 12339u16 , name : "QVision 1280/p" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 12340u16 , name : "QVision 1280/p" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16384u16 , name : "4000 [Triflex]" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16448u16 , name : "Integrated Array" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16456u16 , name : "Compaq Raid LC2" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16496u16 , name : "Smart Array 5300" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16512u16 , name : "Smart Array 5i" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16514u16 , name : "Smart Array 532" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 16515u16 , name : "Smart Array 5312" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 24592u16 , name : "HotPlug PCI Bridge 6010" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 28704u16 , name : "USB Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 41196u16 , name : "Fibre Channel Host Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 41200u16 , name : "Advanced System Management Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45299u16 , name : "ProLiant DL360" }] } , Device { vendor_id : 3601u16 , id : 41203u16 , name : "Triflex PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 41207u16 , name : "PCI Hotplug Controller" , subsystems : & [SubSystem { subvendor : 32902u16 , subdevice : 42u16 , name : "PCI Hotplug Controller A" } , SubSystem { subvendor : 32902u16 , subdevice : 43u16 , name : "PCI Hotplug Controller B" }] } , Device { vendor_id : 3601u16 , id : 41208u16 , name : "ZFMicro Chipset USB" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 41212u16 , name : "FibreChannel HBA Tachyon" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44560u16 , name : "Smart-2/P RAID Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 16432u16 , name : "Smart-2/P Array Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16433u16 , name : "Smart-2SL Array Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16434u16 , name : "Smart Array 3200 Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16435u16 , name : "Smart Array 3100ES Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16436u16 , name : "Smart Array 221 Controller" }] } , Device { vendor_id : 3601u16 , id : 44585u16 , name : "MIS-L" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44586u16 , name : "MPC" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44587u16 , name : "MIS-E" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44593u16 , name : "System Management Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44594u16 , name : "Netelligent 10/100 TX PCI UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44595u16 , name : "Triflex Dual EIDE Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44596u16 , name : "Netelligent 10 T PCI UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44597u16 , name : "Integrated NetFlex-3/P" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44608u16 , name : "Netelligent Dual 10/100 TX PCI UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44611u16 , name : "Netelligent Integrated 10/100 TX UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44649u16 , name : "CETUS-L" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44652u16 , name : "Northstar" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 44653u16 , name : "NorthStar CPU to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45073u16 , name : "Netelligent 10/100 TX Embedded UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45074u16 , name : "Netelligent 10 T/2 PCI UTP/Coax" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45104u16 , name : "Netelligent 10/100 TX UTP" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45152u16 , name : "Smart Array 5300 Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45432u16 , name : "Smart Array 5i/532" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 16512u16 , name : "Smart Array 5i" } , SubSystem { subvendor : 3601u16 , subdevice : 16514u16 , name : "Smart Array 532" } , SubSystem { subvendor : 3601u16 , subdevice : 16515u16 , name : "Smart Array 5312" }] } , Device { vendor_id : 3601u16 , id : 45568u16 , name : "Memory Hot-Plug Controller" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 45571u16 , name : "Integrated Lights Out Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13061u16 , name : "iLO2" }] } , Device { vendor_id : 3601u16 , id : 45572u16 , name : "Integrated Lights Out  Processor" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13061u16 , name : "iLO2" }] } , Device { vendor_id : 3601u16 , id : 49152u16 , name : "Remote Insight Lights-Out Edition" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 61744u16 , name : "NetFlex-3/P ThunderLAN 1.0" , subsystems : & [] } , Device { vendor_id : 3601u16 , id : 61776u16 , name : "NetFlex-3/P ThunderLAN 2.3" , subsystems : & [] }] }),
        (5304, Vendor { id : 5304u16 , name : "Techsoft Technology Co Ltd" , devices : & [] }),
        (26880, Vendor { id : 26880u16 , name : "Red Hat, Inc." , devices : & [] }),
        (6190, Vendor { id : 6190u16 , name : "Raza Microelectronics, Inc." , devices : & [Device { vendor_id : 6190u16 , id : 8u16 , name : "XLR516 Processor" , subsystems : & [] }] }),
        (5595, Vendor { id : 5595u16 , name : "Applied Computing Systems Inc" , devices : & [] }),
        (4195, Vendor { id : 4195u16 , name : "Ocean Office Automation" , devices : & [] }),
        (4185, Vendor { id : 4185u16 , name : "Kontron" , devices : & [Device { vendor_id : 4185u16 , id : 4u16 , name : "FPGA M.2 (K20058)" , subsystems : & [SubSystem { subvendor : 4185u16 , subdevice : 0u16 , name : "FPGA M.2 (K20058)" }] }] }),
        (5064, Vendor { id : 5064u16 , name : "Apptech Inc" , devices : & [] }),
        (7881, Vendor { id : 7881u16 , name : "Wingtech Group(HongKong)Limited" , devices : & [] }),
        (5156, Vendor { id : 5156u16 , name : "Videoserver Connections" , devices : & [] }),
        (6039, Vendor { id : 6039u16 , name : "Intersil Techwell" , devices : & [Device { vendor_id : 6039u16 , id : 22628u16 , name : "TW5864 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26625u16 , name : "TW6802 multimedia video card" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26626u16 , name : "TW6802 multimedia other device" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26640u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26641u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26642u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26643u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26644u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26645u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26646u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26647u16 , name : "TW6816 multimedia video controller" , subsystems : & [] } , Device { vendor_id : 6039u16 , id : 26724u16 , name : "TW6864 multimedia video controller" , subsystems : & [] }] }),
        (16, Vendor { id : 16u16 , name : "Allied Telesis, Inc (Wrong ID)" , devices : & [Device { vendor_id : 16u16 , id : 33081u16 , name : "AT-2500TX V3 Ethernet" , subsystems : & [] }] }),
        (41472, Vendor { id : 41472u16 , name : "NEC Corporation" , devices : & [] }),
        (54577, Vendor { id : 54577u16 , name : "I+ME ACTIA GmbH" , devices : & [] }),
        (61355, Vendor { id : 61355u16 , name : "Enfabrica" , devices : & [] }),
        (5273, Vendor { id : 5273u16 , name : "EMTEC CO., Ltd" , devices : & [] }),
        (6404, Vendor { id : 6404u16 , name : "Hangzhou Silan Microelectronics Co., Ltd." , devices : & [Device { vendor_id : 6404u16 , id : 8241u16 , name : "SC92031 PCI Fast Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 6404u16 , id : 33081u16 , name : "RTL8139D [Realtek] PCI 10/100BaseTX ethernet adaptor" , subsystems : & [] }] }),
        (43523, Vendor { id : 43523u16 , name : "iTuner" , devices : & [] }),
        (6488, Vendor { id : 6488u16 , name : "Faster Technology, LLC." , devices : & [] }),
        (6922, Vendor { id : 6922u16 , name : "Pegatron" , devices : & [Device { vendor_id : 6922u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (5561, Vendor { id : 5561u16 , name : "Maestro Digital Communications" , devices : & [] }),
        (5455, Vendor { id : 5455u16 , name : "STRATABEAM Technology" , devices : & [] }),
        (5586, Vendor { id : 5586u16 , name : "FIC (First International Computer Inc)" , devices : & [] }),
        (8125, Vendor { id : 8125u16 , name : "Shenzhen Enrigin Technology Co., Ltd." , devices : & [Device { vendor_id : 8125u16 , id : 1u16 , name : "D10" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 2u16 , name : "D20 Plus" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 3u16 , name : "D20 Max" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 4u16 , name : "D20 Pro" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 5u16 , name : "D20 Plus N" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 6u16 , name : "D20 Pro N" , subsystems : & [] } , Device { vendor_id : 8125u16 , id : 257u16 , name : "T1" , subsystems : & [] }] }),
        (4144, Vendor { id : 4144u16 , name : "TMC Research" , devices : & [] }),
        (5151, Vendor { id : 5151u16 , name : "Visiontech Ltd" , devices : & [] }),
        (6950, Vendor { id : 6950u16 , name : "Netcope Technologies, a.s." , devices : & [Device { vendor_id : 6950u16 , id : 49458u16 , name : "COMBO-LXT155" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49600u16 , name : "NFB-100G1-e0" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49601u16 , name : "NFB-100G1-e1" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49744u16 , name : "NFB-200G2-master" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49745u16 , name : "NFB-200G2-slave" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49856u16 , name : "NFB-100G2-e0" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 49857u16 , name : "NFB-100G2-e1" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 52000u16 , name : "COMBO-20G" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 52032u16 , name : "COMBO-40G" , subsystems : & [] } , Device { vendor_id : 6950u16 , id : 52096u16 , name : "NFB-40G2" , subsystems : & [] }] }),
        (4251, Vendor { id : 4251u16 , name : "Gemlight Computer Ltd." , devices : & [] }),
        (5432, Vendor { id : 5432u16 , name : "ARALION Inc" , devices : & [Device { vendor_id : 5432u16 , id : 771u16 , name : "ARS106S Ultra ATA 133/100/66 Host Controller" , subsystems : & [] }] }),
        (4696, Vendor { id : 4696u16 , name : "Gilbarco, Inc." , devices : & [] }),
        (4694, Vendor { id : 4694u16 , name : "Perceptive Solutions, Inc." , devices : & [Device { vendor_id : 4694u16 , id : 16897u16 , name : "PCI-2220I" , subsystems : & [] } , Device { vendor_id : 4694u16 , id : 17409u16 , name : "PCI-2240I" , subsystems : & [] } , Device { vendor_id : 4694u16 , id : 20993u16 , name : "PCI-2000" , subsystems : & [] }] }),
        (5174, Vendor { id : 5174u16 , name : "CIS Technology Inc" , devices : & [] }),
        (7783, Vendor { id : 7783u16 , name : "Untether AI" , devices : & [Device { vendor_id : 7783u16 , id : 2u16 , name : "runAI200 AI Inference Accelerator" , subsystems : & [] } , Device { vendor_id : 7783u16 , id : 4u16 , name : "speedAI240 AI Inference Accelerator" , subsystems : & [] }] }),
        (4278, Vendor { id : 4278u16 , name : "Madge Networks" , devices : & [Device { vendor_id : 4278u16 , id : 1u16 , name : "Smart 16/4 PCI Ringnode" , subsystems : & [] } , Device { vendor_id : 4278u16 , id : 2u16 , name : "Smart 16/4 PCI Ringnode Mk2" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 2u16 , name : "Smart 16/4 PCI Ringnode Mk2" } , SubSystem { subvendor : 4278u16 , subdevice : 6u16 , name : "16/4 CardBus Adapter" }] } , Device { vendor_id : 4278u16 , id : 3u16 , name : "Smart 16/4 PCI Ringnode Mk3" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45309u16 , name : "Compaq NC4621 PCI, 4/16, WOL" } , SubSystem { subvendor : 4278u16 , subdevice : 3u16 , name : "Smart 16/4 PCI Ringnode Mk3" } , SubSystem { subvendor : 4278u16 , subdevice : 7u16 , name : "Presto PCI Plus Adapter" }] } , Device { vendor_id : 4278u16 , id : 4u16 , name : "Smart 16/4 PCI Ringnode Mk1" , subsystems : & [] } , Device { vendor_id : 4278u16 , id : 6u16 , name : "16/4 Cardbus Adapter" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 6u16 , name : "16/4 CardBus Adapter" }] } , Device { vendor_id : 4278u16 , id : 7u16 , name : "Presto PCI Adapter" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 7u16 , name : "Presto PCI" }] } , Device { vendor_id : 4278u16 , id : 9u16 , name : "Smart 100/16/4 PCI-HS Ringnode" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 9u16 , name : "Smart 100/16/4 PCI-HS Ringnode" }] } , Device { vendor_id : 4278u16 , id : 10u16 , name : "Token Ring 100/16/4 Ringnode/Ringrunner" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 10u16 , name : "Token Ring 100/16/4 Ringnode/Ringrunner" }] } , Device { vendor_id : 4278u16 , id : 11u16 , name : "16/4 CardBus Adapter Mk2" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 8u16 , name : "16/4 CardBus Adapter Mk2" } , SubSystem { subvendor : 4278u16 , subdevice : 11u16 , name : "16/4 Cardbus Adapter Mk2" }] } , Device { vendor_id : 4278u16 , id : 12u16 , name : "RapidFire 3140V2 16/4 TR Adapter" , subsystems : & [SubSystem { subvendor : 4278u16 , subdevice : 12u16 , name : "RapidFire 3140V2 16/4 TR Adapter" }] } , Device { vendor_id : 4278u16 , id : 4096u16 , name : "Collage 25/155 ATM Client Adapter" , subsystems : & [] } , Device { vendor_id : 4278u16 , id : 4097u16 , name : "Collage 155 ATM Server Adapter" , subsystems : & [] }] }),
        (4510, Vendor { id : 4510u16 , name : "Fujitsu Microelectronics Ltd." , devices : & [Device { vendor_id : 4510u16 , id : 1u16 , name : "FireStream 155" , subsystems : & [] } , Device { vendor_id : 4510u16 , id : 3u16 , name : "FireStream 50" , subsystems : & [] }] }),
        (5022, Vendor { id : 5022u16 , name : "Media4" , devices : & [] }),
        (4318, Vendor { id : 4318u16 , name : "NVIDIA Corporation" , devices : & [Device { vendor_id : 4318u16 , id : 8u16 , name : "NV1 [STG2000X-B Series]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9u16 , name : "NV1 [NV1 Series]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 24u16 , name : "NV3 [Riva 128]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 25u16 , name : "NV3 [Riva 128ZX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 32u16 , name : "NV4 [Riva TNT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "V3400 TNT" } , SubSystem { subvendor : 4168u16 , subdevice : 3096u16 , name : "Erazor II SGRAM" } , SubSystem { subvendor : 4168u16 , subdevice : 3097u16 , name : "Erazor II" } , SubSystem { subvendor : 4168u16 , subdevice : 3099u16 , name : "Erazor II" } , SubSystem { subvendor : 4168u16 , subdevice : 3100u16 , name : "Erazor II" } , SubSystem { subvendor : 4242u16 , subdevice : 1360u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 1362u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18436u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18440u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18448u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18450u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18453u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18464u16 , name : "Viper V550 with TV out" } , SubSystem { subvendor : 4242u16 , subdevice : 18466u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18692u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 18708u16 , name : "Viper V550" } , SubSystem { subvendor : 4242u16 , subdevice : 33317u16 , name : "Viper V550" } , SubSystem { subvendor : 4276u16 , subdevice : 10045u16 , name : "Velocity 4400" } , SubSystem { subvendor : 4276u16 , subdevice : 10046u16 , name : "Velocity 4400" } , SubSystem { subvendor : 4276u16 , subdevice : 10048u16 , name : "Velocity 4400" } , SubSystem { subvendor : 4318u16 , subdevice : 32u16 , name : "Riva TNT" } , SubSystem { subvendor : 4354u16 , subdevice : 4117u16 , name : "Graphics Blaster CT6710" } , SubSystem { subvendor : 4354u16 , subdevice : 4118u16 , name : "Graphics Blaster RIVA TNT" }] } , Device { vendor_id : 4318u16 , id : 40u16 , name : "NV5 [Riva TNT2 / TNT2 Pro]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "AGP-V3800 SGRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 513u16 , name : "AGP-V3800 SDRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 517u16 , name : "PCI-V3800" } , SubSystem { subvendor : 4163u16 , subdevice : 16384u16 , name : "AGP-V3800PRO" } , SubSystem { subvendor : 4168u16 , subdevice : 3105u16 , name : "Synergy II" } , SubSystem { subvendor : 4168u16 , subdevice : 3112u16 , name : "Erazor III" } , SubSystem { subvendor : 4168u16 , subdevice : 3113u16 , name : "Erazor III" } , SubSystem { subvendor : 4168u16 , subdevice : 3114u16 , name : "Erazor III" } , SubSystem { subvendor : 4168u16 , subdevice : 3115u16 , name : "Erazor III" } , SubSystem { subvendor : 4168u16 , subdevice : 3121u16 , name : "Erazor III Pro" } , SubSystem { subvendor : 4168u16 , subdevice : 3122u16 , name : "Erazor III Pro" } , SubSystem { subvendor : 4168u16 , subdevice : 3123u16 , name : "Erazor III Pro" } , SubSystem { subvendor : 4168u16 , subdevice : 3124u16 , name : "Erazor III Pro" } , SubSystem { subvendor : 4221u16 , subdevice : 8500u16 , name : "WinFast 3D S320 II + TV-Out" } , SubSystem { subvendor : 4242u16 , subdevice : 18436u16 , name : "Viper V770" } , SubSystem { subvendor : 4242u16 , subdevice : 18944u16 , name : "Viper V770" } , SubSystem { subvendor : 4242u16 , subdevice : 18946u16 , name : "Viper V770 Ultra" } , SubSystem { subvendor : 4242u16 , subdevice : 23040u16 , name : "RIVA TNT2/TNT2 Pro" } , SubSystem { subvendor : 4242u16 , subdevice : 23104u16 , name : "Viper V770D AGP" } , SubSystem { subvendor : 4242u16 , subdevice : 27138u16 , name : "Viper V770 Ultra" } , SubSystem { subvendor : 4242u16 , subdevice : 31234u16 , name : "Viper V770 Ultra" } , SubSystem { subvendor : 4318u16 , subdevice : 5u16 , name : "RIVA TNT2 Pro" } , SubSystem { subvendor : 4318u16 , subdevice : 15u16 , name : "Compaq NVIDIA TNT2 Pro" } , SubSystem { subvendor : 4354u16 , subdevice : 4128u16 , name : "3D Blaster RIVA TNT2" } , SubSystem { subvendor : 4354u16 , subdevice : 4134u16 , name : "3D Blaster RIVA TNT2 Digital" } , SubSystem { subvendor : 5218u16 , subdevice : 34822u16 , name : "MS-8806 AGPhantom Graphics Card" } , SubSystem { subvendor : 5295u16 , subdevice : 22544u16 , name : "Maxi Gamer Xentor" }] } , Device { vendor_id : 4318u16 , id : 41u16 , name : "NV5 [Riva TNT2 Ultra]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "AGP-V3800 Deluxe" } , SubSystem { subvendor : 4163u16 , subdevice : 513u16 , name : "AGP-V3800 Ultra SDRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 517u16 , name : "PCI-V3800 Ultra" } , SubSystem { subvendor : 4168u16 , subdevice : 3118u16 , name : "Erazor III Ultra" } , SubSystem { subvendor : 4168u16 , subdevice : 3119u16 , name : "Erazor III Ultra" } , SubSystem { subvendor : 4168u16 , subdevice : 3120u16 , name : "Erazor III Ultra" } , SubSystem { subvendor : 4354u16 , subdevice : 4129u16 , name : "3D Blaster RIVA TNT2 Ultra" } , SubSystem { subvendor : 4354u16 , subdevice : 4137u16 , name : "3D Blaster RIVA TNT2 Ultra" } , SubSystem { subvendor : 4354u16 , subdevice : 4143u16 , name : "3D Blaster RIVA TNT2 Ultra" } , SubSystem { subvendor : 5295u16 , subdevice : 22560u16 , name : "Maxi Gamer Xentor 32" } , SubSystem { subvendor : 18499u16 , subdevice : 20276u16 , name : "Dynamite" }] } , Device { vendor_id : 4318u16 , id : 44u16 , name : "NV6 [Vanta LT / Vanta / Vanta-16]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "AGP-V3800 Combat SDRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 513u16 , name : "AGP-V3800 Combat" } , SubSystem { subvendor : 4168u16 , subdevice : 3104u16 , name : "TNT2 Vanta" } , SubSystem { subvendor : 4168u16 , subdevice : 3105u16 , name : "TNT2 Vanta" } , SubSystem { subvendor : 4168u16 , subdevice : 3109u16 , name : "TNT2 Vanta 16MB" } , SubSystem { subvendor : 4242u16 , subdevice : 26656u16 , name : "Viper V730" } , SubSystem { subvendor : 4354u16 , subdevice : 4145u16 , name : "CT6938 VANTA 8MB" } , SubSystem { subvendor : 4354u16 , subdevice : 4148u16 , name : "CT6894 VANTA 16MB" } , SubSystem { subvendor : 5295u16 , subdevice : 20488u16 , name : "Maxi Gamer Phoenix 2" }] } , Device { vendor_id : 4318u16 , id : 45u16 , name : "NV6 [Riva TNT2 Model 64 / Model 64 Pro]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "AGP-V3800M" } , SubSystem { subvendor : 4163u16 , subdevice : 513u16 , name : "AGP-V3800M" } , SubSystem { subvendor : 4168u16 , subdevice : 3130u16 , name : "Erazor III LT" } , SubSystem { subvendor : 4168u16 , subdevice : 3131u16 , name : "Erazor III LT" } , SubSystem { subvendor : 4221u16 , subdevice : 8503u16 , name : "WinFast 3D S325" } , SubSystem { subvendor : 4318u16 , subdevice : 6u16 , name : "RIVA TNT2 Model 64/Model 64 Pro" } , SubSystem { subvendor : 4318u16 , subdevice : 30u16 , name : "M64 AGP4x" } , SubSystem { subvendor : 4354u16 , subdevice : 4131u16 , name : "CT6892 RIVA TNT2 Value" } , SubSystem { subvendor : 4354u16 , subdevice : 4132u16 , name : "CT6932 RIVA TNT2 Value 32Mb" } , SubSystem { subvendor : 4354u16 , subdevice : 4140u16 , name : "CT6931 RIVA TNT2 Value [Jumper]" } , SubSystem { subvendor : 4354u16 , subdevice : 4144u16 , name : "CT6931 RIVA TNT2 Value" } , SubSystem { subvendor : 4362u16 , subdevice : 111u16 , name : "GM1000-16" } , SubSystem { subvendor : 4362u16 , subdevice : 129u16 , name : "GM1000-16" } , SubSystem { subvendor : 5218u16 , subdevice : 34824u16 , name : "MSI-8808" } , SubSystem { subvendor : 5295u16 , subdevice : 22048u16 , name : "Gamer Cougar Video Edition" } , SubSystem { subvendor : 5460u16 , subdevice : 4161u16 , name : "Pixelview RIVA TNT2 M64" } , SubSystem { subvendor : 5481u16 , subdevice : 45u16 , name : "Palit Microsystems Daytona TNT2 M64" }] } , Device { vendor_id : 4318u16 , id : 52u16 , name : "MCP04 SMBus" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 53u16 , name : "MCP04 IDE" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 54u16 , name : "MCP04 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 55u16 , name : "MCP04 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 56u16 , name : "MCP04 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 58u16 , name : "MCP04 AC'97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 59u16 , name : "MCP04 USB Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 60u16 , name : "MCP04 USB Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 61u16 , name : "MCP04 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 62u16 , name : "MCP04 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 64u16 , name : "NV40 [GeForce 6800 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 65u16 , name : "NV40 [GeForce 6800]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33147u16 , name : "V9999 Gamer Edition" } , SubSystem { subvendor : 4221u16 , subdevice : 10642u16 , name : "WinFast A400" } , SubSystem { subvendor : 5208u16 , subdevice : 12559u16 , name : "Geforce 6800 GV-N6812" }] } , Device { vendor_id : 4318u16 , id : 66u16 , name : "NV40 [GeForce 6800 LE]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10651u16 , name : "WinFast A400 LE" }] } , Device { vendor_id : 4318u16 , id : 67u16 , name : "NV40 [GeForce 6800 XE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 68u16 , name : "NV40 [GeForce 6800 XT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 69u16 , name : "NV40 [GeForce 6800 GT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33149u16 , name : "V9999GT" } , SubSystem { subvendor : 5208u16 , subdevice : 12608u16 , name : "GV-N68T256D" }] } , Device { vendor_id : 4318u16 , id : 71u16 , name : "NV40 [GeForce 6800 GS]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 8457u16 , name : "GeForce 6800 GS" }] } , Device { vendor_id : 4318u16 , id : 72u16 , name : "NV40 [GeForce 6800 XT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 78u16 , name : "NV40GL [Quadro FX 4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 80u16 , name : "CK804 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 3089u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 81u16 , name : "CK804 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 ISA Bridge" }] } , Device { vendor_id : 4318u16 , id : 82u16 , name : "CK804 SMBus" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 SMBus" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 3089u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 83u16 , name : "CK804 IDE" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 20482u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 84u16 , name : "CK804 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 Serial ATA" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 45059u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 21505u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 85u16 , name : "CK804 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 Serial ATA" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 45059u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 21505u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 86u16 , name : "CK804 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 87u16 , name : "CK804 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33089u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4318u16 , subdevice : 52100u16 , name : "NF4 Lanparty" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 9473u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 88u16 , name : "CK804 AC'97 Modem" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 89u16 , name : "CK804 AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33066u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5218u16 , subdevice : 30085u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 33297u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 90u16 , name : "CK804 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 onboard USB" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 91u16 , name : "CK804 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 onboard USB" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "MSI K8N Diamond" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 92u16 , name : "CK804 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 93u16 , name : "CK804 PCIE Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 94u16 , name : "CK804 Memory Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 Memory Controller" } , SubSystem { subvendor : 4163u16 , subdevice : 33114u16 , name : "A8N Series Mainboard" } , SubSystem { subvendor : 4318u16 , subdevice : 94u16 , name : "ECS Elitegroup NFORCE3-A939 motherboard." } , SubSystem { subvendor : 4337u16 , subdevice : 10341u16 , name : "Tomcat K8E (S2865)" } , SubSystem { subvendor : 4337u16 , subdevice : 10385u16 , name : "Thunder K8SRE Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28928u16 , name : "K8N Diamond Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28965u16 , name : "K8N Neo4-F Mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7194u16 , name : "KN8-Ultra Mainboard" } , SubSystem { subvendor : 5477u16 , subdevice : 13314u16 , name : "NF4 AM2L Mainboard" }] } , Device { vendor_id : 4318u16 , id : 95u16 , name : "CK804 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 96u16 , name : "nForce2 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32941u16 , name : "A7N8X Mainboard" } , SubSystem { subvendor : 5243u16 , subdevice : 7170u16 , name : "NF7-S/NF7 (nVidia-nForce2) 2.X" } , SubSystem { subvendor : 41120u16 , subdevice : 954u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 100u16 , name : "nForce2 SMBus (MCP)" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7170u16 , name : "NF7-S/NF7 (nVidia-nForce2) 2.X" } , SubSystem { subvendor : 41120u16 , subdevice : 955u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 101u16 , name : "nForce2 IDE" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 3089u16 , name : "nForce 2 EIDE Controller" } , SubSystem { subvendor : 41120u16 , subdevice : 946u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 102u16 , name : "nForce2 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32935u16 , name : "A7N8X Mainboard onboard nForce2 Ethernet" } , SubSystem { subvendor : 4318u16 , subdevice : 3089u16 , name : "nForce MCP-T Networking Adapter" } , SubSystem { subvendor : 41120u16 , subdevice : 947u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 103u16 , name : "nForce2 USB Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 3089u16 , name : "A7N8X Mainboard" } , SubSystem { subvendor : 41120u16 , subdevice : 948u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 104u16 , name : "nForce2 USB Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 3089u16 , name : "A7N8X Mainboard" } , SubSystem { subvendor : 41120u16 , subdevice : 948u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 106u16 , name : "nForce2 AC97 Audio Controller (MCP)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32917u16 , name : "nForce2 AC97 Audio Controller (MCP)" } , SubSystem { subvendor : 41120u16 , subdevice : 772u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 107u16 , name : "nForce Audio Processing Unit" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 107u16 , name : "nForce2 MCP Audio Processing Unit" } , SubSystem { subvendor : 41120u16 , subdevice : 772u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 108u16 , name : "nForce2 External PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 109u16 , name : "nForce2 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 110u16 , name : "nForce2 FireWire (IEEE 1394) Controller" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 774u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 128u16 , name : "MCP2A ISA bridge" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 132u16 , name : "MCP2A SMBus" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 133u16 , name : "MCP2A IDE" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 134u16 , name : "MCP2A Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 135u16 , name : "MCP2A USB Controller" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 136u16 , name : "MCP2A USB Controller" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 138u16 , name : "MCP2S AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 139u16 , name : "MCP2A PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 140u16 , name : "MCP2A Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 142u16 , name : "nForce2 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 144u16 , name : "G70 [GeForce 7800 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 145u16 , name : "G70 [GeForce 7800 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 146u16 , name : "G70 [GeForce 7800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 147u16 , name : "G70 [GeForce 7800 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 148u16 , name : "High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 149u16 , name : "G70 [GeForce 7800 SLI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 151u16 , name : "G70 [GeForce GTS 250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 152u16 , name : "G70M [GeForce Go 7800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 153u16 , name : "G70M [GeForce Go 7800 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 157u16 , name : "G70GL [Quadro FX 4500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 160u16 , name : "NV0A [Aladdin TNT2 IGP]" , subsystems : & [SubSystem { subvendor : 5295u16 , subdevice : 22544u16 , name : "Maxi Gamer Xentor" }] } , Device { vendor_id : 4318u16 , id : 192u16 , name : "NV41 [GeForce 6800 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 193u16 , name : "NV41 [GeForce 6800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 194u16 , name : "NV41 [GeForce 6800 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 195u16 , name : "NV41 [GeForce 6800 XT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 197u16 , name : "NV41" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 198u16 , name : "NV41" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 199u16 , name : "NV41" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 200u16 , name : "NV41M [GeForce Go 6800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 201u16 , name : "NV41M [GeForce Go 6800 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 204u16 , name : "NV41GLM [Quadro FX Go1400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 205u16 , name : "NV42GL [Quadro FX 3450/4000 SDI]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 667u16 , name : "Quadro FX 3450" }] } , Device { vendor_id : 4318u16 , id : 206u16 , name : "NV41GL [Quadro FX 1400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 207u16 , name : "NV41" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 208u16 , name : "nForce3 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 209u16 , name : "nForce3 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 210u16 , name : "nForce3 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 211u16 , name : "CK804 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 212u16 , name : "nForce3 SMBus" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 213u16 , name : "nForce3 IDE" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 214u16 , name : "nForce3 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 215u16 , name : "nForce3 USB 1.1" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 216u16 , name : "nForce3 USB 2.0" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 217u16 , name : "nForce3 Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 218u16 , name : "nForce3 Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 221u16 , name : "nForce3 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 223u16 , name : "CK8S Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32935u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" }] } , Device { vendor_id : 4318u16 , id : 224u16 , name : "nForce3 250Gb LPC Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4318u16 , subdevice : 3089u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 224u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 225u16 , name : "nForce3 250Gb Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 225u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 226u16 , name : "nForce3 250Gb AGP Host to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 227u16 , name : "nForce3 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 227u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 228u16 , name : "nForce 250Gb PCI System Management" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 228u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 229u16 , name : "CK8S Parallel ATA Controller (v2.5)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 229u16 , name : "Motherboard (one of many)" } , SubSystem { subvendor : 63561u16 , subdevice : 229u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 230u16 , name : "CK8S Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 231u16 , name : "CK8S USB Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 231u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 232u16 , name : "nForce3 EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33087u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 28720u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" } , SubSystem { subvendor : 6217u16 , subdevice : 232u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4318u16 , id : 234u16 , name : "nForce3 250Gb AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33181u16 , name : "K8N-E" } , SubSystem { subvendor : 4187u16 , subdevice : 3139u16 , name : "Winfast NF3250K8AA" } , SubSystem { subvendor : 5218u16 , subdevice : 45072u16 , name : "K8N Neo-FSR v2.0" } , SubSystem { subvendor : 5243u16 , subdevice : 7179u16 , name : "NF8 Mainboard" }] } , Device { vendor_id : 4318u16 , id : 237u16 , name : "nForce3 250Gb PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 238u16 , name : "nForce3 Serial ATA Controller 2" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 241u16 , name : "NV43 [GeForce 6600 GT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33190u16 , name : "N6600GT TD 128M AGP" } , SubSystem { subvendor : 4163u16 , subdevice : 33222u16 , name : "N6600GT TD 128M AGP" } , SubSystem { subvendor : 5208u16 , subdevice : 12624u16 , name : "GV-N66T128VP" } , SubSystem { subvendor : 5460u16 , subdevice : 4497u16 , name : "PixelView PV-N43UA (128KD)" } , SubSystem { subvendor : 5762u16 , subdevice : 8473u16 , name : "GeForce 6600 GT AGP" }] } , Device { vendor_id : 4318u16 , id : 242u16 , name : "NV43 [GeForce 6600]" , subsystems : & [SubSystem { subvendor : 5460u16 , subdevice : 4500u16 , name : "PixelView PV-N43AT (256KD)" } , SubSystem { subvendor : 5762u16 , subdevice : 8476u16 , name : "GeForce 6600 256MB DDR DUAL DVI TV" }] } , Device { vendor_id : 4318u16 , id : 243u16 , name : "NV43 [GeForce 6200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 244u16 , name : "NV43 [GeForce 6600 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 245u16 , name : "G70/G71 [GeForce 7800 GS AGP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 246u16 , name : "NV43 [GeForce 6800 GS/XT]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 8574u16 , name : "XFX GeForce 6800 XTreme 256MB DDR3 AGP" }] } , Device { vendor_id : 4318u16 , id : 248u16 , name : "NV45GL [Quadro FX 3400/4400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 249u16 , name : "NV40 [GeForce 6800 GT/GTO/Ultra]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 249u16 , name : "NV40 [GeForce 6800 GT]" } , SubSystem { subvendor : 5762u16 , subdevice : 8480u16 , name : "GEFORCE 6800 GT PCI-E" }] } , Device { vendor_id : 4318u16 , id : 250u16 , name : "NV39 [GeForce PCX 5750]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 251u16 , name : "NV35 [GeForce PCX 5900]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 252u16 , name : "NV37GL [Quadro FX 330/GeForce PCX 5300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 253u16 , name : "NV37GL [Quadro PCI-E Series]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 254u16 , name : "NV38GL [Quadro FX 1300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 255u16 , name : "NV19 [GeForce PCX 4300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 256u16 , name : "NV10 [GeForce 256 SDR]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 512u16 , name : "AGP-V6600 SGRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 513u16 , name : "AGP-V6600 SDRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 16392u16 , name : "AGP-V6600 SGRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 16393u16 , name : "AGP-V6600 SDRAM" } , SubSystem { subvendor : 4168u16 , subdevice : 3137u16 , name : "Erazor X" } , SubSystem { subvendor : 4168u16 , subdevice : 3139u16 , name : "ERAZOR X PCI" } , SubSystem { subvendor : 4168u16 , subdevice : 3144u16 , name : "Synergy Force" } , SubSystem { subvendor : 4354u16 , subdevice : 4141u16 , name : "CT6941 GeForce 256" } , SubSystem { subvendor : 5295u16 , subdevice : 20514u16 , name : "3D Prophet SE" }] } , Device { vendor_id : 4318u16 , id : 257u16 , name : "NV10 [GeForce 256 DDR]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 514u16 , name : "AGP-V6800 DDR" } , SubSystem { subvendor : 4163u16 , subdevice : 16394u16 , name : "AGP-V6800 DDR SGRAM" } , SubSystem { subvendor : 4163u16 , subdevice : 16395u16 , name : "AGP-V6800 DDR SDRAM" } , SubSystem { subvendor : 4168u16 , subdevice : 3138u16 , name : "Erazor X" } , SubSystem { subvendor : 4221u16 , subdevice : 10274u16 , name : "WinFast GeForce 256" } , SubSystem { subvendor : 4354u16 , subdevice : 4142u16 , name : "CT6970/CT6971" } , SubSystem { subvendor : 5295u16 , subdevice : 20513u16 , name : "3D Prophet DDR-DVI" }] } , Device { vendor_id : 4318u16 , id : 259u16 , name : "NV10GL [Quadro]" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 3136u16 , name : "GLoria II-64" } , SubSystem { subvendor : 4168u16 , subdevice : 3140u16 , name : "GLoria II" } , SubSystem { subvendor : 4168u16 , subdevice : 3141u16 , name : "GLoria II" } , SubSystem { subvendor : 4168u16 , subdevice : 3146u16 , name : "GLoria II-64 Pro" } , SubSystem { subvendor : 4168u16 , subdevice : 3147u16 , name : "GLoria II-64 Pro DVII" } , SubSystem { subvendor : 4265u16 , subdevice : 36866u16 , name : "VPro VR3" }] } , Device { vendor_id : 4318u16 , id : 272u16 , name : "NV11 [GeForce2 MX/MX 400]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 16405u16 , name : "AGP-V7100 Pro" } , SubSystem { subvendor : 4163u16 , subdevice : 16417u16 , name : "V7100 Deluxe Combo" } , SubSystem { subvendor : 4163u16 , subdevice : 16433u16 , name : "V7100 Pro with TV output" } , SubSystem { subvendor : 4168u16 , subdevice : 3168u16 , name : "Gladiac MX" } , SubSystem { subvendor : 4168u16 , subdevice : 3169u16 , name : "Gladiac 511PCI" } , SubSystem { subvendor : 4168u16 , subdevice : 3171u16 , name : "Gladiac 511TV-OUT 32MB" } , SubSystem { subvendor : 4168u16 , subdevice : 3172u16 , name : "Gladiac 511TV-OUT 64MB" } , SubSystem { subvendor : 4168u16 , subdevice : 3173u16 , name : "Gladiac 511TWIN" } , SubSystem { subvendor : 4168u16 , subdevice : 3174u16 , name : "Gladiac 311" } , SubSystem { subvendor : 4272u16 , subdevice : 1u16 , name : "GeForce2 MX Jumbo TV" } , SubSystem { subvendor : 4318u16 , subdevice : 145u16 , name : "Dell OEM GeForce 2 MX 400" } , SubSystem { subvendor : 4318u16 , subdevice : 161u16 , name : "Apple OEM GeForce2 MX" } , SubSystem { subvendor : 5218u16 , subdevice : 34083u16 , name : "MS-8852" } , SubSystem { subvendor : 5218u16 , subdevice : 34839u16 , name : "MSI GeForce2 MX400 Pro32S [MS-8817]" } , SubSystem { subvendor : 5295u16 , subdevice : 28930u16 , name : "3D Prophet II MX" } , SubSystem { subvendor : 5295u16 , subdevice : 28931u16 , name : "3D Prophet II MX Dual-Display" } , SubSystem { subvendor : 5445u16 , subdevice : 35u16 , name : "Xtasy Rev. B2" } , SubSystem { subvendor : 5460u16 , subdevice : 4225u16 , name : "MVGA-NVG11AM(400)" }] } , Device { vendor_id : 4318u16 , id : 273u16 , name : "NV11 [GeForce2 MX200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 274u16 , name : "NV11M [GeForce2 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 275u16 , name : "NV11GL [Quadro2 MXR/EX/Go]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 229u16 , name : "Quadro2 Go" }] } , Device { vendor_id : 4318u16 , id : 320u16 , name : "NV43 [GeForce 6600 GT]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 12581u16 , name : "GV-NX66T128D" } , SubSystem { subvendor : 5208u16 , subdevice : 12582u16 , name : "GV-NX66T256DE" } , SubSystem { subvendor : 5218u16 , subdevice : 35129u16 , name : "MS-8983" }] } , Device { vendor_id : 4318u16 , id : 321u16 , name : "NV43 [GeForce 6600]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33200u16 , name : "EN6600 Silencer" } , SubSystem { subvendor : 4221u16 , subdevice : 22842u16 , name : "LR2A22 128MB TV OUT" } , SubSystem { subvendor : 4221u16 , subdevice : 22907u16 , name : "WINFAST PX6600" } , SubSystem { subvendor : 5208u16 , subdevice : 12580u16 , name : "GV-NX66128DP Turbo Force Edition" }] } , Device { vendor_id : 4318u16 , id : 322u16 , name : "NV43 [GeForce 6600 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 323u16 , name : "NV43 [GeForce 6600 VE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 324u16 , name : "NV43M [GeForce Go 6600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 325u16 , name : "NV43 [GeForce 6610 XL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 326u16 , name : "NV43M [GeForce Go6200 TE / 6600 TE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 327u16 , name : "NV43 [GeForce 6700 XL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 328u16 , name : "NV43M [GeForce Go 6600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 329u16 , name : "NV43M [GeForce Go 6600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 330u16 , name : "NV43 [Quadro NVS 440]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 331u16 , name : "NV43" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 333u16 , name : "NV43GL [Quadro FX 550]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 334u16 , name : "NV43GL [Quadro FX 540]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 335u16 , name : "NV43 [GeForce 6200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 336u16 , name : "NV15 [GeForce2 GTS/Pro]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 16406u16 , name : "V7700 AGP Video Card" } , SubSystem { subvendor : 4163u16 , subdevice : 16426u16 , name : "AGP-V7700" } , SubSystem { subvendor : 4168u16 , subdevice : 3152u16 , name : "Gladiac" } , SubSystem { subvendor : 4168u16 , subdevice : 3154u16 , name : "Gladiac-64" } , SubSystem { subvendor : 4221u16 , subdevice : 10304u16 , name : "WinFast GeForce2 GTS with TV output" } , SubSystem { subvendor : 4221u16 , subdevice : 10306u16 , name : "WinFast GeForce 2 Pro" } , SubSystem { subvendor : 4318u16 , subdevice : 46u16 , name : "GeForce2 GTS" } , SubSystem { subvendor : 5218u16 , subdevice : 33114u16 , name : "MS-8815" } , SubSystem { subvendor : 5218u16 , subdevice : 34865u16 , name : "Creative GeForce2 Pro" }] } , Device { vendor_id : 4318u16 , id : 337u16 , name : "NV15 [GeForce2 Ti]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 16479u16 , name : "V7700Ti" } , SubSystem { subvendor : 5218u16 , subdevice : 21766u16 , name : "Creative 3D Blaster GeForce2 Titanium" } , SubSystem { subvendor : 5218u16 , subdevice : 33636u16 , name : "MS-8836" }] } , Device { vendor_id : 4318u16 , id : 338u16 , name : "NV15 [GeForce2 Ultra]" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 3158u16 , name : "GLADIAC Ultra" }] } , Device { vendor_id : 4318u16 , id : 339u16 , name : "NV15GL [Quadro2 Pro]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 352u16 , name : "NV44 [GeForce 6500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 353u16 , name : "NV44 [GeForce 6200 TurboCache]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 354u16 , name : "NV44 [GeForce 6200 SE TurboCache]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 355u16 , name : "NV44 [GeForce 6200 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 356u16 , name : "NV44M [GeForce Go 6200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 357u16 , name : "NV44 [Quadro NVS 285]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 358u16 , name : "NV44M [GeForce Go 6400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 359u16 , name : "NV44M [GeForce Go 6200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 360u16 , name : "NV44M [GeForce Go 6400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 361u16 , name : "NV44 [GeForce 6250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 362u16 , name : "NV44 [GeForce 7100 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 365u16 , name : "NV44" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 366u16 , name : "NV44" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 367u16 , name : "NV44" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 368u16 , name : "NV17 [GeForce4 MX 460]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 34352u16 , name : "MS-8863" }] } , Device { vendor_id : 4318u16 , id : 369u16 , name : "NV17 [GeForce4 MX 440]" , subsystems : & [SubSystem { subvendor : 4272u16 , subdevice : 2u16 , name : "Gainward Pro/600 TV" } , SubSystem { subvendor : 4318u16 , subdevice : 8u16 , name : "Apple OEM GeForce4 MX 440" } , SubSystem { subvendor : 5218u16 , subdevice : 34401u16 , name : "G4MX440-VTP" } , SubSystem { subvendor : 5218u16 , subdevice : 34608u16 , name : "MX440SES-T (MS-8873)" } , SubSystem { subvendor : 5218u16 , subdevice : 34627u16 , name : "MS-8874" } , SubSystem { subvendor : 5218u16 , subdevice : 34898u16 , name : "GeForce4 MX440 PCI" } , SubSystem { subvendor : 5243u16 , subdevice : 36608u16 , name : "Abit Siluro GeForce4MX440" }] } , Device { vendor_id : 4318u16 , id : 370u16 , name : "NV17 [GeForce4 MX 420]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 34608u16 , name : "MS-8873" } , SubSystem { subvendor : 5218u16 , subdevice : 34692u16 , name : "MS-8878" }] } , Device { vendor_id : 4318u16 , id : 371u16 , name : "NV17 [GeForce4 MX 440-SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 372u16 , name : "NV17M [GeForce4 440 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 373u16 , name : "NV17M [GeForce4 420 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 374u16 , name : "NV17M [GeForce4 420 Go 32M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 / Vx9 mainboard" }] } , Device { vendor_id : 4318u16 , id : 375u16 , name : "NV17M [GeForce4 460 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 376u16 , name : "NV17GL [Quadro4 550 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 377u16 , name : "NV17M [GeForce4 440 Go 64M]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 377u16 , name : "GeForce4 MX (Mac)" }] } , Device { vendor_id : 4318u16 , id : 378u16 , name : "NV17GL [Quadro NVS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 379u16 , name : "NV17GL [Quadro4 550 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 380u16 , name : "NV17GL [Quadro4 500 GoGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 383u16 , name : "NV17" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 385u16 , name : "NV18 [GeForce4 MX 440 AGP 8x]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32867u16 , name : "GeForce4 MX 440 AGP 8X" } , SubSystem { subvendor : 4163u16 , subdevice : 32879u16 , name : "V9180 Magic" } , SubSystem { subvendor : 5218u16 , subdevice : 34944u16 , name : "MS-StarForce GeForce4 MX 440 with AGP8X" } , SubSystem { subvendor : 5218u16 , subdevice : 35072u16 , name : "MS-8890 GeForce 4 MX440 AGP8X" } , SubSystem { subvendor : 5218u16 , subdevice : 37712u16 , name : "MSI GeForce4 MX T8X with AGP8X" } , SubSystem { subvendor : 5243u16 , subdevice : 36621u16 , name : "Siluro GF4 MX-8X" } , SubSystem { subvendor : 5460u16 , subdevice : 4369u16 , name : "PixelView MVGA-NVG18A" }] } , Device { vendor_id : 4318u16 , id : 386u16 , name : "NV18 [GeForce4 MX 440SE AGP 8x]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 387u16 , name : "NV18 [GeForce4 MX 420 AGP 8x]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 388u16 , name : "NV18 [GeForce4 MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 389u16 , name : "NV18 [GeForce4 MX 4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 390u16 , name : "NV18M [GeForce4 448 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 391u16 , name : "NV18M [GeForce4 488 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 392u16 , name : "NV18GL [Quadro4 580 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 393u16 , name : "NV18 [GeForce4 MX with AGP8X (Mac)]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 394u16 , name : "NV18GL [Quadro NVS 280 SD]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 395u16 , name : "NV18GL [Quadro4 380 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 396u16 , name : "NV18GL [Quadro NVS 50 PCI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 397u16 , name : "NV18M [GeForce4 448 Go]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 400u16 , name : "G80 [GeForce 8800 GTS / 8800 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 401u16 , name : "G80 [GeForce 8800 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 402u16 , name : "G80 [GeForce 8800 GTS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 403u16 , name : "G80 [GeForce 8800 GTS]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 8381u16 , name : "WinFast PX 8800 GTS TDH" }] } , Device { vendor_id : 4318u16 , id : 404u16 , name : "G80 [GeForce 8800 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 407u16 , name : "G80GL [Tesla C870]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 413u16 , name : "G80GL [Quadro FX 5600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 414u16 , name : "G80GL [Quadro FX 4600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 416u16 , name : "nForce 220/420 NV1A [GeForce2 MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 420u16 , name : "nForce CPU bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 427u16 , name : "nForce 420 Memory Controller (DDR)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 428u16 , name : "nForce 220/420 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 429u16 , name : "nForce 220/420 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 432u16 , name : "nForce Audio Processing Unit" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 433u16 , name : "nForce AC'97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 434u16 , name : "nForce ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 436u16 , name : "nForce PCI System Management" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 439u16 , name : "nForce AGP to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 440u16 , name : "nForce PCI-to-PCI bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 444u16 , name : "nForce IDE" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 449u16 , name : "nForce AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 450u16 , name : "nForce USB Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 451u16 , name : "nForce Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 464u16 , name : "G72 [GeForce 7350 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 465u16 , name : "G72 [GeForce 7300 LE]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 24314u16 , name : "WinFast PX7300LE-TD128" } , SubSystem { subvendor : 4221u16 , subdevice : 24315u16 , name : "WinFast PX7300LE-TD256" } , SubSystem { subvendor : 5218u16 , subdevice : 837u16 , name : "7300LE PCI Express Graphics Adapter" }] } , Device { vendor_id : 4318u16 , id : 466u16 , name : "G72 [GeForce 7550 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 467u16 , name : "G72 [GeForce 7200 GS / 7300 SE]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33283u16 , name : "EN7300SE" } , SubSystem { subvendor : 4163u16 , subdevice : 33360u16 , name : "EN7200GS" }] } , Device { vendor_id : 4318u16 , id : 469u16 , name : "G72" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 470u16 , name : "G72M [GeForce Go 7200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 471u16 , name : "G72M [Quadro NVS 110M/GeForce Go 7300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 472u16 , name : "G72M [GeForce Go 7400]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" }] } , Device { vendor_id : 4318u16 , id : 473u16 , name : "G72M [GeForce Go 7450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 474u16 , name : "G72M [Quadro NVS 110M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 475u16 , name : "G72M [Quadro NVS 120M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 476u16 , name : "G72GLM [Quadro FX 350M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 477u16 , name : "G72 [GeForce 7500 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 478u16 , name : "G72GL [Quadro FX 350]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 476u16 , name : "Quadro  FX Go350M" }] } , Device { vendor_id : 4318u16 , id : 479u16 , name : "G72 [GeForce 7300 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 480u16 , name : "nForce2 IGP2" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7177u16 , name : "NV7 Motherboard" }] } , Device { vendor_id : 4318u16 , id : 488u16 , name : "nForce2 AGP" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 490u16 , name : "nForce2 Memory Controller 0" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 491u16 , name : "nForce2 Memory Controller 1" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 492u16 , name : "nForce2 Memory Controller 2" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 493u16 , name : "nForce2 Memory Controller 3" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 494u16 , name : "nForce2 Memory Controller 4" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 494u16 , name : "MSI Delta-L nForce2 memory controller" } , SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 495u16 , name : "nForce2 Memory Controller 5" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 953u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 496u16 , name : "NV1F C17 [GeForce4 MX IGP]" , subsystems : & [SubSystem { subvendor : 41120u16 , subdevice : 949u16 , name : "UK79G-1394 motherboard" }] } , Device { vendor_id : 4318u16 , id : 512u16 , name : "NV20 [GeForce3]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 16431u16 , name : "AGP-V8200 DDR" } , SubSystem { subvendor : 4168u16 , subdevice : 3184u16 , name : "GLADIAC 920" }] } , Device { vendor_id : 4318u16 , id : 513u16 , name : "NV20 [GeForce3 Ti 200]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 34051u16 , name : "G3Ti200 Pro VT128" }] } , Device { vendor_id : 4318u16 , id : 514u16 , name : "NV20 [GeForce3 Ti 500]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 16475u16 , name : "V8200 T5" } , SubSystem { subvendor : 5445u16 , subdevice : 47u16 , name : "Xtasy 6964" }] } , Device { vendor_id : 4318u16 , id : 515u16 , name : "NV20GL [Quadro DCC]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 529u16 , name : "NV48 [GeForce 6800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 530u16 , name : "NV48 [GeForce 6800 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 533u16 , name : "NV48 [GeForce 6800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 536u16 , name : "NV48 [GeForce 6800 XT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 545u16 , name : "NV44A [GeForce 6200]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33249u16 , name : "N6200/TD/256M/A" } , SubSystem { subvendor : 14402u16 , subdevice : 41793u16 , name : "256A8N341DX" }] } , Device { vendor_id : 4318u16 , id : 546u16 , name : "NV44 [GeForce 6200 A-LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 548u16 , name : "NV44" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 576u16 , name : "C51PV [GeForce 6150]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 577u16 , name : "C51 [GeForce 6150 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 578u16 , name : "C51G [GeForce 6100]" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" }] } , Device { vendor_id : 4318u16 , id : 579u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 580u16 , name : "C51 [GeForce Go 6150]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4318u16 , subdevice : 580u16 , name : "GeForce Go 6150" }] } , Device { vendor_id : 4318u16 , id : 581u16 , name : "C51 [Quadro NVS 210S/GeForce 6150LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 582u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 583u16 , name : "C51 [GeForce Go 6100]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4994u16 , name : "MCP51 PCI-X GeForce Go 6100" }] } , Device { vendor_id : 4318u16 , id : 584u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 585u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 586u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 587u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 588u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 589u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 590u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 591u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 592u16 , name : "NV25 [GeForce4 Ti 4600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 593u16 , name : "NV25 [GeForce4 Ti 4400]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32803u16 , name : "v8440 GeForce 4 Ti4400" } , SubSystem { subvendor : 4318u16 , subdevice : 593u16 , name : "PNY GeForce4 Ti 4400" } , SubSystem { subvendor : 5218u16 , subdevice : 34576u16 , name : "PNY GeForce4 Ti 4400" }] } , Device { vendor_id : 4318u16 , id : 594u16 , name : "NV25 [GeForce4 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 595u16 , name : "NV25 [GeForce4 Ti 4200]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10390u16 , name : "WinFast A250 LE TD (Dual VGA/TV-out/DVI)" } , SubSystem { subvendor : 5243u16 , subdevice : 36617u16 , name : "Siluro (Dual VGA/TV-out/DVI)" }] } , Device { vendor_id : 4318u16 , id : 600u16 , name : "NV25GL [Quadro4 900 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 601u16 , name : "NV25GL [Quadro4 750 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 603u16 , name : "NV25GL [Quadro4 700 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 608u16 , name : "MCP51 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 609u16 , name : "MCP51 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" }] } , Device { vendor_id : 4318u16 , id : 610u16 , name : "MCP51 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 611u16 , name : "MCP51 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 612u16 , name : "MCP51 SMBus" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 613u16 , name : "MCP51 IDE" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" } , SubSystem { subvendor : 61531u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" }] } , Device { vendor_id : 4318u16 , id : 614u16 , name : "MCP51 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 615u16 , name : "MCP51 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 616u16 , name : "MCP51 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 617u16 , name : "MCP51 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33089u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 618u16 , name : "MCP51 MCI" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 619u16 , name : "MCP51 AC97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" }] } , Device { vendor_id : 4318u16 , id : 620u16 , name : "MCP51 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4318u16 , subdevice : 52100u16 , name : "ASUSTeK Computer Inc. A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 621u16 , name : "MCP51 USB Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 622u16 , name : "MCP51 USB Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 623u16 , name : "MCP51 PCI Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" }] } , Device { vendor_id : 4318u16 , id : 624u16 , name : "MCP51 Host Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33212u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" } , SubSystem { subvendor : 5208u16 , subdevice : 20481u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 625u16 , name : "MCP51 PMU" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" }] } , Device { vendor_id : 4318u16 , id : 626u16 , name : "MCP51 Memory Controller 0" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4187u16 , subdevice : 3245u16 , name : "Winfast 6100K8MB" }] } , Device { vendor_id : 4318u16 , id : 638u16 , name : "C51 Memory Controller 2" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 639u16 , name : "C51 Memory Controller 3" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 640u16 , name : "NV28 [GeForce4 Ti 4800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 641u16 , name : "NV28 [GeForce4 Ti 4200 AGP 8x]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 642u16 , name : "NV28 [GeForce4 Ti 4800 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 646u16 , name : "NV28M [GeForce4 Ti 4200 Go AGP 8x]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 648u16 , name : "NV28GL [Quadro4 980 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 649u16 , name : "NV28GL [Quadro4 780 XGL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 652u16 , name : "NV28GLM [Quadro4 Go700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 656u16 , name : "G71 [GeForce 7900 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 657u16 , name : "G71 [GeForce 7900 GT/GTO]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1067u16 , name : "NX7900GTO-T2D512E [7900 GTO]" }] } , Device { vendor_id : 4318u16 , id : 658u16 , name : "G71 [GeForce 7900 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 659u16 , name : "G71 [GeForce 7900 GX2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 660u16 , name : "G71 [GeForce 7950 GX2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 661u16 , name : "G71 [GeForce 7950 GT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33317u16 , name : "GeForce 7950 GT" } , SubSystem { subvendor : 4221u16 , subdevice : 10856u16 , name : "WinFast PX7950GT TDH" } , SubSystem { subvendor : 5218u16 , subdevice : 1635u16 , name : "NX7950GT-VT2D512EZ-HD" }] } , Device { vendor_id : 4318u16 , id : 663u16 , name : "G71M [GeForce Go 7950 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 664u16 , name : "G71M [GeForce Go 7900 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 665u16 , name : "G71M [GeForce Go 7900 GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 666u16 , name : "G71GLM [Quadro FX 2500M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 667u16 , name : "G71GLM [Quadro FX 1500M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 668u16 , name : "G71GL [Quadro FX 5500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 669u16 , name : "G71GL [Quadro FX 3500]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 411u16 , name : "G71GLM [Quadro FX 3500M]" }] } , Device { vendor_id : 4318u16 , id : 670u16 , name : "G71GL [Quadro FX 1500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 671u16 , name : "G71GL [Quadro FX 4500 X2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 672u16 , name : "NV2A [XGPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 677u16 , name : "MCPX CPU Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 678u16 , name : "MCPX Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 736u16 , name : "G73 [GeForce 7600 GT AGP]" , subsystems : & [SubSystem { subvendor : 736u16 , subdevice : 8777u16 , name : "GF 7600GT 560M 256MB DDR3 DUAL DVI TV" }] } , Device { vendor_id : 4318u16 , id : 737u16 , name : "G73 [GeForce 7600 GS AGP]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 8747u16 , name : "PV-T73K-UAL3 (256MB)" } , SubSystem { subvendor : 5762u16 , subdevice : 8775u16 , name : "GF 7600GS 512MB DDR2" }] } , Device { vendor_id : 4318u16 , id : 738u16 , name : "G73 [GeForce 7300 GT AGP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 739u16 , name : "G71 [GeForce 7900 GS AGP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 740u16 , name : "G71 [GeForce 7950 GT AGP]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 8817u16 , name : "PV-T71A-YDF7 (512MB)" }] } , Device { vendor_id : 4318u16 , id : 741u16 , name : "G71 [GeForce 7600 GS AGP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 752u16 , name : "C51 Host Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 753u16 , name : "C51 Host Bridge" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" }] } , Device { vendor_id : 4318u16 , id : 754u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 755u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 756u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 757u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 758u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 759u16 , name : "C51 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 760u16 , name : "C51 Memory Controller 5" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 761u16 , name : "C51 Memory Controller 4" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 762u16 , name : "C51 Memory Controller 0" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 763u16 , name : "C51 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 764u16 , name : "C51 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" }] } , Device { vendor_id : 4318u16 , id : 765u16 , name : "C51 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" }] } , Device { vendor_id : 4318u16 , id : 766u16 , name : "C51 Memory Controller 1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 767u16 , name : "C51 Host Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10804u16 , name : "Pavilion a1677c" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4163u16 , subdevice : 33229u16 , name : "A8N-VM CSM Mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 20480u16 , name : "GA-M55plus-S3G" } , SubSystem { subvendor : 5218u16 , subdevice : 29191u16 , name : "K8NGM2 series" }] } , Device { vendor_id : 4318u16 , id : 768u16 , name : "NV30 [GeForce FX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 769u16 , name : "NV30 [GeForce FX 5800 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 770u16 , name : "NV30 [GeForce FX 5800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 776u16 , name : "NV30GL [Quadro FX 2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 777u16 , name : "NV30GL [Quadro FX 1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 785u16 , name : "NV31 [GeForce FX 5600 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 786u16 , name : "NV31 [GeForce FX 5600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 788u16 , name : "NV31 [GeForce FX 5600XT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33098u16 , name : "V9560XT/TD" }] } , Device { vendor_id : 4318u16 , id : 790u16 , name : "NV31M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 792u16 , name : "NV31GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 794u16 , name : "NV31M [GeForce FX Go5600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 795u16 , name : "NV31M [GeForce FX Go5650]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 796u16 , name : "NV31GLM [Quadro FX Go700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 800u16 , name : "NV34 [GeForce FX 5200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 801u16 , name : "NV34 [GeForce FX 5200 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 802u16 , name : "NV34 [GeForce FX 5200]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 763u16 , name : "V9250 Magic" } , SubSystem { subvendor : 4163u16 , subdevice : 33152u16 , name : "V9520-X/TD/128M" } , SubSystem { subvendor : 4221u16 , subdevice : 10599u16 , name : "WinFast A340T 128MB" } , SubSystem { subvendor : 5218u16 , subdevice : 36979u16 , name : "MS-8907 (FX5200-TDR128)" } , SubSystem { subvendor : 5218u16 , subdevice : 37136u16 , name : "MS-8911 (FX5200-TD128)" } , SubSystem { subvendor : 5218u16 , subdevice : 37233u16 , name : "MS-8917 (FX5200-T128)" } , SubSystem { subvendor : 5218u16 , subdevice : 37728u16 , name : "MS-8936 (FX5200-T128)" } , SubSystem { subvendor : 5762u16 , subdevice : 4945u16 , name : "GeForce FX 5200" }] } , Device { vendor_id : 4318u16 , id : 803u16 , name : "NV34 [GeForce FX 5200LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 804u16 , name : "NV34M [GeForce FX Go5200 64M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "Pavilion ZD7000 laptop" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" }] } , Device { vendor_id : 4318u16 , id : 805u16 , name : "NV34M [GeForce FX Go5250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 806u16 , name : "NV34 [GeForce FX 5500]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 12557u16 , name : "GeForce FX 5500 128 MB" } , SubSystem { subvendor : 5762u16 , subdevice : 8244u16 , name : "GeForce 5500 256 MB" }] } , Device { vendor_id : 4318u16 , id : 807u16 , name : "NV34 [GeForce FX 5100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 808u16 , name : "NV34M [GeForce FX Go5200 32M/64M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 809u16 , name : "NV34M [GeForce FX Go5200]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 16u16 , name : "Powerbook G4" }] } , Device { vendor_id : 4318u16 , id : 810u16 , name : "NV34GL [Quadro NVS 280 PCI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 811u16 , name : "NV34GL [Quadro FX 500/600 PCI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 812u16 , name : "NV34M [GeForce FX Go5300 / Go5350]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 813u16 , name : "NV34M [GeForce FX Go5100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 814u16 , name : "NV34" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 815u16 , name : "NV34 [GeForce FX 5200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 816u16 , name : "NV35 [GeForce FX 5900 Ultra]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33079u16 , name : "V9950 Ultra / 256 MB" }] } , Device { vendor_id : 4318u16 , id : 817u16 , name : "NV35 [GeForce FX 5900]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33093u16 , name : "V9950GE" }] } , Device { vendor_id : 4318u16 , id : 818u16 , name : "NV35 [GeForce FX 5900XT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 819u16 , name : "NV38 [GeForce FX 5950 Ultra]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 820u16 , name : "NV35 [GeForce FX 5900ZT]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 37747u16 , name : "FX5900ZT-VTD128 (MS-8937)" }] } , Device { vendor_id : 4318u16 , id : 824u16 , name : "NV35GL [Quadro FX 3000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 831u16 , name : "NV35GL [Quadro FX 700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 833u16 , name : "NV36 [GeForce FX 5700 Ultra]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 37760u16 , name : "MS-8938 (FX5700U-TD128)" }] } , Device { vendor_id : 4318u16 , id : 834u16 , name : "NV36 [GeForce FX 5700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 835u16 , name : "NV36 [GeForce FX 5700LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 836u16 , name : "NV36 [GeForce FX 5700VE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 839u16 , name : "NV36M [GeForce FX Go5700]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" }] } , Device { vendor_id : 4318u16 , id : 840u16 , name : "NV36M [GeForce FX Go5700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 844u16 , name : "NV36 [Quadro FX Go1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 845u16 , name : "NV36" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 846u16 , name : "NV36GL [Quadro FX 1100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 864u16 , name : "MCP55 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 865u16 , name : "MCP55 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 LPC Bridge" }] } , Device { vendor_id : 4318u16 , id : 866u16 , name : "MCP55 LPC Bridge" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 867u16 , name : "MCP55 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 868u16 , name : "MCP55 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 LPC Bridge" }] } , Device { vendor_id : 4318u16 , id : 869u16 , name : "MCP55 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 870u16 , name : "MCP55 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 871u16 , name : "MCP55 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 872u16 , name : "MCP55 SMBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 524u16 , name : "PowerEdge M605 MCP55 SMBus" } , SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 SMBus" } , SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 873u16 , name : "MCP55 Memory Controller" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 874u16 , name : "MCP55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 875u16 , name : "MCP55 SMU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 876u16 , name : "MCP55 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 524u16 , name : "PowerEdge M605 MCP55 USB Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 USB Controller" } , SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 877u16 , name : "MCP55 USB Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 524u16 , name : "PowerEdge M605 MCP55 USB Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 USB Controller" } , SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 878u16 , name : "MCP55 IDE" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 880u16 , name : "MCP55 PCI bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 881u16 , name : "MCP55 High Definition Audio" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 882u16 , name : "MCP55 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 883u16 , name : "MCP55 Ethernet" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 884u16 , name : "MCP55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 885u16 , name : "MCP55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 886u16 , name : "MCP55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 887u16 , name : "MCP55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 888u16 , name : "MCP55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 890u16 , name : "MCP55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 894u16 , name : "MCP55 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 895u16 , name : "MCP55 SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 MCP55 SATA Controller" } , SubSystem { subvendor : 5243u16 , subdevice : 7204u16 , name : "KN9 series mainboard" }] } , Device { vendor_id : 4318u16 , id : 907u16 , name : "G73 [GeForce 7650 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 912u16 , name : "G73 [GeForce 7650 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 913u16 , name : "G73 [GeForce 7600 GT]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 13351u16 , name : "GV-NX76T128D-RH" } , SubSystem { subvendor : 5218u16 , subdevice : 1106u16 , name : "NX7600GT-VT2D256E" }] } , Device { vendor_id : 4318u16 , id : 914u16 , name : "G73 [GeForce 7600 GS]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 1570u16 , name : "NX7600GS-T2D256EH" }] } , Device { vendor_id : 4318u16 , id : 915u16 , name : "G73 [GeForce 7300 GT]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1042u16 , name : "NX7300GT-TD256EH" } , SubSystem { subvendor : 5218u16 , subdevice : 1042u16 , name : "NX7300GT-TD256EH" }] } , Device { vendor_id : 4318u16 , id : 916u16 , name : "G73 [GeForce 7600 LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 917u16 , name : "G73 [GeForce 7300 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 918u16 , name : "G73" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 919u16 , name : "G73M [GeForce Go 7700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 920u16 , name : "G73M [GeForce Go 7600]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 108u16 , name : "Aspire 9814WKMi" }] } , Device { vendor_id : 4318u16 , id : 921u16 , name : "G73M [GeForce Go 7600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 922u16 , name : "G73M [Quadro NVS 300M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 923u16 , name : "G73M [GeForce Go 7900 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 924u16 , name : "G73GLM [Quadro FX 550M]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 924u16 , name : "Quadro FX 560M" }] } , Device { vendor_id : 4318u16 , id : 925u16 , name : "G73" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 926u16 , name : "G73GL [Quadro FX 560]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 927u16 , name : "G73" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 928u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 929u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 930u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 931u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 932u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 933u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 934u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 935u16 , name : "C55 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 936u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 937u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 938u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 939u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 940u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 941u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 942u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 943u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 944u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 945u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 946u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 947u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 948u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 949u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 950u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 951u16 , name : "C55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 952u16 , name : "C55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 953u16 , name : "C55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 954u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 955u16 , name : "C55 PCI Express bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 956u16 , name : "C55 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 976u16 , name : "C61 [GeForce 6150SE nForce 430]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" }] } , Device { vendor_id : 4318u16 , id : 977u16 , name : "C61 [GeForce 6100 nForce 405]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 978u16 , name : "C61 [GeForce 6100 nForce 400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 981u16 , name : "C61 [GeForce 6100 nForce 420]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 982u16 , name : "C61 [GeForce 7025 / nForce 630a]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 992u16 , name : "MCP61 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 6217u16 , subdevice : 992u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 993u16 , name : "MCP61 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" }] } , Device { vendor_id : 4318u16 , id : 994u16 , name : "MCP61 Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" }] } , Device { vendor_id : 4318u16 , id : 995u16 , name : "MCP61 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 996u16 , name : "MCP61 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 997u16 , name : "MCP61 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 998u16 , name : "MCP61 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 999u16 , name : "MCP61 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1000u16 , name : "MCP61 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 6217u16 , subdevice : 1000u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1001u16 , name : "MCP61 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 6217u16 , subdevice : 1001u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1002u16 , name : "MCP61 Memory Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 6217u16 , subdevice : 1002u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1003u16 , name : "MCP61 SMBus" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1003u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1004u16 , name : "MCP61 IDE" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 914u16 , name : "ET1350" } , SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1004u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1006u16 , name : "MCP61 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1007u16 , name : "MCP61 Ethernet" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 32768u16 , name : "ET1350" } , SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1007u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1008u16 , name : "MCP61 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33813u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 2184u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1009u16 , name : "MCP61 USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1009u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1010u16 , name : "MCP61 USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1010u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1011u16 , name : "MCP61 PCI bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 6217u16 , subdevice : 1011u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1012u16 , name : "MCP61 SMU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1013u16 , name : "MCP61 Memory Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1003u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1014u16 , name : "MCP61 SATA Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Inspiron 531" } , SubSystem { subvendor : 4163u16 , subdevice : 33700u16 , name : "M4N68T series motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1014u16 , name : "939NF6G-VSTA Board" }] } , Device { vendor_id : 4318u16 , id : 1015u16 , name : "MCP61 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1024u16 , name : "G84 [GeForce 8600 GTS]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33345u16 , name : "EN8600GTS" }] } , Device { vendor_id : 4318u16 , id : 1025u16 , name : "G84 [GeForce 8600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1026u16 , name : "G84 [GeForce 8600 GT]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 13397u16 , name : "GV-NX86T512H" } , SubSystem { subvendor : 5218u16 , subdevice : 2320u16 , name : "NX8600GT-T2D256EZ" }] } , Device { vendor_id : 4318u16 , id : 1027u16 , name : "G84 [GeForce 8600 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1028u16 , name : "G84 [GeForce 8400 GS]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 4656u16 , name : "NX8400GS-TD256E" }] } , Device { vendor_id : 4318u16 , id : 1029u16 , name : "G84M [GeForce 9500M GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1030u16 , name : "G84 [GeForce 8300 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1031u16 , name : "G84M [GeForce 8600M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1032u16 , name : "G84M [GeForce 9650M GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1033u16 , name : "G84M [GeForce 8700M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1034u16 , name : "G84GL [Quadro FX 370]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1035u16 , name : "G84GLM [Quadro NVS 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1036u16 , name : "G84GLM [Quadro FX 570M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8409u16 , name : "ThinkPad T61p" }] } , Device { vendor_id : 4318u16 , id : 1037u16 , name : "G84GLM [Quadro FX 1600M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1038u16 , name : "G84GL [Quadro FX 570]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1039u16 , name : "G84GL [Quadro FX 1700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1040u16 , name : "G92 [GeForce GT 330]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1044u16 , name : "G92 [GeForce 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1048u16 , name : "G92 [GeForce GT 330 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1056u16 , name : "G86 [GeForce 8400 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1057u16 , name : "G86 [GeForce 8500 GT]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 2400u16 , name : "NX8500GT-TD512EH/M2" }] } , Device { vendor_id : 4318u16 , id : 1058u16 , name : "G86 [GeForce 8400 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1059u16 , name : "G86 [GeForce 8300 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1060u16 , name : "G86 [GeForce 8400 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1061u16 , name : "G86M [GeForce 8600M GS]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4163u16 , subdevice : 5396u16 , name : "F3SV" }] } , Device { vendor_id : 4318u16 , id : 1062u16 , name : "G86M [GeForce 8400M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1063u16 , name : "G86M [GeForce 8400M GS]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1064u16 , name : "G86M [GeForce 8400M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1065u16 , name : "G86M [Quadro NVS 140M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8408u16 , name : "ThinkPad T61" }] } , Device { vendor_id : 4318u16 , id : 1066u16 , name : "G86M [Quadro NVS 130M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1067u16 , name : "G86M [Quadro NVS 135M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1068u16 , name : "G86 [GeForce 9400 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1069u16 , name : "G86GLM [Quadro FX 360M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1070u16 , name : "G86M [GeForce 9300M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1071u16 , name : "G86 [Quadro NVS 290]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1088u16 , name : "MCP65 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1089u16 , name : "MCP65 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1090u16 , name : "MCP65 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1091u16 , name : "MCP65 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1092u16 , name : "MCP65 Memory Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1093u16 , name : "MCP65 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1094u16 , name : "MCP65 SMBus" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1095u16 , name : "MCP65 SMU" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9500/9600/9700 series" }] } , Device { vendor_id : 4318u16 , id : 1096u16 , name : "MCP65 IDE" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1097u16 , name : "MCP65 PCI bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 52100u16 , name : "HP Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1098u16 , name : "MCP65 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1099u16 , name : "MCP65 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1100u16 , name : "MCP65 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1101u16 , name : "MCP65 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1102u16 , name : "MCP65 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1103u16 , name : "MCP65 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1104u16 , name : "MCP65 Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1105u16 , name : "MCP65 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1106u16 , name : "MCP65 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1107u16 , name : "MCP65 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1108u16 , name : "MCP65 USB 1.1 OHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1109u16 , name : "MCP65 USB 2.0 EHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1110u16 , name : "MCP65 USB Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1111u16 , name : "MCP65 USB Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1112u16 , name : "MCP65 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP65 PCI Express bridge" }] } , Device { vendor_id : 4318u16 , id : 1113u16 , name : "MCP65 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP65 PCI Express bridge" }] } , Device { vendor_id : 4318u16 , id : 1114u16 , name : "MCP65 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP65 PCI Express bridge" }] } , Device { vendor_id : 4318u16 , id : 1115u16 , name : "MCP65 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP65 PCI Express bridge" }] } , Device { vendor_id : 4318u16 , id : 1116u16 , name : "MCP65 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1117u16 , name : "MCP65 SATA Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" }] } , Device { vendor_id : 4318u16 , id : 1118u16 , name : "MCP65 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1119u16 , name : "MCP65 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1329u16 , name : "C67 [GeForce 7150M / nForce 630M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1331u16 , name : "C67 [GeForce 7000M / nForce 610M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1338u16 , name : "C68 [GeForce 7050 PV / nForce 630a]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1339u16 , name : "C68 [GeForce 7050 PV / nForce 630a]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1342u16 , name : "C68 [GeForce 7025 / nForce 630a]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1345u16 , name : "MCP67 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1346u16 , name : "MCP67 SMBus" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1347u16 , name : "MCP67 Co-processor" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1351u16 , name : "MCP67 Memory Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1351u16 , name : "ALiveNF7G-HDready" }] } , Device { vendor_id : 4318u16 , id : 1352u16 , name : "MCP67 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1356u16 , name : "MCP67 Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 1356u16 , name : "ALiveNF7G-HDready, MCP67 Gigabit Ethernet" }] } , Device { vendor_id : 4318u16 , id : 1357u16 , name : "MCP67 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1358u16 , name : "MCP67 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1359u16 , name : "MCP67 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1360u16 , name : "MCP67 AHCI Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1364u16 , name : "MCP67 AHCI Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1365u16 , name : "MCP67 SATA Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1372u16 , name : "MCP67 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33424u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1373u16 , name : "MCP67 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1374u16 , name : "MCP67 OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1375u16 , name : "MCP67 EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1376u16 , name : "MCP67 IDE Controller" , subsystems : & [SubSystem { subvendor : 61507u16 , subdevice : 33544u16 , name : "M2N68-AM Motherboard" }] } , Device { vendor_id : 4318u16 , id : 1377u16 , name : "MCP67 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1378u16 , name : "MCP67 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 1378u16 , name : "ALiveNF7G-HDready" }] } , Device { vendor_id : 4318u16 , id : 1379u16 , name : "MCP67 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1384u16 , name : "MCP78S [GeForce 8200] Memory Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1384u16 , name : "K10N78FullHD-hSLI R3.0 Memory Controller" }] } , Device { vendor_id : 4318u16 , id : 1385u16 , name : "MCP78S [GeForce 8200] PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1385u16 , name : "K10N78FullHD-hSLI R3.0 PCI Express Bridge" }] } , Device { vendor_id : 4318u16 , id : 1386u16 , name : "MCP73 [nForce 630i] USB 2.0 Controller (EHCI)" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 1388u16 , name : "MCP73 IDE Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1389u16 , name : "MCP73 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 4318u16 , subdevice : 52083u16 , name : "MCP73 PCI Bridge" }] } , Device { vendor_id : 4318u16 , id : 1390u16 , name : "MCP73 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP73 PCIe x16 port" }] } , Device { vendor_id : 4318u16 , id : 1391u16 , name : "MCP73 PCI Express bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 4318u16 , subdevice : 0u16 , name : "MCP73 PCIe x1 port" }] } , Device { vendor_id : 4318u16 , id : 1457u16 , name : "NF200 PCIe 2.0 switch" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1464u16 , name : "NF200 PCIe 2.0 switch for GTX 295" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1470u16 , name : "NF200 PCIe 2.0 switch for Quadro Plex S4 / Tesla S870 / Tesla S1070 / Tesla S2050" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1504u16 , name : "GT200b [GeForce GTX 295]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1505u16 , name : "GT200 [GeForce GTX 280]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1506u16 , name : "GT200 [GeForce GTX 260]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1507u16 , name : "GT200b [GeForce GTX 285]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 9360u16 , name : "GX-285N-ZDF" }] } , Device { vendor_id : 4318u16 , id : 1510u16 , name : "GT200b [GeForce GTX 275]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1511u16 , name : "GT200GL [Tesla C1060 / M1060]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1429u16 , name : "Tesla T10 Processor" } , SubSystem { subvendor : 4318u16 , subdevice : 1679u16 , name : "Tesla T10 Processor" } , SubSystem { subvendor : 4318u16 , subdevice : 1687u16 , name : "Tesla M1060" } , SubSystem { subvendor : 4318u16 , subdevice : 1812u16 , name : "Tesla M1060" } , SubSystem { subvendor : 4318u16 , subdevice : 1859u16 , name : "Tesla M1060" }] } , Device { vendor_id : 4318u16 , id : 1514u16 , name : "GT200 [GeForce GTX 260]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1515u16 , name : "GT200 [GeForce GTX 295]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1517u16 , name : "GT200GL [Quadro Plex 2200 D2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1521u16 , name : "GT200 [GeForce GTX 280]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1522u16 , name : "GT200 [GeForce GTX 260]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1528u16 , name : "GT200GL [Quadro Plex 2200 S4]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1529u16 , name : "GT200GL [Quadro CX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1533u16 , name : "GT200GL [Quadro FX 5800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1534u16 , name : "GT200GL [Quadro FX 4800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1535u16 , name : "GT200GL [Quadro FX 3800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1536u16 , name : "G92 [GeForce 8800 GTS 512]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1537u16 , name : "G92 [GeForce 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1538u16 , name : "G92 [GeForce 8800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1539u16 , name : "G92 [GeForce GT 230 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1540u16 , name : "G92 [GeForce 9800 GX2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1541u16 , name : "G92 [GeForce 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1542u16 , name : "G92 [GeForce 8800 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1543u16 , name : "G92 [GeForce GTS 240]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1544u16 , name : "G92M [GeForce 9800M GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1545u16 , name : "G92M [GeForce 8800M GTS]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 167u16 , name : "GeForce 8800 GS" }] } , Device { vendor_id : 4318u16 , id : 1546u16 , name : "G92M [GeForce GTX 280M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1547u16 , name : "G92M [GeForce 9800M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1548u16 , name : "G92M [GeForce 8800M GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1549u16 , name : "G92 [GeForce 8800 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1551u16 , name : "G92M [GeForce GTX 285M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1552u16 , name : "G92 [GeForce 9600 GSO]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 9093u16 , name : "GeForce 9600 GSO 768mb" }] } , Device { vendor_id : 4318u16 , id : 1553u16 , name : "G92 [GeForce 8800 GT]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10928u16 , name : "Winfast PX8800 GT PCI-E" } , SubSystem { subvendor : 5218u16 , subdevice : 4464u16 , name : "NX8800GT series model V117 2xDVI+TV" } , SubSystem { subvendor : 6618u16 , subdevice : 4160u16 , name : "ZT-88TES2P-FSP" }] } , Device { vendor_id : 4318u16 , id : 1554u16 , name : "G92 [GeForce 9800 GTX / 9800 GTX+]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1555u16 , name : "G92 [GeForce 9800 GTX+]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1556u16 , name : "G92 [GeForce 9800 GT]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10931u16 , name : "WinFast PX9800 GT (S-Fanpipe)" }] } , Device { vendor_id : 4318u16 , id : 1557u16 , name : "G92 [GeForce GTS 250]" , subsystems : & [SubSystem { subvendor : 14402u16 , subdevice : 4432u16 , name : "GeForce GTS 250 P/N 512-P3-1150-TR" } , SubSystem { subvendor : 14402u16 , subdevice : 4433u16 , name : "GeForce GTS 250 P/N 512-P3-1151-TR" } , SubSystem { subvendor : 14402u16 , subdevice : 4437u16 , name : "GeForce GTS 250 P/N 01G-P3-1155-TR" } , SubSystem { subvendor : 14402u16 , subdevice : 4438u16 , name : "GeForce GTS 250 P/N 01G-P3-1156-TR" }] } , Device { vendor_id : 4318u16 , id : 1559u16 , name : "G92M [GeForce 9800M GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1560u16 , name : "G92M [GeForce GTX 260M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1561u16 , name : "G92GL [Quadro FX 4700 X2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1562u16 , name : "G92GL [Quadro FX 3700]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1563u16 , name : "G92GL [Quadro VX 200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1564u16 , name : "G92GLM [Quadro FX 3600M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1565u16 , name : "G92GLM [Quadro FX 2800M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1566u16 , name : "G92GLM [Quadro FX 3700M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1567u16 , name : "G92GLM [Quadro FX 3800M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1568u16 , name : "G94 [GeForce 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1569u16 , name : "G94 [GeForce GT 230]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1570u16 , name : "G94 [GeForce 9600 GT]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10945u16 , name : "WinFast PX9600GT 1024MB" } , SubSystem { subvendor : 5208u16 , subdevice : 13441u16 , name : "GV-NX96T512HP" }] } , Device { vendor_id : 4318u16 , id : 1571u16 , name : "G94 [GeForce 9600 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1572u16 , name : "G94 [GeForce 9600 GT Green Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1573u16 , name : "G94 [GeForce 9600 GSO 512]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1574u16 , name : "G94 [GeForce GT 130]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1575u16 , name : "G94 [GeForce GT 140]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1576u16 , name : "G94M [GeForce 9800M GTS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1578u16 , name : "G94M [GeForce 9700M GTS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1579u16 , name : "G94M [GeForce 9800M GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1580u16 , name : "G94M [GeForce 9800M GTS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1581u16 , name : "G94 [GeForce 9600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1582u16 , name : "G94 [GeForce 9600 GT]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 1541u16 , name : "GeForce GT 130" }] } , Device { vendor_id : 4318u16 , id : 1583u16 , name : "G94 [GeForce 9800 S]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1584u16 , name : "G94 [GeForce 9600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1585u16 , name : "G94M [GeForce GTS 160M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1586u16 , name : "G94M [GeForce GTS 150M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1587u16 , name : "G94 [GeForce GT 220]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1589u16 , name : "G94 [GeForce 9600 GSO]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1591u16 , name : "G94 [GeForce 9600 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1592u16 , name : "G94GL [Quadro FX 1800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1594u16 , name : "G94GLM [Quadro FX 2700M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1599u16 , name : "G94 [GeForce 9600 GE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1600u16 , name : "G96C [GeForce 9500 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1601u16 , name : "G96C [GeForce 9400 GT]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 16393u16 , name : "PV-T94G-ZAFG" }] } , Device { vendor_id : 4318u16 , id : 1602u16 , name : "G96 [D9M-10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1603u16 , name : "G96 [GeForce 9500 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1604u16 , name : "G96 [GeForce 9500 GS]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 38400u16 , name : "Geforce 9500GS 512M DDR2 V/D/HDMI" }] } , Device { vendor_id : 4318u16 , id : 1605u16 , name : "G96C [GeForce 9500 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1606u16 , name : "G96C [GeForce GT 120]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1607u16 , name : "G96CM [GeForce 9600M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1608u16 , name : "G96CM [GeForce 9600M GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1609u16 , name : "G96CM [GeForce 9600M GT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 8237u16 , name : "GeForce GT 220M" }] } , Device { vendor_id : 4318u16 , id : 1610u16 , name : "G96M [GeForce 9700M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1611u16 , name : "G96M [GeForce 9500M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1612u16 , name : "G96CM [GeForce 9650M GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1614u16 , name : "G96C [GeForce 9600 GSO / 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1617u16 , name : "G96CM [GeForce G 110M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1618u16 , name : "G96CM [GeForce GT 130M]" , subsystems : & [SubSystem { subvendor : 5421u16 , subdevice : 2128u16 , name : "GeForce GT 240M LE" }] } , Device { vendor_id : 4318u16 , id : 1619u16 , name : "G96CM [GeForce GT 120M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1620u16 , name : "G96CM [GeForce GT 220M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5282u16 , name : "GeForce GT 320M" } , SubSystem { subvendor : 4163u16 , subdevice : 5330u16 , name : "GeForce GT 320M" }] } , Device { vendor_id : 4318u16 , id : 1621u16 , name : "G96 [GeForce GT 120 Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1622u16 , name : "G96 [GeForce GT 120 Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1624u16 , name : "G96GL [Quadro FX 380]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1625u16 , name : "G96CGL [Quadro FX 580]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1626u16 , name : "G96GLM [Quadro FX 1700M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1627u16 , name : "G96C [GeForce 9400 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1628u16 , name : "G96GLM [Quadro FX 770M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1629u16 , name : "G96 [GeForce 9500 GA / 9600 GT / GTS 250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1631u16 , name : "G96C [GeForce G210]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1728u16 , name : "GF100 [GeForce GTX 480]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1732u16 , name : "GF100 [GeForce GTX 465]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1738u16 , name : "GF100M [GeForce GTX 480M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1739u16 , name : "GF100 [GeForce GTX 480]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1741u16 , name : "GF100 [GeForce GTX 470]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1744u16 , name : "GF100GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1745u16 , name : "GF100GL [Tesla C2050 / C2070]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1905u16 , name : "Tesla C2050" } , SubSystem { subvendor : 4318u16 , subdevice : 1906u16 , name : "Tesla C2070" }] } , Device { vendor_id : 4318u16 , id : 1746u16 , name : "GF100GL [Tesla M2070]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1908u16 , name : "Tesla M2070" } , SubSystem { subvendor : 4318u16 , subdevice : 2096u16 , name : "Tesla M2070" } , SubSystem { subvendor : 4318u16 , subdevice : 2114u16 , name : "Tesla M2070" } , SubSystem { subvendor : 4318u16 , subdevice : 2191u16 , name : "Tesla X2070" } , SubSystem { subvendor : 4318u16 , subdevice : 2312u16 , name : "Tesla M2070" }] } , Device { vendor_id : 4318u16 , id : 1752u16 , name : "GF100GL [Quadro 6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1753u16 , name : "GF100GL [Quadro 5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1754u16 , name : "GF100GLM [Quadro 5000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1756u16 , name : "GF100GL [Quadro 6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1757u16 , name : "GF100GL [Quadro 4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1758u16 , name : "GF100GL [Tesla T20 Processor]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1907u16 , name : "Tesla S2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2095u16 , name : "Tesla M2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2112u16 , name : "Tesla X2070" } , SubSystem { subvendor : 4318u16 , subdevice : 2114u16 , name : "Tesla M2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2118u16 , name : "Tesla M2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2150u16 , name : "Tesla M2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2311u16 , name : "Tesla M2050" } , SubSystem { subvendor : 4318u16 , subdevice : 2334u16 , name : "Tesla M2050" }] } , Device { vendor_id : 4318u16 , id : 1759u16 , name : "GF100GL [Tesla M2070-Q]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2125u16 , name : "Tesla M2070-Q" } , SubSystem { subvendor : 4318u16 , subdevice : 2175u16 , name : "Tesla M2070-Q" }] } , Device { vendor_id : 4318u16 , id : 1760u16 , name : "G98 [GeForce 9300 GE]" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 23190u16 , name : "Geforce 9300GE" }] } , Device { vendor_id : 4318u16 , id : 1761u16 , name : "G98 [GeForce 9300 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1762u16 , name : "G98 [GeForce 8400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1763u16 , name : "G98 [GeForce 8300 GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1764u16 , name : "G98 [GeForce 8400 GS Rev. 2]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 13429u16 , name : "GV-NX84S256HE [GeForce 8400 GS]" }] } , Device { vendor_id : 4318u16 , id : 1765u16 , name : "G98M [GeForce 9300M GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1766u16 , name : "G98 [GeForce G 100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1767u16 , name : "G98 [GeForce 9300 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1768u16 , name : "G98M [GeForce 9200M GS]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13835u16 , name : "GeForce 9200M GE" }] } , Device { vendor_id : 4318u16 , id : 1769u16 , name : "G98M [GeForce 9300M GS]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6578u16 , name : "U6V laptop" }] } , Device { vendor_id : 4318u16 , id : 1770u16 , name : "G98M [Quadro NVS 150M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1771u16 , name : "G98M [Quadro NVS 160M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1772u16 , name : "G98M [GeForce G 105M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1773u16 , name : "G98 [GeForce 9600 GT / 9800 GT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1774u16 , name : "G98 [GeForce 9600 GT / 9800 GT / GT 240]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1775u16 , name : "G98M [GeForce G 103M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1777u16 , name : "G98M [GeForce G 105M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1784u16 , name : "G98 [Quadro NVS 420]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1785u16 , name : "G98GL [Quadro FX 370 LP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1786u16 , name : "G98 [Quadro NVS 450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1787u16 , name : "G98GLM [Quadro FX 370M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1789u16 , name : "G98 [Quadro NVS 295]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1791u16 , name : "G98 [HICx16 + Graphics]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 1809u16 , name : "HICx8 + Graphics" }] } , Device { vendor_id : 4318u16 , id : 1873u16 , name : "MCP78S [GeForce 8200] Memory Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1873u16 , name : "K10N78FullHD-hSLI R3.0 Memory Controller" }] } , Device { vendor_id : 4318u16 , id : 1874u16 , name : "MCP78S [GeForce 8200] SMBus" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1874u16 , name : "K10N78FullHD-hSLI R3.0 SMBus" }] } , Device { vendor_id : 4318u16 , id : 1875u16 , name : "MCP78S [GeForce 8200] Co-Processor" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1875u16 , name : "K10N78FullHD-hSLI R3.0 Co-Processor" }] } , Device { vendor_id : 4318u16 , id : 1876u16 , name : "MCP78S [GeForce 8200] Memory Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1876u16 , name : "K10N78FullHD-hSLI R3.0 Memory Controller" }] } , Device { vendor_id : 4318u16 , id : 1881u16 , name : "MCP78S [GeForce 8200] IDE" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1881u16 , name : "K10N78FullHD-hSLI R3.0 IDE" }] } , Device { vendor_id : 4318u16 , id : 1882u16 , name : "MCP78S [GeForce 8200] PCI Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 6217u16 , subdevice : 1882u16 , name : "K10N78FullHD-hSLI R3.0 PCI Bridge" }] } , Device { vendor_id : 4318u16 , id : 1883u16 , name : "MCP78S [GeForce 8200] PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1883u16 , name : "K10N78FullHD-hSLI R3.0 PCI Express Bridge" }] } , Device { vendor_id : 4318u16 , id : 1884u16 , name : "MCP78S [GeForce 8200] LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1884u16 , name : "K10N78FullHD-hSLI R3.0 LPC Bridge" }] } , Device { vendor_id : 4318u16 , id : 1885u16 , name : "MCP78S [GeForce 8200] LPC Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" }] } , Device { vendor_id : 4318u16 , id : 1888u16 , name : "MCP77 Ethernet" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1888u16 , name : "K10N78FullHD-hSLI R3.0 Ethernet" }] } , Device { vendor_id : 4318u16 , id : 1889u16 , name : "MCP77 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1890u16 , name : "MCP77 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1891u16 , name : "MCP77 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1908u16 , name : "MCP72XE/MCP72P/MCP78U/MCP78S High Definition Audio" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33534u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 13922u16 , name : "K10N78FullHD-hSLI R3.0 High Definition Audio" }] } , Device { vendor_id : 4318u16 , id : 1912u16 , name : "MCP78S [GeForce 8200] PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1912u16 , name : "K10N78FullHD-hSLI R3.0 PCI Express Bridge" }] } , Device { vendor_id : 4318u16 , id : 1914u16 , name : "MCP78S [GeForce 8200] PCI Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1914u16 , name : "K10N78FullHD-hSLI R3.0 PCI Bridge" }] } , Device { vendor_id : 4318u16 , id : 1915u16 , name : "MCP78S [GeForce 8200] OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1915u16 , name : "K10N78FullHD-hSLI R3.0 OHCI USB 1.1 Controller" }] } , Device { vendor_id : 4318u16 , id : 1916u16 , name : "MCP78S [GeForce 8200] EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1916u16 , name : "K10N78FullHD-hSLI R3.0 EHCI USB 2.0 Controller" }] } , Device { vendor_id : 4318u16 , id : 1917u16 , name : "MCP78S [GeForce 8200] OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1917u16 , name : "K10N78FullHD-hSLI R3.0 OHCI USB 1.1 Controller" }] } , Device { vendor_id : 4318u16 , id : 1918u16 , name : "MCP78S [GeForce 8200] EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 1918u16 , name : "K10N78FullHD-hSLI R3.0 EHCI USB 2.0 Controller" }] } , Device { vendor_id : 4318u16 , id : 1984u16 , name : "MCP73 Host Bridge" , subsystems : & [SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1985u16 , name : "MCP73 Host Bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" }] } , Device { vendor_id : 4318u16 , id : 1986u16 , name : "MCP73 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1987u16 , name : "MCP73 Host Bridge" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 1989u16 , name : "MCP73 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 1992u16 , name : "MCP73 Memory Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1995u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1997u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1998u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 1999u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2000u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2001u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2002u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2003u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2006u16 , name : "nForce 610i/630i memory controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2007u16 , name : "MCP73 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2008u16 , name : "MCP73 SMBus" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2009u16 , name : "MCP73 Memory Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2010u16 , name : "MCP73 Co-processor" , subsystems : & [SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2012u16 , name : "MCP73 Ethernet" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2013u16 , name : "MCP73 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2014u16 , name : "MCP73 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2015u16 , name : "MCP73 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2016u16 , name : "C73 [GeForce 7150 / nForce 630i]" , subsystems : & [SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2017u16 , name : "C73 [GeForce 7100 / nForce 630i]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" }] } , Device { vendor_id : 4318u16 , id : 2018u16 , name : "C73 [GeForce 7050 / nForce 630i]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2019u16 , name : "C73 [GeForce 7050 / nForce 610i]" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2021u16 , name : "C73 [GeForce 7100 / nForce 620i]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2032u16 , name : "MCP73 SATA Controller (IDE mode)" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2036u16 , name : "GeForce 7100/nForce 630i SATA" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2040u16 , name : "MCP73 SATA RAID Controller" , subsystems : & [SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2044u16 , name : "MCP73 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 4318u16 , subdevice : 2044u16 , name : "MCP73 High Definition Audio" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" }] } , Device { vendor_id : 4318u16 , id : 2046u16 , name : "MCP73 OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 10618u16 , name : "MCP73PVT-SM" } , SubSystem { subvendor : 5243u16 , subdevice : 7230u16 , name : "I-N73V motherboard" } , SubSystem { subvendor : 6906u16 , subdevice : 29008u16 , name : "JW-IN7150-HD" }] } , Device { vendor_id : 4318u16 , id : 2112u16 , name : "C77 [GeForce 8200M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2116u16 , name : "C77 [GeForce 9100M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2117u16 , name : "C77 [GeForce 8200M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2118u16 , name : "C77 [GeForce 9200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2119u16 , name : "C78 [GeForce 9100]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" }] } , Device { vendor_id : 4318u16 , id : 2120u16 , name : "C77 [GeForce 8300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2121u16 , name : "C77 [GeForce 8200]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 2121u16 , name : "K10N78FullHD-hSLI R3.0 GeForce 8200" }] } , Device { vendor_id : 4318u16 , id : 2122u16 , name : "C77 [nForce 730a]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2123u16 , name : "C77 [GeForce 8200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2124u16 , name : "C77 [nForce 780a/980a SLI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2125u16 , name : "C77 [nForce 750a SLI]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D mGPU" }] } , Device { vendor_id : 4318u16 , id : 2127u16 , name : "C77 [GeForce 8100 / nForce 720a]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2144u16 , name : "C79 [GeForce 9300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2145u16 , name : "C79 [GeForce 9400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2146u16 , name : "C79 [GeForce 9400M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2147u16 , name : "C79 [GeForce 9400M]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 170u16 , name : "MacBook5,1" }] } , Device { vendor_id : 4318u16 , id : 2148u16 , name : "C79 [GeForce 9300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2149u16 , name : "C79 [GeForce 9300 / ION]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2150u16 , name : "C79 [GeForce 9400M G]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 177u16 , name : "GeForce 9400M" }] } , Device { vendor_id : 4318u16 , id : 2151u16 , name : "C79 [GeForce 9400]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 173u16 , name : "iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2152u16 , name : "C79 [nForce 760i SLI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2153u16 , name : "MCP7A [GeForce 9400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2154u16 , name : "C79 [GeForce 9400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2156u16 , name : "C79 [GeForce 9300 / nForce 730i]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2157u16 , name : "C79 [GeForce 9200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2158u16 , name : "C79 [GeForce 9100M G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2159u16 , name : "MCP79 [GeForce 8200M G]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5810u16 , name : "F5GL Notebook" }] } , Device { vendor_id : 4318u16 , id : 2160u16 , name : "C79 [GeForce 9400M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2161u16 , name : "C79 [GeForce 9200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2162u16 , name : "C79 [GeForce G102M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6580u16 , name : "GeForce G102M" } , SubSystem { subvendor : 4163u16 , subdevice : 6818u16 , name : "GeForce G102M" } , SubSystem { subvendor : 4163u16 , subdevice : 7170u16 , name : "GeForce G102M" } , SubSystem { subvendor : 4163u16 , subdevice : 7234u16 , name : "GeForce G205M" }] } , Device { vendor_id : 4318u16 , id : 2163u16 , name : "C79 [GeForce G102M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6580u16 , name : "GeForce G102M" } , SubSystem { subvendor : 4163u16 , subdevice : 7186u16 , name : "GeForce G102M" } , SubSystem { subvendor : 4163u16 , subdevice : 7250u16 , name : "GeForce G205M" }] } , Device { vendor_id : 4318u16 , id : 2164u16 , name : "C79 [ION]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2166u16 , name : "C79 [GeForce 9400M / ION]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2170u16 , name : "C79 [GeForce 9400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2173u16 , name : "C79 [ION]" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2174u16 , name : "C79 [ION LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2175u16 , name : "C79 [ION LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2208u16 , name : "MCP89 [GeForce 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2210u16 , name : "MCP89 [GeForce 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2211u16 , name : "MCP89 [GeForce 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2212u16 , name : "MCP89 [GeForce 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2213u16 , name : "MCP89 [GeForce 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2592u16 , name : "GT216 [GeForce GT 220]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33553u16 , name : "ENGT220/DI/1GD3(LP)/V2" }] } , Device { vendor_id : 4318u16 , id : 2593u16 , name : "GT216M [GeForce GT 330M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2594u16 , name : "GT216 [GeForce 315]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2595u16 , name : "GT216 [GeForce 210]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2596u16 , name : "GT216 [GeForce 405]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2598u16 , name : "GT216 [GeForce 405]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2599u16 , name : "GT216 [GeForce 405]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2600u16 , name : "GT216M [GeForce GT 230M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2601u16 , name : "GT216M [GeForce GT 330M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2602u16 , name : "GT216M [GeForce GT 230M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2603u16 , name : "GT216M [GeForce GT 330M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2604u16 , name : "GT216M [NVS 5100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2605u16 , name : "GT216M [GeForce GT 320M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2608u16 , name : "GT216 [GeForce 505]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2610u16 , name : "GT216 [GeForce GT 415]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2612u16 , name : "GT216M [GeForce GT 240M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2613u16 , name : "GT216M [GeForce GT 325M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2616u16 , name : "GT216GL [Quadro 400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2620u16 , name : "GT216GLM [Quadro FX 880M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2656u16 , name : "GT218 [GeForce G210]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2658u16 , name : "GT218 [GeForce 205]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2659u16 , name : "GT218 [GeForce 310]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2660u16 , name : "GT218 [ION]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2661u16 , name : "GT218 [GeForce 210]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33588u16 , name : "EN210 SILENT" } , SubSystem { subvendor : 5208u16 , subdevice : 13993u16 , name : "GV-N210D3-1GI (rev. 6.0/6.1)" } , SubSystem { subvendor : 5218u16 , subdevice : 32916u16 , name : "N210 [Geforce 210] PCIe graphics adapter" } , SubSystem { subvendor : 6618u16 , subdevice : 29218u16 , name : "GeForce 210 1GB [Synergy Edition]" }] } , Device { vendor_id : 4318u16 , id : 2662u16 , name : "GT218 [GeForce 310]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2663u16 , name : "GT218 [GeForce 315]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2664u16 , name : "GT218M [GeForce G 105M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2665u16 , name : "GT218M [GeForce G 105M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2666u16 , name : "GT218M [NVS 2100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2668u16 , name : "GT218M [NVS 3100M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 6058u16 , subdevice : 8514u16 , name : "ThinkPad T410" }] } , Device { vendor_id : 4318u16 , id : 2670u16 , name : "GT218M [GeForce 305M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2671u16 , name : "GT218M [ION]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2672u16 , name : "GT218M [GeForce 310M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2673u16 , name : "GT218M [GeForce 305M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2674u16 , name : "GT218M [GeForce 310M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2675u16 , name : "GT218M [GeForce 305M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2676u16 , name : "GT218M [GeForce G210M]" , subsystems : & [SubSystem { subvendor : 6922u16 , subdevice : 36922u16 , name : "GeForce G210" }] } , Device { vendor_id : 4318u16 , id : 2677u16 , name : "GT218M [GeForce 310M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2678u16 , name : "GT218M [ION 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2680u16 , name : "GT218GL [Quadro FX 380 LP]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2682u16 , name : "GT218M [GeForce 315M]" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 36990u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64592u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64609u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64625u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64656u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64704u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64720u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64738u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64754u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64790u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64832u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64848u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64850u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64865u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64881u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64914u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64918u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64976u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 64978u16 , name : "GeForce 315M" } , SubSystem { subvendor : 4473u16 , subdevice : 65022u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 49314u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 49330u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 50561u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 50567u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 50568u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 50583u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5197u16 , subdevice : 50694u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5218u16 , subdevice : 43601u16 , name : "GeForce 405" } , SubSystem { subvendor : 5218u16 , subdevice : 43608u16 , name : "GeForce 405" } , SubSystem { subvendor : 5218u16 , subdevice : 44145u16 , name : "GeForce 405" } , SubSystem { subvendor : 5218u16 , subdevice : 44161u16 , name : "GeForce 315M" } , SubSystem { subvendor : 5218u16 , subdevice : 44162u16 , name : "GeForce 405" } , SubSystem { subvendor : 5218u16 , subdevice : 44595u16 , name : "GeForce 405" } , SubSystem { subvendor : 5698u16 , subdevice : 14720u16 , name : "GeForce 405" } , SubSystem { subvendor : 6058u16 , subdevice : 14672u16 , name : "GeForce 405M" } , SubSystem { subvendor : 6058u16 , subdevice : 14717u16 , name : "GeForce 405M" } , SubSystem { subvendor : 6922u16 , subdevice : 8337u16 , name : "GeForce 315M" } , SubSystem { subvendor : 6922u16 , subdevice : 37044u16 , name : "GeForce 405" } , SubSystem { subvendor : 7165u16 , subdevice : 3u16 , name : "GeForce 405" } , SubSystem { subvendor : 7165u16 , subdevice : 32774u16 , name : "GeForce 405" } , SubSystem { subvendor : 7165u16 , subdevice : 32775u16 , name : "GeForce 315M" }] } , Device { vendor_id : 4318u16 , id : 2683u16 , name : "GT218 [GeForce 505]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2684u16 , name : "GT218GLM [Quadro FX 380M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2688u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2689u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2690u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2691u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2692u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2693u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2694u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2695u16 , name : "MCP79 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2696u16 , name : "MCP79 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2697u16 , name : "MCP79 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2712u16 , name : "MCP79 Memory Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2720u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2722u16 , name : "MCP79 SMBus" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2723u16 , name : "MCP79 Co-processor" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2724u16 , name : "MCP79 Memory Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2725u16 , name : "MCP79 OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2726u16 , name : "MCP79 EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2727u16 , name : "MCP79 OHCI USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2728u16 , name : "MCP79 OHCI USB 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2729u16 , name : "MCP79 EHCI USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2730u16 , name : "MCP79 EHCI USB 2.0 Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2731u16 , name : "MCP79 PCI Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2732u16 , name : "MCP79 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2733u16 , name : "MCP79 LPC Bridge" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2734u16 , name : "MCP79 LPC Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2735u16 , name : "MCP79 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2736u16 , name : "MCP79 Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4629u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" } , SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2737u16 , name : "MCP79 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2738u16 , name : "MCP79 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2739u16 , name : "MCP79 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2740u16 , name : "MCP79 SATA Controller" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 41251u16 , name : "IONITX-F-E" }] } , Device { vendor_id : 4318u16 , id : 2741u16 , name : "MCP79 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2742u16 , name : "MCP79 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2743u16 , name : "MCP79 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2744u16 , name : "MCP79 AHCI Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" }] } , Device { vendor_id : 4318u16 , id : 2745u16 , name : "MCP79 AHCI Controller" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2746u16 , name : "MCP79 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2747u16 , name : "MCP79 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2748u16 , name : "MCP79 RAID Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2749u16 , name : "MCP79 RAID Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2750u16 , name : "MCP79 RAID Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2751u16 , name : "MCP79 RAID Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2752u16 , name : "MCP79 High Definition Audio" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6403u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2753u16 , name : "MCP79 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2754u16 , name : "MCP79 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2755u16 , name : "MCP79 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2756u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2757u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2758u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2759u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 6791u16 , name : "F5GL Notebook" } , SubSystem { subvendor : 4318u16 , subdevice : 52089u16 , name : "Apple iMac 9,1" }] } , Device { vendor_id : 4318u16 , id : 2760u16 , name : "MCP79 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 2768u16 , name : "MCP78S [GeForce 8200] SATA Controller (non-AHCI mode)" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29960u16 , name : "K9N2GM-FIH" } , SubSystem { subvendor : 6217u16 , subdevice : 2768u16 , name : "K10N78FullHD-hSLI R3.0 IDE" }] } , Device { vendor_id : 4318u16 , id : 2772u16 , name : "MCP78S [GeForce 8200] AHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10910u16 , name : "Pavilion p6310f" } , SubSystem { subvendor : 4163u16 , subdevice : 33512u16 , name : "M3N72-D" } , SubSystem { subvendor : 6217u16 , subdevice : 2772u16 , name : "K10N78FullHD-hSLI R3.0 AHCI Controller" }] } , Device { vendor_id : 4318u16 , id : 2776u16 , name : "MCP78S [GeForce 8200] SATA Controller (RAID mode)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3042u16 , name : "GT216 HDMI Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33553u16 , name : "ENGT220/DI/1GD3(LP)/V2" }] } , Device { vendor_id : 4318u16 , id : 3043u16 , name : "High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 4318u16 , subdevice : 1645u16 , name : "G98 [GeForce 8400GS]" } , SubSystem { subvendor : 5218u16 , subdevice : 32916u16 , name : "N210 [Geforce 210] PCIe graphics adapter" }] } , Device { vendor_id : 4318u16 , id : 3044u16 , name : "High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3045u16 , name : "GF100 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3049u16 , name : "GF106 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 5464u16 , subdevice : 34439u16 , name : "CLEVO/KAPOK W860CU" } , SubSystem { subvendor : 14402u16 , subdevice : 5202u16 , name : "GeForce GTS 450" }] } , Device { vendor_id : 4318u16 , id : 3050u16 , name : "GF108 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 14402u16 , subdevice : 5168u16 , name : "GeForce GT 430" }] } , Device { vendor_id : 4318u16 , id : 3051u16 , name : "GF104 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 8994u16 , name : "N460GTX Cyclone 1GD5/OC" }] } , Device { vendor_id : 4318u16 , id : 3054u16 , name : "GF116 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3056u16 , name : "Tegra2 PCIe x4 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3057u16 , name : "Tegra2 PCIe x2 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3232u16 , name : "GT215 [GeForce GT 330]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3234u16 , name : "GT215 [GeForce GT 320]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3235u16 , name : "GT215 [GeForce GT 240]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3236u16 , name : "GT215 [GeForce GT 340]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3237u16 , name : "GT215 [GeForce GT 220]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3239u16 , name : "GT215 [GeForce GT 330]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3240u16 , name : "GT215M [GeForce GTS 260M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3241u16 , name : "GT215M [GeForce GTS 250M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3244u16 , name : "GT215 [GeForce GT 220/315]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3247u16 , name : "GT215M [GeForce GT 335M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3248u16 , name : "GT215M [GeForce GTS 350M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3249u16 , name : "GT215M [GeForce GTS 360M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3260u16 , name : "GT215GLM [Quadro FX 1800M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3424u16 , name : "MCP89 HOST Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3432u16 , name : "MCP89 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3433u16 , name : "MCP89 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3446u16 , name : "MCP89 PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3449u16 , name : "MCP89 SMBus" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3450u16 , name : "MCP89 Co-Processor" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3451u16 , name : "MCP89 Memory Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3453u16 , name : "MCP89 Ethernet" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3456u16 , name : "MCP89 LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3461u16 , name : "MCP89 SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3464u16 , name : "MCP89 SATA Controller (AHCI mode)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3465u16 , name : "MCP89 SATA Controller (AHCI mode)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3469u16 , name : "MCP89 SATA Controller (RAID mode)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3476u16 , name : "MCP89 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3484u16 , name : "MCP89 OHCI USB 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3485u16 , name : "MCP89 EHCI USB 2.0 Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3520u16 , name : "GF106 [GeForce GT 440]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3524u16 , name : "GF106 [GeForce GTS 450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3525u16 , name : "GF106 [GeForce GTS 450 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3526u16 , name : "GF106 [GeForce GTS 450 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3533u16 , name : "GF106M [GeForce GT 555M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3534u16 , name : "GF106M [GeForce GT 555M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3537u16 , name : "GF106M [GeForce GTX 460M]" , subsystems : & [SubSystem { subvendor : 5464u16 , subdevice : 34439u16 , name : "CLEVO/KAPOK W860CU" }] } , Device { vendor_id : 4318u16 , id : 3538u16 , name : "GF106M [GeForce GT 445M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3539u16 , name : "GF106M [GeForce GT 435M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3542u16 , name : "GF106M [GeForce GT 550M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3544u16 , name : "GF106GL [Quadro 2000]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2324u16 , name : "Quadro 2000D" }] } , Device { vendor_id : 4318u16 , id : 3546u16 , name : "GF106GLM [Quadro 2000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3552u16 , name : "GF108 [GeForce GT 440]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3553u16 , name : "GF108 [GeForce GT 430]" , subsystems : & [SubSystem { subvendor : 14402u16 , subdevice : 5168u16 , name : "GeForce GT 430" }] } , Device { vendor_id : 4318u16 , id : 3554u16 , name : "GF108 [GeForce GT 420]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3555u16 , name : "GF108M [GeForce GT 635M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3556u16 , name : "GF108 [GeForce GT 520]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3557u16 , name : "GF108 [GeForce GT 530]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3559u16 , name : "GF108 [GeForce GT 610]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3560u16 , name : "GF108M [GeForce GT 620M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3561u16 , name : "GF108M [GeForce GT 620M/630M/635M/640M LE]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 1682u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1829u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1832u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1835u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1838u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1875u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1876u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 6058u16 , subdevice : 14711u16 , name : "GeForce GT 640M LE" } , SubSystem { subvendor : 6922u16 , subdevice : 8390u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 6922u16 , subdevice : 8720u16 , name : "GeForce GT 635M" }] } , Device { vendor_id : 4318u16 , id : 3562u16 , name : "GF108M [GeForce 610M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13914u16 , name : "GeForce 615" } , SubSystem { subvendor : 6058u16 , subdevice : 13915u16 , name : "GeForce 615" } , SubSystem { subvendor : 6058u16 , subdevice : 13918u16 , name : "GeForce 615" } , SubSystem { subvendor : 6058u16 , subdevice : 13920u16 , name : "GeForce 615" } , SubSystem { subvendor : 6058u16 , subdevice : 13932u16 , name : "GeForce 615" }] } , Device { vendor_id : 4318u16 , id : 3563u16 , name : "GF108M [GeForce GT 555M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3564u16 , name : "GF108M [GeForce GT 525M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3565u16 , name : "GF108M [GeForce GT 520M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3566u16 , name : "GF108M [GeForce GT 415M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3567u16 , name : "GF108M [NVS 5400M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3568u16 , name : "GF108M [GeForce GT 425M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3569u16 , name : "GF108M [GeForce GT 420M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3570u16 , name : "GF108M [GeForce GT 435M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3571u16 , name : "GF108M [GeForce GT 420M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3572u16 , name : "GF108M [GeForce GT 540M]" , subsystems : & [SubSystem { subvendor : 5421u16 , subdevice : 2386u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 5421u16 , subdevice : 2387u16 , name : "GeForce GT 630M" }] } , Device { vendor_id : 4318u16 , id : 3573u16 , name : "GF108M [GeForce GT 525M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3574u16 , name : "GF108M [GeForce GT 550M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3575u16 , name : "GF108M [GeForce GT 520M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3576u16 , name : "GF108GL [Quadro 600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3577u16 , name : "GF108GLM [Quadro 500M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3578u16 , name : "GF108GLM [Quadro 1000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3580u16 , name : "GF108GLM [NVS 5200M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3592u16 , name : "GF119 HDMI Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33696u16 , name : "ENGT520 SILENT" } , SubSystem { subvendor : 4272u16 , subdevice : 4170u16 , name : "Gainward GeForce GT 610" }] } , Device { vendor_id : 4318u16 , id : 3593u16 , name : "GF110 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3594u16 , name : "GK104 HDMI Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3595u16 , name : "GK106 HDMI Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3596u16 , name : "GF114 HDMI Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3599u16 , name : "GK208 HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3602u16 , name : "TegraK1 PCIe x4 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3603u16 , name : "TegraK1 PCIe x1 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3610u16 , name : "GK110 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3611u16 , name : "GK107 HDMI Audio Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6523u16 , name : "ZBook 15" } , SubSystem { subvendor : 4163u16 , subdevice : 33832u16 , name : "GTX650-DC-1GD5" }] } , Device { vendor_id : 4318u16 , id : 3612u16 , name : "Tegra3+ PCIe x4 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3613u16 , name : "Tegra3+ PCIe x2 Bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3618u16 , name : "GF104 [GeForce GTX 460]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 8994u16 , name : "N460GTX Cyclone 1GD5/OC" }] } , Device { vendor_id : 4318u16 , id : 3619u16 , name : "GF104 [GeForce GTX 460 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3620u16 , name : "GF104 [GeForce GTX 460 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3632u16 , name : "GF104M [GeForce GTX 470M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3633u16 , name : "GF104M [GeForce GTX 485M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3642u16 , name : "GF104GLM [Quadro 3000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3643u16 , name : "GF104GLM [Quadro 4000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3840u16 , name : "GF108 [GeForce GT 630]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3841u16 , name : "GF108 [GeForce GT 620]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3842u16 , name : "GF108 [GeForce GT 730]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3843u16 , name : "GF108 [GeForce GT 610]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 3846u16 , name : "GF108 [GeForce GT 730]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4000u16 , name : "GK11x [GK11x_FPGA]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4005u16 , name : "GK11x" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4007u16 , name : "GK11x [Tegra on x86 (PEATRANS)]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4014u16 , name : "Tegra X1 PCIe x4 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4015u16 , name : "Tegra X1 PCIe x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4016u16 , name : "GM200 High Definition Audio" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4024u16 , name : "GP108 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4025u16 , name : "GP107GL High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4026u16 , name : "GM206 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4027u16 , name : "GM204 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4028u16 , name : "GM107 High Definition Audio Controller [GeForce 940MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4032u16 , name : "GK107 [GeForce GT 640 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4033u16 , name : "GK107 [GeForce GT 640]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4034u16 , name : "GK107 [GeForce GT 630 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4036u16 , name : "GK107 [D14P1-15]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4037u16 , name : "GK107 [GeForce GT 1030]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4038u16 , name : "GK107 [GeForce GTX 650]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33832u16 , name : "GTX650-DC-1GD5" }] } , Device { vendor_id : 4318u16 , id : 4040u16 , name : "GK107 [GeForce GT 740]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4041u16 , name : "GK107 [GeForce GT 730]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4043u16 , name : "GK107 [EXK107]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4044u16 , name : "GK107 [GeForce GT 720]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4045u16 , name : "GK107M [GeForce GT 755M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4046u16 , name : "GK107M [GeForce GT 640M LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4047u16 , name : "GK107 [GEN3 ESI]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4048u16 , name : "GK107 [NB1G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4049u16 , name : "GK107M [GeForce GT 650M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5527u16 , name : "GeForce GT 650M" } , SubSystem { subvendor : 4163u16 , subdevice : 5543u16 , name : "GeForce GT 650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8451u16 , name : "N56VZ" } , SubSystem { subvendor : 4163u16 , subdevice : 8453u16 , name : "GeForce GT 650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8513u16 , name : "GeForce GT 650M" }] } , Device { vendor_id : 4318u16 , id : 4050u16 , name : "GK107M [GeForce GT 640M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1359u16 , name : "GeForce GT 640M" } , SubSystem { subvendor : 4136u16 , subdevice : 1375u16 , name : "GeForce GT 640M" } , SubSystem { subvendor : 4136u16 , subdevice : 1429u16 , name : "GeForce GT 640M LE" } , SubSystem { subvendor : 4136u16 , subdevice : 1458u16 , name : "GeForce GT 640M LE" }] } , Device { vendor_id : 4318u16 , id : 4051u16 , name : "GK107M [GeForce GT 640M LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4052u16 , name : "GK107M [GeForce GTX 660M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4053u16 , name : "GK107M [GeForce GT 650M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4054u16 , name : "GK107M [N13P-GS-W]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4055u16 , name : "GK107 [GK107-GTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4056u16 , name : "GK107M [GeForce GT 640M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4057u16 , name : "GK107M [GeForce GT 645M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4058u16 , name : "GK107 [GK107-ES-A1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4059u16 , name : "GK107 [GK107-ESP-A1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4060u16 , name : "GK107 [GK107-INT22-A1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4061u16 , name : "GK107 [GK107-INT11-A1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4062u16 , name : "GK107 [GK107-ES-KA-E1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4063u16 , name : "GK107M [GeForce GT 740M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4064u16 , name : "GK107M [GeForce GTX 660M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4065u16 , name : "GK107M [GeForce GT 730M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4066u16 , name : "GK107M [GeForce GT 745M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4067u16 , name : "GK107M [GeForce GT 745M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 11030u16 , name : "GeForce GT 745A" } , SubSystem { subvendor : 6058u16 , subdevice : 13941u16 , name : "GeForce GT 745A" }] } , Device { vendor_id : 4318u16 , id : 4068u16 , name : "GK107M [GeForce GT 750M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4069u16 , name : "GK107 [GeForce K340 USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4070u16 , name : "GK107 [GRID K1 NVS USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4071u16 , name : "GK107GL [GRID K100 vGPU]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4126u16 , name : "GRID K100" }] } , Device { vendor_id : 4318u16 , id : 4072u16 , name : "GK107M [N14P-GS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4073u16 , name : "GK107M [GeForce GT 750M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4074u16 , name : "GK107M [GeForce GT 755M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4076u16 , name : "GK107M [GeForce 710A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4077u16 , name : "GK107M [GeForce 820M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4078u16 , name : "GK107M [GeForce 810M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4079u16 , name : "GK107GL [GRID K340]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4080u16 , name : "GK107 [NB1Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4081u16 , name : "GK107 [NVS 1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4082u16 , name : "GK107GL [GRID K1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4083u16 , name : "GK107GL [Quadro K420]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4085u16 , name : "GK107GL [GRID K1 Tesla USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4086u16 , name : "GK107GLM [Quadro K1100M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6523u16 , name : "ZBook 15" }] } , Device { vendor_id : 4318u16 , id : 4087u16 , name : "GK107GL [GRID K140Q vGPU]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4151u16 , name : "GRID K140Q" }] } , Device { vendor_id : 4318u16 , id : 4088u16 , name : "GK107GLM [Quadro K500M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4089u16 , name : "GK107GL [Quadro K2000D]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4090u16 , name : "GK107GL [Quadro K600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4091u16 , name : "GK107GLM [Quadro K2000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4092u16 , name : "GK107GLM [Quadro K1000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4093u16 , name : "GK107 [NVS 510]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4094u16 , name : "GK107GL [Quadro K2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4095u16 , name : "GK107GL [Quadro 410]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4097u16 , name : "GK110B [GeForce GTX TITAN Z]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4099u16 , name : "GK110 [GeForce GTX Titan LE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4100u16 , name : "GK110 [GeForce GTX 780]" , subsystems : & [SubSystem { subvendor : 14402u16 , subdevice : 1924u16 , name : "GK110B [GeForce GTX 780 SC w/ ACX Cooler]" } , SubSystem { subvendor : 14402u16 , subdevice : 6020u16 , name : "GK110B [GeForce GTX 780 Dual FTW w/ ACX Cooler]" } , SubSystem { subvendor : 14402u16 , subdevice : 6024u16 , name : "GK110B [GeForce GTX 780 Dual Classified w/ ACX Cooler]" }] } , Device { vendor_id : 4318u16 , id : 4101u16 , name : "GK110 [GeForce GTX TITAN]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33873u16 , name : "GTXTITAN-6GD5" } , SubSystem { subvendor : 4318u16 , subdevice : 4149u16 , name : "GeForce GTX Titan" } , SubSystem { subvendor : 14402u16 , subdevice : 10128u16 , name : "GeForce GTX Titan" } , SubSystem { subvendor : 14402u16 , subdevice : 10129u16 , name : "GeForce GTX Titan SC" } , SubSystem { subvendor : 14402u16 , subdevice : 10131u16 , name : "GeForce GTX Titan SC Signature" } , SubSystem { subvendor : 14402u16 , subdevice : 10132u16 , name : "GeForce GTX Titan SC Hydro Copper" } , SubSystem { subvendor : 14402u16 , subdevice : 10133u16 , name : "GeForce GTX Titan SC Hydro Copper Signature" }] } , Device { vendor_id : 4318u16 , id : 4103u16 , name : "GK110 [GeForce GTX 780 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4104u16 , name : "GK110 [GeForce GTX 780 Ti 6GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4106u16 , name : "GK110B [GeForce GTX 780 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4108u16 , name : "GK110B [GeForce GTX TITAN Black]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4126u16 , name : "GK110GL [Tesla K20X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4127u16 , name : "GK110GL [Tesla K20]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4128u16 , name : "GK110GL [Tesla K20X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4129u16 , name : "GK110GL [Tesla K20Xm]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4130u16 , name : "GK110GL [Tesla K20c]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4131u16 , name : "GK110BGL [Tesla K40m]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2430u16 , name : "12GB Computational Accelerator" }] } , Device { vendor_id : 4318u16 , id : 4132u16 , name : "GK180GL [Tesla K40c]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4134u16 , name : "GK110GL [Tesla K20s]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4135u16 , name : "GK110BGL [Tesla K40st]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4136u16 , name : "GK110GL [Tesla K20m]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4137u16 , name : "GK110BGL [Tesla K40s]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4138u16 , name : "GK110BGL [Tesla K40t]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4141u16 , name : "GK210GL [Tesla K80]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4142u16 , name : "GK110BGL [Tesla K40d]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4143u16 , name : "GK110BGL [Tesla Stella Solo]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4154u16 , name : "GK110GL [Quadro K6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4156u16 , name : "GK110GL [Quadro K5200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4159u16 , name : "GK110BGL [Tesla Stella SXM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4160u16 , name : "GF119 [GeForce GT 520]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33696u16 , name : "ENGT520 SILENT" }] } , Device { vendor_id : 4318u16 , id : 4162u16 , name : "GF119 [GeForce 510]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4165u16 , name : "GF119" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4168u16 , name : "GF119 [GeForce 605]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4169u16 , name : "GF119 [GeForce GT 620 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4170u16 , name : "GF119 [GeForce GT 610]" , subsystems : & [SubSystem { subvendor : 4272u16 , subdevice : 4170u16 , name : "Gainward GeForce GT 610" }] } , Device { vendor_id : 4318u16 , id : 4171u16 , name : "GF119 [GeForce GT 625 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4172u16 , name : "GF119 [GeForce GT 705]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4173u16 , name : "GF119 [GeForce GT 710]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4176u16 , name : "GF119M [GeForce GT 520M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4177u16 , name : "GF119M [GeForce GT 520MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4178u16 , name : "GF119M [GeForce GT 520M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4180u16 , name : "GF119M [GeForce 410M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4181u16 , name : "GF119M [GeForce 410M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4182u16 , name : "GF119M [NVS 4200M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4183u16 , name : "GF119M [Quadro NVS 4200M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4184u16 , name : "GF119M [GeForce 610M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10989u16 , name : "GeForce 610" } , SubSystem { subvendor : 4156u16 , subdevice : 10993u16 , name : "GeForce 610" } , SubSystem { subvendor : 4163u16 , subdevice : 4268u16 , name : "GeForce GT 610M" } , SubSystem { subvendor : 4163u16 , subdevice : 4284u16 , name : "GeForce GT 610M" } , SubSystem { subvendor : 4163u16 , subdevice : 5714u16 , name : "GeForce GT 610M" } , SubSystem { subvendor : 6058u16 , subdevice : 13946u16 , name : "GeForce 610M" } , SubSystem { subvendor : 6058u16 , subdevice : 13954u16 , name : "GeForce 800A" } , SubSystem { subvendor : 6058u16 , subdevice : 13959u16 , name : "GeForce 800A" } , SubSystem { subvendor : 6058u16 , subdevice : 13970u16 , name : "GeForce 705A" } , SubSystem { subvendor : 6058u16 , subdevice : 13973u16 , name : "GeForce 800A" } , SubSystem { subvendor : 6058u16 , subdevice : 41239u16 , name : "GeForce 610M" }] } , Device { vendor_id : 4318u16 , id : 4185u16 , name : "GF119M [GeForce 610M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4186u16 , name : "GF119M [GeForce 610M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 8465u16 , name : "GeForce GT 610M" } , SubSystem { subvendor : 4163u16 , subdevice : 8466u16 , name : "GeForce GT 610M" }] } , Device { vendor_id : 4318u16 , id : 4187u16 , name : "GF119M [GeForce 705M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 11003u16 , name : "GeForce 705A" } , SubSystem { subvendor : 6058u16 , subdevice : 12445u16 , name : "GeForce 705A" } , SubSystem { subvendor : 6058u16 , subdevice : 12465u16 , name : "GeForce 800A" } , SubSystem { subvendor : 6058u16 , subdevice : 12531u16 , name : "GeForce 705A" } , SubSystem { subvendor : 6058u16 , subdevice : 13985u16 , name : "GeForce 800A" }] } , Device { vendor_id : 4318u16 , id : 4220u16 , name : "GF119 [NVS 315]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4221u16 , name : "GF119 [NVS 310]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4224u16 , name : "GF110 [GeForce GTX 580]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4225u16 , name : "GF110 [GeForce GTX 570]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2174u16 , name : "Leadtek WinFast GTX 570" }] } , Device { vendor_id : 4318u16 , id : 4226u16 , name : "GF110 [GeForce GTX 560 Ti OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4228u16 , name : "GF110 [GeForce GTX 560 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4230u16 , name : "GF110 [GeForce GTX 570 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4231u16 , name : "GF110 [GeForce GTX 560 Ti 448 Cores]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4232u16 , name : "GF110 [GeForce GTX 590]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4233u16 , name : "GF110 [GeForce GTX 580 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4235u16 , name : "GF110 [GeForce GTX 580]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4238u16 , name : "GF110GL [Tesla C2090]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4241u16 , name : "GF110GL [Tesla M2090]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2190u16 , name : "Tesla X2090" } , SubSystem { subvendor : 4318u16 , subdevice : 2193u16 , name : "Tesla X2090" } , SubSystem { subvendor : 4318u16 , subdevice : 2420u16 , name : "Tesla X2090" } , SubSystem { subvendor : 4318u16 , subdevice : 2445u16 , name : "Tesla X2090" }] } , Device { vendor_id : 4318u16 , id : 4244u16 , name : "GF110GL [Tesla M2075]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2184u16 , name : "Tesla M2075" }] } , Device { vendor_id : 4318u16 , id : 4246u16 , name : "GF110GL [Tesla C2050 / C2075]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2320u16 , name : "Tesla C2075" } , SubSystem { subvendor : 4318u16 , subdevice : 2321u16 , name : "Tesla C2050" }] } , Device { vendor_id : 4318u16 , id : 4250u16 , name : "GF100GLM [Quadro 5010M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4251u16 , name : "GF100GL [Quadro 7000]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 2328u16 , name : "Quadro 7000" }] } , Device { vendor_id : 4318u16 , id : 4288u16 , name : "GT218 [GeForce 9300 GS Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4291u16 , name : "GT218 [GeForce 8400 GS Rev. 3]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4293u16 , name : "GT218 [GeForce 405]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4312u16 , name : "GT218 [NVS 300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4325u16 , name : "Parker PCIe x4 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4326u16 , name : "Parker PCIe x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4335u16 , name : "GP102 HDMI Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4336u16 , name : "GP104 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4337u16 , name : "GP106 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34230u16 , name : "DUAL-GTX1060-O6G [GeForce GTX 1060 6GB Dual]" }] } , Device { vendor_id : 4318u16 , id : 4343u16 , name : "TU102 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4344u16 , name : "TU104 HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4345u16 , name : "TU106 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34419u16 , name : "TURBO-RTX2070-8G" }] } , Device { vendor_id : 4318u16 , id : 4416u16 , name : "GF117M [GeForce 610M/710M/810M/820M / GT 620M/625M/630M/720M]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 1945u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4121u16 , subdevice : 39327u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1536u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1542u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1610u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1612u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1658u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1664u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1670u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1673u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1675u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1677u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1678u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1681u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1682u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1684u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1794u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1817u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1829u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1832u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1835u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1838u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1842u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4133u16 , subdevice : 1891u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1907u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1908u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1910u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1914u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1915u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1916u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1917u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1918u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1919u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 1921u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1944u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1945u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1947u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 1948u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2055u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2081u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2083u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2096u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2099u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2103u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4133u16 , subdevice : 2110u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2113u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4133u16 , subdevice : 2132u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2133u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2134u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2135u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2136u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2147u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2152u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2153u16 , name : "GeForce 810M" } , SubSystem { subvendor : 4133u16 , subdevice : 2163u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2168u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2171u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2172u16 , name : "GeForce 810M" } , SubSystem { subvendor : 4133u16 , subdevice : 2177u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2186u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2203u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2319u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2337u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2350u16 , name : "GeForce 810M" } , SubSystem { subvendor : 4133u16 , subdevice : 2351u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2354u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2362u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2364u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2367u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2369u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2373u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2388u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4133u16 , subdevice : 2405u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1357u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1358u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1364u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4136u16 , subdevice : 1367u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4136u16 , subdevice : 1378u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1381u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1384u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1424u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1426u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1428u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1429u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1442u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1457u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1459u16 , name : "GeForce GT 625M" } , SubSystem { subvendor : 4136u16 , subdevice : 1498u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1502u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4136u16 , subdevice : 1504u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4136u16 , subdevice : 1512u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1524u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4136u16 , subdevice : 1551u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4136u16 , subdevice : 1614u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1618u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1619u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1621u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1630u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1634u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1677u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1709u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1710u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1711u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1712u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1728u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4136u16 , subdevice : 1729u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 6383u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4156u16 , subdevice : 6393u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4156u16 , subdevice : 6395u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4156u16 , subdevice : 6397u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4156u16 , subdevice : 6399u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 4156u16 , subdevice : 8586u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8635u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8636u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8718u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8720u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8722u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8724u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8728u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8795u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8797u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8813u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8815u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8914u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 8921u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 9013u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 9015u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4156u16 , subdevice : 10991u16 , name : "GeForce GT 720A" } , SubSystem { subvendor : 4156u16 , subdevice : 11001u16 , name : "GeForce 710A" } , SubSystem { subvendor : 4163u16 , subdevice : 4317u16 , name : "NVS 5200M" } , SubSystem { subvendor : 4163u16 , subdevice : 4333u16 , name : "NVS 5200M" } , SubSystem { subvendor : 4163u16 , subdevice : 4605u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 4685u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 4717u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 4893u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 5117u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 5319u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 5383u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4163u16 , subdevice : 5549u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5613u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5645u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5693u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5741u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5837u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5853u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5901u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 5997u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 6029u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 6045u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 6109u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 8498u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 4163u16 , subdevice : 8502u16 , name : "NVS 5200M" } , SubSystem { subvendor : 4163u16 , subdevice : 8634u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 8698u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 8714u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 8730u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 8762u16 , name : "GeForce GT 710M" } , SubSystem { subvendor : 4163u16 , subdevice : 8778u16 , name : "GeForce GT 710M" } , SubSystem { subvendor : 4163u16 , subdevice : 8826u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 8842u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 9002u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 9018u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 9050u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 9066u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 9098u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34197u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 34282u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 34283u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34284u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34286u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4163u16 , subdevice : 34291u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34318u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34330u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34331u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34344u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34371u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34380u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34386u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34400u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4163u16 , subdevice : 34401u16 , name : "GeForce 820M" } , SubSystem { subvendor : 4187u16 , subdevice : 3500u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4187u16 , subdevice : 3501u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4187u16 , subdevice : 3827u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4210u16 , subdevice : 5421u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4303u16 , subdevice : 6133u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 4473u16 , subdevice : 64001u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64002u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64003u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64005u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64017u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64019u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64024u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64025u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64033u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64035u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64042u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64050u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64051u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64054u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64056u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64066u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64067u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64069u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64071u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64073u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64088u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64089u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64136u16 , name : "GeForce 710M" } , SubSystem { subvendor : 4473u16 , subdevice : 64137u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5197u16 , subdevice : 45202u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 5197u16 , subdevice : 49365u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 5197u16 , subdevice : 49367u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 5197u16 , subdevice : 49378u16 , name : "NVS 5200M" } , SubSystem { subvendor : 5197u16 , subdevice : 49379u16 , name : "NVS 5200M" } , SubSystem { subvendor : 5197u16 , subdevice : 49380u16 , name : "NVS 5200M" } , SubSystem { subvendor : 5197u16 , subdevice : 49421u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "GeForce GT 620M on NP300E5C series laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 50953u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5197u16 , subdevice : 50961u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5197u16 , subdevice : 50998u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5197u16 , subdevice : 50999u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5197u16 , subdevice : 51013u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5197u16 , subdevice : 51024u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5218u16 , subdevice : 4280u16 , name : "GeForce GT 710M" } , SubSystem { subvendor : 5218u16 , subdevice : 4329u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 4374u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5218u16 , subdevice : 43571u16 , name : "GeForce 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 43682u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 43683u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5218u16 , subdevice : 44210u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 44225u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 44641u16 , name : "GeForce 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 44645u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5218u16 , subdevice : 44650u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5218u16 , subdevice : 44657u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5312u16 , subdevice : 131u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5421u16 , subdevice : 2342u16 , name : "GeForce 620M" } , SubSystem { subvendor : 5421u16 , subdevice : 2434u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 5421u16 , subdevice : 2435u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 5421u16 , subdevice : 4101u16 , name : "GeForce GT 820M" } , SubSystem { subvendor : 5421u16 , subdevice : 4114u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5421u16 , subdevice : 4121u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5421u16 , subdevice : 4144u16 , name : "GeForce GT 630M" } , SubSystem { subvendor : 5421u16 , subdevice : 4181u16 , name : "GeForce 710M" } , SubSystem { subvendor : 5421u16 , subdevice : 4199u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5421u16 , subdevice : 4210u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 5421u16 , subdevice : 4230u16 , name : "GeForce 820M" } , SubSystem { subvendor : 5421u16 , subdevice : 4242u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 8704u16 , name : "NVS 5200M" } , SubSystem { subvendor : 6058u16 , subdevice : 8723u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 8736u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 12444u16 , name : "GeForce GT 720A" } , SubSystem { subvendor : 6058u16 , subdevice : 12468u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 12471u16 , name : "GeForce 720A" } , SubSystem { subvendor : 6058u16 , subdevice : 12516u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13851u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13852u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13853u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13910u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 6058u16 , subdevice : 13914u16 , name : "GeForce 705M" } , SubSystem { subvendor : 6058u16 , subdevice : 13918u16 , name : "GeForce 800M" } , SubSystem { subvendor : 6058u16 , subdevice : 13921u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13932u16 , name : "GeForce 800M" } , SubSystem { subvendor : 6058u16 , subdevice : 13957u16 , name : "GeForce 800M" } , SubSystem { subvendor : 6058u16 , subdevice : 13958u16 , name : "GeForce 800M" } , SubSystem { subvendor : 6058u16 , subdevice : 13959u16 , name : "GeForce 705A" } , SubSystem { subvendor : 6058u16 , subdevice : 13974u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13979u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13980u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13981u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13982u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13993u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 13999u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 14000u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 14006u16 , name : "GeForce 820A" } , SubSystem { subvendor : 6058u16 , subdevice : 14336u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14337u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14338u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14339u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14340u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14342u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14344u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14349u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14350u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14351u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14353u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14354u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14355u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14358u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14360u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14362u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14364u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14593u16 , name : "GeForce 610M / GT 620M" } , SubSystem { subvendor : 6058u16 , subdevice : 14594u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 14595u16 , name : "GeForce 610M/710M" } , SubSystem { subvendor : 6058u16 , subdevice : 14596u16 , name : "GeForce GT 620M/625M" } , SubSystem { subvendor : 6058u16 , subdevice : 14597u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14599u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14608u16 , name : "GeForce 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14610u16 , name : "GeForce 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14611u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14613u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 14711u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 14723u16 , name : "GeForce 610M" } , SubSystem { subvendor : 6058u16 , subdevice : 20481u16 , name : "GeForce 610M" } , SubSystem { subvendor : 6058u16 , subdevice : 20483u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 20485u16 , name : "GeForce 705M" } , SubSystem { subvendor : 6058u16 , subdevice : 20493u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 6058u16 , subdevice : 20500u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20503u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20505u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20506u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20511u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 20517u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20519u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20522u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20523u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 20525u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20526u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 20527u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6058u16 , subdevice : 20528u16 , name : "GeForce 705M" } , SubSystem { subvendor : 6058u16 , subdevice : 20529u16 , name : "GeForce 705M" } , SubSystem { subvendor : 6058u16 , subdevice : 20530u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 20531u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 20542u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6058u16 , subdevice : 20543u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6058u16 , subdevice : 20544u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6228u16 , subdevice : 375u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6228u16 , subdevice : 384u16 , name : "GeForce 710M" } , SubSystem { subvendor : 6228u16 , subdevice : 400u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6228u16 , subdevice : 402u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6228u16 , subdevice : 548u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6922u16 , subdevice : 448u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6922u16 , subdevice : 8413u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 6922u16 , subdevice : 8415u16 , name : "GeForce GT 620M" } , SubSystem { subvendor : 6922u16 , subdevice : 8462u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6922u16 , subdevice : 8706u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 6922u16 , subdevice : 37079u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6922u16 , subdevice : 37085u16 , name : "GeForce 820M" } , SubSystem { subvendor : 6992u16 , subdevice : 21808u16 , name : "GeForce 820M" } , SubSystem { subvendor : 7020u16 , subdevice : 21809u16 , name : "GeForce GT 720M" } , SubSystem { subvendor : 7083u16 , subdevice : 262u16 , name : "GeForce 820M" } , SubSystem { subvendor : 7429u16 , subdevice : 4115u16 , name : "GeForce 810M" }] } , Device { vendor_id : 4318u16 , id : 4480u16 , name : "GK104 [GeForce GTX 680]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33777u16 , name : "GTX680-DC2-2GD5" } , SubSystem { subvendor : 14402u16 , subdevice : 13954u16 , name : "GeForce GTX 680 Mac Edition" }] } , Device { vendor_id : 4318u16 , id : 4482u16 , name : "GK104 [GeForce GTX 760 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4483u16 , name : "GK104 [GeForce GTX 660 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4484u16 , name : "GK104 [GeForce GTX 770]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4485u16 , name : "GK104 [GeForce GTX 660 OEM]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4207u16 , name : "GK104 [GeForce GTX 760 OEM]" }] } , Device { vendor_id : 4318u16 , id : 4486u16 , name : "GK104 [GeForce GTX 660 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4487u16 , name : "GK104 [GeForce GTX 760]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4488u16 , name : "GK104 [GeForce GTX 690]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4489u16 , name : "GK104 [GeForce GTX 670]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4212u16 , name : "GK104 [GeForce GTX 760 Ti OEM]" }] } , Device { vendor_id : 4318u16 , id : 4490u16 , name : "GK104GL [GRID K520]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4491u16 , name : "GK104GL [GRID K2 GeForce USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4492u16 , name : "GK104 [GRID K2 NVS USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4493u16 , name : "GK104GL [GRID K200 vGPU]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4125u16 , name : "GRID K200" }] } , Device { vendor_id : 4318u16 , id : 4494u16 , name : "GK104 [GeForce GTX 760 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4495u16 , name : "GK104GL [Tesla K10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4497u16 , name : "GK104 [GeForce GTX 760 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4499u16 , name : "GK104 [GeForce GTX 760 Ti OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4500u16 , name : "GK104GL [Tesla K8]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4501u16 , name : "GK104 [GeForce GTX 660 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4504u16 , name : "GK104M [GeForce GTX 880M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4505u16 , name : "GK104M [GeForce GTX 870M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4506u16 , name : "GK104M [GeForce GTX 860M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4509u16 , name : "GK104M [GeForce GTX 775M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4510u16 , name : "GK104M [GeForce GTX 780M Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4511u16 , name : "GK104M [GeForce GTX 780M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4512u16 , name : "GK104M [GeForce GTX 680M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4513u16 , name : "GK104M [GeForce GTX 670MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4514u16 , name : "GK104M [GeForce GTX 675MX Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4515u16 , name : "GK104M [GeForce GTX 680MX]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 269u16 , name : "iMac 13,2" }] } , Device { vendor_id : 4318u16 , id : 4516u16 , name : "GK104 [GK104-ESA]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4517u16 , name : "GK104 [GK104-ESA]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4519u16 , name : "GK104M [GeForce GTX 675MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4520u16 , name : "GK104GLM [Quadro K5100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4521u16 , name : "GK104M [GeForce GTX 870M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4522u16 , name : "GK104 [GK104-INT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4524u16 , name : "GK104 [GK104-CS]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4527u16 , name : "GK104GLM [GRID IceCube]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4528u16 , name : "GK104GL [GRID K240Q / K260Q vGPU]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4122u16 , name : "GRID K240Q" } , SubSystem { subvendor : 4318u16 , subdevice : 4123u16 , name : "GRID K260Q" }] } , Device { vendor_id : 4318u16 , id : 4529u16 , name : "GK104GL [GRID K2 Tesla USM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4532u16 , name : "GK104GL [Quadro K4200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4534u16 , name : "GK104GLM [Quadro K3100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4535u16 , name : "GK104GLM [Quadro K4100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4536u16 , name : "GK104GLM [Quadro K5100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4537u16 , name : "GK104GLM" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4538u16 , name : "GK104GL [Quadro K5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4539u16 , name : "GK104GL [Quadro 4100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4540u16 , name : "GK104GLM [Quadro K5000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4541u16 , name : "GK104GLM [Quadro K4000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4542u16 , name : "GK104GLM [Quadro K3000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4543u16 , name : "GK104GL [GRID K2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4544u16 , name : "GK106 [GeForce GTX 660]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4545u16 , name : "GK106 [D14P2-30]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4546u16 , name : "GK106 [GeForce GTX 650 Ti Boost]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33883u16 , name : "GeForce GTX 650 Ti Boost DirectCU II OC" } , SubSystem { subvendor : 5218u16 , subdevice : 10356u16 , name : "GeForce GTX 650 Ti Boost TwinFrozr II OC" } , SubSystem { subvendor : 5481u16 , subdevice : 4546u16 , name : "GeForce GTX 650 Ti Boost OC" } , SubSystem { subvendor : 6618u16 , subdevice : 4737u16 , name : "GeForce GTX 650 Ti Boost OC" } , SubSystem { subvendor : 14402u16 , subdevice : 13911u16 , name : "GeForce GTX 650 Ti Boost" } , SubSystem { subvendor : 14402u16 , subdevice : 13912u16 , name : "GeForce GTX 650 Ti Boost Superclocked" }] } , Device { vendor_id : 4318u16 , id : 4547u16 , name : "GK106 [GeForce GTX 650 Ti OEM]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4144u16 , name : "GeForce GTX 650 Ti OEM" }] } , Device { vendor_id : 4318u16 , id : 4548u16 , name : "GK106 [GeForce GTX 645 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4549u16 , name : "GK106 [GeForce GT 740]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4550u16 , name : "GK106 [GeForce GTX 650 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4551u16 , name : "GK106 [GeForce GTX 750 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4552u16 , name : "GK106 [GeForce GTX 650 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4555u16 , name : "GK106 [GeForce GT 740]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4560u16 , name : "GK106 [GK106-INT353]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4561u16 , name : "GK106 [GK106-INT343]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4562u16 , name : "GK106 [GK106-INT232]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4563u16 , name : "GK106 [GK106-ES]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4576u16 , name : "GK106M [GeForce GTX 770M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4577u16 , name : "GK106M [GeForce GTX 765M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4578u16 , name : "GK106M [GeForce GTX 765M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4579u16 , name : "GK106M [GeForce GTX 760M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13955u16 , name : "GeForce GTX 760A" }] } , Device { vendor_id : 4318u16 , id : 4583u16 , name : "GK106M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4602u16 , name : "GK106GL [Quadro K4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4604u16 , name : "GK106GLM [Quadro K2100M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4607u16 , name : "GK106 [NB1Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4608u16 , name : "GF114 [GeForce GTX 560 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4609u16 , name : "GF114 [GeForce GTX 560]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4610u16 , name : "GF114 [GeForce GTX 560 Ti OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4611u16 , name : "GF114 [GeForce GTX 460 SE v2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4613u16 , name : "GF114 [GeForce GTX 460 v2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4614u16 , name : "GF114 [GeForce GTX 555]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4615u16 , name : "GF114 [GeForce GT 645 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4616u16 , name : "GF114 [GeForce GTX 560 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4624u16 , name : "GF114M [GeForce GTX 570M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4625u16 , name : "GF114M [GeForce GTX 580M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4626u16 , name : "GF114M [GeForce GTX 675M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4627u16 , name : "GF114M [GeForce GTX 670M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4673u16 , name : "GF116 [GeForce GT 545 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4675u16 , name : "GF116 [GeForce GT 545]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4676u16 , name : "GF116 [GeForce GTX 550 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4677u16 , name : "GF116 [GeForce GTS 450 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4678u16 , name : "GF116M [GeForce GT 550M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4679u16 , name : "GF116M [GeForce GT 555M/635M]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5970u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 4163u16 , subdevice : 8272u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 4163u16 , subdevice : 8273u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 4163u16 , subdevice : 8490u16 , name : "GeForce GT 635M" } , SubSystem { subvendor : 4163u16 , subdevice : 8491u16 , name : "GeForce GT 635M" } , SubSystem { subvendor : 4163u16 , subdevice : 8492u16 , name : "GeForce GT 635M" } , SubSystem { subvendor : 5421u16 , subdevice : 2352u16 , name : "GeForce GT 635M" }] } , Device { vendor_id : 4318u16 , id : 4680u16 , name : "GF116M [GeForce GT 555M/635M]" , subsystems : & [SubSystem { subvendor : 5421u16 , subdevice : 2352u16 , name : "GeForce GT 635M" } , SubSystem { subvendor : 6080u16 , subdevice : 4327u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6080u16 , subdevice : 4328u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6080u16 , subdevice : 4330u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6228u16 , subdevice : 2192u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6228u16 , subdevice : 2193u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6228u16 , subdevice : 6037u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6228u16 , subdevice : 6038u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 6228u16 , subdevice : 12293u16 , name : "GeForce GT 555M" }] } , Device { vendor_id : 4318u16 , id : 4681u16 , name : "GF116 [GeForce GTS 450 Rev. 3]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4683u16 , name : "GF116 [GeForce GT 640 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4685u16 , name : "GF116M [GeForce GT 555M/635M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1169u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 4136u16 , subdevice : 1392u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 4136u16 , subdevice : 1393u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 5218u16 , subdevice : 4237u16 , name : "GeForce GT 555M" } , SubSystem { subvendor : 5218u16 , subdevice : 4300u16 , name : "GeForce GT 635M" }] } , Device { vendor_id : 4318u16 , id : 4689u16 , name : "GF116M [GeForce GT 560M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4736u16 , name : "GK208 [GeForce GT 635]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4737u16 , name : "GK208 [GeForce GT 710]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4738u16 , name : "GK208 [GeForce GT 640 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4739u16 , name : "GK208 [D15M2-10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4740u16 , name : "GK208 [GeForce GT 630 Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4741u16 , name : "GK208 [GK208-100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4742u16 , name : "GK208 [GeForce GT 720]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4743u16 , name : "GK208B [GeForce GT 730]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4744u16 , name : "GK208B [GeForce GT 720]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4745u16 , name : "GK208 [GeForce GT 710]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4746u16 , name : "GK208B" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4747u16 , name : "GK208B [GeForce GT 710]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34295u16 , name : "GT710-SL-1GD5" } , SubSystem { subvendor : 4163u16 , subdevice : 34672u16 , name : "GT710-4H-SL-2GD5" }] } , Device { vendor_id : 4318u16 , id : 4748u16 , name : "GK208B" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4752u16 , name : "GK208M [GeForce GT 730M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 11002u16 , name : "GeForce GT 730A" } , SubSystem { subvendor : 4156u16 , subdevice : 11012u16 , name : "GeForce GT 730A" } , SubSystem { subvendor : 4163u16 , subdevice : 5037u16 , name : "GeForce GT 730M" } , SubSystem { subvendor : 4163u16 , subdevice : 5069u16 , name : "GeForce GT 730M" }] } , Device { vendor_id : 4318u16 , id : 4753u16 , name : "GK208M [GeForce GT 735M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4754u16 , name : "GK208M [GeForce GT 740M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13941u16 , name : "GeForce GT 740A" } , SubSystem { subvendor : 6058u16 , subdevice : 13948u16 , name : "GeForce GT 740A" } , SubSystem { subvendor : 6058u16 , subdevice : 13956u16 , name : "GeForce GT 740A" }] } , Device { vendor_id : 4318u16 , id : 4755u16 , name : "GK208M [GeForce GT 730M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4756u16 , name : "GK208M [GeForce GT 740M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4757u16 , name : "GK208M [GeForce 710M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 11021u16 , name : "GeForce 710A" } , SubSystem { subvendor : 4156u16 , subdevice : 11023u16 , name : "GeForce 710A" } , SubSystem { subvendor : 4156u16 , subdevice : 11025u16 , name : "GeForce 710A" } , SubSystem { subvendor : 4156u16 , subdevice : 11040u16 , name : "GeForce 810A" } , SubSystem { subvendor : 4156u16 , subdevice : 11041u16 , name : "GeForce 810A" } , SubSystem { subvendor : 4156u16 , subdevice : 11042u16 , name : "GeForce 810A" } , SubSystem { subvendor : 6058u16 , subdevice : 13946u16 , name : "GeForce 805A" } , SubSystem { subvendor : 6058u16 , subdevice : 13948u16 , name : "GeForce 710A" }] } , Device { vendor_id : 4318u16 , id : 4758u16 , name : "GK208M [GeForce 825M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4760u16 , name : "GK208M [GeForce GT 720M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4761u16 , name : "GK208BM [GeForce 920M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12475u16 , name : "GeForce 920A" } , SubSystem { subvendor : 6058u16 , subdevice : 12511u16 , name : "GeForce 920A" } , SubSystem { subvendor : 6058u16 , subdevice : 13991u16 , name : "GeForce 920A" } , SubSystem { subvendor : 6058u16 , subdevice : 13999u16 , name : "GeForce 920M" }] } , Device { vendor_id : 4318u16 , id : 4762u16 , name : "GK208BM [GeForce 910M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4768u16 , name : "GK208" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4781u16 , name : "GK208 [GK208-ES]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4782u16 , name : "GK208 [GK208-CS1-C]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4783u16 , name : "GK208 [GK208-INT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4784u16 , name : "GK208 [GK208-CS-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4785u16 , name : "GK208 [GK208 INT]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4793u16 , name : "GK208GLM [Quadro K610M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4794u16 , name : "GK208GLM [Quadro K510M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4875u16 , name : "GK110 [Q12U-1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4928u16 , name : "GM108M [GeForce 830M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 11051u16 , name : "GeForce 830A" }] } , Device { vendor_id : 4318u16 , id : 4929u16 , name : "GM108M [GeForce 840M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13975u16 , name : "GeForce 840A" } , SubSystem { subvendor : 6058u16 , subdevice : 13977u16 , name : "GeForce 840A" } , SubSystem { subvendor : 6058u16 , subdevice : 13980u16 , name : "GeForce 840A" }] } , Device { vendor_id : 4318u16 , id : 4932u16 , name : "GM108M [GeForce 845M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4934u16 , name : "GM108M [GeForce 930M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4935u16 , name : "GM108M [GeForce 940M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4936u16 , name : "GM108M [GeForce 945M / 945A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4937u16 , name : "GM108M [GeForce 930M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4939u16 , name : "GM108M [GeForce 940MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4941u16 , name : "GM108M [GeForce 940MX]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8776u16 , name : "ThinkPad T570" }] } , Device { vendor_id : 4318u16 , id : 4942u16 , name : "GM108M [GeForce 930MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4943u16 , name : "GM108M [GeForce 920MX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4986u16 , name : "GM108GLM [Quadro K620M / Quadro M500M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 20570u16 , name : "Quadro M500M" }] } , Device { vendor_id : 4318u16 , id : 4987u16 , name : "GM108GLM [Quadro M520 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4989u16 , name : "GM108M [GeForce 940A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4992u16 , name : "GM107 [GeForce GTX 750 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4993u16 , name : "GM107 [GeForce GTX 750]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 4994u16 , name : "GM107 [GeForce GTX 745]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5001u16 , name : "GM107GL [GRID M30]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5008u16 , name : "GM107M [GeForce 845M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5009u16 , name : "GM107M [GeForce GTX 850M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13975u16 , name : "GeForce GTX 850A" } , SubSystem { subvendor : 6058u16 , subdevice : 41253u16 , name : "GeForce GTX 850A" }] } , Device { vendor_id : 4318u16 , id : 5010u16 , name : "GM107M [GeForce GTX 860M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5011u16 , name : "GM107M [GeForce 840M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5016u16 , name : "GM107M [GeForce 845M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5017u16 , name : "GM107M [GeForce 945M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5018u16 , name : "GM107M [GeForce GTX 950M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 13868u16 , name : "GeForce GTX 950A" } , SubSystem { subvendor : 6058u16 , subdevice : 13871u16 , name : "GeForce GTX 950A" } , SubSystem { subvendor : 6058u16 , subdevice : 13887u16 , name : "GeForce GTX 950A" } , SubSystem { subvendor : 6058u16 , subdevice : 13888u16 , name : "GeForce GTX 950A" } , SubSystem { subvendor : 6058u16 , subdevice : 13895u16 , name : "GeForce GTX 950A" } , SubSystem { subvendor : 6058u16 , subdevice : 14009u16 , name : "GeForce GTX 950A" }] } , Device { vendor_id : 4318u16 , id : 5019u16 , name : "GM107M [GeForce GTX 960M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4156u16 , subdevice : 11084u16 , name : "GeForce GTX 960A" }] } , Device { vendor_id : 4318u16 , id : 5020u16 , name : "GM107M [GeForce 940M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5021u16 , name : "GM107M [GeForce GTX 750 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5037u16 , name : "GM204 [GM107 INT52]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5038u16 , name : "GM204 [GM107 CS1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5040u16 , name : "GM107GLM [Quadro M2000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5041u16 , name : "GM107GLM [Quadro M1000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5042u16 , name : "GM107GLM [Quadro M600M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5043u16 , name : "GM107GLM [Quadro K2200M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5044u16 , name : "GM107GLM [Quadro M620 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5046u16 , name : "GM107GLM [Quadro M1200 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5049u16 , name : "GM107GL [NVS 810]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5050u16 , name : "GM107GL [Quadro K2200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5051u16 , name : "GM107GL [Quadro K620]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5052u16 , name : "GM107GL [Quadro K1200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5053u16 , name : "GM107GL [Tesla M10]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4362u16 , name : "GRID M40" } , SubSystem { subvendor : 4318u16 , subdevice : 4448u16 , name : "Tesla M10" } , SubSystem { subvendor : 4318u16 , subdevice : 4562u16 , name : "GRID M10-8Q" }] } , Device { vendor_id : 4318u16 , id : 5054u16 , name : "GM204 [GM107 CS1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5055u16 , name : "GM204 [GM107 INT52]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5056u16 , name : "GM204 [GeForce GTX 980]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34052u16 , name : "GTX980-4GD5" }] } , Device { vendor_id : 4318u16 , id : 5057u16 , name : "GM204" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5058u16 , name : "GM204 [GeForce GTX 970]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5059u16 , name : "GM204" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5060u16 , name : "GM204 [D17U-20]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5079u16 , name : "GM204M [GeForce GTX 980M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5080u16 , name : "GM204M [GeForce GTX 960 OEM / 970M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5081u16 , name : "GM204M [GeForce GTX 965M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5082u16 , name : "GM204M [GeForce GTX 980 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5092u16 , name : "GM204 [Graphics Device ES-A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5095u16 , name : "GM204GL [GeForce GTX 980 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5104u16 , name : "GM204GL [Quadro M5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5105u16 , name : "GM204GL [Quadro M4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5106u16 , name : "GM204GL [Tesla M60]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4429u16 , name : "GRID M60-1Q" } , SubSystem { subvendor : 4318u16 , subdevice : 4430u16 , name : "GRID M60-2Q" } , SubSystem { subvendor : 4318u16 , subdevice : 4432u16 , name : "GRID M60-8Q" } , SubSystem { subvendor : 4318u16 , subdevice : 4528u16 , name : "GRID M60-4A" }] } , Device { vendor_id : 4318u16 , id : 5107u16 , name : "GM204GL [Tesla M6]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4484u16 , name : "GRID M6-8Q" }] } , Device { vendor_id : 4318u16 , id : 5112u16 , name : "GM204GLM [Quadro M5000M / M5000 SE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5113u16 , name : "GM204GLM [Quadro M4000M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5114u16 , name : "GM204GLM [Quadro M3000M]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4553u16 , name : "Quadro M3000 SE" }] } , Device { vendor_id : 4318u16 , id : 5115u16 , name : "GM204GLM [Quadro M5500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5121u16 , name : "GM206 [GeForce GTX 960]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5122u16 , name : "GM206 [GeForce GTX 950]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5124u16 , name : "GM206 [GeForce GTX 960 FAKE]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5126u16 , name : "GM206 [GeForce GTX 960 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5127u16 , name : "GM206 [GeForce GTX 750 v2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5159u16 , name : "GM206M [GeForce GTX 965M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" }] } , Device { vendor_id : 4318u16 , id : 5168u16 , name : "GM206GL [Quadro M2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5169u16 , name : "GM206GL [Tesla M4]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5174u16 , name : "GM206GLM [Quadro M2200 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5570u16 , name : "GP100 [CMP 100-100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5616u16 , name : "GP100GL [Quadro GP100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5617u16 , name : "GP100GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5623u16 , name : "GP100GL [Tesla P100 PCIe 12GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5624u16 , name : "GP100GL [Tesla P100 PCIe 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5625u16 , name : "GP100GL [Tesla P100 SXM2 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5626u16 , name : "GP100GL [DGX Station / PH402 SKU 200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5627u16 , name : "GP100GL [GP100 SKU 200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5628u16 , name : "GP100GL [Tesla P100-DGXS-16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5631u16 , name : "GP100GL [GP100 SKU 15ff]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5651u16 , name : "GM204GL [GRID M60 Service Provider]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5655u16 , name : "GM204M [GeForce GTX 980M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5656u16 , name : "GM204M [GeForce GTX 970M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5657u16 , name : "GM204M [GeForce GTX 965M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5658u16 , name : "GM204M [GeForce GTX 980 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5735u16 , name : "GM204M [GeForce GTX 965M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5750u16 , name : "GM204GLM [Quadro M2200 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5925u16 , name : "GP100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5934u16 , name : "GP100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5935u16 , name : "GP100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5965u16 , name : "GM108M [GeForce MX130]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 5966u16 , name : "GM108M [GeForce MX110]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6025u16 , name : "GM107GL [GRID M3-3020]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6044u16 , name : "GM107 [GeForce 940MX]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4244u16 , name : "Acer Aspire E5-575G" }] } , Device { vendor_id : 4318u16 , id : 6082u16 , name : "GM200 [GeForce GTX TITAN X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6088u16 , name : "GM200 [GeForce GTX 980 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6128u16 , name : "GM200GL [Quadro M6000]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4417u16 , name : "VCA 6000" }] } , Device { vendor_id : 4318u16 , id : 6129u16 , name : "GM200GL [Quadro M6000 24GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6141u16 , name : "GM200GL [Tesla M40]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6864u16 , name : "Tegra PCIe x8 Endpoint" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6865u16 , name : "Tegra PCIe x4/x8 Endpoint/Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6866u16 , name : "Tegra PCIe x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6867u16 , name : "Xavier SATA Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6870u16 , name : "TU102 USB 3.1 Host Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6871u16 , name : "TU102 USB Type-C UCSI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6872u16 , name : "TU104 USB 3.1 Host Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6873u16 , name : "TU104 USB Type-C UCSI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6874u16 , name : "TU106 USB 3.1 Host Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34419u16 , name : "TURBO-RTX2070-8G" }] } , Device { vendor_id : 4318u16 , id : 6875u16 , name : "TU106 USB Type-C UCSI Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34419u16 , name : "TURBO-RTX2070-8G" }] } , Device { vendor_id : 4318u16 , id : 6891u16 , name : "TU116 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6892u16 , name : "TU116 USB 3.1 Host Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6893u16 , name : "TU116 USB Type-C UCSI Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6895u16 , name : "GA102 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6897u16 , name : "GA100 [A100 NVSwitch]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6912u16 , name : "GP102 [TITAN X Pascal]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6913u16 , name : "GP102 [GeForce GTX 1080 Ti 10GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6914u16 , name : "GP102 [TITAN Xp]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6916u16 , name : "GP102" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6918u16 , name : "GP102 [GeForce GTX 1080 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6919u16 , name : "GP102 [P102-100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6960u16 , name : "GP102GL [Quadro P6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6968u16 , name : "GP102GL [Tesla P40]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 6969u16 , name : "GP102GL [Tesla P10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7024u16 , name : "GP102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7032u16 , name : "GP102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7040u16 , name : "GP104 [GeForce GTX 1080]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7041u16 , name : "GP104 [GeForce GTX 1070]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7042u16 , name : "GP104 [GeForce GTX 1070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7043u16 , name : "GP104 [GeForce GTX 1060 6GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7044u16 , name : "GP104 [GeForce GTX 1060 3GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7047u16 , name : "GP104 [P104-100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7072u16 , name : "GP104M [GeForce GTX 1080 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7073u16 , name : "GP104M [GeForce GTX 1070 Mobile]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 5713u16 , name : "GeForce GTX 1070 Max-Q" } , SubSystem { subvendor : 5218u16 , subdevice : 4584u16 , name : "GeForce GTX 1070 Max-Q" } , SubSystem { subvendor : 5218u16 , subdevice : 4585u16 , name : "GeForce GTX 1070 Max-Q" } , SubSystem { subvendor : 5464u16 , subdevice : 38145u16 , name : "GeForce GTX 1070 Max-Q" }] } , Device { vendor_id : 4318u16 , id : 7074u16 , name : "GP104M [GeForce GTX 1070 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7081u16 , name : "GP104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7082u16 , name : "GP104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7085u16 , name : "GP104 [GeForce GTX 1070 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7088u16 , name : "GP104GL [Quadro P5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7089u16 , name : "GP104GL [Quadro P4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7091u16 , name : "GP104GL [Tesla P4]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7092u16 , name : "GP104GL [Tesla P6]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7093u16 , name : "GP104GLM [Quadro P5200 Mobile]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33839u16 , name : "P5200 [Zbook 17 G5 mobile workstation]" }] } , Device { vendor_id : 4318u16 , id : 7094u16 , name : "GP104GLM [Quadro P5000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7095u16 , name : "GP104GLM [Quadro P4000 Mobile]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 4585u16 , name : "Quadro P4000 Max-Q" }] } , Device { vendor_id : 4318u16 , id : 7096u16 , name : "GP104GLM [Quadro P3000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7097u16 , name : "GP104GLM [Quadro P4200 Mobile]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33839u16 , name : "P4200 [Zbook 17 G5 mobile workstation]" }] } , Device { vendor_id : 4318u16 , id : 7099u16 , name : "GP104GLM [Quadro P3200 Mobile]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33839u16 , name : "P3200 [Zbook 17 G5 moble workstation]" }] } , Device { vendor_id : 4318u16 , id : 7111u16 , name : "GP104 [P104-101]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7136u16 , name : "GP104BM [GeForce GTX 1080 Mobile]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1984u16 , name : "GeForce GTX 1080 Max-Q" } , SubSystem { subvendor : 5208u16 , subdevice : 13659u16 , name : "GeForce GTX 1080 Max-Q" }] } , Device { vendor_id : 4318u16 , id : 7137u16 , name : "GP104BM [GeForce GTX 1070 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7168u16 , name : "GP106" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7169u16 , name : "GP106" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7170u16 , name : "GP106 [GeForce GTX 1060 3GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7171u16 , name : "GP106 [GeForce GTX 1060 6GB]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34230u16 , name : "DUAL-GTX1060-O6G [GeForce GTX 1060 6GB Dual]" }] } , Device { vendor_id : 4318u16 , id : 7172u16 , name : "GP106 [GeForce GTX 1060 5GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7174u16 , name : "GP106 [GeForce GTX 1060 6GB Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7175u16 , name : "GP106 [P106-100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7177u16 , name : "GP106 [P106-090]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7200u16 , name : "GP106M [GeForce GTX 1060 Mobile]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14777u16 , name : "GeForce GTX 1060 Max-Q 3GB" }] } , Device { vendor_id : 4318u16 , id : 7201u16 , name : "GP106M [GeForce GTX 1050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7202u16 , name : "GP106M [GeForce GTX 1050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7203u16 , name : "GP106M [GeForce GTX 1060 Mobile Rev. 2]" , subsystems : & [SubSystem { subvendor : 5140u16 , subdevice : 32u16 , name : "GTX 1060 Mobile" }] } , Device { vendor_id : 4318u16 , id : 7213u16 , name : "GP106M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7216u16 , name : "GP106GL [Quadro P2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7217u16 , name : "GP106GL [Quadro P2200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7221u16 , name : "GP106M [Quadro P2000 Mobile / DRIVE PX 2 AutoChauffeur]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7222u16 , name : "GP106 [P106M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7264u16 , name : "GP106BM [GeForce GTX 1060 Mobile 6GB]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33680u16 , name : "GeForce GTX 1060 Max-Q 6GB" }] } , Device { vendor_id : 4318u16 , id : 7265u16 , name : "GP106BM [GeForce GTX 1050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7266u16 , name : "GP106BM [GeForce GTX 1050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7280u16 , name : "GP106GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7297u16 , name : "GP107 [GeForce GTX 1050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7298u16 , name : "GP107 [GeForce GTX 1050 Ti]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34323u16 , name : "PH-GTX1050TI-4G" } , SubSystem { subvendor : 5208u16 , subdevice : 14179u16 , name : "GV-N105TOC-4GD" }] } , Device { vendor_id : 4318u16 , id : 7299u16 , name : "GP107 [GeForce GTX 1050 3GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7308u16 , name : "GP107M [GeForce GTX 1050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7309u16 , name : "GP107M [GeForce GTX 1050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7310u16 , name : "GP107M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7311u16 , name : "GP107M [GeForce GTX 1050 Ti Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7312u16 , name : "GP107M [GeForce MX150]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7313u16 , name : "GP107M [GeForce GTX 1050 3 GB Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7314u16 , name : "GP107M [GeForce GTX 1050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7316u16 , name : "GP107M [GeForce MX350]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7318u16 , name : "GP107M [GeForce MX350]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7335u16 , name : "GP107GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7336u16 , name : "GP107GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7338u16 , name : "GP107GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7345u16 , name : "GP107GL [Quadro P1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7346u16 , name : "GP107GL [Quadro P600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7347u16 , name : "GP107GL [Quadro P400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7350u16 , name : "GP107GL [Quadro P620]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7354u16 , name : "GP107GLM [Quadro P2000 Mobile]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33836u16 , name : "P2000 [Zbook 15 G5 mobile workstation]" } , SubSystem { subvendor : 4156u16 , subdevice : 33839u16 , name : "P2000 [Zbook 17 G5 mobile workstation]" }] } , Device { vendor_id : 4318u16 , id : 7355u16 , name : "GP107GLM [Quadro P1000 Mobile]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33833u16 , name : "P1000 [Zbook Studio G5 mobile workstation]" } , SubSystem { subvendor : 4156u16 , subdevice : 33836u16 , name : "P1000 [Zbook 15 G5 mobile workstation]" } , SubSystem { subvendor : 4156u16 , subdevice : 33839u16 , name : "P1000 [Zbook 17 G5 mobile workstation]" } , SubSystem { subvendor : 4156u16 , subdevice : 33873u16 , name : "P1000 [Zbook Studio x360 G5 mobile workstation]" }] } , Device { vendor_id : 4318u16 , id : 7356u16 , name : "GP107GLM [Quadro P600 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7357u16 , name : "GP107GLM [Quadro P620]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7372u16 , name : "GP107BM [GeForce GTX 1050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7373u16 , name : "GP107BM [GeForce GTX 1050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7418u16 , name : "GP107GL [Quadro P2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7419u16 , name : "GP107GL [Quadro P1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7425u16 , name : "GP108 [GeForce GT 1030]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7426u16 , name : "GP108 [GeForce GT 1010]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7440u16 , name : "GP108M [GeForce MX150]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8798u16 , name : "ThinkPad T480" }] } , Device { vendor_id : 4318u16 , id : 7441u16 , name : "GP108M [GeForce MX230]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7442u16 , name : "GP108M [GeForce MX150]" , subsystems : & [SubSystem { subvendor : 7538u16 , subdevice : 5889u16 , name : "Mi Notebook Pro [GeForce MX150]" }] } , Device { vendor_id : 4318u16 , id : 7443u16 , name : "GP108M [GeForce MX250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7446u16 , name : "GP108M [GeForce MX330]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7475u16 , name : "GP108GLM [Quadro P500 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7476u16 , name : "GP108GLM [Quadro P520]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7506u16 , name : "GP108BM [GeForce MX250]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7510u16 , name : "GP108BM [GeForce MX330]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7553u16 , name : "GV100 [TITAN V]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7555u16 , name : "GV100 [CMP 100-200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7556u16 , name : "GV100 [CMP 100-210]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7600u16 , name : "GV100GL [Tesla GV100 SXM2-16GB SKU 890]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7601u16 , name : "GV100GL [Tesla V100 SXM2 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7602u16 , name : "GV100GL [Tesla V100 DGXS 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7603u16 , name : "GV100GL [Tesla V100 FHHL 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7604u16 , name : "GV100GL [Tesla V100 PCIe 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7605u16 , name : "GV100GL [Tesla V100 SXM2 32GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7606u16 , name : "GV100GL [Tesla V100 PCIe 32GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7607u16 , name : "GV100GL [Tesla V100 DGXS 32GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7608u16 , name : "GV100GL [Tesla V100 SXM3 32GB]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4893u16 , name : "Tesla V100-SXM3-32GB-H" }] } , Device { vendor_id : 4318u16 , id : 7610u16 , name : "GV100GL [Quadro GV100]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4843u16 , name : "TITAN V CEO Edition" }] } , Device { vendor_id : 4318u16 , id : 7613u16 , name : "GV100GL [Tesla GV100 DGX1-V]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7614u16 , name : "GV100 Engineering Sample" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7617u16 , name : "GV100 [CMP 100-200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7664u16 , name : "GV100GL [Tesla PG500-216]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7666u16 , name : "GV100GL [Tesla PG503-216]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7668u16 , name : "GV100 [CMP 100-210]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7669u16 , name : "GV100GL [Tesla V100 SXM2 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7670u16 , name : "GV100GL [Tesla V100S PCIe 32GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7682u16 , name : "TU102 [TITAN RTX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7683u16 , name : "TU102 [GeForce RTX 2080 Ti 12GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7684u16 , name : "TU102 [GeForce RTX 2080 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7687u16 , name : "TU102 [GeForce RTX 2080 Ti Rev. A]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 14101u16 , name : "RTX 2080 Ti GAMING X TRIO" }] } , Device { vendor_id : 4318u16 , id : 7689u16 , name : "TU102 [CMP 50HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7725u16 , name : "TU102 [GeForce RTX 2080 Ti Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7726u16 , name : "TU102 [GeForce RTX 2080 Ti 12GB Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7728u16 , name : "TU102GL [Quadro RTX 6000/8000]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4766u16 , name : "Quadro RTX 8000" } , SubSystem { subvendor : 4318u16 , subdevice : 4794u16 , name : "Quadro RTX 6000" }] } , Device { vendor_id : 4318u16 , id : 7733u16 , name : "TU102GL [Tesla T10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7734u16 , name : "TU102GL [Quadro RTX 6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7735u16 , name : "TU102GL [Tesla T10 16GB / GRID RTX T10-2/T10-4/T10-8]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 4868u16 , name : "Tesla T10 16GB" } , SubSystem { subvendor : 4318u16 , subdevice : 4935u16 , name : "GRID RTX T10-8" } , SubSystem { subvendor : 4318u16 , subdevice : 4936u16 , name : "GRID RTX T10-4" } , SubSystem { subvendor : 4318u16 , subdevice : 4937u16 , name : "GRID RTX T10-2" } , SubSystem { subvendor : 4318u16 , subdevice : 4976u16 , name : "Tesla T10 16GB" } , SubSystem { subvendor : 4318u16 , subdevice : 5029u16 , name : "GRID RTX T10-8" }] } , Device { vendor_id : 4318u16 , id : 7736u16 , name : "TU102GL [Tesla T40 24GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7740u16 , name : "TU102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7741u16 , name : "TU102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7742u16 , name : "TU102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7800u16 , name : "TU102GL [Quadro RTX 6000/8000]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 5080u16 , name : "Quadro RTX 8000" } , SubSystem { subvendor : 4318u16 , subdevice : 5081u16 , name : "Quadro RTX 6000" }] } , Device { vendor_id : 4318u16 , id : 7809u16 , name : "TU104 [GeForce RTX 2080 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7810u16 , name : "TU104 [GeForce RTX 2080]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7812u16 , name : "TU104 [GeForce RTX 2070 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7815u16 , name : "TU104 [GeForce RTX 2080 Rev. A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7817u16 , name : "TU104 [GeForce RTX 2060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7824u16 , name : "TU104M [GeForce RTX 2080 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7825u16 , name : "TU104M [GeForce RTX 2070 SUPER Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7827u16 , name : "TU104M [GeForce RTX 2080 SUPER Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7851u16 , name : "TU104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7854u16 , name : "TU104M [GeForce GTX 2080 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7856u16 , name : "TU104GL [Quadro RTX 5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7857u16 , name : "TU104GL [Quadro RTX 4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7860u16 , name : "TU104GL [T4G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7861u16 , name : "TU104GLM [Quadro RTX 5000 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7862u16 , name : "TU104GLM [Quadro RTX 4000 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7864u16 , name : "TU104GL [Tesla T4]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7865u16 , name : "TU104GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7866u16 , name : "TU104GL [PG189 SKU600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7870u16 , name : "TU104GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7874u16 , name : "TU104 [GeForce RTX 2070 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7879u16 , name : "TU104 [GeForce RTX 2070 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7888u16 , name : "TU104BM [GeForce RTX 2080 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7889u16 , name : "TU104BM [GeForce RTX 2070 SUPER Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7891u16 , name : "TU104BM [GeForce RTX 2080 SUPER Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7925u16 , name : "TU104GLM [Quadro RTX 5000 Mobile Refresh]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7938u16 , name : "TU106 [GeForce RTX 2070]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34419u16 , name : "TURBO RTX 2070" }] } , Device { vendor_id : 4318u16 , id : 7939u16 , name : "TU106 [GeForce RTX 2060 12GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7940u16 , name : "TU106" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7942u16 , name : "TU106 [GeForce RTX 2060 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7943u16 , name : "TU106 [GeForce RTX 2070 Rev. A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7944u16 , name : "TU106 [GeForce RTX 2060 Rev. A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7945u16 , name : "TU106 [GeForce GTX 1660 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7946u16 , name : "TU106 [GeForce GTX 1650]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7947u16 , name : "TU106 [CMP 40HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7952u16 , name : "TU106M [GeForce RTX 2070 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7953u16 , name : "TU106M [GeForce RTX 2060 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7954u16 , name : "TU106M [GeForce RTX 2060 Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7956u16 , name : "TU106M [GeForce RTX 2070 Mobile / Max-Q Refresh]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7957u16 , name : "TU106M [GeForce RTX 2060 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7982u16 , name : "TU106M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 7990u16 , name : "TU106GLM [Quadro RTX 3000 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8002u16 , name : "TU106 [GeForce RTX 2060 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8007u16 , name : "TU106 [GeForce RTX 2060 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8016u16 , name : "TU106BM [GeForce RTX 2070 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8017u16 , name : "TU106BM [GeForce RTX 2060 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8020u16 , name : "TU106BM [GeForce RTX 2070 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8021u16 , name : "TU106BM [GeForce RTX 2060 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8054u16 , name : "TU106GLM [Quadro RTX 3000 Mobile Refresh]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8065u16 , name : "TU117" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8066u16 , name : "TU117 [GeForce GTX 1650]" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 13717u16 , name : "GeForce GTX 1650 OC GDDR6" }] } , Device { vendor_id : 4318u16 , id : 8067u16 , name : "TU117 [GeForce GTX 1630]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8081u16 , name : "TU117M [GeForce GTX 1650 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8082u16 , name : "TU117M [GeForce GTX 1650 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8084u16 , name : "TU117M [GeForce GTX 1650 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8085u16 , name : "TU117M [GeForce GTX 1650 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8086u16 , name : "TU117M [GeForce GTX 1650 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8087u16 , name : "TU117M [GeForce MX450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8088u16 , name : "TU117M [GeForce MX450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8089u16 , name : "TU117M [GeForce GTX 1650 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8092u16 , name : "TU117M [GeForce MX450]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8093u16 , name : "TU117M [GeForce GTX 1650 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8095u16 , name : "TU117M [GeForce MX550]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8096u16 , name : "TU117M [GeForce MX550]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8097u16 , name : "TU117M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8110u16 , name : "TU117GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8112u16 , name : "TU117GLM [Quadro T1000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8113u16 , name : "TU117GL [T600]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8114u16 , name : "TU117GLM [Quadro T400 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8118u16 , name : "TU117GLM [T600 Laptop GPU]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" }] } , Device { vendor_id : 4318u16 , id : 8119u16 , name : "TU117GLM [T550 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8120u16 , name : "TU117GLM [Quadro T2000 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8121u16 , name : "TU117GLM [Quadro T1000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8122u16 , name : "TU117GLM [T600 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8123u16 , name : "TU117GLM [Quadro T500 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8124u16 , name : "TU117GLM [T1200 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8127u16 , name : "TU117GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8153u16 , name : "TU117BM [GeForce GTX 1650 Mobile Refresh]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8157u16 , name : "TU117BM [GeForce GTX 1650 Mobile Refresh]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8176u16 , name : "TU117GL [T1000 8GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8178u16 , name : "TU117GL [T400 4GB / T400E]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8185u16 , name : "TU117GLM [Quadro T1000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8320u16 , name : "GA100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8321u16 , name : "GA100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8322u16 , name : "GA100 [CMP 170HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8368u16 , name : "GA100 [A100 SXM4 40GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8369u16 , name : "GA100 [A100 PCIe 40GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8370u16 , name : "GA100 [A100 SXM4 80GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8371u16 , name : "GA100 [A100-SXM-64GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8372u16 , name : "GA100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8373u16 , name : "GA100 [A100 PCIe 80GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8374u16 , name : "GA100GL [PG506-232]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8375u16 , name : "GA100GL [A30 PCIe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8376u16 , name : "GA100 [A100X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8377u16 , name : "GA100 [A30X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8379u16 , name : "GA100 [DRIVE A100 PROD]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8381u16 , name : "GA100 [A800 SXM4 40GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8382u16 , name : "GA100 [GRID A100A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8383u16 , name : "GA100 [GRID A100B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8384u16 , name : "GA100 [Initial DevID]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8386u16 , name : "GA100 [CMP 170HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8432u16 , name : "GA100 [A100-PG506-207]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8433u16 , name : "GA100 [A100 PCIe 40GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8434u16 , name : "GA100 [A100-PG506-217]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8435u16 , name : "GA100 [A800-SXM4-80GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8437u16 , name : "GA100 [A800 80GB PCIe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8438u16 , name : "GA100 [A800 40GB PCIe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8445u16 , name : "GA100 [AX800 Converged Accelerator]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8446u16 , name : "GA100 [INT SKU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8447u16 , name : "GA100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8578u16 , name : "TU116 [GeForce GTX 1660 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8579u16 , name : "TU116" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8580u16 , name : "TU116 [GeForce GTX 1660]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8582u16 , name : "TU116 [PG160 SKU18]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8583u16 , name : "TU116 [GeForce GTX 1650 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8584u16 , name : "TU116 [GeForce GTX 1650]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8585u16 , name : "TU116 [CMP 30HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8593u16 , name : "TU116M [GeForce GTX 1660 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8594u16 , name : "TU116M [GeForce GTX 1650 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8622u16 , name : "TU116GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8639u16 , name : "TU116GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8642u16 , name : "TU116" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8643u16 , name : "TU116" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8644u16 , name : "TU116 [GeForce GTX 1660 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8657u16 , name : "TU116BM [GeForce GTX 1660 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8704u16 , name : "GA102" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8707u16 , name : "GA102 [GeForce RTX 3090 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8708u16 , name : "GA102 [GeForce RTX 3090]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 5245u16 , name : "GeForce RTX 3090 Founders Edition" } , SubSystem { subvendor : 5218u16 , subdevice : 14465u16 , name : "MSI RTX 3090 VENTUS 3X OC" } , SubSystem { subvendor : 14402u16 , subdevice : 14707u16 , name : "GeForce RTX 3090 XC3" }] } , Device { vendor_id : 4318u16 , id : 8709u16 , name : "GA102 [GeForce RTX 3080 Ti 20GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8710u16 , name : "GA102 [GeForce RTX 3080]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 5223u16 , name : "GA102 [GeForce RTX 3080]" } , SubSystem { subvendor : 4318u16 , subdevice : 5229u16 , name : "GA102 [GeForce RTX 3080 20GB]" } , SubSystem { subvendor : 5218u16 , subdevice : 14482u16 , name : "RTX 3080 10GB GAMING X TRIO" }] } , Device { vendor_id : 4318u16 , id : 8711u16 , name : "GA102 [GeForce RTX 3070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8712u16 , name : "GA102 [GeForce RTX 3080 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8714u16 , name : "GA102 [GeForce RTX 3080 12GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8717u16 , name : "GA102 [CMP 90HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8726u16 , name : "GA102 [GeForce RTX 3080 Lite Hash Rate]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8747u16 , name : "GA102 [GeForce RTX 3090 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8751u16 , name : "GA102 [GeForce RTX 3080 11GB / 12GB Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8752u16 , name : "GA102GL [RTX A6000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8753u16 , name : "GA102GL [RTX A5000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8754u16 , name : "GA102GL [RTX A4500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8755u16 , name : "GA102GL [RTX A5500]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8757u16 , name : "GA102GL [A40]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8758u16 , name : "GA102GL [A10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8759u16 , name : "GA102GL [A10G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8760u16 , name : "GA102GL [A10M]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8767u16 , name : "GA102GL" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8843u16 , name : "GA104 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8846u16 , name : "GA106 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8849u16 , name : "GA107 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8854u16 , name : "Tegra PCIe Endpoint Virtual Network" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8858u16 , name : "Orin PCIe x8 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8860u16 , name : "Orin PCIe x4/x8 Endpoint/Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8862u16 , name : "Orin PCIe x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8867u16 , name : "GH100 [H100 NVSwitch]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8890u16 , name : "AD102 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8891u16 , name : "AD103 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8892u16 , name : "AD104 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8893u16 , name : "AD106M High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8894u16 , name : "AD107 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8910u16 , name : "GB10 GEN5 X4 PCIe host" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8912u16 , name : "GB10 GEN4 X1 PCIe host" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8913u16 , name : "GB20B PCI bridge" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8920u16 , name : "THOR Processor PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8934u16 , name : "THOR Processor PCI Express x16 Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8936u16 , name : "GB202 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8937u16 , name : "GB203 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8939u16 , name : "GB206 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8940u16 , name : "GB207 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8962u16 , name : "GH100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8972u16 , name : "GH100 [H20 NVL16]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8974u16 , name : "GH100 [H20 NVL16]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8979u16 , name : "GH100 [H100 CNX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8993u16 , name : "GH100 [H100L 94GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8994u16 , name : "GH100 [H800 PCIe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 8996u16 , name : "GH100 [H800]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9000u16 , name : "GH100 [H20B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9001u16 , name : "GH100 [H20]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9004u16 , name : "GH100 [H20 HBM3e]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9008u16 , name : "GH100 [H100 SXM5 80GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9009u16 , name : "GH100 [H100 PCIe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9013u16 , name : "GH100 [H200 SXM 141GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9014u16 , name : "GH100 [H100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9015u16 , name : "GH100 [H100 SXM5 64GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9016u16 , name : "GH100 [H100 SXM5 96GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9017u16 , name : "GH100 [H100 SXM5 94GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9018u16 , name : "GH100 [H800L 94GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9019u16 , name : "GH100 [H200 NVL]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9021u16 , name : "GH100 [H100 96GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9026u16 , name : "GH100 [GH200 120GB / 480GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9027u16 , name : "GH100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9029u16 , name : "GH100 [GH100-88K-A1]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9032u16 , name : "GH100 [GH200 144G HBM3e]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9086u16 , name : "GH100 [H100 GH3]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9087u16 , name : "GH100 [Skinny Joe]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9136u16 , name : "GH100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9200u16 , name : "GH100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9236u16 , name : "GA103 [GeForce RTX 3060 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9248u16 , name : "GA103M [GeForce RTX 3080 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9272u16 , name : "GA103GLM [RTX A5500 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9312u16 , name : "GA103M [GeForce RTX 3080 Ti Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9344u16 , name : "GA104 [Reserved Dev ID A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9346u16 , name : "GA104 [GeForce RTX 3070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9347u16 , name : "GA104" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9348u16 , name : "GA104 [GeForce RTX 3070]" , subsystems : & [SubSystem { subvendor : 4318u16 , subdevice : 5227u16 , name : "GA104 [GeForce RTX 3070]" } , SubSystem { subvendor : 4318u16 , subdevice : 5294u16 , name : "GA104 [GeForce RTX 3070 16GB]" }] } , Device { vendor_id : 4318u16 , id : 9350u16 , name : "GA104 [GeForce RTX 3060 Ti]" , subsystems : & [SubSystem { subvendor : 6618u16 , subdevice : 26160u16 , name : "ZT-A30610H-10M [RTX 3060 Ti Twin Edge OC]" }] } , Device { vendor_id : 4318u16 , id : 9351u16 , name : "GA104 [GeForce RTX 3060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9352u16 , name : "GA104 [GeForce RTX 3070 Lite Hash Rate]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9353u16 , name : "GA104 [GeForce RTX 3060 Ti Lite Hash Rate]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9354u16 , name : "GA104 [CMP 70HX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9356u16 , name : "GA104 [GeForce RTX 3070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9357u16 , name : "GA104 [GeForce RTX 3070]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9358u16 , name : "GA104 [GeForce RTX 3060 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9372u16 , name : "GA104M [GeForce RTX 3080 Mobile / Max-Q 8GB/16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9373u16 , name : "GA104M [GeForce RTX 3070 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9375u16 , name : "GA104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9376u16 , name : "GA104 [Geforce RTX 3070 Ti Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9380u16 , name : "GA104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9388u16 , name : "GA104 [GeForce RTX 30x0 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9389u16 , name : "GA104 [GeForce RTX 3060 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9391u16 , name : "GA104 [GeForce RTX 3070 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9392u16 , name : "GA104GL [RTX A4000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9393u16 , name : "GA104GL [RTX A4000H]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9398u16 , name : "GA104GLM [RTX A5000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9399u16 , name : "GA104GLM [RTX A4000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9400u16 , name : "GA104GLM [RTX A3000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9401u16 , name : "GA104GLM [RTX A3000 12GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9402u16 , name : "GA104GLM [RTX A4500 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9403u16 , name : "GA104GLM [RTX A3000 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9407u16 , name : "GA104 [GeForce RTX 3070 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9408u16 , name : "GA104 [Initial Dev ID B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9415u16 , name : "GA104 [GeForce RTX 3060 8GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9416u16 , name : "GA104 [GeForce RTX 3070 GDDR6X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9417u16 , name : "GA104 [GeForce RTX 3060 Ti GDDR6X]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9436u16 , name : "GA104M [GeForce RTX 3080 Mobile / Max-Q 8GB/16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9437u16 , name : "GA104M [GeForce RTX 3070 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9439u16 , name : "GA104M" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9440u16 , name : "GA104M [Geforce RTX 3070 Ti Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9466u16 , name : "GA104 [RTX A4500 Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9473u16 , name : "GA106 [GeForce RTX 3060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9475u16 , name : "GA106 [GeForce RTX 3060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9476u16 , name : "GA106 [GeForce RTX 3060 Lite Hash Rate]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9477u16 , name : "GA106" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9479u16 , name : "GA106 [Geforce RTX 3050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9480u16 , name : "GA106 [GeForce RTX 3050 OEM]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9481u16 , name : "GA106 [GeForce RTX 3060 12GB Rev. 2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9504u16 , name : "GA106M [GeForce RTX 3060 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9505u16 , name : "GA106M [GeForce RTX 3060 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9507u16 , name : "GA106M [GeForce RTX 3050 Ti Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9519u16 , name : "GA106 [GeForce RTX 3060 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9521u16 , name : "GA106 [RTX A2000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9540u16 , name : "GA106 [GeForce RTX 3060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9568u16 , name : "GA106M [GeForce RTX 3060 Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9569u16 , name : "GA106M [GeForce RTX 3060 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9571u16 , name : "GA106M [GeForce RTX 3050 Ti Mobile / Max-Q]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9585u16 , name : "GA106 [RTX A2000 12GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9602u16 , name : "GA107 [GeForce RTX 3050 8GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9603u16 , name : "GA107 [GeForce RTX 3050 4GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9604u16 , name : "GA107 [GeForce RTX 3050 6GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9632u16 , name : "GA107M [GeForce RTX 3050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9634u16 , name : "GA107M [GeForce RTX 3050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9635u16 , name : "GA107" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9636u16 , name : "GA107" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9637u16 , name : "GA107M [GeForce RTX 3050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9638u16 , name : "GA107M [GeForce MX570]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9639u16 , name : "GA107M [GeForce MX570]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9641u16 , name : "GA107M [GeForce RTX 2050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9642u16 , name : "GA107M [GeForce MX570 A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9643u16 , name : "GA107M [GeForce RTX 3050 4GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9644u16 , name : "GA107BM / GN20-P0-R-K2 [GeForce RTX 3050 6GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9645u16 , name : "GA107 [GeForce RTX 2050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9647u16 , name : "GA107 [GeForce RTX 3050 Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9648u16 , name : "GA107GL [RTX A1000]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9650u16 , name : "GA107GL [RTX A400]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9653u16 , name : "GA107GLM [RTX A4 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9654u16 , name : "GA107GL [A2 / A16]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9656u16 , name : "GA107GLM [RTX A2000 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9657u16 , name : "GA107GLM [RTX A1000 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9658u16 , name : "GA107GLM [RTX A2000 8GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9659u16 , name : "GA107GLM [RTX A500 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9660u16 , name : "GA107GLM [RTX A1000 6GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9661u16 , name : "GA107GLM [RTX A500 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9696u16 , name : "GA107BM [GeForce RTX 3050 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9698u16 , name : "GA107BM [GeForce RTX 3050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9701u16 , name : "GA107BM [GeForce RTX 3050 Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9708u16 , name : "GA107BM / GN20-P0-R-K2 [GeForce RTX 3050 6GB Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9709u16 , name : "GA107 [GeForce RTX 2050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9721u16 , name : "GA107 [RTX A1000 Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9722u16 , name : "GA107 [RTX A2000 Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9723u16 , name : "GA107 [RTX A500 Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9857u16 , name : "AD102 [RTX TITAN Ada]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9860u16 , name : "AD102 [GeForce RTX 4090]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9861u16 , name : "AD102 [GeForce RTX 4090 D]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9865u16 , name : "AD102 [GeForce RTX 4070 Ti SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9903u16 , name : "AD102 [PG137]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9905u16 , name : "AD102GL [RTX 6000 Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9906u16 , name : "AD102GL [RTX 5000 Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9907u16 , name : "AD102GL [RTX 5880 Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9909u16 , name : "AD102GL [L40]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9911u16 , name : "AD102GL [L20]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9912u16 , name : "AD102GL [L40G]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9913u16 , name : "AD102GL [L40S]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9914u16 , name : "AD102GL [L20]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9915u16 , name : "AD102GL [L30]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9973u16 , name : "AD102GL [L40 CNX]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9986u16 , name : "AD103 [GeForce RTX 4080 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9987u16 , name : "AD103 [GeForce RTX 4080 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9988u16 , name : "AD103 [GeForce RTX 4080]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9989u16 , name : "AD103 [GeForce RTX 4070 Ti SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 9993u16 , name : "AD103 [GeForce RTX 4070]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10007u16 , name : "AD103M / GN21-X11 [GeForce RTX 4090 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10032u16 , name : "AD103GLM [RTX 5000 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10071u16 , name : "AD103M / GN21-X11 [GeForce RTX 4090 Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10096u16 , name : "AD103GLM [RTX 5000 Ada Generation Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10114u16 , name : "AD104 [GeForce RTX 4070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10115u16 , name : "AD104 [GeForce RTX 4070 SUPER]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10117u16 , name : "AD104 [AC AD104 20GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10118u16 , name : "AD104 [GeForce RTX 4070]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10120u16 , name : "AD104 [GeForce RTX 4060 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10144u16 , name : "AD104M [GeForce RTX 4080 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10160u16 , name : "AD104GL [RTX 4000 SFF Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10161u16 , name : "AD104GL [RTX 4500 Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10162u16 , name : "AD104GL [RTX 4000 Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10166u16 , name : "AD104GL [L2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10167u16 , name : "AD104GL [L16]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10168u16 , name : "AD104GL [L4]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10170u16 , name : "AD104GLM [RTX 4000 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10171u16 , name : "AD104GLM [RTX 3500 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10208u16 , name : "AD104M [GeForce RTX 4080 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10234u16 , name : "AD104GLM [RTX 4000 Ada Generation Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10235u16 , name : "AD104GLM [RTX 3500 Ada Generation Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10243u16 , name : "AD106 [GeForce RTX 4060 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10245u16 , name : "AD106 [GeForce RTX 4060 Ti 16GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10248u16 , name : "AD106 [GeForce RTX 4060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10272u16 , name : "AD106M [GeForce RTX 4070 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10274u16 , name : "AD106M [GeForce RTX 3050 A Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10296u16 , name : "AD106GLM [RTX 3000 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10336u16 , name : "AD106M [GeForce RTX 4070 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10360u16 , name : "AD106GLM [RTX 3000 Ada Generation Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10370u16 , name : "AD107 [GeForce RTX 4060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10400u16 , name : "AD107M [GeForce RTX 4060 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10401u16 , name : "AD107M [GeForce RTX 4050 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10403u16 , name : "AD107M [GeForce RTX 3050 A Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10416u16 , name : "AD107GL [RTX 2000 / 2000E Ada Generation]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10424u16 , name : "AD107GLM [RTX 2000 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10425u16 , name : "AD107GLM [RTX 1000 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10426u16 , name : "AD107GLM [RTX 500 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10427u16 , name : "AD107GLM [RTX 500 Ada Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10464u16 , name : "AD107M [GeForce RTX 4060 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10465u16 , name : "AD107M [GeForce RTX 4050 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10467u16 , name : "AD107M [GeForce RTX 3050 A Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10488u16 , name : "AD107GLM [RTX 2000 Ada Generation Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10496u16 , name : "GB100 [Reserved Dev ID A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10497u16 , name : "GB100 [B200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10505u16 , name : "GB100 [HGX B200 168GB]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10528u16 , name : "GB100 [TS4 / B100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10532u16 , name : "GB100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10533u16 , name : "GB100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10557u16 , name : "GB100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10560u16 , name : "GB100 [Reserved Dev ID B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10561u16 , name : "GB100 [HGX GB200]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10622u16 , name : "GB100" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10624u16 , name : "GB102 [Reserved Dev ID A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10683u16 , name : "GB102 [DRIVE P2021]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10684u16 , name : "GB102 [B100]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10688u16 , name : "GB102 [Reserved Dev ID B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 10737u16 , name : "GB102" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11008u16 , name : "GB10B [Jetson AGX Thor]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11141u16 , name : "GB202 [GeForce RTX 5090]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11143u16 , name : "GB202 [GeForce RTX 5090 D]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11148u16 , name : "GB202 [GeForce RTX 5090 D V2]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11185u16 , name : "GB202GL [RTX PRO 6000 Blackwell Workstation Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11186u16 , name : "GB202GL [RTX PRO 6000D Blackwell Workstation Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11187u16 , name : "GB202GL [RTX PRO 5000 Blackwell / RTX PRO 5000 72GB Blackwell]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11188u16 , name : "GB202GL [RTX PRO 6000 Blackwell Max-Q Workstation Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11189u16 , name : "GB202GL [RTX PRO 6000 Blackwell Server Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11193u16 , name : "GB202GL [RTX 6000D]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11196u16 , name : "GB202GL [RTX PRO 6000D Blackwell Max-Q Workstation Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11266u16 , name : "GB203 [GeForce RTX 5080]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11269u16 , name : "GB203 [GeForce RTX 5070 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11288u16 , name : "GB203M / GN22 [GeForce RTX 5090 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11289u16 , name : "GB203M / GN22 [GeForce RTX 5080 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11308u16 , name : "GB6-256(N22W-ES-A1)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11313u16 , name : "GB203GL [RTX PRO 4500 Blackwell]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11315u16 , name : "GB203GL [RTX PRO 4000 Blackwell SFF Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11316u16 , name : "GB203GL [RTX PRO 4000 Blackwell]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11320u16 , name : "GB203GLM [RTX PRO 5000 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11321u16 , name : "GB203GLM [RTX PRO 4000 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11322u16 , name : "GB203GL [RTX PRO 4500 Blackwell Server Edition]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11352u16 , name : "GB203M / GN22-X11 [GeForce RTX 5090 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11353u16 , name : "GB203M / GN22-X9 [GeForce RTX 5080 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11383u16 , name : "GB203GLM [RTX PRO 5000 Blackwell Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11385u16 , name : "GB203GLM [RTX PRO 4000 Blackwell Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11524u16 , name : "GB206 [GeForce RTX 5060 Ti]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11525u16 , name : "GB206 [GeForce RTX 5060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11544u16 , name : "GB206M [GeForce RTX 5070 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11545u16 , name : "GB206M [GeForce RTX 5060 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11564u16 , name : "GB6-128 (N22Y-ES-A1)" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11568u16 , name : "GB206GL [RTX PRO 2000 Blackwell]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11577u16 , name : "GB206GLM [RTX PRO 2000 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11608u16 , name : "GB206M [GeForce RTX 5070 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11609u16 , name : "GB206M [GeForce RTX 5060 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11641u16 , name : "GB206GLM [RTX PRO 2000 Blackwell Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11651u16 , name : "GB207 [GeForce RTX 5050]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11672u16 , name : "GB207M [GeForce RTX 5050 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11704u16 , name : "GB207GLM [RTX PRO 1000 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11705u16 , name : "GB207GLM [RTX PRO 500 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11736u16 , name : "GB207M [GeForce RTX 5050 Max-Q / Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11769u16 , name : "GB207GLM [RTX PRO 500 Blackwell Embedded GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11779u16 , name : "GB20B [RTX Spark N1X (6144-core Blackwell RTX GPU)]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11782u16 , name : "GB20B [RTX Spark N1X (5120-core Blackwell RTX GPU)]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11794u16 , name : "GB20B [GB10]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11795u16 , name : "GB20B [Desktop Device]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 11818u16 , name : "GB20B [JMJWOA-Generic-GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12036u16 , name : "GB205 [GeForce RTX 5070]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12038u16 , name : "GB205 [GeForce RTX 5060]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12056u16 , name : "GB205M [GeForce RTX 5070 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12088u16 , name : "GB205GLM [RTX PRO 3000 Blackwell Generation Laptop GPU]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12120u16 , name : "GB205M [GeForce RTX 5070 Ti Mobile]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12160u16 , name : "GB205 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12181u16 , name : "Vera Rubin CPU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12182u16 , name : "Vera Rubin CPU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12183u16 , name : "Vera Rubin CPU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12184u16 , name : "Vera Rubin CPU" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12288u16 , name : "GR100 [Reserved Dev ID A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12321u16 , name : "GR100 [PCIe Faber]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12352u16 , name : "GR100 [Reserved Dev ID B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12480u16 , name : "GR102" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12672u16 , name : "GB110 [Reserved Dev ID A]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12674u16 , name : "GB110 [B300 SXM6 AC]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12705u16 , name : "GB110 [GB300 MaxQ]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12736u16 , name : "GB110 [Reserved Dev ID B]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12738u16 , name : "GB110 [GB300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12739u16 , name : "GB110 [GB300]" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12798u16 , name : "GB110" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12800u16 , name : "GB112" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12836u16 , name : "GB112" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 12862u16 , name : "GB112" , subsystems : & [] } , Device { vendor_id : 4318u16 , id : 13120u16 , name : "GB120" , subsystems : & [] }] }),
        (7579, Vendor { id : 7579u16 , name : "Meta Platforms, Inc." , devices : & [Device { vendor_id : 7579u16 , id : 16u16 , name : "Networking DOM Engine" , subsystems : & [] } , Device { vendor_id : 7579u16 , id : 17u16 , name : "IO Bridge" , subsystems : & [] } , Device { vendor_id : 7579u16 , id : 19u16 , name : "Host Network Interface" , subsystems : & [] } , Device { vendor_id : 7579u16 , id : 1024u16 , name : "Time Card" , subsystems : & [] }] }),
        (4370, Vendor { id : 4370u16 , name : "Osicom Technologies Inc" , devices : & [Device { vendor_id : 4370u16 , id : 8704u16 , name : "FDDI Adapter" , subsystems : & [] } , Device { vendor_id : 4370u16 , id : 8960u16 , name : "Fast Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4370u16 , id : 9024u16 , name : "4 Port Fast Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4370u16 , id : 9216u16 , name : "ATM Adapter" , subsystems : & [] }] }),
        (4435, Vendor { id : 4435u16 , name : "Land Win Electronic Corp" , devices : & [] }),
        (4405, Vendor { id : 4405u16 , name : "FUJIFILM Business Innovation Corp." , devices : & [Device { vendor_id : 4405u16 , id : 1u16 , name : "Printer controller" , subsystems : & [] }] }),
        (4171, Vendor { id : 4171u16 , name : "BusLogic" , devices : & [Device { vendor_id : 4171u16 , id : 320u16 , name : "BT-946C (old) [multimaster  01]" , subsystems : & [] } , Device { vendor_id : 4171u16 , id : 4160u16 , name : "BT-946C (BA80C30) [MultiMaster 10]" , subsystems : & [] } , Device { vendor_id : 4171u16 , id : 33072u16 , name : "Flashpoint LT" , subsystems : & [] }] }),
        (6838, Vendor { id : 6838u16 , name : "CalDigit, Inc." , devices : & [Device { vendor_id : 6838u16 , id : 25089u16 , name : "RAID Card" , subsystems : & [] }] }),
        (4993, Vendor { id : 4993u16 , name : "Brains Co. Ltd" , devices : & [] }),
        (4548, Vendor { id : 4548u16 , name : "Document Technologies, Inc" , devices : & [] }),
        (4819, Vendor { id : 4819u16 , name : "Vingmed Sound A/S" , devices : & [] }),
        (5244, Vendor { id : 5244u16 , name : "AWARE, Inc." , devices : & [] }),
        (6626, Vendor { id : 6626u16 , name : "Vector Informatik GmbH" , devices : & [] }),
        (5447, Vendor { id : 5447u16 , name : "MITUTOYO Corp" , devices : & [] }),
        (5517, Vendor { id : 5517u16 , name : "Point Multimedia Systems" , devices : & [] }),
        (4650, Vendor { id : 4650u16 , name : "Integrated Telecom" , devices : & [] }),
        (7888, Vendor { id : 7888u16 , name : "Hosin Global Electronics" , devices : & [Device { vendor_id : 7888u16 , id : 8835u16 , name : "Patriot P300 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5479, Vendor { id : 5479u16 , name : "Jungsoft" , devices : & [] }),
        (5620, Vendor { id : 5620u16 , name : "Valuesoft" , devices : & [] }),
        (5333, Vendor { id : 5333u16 , name : "Nitsuko Corporation" , devices : & [] }),
        (4600, Vendor { id : 4600u16 , name : "Microchip Technology" , devices : & [Device { vendor_id : 4600u16 , id : 16384u16 , name : "PM40100 Switchtec PFX 100xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16424u16 , name : "PM40028 Switchtec PFX 28xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16438u16 , name : "PM40036 Switchtec PFX 36xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16466u16 , name : "PM40052 Switchtec PFX 52xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16488u16 , name : "PM40068 Switchtec PFX 68xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16516u16 , name : "PM40084 Switchtec PFX 84xG4 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16640u16 , name : "PM41100 Switchtec PSX 100xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16680u16 , name : "PM41028 Switchtec PSX 28xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16694u16 , name : "PM41036 Switchtec PSX 36xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16722u16 , name : "PM41052 Switchtec PSX 52xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16744u16 , name : "PM41068 Switchtec PSX 68xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16772u16 , name : "PM41084 Switchtec PSX 84xG4 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16896u16 , name : "PM42100 Switchtec PAX 100xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16936u16 , name : "PM42028 Switchtec PAX 28xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16950u16 , name : "PM42036 Switchtec PAX 36xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 16978u16 , name : "PM42052 Switchtec PAX 52xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17000u16 , name : "PM42068 Switchtec PAX 68xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17028u16 , name : "PM42084 Switchtec PAX 84xG4 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17192u16 , name : "PM43028 Switchtec PFXA 28xG4 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17206u16 , name : "PM43036 Switchtec PFXA 36xG4 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17234u16 , name : "PM43052 Switchtec PFXA 52xG4 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17448u16 , name : "PM44028 Switchtec PSXA 28xG4 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17462u16 , name : "PM44036 Switchtec PSXA 36xG4 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17490u16 , name : "PM44052 Switchtec PSXA 52xG4 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17704u16 , name : "PM45028 Switchtec PAXA 28xG4 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17718u16 , name : "PM45036 Switchtec PAXA 36xG4 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 17746u16 , name : "PM45052 Switchtec PAXA 52xG4 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20480u16 , name : "PM50100 Switchtec PFX 100xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20520u16 , name : "PM50028 Switchtec PFX 28xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20534u16 , name : "PM50036 Switchtec PFX 36xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20562u16 , name : "PM50052 Switchtec PFX 52xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20584u16 , name : "PM50068 Switchtec PFX 68xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20612u16 , name : "PM50084 Switchtec PFX 84xG5 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20736u16 , name : "PM51100 Switchtec PSX 100xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20776u16 , name : "PM51028 Switchtec PSX 28xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20790u16 , name : "PM51036 Switchtec PSX 36xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20818u16 , name : "PM51052 Switchtec PSX 52xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20840u16 , name : "PM51068 Switchtec PSX 68xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20868u16 , name : "PM51084 Switchtec PSX 84xG5 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 20992u16 , name : "PM52100 Switchtec PAX 100xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21024u16 , name : "BR522x [PMC-Sierra maxRAID SAS Controller]" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21032u16 , name : "PM52028 Switchtec PAX 28xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21046u16 , name : "PM52036 Switchtec PAX 36xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21074u16 , name : "PM52052 Switchtec PAX 52xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21096u16 , name : "PM52068 Switchtec PAX 68xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21124u16 , name : "PM52084 Switchtec PAX 84xG5 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21248u16 , name : "PM53100 Switchtec PFXA 100xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21288u16 , name : "PM53028 Switchtec PFXA 28xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21302u16 , name : "PM53036 Switchtec PFXA 36xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21330u16 , name : "PM53052 Switchtec PFXA 52xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21352u16 , name : "PM53068 Switchtec PFXA 68xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21380u16 , name : "PM53084 Switchtec PFXA 84xG5 Automotive Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21504u16 , name : "PM54100 Switchtec PSXA 100xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21544u16 , name : "PM54028 Switchtec PSXA 28xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21558u16 , name : "PM54036 Switchtec PSXA 36xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21586u16 , name : "PM54052 Switchtec PSXA 52xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21608u16 , name : "PM54068 Switchtec PSXA 68xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21636u16 , name : "PM54084 Switchtec PSXA 84xG5 Automotive Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21760u16 , name : "PM55100 Switchtec PAXA 100xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21800u16 , name : "PM55028 Switchtec PAXA 28xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21814u16 , name : "PM55036 Switchtec PAXA 36xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21842u16 , name : "PM55052 Switchtec PAXA 52xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21864u16 , name : "PM55068 Switchtec PAXA 68xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 21892u16 , name : "PM55084 Switchtec PAXA 84xG5 Automotive Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24644u16 , name : "PM60144 Switchtec PFXs 144xG6 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24648u16 , name : "PM60048 Switchtec PFXs 48xG6 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24672u16 , name : "PM60160 Switchtec PFXs 160xG6 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24676u16 , name : "PM60064 Switchtec PFXs 64xG6 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24900u16 , name : "PM61144 Switchtec PSXs 144xG6 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24904u16 , name : "PM61048 Switchtec PSXs 48xG6 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24928u16 , name : "PM61160 Switchtec PSXs 160xG6 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 24932u16 , name : "PM61064 Switchtec PSXs 64xG6 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25156u16 , name : "PM62144 Switchtec PFX 144xG6 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25160u16 , name : "PM62048 Switchtec PFX 48xG6 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25184u16 , name : "PM62160 Switchtec PFX 160xG6 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25188u16 , name : "PM62064 Switchtec PFX 64xG6 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25412u16 , name : "PM63144 Switchtec PSX 144xG6 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25416u16 , name : "PM63048 Switchtec PSX 48xG6 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25440u16 , name : "PM63160 Switchtec PSX 160xG6 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 25444u16 , name : "PM63064 Switchtec PSX 64xG6 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 28740u16 , name : "PM70144 Switchtec PFXs 144xG7 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 28744u16 , name : "PM70048 Switchtec PFXs 48xG7 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 28768u16 , name : "PM70160 Switchtec PFXs 160xG7 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 28772u16 , name : "PM70064 Switchtec PFXs 64xG7 Secure-capable Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 28996u16 , name : "PM71144 Switchtec PSXs 144xG7 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29000u16 , name : "PM71048 Switchtec PSXs 48xG7 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29024u16 , name : "PM71160 Switchtec PSXs 160xG7 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29028u16 , name : "PM71064 Switchtec PSXs 64xG7 Secure-capable Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29252u16 , name : "PM72144 Switchtec PFX 144xG7 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29256u16 , name : "PM72048 Switchtec PFX 48xG7 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29280u16 , name : "PM72160 Switchtec PFX 160xG7 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29284u16 , name : "PM72064 Switchtec PFX 64xG7 Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29540u16 , name : "PM7364 [FREEDM - 32 Frame Engine & Datalink Mgr]" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29557u16 , name : "PM7375 [LASAR-155 ATM SAR]" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29572u16 , name : "PM7384 [FREEDM - 84P672 Frm Engine & Datalink Mgr]" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29764u16 , name : "PM74144 Switchtec PSX 144xG7 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29768u16 , name : "PM74048 Switchtec PSX 48xG7 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29792u16 , name : "PM74160 Switchtec PSX 160xG7 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 29796u16 , name : "PM74064 Switchtec PSX 64xG7 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32768u16 , name : "PM8000  [SPC - SAS Protocol Controller]" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32777u16 , name : "PM8009 SPCve 8x6G" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32792u16 , name : "PM8018 Adaptec SAS Adaptor ASA-70165H PCIe Gen3 x8 6 Gbps 16-lane 4x SFF-8644" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32818u16 , name : "PM8032 Tachyon QE8" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 58u16 , name : "Celerity FC-81EN Fibre Channel Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 59u16 , name : "Celerity FC-82EN Fibre Channel Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 60u16 , name : "Celerity FC-84EN Fibre Channel Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16443u16 , name : "ThunderLink FC 1082 Fibre Channel Adapter" }] } , Device { vendor_id : 4600u16 , id : 32851u16 , name : "PM8053 SXP 12G 24-port SAS/SATA expander" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32852u16 , name : "PM8054 SXP 12G 36-port SAS/SATA expander" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32853u16 , name : "PM8055 SXP 12G 48-port SAS/SATA expander" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32854u16 , name : "PM8056 SXP 12G 68-port SAS/SATA expander" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32864u16 , name : "PM8060 SRCv 12G eight-port SAS/SATA RoC" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32867u16 , name : "PM8063 SRCv 12G 16-port SAS/SATA RoC" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32880u16 , name : "PM8070 Tachyon SPCv 12G eight-port SAS/SATA controller" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32881u16 , name : "PM8071 Tachyon SPCve 12G eight-port SAS/SATA controller" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32882u16 , name : "PM8072 Tachyon SPCv 12G 16-port SAS/SATA controller" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 32883u16 , name : "PM8073 Tachyon SPCve 12G 16-port SAS/SATA controller" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34097u16 , name : "PM8531 PFX 24xG3 PCIe Fanout Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34098u16 , name : "PM8532 PFX 32xG3 PCIe Fanout Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34099u16 , name : "PM8533 PFX 48xG3 PCIe Fanout Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34100u16 , name : "PM8534 PFX 64xG3 PCIe Fanout Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34101u16 , name : "PM8535 PFX 80xG3 PCIe Fanout Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34102u16 , name : "PM8536 PFX 96xG3 PCIe Fanout Switch" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 129u16 , name : "PM8536 PFX 96xG3 PCIe Fanout Switch" }] } , Device { vendor_id : 4600u16 , id : 34113u16 , name : "PM8541 PSX 24xG3 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34114u16 , name : "PM8542 PSX 32xG3 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34115u16 , name : "PM8543 PSX 48xG3 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34116u16 , name : "PM8544 PSX 64xG3 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34117u16 , name : "PM8545 PSX 80xG3 Programmable PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34118u16 , name : "PM8546 B-FEIP PSX 96xG3 PCIe Storage Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34129u16 , name : "PM8551 PAX 24xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34130u16 , name : "PM8552 PAX 32xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34131u16 , name : "PM8553 PAX 48xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34132u16 , name : "PM8554 PAX 64xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34133u16 , name : "PM8555 PAX 80xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34134u16 , name : "PM8556 PAX 96xG3 Programmable Advanced Fabric PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34145u16 , name : "PM8561 Switchtec PFX-L 24xG3 Fanout-Lite PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34146u16 , name : "PM8562 Switchtec PFX-L 32xG3 Fanout-Lite PCIe Gen3 Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34147u16 , name : "PM8563 Switchtec PFX-L 48xG3 Fanout-Lite PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34148u16 , name : "PM8564 Switchtec PFX-L 64xG3 Fanout-Lite PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34149u16 , name : "PM8565 Switchtec PFX-L 80xG3 Fanout-Lite PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34150u16 , name : "PM8566 Switchtec PFX-L 96xG3 Fanout-Lite PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34161u16 , name : "PM8571 Switchtec PFX-I 24xG3 Industrial Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34162u16 , name : "PM8572 Switchtec PFX-I 32xG3 Industrial Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34163u16 , name : "PM8573 Switchtec PFX-I 48xG3 Industrial Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34164u16 , name : "PM8574 Switchtec PFX-I 64xG3 Industrial Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34165u16 , name : "PM8575 Switchtec PFX-I 80xG3 Industrial Fanout PCIe Switch" , subsystems : & [] } , Device { vendor_id : 4600u16 , id : 34166u16 , name : "PM8576 Switchtec PFX-I 96xG3 Industrial Fanout PCIe Switch" , subsystems : & [] }] }),
        (4281, Vendor { id : 4281u16 , name : "ULi Electronics Inc." , devices : & [Device { vendor_id : 4281u16 , id : 257u16 , name : "CMI8338/C3DX PCI Audio Device" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 273u16 , name : "C-Media CMI8738/C3DX Audio Device (OEM)" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 273u16 , name : "C-Media CMI8738/C3DX Audio Device (OEM)" }] } , Device { vendor_id : 4281u16 , id : 1920u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 1922u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5173u16 , name : "M1435" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5189u16 , name : "M1445" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5193u16 , name : "M1449" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5201u16 , name : "M1451" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5217u16 , name : "M1461" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5257u16 , name : "M1489 Cache-Memory PCI Controller (CMP) [FinALi 486] CPU to PCI bridge" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5393u16 , name : "M1511 [Aladdin]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5394u16 , name : "M1512 [Aladdin]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5395u16 , name : "M1513 [Aladdin]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5409u16 , name : "M1521 [Aladdin III]" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5409u16 , name : "ALI M1521 Aladdin III CPU Bridge" }] } , Device { vendor_id : 4281u16 , id : 5411u16 , name : "M1523" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5411u16 , name : "ALI M1523 ISA Bridge" }] } , Device { vendor_id : 4281u16 , id : 5425u16 , name : "M1531 [Aladdin IV]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5427u16 , name : "M1533/M1535/M1543 PCI to ISA Bridge [Aladdin IV/V/V+]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1339u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4281u16 , subdevice : 5427u16 , name : "ALi M1533 Aladdin IV/V ISA Bridge" }] } , Device { vendor_id : 4281u16 , id : 5441u16 , name : "M1541" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5441u16 , name : "ALI M1541 Aladdin V/V+ AGP System Controller" }] } , Device { vendor_id : 4281u16 , id : 5443u16 , name : "M1543" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5475u16 , name : "M1563 HyperTransport South Bridge" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5475u16 , name : "ASRock 939Dual-SATA2 Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 5475u16 , name : "ASRock 939Dual-SATA2 Motherboard" }] } , Device { vendor_id : 4281u16 , id : 5491u16 , name : "PCI to LPC Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5493u16 , name : "M1575 South Bridge" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5665u16 , name : "M1621" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5681u16 , name : "ALI M1631 PCI North Bridge Aladdin Pro III" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5682u16 , name : "M1632M Northbridge+Trident" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5697u16 , name : "ALI M1641 PCI North Bridge Aladdin Pro IV" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5700u16 , name : "M1644/M1644T Northbridge+Trident" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5702u16 , name : "M1646 Northbridge+Trident" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5703u16 , name : "M1647 Northbridge [MAGiK 1 / MobileMAGiK 1]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5713u16 , name : "M1651/M1651T Northbridge [Aladdin-Pro 5/5M,Aladdin-Pro 5T/5TM]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5745u16 , name : "M1671 Super P4 Northbridge [AGP4X,PCI and SDR/DDR]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5746u16 , name : "M1672 Northbridge [CyberALADDiN-P4]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5761u16 , name : "M1681 P4 Northbridge [AGP8X,HyperTransport and SDR/DDR]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5767u16 , name : "M1687 K8 Northbridge [AGP8X and HyperTransport]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5769u16 , name : "M1689 K8 Northbridge [Super K8 Single Chip]" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5781u16 , name : "M1695 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 5783u16 , name : "M1697 HTT Host Bridge" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12609u16 , name : "M3141" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12611u16 , name : "M3143" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12613u16 , name : "M3145" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12615u16 , name : "M3147" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12617u16 , name : "M3149" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 12625u16 , name : "M3151" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 13063u16 , name : "M3307" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 13065u16 , name : "M3309" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 13091u16 , name : "M3325 Video/Audio Decoder" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21010u16 , name : "M4803" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21013u16 , name : "MS4803" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21015u16 , name : "M5217H" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21017u16 , name : "M5219" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21029u16 , name : "M5225" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21032u16 , name : "M5228 ALi ATA/RAID Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21033u16 , name : "M5229 IDE" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1295u16 , name : "ThinkPad R30" } , SubSystem { subvendor : 4116u16 , subdevice : 1341u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin IDE" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" } , SubSystem { subvendor : 4163u16 , subdevice : 32851u16 , name : "A7A266 Motherboard IDE" } , SubSystem { subvendor : 6217u16 , subdevice : 21033u16 , name : "ASRock 939Dual-SATA2 Motherboard IDE (PATA)" }] } , Device { vendor_id : 4281u16 , id : 21045u16 , name : "M5225" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21047u16 , name : "USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1344u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin USB" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" } , SubSystem { subvendor : 4173u16 , subdevice : 33039u16 , name : "VAIO PCG-U1 USB/OHCI Revision 1.0" } , SubSystem { subvendor : 4281u16 , subdevice : 21047u16 , name : "M5273 USB OHCI PCI adapter" } , SubSystem { subvendor : 6217u16 , subdevice : 21047u16 , name : "ASRock 939Dual-SATA2 Motherboard" }] } , Device { vendor_id : 4281u16 , id : 21049u16 , name : "USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 21049u16 , name : "M5273 USB PCI adapter" } , SubSystem { subvendor : 6217u16 , subdevice : 21049u16 , name : "ASRock 939Dual-SATA2 Motherboard" }] } , Device { vendor_id : 4281u16 , id : 21059u16 , name : "M1541 PCI to AGP Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21062u16 , name : "AGP8X Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21063u16 , name : "PCI to AGP Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21065u16 , name : "M5249 HTT to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21067u16 , name : "PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21068u16 , name : "PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21069u16 , name : "PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21070u16 , name : "PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21073u16 , name : "M5251 P1394 OHCI 1.0 Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21075u16 , name : "M5253 P1394 OHCI 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21089u16 , name : "M5261 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21091u16 , name : "ULi 1689,1573 integrated ethernet." , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21121u16 , name : "ALi M5281 Serial ATA / RAID Host Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21127u16 , name : "ULi 5287 SATA" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21128u16 , name : "ULi M5288 SATA" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32854u16 , name : "A8R-MVP Mainboard" }] } , Device { vendor_id : 4281u16 , id : 21129u16 , name : "ULi 5289 SATA" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21584u16 , name : "Lucent Technologies Soft Modem AMR" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21585u16 , name : "M5451 PCI AC-Link Controller Audio Device" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1286u16 , name : "ThinkPad R30" } , SubSystem { subvendor : 4116u16 , subdevice : 1342u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin Audio" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" }] } , Device { vendor_id : 4281u16 , id : 21587u16 , name : "M5453 PCI AC-Link Controller Modem Device" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21589u16 , name : "M5455 PCI AC-Link Controller Audio Device" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 21589u16 , name : "ASRock 939Dual-SATA2 Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 2128u16 , name : "ASRock 939Dual-SATA2 Motherboard" }] } , Device { vendor_id : 4281u16 , id : 21591u16 , name : "M5457 AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1333u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin Modem Device" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" }] } , Device { vendor_id : 4281u16 , id : 21593u16 , name : "SmartLink SmartPCI561 56K Modem" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21594u16 , name : "SmartLink SmartPCI563 56K Modem" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21601u16 , name : "HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21617u16 , name : "M5471 Memory Stick Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 21619u16 , name : "M5473 SD-MMC Controller" , subsystems : & [] } , Device { vendor_id : 4281u16 , id : 28929u16 , name : "M7101 Power Management Controller [PMU]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1296u16 , name : "ThinkPad R30" } , SubSystem { subvendor : 4116u16 , subdevice : 1340u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" } , SubSystem { subvendor : 6217u16 , subdevice : 28929u16 , name : "ASRock 939Dual-SATA2 Motherboard" }] }] }),
        (5114, Vendor { id : 5114u16 , name : "Pentland Systems Ltd" , devices : & [] }),
        (4720, Vendor { id : 4720u16 , name : "Olympus Optical Co., Ltd." , devices : & [] }),
        (4369, Vendor { id : 4369u16 , name : "Santa Cruz Operation" , devices : & [] }),
        (5102, Vendor { id : 5102u16 , name : "Hayes Microcomputer Products Inc" , devices : & [] }),
        (4947, Vendor { id : 4947u16 , name : "dbeeSet Technology" , devices : & [Device { vendor_id : 4947u16 , id : 2u16 , name : "Proserver" , subsystems : & [] } , Device { vendor_id : 4947u16 , id : 3u16 , name : "PCI-FUT" , subsystems : & [] } , Device { vendor_id : 4947u16 , id : 4u16 , name : "PCI-S0" , subsystems : & [] } , Device { vendor_id : 4947u16 , id : 5u16 , name : "PCI-FUT-S0" , subsystems : & [] } , Device { vendor_id : 4947u16 , id : 6u16 , name : "OTDU-1U (FPGA Zynq-7000)" , subsystems : & [] } , Device { vendor_id : 4947u16 , id : 7u16 , name : "OTDU-EX" , subsystems : & [] }] }),
        (4499, Vendor { id : 4499u16 , name : "Zeitnet Inc." , devices : & [Device { vendor_id : 4499u16 , id : 1u16 , name : "1221" , subsystems : & [] } , Device { vendor_id : 4499u16 , id : 2u16 , name : "1225" , subsystems : & [] }] }),
        (4401, Vendor { id : 4401u16 , name : "Philips Semiconductors" , devices : & [Device { vendor_id : 4401u16 , id : 5473u16 , name : "USB 1.1 Host Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 49664u16 , name : "C2K onboard USB 1.1 host controller" }] } , Device { vendor_id : 4401u16 , id : 5474u16 , name : "USB 2.0 Host Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 49664u16 , name : "C2K onboard USB 2.0 host controller" }] } , Device { vendor_id : 4401u16 , id : 13312u16 , name : "SmartPCI56(UCB1500) 56K Modem" , subsystems : & [] } , Device { vendor_id : 4401u16 , id : 21504u16 , name : "TriMedia TM1000/1100" , subsystems : & [SubSystem { subvendor : 4810u16 , subdevice : 0u16 , name : "BlueICE" }] } , Device { vendor_id : 4401u16 , id : 21506u16 , name : "TriMedia TM1300" , subsystems : & [SubSystem { subvendor : 4676u16 , subdevice : 3840u16 , name : "Fritz!Card DSL" } , SubSystem { subvendor : 5611u16 , subdevice : 4864u16 , name : "DT1300" } , SubSystem { subvendor : 5611u16 , subdevice : 4866u16 , name : "DT1302" } , SubSystem { subvendor : 5611u16 , subdevice : 4868u16 , name : "DT1304" } , SubSystem { subvendor : 5611u16 , subdevice : 4869u16 , name : "DT1305" } , SubSystem { subvendor : 5611u16 , subdevice : 4870u16 , name : "PMCDT1306" } , SubSystem { subvendor : 5611u16 , subdevice : 4872u16 , name : "DT1308" } , SubSystem { subvendor : 5611u16 , subdevice : 4913u16 , name : "DT1301 with SAA7121" } , SubSystem { subvendor : 5611u16 , subdevice : 4919u16 , name : "DT1301 with SAA7127" } , SubSystem { subvendor : 5611u16 , subdevice : 11581u16 , name : "X3D" } , SubSystem { subvendor : 5611u16 , subdevice : 28706u16 , name : "PTM1300" }] } , Device { vendor_id : 4401u16 , id : 21509u16 , name : "TriMedia TM1500" , subsystems : & [SubSystem { subvendor : 4406u16 , subdevice : 5u16 , name : "LCP-1500" }] } , Device { vendor_id : 4401u16 , id : 21510u16 , name : "TriMedia TM1700" , subsystems : & [] } , Device { vendor_id : 4401u16 , id : 21515u16 , name : "PNX1005 Media Processor" , subsystems : & [SubSystem { subvendor : 4401u16 , subdevice : 32u16 , name : "PNXLite PCI Demo Board" }] } , Device { vendor_id : 4401u16 , id : 28976u16 , name : "SAA7130 Video Broadcast Decoder" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 16406u16 , name : "Behold TV 401" } , SubSystem { subvendor : 0u16 , subdevice : 16465u16 , name : "Behold TV 405 FM" } , SubSystem { subvendor : 0u16 , subdevice : 20561u16 , name : "Behold TV 505 RDS" } , SubSystem { subvendor : 0u16 , subdevice : 20571u16 , name : "Behold TV 505 RDS" } , SubSystem { subvendor : 4139u16 , subdevice : 18640u16 , name : "Matrox CronosPlus" } , SubSystem { subvendor : 4168u16 , subdevice : 8811u16 , name : "ELSA EX-VISION 300TV" } , SubSystem { subvendor : 4221u16 , subdevice : 26197u16 , name : "WinFast DTV1000S" } , SubSystem { subvendor : 4401u16 , subdevice : 0u16 , name : "SAA7130-based TV tuner card" } , SubSystem { subvendor : 4401u16 , subdevice : 8193u16 , name : "10MOONS PCI TV CAPTURE CARD" } , SubSystem { subvendor : 4401u16 , subdevice : 8197u16 , name : "Techcom (India) TV Tuner Card (SSD-TV-670)" } , SubSystem { subvendor : 5208u16 , subdevice : 36870u16 , name : "GT-PS700 DVB-S tuner" } , SubSystem { subvendor : 5217u16 , subdevice : 1292u16 , name : "Nagase Sangyo TransGear 3000TV" } , SubSystem { subvendor : 5217u16 , subdevice : 4351u16 , name : "AVerMedia DVD EZMaker" } , SubSystem { subvendor : 5217u16 , subdevice : 8456u16 , name : "AverMedia AverTV/305" } , SubSystem { subvendor : 5217u16 , subdevice : 8469u16 , name : "AverMedia AverTV Studio 305" } , SubSystem { subvendor : 5435u16 , subdevice : 4434u16 , name : "Terratec Cinergy 200 TV" } , SubSystem { subvendor : 6235u16 , subdevice : 49408u16 , name : "Compro VideoMate TV PVR/FM" } , SubSystem { subvendor : 6235u16 , subdevice : 51457u16 , name : "Videomate DVB-T200" } , SubSystem { subvendor : 20840u16 , subdevice : 312u16 , name : "LifeView FlyVIDEO2000" } , SubSystem { subvendor : 23246u16 , subdevice : 20496u16 , name : "Behold TV 501" } , SubSystem { subvendor : 23246u16 , subdevice : 20560u16 , name : "Behold TV 505 FM" }] } , Device { vendor_id : 4401u16 , id : 28979u16 , name : "SAA7131/SAA7133/SAA7135 Video Broadcast Decoder" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 16529u16 , name : "Beholder BeholdTV 409 FM" } , SubSystem { subvendor : 0u16 , subdevice : 20593u16 , name : "Behold TV 507 RDS" } , SubSystem { subvendor : 0u16 , subdevice : 20603u16 , name : "Behold TV 507 RDS" } , SubSystem { subvendor : 0u16 , subdevice : 20993u16 , name : "Behold TV Columbus" } , SubSystem { subvendor : 112u16 , subdevice : 26369u16 , name : "WinTV HVR-1110" } , SubSystem { subvendor : 4121u16 , subdevice : 19637u16 , name : "Elitegroup ECS TVP3XP FM1236 Tuner Card (NTSC,FM)" } , SubSystem { subvendor : 4163u16 , subdevice : 528u16 , name : "FlyTV mini Asus Digimatrix" } , SubSystem { subvendor : 4163u16 , subdevice : 18499u16 , name : "TV-FM 7133" } , SubSystem { subvendor : 4163u16 , subdevice : 18501u16 , name : "TV-FM 7135" } , SubSystem { subvendor : 4163u16 , subdevice : 18530u16 , name : "P7131 Dual" } , SubSystem { subvendor : 4163u16 , subdevice : 18550u16 , name : "My Cinema-P7131 Hybrid" } , SubSystem { subvendor : 4401u16 , subdevice : 0u16 , name : "SAA713x-based TV tuner card" } , SubSystem { subvendor : 4401u16 , subdevice : 8193u16 , name : "Proteus Pro [philips reference design]" } , SubSystem { subvendor : 4401u16 , subdevice : 8216u16 , name : "Tiger reference design" } , SubSystem { subvendor : 4401u16 , subdevice : 20201u16 , name : "MonsterTV Mobile" } , SubSystem { subvendor : 4401u16 , subdevice : 28979u16 , name : "Pinnacle PCTV 301i" } , SubSystem { subvendor : 4541u16 , subdevice : 43u16 , name : "PCTV Stereo" } , SubSystem { subvendor : 4541u16 , subdevice : 46u16 , name : "PCTV 110i (saa7133)" } , SubSystem { subvendor : 4779u16 , subdevice : 2048u16 , name : "PURPLE TV" } , SubSystem { subvendor : 5058u16 , subdevice : 10244u16 , name : "Technotrend Budget T-3000 Hybrid" } , SubSystem { subvendor : 5153u16 , subdevice : 821u16 , name : "Instant TV DVB-T Cardbus" } , SubSystem { subvendor : 5153u16 , subdevice : 4976u16 , name : "Instant TV (saa7135)" } , SubSystem { subvendor : 5173u16 , subdevice : 29488u16 , name : "VFG7330" } , SubSystem { subvendor : 5173u16 , subdevice : 29520u16 , name : "VFG7350" } , SubSystem { subvendor : 5208u16 , subdevice : 36865u16 , name : "GC-PTV-TAF Hybrid TV card" } , SubSystem { subvendor : 5208u16 , subdevice : 36866u16 , name : "GT-PTV-TAF-RH DVB-T/Analog TV/FM tuner" } , SubSystem { subvendor : 5208u16 , subdevice : 36867u16 , name : "GT-PTV-AF-RH Analog TV/FM tuner" } , SubSystem { subvendor : 5208u16 , subdevice : 36868u16 , name : "GT-P8000 DVB-T/Analog TV/FM tuner" } , SubSystem { subvendor : 5208u16 , subdevice : 36869u16 , name : "GT-P6000 Analog TV/FM tuner" } , SubSystem { subvendor : 5208u16 , subdevice : 36872u16 , name : "GT-P5100 Analog TV tuner" } , SubSystem { subvendor : 5217u16 , subdevice : 4164u16 , name : "AVerTVHD MCE A180" } , SubSystem { subvendor : 5217u16 , subdevice : 18486u16 , name : "M10D Hybrid DVBT" } , SubSystem { subvendor : 5217u16 , subdevice : 34334u16 , name : "M105 PAL/SECAM/NTSC/FM Tuner" } , SubSystem { subvendor : 5217u16 , subdevice : 41291u16 , name : "AVerTV Studio 509" } , SubSystem { subvendor : 5217u16 , subdevice : 43062u16 , name : "M115 DVB-T, PAL/SECAM/NTSC Tuner" } , SubSystem { subvendor : 5217u16 , subdevice : 61469u16 , name : "DVB-T Super 007" } , SubSystem { subvendor : 5217u16 , subdevice : 62239u16 , name : "Avermedia AVerTV GO 007 FM" } , SubSystem { subvendor : 5217u16 , subdevice : 63798u16 , name : "Hybrid+FM PCI (rev A16D)" } , SubSystem { subvendor : 5218u16 , subdevice : 25137u16 , name : "TV@nywhere Plus" } , SubSystem { subvendor : 5257u16 , subdevice : 532u16 , name : "LifeView FlyTV Platinum FM" } , SubSystem { subvendor : 5312u16 , subdevice : 4626u16 , name : "LifeView FlyTV Platinum Mini2" } , SubSystem { subvendor : 5435u16 , subdevice : 4448u16 , name : "Cinergy 250 PCI TV" } , SubSystem { subvendor : 5435u16 , subdevice : 4450u16 , name : "Terratec Cinergy 400 mobile" } , SubSystem { subvendor : 6110u16 , subdevice : 29270u16 , name : "PlusTV All In One PI610 card" } , SubSystem { subvendor : 6110u16 , subdevice : 29520u16 , name : "ATSC 110 Digital / Analog HDTV Tuner" } , SubSystem { subvendor : 6110u16 , subdevice : 29522u16 , name : "ATSC 115 Digital / Analog HDTV Tuner" } , SubSystem { subvendor : 6235u16 , subdevice : 49408u16 , name : "VideoMate TV" } , SubSystem { subvendor : 6235u16 , subdevice : 51456u16 , name : "VideoMate T750" } , SubSystem { subvendor : 20840u16 , subdevice : 774u16 , name : "LifeView FlyDVB-T DUO" } , SubSystem { subvendor : 20840u16 , subdevice : 793u16 , name : "LifeView FlyDVB Trio" } , SubSystem { subvendor : 20840u16 , subdevice : 1282u16 , name : "LifeView FlyDVB-T Duo CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 1312u16 , name : "LifeView FlyDVB Trio CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 5378u16 , name : "LifeView FlyTV CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 9474u16 , name : "LifeView FlyDVB-T CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 9504u16 , name : "LifeView FlyDVB-S Duo CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 13570u16 , name : "LifeView FlyDVB-T Hybrid CardBus" } , SubSystem { subvendor : 20840u16 , subdevice : 13600u16 , name : "LifeView FlyDVB Trio N CardBus" } , SubSystem { subvendor : 23246u16 , subdevice : 20528u16 , name : "Behold TV 503 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 20624u16 , name : "Behold TV 509 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24720u16 , name : "Behold TV 609 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24721u16 , name : "Behold TV 609 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24722u16 , name : "Behold TV 609 RDS" } , SubSystem { subvendor : 23246u16 , subdevice : 24723u16 , name : "Behold TV 609 RDS" } , SubSystem { subvendor : 23246u16 , subdevice : 24976u16 , name : "Behold TV M6" } , SubSystem { subvendor : 23246u16 , subdevice : 24977u16 , name : "Behold TV M63" } , SubSystem { subvendor : 23246u16 , subdevice : 24979u16 , name : "Behold TV M6 Extra" } , SubSystem { subvendor : 23246u16 , subdevice : 25232u16 , name : "Behold TV H6" } , SubSystem { subvendor : 23246u16 , subdevice : 28816u16 , name : "Behold TV A7" } , SubSystem { subvendor : 23246u16 , subdevice : 29008u16 , name : "Behold TV H75" } , SubSystem { subvendor : 23246u16 , subdevice : 29009u16 , name : "Behold TV H75" } , SubSystem { subvendor : 23246u16 , subdevice : 29072u16 , name : "Behold TV H7" } , SubSystem { subvendor : 23246u16 , subdevice : 29073u16 , name : "Behold TV H7" } , SubSystem { subvendor : 23246u16 , subdevice : 29328u16 , name : "Behold TV T7" } , SubSystem { subvendor : 23246u16 , subdevice : 30097u16 , name : "Behold TV X7" } , SubSystem { subvendor : 23246u16 , subdevice : 30101u16 , name : "Behold TV X7" }] } , Device { vendor_id : 4401u16 , id : 28980u16 , name : "SAA7134/SAA7135HL Video Broadcast Decoder" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 16438u16 , name : "Behold TV 403" } , SubSystem { subvendor : 0u16 , subdevice : 16439u16 , name : "Behold TV 403 FM" } , SubSystem { subvendor : 0u16 , subdevice : 16497u16 , name : "Behold TV 407 FM" } , SubSystem { subvendor : 4121u16 , subdevice : 19636u16 , name : "Elitegroup ECS TVP3XP FM1216 Tuner Card(PAL-BG,FM)" } , SubSystem { subvendor : 4163u16 , subdevice : 528u16 , name : "Digimatrix TV" } , SubSystem { subvendor : 4163u16 , subdevice : 18496u16 , name : "TV-FM 7134" } , SubSystem { subvendor : 4163u16 , subdevice : 18498u16 , name : "TV-FM 7134" } , SubSystem { subvendor : 4401u16 , subdevice : 0u16 , name : "SAA713x-based TV tuner card" } , SubSystem { subvendor : 4401u16 , subdevice : 8196u16 , name : "EUROPA V3 reference design" } , SubSystem { subvendor : 4401u16 , subdevice : 20101u16 , name : "SKNet Monster TV" } , SubSystem { subvendor : 4401u16 , subdevice : 26450u16 , name : "EMPRESS" } , SubSystem { subvendor : 4541u16 , subdevice : 43u16 , name : "PCTV Stereo" } , SubSystem { subvendor : 4541u16 , subdevice : 45u16 , name : "PCTV 300i DVB-T + PAL" } , SubSystem { subvendor : 5217u16 , subdevice : 11264u16 , name : "AverTV Hybrid+FM PCI" } , SubSystem { subvendor : 5217u16 , subdevice : 38677u16 , name : "AVerTV Studio 307" } , SubSystem { subvendor : 5217u16 , subdevice : 42762u16 , name : "Avermedia AVerTV 307" } , SubSystem { subvendor : 5217u16 , subdevice : 42763u16 , name : "AverMedia M156 / Medion 2819" } , SubSystem { subvendor : 5217u16 , subdevice : 55022u16 , name : "Cardbus TV/Radio (E500)" } , SubSystem { subvendor : 5233u16 , subdevice : 47081u16 , name : "AVerTV Cardbus plus" } , SubSystem { subvendor : 5435u16 , subdevice : 4418u16 , name : "Terratec Cinergy 400 TV" } , SubSystem { subvendor : 5435u16 , subdevice : 4419u16 , name : "Terratec Cinergy 600 TV" } , SubSystem { subvendor : 5435u16 , subdevice : 4440u16 , name : "Terratec Cinergy 600 TV MK3" } , SubSystem { subvendor : 5440u16 , subdevice : 38180u16 , name : "ProVideo PV952" } , SubSystem { subvendor : 5822u16 , subdevice : 3u16 , name : "Medion 7134" } , SubSystem { subvendor : 6235u16 , subdevice : 49664u16 , name : "Compro VideoMate Gold+ Pal" } , SubSystem { subvendor : 6235u16 , subdevice : 51456u16 , name : "Videomate DVB-T300" } , SubSystem { subvendor : 6292u16 , subdevice : 40966u16 , name : "KNC One TV-Station DVR" } , SubSystem { subvendor : 6292u16 , subdevice : 65025u16 , name : "KNC One TV-Station RDS / Typhoon TV Tuner RDS" } , SubSystem { subvendor : 20840u16 , subdevice : 312u16 , name : "FLY TV PRIME 34FM" } , SubSystem { subvendor : 20840u16 , subdevice : 768u16 , name : "FlyDVB-S" } , SubSystem { subvendor : 23246u16 , subdevice : 20592u16 , name : "Behold TV 507 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24688u16 , name : "Behold TV 607 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24689u16 , name : "Behold TV 607 FM" } , SubSystem { subvendor : 23246u16 , subdevice : 24690u16 , name : "Behold TV 607 RDS" } , SubSystem { subvendor : 23246u16 , subdevice : 24691u16 , name : "Behold TV 607 RDS" }] } , Device { vendor_id : 4401u16 , id : 28997u16 , name : "SAA7145" , subsystems : & [] } , Device { vendor_id : 4401u16 , id : 28998u16 , name : "SAA7146" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 0u16 , name : "Fujitsu/Siemens DVB-C card rev1.5" } , SubSystem { subvendor : 4362u16 , subdevice : 65535u16 , name : "Fujitsu/Siemens DVB-C card rev1.5" } , SubSystem { subvendor : 4388u16 , subdevice : 9601u16 , name : "Leutron Vision PicPort" } , SubSystem { subvendor : 4401u16 , subdevice : 20310u16 , name : "KNC1 DVB-S Budget" } , SubSystem { subvendor : 4401u16 , subdevice : 20320u16 , name : "Fujitsu-Siemens Activy DVB-S Budget Rev AL" } , SubSystem { subvendor : 4401u16 , subdevice : 20321u16 , name : "Activy DVB-S Budget Rev GR" } , SubSystem { subvendor : 4401u16 , subdevice : 24417u16 , name : "Activy DVB-T Budget" } , SubSystem { subvendor : 4427u16 , subdevice : 8195u16 , name : "DVRaptor Video Edit/Capture Card" } , SubSystem { subvendor : 4441u16 , subdevice : 64u16 , name : "MuTech M-Vision 500 (MV-500 rev. E)" } , SubSystem { subvendor : 4441u16 , subdevice : 80u16 , name : "MuTech M-Vision 500 (MV-500 rev. F)" } , SubSystem { subvendor : 4541u16 , subdevice : 6u16 , name : "DV500 Overlay" } , SubSystem { subvendor : 4541u16 , subdevice : 10u16 , name : "DV500 Overlay" } , SubSystem { subvendor : 4541u16 , subdevice : 15u16 , name : "DV500 Overlay" } , SubSystem { subvendor : 5058u16 , subdevice : 0u16 , name : "Siemens/Technotrend/Hauppauge DVB card rev1.3 or rev1.5" } , SubSystem { subvendor : 5058u16 , subdevice : 1u16 , name : "Technotrend/Hauppauge DVB card rev1.3 or rev1.6" } , SubSystem { subvendor : 5058u16 , subdevice : 2u16 , name : "Technotrend/Hauppauge DVB card rev2.1" } , SubSystem { subvendor : 5058u16 , subdevice : 3u16 , name : "Technotrend/Hauppauge DVB card rev2.1" } , SubSystem { subvendor : 5058u16 , subdevice : 4u16 , name : "Technotrend/Hauppauge DVB card rev2.1" } , SubSystem { subvendor : 5058u16 , subdevice : 6u16 , name : "Technotrend/Hauppauge DVB card rev1.3 or rev1.6" } , SubSystem { subvendor : 5058u16 , subdevice : 8u16 , name : "Technotrend/Hauppauge DVB-T" } , SubSystem { subvendor : 5058u16 , subdevice : 10u16 , name : "Octal/Technotrend DVB-C for iTV" } , SubSystem { subvendor : 5058u16 , subdevice : 14u16 , name : "Technotrend/Hauppauge DVB card rev2.3" } , SubSystem { subvendor : 5058u16 , subdevice : 4099u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-S DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4100u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-C DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4101u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-T DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4108u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-CI DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4111u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-CI DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4112u16 , name : "DVB C-1500" } , SubSystem { subvendor : 5058u16 , subdevice : 4113u16 , name : "Technotrend-Budget/Hauppauge WinTV-NOVA-T DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4114u16 , name : "DVB T-1500" } , SubSystem { subvendor : 5058u16 , subdevice : 4115u16 , name : "SATELCO Multimedia DVB" } , SubSystem { subvendor : 5058u16 , subdevice : 4118u16 , name : "WinTV-NOVA-SE DVB card" } , SubSystem { subvendor : 5058u16 , subdevice : 4120u16 , name : "DVB S-1401" } , SubSystem { subvendor : 5058u16 , subdevice : 4121u16 , name : "S2-3200" } , SubSystem { subvendor : 5058u16 , subdevice : 4354u16 , name : "Technotrend/Hauppauge DVB card rev2.1" } , SubSystem { subvendor : 5435u16 , subdevice : 4437u16 , name : "Cinergy 1200 DVB-S" } , SubSystem { subvendor : 5435u16 , subdevice : 4438u16 , name : "Cinergy 1200 DVB-C" } , SubSystem { subvendor : 5435u16 , subdevice : 4439u16 , name : "Cinergy 1200 DVB-T" } , SubSystem { subvendor : 5435u16 , subdevice : 4470u16 , name : "Cinergy 1200 DVB-C (MK3)" } , SubSystem { subvendor : 6292u16 , subdevice : 32u16 , name : "KNC One DVB-C V1.0" } , SubSystem { subvendor : 6292u16 , subdevice : 35u16 , name : "TV-Station DVB-C plus" } , SubSystem { subvendor : 6292u16 , subdevice : 84u16 , name : "TV-Station DVB-S" }] } , Device { vendor_id : 4401u16 , id : 29024u16 , name : "SAA7160" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 36873u16 , name : "E8000 DVB-T/Analog TV/FM tuner" } , SubSystem { subvendor : 5217u16 , subdevice : 5205u16 , name : "AVerTV Hybrid Speedy PCI-E (H788)" }] } , Device { vendor_id : 4401u16 , id : 29026u16 , name : "SAA7162" , subsystems : & [SubSystem { subvendor : 4541u16 , subdevice : 257u16 , name : "Pinnacle PCTV 7010iX TV Card" }] } , Device { vendor_id : 4401u16 , id : 29028u16 , name : "SAA7164" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 34816u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34832u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34897u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34899u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34944u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34961u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34976u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 34977u16 , name : "WinTV HVR-2250" } , SubSystem { subvendor : 112u16 , subdevice : 35072u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35073u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35136u16 , name : "WinTV HVR-2200 (submodel 89619)" } , SubSystem { subvendor : 112u16 , subdevice : 35153u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35155u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35200u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35217u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35219u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35232u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 35233u16 , name : "WinTV HVR-2200" } , SubSystem { subvendor : 112u16 , subdevice : 61728u16 , name : "WinTV HVR-2205" } , SubSystem { subvendor : 112u16 , subdevice : 61731u16 , name : "WinTV HVR-2215" }] } , Device { vendor_id : 4401u16 , id : 29233u16 , name : "SAA7231" , subsystems : & [SubSystem { subvendor : 23246u16 , subdevice : 32768u16 , name : "Behold TV H8" } , SubSystem { subvendor : 23246u16 , subdevice : 32769u16 , name : "Behold TV H8" } , SubSystem { subvendor : 23246u16 , subdevice : 32848u16 , name : "Behold TV H85" } , SubSystem { subvendor : 23246u16 , subdevice : 32849u16 , name : "Behold TV H85" } , SubSystem { subvendor : 23246u16 , subdevice : 33024u16 , name : "Behold TV A8" } , SubSystem { subvendor : 23246u16 , subdevice : 33025u16 , name : "Behold TV A8" } , SubSystem { subvendor : 23246u16 , subdevice : 33104u16 , name : "Behold TV A85" } , SubSystem { subvendor : 23246u16 , subdevice : 33105u16 , name : "Behold TV A85" } , SubSystem { subvendor : 23246u16 , subdevice : 33281u16 , name : "Behold TV T8" }] } , Device { vendor_id : 4401u16 , id : 38704u16 , name : "SAA9730 Integrated Multimedia and Peripheral Controller" , subsystems : & [SubSystem { subvendor : 4401u16 , subdevice : 0u16 , name : "Integrated Multimedia and Peripheral Controller" }] }] }),
        (5624, Vendor { id : 5624u16 , name : "Koga Electronics Co" , devices : & [] }),
        (4820, Vendor { id : 4820u16 , name : "Ulticom (Formerly DGM&S)" , devices : & [Device { vendor_id : 4820u16 , id : 512u16 , name : "T1 Card" , subsystems : & [] }] }),
        (5025, Vendor { id : 5025u16 , name : "Kawasaki Heavy Industries Ltd" , devices : & [] }),
        (5770, Vendor { id : 5770u16 , name : "Utimaco IS GmbH" , devices : & [Device { vendor_id : 5770u16 , id : 8326u16 , name : "CryptoServer Se-Series Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 5770u16 , id : 49216u16 , name : "CryptoServer CSe-Series Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 5770u16 , id : 49233u16 , name : "CryptoServer Se-Series Gen2 Hardware Security Module" , subsystems : & [] } , Device { vendor_id : 5770u16 , id : 49264u16 , name : "u.trust Anchor Hardware Security Module cs7.2 Series" , subsystems : & [] } , Device { vendor_id : 5770u16 , id : 49265u16 , name : "u.trust Anchor Hardware Security Module cs7.3 Series" , subsystems : & [] } , Device { vendor_id : 5770u16 , id : 49266u16 , name : "u.trust Anchor Hardware Security Module cs7.3 Series Virtual Function" , subsystems : & [] }] }),
        (22597, Vendor { id : 22597u16 , name : "X-ES, Inc." , devices : & [] }),
        (4495, Vendor { id : 4495u16 , name : "Green Logic" , devices : & [] }),
        (4170, Vendor { id : 4170u16 , name : "STMicroelectronics" , devices : & [Device { vendor_id : 4170u16 , id : 0u16 , name : "STLS2F Host Bridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 8u16 , name : "STG 2000X" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 9u16 , name : "STG 1764X" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 16u16 , name : "STG4000 [3D Prophet Kyro Series]" , subsystems : & [SubSystem { subvendor : 4170u16 , subdevice : 16408u16 , name : "ST PowerVR Kyro (64MB AGP TVO)" } , SubSystem { subvendor : 5761u16 , subdevice : 16u16 , name : "PowerVR Kyro II [3D Prophet 4500]" } , SubSystem { subvendor : 5761u16 , subdevice : 40u16 , name : "3D Prophet 4000XT" } , SubSystem { subvendor : 5761u16 , subdevice : 49168u16 , name : "3D Prophet 4500 TV-Out" } , SubSystem { subvendor : 5761u16 , subdevice : 49257u16 , name : "3D Prophet 4000XT" }] } , Device { vendor_id : 4170u16 , id : 513u16 , name : "STPC Vega Northbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 521u16 , name : "STPC Consumer/Industrial North- and Southbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 522u16 , name : "STPC Atlas/ConsumerS/Consumer IIA Northbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 523u16 , name : "STPC Consumer II ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 528u16 , name : "STPC Atlas ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 538u16 , name : "STPC Consumer S Southbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 539u16 , name : "STPC Consumer IIA Southbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 544u16 , name : "STPC Industrial PCI to PCCard bridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 552u16 , name : "STPC Atlas IDE" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 553u16 , name : "STPC Vega IDE" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 560u16 , name : "STPC Atlas/Vega OHCI USB Controller" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 568u16 , name : "STPC Vega LAN" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 1280u16 , name : "ST70137 [Unicorn] ADSL DMT Transceiver" , subsystems : & [SubSystem { subvendor : 4170u16 , subdevice : 1280u16 , name : "BeWAN ADSL PCI st" }] } , Device { vendor_id : 4170u16 , id : 1380u16 , name : "STPC Client Northbridge" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 2433u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 5958u16 , name : "STG 1764X" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 10100u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 13600u16 , name : "MPEG-II decoder card" , subsystems : & [] } , Device { vendor_id : 4170u16 , id : 21964u16 , name : "STPC Client Southbridge" , subsystems : & [] }] }),
        (7091, Vendor { id : 7091u16 , name : "Bluecherry" , devices : & [Device { vendor_id : 7091u16 , id : 17156u16 , name : "BC-04120A MPEG4 4 port video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 17161u16 , name : "BC-08240A MPEG4 4 port video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 17168u16 , name : "BC-16480A MPEG4 16 port video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 19972u16 , name : "BC-04120A 4 port MPEG4 video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 19977u16 , name : "BC-08240A 8 port MPEG4 video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 19984u16 , name : "BC-16480A 16 port MPEG4 video encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 21252u16 , name : "BC-H04120A 4 port H.264 video and audio encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 21256u16 , name : "BC-H08240A 8 port H.264 video and audio encoder / decoder" , subsystems : & [] } , Device { vendor_id : 7091u16 , id : 21264u16 , name : "BC-H16480A 16 port H.264 video and audio encoder / decoder" , subsystems : & [] }] }),
        (5376, Vendor { id : 5376u16 , name : "DELTA Electronics, Inc" , devices : & [Device { vendor_id : 5376u16 , id : 4960u16 , name : "RTL81xx RealTek Ethernet" , subsystems : & [] }] }),
        (7891, Vendor { id : 7891u16 , name : "Yeston" , devices : & [] }),
        (7375, Vendor { id : 7375u16 , name : "Zoom Corporation" , devices : & [Device { vendor_id : 7375u16 , id : 1u16 , name : "TAC-2 Thunderbolt Audio Converter" , subsystems : & [] }] }),
        (7828, Vendor { id : 7828u16 , name : "Calian SED" , devices : & [] }),
        (5490, Vendor { id : 5490u16 , name : "Otis Elevator Company" , devices : & [] }),
        (4291, Vendor { id : 4291u16 , name : "Samsung Semiconductors, Inc." , devices : & [] }),
        (4315, Vendor { id : 4315u16 , name : "Rohm LSI Systems, Inc." , devices : & [] }),
        (4148, Vendor { id : 4148u16 , name : "Framatome Connectors USA Inc." , devices : & [] }),
        (34064, Vendor { id : 34064u16 , name : "Sietium Semiconductor Co., Ltd." , devices : & [Device { vendor_id : 34064u16 , id : 513u16 , name : "GenBu02 Series GPU" , subsystems : & [SubSystem { subvendor : 34064u16 , subdevice : 1u16 , name : "GB2062-PUB-LPDDR" } , SubSystem { subvendor : 34064u16 , subdevice : 2u16 , name : "GB2062-PCIe-C20" } , SubSystem { subvendor : 34064u16 , subdevice : 3u16 , name : "GB2062-PCIe-C41" } , SubSystem { subvendor : 34064u16 , subdevice : 4u16 , name : "GB2062-PCIe-HIEILP4" } , SubSystem { subvendor : 34064u16 , subdevice : 5u16 , name : "GB02-PCIe-C21" } , SubSystem { subvendor : 34064u16 , subdevice : 6u16 , name : "CQ2020-C2X Series" } , SubSystem { subvendor : 34064u16 , subdevice : 7u16 , name : "GB02-PCIe-C46-4HDMI" } , SubSystem { subvendor : 34064u16 , subdevice : 8u16 , name : "GB02-MXM-M60" } , SubSystem { subvendor : 34064u16 , subdevice : 9u16 , name : "GB02-PCIe-C46-4HDMI-2DP" } , SubSystem { subvendor : 34064u16 , subdevice : 10u16 , name : "GB02-MXM-M62-A-DP" } , SubSystem { subvendor : 34064u16 , subdevice : 11u16 , name : "GB02-PCIe-C42" } , SubSystem { subvendor : 34064u16 , subdevice : 12u16 , name : "GB02-PUB" } , SubSystem { subvendor : 34064u16 , subdevice : 13u16 , name : "GB02-PCIe-C46-6HDMI" } , SubSystem { subvendor : 34064u16 , subdevice : 14u16 , name : "GB02-MXM-M62-A-HDMI" } , SubSystem { subvendor : 34064u16 , subdevice : 16u16 , name : "GB02-MXM-M66-3HDMI-3DP" } , SubSystem { subvendor : 34064u16 , subdevice : 17u16 , name : "GB02-MXM-M66-2HDMI-4DP" } , SubSystem { subvendor : 34064u16 , subdevice : 18u16 , name : "GB02-PCIe-C25-HD" } , SubSystem { subvendor : 34064u16 , subdevice : 19u16 , name : "GB02-PCIe-C25-HH" } , SubSystem { subvendor : 34064u16 , subdevice : 20u16 , name : "GB2064-VPX-V40" } , SubSystem { subvendor : 34064u16 , subdevice : 513u16 , name : "GB2062-PUB-DDR" }] } , Device { vendor_id : 34064u16 , id : 769u16 , name : "GenBu03 Series GPU" , subsystems : & [] }] }),
        (43522, Vendor { id : 43522u16 , name : "iTuner" , devices : & [] }),
        (64183, Vendor { id : 64183u16 , name : "Fabric7 Systems, Inc." , devices : & [] }),
        (5103, Vendor { id : 5103u16 , name : "Coppercom Inc" , devices : & [] }),
        (4562, Vendor { id : 4562u16 , name : "Intercom Inc." , devices : & [] }),
        (5506, Vendor { id : 5506u16 , name : "Cytec Corp" , devices : & [] }),
        (4372, Vendor { id : 4372u16 , name : "Atmel Corporation" , devices : & [Device { vendor_id : 4372u16 , id : 1286u16 , name : "at76c506 802.11b Wireless Network Adaptor" , subsystems : & [] }] }),
        (5505, Vendor { id : 5505u16 , name : "SEH Computertechnik GmbH" , devices : & [] }),
        (6270, Vendor { id : 6270u16 , name : "ZyXEL Communications Corporation" , devices : & [] }),
        (41561, Vendor { id : 41561u16 , name : "Hewlett Packard" , devices : & [] }),
        (6442, Vendor { id : 6442u16 , name : "BiTMICRO Networks Inc." , devices : & [Device { vendor_id : 6442u16 , id : 8u16 , name : "RAMPART" , subsystems : & [] }] }),
        (4885, Vendor { id : 4885u16 , name : "Wavesat" , devices : & [] }),
        (4276, Vendor { id : 4276u16 , name : "STB Systems Inc" , devices : & [Device { vendor_id : 4276u16 , id : 6941u16 , name : "Velocity 128 3D" , subsystems : & [SubSystem { subvendor : 4276u16 , subdevice : 9086u16 , name : "Velocity 4400" }] }] }),
        (4424, Vendor { id : 4424u16 , name : "SysKonnect" , devices : & [Device { vendor_id : 4424u16 , id : 16384u16 , name : "FDDI Adapter" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45115u16 , name : "Netelligent 100 FDDI DAS Fibre SC" } , SubSystem { subvendor : 3601u16 , subdevice : 45116u16 , name : "Netelligent 100 FDDI SAS Fibre SC" } , SubSystem { subvendor : 3601u16 , subdevice : 45117u16 , name : "Netelligent 100 FDDI DAS UTP" } , SubSystem { subvendor : 3601u16 , subdevice : 45118u16 , name : "Netelligent 100 FDDI SAS UTP" } , SubSystem { subvendor : 3601u16 , subdevice : 45119u16 , name : "Netelligent 100 FDDI SAS Fibre MIC" } , SubSystem { subvendor : 4424u16 , subdevice : 21793u16 , name : "FDDI SK-5521 (SK-NET FDDI-UP)" } , SubSystem { subvendor : 4424u16 , subdevice : 21794u16 , name : "FDDI SK-5522 (SK-NET FDDI-UP DAS)" } , SubSystem { subvendor : 4424u16 , subdevice : 21825u16 , name : "FDDI SK-5541 (SK-NET FDDI-FP)" } , SubSystem { subvendor : 4424u16 , subdevice : 21827u16 , name : "FDDI SK-5543 (SK-NET FDDI-LP)" } , SubSystem { subvendor : 4424u16 , subdevice : 21828u16 , name : "FDDI SK-5544 (SK-NET FDDI-LP DAS)" } , SubSystem { subvendor : 4424u16 , subdevice : 22561u16 , name : "FDDI SK-5821 (SK-NET FDDI-UP64)" } , SubSystem { subvendor : 4424u16 , subdevice : 22562u16 , name : "FDDI SK-5822 (SK-NET FDDI-UP64 DAS)" } , SubSystem { subvendor : 4424u16 , subdevice : 22593u16 , name : "FDDI SK-5841 (SK-NET FDDI-FP64)" } , SubSystem { subvendor : 4424u16 , subdevice : 22595u16 , name : "FDDI SK-5843 (SK-NET FDDI-LP64)" } , SubSystem { subvendor : 4424u16 , subdevice : 22596u16 , name : "FDDI SK-5844 (SK-NET FDDI-LP64 DAS)" }] } , Device { vendor_id : 4424u16 , id : 16896u16 , name : "Token Ring adapter" , subsystems : & [] } , Device { vendor_id : 4424u16 , id : 17152u16 , name : "SK-9872 Gigabit Ethernet Server Adapter (SK-NET GE-ZX dual link)" , subsystems : & [SubSystem { subvendor : 4424u16 , subdevice : 38945u16 , name : "SK-9821 Gigabit Ethernet Server Adapter (SK-NET GE-T)" } , SubSystem { subvendor : 4424u16 , subdevice : 38946u16 , name : "SK-9822 Gigabit Ethernet Server Adapter (SK-NET GE-T dual link)" } , SubSystem { subvendor : 4424u16 , subdevice : 38977u16 , name : "SK-9841 Gigabit Ethernet Server Adapter (SK-NET GE-LX)" } , SubSystem { subvendor : 4424u16 , subdevice : 38978u16 , name : "SK-9842 Gigabit Ethernet Server Adapter (SK-NET GE-LX dual link)" } , SubSystem { subvendor : 4424u16 , subdevice : 38979u16 , name : "SK-9843 Gigabit Ethernet Server Adapter (SK-NET GE-SX)" } , SubSystem { subvendor : 4424u16 , subdevice : 38980u16 , name : "SK-9844 Gigabit Ethernet Server Adapter (SK-NET GE-SX dual link)" } , SubSystem { subvendor : 4424u16 , subdevice : 39009u16 , name : "SK-9861 Gigabit Ethernet Server Adapter (SK-NET GE-SX Volition)" } , SubSystem { subvendor : 4424u16 , subdevice : 39010u16 , name : "SK-9862 Gigabit Ethernet Server Adapter (SK-NET GE-SX Volition dual link)" } , SubSystem { subvendor : 4424u16 , subdevice : 39025u16 , name : "SK-9871 Gigabit Ethernet Server Adapter (SK-NET GE-ZX)" } , SubSystem { subvendor : 4424u16 , subdevice : 39026u16 , name : "SK-9872 Gigabit Ethernet Server Adapter (SK-NET GE-ZX dual link)" } , SubSystem { subvendor : 4697u16 , subdevice : 10608u16 , name : "AT-2970SX Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10609u16 , name : "AT-2970LX Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10610u16 , name : "AT-2970TX Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10611u16 , name : "AT-2971SX Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10612u16 , name : "AT-2971T Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10613u16 , name : "AT-2970SX/2SC Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10614u16 , name : "AT-2970LX/2SC Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10615u16 , name : "AT-2970TX/2TX Gigabit Ethernet Adapter" }] } , Device { vendor_id : 4424u16 , id : 17184u16 , name : "SK-98xx V2.0 Gigabit Ethernet Adapter [Marvell 88E8001]" , subsystems : & [SubSystem { subvendor : 4424u16 , subdevice : 289u16 , name : "Marvell RDK-8001 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 545u16 , name : "Marvell RDK-8002 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 801u16 , name : "Marvell RDK-8003 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 1057u16 , name : "Marvell RDK-8004 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 1569u16 , name : "Marvell RDK-8006 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 1825u16 , name : "Marvell RDK-8007 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 2081u16 , name : "Marvell RDK-8008 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 2337u16 , name : "Marvell RDK-8009 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 4385u16 , name : "Marvell RDK-8011 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 4641u16 , name : "Marvell RDK-8012 Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 12833u16 , name : "SK-9521 V2.0 10/100/1000Base-T Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20513u16 , name : "SK-9821 V2.0 Gigabit Ethernet 10/100/1000Base-T Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20545u16 , name : "SK-9841 V2.0 Gigabit Ethernet 1000Base-LX Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20547u16 , name : "SK-9843 V2.0 Gigabit Ethernet 1000Base-SX Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20561u16 , name : "SK-9851 V2.0 Gigabit Ethernet 1000Base-SX Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20577u16 , name : "SK-9861 V2.0 Gigabit Ethernet 1000Base-SX Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 20593u16 , name : "SK-9871 V2.0 Gigabit Ethernet 1000Base-ZX Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 38177u16 , name : "SK-9521 10/100/1000Base-T Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10518u16 , name : "AT-2916T" }] } , Device { vendor_id : 4424u16 , id : 17408u16 , name : "SK-9Dxx Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4424u16 , id : 17664u16 , name : "SK-9Mxx Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4424u16 , id : 36864u16 , name : "SK-9Sxx Gigabit Ethernet Server Adapter PCI-X [Marvell 88E8022]" , subsystems : & [SubSystem { subvendor : 4424u16 , subdevice : 8448u16 , name : "SK-9S21 10/100/1000Base-T Server Adapter, PCI-X, Copper RJ-45" } , SubSystem { subvendor : 4424u16 , subdevice : 8704u16 , name : "SK-9S22 10/100/1000Base-T Dual Port Server Adapter, PCI-X, 2 Copper RJ-45" } , SubSystem { subvendor : 4424u16 , subdevice : 8720u16 , name : "SK-9P22 10/100/1000 Base-T Dual Port PMC card" } , SubSystem { subvendor : 4424u16 , subdevice : 8736u16 , name : "TPMC-GBE-CO" } , SubSystem { subvendor : 4424u16 , subdevice : 33024u16 , name : "SK-9S81 1000Base-SX Server Adapter,PCI-X, Fiber SX/LC" } , SubSystem { subvendor : 4424u16 , subdevice : 33280u16 , name : "SK-9S82 1000Base-SX Dual Port Server Adapter, PCI-X, 2 Fiber SX/LC" } , SubSystem { subvendor : 4424u16 , subdevice : 33296u16 , name : "SK-9P82 1000 Base-SX Dual Port PMC card" } , SubSystem { subvendor : 4424u16 , subdevice : 33312u16 , name : "TPMC-GBE-FI" } , SubSystem { subvendor : 4424u16 , subdevice : 37120u16 , name : "SK-9S91 1000Base-LX Server Adapter,PCI-X, Fiber LX/LC" } , SubSystem { subvendor : 4424u16 , subdevice : 37376u16 , name : "SK-9S92 1000Base-LX Dual Port Server Adapter, PCI-X, 2 Fiber LX/LC" } , SubSystem { subvendor : 4697u16 , subdevice : 10611u16 , name : "AT-2971SX v2 Gigabit Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10612u16 , name : "AT-2971T v2 Gigabit Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 10616u16 , name : "AT-2971LX Gigabit Adapter" }] } , Device { vendor_id : 4424u16 , id : 38979u16 , name : "[Fujitsu] Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4424u16 , id : 40448u16 , name : "SK-9E21D 10/100/1000Base-T Adapter, Copper RJ-45" , subsystems : & [SubSystem { subvendor : 4424u16 , subdevice : 8448u16 , name : "SK-9E21 Server Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 8656u16 , name : "SK-9E21D 10/100/1000Base-T Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 8704u16 , name : "SK-9E22 Server Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 33024u16 , name : "SK-9E81 Server Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 33280u16 , name : "SK-9E82 Server Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 37120u16 , name : "SK-9E91 Server Adapter" } , SubSystem { subvendor : 4424u16 , subdevice : 37376u16 , name : "SK-9E92 Server Adapter" }] } , Device { vendor_id : 4424u16 , id : 40449u16 , name : "SK-9E21M 10/100/1000Base-T Adapter" , subsystems : & [] }] }),
        (4121, Vendor { id : 4121u16 , name : "Elitegroup Computer Systems" , devices : & [Device { vendor_id : 4121u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (5899, Vendor { id : 5899u16 , name : "NetOctave" , devices : & [Device { vendor_id : 5899u16 , id : 256u16 , name : "NSP2000-SSL crypto accelerator" , subsystems : & [] }] }),
        (7246, Vendor { id : 7246u16 , name : "TECHWAY" , devices : & [Device { vendor_id : 7246u16 , id : 28708u16 , name : "Default PCIe Endpoint ID" , subsystems : & [] }] }),
        (6827, Vendor { id : 6827u16 , name : "Silver Creations AG" , devices : & [Device { vendor_id : 6827u16 , id : 30544u16 , name : "Sceye 10L" , subsystems : & [] }] }),
        (8643, Vendor { id : 8643u16 , name : "21st Century Computer Corp." , devices : & [] }),
        (5163, Vendor { id : 5163u16 , name : "Radiolan" , devices : & [] }),
        (5034, Vendor { id : 5034u16 , name : "Broadband Networks Inc" , devices : & [] }),
        (4237, Vendor { id : 4237u16 , name : "Olicom" , devices : & [Device { vendor_id : 4237u16 , id : 1u16 , name : "Token-Ring 16/4 PCI Adapter (3136/3137)" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 2u16 , name : "16/4 Token Ring" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 4u16 , name : "RapidFire OC-3139/3140 Token-Ring 16/4 PCI Adapter" , subsystems : & [SubSystem { subvendor : 4237u16 , subdevice : 4u16 , name : "OC-3139/3140 RapidFire Token-Ring 16/4 Adapter" }] } , Device { vendor_id : 4237u16 , id : 5u16 , name : "GoCard 3250 Token-Ring 16/4 CardBus PC Card" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 6u16 , name : "OC-3530 RapidFire Token-Ring 100" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 7u16 , name : "RapidFire 3141 Token-Ring 16/4 PCI Fiber Adapter" , subsystems : & [SubSystem { subvendor : 4237u16 , subdevice : 7u16 , name : "OC-3141 RapidFire Token-Ring 16/4 Adapter" }] } , Device { vendor_id : 4237u16 , id : 8u16 , name : "RapidFire 3540 HSTR 100/16/4 PCI Adapter" , subsystems : & [SubSystem { subvendor : 4237u16 , subdevice : 8u16 , name : "OC-3540 RapidFire HSTR 100/16/4 Adapter" }] } , Device { vendor_id : 4237u16 , id : 17u16 , name : "OC-2315" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 18u16 , name : "OC-2325" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 19u16 , name : "OC-2183/2185" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 20u16 , name : "OC-2326" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 25u16 , name : "OC-2327/2250 10/100 Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4237u16 , subdevice : 22u16 , name : "OC-2327 Rapidfire 10/100 Ethernet Adapter" } , SubSystem { subvendor : 4237u16 , subdevice : 23u16 , name : "OC-2250 GoCard 10/100 Ethernet Adapter" }] } , Device { vendor_id : 4237u16 , id : 33u16 , name : "OC-6151/6152 [RapidFire ATM 155]" , subsystems : & [] } , Device { vendor_id : 4237u16 , id : 34u16 , name : "ATM Adapter" , subsystems : & [] }] }),
        (4733, Vendor { id : 4733u16 , name : "Vela Research" , devices : & [] }),
        (5004, Vendor { id : 5004u16 , name : "Silicon Reality" , devices : & [] }),
        (4578, Vendor { id : 4578u16 , name : "Samsung Information Systems America" , devices : & [] }),
        (4298, Vendor { id : 4298u16 , name : "Fujitsu Microelectr., Inc." , devices : & [] }),
        (5135, Vendor { id : 5135u16 , name : "Salient Systems Corp" , devices : & [] }),
        (4107, Vendor { id : 4107u16 , name : "National Semiconductor Corporation" , devices : & [Device { vendor_id : 4107u16 , id : 1u16 , name : "DP83810" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 2u16 , name : "87415/87560 IDE" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 14u16 , name : "87560 Legacy I/O" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 15u16 , name : "FireWire Controller" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 17u16 , name : "NS87560 National PCI System I/O" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 18u16 , name : "USB Controller" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 32u16 , name : "DP83815 (MacPhyter) Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin Network" } , SubSystem { subvendor : 4825u16 , subdevice : 12u16 , name : "Aculab E1/T1 PMXc cPCI carrier card" } , SubSystem { subvendor : 4997u16 , subdevice : 62225u16 , name : "FA311 / FA312 (FA311 with WoL HW)" } , SubSystem { subvendor : 4997u16 , subdevice : 62226u16 , name : "FA312 (rev. A1) Fast Ethernet PCI Adapter" }] } , Device { vendor_id : 4107u16 , id : 33u16 , name : "PC87200 PCI to ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 34u16 , name : "DP83820 10/100/1000 Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 18688u16 , name : "DGE-500T" } , SubSystem { subvendor : 4997u16 , subdevice : 25114u16 , name : "GA621" } , SubSystem { subvendor : 4997u16 , subdevice : 25130u16 , name : "GA622T" }] } , Device { vendor_id : 4107u16 , id : 40u16 , name : "Geode GX2 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 42u16 , name : "CS5535 South Bridge" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 43u16 , name : "CS5535 ISA bridge" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 45u16 , name : "CS5535 IDE" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 46u16 , name : "CS5535 Audio" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 47u16 , name : "CS5535 USB" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 48u16 , name : "Geode GX2 Graphics Processor" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 53u16 , name : "DP83065 [Saturn] 10/100/1000 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 1280u16 , name : "SCx200 Bridge" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 1281u16 , name : "SCx200 SMI" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 1282u16 , name : "SCx200, SC1100 IDE controller" , subsystems : & [SubSystem { subvendor : 4107u16 , subdevice : 1282u16 , name : "IDE Controller" }] } , Device { vendor_id : 4107u16 , id : 1283u16 , name : "SCx200, SC1100 Audio Controller" , subsystems : & [SubSystem { subvendor : 4107u16 , subdevice : 1283u16 , name : "XpressAudio controller" }] } , Device { vendor_id : 4107u16 , id : 1284u16 , name : "SCx200 Video" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 1285u16 , name : "SCx200 XBus" , subsystems : & [] } , Device { vendor_id : 4107u16 , id : 1296u16 , name : "SC1100 Bridge" , subsystems : & [SubSystem { subvendor : 4107u16 , subdevice : 1280u16 , name : "GPIO and LPC support bridge" }] } , Device { vendor_id : 4107u16 , id : 1297u16 , name : "SC1100 SMI & ACPI" , subsystems : & [SubSystem { subvendor : 4107u16 , subdevice : 1281u16 , name : "SC1100 SMI & ACPI bridge" }] } , Device { vendor_id : 4107u16 , id : 1301u16 , name : "SC1100 XBus" , subsystems : & [SubSystem { subvendor : 4107u16 , subdevice : 1285u16 , name : "SC1100 PCI to XBus bridge" }] } , Device { vendor_id : 4107u16 , id : 53249u16 , name : "87410 IDE" , subsystems : & [] }] }),
        (8284, Vendor { id : 8284u16 , name : "Zhejiang VMing Semiconductor Co., Ltd." , devices : & [Device { vendor_id : 8284u16 , id : 5396u16 , name : "EP9410 U.2 1.92TB NVME SSD" , subsystems : & [] } , Device { vendor_id : 8284u16 , id : 5397u16 , name : "EP9410 U.2 3.84TB NVME SSD" , subsystems : & [] } , Device { vendor_id : 8284u16 , id : 5428u16 , name : "EP9430 U.2 1.6TB NVME SSD" , subsystems : & [] } , Device { vendor_id : 8284u16 , id : 5429u16 , name : "EP9430 U.2 3.2TB NVME SSD" , subsystems : & [] } , Device { vendor_id : 8284u16 , id : 5430u16 , name : "EP9430 U.2 6.4TB NVME SSD" , subsystems : & [] }] }),
        (4590, Vendor { id : 4590u16 , name : "Dome Imaging Systems Inc" , devices : & [] }),
        (4605, Vendor { id : 4605u16 , name : "High Street Consultants" , devices : & [] }),
        (25609, Vendor { id : 25609u16 , name : "Logitec Corp." , devices : & [] }),
        (4683, Vendor { id : 4683u16 , name : "SBS/Greenspring Modular I/O" , devices : & [Device { vendor_id : 4683u16 , id : 64u16 , name : "PCI-40A or cPCI-200 Quad IndustryPack carrier" , subsystems : & [SubSystem { subvendor : 4683u16 , subdevice : 36992u16 , name : "PCI9080 Bridge" }] }] }),
        (44093, Vendor { id : 44093u16 , name : "Actuality Systems" , devices : & [] }),
        (7193, Vendor { id : 7193u16 , name : "LaCie" , devices : & [Device { vendor_id : 7193u16 , id : 3u16 , name : "6big and 12big Thunderbolt 3" , subsystems : & [SubSystem { subvendor : 7193u16 , subdevice : 3u16 , name : "6big Thunderbolt 3" } , SubSystem { subvendor : 7193u16 , subdevice : 4u16 , name : "12big Thunderbolt 3" }] }] }),
        (5394, Vendor { id : 5394u16 , name : "ROSUN Technologies Inc" , devices : & [] }),
        (789, Vendor { id : 789u16 , name : "SK-Electronics Co., Ltd." , devices : & [] }),
        (7629, Vendor { id : 7629u16 , name : "Liqid Inc." , devices : & [] }),
        (4633, Vendor { id : 4633u16 , name : "First Virtual Corporation" , devices : & [] }),
        (4567, Vendor { id : 4567u16 , name : "Trenton Technology, Inc." , devices : & [] }),
        (5367, Vendor { id : 5367u16 , name : "WOLF Technology Inc" , devices : & [] }),
        (5186, Vendor { id : 5186u16 , name : "Phoenix Contact GmbH & Co." , devices : & [] }),
        (6622, Vendor { id : 6622u16 , name : "Pico Computing" , devices : & [] }),
        (4514, Vendor { id : 4514u16 , name : "Sierra Research and Technology" , devices : & [] }),
        (7755, Vendor { id : 7755u16 , name : "MAXIO Technology (Hangzhou) Ltd." , devices : & [Device { vendor_id : 7755u16 , id : 4097u16 , name : "NVMe SSD Controller MAP1001" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 4098u16 , name : "NVMe SSD Controller MAP1002 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 4099u16 , name : "NVMe SSD Controller MAP1003" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 4609u16 , name : "NVMe SSD Controller MAP1201" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 4610u16 , name : "NVMe SSD Controller MAP1202 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 5633u16 , name : "NVMe SSD Controller MAP1601" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 5634u16 , name : "NVMe SSD Controller MAP1602 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7755u16 , id : 5640u16 , name : "NVMe SSD Controller MAP1608 (DRAM-less)" , subsystems : & [] }] }),
        (4550, Vendor { id : 4550u16 , name : "Dainippon Screen Mfg. Co. Ltd" , devices : & [] }),
        (4976, Vendor { id : 4976u16 , name : "ATL Products" , devices : & [] }),
        (19540, Vendor { id : 19540u16 , name : "Lisuan Technology Co., Ltd." , devices : & [Device { vendor_id : 19540u16 , id : 20480u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20481u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20482u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20483u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20484u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20485u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] } , Device { vendor_id : 19540u16 , id : 20486u16 , name : "LISUAN 7G100 Series Graphics" , subsystems : & [] }] }),
        (4187, Vendor { id : 4187u16 , name : "Foxconn International, Inc." , devices : & [Device { vendor_id : 4187u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57515u16 , name : "T99W175 5G Modem [Snapdragon X55]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57520u16 , name : "DW5930e 5G Modem [Snapdragon X55]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57521u16 , name : "DW5930e 5G Modem [Snapdragon X55]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57535u16 , name : "T99W175 5G Modem [Snapdragon X55]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57539u16 , name : "T99W175 5G Modem [Snapdragon X55]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57560u16 , name : "T99W368 5G Modem [Snapdragon X65]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57561u16 , name : "T99W373 5G Modem [Snapdragon X62]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57584u16 , name : "T99W510 4G Modem [Snapdragon X24]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57585u16 , name : "T99W510 4G Modem [Snapdragon X24]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57586u16 , name : "T99W510 4G Modem [Snapdragon X24]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57589u16 , name : "DW5932e-eSIM 5G Modem [Snapdragon X62]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57593u16 , name : "DW5932e 5G Modem [Snapdragon X62]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57624u16 , name : "T99W640 5G Modem [Snapdragon X72]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57629u16 , name : "DW5934e-eSIM 5G Modem [Snapdragon X72]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57630u16 , name : "DW5934e 5G Modem [Snapdragon X72]" , subsystems : & [] } , Device { vendor_id : 4187u16 , id : 57635u16 , name : "T99W760 5G Redcap Modem [Snapdragon X35]" , subsystems : & [] }] }),
        (4395, Vendor { id : 4395u16 , name : "Heidelberger Druckmaschinen AG" , devices : & [Device { vendor_id : 4395u16 , id : 1u16 , name : "SCU5" , subsystems : & [] }] }),
        (4575, Vendor { id : 4575u16 , name : "New Wave PDG" , devices : & [] }),
        (4568, Vendor { id : 4568u16 , name : "Image Technologies Development" , devices : & [] }),
        (58718, Vendor { id : 58718u16 , name : "Essence Technology, Inc." , devices : & [] }),
        (4441, Vendor { id : 4441u16 , name : "MuTech Corporation" , devices : & [Device { vendor_id : 4441u16 , id : 1u16 , name : "MV-1000" , subsystems : & [] } , Device { vendor_id : 4441u16 , id : 2u16 , name : "MV-1500" , subsystems : & [] }] }),
        (19796, Vendor { id : 19796u16 , name : "Microtechnica Co Ltd" , devices : & [] }),
        (4564, Vendor { id : 4564u16 , name : "Analog Devices" , devices : & [Device { vendor_id : 4564u16 , id : 5429u16 , name : "Blackfin BF535 processor" , subsystems : & [] } , Device { vendor_id : 4564u16 , id : 6149u16 , name : "SM56 PCI modem" , subsystems : & [] }] }),
        (36869, Vendor { id : 36869u16 , name : "Adaptec" , devices : & [Device { vendor_id : 36869u16 , id : 16u16 , name : "AHA-2940U2/U2W" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 8576u16 , name : "AHA-2940U2 SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 33024u16 , name : "AHA-2940U2B SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 41216u16 , name : "AHA-2940U2B SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 41344u16 , name : "AIC-3860Q [AHA-2940U2W/GE] SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 57600u16 , name : "AHA-2950U2B SCSI Controller" }] } , Device { vendor_id : 36869u16 , id : 17u16 , name : "AHA-2930U2" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 19u16 , name : "78902" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 3u16 , name : "AAA-131U2 Array1000 1 Channel RAID Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 15u16 , name : "AIC7890_ARO" }] } , Device { vendor_id : 36869u16 , id : 31u16 , name : "AHA-2940U2/U2W / 7890/7891" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 15u16 , name : "2940U2W SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 41344u16 , name : "2940U2W SCSI Controller" }] } , Device { vendor_id : 36869u16 , id : 32u16 , name : "AIC-7890" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 47u16 , name : "AIC-7890" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 48u16 , name : "AIC-7890" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 63u16 , name : "AIC-7890" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 80u16 , name : "AHA-3940U2x/395U2x" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 62720u16 , name : "AHA-3950U2B" } , SubSystem { subvendor : 36869u16 , subdevice : 65535u16 , name : "AHA-3950U2B" }] } , Device { vendor_id : 36869u16 , id : 81u16 , name : "AHA-3950U2D" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 46336u16 , name : "AHA-3950U2D" }] } , Device { vendor_id : 36869u16 , id : 83u16 , name : "AIC-7896 SCSI Controller" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 65535u16 , name : "AIC-7896 SCSI Controller mainboard implementation" }] } , Device { vendor_id : 36869u16 , id : 95u16 , name : "AIC-7896U2/7897U2" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 128u16 , name : "AIC-7892A U160/m" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 58016u16 , name : "Compaq 64-Bit/66MHz Wide Ultra3 SCSI Adapter" } , SubSystem { subvendor : 36869u16 , subdevice : 25120u16 , name : "AHA-29160C" } , SubSystem { subvendor : 36869u16 , subdevice : 25248u16 , name : "29160N Ultra160 SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 57888u16 , name : "29160LP Low Profile Ultra160 SCSI Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 58016u16 , name : "29160 Ultra160 SCSI Controller" }] } , Device { vendor_id : 36869u16 , id : 129u16 , name : "AIC-7892B U160/m" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 25249u16 , name : "19160 Ultra160 SCSI Controller" }] } , Device { vendor_id : 36869u16 , id : 131u16 , name : "AIC-7892D U160/m" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 143u16 , name : "AIC-7892P U160/m" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" } , SubSystem { subvendor : 5593u16 , subdevice : 36869u16 , name : "Onboard SCSI Host Adapter" }] } , Device { vendor_id : 36869u16 , id : 146u16 , name : "AVC-2010 [VideoH!]" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 147u16 , name : "AVC-2410 [VideoH!]" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 192u16 , name : "AHA-3960D / AIC-7899A U160/m" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 63008u16 , name : "Compaq 64-Bit/66MHz Dual Channel Wide Ultra3 SCSI Adapter" } , SubSystem { subvendor : 36869u16 , subdevice : 63008u16 , name : "AHA-3960D U160/m" }] } , Device { vendor_id : 36869u16 , id : 193u16 , name : "AIC-7899B U160/m" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 195u16 , name : "AIC-7899D U160/m" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 197u16 , name : "RAID subsystem HBA" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 197u16 , name : "PowerEdge 2400,2500,2550,4400" }] } , Device { vendor_id : 36869u16 , id : 207u16 , name : "AIC-7899P U160/m" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 206u16 , name : "PowerEdge 1400" } , SubSystem { subvendor : 4136u16 , subdevice : 209u16 , name : "PowerEdge 2550" } , SubSystem { subvendor : 4136u16 , subdevice : 217u16 , name : "PowerEdge 2500" } , SubSystem { subvendor : 4337u16 , subdevice : 9314u16 , name : "Thunder K7 S2462" } , SubSystem { subvendor : 5593u16 , subdevice : 36869u16 , name : "Onboard SCSI Host Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 13329u16 , name : "SDS2 Mainboard" }] } , Device { vendor_id : 36869u16 , id : 577u16 , name : "Serial ATA II RAID 1420SA" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 578u16 , name : "Serial ATA II RAID 1220SA" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 579u16 , name : "Serial ATA II RAID 1430SA" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 580u16 , name : "eSATA II RAID 1225SA" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 592u16 , name : "ServeRAID Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 633u16 , name : "ServeRAID 6M" } , SubSystem { subvendor : 4116u16 , subdevice : 652u16 , name : "ServeRAID 6i/6i+" } , SubSystem { subvendor : 4116u16 , subdevice : 654u16 , name : "ServeRAID 7k" }] } , Device { vendor_id : 36869u16 , id : 633u16 , name : "ServeRAID 6M" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 643u16 , name : "AAC-RAID" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 643u16 , name : "Catapult" }] } , Device { vendor_id : 36869u16 , id : 644u16 , name : "AAC-RAID" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 644u16 , name : "Tomcat" }] } , Device { vendor_id : 36869u16 , id : 645u16 , name : "AAC-RAID" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 661u16 , name : "SATA 6Ch (Bearcat)" } , SubSystem { subvendor : 4116u16 , subdevice : 754u16 , name : "ServeRAID 8i" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge Expandable RAID Controller 320/DC" } , SubSystem { subvendor : 4136u16 , subdevice : 657u16 , name : "CERC SATA RAID 2 PCI SATA 6ch (DellCorsair)" } , SubSystem { subvendor : 4156u16 , subdevice : 12839u16 , name : "AAR-2610SA" } , SubSystem { subvendor : 4238u16 , subdevice : 646u16 , name : "Sun StorageTek SAS RAID HBA, Internal" } , SubSystem { subvendor : 4238u16 , subdevice : 647u16 , name : "STK RAID EXT" } , SubSystem { subvendor : 4238u16 , subdevice : 31404u16 , name : "STK RAID REM" } , SubSystem { subvendor : 4238u16 , subdevice : 31406u16 , name : "STK RAID EX" } , SubSystem { subvendor : 5593u16 , subdevice : 693u16 , name : "SMC AOC-USAS-S4i" } , SubSystem { subvendor : 5593u16 , subdevice : 694u16 , name : "SMC AOC-USAS-S8i" } , SubSystem { subvendor : 5593u16 , subdevice : 713u16 , name : "SMC AOC-USAS-S4iR" } , SubSystem { subvendor : 5593u16 , subdevice : 714u16 , name : "SMC AOC-USAS-S8iR" } , SubSystem { subvendor : 5593u16 , subdevice : 722u16 , name : "SMC AOC-USAS-S8i-LP" } , SubSystem { subvendor : 5593u16 , subdevice : 723u16 , name : "SMC AOC-USAS-S8iR-LP" } , SubSystem { subvendor : 6058u16 , subdevice : 646u16 , name : "Legend S220 (Legend Crusader)" } , SubSystem { subvendor : 6058u16 , subdevice : 647u16 , name : "Legend S230 (Legend Vulcan)" } , SubSystem { subvendor : 36869u16 , subdevice : 645u16 , name : "2200S (Vulcan)" } , SubSystem { subvendor : 36869u16 , subdevice : 646u16 , name : "2120S (Crusader)" } , SubSystem { subvendor : 36869u16 , subdevice : 647u16 , name : "2200S (Vulcan-2m)" } , SubSystem { subvendor : 36869u16 , subdevice : 648u16 , name : "3230S (Harrier)" } , SubSystem { subvendor : 36869u16 , subdevice : 649u16 , name : "3240S (Tornado)" } , SubSystem { subvendor : 36869u16 , subdevice : 650u16 , name : "ASR-2020ZCR" } , SubSystem { subvendor : 36869u16 , subdevice : 651u16 , name : "ASR-2025ZCR (Terminator)" } , SubSystem { subvendor : 36869u16 , subdevice : 654u16 , name : "ASR-2020SA (Skyhawk)" } , SubSystem { subvendor : 36869u16 , subdevice : 655u16 , name : "ASR-2025SA" } , SubSystem { subvendor : 36869u16 , subdevice : 656u16 , name : "AAR-2410SA PCI SATA 4ch (Jaguar II)" } , SubSystem { subvendor : 36869u16 , subdevice : 658u16 , name : "AAR-2810SA PCI SATA 8ch (Corsair-8)" } , SubSystem { subvendor : 36869u16 , subdevice : 659u16 , name : "AAR-21610SA PCI SATA 16ch (Corsair-16)" } , SubSystem { subvendor : 36869u16 , subdevice : 660u16 , name : "ESD SO-DIMM PCI-X SATA ZCR (Prowler)" } , SubSystem { subvendor : 36869u16 , subdevice : 662u16 , name : "ASR-2240S" } , SubSystem { subvendor : 36869u16 , subdevice : 663u16 , name : "ASR-4005SAS" } , SubSystem { subvendor : 36869u16 , subdevice : 664u16 , name : "ASR-4000" } , SubSystem { subvendor : 36869u16 , subdevice : 665u16 , name : "ASR-4800SAS" } , SubSystem { subvendor : 36869u16 , subdevice : 666u16 , name : "4805SAS" } , SubSystem { subvendor : 36869u16 , subdevice : 676u16 , name : "ICP ICP9085LI" } , SubSystem { subvendor : 36869u16 , subdevice : 677u16 , name : "ICP ICP5085BR" } , SubSystem { subvendor : 36869u16 , subdevice : 693u16 , name : "ASR5800" } , SubSystem { subvendor : 36869u16 , subdevice : 694u16 , name : "ASR5805" } , SubSystem { subvendor : 36869u16 , subdevice : 695u16 , name : "ASR5808" } , SubSystem { subvendor : 36869u16 , subdevice : 696u16 , name : "ICP5445SL" } , SubSystem { subvendor : 36869u16 , subdevice : 697u16 , name : "ICP5085SL" } , SubSystem { subvendor : 36869u16 , subdevice : 698u16 , name : "ICP5805SL" } , SubSystem { subvendor : 36869u16 , subdevice : 699u16 , name : "3405" } , SubSystem { subvendor : 36869u16 , subdevice : 700u16 , name : "3805" } , SubSystem { subvendor : 36869u16 , subdevice : 701u16 , name : "31205" } , SubSystem { subvendor : 36869u16 , subdevice : 702u16 , name : "31605" } , SubSystem { subvendor : 36869u16 , subdevice : 703u16 , name : "ICP ICP5045BL" } , SubSystem { subvendor : 36869u16 , subdevice : 704u16 , name : "ICP ICP5085BL" } , SubSystem { subvendor : 36869u16 , subdevice : 705u16 , name : "ICP ICP5125BR" } , SubSystem { subvendor : 36869u16 , subdevice : 706u16 , name : "ICP ICP5165BR" } , SubSystem { subvendor : 36869u16 , subdevice : 707u16 , name : "51205" } , SubSystem { subvendor : 36869u16 , subdevice : 708u16 , name : "51605" } , SubSystem { subvendor : 36869u16 , subdevice : 709u16 , name : "ICP ICP5125SL" } , SubSystem { subvendor : 36869u16 , subdevice : 710u16 , name : "ICP ICP5165SL" } , SubSystem { subvendor : 36869u16 , subdevice : 711u16 , name : "3085" } , SubSystem { subvendor : 36869u16 , subdevice : 712u16 , name : "ICP5805BL" } , SubSystem { subvendor : 36869u16 , subdevice : 718u16 , name : "51245" } , SubSystem { subvendor : 36869u16 , subdevice : 719u16 , name : "51645" } , SubSystem { subvendor : 36869u16 , subdevice : 720u16 , name : "52445" } , SubSystem { subvendor : 36869u16 , subdevice : 721u16 , name : "5405" } , SubSystem { subvendor : 36869u16 , subdevice : 724u16 , name : "ASR-2045" } , SubSystem { subvendor : 36869u16 , subdevice : 725u16 , name : "ASR-2405" } , SubSystem { subvendor : 36869u16 , subdevice : 726u16 , name : "ASR-2445" } , SubSystem { subvendor : 36869u16 , subdevice : 727u16 , name : "ASR-2805" } , SubSystem { subvendor : 36869u16 , subdevice : 728u16 , name : "5405G" } , SubSystem { subvendor : 36869u16 , subdevice : 729u16 , name : "5445G" } , SubSystem { subvendor : 36869u16 , subdevice : 730u16 , name : "5805G" } , SubSystem { subvendor : 36869u16 , subdevice : 731u16 , name : "5085G" } , SubSystem { subvendor : 36869u16 , subdevice : 732u16 , name : "51245G" } , SubSystem { subvendor : 36869u16 , subdevice : 733u16 , name : "51645G" } , SubSystem { subvendor : 36869u16 , subdevice : 734u16 , name : "52445G" } , SubSystem { subvendor : 36869u16 , subdevice : 735u16 , name : "ASR-2045G" } , SubSystem { subvendor : 36869u16 , subdevice : 736u16 , name : "ASR-2405G" } , SubSystem { subvendor : 36869u16 , subdevice : 737u16 , name : "ASR-2445G" } , SubSystem { subvendor : 36869u16 , subdevice : 738u16 , name : "ASR-2805G" }] } , Device { vendor_id : 36869u16 , id : 646u16 , name : "AAC-RAID (Rocket)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 845u16 , name : "8s" } , SubSystem { subvendor : 4116u16 , subdevice : 38208u16 , name : "ServeRAID 8k/8k-l4" } , SubSystem { subvendor : 4116u16 , subdevice : 38272u16 , name : "ServeRAID 8k/8k-l8" } , SubSystem { subvendor : 36869u16 , subdevice : 652u16 , name : "ASR-2230S + ASR-2230SLP PCI-X (Lancer)" } , SubSystem { subvendor : 36869u16 , subdevice : 653u16 , name : "ASR-2130S" } , SubSystem { subvendor : 36869u16 , subdevice : 667u16 , name : "ASR-2820SA" } , SubSystem { subvendor : 36869u16 , subdevice : 668u16 , name : "ASR-2620SA" } , SubSystem { subvendor : 36869u16 , subdevice : 669u16 , name : "ASR-2420SA" } , SubSystem { subvendor : 36869u16 , subdevice : 670u16 , name : "ICP ICP9024R0" } , SubSystem { subvendor : 36869u16 , subdevice : 671u16 , name : "ICP ICP9014R0" } , SubSystem { subvendor : 36869u16 , subdevice : 672u16 , name : "ICP ICP9047MA" } , SubSystem { subvendor : 36869u16 , subdevice : 673u16 , name : "ICP ICP9087MA" } , SubSystem { subvendor : 36869u16 , subdevice : 674u16 , name : "3800" } , SubSystem { subvendor : 36869u16 , subdevice : 675u16 , name : "ICP ICP5445AU" } , SubSystem { subvendor : 36869u16 , subdevice : 676u16 , name : "ICP ICP9085LI" } , SubSystem { subvendor : 36869u16 , subdevice : 677u16 , name : "ICP ICP5085BR" } , SubSystem { subvendor : 36869u16 , subdevice : 678u16 , name : "ICP9067MA" } , SubSystem { subvendor : 36869u16 , subdevice : 679u16 , name : "3805" } , SubSystem { subvendor : 36869u16 , subdevice : 680u16 , name : "3400" } , SubSystem { subvendor : 36869u16 , subdevice : 681u16 , name : "ICP ICP5085AU" } , SubSystem { subvendor : 36869u16 , subdevice : 682u16 , name : "ICP ICP5045AU" } , SubSystem { subvendor : 36869u16 , subdevice : 684u16 , name : "1800" } , SubSystem { subvendor : 36869u16 , subdevice : 691u16 , name : "2400" } , SubSystem { subvendor : 36869u16 , subdevice : 692u16 , name : "ICP ICP5045AL" } , SubSystem { subvendor : 36869u16 , subdevice : 2048u16 , name : "Callisto" }] } , Device { vendor_id : 36869u16 , id : 651u16 , name : "Series 6 - 6G SAS/PCIe 2" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 512u16 , name : "Series 6 Entry Level - ASR-6405E - 4 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 513u16 , name : "Series 6 Entry Level - ASR-6805E - 8 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 768u16 , name : "Series 6 - ASR-6405 - 4 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 769u16 , name : "Series 6 - ASR-6805 - 8 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 770u16 , name : "Series 6 - ASR-6445 - 4 internal and 4 external 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 784u16 , name : "Series 6 Connectors on Top - ASR-6405T - 4 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 785u16 , name : "Series 6 Connectors on Top - ASR-6805T - 8 internal 6G SAS" } , SubSystem { subvendor : 36869u16 , subdevice : 1024u16 , name : "Series 6 - ASR-61205 - 12 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 1025u16 , name : "Series 6 - ASR-61605 - 16 internal 6G SAS ports" } , SubSystem { subvendor : 36869u16 , subdevice : 1027u16 , name : "Series 6 - ASR-62405 - 24 internal 6G SAS ports" }] } , Device { vendor_id : 36869u16 , id : 652u16 , name : "Series 7 6G SAS/PCIe 3" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1280u16 , name : "Series 7 - ASR-7805 - 8 internal 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1281u16 , name : "Series 7 - ASR-71605 - 16 internal 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1282u16 , name : "Series 7 - ASR-71685 - 16 internal 8 external 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1283u16 , name : "Series 7 - ASR-72405 - 24 internal 0 external 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1284u16 , name : "Series 7 - ASR-7885 - 8 internal 8 external 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1285u16 , name : "Series 7 Entry Level - ASR-71685E - 16 internal 8 external 6G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1286u16 , name : "Series 7 Entry Level - ASR-72405E - 24 internal 0 external 6G SAS Port/PCIe 3.0" }] } , Device { vendor_id : 36869u16 , id : 653u16 , name : "Series 8 12G SAS/PCIe 3" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1360u16 , name : "Series 8 - ASR-82405 - 24 internal 0 external 12G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1361u16 , name : "Series 8 - ASR-81605 - 16 internal 0 external 12G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1362u16 , name : "Series 8 - ASR-8805 - 8 internal 0 external 12G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1363u16 , name : "Series 8 - ASR-8085 - 0 internal 8 external 12G SAS Port/PCIe 3.0" } , SubSystem { subvendor : 36869u16 , subdevice : 1364u16 , name : "Series 8 - ASR-8885 - 8 internal 8 external 12G SAS Port/PCIe 3.0" }] } , Device { vendor_id : 36869u16 , id : 655u16 , name : "Smart Storage PQI SAS" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 1536u16 , name : "Smart Array P408i-p SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1537u16 , name : "Smart Array P408e-p SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1538u16 , name : "Smart Array P408i-a SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1539u16 , name : "Smart Array P408i-c SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1616u16 , name : "Smart Array E208i-p SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1617u16 , name : "Smart Array E208e-p SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1618u16 , name : "Smart Array E208i-c SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1620u16 , name : "Smart Array E208i-a SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1621u16 , name : "Smart Array P408e-m SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1792u16 , name : "Smart Array P204i-c SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 1793u16 , name : "Smart Array P204i-b SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 4352u16 , name : "Smart Array P816i-a SR Gen10" } , SubSystem { subvendor : 4156u16 , subdevice : 4353u16 , name : "Smart Array P416ie-m SR G10" } , SubSystem { subvendor : 4187u16 , subdevice : 4625u16 , name : "HBA 8238-16i" } , SubSystem { subvendor : 4187u16 , subdevice : 4897u16 , name : "HBA 8242-24i" } , SubSystem { subvendor : 4407u16 , subdevice : 760u16 , name : "24G TriMode M1 RAID 4GB FBWC 32D" } , SubSystem { subvendor : 4407u16 , subdevice : 761u16 , name : "24G TriMode M1 RAID 4GB FBWC 16D" } , SubSystem { subvendor : 4407u16 , subdevice : 762u16 , name : "24G TriMode M1 HBA 16D" } , SubSystem { subvendor : 5118u16 , subdevice : 33554u16 , name : "SKY-9200 MIC-8312BridgeB" } , SubSystem { subvendor : 5421u16 , subdevice : 35362u16 , name : "QS-8204-8i" } , SubSystem { subvendor : 5421u16 , subdevice : 35363u16 , name : "QS-8238-16i" } , SubSystem { subvendor : 5421u16 , subdevice : 35364u16 , name : "QS-8236-16i" } , SubSystem { subvendor : 5421u16 , subdevice : 35382u16 , name : "QS-8240-24i" } , SubSystem { subvendor : 5421u16 , subdevice : 35383u16 , name : "QS-8242-24i" } , SubSystem { subvendor : 5520u16 , subdevice : 660u16 , name : "SR932i-p Gen10+" } , SubSystem { subvendor : 5520u16 , subdevice : 732u16 , name : "SR416i-a Gen10+" } , SubSystem { subvendor : 6461u16 , subdevice : 4356u16 , name : "RAID P2404-Mf-4i-2GB" } , SubSystem { subvendor : 6461u16 , subdevice : 4357u16 , name : "RAID P4408-Mf-8i-2GB" } , SubSystem { subvendor : 6461u16 , subdevice : 4358u16 , name : "RAID P2404-Mf-4i-1GB" } , SubSystem { subvendor : 6461u16 , subdevice : 4359u16 , name : "RAID P4408-Mf-8i-4GB" } , SubSystem { subvendor : 6461u16 , subdevice : 33888u16 , name : "HBA H460-M1" } , SubSystem { subvendor : 6461u16 , subdevice : 33889u16 , name : "HBA H460-B1" } , SubSystem { subvendor : 6461u16 , subdevice : 50272u16 , name : "RAID P460-M2" } , SubSystem { subvendor : 6461u16 , subdevice : 50273u16 , name : "RAID P460-B2" } , SubSystem { subvendor : 6461u16 , subdevice : 62560u16 , name : "RAID P460-M4" } , SubSystem { subvendor : 6461u16 , subdevice : 62561u16 , name : "RAID P460-B4" } , SubSystem { subvendor : 6629u16 , subdevice : 53799u16 , name : "SmartROC-HD SR465C-M 4G" } , SubSystem { subvendor : 6629u16 , subdevice : 53800u16 , name : "SmartROC SR455C-M 2G" } , SubSystem { subvendor : 6629u16 , subdevice : 53801u16 , name : "SmartIOC SR155-M" } , SubSystem { subvendor : 6629u16 , subdevice : 53802u16 , name : "SmartIOC-HD SR765-M" } , SubSystem { subvendor : 6629u16 , subdevice : 53803u16 , name : "SmartROC-e SR455C-ME 4G" } , SubSystem { subvendor : 6629u16 , subdevice : 53804u16 , name : "SmartROC SR455C-M 4G" } , SubSystem { subvendor : 7124u16 , subdevice : 69u16 , name : "SMART-HBA 8242-24i" } , SubSystem { subvendor : 7124u16 , subdevice : 70u16 , name : "RAID 8236-16i" } , SubSystem { subvendor : 7124u16 , subdevice : 71u16 , name : "RAID 8240-24i" } , SubSystem { subvendor : 7124u16 , subdevice : 72u16 , name : "SMART-HBA 8238-16i" } , SubSystem { subvendor : 7124u16 , subdevice : 74u16 , name : "PM8222-SHBA" } , SubSystem { subvendor : 7124u16 , subdevice : 75u16 , name : "RAID PM8204-2GB" } , SubSystem { subvendor : 7124u16 , subdevice : 76u16 , name : "RAID PM8204-4GB" } , SubSystem { subvendor : 7124u16 , subdevice : 79u16 , name : "PM8222-HBA" } , SubSystem { subvendor : 7124u16 , subdevice : 107u16 , name : "RS0800M5H24I" } , SubSystem { subvendor : 7124u16 , subdevice : 108u16 , name : "RS0800M5E8i" } , SubSystem { subvendor : 7124u16 , subdevice : 109u16 , name : "RS0800M5H8i" } , SubSystem { subvendor : 7124u16 , subdevice : 111u16 , name : "RS0804M5R16i" } , SubSystem { subvendor : 7124u16 , subdevice : 112u16 , name : "RS0800M5E24i" } , SubSystem { subvendor : 7124u16 , subdevice : 113u16 , name : "RS0800M5H16i" } , SubSystem { subvendor : 7124u16 , subdevice : 114u16 , name : "RS0800M5E16i" } , SubSystem { subvendor : 7124u16 , subdevice : 119u16 , name : "RS0800M5E16iM" } , SubSystem { subvendor : 7124u16 , subdevice : 120u16 , name : "RS0800M5E24iM" } , SubSystem { subvendor : 7124u16 , subdevice : 121u16 , name : "RS0800M5H24iM" } , SubSystem { subvendor : 7124u16 , subdevice : 128u16 , name : "RS0804M5R16iM" } , SubSystem { subvendor : 7124u16 , subdevice : 134u16 , name : "RT0800M7E" } , SubSystem { subvendor : 7124u16 , subdevice : 135u16 , name : "RT0800M7H" } , SubSystem { subvendor : 7124u16 , subdevice : 136u16 , name : "RT0804M7R" } , SubSystem { subvendor : 7124u16 , subdevice : 137u16 , name : "RT0808M7R" } , SubSystem { subvendor : 7364u16 , subdevice : 257u16 , name : "Ramaxel FBGF-RAD PM8204" } , SubSystem { subvendor : 7364u16 , subdevice : 513u16 , name : "Ramaxel FBGF-RAD PM8222" } , SubSystem { subvendor : 7497u16 , subdevice : 544u16 , name : "ThinkSystem 4350-8i SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 545u16 , name : "ThinkSystem 4350-16i SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 7497u16 , subdevice : 1312u16 , name : "ThinkSystem RAID 5350-8i PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1314u16 , name : "ThinkSystem RAID 5350-8i PCIe 12Gb Internal Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1568u16 , name : "ThinkSystem RAID 9350-8i 2GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1569u16 , name : "ThinkSystem RAID 9350-8i 2GB Flash PCIe 12Gb Internal Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1570u16 , name : "ThinkSystem RAID 9350-16i 4GB Flash PCIe 12Gb Adapter" } , SubSystem { subvendor : 7497u16 , subdevice : 1571u16 , name : "ThinkSystem RAID 9350-16i 4GB Flash PCIe 12Gb Internal Adapter" } , SubSystem { subvendor : 7999u16 , subdevice : 1552u16 , name : "3S610-8i, SAS/SATA 12Gb HBA" } , SubSystem { subvendor : 8317u16 , subdevice : 16704u16 , name : "TrustRAID D3152s-8i" } , SubSystem { subvendor : 8317u16 , subdevice : 16960u16 , name : "TrustRAID D3154s-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 1544u16 , name : "SmartRAID 3162-8i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 2048u16 , name : "SmartRAID 3154-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2049u16 , name : "SmartRAID 3152-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2050u16 , name : "SmartRAID 3151-4i" } , SubSystem { subvendor : 36869u16 , subdevice : 2051u16 , name : "SmartRAID 3101-4i" } , SubSystem { subvendor : 36869u16 , subdevice : 2052u16 , name : "SmartRAID 3154-8e" } , SubSystem { subvendor : 36869u16 , subdevice : 2053u16 , name : "SmartRAID 3102-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2054u16 , name : "SmartRAID 3100" } , SubSystem { subvendor : 36869u16 , subdevice : 2055u16 , name : "SmartRAID 3162-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2056u16 , name : "SmartRAID 3101E-4i" } , SubSystem { subvendor : 36869u16 , subdevice : 2057u16 , name : "SmartRAID 3102E-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2304u16 , name : "SmartHBA 2100-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2305u16 , name : "SmartHBA 2100-4i" } , SubSystem { subvendor : 36869u16 , subdevice : 2306u16 , name : "HBA 1100-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 2307u16 , name : "HBA 1100-4i" } , SubSystem { subvendor : 36869u16 , subdevice : 2308u16 , name : "SmartHBA 2100-8e" } , SubSystem { subvendor : 36869u16 , subdevice : 2309u16 , name : "HBA 1100-8e" } , SubSystem { subvendor : 36869u16 , subdevice : 2310u16 , name : "SmartHBA 2100-4i4e" } , SubSystem { subvendor : 36869u16 , subdevice : 2311u16 , name : "HBA 1100" } , SubSystem { subvendor : 36869u16 , subdevice : 2312u16 , name : "SmartHBA 2100" } , SubSystem { subvendor : 36869u16 , subdevice : 2314u16 , name : "SmartHBA 2100A-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 4608u16 , name : "SmartRAID 3154-24i" } , SubSystem { subvendor : 36869u16 , subdevice : 4609u16 , name : "SmartRAID 3154-8i16e" } , SubSystem { subvendor : 36869u16 , subdevice : 4610u16 , name : "SmartRAID 3154-8i8e" } , SubSystem { subvendor : 36869u16 , subdevice : 4736u16 , name : "HBA 1100-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 4737u16 , name : "HBA 1100-16e" } , SubSystem { subvendor : 36869u16 , subdevice : 4738u16 , name : "SmartHBA 2100-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 4864u16 , name : "HBA 1100-8i8e" } , SubSystem { subvendor : 36869u16 , subdevice : 4865u16 , name : "HBA 1100-24i" } , SubSystem { subvendor : 36869u16 , subdevice : 4866u16 , name : "SmartHBA 2100-8i8e" } , SubSystem { subvendor : 36869u16 , subdevice : 4867u16 , name : "SmartHBA 2100-24i" } , SubSystem { subvendor : 36869u16 , subdevice : 4992u16 , name : "SmartRAID 3154-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 5120u16 , name : "SmartRAID Ultra 3258p-16i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5122u16 , name : "HBA Ultra 1200p-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 5136u16 , name : "HBA Ultra 1200-16e" } , SubSystem { subvendor : 36869u16 , subdevice : 5168u16 , name : "SmartRAID Ultra 3254-16e /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5185u16 , name : "HBA Ultra 1200p-32i" } , SubSystem { subvendor : 36869u16 , subdevice : 5200u16 , name : "SmartRAID Ultra 3258p-32i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5218u16 , name : "HBA 1200-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 5233u16 , name : "SmartRAID 3254-16i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5234u16 , name : "SmartRAID 3258-16i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5237u16 , name : "SmartRAID 3254-16e /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5280u16 , name : "SmartRAID 3254-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 5281u16 , name : "SmartRAID 3204-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 5282u16 , name : "SmartRAID 3252-8i" } , SubSystem { subvendor : 36869u16 , subdevice : 5284u16 , name : "SmartRAID 3254-8i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5285u16 , name : "SmartRAID 3252-8i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5286u16 , name : "SmartRAID 3204-8i /e" } , SubSystem { subvendor : 36869u16 , subdevice : 5312u16 , name : "SmartHBA 2200-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 5313u16 , name : "HBA 1200-16i" } , SubSystem { subvendor : 36869u16 , subdevice : 5315u16 , name : "HBA 1200-16e" } , SubSystem { subvendor : 36869u16 , subdevice : 5316u16 , name : "HBA 1200-8e" }] } , Device { vendor_id : 36869u16 , id : 1040u16 , name : "AIC-9410W SAS (Razor HBA RAID)" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1040u16 , name : "ASC-48300(Spirit RAID)" } , SubSystem { subvendor : 36869u16 , subdevice : 1041u16 , name : "ASC-58300 (Oakmont RAID)" }] } , Device { vendor_id : 36869u16 , id : 1042u16 , name : "AIC-9410W SAS (Razor HBA non-RAID)" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1042u16 , name : "ASC-48300 (Spirit non-RAID)" } , SubSystem { subvendor : 36869u16 , subdevice : 1043u16 , name : "ASC-58300 (Oakmont non-RAID)" }] } , Device { vendor_id : 36869u16 , id : 1045u16 , name : "ASC-58300 SAS (Razor-External HBA RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1046u16 , name : "ASC-58300 SAS (Razor-External HBA non-RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1054u16 , name : "AIC-9410W SAS (Razor ASIC non-RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1055u16 , name : "AIC-9410W SAS (Razor ASIC RAID)" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1055u16 , name : "AIC-9410W SAS (Razor ASIC RAID)" }] } , Device { vendor_id : 36869u16 , id : 1071u16 , name : "VSC7250/7251 SAS (Aurora ASIC non-RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1072u16 , name : "AIC-9405W SAS (Razor-Lite HBA RAID)" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1072u16 , name : "ASC-44300 (Spirit-Lite RAID)" }] } , Device { vendor_id : 36869u16 , id : 1074u16 , name : "AIC-9405W SAS (Razor-Lite HBA non-RAID)" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1074u16 , name : "ASC-44300 (Spirit-Lite non-RAID)" }] } , Device { vendor_id : 36869u16 , id : 1086u16 , name : "AIC-9405W SAS (Razor-Lite ASIC non-RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1087u16 , name : "AIC-9405W SAS (Razor-Lite ASIC RAID)" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1104u16 , name : "ASC-1405 Unified Serial HBA" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 1280u16 , name : "Obsidian chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 705u16 , name : "PCI-X DDR 3Gb SAS Adapter (572A/572C)" } , SubSystem { subvendor : 4116u16 , subdevice : 706u16 , name : "PCI-X DDR 3Gb SAS RAID Adapter (572B/572D)" }] } , Device { vendor_id : 36869u16 , id : 1283u16 , name : "Scamp chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 703u16 , name : "Quad Channel PCI-X DDR U320 SCSI RAID Adapter (571E)" } , SubSystem { subvendor : 4116u16 , subdevice : 707u16 , name : "PCI-X DDR 3Gb SAS RAID Adapter (572F)" } , SubSystem { subvendor : 4116u16 , subdevice : 725u16 , name : "Quad Channel PCI-X DDR U320 SCSI RAID Adapter (571F)" }] } , Device { vendor_id : 36869u16 , id : 2320u16 , name : "AUA-3100B" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 2334u16 , name : "AUA-3100B" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32768u16 , name : "ASC-29320A U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32783u16 , name : "AIC-7901 U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32784u16 , name : "ASC-39320 U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32785u16 , name : "ASC-39320D" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 172u16 , name : "ASC-39320D U320" } , SubSystem { subvendor : 36869u16 , subdevice : 65u16 , name : "ASC-39320D U320" }] } , Device { vendor_id : 36869u16 , id : 32786u16 , name : "ASC-29320 U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32787u16 , name : "ASC-29320B U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32788u16 , name : "ASC-29320LP U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32789u16 , name : "ASC-39320B U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32790u16 , name : "ASC-39320A U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32791u16 , name : "ASC-29320ALP U320" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 68u16 , name : "ASC-29320ALP PCIx U320" } , SubSystem { subvendor : 36869u16 , subdevice : 69u16 , name : "ASC-29320LPE PCIe U320" }] } , Device { vendor_id : 36869u16 , id : 32796u16 , name : "ASC-39320D U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32797u16 , name : "AIC-7902B U320" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 716u16 , name : "ServeRAID 7e" }] } , Device { vendor_id : 36869u16 , id : 32798u16 , name : "AIC-7901A U320" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32799u16 , name : "AIC-7902 U320" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4113u16 , name : "PRIMERGY RX300 onboard SCSI" }] } , Device { vendor_id : 36869u16 , id : 32896u16 , name : "ASC-29320A U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32897u16 , name : "PMC-Sierra PM8001 SAS HBA [Series 6H]" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 1024u16 , name : "Adaptec SAS HBA 6405H" } , SubSystem { subvendor : 36869u16 , subdevice : 2048u16 , name : "Adaptec SAS HBA 6805H" }] } , Device { vendor_id : 36869u16 , id : 32904u16 , name : "PMC-Sierra PM8018 SAS HBA [Series 7H]" , subsystems : & [SubSystem { subvendor : 36869u16 , subdevice : 8u16 , name : "Adaptec SAS HBA 7085H" } , SubSystem { subvendor : 36869u16 , subdevice : 22u16 , name : "Adaptec SAS HBA 70165H" } , SubSystem { subvendor : 36869u16 , subdevice : 2048u16 , name : "Adaptec SAS HBA 7805H" } , SubSystem { subvendor : 36869u16 , subdevice : 5632u16 , name : "Adaptec SAS HBA 71605H" }] } , Device { vendor_id : 36869u16 , id : 32905u16 , name : "PMC-Sierra PM8019 SAS encryption HBA [Series 7He]" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32911u16 , name : "AIC-7901 U320 w/HostRAID" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" }] } , Device { vendor_id : 36869u16 , id : 32912u16 , name : "ASC-39320 U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32913u16 , name : "ASC-39320D U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32914u16 , name : "ASC-29320 U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32915u16 , name : "ASC-29320B U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32916u16 , name : "ASC-29320LP U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32917u16 , name : "ASC-39320(B) U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32918u16 , name : "ASC-39320A U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32919u16 , name : "ASC-29320ALP U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32924u16 , name : "ASC-39320D(B) U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32925u16 , name : "AIC-7902(B) U320 w/HostRAID" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 716u16 , name : "ServeRAID 7e" }] } , Device { vendor_id : 36869u16 , id : 32926u16 , name : "AIC-7901A U320 w/HostRAID" , subsystems : & [] } , Device { vendor_id : 36869u16 , id : 32927u16 , name : "AIC-7902 U320 w/HostRAID" , subsystems : & [] }] }),
        (4518, Vendor { id : 4518u16 , name : "Pure Data Ltd." , devices : & [] }),
        (4656, Vendor { id : 4656u16 , name : "Fishcamp Engineering" , devices : & [] }),
        (5579, Vendor { id : 5579u16 , name : "AG Electronics Ltd" , devices : & [] }),
        (4647, Vendor { id : 4647u16 , name : "EIZO Rugged Solutions" , devices : & [Device { vendor_id : 4647u16 , id : 6u16 , name : "Raptor GFX 8P" , subsystems : & [] } , Device { vendor_id : 4647u16 , id : 35u16 , name : "Raptor GFX [1100T]" , subsystems : & [] } , Device { vendor_id : 4647u16 , id : 69u16 , name : "Raptor 4000-L [Linux version]" , subsystems : & [] } , Device { vendor_id : 4647u16 , id : 74u16 , name : "Raptor 4000-LR-L [Linux version]" , subsystems : & [] }] }),
        (7045, Vendor { id : 7045u16 , name : "OCZ Technology Group, Inc." , devices : & [Device { vendor_id : 7045u16 , id : 4129u16 , name : "RevoDrive 3 X2 PCIe SSD 240 GB (Marvell SAS Controller)" , subsystems : & [] } , Device { vendor_id : 7045u16 , id : 4161u16 , name : "RevoDrive 3 X2 PCIe SSD 240 GB (Marvell SAS Controller)" , subsystems : & [] } , Device { vendor_id : 7045u16 , id : 16408u16 , name : "Z Drive 6000/6300 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7045u16 , id : 24600u16 , name : "RD400/400A SSD" , subsystems : & [] } , Device { vendor_id : 7045u16 , id : 34696u16 , name : "RevoDrive Hybrid" , subsystems : & [] }] }),
        (5369, Vendor { id : 5369u16 , name : "AG COMMUNICATIONS" , devices : & [] }),
        (5373, Vendor { id : 5373u16 , name : "JAPAN Computer Industry Inc" , devices : & [] }),
        (4953, Vendor { id : 4953u16 , name : "Prisa Networks" , devices : & [] }),
        (4532, Vendor { id : 4532u16 , name : "Leitch Technology International" , devices : & [] }),
        (4847, Vendor { id : 4847u16 , name : "Vienna Systems" , devices : & [] }),
        (8093, Vendor { id : 8093u16 , name : "Axelera AI" , devices : & [Device { vendor_id : 8093u16 , id : 1u16 , name : "Europa AIPU" , subsystems : & [] } , Device { vendor_id : 8093u16 , id : 4352u16 , name : "Metis AIPU (rev 02)" , subsystems : & [] } , Device { vendor_id : 8093u16 , id : 4522u16 , name : "Metis AIPU (rev 01)" , subsystems : & [] }] }),
        (5639, Vendor { id : 5639u16 , name : "Lava Semiconductor Manufacturing Inc" , devices : & [] }),
        (4155, Vendor { id : 4155u16 , name : "Tatung Corp. Of America" , devices : & [] }),
        (5868, Vendor { id : 5868u16 , name : "U.S. Robotics" , devices : & [Device { vendor_id : 5868u16 , id : 237u16 , name : "USR997900" , subsystems : & [] } , Device { vendor_id : 5868u16 , id : 278u16 , name : "USR997902 10/100/1000 Mbps PCI Network Card" , subsystems : & [] } , Device { vendor_id : 5868u16 , id : 12032u16 , name : "USR5660A (USR265660A, USR5660A-BP) 56K PCI Faxmodem" , subsystems : & [] } , Device { vendor_id : 5868u16 , id : 13957u16 , name : "Wireless Access PCI Adapter Model 022415" , subsystems : & [] } , Device { vendor_id : 5868u16 , id : 17184u16 , name : "USR997904 10/100/1000 64-bit NIC (Marvell Yukon)" , subsystems : & [] } , Device { vendor_id : 5868u16 , id : 43782u16 , name : "USR997901A 10/100 Cardbus NIC" , subsystems : & [] }] }),
        (4295, Vendor { id : 4295u16 , name : "Media Vision" , devices : & [] }),
        (7447, Vendor { id : 7447u16 , name : "Zhaoxin" , devices : & [Device { vendor_id : 7447u16 , id : 1807u16 , name : "ZX-100 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1808u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1809u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1810u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1811u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1812u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1813u16 , name : "ZX-100/ZX-200 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1814u16 , name : "ZX-D PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1815u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1816u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1817u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1818u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1819u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1820u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1821u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1822u16 , name : "KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1823u16 , name : "ZX-200 Upstream Port of PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1824u16 , name : "ZX-200 PCIE RC6 controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1825u16 , name : "ZX-200 Downstream Port of PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1826u16 , name : "ZX-200 PCIE P2C bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1841u16 , name : "KH-40000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1842u16 , name : "KH-40000/KX-7000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1843u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1844u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1845u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1846u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1847u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1848u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1849u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1850u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1851u16 , name : "KX-7000 PCIE Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1852u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1853u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1854u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1855u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1856u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1857u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1858u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1859u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1860u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1861u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1862u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1863u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1864u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1865u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1866u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1867u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1868u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1869u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1870u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1871u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1872u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1873u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1874u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1879u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1880u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1881u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1882u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1883u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1884u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1885u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 1886u16 , name : "KH-50000 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4096u16 , name : "ZX-D Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4097u16 , name : "ZX-D/ZX-E/KH-40000/KX-7000/KH-50000 Miscellaneous Bus" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4099u16 , name : "ZX-E Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4101u16 , name : "KH-40000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4102u16 , name : "KX-6000G Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4103u16 , name : "KX-7000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 4104u16 , name : "KH-50000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12289u16 , name : "ZX-100 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12298u16 , name : "ZX-100 Miscellaneous Bus" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12344u16 , name : "ZX-100/ZX-200/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Standard Universal PCI to USB Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12548u16 , name : "ZX-100/ZX-200/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Standard Enhanced PCI to USB Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12720u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12721u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12722u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000 DRAM Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12723u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Power Management Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12724u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 I/O APIC" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12725u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Scratch Device" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12727u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 Standard Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12728u16 , name : "ZX-100/ZX-D PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12800u16 , name : "KX-7000 Host Bridge" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 12936u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 13403u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000 Miscellaneous Bus" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 14850u16 , name : "ZX-100 C-320 GPU" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 14851u16 , name : "ZX-D C-860 GPU" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 14852u16 , name : "KX-6000 C-960 GPU" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 14853u16 , name : "KX-7000 C-1190 GPU" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15105u16 , name : "KH-40000 OPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15106u16 , name : "KH-40000 OPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15107u16 , name : "KH-40000 OPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15108u16 , name : "KH-40000 OPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15360u16 , name : "KH-40000/KX-7000 DRAM Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15362u16 , name : "KX-6000G DRAM Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 15617u16 , name : "KX-6000G C-1080 GPU" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36866u16 , name : "ZX-100/ZX-200/KH-40000/KX-7000/KH-50000 EIDE Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36867u16 , name : "ZX-100/KX-6000/KX-6000G EIDE Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36931u16 , name : "KX-6000G/KH-40000/KX-7000/KH-50000 RAID Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36933u16 , name : "ZX-100/ZX-D/ZX-E RAID Accelerator 0" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36934u16 , name : "ZX-D/ZX-E RAID Accelerator 1" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36995u16 , name : "ZX-100/ZX-200/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 StorX AHCI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 36996u16 , name : "ZX-100 StorX AHCI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37120u16 , name : "ZX-200 Cross bus" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37121u16 , name : "ZX-200 Traffic Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37185u16 , name : "ZX-100 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37186u16 , name : "ZX-D High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37188u16 , name : "ZX-E High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37189u16 , name : "KX-6000G High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37190u16 , name : "KX-7000 High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37248u16 , name : "ZX-200 Networking Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37313u16 , name : "KH-40000 ZPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37314u16 , name : "KH-40000 ZPI Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37378u16 , name : "ZX-100 USB eXtensible Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37379u16 , name : "ZX-200 USB eXtensible Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37380u16 , name : "KX-6000/KX-6000G/KX-7000 USB3 xHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37381u16 , name : "KH-40000 USB eXtensible Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37382u16 , name : "KX-7000 USB4 Contoller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37383u16 , name : "KH-50000 USB eXtensible Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37510u16 , name : "ZX-D eMMC Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 37632u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 eSPI Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 38144u16 , name : "KX-7000 I2S Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 38145u16 , name : "KX-7000 I2S Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 38352u16 , name : "ZX-100 Universal SD Host Controller" , subsystems : & [] } , Device { vendor_id : 7447u16 , id : 62480u16 , name : "ZX-100/KX-5000/KX-6000/KX-6000G/KH-40000/KX-7000/KH-50000 PCI Com Port" , subsystems : & [] }] }),
        (5297, Vendor { id : 5297u16 , name : "Nextcom K.K." , devices : & [] }),
        (5166, Vendor { id : 5166u16 , name : "Vitec Multimedia" , devices : & [Device { vendor_id : 5166u16 , id : 16416u16 , name : "VM2-2 [Video Maker 2] MPEG1/2 Encoder" , subsystems : & [] } , Device { vendor_id : 5166u16 , id : 17207u16 , name : "VM2-2-C7 [Video Maker 2 rev. C7] MPEG1/2 Encoder" , subsystems : & [] }] }),
        (43534, Vendor { id : 43534u16 , name : "iTuner" , devices : & [] }),
        (5125, Vendor { id : 5125u16 , name : "Excalibur Systems Inc" , devices : & [] }),
        (5563, Vendor { id : 5563u16 , name : "Portwell Inc" , devices : & [] }),
        (5562, Vendor { id : 5562u16 , name : "Impacct Technology Corp" , devices : & [] }),
        (4723, Vendor { id : 4723u16 , name : "Hughes Network Systems" , devices : & [Device { vendor_id : 4723u16 , id : 2u16 , name : "DirecPC" , subsystems : & [] }] }),
        (4512, Vendor { id : 4512u16 , name : "Convex Computer Corporation" , devices : & [] }),
        (45054, Vendor { id : 45054u16 , name : "Sirrix AG security technologies" , devices : & [Device { vendor_id : 45054u16 , id : 481u16 , name : "PCI1E1 1-port ISDN E1 interface" , subsystems : & [] } , Device { vendor_id : 45054u16 , id : 737u16 , name : "PCI2E1 2-port ISDN E1 interface" , subsystems : & [] } , Device { vendor_id : 45054u16 , id : 17678u16 , name : "PCI4S0EC 4-port ISDN S0 interface" , subsystems : & [] } , Device { vendor_id : 45054u16 , id : 57005u16 , name : "Sirrix.PCI4S0 4-port ISDN S0 interface" , subsystems : & [] }] }),
        (5762, Vendor { id : 5762u16 , name : "XFX Pine Group Inc." , devices : & [] }),
        (5053, Vendor { id : 5053u16 , name : "Sharp corporation" , devices : & [] }),
        (41201, Vendor { id : 41201u16 , name : "UNISYS Corporation" , devices : & [] }),
        (5557, Vendor { id : 5557u16 , name : "Cimetrics Inc" , devices : & [] }),
        (5141, Vendor { id : 5141u16 , name : "Oxford Semiconductor Ltd" , devices : & [Device { vendor_id : 5141u16 , id : 33793u16 , name : "OX9162 Mode 1 (8-bit bus)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 33795u16 , name : "OX9162 Mode 0 (parallel port)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38144u16 , name : "OX16PCI954 (Quad 16950 UART) function 0 (Disabled)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38145u16 , name : "OX16PCI954 (Quad 16950 UART) function 0 (Uart)" , subsystems : & [SubSystem { subvendor : 4804u16 , subdevice : 513u16 , name : "Titan/cPCI (2 port)" } , SubSystem { subvendor : 4804u16 , subdevice : 514u16 , name : "Titan/cPCI (4 port)" } , SubSystem { subvendor : 4804u16 , subdevice : 515u16 , name : "Titan/cPCI (8 port)" } , SubSystem { subvendor : 4804u16 , subdevice : 528u16 , name : "Titan/104-Plus (8 port, p1-4)" } , SubSystem { subvendor : 4895u16 , subdevice : 8272u16 , name : "CyberPro (4-port)" } , SubSystem { subvendor : 4895u16 , subdevice : 8273u16 , name : "CyberSerial 4S Plus" } , SubSystem { subvendor : 5613u16 , subdevice : 8192u16 , name : "MCCR Serial p0-3 of 8" } , SubSystem { subvendor : 5613u16 , subdevice : 8193u16 , name : "MCCR Serial p0-3 of 16" }] } , Device { vendor_id : 5141u16 , id : 38149u16 , name : "OXuPCI952 (Dual 16C950 UART)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38154u16 , name : "EXSYS EX-41092 Dual 16950 Serial adapter" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38155u16 , name : "OXCB950 Cardbus 16950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38160u16 , name : "OX16PCI954 (Quad 16950 UART) function 1 (Disabled)" , subsystems : & [SubSystem { subvendor : 4804u16 , subdevice : 512u16 , name : "Titan/cPCI (Unused)" }] } , Device { vendor_id : 5141u16 , id : 38161u16 , name : "OX16PCI954 (Quad 16950 UART) function 1 (8bit bus)" , subsystems : & [SubSystem { subvendor : 4804u16 , subdevice : 529u16 , name : "Titan/104-Plus (8 port, p5-8)" } , SubSystem { subvendor : 5613u16 , subdevice : 8192u16 , name : "MCCR Serial p4-7 of 8" } , SubSystem { subvendor : 5613u16 , subdevice : 8193u16 , name : "MCCR Serial p4-15 of 16" }] } , Device { vendor_id : 5141u16 , id : 38162u16 , name : "OX16PCI954 (Quad 16950 UART) function 1 (32bit bus)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38163u16 , name : "OX16PCI954 (Quad 16950 UART) function 1 (parallel port)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38177u16 , name : "OX16PCI952 (Dual 16950 UART)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 38179u16 , name : "OX16PCI952 Integrated Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49152u16 , name : "OXPCIe840 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49156u16 , name : "OXPCIe840 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49158u16 , name : "OXPCIe840 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49408u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49409u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49412u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49413u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49414u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49416u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49417u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49420u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49421u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49422u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49424u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49428u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49432u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49435u16 , name : "OXPCIe952 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49436u16 , name : "OXPCIe952 Parallel Port" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49438u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49439u16 , name : "OXPCIe952 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49440u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49444u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49446u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49448u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49452u16 , name : "OXPCIe952 Legacy 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49454u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49460u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49464u16 , name : "OXPCIe952 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49468u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49469u16 , name : "OXPCIe952 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49472u16 , name : "OXPCIe952 Legacy 950 UART #1" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49473u16 , name : "OXPCIe952 Legacy 950 UART #2" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49476u16 , name : "OXPCIe952 Legacy 950 UART #1" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49477u16 , name : "OXPCIe952 Legacy 950 UART #2" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49478u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49480u16 , name : "OXPCIe952 Legacy 950 UART #1" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49481u16 , name : "OXPCIe952 Legacy 950 UART #2" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49484u16 , name : "OXPCIe952 Legacy 950 UART #1" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49485u16 , name : "OXPCIe952 Legacy 950 UART #2" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49486u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49492u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49496u16 , name : "OXPCIe952 Dual Native 950 UART" , subsystems : & [SubSystem { subvendor : 58559u16 , subdevice : 50436u16 , name : "CP4-SCAT Wireless Technologies Carrier Board" } , SubSystem { subvendor : 58559u16 , subdevice : 54609u16 , name : "DU1-MUSTANG Dual-Port RS-485 Interface" }] } , Device { vendor_id : 5141u16 , id : 49500u16 , name : "OXPCIe952 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49501u16 , name : "OXPCIe952 Dual Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49668u16 , name : "OXPCIe954 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49672u16 , name : "OXPCIe954 Quad Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49676u16 , name : "OXPCIe954 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49677u16 , name : "OXPCIe954 Quad Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49924u16 , name : "OXPCIe958 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49928u16 , name : "OXPCIe958 Quad Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49932u16 , name : "OXPCIe958 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 49933u16 , name : "OXPCIe958 Quad Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50480u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50481u16 , name : "OXPCIe200 Dual EHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50484u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50485u16 , name : "OXPCIe200 Dual EHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50486u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50488u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50489u16 , name : "OXPCIe200 Dual EHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50491u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50492u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50493u16 , name : "OXPCIe200 Dual EHCI USB Controller (ULPI/R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50494u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50495u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50496u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50497u16 , name : "OXPCIe200 Dual EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50500u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50501u16 , name : "OXPCIe200 Dual EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50502u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50504u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50505u16 , name : "OXPCIe200 Dual EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50507u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50508u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50509u16 , name : "OXPCIe200 Dual EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50510u16 , name : "OXPCIe200 Dual GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50511u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50528u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50529u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50532u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50533u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50534u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50536u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50537u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50539u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50540u16 , name : "OXPCIe200 Dual OHCI USB Controller (ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50541u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50542u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50543u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50544u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50545u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50548u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50549u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50550u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50552u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50553u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50555u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50556u16 , name : "OXPCIe200 Dual OHCI USB Controller (R-ULPI/analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50557u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50558u16 , name : "OXPCIe200 GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50559u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50592u16 , name : "OXPCIe200 OHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50593u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50594u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50596u16 , name : "OXPCIe200 OHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50597u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50598u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50600u16 , name : "OXPCIe200 OHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50601u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50602u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50603u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50604u16 , name : "OXPCIe200 OHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50605u16 , name : "OXPCIe200 EHCI USB Controller (ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50606u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50607u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50608u16 , name : "OXPCIe200 OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50609u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50610u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50612u16 , name : "OXPCIe200 OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50613u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50614u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50616u16 , name : "OXPCIe200 OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50617u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50618u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50619u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50620u16 , name : "OXPCIe200 OHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50621u16 , name : "OXPCIe200 EHCI USB Controller (R-ULPI)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50622u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50623u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50624u16 , name : "OXPCIe200 OHCI USB Controller (analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50626u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50628u16 , name : "OXPCIe200 OHCI USB Controller (analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50630u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50632u16 , name : "OXPCIe200 OHCI USB Controller (analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50634u16 , name : "OXPCIe200 Programmable Memory Interface" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50635u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50636u16 , name : "OXPCIe200 OHCI USB Controller (analog)" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50638u16 , name : "OXPCIe200 Programmable Memory Interface & GPIO" , subsystems : & [] } , Device { vendor_id : 5141u16 , id : 50639u16 , name : "OXPCIe200 Native 950 UART" , subsystems : & [] }] }),
        (5597, Vendor { id : 5597u16 , name : "Sigmatel Inc" , devices : & [] }),
        (5107, Vendor { id : 5107u16 , name : "Brocade Communications Systems, Inc." , devices : & [] }),
        (4945, Vendor { id : 4945u16 , name : "Sonix Inc" , devices : & [] }),
        (6750, Vendor { id : 6750u16 , name : "Aprius Inc." , devices : & [] }),
        (5425, Vendor { id : 5425u16 , name : "CHRYON Corp" , devices : & [] }),
        (7585, Vendor { id : 7585u16 , name : "Teko Telecom S.r.l." , devices : & [] }),
        (7457, Vendor { id : 7457u16 , name : "Allo" , devices : & [] }),
        (5547, Vendor { id : 5547u16 , name : "Bluesteel Networks Inc" , devices : & [] }),
        (1175, Vendor { id : 1175u16 , name : "Dell Inc. (wrong ID)" , devices : & [] }),
        (14575, Vendor { id : 14575u16 , name : "4Links" , devices : & [] }),
        (4525, Vendor { id : 4525u16 , name : "Lite-On Communications Inc" , devices : & [Device { vendor_id : 4525u16 , id : 2u16 , name : "LNE100TX" , subsystems : & [SubSystem { subvendor : 4525u16 , subdevice : 2u16 , name : "LNE100TX" } , SubSystem { subvendor : 4525u16 , subdevice : 3u16 , name : "LNE100TX" } , SubSystem { subvendor : 4525u16 , subdevice : 61443u16 , name : "LNE100TX" } , SubSystem { subvendor : 4525u16 , subdevice : 65535u16 , name : "LNE100TX" } , SubSystem { subvendor : 4997u16 , subdevice : 61444u16 , name : "FA310/TX LAN 10/100 PCI Ethernet Adapter" } , SubSystem { subvendor : 9798u16 , subdevice : 61442u16 , name : "KNE110TX EtheRx Fast Ethernet" }] } , Device { vendor_id : 4525u16 , id : 49429u16 , name : "LNE100TX [Linksys EtherFast 10/100]" , subsystems : & [SubSystem { subvendor : 4525u16 , subdevice : 49153u16 , name : "LNE100TX [ver 2.0]" } , SubSystem { subvendor : 9798u16 , subdevice : 11u16 , name : "KNE111TX" }] }] }),
        (5268, Vendor { id : 5268u16 , name : "WINTOP Technology, Inc." , devices : & [] }),
        (5048, Vendor { id : 5048u16 , name : "Nokia Telecommunications oy" , devices : & [] }),
        (5594, Vendor { id : 5594u16 , name : "Cyberfirm Inc" , devices : & [] }),
        (4634, Vendor { id : 4634u16 , name : "3Dfx Interactive, Inc." , devices : & [Device { vendor_id : 4634u16 , id : 1u16 , name : "Voodoo" , subsystems : & [] } , Device { vendor_id : 4634u16 , id : 2u16 , name : "Voodoo 2" , subsystems : & [] } , Device { vendor_id : 4634u16 , id : 3u16 , name : "Voodoo Banshee" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 3u16 , name : "Monster Fusion" } , SubSystem { subvendor : 4242u16 , subdevice : 16384u16 , name : "Monster Fusion" } , SubSystem { subvendor : 4242u16 , subdevice : 16386u16 , name : "Monster Fusion" } , SubSystem { subvendor : 4242u16 , subdevice : 18433u16 , name : "Monster Fusion AGP" } , SubSystem { subvendor : 4242u16 , subdevice : 18435u16 , name : "Monster Fusion AGP" } , SubSystem { subvendor : 4242u16 , subdevice : 32816u16 , name : "Monster Fusion" } , SubSystem { subvendor : 4242u16 , subdevice : 32821u16 , name : "Monster Fusion AGP" } , SubSystem { subvendor : 4272u16 , subdevice : 1u16 , name : "Dragon 4000" } , SubSystem { subvendor : 4354u16 , subdevice : 4119u16 , name : "3D Blaster Banshee PCI (CT6760)" } , SubSystem { subvendor : 4354u16 , subdevice : 4120u16 , name : "3D Blaster Banshee VE" } , SubSystem { subvendor : 4634u16 , subdevice : 1u16 , name : "Voodoo Banshee AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 3u16 , name : "Voodoo Banshee AGP SGRAM" } , SubSystem { subvendor : 4634u16 , subdevice : 4u16 , name : "Voodoo Banshee" } , SubSystem { subvendor : 5020u16 , subdevice : 22u16 , name : "Raven" } , SubSystem { subvendor : 5020u16 , subdevice : 23u16 , name : "Raven" } , SubSystem { subvendor : 5295u16 , subdevice : 2u16 , name : "Maxi Gamer Phoenix" }] } , Device { vendor_id : 4634u16 , id : 4u16 , name : "Voodoo Banshee [Velocity 100]" , subsystems : & [] } , Device { vendor_id : 4634u16 , id : 5u16 , name : "Voodoo 3" , subsystems : & [SubSystem { subvendor : 4634u16 , subdevice : 4u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 48u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 49u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 52u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 54u16 , name : "Voodoo3 2000 PCI" } , SubSystem { subvendor : 4634u16 , subdevice : 55u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 56u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 58u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 68u16 , name : "Voodoo3" } , SubSystem { subvendor : 4634u16 , subdevice : 75u16 , name : "Velocity 100" } , SubSystem { subvendor : 4634u16 , subdevice : 76u16 , name : "Velocity 200" } , SubSystem { subvendor : 4634u16 , subdevice : 77u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 78u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 81u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 82u16 , name : "Voodoo3 AGP" } , SubSystem { subvendor : 4634u16 , subdevice : 87u16 , name : "Voodoo3 3000 PCI" } , SubSystem { subvendor : 4634u16 , subdevice : 96u16 , name : "Voodoo3 3500 TV (NTSC)" } , SubSystem { subvendor : 4634u16 , subdevice : 97u16 , name : "Voodoo3 3500 TV (PAL)" } , SubSystem { subvendor : 4634u16 , subdevice : 98u16 , name : "Voodoo3 3500 TV (SECAM)" }] } , Device { vendor_id : 4634u16 , id : 9u16 , name : "Voodoo 4 / Voodoo 5" , subsystems : & [SubSystem { subvendor : 4634u16 , subdevice : 3u16 , name : "Voodoo5 PCI 5500" } , SubSystem { subvendor : 4634u16 , subdevice : 9u16 , name : "Voodoo5 AGP 5500/6000" }] } , Device { vendor_id : 4634u16 , id : 87u16 , name : "Voodoo 3/3000 [Avenger]" , subsystems : & [] }] }),
        (29808, Vendor { id : 29808u16 , name : "TP-LINK Technologies Co., Ltd." , devices : & [] }),
        (6469, Vendor { id : 6469u16 , name : "MERA" , devices : & [Device { vendor_id : 6469u16 , id : 25088u16 , name : "PXI/PXIe measurement module" , subsystems : & [] }] }),
        (4335, Vendor { id : 4335u16 , name : "Racore Computer Products, Inc." , devices : & [Device { vendor_id : 4335u16 , id : 33108u16 , name : "M815x Token Ring Adapter" , subsystems : & [] }] }),
        (1941, Vendor { id : 1941u16 , name : "Wired Inc." , devices : & [Device { vendor_id : 1941u16 , id : 26211u16 , name : "Butane II (MPEG2 encoder board)" , subsystems : & [] } , Device { vendor_id : 1941u16 , id : 26214u16 , name : "MediaPress (MPEG2 encoder board)" , subsystems : & [] }] }),
        (4850, Vendor { id : 4850u16 , name : "Gammagraphx, Inc." , devices : & [] }),
        (5360, Vendor { id : 5360u16 , name : "CANON RESEACH CENTRE FRANCE" , devices : & [] }),
        (4709, Vendor { id : 4709u16 , name : "Casio Computer Co., Ltd." , devices : & [] }),
        (7729, Vendor { id : 7729u16 , name : "SORD CORPORATION" , devices : & [] }),
        (5137, Vendor { id : 5137u16 , name : "Ikos Systems Inc" , devices : & [] }),
        (4357, Vendor { id : 4357u16 , name : "Sigma Designs, Inc." , devices : & [Device { vendor_id : 4357u16 , id : 4357u16 , name : "REALmagic Xcard MPEG 1/2/3/4 DVD Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33536u16 , name : "REALmagic Hollywood Plus DVD Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33792u16 , name : "EM840x REALmagic DVD/MPEG-2 Audio/Video Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33793u16 , name : "EM8401 REALmagic DVD/MPEG-2 A/V Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33904u16 , name : "EM8470 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33905u16 , name : "EM8471 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33909u16 , name : "EM8475 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [SubSystem { subvendor : 4357u16 , subdevice : 1u16 , name : "REALmagic X-Card" }] } , Device { vendor_id : 4357u16 , id : 33910u16 , name : "EM8476 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [SubSystem { subvendor : 4733u16 , subdevice : 0u16 , name : "CineView II" }] } , Device { vendor_id : 4357u16 , id : 33925u16 , name : "EM8485 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 33926u16 , name : "EM8486 REALmagic DVD/MPEG-4 A/V Decoder" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 50721u16 , name : "EM8621L Digital Media Processor" , subsystems : & [] } , Device { vendor_id : 4357u16 , id : 50722u16 , name : "EM8622L MPEG-4.10 (H.264) and SMPTE 421M (VC-1) A/V Decoder" , subsystems : & [] }] }),
        (4210, Vendor { id : 4210u16 , name : "GIT Co Ltd" , devices : & [] }),
        (4655, Vendor { id : 4655u16 , name : "Andrew Corporation" , devices : & [] }),
        (4665, Vendor { id : 4665u16 , name : "3DO Company" , devices : & [] }),
        (6203, Vendor { id : 6203u16 , name : "MikroM GmbH" , devices : & [Device { vendor_id : 6203u16 , id : 2215u16 , name : "MVC100 DVI" , subsystems : & [] } , Device { vendor_id : 6203u16 , id : 2216u16 , name : "MVC101 SDI" , subsystems : & [] } , Device { vendor_id : 6203u16 , id : 2217u16 , name : "MVC102 DVI+Audio" , subsystems : & [] } , Device { vendor_id : 6203u16 , id : 2224u16 , name : "MVC200-DC" , subsystems : & [] }] }),
        (5110, Vendor { id : 5110u16 , name : "C-Media Electronics Inc" , devices : & [Device { vendor_id : 5110u16 , id : 17u16 , name : "CMI8738" , subsystems : & [] } , Device { vendor_id : 5110u16 , id : 256u16 , name : "CM8338A" , subsystems : & [SubSystem { subvendor : 5110u16 , subdevice : 65535u16 , name : "CMI8338/C3DX PCI Audio Device" }] } , Device { vendor_id : 5110u16 , id : 257u16 , name : "CM8338B" , subsystems : & [SubSystem { subvendor : 5110u16 , subdevice : 257u16 , name : "CMI8338-031 PCI Audio Device" }] } , Device { vendor_id : 5110u16 , id : 273u16 , name : "CMI8738/CMI8768 PCI Audio" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2416u16 , name : "P6STP-FL motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32821u16 , name : "CUSI-FX motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32887u16 , name : "CMI8738 6-channel audio controller" } , SubSystem { subvendor : 4163u16 , subdevice : 32994u16 , name : "CMI8738 6ch-MX" } , SubSystem { subvendor : 5110u16 , subdevice : 273u16 , name : "CMI8738/C3DX PCI Audio Device" } , SubSystem { subvendor : 5110u16 , subdevice : 38753u16 , name : "Theatron Agrippa" } , SubSystem { subvendor : 5435u16 , subdevice : 4420u16 , name : "Aureon 5.1" } , SubSystem { subvendor : 5435u16 , subdevice : 4464u16 , name : "Aureon 7.1" } , SubSystem { subvendor : 5761u16 , subdevice : 40960u16 , name : "Gamesurround MUSE XL" } , SubSystem { subvendor : 6059u16 , subdevice : 1540u16 , name : "PSC604 Dynamic Edge" } , SubSystem { subvendor : 6059u16 , subdevice : 1541u16 , name : "PSC605 Sonic Edge" } , SubSystem { subvendor : 6059u16 , subdevice : 30583u16 , name : "PSC605 Sonic Edge" } , SubSystem { subvendor : 9999u16 , subdevice : 4355u16 , name : "CT-7NJS Ultra motherboard" } , SubSystem { subvendor : 9999u16 , subdevice : 62562u16 , name : "7NJL1 motherboard" } , SubSystem { subvendor : 22605u16 , subdevice : 14129u16 , name : "Digital X-Mystique" } , SubSystem { subvendor : 22605u16 , subdevice : 14145u16 , name : "X-Plosion 7.1" } , SubSystem { subvendor : 22605u16 , subdevice : 14161u16 , name : "X-Raider 7.1" } , SubSystem { subvendor : 22605u16 , subdevice : 14177u16 , name : "X-Mystique 7.1 LP" } , SubSystem { subvendor : 22605u16 , subdevice : 14193u16 , name : "X-Mystique 7.1 LP Value" } , SubSystem { subvendor : 29316u16 , subdevice : 33668u16 , name : "Striker 7.1" }] } , Device { vendor_id : 5110u16 , id : 529u16 , name : "CM8738" , subsystems : & [] } , Device { vendor_id : 5110u16 , id : 20497u16 , name : "CM8888 [Oxygen Express]" , subsystems : & [SubSystem { subvendor : 5110u16 , subdevice : 20497u16 , name : "HDA Controller" }] } , Device { vendor_id : 5110u16 , id : 34696u16 , name : "CMI8788 [Oxygen HD Audio]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33385u16 , name : "Virtuoso 200 (Xonar D2)" } , SubSystem { subvendor : 4163u16 , subdevice : 33397u16 , name : "Virtuoso 100 (Xonar DX)" } , SubSystem { subvendor : 4163u16 , subdevice : 33463u16 , name : "Virtuoso 200 (Xonar D2X)" } , SubSystem { subvendor : 4163u16 , subdevice : 33556u16 , name : "Virtuoso 200 (Xonar HDAV1.3)" } , SubSystem { subvendor : 4163u16 , subdevice : 33575u16 , name : "Virtuoso 100 (Xonar DX)" } , SubSystem { subvendor : 4163u16 , subdevice : 33615u16 , name : "Virtuoso 100 (Xonar D1)" } , SubSystem { subvendor : 4163u16 , subdevice : 33628u16 , name : "Virtuoso 100 (Xonar Essence STX)" } , SubSystem { subvendor : 4163u16 , subdevice : 33629u16 , name : "Virtuoso 100 (Xonar ST)" } , SubSystem { subvendor : 4163u16 , subdevice : 33630u16 , name : "Virtuoso 200 (Xonar HDAV1.3 Slim)" } , SubSystem { subvendor : 4163u16 , subdevice : 33678u16 , name : "Virtuoso 66 (Xonar DS)" } , SubSystem { subvendor : 4163u16 , subdevice : 33832u16 , name : "Virtuoso 100 (Xonar Xense)" } , SubSystem { subvendor : 4163u16 , subdevice : 33895u16 , name : "CMI8786 (Xonar DG)" } , SubSystem { subvendor : 4163u16 , subdevice : 34081u16 , name : "CMI8786 (Xonar DGX)" } , SubSystem { subvendor : 4163u16 , subdevice : 34082u16 , name : "Xonar DSX" } , SubSystem { subvendor : 4163u16 , subdevice : 34292u16 , name : "Virtuoso 100 (Xonar Essence STX II)" } , SubSystem { subvendor : 5110u16 , subdevice : 34690u16 , name : "PCI 2.0 HD Audio" } , SubSystem { subvendor : 5110u16 , subdevice : 65535u16 , name : "CMI8787-HG2PCI" } , SubSystem { subvendor : 5315u16 , subdevice : 5904u16 , name : "HiFier Fantasia" } , SubSystem { subvendor : 5315u16 , subdevice : 5905u16 , name : "HiFier Serenade" } , SubSystem { subvendor : 5315u16 , subdevice : 5907u16 , name : "HiFier Serenade III" } , SubSystem { subvendor : 6744u16 , subdevice : 2320u16 , name : "Barracuda AC-1" } , SubSystem { subvendor : 16730u16 , subdevice : 21553u16 , name : "X-Meridian 7.1" } , SubSystem { subvendor : 21553u16 , subdevice : 378u16 , name : "X-Meridian 7.1 2G" } , SubSystem { subvendor : 22605u16 , subdevice : 14209u16 , name : "HDA X-Purity 7.1 Platinum" } , SubSystem { subvendor : 29316u16 , subdevice : 38753u16 , name : "CLARO" } , SubSystem { subvendor : 29316u16 , subdevice : 38785u16 , name : "CLARO halo" } , SubSystem { subvendor : 29316u16 , subdevice : 38787u16 , name : "eCLARO" } , SubSystem { subvendor : 29316u16 , subdevice : 38791u16 , name : "CLARO II" }] }] }),
        (4651, Vendor { id : 4651u16 , name : "LG Industrial Systems Co., Ltd" , devices : & [] }),
        (5068, Vendor { id : 5068u16 , name : "BARCO" , devices : & [] }),
        (4365, Vendor { id : 4365u16 , name : "Znyx Advanced Systems" , devices : & [] }),
        (4903, Vendor { id : 4903u16 , name : "Voss scientific" , devices : & [] }),
        (5290, Vendor { id : 5290u16 , name : "Advanced MOS Technology Inc" , devices : & [] }),
        (8007, Vendor { id : 8007u16 , name : "YUSUR Technology" , devices : & [Device { vendor_id : 8007u16 , id : 4097u16 , name : "K2-Pro Family [FLEXFLOW-2200T]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "FLEXFLOW-2200T 2*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "FLEXFLOW-2200T 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "FLEXFLOW-2200T 2*40GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 100G 1P FLEXFLOW-2200T" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "FLEXFLOW-2200T 2*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 10G 2P FLEXFLOW-2200T" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 25G 2P FLEXFLOW-2200T" } , SubSystem { subvendor : 8007u16 , subdevice : 8u16 , name : "Ethernet 40G 2P FLEXFLOW-2200T" } , SubSystem { subvendor : 8007u16 , subdevice : 9u16 , name : "Ethernet 100G 1P FLEXFLOW-2200T" } , SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "Ethernet 100G 2P FLEXFLOW-2200T" }] } , Device { vendor_id : 8007u16 , id : 4098u16 , name : "K2-Pro Family [FLEXFLOW-2200T Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4099u16 , name : "K2-Pro Family [FLEXFLOW-2200T MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4100u16 , name : "K2-Pro Family [FLEXFLOW-2200T DATA Offload Engine]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4101u16 , name : "CONFLUX-2200P NVMe Controller" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4113u16 , name : "K3 Family" , subsystems : & [SubSystem { subvendor : 7631u16 , subdevice : 867u16 , name : "K3_F2_10G_PCIE 2*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "F31TX2S 2*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "F31TX4S 4*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "F31TA2S 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "F31TA4S 4*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "FLEXFLOW-3100T 1*40GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "F31TC1S 1*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "F31TA2K 2*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 8u16 , name : "FLEXFLOW-3100T 4*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 9u16 , name : "F31TA2K 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "FLEXFLOW-3100T 4*25GE Ethernet Adapter" }] } , Device { vendor_id : 8007u16 , id : 4114u16 , name : "K3 Family [Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4115u16 , name : "K3 Family [MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 4357u16 , name : "CONFLUX-2200P NVMe Controller [Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 8216u16 , name : "DPU Card" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 8224u16 , name : "DPU" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 12305u16 , name : "K3 Family [FLEXFLOW-3100R]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "FLEXFLOW-3100R 1*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 11u16 , name : "FLEXFLOW-3100R 2*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 12u16 , name : "FLEXFLOW-3100R 4*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 13u16 , name : "FLEXFLOW-3100R 8*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 14u16 , name : "FLEXFLOW-3100R 1*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 15u16 , name : "FLEXFLOW-3100R 2*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 16u16 , name : "FLEXFLOW-3100R 4*100GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 17u16 , name : "FLEXFLOW-3100R 8*100GE Ethernet Adapter" }] } , Device { vendor_id : 8007u16 , id : 12306u16 , name : "K3 Family [FLEXFLOW-3100R Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 12307u16 , name : "K3 Family [FLEXFLOW-3100R MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 12545u16 , name : "FLEXFLOW-2100R Ethernet Controller" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "Ethernet 10G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "Ethernet 25G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "Ethernet 50G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 100G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "Ethernet 10G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 25G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 50G 2P FLEXFLOW-2100R" } , SubSystem { subvendor : 8007u16 , subdevice : 8u16 , name : "Ethernet 100G 2P FLEXFLOW-2100R" }] } , Device { vendor_id : 8007u16 , id : 12801u16 , name : "FLEXFLOW-2200R Ethernet Controller" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "Ethernet 10G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "Ethernet 25G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "Ethernet 50G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 100G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "Ethernet 10G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 25G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 50G 2P FLEXFLOW-2200R" } , SubSystem { subvendor : 8007u16 , subdevice : 8u16 , name : "Ethernet 100G 2P FLEXFLOW-2200R" }] } , Device { vendor_id : 8007u16 , id : 16385u16 , name : "K2-Pro Family [CONFLUX-2200E]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "Ethernet 25G 2P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "Ethernet 40G 2P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "Ethernet 100G 1P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 100G 2P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "CONFLUX-2200E 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 40G 2P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 100G 1P CONFLUX-2200E" } , SubSystem { subvendor : 8007u16 , subdevice : 8u16 , name : "CONFLUX-2200E 2*100GE Ethernet Adapter" }] } , Device { vendor_id : 8007u16 , id : 16386u16 , name : "K2-Pro Family [CONFLUX-2200E Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 16387u16 , name : "K2-Pro Family [CONFLUX-2200E MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 16388u16 , name : "K2-Pro Family [CONFLUX-2200E DATA Offload Engine]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 16401u16 , name : "K3 Family [CONFLUX-3100E]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "CONFLUX-3100E 2*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-3100E 4*10GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "CONFLUX-3100E 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "CONFLUX-3100E 4*25GE Ethernet Adapter" }] } , Device { vendor_id : 8007u16 , id : 16402u16 , name : "K3 Family [CONFLUX-3100E Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 16403u16 , name : "K3 Family [CONFLUX-3100E MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 20481u16 , name : "CONFLUX-2200P Ethernet Controller" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 9u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 12u16 , name : "Ethernet 100G 2P CONFLUX-2200P" }] } , Device { vendor_id : 8007u16 , id : 20482u16 , name : "CONFLUX-2200P Ethernet Controller [Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 20483u16 , name : "CONFLUX-2200P NVMe Controller" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 7u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 9u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 12u16 , name : "Ethernet 100G 2P CONFLUX-2200P" }] } , Device { vendor_id : 8007u16 , id : 20484u16 , name : "CONFLUX-2200P NVMe Controller [Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 20485u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function]" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function] DPU" }] } , Device { vendor_id : 8007u16 , id : 20486u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function]" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function] DPU" }] } , Device { vendor_id : 8007u16 , id : 20487u16 , name : "CONFLUX-2200P DATA OFFLOAD ENGINE" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 20497u16 , name : "CONFLUX-2200P Ethernet Controller" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 4u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 6u16 , name : "Ethernet 100G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 10u16 , name : "Ethernet 25G 2P CONFLUX-2200P" } , SubSystem { subvendor : 8007u16 , subdevice : 12u16 , name : "Ethernet 100G 2P CONFLUX-2200P" }] } , Device { vendor_id : 8007u16 , id : 20501u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function]" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-2200P Ethernet Controller [MGMT Function] DPU" }] } , Device { vendor_id : 8007u16 , id : 20502u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 1u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function]" } , SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-2200P Ethernet Controller [YDMI Function] DPU" }] } , Device { vendor_id : 8007u16 , id : 24577u16 , name : "K2-Pro Family [CONFLUX-2200X]" , subsystems : & [SubSystem { subvendor : 8007u16 , subdevice : 2u16 , name : "CONFLUX-2200X 2*25GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 3u16 , name : "CONFLUX-2200X 2*40GE Ethernet Adapter" } , SubSystem { subvendor : 8007u16 , subdevice : 5u16 , name : "CONFLUX-2200X 2*100GE Ethernet Adapter" }] } , Device { vendor_id : 8007u16 , id : 24578u16 , name : "K2-Pro Family [CONFLUX-2200X Virtual Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 24579u16 , name : "K2-Pro Family [CONFLUX-2200X MGMT Function]" , subsystems : & [] } , Device { vendor_id : 8007u16 , id : 24580u16 , name : "K2-Pro Family [CONFLUX-2200X DATA Offload Engine]" , subsystems : & [] }] }),
        (4286, Vendor { id : 4286u16 , name : "Tseng Labs International Co." , devices : & [] }),
        (5208, Vendor { id : 5208u16 , name : "Gigabyte Technology Co., Ltd" , devices : & [Device { vendor_id : 5208u16 , id : 4098u16 , name : "Aourus Radeon RX6900XT Xtreme Waterforce WB 16G" , subsystems : & [] } , Device { vendor_id : 5208u16 , id : 8936u16 , name : "Ellesmere [Radeon RX 480]" , subsystems : & [] } , Device { vendor_id : 5208u16 , id : 9253u16 , name : "Radeon RX 9070 GAMING OC 16G" , subsystems : & [] } , Device { vendor_id : 5208u16 , id : 13443u16 , name : "USB 3.0 Controller (VIA VL80x-based xHCI Controller)" , subsystems : & [] }] }),
        (4497, Vendor { id : 4497u16 , name : "Artop Electronic Corp" , devices : & [Device { vendor_id : 4497u16 , id : 3u16 , name : "SCSI Cache Host Adapter" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 4u16 , name : "ATP8400" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 5u16 , name : "ATP850UF" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 6u16 , name : "ATP860 NO-BIOS" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 7u16 , name : "ATP860" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 8u16 , name : "ATP865 NO-ROM" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 9u16 , name : "ATP865" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 10u16 , name : "ATP867-A" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 11u16 , name : "ATP867-B" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 13u16 , name : "ATP8620" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 14u16 , name : "ATP8620" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 17u16 , name : "ATP865-B" , subsystems : & [SubSystem { subvendor : 4497u16 , subdevice : 17u16 , name : "ACARD AEC-6280" }] } , Device { vendor_id : 4497u16 , id : 32770u16 , name : "AEC6710 SCSI-2 Host Adapter" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32784u16 , name : "AEC6712UW SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32800u16 , name : "AEC6712U SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32816u16 , name : "AEC6712S SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32832u16 , name : "AEC6712D SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32848u16 , name : "AEC6712SUW SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32864u16 , name : "AEC6712 SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32896u16 , name : "AEC67160 SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32897u16 , name : "AEC67160S SCSI" , subsystems : & [] } , Device { vendor_id : 4497u16 , id : 32906u16 , name : "AEC67162 2-ch. LVD SCSI" , subsystems : & [] }] }),
        (4486, Vendor { id : 4486u16 , name : "D-Link System Inc" , devices : & [Device { vendor_id : 4486u16 , id : 4098u16 , name : "DL10050 Sundance Ethernet" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 4098u16 , name : "DFE-550TX/FX" } , SubSystem { subvendor : 4486u16 , subdevice : 4114u16 , name : "DFE-580TX" }] } , Device { vendor_id : 4486u16 , id : 4133u16 , name : "AirPlus Xtreme G DWL-G650 Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 4134u16 , name : "AirXpert DWL-AG650 Wireless Cardbus Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 4163u16 , name : "AirXpert DWL-AG650 Wireless Cardbus Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 4864u16 , name : "RTL8139 Ethernet" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 4864u16 , name : "DFE-538TX 10/100 Ethernet Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 4865u16 , name : "DFE-530TX+ 10/100 Ethernet Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 4867u16 , name : "DFE-528TX 10/100 Fast Ethernet PCI Adapter" }] } , Device { vendor_id : 4486u16 , id : 4928u16 , name : "DFE-690TXD CardBus PC Card" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 5440u16 , name : "DFE-680TX" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 5441u16 , name : "DFE-680TXD CardBus PC Card" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 5473u16 , name : "DRP-32TXD Cardbus PC Card" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 13056u16 , name : "DWL-510 / DWL-610 802.11b [Realtek RTL8180L]" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 13056u16 , name : "DWL-610 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 13057u16 , name : "DWL-510 Wireless PCI Adapter" }] } , Device { vendor_id : 4486u16 , id : 14864u16 , name : "AirXpert DWL-AG650 Wireless Cardbus Adapter(rev.B)" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 14865u16 , name : "AirXpert DWL-AG520 Wireless PCI Adapter(rev.B)" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 16384u16 , name : "DL2000-based Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 16385u16 , name : "DGE-550SX PCI-X Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 16896u16 , name : "DFE-520TX Fast Ethernet PCI Adapter" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 4355u16 , name : "DFE-520TX Fast Ethernet PCI Adapter (rev. C1)" }] } , Device { vendor_id : 4486u16 , id : 17152u16 , name : "DGE-528T Gigabit Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 17152u16 , name : "DGE-528T PCI Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 19216u16 , name : "DGE-560T PCI Express (x1) Gigabit Ethernet Adapter" }] } , Device { vendor_id : 4486u16 , id : 17154u16 , name : "DGE-530T Gigabit Ethernet Adapter (rev.C1) [Realtek RTL8169]" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 19200u16 , name : "DGE-560T PCI Express Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 19201u16 , name : "DGE-530T Gigabit Ethernet Adapter (rev 11)" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 19202u16 , name : "DGE-560SX PCI Express Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 19203u16 , name : "DGE-550T Gigabit Ethernet Adapter V.B1" , subsystems : & [] } , Device { vendor_id : 4486u16 , id : 19456u16 , name : "Gigabit Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 19456u16 , name : "DGE-530T Gigabit Ethernet Adapter" }] } , Device { vendor_id : 4486u16 , id : 33792u16 , name : "D-Link DWL-650+ CardBus PC Card" , subsystems : & [] }] }),
        (5209, Vendor { id : 5209u16 , name : "DOOIN Electronics" , devices : & [] }),
        (8351, Vendor { id : 8351u16 , name : "Mobilint, Inc." , devices : & [] }),
        (6559, Vendor { id : 6559u16 , name : "Auvitek" , devices : & [Device { vendor_id : 6559u16 , id : 34049u16 , name : "AU85X1 PCI REV1.1" , subsystems : & [] } , Device { vendor_id : 6559u16 , id : 34081u16 , name : "AU8521 TV card" , subsystems : & [] }] }),
        (13192, Vendor { id : 13192u16 , name : "Hint Corp" , devices : & [Device { vendor_id : 13192u16 , id : 19u16 , name : "HiNT HC4 PCI to ISDN bridge, Multimedia audio controller" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 20u16 , name : "HiNT HC4 PCI to ISDN bridge, Network controller" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 32u16 , name : "HB6 Universal PCI-PCI bridge (transparent mode)" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 33u16 , name : "HB6 Universal PCI-PCI bridge (non-transparent mode)" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 49664u16 , name : "C2K CompactPCI interface bridge" } , SubSystem { subvendor : 6005u16 , subdevice : 52880u16 , name : "CE9" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4240u16 , name : "Cx9 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4256u16 , name : "CA3/CR3 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 12304u16 , name : "PPCI mezzanine (32-bit PMC)" } , SubSystem { subvendor : 19539u16 , subdevice : 12305u16 , name : "PPCI mezzanine (64-bit PMC)" } , SubSystem { subvendor : 19539u16 , subdevice : 16384u16 , name : "PMCCARR1 carrier board" }] } , Device { vendor_id : 13192u16 , id : 34u16 , name : "HiNT HB4 PCI-PCI Bridge (PCI6150)" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 38u16 , name : "HB2 PCI-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 4116u16 , name : "AudioTrak Maya" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 4120u16 , name : "Audiotrak INCA88" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 4121u16 , name : "Miditrak 2120" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 4122u16 , name : "E.Band [AudioTrak Inca88]" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 4123u16 , name : "E.Band [AudioTrak Inca88]" , subsystems : & [] } , Device { vendor_id : 13192u16 , id : 32785u16 , name : "VXPro II Chipset" , subsystems : & [SubSystem { subvendor : 13192u16 , subdevice : 32785u16 , name : "VXPro II Chipset CPU to PCI Bridge" }] } , Device { vendor_id : 13192u16 , id : 32786u16 , name : "VXPro II Chipset" , subsystems : & [SubSystem { subvendor : 13192u16 , subdevice : 32786u16 , name : "VXPro II Chipset PCI to ISA Bridge" }] } , Device { vendor_id : 13192u16 , id : 32787u16 , name : "VXPro II IDE" , subsystems : & [SubSystem { subvendor : 13192u16 , subdevice : 32787u16 , name : "VXPro II Chipset EIDE Controller" }] } , Device { vendor_id : 13192u16 , id : 41219u16 , name : "Blackmagic Design DeckLink HD Pro" , subsystems : & [] }] }),
        (4250, Vendor { id : 4250u16 , name : "Packard Bell" , devices : & [] }),
        (5131, Vendor { id : 5131u16 , name : "Abaco Systems, Inc." , devices : & [] }),
        (4583, Vendor { id : 4583u16 , name : "Toshiba America, Elec. Company" , devices : & [] }),
        (4196, Vendor { id : 4196u16 , name : "Alcatel" , devices : & [Device { vendor_id : 4196u16 , id : 4354u16 , name : "Dynamite 2840 (ADSL PCI modem)" , subsystems : & [] }] }),
        (5665, Vendor { id : 5665u16 , name : "Lynx Studio Technology, Inc." , devices : & [Device { vendor_id : 5665u16 , id : 32u16 , name : "LynxTWO-A" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 33u16 , name : "LynxTWO-B" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 34u16 , name : "LynxTWO-C" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 35u16 , name : "Lynx L22" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 36u16 , name : "Lynx AES16" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 37u16 , name : "Lynx AES16-SRC" , subsystems : & [] } , Device { vendor_id : 5665u16 , id : 40u16 , name : "Lynx AES16e" , subsystems : & [] }] }),
        (33669, Vendor { id : 33669u16 , name : "Suzhou Yige Technology Co., Ltd." , devices : & [] }),
        (5032, Vendor { id : 5032u16 , name : "Exar Corp." , devices : & [Device { vendor_id : 5032u16 , id : 338u16 , name : "XR17C/D152 Dual PCI UART" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 340u16 , name : "XR17C154 Quad UART" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 344u16 , name : "XR17C158 Octal UART" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 594u16 , name : "XR17V252 Dual UART PCI controller" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 596u16 , name : "XR17V254 Quad UART PCI controller" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 600u16 , name : "XR17V258 Octal UART PCI controller" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 850u16 , name : "XR17V352 High Performance Dual PCI Express UART" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 37458u16 , name : "LRUS9252H 2-Port RS232 Serial Adapter" }] } , Device { vendor_id : 5032u16 , id : 852u16 , name : "XR17V354 High Performance Quad PCI Express UART" , subsystems : & [] } , Device { vendor_id : 5032u16 , id : 856u16 , name : "XR17V358 High Performance Octal PCI Express UART" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 33330u16 , name : "8-Channel RS232 Card" }] } , Device { vendor_id : 5032u16 , id : 33624u16 , name : "Twin XR17V358 High Performance Octal PCI Express UARTs" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 33330u16 , name : "16-Channel RS232 Card" }] }] }),
        (5118, Vendor { id : 5118u16 , name : "Advantech Co. Ltd" , devices : & [Device { vendor_id : 5118u16 , id : 113u16 , name : "PCIE-1761H, 8-ch Relay and 8-ch Isolated Digital Input Card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 4672u16 , name : "PCI-1240 4-channel stepper motor controller card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5632u16 , name : "PCI-16xx series PCI multiport serial board (function 0)" , subsystems : & [SubSystem { subvendor : 5633u16 , subdevice : 2u16 , name : "PCI-1601 2-port unisolated RS-422/485" } , SubSystem { subvendor : 5634u16 , subdevice : 2u16 , name : "PCI-1602 2-port isolated RS-422/485" } , SubSystem { subvendor : 5650u16 , subdevice : 4u16 , name : "PCI-1612 4-port RS-232/422/485" }] } , Device { vendor_id : 5118u16 , id : 5635u16 , name : "PCI-1603 2-port isolated RS-232/current loop" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5636u16 , name : "PCI-1604 2-port RS-232" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5760u16 , name : "PCI-1680 Rev.A1 2-port CAN-bus with isolation protection" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5887u16 , name : "PCI-16xx series PCI multiport serial board (function 1: RX/TX steering CPLD)" , subsystems : & [SubSystem { subvendor : 5633u16 , subdevice : 0u16 , name : "PCI-1601 2-port unisolated RS-422/485 PCI communications card" } , SubSystem { subvendor : 5634u16 , subdevice : 0u16 , name : "PCI-1602 2-port isolated RS-422/485" } , SubSystem { subvendor : 5650u16 , subdevice : 0u16 , name : "PCI-1612 4-port RS-232/422/485" }] } , Device { vendor_id : 5118u16 , id : 5905u16 , name : "PCI-1711 16-channel data acquisition card 12-bit, 100kS/s" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5907u16 , name : "PCI-1713 32-channel isolated analog input card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5939u16 , name : "PCI-1733 32-channel isolated digital input card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5940u16 , name : "PCI-1734 32-channel isolated digital output card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5970u16 , name : "PCI-1752 64-channel Isolated Digital Output Card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5972u16 , name : "PCI-1754 64-channel Isolated Digital Input Card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 5974u16 , name : "PCI-1756 64-ch Isolated Digital I/O PCI Card" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 40964u16 , name : "PCI-1612 4-port RS-232/422/485" , subsystems : & [] } , Device { vendor_id : 5118u16 , id : 49922u16 , name : "MIOe-3680 2-Port CAN-Bus MIOe Module with Isolation Protection" , subsystems : & [] }] }),
        (60608, Vendor { id : 60608u16 , name : "Echo Digital Audio Corporation" , devices : & [] }),
        (5606, Vendor { id : 5606u16 , name : "Agere Inc" , devices : & [] }),
        (7267, Vendor { id : 7267u16 , name : "Science and Research Centre of Computer Technology (JSC \"NICEVT\")" , devices : & [Device { vendor_id : 7267u16 , id : 8u16 , name : "K1927BB1Ya [EC8430] Angara Interconnection Network Adapter" , subsystems : & [] }] }),
        (4731, Vendor { id : 4731u16 , name : "Pixera Corporation" , devices : & [] }),
        (5830, Vendor { id : 5830u16 , name : "Micrel-Kendin" , devices : & [Device { vendor_id : 5830u16 , id : 34453u16 , name : "Centaur KS8695 ARM processor" , subsystems : & [] } , Device { vendor_id : 5830u16 , id : 34882u16 , name : "KSZ8842-PMQL 2-Port Ethernet Switch" , subsystems : & [] }] }),
        (5476, Vendor { id : 5476u16 , name : "Yamakatsu Electronics Industry Co Ltd" , devices : & [] }),
        (5521, Vendor { id : 5521u16 , name : "ARN" , devices : & [] }),
        (4299, Vendor { id : 4299u16 , name : "Omron Corporation" , devices : & [] }),
        (5485, Vendor { id : 5485u16 , name : "Alpha-Top Corp" , devices : & [] }),
        (6775, Vendor { id : 6775u16 , name : "Lightfleet Corporation" , devices : & [] }),
        (5182, Vendor { id : 5182u16 , name : "Jones Futurex Inc" , devices : & [] }),
        (4158, Vendor { id : 4158u16 , name : "Solliday Engineering" , devices : & [] }),
        (4632, Vendor { id : 4632u16 , name : "Hybricon Corp." , devices : & [] }),
        (4749, Vendor { id : 4749u16 , name : "G2 Networks, Inc." , devices : & [Device { vendor_id : 4749u16 , id : 33u16 , name : "ATM155 Adapter" , subsystems : & [] }] }),
        (7813, Vendor { id : 7813u16 , name : "HEITEC AG" , devices : & [] }),
        (5277, Vendor { id : 5277u16 , name : "NEWTEK Inc" , devices : & [Device { vendor_id : 5277u16 , id : 1u16 , name : "Video Toaster for PC" , subsystems : & [] }] }),
        (7429, Vendor { id : 7429u16 , name : "AIstone Global Limited" , devices : & [Device { vendor_id : 7429u16 , id : 24615u16 , name : "B760-N2D5 motherboard" , subsystems : & [] } , Device { vendor_id : 7429u16 , id : 28673u16 , name : "H610-N2 motherboard" , subsystems : & [] }] }),
        (4444, Vendor { id : 4444u16 , name : "Photron Ltd." , devices : & [] }),
        (6983, Vendor { id : 6983u16 , name : "Numascale AS" , devices : & [Device { vendor_id : 6983u16 , id : 1537u16 , name : "NumaChip N601" , subsystems : & [] } , Device { vendor_id : 6983u16 , id : 1538u16 , name : "NumaChip N602" , subsystems : & [] }] }),
        (22611, Vendor { id : 22611u16 , name : "XenSource, Inc." , devices : & [Device { vendor_id : 22611u16 , id : 1u16 , name : "Xen Platform Device" , subsystems : & [] } , Device { vendor_id : 22611u16 , id : 49152u16 , name : "Citrix XenServer PCI Device for Windows Update" , subsystems : & [] } , Device { vendor_id : 22611u16 , id : 49424u16 , name : "Virtualized HID" , subsystems : & [] } , Device { vendor_id : 22611u16 , id : 49479u16 , name : "Virtualized Graphics Device" , subsystems : & [] } , Device { vendor_id : 22611u16 , id : 49664u16 , name : "XCP-ng Project PCI Device for Windows Update" , subsystems : & [] }] }),
        (5522, Vendor { id : 5522u16 , name : "Syba Tech Ltd" , devices : & [Device { vendor_id : 5522u16 , id : 1921u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1922u16 , name : "Parallel Port Card 2xEPP" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1923u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1925u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1926u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1927u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1928u16 , name : "Multi-IO Card" , subsystems : & [] } , Device { vendor_id : 5522u16 , id : 1930u16 , name : "Multi-IO Card" , subsystems : & [] }] }),
        (4715, Vendor { id : 4715u16 , name : "Adax, Inc." , devices : & [] }),
        (6374, Vendor { id : 6374u16 , name : "MPL AG" , devices : & [Device { vendor_id : 6374u16 , id : 1u16 , name : "OSCI [Octal Serial Communication Interface]" , subsystems : & [] }] }),
        (4594, Vendor { id : 4594u16 , name : "Picture Tel Japan K.K." , devices : & [] }),
        (6699, Vendor { id : 6699u16 , name : "Ascom AG" , devices : & [Device { vendor_id : 6699u16 , id : 0u16 , name : "GESP v1.2" , subsystems : & [] } , Device { vendor_id : 6699u16 , id : 1u16 , name : "GESP v1.3" , subsystems : & [] } , Device { vendor_id : 6699u16 , id : 2u16 , name : "ECOMP v1.3" , subsystems : & [] } , Device { vendor_id : 6699u16 , id : 5u16 , name : "ETP v1.4" , subsystems : & [] } , Device { vendor_id : 6699u16 , id : 10u16 , name : "ETP-104 v1.1" , subsystems : & [] } , Device { vendor_id : 6699u16 , id : 14u16 , name : "DSLP-104 v1.1" , subsystems : & [] }] }),
        (5577, Vendor { id : 5577u16 , name : "Serome Technology Inc" , devices : & [] }),
        (4118, Vendor { id : 4118u16 , name : "ICL Personal Systems" , devices : & [] }),
        (8137, Vendor { id : 8137u16 , name : "Tehuti Networks Ltd." , devices : & [Device { vendor_id : 8137u16 , id : 12297u16 , name : "10-Giga TOE SmartNIC" , subsystems : & [] } , Device { vendor_id : 8137u16 , id : 12304u16 , name : "10-Giga TOE SmartNIC" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12290u16 , name : "10-Giga TOE Single Port XFP SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12292u16 , name : "10-Giga TOE Single Port SFP+ SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12296u16 , name : "10-Giga TOE Single Port CX4 SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 12308u16 , name : "10-Giga TOE SmartNIC 2-Port" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12291u16 , name : "10-Giga TOE Dual Port XFP Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12293u16 , name : "10-Giga TOE Dual Port SFP+ Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12308u16 , name : "10-Giga TOE Dual Port CX4 Low Profile SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 12560u16 , name : "10-Giga TOE Single Port SmartNIC" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12292u16 , name : "10-Giga TOE Single Port SFP+ SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 12564u16 , name : "10-Giga TOE Dual Port Low Profile SmartNIC" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12293u16 , name : "10-Giga TOE Dual Port SFP+ Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12305u16 , name : "10-Giga TOE Dual Port SFP+/CX4 Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12306u16 , name : "10-Giga TOE Dual Port CX4/SFP+ Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12308u16 , name : "10-Giga TOE Dual Port CX4 Low Profile SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 13072u16 , name : "10-Giga TOE SFP+ Single Port SmartNIC" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12292u16 , name : "10-Giga TOE Single Port SFP+ SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 13076u16 , name : "10-Giga TOE Dual Port Low Profile SmartNIC" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 12293u16 , name : "10-Giga TOE Dual Port SFP+ Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12305u16 , name : "10-Giga TOE Dual Port SFP+/CX4 Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12306u16 , name : "10-Giga TOE Dual Port CX4/SFP+ Low Profile SmartNIC" } , SubSystem { subvendor : 0u16 , subdevice : 12308u16 , name : "10-Giga TOE Dual Port CX4 Low Profile SmartNIC" }] } , Device { vendor_id : 8137u16 , id : 16400u16 , name : "TN4010 Clean SROM" , subsystems : & [] } , Device { vendor_id : 8137u16 , id : 16416u16 , name : "TN9030 10GbE CX4 Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 6156u16 , subdevice : 8256u16 , name : "Mustang-200 10GbE Ethernet Adapter" }] } , Device { vendor_id : 8137u16 , id : 16418u16 , name : "TN9310 10GbE SFP+ Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34569u16 , name : "XG-C100F 10GbE SFP+ Ethernet Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 19712u16 , name : "DXE-810S 10GbE SFP+ Ethernet Adapter" } , SubSystem { subvendor : 5170u16 , subdevice : 33027u16 , name : "10 Gigabit Ethernet SFP+ PCI Express Adapter" } , SubSystem { subvendor : 8137u16 , subdevice : 12309u16 , name : "Ethernet Adapter" }] } , Device { vendor_id : 8137u16 , id : 16420u16 , name : "TN9210 10GBase-T Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 8137u16 , id : 16421u16 , name : "TN9510 10GBase-T/NBASE-T Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4186u16 , subdevice : 29187u16 , name : "SANLink3 NBase-T1" } , SubSystem { subvendor : 4486u16 , subdevice : 10496u16 , name : "DXE-810T 10GBase-T Ethernet Adapter" } , SubSystem { subvendor : 5170u16 , subdevice : 33026u16 , name : "10 Gigabit Ethernet PCI Express Adapter" } , SubSystem { subvendor : 8137u16 , subdevice : 12309u16 , name : "Ethernet Adapter" }] } , Device { vendor_id : 8137u16 , id : 16422u16 , name : "TN9610 10GbE SFP+ Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 4096u16 , name : "LREC6860AF 10 Gigabit Ethernet Adapter" }] } , Device { vendor_id : 8137u16 , id : 16423u16 , name : "TN9710P 10GBase-T/NBASE-T Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4436u16 , subdevice : 872u16 , name : "LGY-PCIE-MG" } , SubSystem { subvendor : 5170u16 , subdevice : 33028u16 , name : "10 Gigabit Ethernet PCI Express Adapter" } , SubSystem { subvendor : 5446u16 , subdevice : 16423u16 , name : "GE10-PCIE4XG202P 10Gbase-T/NBASE-T Ethernet Adapter" } , SubSystem { subvendor : 7082u16 , subdevice : 13072u16 , name : "PCIe Expansion Card" } , SubSystem { subvendor : 8137u16 , subdevice : 12309u16 , name : "Ethernet Adapter" } , SubSystem { subvendor : 19538u16 , subdevice : 4097u16 , name : "LREC6860BT 10 Gigabit Ethernet Adapter" }] } , Device { vendor_id : 8137u16 , id : 17703u16 , name : "TN9710Q 5GBase-T/NBASE-T Ethernet Adapter" , subsystems : & [] }] }),
        (4679, Vendor { id : 4679u16 , name : "Xylon Research, Inc." , devices : & [] }),
        (4129, Vendor { id : 4129u16 , name : "OKI Electric Industry Co. Ltd." , devices : & [] }),
        (4342, Vendor { id : 4342u16 , name : "Creative Electronic Systems SA" , devices : & [] }),
        (7669, Vendor { id : 7669u16 , name : "Shenzhen TIGO Semiconductor" , devices : & [Device { vendor_id : 7669u16 , id : 4610u16 , name : "kimtigo NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7669u16 , id : 8803u16 , name : "kimtigo MG931K NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (6295, Vendor { id : 6295u16 , name : "AMtek" , devices : & [] }),
        (5612, Vendor { id : 5612u16 , name : "Beckhoff GmbH" , devices : & [Device { vendor_id : 5612u16 , id : 12545u16 , name : "FC3101 Profibus DP 1 Channel PCI" , subsystems : & [] } , Device { vendor_id : 5612u16 , id : 20738u16 , name : "FC5102" , subsystems : & [] }] }),
        (6505, Vendor { id : 6505u16 , name : "Qualcomm Atheros" , devices : & [Device { vendor_id : 6505u16 , id : 4134u16 , name : "AR8121/AR8113/AR8114 Gigabit or Fast Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33540u16 , name : "P5KPL-CM Motherboard" }] } , Device { vendor_id : 6505u16 , id : 4168u16 , name : "Attansic L1 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33318u16 , name : "P5B-MX/WiFi-AP, P5KPL-VM Motherboard" }] } , Device { vendor_id : 6505u16 , id : 4194u16 , name : "AR8132 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4195u16 , name : "AR8131 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "GA-G31M-ES2L Motherboard" } , SubSystem { subvendor : 6080u16 , subdevice : 4306u16 , name : "Medion Akoya E7214 Notebook PC [MD98410]" }] } , Device { vendor_id : 6505u16 , id : 4198u16 , name : "Attansic L2c Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4199u16 , name : "Attansic L1c Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4211u16 , name : "AR8151 v1.0 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4227u16 , name : "AR8151 v2.0 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4240u16 , name : "AR8162 Fast Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4237u16 , name : "VivoBook X202E, X202EV" }] } , Device { vendor_id : 6505u16 , id : 4241u16 , name : "AR8161 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5239u16 , name : "N56VZ" }] } , Device { vendor_id : 6505u16 , id : 4256u16 , name : "QCA8172 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 4257u16 , name : "QCA8171 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 8208u16 , name : "QCA8175 card reader controller" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 8264u16 , name : "Attansic L2 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 8288u16 , name : "AR8152 v1.1 Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 8290u16 , name : "AR8152 v2.0 Fast Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33896u16 , name : "Eee PC 1015PX" }] } , Device { vendor_id : 6505u16 , id : 12304u16 , name : "QCA8175 SD controller" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 57489u16 , name : "Killer E220x Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 57505u16 , name : "Killer E2400 Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 6505u16 , id : 57521u16 , name : "Killer E2500 Gigabit Ethernet Controller" , subsystems : & [] }] }),
        (5020, Vendor { id : 5020u16 , name : "Quantum 3d Inc" , devices : & [] }),
        (4452, Vendor { id : 4452u16 , name : "Advanced Peripherals Technologies" , devices : & [] }),
        (4151, Vendor { id : 4151u16 , name : "Hitachi Micro Systems" , devices : & [] }),
        (5066, Vendor { id : 5066u16 , name : "Iomega Corporation" , devices : & [] }),
        (7009, Vendor { id : 7009u16 , name : "Byd Precision Manufacture Co.,Ltd" , devices : & [] }),
        (5347, Vendor { id : 5347u16 , name : "AMTELCO" , devices : & [] }),
        (5343, Vendor { id : 5343u16 , name : "ASIC Communications Corp" , devices : & [] }),
        (4609, Vendor { id : 4609u16 , name : "Vista Controls Corp" , devices : & [] }),
        (5318, Vendor { id : 5318u16 , name : "Data Race Inc" , devices : & [] }),
        (5349, Vendor { id : 5349u16 , name : "Pixelfusion Ltd" , devices : & [] }),
        (8051, Vendor { id : 8051u16 , name : "Shenzhen Quanxing Tech Co., Ltd." , devices : & [] }),
        (6824, Vendor { id : 6824u16 , name : "Ciprico, Inc." , devices : & [Device { vendor_id : 6824u16 , id : 9u16 , name : "RAIDCore Controller" , subsystems : & [] } , Device { vendor_id : 6824u16 , id : 10u16 , name : "RAIDCore Controller" , subsystems : & [] }] }),
        (6890, Vendor { id : 6890u16 , name : "Alcor Micro" , devices : & [Device { vendor_id : 6890u16 , id : 26113u16 , name : "AU6601 PCI-E Flash card reader controller" , subsystems : & [] } , Device { vendor_id : 6890u16 , id : 26145u16 , name : "AU6621 PCI-E Flash card reader controller" , subsystems : & [] } , Device { vendor_id : 6890u16 , id : 26149u16 , name : "AU6625 PCI-E Flash card reader controller" , subsystems : & [] }] }),
        (7500, Vendor { id : 7500u16 , name : "Diamanti, Inc." , devices : & [] }),
        (7878, Vendor { id : 7878u16 , name : "Vastai Technologies" , devices : & [Device { vendor_id : 7878u16 , id : 256u16 , name : "SV100" , subsystems : & [SubSystem { subvendor : 7878u16 , subdevice : 16u16 , name : "VA1V" } , SubSystem { subvendor : 7878u16 , subdevice : 17u16 , name : "VA1A" } , SubSystem { subvendor : 7878u16 , subdevice : 18u16 , name : "VA1-16G" } , SubSystem { subvendor : 7878u16 , subdevice : 19u16 , name : "VA1E" } , SubSystem { subvendor : 7878u16 , subdevice : 20u16 , name : "VA1L" } , SubSystem { subvendor : 7878u16 , subdevice : 21u16 , name : "VA1-32G" } , SubSystem { subvendor : 7878u16 , subdevice : 32u16 , name : "VA10-32G" } , SubSystem { subvendor : 7878u16 , subdevice : 33u16 , name : "VA12-32G" } , SubSystem { subvendor : 7878u16 , subdevice : 34u16 , name : "VA16" } , SubSystem { subvendor : 7878u16 , subdevice : 35u16 , name : "VA10-64G" } , SubSystem { subvendor : 7878u16 , subdevice : 36u16 , name : "VA10L" } , SubSystem { subvendor : 7878u16 , subdevice : 37u16 , name : "VA12-64G" } , SubSystem { subvendor : 7878u16 , subdevice : 38u16 , name : "VA12-128G" } , SubSystem { subvendor : 7878u16 , subdevice : 48u16 , name : "VE1P" } , SubSystem { subvendor : 7878u16 , subdevice : 49u16 , name : "VE1S" } , SubSystem { subvendor : 7878u16 , subdevice : 50u16 , name : "VE1V" } , SubSystem { subvendor : 7878u16 , subdevice : 51u16 , name : "VE1M" } , SubSystem { subvendor : 7878u16 , subdevice : 52u16 , name : "VE1E" } , SubSystem { subvendor : 7878u16 , subdevice : 53u16 , name : "VE1C" } , SubSystem { subvendor : 7878u16 , subdevice : 54u16 , name : "VE1MT" } , SubSystem { subvendor : 7878u16 , subdevice : 58u16 , name : "VR1S-8G" } , SubSystem { subvendor : 7878u16 , subdevice : 59u16 , name : "VR1S-16G" } , SubSystem { subvendor : 7878u16 , subdevice : 60u16 , name : "VR1M-8G" } , SubSystem { subvendor : 7878u16 , subdevice : 61u16 , name : "VR1M-16G" } , SubSystem { subvendor : 7878u16 , subdevice : 62u16 , name : "VR1MT-8G" } , SubSystem { subvendor : 7878u16 , subdevice : 63u16 , name : "VR1MT-16G" } , SubSystem { subvendor : 7878u16 , subdevice : 64u16 , name : "VA1M" } , SubSystem { subvendor : 7878u16 , subdevice : 65u16 , name : "VA1U" } , SubSystem { subvendor : 7878u16 , subdevice : 80u16 , name : "VAM" }] } , Device { vendor_id : 7878u16 , id : 512u16 , name : "SG100" , subsystems : & [SubSystem { subvendor : 7878u16 , subdevice : 96u16 , name : "VG1200" } , SubSystem { subvendor : 7878u16 , subdevice : 97u16 , name : "VG1600" } , SubSystem { subvendor : 7878u16 , subdevice : 98u16 , name : "VG1800" } , SubSystem { subvendor : 7878u16 , subdevice : 99u16 , name : "VG1000-32G" } , SubSystem { subvendor : 7878u16 , subdevice : 100u16 , name : "VG600" } , SubSystem { subvendor : 7878u16 , subdevice : 101u16 , name : "VG800" } , SubSystem { subvendor : 7878u16 , subdevice : 102u16 , name : "VG1000-64G" } , SubSystem { subvendor : 7878u16 , subdevice : 103u16 , name : "VG1000-24G" } , SubSystem { subvendor : 7878u16 , subdevice : 104u16 , name : "VG1000c-24G" } , SubSystem { subvendor : 7878u16 , subdevice : 105u16 , name : "VG1000c-32G" } , SubSystem { subvendor : 7878u16 , subdevice : 106u16 , name : "VG1000c-64G" } , SubSystem { subvendor : 7878u16 , subdevice : 107u16 , name : "VG1000c-16G" } , SubSystem { subvendor : 7878u16 , subdevice : 108u16 , name : "VG1000c-48G" } , SubSystem { subvendor : 7878u16 , subdevice : 109u16 , name : "VG600c" } , SubSystem { subvendor : 7878u16 , subdevice : 110u16 , name : "VG800c" }] } , Device { vendor_id : 7878u16 , id : 527u16 , name : "SG100 vGPU" , subsystems : & [] } , Device { vendor_id : 7878u16 , id : 768u16 , name : "SR100" , subsystems : & [] }] }),
        (7177, Vendor { id : 7177u16 , name : "CSP, Inc." , devices : & [Device { vendor_id : 7177u16 , id : 16980u16 , name : "10G-PCIE3-8D-2S" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16981u16 , name : "10G-PCIE3-8D-Q" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16982u16 , name : "10G-PCIE3-8D-2S" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16984u16 , name : "10G-PCIE3-8E-2S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16992u16 , name : "10G-PCIE3-8E-4S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16993u16 , name : "10G-PCIE3-8E-4S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16994u16 , name : "10G-PCIE3-8E-4S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16995u16 , name : "10G-PCIE3-8E-4S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16996u16 , name : "10G-PCIE3-8E-2S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 16997u16 , name : "10G-PCIE3-8E-2S Network Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 20480u16 , name : "25G-PCIE3-8A-2S Security Intelligent Adapter" , subsystems : & [] } , Device { vendor_id : 7177u16 , id : 20481u16 , name : "25G-PCIE3-8B-2S Security Intelligent Adapter" , subsystems : & [] }] }),
        (17476, Vendor { id : 17476u16 , name : "Internext Compression Inc" , devices : & [Device { vendor_id : 17476u16 , id : 22u16 , name : "iTVC16 (CX23416) Video Decoder" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 3u16 , name : "WinTV PVR 250" } , SubSystem { subvendor : 112u16 , subdevice : 9u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 2049u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 2055u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 16385u16 , name : "WinTV PVR 250" } , SubSystem { subvendor : 112u16 , subdevice : 16393u16 , name : "WinTV PVR 250" } , SubSystem { subvendor : 112u16 , subdevice : 18433u16 , name : "WinTV PVR 250" } , SubSystem { subvendor : 112u16 , subdevice : 18435u16 , name : "WinTV PVR 250" } , SubSystem { subvendor : 112u16 , subdevice : 32771u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 34817u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 51201u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 112u16 , subdevice : 59399u16 , name : "WinTV PVR 500 (1st unit)" } , SubSystem { subvendor : 112u16 , subdevice : 59415u16 , name : "WinTV PVR 500 (2nd unit)" } , SubSystem { subvendor : 112u16 , subdevice : 65426u16 , name : "WiNTV PVR-550" } , SubSystem { subvendor : 624u16 , subdevice : 2049u16 , name : "WinTV PVR 150" } , SubSystem { subvendor : 4173u16 , subdevice : 317u16 , name : "ENX-26 TV Encoder" } , SubSystem { subvendor : 4348u16 , subdevice : 53304u16 , name : "GV-MVP/RX2W (1st unit)" } , SubSystem { subvendor : 4348u16 , subdevice : 53305u16 , name : "GV-MVP/RX2W (2nd unit)" } , SubSystem { subvendor : 4779u16 , subdevice : 65523u16 , name : "MPG600" } , SubSystem { subvendor : 4779u16 , subdevice : 65535u16 , name : "MPG600" } , SubSystem { subvendor : 5217u16 , subdevice : 49162u16 , name : "M113 PCI Analog TV (PAL/SECAM, Philips FQ1216MK3 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49163u16 , name : "M113 PCI Analog TV (PAL/SECAM+FM, Philips FM1216MK3 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49164u16 , name : "M113 PCI Analog TV (NTSC, JAPAN version, Philips FI1286MK2 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49168u16 , name : "M113 PCI Analog TV (NTSC, Philips FI1236MK3 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49169u16 , name : "M113 PCI Analog TV (NTSC+FM, Philips FM1236MK3 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49176u16 , name : "M113 PCI Analog TV (NTSC, Philips FQ1236MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49177u16 , name : "UltraTV 1500 MCE, a.k.a. M113 PCI Analog TV (NTSC+FM, Philips FQ1236MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49178u16 , name : "M113 PCI Analog TV (PAL/SECAM, Philips FQ1216MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49179u16 , name : "M113 PCI Analog TV (PAL/SECAM+FM, Philips FM1216MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49200u16 , name : "M113 PCI Analog TV (NTSC-J, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49201u16 , name : "M113 PCI Analog TV (NTSC-J+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49202u16 , name : "M113 PCI Analog TV (PAL/SECAM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49203u16 , name : "M113 PCI Analog TV (PAL/SECAM+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49204u16 , name : "M113 PCI Analog TV (NTSC, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49205u16 , name : "M113 PCI Analog TV (NTSC+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49215u16 , name : "C115 PCI video capture card (no tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49462u16 , name : "M104 mini-PCI Analog TV" } , SubSystem { subvendor : 5217u16 , subdevice : 49674u16 , name : "M755 AVerTV Video Capture (PAL/SECAM, Philips FQ1216MK3 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49688u16 , name : "M755 AVerTV Video Capture (NTSC, Philips FQ1236MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49689u16 , name : "M755 AVerTV Video Capture (NTSC+FM, Philips FQ1236MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49690u16 , name : "M755 AVerTV Video Capture (PAL/SECAM, Philips FQ1216MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49691u16 , name : "M755 AVerTV Video Capture (PAL/SECAM+FM, Philips FM1216MK5 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49712u16 , name : "M755 AVerTV Video Capture (NTSC-J, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49713u16 , name : "M755 AVerTV Video Capture (NTSC-J+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49714u16 , name : "M755 AVerTV Video Capture (PAL/SECAM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49715u16 , name : "M755 AVerTV Video Capture (PAL/SECAM+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49716u16 , name : "M755 AVerTV Video Capture (NTSC, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49717u16 , name : "M755 AVerTV Video Capture (NTSC+FM, Partsnic tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 49975u16 , name : "E106 AVerMedia AVerTV Video Capture" } , SubSystem { subvendor : 5217u16 , subdevice : 50233u16 , name : "M116 AVerMedia AVerTV MCE 116 Plus (NTSC/PAL/SECAM+FM+REMOTE, Xceive 2028 tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 50687u16 , name : "C755 AVerTV Video Capture card (no tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 50943u16 , name : "C115 PCI video capture card (no tuner)" } , SubSystem { subvendor : 5217u16 , subdevice : 51001u16 , name : "M785 AVerMedia PCI Analog TV (NTSC/PAL/SECAM+FM, Xceive 2028 tuner)" } , SubSystem { subvendor : 36869u16 , subdevice : 146u16 , name : "VideOh! AVC-2010" } , SubSystem { subvendor : 36869u16 , subdevice : 147u16 , name : "VideOh! AVC-2410" }] } , Device { vendor_id : 17476u16 , id : 2051u16 , name : "iTVC15 (CX23415) Video Decoder" , subsystems : & [SubSystem { subvendor : 112u16 , subdevice : 16384u16 , name : "WinTV PVR-350" } , SubSystem { subvendor : 112u16 , subdevice : 16385u16 , name : "WinTV PVR-250" } , SubSystem { subvendor : 112u16 , subdevice : 18432u16 , name : "WinTV PVR-350 (V1)" } , SubSystem { subvendor : 4779u16 , subdevice : 0u16 , name : "MPG160" } , SubSystem { subvendor : 5217u16 , subdevice : 41934u16 , name : "M179" } , SubSystem { subvendor : 5217u16 , subdevice : 41935u16 , name : "M179" }] }] }),
        (5583, Vendor { id : 5583u16 , name : "Hilscher Gesellschaft für Systemautomation mbH" , devices : & [Device { vendor_id : 5583u16 , id : 0u16 , name : "CIFX PCI/PCIe" , subsystems : & [] } , Device { vendor_id : 5583u16 , id : 144u16 , name : "CIFX PCIe" , subsystems : & [] } , Device { vendor_id : 5583u16 , id : 2304u16 , name : "CIFX PCIe" , subsystems : & [] }] }),
        (5609, Vendor { id : 5609u16 , name : "Pacific Digital Corp" , devices : & [Device { vendor_id : 5609u16 , id : 6209u16 , name : "ADMA-100 DiscStaQ ATA Controller" , subsystems : & [] }] }),
        (4301, Vendor { id : 4301u16 , name : "Advanced System Products, Inc" , devices : & [Device { vendor_id : 4301u16 , id : 4352u16 , name : "ASC1100" , subsystems : & [] } , Device { vendor_id : 4301u16 , id : 4608u16 , name : "ASC1200 [(abp940) Fast SCSI-II]" , subsystems : & [] } , Device { vendor_id : 4301u16 , id : 4864u16 , name : "ASC1300 / ASC3030 [ABP940-U / ABP960-U / ABP3925]" , subsystems : & [SubSystem { subvendor : 4301u16 , subdevice : 4880u16 , name : "ASC1300/3030 SCSI adapter" } , SubSystem { subvendor : 4501u16 , subdevice : 4896u16 , name : "Ultra-SCSI CardBus PC Card REX CB31" }] } , Device { vendor_id : 4301u16 , id : 8960u16 , name : "ABP940-UW" , subsystems : & [] } , Device { vendor_id : 4301u16 , id : 9472u16 , name : "ABP940-U2W" , subsystems : & [] } , Device { vendor_id : 4301u16 , id : 9984u16 , name : "ABP3950-U3W" , subsystems : & [] }] }),
        (4571, Vendor { id : 4571u16 , name : "Sega Enterprises Ltd" , devices : & [] }),
        (5878, Vendor { id : 5878u16 , name : "VideoTele.com, Inc." , devices : & [] }),
        (4133, Vendor { id : 4133u16 , name : "Acer Incorporated [ALI]" , devices : & [Device { vendor_id : 4133u16 , id : 5173u16 , name : "M1435" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5189u16 , name : "M1445" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5193u16 , name : "M1449" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5201u16 , name : "M1451" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5217u16 , name : "M1461" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5257u16 , name : "M1489" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5393u16 , name : "M1511" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5394u16 , name : "ALI M1512 Aladdin" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5395u16 , name : "M1513" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5409u16 , name : "ALI M1521 Aladdin III CPU Bridge" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5409u16 , name : "ALI M1521 Aladdin III CPU Bridge" }] } , Device { vendor_id : 4133u16 , id : 5411u16 , name : "ALI M1523 ISA Bridge" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5411u16 , name : "ALI M1523 ISA Bridge" }] } , Device { vendor_id : 4133u16 , id : 5425u16 , name : "M1531 Northbridge [Aladdin IV/IV+]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5427u16 , name : "M1533 PCI-to-ISA Bridge" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5427u16 , name : "ALI M1533 Aladdin IV/V ISA South Bridge" }] } , Device { vendor_id : 4133u16 , id : 5429u16 , name : "M1535 PCI Bridge + Super I/O + FIR" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5441u16 , name : "M1541 Northbridge [Aladdin V]" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 5441u16 , name : "ALI M1541 Aladdin V/V+ AGP+PCI North Bridge" }] } , Device { vendor_id : 4133u16 , id : 5442u16 , name : "M1542 Northbridge [Aladdin V]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5443u16 , name : "M1543 PCI-to-ISA Bridge + Super I/O + FIR" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5473u16 , name : "M1561 Northbridge [Aladdin 7]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5665u16 , name : "M1621 Northbridge [Aladdin-Pro II]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5681u16 , name : "M1631 Northbridge+3D Graphics [Aladdin TNT2]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5697u16 , name : "M1641 Northbridge [Aladdin-Pro IV]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5703u16 , name : "M1647 [MaGiK1] PCI North Bridge" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5745u16 , name : "M1671 Northbridge [ALADDiN-P4]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 5746u16 , name : "Northbridge [CyberALADDiN-P4]" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12609u16 , name : "M3141" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12611u16 , name : "M3143" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12613u16 , name : "M3145" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12615u16 , name : "M3147" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12617u16 , name : "M3149" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 12625u16 , name : "M3151" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 13063u16 , name : "M3307 MPEG-I Video Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 13065u16 , name : "M3309 MPEG-II Video w/ Software Audio Decoder" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 13089u16 , name : "M3321 MPEG-II Audio/Video Decoder" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21010u16 , name : "M4803" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21013u16 , name : "ALI PCI EIDE Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21015u16 , name : "M5217H" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21017u16 , name : "M5219" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21029u16 , name : "M5225" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21033u16 , name : "M5229" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21045u16 , name : "M5235" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21047u16 , name : "OHCI USB Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21049u16 , name : "EHCI USB Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21056u16 , name : "EIDE Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21057u16 , name : "PCMCIA Bridge" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21058u16 , name : "General Purpose Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21059u16 , name : "PCI to PCI Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21060u16 , name : "Floppy Disk Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21063u16 , name : "M1541 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21073u16 , name : "M5251 P1394 Controller" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21543u16 , name : "PCI to AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21585u16 , name : "M5451 PCI AC-Link Controller Audio Device" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 21587u16 , name : "M5453 PCI AC-Link Controller Modem Device" , subsystems : & [] } , Device { vendor_id : 4133u16 , id : 28929u16 , name : "M7101 PCI PMU Power Management Controller" , subsystems : & [SubSystem { subvendor : 4281u16 , subdevice : 28929u16 , name : "M7101 PCI PMU Power Management Controller" }] } , Device { vendor_id : 4133u16 , id : 38402u16 , name : "AMD RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (17812, Vendor { id : 17812u16 , name : "Cogetec Informatique Inc" , devices : & [] }),
        (7006, Vendor { id : 7006u16 , name : "STAR-Dundee Ltd." , devices : & [Device { vendor_id : 7006u16 , id : 1u16 , name : "SpaceWire PCI Mk2" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 2u16 , name : "SpaceWire PCIe Mk1" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 3u16 , name : "SpaceWire cPCI Mk2" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 4u16 , name : "SpaceWire PXI Recorder Mk1" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 5u16 , name : "SpaceWire PXI Interface Mk1" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 6u16 , name : "SpaceWire PXI Interface Mk1 with RMAP Target" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 8u16 , name : "SpaceWire PXI Router Mk1" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 11u16 , name : "SpaceWire PXI Interface Mk2" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 12u16 , name : "SpaceWire PXI Interface Mk2 with RMAP Target" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 13u16 , name : "SpaceWire PXI Router Mk2" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 14u16 , name : "SpaceWire PXI Recorder Mk2" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 256u16 , name : "STAR-Ultra PCIe" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 258u16 , name : "STAR-Ultra Single-Lane Router" , subsystems : & [] } , Device { vendor_id : 7006u16 , id : 512u16 , name : "SpaceWire PCIe Mk2" , subsystems : & [] }] }),
        (5806, Vendor { id : 5806u16 , name : "SafeNet Inc" , devices : & [Device { vendor_id : 5806u16 , id : 1u16 , name : "SafeXcel 1140" , subsystems : & [] } , Device { vendor_id : 5806u16 , id : 10u16 , name : "SafeXcel 1841" , subsystems : & [] } , Device { vendor_id : 5806u16 , id : 4417u16 , name : "SafeXcel 1141" , subsystems : & [SubSystem { subvendor : 1u16 , subdevice : 1u16 , name : "SafeXcel 1141 v. 1.1" }] } , Device { vendor_id : 5806u16 , id : 6209u16 , name : "SafeXcel 1842" , subsystems : & [] }] }),
        (8285, Vendor { id : 8285u16 , name : "Shanghai Zijing Xinjie Intelligent Technology Co., Ltd." , devices : & [] }),
        (6900, Vendor { id : 6900u16 , name : "Red Hat, Inc." , devices : & [Device { vendor_id : 6900u16 , id : 4096u16 , name : "Virtio network device" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65531u16 , name : "Propolis Virtio network device" }] } , Device { vendor_id : 6900u16 , id : 4097u16 , name : "Virtio block device" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65530u16 , name : "Propolis Virtio block device" }] } , Device { vendor_id : 6900u16 , id : 4098u16 , name : "Virtio memory balloon" , subsystems : & [] } , Device { vendor_id : 6900u16 , id : 4099u16 , name : "Virtio console" , subsystems : & [] } , Device { vendor_id : 6900u16 , id : 4100u16 , name : "Virtio SCSI" , subsystems : & [] } , Device { vendor_id : 6900u16 , id : 4101u16 , name : "Virtio RNG" , subsystems : & [] } , Device { vendor_id : 6900u16 , id : 4105u16 , name : "Virtio filesystem" , subsystems : & [] } , Device { vendor_id : 6900u16 , id : 4161u16 , name : "Virtio 1.0 network device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4162u16 , name : "Virtio 1.0 block device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4163u16 , name : "Virtio 1.0 console" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4164u16 , name : "Virtio 1.0 RNG" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4165u16 , name : "Virtio 1.0 balloon" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4166u16 , name : "Virtio 1.0 ioMemory" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4167u16 , name : "Virtio 1.0 remote processor messaging" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4168u16 , name : "Virtio 1.0 SCSI" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4169u16 , name : "Virtio 9P transport" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4170u16 , name : "Virtio 1.0 WLAN MAC" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4171u16 , name : "Virtio 1.0 remoteproc serial link" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4173u16 , name : "Virtio 1.0 memory balloon" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4176u16 , name : "Virtio 1.0 GPU" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4177u16 , name : "Virtio 1.0 clock/timer" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4178u16 , name : "Virtio 1.0 input" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4179u16 , name : "Virtio 1.0 socket" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4180u16 , name : "Virtio 1.0 crypto" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4181u16 , name : "Virtio 1.0 signal distribution device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4182u16 , name : "Virtio 1.0 pstore device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4183u16 , name : "Virtio 1.0 IOMMU" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4184u16 , name : "Virtio 1.0 mem" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4185u16 , name : "Virtio 1.0 sound" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4186u16 , name : "Virtio 1.0 file system" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4187u16 , name : "Virtio 1.0 pmem" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4188u16 , name : "Virtio 1.0 rpmb" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4189u16 , name : "Virtio 1.0 mac80211-hwsim" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4190u16 , name : "Virtio 1.0 video encoder" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4191u16 , name : "Virtio 1.0 video decoder" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4192u16 , name : "Virtio 1.0 SCMI" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4193u16 , name : "Virtio 1.0 nitro secure module" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4194u16 , name : "Virtio 1.0 I2C adapter" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4195u16 , name : "Virtio 1.0 watchdog" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4196u16 , name : "Virtio 1.0 can" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4197u16 , name : "Virtio 1.0 dmabuf" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4198u16 , name : "Virtio 1.0 parameter server" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4199u16 , name : "Virtio 1.0 audio policy" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4200u16 , name : "Virtio 1.0 Bluetooth" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4201u16 , name : "Virtio 1.0 GPIO" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] } , Device { vendor_id : 6900u16 , id : 4368u16 , name : "QEMU Inter-VM shared memory device" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU" }] }] }),
        (4176, Vendor { id : 4176u16 , name : "Winbond Electronics Corp" , devices : & [Device { vendor_id : 4176u16 , id : 0u16 , name : "NE2000" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 1u16 , name : "W83769F" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 51u16 , name : "W89C33D 802.11 a/b/g BB/MAC" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 261u16 , name : "W82C105" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 1576u16 , name : "W83628F/W83629D PCI to ISA Bridge Set" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 2112u16 , name : "W89C840" , subsystems : & [SubSystem { subvendor : 4176u16 , subdevice : 1u16 , name : "W89C840 Ethernet Adapter" } , SubSystem { subvendor : 4176u16 , subdevice : 2112u16 , name : "W89C840 Ethernet Adapter" }] } , Device { vendor_id : 4176u16 , id : 2368u16 , name : "W89C940" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 23130u16 , name : "W89C940 Twisted-pair Ether-LAN Controller With PCI Interface [ELANC-PCI]" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 26258u16 , name : "W6692 PCI ISDN S/T-Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5890u16 , name : "ISDN Adapter (PCI Bus, D, W)" } , SubSystem { subvendor : 4163u16 , subdevice : 5891u16 , name : "ISDN Adapter (PCI Bus, DV, W)" } , SubSystem { subvendor : 4163u16 , subdevice : 5895u16 , name : "ISDN Adapter (PCI Bus, DV, W)" } , SubSystem { subvendor : 5199u16 , subdevice : 5890u16 , name : "ISDN Adapter (PCI Bus, D, W)" } , SubSystem { subvendor : 5199u16 , subdevice : 5891u16 , name : "ISDN Adapter (PCI Bus, DV, W)" } , SubSystem { subvendor : 5199u16 , subdevice : 5895u16 , name : "ISDN Adapter (PCI Bus, DV, W)" }] } , Device { vendor_id : 4176u16 , id : 39201u16 , name : "W99200F MPEG-1 Video Encoder" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 39202u16 , name : "W99200F/W9922PF MPEG-1/2 Video Encoder" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 39264u16 , name : "W9960CF Video Codec" , subsystems : & [] } , Device { vendor_id : 4176u16 , id : 39280u16 , name : "W9970CF" , subsystems : & [] }] }),
        (5529, Vendor { id : 5529u16 , name : "Delta Electronics Inc" , devices : & [] }),
        (5261, Vendor { id : 5261u16 , name : "DIGICOM Systems, Inc." , devices : & [Device { vendor_id : 5261u16 , id : 4099u16 , name : "HCF 56k Data/Fax Modem" , subsystems : & [] }] }),
        (4859, Vendor { id : 4859u16 , name : "Spectrum Signal Processing" , devices : & [Device { vendor_id : 4859u16 , id : 1u16 , name : "PMC-MAI" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 245u16 , name : "F5 Dakar" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 685u16 , name : "PMC-2MAI" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 10972u16 , name : "ePMC-2ADC" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 12544u16 , name : "PRO-3100" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 13568u16 , name : "PRO-3500" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 19791u16 , name : "Modena" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 33056u16 , name : "ePMC-8120" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 55906u16 , name : "Daytona C6201 PCI (Hurricane)" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 56162u16 , name : "Ingliston XBIF" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 56418u16 , name : "Ingliston PLX9054" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 56674u16 , name : "Ingliston JTAG/ISP" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 60892u16 , name : "ePMC-MSDDC" , subsystems : & [] } , Device { vendor_id : 4859u16 , id : 64001u16 , name : "ePMC-FPGA" , subsystems : & [] }] }),
        (8112, Vendor { id : 8112u16 , name : "ICube Corporation Limited" , devices : & [Device { vendor_id : 8112u16 , id : 4096u16 , name : "NF1000 Series GPU" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 4097u16 , name : "NF1001" } , SubSystem { subvendor : 8112u16 , subdevice : 4098u16 , name : "NF1002" } , SubSystem { subvendor : 8112u16 , subdevice : 4099u16 , name : "NF1003" }] } , Device { vendor_id : 8112u16 , id : 4097u16 , name : "NF1001" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 4097u16 , name : "NF1001" } , SubSystem { subvendor : 8112u16 , subdevice : 4098u16 , name : "NF1002" } , SubSystem { subvendor : 8112u16 , subdevice : 4099u16 , name : "NF1003" }] } , Device { vendor_id : 8112u16 , id : 8192u16 , name : "NF2000 Series GPU" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 8193u16 , name : "NF2001" } , SubSystem { subvendor : 8112u16 , subdevice : 8194u16 , name : "NF 2002" } , SubSystem { subvendor : 8112u16 , subdevice : 8195u16 , name : "NF 2003" }] } , Device { vendor_id : 8112u16 , id : 12288u16 , name : "NF3000 Series GPU" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 12289u16 , name : "NF3001" } , SubSystem { subvendor : 8112u16 , subdevice : 12290u16 , name : "NF3002" } , SubSystem { subvendor : 8112u16 , subdevice : 12291u16 , name : "NF3003" }] } , Device { vendor_id : 8112u16 , id : 16384u16 , name : "NF4000 Series GPU" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 16385u16 , name : "NF4001" } , SubSystem { subvendor : 8112u16 , subdevice : 16386u16 , name : "NF4002" } , SubSystem { subvendor : 8112u16 , subdevice : 16387u16 , name : "NF4003" }] } , Device { vendor_id : 8112u16 , id : 20480u16 , name : "NF5000 Series GPU" , subsystems : & [SubSystem { subvendor : 8112u16 , subdevice : 20481u16 , name : "NF 5001" } , SubSystem { subvendor : 8112u16 , subdevice : 20482u16 , name : "NF 5002" } , SubSystem { subvendor : 8112u16 , subdevice : 20483u16 , name : "NF5003" }] }] }),
        (4541, Vendor { id : 4541u16 , name : "Pinnacle Systems Inc." , devices : & [Device { vendor_id : 4541u16 , id : 21u16 , name : "FireWire Controller" , subsystems : & [] } , Device { vendor_id : 4541u16 , id : 35u16 , name : "Studio 700" , subsystems : & [] } , Device { vendor_id : 4541u16 , id : 46u16 , name : "PCTV 40i" , subsystems : & [] } , Device { vendor_id : 4541u16 , id : 64u16 , name : "Royal TS Function 1" , subsystems : & [SubSystem { subvendor : 4541u16 , subdevice : 68u16 , name : "PCTV 2000i Dual DVB-T Pro PCI Tuner 1" } , SubSystem { subvendor : 4541u16 , subdevice : 69u16 , name : "PCTV Dual Sat Pro PCI 4000i Tuner 1" }] } , Device { vendor_id : 4541u16 , id : 65u16 , name : "RoyalTS Function 2" , subsystems : & [SubSystem { subvendor : 4541u16 , subdevice : 68u16 , name : "PCTV 2000i Dual DVB-T Pro PCI Tuner 2" } , SubSystem { subvendor : 4541u16 , subdevice : 69u16 , name : "PCTV Dual Sat Pro PCI 4000i Tuner 2" }] } , Device { vendor_id : 4541u16 , id : 66u16 , name : "Royal TS Function 3" , subsystems : & [SubSystem { subvendor : 4541u16 , subdevice : 68u16 , name : "PCTV 2000i Dual DVB-T Pro PCI Common" } , SubSystem { subvendor : 4541u16 , subdevice : 69u16 , name : "PCTV Dual Sat Pro PCI 4000i Common" }] } , Device { vendor_id : 4541u16 , id : 81u16 , name : "PCTV HD 800i" , subsystems : & [] } , Device { vendor_id : 4541u16 , id : 48862u16 , name : "AV/DV Studio Capture Card" , subsystems : & [] }] }),
        (6048, Vendor { id : 6048u16 , name : "Genesys Logic, Inc" , devices : & [Device { vendor_id : 6048u16 , id : 29027u16 , name : "GL9701 PCIe to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 32899u16 , name : "GL880 USB 1.1 UHCI controller" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 32900u16 , name : "GL880 USB 2.0 EHCI controller" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 38736u16 , name : "GL9750 SD Host Controller" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 38741u16 , name : "GL9755 SD Host Controller" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 38759u16 , name : "GL9767 PCIe SD UHS-II & SD Express Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 6048u16 , id : 59235u16 , name : "GL9763E eMMC Controller" , subsystems : & [] }] }),
        (4217, Vendor { id : 4217u16 , name : "I-Bus" , devices : & [] }),
        (4483, Vendor { id : 4483u16 , name : "Fujikura Ltd" , devices : & [] }),
        (8448, Vendor { id : 8448u16 , name : "Shenzhen Kimviking Semiconductor Co., Ltd." , devices : & [] }),
        (8286, Vendor { id : 8286u16 , name : "SDTECH" , devices : & [Device { vendor_id : 8286u16 , id : 0u16 , name : "SD85xxx PCIe Gen5 Switch" , subsystems : & [SubSystem { subvendor : 8286u16 , subdevice : 72u16 , name : "SD85048 PCIe Gen 5 48 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8286u16 , subdevice : 150u16 , name : "SD85096 PCIe Gen 5 96 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8286u16 , subdevice : 260u16 , name : "SD85104 PCIe Gen 5 104 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8286u16 , subdevice : 261u16 , name : "SD85104 PCIe Gen 5 104 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8286u16 , subdevice : 324u16 , name : "SD85144 PCIe Gen 5 144 lane Switch Upstream/Downstream Port" } , SubSystem { subvendor : 8286u16 , subdevice : 8196u16 , name : "SD85000 PCIe NT Endpoint" } , SubSystem { subvendor : 8286u16 , subdevice : 8197u16 , name : "SD85000 PCIe gDMA Endpoint" } , SubSystem { subvendor : 8286u16 , subdevice : 8198u16 , name : "SD85000 PCIe Management Endpoint" }] } , Device { vendor_id : 8286u16 , id : 20552u16 , name : "SD85048 PCIe Gen 5 48 lane Switch Upstream/Downstream Port" , subsystems : & [] } , Device { vendor_id : 8286u16 , id : 20630u16 , name : "SD85096 PCIe Gen 5 96 lane Switch Upstream/Downstream Port" , subsystems : & [] } , Device { vendor_id : 8286u16 , id : 20740u16 , name : "SD85104 PCIe Gen 5 104 lane Switch Upstream/Downstream Port" , subsystems : & [] } , Device { vendor_id : 8286u16 , id : 20741u16 , name : "SD85104 PCIe Gen 5 104 lane Switch Upstream/Downstream Port" , subsystems : & [] } , Device { vendor_id : 8286u16 , id : 20804u16 , name : "SD85144 PCIe Gen 5 144 lane Switch Upstream/Downstream Port" , subsystems : & [] }] }),
        (5065, Vendor { id : 5065u16 , name : "Eaton Corporation" , devices : & [] }),
        (4937, Vendor { id : 4937u16 , name : "Sumitomo Electric Industries, Ltd." , devices : & [] }),
        (4274, Vendor { id : 4274u16 , name : "Raytheon Company" , devices : & [] }),
        (4732, Vendor { id : 4732u16 , name : "Crosspoint Solutions, Inc." , devices : & [] }),
        (7418, Vendor { id : 7418u16 , name : "Corsair Memory, Inc" , devices : & [] }),
        (1825, Vendor { id : 1825u16 , name : "Sapphire, Inc." , devices : & [] }),
        (7896, Vendor { id : 7896u16 , name : "Digiteq Automotive" , devices : & [Device { vendor_id : 7896u16 , id : 257u16 , name : "FG4 PCIe Frame Grabber (T100)" , subsystems : & [] } , Device { vendor_id : 7896u16 , id : 513u16 , name : "FG4 PCIe Frame Grabber (T200)" , subsystems : & [] }] }),
        (5049, Vendor { id : 5049u16 , name : "Elecom Co Ltd" , devices : & [] }),
        (4429, Vendor { id : 4429u16 , name : "IC Corporation" , devices : & [] }),
        (798, Vendor { id : 798u16 , name : "ANDES" , devices : & [] }),
        (5781, Vendor { id : 5781u16 , name : "EPoX Computer Co., Ltd." , devices : & [] }),
        (5138, Vendor { id : 5138u16 , name : "VIA Technologies Inc." , devices : & [Device { vendor_id : 5138u16 , id : 5906u16 , name : "ICE1712 [Envy24] PCI Multi-Channel I/O Controller" , subsystems : & [SubSystem { subvendor : 5138u16 , subdevice : 5906u16 , name : "Hoontech ST Audio DSP 24" } , SubSystem { subvendor : 5138u16 , subdevice : 54832u16 , name : "M-Audio Delta 1010" } , SubSystem { subvendor : 5138u16 , subdevice : 54833u16 , name : "M-Audio Delta DiO" } , SubSystem { subvendor : 5138u16 , subdevice : 54834u16 , name : "M-Audio Delta 66" } , SubSystem { subvendor : 5138u16 , subdevice : 54835u16 , name : "M-Audio Delta 44" } , SubSystem { subvendor : 5138u16 , subdevice : 54836u16 , name : "M-Audio Delta Audiophile 2496" } , SubSystem { subvendor : 5138u16 , subdevice : 54837u16 , name : "M-Audio Delta TDIF" } , SubSystem { subvendor : 5138u16 , subdevice : 54839u16 , name : "M-Audio Delta RBUS" } , SubSystem { subvendor : 5138u16 , subdevice : 54840u16 , name : "M-Audio Delta 410" } , SubSystem { subvendor : 5138u16 , subdevice : 54843u16 , name : "M-Audio Delta 1010LT" } , SubSystem { subvendor : 5138u16 , subdevice : 54844u16 , name : "Digigram VX442" } , SubSystem { subvendor : 5142u16 , subdevice : 5906u16 , name : "Hoontech ST Audio DSP 24 Media 7.1" } , SubSystem { subvendor : 5435u16 , subdevice : 4373u16 , name : "EWS88 MT" } , SubSystem { subvendor : 5435u16 , subdevice : 4389u16 , name : "EWS88 MT (Master)" } , SubSystem { subvendor : 5435u16 , subdevice : 4395u16 , name : "EWS88 D" } , SubSystem { subvendor : 5435u16 , subdevice : 4396u16 , name : "EWS88 D (Master)" } , SubSystem { subvendor : 5435u16 , subdevice : 4400u16 , name : "EWX 24/96" } , SubSystem { subvendor : 5435u16 , subdevice : 4408u16 , name : "DMX 6fire 24/96" } , SubSystem { subvendor : 5435u16 , subdevice : 4433u16 , name : "PHASE88" } , SubSystem { subvendor : 5838u16 , subdevice : 4160u16 , name : "Edirol DA-2496" }] } , Device { vendor_id : 5138u16 , id : 5924u16 , name : "VT1720/24 [Envy24PT/HT] PCI Multi-Channel Audio Controller" , subsystems : & [SubSystem { subvendor : 4272u16 , subdevice : 512u16 , name : "Hollywood@Home 7.1" } , SubSystem { subvendor : 5138u16 , subdevice : 5924u16 , name : "Albatron PX865PE 7.1" } , SubSystem { subvendor : 5138u16 , subdevice : 13872u16 , name : "M-Audio Revolution 7.1" } , SubSystem { subvendor : 5138u16 , subdevice : 13873u16 , name : "M-Audio Revolution 5.1" } , SubSystem { subvendor : 5138u16 , subdevice : 13874u16 , name : "M-Audio Audiophile 192" } , SubSystem { subvendor : 5435u16 , subdevice : 4421u16 , name : "Aureon 7.1 Space" } , SubSystem { subvendor : 5435u16 , subdevice : 4423u16 , name : "Aureon 5.1 Sky" } , SubSystem { subvendor : 5435u16 , subdevice : 4432u16 , name : "PHASE 22" } , SubSystem { subvendor : 5435u16 , subdevice : 4435u16 , name : "Aureon 7.1 Universe" } , SubSystem { subvendor : 6059u16 , subdevice : 6406u16 , name : "PSC 724 [Ultimate Edge]" } , SubSystem { subvendor : 9999u16 , subdevice : 63041u16 , name : "ZNF3-150" } , SubSystem { subvendor : 9999u16 , subdevice : 63045u16 , name : "ZNF3-250" } , SubSystem { subvendor : 12592u16 , subdevice : 16724u16 , name : "MAYA 44 MKII" }] }] }),
        (8142, Vendor { id : 8142u16 , name : "Cognio Inc." , devices : & [Device { vendor_id : 8142u16 , id : 1u16 , name : "Spectrum Analyzer PC Card (SAgE)" , subsystems : & [] }] }),
        (7602, Vendor { id : 7602u16 , name : "ATP Electronics, Inc." , devices : & [] }),
        (4740, Vendor { id : 4740u16 , name : "Sahara Networks, Inc." , devices : & [] }),
        (4796, Vendor { id : 4796u16 , name : "Array Microsystems" , devices : & [] }),
        (6906, Vendor { id : 6906u16 , name : "J & W Electronics Co., Ltd." , devices : & [] }),
        (5321, Vendor { id : 5321u16 , name : "ODIN Telesystems Inc" , devices : & [] }),
        (4323, Vendor { id : 4323u16 , name : "Tundra Semiconductor Corp." , devices : & [Device { vendor_id : 4323u16 , id : 0u16 , name : "CA91C042 [Universe]" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 264u16 , name : "Tsi108 Host Bridge for Single PowerPC" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 328u16 , name : "Tsi148 [Tempe]" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "VR11 Single Board Computer" }] } , Device { vendor_id : 4323u16 , id : 2144u16 , name : "CA91C860 [QSpan]" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 2146u16 , name : "CA91C862A [QSpan-II]" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 33041u16 , name : "Tsi381 PCIe to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 33043u16 , name : "89HPEB383 PCIe-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 33376u16 , name : "CA91L8200B [Dual PCI PowerSpan II]" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 33377u16 , name : "CA91L8260B [Single PCI PowerSpan II]" , subsystems : & [] } , Device { vendor_id : 4323u16 , id : 41224u16 , name : "Tsi109 Host Bridge for Dual PowerPC" , subsystems : & [] }] }),
        (5534, Vendor { id : 5534u16 , name : "A-Max Technology Co Ltd" , devices : & [] }),
        (4113, Vendor { id : 4113u16 , name : "Digital Equipment Corporation" , devices : & [Device { vendor_id : 4113u16 , id : 1u16 , name : "DECchip 21050" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 2u16 , name : "DECchip 21040 [Tulip]" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 4u16 , name : "DECchip 21030 [TGA]" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 7u16 , name : "NVRAM [Zephyr NVRAM]" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 8u16 , name : "KZPSA [KZPSA]" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 9u16 , name : "DECchip 21140 [FasterNet]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 784u16 , name : "21140 Fast Ethernet" } , SubSystem { subvendor : 4280u16 , subdevice : 8193u16 , name : "SMC9332BDT EtherPower 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 8194u16 , name : "SMC9332BVT EtherPower T4 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 8195u16 , name : "SMC9334BDT EtherPower 10/100 (1-port)" } , SubSystem { subvendor : 4361u16 , subdevice : 9216u16 , name : "ANA-6944A/TX Fast Ethernet" } , SubSystem { subvendor : 4370u16 , subdevice : 8960u16 , name : "RNS2300 Fast Ethernet" } , SubSystem { subvendor : 4370u16 , subdevice : 8992u16 , name : "RNS2320 Fast Ethernet" } , SubSystem { subvendor : 4370u16 , subdevice : 9024u16 , name : "RNS2340 Fast Ethernet" } , SubSystem { subvendor : 4371u16 , subdevice : 4615u16 , name : "EN-1207-TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4352u16 , name : "DFE-500TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4370u16 , name : "DFE-570TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4416u16 , name : "DFE-660 Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4486u16 , subdevice : 4418u16 , name : "DFE-660 Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4598u16 , subdevice : 1283u16 , name : "Freedomline Fast Ethernet" } , SubSystem { subvendor : 4738u16 , subdevice : 37120u16 , name : "AEF-380TXD Fast Ethernet" } , SubSystem { subvendor : 4997u16 , subdevice : 4352u16 , name : "FA310TX Fast Ethernet" } , SubSystem { subvendor : 9798u16 , subdevice : 1u16 , name : "KNE100TX Fast Ethernet" }] } , Device { vendor_id : 4113u16 , id : 10u16 , name : "21230 Video Codec" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 13u16 , name : "PBXGB [TGA2]" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 15u16 , name : "PCI-to-PDQ Interface Chip [PFI] FDDI (DEFPA)" , subsystems : & [SubSystem { subvendor : 4113u16 , subdevice : 57073u16 , name : "FDDIcontroller/PCI (DEFPA)" } , SubSystem { subvendor : 4156u16 , subdevice : 57073u16 , name : "FDDIcontroller/PCI (3X-DEFPA)" }] } , Device { vendor_id : 4113u16 , id : 20u16 , name : "DECchip 21041 [Tulip Pass 3]" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 256u16 , name : "DE-530+" }] } , Device { vendor_id : 4113u16 , id : 22u16 , name : "ATMworks 350 Adapter [OPPO] (DGLPB)" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 23u16 , name : "PV-PCI Graphics Controller (ZLXp-L)" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 24u16 , name : "Memory Channel interface" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 25u16 , name : "DECchip 21142/43" , subsystems : & [SubSystem { subvendor : 4113u16 , subdevice : 20490u16 , name : "DE500A Fast Ethernet" } , SubSystem { subvendor : 4113u16 , subdevice : 20491u16 , name : "DE500B Fast Ethernet" } , SubSystem { subvendor : 4116u16 , subdevice : 1u16 , name : "10/100 EtherJet Cardbus" } , SubSystem { subvendor : 4133u16 , subdevice : 789u16 , name : "ALN315 Fast Ethernet" } , SubSystem { subvendor : 4147u16 , subdevice : 32780u16 , name : "PC-9821-CS01 100BASE-TX Interface Card" } , SubSystem { subvendor : 4147u16 , subdevice : 32781u16 , name : "PC-9821NR-B06 100BASE-TX Interface Card" } , SubSystem { subvendor : 4156u16 , subdevice : 4698u16 , name : "10/100Base-TX (PCI) [A5506B]" } , SubSystem { subvendor : 4237u16 , subdevice : 22u16 , name : "Rapidfire 2327 10/100 Ethernet" } , SubSystem { subvendor : 4237u16 , subdevice : 23u16 , name : "GoCard 2250 Ethernet 10/100 Cardbus" } , SubSystem { subvendor : 4280u16 , subdevice : 8197u16 , name : "SMC8032DT Extreme Ethernet 10/100" } , SubSystem { subvendor : 4280u16 , subdevice : 32820u16 , name : "SMC8034 Extreme Ethernet 10/100" } , SubSystem { subvendor : 4335u16 , subdevice : 33129u16 , name : "Cardbus Fast Ethernet" } , SubSystem { subvendor : 4361u16 , subdevice : 10752u16 , name : "ANA-6911A/TX Fast Ethernet" } , SubSystem { subvendor : 4361u16 , subdevice : 11008u16 , name : "ANA-6911A/TXC Fast Ethernet" } , SubSystem { subvendor : 4361u16 , subdevice : 12288u16 , name : "ANA-6922/TX Fast Ethernet" } , SubSystem { subvendor : 4371u16 , subdevice : 4615u16 , name : "Cheetah Fast Ethernet" } , SubSystem { subvendor : 4371u16 , subdevice : 8736u16 , name : "Cardbus Fast Ethernet" } , SubSystem { subvendor : 4445u16 , subdevice : 2u16 , name : "Cardbus Ethernet 10/100" } , SubSystem { subvendor : 4473u16 , subdevice : 515u16 , name : "Fast Ethernet" } , SubSystem { subvendor : 4473u16 , subdevice : 516u16 , name : "Cardbus Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4352u16 , name : "DFE-500TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4353u16 , name : "DFE-500TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4354u16 , name : "DFE-500TX Fast Ethernet" } , SubSystem { subvendor : 4486u16 , subdevice : 4370u16 , name : "DFE-570TX Quad Fast Ethernet" } , SubSystem { subvendor : 4592u16 , subdevice : 16949u16 , name : "21143 [FASTLine-II UTP 10/100]" } , SubSystem { subvendor : 4697u16 , subdevice : 10240u16 , name : "AT-2800Tx Fast Ethernet" } , SubSystem { subvendor : 4710u16 , subdevice : 4u16 , name : "Eagle Fast EtherMAX" } , SubSystem { subvendor : 4783u16 , subdevice : 25u16 , name : "NetFlyer Cardbus Fast Ethernet" } , SubSystem { subvendor : 4980u16 , subdevice : 1u16 , name : "Cardbus Ethernet Card 10/100" } , SubSystem { subvendor : 4980u16 , subdevice : 2u16 , name : "Cardbus Ethernet Card 10/100" } , SubSystem { subvendor : 4980u16 , subdevice : 7u16 , name : "Cardbus Ethernet Card 10/100" } , SubSystem { subvendor : 4980u16 , subdevice : 8u16 , name : "Cardbus Ethernet Card 10/100" } , SubSystem { subvendor : 4997u16 , subdevice : 8448u16 , name : "FA510" } , SubSystem { subvendor : 5013u16 , subdevice : 1u16 , name : "10/100 Ethernet CardBus PC Card" } , SubSystem { subvendor : 5073u16 , subdevice : 43777u16 , name : "EtherFast 10/100 Cardbus (PCMPC200)" } , SubSystem { subvendor : 5272u16 , subdevice : 10u16 , name : "TPMC880-10 10/100Base-T and 10Base2 PMC Ethernet Adapter" } , SubSystem { subvendor : 5272u16 , subdevice : 11u16 , name : "TPMC880-11 Single 10/100Base-T PMC Ethernet Adapter" } , SubSystem { subvendor : 5272u16 , subdevice : 12u16 , name : "TPMC880-12 Single 10Base2 PMC Ethernet Adapter" } , SubSystem { subvendor : 5323u16 , subdevice : 256u16 , name : "LNDL-100N 100Base-TX Ethernet PC Card" } , SubSystem { subvendor : 5736u16 , subdevice : 8192u16 , name : "FastNet Pro (PE2000)" } , SubSystem { subvendor : 9798u16 , subdevice : 1u16 , name : "KNE100TX" } , SubSystem { subvendor : 9798u16 , subdevice : 2u16 , name : "KNE-CB4TX" } , SubSystem { subvendor : 32902u16 , subdevice : 1u16 , name : "EtherExpress PRO/100 Mobile CardBus 32" }] } , Device { vendor_id : 4113u16 , id : 26u16 , name : "Farallon PN9000SX Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 33u16 , name : "DECchip 21052" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 34u16 , name : "DECchip 21150" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 35u16 , name : "DECchip 21150" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 36u16 , name : "DECchip 21152" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 37u16 , name : "DECchip 21153" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 38u16 , name : "DECchip 21154" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 52u16 , name : "56k Modem Cardbus" , subsystems : & [SubSystem { subvendor : 4980u16 , subdevice : 3u16 , name : "56k Modem Cardbus" }] } , Device { vendor_id : 4113u16 , id : 69u16 , name : "DECchip 21553" , subsystems : & [] } , Device { vendor_id : 4113u16 , id : 70u16 , name : "DECchip 21554" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 16464u16 , name : "Smart Array 4200 Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16465u16 , name : "Smart Array 4250ES Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 16472u16 , name : "Smart Array 431 Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 4290u16 , name : "NetRAID-4M" } , SubSystem { subvendor : 4825u16 , subdevice : 10u16 , name : "IP Telephony card" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" } , SubSystem { subvendor : 36869u16 , subdevice : 868u16 , name : "5400S (Mustang)" } , SubSystem { subvendor : 36869u16 , subdevice : 869u16 , name : "5400S (Mustang)" } , SubSystem { subvendor : 36869u16 , subdevice : 4964u16 , name : "Dell PowerEdge RAID Controller 2" } , SubSystem { subvendor : 36869u16 , subdevice : 4965u16 , name : "Dell PowerEdge RAID Controller 2" } , SubSystem { subvendor : 58559u16 , subdevice : 4096u16 , name : "CC8-1-BLUES" }] } , Device { vendor_id : 4113u16 , id : 4197u16 , name : "StrongARM DC21285" , subsystems : & [SubSystem { subvendor : 4201u16 , subdevice : 32u16 , name : "DAC960P / DAC1164P" }] }] }),
        (606, Vendor { id : 606u16 , name : "Solidigm" , devices : & [Device { vendor_id : 606u16 , id : 2912u16 , name : "NVMe DC SSD [Sentinel Rock Plus controller]" , subsystems : & [SubSystem { subvendor : 606u16 , subdevice : 32776u16 , name : "NVMe DC SSD U.2 15mm [D7-P5510]" } , SubSystem { subvendor : 606u16 , subdevice : 33288u16 , name : "NVMe DC SSD U.2 15mm [D7-P5810]" } , SubSystem { subvendor : 606u16 , subdevice : 36125u16 , name : "NVMe DC SSD E1.L 9.5mm [D5-P5316]" } , SubSystem { subvendor : 606u16 , subdevice : 36872u16 , name : "NVMe DC SSD U.2 15mm [D7-P5520]" } , SubSystem { subvendor : 606u16 , subdevice : 36876u16 , name : "NVMe DC SSD E1.S 9.5mm [D7-P5520]" } , SubSystem { subvendor : 606u16 , subdevice : 36877u16 , name : "NVMe DC SSD E1.S 15mm [D7-P5520]" } , SubSystem { subvendor : 606u16 , subdevice : 36892u16 , name : "NVMe DC SSD E1.L 9.5mm [D7-P5520]" } , SubSystem { subvendor : 606u16 , subdevice : 37128u16 , name : "NVMe DC SSD U.2 15mm [D7-P5620]" } , SubSystem { subvendor : 606u16 , subdevice : 49160u16 , name : "NVMe DC SSD U.2 15mm [D5-P5530]" } , SubSystem { subvendor : 606u16 , subdevice : 54280u16 , name : "NVMe DC SSD U.2 15mm [D5-P5430]" } , SubSystem { subvendor : 606u16 , subdevice : 54284u16 , name : "NVMe DC SSD E1.S 9.5mm [D5-P5430]" } , SubSystem { subvendor : 606u16 , subdevice : 54297u16 , name : "NVMe DC SSD E3.S 7.5mm [D5-P5430]" } , SubSystem { subvendor : 606u16 , subdevice : 55304u16 , name : "NVMe DC SSD U.2 15mm [D5-P5336]" } , SubSystem { subvendor : 606u16 , subdevice : 55321u16 , name : "NVMe DC SSD E3.S 7.5mm [D5-P5336]" } , SubSystem { subvendor : 606u16 , subdevice : 55324u16 , name : "NVMe DC SSD E1.L 18mm [D5-P5336]" } , SubSystem { subvendor : 606u16 , subdevice : 55325u16 , name : "NVMe DC SSD E1.L 9.5mm [D5-P5336]" }] } , Device { vendor_id : 606u16 , id : 2928u16 , name : "NVMe DC SSD [Yorktown controller]" , subsystems : & [] } , Device { vendor_id : 606u16 , id : 11097u16 , name : "NVMe DC SSD [Atomos Prime]" , subsystems : & [SubSystem { subvendor : 606u16 , subdevice : 8u16 , name : "NVMe DC SSD U.2-SFF 15mm [D7-PS1010]" } , SubSystem { subvendor : 606u16 , subdevice : 12u16 , name : "NVMe DC SSD E1.S 9.5mm [D7-PS1010]" } , SubSystem { subvendor : 606u16 , subdevice : 13u16 , name : "NVMe DC SSD E1.S 15mm [D7-PS1010]" } , SubSystem { subvendor : 606u16 , subdevice : 25u16 , name : "NVMe DC SSD E3.S-1T 7.5mm [D7-PS1010]" } , SubSystem { subvendor : 606u16 , subdevice : 264u16 , name : "NVMe DC SSD U.2-SFF 15mm [D7-PS1030]" } , SubSystem { subvendor : 606u16 , subdevice : 281u16 , name : "NVMe DC SSD E3.S-1T 7.5mm [D7-PS1030]" } , SubSystem { subvendor : 4238u16 , subdevice : 18592u16 , name : "NVMe DC SSD U.2-SFF 15mm 3.84TB [D7-PS1010 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18593u16 , name : "NVMe DC SSD U.2-SFF 15mm 7.68TB [D7-PS1010 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18594u16 , name : "NVMe DC SSD U.2-SFF 15mm 15.36TB [D7-PS1010 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18595u16 , name : "NVMe DC SSD Add-In-Card [D7-PS1030 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18596u16 , name : "NVMe DC SSD E3.S-1T 7.5mm 3.84TB [D7-PS1010 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18597u16 , name : "NVMe DC SSD E3.S-1T 7.5mm 7.68TB [D7-PS1010 Custom]" } , SubSystem { subvendor : 4238u16 , subdevice : 18598u16 , name : "NVMe DC SSD E3.S-1T 7.5mm 15.36TB [D7-PS1010 Custom]" }] } , Device { vendor_id : 606u16 , id : 61867u16 , name : "P41 Plus NVMe SSD (DRAM-less) [Echo Harbor]" , subsystems : & [] } , Device { vendor_id : 606u16 , id : 61868u16 , name : "P44 Pro NVMe SSD [Hollywood Beach]" , subsystems : & [] }] }),
        (4353, Vendor { id : 4353u16 , name : "Initio Corporation" , devices : & [Device { vendor_id : 4353u16 , id : 2u16 , name : "INI-920 Ultra SCSI Adapter" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 4192u16 , name : "INI-A100U2W" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 5666u16 , name : "INI-1623 PCI SATA-II Controller" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 37120u16 , name : "INI-9100/9100W" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 37888u16 , name : "INI-940 Fast Wide SCSI Adapter" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 37889u16 , name : "INI-935 Fast Wide SCSI Adapter" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 38144u16 , name : "INI-950 SCSI Adapter" , subsystems : & [] } , Device { vendor_id : 4353u16 , id : 38146u16 , name : "INI-950P Ultra Wide SCSI Adapter" , subsystems : & [] }] }),
        (10114, Vendor { id : 10114u16 , name : "Emdoor Digital Technology Co., Ltd" , devices : & [] }),
        (52701, Vendor { id : 52701u16 , name : "Tyzx, Inc." , devices : & [Device { vendor_id : 52701u16 , id : 257u16 , name : "DeepSea 1 High Speed Stereo Vision Frame Grabber" , subsystems : & [] } , Device { vendor_id : 52701u16 , id : 512u16 , name : "DeepSea 2 High Speed Stereo Vision Frame Grabber" , subsystems : & [] }] }),
        (4508, Vendor { id : 4508u16 , name : "Information Technology Inst." , devices : & [] }),
        (16707, Vendor { id : 16707u16 , name : "Digital Equipment Corp" , devices : & [] }),
        (4857, Vendor { id : 4857u16 , name : "Four Fold Ltd" , devices : & [] }),
        (5272, Vendor { id : 5272u16 , name : "TEWS Technologies GmbH" , devices : & [Device { vendor_id : 5272u16 , id : 816u16 , name : "TPMC816 2 Channel CAN bus controller." , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 861u16 , name : "TPMC861 4-Channel Isolated Serial Interface RS422/RS485" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 901u16 , name : "TPMC901 Extended CAN bus with 2/4/6 CAN controller" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 8652u16 , name : "TCP460 CompactPCI 16 Channel Serial Interface RS232/RS422" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 8653u16 , name : "TCP461 CompactPCI 8 Channel Serial Interface RS232/RS422" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 12388u16 , name : "TPCI100 (2 Slot IndustryPack PCI Carrier)" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 12488u16 , name : "TPCI200 4 Slot IndustryPack PCI Carrier" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 28872u16 , name : "TPCE200 4 Slot IndustryPack PCIe Carrier" , subsystems : & [] } , Device { vendor_id : 5272u16 , id : 37239u16 , name : "TXMC375 8 channel RS232/RS422/RS485 programmable serial interface" , subsystems : & [] }] }),
        (5399, Vendor { id : 5399u16 , name : "ECHOTEK Corp" , devices : & [] }),
        (36864, Vendor { id : 36864u16 , name : "C*Core Technology Co., Ltd." , devices : & [] }),
        (4421, Vendor { id : 4421u16 , name : "Workbit Corporation" , devices : & [Device { vendor_id : 4421u16 , id : 32775u16 , name : "NinjaSCSI-32 Workbit" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61447u16 , name : "NinjaSCSI-32 KME" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61456u16 , name : "NinjaSCSI-32 Workbit" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61458u16 , name : "NinjaSCSI-32 Logitec" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61459u16 , name : "NinjaSCSI-32 Logitec" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61461u16 , name : "NinjaSCSI-32 Melco" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61472u16 , name : "NinjaSCSI-32 Sony PCGA-DVD51" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61473u16 , name : "NinjaPATA-32 Delkin Cardbus UDMA" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61476u16 , name : "NinjaPATA-32 Delkin Cardbus UDMA" , subsystems : & [] } , Device { vendor_id : 4421u16 , id : 61699u16 , name : "NinjaPATA-32 Delkin Cardbus UDMA" , subsystems : & [] }] }),
        (5340, Vendor { id : 5340u16 , name : "Amplicon Liveline Ltd" , devices : & [Device { vendor_id : 5340u16 , id : 0u16 , name : "PCI230" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 1u16 , name : "PCI242" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 2u16 , name : "PCI244" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 3u16 , name : "PCI247" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 4u16 , name : "PCI248" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 5u16 , name : "PCI249" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 6u16 , name : "PCI260" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 7u16 , name : "PCI224" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 8u16 , name : "PCI234" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 9u16 , name : "PCI236" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 10u16 , name : "PCI272" , subsystems : & [] } , Device { vendor_id : 5340u16 , id : 11u16 , name : "PCI215" , subsystems : & [] }] }),
        (7767, Vendor { id : 7767u16 , name : "Beijing Panyi Technology Co., Ltd" , devices : & [Device { vendor_id : 7767u16 , id : 256u16 , name : "The device has already been deleted." , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 256u16 , name : "PY8800 64GB Accelerator" }] }] }),
        (4923, Vendor { id : 4923u16 , name : "Softcom Microsystems" , devices : & [] }),
        (4987, Vendor { id : 4987u16 , name : "PPT Vision" , devices : & [] }),
        (5082, Vendor { id : 5082u16 , name : "Intresource Systems pte Ltd" , devices : & [] }),
        (4557, Vendor { id : 4557u16 , name : "HAL Computer Systems, Inc." , devices : & [] }),
        (5203, Vendor { id : 5203u16 , name : "MYCOM Inc" , devices : & [] }),
        (4312, Vendor { id : 4312u16 , name : "Advanced Peripherals Labs" , devices : & [] }),
        (57689, Vendor { id : 57689u16 , name : "Tiger Jet Network Inc. / ICP DAS" , devices : & [Device { vendor_id : 57689u16 , id : 1u16 , name : "Tiger3XX Modem/ISDN interface" , subsystems : & [SubSystem { subvendor : 89u16 , subdevice : 1u16 , name : "128k ISDN-S/T Adapter" } , SubSystem { subvendor : 89u16 , subdevice : 3u16 , name : "128k ISDN-U Adapter" } , SubSystem { subvendor : 167u16 , subdevice : 1u16 , name : "TELES.S0/PCI 2.x ISDN Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 3u16 , name : "Digium X100P/X101P analogue PSTN FXO interface" } , SubSystem { subvendor : 45312u16 , subdevice : 3u16 , name : "OpenVox A400P 4-port analog card" } , SubSystem { subvendor : 45529u16 , subdevice : 3u16 , name : "AX400P 4-port analog card" }] } , Device { vendor_id : 57689u16 , id : 2u16 , name : "Tiger100APC ISDN chipset" , subsystems : & [] }] }),
        (4310, Vendor { id : 4310u16 , name : "Cetia" , devices : & [] }),
        (5050, Vendor { id : 5050u16 , name : "Oxford Instruments" , devices : & [] }),
        (5003, Vendor { id : 5003u16 , name : "Tokimec Inc" , devices : & [] }),
        (4998, Vendor { id : 4998u16 , name : "Video Domain Technologies" , devices : & [] }),
        (7383, Vendor { id : 7383u16 , name : "Nanjing Magewell Electronics Co., Ltd." , devices : & [Device { vendor_id : 7383u16 , id : 2u16 , name : "Pro Capture AIO" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 16u16 , name : "Pro Capture Endpoint" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 20u16 , name : "PRO CAPTURE AIO 4K PLUS" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 21u16 , name : "Pro Capture HDMI 4K +" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 23u16 , name : "PRO CAPTURE AIO 4K" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 81u16 , name : "Eco Capture Dual HDMI M.2" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 82u16 , name : "Eco Capture HDMI 4K M.2" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 83u16 , name : "Eco Capture Dual SDI M.2" , subsystems : & [] } , Device { vendor_id : 7383u16 , id : 84u16 , name : "Eco Capture Quad SDI M.2" , subsystems : & [] }] }),
        (8107, Vendor { id : 8107u16 , name : "Unifabrix Ltd." , devices : & [Device { vendor_id : 8107u16 , id : 0u16 , name : "Nexus Alpha IVPU" , subsystems : & [] } , Device { vendor_id : 8107u16 , id : 256u16 , name : "NoX Gamma" , subsystems : & [] } , Device { vendor_id : 8107u16 , id : 509u16 , name : "Smart Memory Node Generic CXL Port (T1)" , subsystems : & [] } , Device { vendor_id : 8107u16 , id : 510u16 , name : "Smart Memory Node Generic CXL Port (T2)" , subsystems : & [] } , Device { vendor_id : 8107u16 , id : 511u16 , name : "Smart Memory Node Generic CXL Port (T3)" , subsystems : & [] } , Device { vendor_id : 8107u16 , id : 6912u16 , name : "MAX Host Device" , subsystems : & [] }] }),
        (5010, Vendor { id : 5010u16 , name : "Medialight Inc" , devices : & [] }),
        (5657, Vendor { id : 5657u16 , name : "FarSite Communications Ltd" , devices : & [Device { vendor_id : 5657u16 , id : 1024u16 , name : "FarSync T2P (2 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 1088u16 , name : "FarSync T4P (4 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 1552u16 , name : "FarSync T1U (1 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 1568u16 , name : "FarSync T2U (2 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 1600u16 , name : "FarSync T4U (4 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 5648u16 , name : "FarSync TE1 (T1,E1)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 5650u16 , name : "FarSync TE1 PCI Express (T1,E1)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 9744u16 , name : "FarSync DSL-S1 (SHDSL)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 13888u16 , name : "FarSync T4E (4-port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 17952u16 , name : "FarSync T2Ue PCI Express (2-port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 17984u16 , name : "FarSync T4Ue PCI Express (4-port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 22049u16 , name : "FarSync T2Ee PCI Express (2 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 22081u16 , name : "FarSync T4Ee PCI Express (4 port X.21/V.35/V.24)" , subsystems : & [] } , Device { vendor_id : 5657u16 , id : 26144u16 , name : "FarSync T2U-PMC PCI Express (2 port X.21/V.35/V.24)" , subsystems : & [] }] }),
        (7773, Vendor { id : 7773u16 , name : "ASR Microelectronics" , devices : & [Device { vendor_id : 7773u16 , id : 28672u16 , name : "AI controller A7000" , subsystems : & [] } , Device { vendor_id : 7773u16 , id : 28688u16 , name : "AI controller A7010" , subsystems : & [] }] }),
        (4213, Vendor { id : 4213u16 , name : "Advanced Integrations Research" , devices : & [] }),
        (21845, Vendor { id : 21845u16 , name : "Genroco, Inc" , devices : & [Device { vendor_id : 21845u16 , id : 3u16 , name : "TURBOstor HFP-832 [HiPPI NIC]" , subsystems : & [] } , Device { vendor_id : 21845u16 , id : 4u16 , name : "Torrent QN16e [16-128 Channel MPEG QAM Modulator for DVB-C]" , subsystems : & [] } , Device { vendor_id : 21845u16 , id : 15104u16 , name : "Epiphan DVI2PCIe video capture card" , subsystems : & [] }] }),
        (6328, Vendor { id : 6328u16 , name : "Ammasso" , devices : & [Device { vendor_id : 6328u16 , id : 45057u16 , name : "AMSO 1100 iWARP/RDMA Gigabit Ethernet Coprocessor" , subsystems : & [] }] }),
        (4317, Vendor { id : 4317u16 , name : "Evans & Sutherland" , devices : & [Device { vendor_id : 4317u16 , id : 256u16 , name : "Lightning 1200" , subsystems : & [SubSystem { subvendor : 4317u16 , subdevice : 35u16 , name : "Lightning 1200 15+16M" }] }] }),
        (5017, Vendor { id : 5017u16 , name : "Rios systems Co Ltd" , devices : & [] }),
        (6360, Vendor { id : 6360u16 , name : "Dialogue Technology Corp." , devices : & [] }),
        (4341, Vendor { id : 4341u16 , name : "NKK Corporation" , devices : & [Device { vendor_id : 4341u16 , id : 40961u16 , name : "NDR4000 [NR4600 Bridge]" , subsystems : & [] }] }),
        (32624, Vendor { id : 32624u16 , name : "Arrow Lake-S PCH CNVi WiFi" , devices : & [Device { vendor_id : 32624u16 , id : 148u16 , name : "WiFi 6E AX211 160MHz" , subsystems : & [] }] }),
        (5039, Vendor { id : 5039u16 , name : "T.Sqware" , devices : & [] }),
        (5859, Vendor { id : 5859u16 , name : "European Space Agency" , devices : & [Device { vendor_id : 5859u16 , id : 7695u16 , name : "LEON2FT Processor" , subsystems : & [] }] }),
        (5132, Vendor { id : 5132u16 , name : "Elmic Systems Inc" , devices : & [] }),
        (5507, Vendor { id : 5507u16 , name : "Inet Technologies Inc" , devices : & [] }),
        (20, Vendor { id : 20u16 , name : "Loongson Technology LLC" , devices : & [Device { vendor_id : 20u16 , id : 15119u16 , name : "DMA Adress Translation Unit [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 15369u16 , name : "Internal PCI to PCI Bridge [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 15375u16 , name : "DMA Adress Translation Unit [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 15385u16 , name : "PCI Express x16 Root Port [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 15401u16 , name : "PCI Express x8 Root Port [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 15417u16 , name : "PCI Express x4 Root Port [Loongson 3 Processor Family]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31232u16 , name : "7A1000 Chipset Hyper Transport Bridge Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31234u16 , name : "2K1000 / 7A1000 Chipset Advanced Peripheral Bus Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31235u16 , name : "2K1000/2000 / 7A1000 Chipset Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31236u16 , name : "2K1000 / 7A1000 Chipset OTG USB Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31237u16 , name : "2K1000 Vivante GC1000 GPU" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31238u16 , name : "2K1000 / 7A1000 Chipset Display Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31239u16 , name : "2K1000/2000/3000 / 3B6000M / 7A1000/2000 Chipset HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31240u16 , name : "2K1000 / 7A1000 Chipset 3Gb/s SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31241u16 , name : "2K1000 / 7A1000 Chipset PCIe x1 Bridge" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31243u16 , name : "7A1000 Chipset SPI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31244u16 , name : "2K2000 / 7A1000/2000 Chipset LPC Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31246u16 , name : "2K2000 AES Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31247u16 , name : "2K1000 / 7A1000 Chipset DMA Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31248u16 , name : "7A2000 Chipset Hyper Transport Bridge Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31251u16 , name : "2K2000 / 7A2000 Chipset Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31252u16 , name : "2K1000 / 7A1000/2000 Chipset USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31253u16 , name : "7A1000 Chipset Vivante GC1000 GPU" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31254u16 , name : "2K1000/2000 VPU Decoder" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31255u16 , name : "7A1000 Chipset AC97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31256u16 , name : "2K2000/3000 / 3B6000M / 7A2000 Chipset 6Gb/s SATA AHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31257u16 , name : "PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31258u16 , name : "2K2000 Configuration Bus" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31259u16 , name : "2K2000/3000 / 3B6000M / 7A2000 Chipset SPI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31261u16 , name : "2K2000 / 2K3000 / 3B6000M RapidIO Interface" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31262u16 , name : "2K2000 DES Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31266u16 , name : "2K2000 Advanced Peripheral Bus Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31267u16 , name : "Gigabit Ethernet Controller [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31268u16 , name : "2K1000 / 7A1000/2000 Chipset USB OHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31269u16 , name : "2K2000 / 7A2000 Chipset LG100 GPU" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31270u16 , name : "2K1000 Camera Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31271u16 , name : "2K2000/3000 / 3B6000M / 7A2000 Chipset I2S Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31273u16 , name : "7A1000 Chipset PCIe x8 Bridge" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31278u16 , name : "2K2000 RSA Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31279u16 , name : "2K2000 DMA Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31284u16 , name : "2K2000/3000 / 3B6000M / 7A2000 Chipset USB 3.0 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31285u16 , name : "LG200 GPU" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31286u16 , name : "2K2000 / 7A2000 Chipset Display Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31287u16 , name : "2K2000/3000 / 3B6000M / 7A2000 Chipset HDMI Audio Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31289u16 , name : "2K2000 / 7A2000 Chipset PCIe x1 Root Port" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31294u16 , name : "2K2000 RNG Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31298u16 , name : "Advanced Peripheral Bus Controller [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31300u16 , name : "2K2000/3000 / 3B6000M USB 2.0 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31302u16 , name : "Video Display Controller [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31303u16 , name : "Pulse-Code Modulation [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31304u16 , name : "2K2000/3000 / 3B6000M SDIO Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31305u16 , name : "2K2000 / 7A2000 Chipset PCIe x4 Root Port" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31316u16 , name : "2K2000 OTG USB Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31318u16 , name : "Video Processing Unit Decoder [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31321u16 , name : "7A2000 Chipset PCIe x8 Root Port" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31334u16 , name : "Video Processing Unit Encoder [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31337u16 , name : "7A2000 Chipset PCIe x16 Root Port" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31353u16 , name : "2K2000 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31368u16 , name : "2K2000/3000 / 3B6000M eMMC Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31369u16 , name : "PCI Express x1 Root Port [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31374u16 , name : "2K2000 SE Controller" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31385u16 , name : "PCI Express x4 Root Port [Chipset / CPU inside]" , subsystems : & [] } , Device { vendor_id : 20u16 , id : 31481u16 , name : "2K2000 PCIe Endpoint" , subsystems : & [] }] }),
        (5439, Vendor { id : 5439u16 , name : "MIPS Technologies, Inc." , devices : & [Device { vendor_id : 5439u16 , id : 1u16 , name : "SOC-it 101 System Controller" , subsystems : & [] }] }),
        (5441, Vendor { id : 5441u16 , name : "MACHONE Communications" , devices : & [] }),
        (4728, Vendor { id : 4728u16 , name : "Transtech Parallel Systems Ltd." , devices : & [Device { vendor_id : 4728u16 , id : 1793u16 , name : "TPE3/TM3 PowerPC Node" , subsystems : & [] } , Device { vendor_id : 4728u16 , id : 1808u16 , name : "TPE5 PowerPC PCI board" , subsystems : & [] } , Device { vendor_id : 4728u16 , id : 4352u16 , name : "PMC-FPGA02" , subsystems : & [] } , Device { vendor_id : 4728u16 , id : 4353u16 , name : "TS-C43 card with 4 ADSP-TS101 processors" , subsystems : & [] }] }),
        (4629, Vendor { id : 4629u16 , name : "Interware Co., Ltd" , devices : & [] }),
        (4843, Vendor { id : 4843u16 , name : "Aureal Semiconductor" , devices : & [Device { vendor_id : 4843u16 , id : 1u16 , name : "Vortex 1" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 768u16 , name : "Terasound A3D PCI" } , SubSystem { subvendor : 4173u16 , subdevice : 32822u16 , name : "AU8820 Vortex Digital Audio Processor" } , SubSystem { subvendor : 4242u16 , subdevice : 8192u16 , name : "Sonic Impact A3D" } , SubSystem { subvendor : 4242u16 , subdevice : 8448u16 , name : "Sonic Impact A3D" } , SubSystem { subvendor : 4242u16 , subdevice : 8464u16 , name : "Sonic Impact A3D" } , SubSystem { subvendor : 4242u16 , subdevice : 8704u16 , name : "Sonic Impact A3D" } , SubSystem { subvendor : 4653u16 , subdevice : 4098u16 , name : "SC 338-A3D" } , SubSystem { subvendor : 4843u16 , subdevice : 1u16 , name : "AU8820 Vortex Digital Audio Processor" } , SubSystem { subvendor : 20563u16 , subdevice : 13141u16 , name : "Montego" } , SubSystem { subvendor : 20658u16 , subdevice : 4369u16 , name : "XLerate" }] } , Device { vendor_id : 4843u16 , id : 2u16 , name : "Vortex 2" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32841u16 , name : "AU8830 Vortex 3D Digital Audio Processor" } , SubSystem { subvendor : 4173u16 , subdevice : 32891u16 , name : "AU8830 Vortex 3D Digital Audio Processor" } , SubSystem { subvendor : 4242u16 , subdevice : 12288u16 , name : "Monster Sound II" } , SubSystem { subvendor : 4242u16 , subdevice : 12289u16 , name : "Monster Sound II" } , SubSystem { subvendor : 4242u16 , subdevice : 12290u16 , name : "Monster Sound II" } , SubSystem { subvendor : 4242u16 , subdevice : 12291u16 , name : "Monster Sound II" } , SubSystem { subvendor : 4242u16 , subdevice : 12292u16 , name : "Monster Sound II" } , SubSystem { subvendor : 4843u16 , subdevice : 2u16 , name : "AU8830 Vortex 3D Digital Audio Processor" } , SubSystem { subvendor : 4843u16 , subdevice : 136u16 , name : "AU8830 Vortex 3D Digital Audio Processor" } , SubSystem { subvendor : 5197u16 , subdevice : 13584u16 , name : "AU8830 Vortex 3D Digital Audio Processor" } , SubSystem { subvendor : 20563u16 , subdevice : 13142u16 , name : "Montego II" }] } , Device { vendor_id : 4843u16 , id : 3u16 , name : "AU8810 Vortex Digital Audio Processor" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32841u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 4173u16 , subdevice : 32887u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 4255u16 , subdevice : 4096u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 4843u16 , subdevice : 3u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 5218u16 , subdevice : 26496u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 5284u16 , subdevice : 8307u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 5284u16 , subdevice : 8337u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 5284u16 , subdevice : 8452u16 , name : "AU8810 Vortex Digital Audio Processor" } , SubSystem { subvendor : 5284u16 , subdevice : 8454u16 , name : "AU8810 Vortex Digital Audio Processor" }] } , Device { vendor_id : 4843u16 , id : 34819u16 , name : "Vortex 56k Software Modem" , subsystems : & [SubSystem { subvendor : 4843u16 , subdevice : 34819u16 , name : "Vortex 56k Software Modem" }] }] }),
        (6888, Vendor { id : 6888u16 , name : "Basler AG" , devices : & [Device { vendor_id : 6888u16 , id : 1873u16 , name : "mE5 marathon VCL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1874u16 , name : "mE5 marathon AF2" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1875u16 , name : "mE5 marathon ACX QP" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1876u16 , name : "mE5 marathon ACL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1877u16 , name : "mE5 marathon ACX SP" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1878u16 , name : "mE5 marathon ACX DP" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1879u16 , name : "mE5 marathon VCX QP" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1880u16 , name : "mE5 marathon VF2" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 1881u16 , name : "mE5 marathon VCLx" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2624u16 , name : "microEnable IV AD1-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2625u16 , name : "microEnable IV VD1-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2626u16 , name : "microEnable IV AD4-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2628u16 , name : "microEnable IV VD4-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2629u16 , name : "microEnable IV AS1-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2643u16 , name : "microEnable 5 AQ8-CXP6B" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2644u16 , name : "microEnable 5 VQ8-CXP6B" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2646u16 , name : "microEnable 5 VQ8-CXP6D" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2647u16 , name : "microEnable 5 AQ8-CXP6D" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2648u16 , name : "microEnable 5 VD8-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2650u16 , name : "microEnable 5 AD8-CL" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2658u16 , name : "imaFlex CXP-12 Quad" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2660u16 , name : "imaWorx CXP-12 Quad" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2665u16 , name : "imaFlex CXP-12 Penta" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2898u16 , name : "mE5 Abacus 4G Base" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2899u16 , name : "mE5 Abacus 4G Base II" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2913u16 , name : "mE6 Abacus 4TG" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2915u16 , name : "CXP-12 Interface Card 1C" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2916u16 , name : "CXP-12 Interface Card 4C" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2917u16 , name : "CXP-12 Interface Card 2C" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 2918u16 , name : "CXP-12 LightBridge 2C" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 3650u16 , name : "microEnable IV AQ4-GE" , subsystems : & [] } , Device { vendor_id : 6888u16 , id : 3652u16 , name : "microEnable IV VQ4-GE" , subsystems : & [] }] }),
        (40960, Vendor { id : 40960u16 , name : "Asix Electronics Corporation (Wrong ID)" , devices : & [] }),
        (8004, Vendor { id : 8004u16 , name : "VVDN Technologies Private Limited" , devices : & [] }),
        (4714, Vendor { id : 4714u16 , name : "Lexmark International, Inc." , devices : & [] }),
        (4297, Vendor { id : 4297u16 , name : "Dataexpert Corporation" , devices : & [] }),
        (5062, Vendor { id : 5062u16 , name : "Condor Engineering Inc" , devices : & [Device { vendor_id : 5062u16 , id : 1312u16 , name : "CEI-520 A429 Card" , subsystems : & [] } , Device { vendor_id : 5062u16 , id : 1568u16 , name : "CEI-620 A429 Card" , subsystems : & [] } , Device { vendor_id : 5062u16 , id : 2080u16 , name : "CEI-820 A429 Card" , subsystems : & [] } , Device { vendor_id : 5062u16 , id : 2096u16 , name : "CEI-830 A429 Card" , subsystems : & [] } , Device { vendor_id : 5062u16 , id : 4100u16 , name : "P-SER Multi-channel PMC to RS-485/422/232 adapter" , subsystems : & [] }] }),
        (4239, Vendor { id : 4239u16 , name : "Systemsoft" , devices : & [] }),
        (5834, Vendor { id : 5834u16 , name : "CENATEK Inc" , devices : & [Device { vendor_id : 5834u16 , id : 1u16 , name : "Rocket Drive DL" , subsystems : & [] }] }),
        (5600, Vendor { id : 5600u16 , name : "Cacheflow Inc" , devices : & [] }),
        (5331, Vendor { id : 5331u16 , name : "CIRTECH (UK) Ltd" , devices : & [Device { vendor_id : 5331u16 , id : 2u16 , name : "DTL-T14000 Rev. 1 [PS2 TOOL CD/DVD Emulator]" , subsystems : & [] } , Device { vendor_id : 5331u16 , id : 3u16 , name : "DTL-T14000 Rev. 2 [PS2 TOOL CD/DVD Emulator]" , subsystems : & [] } , Device { vendor_id : 5331u16 , id : 4u16 , name : "DTL-T14000 Rev. 3 [PS2 TOOL CD/DVD Emulator]" , subsystems : & [] }] }),
        (5616, Vendor { id : 5616u16 , name : "B-Tree Systems Inc" , devices : & [] }),
        (3277, Vendor { id : 3277u16 , name : "Preferred Networks, Inc." , devices : & [Device { vendor_id : 3277u16 , id : 272u16 , name : "MN-Core" , subsystems : & [] } , Device { vendor_id : 3277u16 , id : 288u16 , name : "MN-Core 2" , subsystems : & [SubSystem { subvendor : 3277u16 , subdevice : 0u16 , name : "MN-Core 2 16GB" } , SubSystem { subvendor : 3277u16 , subdevice : 16u16 , name : "MN-Core 2 32GB" }] } , Device { vendor_id : 3277u16 , id : 512u16 , name : "MN-Core Direct Connect" , subsystems : & [] } , Device { vendor_id : 3277u16 , id : 513u16 , name : "MN-Core 2 Middle-plane" , subsystems : & [] }] }),
        (6704, Vendor { id : 6704u16 , name : "Lantiq" , devices : & [Device { vendor_id : 6704u16 , id : 1664u16 , name : "MtW8171 [Hyperion II]" , subsystems : & [] } , Device { vendor_id : 6704u16 , id : 1792u16 , name : "Wave300 PSB8224 [Hyperion III]" , subsystems : & [] } , Device { vendor_id : 6704u16 , id : 1808u16 , name : "Wave300 PSB8231 [Hyperion III]" , subsystems : & [] }] }),
        (5393, Vendor { id : 5393u16 , name : "CENTILLIUM Technology Corp" , devices : & [] }),
        (7522, Vendor { id : 7522u16 , name : "Nebbiolo Technologies" , devices : & [] }),
        (49374, Vendor { id : 49374u16 , name : "Motorola" , devices : & [] }),
        (6697, Vendor { id : 6697u16 , name : "Fortinet, Inc." , devices : & [Device { vendor_id : 6697u16 , id : 17208u16 , name : "CP8 Content Processor ASIC" , subsystems : & [] } , Device { vendor_id : 6697u16 , id : 17312u16 , name : "CP9 Content Processor ASIC" , subsystems : & [] } , Device { vendor_id : 6697u16 , id : 20022u16 , name : "NP6 Network Processor" , subsystems : & [] } , Device { vendor_id : 6697u16 , id : 20023u16 , name : "NP7 Network Processor" , subsystems : & [] }] }),
        (539, Vendor { id : 539u16 , name : "Compaq Computer Corporation" , devices : & [Device { vendor_id : 539u16 , id : 33081u16 , name : "HNE-300 (RealTek RTL8139c) [iPaq Networking]" , subsystems : & [] }] }),
        (39170, Vendor { id : 39170u16 , name : "Stargen Inc." , devices : & [Device { vendor_id : 39170u16 , id : 1u16 , name : "SG2010 PCI over Starfabric Bridge" , subsystems : & [] } , Device { vendor_id : 39170u16 , id : 2u16 , name : "SG2010 PCI to Starfabric Gateway" , subsystems : & [] } , Device { vendor_id : 39170u16 , id : 3u16 , name : "SG1010 Starfabric Switch and PCI Bridge" , subsystems : & [] }] }),
        (4167, Vendor { id : 4167u16 , name : "Genoa Systems Corp" , devices : & [] }),
        (25430, Vendor { id : 25430u16 , name : "UltraStor" , devices : & [] }),
        (17230, Vendor { id : 17230u16 , name : "Cornelis Networks" , devices : & [Device { vendor_id : 17230u16 , id : 1u16 , name : "CN5000 HFI Silicon, Dual Port, BGA [discrete]" , subsystems : & [SubSystem { subvendor : 17230u16 , subdevice : 1u16 , name : "CN5000 SuperNIC, Single Port, QSFP, x16 PCIe Gen 5" } , SubSystem { subvendor : 17230u16 , subdevice : 2u16 , name : "CN5000 SuperNIC, Dual Port, QSFP-DD, x16 PCIe Gen 5" } , SubSystem { subvendor : 17230u16 , subdevice : 3u16 , name : "CN5000 SuperNIC, Single Port, QSFP, x16 PCIe Gen 5, II" } , SubSystem { subvendor : 17230u16 , subdevice : 4u16 , name : "CN5000 SuperNIC, Dual Port, QSFP-DD, x16 PCIe Gen 5, II" }] } , Device { vendor_id : 17230u16 , id : 2u16 , name : "CN6000 HFI Silicon, Dual Port, BGA [discrete]" , subsystems : & [SubSystem { subvendor : 17230u16 , subdevice : 1u16 , name : "CN6000 SuperNIC, Single Port, QSFP-DD, x16 PCIe Gen 6" }] } , Device { vendor_id : 17230u16 , id : 32769u16 , name : "CN5000 Switch Silicon, 48 Port, BGA" , subsystems : & [SubSystem { subvendor : 17230u16 , subdevice : 257u16 , name : "CN5000 Switch" } , SubSystem { subvendor : 17230u16 , subdevice : 259u16 , name : "CN5000 Director Class Switch Spine" } , SubSystem { subvendor : 17230u16 , subdevice : 260u16 , name : "CN5000 Director Class Switch Leaf" } , SubSystem { subvendor : 17230u16 , subdevice : 262u16 , name : "CN6000 Switch" }] }] }),
        (7718, Vendor { id : 7718u16 , name : "Fujitsu Client Computing Limited" , devices : & [] }),
        (5258, Vendor { id : 5258u16 , name : "OPTO" , devices : & [] }),
        (16748, Vendor { id : 16748u16 , name : "Aladdin Knowledge Systems" , devices : & [Device { vendor_id : 16748u16 , id : 256u16 , name : "AladdinCARD" , subsystems : & [] } , Device { vendor_id : 16748u16 , id : 512u16 , name : "CPC" , subsystems : & [] }] }),
        (5772, Vendor { id : 5772u16 , name : "Qualcomm Atheros" , devices : & [Device { vendor_id : 5772u16 , id : 7u16 , name : "AR5210 Wireless Network Adapter [AR5000 802.11a]" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 14848u16 , name : "DWL-A650 5GHz Wireless CardBus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14850u16 , name : "DWL-A520 5GHz Wireless PCI Adapter" } , SubSystem { subvendor : 5736u16 , subdevice : 1065u16 , name : "802CA Wireless PC Card" } , SubSystem { subvendor : 5943u16 , subdevice : 7u16 , name : "WPC54A Wireless PC Card" } , SubSystem { subvendor : 6983u16 , subdevice : 256u16 , name : "Harmony 8450CN Wireless CardBus Module" } , SubSystem { subvendor : 6983u16 , subdevice : 272u16 , name : "Skyline 4030 / Harmony 8450 802.11a Wireless CardBus Adapter" } , SubSystem { subvendor : 6983u16 , subdevice : 1024u16 , name : "PC50E-8-FC/A Wireless Access Point Radio Card" } , SubSystem { subvendor : 32902u16 , subdevice : 9472u16 , name : "PRO/Wireless 5000 LAN CardBus Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 9473u16 , name : "PRO/Wireless 5000 LAN PCI Adapter Module" }] } , Device { vendor_id : 5772u16 , id : 17u16 , name : "AR5211 Wireless Network Adapter [AR5001A 802.11a]" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 18u16 , name : "AR5211 Wireless Network Adapter [AR5001X 802.11ab]" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 14851u16 , name : "AirPro DWL-A650 Wireless Cardbus Adapter (rev.B)" } , SubSystem { subvendor : 4486u16 , subdevice : 14852u16 , name : "AirPro DWL-AB650 Multimode Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14853u16 , name : "AirPro DWL-AB520 Multimode Wireless PCI Adapter" } , SubSystem { subvendor : 4716u16 , subdevice : 32817u16 , name : "2201 Mobile Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 17408u16 , name : "WAB501 802.11ab Wireless CardBus Card" } , SubSystem { subvendor : 6983u16 , subdevice : 43520u16 , name : "8460 802.11ab Wireless CardBus Adapter" }] } , Device { vendor_id : 5772u16 , id : 19u16 , name : "AR5212/5213/2414 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 776u16 , subdevice : 13314u16 , name : "AG-100 802.11ag Wireless Cardbus Adapter" } , SubSystem { subvendor : 776u16 , subdevice : 13317u16 , name : "G-102 v2 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 776u16 , subdevice : 13320u16 , name : "G-170S 802.11g Wireless CardBus Adapter" } , SubSystem { subvendor : 3601u16 , subdevice : 229u16 , name : "NC6000/NC8000 laptop" } , SubSystem { subvendor : 4279u16 , subdevice : 24578u16 , name : "3CRWE154A72 802.11abg Cardbus Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 54017u16 , name : "Philips CPWNA100 Wireless CardBus adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 60963u16 , name : "SMCWPCIT-G 108Mbps Wireless PCI adapter" } , SubSystem { subvendor : 4436u16 , subdevice : 827u16 , name : "Buffalo WLI-CB-AMG54" } , SubSystem { subvendor : 4436u16 , subdevice : 846u16 , name : "Buffalo WLI-CB-AG108HP 802.11abg Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 12802u16 , name : "DWL-G650 (Rev B3,B5) Wireless cardbus adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 12803u16 , name : "AirPlus DWL-G520 Wireless PCI Adapter (rev. A)" } , SubSystem { subvendor : 4486u16 , subdevice : 14855u16 , name : "AirXpert DWL-AG650 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14856u16 , name : "AirXpert DWL-AG520 Wireless PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14866u16 , name : "D-Link AirPlus DWL-G650 Wireless Cardbus Adapter(rev.C)" } , SubSystem { subvendor : 4486u16 , subdevice : 14867u16 , name : "AirPlus DWL-G520 Wireless PCI Adapter (rev. B)" } , SubSystem { subvendor : 4486u16 , subdevice : 14868u16 , name : "AirPremier AG DWL-AG530 Wireless PCI Adapter (rev.A)" } , SubSystem { subvendor : 4486u16 , subdevice : 14871u16 , name : "D-Link AirPremier DWL-G680 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14872u16 , name : "D-Link AirPremier DWL-G550 Wireless PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14874u16 , name : "WNA-2330 802.11bg Wireless CardBus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14947u16 , name : "D-Link AirPremier DWL-AG660 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14995u16 , name : "Conceptronic C54I Wireless 801.11g PCI card" } , SubSystem { subvendor : 4486u16 , subdevice : 14996u16 , name : "Conceptronic C54C 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 15024u16 , name : "Allnet ALL0281 Wireless PCI Card" } , SubSystem { subvendor : 4997u16 , subdevice : 17920u16 , name : "WAG511 802.11a/b/g Dual Band Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 17936u16 , name : "WAG511 802.11a/b/g Dual Band Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 18688u16 , name : "WG311v1 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 18944u16 , name : "WAG311 802.11a/g Wireless PCI Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 19200u16 , name : "WG511T 108 Mbps Wireless PC Card (rev.A/B)" } , SubSystem { subvendor : 4997u16 , subdevice : 19712u16 , name : "WG311T 108 Mbps Wireless PCI Adapter (rev.A2)" } , SubSystem { subvendor : 4997u16 , subdevice : 20224u16 , name : "WG511U Double 108 Mbps Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 23040u16 , name : "WG311T (rev.A3 v1h3/v1h4) 108 Mbps Wireless PCI Adapter [AR2412]" } , SubSystem { subvendor : 4997u16 , subdevice : 23296u16 , name : "WG511T 108 Mbps Wireless PC Card (rev.C)" } , SubSystem { subvendor : 4997u16 , subdevice : 23808u16 , name : "WPN511 RangeMax Wireless PC Card" } , SubSystem { subvendor : 5208u16 , subdevice : 59665u16 , name : "GN-WIAG02" } , SubSystem { subvendor : 5224u16 , subdevice : 1027u16 , name : "U10H014 802.11g Cardbus Adapter" } , SubSystem { subvendor : 5224u16 , subdevice : 1032u16 , name : "ThinkPad 11b/g Wireless LAN Mini PCI Adapter" } , SubSystem { subvendor : 5303u16 , subdevice : 2576u16 , name : "8480-WD 802.11abg Cardbus Adapter" } , SubSystem { subvendor : 5303u16 , subdevice : 2656u16 , name : "8482-WD ORiNOCO 11a/b/g Wireless PCI Adapter" } , SubSystem { subvendor : 5303u16 , subdevice : 43568u16 , name : "8800-FC 802.11bg Cardbus Adapter" } , SubSystem { subvendor : 5303u16 , subdevice : 43584u16 , name : "8470-WD 802.11bg Cardbus Adapter" } , SubSystem { subvendor : 5305u16 , subdevice : 52001u16 , name : "CB21 802.11a/b/g Cardbus Adapter" } , SubSystem { subvendor : 5736u16 , subdevice : 4134u16 , name : "IBM HighRate 11 a/b/g Wireless CardBus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 19u16 , name : "AirPlus XtremeG DWL-G650 Wireless PCMCIA Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 4133u16 , name : "DWL-G650B2 Wireless CardBus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 4135u16 , name : "Engenius NL-3054CB ARIES b/g CardBus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 4162u16 , name : "Ubiquiti Networks SuperRange a/b/g Cardbus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 4177u16 , name : "EZ Connect g 802.11g 108Mbps Wireless PCI Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8230u16 , name : "Netgate 5354MP ARIES a(108Mb turbo)/b/g MiniPCI Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8231u16 , name : "D-Link AirPlus DWL-G520 Wireless PCI Adapter (rev. A)" } , SubSystem { subvendor : 5772u16 , subdevice : 8257u16 , name : "Engenius 5354MP Plus ARIES2 b/g MiniPCI Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8258u16 , name : "Engenius 5354MP Plus ARIES2 a/b/g MiniPCI Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8273u16 , name : "TRENDnet TEW-443PI Wireless PCI Adapter" } , SubSystem { subvendor : 5797u16 , subdevice : 5642u16 , name : "BWP712 802.11bg Wireless CardBus Adapter" } , SubSystem { subvendor : 5803u16 , subdevice : 29442u16 , name : "Trust Speedshare Turbo Pro Wireless PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 23u16 , name : "WPC55AG" } , SubSystem { subvendor : 5943u16 , subdevice : 38u16 , name : "WMP55AG v1.1" } , SubSystem { subvendor : 5943u16 , subdevice : 53u16 , name : "WPC55AG v1.2 802.11abg Cardbus Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 54u16 , name : "WMP55AG v1.2 802.11abg PCI Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 12288u16 , name : "F6D3000 Dual-Band Wireless A+G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 12304u16 , name : "F6D3010 Dual-Band Wireless A+G Notebook Card" } , SubSystem { subvendor : 6095u16 , subdevice : 66u16 , name : "Z-COMAX Highpower XG-622H (400mw) 802.11b/g mini-PCI Adapter" } , SubSystem { subvendor : 6239u16 , subdevice : 4114u16 , name : "CM9 Wireless a/b/g MiniPCI Adapter" } , SubSystem { subvendor : 6239u16 , subdevice : 8210u16 , name : "Wistron NeWeb WLAN a+b+g model CB9" } , SubSystem { subvendor : 42791u16 , subdevice : 26625u16 , name : "3CRXJK10075 OfficeConnect Wireless 108Mbps 11g XJACK PC Card" }] } , Device { vendor_id : 5772u16 , id : 26u16 , name : "AR2413/AR2414 Wireless Network Adapter [AR5005G(S) 802.11bg]" , subsystems : & [SubSystem { subvendor : 4178u16 , subdevice : 5772u16 , name : "Sweex Wireless Lan PC Card 54Mbps" } , SubSystem { subvendor : 4371u16 , subdevice : 60960u16 , name : "SMC Wireless CardBus Adapter 802.11g (SMCWCB-G EU)" } , SubSystem { subvendor : 4371u16 , subdevice : 60964u16 , name : "SMC Wireless PCI Card WPCI-G" } , SubSystem { subvendor : 4486u16 , subdevice : 14869u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.D1)" } , SubSystem { subvendor : 4486u16 , subdevice : 14870u16 , name : "AirPlus G DWL-G510 Wireless PCI Adapter(rev.B)" } , SubSystem { subvendor : 4486u16 , subdevice : 14876u16 , name : "WNA-1330 Notebook Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14877u16 , name : "WDA-1320 Desktop Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14883u16 , name : "AirPlus G DWL-G520+A Wireless PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14884u16 , name : "AirPlus G DWL-G650+A Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 15112u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.C1)" } , SubSystem { subvendor : 5772u16 , subdevice : 26u16 , name : "Belkin FD7000" } , SubSystem { subvendor : 5772u16 , subdevice : 4178u16 , name : "TP-Link TL-WN510G Wireless CardBus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8274u16 , name : "Compex Wireless 802.11 b/g  MiniPCI Adapter, Rev A1 [WLM54G]" } , SubSystem { subvendor : 5868u16 , subdevice : 290u16 , name : "Wireless PCI Adapter Model 5418" } , SubSystem { subvendor : 5943u16 , subdevice : 83u16 , name : "WPC54G v7 802.11g Wireless-G Notebook Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 28684u16 , name : "F5D7000 v5000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28701u16 , name : "F5D7010 v5000 Wireless G Notebook Card" } , SubSystem { subvendor : 6137u16 , subdevice : 8u16 , name : "DX-WGNBC 802.11bg Wireless CardBus Adapter" } , SubSystem { subvendor : 6137u16 , subdevice : 24u16 , name : "DX-WGDTC 802.11bg Wireless PCI Adapter" }] } , Device { vendor_id : 5772u16 , id : 27u16 , name : "AR5413/AR5414 Wireless Network Adapter [AR5006X(S) 802.11abg]" , subsystems : & [SubSystem { subvendor : 1911u16 , subdevice : 4359u16 , name : "UB5 802.11a Wireless Mini PCI Adapter" } , SubSystem { subvendor : 1911u16 , subdevice : 12290u16 , name : "XR2 802.11g Wireless Mini PCI Adapter" } , SubSystem { subvendor : 1911u16 , subdevice : 12293u16 , name : "XR5 802.11a Wireless Mini PCI Adapter" } , SubSystem { subvendor : 1911u16 , subdevice : 12297u16 , name : "XR9 900MHz Wireless Mini PCI Adapter" } , SubSystem { subvendor : 4436u16 , subdevice : 846u16 , name : "WLI-CB-AG108HP 802.11abg Wireless CardBus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14873u16 , name : "D-Link AirPremier AG DWL-AG660 Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14882u16 , name : "AirPremier AG DWL-AG530 Wireless PCI Adapter (rev.B)" } , SubSystem { subvendor : 4525u16 , subdevice : 20481u16 , name : "WN5301A 802.11bg Wireless PCI Adapter" } , SubSystem { subvendor : 5208u16 , subdevice : 59649u16 , name : "GN-WI01HT Wireless a/b/g MiniPCI Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 27u16 , name : "Wireless LAN PCI LiteOn" } , SubSystem { subvendor : 5772u16 , subdevice : 4194u16 , name : "IPN-W100CB 802.11abg Wireless CardBus Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8290u16 , name : "EnGenius EMP-8602 (400mw) or Compex WLM54AG (SuperAG)" } , SubSystem { subvendor : 5772u16 , subdevice : 8291u16 , name : "EnGenius EMP-8602 (400mw) or Compex WLM54AG" } , SubSystem { subvendor : 6137u16 , subdevice : 11u16 , name : "WL-711A 802.11abg Wireless CardBus Adapter" } , SubSystem { subvendor : 6137u16 , subdevice : 12u16 , name : "WPIA-112AG 802.11abg Wireless PCI Adapter" } , SubSystem { subvendor : 6137u16 , subdevice : 13u16 , name : "PC-686X 802.11abg Wireless Mini PCI Adapter" } , SubSystem { subvendor : 6239u16 , subdevice : 5632u16 , name : "DCMA-82 High Power WLAN 802.11a/b/g mini-PCI Module (Super A/G, eXtended Range, 400mW)" } , SubSystem { subvendor : 6472u16 , subdevice : 15034u16 , name : "RBTBJ-AW 802.11abg Wireless Cardbus Adapter" } , SubSystem { subvendor : 42791u16 , subdevice : 26628u16 , name : "Wireless 11a/b/g PC Card with XJACK(r) Antenna" }] } , Device { vendor_id : 5772u16 , id : 28u16 , name : "AR242x / AR542x Wireless Network Adapter (PCI-Express)" , subsystems : & [SubSystem { subvendor : 1911u16 , subdevice : 12294u16 , name : "SRX 802.11abg Wireless ExpressCard Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4986u16 , name : "AR5BXB63 (Foxconn) 802.11bg Mini PCIe NIC" } , SubSystem { subvendor : 4203u16 , subdevice : 134u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 5199u16 , subdevice : 28934u16 , name : "WLL3140 (Toshiba PA3501U-1MPC) 802.11bg Wireless Mini PCIe Card" } , SubSystem { subvendor : 5199u16 , subdevice : 28968u16 , name : "WLL3141 (Toshiba PA3613U-1MPC) 802.11bg Wireless Mini PCIe Card" } , SubSystem { subvendor : 5224u16 , subdevice : 1064u16 , name : "AR5BXB63 802.11bg NIC" } , SubSystem { subvendor : 5224u16 , subdevice : 1066u16 , name : "AR5007EG 802.11bg NIC" } , SubSystem { subvendor : 5243u16 , subdevice : 4147u16 , name : "AirPace Wi-Fi" } , SubSystem { subvendor : 5772u16 , subdevice : 28u16 , name : "AR242x 802.11abg NIC (PCI Express)" } , SubSystem { subvendor : 5772u16 , subdevice : 12385u16 , name : "AR5006EGS 802.11bg NIC (2.4GHz, PCI Express)" } , SubSystem { subvendor : 5772u16 , subdevice : 12386u16 , name : "AR5006EXS 802.11abg NIC (2.4/5.0GHz, PCI Express)" } , SubSystem { subvendor : 5772u16 , subdevice : 12387u16 , name : "AR5006EX 802.11abg NIC (2.4/5.0GHz, PCI Express)" } , SubSystem { subvendor : 5772u16 , subdevice : 12389u16 , name : "AR5006EG 802.11bg NIC (2.4GHz, PCI Express)" } , SubSystem { subvendor : 5772u16 , subdevice : 12391u16 , name : "AR242x 802.11abg Wireless PCI Express Adapter (rev 01)" } , SubSystem { subvendor : 6715u16 , subdevice : 4134u16 , name : "AW-GE780 802.11bg Wireless Mini PCIe Card" }] } , Device { vendor_id : 5772u16 , id : 29u16 , name : "AR2417 Wireless Network Adapter [AR5007G 802.11bg]" , subsystems : & [SubSystem { subvendor : 6041u16 , subdevice : 29195u16 , name : "F5D7000 v8000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 29211u16 , name : "F5D7010 v8000 Wireless G Notebook Card" }] } , Device { vendor_id : 5772u16 , id : 32u16 , name : "AR5513 802.11abg Wireless NIC" , subsystems : & [SubSystem { subvendor : 776u16 , subdevice : 13319u16 , name : "M-102 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14951u16 , name : "DWL-G650M Super G MIMO Wireless Notebook Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14952u16 , name : "DWL-G520M Wireless 108G MIMO Desktop Adapter" } , SubSystem { subvendor : 6270u16 , subdevice : 13326u16 , name : "M-302 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 6518u16 , subdevice : 5632u16 , name : "TEW-603PI 108Mbps 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 6518u16 , subdevice : 8195u16 , name : "TEW-601PC 802.11g Wireless CardBus Adapter" }] } , Device { vendor_id : 5772u16 , id : 35u16 , name : "AR5416 Wireless Network Adapter [AR5008 802.11(a)bgn]" , subsystems : & [SubSystem { subvendor : 776u16 , subdevice : 13323u16 , name : "NWD-170N 802.11bgn Wireless CardBus Adapter" } , SubSystem { subvendor : 4436u16 , subdevice : 869u16 , name : "Buffalo WLP-CB-AG300 802.11abgn Cardbus Adapter" } , SubSystem { subvendor : 4436u16 , subdevice : 871u16 , name : "WLI-CB-AG301N 802.11abgn Wireless CardBus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14954u16 , name : "DWA-642 802.11n RangeBooster N CardBus Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14955u16 , name : "DWA-547 802.11n RangeBooster N 650 DeskTop Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14957u16 , name : "DWA-552 802.11n Xtreme N Desktop Adapter (rev A1)" } , SubSystem { subvendor : 4486u16 , subdevice : 14966u16 , name : "DWA-645 802.11n RangeBooster N 650 Notebook Adapter (rev A1)" } , SubSystem { subvendor : 5943u16 , subdevice : 89u16 , name : "WPC300N v2 Wireless-N Notebook Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 105u16 , name : "WPC100 v1 802.11n RangePlus Wireless Notebook Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 114u16 , name : "WMP110 v1 802.11n RangePlus Wireless PCI Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 32785u16 , name : "F5D8011 v1 802.11n N1 Wireless Notebook Card" } , SubSystem { subvendor : 6270u16 , subdevice : 13329u16 , name : "NWD-370N 802.11n Wireless PCI Adapter" } , SubSystem { subvendor : 6518u16 , subdevice : 8200u16 , name : "TEW-621PC 802.11bgn Wireless CardBus Adapter" }] } , Device { vendor_id : 5772u16 , id : 36u16 , name : "AR5418 Wireless Network Adapter [AR5008E 802.11(a)bgn] (PCI-Express)" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 135u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4436u16 , subdevice : 870u16 , name : "WLP-EXC-AG300 802.11abgn ExpressCard" } , SubSystem { subvendor : 4486u16 , subdevice : 14959u16 , name : "DWA-643 Xtreme N Notebook ExpressCard" } , SubSystem { subvendor : 4486u16 , subdevice : 14960u16 , name : "DWA-556 Xtreme N PCI Express Desktop Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 32881u16 , name : "F5D8071 v1 N1 Wireless ExpressCard" }] } , Device { vendor_id : 5772u16 , id : 39u16 , name : "AR9160 Wireless Network Adapter [AR9001 802.11(a)bgn]" , subsystems : & [SubSystem { subvendor : 1911u16 , subdevice : 16514u16 , name : "SR71-A 802.11abgn Wireless Mini PCI Adapter" }] } , Device { vendor_id : 5772u16 , id : 41u16 , name : "AR922X Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 1911u16 , subdevice : 16389u16 , name : "SR71-15 802.11an Mini PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 14970u16 , name : "DWA-552 802.11n Xtreme N Desktop Adapter (rev A2)" } , SubSystem { subvendor : 4486u16 , subdevice : 14973u16 , name : "DWA-552 802.11n Xtreme N Desktop Adapter (rev A3)" } , SubSystem { subvendor : 5772u16 , subdevice : 41u16 , name : "AR922X Wireless Network Adapter" } , SubSystem { subvendor : 5772u16 , subdevice : 8342u16 , name : "Compex WLM200NX / Wistron DNMA-92" }] } , Device { vendor_id : 5772u16 , id : 42u16 , name : "AR928X Wireless Network Adapter (PCI-Express)" , subsystems : & [SubSystem { subvendor : 1911u16 , subdevice : 20229u16 , name : "SR71-X 802.11abgn Wireless ExpressCard Adapter [AR9280]" } , SubSystem { subvendor : 4156u16 , subdevice : 12353u16 , name : "AR5BHB92-H 802.11abgn Wireless Half-size Mini PCIe Card [AR9280]" } , SubSystem { subvendor : 4156u16 , subdevice : 12354u16 , name : "AzureWave AW-NE773 802.11abgn Wireless Half-size Mini PCIe Card [AR9280]" } , SubSystem { subvendor : 4187u16 , subdevice : 57350u16 , name : "T77H053.00 802.11bgn Wireless Mini PCIe Card [AR9281]" } , SubSystem { subvendor : 4187u16 , subdevice : 57375u16 , name : "T77H047.31 802.11bgn Wireless Half-size Mini PCIe Card [AR9283]" } , SubSystem { subvendor : 4203u16 , subdevice : 143u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4525u16 , subdevice : 26112u16 , name : "WN6600A 802.11bgn Wireless Mini PCIe Card [AR9281]" } , SubSystem { subvendor : 5199u16 , subdevice : 28993u16 , name : "WLL6080 802.11bgn Wireless Mini PCIe Card [AR9281]" } , SubSystem { subvendor : 5772u16 , subdevice : 515u16 , name : "DW1525 802.11abgn WLAN PCIe Card [AR9280]" } , SubSystem { subvendor : 6706u16 , subdevice : 771u16 , name : "EM303 802.11bgn Wireless Mini PCIe Card [AR9281]" } , SubSystem { subvendor : 6706u16 , subdevice : 774u16 , name : "EM306 802.11bgn Wireless Half-size Mini PCIe Card [AR9283]" } , SubSystem { subvendor : 6715u16 , subdevice : 4199u16 , name : "AW-NE771 802.11bgn Wireless Mini PCIe Card [AR9281]" } , SubSystem { subvendor : 6715u16 , subdevice : 4209u16 , name : "AW-NE772 802.11abgn Wireless Mini PCIe Card [AR9280]" } , SubSystem { subvendor : 6715u16 , subdevice : 4225u16 , name : "AW-NE773 802.11abgn Wireless Half-size Mini PCIe Card [AR9280]" }] } , Device { vendor_id : 5772u16 , id : 43u16 , name : "AR9285 Wireless Network Adapter (PCI-Express)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 516u16 , name : "Wireless 1502 802.11bgn Half-size Mini PCIe Card" } , SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "Wireless 1702 802.11bgn Half-size Mini PCIe Card [AR9002WB-1NGCD]" } , SubSystem { subvendor : 4156u16 , subdevice : 12351u16 , name : "U98Z062.10 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 4156u16 , subdevice : 12352u16 , name : "U98Z062.12 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 4187u16 , subdevice : 57367u16 , name : "T77H126.00 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 4187u16 , subdevice : 57379u16 , name : "T77H121.04 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 4187u16 , subdevice : 57381u16 , name : "T77H121.05 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 4371u16 , subdevice : 59409u16 , name : "WN7811A (Toshiba PA3722U-1MPC) 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 6239u16 , subdevice : 12463u16 , name : "DNXA-95 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 6449u16 , subdevice : 35u16 , name : "Option GTM67x PCIe WiFi Adapter" } , SubSystem { subvendor : 6715u16 , subdevice : 4233u16 , name : "AW-NE785 / AW-NE785H 802.11bgn Wireless Full or Half-size Mini PCIe Card" } , SubSystem { subvendor : 6715u16 , subdevice : 11319u16 , name : "AW-NB037H 802.11bgn Wireless Half-size Mini PCIe Card [AR9002WB-1NGCD]" } , SubSystem { subvendor : 7066u16 , subdevice : 1025u16 , name : "XW204E 802.11bgn Wireless Half-size Mini PCIe Card" } , SubSystem { subvendor : 7066u16 , subdevice : 3075u16 , name : "WB214E 802.11bgn Wireless Half-size Mini PCIe Card [AR9002WB-1NGCD]" }] } , Device { vendor_id : 5772u16 , id : 44u16 , name : "AR2427 802.11bg Wireless Network Adapter (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 45u16 , name : "AR9227 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 46u16 , name : "AR9287 Wireless Network Adapter (PCI-Express)" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 57396u16 , name : "T77H167.00" }] } , Device { vendor_id : 5772u16 , id : 48u16 , name : "AR93xx Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5671u16 , name : "AR9380/HB112 802.11abgn 3x3 Wi-Fi Adapter" } , SubSystem { subvendor : 4203u16 , subdevice : 154u16 , name : "AirPort Extreme" } , SubSystem { subvendor : 4486u16 , subdevice : 14974u16 , name : "DWA-566 Wireless N 300 Dual Band PCIe Desktop Adapter" } , SubSystem { subvendor : 6742u16 , subdevice : 8192u16 , name : "Killer Wireless-N 1102 Half-size Mini PCIe Card [AR9382]" } , SubSystem { subvendor : 6742u16 , subdevice : 8193u16 , name : "Killer Wireless-N 1103 Half-size Mini PCIe Card [AR9380]" }] } , Device { vendor_id : 5772u16 , id : 50u16 , name : "AR9485 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 520u16 , name : "Wireless 1506 WLAN Half Mini-Card" } , SubSystem { subvendor : 4156u16 , subdevice : 6200u16 , name : "AR9485/HB125 802.11bgn 1x1 Wi-Fi Adapter" } , SubSystem { subvendor : 4187u16 , subdevice : 57412u16 , name : "Unex DHXA-225" } , SubSystem { subvendor : 5197u16 , subdevice : 16654u16 , name : "AR9485WB-EG 802.11b/g/n mini-PCIe card on a series 3 laptop" } , SubSystem { subvendor : 6715u16 , subdevice : 4486u16 , name : "AW-NE186H" }] } , Device { vendor_id : 5772u16 , id : 51u16 , name : "AR958x 802.11abgn Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 5772u16 , subdevice : 41248u16 , name : "AR9582 802.11a/n WLAN Mini-PCIe Adapter" }] } , Device { vendor_id : 5772u16 , id : 52u16 , name : "AR9462 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "Wireless 1601 802.11abgn Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 768u16 , name : "Wireless 1802 802.11abgn Adapter" } , SubSystem { subvendor : 6742u16 , subdevice : 8195u16 , name : "Killer Wireless-N 1202 Half-size Mini PCIe Card" }] } , Device { vendor_id : 5772u16 , id : 54u16 , name : "QCA9565 / AR9565 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 526u16 , name : "Vostro 3470" }] } , Device { vendor_id : 5772u16 , id : 55u16 , name : "AR9485 Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 6715u16 , subdevice : 8448u16 , name : "AW-NB100H 802.11n Wireless Mini PCIe Card" }] } , Device { vendor_id : 5772u16 , id : 60u16 , name : "QCA986x/988x 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 62u16 , name : "QCA6174 802.11ac Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 5178u16 , name : "Killer 1435 Wireless-AC" } , SubSystem { subvendor : 6742u16 , subdevice : 5413u16 , name : "Killer N1525 Wireless-AC" } , SubSystem { subvendor : 6742u16 , subdevice : 5429u16 , name : "Killer Wireless-n/a/ac 1535 Wireless Network Adapter" }] } , Device { vendor_id : 5772u16 , id : 64u16 , name : "QCA9980/9990 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 65u16 , name : "QCA6164 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 66u16 , name : "QCA9377 802.11ac Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 6160u16 , name : "Dell Wireless DW1810" } , SubSystem { subvendor : 4525u16 , subdevice : 2214u16 , name : "Qualcomm Atheros QCA9377 802.11ac Wireless Network Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 2305u16 , name : "Qualcomm Atheros QCA9377 Wireless Network Adapter" }] } , Device { vendor_id : 5772u16 , id : 70u16 , name : "QCA9984 802.11ac Wave 2 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 80u16 , name : "QCA9887 802.11ac Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 519u16 , name : "AR5210 Wireless Network Adapter [AR5000 802.11a]" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 4116u16 , name : "AR5212 802.11abg NIC" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1418u16 , name : "ThinkPad 11a/b/g Wireless LAN Mini Express Adapter (AR5BXB6)" }] } , Device { vendor_id : 5772u16 , id : 36883u16 , name : "AR5002X Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 65305u16 , name : "AR5006X Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 65307u16 , name : "AR2425 Wireless Network Adapter [AR5007EG 802.11bg]" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 65308u16 , name : "AR5008 Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 5772u16 , id : 65309u16 , name : "AR922x Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 5772u16 , subdevice : 60956u16 , name : "AR9220-AC1A [AVM Fritz!Box FON WLAN 7270 v3]" }] }] }),
        (4504, Vendor { id : 4504u16 , name : "Lambda Systems Inc" , devices : & [] }),
        (7124, Vendor { id : 7124u16 , name : "IEIT SYSTEMS Co., Ltd" , devices : & [Device { vendor_id : 7124u16 , id : 2321u16 , name : "Arria10_PCIe_F10A1150" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4096u16 , name : "NS8600G1U160 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4097u16 , name : "NS8600G1U320 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4098u16 , name : "NS8600G1U640 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4099u16 , name : "NS8500G1U192 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4100u16 , name : "NS8500G1U384 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4101u16 , name : "NS8500G1U768 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4102u16 , name : "NS6610G1U160, NS6510G1U192 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4103u16 , name : "NS6610G1U320, NS6510G1U384 NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4108u16 , name : "NS8510G1Uxxx, NS8610G1Uxxx NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4110u16 , name : "NS8500G2Uxxxx, NS8600G2Uxxxx NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 4111u16 , name : "NS6500G2UXXXX,NS6600G2UXXXX NVME SSD" , subsystems : & [] } , Device { vendor_id : 7124u16 , id : 10625u16 , name : "32G SFP28 EP2744 Apatter" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 176u16 , name : "32G SFP28 EP2744 Apatter" }] }] }),
        (6255, Vendor { id : 6255u16 , name : "WiNRADiO Communications" , devices : & [] }),
        (4610, Vendor { id : 4610u16 , name : "Network General Corp." , devices : & [Device { vendor_id : 4610u16 , id : 17152u16 , name : "Gigabit Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4610u16 , subdevice : 38977u16 , name : "SK-9841 LX" } , SubSystem { subvendor : 4610u16 , subdevice : 38978u16 , name : "SK-9841 LX dual link" } , SubSystem { subvendor : 4610u16 , subdevice : 38979u16 , name : "SK-9843 SX" } , SubSystem { subvendor : 4610u16 , subdevice : 38980u16 , name : "SK-9843 SX dual link" }] }] }),
        (4805, Vendor { id : 4805u16 , name : "Picture Elements Incorporated" , devices : & [Device { vendor_id : 4805u16 , id : 126u16 , name : "Imaging/Scanning Subsystem Engine" , subsystems : & [] } , Device { vendor_id : 4805u16 , id : 127u16 , name : "Imaging/Scanning Subsystem Engine" , subsystems : & [] } , Device { vendor_id : 4805u16 , id : 129u16 , name : "PCIVST [Grayscale Thresholding Engine]" , subsystems : & [] } , Device { vendor_id : 4805u16 , id : 133u16 , name : "Video Simulator/Sender" , subsystems : & [] } , Device { vendor_id : 4805u16 , id : 134u16 , name : "THR2 Multi-scale Thresholder" , subsystems : & [] }] }),
        (5925, Vendor { id : 5925u16 , name : "Vitesse Semiconductor" , devices : & [Device { vendor_id : 5925u16 , id : 29044u16 , name : "VSC7174 PCI/PCI-X Serial ATA Host Bus Controller" , subsystems : & [] }] }),
        (4863, Vendor { id : 4863u16 , name : "Lexicon" , devices : & [] }),
        (5538, Vendor { id : 5538u16 , name : "Catalyst Enterprises Inc" , devices : & [Device { vendor_id : 5538u16 , id : 1u16 , name : "TA700 PCI Bus Analyzer/Exerciser" , subsystems : & [] }] }),
        (4153, Vendor { id : 4153u16 , name : "Silicon Integrated Systems [SiS]" , devices : & [Device { vendor_id : 4153u16 , id : 1u16 , name : "AGP Port (virtual PCI-to-PCI bridge)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2u16 , name : "AGP Port (virtual PCI-to-PCI bridge)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 3u16 , name : "AGP Port (virtual PCI-to-PCI bridge)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 4u16 , name : "PCI-to-PCI bridge" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 0u16 , name : "PCIe x16 port" }] } , Device { vendor_id : 4153u16 , id : 6u16 , name : "85C501/2/3" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 8u16 , name : "SiS85C503/5513 (LPC Bridge)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 9u16 , name : "5595 Power Management Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 10u16 , name : "PCI-to-PCI bridge" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 0u16 , name : "PCIe x1 port" }] } , Device { vendor_id : 4153u16 , id : 22u16 , name : "SiS961/2/3 SMBus controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 24u16 , name : "SiS85C503/5513 (LPC Bridge)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 355u16 , name : "163 802.11b/g Wireless LAN Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 384u16 , name : "RAID bus controller 180 SATA/PATA  [SiS]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 385u16 , name : "SATA" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 386u16 , name : "182 SATA/RAID Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4245u16 , name : "D2030-A1" }] } , Device { vendor_id : 4153u16 , id : 390u16 , name : "AHCI Controller (0106)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 400u16 , name : "190 Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 401u16 , name : "191 Gigabit Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33081u16 , name : "P5SD2-FM/S mainboard" }] } , Device { vendor_id : 4153u16 , id : 512u16 , name : "5597/5598/6326 VGA" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 0u16 , name : "SiS5597 SVGA (Shared RAM)" }] } , Device { vendor_id : 4153u16 , id : 516u16 , name : "82C204" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 517u16 , name : "SG86C205" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 768u16 , name : "300/305 PCI/AGP VGA Display Adapter" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 10016u16 , name : "Leadtek WinFast VR300" }] } , Device { vendor_id : 4153u16 , id : 784u16 , name : "315H PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 789u16 , name : "315 PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 805u16 , name : "315PRO PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 816u16 , name : "330 [Xabre] PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1030u16 , name : "85C501/2" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1174u16 , name : "SiS85C496 PCI & CPU Memory Controller (PCM)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1328u16 , name : "530 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1344u16 , name : "540 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1360u16 , name : "550 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1431u16 , name : "5513C" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1537u16 , name : "85C601" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1568u16 , name : "620 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1584u16 , name : "630 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1587u16 , name : "633 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1589u16 , name : "635 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1605u16 , name : "SiS645 Host & Memory & AGP Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1606u16 , name : "SiS645DX Host & Memory & AGP Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1608u16 , name : "645xx" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1609u16 , name : "SiS649 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1616u16 , name : "650/M650 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1617u16 , name : "651 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1621u16 , name : "655 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1632u16 , name : "660 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1633u16 , name : "661FX/M661FX/M661MX Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1634u16 , name : "662 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1649u16 , name : "671MX" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1840u16 , name : "730 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1843u16 , name : "733 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1845u16 , name : "735 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1856u16 , name : "740 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1857u16 , name : "741/741GX/M741 Host" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 1857u16 , name : "K7S41/K7S41GX motherboard" }] } , Device { vendor_id : 4153u16 , id : 1861u16 , name : "745 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1862u16 , name : "746 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1877u16 , name : "755 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1888u16 , name : "760/M760 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 1889u16 , name : "761/M761 Host" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4249u16 , name : "D2030-A1 Motherboard" }] } , Device { vendor_id : 4153u16 , id : 2304u16 , name : "SiS900 PCI Fast Ethernet" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2580u16 , name : "K7S5A motherboard" } , SubSystem { subvendor : 4153u16 , subdevice : 2304u16 , name : "SiS900 10/100 Ethernet Adapter onboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32821u16 , name : "CUSI-FX motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32935u16 , name : "Motherboard P4S800D-X" } , SubSystem { subvendor : 5218u16 , subdevice : 2304u16 , name : "MS-6701 motherboard" }] } , Device { vendor_id : 4153u16 , id : 2401u16 , name : "SiS961 [MuTIOL Media IO]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2402u16 , name : "SiS962 [MuTIOL Media IO] LPC Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2403u16 , name : "SiS963 [MuTIOL Media IO] LPC Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2404u16 , name : "SiS964 [MuTIOL Media IO] LPC Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2405u16 , name : "SiS965 [MuTIOL Media IO]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2406u16 , name : "SiS966 [MuTIOL Media IO]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 2408u16 , name : "SiS968 [MuTIOL Media IO]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 4480u16 , name : "SATA Controller / IDE mode" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 4482u16 , name : "SATA Controller / RAID mode" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 384u16 , name : "SiS 966 4-port SATA controller" }] } , Device { vendor_id : 4153u16 , id : 4483u16 , name : "SATA Controller / IDE mode" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 384u16 , name : "SiS 966 4-port SATA controller" }] } , Device { vendor_id : 4153u16 , id : 4484u16 , name : "AHCI Controller / RAID mode" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 4485u16 , name : "AHCI IDE Controller (0106)" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 13826u16 , name : "83C602" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 20743u16 , name : "5107" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21248u16 , name : "SiS540 PCI Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21269u16 , name : "550 PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21505u16 , name : "486 PCI Chipset" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21777u16 , name : "5511/5512" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21779u16 , name : "5513 IDE Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2416u16 , name : "P6STP-FL motherboard" } , SubSystem { subvendor : 4153u16 , subdevice : 21779u16 , name : "SiS5513 EIDE Controller (A,B step)" } , SubSystem { subvendor : 4163u16 , subdevice : 32821u16 , name : "CUSI-FX motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28688u16 , name : "MS-6701 motherboard" } , SubSystem { subvendor : 5681u16 , subdevice : 21779u16 , name : "GA-8SIML Rev1.0 Motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4245u16 , name : "D2030-A1 Motherboard" }] } , Device { vendor_id : 4153u16 , id : 21783u16 , name : "5517" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21873u16 , name : "5571" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21889u16 , name : "5581 Pentium Chipset" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21890u16 , name : "5582" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21905u16 , name : "5591/5592 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21910u16 , name : "5596 Pentium Chipset" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 21911u16 , name : "5597 [SiS5582]" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 22016u16 , name : "5600 Host" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25092u16 , name : "Video decoder & MPEG interface" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25093u16 , name : "VGA Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25142u16 , name : "6236 3D-AGP" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25344u16 , name : "630/730 PCI/AGP VGA Display Adapter" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2416u16 , name : "P6STP-FL motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 32821u16 , name : "CUSI-FX motherboard" } , SubSystem { subvendor : 4173u16 , subdevice : 32994u16 , name : "VAIO PCV-J200" }] } , Device { vendor_id : 4153u16 , id : 25350u16 , name : "530/620 PCI/AGP VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25381u16 , name : "65x/M650/740 PCI/AGP VGA Display Adapter" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 25381u16 , name : "SiS 651 onboard [Asus P4SC-EA]" } , SubSystem { subvendor : 5681u16 , subdevice : 4100u16 , name : "SiS 651C onboard [Gigabyte GA-8SIML Rev1.0]" }] } , Device { vendor_id : 4153u16 , id : 25382u16 , name : "86C326 5598/6326" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 25382u16 , name : "SiS6326 GUI Accelerator" } , SubSystem { subvendor : 4242u16 , subdevice : 2640u16 , name : "SpeedStar A50" } , SubSystem { subvendor : 4242u16 , subdevice : 2672u16 , name : "SpeedStar A70" } , SubSystem { subvendor : 4242u16 , subdevice : 18704u16 , name : "SpeedStar A70" } , SubSystem { subvendor : 4242u16 , subdevice : 18720u16 , name : "SpeedStar A70" } , SubSystem { subvendor : 4272u16 , subdevice : 25382u16 , name : "S6110-B (AGP)" } , SubSystem { subvendor : 5481u16 , subdevice : 25382u16 , name : "SiS6326 GUI Accelerator" }] } , Device { vendor_id : 4153u16 , id : 25392u16 , name : "661/741/760 PCI/AGP or 662/761Gx PCIE VGA Display Adapter" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 25392u16 , name : "[M]661xX/[M]741[GX]/[M]760 PCI/AGP VGA Adapter" } , SubSystem { subvendor : 4163u16 , subdevice : 33043u16 , name : "SiS Real 256E (ASUS P5S800-VM motherboard)" } , SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "SiS661FX GUI 2D/3D Accelerator" } , SubSystem { subvendor : 5940u16 , subdevice : 4249u16 , name : "D2030-A1" }] } , Device { vendor_id : 4153u16 , id : 25424u16 , name : "770/670 PCIE VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 25425u16 , name : "771/671 PCIE VGA Display Adapter" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 28673u16 , name : "USB 1.1 Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 2580u16 , name : "K7S5A motherboard" } , SubSystem { subvendor : 4153u16 , subdevice : 28672u16 , name : "Onboard USB Controller" } , SubSystem { subvendor : 5218u16 , subdevice : 21616u16 , name : "ECS K7SOM+ motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28688u16 , name : "MS-6701 motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4245u16 , name : "D2030-A1 Motherboard" }] } , Device { vendor_id : 4153u16 , id : 28674u16 , name : "USB 2.0 Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 21616u16 , name : "K7SOM+ 5.2C Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28688u16 , name : "MS-6701 motherboard" } , SubSystem { subvendor : 5385u16 , subdevice : 28674u16 , name : "Onboard USB Controller" } , SubSystem { subvendor : 5940u16 , subdevice : 4245u16 , name : "D2030-A1" }] } , Device { vendor_id : 4153u16 , id : 28679u16 , name : "FireWire Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 28701u16 , name : "MS-6701" }] } , Device { vendor_id : 4153u16 , id : 28690u16 , name : "SiS7012 AC'97 Sound Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 3845u16 , name : "A928 (i-Buddie)" } , SubSystem { subvendor : 4153u16 , subdevice : 28690u16 , name : "SiS 7012 onboard [Asus P4SC-EA] AC'97 Sound Controller" } , SubSystem { subvendor : 4163u16 , subdevice : 33167u16 , name : "A8S-X Motherboard" } , SubSystem { subvendor : 5110u16 , subdevice : 768u16 , name : "CMI9739(A) on ECS K7S series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 22608u16 , name : "MSI 648 Max (MS-6585)" } , SubSystem { subvendor : 5218u16 , subdevice : 28688u16 , name : "MS-6701 motherboard" } , SubSystem { subvendor : 5565u16 , subdevice : 4097u16 , name : "DFI 661FX motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4255u16 , name : "D2030-A1 Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 28690u16 , name : "K7S41GX motherboard" }] } , Device { vendor_id : 4153u16 , id : 28691u16 , name : "AC'97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 28694u16 , name : "SiS7016 PCI Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4153u16 , subdevice : 28694u16 , name : "SiS7016 10/100 Ethernet Adapter" }] } , Device { vendor_id : 4153u16 , id : 28696u16 , name : "SiS PCI Audio Accelerator" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 438u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4116u16 , subdevice : 439u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4121u16 , subdevice : 28696u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4133u16 , subdevice : 14u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4133u16 , subdevice : 24u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4153u16 , subdevice : 28696u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4163u16 , subdevice : 5203u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4163u16 , subdevice : 32779u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4173u16 , subdevice : 32994u16 , name : "VAIO PCV-J200" } , SubSystem { subvendor : 4180u16 , subdevice : 28696u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4221u16 , subdevice : 21296u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4221u16 , subdevice : 21328u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 4464u16 , subdevice : 12809u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5218u16 , subdevice : 16394u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5284u16 , subdevice : 8329u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5325u16 , subdevice : 8596u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5375u16 , subdevice : 4352u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5421u16 , subdevice : 34824u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5464u16 , subdevice : 4355u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5464u16 , subdevice : 8704u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5475u16 , subdevice : 28696u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 5573u16 , subdevice : 273u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 9999u16 , subdevice : 41329u16 , name : "SiS PCI Audio Accelerator" } , SubSystem { subvendor : 41120u16 , subdevice : 34u16 , name : "SiS PCI Audio Accelerator" }] } , Device { vendor_id : 4153u16 , id : 28697u16 , name : "SiS7019 Audio Accelerator" , subsystems : & [] } , Device { vendor_id : 4153u16 , id : 29954u16 , name : "Azalia Audio Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33185u16 , name : "P5SD2-FM/S mainboard" }] }] }),
        (4420, Vendor { id : 4420u16 , name : "Cincinnati Milacron" , devices : & [Device { vendor_id : 4420u16 , id : 1u16 , name : "Noservo controller" , subsystems : & [] }] }),
        (2084, Vendor { id : 2084u16 , name : "T1042 [Freescale]" , devices : & [] }),
        (4638, Vendor { id : 4638u16 , name : "CSPI" , devices : & [Device { vendor_id : 4638u16 , id : 513u16 , name : "Myrinet 2000 Scalable Cluster Interconnect" , subsystems : & [] }] }),
        (4791, Vendor { id : 4791u16 , name : "Cognex Corporation" , devices : & [] }),
        (19932, Vendor { id : 19932u16 , name : "ILC Data Device Corp" , devices : & [Device { vendor_id : 19932u16 , id : 256u16 , name : "DD-42924I5-300 (ARINC 429 Data Bus)" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 768u16 , name : "SB-3620 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 832u16 , name : "SB-3623 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 1024u16 , name : "SB-3622 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 1280u16 , name : "SB-3621 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 1296u16 , name : "SB-3624 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2049u16 , name : "BU-65570I1 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2050u16 , name : "BU-65570I2 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2065u16 , name : "BU-65572I1 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2066u16 , name : "BU-65572I2 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2177u16 , name : "BU-65570T1 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2178u16 , name : "BU-65570T2 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2193u16 , name : "BU-65572T1 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2194u16 , name : "BU-65572T2 MIL-STD-1553 Test and Simulation" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2305u16 , name : "BU-65565C1 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2306u16 , name : "BU-65565C2 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2307u16 , name : "BU-65565C3 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2308u16 , name : "BU-65565C4 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2817u16 , name : "BU-65569I1 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2818u16 , name : "BU-65569I2 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2819u16 , name : "BU-65569I3 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 2820u16 , name : "BU-65569I4 MIL-STD-1553 Data Bus" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 3329u16 , name : "SB-3641 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 3344u16 , name : "SB-365x Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 12032u16 , name : "SB-3642 Motion Feedback Device" , subsystems : & [] } , Device { vendor_id : 19932u16 , id : 12288u16 , name : "SB-3644 Motion Feedback Device" , subsystems : & [] }] }),
        (4618, Vendor { id : 4618u16 , name : "Synaptel" , devices : & [] }),
        (4440, Vendor { id : 4440u16 , name : "Voarx R & D Inc" , devices : & [Device { vendor_id : 4440u16 , id : 12305u16 , name : "Tokenet/vg 1001/10m anylan" , subsystems : & [] } , Device { vendor_id : 4440u16 , id : 36944u16 , name : "Lanfleet/Truevalue" , subsystems : & [] } , Device { vendor_id : 4440u16 , id : 36945u16 , name : "Lanfleet/Truevalue" , subsystems : & [] }] }),
        (4431, Vendor { id : 4431u16 , name : "Digi International" , devices : & [Device { vendor_id : 4431u16 , id : 2u16 , name : "AccelePort EPC" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 3u16 , name : "RightSwitch SE-6" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 4u16 , name : "AccelePort Xem" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 5u16 , name : "AccelePort Xr" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 6u16 , name : "AccelePort Xr,C/X" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 9u16 , name : "AccelePort Xr/J" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 10u16 , name : "AccelePort EPC/J" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 12u16 , name : "DataFirePRIme T1 (1-port)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 13u16 , name : "SyncPort 2-Port (x.25/FR)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 17u16 , name : "AccelePort 8r EIA-232 (IBM)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 18u16 , name : "AccelePort 8r EIA-422" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 19u16 , name : "AccelePort Xr" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 20u16 , name : "AccelePort 8r EIA-422" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 21u16 , name : "AccelePort Xem" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 22u16 , name : "AccelePort EPC/X" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 23u16 , name : "AccelePort C/X" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 26u16 , name : "DataFirePRIme E1 (1-port)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 27u16 , name : "AccelePort C/X (IBM)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 28u16 , name : "AccelePort Xr (SAIP)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 29u16 , name : "DataFire RAS T1/E1/PRI" , subsystems : & [SubSystem { subvendor : 4431u16 , subdevice : 80u16 , name : "DataFire RAS E1 Adapter" } , SubSystem { subvendor : 4431u16 , subdevice : 81u16 , name : "DataFire RAS Dual E1 Adapter" } , SubSystem { subvendor : 4431u16 , subdevice : 82u16 , name : "DataFire RAS T1 Adapter" } , SubSystem { subvendor : 4431u16 , subdevice : 83u16 , name : "DataFire RAS Dual T1 Adapter" }] } , Device { vendor_id : 4431u16 , id : 35u16 , name : "AccelePort RAS" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 36u16 , name : "DataFire RAS B4 ST/U" , subsystems : & [SubSystem { subvendor : 4431u16 , subdevice : 48u16 , name : "DataFire RAS BRI U Adapter" } , SubSystem { subvendor : 4431u16 , subdevice : 49u16 , name : "DataFire RAS BRI S/T Adapter" }] } , Device { vendor_id : 4431u16 , id : 38u16 , name : "AccelePort 4r 920" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 39u16 , name : "AccelePort Xr 920" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 40u16 , name : "ClassicBoard 4" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 41u16 , name : "ClassicBoard 8" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 52u16 , name : "AccelePort 2r 920" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 53u16 , name : "DataFire DSP T1/E1/PRI cPCI" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 64u16 , name : "AccelePort Xp" , subsystems : & [SubSystem { subvendor : 4431u16 , subdevice : 66u16 , name : "AccelePort 2p PCI" } , SubSystem { subvendor : 4431u16 , subdevice : 67u16 , name : "AccelePort 4p PCI" } , SubSystem { subvendor : 4431u16 , subdevice : 68u16 , name : "AccelePort 8p PCI" } , SubSystem { subvendor : 4431u16 , subdevice : 69u16 , name : "AccelePort 16p PCI" } , SubSystem { subvendor : 4431u16 , subdevice : 78u16 , name : "AccelePort 32p PCI" }] } , Device { vendor_id : 4431u16 , id : 66u16 , name : "AccelePort 2p" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 67u16 , name : "AccelePort 4p" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 68u16 , name : "AccelePort 8p" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 69u16 , name : "AccelePort 16p" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 78u16 , name : "AccelePort 32p" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 112u16 , name : "Datafire Micro V IOM2 (Europe)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 113u16 , name : "Datafire Micro V (Europe)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 114u16 , name : "Datafire Micro V IOM2 (North America)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 115u16 , name : "Datafire Micro V (North America)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 176u16 , name : "Digi Neo 4" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 177u16 , name : "Digi Neo 8" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 200u16 , name : "Digi Neo 2 DB9" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 201u16 , name : "Digi Neo 2 DB9 PRI" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 202u16 , name : "Digi Neo 2 RJ45" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 203u16 , name : "Digi Neo 2 RJ45 PRI" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 204u16 , name : "Digi Neo 1 422" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 205u16 , name : "Digi Neo 1 422 485" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 206u16 , name : "Digi Neo 2 422 485" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 208u16 , name : "ClassicBoard 4 422" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 209u16 , name : "ClassicBoard 8 422" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 241u16 , name : "Digi Neo PCI-E 4 port" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 244u16 , name : "Digi Neo 4 (IBM version)" , subsystems : & [] } , Device { vendor_id : 4431u16 , id : 24577u16 , name : "Avanstar" , subsystems : & [] }] }),
        (4797, Vendor { id : 4797u16 , name : "Computerm Corp." , devices : & [] }),
        (43528, Vendor { id : 43528u16 , name : "iTuner" , devices : & [] }),
        (7027, Vendor { id : 7027u16 , name : "Fresco Logic" , devices : & [Device { vendor_id : 7027u16 , id : 4096u16 , name : "FL1000G USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 7516u16 , subdevice : 4096u16 , name : "Anker USB 3.0 Express Card" }] } , Device { vendor_id : 7027u16 , id : 4105u16 , name : "FL1009 USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 7027u16 , id : 4352u16 , name : "FL1100 USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 5816u16 , subdevice : 28209u16 , name : "Allegro Pro USB 3.0 PCIe" }] } , Device { vendor_id : 7027u16 , id : 5120u16 , name : "USB 3.0 Host Controller" , subsystems : & [] }] }),
        (5541, Vendor { id : 5541u16 , name : "Toyota Macs Inc" , devices : & [] }),
        (5536, Vendor { id : 5536u16 , name : "Compumaster SRL" , devices : & [] }),
        (1550, Vendor { id : 1550u16 , name : "Lightelligence" , devices : & [Device { vendor_id : 1550u16 , id : 1u16 , name : "Hummingbird ES" , subsystems : & [] }] }),
        (8076, Vendor { id : 8076u16 , name : "Exascend,INC." , devices : & [Device { vendor_id : 8076u16 , id : 32776u16 , name : "SM8008 NVMe SSD [Px5 Series SSD]" , subsystems : & [SubSystem { subvendor : 8076u16 , subdevice : 1u16 , name : "PD5 Series General" } , SubSystem { subvendor : 8076u16 , subdevice : 2u16 , name : "Px5 Series General" }] } , Device { vendor_id : 8076u16 , id : 33638u16 , name : "SM8366 NVMe SSD [PD5 Series SSD]" , subsystems : & [SubSystem { subvendor : 8076u16 , subdevice : 1u16 , name : "PD5 Series General" } , SubSystem { subvendor : 8076u16 , subdevice : 2u16 , name : "Px5 Series General" }] }] }),
        (45491, Vendor { id : 45491u16 , name : "Shiva Europe Limited" , devices : & [] }),
        (5463, Vendor { id : 5463u16 , name : "MEDIASTAR Co Ltd" , devices : & [] }),
        (6316, Vendor { id : 6316u16 , name : "DViCO Corporation" , devices : & [Device { vendor_id : 6316u16 , id : 54528u16 , name : "FusionHDTV 5" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 55296u16 , name : "FusionHDTV 3 Gold" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 55312u16 , name : "FusionHDTV 3 Gold-Q" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 55328u16 , name : "FusionHDTV 3 Gold-T" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 56112u16 , name : "FusionHDTV DVB-T Pro" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 56128u16 , name : "FusionHDTV DVB-T Hybrid" , subsystems : & [] } , Device { vendor_id : 6316u16 , id : 56184u16 , name : "FusionHDTV DVB-T Dual Express" , subsystems : & [] }] }),
        (5585, Vendor { id : 5585u16 , name : "Infineon Technologies AG" , devices : & [] }),
        (15677, Vendor { id : 15677u16 , name : "3DLabs" , devices : & [Device { vendor_id : 15677u16 , id : 1u16 , name : "GLINT 300SX" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 2u16 , name : "GLINT 500TX" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 0u16 , name : "GLoria L" }] } , Device { vendor_id : 15677u16 , id : 3u16 , name : "GLINT Delta" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 0u16 , name : "GLoria XL" }] } , Device { vendor_id : 15677u16 , id : 4u16 , name : "Permedia" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 5u16 , name : "Permedia" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 6u16 , name : "GLINT MX" , subsystems : & [SubSystem { subvendor : 0u16 , subdevice : 0u16 , name : "GLoria XL" } , SubSystem { subvendor : 4168u16 , subdevice : 2626u16 , name : "GLoria XXL" }] } , Device { vendor_id : 15677u16 , id : 7u16 , name : "3D Extreme" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 8u16 , name : "GLINT Gamma G1" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 2626u16 , name : "GLoria XXL" }] } , Device { vendor_id : 15677u16 , id : 9u16 , name : "Permedia II 2D+3D" , subsystems : & [SubSystem { subvendor : 4160u16 , subdevice : 17u16 , name : "AccelStar II" } , SubSystem { subvendor : 4168u16 , subdevice : 2626u16 , name : "GLoria XXL" } , SubSystem { subvendor : 5097u16 , subdevice : 4096u16 , name : "6221L-4U" } , SubSystem { subvendor : 15677u16 , subdevice : 256u16 , name : "AccelStar II 3D Accelerator" } , SubSystem { subvendor : 15677u16 , subdevice : 273u16 , name : "Permedia 3:16" } , SubSystem { subvendor : 15677u16 , subdevice : 276u16 , name : "Santa Ana" } , SubSystem { subvendor : 15677u16 , subdevice : 278u16 , name : "Oxygen GVX1" } , SubSystem { subvendor : 15677u16 , subdevice : 281u16 , name : "Scirocco" } , SubSystem { subvendor : 15677u16 , subdevice : 288u16 , name : "Santa Ana PCL" } , SubSystem { subvendor : 15677u16 , subdevice : 293u16 , name : "Oxygen VX1" } , SubSystem { subvendor : 15677u16 , subdevice : 295u16 , name : "Permedia3 Create!" }] } , Device { vendor_id : 15677u16 , id : 10u16 , name : "GLINT R3" , subsystems : & [SubSystem { subvendor : 15677u16 , subdevice : 289u16 , name : "Oxygen VX1" }] } , Device { vendor_id : 15677u16 , id : 12u16 , name : "GLINT R3 [Oxygen VX1]" , subsystems : & [SubSystem { subvendor : 15677u16 , subdevice : 324u16 , name : "Oxygen VX1-4X AGP [Permedia 4]" }] } , Device { vendor_id : 15677u16 , id : 13u16 , name : "GLint R4 rev A" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 14u16 , name : "GLINT Gamma G2" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 17u16 , name : "GLint R4 rev B" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 18u16 , name : "GLint R5 rev A" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 19u16 , name : "GLint R5 rev B" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 32u16 , name : "VP10 visual processor" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 34u16 , name : "VP10 visual processor" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 36u16 , name : "VP9 visual processor" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 44u16 , name : "Wildcat Realizm 100/200" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 48u16 , name : "Wildcat Realizm 800" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 50u16 , name : "Wildcat Realizm 500" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 256u16 , name : "Permedia II 2D+3D" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 1953u16 , name : "Wildcat III 6210" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 1954u16 , name : "Sun XVR-500 Graphics Accelerator" , subsystems : & [SubSystem { subvendor : 15677u16 , subdevice : 4167u16 , name : "Sun XVR-600 Graphics Accelerator" }] } , Device { vendor_id : 15677u16 , id : 1955u16 , name : "Wildcat IV 7210" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 4100u16 , name : "Permedia" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 15620u16 , name : "Permedia" , subsystems : & [] } , Device { vendor_id : 15677u16 , id : 65535u16 , name : "Glint VGA" , subsystems : & [] }] }),
        (8291, Vendor { id : 8291u16 , name : "Hubei Yangtze Mason Semiconductor Technology Co., Ltd." , devices : & [Device { vendor_id : 8291u16 , id : 3000u16 , name : "MC3000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 3001u16 , name : "RC3000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 3100u16 , name : "MC3100" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 3101u16 , name : "MC3100T" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 3200u16 , name : "RC3200" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 5000u16 , name : "MC5000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 5001u16 , name : "RC5000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 5126u16 , name : "ME7000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 7000u16 , name : "MC7000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 7001u16 , name : "RC7000" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 7200u16 , name : "LC7200" , subsystems : & [] } , Device { vendor_id : 8291u16 , id : 14000u16 , name : "ME14000 NVMe SSD" , subsystems : & [SubSystem { subvendor : 8291u16 , subdevice : 4096u16 , name : "NVMe SSD ME14000 U.2 3.2TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4097u16 , name : "NVMe SSD ME14000 U.2 3.84TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4098u16 , name : "NVMe SSD ME14000 U.2 6.4TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4099u16 , name : "NVMe SSD ME14000 U.2 7.68TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4100u16 , name : "NVMe SSD ME14000 U.2 12.8TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4101u16 , name : "NVMe SSD ME14000 U.2 15.36TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4102u16 , name : "NVMe SSD ME14000 U.2 25.6TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4103u16 , name : "NVMe SSD ME14000 U.2 30.72TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4104u16 , name : "NVMe SSD ME14000 U.2 61.44TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4105u16 , name : "NVMe SSD ME14000 U.2 122.88TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4112u16 , name : "NVMe SSD ME14000 U.2 3.2TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4113u16 , name : "NVMe SSD ME14000 U.2 3.84TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4114u16 , name : "NVMe SSD ME14000 U.2 6.4TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4115u16 , name : "NVMe SSD ME14000 U.2 7.68TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4116u16 , name : "NVMe SSD ME14000 U.2 12.8TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4117u16 , name : "NVMe SSD ME14000 U.2 15.36TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4118u16 , name : "NVMe SSD ME14000 U.2 25.6TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4119u16 , name : "NVMe SSD ME14000 U.2 30.72TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4120u16 , name : "NVMe SSD ME14000 U.2 61.44TB" } , SubSystem { subvendor : 8291u16 , subdevice : 4121u16 , name : "NVMe SSD ME14000 U.2 122.88TB" }] }] }),
        (4367, Vendor { id : 4367u16 , name : "Ross Technology" , devices : & [] }),
        (6974, Vendor { id : 6974u16 , name : "Teradata Corp." , devices : & [Device { vendor_id : 6974u16 , id : 8104u16 , name : "BYNET BIC2SE/X" , subsystems : & [SubSystem { subvendor : 6974u16 , subdevice : 163u16 , name : "BYNET BIC2SX" } , SubSystem { subvendor : 6974u16 , subdevice : 195u16 , name : "BYNET BIC2SE" }] }] }),
        (6664, Vendor { id : 6664u16 , name : "Sierra semiconductor" , devices : & [Device { vendor_id : 6664u16 , id : 0u16 , name : "SC15064" , subsystems : & [] }] }),
        (54277, Vendor { id : 54277u16 , name : "LeapIO" , devices : & [Device { vendor_id : 54277u16 , id : 33280u16 , name : "LeapHBA" , subsystems : & [] } , Device { vendor_id : 54277u16 , id : 33281u16 , name : "LeapRAID" , subsystems : & [] }] }),
        (5514, Vendor { id : 5514u16 , name : "Digalog Systems Inc" , devices : & [] }),
        (4430, Vendor { id : 4430u16 , name : "Nikon Systems Inc" , devices : & [] }),
        (167, Vendor { id : 167u16 , name : "Teles AG (Wrong ID)" , devices : & [] }),
        (33124, Vendor { id : 33124u16 , name : "HippStor Technology Co., Ltd" , devices : & [Device { vendor_id : 33124u16 , id : 5632u16 , name : "HP600 U.2 NVME SSD" , subsystems : & [] }] }),
        (7222, Vendor { id : 7222u16 , name : "Annapurna Labs Ltd." , devices : & [Device { vendor_id : 7222u16 , id : 1u16 , name : "Gigabit Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 7222u16 , id : 2u16 , name : "SFP+ 10G Ethernet Adapter" , subsystems : & [] }] }),
        (4249, Vendor { id : 4249u16 , name : "Samsung Electronics Co., Ltd" , devices : & [] }),
        (4813, Vendor { id : 4813u16 , name : "Aims Lab" , devices : & [] }),
        (4967, Vendor { id : 4967u16 , name : "Hitachi Zosen Corporation" , devices : & [] }),
        (7949, Vendor { id : 7949u16 , name : "DeGirum Corp." , devices : & [Device { vendor_id : 7949u16 , id : 256u16 , name : "AI Accelerator [ORCA]" , subsystems : & [] }] }),
        (4563, Vendor { id : 4563u16 , name : "Trancell Systems Inc" , devices : & [] }),
        (4757, Vendor { id : 4757u16 , name : "Imagenation Corporation" , devices : & [Device { vendor_id : 4757u16 , id : 2048u16 , name : "PXR800" , subsystems : & [] } , Device { vendor_id : 4757u16 , id : 4096u16 , name : "PXD1000" , subsystems : & [] }] }),
        (1841, Vendor { id : 1841u16 , name : "Jingjia Microelectronics Co Ltd" , devices : & [Device { vendor_id : 1841u16 , id : 256u16 , name : "JMN100" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 257u16 , name : "JMN100 16G" }] } , Device { vendor_id : 1841u16 , id : 4176u16 , name : "JM1050" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 4177u16 , name : "JM1050 8G" }] } , Device { vendor_id : 1841u16 , id : 4352u16 , name : "JM1100" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 4353u16 , name : "JM1100-C" } , SubSystem { subvendor : 1841u16 , subdevice : 4354u16 , name : "JM1100-II" } , SubSystem { subvendor : 1841u16 , subdevice : 4355u16 , name : "JM1100-I" } , SubSystem { subvendor : 1841u16 , subdevice : 4356u16 , name : "JM1100-M" } , SubSystem { subvendor : 1841u16 , subdevice : 4357u16 , name : "JM1100-Y" } , SubSystem { subvendor : 1841u16 , subdevice : 4358u16 , name : "JM1100-EI" } , SubSystem { subvendor : 1841u16 , subdevice : 4359u16 , name : "JM1100-EM" } , SubSystem { subvendor : 1841u16 , subdevice : 4360u16 , name : "JY1008" } , SubSystem { subvendor : 1841u16 , subdevice : 4361u16 , name : "JY1032" }] } , Device { vendor_id : 1841u16 , id : 4354u16 , name : "JM1100-Y" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 4385u16 , name : "JY1032 LE 64G" } , SubSystem { subvendor : 1841u16 , subdevice : 4386u16 , name : "JY1032 LE 32G" }] } , Device { vendor_id : 1841u16 , id : 29184u16 , name : "JM7200 Series GPU" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 29185u16 , name : "JM7201" } , SubSystem { subvendor : 1841u16 , subdevice : 29186u16 , name : "JM7202" } , SubSystem { subvendor : 1841u16 , subdevice : 29192u16 , name : "JM7200" } , SubSystem { subvendor : 1841u16 , subdevice : 29202u16 , name : "JM7200" } , SubSystem { subvendor : 1841u16 , subdevice : 29204u16 , name : "JM7500" } , SubSystem { subvendor : 1841u16 , subdevice : 29205u16 , name : "JM7200" }] } , Device { vendor_id : 1841u16 , id : 37120u16 , name : "JM9100" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37121u16 , name : "JM9100" } , SubSystem { subvendor : 1841u16 , subdevice : 37122u16 , name : "JM9100-I" } , SubSystem { subvendor : 1841u16 , subdevice : 37123u16 , name : "JM9100M" }] } , Device { vendor_id : 1841u16 , id : 37130u16 , name : "JH910" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37130u16 , name : "JH910" } , SubSystem { subvendor : 1841u16 , subdevice : 37131u16 , name : "JH910-I" } , SubSystem { subvendor : 1841u16 , subdevice : 37132u16 , name : "JH910-M" }] } , Device { vendor_id : 1841u16 , id : 37376u16 , name : "JM9200" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 37386u16 , name : "JH920" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37386u16 , name : "JH920" } , SubSystem { subvendor : 1841u16 , subdevice : 37387u16 , name : "JH920-I" } , SubSystem { subvendor : 1841u16 , subdevice : 37388u16 , name : "JH920-M" } , SubSystem { subvendor : 1841u16 , subdevice : 37389u16 , name : "JH920-II" }] } , Device { vendor_id : 1841u16 , id : 37387u16 , name : "JH920-I" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 37388u16 , name : "JH920-M" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 37392u16 , name : "JM9210" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37392u16 , name : "JM9210" } , SubSystem { subvendor : 1841u16 , subdevice : 37393u16 , name : "JM9210-I" }] } , Device { vendor_id : 1841u16 , id : 37393u16 , name : "JM9210-I" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 37424u16 , name : "JM9230" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37424u16 , name : "JM9230" } , SubSystem { subvendor : 1841u16 , subdevice : 37425u16 , name : "JM9230-I" }] } , Device { vendor_id : 1841u16 , id : 37425u16 , name : "JM9231-I" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 37456u16 , name : "JM9250" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37456u16 , name : "JM9250" }] } , Device { vendor_id : 1841u16 , id : 37642u16 , name : "JH930-I" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 37642u16 , name : "JH930-I" } , SubSystem { subvendor : 1841u16 , subdevice : 37643u16 , name : "JH930-M" } , SubSystem { subvendor : 1841u16 , subdevice : 37644u16 , name : "JH930" } , SubSystem { subvendor : 1841u16 , subdevice : 37645u16 , name : "JH930-II" }] } , Device { vendor_id : 1841u16 , id : 37643u16 , name : "JH930-M" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 61457u16 , name : "JM1100-IV" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 61713u16 , name : "JM1100-MV" , subsystems : & [] } , Device { vendor_id : 1841u16 , id : 65297u16 , name : "JM1100-YV" , subsystems : & [SubSystem { subvendor : 1841u16 , subdevice : 4357u16 , name : "JM11 Series" } , SubSystem { subvendor : 1841u16 , subdevice : 4360u16 , name : "JY1008 vGPU" } , SubSystem { subvendor : 1841u16 , subdevice : 4361u16 , name : "JY1032 vGPU" } , SubSystem { subvendor : 1841u16 , subdevice : 4385u16 , name : "JY1032 LE vGPU" } , SubSystem { subvendor : 1841u16 , subdevice : 4386u16 , name : "JY1032 LE vGPU" }] }] }),
        (7219, Vendor { id : 7219u16 , name : "Daktronics, Inc" , devices : & [] }),
        (4382, Vendor { id : 4382u16 , name : "Eldec" , devices : & [] }),
        (24, Vendor { id : 24u16 , name : "Fn-Link Technology Limited" , devices : & [Device { vendor_id : 24u16 , id : 25170u16 , name : "6252CPUB 802.11ax PCIe Wireless Network Adapter" , subsystems : & [] }] }),
        (5044, Vendor { id : 5044u16 , name : "Wellbean Co Inc" , devices : & [] }),
        (6217, Vendor { id : 6217u16 , name : "ASRock Incorporation" , devices : & [Device { vendor_id : 6217u16 , id : 4432u16 , name : "ASPEED AST1150 PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6217u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (5560, Vendor { id : 5560u16 , name : "ADDI-DATA GmbH" , devices : & [Device { vendor_id : 5560u16 , id : 4097u16 , name : "APCI1516 SP controller (16 digi outputs)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 4099u16 , name : "APCI1032 SP controller (32 digi inputs w/ opto coupler)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 4100u16 , name : "APCI2032 SP controller (32 digi outputs)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 4101u16 , name : "APCI2200 SP controller (8/16 digi outputs (relay))" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 4102u16 , name : "APCI1564 SP controller (32 digi ins, 32 digi outs)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 4106u16 , name : "APCI1696 SP controller (96 TTL I/Os)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 12289u16 , name : "APCI3501 SP controller (analog output board)" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 12303u16 , name : "APCI3600 Noise and vibration measurement board" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 28673u16 , name : "APCI7420 2-port Serial Controller" , subsystems : & [] } , Device { vendor_id : 5560u16 , id : 28674u16 , name : "APCI7300 Serial Controller" , subsystems : & [] }] }),
        (61507, Vendor { id : 61507u16 , name : "ASUSTeK Computer Inc. (Wrong ID)" , devices : & [] }),
        (8438, Vendor { id : 8438u16 , name : "Shenzhen Zhishi Network Technology Co., Ltd." , devices : & [Device { vendor_id : 8438u16 , id : 1u16 , name : "MPU H1" , subsystems : & [] }] }),
        (44062, Vendor { id : 44062u16 , name : "Digital Receiver Technology Inc" , devices : & [] }),
        (6629, Vendor { id : 6629u16 , name : "Huawei Technologies Co., Ltd." , devices : & [Device { vendor_id : 6629u16 , id : 291u16 , name : "ES3000 V3 NVMe PCIe SSD" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 12322u16 , name : "NVMe SSD ES3600P V3 800GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 12323u16 , name : "NVMe SSD ES3600P V3 1200GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 12324u16 , name : "NVMe SSD ES3600P V3 1600GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 12325u16 , name : "NVMe SSD ES3600P V3 2000GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 12326u16 , name : "NVMe SSD ES3600P V3 3200GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 12339u16 , name : "NVMe SSD ES3600C V3 1200GB HHHL AIC" } , SubSystem { subvendor : 6629u16 , subdevice : 12340u16 , name : "NVMe SSD ES3600C V3 1600GB HHHL AIC" } , SubSystem { subvendor : 6629u16 , subdevice : 12342u16 , name : "NVMe SSD ES3600C V3 3200GB HHHL AIC" }] } , Device { vendor_id : 6629u16 , id : 512u16 , name : "Hi1822 Family (2*100GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53561u16 , name : "Hi1822 SP572 (2*100GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53565u16 , name : "Hi1822 SC371 (2*100GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53575u16 , name : "Hi1822 SP573 (2*100GE)" }] } , Device { vendor_id : 6629u16 , id : 514u16 , name : "Hi1822 Family (2*32G FC)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53577u16 , name : "Hi1822 SP528 (2*32G FC)" } , SubSystem { subvendor : 6629u16 , subdevice : 54018u16 , name : "Hi1822 SP521 (2*32G FC)" } , SubSystem { subvendor : 6629u16 , subdevice : 54020u16 , name : "Hi1822 SP526 (2*32G FC)" }] } , Device { vendor_id : 6629u16 , id : 515u16 , name : "Hi1822 Family (2*16G FC)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53576u16 , name : "Hi1822 SP527 (2*16G FC)" } , SubSystem { subvendor : 6629u16 , subdevice : 54017u16 , name : "Hi1822 SP520 (2*16G FC)" } , SubSystem { subvendor : 6629u16 , subdevice : 54021u16 , name : "Hi1822 SP525 (2*16G FC)" }] } , Device { vendor_id : 6629u16 , id : 516u16 , name : "Hi1822 Family (4*10GE)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 517u16 , name : "Hi1822 Family (2*100GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 57127u16 , name : "Hi1822 MZ731 MEZZ (2*100GE)" }] } , Device { vendor_id : 6629u16 , id : 518u16 , name : "Hi1822 Family (2*25GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53560u16 , name : "Hi1822 SP582 (2*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53562u16 , name : "Hi1822 SC381 (2*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53573u16 , name : "Hi1822 SP586 (2*25GE)" }] } , Device { vendor_id : 6629u16 , id : 520u16 , name : "Hi1822 Family (2*100GE)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 523u16 , name : "Hi1822 Family (4*25GE)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 524u16 , name : "Hi1822 Family (4*32G FC)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 525u16 , name : "Hi1822 Family (2*40GE)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 528u16 , name : "Hi1822 Family (4*25GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 57134u16 , name : "Hi1822 MZ532 MEZZ (4*25GE)" }] } , Device { vendor_id : 6629u16 , id : 529u16 , name : "Hi1822 Family (4*25GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53551u16 , name : "Hi1822 SP571 (4*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53559u16 , name : "Hi1822 SP581 (4*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53570u16 , name : "Hi1822 SP583 (4*25GE)" }] } , Device { vendor_id : 6629u16 , id : 530u16 , name : "Hi1822 Family (2*8G FC)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 54019u16 , name : "Hi1822 SP522 (2*8G FC)" } , SubSystem { subvendor : 6629u16 , subdevice : 54022u16 , name : "Hi1822 SP523 (2*8G FC)" }] } , Device { vendor_id : 6629u16 , id : 546u16 , name : "Hi1822 Family" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 81u16 , name : "Hi1822 SP681 (2*25/10GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 82u16 , name : "Hi1822 SP680 (4*25/10GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 161u16 , name : "Hi1822 SP670 (2*100GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 338u16 , name : "Hi1822 SP623Q (4*25/10GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 417u16 , name : "Hi1822 SP625D (2*100GE)" }] } , Device { vendor_id : 6629u16 , id : 553u16 , name : "Hi1872 Family" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 554u16 , name : "Hi1872 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 560u16 , name : "Hi1825 Family" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 561u16 , name : "Hi1825 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 5904u16 , name : "iBMA Virtual Network Adapter" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 5905u16 , name : "Hi171x Series [iBMC Intelligent Management system chip w/VGA support]" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 5906u16 , name : "Intelligent Management system chip Virtual Network Adapter" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 6178u16 , name : "Hi1822 Family (4*25GE)" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 53545u16 , name : "Hi1822 SP570 (4*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53558u16 , name : "Hi1822 SP580 (4*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53569u16 , name : "Hi1822 SP583 (4*25GE)" } , SubSystem { subvendor : 6629u16 , subdevice : 53574u16 , name : "Hi1822 SP585 (4*25GE)" }] } , Device { vendor_id : 6629u16 , id : 14064u16 , name : "Intelligent Management system chip SATA AHCI support" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14100u16 , name : "ES3000 V5 NVMe PCIe SSD" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 21266u16 , name : "NVMe SSD ES3500P V5 2000GB 2.5\" U.2" }] } , Device { vendor_id : 6629u16 , id : 14110u16 , name : "Hi1822 Family Virtual Bridge" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14128u16 , name : "Intelligent Management system chip GE support" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14164u16 , name : "ES3000 V6 NVMe PCIe SSD" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 24866u16 , name : "NVMe SSD ES3600P V6 1600GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 24867u16 , name : "NVMe SSD ES3600P V6 3200GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 24868u16 , name : "NVMe SSD ES3600P V6 6400GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 24897u16 , name : "NVMe SSD ES3800P V6 800GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 24898u16 , name : "NVMe SSD ES3800P V6 1600GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 25106u16 , name : "NVMe SSD ES3500P V6 1920GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 25107u16 , name : "NVMe SSD ES3500P V6 3840GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 25108u16 , name : "NVMe SSD ES3500P V6 7680GB 2.5\" U.2" } , SubSystem { subvendor : 6629u16 , subdevice : 25109u16 , name : "NVMe SSD ES3500P V6 15360GB 2.5\" U.2" }] } , Device { vendor_id : 6629u16 , id : 14168u16 , name : "SP686C RAID Controller Card" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 389u16 , name : "RAID SP686C-M-16i 2G" } , SubSystem { subvendor : 6629u16 , subdevice : 417u16 , name : "RAID SP686C-M-40i 2G" } , SubSystem { subvendor : 6629u16 , subdevice : 420u16 , name : "RAID SP686C-M-16i 4G" } , SubSystem { subvendor : 6629u16 , subdevice : 424u16 , name : "RAID SP686C-MH-32i 4G" } , SubSystem { subvendor : 6629u16 , subdevice : 429u16 , name : "RAID SP686C-M-40i 4G" }] } , Device { vendor_id : 6629u16 , id : 14174u16 , name : "Hi1822 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14175u16 , name : "Hi1822 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14192u16 , name : "Intelligent Management system chip Virtual Feature USB2.0 HOST support" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14238u16 , name : "Hi1822 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14239u16 , name : "Hi1822 Family Virtual Function" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14256u16 , name : "Intelligent Management system chip USB3.0 HOST support" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14320u16 , name : "Intelligent Management system chip USB2.0 HOST support" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14384u16 , name : "Intelligent Management system chip Memory-Mapped Buffer Interface" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 14424u16 , name : "SP186 HBA Controller Card" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 288u16 , name : "HBA SP186-M-32i" } , SubSystem { subvendor : 6629u16 , subdevice : 293u16 , name : "HBA SP186-M-40i" } , SubSystem { subvendor : 6629u16 , subdevice : 384u16 , name : "HBA SP186-M-16i" } , SubSystem { subvendor : 6629u16 , subdevice : 392u16 , name : "HBA SP186-M-8i" }] } , Device { vendor_id : 6629u16 , id : 41248u16 , name : "HiSilicon PCIe Root Port" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41249u16 , name : "HiSilicon PCI-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41250u16 , name : "HiSilicon Embedded DMA Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41252u16 , name : "HiSilicon Internal SDI Function Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41253u16 , name : "HiSilicon SDI Network Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41254u16 , name : "HiSilicon SDI NVMe Storage Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41255u16 , name : "HiSilicon SDI Accelerator" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41258u16 , name : "HiSilicon Add-on PCI-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41260u16 , name : "HiSilicon Embedded PCIe DFX" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41261u16 , name : "HiSilicon Embedded PMU" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41262u16 , name : "HiSilicon Embedded PCIe PTT" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41263u16 , name : "HiSilicon DFX Registers" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41504u16 , name : "HNS GE Network Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41505u16 , name : "HNS GE/10GE/25GE Network Controller" , subsystems : & [SubSystem { subvendor : 6629u16 , subdevice : 1108u16 , name : "TM280" } , SubSystem { subvendor : 6629u16 , subdevice : 1228u16 , name : "TM210" } , SubSystem { subvendor : 6629u16 , subdevice : 53578u16 , name : "TM280 4*25G" } , SubSystem { subvendor : 6629u16 , subdevice : 53579u16 , name : "TM210 4*GE" }] } , Device { vendor_id : 6629u16 , id : 41506u16 , name : "HNS GE/10GE/25GE RDMA Network Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41508u16 , name : "HNS GE/10GE/25GE/50GE RDMA Network Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41510u16 , name : "HNS GE/10GE/25GE/50GE/100GE RDMA Network Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41514u16 , name : "HiSilicon Network For SDI" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41518u16 , name : "HNS Network Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41519u16 , name : "HNS RDMA Network Controller (Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41520u16 , name : "HiSilicon SAS 3.0 HBA" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41525u16 , name : "HiSilicon AHCI HBA" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41528u16 , name : "HiSilicon USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41529u16 , name : "HiSilicon USB 2.0 2-port Host Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41530u16 , name : "HiSilicon USB 2.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41531u16 , name : "HiSilicon USB 1.1 Host Controller" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41552u16 , name : "HiSilicon ZIP Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41553u16 , name : "HiSilicon ZIP Engine(Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41557u16 , name : "HiSilicon SEC Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41558u16 , name : "HiSilicon SEC Engine(Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41560u16 , name : "HiSilicon HPRE Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41561u16 , name : "HiSilicon HPRE Engine(Virtual Function)" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41562u16 , name : "HiSilicon RDE Engine" , subsystems : & [] } , Device { vendor_id : 6629u16 , id : 41563u16 , name : "HiSilicon RDE Engine(Virtual Function)" , subsystems : & [] }] }),
        (7649, Vendor { id : 7649u16 , name : "Tekram Technology Co.,Ltd." , devices : & [Device { vendor_id : 7649u16 , id : 913u16 , name : "TRM-S1040 [DC-315 / DC-395 series]" , subsystems : & [] } , Device { vendor_id : 7649u16 , id : 8224u16 , name : "DC-390 Series SCSI Adapter [AMD Am53C974]" , subsystems : & [] } , Device { vendor_id : 7649u16 , id : 26892u16 , name : "690c" , subsystems : & [] } , Device { vendor_id : 7649u16 , id : 56361u16 , name : "DC290" , subsystems : & [] }] }),
        (5458, Vendor { id : 5458u16 , name : "RACAL AIRTECH Ltd" , devices : & [] }),
        (7592, Vendor { id : 7592u16 , name : "Corigine, Inc." , devices : & [Device { vendor_id : 7592u16 , id : 14336u16 , name : "Network Flow Processor 3800" , subsystems : & [] } , Device { vendor_id : 7592u16 , id : 14339u16 , name : "Network Flow Processor 3800 Virtual Function" , subsystems : & [] }] }),
        (4152, Vendor { id : 4152u16 , name : "AMP, Inc" , devices : & [] }),
        (7199, Vendor { id : 7199u16 , name : "SoftLab-NSK" , devices : & [Device { vendor_id : 7199u16 , id : 21u16 , name : "FD842" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 25u16 , name : "FD722" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 26u16 , name : "FD788" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 27u16 , name : "FD720" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 28u16 , name : "FD922" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 29u16 , name : "Vega" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 31u16 , name : "FD940" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 32u16 , name : "FD2110" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 33u16 , name : "FD722" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 34u16 , name : "FD788" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 35u16 , name : "FD722-M2" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 36u16 , name : "FD722 with bypass" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 37u16 , name : "FD922" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 38u16 , name : "FD788" , subsystems : & [] } , Device { vendor_id : 7199u16 , id : 39u16 , name : "FD940" , subsystems : & [] }] }),
        (65535, Vendor { id : 65535u16 , name : "Illegal Vendor ID" , devices : & [] }),
        (4172, Vendor { id : 4172u16 , name : "Texas Instruments" , devices : & [Device { vendor_id : 4172u16 , id : 1280u16 , name : "100 MBit LAN Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 1288u16 , name : "TMS380C2X Compressor Interface" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 4096u16 , name : "Eagle i/f AS" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 4172u16 , name : "PCI1510 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 15620u16 , name : "TVP4010 [Permedia]" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 15623u16 , name : "TVP4020 [Permedia 2]" , subsystems : & [SubSystem { subvendor : 4113u16 , subdevice : 19728u16 , name : "Comet" } , SubSystem { subvendor : 4160u16 , subdevice : 15u16 , name : "AccelStar II" } , SubSystem { subvendor : 4160u16 , subdevice : 17u16 , name : "AccelStar II" } , SubSystem { subvendor : 4168u16 , subdevice : 2609u16 , name : "WINNER 2000" } , SubSystem { subvendor : 4168u16 , subdevice : 2610u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4168u16 , subdevice : 2612u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4168u16 , subdevice : 2613u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4168u16 , subdevice : 2614u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4168u16 , subdevice : 2627u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4168u16 , subdevice : 2628u16 , name : "GLoria Synergy" } , SubSystem { subvendor : 4221u16 , subdevice : 9779u16 , name : "WinFast 3D L2300" } , SubSystem { subvendor : 4242u16 , subdevice : 294u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 295u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 310u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 321u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 326u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 328u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 329u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 338u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 340u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 341u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 342u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4242u16 , subdevice : 343u16 , name : "FIRE GL 1000 PRO" } , SubSystem { subvendor : 4247u16 , subdevice : 15617u16 , name : "Jeronimo Pro" } , SubSystem { subvendor : 4354u16 , subdevice : 4111u16 , name : "Graphics Blaster Extreme" } , SubSystem { subvendor : 15677u16 , subdevice : 256u16 , name : "Reference Permedia 2 3D" }] } , Device { vendor_id : 4172u16 , id : 32768u16 , name : "PCILynx/PCILynx2 IEEE 1394 Link Layer Controller" , subsystems : & [SubSystem { subvendor : 4190u16 , subdevice : 32771u16 , name : "FireBoard200" } , SubSystem { subvendor : 5187u16 , subdevice : 32771u16 , name : "FireBoard200" } , SubSystem { subvendor : 5187u16 , subdevice : 32773u16 , name : "FireBoard400" } , SubSystem { subvendor : 5187u16 , subdevice : 32774u16 , name : "FireBoard400" } , SubSystem { subvendor : 58559u16 , subdevice : 4112u16 , name : "CF1-1-SNARE" } , SubSystem { subvendor : 58559u16 , subdevice : 4128u16 , name : "CF1-2-SNARE" } , SubSystem { subvendor : 58559u16 , subdevice : 4160u16 , name : "FireCompact400" }] } , Device { vendor_id : 4172u16 , id : 32777u16 , name : "TSB12LV22 IEEE-1394 Controller" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32818u16 , name : "8032 OHCI i.LINK (IEEE 1394) Controller" } , SubSystem { subvendor : 5187u16 , subdevice : 32784u16 , name : "FireBoard400-OHCI" }] } , Device { vendor_id : 4172u16 , id : 32791u16 , name : "PCI4410 FireWire Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 32793u16 , name : "TSB12LV23 IEEE-1394 Controller" , subsystems : & [SubSystem { subvendor : 4541u16 , subdevice : 10u16 , name : "Studio DV500-1394" } , SubSystem { subvendor : 4541u16 , subdevice : 14u16 , name : "Studio DV" } , SubSystem { subvendor : 5187u16 , subdevice : 32784u16 , name : "FireBoard400-OHCI" } , SubSystem { subvendor : 58559u16 , subdevice : 4112u16 , name : "CF2-1-CYMBAL" }] } , Device { vendor_id : 4172u16 , id : 32800u16 , name : "TSB12LV26 IEEE-1394 Controller (Link)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4173u16 , subdevice : 32994u16 , name : "VAIO PCV-J200" } , SubSystem { subvendor : 4541u16 , subdevice : 15u16 , name : "Studio DV500-1394" } , SubSystem { subvendor : 4541u16 , subdevice : 28u16 , name : "Excalibur 4.1" } , SubSystem { subvendor : 5187u16 , subdevice : 32784u16 , name : "FireBoard400-OHCI" }] } , Device { vendor_id : 4172u16 , id : 32801u16 , name : "TSB43AA22 IEEE-1394 Controller (PHY/Link Integrated)" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" }] } , Device { vendor_id : 4172u16 , id : 32802u16 , name : "TSB43AB22 IEEE-1394a-2000 Controller (PHY/Link) [iOHCI-Lynx]" , subsystems : & [SubSystem { subvendor : 4172u16 , subdevice : 32803u16 , name : "TSB43AB22/A IEEE-1394a-2000 Controller (PHY/Link)" }] } , Device { vendor_id : 4172u16 , id : 32803u16 , name : "TSB43AB22A IEEE-1394a-2000 Controller (PHY/Link) [iOHCI-Lynx]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 360u16 , name : "Precision Workstation 670 Mainboard" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 32907u16 , name : "K8N4/A8N Series Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33115u16 , name : "P5W DH Deluxe Motherboard" } , SubSystem { subvendor : 5187u16 , subdevice : 32803u16 , name : "FireCard400" } , SubSystem { subvendor : 32902u16 , subdevice : 20548u16 , name : "Desktop Board DP35DP" }] } , Device { vendor_id : 4172u16 , id : 32804u16 , name : "TSB43AB23 IEEE-1394a-2000 Controller (PHY/Link)" , subsystems : & [SubSystem { subvendor : 4221u16 , subdevice : 26144u16 , name : "Winfast DV2000 FireWire Controller" } , SubSystem { subvendor : 5187u16 , subdevice : 32804u16 , name : "FireBoard Blue" } , SubSystem { subvendor : 5208u16 , subdevice : 4096u16 , name : "Motherboard" }] } , Device { vendor_id : 4172u16 , id : 32805u16 , name : "TSB82AA2 IEEE-1394b Link Layer Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33084u16 , name : "P5P series mainboard" } , SubSystem { subvendor : 5187u16 , subdevice : 32805u16 , name : "FireBoard800" } , SubSystem { subvendor : 5208u16 , subdevice : 4096u16 , name : "GA-K8N Ultra-9 Mainboard" } , SubSystem { subvendor : 5446u16 , subdevice : 32805u16 , name : "FWB-PCI01" } , SubSystem { subvendor : 6140u16 , subdevice : 32805u16 , name : "GIC3800" }] } , Device { vendor_id : 4172u16 , id : 32806u16 , name : "TSB43AB21 IEEE-1394a-2000 Controller (PHY/Link)" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 53u16 , name : "TravelMate 660" } , SubSystem { subvendor : 4133u16 , subdevice : 60u16 , name : "Aspire 2001WLCi (Compaq CL50 motherboard)" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4163u16 , subdevice : 32909u16 , name : "A7V333 mainboard." }] } , Device { vendor_id : 4172u16 , id : 32807u16 , name : "PCI4451 IEEE-1394 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 229u16 , name : "Latitude C810" } , SubSystem { subvendor : 4136u16 , subdevice : 230u16 , name : "PCI4451 IEEE-1394 Controller (Dell Inspiron 8100)" }] } , Device { vendor_id : 4172u16 , id : 32809u16 , name : "PCI4510 IEEE-1394 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2900" }] } , Device { vendor_id : 4172u16 , id : 32811u16 , name : "PCI7410,7510,7610 OHCI-Lynx Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 334u16 , name : "PCI7410,7510,7610 OHCI-Lynx Controller (Latitude D800)" }] } , Device { vendor_id : 4172u16 , id : 32814u16 , name : "PCI7x20 1394a-2000 OHCI Two-Port PHY/Link-Layer Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" }] } , Device { vendor_id : 4172u16 , id : 32817u16 , name : "PCIxx21/PCIxx11/PCIx515 PC Card Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" }] } , Device { vendor_id : 4172u16 , id : 32818u16 , name : "OHCI Compliant IEEE 1394 Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" }] } , Device { vendor_id : 4172u16 , id : 32819u16 , name : "PCIxx21/PCIxx11 Flash Media Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop" } , SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" }] } , Device { vendor_id : 4172u16 , id : 32820u16 , name : "PCIxx21/PCIxx11 SD Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" }] } , Device { vendor_id : 4172u16 , id : 32821u16 , name : "PCIxx21/PCIxx11 Smart Card Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "Compaq nw8240/nx8220" } , SubSystem { subvendor : 4156u16 , subdevice : 2372u16 , name : "Compaq nc6220 Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 2460u16 , name : "NX6110/NC6120" }] } , Device { vendor_id : 4172u16 , id : 32822u16 , name : "PCI6515 Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 32824u16 , name : "PCI6515 SmartCard Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 32825u16 , name : "PCIxx12 Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" }] } , Device { vendor_id : 4172u16 , id : 32826u16 , name : "PCIxx12 OHCI Compliant IEEE 1394 Host Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "nx9420" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" }] } , Device { vendor_id : 4172u16 , id : 32827u16 , name : "PCIxx12 Flash Media Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "nx9420" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" } , SubSystem { subvendor : 4173u16 , subdevice : 33298u16 , name : "VAIO VGN-N21E" } , SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" }] } , Device { vendor_id : 4172u16 , id : 32828u16 , name : "PCIxx12 SDA Standard Compliant SD Host Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "nx9420" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq nw8440" }] } , Device { vendor_id : 4172u16 , id : 32829u16 , name : "PCIxx12 GemCore based SmartCard controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12449u16 , name : "NC2400" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "nc8430" } , SubSystem { subvendor : 4156u16 , subdevice : 12458u16 , name : "nc6310" }] } , Device { vendor_id : 4172u16 , id : 33025u16 , name : "TSB43DB42 IEEE-1394a-2000 Controller (PHY/Link)" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33281u16 , name : "PCI1620 Firmware Loading Function" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33284u16 , name : "PCI7410/7510/7610 PCI Firmware Loading Function" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 334u16 , name : "Latitude D800" }] } , Device { vendor_id : 4172u16 , id : 33329u16 , name : "XIO2000(A)/XIO2200A PCI Express-to-PCI Bridge" , subsystems : & [SubSystem { subvendor : 22136u16 , subdevice : 4660u16 , name : "DC-1394 PCIe" }] } , Device { vendor_id : 4172u16 , id : 33330u16 , name : "XIO3130 PCI Express Switch (Upstream)" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33331u16 , name : "XIO3130 PCI Express Switch (Downstream)" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33333u16 , name : "XIO2200A IEEE-1394a-2000 Controller (PHY/Link)" , subsystems : & [SubSystem { subvendor : 22136u16 , subdevice : 4660u16 , name : "DC-1394 PCIe" }] } , Device { vendor_id : 4172u16 , id : 33342u16 , name : "XIO2213A/B/XIO2221 PCI Express to PCI Bridge [Cheetah Express]" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33343u16 , name : "XIO2213A/B/XIO2221 IEEE-1394b OHCI Controller [Cheetah Express]" , subsystems : & [SubSystem { subvendor : 5446u16 , subdevice : 32828u16 , name : "FWB-PCIE1X11B" }] } , Device { vendor_id : 4172u16 , id : 33344u16 , name : "XIO2001 PCI Express-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 33345u16 , name : "TUSB73x0 SuperSpeed USB 3.0 xHCI Host Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1202u16 , name : "S824 (8286-42A)" }] } , Device { vendor_id : 4172u16 , id : 33792u16 , name : "ACX 100 22Mbps Wireless Interface" , subsystems : & [SubSystem { subvendor : 4378u16 , subdevice : 4129u16 , name : "SpeedStream 1021 (SS1021) Wireless Cardbus PC Card" } , SubSystem { subvendor : 4486u16 , subdevice : 15104u16 , name : "DWL-650+ PC Card cardbus 22Mbs Wireless Adapter [AirPlus]" } , SubSystem { subvendor : 4486u16 , subdevice : 15105u16 , name : "DWL-520+ 22Mbps PCI Wireless Adapter" } , SubSystem { subvendor : 5013u16 , subdevice : 8705u16 , name : "WL22-PC" } , SubSystem { subvendor : 5757u16 , subdevice : 45616u16 , name : "Netopia TER/WPC11N1 Wireless LAN Card" } , SubSystem { subvendor : 5797u16 , subdevice : 6145u16 , name : "WE302-TF" } , SubSystem { subvendor : 5803u16 , subdevice : 34049u16 , name : "WL-8305 IEEE802.11b+ Wireless LAN PCI Adapter" }] } , Device { vendor_id : 4172u16 , id : 33793u16 , name : "ACX 100 22Mbps Wireless Interface" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 34952u16 , name : "Multicore DSP+ARM KeyStone II SOC" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 36864u16 , name : "Wireless Interface (of unknown type)" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 36965u16 , name : "TMS320DM642" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 36966u16 , name : "ACX 111 54Mbps Wireless Interface" , subsystems : & [SubSystem { subvendor : 776u16 , subdevice : 13316u16 , name : "G-102 v1 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 776u16 , subdevice : 13318u16 , name : "G-162 v2 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 4172u16 , subdevice : 36966u16 , name : "WL212 Sitecom Wireless Network PCI-Card 100M (Version 1)" } , SubSystem { subvendor : 4172u16 , subdevice : 36967u16 , name : "TNETW1130GVF" } , SubSystem { subvendor : 4172u16 , subdevice : 37014u16 , name : "Trendnet TEW-412PC Wireless PCI Adapter (Version A)" } , SubSystem { subvendor : 4436u16 , subdevice : 812u16 , name : "WLI-CB-G54L" } , SubSystem { subvendor : 4486u16 , subdevice : 15108u16 , name : "DWL-G520+ Wireless PCI Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 15109u16 , name : "DWL-G650+ AirPlusG+ CardBus Wireless LAN" } , SubSystem { subvendor : 4486u16 , subdevice : 15112u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.B1)" } , SubSystem { subvendor : 4997u16 , subdevice : 19456u16 , name : "WG311v2 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 5073u16 , subdevice : 43936u16 , name : "SWLMP-54108 108Mbps Wireless mini PCI card 802.11g+" } , SubSystem { subvendor : 5354u16 , subdevice : 43783u16 , name : "GW-NS54GM Wireless Cardbus Adapter" } , SubSystem { subvendor : 5757u16 , subdevice : 45664u16 , name : "SWL-2600N" } , SubSystem { subvendor : 5803u16 , subdevice : 36967u16 , name : "A90-200WG-01 802.11g Wireless PC Card" } , SubSystem { subvendor : 5868u16 , subdevice : 269u16 , name : "USR5416 802.11g Wireless Turbo PCI Adapter" } , SubSystem { subvendor : 5868u16 , subdevice : 270u16 , name : "USR5410 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 51u16 , name : "WPC54G v2 802.11g Wireless-G Notebook Adapter" } , SubSystem { subvendor : 6095u16 , subdevice : 50u16 , name : "G-162 v1 802.11g Wireless Cardbus Adapter" } , SubSystem { subvendor : 6095u16 , subdevice : 51u16 , name : "Z-Com XG650 Wireless miniPCI 802.11b/g" } , SubSystem { subvendor : 6270u16 , subdevice : 13323u16 , name : "G-302 v2 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 6270u16 , subdevice : 13324u16 , name : "G-360 v2 802.11g Wireless PCI Adapter" }] } , Device { vendor_id : 4172u16 , id : 40961u16 , name : "TDC1570" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 41216u16 , name : "TDC1561" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 41218u16 , name : "TNETA1575 HyperSAR Plus w/PCI Host i/f & UTOPIA i/f" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 41222u16 , name : "TMS320C6414 TMS320C6415 TMS320C6416" , subsystems : & [SubSystem { subvendor : 5980u16 , subdevice : 20480u16 , name : "ASI50xx Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 25600u16 , name : "ASI6400 Cobranet series" } , SubSystem { subvendor : 5980u16 , subdevice : 34560u16 , name : "ASI87xx Radio Tuner card" }] } , Device { vendor_id : 4172u16 , id : 44048u16 , name : "PCI1050" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44049u16 , name : "PCI1053" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44050u16 , name : "PCI1130" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44051u16 , name : "PCI1031" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44053u16 , name : "PCI1131" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44054u16 , name : "PCI1250" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 146u16 , name : "ThinkPad 600" }] } , Device { vendor_id : 4172u16 , id : 44055u16 , name : "PCI1220" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44056u16 , name : "PCI1260" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44057u16 , name : "PCI1221" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44058u16 , name : "PCI1210" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44059u16 , name : "PCI1450" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45331u16 , name : "Armada M700" } , SubSystem { subvendor : 4116u16 , subdevice : 304u16 , name : "ThinkPad 600X/A21m/T20/T22" }] } , Device { vendor_id : 4172u16 , id : 44060u16 , name : "PCI1225" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45345u16 , name : "Armada E500" } , SubSystem { subvendor : 4136u16 , subdevice : 136u16 , name : "Latitude CPi A400XT" }] } , Device { vendor_id : 4172u16 , id : 44061u16 , name : "PCI1251A" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44062u16 , name : "PCI1211" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44063u16 , name : "PCI1251B" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44064u16 , name : "TI 2030" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44065u16 , name : "PCI2031" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44066u16 , name : "PCI2032 PCI Docking Bridge" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44067u16 , name : "PCI2250 PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44072u16 , name : "PCI2050 PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44076u16 , name : "PCI2060 PCI-to-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44080u16 , name : "PCI1260 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44096u16 , name : "PCI4450 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44097u16 , name : "PCI4410 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44098u16 , name : "PCI4451 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 230u16 , name : "PCI4451 PC card CardBus Controller (Inspiron 8100)" }] } , Device { vendor_id : 4172u16 , id : 44100u16 , name : "PCI4510 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 329u16 , name : "Inspiron 5100" } , SubSystem { subvendor : 4136u16 , subdevice : 355u16 , name : "Latitude D505" } , SubSystem { subvendor : 4136u16 , subdevice : 406u16 , name : "Inspiron 5160" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" }] } , Device { vendor_id : 4172u16 , id : 44102u16 , name : "PCI4520 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1362u16 , name : "ThinkPad" }] } , Device { vendor_id : 4172u16 , id : 44103u16 , name : "PCI7510 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 319u16 , name : "Precision M60" } , SubSystem { subvendor : 4136u16 , subdevice : 334u16 , name : "Latitude D800" }] } , Device { vendor_id : 4172u16 , id : 44104u16 , name : "PCI7610 PC Card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44105u16 , name : "PCI7410 PC Card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44106u16 , name : "PCI7510/7610 CardBus Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 313u16 , name : "Latitude D400" } , SubSystem { subvendor : 4136u16 , subdevice : 334u16 , name : "Latitude D800" }] } , Device { vendor_id : 4172u16 , id : 44107u16 , name : "PCI7610 SD/MMC controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44108u16 , name : "PCI7610 Memory Stick controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44112u16 , name : "PCI1410 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44113u16 , name : "PCI1420 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 78u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 328u16 , name : "ThinkPad A20m" } , SubSystem { subvendor : 4116u16 , subdevice : 571u16 , name : "ThinkPad T23" } , SubSystem { subvendor : 4136u16 , subdevice : 177u16 , name : "Latitude C600" } , SubSystem { subvendor : 4136u16 , subdevice : 298u16 , name : "Latitude C640" } , SubSystem { subvendor : 4147u16 , subdevice : 32973u16 , name : "Versa Note VXi" } , SubSystem { subvendor : 4303u16 , subdevice : 4245u16 , name : "Lifebook S-4510/C6155" } , SubSystem { subvendor : 58559u16 , subdevice : 4096u16 , name : "CP2-2-HIPHOP" }] } , Device { vendor_id : 4172u16 , id : 44114u16 , name : "PCI1451 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44115u16 , name : "PCI1421 PC card Cardbus Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44116u16 , name : "PCI1620 PC Card Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2224u16 , name : "tc1100 tablet" }] } , Device { vendor_id : 4172u16 , id : 44117u16 , name : "PCI1520 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1298u16 , name : "ThinkPad T30/T40" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" }] } , Device { vendor_id : 4172u16 , id : 44118u16 , name : "PCI1510 PC card Cardbus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1298u16 , name : "ThinkPad R50e" } , SubSystem { subvendor : 4116u16 , subdevice : 1320u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 6058u16 , subdevice : 8210u16 , name : "ThinkPad T60/R60 series" }] } , Device { vendor_id : 4172u16 , id : 44128u16 , name : "PCI2040 PCI to DSP Bridge Controller" , subsystems : & [SubSystem { subvendor : 5980u16 , subdevice : 20736u16 , name : "ASI51xx Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 24832u16 , name : "ASI61xx Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 25088u16 , name : "ASI62xx Audio Adapter" } , SubSystem { subvendor : 5980u16 , subdevice : 34816u16 , name : "ASI88xx Audio Adapter" } , SubSystem { subvendor : 6255u16 , subdevice : 12289u16 , name : "WR-G303 PCI radio receiver" } , SubSystem { subvendor : 6255u16 , subdevice : 12293u16 , name : "WR-G305 PCI radio receiver" } , SubSystem { subvendor : 6255u16 , subdevice : 12545u16 , name : "WR-G313 PCI radio receiver" } , SubSystem { subvendor : 6255u16 , subdevice : 12549u16 , name : "WR-G315 PCI radio receiver" }] } , Device { vendor_id : 4172u16 , id : 44173u16 , name : "PCI 7620" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 44174u16 , name : "PCI7420 CardBus Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" }] } , Device { vendor_id : 4172u16 , id : 44175u16 , name : "PCI7420/7620 SD/MS-Pro Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 397u16 , name : "Inspiron 700m/710m" }] } , Device { vendor_id : 4172u16 , id : 45057u16 , name : "TMS320C6424" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 45074u16 , name : "TDA4VH PCIE Root Complex" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 65024u16 , name : "FireWire Host Controller" , subsystems : & [] } , Device { vendor_id : 4172u16 , id : 65027u16 , name : "12C01A FireWire Host Controller" , subsystems : & [] }] }),
        (4528, Vendor { id : 4528u16 , name : "V3 Semiconductor Inc." , devices : & [Device { vendor_id : 4528u16 , id : 2u16 , name : "V300PSC" , subsystems : & [] } , Device { vendor_id : 4528u16 , id : 658u16 , name : "V292PBC [Am29030/40 Bridge]" , subsystems : & [] } , Device { vendor_id : 4528u16 , id : 2400u16 , name : "V96xPBC" , subsystems : & [] } , Device { vendor_id : 4528u16 , id : 34826u16 , name : "Deltacast Delta-HD-22" , subsystems : & [] } , Device { vendor_id : 4528u16 , id : 51552u16 , name : "V96DPC" , subsystems : & [] }] }),
        (7946, Vendor { id : 7946u16 , name : "Motorcomm Microelectronics." , devices : & [Device { vendor_id : 7946u16 , id : 26625u16 , name : "YT6801 Gigabit Ethernet Controller" , subsystems : & [] }] }),
        (5635, Vendor { id : 5635u16 , name : "Nokia Wireless Communications" , devices : & [] }),
        (5423, Vendor { id : 5423u16 , name : "PHILIPS - CRYPTO" , devices : & [] }),
        (5729, Vendor { id : 5729u16 , name : "Worldspace Corp." , devices : & [] }),
        (5111, Vendor { id : 5111u16 , name : "Wildfire Communications" , devices : & [] }),
        (4877, Vendor { id : 4877u16 , name : "Accelerix Inc" , devices : & [] }),
        (7466, Vendor { id : 7466u16 , name : "KAYA Instruments" , devices : & [Device { vendor_id : 7466u16 , id : 4096u16 , name : "Frame grabber" , subsystems : & [] }] }),
        (4336, Vendor { id : 4336u16 , name : "Peritek Corporation" , devices : & [] }),
        (4206, Vendor { id : 4206u16 , name : "DFI, Inc" , devices : & [] }),
        (8384, Vendor { id : 8384u16 , name : "ExpectedIT GmbH" , devices : & [] }),
        (4783, Vendor { id : 4783u16 , name : "TDK Corporation" , devices : & [Device { vendor_id : 4783u16 , id : 22577u16 , name : "GBDriver GX1 x2 NVMe SSD Controller (DRAM-less)" , subsystems : & [] }] }),
        (5855, Vendor { id : 5855u16 , name : "PIKA Technologies Inc." , devices : & [] }),
        (33570, Vendor { id : 33570u16 , name : "Sodick America Corp." , devices : & [] }),
        (5267, Vendor { id : 5267u16 , name : "MAKER Communications" , devices : & [] }),
        (6562, Vendor { id : 6562u16 , name : "Emulex Corporation" , devices : & [Device { vendor_id : 6562u16 , id : 288u16 , name : "x1 PCIe Gen2 Bridge[Pilot4]" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 512u16 , name : "BladeEngine 10Gb PCI-E iSCSI adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 513u16 , name : "BladeEngine 10Gb PCIe Network Adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 529u16 , name : "BladeEngine2 10Gb Gen2 PCIe Network Adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 530u16 , name : "BladeEngine2 10Gb Gen2 PCIe iSCSI Adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 545u16 , name : "BladeEngine3 10Gb Gen2 PCIe Network Adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 546u16 , name : "BladeEngine3 10Gb Gen2 PCIe iSCSI Adapter" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 1792u16 , name : "OneConnect OCe10100/OCe10102 Series 10 GbE" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5959u16 , name : "NC550SFP DualPort 10GbE Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5961u16 , name : "NC550SFP Dual Port Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5962u16 , name : "NC551m Dual Port FlexFabric 10Gb Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 5963u16 , name : "StorageWorks NC550 DualPort Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13076u16 , name : "NC551i Dual Port FlexFabric 10Gb Adapter" }] } , Device { vendor_id : 6562u16 , id : 1794u16 , name : "OneConnect 10Gb iSCSI Initiator" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 1796u16 , name : "OneConnect OCe10100/OCe10102 Series 10 GbE CNA" , subsystems : & [SubSystem { subvendor : 4319u16 , subdevice : 58882u16 , name : "OneConnect OCe10100 10Gb CNA" } , SubSystem { subvendor : 4319u16 , subdevice : 58928u16 , name : "OneConnect OCe10102-FM-E / OCe10102-FX-E for EMC VNX Symmetrix" }] } , Device { vendor_id : 6562u16 , id : 1808u16 , name : "OneConnect 10Gb NIC (be3)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 976u16 , name : "PCIe2 2-port 10GbE SR Adapter for POWER" } , SubSystem { subvendor : 4116u16 , subdevice : 977u16 , name : "PCIe2 2-port 10GbE SFP+ Copper Adapter for POWER" } , SubSystem { subvendor : 4116u16 , subdevice : 1033u16 , name : "Integrated Multifunction Card with Dual 10GbE SR Optical + Dual 1GbE for Power 750/760" } , SubSystem { subvendor : 4116u16 , subdevice : 1034u16 , name : "Integrated Multifunction Card with Dual 10GbE SR Copper + Dual 1GbE for Power 750/760" } , SubSystem { subvendor : 4156u16 , subdevice : 13077u16 , name : "NC553i 10Gb 2-port FlexFabric Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13120u16 , name : "NC552SFP 2-port 10Gb Server Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13121u16 , name : "NC552m 10Gb 2-port FlexFabric Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13125u16 , name : "NC553m 10Gb 2-port FlexFabric Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13179u16 , name : "NC554FLB 10Gb 2-port FlexFabric Converged Network Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 59187u16 , name : "Flex System EN4054 4-port 10Gb Ethernet Mezzanine Adapter" }] } , Device { vendor_id : 6562u16 , id : 1810u16 , name : "OneConnect 10Gb iSCSI Initiator (be3)" , subsystems : & [] } , Device { vendor_id : 6562u16 , id : 1812u16 , name : "OneConnect 10Gb FCoE Initiator (be3)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13077u16 , name : "NC553i 10Gb 2-port FlexFabric Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13179u16 , name : "NC554FLB 10Gb 2-port FlexFabric Converged Network Adapter" }] } , Device { vendor_id : 6562u16 , id : 2048u16 , name : "ServerView iRMC HTI" , subsystems : & [] }] }),
        (4117, Vendor { id : 4117u16 , name : "LSI Logic Corp of Canada" , devices : & [] }),
        (12273, Vendor { id : 12273u16 , name : "Maginfra Co., LTD" , devices : & [] }),
        (8456, Vendor { id : 8456u16 , name : "HuiLink Technologies (Xiamen) Co., Ltd." , devices : & [Device { vendor_id : 8456u16 , id : 9217u16 , name : "PCIe4.0 to USB3.2 Gen2 Host Controller" , subsystems : & [] }] }),
        (4115, Vendor { id : 4115u16 , name : "Cirrus Logic" , devices : & [Device { vendor_id : 4115u16 , id : 56u16 , name : "GD 7548" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 64u16 , name : "GD 7555 Flat Panel GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 76u16 , name : "GD 7556 Video/Graphics LCD/CRT Ctrlr" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 160u16 , name : "GD 5430/40 [Alpine]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 162u16 , name : "GD 5432 [Alpine]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 164u16 , name : "GD 5434-4 [Alpine]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 168u16 , name : "GD 5434-8 [Alpine]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 172u16 , name : "GD 5436 [Alpine]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 176u16 , name : "GD 5440" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 184u16 , name : "GD 5446" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 4115u16 , id : 188u16 , name : "GD 5480" , subsystems : & [SubSystem { subvendor : 4115u16 , subdevice : 188u16 , name : "CL-GD5480" }] } , Device { vendor_id : 4115u16 , id : 208u16 , name : "GD 5462" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 210u16 , name : "GD 5462 [Laguna I]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 212u16 , name : "GD 5464 [Laguna]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 213u16 , name : "GD 5464 BD [Laguna]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 214u16 , name : "GD 5465 [Laguna]" , subsystems : & [SubSystem { subvendor : 5070u16 , subdevice : 32817u16 , name : "Barco Metheus 2 Megapixel, Dual Head" } , SubSystem { subvendor : 5071u16 , subdevice : 32817u16 , name : "Barco Metheus 2 Megapixel, Dual Head" }] } , Device { vendor_id : 4115u16 , id : 232u16 , name : "GD 5436U" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4352u16 , name : "CL 6729" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4368u16 , name : "PD 6832 PCMCIA/CardBus Ctrlr" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4370u16 , name : "PD 6834 PCMCIA/CardBus Ctrlr" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4371u16 , name : "PD 6833 PCMCIA/CardBus Ctrlr" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4608u16 , name : "GD 7542 [Nordic]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4610u16 , name : "GD 7543 [Viking]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 4612u16 , name : "GD 7541 [Nordic Light]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 16384u16 , name : "MD 5620 [CLM Data Fax Voice]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 17408u16 , name : "CD 4400" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 24577u16 , name : "CS 4610/11 [CrystalClear SoundFusion Audio Accelerator]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 4112u16 , name : "CS4610 SoundFusion Audio Accelerator" }] } , Device { vendor_id : 4115u16 , id : 24579u16 , name : "CS 4614/22/24/30 [CrystalClear SoundFusion Audio Accelerator]" , subsystems : & [SubSystem { subvendor : 4115u16 , subdevice : 17024u16 , name : "Crystal SoundFusion PCI Audio Accelerator" } , SubSystem { subvendor : 4116u16 , subdevice : 339u16 , name : "ThinkPad 600X/A20m" } , SubSystem { subvendor : 5435u16 , subdevice : 4398u16 , name : "DMX XFire 1024" } , SubSystem { subvendor : 5435u16 , subdevice : 4406u16 , name : "SiXPack 5.1+" } , SubSystem { subvendor : 5761u16 , subdevice : 80u16 , name : "Game Theater XP" } , SubSystem { subvendor : 5761u16 , subdevice : 40976u16 , name : "Gamesurround Fortissimo II" } , SubSystem { subvendor : 5761u16 , subdevice : 40977u16 , name : "Gamesurround Fortissimo III 7.1" } , SubSystem { subvendor : 20563u16 , subdevice : 13143u16 , name : "Santa Cruz" }] } , Device { vendor_id : 4115u16 , id : 24580u16 , name : "CS 4614/22/24 [CrystalClear SoundFusion Audio Accelerator]" , subsystems : & [] } , Device { vendor_id : 4115u16 , id : 24581u16 , name : "Crystal CS4281 PCI Audio" , subsystems : & [SubSystem { subvendor : 4115u16 , subdevice : 17025u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4264u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4265u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4266u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4267u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4268u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4269u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4303u16 , subdevice : 4276u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Crystal CS4281 PCI Audio" } , SubSystem { subvendor : 5312u16 , subdevice : 12u16 , name : "Crystal CS4281 PCI Audio" }] }] }),
        (5253, Vendor { id : 5253u16 , name : "ERMA - Electronic GmBH" , devices : & [] }),
        (5063, Vendor { id : 5063u16 , name : "Blue Chip Technology Ltd" , devices : & [Device { vendor_id : 5063u16 , id : 2780u16 , name : "PCI-ADC" , subsystems : & [] } , Device { vendor_id : 5063u16 , id : 2832u16 , name : "PCI-PIO" , subsystems : & [] } , Device { vendor_id : 5063u16 , id : 3344u16 , name : "PCI-DIO" , subsystems : & [] } , Device { vendor_id : 5063u16 , id : 21068u16 , name : "PCI-RLY" , subsystems : & [] } , Device { vendor_id : 5063u16 , id : 22340u16 , name : "PCI-WDT" , subsystems : & [] }] }),
        (4925, Vendor { id : 4925u16 , name : "SS Technologies" , devices : & [] }),
        (6749, Vendor { id : 6749u16 , name : "Celoxica" , devices : & [] }),
        (51713, Vendor { id : 51713u16 , name : "I-TEK OptoElectronics Co., LTD." , devices : & [Device { vendor_id : 51713u16 , id : 530u16 , name : "Vulcan-CL CameraLink Frame Grabber" , subsystems : & [] }] }),
        (4725, Vendor { id : 4725u16 , name : "Network Appliance Corporation" , devices : & [] }),
        (4792, Vendor { id : 4792u16 , name : "Korg" , devices : & [] }),
        (6659, Vendor { id : 6659u16 , name : "ASPEED Technology, Inc." , devices : & [Device { vendor_id : 6659u16 , id : 4432u16 , name : "AST1150 PCI-to-PCI Bridge" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS" } , SubSystem { subvendor : 5593u16 , subdevice : 7504u16 , name : "X14DBG-AP" }] } , Device { vendor_id : 6659u16 , id : 8192u16 , name : "ASPEED Graphics Family" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 2081u16 , name : "X10DRW-i (AST2400 BMC)" } , SubSystem { subvendor : 5593u16 , subdevice : 2098u16 , name : "X10SRL-F (AST2400 BMC)" } , SubSystem { subvendor : 5593u16 , subdevice : 2155u16 , name : "X10DRS (AST2400 BMC)" } , SubSystem { subvendor : 5593u16 , subdevice : 2157u16 , name : "X10SDV (AST2400 BMC)" } , SubSystem { subvendor : 5593u16 , subdevice : 7061u16 , name : "H12SSL-i (AST2500 BMC)" } , SubSystem { subvendor : 5593u16 , subdevice : 7504u16 , name : "X14DBG-AP (AST2600 BMC)" } , SubSystem { subvendor : 6217u16 , subdevice : 8192u16 , name : "Onboard Graphics" }] }] }),
        (5607, Vendor { id : 5607u16 , name : "Get Engineering Corp" , devices : & [] }),
        (5388, Vendor { id : 5388u16 , name : "KYOPAL CO Ltd" , devices : & [] }),
        (6516, Vendor { id : 6516u16 , name : "Star Electronics GmbH & Co. KG" , devices : & [Device { vendor_id : 6516u16 , id : 9u16 , name : "FlexCard PMC-II" , subsystems : & [] } , Device { vendor_id : 6516u16 , id : 17u16 , name : "FlexCard PMC-II Ethernet" , subsystems : & [] } , Device { vendor_id : 6516u16 , id : 24u16 , name : "FlexCard PXIe3" , subsystems : & [] } , Device { vendor_id : 6516u16 , id : 25u16 , name : "FlexCard PCIe3" , subsystems : & [] } , Device { vendor_id : 6516u16 , id : 26u16 , name : "FlexCard PXIe Ethernet" , subsystems : & [] } , Device { vendor_id : 6516u16 , id : 27u16 , name : "FlexCard PCIe Ethernet" , subsystems : & [] }] }),
        (4136, Vendor { id : 4136u16 , name : "Dell" , devices : & [Device { vendor_id : 4136u16 , id : 1u16 , name : "PowerEdge Expandable RAID Controller 2/Si" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1u16 , name : "PowerEdge 2400" }] } , Device { vendor_id : 4136u16 , id : 2u16 , name : "PowerEdge Expandable RAID Controller 3/Di" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2u16 , name : "PowerEdge 4400" } , SubSystem { subvendor : 4136u16 , subdevice : 209u16 , name : "PERC 3/DiV [Viper]" } , SubSystem { subvendor : 4136u16 , subdevice : 217u16 , name : "PERC 3/DiL [Lexus]" }] } , Device { vendor_id : 4136u16 , id : 3u16 , name : "PowerEdge Expandable RAID Controller 3/Si" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 3u16 , name : "PowerEdge 2450" }] } , Device { vendor_id : 4136u16 , id : 4u16 , name : "PowerEdge Expandable RAID Controller 3/Di [Iguana]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 4u16 , name : "PERC 3/DiF [Iguana]" }] } , Device { vendor_id : 4136u16 , id : 6u16 , name : "PowerEdge Expandable RAID Controller 3/Di" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 7u16 , name : "Remote Access Card III" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 8u16 , name : "Remote Access Card III" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 9u16 , name : "Remote Access Card III: BMC/SMIC device not present" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 10u16 , name : "PowerEdge Expandable RAID Controller 3/Di" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 262u16 , name : "PERC 3/DiJ [Jaguar]" } , SubSystem { subvendor : 4136u16 , subdevice : 283u16 , name : "PERC 3/DiD [Dagger]" } , SubSystem { subvendor : 4136u16 , subdevice : 289u16 , name : "PERC 3/DiB [Boxster]" }] } , Device { vendor_id : 4136u16 , id : 12u16 , name : "Embedded Remote Access or ERA/O" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 13u16 , name : "Embedded Remote Access: BMC/SMIC device" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 14u16 , name : "PowerEdge Expandable RAID controller 4/Di" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 15u16 , name : "PowerEdge Expandable RAID controller 4/Di" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 330u16 , name : "PowerEdge 1750" }] } , Device { vendor_id : 4136u16 , id : 16u16 , name : "Remote Access Card 4" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 17u16 , name : "Remote Access Card 4 Daughter Card" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 18u16 , name : "Remote Access Card 4 Daughter Card Virtual UART" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 19u16 , name : "PowerEdge Expandable RAID controller 4" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge Expandable RAID Controller 4e/Si" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge Expandable RAID Controller 4e/Di" } , SubSystem { subvendor : 4136u16 , subdevice : 366u16 , name : "PowerEdge Expandable RAID Controller 4e/Di" } , SubSystem { subvendor : 4136u16 , subdevice : 367u16 , name : "PowerEdge Expandable RAID Controller 4e/Di" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge Expandable RAID Controller 4e/Di" }] } , Device { vendor_id : 4136u16 , id : 20u16 , name : "Remote Access Card 4 Daughter Card SMIC interface" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 21u16 , name : "PowerEdge Expandable RAID controller 5" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7937u16 , name : "PERC 5/E Adapter RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7938u16 , name : "PERC 5/i Adapter RAID Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 7939u16 , name : "PERC 5/i Integrated RAID Controller" }] } , Device { vendor_id : 4136u16 , id : 22u16 , name : "PowerEdge Expandable RAID controller S300" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 7972u16 , name : "PERC S300 Controller" }] } , Device { vendor_id : 4136u16 , id : 115u16 , name : "NV-RAM Adapter" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 3142u16 , name : "PERC S160 RAID controller" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 4136u16 , name : "PCIe Bridge riser" , subsystems : & [] } , Device { vendor_id : 4136u16 , id : 9728u16 , name : "ENT NVMe RT1" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8542u16 , name : "ENT NVMe RT1 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8543u16 , name : "ENT NVMe RT1 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8544u16 , name : "ENT NVMe RT1 FIPS RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8545u16 , name : "ENT NVMe RT1 FIPS RI 7.68TB" }] } , Device { vendor_id : 4136u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (4516, Vendor { id : 4516u16 , name : "Barco Graphics NV" , devices : & [] }),
        (6224, Vendor { id : 6224u16 , name : "Advantest Corporation" , devices : & [Device { vendor_id : 6224u16 , id : 72u16 , name : "EK220-66401 Computer Interface Card" , subsystems : & [] }] }),
        (12288, Vendor { id : 12288u16 , name : "Hansol Electronics Inc." , devices : & [] }),
        (8022, Vendor { id : 8022u16 , name : "SAPEON Inc." , devices : & [] }),
        (20550, Vendor { id : 20550u16 , name : "GemTek Technology Corporation" , devices : & [Device { vendor_id : 20550u16 , id : 4097u16 , name : "PCI Radio" , subsystems : & [] }] }),
        (4455, Vendor { id : 4455u16 , name : "Mutoh Industries Inc" , devices : & [] }),
        (5533, Vendor { id : 5533u16 , name : "Ningbo Harrison Electronics Co Ltd" , devices : & [] }),
        (4142, Vendor { id : 4142u16 , name : "Olivetti Advanced Technology" , devices : & [] }),
        (4663, Vendor { id : 4663u16 , name : "Alta Technology Corporation" , devices : & [] }),
        (5308, Vendor { id : 5308u16 , name : "Globespan Semiconductor Inc." , devices : & [Device { vendor_id : 5308u16 , id : 53250u16 , name : "Pulsar [PCI ADSL Card]" , subsystems : & [] } , Device { vendor_id : 5308u16 , id : 53263u16 , name : "Pulsar [PCI ADSL Card]" , subsystems : & [] }] }),
        (18001, Vendor { id : 18001u16 , name : "TXIC" , devices : & [] }),
        (4747, Vendor { id : 4747u16 , name : "Transwitch Corporation" , devices : & [] }),
        (4753, Vendor { id : 4753u16 , name : "NCS Computer Italia" , devices : & [] }),
        (5390, Vendor { id : 5390u16 , name : "C-PORT Corp" , devices : & [] }),
        (5789, Vendor { id : 5789u16 , name : "Club-3D VB (Wrong ID)" , devices : & [] }),
        (7784, Vendor { id : 7784u16 , name : "Jiangsu Xinsheng Intelligent Technology Co., Ltd" , devices : & [Device { vendor_id : 7784u16 , id : 33041u16 , name : "EP2000Pro PCIe 3 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (4162, Vendor { id : 4162u16 , name : "Micron" , devices : & [Device { vendor_id : 4162u16 , id : 4096u16 , name : "PC Tech RZ1000" , subsystems : & [] } , Device { vendor_id : 4162u16 , id : 4097u16 , name : "PC Tech RZ1001" , subsystems : & [] } , Device { vendor_id : 4162u16 , id : 12288u16 , name : "Samurai_0" , subsystems : & [] } , Device { vendor_id : 4162u16 , id : 12304u16 , name : "Samurai_1" , subsystems : & [] } , Device { vendor_id : 4162u16 , id : 12320u16 , name : "Samurai_IDE" , subsystems : & [] }] }),
        (4529, Vendor { id : 4529u16 , name : "Apricot Computers" , devices : & [] }),
        (5564, Vendor { id : 5564u16 , name : "Agilent Technologies" , devices : & [Device { vendor_id : 5564u16 , id : 256u16 , name : "HPFC-5600 Tachyon DX2+ FC" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 259u16 , name : "QX4 PCI Express quad 4-gigabit Fibre Channel controller" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 261u16 , name : "Celerity FC-44XS/FC-42XS/FC-41XS/FC-44ES/FC-42ES/FC-41ES" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 34u16 , name : "Celerity FC-42XS Fibre Channel Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 37u16 , name : "Celerity FC-44ES Fibre Channel Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 38u16 , name : "Celerity FC-42ES Fibre Channel Adapter" }] } , Device { vendor_id : 5564u16 , id : 1280u16 , name : "Infiniium Memory Controller Interface" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 1281u16 , name : "Infiniium Acquisition System Interface" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 1287u16 , name : "Infiniium Acquisition System (80000 series)" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 1288u16 , name : "Infiniium Acquisition Support" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 2817u16 , name : "82350B PCI GPIB" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 4352u16 , name : "E8001-66442 PCI Express CIC" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 4632u16 , name : "82351A PCI Express GPIB" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 4822u16 , name : "82350C PCI GPIB" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 4823u16 , name : "82351B PCI Express GPIB" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 10530u16 , name : "64 Bit, 133MHz PCI-X Exerciser & Protocol Checker" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 10536u16 , name : "64 Bit, 66MHz PCI Exerciser & Analyzer" , subsystems : & [] } , Device { vendor_id : 5564u16 , id : 10537u16 , name : "64 Bit, 133MHz PCI-X Analyzer & Exerciser" , subsystems : & [] }] }),
        (21249, Vendor { id : 21249u16 , name : "Alliance Semiconductor Corp." , devices : & [Device { vendor_id : 21249u16 , id : 1u16 , name : "ProMotion aT3D" , subsystems : & [] }] }),
        (5532, Vendor { id : 5532u16 , name : "Stratus Computer Systems" , devices : & [] }),
        (5617, Vendor { id : 5617u16 , name : "Times N Systems Inc" , devices : & [] }),
        (5168, Vendor { id : 5168u16 , name : "ITT Aerospace/Communications Division" , devices : & [] }),
        (5384, Vendor { id : 5384u16 , name : "HONDA CONNECTORS/MHOTRONICS Inc" , devices : & [] }),
        (17512, Vendor { id : 17512u16 , name : "Bridgeport machines" , devices : & [] }),
        (5037, Vendor { id : 5037u16 , name : "Nexus Inc" , devices : & [] }),
        (5604, Vendor { id : 5604u16 , name : "VSN Systemen BV" , devices : & [] }),
        (20658, Vendor { id : 20658u16 , name : "TerraTec Electronic GmbH" , devices : & [] }),
        (7103, Vendor { id : 7103u16 , name : "Maxeler Technologies Ltd." , devices : & [Device { vendor_id : 7103u16 , id : 3u16 , name : "MAX3" , subsystems : & [] } , Device { vendor_id : 7103u16 , id : 4u16 , name : "MAX4" , subsystems : & [] }] }),
        (5257, Vendor { id : 5257u16 , name : "KYE Systems Corporation" , devices : & [] }),
        (4208, Vendor { id : 4208u16 , name : "Daewoo Telecom Ltd" , devices : & [] }),
        (4456, Vendor { id : 4456u16 , name : "Thine Electronics Inc" , devices : & [] }),
        (4125, Vendor { id : 4125u16 , name : "Maxim Integrated Products" , devices : & [] }),
        (4521, Vendor { id : 4521u16 , name : "InnoSys Inc." , devices : & [Device { vendor_id : 4521u16 , id : 16960u16 , name : "AMCC S933Q Intelligent Serial Card" , subsystems : & [] }] }),
        (4398, Vendor { id : 4398u16 , name : "Infomedia Microelectronics Inc." , devices : & [] }),
        (8131, Vendor { id : 8131u16 , name : "Emerson" , devices : & [Device { vendor_id : 8131u16 , id : 4u16 , name : "IC695ETM001 RX3i Ethernet PCI Module" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 5u16 , name : "IC695ALG600 RX3i Universal Analog Input" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 6u16 , name : "IC695ALG240 RX3i Isolated Analog Input" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 7u16 , name : "IC695ALG331 RX3i Isolated Analog Output" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 8u16 , name : "IC695PBM300 RX3i Profibus Master" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 9u16 , name : "IC695PBS301 RX3i Profibus Slave" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 10u16 , name : "IC695ALG616 RX3i 16 Channel Input Analog Voltage/Current" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 11u16 , name : "IC695ALG708 RX3i 8 Channel Output Analog Voltage/Current" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 12u16 , name : "IC695ALG626 RX3i 16 Channel Input Analog Voltage/Current with HART" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 13u16 , name : "IC695ALG608 RX3i 8 Channel Input Analog Voltage/Current" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 14u16 , name : "IC695ALG628 RX3i 8 Channel Input Analog Voltage/Current with HART" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 15u16 , name : "IC695ALG728 RX3i 8 Channel Output Analog Voltage/Current with HART" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 16u16 , name : "IC695ALG704 RX3i 4 Channel Output Analog Voltage/Current" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 17u16 , name : "IC695HSC304 RX3i High Speed Counter (4 counters)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 18u16 , name : "IC695HSC308 RX3i High Speed Counter (8 counters)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 19u16 , name : "IC695PMM335 PACMotion Digital Motion Control Module" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 20u16 , name : "IC695CMM002 RX3i Serial Communications Module (2 ports)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 21u16 , name : "IC695CMM004 RX3i Serial Communications Module (4 ports)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 22u16 , name : "IC695PNC001 RX3i Profinet Controller – XH4 version" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 23u16 , name : "IC695PNC001 RX3i Profinet Controller – Atom-based version" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 24u16 , name : "IC695CMX128 RX3i 128MB Control Memory Xchange" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 25u16 , name : "IC695RMX128 RX3i 128MB Redundancy Memory Xchange" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 26u16 , name : "IC695ALG106 RX3i 6 Channel Isolated Analog Input" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 27u16 , name : "IC695ALG112 RX3i 12 Channel Isolated Analog Input" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 28u16 , name : "IC695ALG808 RX3i 8 Channel Isolated Analog Output" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 29u16 , name : "IC695ALG306 RX3i Isolated Thermocouple Module (6 Channel)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 30u16 , name : "IC695ALG312 RX3i Isolated Thermocouple Module (12 Channel)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 31u16 , name : "IC695ALG508 RX3i Isolated RTD Module (8 Channel)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32u16 , name : "IC695ALG412 RX3i Isolated High-Speed Thermocouple Module (12 Channel)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 33u16 , name : "IC695PRS015 RX3i High-Speed Serial Pressure Transducer Module (1 Port)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 34u16 , name : "IC695MDL664 RX3i 16 Circuit Smart Input 24 VDC" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 35u16 , name : "IC695MDL764 RX3i 16 Circuit Smart Output 24/125 VDC 2A (8 Positive / 8 Negative)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 36u16 , name : "IC695MDL765 RX3i 16 Circuit Smart Output 24/125 VDC 2A (16 Positive)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 37u16 , name : "IC695CNM001 RX3i CANopen Master" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 38u16 , name : "Advanced I/O R&D Device" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 39u16 , name : "IC695ECM850 RX3i Ethernet Communication Module" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 40u16 , name : "IC695RMX228 RX3i 128MB Redundancy Memory Xchange with Single Fiber transceiver" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 41u16 , name : "IC695PMM345 PACMotion Digital Motion Control Module (EtherCAT)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 42u16 , name : "IC695HSC318 RX3i High Speed Counter (8 counters)" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32771u16 , name : "IC695CPU310/CMU310/NIU001 RX3i Controller 300 MHz Celeron" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32772u16 , name : "IC695CPU315/320/CRU320 RX3i Controller 1.0 GHz Celeron-M" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32773u16 , name : "IC695PNS001/101 RX3i Profinet Scanner" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32774u16 , name : "IC695CPE/CRE330 RX3i Controller 1.0 GHz G-series" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32775u16 , name : "IC695CPE302/305/310/NIU001+ RX3i Controller 1.1 GHz Atom Z510" , subsystems : & [] } , Device { vendor_id : 8131u16 , id : 32781u16 , name : "IC695CPE302/305/310/NIU001+ RX3i Controller Atom x3930" , subsystems : & [] }] }),
        (6967, Vendor { id : 6967u16 , name : "Signal Processing Devices Sweden AB" , devices : & [Device { vendor_id : 6967u16 , id : 1u16 , name : "ADQ214" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 3u16 , name : "ADQ114" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 5u16 , name : "ADQ112" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 14u16 , name : "ADQ108" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 15u16 , name : "ADQDSP" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 20u16 , name : "ADQ412" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 21u16 , name : "ADQ212" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 27u16 , name : "SDR14" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 28u16 , name : "ADQ1600" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 30u16 , name : "ADQ208" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 31u16 , name : "DSU" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 32u16 , name : "ADQ14" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 35u16 , name : "ADQ7" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 38u16 , name : "ADQ8" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 49u16 , name : "ADQ32/ADQ33" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 50u16 , name : "ADQ System Manager" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 51u16 , name : "ADQ36" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 52u16 , name : "ADQ30" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 53u16 , name : "ADQ35" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 8212u16 , name : "TX320" , subsystems : & [] } , Device { vendor_id : 6967u16 , id : 8217u16 , name : "S6000" , subsystems : & [] }] }),
        (4983, Vendor { id : 4983u16 , name : "Electronic Equipment Production & Distribution GmbH" , devices : & [] }),
        (20044, Vendor { id : 20044u16 , name : "NieL TechSolution" , devices : & [] }),
        (4180, Vendor { id : 4180u16 , name : "Hitachi, Ltd" , devices : & [Device { vendor_id : 4180u16 , id : 12297u16 , name : "2Gbps Fibre Channel to PCI HBA 3009" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12298u16 , name : "4Gbps Fibre Channel to PCI-X HBA 300a" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12299u16 , name : "4Gbps Fibre Channel to PCI-X HBA 300b" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12303u16 , name : "ColdFusion 3 Chipset Processor to I/O Controller" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12304u16 , name : "ColdFusion 3 Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12305u16 , name : "ColdFusion 3e Chipset Processor to I/O Controller" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12306u16 , name : "ColdFusion 3e Chipset Memory Controller Hub" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12311u16 , name : "Unassigned Hitachi Shared FC Device 3017" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12315u16 , name : "Virtual VGA Device" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12317u16 , name : "PCIe-to-PCIe Bridge with Virtualization IO Assist Feature" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12320u16 , name : "FIVE-EX based Fibre Channel to PCIe HBA" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12332u16 , name : "M001 PCI Express Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12333u16 , name : "M001 PCI Express Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 12400u16 , name : "Hitachi FIVE-FX Fibre Channel to PCIe HBA" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 13573u16 , name : "SH7751 PCI Controller (PCIC)" , subsystems : & [] } , Device { vendor_id : 4180u16 , id : 13582u16 , name : "SH7751R PCI Controller (PCIC)" , subsystems : & [] }] }),
        (7884, Vendor { id : 7884u16 , name : "Sunrise AI Inc." , devices : & [Device { vendor_id : 7884u16 , id : 512u16 , name : "Sun S2 Graphics Controller[Sun S2-X1 PCIe]" , subsystems : & [] }] }),
        (52922, Vendor { id : 52922u16 , name : "KEBA AG" , devices : & [] }),
        (7462, Vendor { id : 7462u16 , name : "Kalray Inc." , devices : & [Device { vendor_id : 7462u16 , id : 64u16 , name : "Turbocard2 Accelerator" , subsystems : & [] } , Device { vendor_id : 7462u16 , id : 128u16 , name : "Open Network Interface Card 80G" , subsystems : & [] } , Device { vendor_id : 7462u16 , id : 192u16 , name : "Turbocard3 Accelerator" , subsystems : & [] } , Device { vendor_id : 7462u16 , id : 320u16 , name : "Open Network Interface Card 40G" , subsystems : & [] } , Device { vendor_id : 7462u16 , id : 57348u16 , name : "AB01/EMB01 Development Board" , subsystems : & [] }] }),
        (5291, Vendor { id : 5291u16 , name : "Siemens Industry Software Inc." , devices : & [] }),
        (4961, Vendor { id : 4961u16 , name : "Soliton Systems K.K." , devices : & [] }),
        (8089, Vendor { id : 8089u16 , name : "Shenzhen Techwinsemi Technology Co., Ltd." , devices : & [Device { vendor_id : 8089u16 , id : 4610u16 , name : "TE3420 series / Patriot P320 M.2 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 5634u16 , name : "PCIe Gen4 x4 M.2 2280 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 5640u16 , name : "PCIe Gen4 x4 M.2 2280" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 8072u16 , name : "TE3420 PCIe Gen3 x4 M.2 2280" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 8809u16 , name : "XE4403 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12545u16 , name : "EP1021 Series M.2 2242 NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12549u16 , name : "EP2021 Series M.2 2242 NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12801u16 , name : "EP1021 Series M.2 2280 NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12802u16 , name : "EP1031 Series M.2 2280 NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12805u16 , name : "EP2021 Series M.2 2280 NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12806u16 , name : "EP2031 Series M.2 2280 NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 12897u16 , name : "DP1421 Series M.2 2280 NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 13344u16 , name : "PCIe Gen3 x4 M.2 2280" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24832u16 , name : "TE3420 Series NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24833u16 , name : "XE3420 Series NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24834u16 , name : "TE4407 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24835u16 , name : "XE4407 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24836u16 , name : "TE4408 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24837u16 , name : "XE4408 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24838u16 , name : "TE4402 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24839u16 , name : "XE4402 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24840u16 , name : "TE4405 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24841u16 , name : "XE4405 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24842u16 , name : "TE4406 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24843u16 , name : "XE4406 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24844u16 , name : "TE3410 Series NVMe PCIe Gen3x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24845u16 , name : "TE4403 Series NVMe PCIe Gen4x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24846u16 , name : "TE5440 Series NVMe PCIe Gen5x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24847u16 , name : "TE5441 Series NVMe PCIe Gen5x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24848u16 , name : "TE5480 Series NVMe PCIe Gen5x4 SSD" , subsystems : & [] } , Device { vendor_id : 8089u16 , id : 24849u16 , name : "TE5481 Series NVMe PCIe Gen5x4 SSD" , subsystems : & [] }] }),
        (6093, Vendor { id : 6093u16 , name : "Cadence Design Systems, Inc." , devices : & [Device { vendor_id : 6093u16 , id : 256u16 , name : "CIX PCIe Root Port" , subsystems : & [] } , Device { vendor_id : 6093u16 , id : 56321u16 , name : "Phytium PCIe 3.0 x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 6093u16 , id : 56328u16 , name : "Phytium PCIe 3.0 x8 Root Complex" , subsystems : & [] } , Device { vendor_id : 6093u16 , id : 56342u16 , name : "Phytium PCIe 3.0 x8 Root Complex" , subsystems : & [] } , Device { vendor_id : 6093u16 , id : 64513u16 , name : "Phytium PCIe 3.0 x1 Root Complex" , subsystems : & [] } , Device { vendor_id : 6093u16 , id : 64534u16 , name : "Phytium PCIe 3.0 x8/x16 Root Complex" , subsystems : & [] }] }),
        (5582, Vendor { id : 5582u16 , name : "Genrad Inc" , devices : & [] }),
        (4279, Vendor { id : 4279u16 , name : "3Com Corporation" , devices : & [Device { vendor_id : 4279u16 , id : 1u16 , name : "3c985 1000BaseSX (SX/TX)" , subsystems : & [SubSystem { subvendor : 38992u16 , subdevice : 1u16 , name : "3c985B-SX" }] } , Device { vendor_id : 4279u16 , id : 19u16 , name : "AR5212 802.11abg NIC (3CRDAG675)" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 8241u16 , name : "3CRDAG675 11a/b/g Wireless PCI Adapter" }] } , Device { vendor_id : 4279u16 , id : 2320u16 , name : "3C910-A01" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 4102u16 , name : "MINI PCI type 3B Data Fax Modem" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 4103u16 , name : "Mini PCI 56k Winmodem" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 24923u16 , name : "Mini PCI 56K Modem" } , SubSystem { subvendor : 4279u16 , subdevice : 24924u16 , name : "Mini PCI 56K Modem" }] } , Device { vendor_id : 4279u16 , id : 4609u16 , name : "3c982-TXM 10/100baseTX Dual Port A [Hydra]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 4610u16 , name : "3c982-TXM 10/100baseTX Dual Port B [Hydra]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 5888u16 , name : "3c940 10/100/1000Base-T [Marvell]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33003u16 , name : "A7V600/P4P800/K8V motherboard" } , SubSystem { subvendor : 4279u16 , subdevice : 16u16 , name : "3C940 Gigabit LOM Ethernet Adapter" } , SubSystem { subvendor : 4279u16 , subdevice : 32u16 , name : "3C941 Gigabit LOM Ethernet Adapter" } , SubSystem { subvendor : 5243u16 , subdevice : 5127u16 , name : "KV8-MAX3 motherboard" }] } , Device { vendor_id : 4279u16 , id : 13200u16 , name : "3c339 TokenLink Velocity" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 13712u16 , name : "3c359 TokenLink Velocity XL" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 13712u16 , name : "TokenLink Velocity XL Adapter (3C359/359B)" }] } , Device { vendor_id : 4279u16 , id : 17664u16 , name : "3c450 HomePNA [Tornado]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 20565u16 , name : "3c555 Laptop Hurricane" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 20567u16 , name : "3c575 Megahertz 10/100 LAN CardBus [Boomerang]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 23127u16 , name : "3C575 Megahertz 10/100 LAN Cardbus PC Card" }] } , Device { vendor_id : 4279u16 , id : 20823u16 , name : "3cCFE575BT Megahertz 10/100 LAN CardBus [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 23383u16 , name : "3C575 Megahertz 10/100 LAN Cardbus PC Card" }] } , Device { vendor_id : 4279u16 , id : 21079u16 , name : "3cCFE575CT CardBus [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 23639u16 , name : "FE575C-3Com 10/100 LAN CardBus-Fast Ethernet" }] } , Device { vendor_id : 4279u16 , id : 22784u16 , name : "3c590 10BaseT [Vortex]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 22816u16 , name : "3c592 EISA 10mbps Demon/Vortex" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 22864u16 , name : "3c595 100BaseTX [Vortex]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 22865u16 , name : "3c595 100BaseT4 [Vortex]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 22866u16 , name : "3c595 100Base-MII [Vortex]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 22896u16 , name : "3c597 EISA Fast Demon/Vortex" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 23383u16 , name : "3c595 Megahertz 10/100 LAN CardBus [Boomerang]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 23383u16 , name : "3C575 Megahertz 10/100 LAN Cardbus PC Card" }] } , Device { vendor_id : 4279u16 , id : 24576u16 , name : "3CRSHPW796 [OfficeConnect Wireless CardBus]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 24577u16 , name : "3com 3CRWE154G72 [Office Connect Wireless LAN Adapter]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 24661u16 , name : "3c556 Hurricane CardBus [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 24662u16 , name : "3c556B CardBus [Tornado]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 25942u16 , name : "10/100 Mini PCI Ethernet Adapter" }] } , Device { vendor_id : 4279u16 , id : 25952u16 , name : "3cCFE656 CardBus [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 25962u16 , name : "3CCFEM656 10/100 LAN+56K Modem CardBus" }] } , Device { vendor_id : 4279u16 , id : 25953u16 , name : "3cCFEM656 10/100 LAN+56K Modem CardBus" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 25963u16 , name : "3CCFEM656 10/100 LAN+56K Modem CardBus" }] } , Device { vendor_id : 4279u16 , id : 25954u16 , name : "3cCFEM656B 10/100 LAN+Winmodem CardBus [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 25963u16 , name : "3CCFEM656B 10/100 LAN+56K Modem CardBus" }] } , Device { vendor_id : 4279u16 , id : 25955u16 , name : "3cCFEM656B 10/100 LAN+56K Modem CardBus" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 25963u16 , name : "3CCFEM656 10/100 LAN+56K Modem CardBus" }] } , Device { vendor_id : 4279u16 , id : 25956u16 , name : "3cXFEM656C 10/100 LAN+Winmodem CardBus [Tornado]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 30278u16 , name : "3cSOHO100-TX Hurricane" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 30576u16 , name : "3CRWE777 PCI Wireless Adapter [Airconnect]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 31040u16 , name : "3c803 FDDILink UTP Controller" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 31104u16 , name : "3c804 FDDILink SAS Controller" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 31120u16 , name : "3c805 FDDILink DAS Controller" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 33003u16 , name : "3c940B 10/100/1000Base-T" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 34833u16 , name : "Token ring" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36864u16 , name : "3c900 10BaseT [Boomerang]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36865u16 , name : "3c900 10Mbps Combo [Boomerang]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36868u16 , name : "3c900B-TPO Etherlink XL [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 36868u16 , name : "3C900B-TPO Etherlink XL TPO 10Mb" }] } , Device { vendor_id : 4279u16 , id : 36869u16 , name : "3c900B-Combo Etherlink XL [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 36869u16 , name : "3C900B-Combo Etherlink XL Combo" }] } , Device { vendor_id : 4279u16 , id : 36870u16 , name : "3c900B-TPC Etherlink XL [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36874u16 , name : "3c900B-FL 10base-FL [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36944u16 , name : "3c905 100BaseTX [Boomerang]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36945u16 , name : "3c905 100BaseT4 [Boomerang]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36948u16 , name : "3C905B-TX Fast Etherlink XL PCI" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 36948u16 , name : "3C905B-TX Fast Etherlink XL PCI" }] } , Device { vendor_id : 4279u16 , id : 36949u16 , name : "3c905B 100BaseTX [Cyclone]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 128u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 129u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 130u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 131u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 132u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 133u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 134u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 135u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 136u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 137u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 144u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 145u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 146u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 147u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 148u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 149u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 150u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 151u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 152u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4136u16 , subdevice : 153u16 , name : "3C905B Fast Etherlink XL 10/100" } , SubSystem { subvendor : 4279u16 , subdevice : 36949u16 , name : "3C905B Fast Etherlink XL 10/100" }] } , Device { vendor_id : 4279u16 , id : 36950u16 , name : "3c905B-T4 Fast EtherLink XL [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36952u16 , name : "3c905B Deluxe Etherlink 10/100/BNC [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 36954u16 , name : "3c905B-FX Fast Etherlink XL FX 100baseFx [Cyclone]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 37376u16 , name : "3c905C-TX/TX-M [Tornado]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 149u16 , name : "3C920 Integrated Fast Ethernet Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 151u16 , name : "3C920 Integrated Fast Ethernet Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 180u16 , name : "OptiPlex GX110" } , SubSystem { subvendor : 4136u16 , subdevice : 216u16 , name : "Precision 530" } , SubSystem { subvendor : 4136u16 , subdevice : 254u16 , name : "Optiplex GX240" } , SubSystem { subvendor : 4136u16 , subdevice : 298u16 , name : "3C920 Integrated Fast Ethernet Controller [Latitude C640]" } , SubSystem { subvendor : 4279u16 , subdevice : 4096u16 , name : "3C905CX-TX/TX-M Fast Etherlink for PC Management NIC" } , SubSystem { subvendor : 4279u16 , subdevice : 28672u16 , name : "10/100 Mini PCI Ethernet Adapter" } , SubSystem { subvendor : 4337u16 , subdevice : 9318u16 , name : "Tiger MPX S2466 (3C920 Integrated Fast Ethernet Controller)" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" }] } , Device { vendor_id : 4279u16 , id : 37377u16 , name : "3C920B-EMB Integrated Fast Ethernet Controller [Tornado]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32939u16 , name : "A7N8X Deluxe onboard 3C920B-EMB Integrated Fast Ethernet Controller" }] } , Device { vendor_id : 4279u16 , id : 37378u16 , name : "3Com 3C920B-EMB-WNM Integrated Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 37392u16 , name : "3C920B-EMB-WNM Integrated Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 37632u16 , name : "3CSOHO100B-TX 910-A01 [tulip]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 38912u16 , name : "3c980-TX Fast Etherlink XL Server Adapter [Cyclone]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 38912u16 , name : "3c980-TX Fast Etherlink XL Server Adapter" }] } , Device { vendor_id : 4279u16 , id : 38917u16 , name : "3c980-C 10/100baseTX NIC [Python-T]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 4609u16 , name : "EtherLink Server 10/100 Dual Port A" } , SubSystem { subvendor : 4279u16 , subdevice : 4610u16 , name : "EtherLink Server 10/100 Dual Port B" } , SubSystem { subvendor : 4279u16 , subdevice : 38917u16 , name : "3c980 10/100baseTX NIC [Python-T]" } , SubSystem { subvendor : 4337u16 , subdevice : 9314u16 , name : "Thunder K7 S2462" }] } , Device { vendor_id : 4279u16 , id : 39168u16 , name : "3C990-TX [Typhoon]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39170u16 , name : "3CR990-TX-95 [Typhoon 56-bit]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39171u16 , name : "3CR990-TX-97 [Typhoon 168-bit]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39172u16 , name : "3C990B-TX-M/3C990BSVR [Typhoon2]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 4096u16 , name : "3CR990B-TX-M [Typhoon2]" } , SubSystem { subvendor : 4279u16 , subdevice : 8192u16 , name : "3CR990BSVR [Typhoon2 Server]" }] } , Device { vendor_id : 4279u16 , id : 39173u16 , name : "3CR990-FX-95/97/95 [Typhon Fiber]" , subsystems : & [SubSystem { subvendor : 4279u16 , subdevice : 4353u16 , name : "3CR990-FX-95 [Typhoon Fiber 56-bit]" } , SubSystem { subvendor : 4279u16 , subdevice : 4354u16 , name : "3CR990-FX-97 [Typhoon Fiber 168-bit]" } , SubSystem { subvendor : 4279u16 , subdevice : 8449u16 , name : "3CR990-FX-95 Server [Typhoon Fiber 56-bit]" } , SubSystem { subvendor : 4279u16 , subdevice : 8450u16 , name : "3CR990-FX-97 Server [Typhoon Fiber 168-bit]" }] } , Device { vendor_id : 4279u16 , id : 39176u16 , name : "3CR990SVR95 [Typhoon Server 56-bit]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39177u16 , name : "3CR990SVR97 [Typhoon Server 168-bit]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39178u16 , name : "3C990SVR [Typhoon Server]" , subsystems : & [] } , Device { vendor_id : 4279u16 , id : 39179u16 , name : "3C990SVR [Typhoon Server]" , subsystems : & [] }] }),
        (5692, Vendor { id : 5692u16 , name : "Smart Link Ltd." , devices : & [Device { vendor_id : 5692u16 , id : 12370u16 , name : "SmartLink SmartPCI562 56K Modem" , subsystems : & [] } , Device { vendor_id : 5692u16 , id : 21577u16 , name : "SmartPCI561 Modem" , subsystems : & [] }] }),
        (5630, Vendor { id : 5630u16 , name : "Kinpo Electronics Inc" , devices : & [] }),
        (14297, Vendor { id : 14297u16 , name : "ITD Firm ltd." , devices : & [Device { vendor_id : 14297u16 , id : 4408u16 , name : "SCHD-PH-8 Phase detector" , subsystems : & [] } , Device { vendor_id : 14297u16 , id : 4416u16 , name : "VR-12-PCI 12-ch Relay Actuator Card" , subsystems : & [] } , Device { vendor_id : 14297u16 , id : 4417u16 , name : "PCI-485(422) Multi-port Serial Board" , subsystems : & [] } , Device { vendor_id : 14297u16 , id : 4418u16 , name : "PCI-CAN2" , subsystems : & [] }] }),
        (8106, Vendor { id : 8106u16 , name : "Hexaflake (Shanghai) Information Technology Co., Ltd." , devices : & [Device { vendor_id : 8106u16 , id : 3088u16 , name : "Compass C10 PF" , subsystems : & [] } , Device { vendor_id : 8106u16 , id : 3089u16 , name : "Compass C10 VF" , subsystems : & [] } , Device { vendor_id : 8106u16 , id : 3200u16 , name : "Compass2 C80 PF" , subsystems : & [] } , Device { vendor_id : 8106u16 , id : 3201u16 , name : "Compass2 C80 VF" , subsystems : & [] }] }),
        (4582, Vendor { id : 4582u16 , name : "Mitsui-Zosen System Research" , devices : & [] }),
        (5124, Vendor { id : 5124u16 , name : "Fundamental Software Inc" , devices : & [] }),
        (4209, Vendor { id : 4209u16 , name : "Mitac" , devices : & [Device { vendor_id : 4209u16 , id : 33120u16 , name : "Mitac 8060B Mobile Platform" , subsystems : & [] }] }),
        (4396, Vendor { id : 4396u16 , name : "Zenith Data Systems" , devices : & [] }),
        (26000, Vendor { id : 26000u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (4161, Vendor { id : 4161u16 , name : "Computrend" , devices : & [] }),
        (61, Vendor { id : 61u16 , name : "Lockheed Martin-Marietta Corp" , devices : & [] }),
        (43532, Vendor { id : 43532u16 , name : "iTuner" , devices : & [] }),
        (5097, Vendor { id : 5097u16 , name : "Intraserver Technology Inc" , devices : & [] }),
        (4706, Vendor { id : 4706u16 , name : "ES Computer Company, Ltd." , devices : & [] }),
        (5191, Vendor { id : 5191u16 , name : "AIM GmBH" , devices : & [] }),
        (5248, Vendor { id : 5248u16 , name : "SCII Telecom" , devices : & [] }),
        (4738, Vendor { id : 4738u16 , name : "Davicom Semiconductor, Inc." , devices : & [Device { vendor_id : 4738u16 , id : 25989u16 , name : "DM562P V90 Modem" , subsystems : & [] } , Device { vendor_id : 4738u16 , id : 36873u16 , name : "DM9009 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4738u16 , id : 37120u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 4738u16 , id : 37122u16 , name : "DM9102 Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 657u16 , subdevice : 33298u16 , name : "DM9102A (DM9102AE, SM9102AF) Ethernet 100/10 MBit" }] } , Device { vendor_id : 4738u16 , id : 37170u16 , name : "Ethernet 100/10 MBit" , subsystems : & [] }] }),
        (5688, Vendor { id : 5688u16 , name : "Standard Microsystems Corp [SMC]" , devices : & [Device { vendor_id : 5688u16 , id : 4352u16 , name : "SMC2602W EZConnect / Addtron AWA-100 / Eumitcom PCI WL11000" , subsystems : & [] }] }),
        (4179, Vendor { id : 4179u16 , name : "Young Micro Systems" , devices : & [] }),
        (4691, Vendor { id : 4691u16 , name : "Guzik Technical Enterprises" , devices : & [] }),
        (5223, Vendor { id : 5223u16 , name : "DIGICOM SPA" , devices : & [] }),
        (4842, Vendor { id : 4842u16 , name : "Zuken" , devices : & [] }),
        (4681, Vendor { id : 4681u16 , name : "Samsung Electronics Co., Ltd." , devices : & [] }),
        (4203, Vendor { id : 4203u16 , name : "Apple Inc." , devices : & [Device { vendor_id : 4203u16 , id : 1u16 , name : "Bandit PowerPC host bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 2u16 , name : "Grand Central I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 3u16 , name : "Control Video" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 4u16 , name : "PlanB Video-In" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 7u16 , name : "O'Hare I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 11u16 , name : "Apple Camera" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 12u16 , name : "DOS on Mac" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 14u16 , name : "Hydra Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 16u16 , name : "Heathrow Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 23u16 , name : "Paddington Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 24u16 , name : "UniNorth FireWire" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 25u16 , name : "KeyLargo USB" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 30u16 , name : "UniNorth Internal PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 31u16 , name : "UniNorth PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 32u16 , name : "UniNorth AGP" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 33u16 , name : "UniNorth GMAC (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 34u16 , name : "KeyLargo Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 36u16 , name : "UniNorth/Pangea GMAC (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 37u16 , name : "KeyLargo/Pangea Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 38u16 , name : "KeyLargo/Pangea USB" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 39u16 , name : "UniNorth/Pangea AGP" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 40u16 , name : "UniNorth/Pangea PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 41u16 , name : "UniNorth/Pangea Internal PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 45u16 , name : "UniNorth 1.5 AGP" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 46u16 , name : "UniNorth 1.5 PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 47u16 , name : "UniNorth 1.5 Internal PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 48u16 , name : "UniNorth/Pangea FireWire" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 49u16 , name : "UniNorth 2 FireWire" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 22545u16 , name : "iBook G4 2004" }] } , Device { vendor_id : 4203u16 , id : 50u16 , name : "UniNorth 2 GMAC (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 51u16 , name : "UniNorth 2 ATA/100" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 52u16 , name : "UniNorth 2 AGP" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 53u16 , name : "UniNorth 2 PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 54u16 , name : "UniNorth 2 Internal PCI" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 59u16 , name : "UniNorth/Intrepid ATA/100" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 62u16 , name : "KeyLargo/Intrepid Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 63u16 , name : "KeyLargo/Intrepid USB" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 4203u16 , id : 64u16 , name : "K2 KeyLargo USB" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 65u16 , name : "K2 KeyLargo Mac/IO" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 66u16 , name : "K2 FireWire" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 67u16 , name : "K2 ATA/100" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 69u16 , name : "K2 HT-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 70u16 , name : "K2 HT-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 71u16 , name : "K2 HT-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 72u16 , name : "K2 HT-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 73u16 , name : "K2 HT-PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 74u16 , name : "CPC945 HT Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 75u16 , name : "U3 AGP" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 76u16 , name : "K2 GMAC (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 79u16 , name : "Shasta Mac I/O" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 80u16 , name : "Shasta IDE" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 81u16 , name : "Shasta (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 82u16 , name : "Shasta Firewire" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 83u16 , name : "Shasta PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 84u16 , name : "Shasta PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 85u16 , name : "Shasta PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 86u16 , name : "U4 PCIe" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 87u16 , name : "U3 HT Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 88u16 , name : "U3L AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 89u16 , name : "U3H AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 91u16 , name : "CPC945 PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 102u16 , name : "Intrepid2 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 103u16 , name : "Intrepid2 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 104u16 , name : "Intrepid2 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 105u16 , name : "Intrepid2 ATA/100" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 106u16 , name : "Intrepid2 Firewire" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 107u16 , name : "Intrepid2 GMAC (Sun GEM)" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 116u16 , name : "U4 HT Bridge" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 4108u16 , name : "Apple Silicon PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 4112u16 , name : "Apple Silicon USB4/Thunderbolt PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 5701u16 , name : "Broadcom NetXtreme BCM5701 Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 6145u16 , name : "T2 Bridge Controller" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 6146u16 , name : "T2 Secure Enclave Processor" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 6147u16 , name : "Apple Audio Device" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 8193u16 , name : "S1X NVMe Controller" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 8194u16 , name : "S3ELab NVMe Controller" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 8195u16 , name : "S3X NVMe Controller" , subsystems : & [] } , Device { vendor_id : 4203u16 , id : 8197u16 , name : "ANS2 NVMe Controller" , subsystems : & [] }] }),
        (5736, Vendor { id : 5736u16 , name : "Actiontec Electronics Inc" , devices : & [Device { vendor_id : 5736u16 , id : 256u16 , name : "Mini-PCI bridge" , subsystems : & [] }] }),
        (5176, Vendor { id : 5176u16 , name : "Atmel-dream" , devices : & [] }),
        (4994, Vendor { id : 4994u16 , name : "Marian - Electronic & Software" , devices : & [Device { vendor_id : 4994u16 , id : 1u16 , name : "ARC88 audio recording card" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 8200u16 , name : "Prodif 96 Pro sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 8264u16 , name : "Prodif Plus sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 8328u16 , name : "Marc 8 Midi sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 8392u16 , name : "Marc A sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 16392u16 , name : "Marc 2 sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 16400u16 , name : "Marc 2 Pro sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 16456u16 , name : "Marc 4 MIDI sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 16520u16 , name : "Marc 4 Digi sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 16968u16 , name : "Marc X sound system" , subsystems : & [] } , Device { vendor_id : 4994u16 , id : 17444u16 , name : "TRACE D4 Sound System" , subsystems : & [] }] }),
        (4256, Vendor { id : 4256u16 , name : "Meidensha Corporation" , devices : & [] }),
        (5201, Vendor { id : 5201u16 , name : "SP3D Chip Design GmBH" , devices : & [] }),
        (4919, Vendor { id : 4919u16 , name : "Third Planet Publishing" , devices : & [] }),
        (5420, Vendor { id : 5420u16 , name : "MACRAIGOR Systems LLC" , devices : & [] }),
        (22612, Vendor { id : 22612u16 , name : "GoTView" , devices : & [] }),
        (5134, Vendor { id : 5134u16 , name : "Goepel Electronic GmbH" , devices : & [] }),
        (4245, Vendor { id : 4245u16 , name : "Silicon Image, Inc." , devices : & [Device { vendor_id : 4245u16 , id : 576u16 , name : "Adaptec AAR-1210SA SATA HostRAID Controller" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 578u16 , name : "AAR-1220SA SATA RAID Controller" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1600u16 , name : "PCI0640" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1603u16 , name : "PCI0643" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1606u16 , name : "PCI0646" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1607u16 , name : "PCI0647" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1608u16 , name : "PCI0648" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32805u16 , name : "CUBX motherboard" }] } , Device { vendor_id : 4245u16 , id : 1609u16 , name : "SiI 0649 Ultra ATA/100 PCI to ATA Host Controller" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 93u16 , name : "Integrated Ultra ATA-100 Dual Channel Controller" } , SubSystem { subvendor : 3601u16 , subdevice : 126u16 , name : "Integrated Ultra ATA-100 IDE RAID Controller" } , SubSystem { subvendor : 4126u16 , subdevice : 1609u16 , name : "AMI MegaRAID IDE 100 Controller" }] } , Device { vendor_id : 4245u16 , id : 1616u16 , name : "PBC0650A" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1648u16 , name : "USB0670" , subsystems : & [SubSystem { subvendor : 4245u16 , subdevice : 1648u16 , name : "USB0670" }] } , Device { vendor_id : 4245u16 , id : 1651u16 , name : "USB0673" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 1664u16 , name : "PCI0680 Ultra ATA-133 Host Controller" , subsystems : & [SubSystem { subvendor : 4245u16 , subdevice : 1664u16 , name : "SiI 0680 ATA/133 Controller" } , SubSystem { subvendor : 4245u16 , subdevice : 13952u16 , name : "Winic W-680 (Silicon Image 680 based)" }] } , Device { vendor_id : 4245u16 , id : 12562u16 , name : "SiI 3112 [SATALink/SATARaid] Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4245u16 , subdevice : 12562u16 , name : "SiI 3112 SATALink Controller" } , SubSystem { subvendor : 4245u16 , subdevice : 24850u16 , name : "SiI 3112 SATARaid Controller" } , SubSystem { subvendor : 36869u16 , subdevice : 592u16 , name : "SATAConnect 1205SA Host Controller" }] } , Device { vendor_id : 4245u16 , id : 12564u16 , name : "SiI 3114 [SATALink/SATARaid] Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33127u16 , name : "A8N-SLI Deluxe/Premium Mainboard" } , SubSystem { subvendor : 4245u16 , subdevice : 12564u16 , name : "SiI 3114 SATALink Controller" } , SubSystem { subvendor : 4245u16 , subdevice : 24852u16 , name : "SiI 3114 SATARaid Controller" }] } , Device { vendor_id : 4245u16 , id : 12580u16 , name : "SiI 3124 PCI-X Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4245u16 , subdevice : 12580u16 , name : "SiI 3124 PCI-X Serial ATA Controller" }] } , Device { vendor_id : 4245u16 , id : 12594u16 , name : "SiI 3132 Serial ATA Raid II Controller" , subsystems : & [] } , Device { vendor_id : 4245u16 , id : 13586u16 , name : "SiI 3512 [SATALink/SATARaid] Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4245u16 , subdevice : 13586u16 , name : "SiI 3512 SATALink Controller" } , SubSystem { subvendor : 4245u16 , subdevice : 25874u16 , name : "SiI 3512 SATARaid Controller" }] } , Device { vendor_id : 4245u16 , id : 13617u16 , name : "SiI 3531 [SATALink/SATARaid] Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 6080u16 , subdevice : 16515u16 , name : "Medion WIM 2210 Notebook PC [MD96850]" }] }] }),
        (5218, Vendor { id : 5218u16 , name : "Micro-Star International Co., Ltd. [MSI]" , devices : & [Device { vendor_id : 5218u16 , id : 13443u16 , name : "MSI USB 3.0 (VIA VL80x-based xHCI USB Controller)" , subsystems : & [] }] }),
        (6618, Vendor { id : 6618u16 , name : "ZOTAC International (MCO) Ltd." , devices : & [] }),
        (5641, Vendor { id : 5641u16 , name : "Scimetric Instruments Inc" , devices : & [] }),
        (5623, Vendor { id : 5623u16 , name : "Banctec" , devices : & [] }),
        (4359, Vendor { id : 4359u16 , name : "Stratus Computers" , devices : & [Device { vendor_id : 4359u16 , id : 1398u16 , name : "VIA VT82C570MV [Apollo] (Wrong vendor ID!)" , subsystems : & [] }] }),
        (5957, Vendor { id : 5957u16 , name : "ViXS Systems, Inc." , devices : & [Device { vendor_id : 5957u16 , id : 8224u16 , name : "XCode II Series" , subsystems : & [] } , Device { vendor_id : 5957u16 , id : 8448u16 , name : "XCode 2100 Series" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 18633u16 , name : "My Cinema PE6200 Analoog" }] }] }),
        (5428, Vendor { id : 5428u16 , name : "ROAD Corp" , devices : & [] }),
        (112, Vendor { id : 112u16 , name : "Hauppauge computer works Inc." , devices : & [] }),
        (6860, Vendor { id : 6860u16 , name : "Point of View BV" , devices : & [] }),
        (5807, Vendor { id : 5807u16 , name : "SparkLAN Communications, Inc." , devices : & [] }),
        (6082, Vendor { id : 6082u16 , name : "Newisys, Inc." , devices : & [] }),
        (4231, Vendor { id : 4231u16 , name : "Cache Computer" , devices : & [] }),
        (38992, Vendor { id : 38992u16 , name : "3Com (wrong ID)" , devices : & [] }),
        (4160, Vendor { id : 4160u16 , name : "Accelgraphics Inc." , devices : & [] }),
        (4463, Vendor { id : 4463u16 , name : "Workstation Technology" , devices : & [] }),
        (5236, Vendor { id : 5236u16 , name : "Doug Carson & Associates" , devices : & [] }),
        (6638, Vendor { id : 6638u16 , name : "Netronome Systems, Inc." , devices : & [] }),
        (5543, Vendor { id : 5543u16 , name : "SSE Telecom Inc" , devices : & [] }),
        (7521, Vendor { id : 7521u16 , name : "Technobox, Inc." , devices : & [] }),
        (5537, Vendor { id : 5537u16 , name : "Geocast Network Systems" , devices : & [] }),
        (4139, Vendor { id : 4139u16 , name : "Matrox Electronics Systems Ltd." , devices : & [Device { vendor_id : 4139u16 , id : 16u16 , name : "MGA-I [Impression?]" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 256u16 , name : "MGA 1064SG [Mystique]" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 1304u16 , name : "MGA-II [Athena]" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 1305u16 , name : "MGA 2064W [Millennium]" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 1306u16 , name : "MGA 1064SG [Mystique]" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 256u16 , name : "MGA-1064SG Mystique" } , SubSystem { subvendor : 4139u16 , subdevice : 1306u16 , name : "MGA-1164SG Mystique 220" } , SubSystem { subvendor : 4139u16 , subdevice : 4352u16 , name : "MGA-1084SG Mystique" } , SubSystem { subvendor : 4139u16 , subdevice : 4608u16 , name : "MGA-1084SG Mystique" } , SubSystem { subvendor : 4352u16 , subdevice : 4139u16 , name : "MGA-1084SG Mystique" } , SubSystem { subvendor : 4362u16 , subdevice : 24u16 , name : "Scenic Pro C5 (D1025)" }] } , Device { vendor_id : 4139u16 , id : 1307u16 , name : "MGA 2164W [Millennium II]" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 1307u16 , name : "MGA-2164W Millennium II" } , SubSystem { subvendor : 4139u16 , subdevice : 4352u16 , name : "MGA-2164W Millennium II" } , SubSystem { subvendor : 4139u16 , subdevice : 4608u16 , name : "MGA-2164W Millennium II" } , SubSystem { subvendor : 4139u16 , subdevice : 8448u16 , name : "MGA-2164W Millennium II" }] } , Device { vendor_id : 4139u16 , id : 1310u16 , name : "MGA 1064SG [Mystique] AGP" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 1311u16 , name : "MGA 2164W [Millennium II] AGP" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 8448u16 , name : "MGA-2164WA [Millennium II A]" }] } , Device { vendor_id : 4139u16 , id : 1312u16 , name : "MGA G200" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 56258u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56264u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56290u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56296u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 65283u16 , name : "Millennium G200 SD" } , SubSystem { subvendor : 4139u16 , subdevice : 65284u16 , name : "Marvel G200" }] } , Device { vendor_id : 4139u16 , id : 1313u16 , name : "MGA G200 AGP" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 65283u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 18665u16 , name : "Mystique G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 18680u16 , name : "Millennium G200 SD AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 19040u16 , name : "Millennium G200 LE AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 19044u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 51516u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 51632u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 51644u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 51808u16 , name : "Millennium G250 LE AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 51820u16 , name : "Millennium G250 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 56252u16 , name : "Millennium G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 56258u16 , name : "Millennium G200 MMS (Dual G200)" } , SubSystem { subvendor : 4139u16 , subdevice : 56259u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56264u16 , name : "Millennium G200 MMS (Dual G200)" } , SubSystem { subvendor : 4139u16 , subdevice : 56274u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56275u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56276u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56277u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56280u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56281u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56290u16 , name : "Millennium G200 MMS (Quad G200)" } , SubSystem { subvendor : 4139u16 , subdevice : 56291u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56296u16 , name : "Millennium G200 MMS (Quad G200)" } , SubSystem { subvendor : 4139u16 , subdevice : 56306u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56307u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56308u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56309u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56312u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 56313u16 , name : "G200 Multi-Monitor" } , SubSystem { subvendor : 4139u16 , subdevice : 63494u16 , name : "Mystique G200 Video AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65280u16 , name : "MGA-G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65282u16 , name : "Mystique G200 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65283u16 , name : "Millennium G200A AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65284u16 , name : "Marvel G200 AGP" } , SubSystem { subvendor : 4362u16 , subdevice : 50u16 , name : "MGA-G200 AGP" }] } , Device { vendor_id : 4139u16 , id : 1314u16 , name : "MGA G200e [Pilot] ServerEngines (SEP1)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12794u16 , name : "ProLiant DL140 G3" }] } , Device { vendor_id : 4139u16 , id : 1317u16 , name : "MGA G400/G450" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45423u16 , name : "MGA-G400 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 808u16 , name : "Millennium G400 16Mb SDRAM" } , SubSystem { subvendor : 4139u16 , subdevice : 824u16 , name : "Millennium G400 16Mb SDRAM" } , SubSystem { subvendor : 4139u16 , subdevice : 888u16 , name : "Millennium G400 32Mb SDRAM" } , SubSystem { subvendor : 4139u16 , subdevice : 1345u16 , name : "Millennium G450 Dual Head" } , SubSystem { subvendor : 4139u16 , subdevice : 1346u16 , name : "Millennium G450 Dual Head LX" } , SubSystem { subvendor : 4139u16 , subdevice : 1347u16 , name : "Millennium G450 Single Head LX" } , SubSystem { subvendor : 4139u16 , subdevice : 1601u16 , name : "Millennium G450 32Mb SDRAM Dual Head" } , SubSystem { subvendor : 4139u16 , subdevice : 1602u16 , name : "Millennium G450 32Mb SDRAM Dual Head LX" } , SubSystem { subvendor : 4139u16 , subdevice : 1603u16 , name : "Millennium G450 32Mb SDRAM Single Head LX" } , SubSystem { subvendor : 4139u16 , subdevice : 1984u16 , name : "Millennium G450 Dual Head LE" } , SubSystem { subvendor : 4139u16 , subdevice : 1985u16 , name : "Millennium G450 SDR Dual Head LE" } , SubSystem { subvendor : 4139u16 , subdevice : 3393u16 , name : "Millennium G450 Dual Head PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 3394u16 , name : "Millennium G450 Dual Head LX PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 3395u16 , name : "Millennium G450 32Mb Dual Head PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 3584u16 , name : "Marvel G450 eTV" } , SubSystem { subvendor : 4139u16 , subdevice : 3585u16 , name : "Marvel G450 eTV" } , SubSystem { subvendor : 4139u16 , subdevice : 3586u16 , name : "Marvel G450 eTV" } , SubSystem { subvendor : 4139u16 , subdevice : 3587u16 , name : "Marvel G450 eTV" } , SubSystem { subvendor : 4139u16 , subdevice : 3968u16 , name : "Millennium G450 Low Profile" } , SubSystem { subvendor : 4139u16 , subdevice : 3969u16 , name : "Millennium G450 Low Profile" } , SubSystem { subvendor : 4139u16 , subdevice : 3970u16 , name : "Millennium G450 Low Profile DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 3971u16 , name : "Millennium G450 Low Profile DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 6616u16 , name : "Millennium G400 16Mb SGRAM" } , SubSystem { subvendor : 4139u16 , subdevice : 6648u16 , name : "Millennium G400 32Mb SGRAM" } , SubSystem { subvendor : 4139u16 , subdevice : 8537u16 , name : "Millennium G400 Dual Head 16Mb" } , SubSystem { subvendor : 4139u16 , subdevice : 8569u16 , name : "Millennium G400 MAX/Dual Head 32Mb" } , SubSystem { subvendor : 4139u16 , subdevice : 8573u16 , name : "Millennium G400 Dual Head Max" } , SubSystem { subvendor : 4139u16 , subdevice : 9152u16 , name : "Millennium G450" } , SubSystem { subvendor : 4139u16 , subdevice : 9153u16 , name : "Millennium G450" } , SubSystem { subvendor : 4139u16 , subdevice : 9154u16 , name : "Millennium G450 DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 9155u16 , name : "Millennium G450 DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 12120u16 , name : "Millennium G400" } , SubSystem { subvendor : 4139u16 , subdevice : 12152u16 , name : "Millennium G400" } , SubSystem { subvendor : 4139u16 , subdevice : 13971u16 , name : "Marvel G400 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 24016u16 , name : "4Sight II" } , SubSystem { subvendor : 4139u16 , subdevice : 24400u16 , name : "4Sight II" } , SubSystem { subvendor : 4139u16 , subdevice : 24401u16 , name : "4Sight II" } , SubSystem { subvendor : 4139u16 , subdevice : 24402u16 , name : "4Sight II" } , SubSystem { subvendor : 4139u16 , subdevice : 36880u16 , name : "Millennium G400 Dual Head" } , SubSystem { subvendor : 5208u16 , subdevice : 1024u16 , name : "GA-G400" } , SubSystem { subvendor : 5893u16 , subdevice : 1u16 , name : "Millennium G450 32MB SGRAM" } , SubSystem { subvendor : 5893u16 , subdevice : 2u16 , name : "Millennium G450 16MB SGRAM" } , SubSystem { subvendor : 5893u16 , subdevice : 3u16 , name : "Millennium G450 32MB" } , SubSystem { subvendor : 5893u16 , subdevice : 4u16 , name : "Millennium G450 16MB" }] } , Device { vendor_id : 4139u16 , id : 1319u16 , name : "Parhelia" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 2112u16 , name : "Parhelia 128Mb" } , SubSystem { subvendor : 4139u16 , subdevice : 2128u16 , name : "Parhelia 256MB" } , SubSystem { subvendor : 4139u16 , subdevice : 2160u16 , name : "MED2mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 2176u16 , name : "P-256 Edge Overlap Controller" }] } , Device { vendor_id : 4139u16 , id : 1320u16 , name : "Parhelia" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 4128u16 , name : "Parhelia 128MB" } , SubSystem { subvendor : 4139u16 , subdevice : 4144u16 , name : "Parhelia 256 MB Dual DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 4160u16 , name : "MED2mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 4176u16 , name : "Sono S20" } , SubSystem { subvendor : 4139u16 , subdevice : 4192u16 , name : "PJ-30L" } , SubSystem { subvendor : 4139u16 , subdevice : 4208u16 , name : "PJ-40L" } , SubSystem { subvendor : 4139u16 , subdevice : 5153u16 , name : "MED5mp" } , SubSystem { subvendor : 4139u16 , subdevice : 5169u16 , name : "MED3mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5201u16 , name : "MED5mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5265u16 , name : "MED2mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5297u16 , name : "MED3mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5313u16 , name : "MED5mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5345u16 , name : "Parhelia PCI 256MB" } , SubSystem { subvendor : 4139u16 , subdevice : 5361u16 , name : "Parhelia Precision SGT" } , SubSystem { subvendor : 4139u16 , subdevice : 5377u16 , name : "ATC-4MP" } , SubSystem { subvendor : 4139u16 , subdevice : 5393u16 , name : "ATC-4MP" } , SubSystem { subvendor : 4139u16 , subdevice : 5409u16 , name : "TheatreVUE T30" } , SubSystem { subvendor : 4139u16 , subdevice : 5425u16 , name : "TheatreVUE T20" } , SubSystem { subvendor : 4139u16 , subdevice : 5441u16 , name : "MED2mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5457u16 , name : "MED3mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5473u16 , name : "MED5mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 5489u16 , name : "Parhelia DL256 PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 5521u16 , name : "Parhelia Precision SDT" } , SubSystem { subvendor : 4139u16 , subdevice : 5537u16 , name : "MED4mp-DVI" } , SubSystem { subvendor : 4139u16 , subdevice : 8209u16 , name : "Parhelia HR256" } , SubSystem { subvendor : 4139u16 , subdevice : 8225u16 , name : "QID Pro" } , SubSystem { subvendor : 4139u16 , subdevice : 8289u16 , name : "PJ-40LP" } , SubSystem { subvendor : 4139u16 , subdevice : 8321u16 , name : "EWS Quad" } , SubSystem { subvendor : 4139u16 , subdevice : 9233u16 , name : "PPX-OUT8" } , SubSystem { subvendor : 4139u16 , subdevice : 9249u16 , name : "VPX-OUT8" } , SubSystem { subvendor : 4139u16 , subdevice : 9281u16 , name : "PPX-OUT4" } , SubSystem { subvendor : 4139u16 , subdevice : 9297u16 , name : "VPX-OUT4" } , SubSystem { subvendor : 4139u16 , subdevice : 9361u16 , name : "LPX-OUT4" }] } , Device { vendor_id : 4139u16 , id : 1328u16 , name : "MGA G200EV" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 1330u16 , name : "MGA G200eW WPCM450" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 565u16 , name : "PowerEdge R710 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 566u16 , name : "PowerEdge R610 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 567u16 , name : "PowerEdge T610 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 647u16 , name : "PowerEdge M610 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 652u16 , name : "PowerEdge R410 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 653u16 , name : "PowerEdge T410 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 668u16 , name : "PowerEdge M710 MGA G200eW WPCM450" } , SubSystem { subvendor : 4136u16 , subdevice : 676u16 , name : "PowerEdge T310 MGA G200eW WPCM450" } , SubSystem { subvendor : 5593u16 , subdevice : 1541u16 , name : "X8SIL" } , SubSystem { subvendor : 5593u16 , subdevice : 1572u16 , name : "X9SCM-F Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 1643u16 , name : "X9SRL-F" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" }] } , Device { vendor_id : 4139u16 , id : 1331u16 , name : "MGA G200EH" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4139u16 , id : 1332u16 , name : "G200eR2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1271u16 , name : "PowerEdge R320 server" }] } , Device { vendor_id : 4139u16 , id : 1334u16 , name : "Integrated Matrox G200eW3 Graphics Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2667u16 , name : "PowerEdge R760" }] } , Device { vendor_id : 4139u16 , id : 1336u16 , name : "MGA G200eH3" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 228u16 , name : "iLO5 VGA" }] } , Device { vendor_id : 4139u16 , id : 1344u16 , name : "M91XX" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 8320u16 , name : "M9140 LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8384u16 , name : "Xenia" } , SubSystem { subvendor : 4139u16 , subdevice : 8385u16 , name : "Xenia Pro" } , SubSystem { subvendor : 4139u16 , subdevice : 8448u16 , name : "M9120 PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8512u16 , name : "M9125 PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8576u16 , name : "M9120 Plus LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8640u16 , name : "M9120 Plus LP PCIe x1" } , SubSystem { subvendor : 4139u16 , subdevice : 8704u16 , name : "VDA1164 Output Board" } , SubSystem { subvendor : 4139u16 , subdevice : 8768u16 , name : "M9148 LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8769u16 , name : "M9138 LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8832u16 , name : "M9188 ATX PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 8896u16 , name : "M9128 LP PCIe x16" }] } , Device { vendor_id : 4139u16 , id : 1360u16 , name : "SV2" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 192u16 , name : "MURA-IPX-I4EF" } , SubSystem { subvendor : 4139u16 , subdevice : 193u16 , name : "MURA-IPX-I4DF" } , SubSystem { subvendor : 4139u16 , subdevice : 195u16 , name : "MURA-IPX-I4DHF" } , SubSystem { subvendor : 4139u16 , subdevice : 197u16 , name : "MURA-IPX-I4EHF" }] } , Device { vendor_id : 4139u16 , id : 3344u16 , name : "MGA Ultima/Impression" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 4096u16 , name : "MGA G100 [Productiva]" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 65281u16 , name : "Productiva G100" } , SubSystem { subvendor : 4139u16 , subdevice : 65285u16 , name : "Productiva G100 Multi-Monitor" }] } , Device { vendor_id : 4139u16 , id : 4097u16 , name : "MGA G100 [Productiva] AGP" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 4097u16 , name : "MGA-G100 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65280u16 , name : "MGA-G100 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65281u16 , name : "MGA-G100 Productiva AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65283u16 , name : "Millennium G100 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65284u16 , name : "MGA-G100 AGP" } , SubSystem { subvendor : 4139u16 , subdevice : 65285u16 , name : "MGA-G100 Productiva AGP Multi-Monitor" } , SubSystem { subvendor : 4362u16 , subdevice : 30u16 , name : "MGA-G100 AGP" }] } , Device { vendor_id : 4139u16 , id : 8199u16 , name : "MGA Mistral" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 9511u16 , name : "Millennium G550" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 3906u16 , name : "Matrox G550 Low Profile PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 3971u16 , name : "Millennium G550" } , SubSystem { subvendor : 4139u16 , subdevice : 3972u16 , name : "Millennium G550 Dual Head DDR 32Mb" } , SubSystem { subvendor : 4139u16 , subdevice : 7745u16 , name : "Millennium G550" } , SubSystem { subvendor : 4139u16 , subdevice : 8896u16 , name : "G550 PCIe" } , SubSystem { subvendor : 4139u16 , subdevice : 8960u16 , name : "Millennium G550 LP PCIE" }] } , Device { vendor_id : 4139u16 , id : 9527u16 , name : "Millennium P650/P750" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 6176u16 , name : "Millennium P750 64MB" } , SubSystem { subvendor : 4139u16 , subdevice : 6192u16 , name : "Millennium P650 64MB" } , SubSystem { subvendor : 4139u16 , subdevice : 6224u16 , name : "RAD2mp" } , SubSystem { subvendor : 4139u16 , subdevice : 6240u16 , name : "RAD3mp" } , SubSystem { subvendor : 4139u16 , subdevice : 6272u16 , name : "Sono S10" } , SubSystem { subvendor : 4139u16 , subdevice : 7184u16 , name : "QID 128MB" } , SubSystem { subvendor : 4139u16 , subdevice : 10257u16 , name : "Millennium P650 Low-profile PCI 64MB" } , SubSystem { subvendor : 4139u16 , subdevice : 10273u16 , name : "Millennium P650 Low-profile PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 10305u16 , name : "RAD PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 10321u16 , name : "Spectrum PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 10353u16 , name : "EpicA TC2" } , SubSystem { subvendor : 4139u16 , subdevice : 11281u16 , name : "QID Low-profile PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 11297u16 , name : "QID LP PCI LW" } , SubSystem { subvendor : 4139u16 , subdevice : 11313u16 , name : "QID LP PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 11329u16 , name : "EpicA TC4" } , SubSystem { subvendor : 4139u16 , subdevice : 12289u16 , name : "Extio F1400" } , SubSystem { subvendor : 4139u16 , subdevice : 12305u16 , name : "Extio F1220" } , SubSystem { subvendor : 4139u16 , subdevice : 12353u16 , name : "RG-200DL" } , SubSystem { subvendor : 4139u16 , subdevice : 12369u16 , name : "RG-400SL" } , SubSystem { subvendor : 4139u16 , subdevice : 12385u16 , name : "Extio F1420" } , SubSystem { subvendor : 4139u16 , subdevice : 12417u16 , name : "Extio F1240" }] } , Device { vendor_id : 4139u16 , id : 9528u16 , name : "Millennium P650 PCIe" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 2119u16 , name : "RAD PCIe" } , SubSystem { subvendor : 4139u16 , subdevice : 2247u16 , name : "Millennium P650 PCIe 128MB" } , SubSystem { subvendor : 4139u16 , subdevice : 2311u16 , name : "Millennium P650 PCIe 64MB" } , SubSystem { subvendor : 4139u16 , subdevice : 2375u16 , name : "Parhelia APVe" } , SubSystem { subvendor : 4139u16 , subdevice : 2439u16 , name : "ATC PCIe 4MP" } , SubSystem { subvendor : 4139u16 , subdevice : 4167u16 , name : "Millennium P650 LP PCIe 128MB" } , SubSystem { subvendor : 4139u16 , subdevice : 4231u16 , name : "Millennium P650 LP PCIe 64MB" } , SubSystem { subvendor : 4139u16 , subdevice : 6145u16 , name : "Millennium P650 PCIe x1" } , SubSystem { subvendor : 4139u16 , subdevice : 9528u16 , name : "Parhelia APVe" } , SubSystem { subvendor : 4139u16 , subdevice : 12295u16 , name : "QID Low-profile PCIe" } , SubSystem { subvendor : 4139u16 , subdevice : 12423u16 , name : "Aurora VX3mp" } , SubSystem { subvendor : 4139u16 , subdevice : 12487u16 , name : "QID LP PCIe" }] } , Device { vendor_id : 4139u16 , id : 9529u16 , name : "Millennium P690" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 64u16 , name : "Millennium P690 PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 66u16 , name : "ONYX" } , SubSystem { subvendor : 4139u16 , subdevice : 67u16 , name : "SPECTRA" } , SubSystem { subvendor : 4139u16 , subdevice : 128u16 , name : "Millennium P690 Plus LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 129u16 , name : "Millennium P690 LP PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 130u16 , name : "RAD LPX PCIe x16" } , SubSystem { subvendor : 4139u16 , subdevice : 192u16 , name : "Millennium P690 Plus LP PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 194u16 , name : "Millennium P690 LP PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 195u16 , name : "RAD LPX PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 257u16 , name : "Millennium P690 PCI" } , SubSystem { subvendor : 4139u16 , subdevice : 320u16 , name : "Millennium P690 LP PCIe x1" } , SubSystem { subvendor : 4139u16 , subdevice : 384u16 , name : "Display Wall IP Decode 128 MB" }] } , Device { vendor_id : 4139u16 , id : 16740u16 , name : "Morphis QxT frame grabber" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 17332u16 , name : "Morphis Qxt encoding engine" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 17680u16 , name : "Morphis COM port" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 17718u16 , name : "VIA Framegrabber" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 18054u16 , name : "Concord GX (customized Intel 82541)" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 18267u16 , name : "Solios eCL/XCL-B frame grabber" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 18269u16 , name : "Vio frame grabber family" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 19344u16 , name : "Vio Duo frame grabber (single channel)" } , SubSystem { subvendor : 4139u16 , subdevice : 19345u16 , name : "Vio Duo frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19346u16 , name : "Vio Analog frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19347u16 , name : "Vio SDI Frame Grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19348u16 , name : "Vio DVI-A frame grabber" }] } , Device { vendor_id : 4139u16 , id : 18271u16 , name : "Solios (single-Full) CL frame grabber" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 18271u16 , name : "Solios eCL/XCL-F frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19807u16 , name : "Solios eV-CL (single-Full) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 20063u16 , name : "Solios eM-CL (single-Full) frame grabber" }] } , Device { vendor_id : 4139u16 , id : 18337u16 , name : "Solios eA/XA frame grabber" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 19424u16 , name : "Solios eA/XA (single) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19425u16 , name : "Solios eA/XA (dual) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19426u16 , name : "Solios eA/XA (quad) frame grabber" }] } , Device { vendor_id : 4139u16 , id : 18338u16 , name : "Solios COM port" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 18369u16 , name : "Solios (dual-Base/single-Medium) CL frame grabber" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 0u16 , name : "Solios frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19328u16 , name : "Solios eCL/XCL (single-Medium) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19329u16 , name : "Solios eCL/XCL (dual-Base) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19840u16 , name : "Solios eV-CL (single-Medium) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 19841u16 , name : "Solios eV-CL (dual-Base) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 20096u16 , name : "Solios eM-CL (single-Medium) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 20097u16 , name : "Solios eM-CL (dual-Base) frame grabber" }] } , Device { vendor_id : 4139u16 , id : 18370u16 , name : "Solios COM port" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 18761u16 , name : "Radient frame grabber family" , subsystems : & [SubSystem { subvendor : 4139u16 , subdevice : 16u16 , name : "Radient eCL (Single-full) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 17u16 , name : "Radient eCLV (Single-full) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 32u16 , name : "Radient eCL (Dual-base) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 48u16 , name : "Radient eCL (Dual-full) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 64u16 , name : "Radient eCL (Quad-base) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 80u16 , name : "Radient eCL (Golden) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 4112u16 , name : "Radient eV-CXP (quad CXP-6) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 4117u16 , name : "Radient eV-CXP (dual CXP-6) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 4128u16 , name : "Radient eV-CXP (quad CXP-3) frame grabber" } , SubSystem { subvendor : 4139u16 , subdevice : 4176u16 , name : "Radient eV-CXP (Golden) frame grabber" }] } , Device { vendor_id : 4139u16 , id : 19676u16 , name : "Morphis JPEG2000 accelerator" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 20308u16 , name : "Morphis (e)Quad frame grabber" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 20421u16 , name : "Morphis (e)Dual frame grabber" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 24080u16 , name : "Morphis aux I/O" , subsystems : & [] } , Device { vendor_id : 4139u16 , id : 25971u16 , name : "Shark 10/100 Multiport SwitchNIC" , subsystems : & [] }] }),
        (5286, Vendor { id : 5286u16 , name : "INOVA Computers GmBH & Co KG" , devices : & [] }),
        (5422, Vendor { id : 5422u16 , name : "MELEC Inc" , devices : & [] }),
        (6164, Vendor { id : 6164u16 , name : "Ralink corp." , devices : & [Device { vendor_id : 6164u16 , id : 257u16 , name : "Wireless PCI Adapter RT2400 / RT2460" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 295u16 , name : "WiFi-b add-on Card" } , SubSystem { subvendor : 4977u16 , subdevice : 16u16 , name : "Minitar MNW2BPCI Wireless PCI Card" } , SubSystem { subvendor : 5218u16 , subdevice : 26664u16 , name : "PC11B2 (MS-6828) Wireless 11b PCI Card" }] } , Device { vendor_id : 6164u16 , id : 512u16 , name : "RT2500 802.11g PCI [PC54G2]" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 513u16 , name : "RT2500 Wireless 802.11bg" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 4879u16 , name : "WL-130g" } , SubSystem { subvendor : 4486u16 , subdevice : 15360u16 , name : "DWL-G650X Wireless 11g CardBus Adapter" } , SubSystem { subvendor : 4977u16 , subdevice : 30u16 , name : "CWC-854 Wireless-G CardBus Adapter" } , SubSystem { subvendor : 4977u16 , subdevice : 31u16 , name : "CWM-854 Wireless-G Mini PCI Adapter" } , SubSystem { subvendor : 4977u16 , subdevice : 32u16 , name : "CWP-854 Wireless-G PCI Adapter" } , SubSystem { subvendor : 5208u16 , subdevice : 58241u16 , name : "GN-WMKG 802.11b/g Wireless CardBus Adapter" } , SubSystem { subvendor : 5208u16 , subdevice : 59697u16 , name : "GN-WIKG 802.11b/g mini-PCI Adapter" } , SubSystem { subvendor : 5218u16 , subdevice : 26675u16 , name : "Unknown 802.11g mini-PCI Adapter" } , SubSystem { subvendor : 5218u16 , subdevice : 26677u16 , name : "Wireless 11G CardBus CB54G2" } , SubSystem { subvendor : 5943u16 , subdevice : 50u16 , name : "WMP54G v4.0 PCI Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 28682u16 , name : "F5D7000 v2000/v3000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28698u16 , name : "F5D7010 v2000/v3000 Wireless G Notebook Card" } , SubSystem { subvendor : 6164u16 , subdevice : 9568u16 , name : "RT2500 Wireless 802.11bg" } , SubSystem { subvendor : 6189u16 , subdevice : 36979u16 , name : "WL-115 Wireless Network PCI Adapter" } , SubSystem { subvendor : 6239u16 , subdevice : 8864u16 , name : "CN-WF513 Wireless Cardbus Adapter" } , SubSystem { subvendor : 6379u16 , subdevice : 21266u16 , name : "WL531P IEEE 802.11g PCI Card-EU" } , SubSystem { subvendor : 6472u16 , subdevice : 15360u16 , name : "C54RC v1 Wireless 11g CardBus Adapter" } , SubSystem { subvendor : 6472u16 , subdevice : 15361u16 , name : "C54Ri v1 Wireless 11g PCI Adapter" }] } , Device { vendor_id : 6164u16 , id : 768u16 , name : "Wireless Adapter Canyon CN-WF511" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 769u16 , name : "RT2561/RT61 802.11g PCI" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 15368u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.E1)" } , SubSystem { subvendor : 4486u16 , subdevice : 15369u16 , name : "DWL-G510 Rev C" } , SubSystem { subvendor : 4697u16 , subdevice : 49443u16 , name : "CG-WLCB54GPX" } , SubSystem { subvendor : 5073u16 , subdevice : 44003u16 , name : "miniPCI Pluscom 802.11 a/b/g" } , SubSystem { subvendor : 5208u16 , subdevice : 59699u16 , name : "GN-WI01GS" } , SubSystem { subvendor : 5208u16 , subdevice : 59700u16 , name : "GN-WP01GS" } , SubSystem { subvendor : 5218u16 , subdevice : 47155u16 , name : "MP54G5 (MS-6833B)" } , SubSystem { subvendor : 5218u16 , subdevice : 47156u16 , name : "PC60G (MS-6834B) Wireless 11g Turbo G PCI Card" } , SubSystem { subvendor : 5943u16 , subdevice : 85u16 , name : "WMP54G v4.1" } , SubSystem { subvendor : 6041u16 , subdevice : 28686u16 , name : "F5D7000 v6000 Wireless G Desktop Card" } , SubSystem { subvendor : 6041u16 , subdevice : 28702u16 , name : "F5D7010 v6000 Wireless G Notebook Card" } , SubSystem { subvendor : 6137u16 , subdevice : 18u16 , name : "AWLC3026T 802.11g Wireless CardBus Adapter" } , SubSystem { subvendor : 6164u16 , subdevice : 9569u16 , name : "EW-7108PCg/EW-7128g" } , SubSystem { subvendor : 6189u16 , subdevice : 37034u16 , name : "WL-170 Wireless Network Cardbus Card" }] } , Device { vendor_id : 6164u16 , id : 770u16 , name : "RT2561/RT61 rev B 802.11g" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 14961u16 , name : "DWA-510 Wireless G Desktop Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 15368u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.E2)" } , SubSystem { subvendor : 4486u16 , subdevice : 15369u16 , name : "AirPlus G DWL-G510 Wireless Network Adapter (Rev.C)" } , SubSystem { subvendor : 5218u16 , subdevice : 47156u16 , name : "PC54G3 Wireless 11g PCI Card" } , SubSystem { subvendor : 6472u16 , subdevice : 15395u16 , name : "C54RC v2 Wireless 11g CardBus Adapter" } , SubSystem { subvendor : 6472u16 , subdevice : 15396u16 , name : "C54Ri v2 Wireless 11g PCI Adapter" }] } , Device { vendor_id : 6164u16 , id : 1025u16 , name : "RT2600 802.11 MIMO" , subsystems : & [SubSystem { subvendor : 5943u16 , subdevice : 82u16 , name : "WPC54GR v1 802.11g Wireless-G Notebook Adapter with RangeBooster" } , SubSystem { subvendor : 6137u16 , subdevice : 17u16 , name : "WPCR-137G 802.11bg Wireless CardBus Adapter" } , SubSystem { subvendor : 6137u16 , subdevice : 22u16 , name : "WPIR-119GH 802.11bg Wireless Desktop Adapter" }] } , Device { vendor_id : 6164u16 , id : 1537u16 , name : "RT2800 802.11n PCI" , subsystems : & [SubSystem { subvendor : 6041u16 , subdevice : 32796u16 , name : "F5D8011 v3 802.11n N1 Wireless Notebook Card" } , SubSystem { subvendor : 6270u16 , subdevice : 13330u16 , name : "NWD-310N 802.11n Wireless PCI Adapter" }] } , Device { vendor_id : 6164u16 , id : 1665u16 , name : "RT2890 Wireless 802.11n PCIe" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 59705u16 , name : "GN-WS30N-RH 802.11bgn Mini PCIe Card" } , SubSystem { subvendor : 6041u16 , subdevice : 32883u16 , name : "F5D8073 v1 N Wireless ExpressCard Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 32892u16 , name : "F5D8071 v3 N1 Wireless ExpressCard" }] } , Device { vendor_id : 6164u16 , id : 1793u16 , name : "RT2760 Wireless 802.11n 1T/2R" , subsystems : & [SubSystem { subvendor : 5943u16 , subdevice : 116u16 , name : "WMP110 v2 802.11n RangePlus Wireless PCI Adapter" }] } , Device { vendor_id : 6164u16 , id : 1921u16 , name : "RT2790 Wireless 802.11n 1T/2R PCIe" , subsystems : & [SubSystem { subvendor : 4525u16 , subdevice : 30208u16 , name : "HP WN7600R" } , SubSystem { subvendor : 6041u16 , subdevice : 33148u16 , name : "F5D8073 v3 N Wireless ExpressCard Adapter" } , SubSystem { subvendor : 6164u16 , subdevice : 10128u16 , name : "RT2790 Wireless 802.11n 1T/2R PCIe" } , SubSystem { subvendor : 6518u16 , subdevice : 9728u16 , name : "TEW-642EC Wireless N ExpressCard" }] } , Device { vendor_id : 6164u16 , id : 12384u16 , name : "RT3060 Wireless 802.11n 1T/1R" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 15364u16 , name : "DWA-525 Wireless N 150 Desktop Adapter (rev.A1)" }] } , Device { vendor_id : 6164u16 , id : 12386u16 , name : "RT3062 Wireless 802.11n 2T/2R" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 12432u16 , name : "RT3090 Wireless 802.11n 1T/1R PCIe" , subsystems : & [SubSystem { subvendor : 5053u16 , subdevice : 4183u16 , name : "GN-WS32L-RH Half-size Mini PCIe Card" }] } , Device { vendor_id : 6164u16 , id : 12433u16 , name : "RT3091 Wireless 802.11n 1T/2R PCIe" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 12434u16 , name : "RT3092 Wireless 802.11n 2T/2R PCIe" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 12944u16 , name : "RT3290 Wireless 802.11n 1T/1R PCIe" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6380u16 , name : "Ralink RT3290LE 802.11bgn 1x1 Wi-Fi and Bluetooth 4.0 Combo Adapter" }] } , Device { vendor_id : 6164u16 , id : 12952u16 , name : "RT3290 Bluetooth" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6380u16 , name : "Ralink RT3290LE 802.11bgn 1x1 Wi-Fi and Bluetooth 4.0 Combo Adapter" }] } , Device { vendor_id : 6164u16 , id : 13714u16 , name : "RT3592 Wireless 802.11abgn 2T/2R PCIe" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 13727u16 , name : "RT3592 PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 21344u16 , name : "RT5360 Wireless 802.11n 1T/1R" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 15365u16 , name : "DWA-525 Wireless N 150 Desktop Adapter (rev.A2)" } , SubSystem { subvendor : 8436u16 , subdevice : 28730u16 , name : "TEW-703PI N150 Wireless PCI Adapter" }] } , Device { vendor_id : 6164u16 , id : 21346u16 , name : "RT5362 PCI 802.11n Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 21392u16 , name : "RT5390 Wireless 802.11n 1T/1R PCIe" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5686u16 , name : "U98Z077.00 Half-size Mini PCIe Card" }] } , Device { vendor_id : 6164u16 , id : 21394u16 , name : "RT5392 PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 21403u16 , name : "RT5390R 802.11bgn PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 21407u16 , name : "RT5390 [802.11 b/g/n 1T1R G-band PCI Express Single Chip]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5687u16 , name : "Pavilion DM1Z-3000 PCIe wireless card" }] } , Device { vendor_id : 6164u16 , id : 21906u16 , name : "RT5592 PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 6164u16 , id : 59698u16 , name : "RT2560F 802.11 b/g PCI" , subsystems : & [] }] }),
        (5255, Vendor { id : 5255u16 , name : "MARQUETTE Medical Systems" , devices : & [] }),
        (4156, Vendor { id : 4156u16 , name : "Hewlett-Packard Company" , devices : & [Device { vendor_id : 4156u16 , id : 4101u16 , name : "A4977A Visualize EG" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4104u16 , name : "Visualize FX" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4128u16 , name : "548XX Scope Interface" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4136u16 , name : "Tach TL Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4137u16 , name : "Tach XL2 Fibre Channel Host Adapter" , subsystems : & [SubSystem { subvendor : 4222u16 , subdevice : 15u16 , name : "Interphase 5560 Fibre Channel Adapter" } , SubSystem { subvendor : 36868u16 , subdevice : 37392u16 , name : "1Gb/2Gb Family Fibre Channel Controller" } , SubSystem { subvendor : 36868u16 , subdevice : 37393u16 , name : "1Gb/2Gb Family Fibre Channel Controller" }] } , Device { vendor_id : 4156u16 , id : 4138u16 , name : "Tach TS Fibre Channel Host Adapter" , subsystems : & [SubSystem { subvendor : 4222u16 , subdevice : 14u16 , name : "Interphase 5540/5541 Fibre Channel Adapter" } , SubSystem { subvendor : 36868u16 , subdevice : 37136u16 , name : "1Gb/2Gb Family Fibre Channel Controller" } , SubSystem { subvendor : 36868u16 , subdevice : 37137u16 , name : "1Gb/2Gb Family Fibre Channel Controller" }] } , Device { vendor_id : 4156u16 , id : 4144u16 , name : "J2585A DeskDirect 10/100VG NIC" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4145u16 , name : "J2585B HP 10/100VG PCI LAN Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4160u16 , name : "J2973A DeskDirect 10BaseT NIC" } , SubSystem { subvendor : 4156u16 , subdevice : 4161u16 , name : "J2585B DeskDirect 10/100VG NIC" } , SubSystem { subvendor : 4156u16 , subdevice : 4162u16 , name : "J2970A DeskDirect 10BaseT/2 NIC" }] } , Device { vendor_id : 4156u16 , id : 4160u16 , name : "J2973A DeskDirect 10BaseT NIC" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4161u16 , name : "J2585B DeskDirect 10/100 NIC" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4162u16 , name : "J2970A DeskDirect 10BaseT/2 NIC" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4168u16 , name : "Diva Serial [GSP] Multiport UART" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4169u16 , name : "Tosca Console" } , SubSystem { subvendor : 4156u16 , subdevice : 4170u16 , name : "Tosca Secondary" } , SubSystem { subvendor : 4156u16 , subdevice : 4171u16 , name : "Maestro SP2" } , SubSystem { subvendor : 4156u16 , subdevice : 4643u16 , name : "Superdome Console" } , SubSystem { subvendor : 4156u16 , subdevice : 4646u16 , name : "Keystone SP2" } , SubSystem { subvendor : 4156u16 , subdevice : 4647u16 , name : "Powerbar SP2" } , SubSystem { subvendor : 4156u16 , subdevice : 4738u16 , name : "Everest SP2" } , SubSystem { subvendor : 4156u16 , subdevice : 4865u16 , name : "Diva RMP3" }] } , Device { vendor_id : 4156u16 , id : 4180u16 , name : "PCI Local Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4196u16 , name : "79C970 PCnet Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4235u16 , name : "Visualize FXe" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4289u16 , name : "NetServer Smart IRQ Router" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4333u16 , name : "TopTools Remote Control" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4336u16 , name : "rio System Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4337u16 , name : "rio I/O Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4633u16 , name : "NetServer PCI Hot-Plug Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4634u16 , name : "NetServer SMIC Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4635u16 , name : "NetServer Legacy COM Port Decoder" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4636u16 , name : "NetServer PCI COM Port Decoder" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4649u16 , name : "zx1 System Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4650u16 , name : "zx1 I/O Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4654u16 , name : "PCI-X Local Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4731u16 , name : "sx1000 System Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4732u16 , name : "sx1000 I/O Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4749u16 , name : "Diva [GSP] Management Board" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4752u16 , name : "Auxiliary Diva Serial Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4753u16 , name : "Diva SP2" }] } , Device { vendor_id : 4156u16 , id : 4753u16 , name : "Auxiliary Diva Serial Port" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4788u16 , name : "zx1 QuickSilver AGP8x Local Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4843u16 , name : "sx2000 System Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4844u16 , name : "sx2000 I/O Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4846u16 , name : "PCI-X 2.0 Local Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4866u16 , name : "RMP-3 Shared Memory Driver" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 4867u16 , name : "RMP-3 (Remote Management Processor)" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 8950u16 , name : "iLO5 Virtual USB Controller" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 228u16 , name : "iLO5 Standard Virtual USB Controller" }] } , Device { vendor_id : 4156u16 , id : 10512u16 , name : "E2910A PCIBus Exerciser" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 10533u16 , name : "E2925A 32 Bit, 33 MHzPCI Exerciser & Analyzer" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 12806u16 , name : "Adaptec Embedded Serial ATA HostRAID" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 12832u16 , name : "Smart Array P600" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12837u16 , name : "3 Gb/s SAS RAID" }] } , Device { vendor_id : 4156u16 , id : 12848u16 , name : "Smart Array Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12835u16 , name : "Smart Array P800" } , SubSystem { subvendor : 4156u16 , subdevice : 12852u16 , name : "P400 SAS Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 12853u16 , name : "P400i SAS Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 12855u16 , name : "E500 SAS Controller" } , SubSystem { subvendor : 4156u16 , subdevice : 12861u16 , name : "P700m SAS Controller" }] } , Device { vendor_id : 4156u16 , id : 12856u16 , name : "Smart Array E200i (SAS Controller)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12817u16 , name : "Smart Array E200i" } , SubSystem { subvendor : 4156u16 , subdevice : 12818u16 , name : "Smart Array E200" }] } , Device { vendor_id : 4156u16 , id : 12857u16 , name : "Smart Array Gen9 Controllers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 8637u16 , name : "P244br" } , SubSystem { subvendor : 4156u16 , subdevice : 8638u16 , name : "P741m" } , SubSystem { subvendor : 4156u16 , subdevice : 8639u16 , name : "H240ar" } , SubSystem { subvendor : 4156u16 , subdevice : 8640u16 , name : "P440ar" } , SubSystem { subvendor : 4156u16 , subdevice : 8641u16 , name : "P840ar" } , SubSystem { subvendor : 4156u16 , subdevice : 8642u16 , name : "P440" } , SubSystem { subvendor : 4156u16 , subdevice : 8643u16 , name : "P441" } , SubSystem { subvendor : 4156u16 , subdevice : 8644u16 , name : "Smart Array" } , SubSystem { subvendor : 4156u16 , subdevice : 8645u16 , name : "P841" } , SubSystem { subvendor : 4156u16 , subdevice : 8646u16 , name : "H244br" } , SubSystem { subvendor : 4156u16 , subdevice : 8647u16 , name : "H240" } , SubSystem { subvendor : 4156u16 , subdevice : 8648u16 , name : "H241" } , SubSystem { subvendor : 4156u16 , subdevice : 8649u16 , name : "Smart Array" } , SubSystem { subvendor : 4156u16 , subdevice : 8650u16 , name : "P246br" } , SubSystem { subvendor : 4156u16 , subdevice : 8651u16 , name : "P840" } , SubSystem { subvendor : 4156u16 , subdevice : 8652u16 , name : "Smart Array" } , SubSystem { subvendor : 4156u16 , subdevice : 8653u16 , name : "P240nr" } , SubSystem { subvendor : 4156u16 , subdevice : 8654u16 , name : "H240nr" }] } , Device { vendor_id : 4156u16 , id : 12858u16 , name : "Smart Array G6 controllers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12865u16 , name : "Smart Array P212" } , SubSystem { subvendor : 4156u16 , subdevice : 12867u16 , name : "Smart Array P410" } , SubSystem { subvendor : 4156u16 , subdevice : 12869u16 , name : "Smart Array P410i" } , SubSystem { subvendor : 4156u16 , subdevice : 12871u16 , name : "Smart Array P411" } , SubSystem { subvendor : 4156u16 , subdevice : 12873u16 , name : "Smart Array P812" } , SubSystem { subvendor : 4156u16 , subdevice : 12874u16 , name : "Smart Array 712m (Mezzanine RAID controller)" } , SubSystem { subvendor : 4156u16 , subdevice : 12875u16 , name : "Smart Array P711m (Mezzanine RAID controller)" }] } , Device { vendor_id : 4156u16 , id : 12859u16 , name : "Smart Array Gen8 Controllers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13136u16 , name : "P222" } , SubSystem { subvendor : 4156u16 , subdevice : 13137u16 , name : "P420" } , SubSystem { subvendor : 4156u16 , subdevice : 13138u16 , name : "P421" } , SubSystem { subvendor : 4156u16 , subdevice : 13140u16 , name : "P420i" } , SubSystem { subvendor : 4156u16 , subdevice : 13141u16 , name : "P220i" }] } , Device { vendor_id : 4156u16 , id : 12860u16 , name : "Smart Array Gen8+ Controllers" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6432u16 , name : "P430i" } , SubSystem { subvendor : 4156u16 , subdevice : 6433u16 , name : "P830i" } , SubSystem { subvendor : 4156u16 , subdevice : 6434u16 , name : "P430" } , SubSystem { subvendor : 4156u16 , subdevice : 6435u16 , name : "P431" } , SubSystem { subvendor : 4156u16 , subdevice : 6436u16 , name : "P830" } , SubSystem { subvendor : 4156u16 , subdevice : 6437u16 , name : "Smart Array" } , SubSystem { subvendor : 4156u16 , subdevice : 6438u16 , name : "P731m" } , SubSystem { subvendor : 4156u16 , subdevice : 6440u16 , name : "P230i" }] } , Device { vendor_id : 4156u16 , id : 13056u16 , name : "Integrated Lights-Out Standard Virtual USB Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13060u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13061u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13065u16 , name : "iLO2 GXL/iLO3 GXE" } , SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4156u16 , id : 13057u16 , name : "Integrated Lights-Out Standard Serial Port" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13060u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13061u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4156u16 , id : 13058u16 , name : "Integrated Lights-Out Standard KCS Interface" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13060u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13061u16 , name : "iLO2" } , SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4156u16 , id : 13061u16 , name : "Integrated Lights-Out (iLO2) Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 13062u16 , name : "Integrated Lights-Out Standard Slave Instrumentation & System Support" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" } , SubSystem { subvendor : 5520u16 , subdevice : 228u16 , name : "iLO5" }] } , Device { vendor_id : 4156u16 , id : 13063u16 , name : "Integrated Lights-Out Standard Management Processor Support and Messaging" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13065u16 , name : "iLO 2" } , SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4156u16 , id : 13064u16 , name : "Integrated Lights-Out Standard MS Watchdog Timer" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13070u16 , name : "iLO3" } , SubSystem { subvendor : 4156u16 , subdevice : 13185u16 , name : "iLO4" }] } , Device { vendor_id : 4156u16 , id : 16432u16 , name : "zx2 System Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 16433u16 , name : "zx2 I/O Controller" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 16439u16 , name : "PCIe Local Bus Adapter" , subsystems : & [] } , Device { vendor_id : 4156u16 , id : 38402u16 , name : "AMD RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" }] }] }),
        (6292, Vendor { id : 6292u16 , name : "KNC One" , devices : & [] }),
        (4165, Vendor { id : 4165u16 , name : "OPTi Inc." , devices : & [Device { vendor_id : 4165u16 , id : 41208u16 , name : "82C750 [Vendetta] USB Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 49409u16 , name : "82C264 GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 49528u16 , name : "92C178" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50518u16 , name : "82X556 [Viper]" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50519u16 , name : "82C557 [Viper-M]" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50520u16 , name : "82C558 [Viper-M ISA+IDE]" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50535u16 , name : "82C750 [Vendetta] Host Bridge" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50536u16 , name : "82C750 [Vendetta] ISA Bridge" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50537u16 , name : "82C579 [Viper XPress+ Chipset]" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50721u16 , name : "82C621A PCI IDE Contoller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50944u16 , name : "82C700 [FireStar]" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 50945u16 , name : "82C700 [FireStar] Host Bridge" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51220u16 , name : "82C814 [FireBridge II] Docking Stration Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51234u16 , name : "82C822" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51236u16 , name : "82C824 [FireFox] 32-Bit PC Card Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51237u16 , name : "82C825 [FireBridge II] Docking Stration Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51250u16 , name : "82C832" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51297u16 , name : "82C861/2/3 [FireLink] PCI-USB Host Bridge" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51329u16 , name : "82C881 [FireLink] 1394 OHCI Link Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51349u16 , name : "82C895" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 51509u16 , name : "82С935 [MachOne] Integrated PCI Audio Processor" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 54632u16 , name : "82C700 [FireStar] PCI IDE Controller" , subsystems : & [] } , Device { vendor_id : 4165u16 , id : 55073u16 , name : "82C700 [FireStar] PCI IDE Controller" , subsystems : & [] }] }),
        (5192, Vendor { id : 5192u16 , name : "Alesis Studio Electronics" , devices : & [] }),
        (25093, Vendor { id : 25093u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (4830, Vendor { id : 4830u16 , name : "Rainbow Technologies" , devices : & [Device { vendor_id : 4830u16 , id : 512u16 , name : "CryptoSwift CS200" , subsystems : & [] }] }),
        (4494, Vendor { id : 4494u16 , name : "Hermstedt GmbH" , devices : & [] }),
        (6111, Vendor { id : 6111u16 , name : "Dini Group" , devices : & [Device { vendor_id : 6111u16 , id : 6244u16 , name : "Virtex4 PCI Board w/ QL5064 Bridge [DN7000K10PCI/DN8000K10PCI/DN8000K10PSX/NOTUS]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6245u16 , name : "Virtex4 ASIC Emulator [DN8000K10PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6246u16 , name : "Virtex4 ASIC Emulator Cable Connection [DN8000K10PCI]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6247u16 , name : "Virtex4 ASIC Emulator Cable Connection [DN8000K10PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6248u16 , name : "Virtex4 ASIC Emulator [DN8000K10PCIe-8]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6400u16 , name : "Virtex5 PCIe ASIC Emulator [DN9000K10PCIe8T/DN9002K10PCIe8T/DN9200K10PCIe8T/DN7006K10PCIe8T/DN7406K10PCIe8T]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6401u16 , name : "Virtex5 PCIe ASIC Emulator Large BARs [DN9000K10PCIe8T/DN9002K10PCIe8T/DN9200K10PCIe8T/DN7006K10PCIe8T/DN7406K10PCIe8T]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6402u16 , name : "Virtex5 PCIe ASIC Emulator Low Power [Interceptor]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6403u16 , name : "Spartan6 PCIe FPGA Accelerator Board [DNBFCS12PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6404u16 , name : "Virtex6 PCIe ASIC Emulation Board [DNDUALV6_PCIe4]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6405u16 , name : "Virtex6 PCIe ASIC Emulation Board [DNV6F6PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6406u16 , name : "Virtex6 PCIe ASIC Emulation Board [DN2076K10]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6407u16 , name : "Virtex6 PCIe ASIC Emulation Board [DNV6F2PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6408u16 , name : "Virtex6 PCIe ASIC Emulation Board Large BARs[DNV6F2PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6409u16 , name : "Kintex7 PCIe FPGA Accelerator Board [DNK7F5PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6410u16 , name : "Virtex7 PCIe ASIC Emulation Board [DNV7F1A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6411u16 , name : "Stratix5 PCIe ASIC Emulation Board [DNS5GXF2]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6412u16 , name : "Virtex7 PCIe ASIC Emulation Board [DNV7F2A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6413u16 , name : "Virtex7 PCIe ASIC Emulation Board [DNV7F4A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6414u16 , name : "Virtex7 PCIe ASIC Emulation Board [DNV7F2B]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6415u16 , name : "KintexUS PCIe MainRef Design [DNPCIE_40G_KU_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6416u16 , name : "VirtexUS ASIC Emulation Board [DNVUF4A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6417u16 , name : "VirtexUS PCIe ASIC Emulation Board [DNVU_F2PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6418u16 , name : "KintexUS PCIe MainRef Design [DNPCIe_40G_KU_LL_QSFP]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6419u16 , name : "VirtexUS ASIC Emulation Board [DNVUF1A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6420u16 , name : "VirtexUS ASIC Emulation Board [DNVUF2A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6421u16 , name : "Arria10 PCIe MainRef Design [DNPCIe_80G_A10_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6422u16 , name : "VirtexUS PCIe Accelerator Board [DNVUF2_HPC_PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6423u16 , name : "UltrascalePlus PCIe Accelerator Board [DNPCIe_400G_VU_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6424u16 , name : "VirtexUS+ ASIC Emulation Board [DNVUPF4A]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6425u16 , name : "UltrascalePlus PCIe Accelerator Board [DNPCIe_400G_VUP_HBM_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6656u16 , name : "Virtex6 PCIe DMA Netlist Design" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6657u16 , name : "Virtex6 PCIe Darklite Design [DNPCIe_HXT_10G_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6658u16 , name : "Virtex7 PCIe DMA Netlist Design" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6659u16 , name : "Kintex7 PCIe Darklite Design [DNPCIe_K7_10G_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6661u16 , name : "Stratix5 PCIe Darklite Design [DNS5GX_F2]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6662u16 , name : "VirtexUS PCIe DMA Netlist Design" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6663u16 , name : "KintexUS PCIe Darklite Design [DNPCIe_40G_KU_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6664u16 , name : "KintexUS PCIe Darklite Design [DNPCIe_40G_KU_LL_QSFP]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6665u16 , name : "Arria10 PCIe Darklite Design [DNPCIe_80G_A10_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6666u16 , name : "VirtexUS PCIe Darklite Design [DNVUF2_HPC_PCIe]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6667u16 , name : "UltrascalePlus PCIe Darklite Design [DNPCIe_400G_VU_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6668u16 , name : "KintexUS PCIe DRAM Packet Capture Design [DNPCIe_40G_KU_LL]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6669u16 , name : "KintexUS PCIe DRAM Packet Capture Design [DNPCIe_40G_KU_LL_2QSFP]" , subsystems : & [] } , Device { vendor_id : 6111u16 , id : 6670u16 , name : "UltrascalePlus PCIe Darklite Design [DNPCIe_400G_VUP_HBM_LL]" , subsystems : & [] }] }),
        (5161, Vendor { id : 5161u16 , name : "Unex Technology Corp." , devices : & [] }),
        (4913, Vendor { id : 4913u16 , name : "RadiSys Corporation" , devices : & [Device { vendor_id : 4913u16 , id : 48u16 , name : "ENP-2611" , subsystems : & [] } , Device { vendor_id : 4913u16 , id : 33280u16 , name : "82600 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4913u16 , id : 33281u16 , name : "82600 IDE" , subsystems : & [] } , Device { vendor_id : 4913u16 , id : 33282u16 , name : "82600 USB" , subsystems : & [] } , Device { vendor_id : 4913u16 , id : 33296u16 , name : "82600 PCI Bridge" , subsystems : & [] }] }),
        (8419, Vendor { id : 8419u16 , name : "Elix Systems SA" , devices : & [] }),
        (4585, Vendor { id : 4585u16 , name : "Highwater Designs Ltd." , devices : & [] }),
        (4802, Vendor { id : 4802u16 , name : "Mentec Limited" , devices : & [] }),
        (4905, Vendor { id : 4905u16 , name : "Productivity Enhancement" , devices : & [] }),
        (5982, Vendor { id : 5982u16 , name : "Sanera Systems, Inc." , devices : & [] }),
        (5445, Vendor { id : 5445u16 , name : "VISIONTEK" , devices : & [] }),
        (4192, Vendor { id : 4192u16 , name : "United Microelectronics [UMC]" , devices : & [Device { vendor_id : 4192u16 , id : 1u16 , name : "UM82C881" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 2u16 , name : "UM82C886" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 257u16 , name : "UM8673F" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 2177u16 , name : "UM8881" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 2182u16 , name : "UM8886F" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 2193u16 , name : "UM8891A" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 4097u16 , name : "UM886A" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 26426u16 , name : "UM8886BF" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 26427u16 , name : "EIDE Master/DMA" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34576u16 , name : "UM8710" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34922u16 , name : "UM8886A" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34945u16 , name : "UM8881F" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34950u16 , name : "UM8886F" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34954u16 , name : "UM8886A" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 34961u16 , name : "UM8891A" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 36887u16 , name : "UM9017F" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 36888u16 , name : "UM9018" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 36902u16 , name : "UM9026" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 59521u16 , name : "UM8881N" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 59526u16 , name : "UM8886N" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 59530u16 , name : "UM8886N" , subsystems : & [] } , Device { vendor_id : 4192u16 , id : 59537u16 , name : "UM8891N" , subsystems : & [] }] }),
        (5262, Vendor { id : 5262u16 , name : "OSI Plus Corporation" , devices : & [] }),
        (5573, Vendor { id : 5573u16 , name : "Procomp Informatics Ltd" , devices : & [Device { vendor_id : 5573u16 , id : 32784u16 , name : "1394b - 1394 Firewire 3-Port Host Adapter Card" , subsystems : & [] }] }),
        (5237, Vendor { id : 5237u16 , name : "PICAZO Communications" , devices : & [] }),
        (6191, Vendor { id : 6191u16 , name : "Broadcom" , devices : & [Device { vendor_id : 6191u16 , id : 11u16 , name : "BCM5785 [HT1000] SATA (RAID Mode)" , subsystems : & [] }] }),
        (5334, Vendor { id : 5334u16 , name : "Accusys Inc" , devices : & [Device { vendor_id : 5334u16 , id : 24833u16 , name : "ACS-61xxx, PCIe to SAS/SATA RAID HBA" , subsystems : & [] } , Device { vendor_id : 5334u16 , id : 25089u16 , name : "ACS-62xxx, External PCIe to SAS/SATA RAID controller" , subsystems : & [] }] }),
        (4379, Vendor { id : 4379u16 , name : "Teledyne Electronic Systems" , devices : & [] }),
        (8196, Vendor { id : 8196u16 , name : "Smart Link Ltd." , devices : & [] }),
        (4347, Vendor { id : 4347u16 , name : "Thesys Gesellschaft fuer Mikroelektronik mbH" , devices : & [Device { vendor_id : 4347u16 , id : 6255u16 , name : "TH 6255" , subsystems : & [] }] }),
        (4631, Vendor { id : 4631u16 , name : "O2 Micro, Inc." , devices : & [Device { vendor_id : 4631u16 , id : 247u16 , name : "Firewire (IEEE 1394)" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" }] } , Device { vendor_id : 4631u16 , id : 4343u16 , name : "1394 OHCI Compliant Host Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 4599u16 , name : "OZ600 1394a-2000 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" }] } , Device { vendor_id : 4631u16 , id : 5111u16 , name : "1394 OHCI Compliant Host Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26409u16 , name : "OZ6729" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26426u16 , name : "OZ6730" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26674u16 , name : "OZ6832/6833 CardBus Controller [Saturn]" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26678u16 , name : "OZ6836/6860 CardBus Controller [Mercury]" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26738u16 , name : "OZ6812 CardBus Controller [Challenger]" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26917u16 , name : "OZ6922 CardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 26931u16 , name : "OZ6933/711E1 CardBus/SmartCardBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4118u16 , name : "Travelmate 612 TX" }] } , Device { vendor_id : 4631u16 , id : 26994u16 , name : "OZ601/6912/711E0 CardBus/SmartCardBus Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 524u16 , name : "ThinkPad R30" } , SubSystem { subvendor : 4136u16 , subdevice : 338u16 , name : "Latitude D500" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Magnia Z310" }] } , Device { vendor_id : 4631u16 , id : 28944u16 , name : "OZ711Mx 4-in-1 MemoryCardBus Accelerator" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 5940u16 , subdevice : 4204u16 , name : "Amilo A1645" }] } , Device { vendor_id : 4631u16 , id : 28946u16 , name : "OZ711EC1/M1 SmartCardBus/MemoryCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 28947u16 , name : "OZ711EC1 SmartCardBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 53u16 , name : "TravelMate 660" }] } , Device { vendor_id : 4631u16 , id : 28948u16 , name : "OZ711M1/MC1 4-in-1 MemoryCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 28960u16 , name : "Integrated MMC/SD Controller" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" }] } , Device { vendor_id : 4631u16 , id : 28976u16 , name : "Integrated MS/xD Controller" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" }] } , Device { vendor_id : 4631u16 , id : 28980u16 , name : "OZ711MP1/MS1 MemoryCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 28981u16 , name : "Cardbus bridge" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 28982u16 , name : "OZ711SP1 Memory CardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 29154u16 , name : "OZ711E2 SmartCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 29202u16 , name : "OZ711M2 4-in-1 MemoryCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 29203u16 , name : "OZ6933E CardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 29219u16 , name : "OZ711M3/MC3 4-in-1 MemoryCardBus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 4303u16 , subdevice : 4548u16 , name : "Lifebook P5020D Laptop" }] } , Device { vendor_id : 4631u16 , id : 29235u16 , name : "OZ711MP3/MS3 4-in-1 MemoryCardBus Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33056u16 , name : "OZ888 SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33072u16 , name : "Integrated MS/MSPRO/xD Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33312u16 , name : "OZ600FJ1/OZ900FJ1 SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33313u16 , name : "OZ600FJ0/OZ900FJ0/OZ600FJS SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33568u16 , name : "OZ600RJ1/OZ900RJ1 SD/MMC Card Reader Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" }] } , Device { vendor_id : 4631u16 , id : 33569u16 , name : "OZ600RJ0/OZ900RJ0/OZ600RJS SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 33584u16 , name : "OZ600 MS/xD Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" }] } , Device { vendor_id : 4631u16 , id : 33585u16 , name : "O2 Flash Memory Card" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 34080u16 , name : "OZ777 SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 34336u16 , name : "OZ620 SD/MMC Card Reader Controller" , subsystems : & [] } , Device { vendor_id : 4631u16 , id : 34337u16 , name : "OZ711 SD/MMC Card Reader Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 20584u16 , name : "Thinkpad E480/E580" } , SubSystem { subvendor : 7748u16 , subdevice : 6006u16 , name : "Steam Deck" }] } , Device { vendor_id : 4631u16 , id : 34656u16 , name : "FORESEE E2M2 NVMe SSD" , subsystems : & [] }] }),
        (5589, Vendor { id : 5589u16 , name : "Tatung Co" , devices : & [] }),
        (5570, Vendor { id : 5570u16 , name : "Newer Technology Inc" , devices : & [] }),
        (4635, Vendor { id : 4635u16 , name : "Advanced Telecommunications Modules" , devices : & [] }),
        (8358, Vendor { id : 8358u16 , name : "XCENA, Inc." , devices : & [] }),
        (4474, Vendor { id : 4474u16 , name : "A-Trend Technology" , devices : & [] }),
        (4188, Vendor { id : 4188u16 , name : "Wipro Infotech Limited" , devices : & [] }),
        (26372, Vendor { id : 26372u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (6130, Vendor { id : 6130u16 , name : "Albatron Corp." , devices : & [] }),
        (5341, Vendor { id : 5341u16 , name : "Boulder Design Labs Inc" , devices : & [] }),
        (5930, Vendor { id : 5930u16 , name : "Accelerated Encryption" , devices : & [Device { vendor_id : 5930u16 , id : 5064u16 , name : "AEP SureWare Runner 1000V3" , subsystems : & [] }] }),
        (4534, Vendor { id : 4534u16 , name : "United Video Corp" , devices : & [] }),
        (4232, Vendor { id : 4232u16 , name : "Microcomputer Systems (M) Son" , devices : & [] }),
        (7857, Vendor { id : 7857u16 , name : "VeriSilicon Inc" , devices : & [Device { vendor_id : 7857u16 , id : 4097u16 , name : "Video Accelerator" , subsystems : & [] }] }),
        (5210, Vendor { id : 5210u16 , name : "Escalate Networks Inc" , devices : & [] }),
        (4262, Vendor { id : 4262u16 , name : "Informtech Industrial Ltd." , devices : & [] }),
        (5572, Vendor { id : 5572u16 , name : "EVSX Inc" , devices : & [] }),
        (7554, Vendor { id : 7554u16 , name : "NETINT Technologies Inc." , devices : & [Device { vendor_id : 7554u16 , id : 257u16 , name : "Codensity D400 SSD" , subsystems : & [] } , Device { vendor_id : 7554u16 , id : 258u16 , name : "Codensity D408 PCIe Gen4 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7554u16 , id : 514u16 , name : "Codensity T408 Video Encoding-Decoding Accelerator" , subsystems : & [] } , Device { vendor_id : 7554u16 , id : 1025u16 , name : "Quadra Video Processing Unit (VPU)" , subsystems : & [] } , Device { vendor_id : 7554u16 , id : 1041u16 , name : "Quadra Video Processing Unit (VPU) Virtual Function" , subsystems : & [] }] }),
        (12715, Vendor { id : 12715u16 , name : "Zonet" , devices : & [Device { vendor_id : 12715u16 , id : 8106u16 , name : "ZEW1602 802.11b/g Wireless Adapter" , subsystems : & [] }] }),
        (4269, Vendor { id : 4269u16 , name : "Winbond Electronics Corp / Symphony Labs" , devices : & [Device { vendor_id : 4269u16 , id : 1u16 , name : "W83769F" , subsystems : & [] } , Device { vendor_id : 4269u16 , id : 3u16 , name : "SL82C103" , subsystems : & [] } , Device { vendor_id : 4269u16 , id : 5u16 , name : "SL82C105" , subsystems : & [] } , Device { vendor_id : 4269u16 , id : 259u16 , name : "SL82c103" , subsystems : & [] } , Device { vendor_id : 4269u16 , id : 261u16 , name : "SL82C105/W83C55xF Bus Master IDE" , subsystems : & [] } , Device { vendor_id : 4269u16 , id : 1381u16 , name : "W83C553F/554F ISA bridge" , subsystems : & [] }] }),
        (7119, Vendor { id : 7119u16 , name : "NEC Corporation" , devices : & [Device { vendor_id : 7119u16 , id : 28u16 , name : "Vector Engine 1.0" , subsystems : & [] }] }),
        (5296, Vendor { id : 5296u16 , name : "BST Communication Technology Ltd" , devices : & [] }),
        (6574, Vendor { id : 6574u16 , name : "Progeny Systems Corporation" , devices : & [Device { vendor_id : 6574u16 , id : 1312u16 , name : "4135 HFT Interface Controller" , subsystems : & [] } , Device { vendor_id : 6574u16 , id : 1313u16 , name : "Decimator" , subsystems : & [] }] }),
        (7865, Vendor { id : 7865u16 , name : "Senscomm Semiconductor, Inc" , devices : & [Device { vendor_id : 7865u16 , id : 8224u16 , name : "SCM2625 Wi-Fi6 Network Adapter" , subsystems : & [] }] }),
        (5494, Vendor { id : 5494u16 , name : "Viewcast COM" , devices : & [] }),
        (32776, Vendor { id : 32776u16 , name : "Quancom Electronic GmbH" , devices : & [Device { vendor_id : 32776u16 , id : 16u16 , name : "WDOG1 [PCI-Watchdog 1]" , subsystems : & [] } , Device { vendor_id : 32776u16 , id : 17u16 , name : "PWDOG2 [PCI-Watchdog 2]" , subsystems : & [] } , Device { vendor_id : 32776u16 , id : 21u16 , name : "Clock77/PCI & Clock77/PCIe (DCF-77 receiver)" , subsystems : & [] }] }),
        (5869, Vendor { id : 5869u16 , name : "Sycron N. V." , devices : & [Device { vendor_id : 5869u16 , id : 4097u16 , name : "UMIO communication card" , subsystems : & [] }] }),
        (18710, Vendor { id : 18710u16 , name : "RedCreek Communications Inc" , devices : & [Device { vendor_id : 18710u16 , id : 6496u16 , name : "RedCreek PCI adapter" , subsystems : & [] }] }),
        (4211, Vendor { id : 4211u16 , name : "Yamaha Corporation" , devices : & [Device { vendor_id : 4211u16 , id : 1u16 , name : "3D GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 2u16 , name : "YGV615 [RPA3 3D-Graphics Controller]" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 3u16 , name : "YMF-740" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 4u16 , name : "YMF-724" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 4u16 , name : "YMF724-Based PCI Audio Adapter" }] } , Device { vendor_id : 4211u16 , id : 5u16 , name : "DS1 Audio" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 5u16 , name : "DS-XG PCI Audio CODEC" }] } , Device { vendor_id : 4211u16 , id : 6u16 , name : "DS1 Audio" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 8u16 , name : "DS1 Audio" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 8u16 , name : "DS-XG PCI Audio CODEC" }] } , Device { vendor_id : 4211u16 , id : 10u16 , name : "DS1L Audio" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 4u16 , name : "DS-XG PCI Audio CODEC" } , SubSystem { subvendor : 4211u16 , subdevice : 10u16 , name : "DS-XG PCI Audio CODEC" } , SubSystem { subvendor : 32902u16 , subdevice : 19797u16 , name : "DS-XG PCI Audio CODEC [Intel MU440EX]" }] } , Device { vendor_id : 4211u16 , id : 12u16 , name : "YMF-740C [DS-1L Audio Controller]" , subsystems : & [SubSystem { subvendor : 4218u16 , subdevice : 12u16 , name : "DS-XG PCI Audio CODEC" }] } , Device { vendor_id : 4211u16 , id : 13u16 , name : "YMF-724F [DS-1 Audio Controller]" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 13u16 , name : "DS-XG PCI Audio CODEC" }] } , Device { vendor_id : 4211u16 , id : 16u16 , name : "YMF-744B [DS-1S Audio Controller]" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 6u16 , name : "DS-XG PCI Audio CODEC" } , SubSystem { subvendor : 4211u16 , subdevice : 16u16 , name : "DS-XG PCI Audio CODEC" }] } , Device { vendor_id : 4211u16 , id : 18u16 , name : "YMF-754 [DS-1E Audio Controller]" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 18u16 , name : "DS-XG PCI Audio Codec" }] } , Device { vendor_id : 4211u16 , id : 32u16 , name : "DS-1 Audio" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 4096u16 , name : "SW1000XG [XG Factory]" , subsystems : & [] } , Device { vendor_id : 4211u16 , id : 8192u16 , name : "DS2416 Digital Mixing Card" , subsystems : & [SubSystem { subvendor : 4211u16 , subdevice : 8192u16 , name : "DS2416 Digital Mixing Card" }] }] }),
        (5913, Vendor { id : 5913u16 , name : "EZChip Technologies" , devices : & [Device { vendor_id : 5913u16 , id : 4096u16 , name : "NPA Access Network Processor Family" , subsystems : & [] }] }),
        (5440, Vendor { id : 5440u16 , name : "PROVIDEO MULTIMEDIA Co Ltd" , devices : & [] }),
        (5697, Vendor { id : 5697u16 , name : "MKNet Corp." , devices : & [] }),
        (4109, Vendor { id : 4109u16 , name : "AST Research Inc" , devices : & [] }),
        (5698, Vendor { id : 5698u16 , name : "Bitland(ShenZhen) Information Technology Co., Ltd." , devices : & [] }),
        (5489, Vendor { id : 5489u16 , name : "Contemporary Controls" , devices : & [Device { vendor_id : 5489u16 , id : 40961u16 , name : "CCSI PCI20-485 ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40962u16 , name : "CCSI PCI20-485D ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40963u16 , name : "CCSI PCI20-485X ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40964u16 , name : "CCSI PCI20-CXB ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40965u16 , name : "CCSI PCI20-CXS ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40966u16 , name : "CCSI PCI20-FOG-SMA ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40967u16 , name : "CCSI PCI20-FOG-ST ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40968u16 , name : "CCSI PCI20-TB5 ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40969u16 , name : "CCSI PCI20-5-485 5Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40970u16 , name : "CCSI PCI20-5-485D 5Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40971u16 , name : "CCSI PCI20-5-485X 5Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40972u16 , name : "CCSI PCI20-5-FOG-ST 5Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 40973u16 , name : "CCSI PCI20-5-FOG-SMA 5Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41473u16 , name : "CCSI PCI22-485 10Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41474u16 , name : "CCSI PCI22-485D 10Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41475u16 , name : "CCSI PCI22-485X 10Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41476u16 , name : "CCSI PCI22-CHB 10Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41477u16 , name : "CCSI PCI22-FOG_ST 10Mbit ARCnet" , subsystems : & [] } , Device { vendor_id : 5489u16 , id : 41478u16 , name : "CCSI PCI22-THB 10Mbit ARCnet" , subsystems : & [] }] }),
        (4116, Vendor { id : 4116u16 , name : "IBM" , devices : & [Device { vendor_id : 4116u16 , id : 2u16 , name : "PCI to MCA Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 5u16 , name : "Processor to I/O Controller [Alta Lite]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 7u16 , name : "Processor to I/O Controller [Alta MP]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 10u16 , name : "PCI to ISA Bridge (IBM27-82376) [Fire Coral]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 23u16 , name : "CPU to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 24u16 , name : "TR Auto LANstreamer" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 27u16 , name : "GXT-150P" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 28u16 , name : "Carrera" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 29u16 , name : "SCSI-2 FAST PCI Adapter (82G2675)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 32u16 , name : "GXT1000 Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 34u16 , name : "PCI to PCI Bridge (IBM27-82351)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 45u16 , name : "Processor to I/O Controller [Python]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 46u16 , name : "SCSI RAID Adapter [ServeRAID]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 46u16 , name : "ServeRAID-3x" } , SubSystem { subvendor : 4116u16 , subdevice : 558u16 , name : "ServeRAID-4H" }] } , Device { vendor_id : 4116u16 , id : 49u16 , name : "2 Port Serial Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 49u16 , name : "2721 WAN IOA - 2 Port Sync Serial Adapter" }] } , Device { vendor_id : 4116u16 , id : 54u16 , name : "PCI to 32-bit LocalBus Bridge [Miami]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 55u16 , name : "PowerPC to PCI Bridge (IBM27-82660)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 58u16 , name : "CPU to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 60u16 , name : "GXT250P/GXT255P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 62u16 , name : "16/4 Token ring UTP/STP controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 62u16 , name : "Token-Ring Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 205u16 , name : "Token-Ring Adapter + Wake-On-LAN" } , SubSystem { subvendor : 4116u16 , subdevice : 206u16 , name : "16/4 Token-Ring Adapter 2" } , SubSystem { subvendor : 4116u16 , subdevice : 207u16 , name : "16/4 Token-Ring Adapter Special" } , SubSystem { subvendor : 4116u16 , subdevice : 228u16 , name : "High-Speed 100/16/4 Token-Ring Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 229u16 , name : "16/4 Token-Ring Adapter 2 + Wake-On-LAN" } , SubSystem { subvendor : 4116u16 , subdevice : 365u16 , name : "iSeries 2744 Card" }] } , Device { vendor_id : 4116u16 , id : 69u16 , name : "SSA Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 70u16 , name : "MPIC interrupt controller" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 71u16 , name : "PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 72u16 , name : "PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 73u16 , name : "Warhead SCSI Controller" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 78u16 , name : "ATM Controller (14104e00)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 79u16 , name : "ATM Controller (14104f00)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 80u16 , name : "ATM Controller (14105000)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 83u16 , name : "25 MBit ATM Controller" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 84u16 , name : "GXT500P/GXT550P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 87u16 , name : "MPEG PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 88u16 , name : "SSA Adapter [Advanced SerialRAID/X]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 94u16 , name : "GXT800P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 124u16 , name : "ATM Controller (14107c00)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 125u16 , name : "3780IDSP [MWave]" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 139u16 , name : "EADS PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 142u16 , name : "GXT3000P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 144u16 , name : "GXT 3000P" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 142u16 , name : "GXT-3000P" }] } , Device { vendor_id : 4116u16 , id : 145u16 , name : "SSA Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 149u16 , name : "20H2999 PCI Docking Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 150u16 , name : "Chukar chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 151u16 , name : "iSeries 2778 DASD IOA" } , SubSystem { subvendor : 4116u16 , subdevice : 152u16 , name : "iSeries 2763 DASD IOA" } , SubSystem { subvendor : 4116u16 , subdevice : 153u16 , name : "iSeries 2748 DASD IOA" }] } , Device { vendor_id : 4116u16 , id : 159u16 , name : "PCI 4758 Cryptographic Accelerator" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 165u16 , name : "ATM Controller (1410a500)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 166u16 , name : "ATM 155MBPS MM Controller (1410a600)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 183u16 , name : "GXT2000P Graphics Adapter" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 184u16 , name : "FireGL1 AGP 32Mb" }] } , Device { vendor_id : 4116u16 , id : 184u16 , name : "GXT2000P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 190u16 , name : "ATM 622MBPS Controller (1410be00)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 220u16 , name : "Advanced Systems Management Adapter (ASMA)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 252u16 , name : "CPC710 Dual Bridge and Memory Controller (PCI-64)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 261u16 , name : "CPC710 Dual Bridge and Memory Controller (PCI-32)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 271u16 , name : "Remote Supervisor Adapter (RSA)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 322u16 , name : "Yotta Video Compositor Input" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 323u16 , name : "Yotta Input Controller (ytin)" }] } , Device { vendor_id : 4116u16 , id : 324u16 , name : "Yotta Video Compositor Output" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 325u16 , name : "Yotta Output Controller (ytout)" }] } , Device { vendor_id : 4116u16 , id : 342u16 , name : "405GP PLB to PCI Bridge" , subsystems : & [SubSystem { subvendor : 8131u16 , subdevice : 4u16 , name : "IC695ETM001 RX3i Ethernet PCI Module" }] } , Device { vendor_id : 4116u16 , id : 350u16 , name : "622Mbps ATM PCI Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 352u16 , name : "64bit/66MHz PCI ATM 155 MMF" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 366u16 , name : "GXT4000P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 368u16 , name : "GXT6000P Graphics Adapter" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 370u16 , name : "Fire GL2" } , SubSystem { subvendor : 4242u16 , subdevice : 371u16 , name : "Fire GL3" } , SubSystem { subvendor : 4242u16 , subdevice : 372u16 , name : "Fire GL4" } , SubSystem { subvendor : 4242u16 , subdevice : 388u16 , name : "Fire GL4s" }] } , Device { vendor_id : 4116u16 , id : 381u16 , name : "GXT300P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 384u16 , name : "Snipe chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 577u16 , name : "iSeries 2757 DASD IOA" } , SubSystem { subvendor : 4116u16 , subdevice : 612u16 , name : "Quad Channel PCI-X U320 SCSI RAID Adapter (2780)" }] } , Device { vendor_id : 4116u16 , id : 392u16 , name : "EADS-X PCI-X to PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 423u16 , name : "PCI-X to PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 445u16 , name : "ServeRAID Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 445u16 , name : "ServeRAID 4Lx" } , SubSystem { subvendor : 4116u16 , subdevice : 446u16 , name : "ServeRAID-4M" } , SubSystem { subvendor : 4116u16 , subdevice : 447u16 , name : "ServeRAID-4L" } , SubSystem { subvendor : 4116u16 , subdevice : 520u16 , name : "ServeRAID-4Mx" } , SubSystem { subvendor : 4116u16 , subdevice : 526u16 , name : "ServeRAID-4Lx" } , SubSystem { subvendor : 4116u16 , subdevice : 558u16 , name : "ServeRAID-4H" } , SubSystem { subvendor : 4116u16 , subdevice : 600u16 , name : "ServeRAID-5i" } , SubSystem { subvendor : 4116u16 , subdevice : 601u16 , name : "ServeRAID-5i" }] } , Device { vendor_id : 4116u16 , id : 449u16 , name : "64bit/66MHz PCI ATM 155 UTP" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 486u16 , name : "Cryptographic Accelerator" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 495u16 , name : "PowerPC 440GP PCI Bridge" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4139u16 , name : "PCEAS PCI-X Dual Port ESCON Adapter" } , SubSystem { subvendor : 5940u16 , subdevice : 4344u16 , name : "PCEAT PCI-Express Dual Port ESCON Adapter" }] } , Device { vendor_id : 4116u16 , id : 511u16 , name : "10/100 Mbps Ethernet" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 537u16 , name : "Multiport Serial Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 538u16 , name : "Dual RVX" } , SubSystem { subvendor : 4116u16 , subdevice : 593u16 , name : "Internal Modem/RVX" } , SubSystem { subvendor : 4116u16 , subdevice : 594u16 , name : "Quad Internal Modem" }] } , Device { vendor_id : 4116u16 , id : 539u16 , name : "GXT6500P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 540u16 , name : "GXT4500P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 563u16 , name : "GXT135P Graphics Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 602u16 , name : "Drone card" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 652u16 , name : "Citrine chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 653u16 , name : "Dual Channel PCI-X DDR SAS RAID Adapter (572E)" } , SubSystem { subvendor : 4116u16 , subdevice : 702u16 , name : "Dual Channel PCI-X DDR U320 SCSI RAID Adapter (571B)" } , SubSystem { subvendor : 4116u16 , subdevice : 704u16 , name : "Dual Channel PCI-X DDR U320 SCSI Adapter (571A)" } , SubSystem { subvendor : 4116u16 , subdevice : 781u16 , name : "PCI-X DDR Auxiliary Cache Adapter (575B)" }] } , Device { vendor_id : 4116u16 , id : 673u16 , name : "Calgary PCI-X Host Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 701u16 , name : "Obsidian chipset SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 705u16 , name : "PCI-X DDR 3Gb SAS Adapter (572A/572C)" } , SubSystem { subvendor : 4116u16 , subdevice : 706u16 , name : "PCI-X DDR 3Gb SAS RAID Adapter (572B/571D)" } , SubSystem { subvendor : 4116u16 , subdevice : 824u16 , name : "PCI-X DDR Auxiliary Cache Adapter (575C)" }] } , Device { vendor_id : 4116u16 , id : 770u16 , name : "Winnipeg PCI-X Host Bridge" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 776u16 , name : "CalIOC2 PCI-E Root Port" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 785u16 , name : "FC 5740/1954 4-Port 10/100/1000 Base-TX PCI-X Adapter for POWER" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 788u16 , name : "ZISC 036 Neural accelerator card" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 813u16 , name : "Axon - Cell Companion Chip" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 929u16 , name : "PCIe PowerXCell 8i Cell Accelerator Board" }] } , Device { vendor_id : 4116u16 , id : 825u16 , name : "Obsidian-E PCI-E SCSI controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 778u16 , name : "PCIe 3Gb SAS RAID Adapter (574E)" } , SubSystem { subvendor : 4116u16 , subdevice : 826u16 , name : "PCIe 3Gb SAS Adapter (57B3)" } , SubSystem { subvendor : 4116u16 , subdevice : 860u16 , name : "PCIe x8 Internal 3Gb SAS adapter (57CC)" } , SubSystem { subvendor : 4116u16 , subdevice : 864u16 , name : "PCI-E Auxiliary Cache Adapter (57B7)" }] } , Device { vendor_id : 4116u16 , id : 829u16 , name : "PCI-E IPR SAS Adapter (FPGA)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 828u16 , name : "PCIe2 1.8GB Cache 6Gb SAS RAID Adapter Tri-port (57B5)" } , SubSystem { subvendor : 4116u16 , subdevice : 851u16 , name : "PCIe2 3.1GB Cache 6Gb SAS RAID Enclosure (57C3)" } , SubSystem { subvendor : 4116u16 , subdevice : 852u16 , name : "PCIe2 6Gb SAS Adapter Dual-port (57C4)" } , SubSystem { subvendor : 4116u16 , subdevice : 854u16 , name : "PCIe2 1.8GB Cache 6Gb SAS RAID & SSD Adapter (574D)" } , SubSystem { subvendor : 4116u16 , subdevice : 863u16 , name : "PCIe2 6Gb SAS Adapter Quad-port (57B2)" }] } , Device { vendor_id : 4116u16 , id : 842u16 , name : "PCI-E IPR SAS Adapter (ASIC)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 827u16 , name : "PCIe2 6Gb SAS RAID Adapter Quad-port (57B4)" } , SubSystem { subvendor : 4116u16 , subdevice : 853u16 , name : "PCIe2 3.6GB Cache 6Gb SAS RAID Adapter Quad-port (57B1)" } , SubSystem { subvendor : 4116u16 , subdevice : 855u16 , name : "PCIe2 6Gb SAS Adapter Quad-port (57C6)" } , SubSystem { subvendor : 4116u16 , subdevice : 861u16 , name : "PCIe3 1.8GB Cache RAID SAS Adapter Quad-port 6GB (57C8)" } , SubSystem { subvendor : 4116u16 , subdevice : 862u16 , name : "PCIe2 3.6GB Cache 6Gb SAS RAID Adapter Quad-port (57CE)" } , SubSystem { subvendor : 4116u16 , subdevice : 1019u16 , name : "PCIe3 28GB Cache RAID SAS Enclosure 6Gb x 16 (57D5)" } , SubSystem { subvendor : 4116u16 , subdevice : 1022u16 , name : "PCIe3 x8 Cache SAS RAID Internal Adapter 6Gb (57D8)" } , SubSystem { subvendor : 4116u16 , subdevice : 1023u16 , name : "PCIe3 x8 SAS RAID Internal Adapter 6Gb (57D7)" } , SubSystem { subvendor : 4116u16 , subdevice : 1140u16 , name : "PCIe3 x16 Cache SAS RAID Internal Adapter 6Gb (57EB)" } , SubSystem { subvendor : 4116u16 , subdevice : 1141u16 , name : "PCIe3 x16 SAS RAID Internal Adapter 6Gb (57EC)" } , SubSystem { subvendor : 4116u16 , subdevice : 1177u16 , name : "PCIe3 x16 Cache SAS RAID Internal Adapter 6Gb (57ED)" } , SubSystem { subvendor : 4116u16 , subdevice : 1178u16 , name : "PCIe3 x16 SAS RAID Internal Adapter 6Gb (57EE)" } , SubSystem { subvendor : 4116u16 , subdevice : 1223u16 , name : "PCIe3 x 8 Cache SAS RAID Internal Adapter 6GB(2CCA)" } , SubSystem { subvendor : 4116u16 , subdevice : 1224u16 , name : "PCIe3 x 8 Cache SAS RAID Internal Adapter 6GB(2CD2)" } , SubSystem { subvendor : 4116u16 , subdevice : 1225u16 , name : "PCIe3 x 8 Cache SAS RAID Internal Adapter 6GB(2CCD)" }] } , Device { vendor_id : 4116u16 , id : 988u16 , name : "POWER8 Host Bridge (PHB3)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1099u16 , name : "GenWQE Accelerator Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1194u16 , name : "Flash Adapter 90 (PCIe2 0.9TB)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1217u16 , name : "POWER9 Host Bridge (PHB4)" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1242u16 , name : "PCI-E IPR SAS+ Adapter (ASIC)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1275u16 , name : "PCIe3 x16 20GB Cache 12Gb Quad SAS RAID+ Adapter(580B)" } , SubSystem { subvendor : 4116u16 , subdevice : 1276u16 , name : "PCIe3 x8 12Gb Quad SAS RAID+ Adapter(580A)" }] } , Device { vendor_id : 4116u16 , id : 1261u16 , name : "Internal Shared Memory (ISM) virtual PCI device" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1553u16 , name : "4769 Cryptographic Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1698u16 , name : "4770 Cryptographic Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1703u16 , name : "Spyre Accelerator" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 1704u16 , name : "Spyre Accelerator Virtual Function" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 12322u16 , name : "QLA3022 Network Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 16418u16 , name : "QLA3022 Network Adapter" , subsystems : & [] } , Device { vendor_id : 4116u16 , id : 65535u16 , name : "MPIC-2 interrupt controller" , subsystems : & [] }] }),
        (291, Vendor { id : 291u16 , name : "General Dynamics" , devices : & [] }),
        (5140, Vendor { id : 5140u16 , name : "Microsoft Corporation" , devices : & [Device { vendor_id : 5140u16 , id : 1u16 , name : "MN-120 (ADMtek Centaur-C based)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 2u16 , name : "MN-130 (ADMtek Centaur-P based)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 140u16 , name : "Basic Render Driver" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 142u16 , name : "Basic Render Driver" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 21331u16 , name : "Hyper-V virtual VGA" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22529u16 , name : "XMA Decoder (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22530u16 , name : "SATA Controller - CdRom (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22531u16 , name : "SATA Controller - Disk (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22532u16 , name : "OHCI Controller 0 (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22533u16 , name : "EHCI Controller 0 (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22534u16 , name : "OHCI Controller 1 (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22535u16 , name : "EHCI Controller 1 (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22538u16 , name : "Fast Ethernet Adapter (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22539u16 , name : "Secure Flash Controller (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22541u16 , name : "System Management Controller (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22545u16 , name : "Xenos GPU (Xenon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22561u16 , name : "Xenos GPU (Zephyr/Falcon)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22577u16 , name : "Xenos GPU (Jasper)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 22593u16 , name : "Xenos GPU (Slim)" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 49200u16 , name : "OpenVMM PCIe Root Port" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 49201u16 , name : "OpenVMM PCIe Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 49202u16 , name : "OpenVMM PCIe Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 5140u16 , id : 49214u16 , name : "OpenVMM NVMe Controller" , subsystems : & [] }] }),
        (5235, Vendor { id : 5235u16 , name : "ZAPEX Technologies Inc" , devices : & [] }),
        (13429, Vendor { id : 13429u16 , name : "Arista Networks, Inc." , devices : & [] }),
        (4428, Vendor { id : 4428u16 , name : "Annabooks" , devices : & [] }),
        (6339, Vendor { id : 6339u16 , name : "Micronas Semiconductor Holding AG" , devices : & [Device { vendor_id : 6339u16 , id : 1824u16 , name : "nGene PCI-Express Multimedia Controller" , subsystems : & [SubSystem { subvendor : 5217u16 , subdevice : 814u16 , name : "Hybrid M779 PCI-E" }] }] }),
        (5216, Vendor { id : 5216u16 , name : "DYNARC INC" , devices : & [] }),
        (4798, Vendor { id : 4798u16 , name : "Anchor Chips Inc." , devices : & [Device { vendor_id : 4798u16 , id : 12353u16 , name : "AN3041Q CO-MEM" , subsystems : & [] } , Device { vendor_id : 4798u16 , id : 12354u16 , name : "AN3042Q CO-MEM Lite" , subsystems : & [SubSystem { subvendor : 4798u16 , subdevice : 12354u16 , name : "Anchor Chips Lite Evaluation Board" }] }] }),
        (4748, Vendor { id : 4748u16 , name : "Retix Corporation" , devices : & [] }),
        (5179, Vendor { id : 5179u16 , name : "Canon Research Center, America" , devices : & [] }),
        (4708, Vendor { id : 4708u16 , name : "Aval Nagasaki Corporation" , devices : & [] }),
        (4992, Vendor { id : 4992u16 , name : "Sanritz Automation Co Ltd" , devices : & [] }),
        (4553, Vendor { id : 4553u16 , name : "Magma" , devices : & [Device { vendor_id : 4553u16 , id : 16u16 , name : "16-line serial port w/- DMA" , subsystems : & [] } , Device { vendor_id : 4553u16 , id : 17u16 , name : "4-line serial port w/- DMA" , subsystems : & [] }] }),
        (19216, Vendor { id : 19216u16 , name : "Buslogic Inc." , devices : & [] }),
        (4722, Vendor { id : 4722u16 , name : "Telematics International" , devices : & [] }),
        (26883, Vendor { id : 26883u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (5274, Vendor { id : 5274u16 , name : "ANDOR Technology Ltd" , devices : & [] }),
        (4944, Vendor { id : 4944u16 , name : "Systec Co. Ltd" , devices : & [] }),
        (43525, Vendor { id : 43525u16 , name : "iTuner" , devices : & [] }),
        (5288, Vendor { id : 5288u16 , name : "FEATRON Technologies Corporation" , devices : & [] }),
        (4928, Vendor { id : 4928u16 , name : "Atalla Corp" , devices : & [] }),
        (4485, Vendor { id : 4485u16 , name : "Dataworld International Ltd" , devices : & [] }),
        (5307, Vendor { id : 5307u16 , name : "SEMTECH Corporation" , devices : & [] }),
        (21553, Vendor { id : 21553u16 , name : "AuzenTech, Inc." , devices : & [] }),
        (4989, Vendor { id : 4989u16 , name : "Dynachip Corporation" , devices : & [] }),
        (5317, Vendor { id : 5317u16 , name : "Automation Products AB" , devices : & [] }),
        (4178, Vendor { id : 4178u16 , name : "?Young Micro Systems" , devices : & [] }),
        (8264, Vendor { id : 8264u16 , name : "Beijing SpaceControl Technology Co.Ltd" , devices : & [] }),
        (5009, Vendor { id : 5009u16 , name : "Development Concepts Inc" , devices : & [] }),
        (5054, Vendor { id : 5054u16 , name : "Miroku Jyoho Service Co. Ltd" , devices : & [] }),
        (5571, Vendor { id : 5571u16 , name : "Taiwan Mycomp Co Ltd" , devices : & [] }),
        (6535, Vendor { id : 6535u16 , name : "Phison Electronics Corporation" , devices : & [Device { vendor_id : 6535u16 , id : 20487u16 , name : "E7 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20488u16 , name : "E8 PCIe3 x2 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20498u16 , name : "E12 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20499u16 , name : "PS5013-E13 PCIe3 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20501u16 , name : "PS5015-E15 PCIe3 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20502u16 , name : "E16 PCIe4 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20504u16 , name : "E18 PCIe4 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20505u16 , name : "PS5019-E19 PCIe4 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20512u16 , name : "PS5020-E20 (X1) PCIe4 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20513u16 , name : "PS5021-E21 PCIe4 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20518u16 , name : "PS5026-E26 PCIe5 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20519u16 , name : "PS5027-E27T PCIe4 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20520u16 , name : "PS5028-E28 PCIe5 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20521u16 , name : "PS5029-E29T PCIe4 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20529u16 , name : "PS5031-E31T PCIe5 NVMe Controller" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 20535u16 , name : "PS5037-E37T PCIe5 NVMe Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6535u16 , id : 21250u16 , name : "PS5302-X2 PCIe5 NVMe Controller" , subsystems : & [] }] }),
        (5515, Vendor { id : 5515u16 , name : "Allied Data Technologies" , devices : & [] }),
        (6901, Vendor { id : 6901u16 , name : "Netezza Corp." , devices : & [] }),
        (29990, Vendor { id : 29990u16 , name : "HongQin (Beijing) Technology Co., Ltd." , devices : & [Device { vendor_id : 29990u16 , id : 130u16 , name : "HQ SSD M.2" , subsystems : & [] } , Device { vendor_id : 29990u16 , id : 131u16 , name : "HQ SSD 2TB M.2 NVMe" , subsystems : & [] }] }),
        (4719, Vendor { id : 4719u16 , name : "Silicon Motion, Inc." , devices : & [Device { vendor_id : 4719u16 , id : 1281u16 , name : "SM501 VoyagerGX Rev. AA" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1296u16 , name : "SM501 VoyagerGX Rev. B" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1808u16 , name : "SM710 LynxEM" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1810u16 , name : "SM712 LynxEM+" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1816u16 , name : "SM718 LynxSE+" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1824u16 , name : "SM720 Lynx3DM" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1840u16 , name : "SM731 Cougar3DR" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1872u16 , name : "SM750" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 1896u16 , name : "SM768" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 2064u16 , name : "SM810 LynxE" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 2065u16 , name : "SM811 LynxE" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 2080u16 , name : "SM820 Lynx3D" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 2320u16 , name : "SM910" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8800u16 , name : "SM2260 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8801u16 , name : "SM2261XT x2 NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8802u16 , name : "SM2262/SM2262EN SSD Controller" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8803u16 , name : "SM2263EN/SM2263XT (DRAM-less) NVMe SSD Controllers" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8808u16 , name : "SM2268XT (DRAM-less) NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 8809u16 , name : "SM2269XT (DRAM-less) NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 9476u16 , name : "SM2504XT NVMe 2.0 SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 9480u16 , name : "SM2508 NVMe 2.0 SSD Controller" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 9992u16 , name : "SM2708 x2 NVMe SD Express SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 33638u16 , name : "SM8366 NVMe SSD Controller [MonTitan]" , subsystems : & [] } , Device { vendor_id : 4719u16 , id : 33672u16 , name : "SM8388 NVMe SSD Controller [MonTitan]" , subsystems : & [] }] }),
        (5508, Vendor { id : 5508u16 , name : "Uniwill Computer Corp" , devices : & [] }),
        (4652, Vendor { id : 4652u16 , name : "Sican GmbH" , devices : & [] }),
        (5199, Vendor { id : 5199u16 , name : "Askey Computer Corp." , devices : & [] }),
        (5822, Vendor { id : 5822u16 , name : "Creatix Polymedia GmbH" , devices : & [] }),
        (4425, Vendor { id : 4425u16 , name : "Win System Corporation" , devices : & [] }),
        (4556, Vendor { id : 4556u16 , name : "Michels & Kleberhoff Computer GmbH" , devices : & [] }),
        (6825, Vendor { id : 6825u16 , name : "Schweitzer Engineering Laboratories" , devices : & [Device { vendor_id : 6825u16 , id : 13u16 , name : "SEL-3390S8 Serial Adapter" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 14u16 , name : "SEL-3390E4 Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 20u16 , name : "SEL-3390T Time and Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 21u16 , name : "SEL-3350 Mainboard" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 22u16 , name : "SEL-3350 Serial Expansion Board" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 23u16 , name : "SEL-3350 GPIO Expansion Board" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 24u16 , name : "SEL-3390E4 Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 25u16 , name : "SEL-2241-2/SEL-3361 Mainboard" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 28u16 , name : "SEL-3390E4 Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 29u16 , name : "SEL-3350 GPIO Expansion Board" , subsystems : & [] } , Device { vendor_id : 6825u16 , id : 30u16 , name : "SEL-3350 Serial Expansion Board" , subsystems : & [] }] }),
        (5185, Vendor { id : 5185u16 , name : "AGIE Ltd" , devices : & [] }),
        (4705, Vendor { id : 4705u16 , name : "Matsushita-Kotobuki Electronics Industries, Ltd." , devices : & [] }),
        (57050, Vendor { id : 57050u16 , name : "XIMEA" , devices : & [Device { vendor_id : 57050u16 , id : 16385u16 , name : "CB or MX camera" , subsystems : & [] } , Device { vendor_id : 57050u16 , id : 16417u16 , name : "MT camera" , subsystems : & [] }] }),
        (6547, Vendor { id : 6547u16 , name : "Innominate Security Technologies AG" , devices : & [] }),
        (4712, Vendor { id : 4712u16 , name : "Tektronix" , devices : & [] }),
        (5207, Vendor { id : 5207u16 , name : "Nuera Communications Inc" , devices : & [] }),
        (5408, Vendor { id : 5408u16 , name : "CHAPLET System Inc" , devices : & [] }),
        (4603, Vendor { id : 4603u16 , name : "Datel Inc" , devices : & [] }),
        (6771, Vendor { id : 6771u16 , name : "Violin Memory, Inc" , devices : & [Device { vendor_id : 6771u16 , id : 1u16 , name : "Mozart [Memory Appliance 1010]" , subsystems : & [] }] }),
        (6280, Vendor { id : 6280u16 , name : "Varisys Ltd" , devices : & [Device { vendor_id : 6280u16 , id : 769u16 , name : "VMFX1 FPGA PMC module" , subsystems : & [] } , Device { vendor_id : 6280u16 , id : 1537u16 , name : "VSM2 dual PMC carrier" , subsystems : & [] } , Device { vendor_id : 6280u16 , id : 1808u16 , name : "VS14x series PowerPC PCI board" , subsystems : & [] } , Device { vendor_id : 6280u16 , id : 1824u16 , name : "VS24x series PowerPC PCI board" , subsystems : & [] }] }),
        (5380, Vendor { id : 5380u16 , name : "KAISER Electronics" , devices : & [] }),
        (4640, Vendor { id : 4640u16 , name : "Ariel Corporation" , devices : & [Device { vendor_id : 4640u16 , id : 4640u16 , name : "AMCC 5933 TMS320C80 DSP/Imaging board" , subsystems : & [] }] }),
        (4378, Vendor { id : 4378u16 , name : "Efficient Networks, Inc" , devices : & [Device { vendor_id : 4378u16 , id : 0u16 , name : "155P-MF1 (FPGA)" , subsystems : & [] } , Device { vendor_id : 4378u16 , id : 2u16 , name : "155P-MF1 (ASIC)" , subsystems : & [] } , Device { vendor_id : 4378u16 , id : 3u16 , name : "ENI-25P ATM" , subsystems : & [SubSystem { subvendor : 4378u16 , subdevice : 0u16 , name : "ENI-25p Miniport ATM Adapter" }] } , Device { vendor_id : 4378u16 , id : 5u16 , name : "SpeedStream (LANAI)" , subsystems : & [SubSystem { subvendor : 4378u16 , subdevice : 1u16 , name : "ENI-3010 ATM" } , SubSystem { subvendor : 4378u16 , subdevice : 9u16 , name : "ENI-3060 ADSL (VPI=0)" } , SubSystem { subvendor : 4378u16 , subdevice : 257u16 , name : "ENI-3010 ATM" } , SubSystem { subvendor : 4378u16 , subdevice : 265u16 , name : "ENI-3060CO ADSL (VPI=0)" } , SubSystem { subvendor : 4378u16 , subdevice : 2057u16 , name : "ENI-3060 ADSL (VPI=0 or 8)" } , SubSystem { subvendor : 4378u16 , subdevice : 2313u16 , name : "ENI-3060CO ADSL (VPI=0 or 8)" } , SubSystem { subvendor : 4378u16 , subdevice : 2569u16 , name : "ENI-3060 ADSL (VPI=<0..15>)" }] } , Device { vendor_id : 4378u16 , id : 7u16 , name : "SpeedStream ADSL" , subsystems : & [SubSystem { subvendor : 4378u16 , subdevice : 4097u16 , name : "ENI-3061 ADSL [ASIC]" }] } , Device { vendor_id : 4378u16 , id : 4128u16 , name : "SpeedStream PCI 10/100 Network Card" , subsystems : & [] } , Device { vendor_id : 4378u16 , id : 4611u16 , name : "SpeedStream 1023 Wireless PCI Adapter" , subsystems : & [] }] }),
        (5850, Vendor { id : 5850u16 , name : "Advantech Co., Ltd." , devices : & [Device { vendor_id : 5850u16 , id : 17u16 , name : "INES GPIB-PCI" , subsystems : & [] }] }),
        (6510, Vendor { id : 6510u16 , name : "PNY" , devices : & [] }),
        (60110, Vendor { id : 60110u16 , name : "Endace Measurement Systems, Ltd" , devices : & [Device { vendor_id : 60110u16 , id : 12544u16 , name : "DAG 3.10 OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 12800u16 , name : "DAG 3.2x OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 12814u16 , name : "DAG 3.2E Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13326u16 , name : "DAG 3.4E Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13342u16 , name : "DAG 3.41E Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13568u16 , name : "DAG 3.5 OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13596u16 , name : "DAG 3.5ECM Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13837u16 , name : "DAG 3.6D DS3" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13838u16 , name : "DAG 3.6E Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 13966u16 , name : "DAG 3.6E Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 14087u16 , name : "DAG 3.7T T1/E1/J1" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 14093u16 , name : "DAG 3.7D DS3/E3" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 14222u16 , name : "DAG 3.7G Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 14336u16 , name : "DAG 3.8S OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16640u16 , name : "DAG 4.10 OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16656u16 , name : "DAG 4.11 OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16928u16 , name : "DAG 4.2 OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16942u16 , name : "DAG 4.2GE Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16944u16 , name : "DAG 4.2S OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 16958u16 , name : "DAG 4.2GE Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17152u16 , name : "DAG 4.3S OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17166u16 , name : "DAG 4.3GE Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17710u16 , name : "DAG 4.5G2 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17742u16 , name : "DAG 4.5G4 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17848u16 , name : "DAG 4.5Z8 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 17854u16 , name : "DAG 4.5Z2 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21006u16 , name : "DAG 5.2X 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21018u16 , name : "DAG 5.2SXA 10G Ethernet/OC-192" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21504u16 , name : "DAG 5.4S-12 OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21505u16 , name : "DAG 5.4SG-48 Gig Ethernet/OC-3/OC-12/OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21514u16 , name : "DAG 5.4GA Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21530u16 , name : "DAG 5.4SA-12 OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 21546u16 , name : "DAG 5.4SGA-48 Gig Ethernet/OC-3/OC-12/OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 24576u16 , name : "DAG 6.0SE 10G Ethernet/OC-192" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 24832u16 , name : "DAG 6.1SE 10G Ethernet/OC-192" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 25088u16 , name : "DAG 6.2SE 10G Ethernet/OC-192" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 28928u16 , name : "DAG 7.1S OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 29696u16 , name : "DAG 7.4S OC-3/OC-12" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 29697u16 , name : "DAG 7.4S48 OC-48" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 29998u16 , name : "DAG 7.5G2 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 30030u16 , name : "DAG 7.5G4 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33024u16 , name : "DAG 8.1X 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33025u16 , name : "DAG 8.1SX 10G Ethernet/OC-192" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33026u16 , name : "DAG 8.1X 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33294u16 , name : "DAG 8.2X 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33295u16 , name : "DAG 8.2X 10G Ethernet (2nd bus)" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 33792u16 , name : "DAG 8.4I Infiniband x4 SDR" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 34048u16 , name : "DAG 8.5I Infiniband x4 DDR" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 37376u16 , name : "DAG 9.2SX2 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 37390u16 , name : "DAG 9.2X2 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 38208u16 , name : "DAG 9.5G4 Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 38223u16 , name : "DAG 9.5G4F Gig Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 41248u16 , name : "DAG 10X2-P 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 41262u16 , name : "DAG 10X2-S 10G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 41280u16 , name : "DAG 10X4-P 10/40G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 41294u16 , name : "DAG 10X4-S 10/40G Ethernet" , subsystems : & [] } , Device { vendor_id : 60110u16 , id : 60110u16 , name : "vDAG virtual device" , subsystems : & [] }] }),
        (4530, Vendor { id : 4530u16 , name : "Eastman Kodak" , devices : & [] }),
        (4698, Vendor { id : 4698u16 , name : "ABB Power Systems" , devices : & [] }),
        (65505, Vendor { id : 65505u16 , name : "Suzhou XiongLi Technology Inc." , devices : & [Device { vendor_id : 65505u16 , id : 53760u16 , name : "XL82101/82102 PCI Express Gigabit Ethernet Controller" , subsystems : & [] }] }),
        (20840, Vendor { id : 20840u16 , name : "Animation Technologies Inc." , devices : & [Device { vendor_id : 20840u16 , id : 768u16 , name : "FlyDVB-S" , subsystems : & [] } , Device { vendor_id : 20840u16 , id : 769u16 , name : "FlyDVB-T" , subsystems : & [] }] }),
        (7667, Vendor { id : 7667u16 , name : "Ethernity Networks" , devices : & [Device { vendor_id : 7667u16 , id : 513u16 , name : "ACE-NIC40 Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA1040" } , SubSystem { subvendor : 7667u16 , subdevice : 2u16 , name : "ENA1044" } , SubSystem { subvendor : 7667u16 , subdevice : 3u16 , name : "ENA1044S" }] } , Device { vendor_id : 7667u16 , id : 514u16 , name : "ACE-NIC50 Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2050F" } , SubSystem { subvendor : 7667u16 , subdevice : 2u16 , name : "ENA2050FS" }] } , Device { vendor_id : 7667u16 , id : 515u16 , name : "ACE-NIC100 Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 0u16 , name : "Maintenance Mode" } , SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2080F" } , SubSystem { subvendor : 7667u16 , subdevice : 2u16 , name : "ENA2080FS" } , SubSystem { subvendor : 7667u16 , subdevice : 3u16 , name : "ENA2100F" } , SubSystem { subvendor : 7667u16 , subdevice : 4u16 , name : "ENA2040F" }] } , Device { vendor_id : 7667u16 , id : 516u16 , name : "ACE-NIC-NID Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA1020Z" } , SubSystem { subvendor : 7667u16 , subdevice : 2u16 , name : "ENA1020ZS" }] } , Device { vendor_id : 7667u16 , id : 517u16 , name : "ACE-NIC250 Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 0u16 , name : "Maintenance Mode" } , SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2250F" }] } , Device { vendor_id : 7667u16 , id : 518u16 , name : "ACE-NIC200 Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 0u16 , name : "Maintenance Mode" } , SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2200F" }] } , Device { vendor_id : 7667u16 , id : 519u16 , name : "ACE-NIC50RN Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 0u16 , name : "Maintenance Mode" } , SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2050RN" }] } , Device { vendor_id : 7667u16 , id : 520u16 , name : "ACE-NIC100RN Programmable Network Accelerator" , subsystems : & [SubSystem { subvendor : 7667u16 , subdevice : 0u16 , name : "Maintenance Mode" } , SubSystem { subvendor : 7667u16 , subdevice : 1u16 , name : "ENA2100RN" }] }] }),
        (5099, Vendor { id : 5099u16 , name : "Hauppauge Computer Works Inc" , devices : & [] }),
        (5569, Vendor { id : 5569u16 , name : "Quantel" , devices : & [] }),
        (4216, Vendor { id : 4216u16 , name : "Cyrix Corporation" , devices : & [Device { vendor_id : 4216u16 , id : 0u16 , name : "5510 [Grappa]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 1u16 , name : "PCI Master" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 2u16 , name : "5520 [Cognac]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 256u16 , name : "5530 Legacy [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 257u16 , name : "5530 SMI [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 258u16 , name : "5530 IDE [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 259u16 , name : "5530 Audio [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 260u16 , name : "5530 Video [Kahlua]" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 1024u16 , name : "ZFMicro PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 1025u16 , name : "ZFMicro Chipset SMI" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 1026u16 , name : "ZFMicro Chipset IDE" , subsystems : & [] } , Device { vendor_id : 4216u16 , id : 1027u16 , name : "ZFMicro Expansion Bus" , subsystems : & [] }] }),
        (7806, Vendor { id : 7806u16 , name : "Pliops" , devices : & [Device { vendor_id : 7806u16 , id : 512u16 , name : "XDP [Extreme Data Processor PRO 1.0]" , subsystems : & [] } , Device { vendor_id : 7806u16 , id : 36916u16 , name : "XDP [Extreme Data Processor FLEX 1.0]" , subsystems : & [] }] }),
        (43535, Vendor { id : 43535u16 , name : "iTuner" , devices : & [] }),
        (9995, Vendor { id : 9995u16 , name : "Xantel Corporation" , devices : & [] }),
        (4224, Vendor { id : 4224u16 , name : "Contaq Microsystems" , devices : & [Device { vendor_id : 4224u16 , id : 1536u16 , name : "82C599" , subsystems : & [] } , Device { vendor_id : 4224u16 , id : 50833u16 , name : "Cypress CY82C691" , subsystems : & [] } , Device { vendor_id : 4224u16 , id : 50835u16 , name : "82c693" , subsystems : & [] }] }),
        (4481, Vendor { id : 4481u16 , name : "Telmatics International" , devices : & [] }),
        (4426, Vendor { id : 4426u16 , name : "VMIC" , devices : & [Device { vendor_id : 4426u16 , id : 21861u16 , name : "GE-IP PCI5565,PMC5565 Reflective Memory Node" , subsystems : & [] } , Device { vendor_id : 4426u16 , id : 21881u16 , name : "VMIPCI-5579 (Reflective Memory Card)" , subsystems : & [] } , Device { vendor_id : 4426u16 , id : 21895u16 , name : "VMIPCI-5587 (Reflective Memory Card)" , subsystems : & [] } , Device { vendor_id : 4426u16 , id : 25860u16 , name : "VMIC PCI 7755 FPGA" , subsystems : & [] } , Device { vendor_id : 4426u16 , id : 30087u16 , name : "VMIVME-7587" , subsystems : & [] }] }),
        (4114, Vendor { id : 4114u16 , name : "Micronics Computers Inc" , devices : & [] }),
        (4899, Vendor { id : 4899u16 , name : "Dome Inc" , devices : & [] }),
        (4896, Vendor { id : 4896u16 , name : "Crypto AG" , devices : & [] }),
        (56080, Vendor { id : 56080u16 , name : "Diablo Technologies" , devices : & [] }),
        (7839, Vendor { id : 7839u16 , name : "Lynxi Technologies Co., Ltd." , devices : & [] }),
        (4434, Vendor { id : 4434u16 , name : "Megatek" , devices : & [] }),
        (6131, Vendor { id : 6131u16 , name : "RDC Semiconductor, Inc." , devices : & [Device { vendor_id : 6131u16 , id : 4112u16 , name : "R1010 IDE Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 4113u16 , name : "R1011 IDE Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 4114u16 , name : "R1012 IDE Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 4145u16 , name : "PCI/PCI-X to PCI-E Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 4208u16 , name : "CAN Bus Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 4913u16 , name : "Motion Control Interface" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 6448u16 , name : "Hybrid Function Control Register" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 8208u16 , name : "RDC M2010 VGA-compatible graphics adapter" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 8210u16 , name : "M2012/R3308 VGA-compatible graphics adapter" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 8213u16 , name : "RDC M2015 VGA-compatible graphics adapter" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 12304u16 , name : "R3010 HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24593u16 , name : "R6011 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24595u16 , name : "R6013 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24608u16 , name : "R6020 North Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24609u16 , name : "R6021 Host Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24610u16 , name : "R6022 Host Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24611u16 , name : "R6023 Host Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24613u16 , name : "R6025 Host Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24614u16 , name : "R6026 Host Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24624u16 , name : "R6030 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24625u16 , name : "R6031 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24629u16 , name : "R6035 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24630u16 , name : "R6036 ISA Bridge" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24640u16 , name : "R6040 MAC Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24672u16 , name : "R6060 USB 1.1 Controller" , subsystems : & [] } , Device { vendor_id : 6131u16 , id : 24673u16 , name : "R6061 USB 2.0 Controller" , subsystems : & [] }] }),
        (4137, Vendor { id : 4137u16 , name : "Siemens Nixdorf IS" , devices : & [] }),
        (6142, Vendor { id : 6142u16 , name : "InProComm Inc." , devices : & [Device { vendor_id : 6142u16 , id : 8480u16 , name : "IPN 2120 802.11b" , subsystems : & [SubSystem { subvendor : 5943u16 , subdevice : 32u16 , name : "WMP11 v4 802.11b Wireless-B PCI Adapter" }] } , Device { vendor_id : 6142u16 , id : 8736u16 , name : "IPN 2220 802.11g" , subsystems : & [SubSystem { subvendor : 4436u16 , subdevice : 820u16 , name : "WLI2-CB-G54L" } , SubSystem { subvendor : 5224u16 , subdevice : 773u16 , name : "T60N871 802.11g Mini PCI Wireless Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 41u16 , name : "WPC54G v4 802.11g Wireless-G Notebook Adapter" }] }] }),
        (5471, Vendor { id : 5471u16 , name : "Perle Systems Ltd" , devices : & [] }),
        (5611, Vendor { id : 5611u16 , name : "DResearch Digital Media Systems GmbH" , devices : & [] }),
        (5598, Vendor { id : 5598u16 , name : "Malleable Technologies Inc" , devices : & [] }),
        (4097, Vendor { id : 4097u16 , name : "Kolter Electronic" , devices : & [Device { vendor_id : 4097u16 , id : 16u16 , name : "PCI 1616 Measurement card with 32 digital I/O lines" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 17u16 , name : "OPTO-PCI Opto-Isolated digital I/O board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 18u16 , name : "PCI-AD/DA Analogue I/O board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 19u16 , name : "PCI-OPTO-RELAIS Digital I/O board with relay outputs" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 20u16 , name : "PCI-Counter/Timer Counter Timer board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 21u16 , name : "PCI-DAC416 Analogue output board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 22u16 , name : "PCI-MFB Analogue I/O board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 23u16 , name : "PROTO-3 PCI Prototyping board" , subsystems : & [] } , Device { vendor_id : 4097u16 , id : 37120u16 , name : "INI-9100/9100W SCSI Host" , subsystems : & [] }] }),
        (22605, Vendor { id : 22605u16 , name : "AuzenTech Co., Ltd." , devices : & [] }),
        (13329, Vendor { id : 13329u16 , name : "Quantum Designs (H.K.) Inc" , devices : & [] }),
        (25097, Vendor { id : 25097u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (5164, Vendor { id : 5164u16 , name : "Minton Optic Industry Co Ltd" , devices : & [] }),
        (5833, Vendor { id : 5833u16 , name : "EONIC B.V. The Netherlands" , devices : & [] }),
        (6523, Vendor { id : 6523u16 , name : "JMicron Technology Corp." , devices : & [Device { vendor_id : 6523u16 , id : 592u16 , name : "JMC250 PCI Express Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 608u16 , name : "JMC260 PCI Express Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 872u16 , name : "JMB368 IDE controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 1413u16 , name : "JMB58x AHCI SATA controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9056u16 , name : "JMB360 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9057u16 , name : "JMB361 AHCI/IDE" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29237u16 , name : "P965 Neo MS-7235 mainboard" }] } , Device { vendor_id : 6523u16 , id : 9058u16 , name : "JMB362 SATA Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33888u16 , name : "P8P67 Deluxe Motherboard" }] } , Device { vendor_id : 6523u16 , id : 9059u16 , name : "JMB363 SATA/IDE Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33252u16 , name : "P5B [JMB363]" } , SubSystem { subvendor : 5208u16 , subdevice : 45056u16 , name : "Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 9059u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 6523u16 , id : 9060u16 , name : "JMB364 AHCI Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9061u16 , name : "JMB365 AHCI/IDE" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9062u16 , name : "JMB366 AHCI/IDE" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9064u16 , name : "JMB368 IDE controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9065u16 , name : "JMB369 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9088u16 , name : "IEEE 1394 Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9089u16 , name : "Standard SD Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9090u16 , name : "SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9091u16 , name : "MS Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9092u16 , name : "xD Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9094u16 , name : "Standard SD Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9095u16 , name : "SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9096u16 , name : "MS Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9097u16 , name : "xD Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9105u16 , name : "Standard SD Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9106u16 , name : "SD/MMC Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9107u16 , name : "MS Host Controller" , subsystems : & [] } , Device { vendor_id : 6523u16 , id : 9108u16 , name : "xD Host Controller" , subsystems : & [] }] }),
        (5232, Vendor { id : 5232u16 , name : "Bay Networks" , devices : & [] }),
        (6487, Vendor { id : 6487u16 , name : "Freescale Semiconductor Inc" , devices : & [Device { vendor_id : 6487u16 , id : 18u16 , name : "MPC8548E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 19u16 , name : "MPC8548" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 20u16 , name : "MPC8543E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 21u16 , name : "MPC8543" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 24u16 , name : "MPC8547E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 25u16 , name : "MPC8545E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 26u16 , name : "MPC8545" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 32u16 , name : "MPC8568E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 33u16 , name : "MPC8568" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 34u16 , name : "MPC8567E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 35u16 , name : "MPC8567" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 48u16 , name : "MPC8533E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 49u16 , name : "MPC8533" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 50u16 , name : "MPC8544E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 51u16 , name : "MPC8544" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 64u16 , name : "MPC8572E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 65u16 , name : "MPC8572" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 80u16 , name : "MPC8536E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 81u16 , name : "MPC8536" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 82u16 , name : "MPC8535E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 83u16 , name : "MPC8535" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 96u16 , name : "MPC8569" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 97u16 , name : "MPC8569E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 112u16 , name : "P2020E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 113u16 , name : "P2020" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 120u16 , name : "P2010E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 121u16 , name : "P2010" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 128u16 , name : "MPC8349E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 129u16 , name : "MPC8349" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 130u16 , name : "MPC8347E TBGA" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 131u16 , name : "MPC8347 TBGA" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 132u16 , name : "MPC8347E PBGA" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 16500u16 , name : "SIMATIC NET CP 1628" }] } , Device { vendor_id : 6487u16 , id : 133u16 , name : "MPC8347 PBGA" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 16454u16 , name : "SIMATIC NET CP 1623" }] } , Device { vendor_id : 6487u16 , id : 134u16 , name : "MPC8343E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 135u16 , name : "MPC8343" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 180u16 , name : "MPC8315E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 182u16 , name : "MPC8314E" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 4353u16 , name : "Bigfoot Killer Xeno Pro Gigabit Ethernet Controller" }] } , Device { vendor_id : 6487u16 , id : 194u16 , name : "MPC8379E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 195u16 , name : "MPC8379" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 196u16 , name : "MPC8378E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 197u16 , name : "MPC8378" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 198u16 , name : "MPC8377E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 199u16 , name : "MPC8377" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 256u16 , name : "P1020E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 257u16 , name : "P1020" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 258u16 , name : "P1021E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 259u16 , name : "P1021" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 264u16 , name : "P1011E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 265u16 , name : "P1011" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 266u16 , name : "P1012E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 267u16 , name : "P1012" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 272u16 , name : "P1022E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 273u16 , name : "P1022" , subsystems : & [SubSystem { subvendor : 7295u16 , subdevice : 20992u16 , name : "EB5200" }] } , Device { vendor_id : 6487u16 , id : 280u16 , name : "P1013E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 281u16 , name : "P1013" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 296u16 , name : "P1010" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1024u16 , name : "P4080E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1025u16 , name : "P4080" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1032u16 , name : "P4040E" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1033u16 , name : "P4040" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1055u16 , name : "P3041" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1056u16 , name : "QorIQ P5020 Communications Processor with security engine" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1057u16 , name : "QorIQ P5020 Communications Processor without security engine" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1064u16 , name : "QorIQ P5010 Communications Processor with security engine" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1065u16 , name : "QorIQ P5010 Communications Processor without security engine" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1088u16 , name : "T4240 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1089u16 , name : "T4240 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1094u16 , name : "T4160 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 1095u16 , name : "T4160 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2080u16 , name : "T1040 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2081u16 , name : "T1040 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2084u16 , name : "T1042 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2085u16 , name : "T1042 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2088u16 , name : "T1020 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2089u16 , name : "T1020 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2092u16 , name : "T1022 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2093u16 , name : "T1022 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2096u16 , name : "T2080 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2097u16 , name : "T2080 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2104u16 , name : "T2081 with security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 2105u16 , name : "T2081 without security" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 22540u16 , name : "MPC5121e" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 28688u16 , name : "MPC8641 PCI Host Bridge" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 28689u16 , name : "MPC8641D PCI Host Bridge" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 28696u16 , name : "MPC8610" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 33216u16 , name : "LS1046A PCI Express Bridge" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 49158u16 , name : "MPC8308" , subsystems : & [SubSystem { subvendor : 6742u16 , subdevice : 4609u16 , name : "Bigfoot Killer E2100 Gigabit Ethernet Controller" }] } , Device { vendor_id : 6487u16 , id : 64514u16 , name : "RedStone" , subsystems : & [] } , Device { vendor_id : 6487u16 , id : 64515u16 , name : "CFI" , subsystems : & [] }] }),
        (5633, Vendor { id : 5633u16 , name : "Tenta Technology" , devices : & [] }),
        (7695, Vendor { id : 7695u16 , name : "KIOXIA Corporation" , devices : & [Device { vendor_id : 7695u16 , id : 1u16 , name : "NVMe SSD Controller BG4 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 7u16 , name : "NVMe SSD Controller Cx6" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8312u16 , name : "DC NVMe CD6 RI 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8313u16 , name : "DC NVMe CD6 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8314u16 , name : "DC NVMe CD6 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8315u16 , name : "DC NVMe CD6 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8316u16 , name : "DC NVMe CD6 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8318u16 , name : "Dell Ent NVMe CM6 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8319u16 , name : "Dell Ent NVMe CM6 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8320u16 , name : "Dell Ent NVMe CM6 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8321u16 , name : "Dell Ent NVMe CM6 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8324u16 , name : "Dell Ent NVMe CM6 MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8325u16 , name : "Dell Ent NVMe CM6 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8326u16 , name : "Dell Ent NVMe CM6 MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8458u16 , name : "Dell Ent NVMe FIPS CM6 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8459u16 , name : "Dell Ent NVMe FIPS CM6 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8460u16 , name : "Dell Ent NVMe FIPS CM6 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8461u16 , name : "Dell Ent NVMe FIPS CM6 RI15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8462u16 , name : "Dell Ent NVMe FIPS CM6 MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8463u16 , name : "Dell Ent NVMe FIPS CM6 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8464u16 , name : "Dell Ent NVMe FIPS CM6 MU 6.4TB" } , SubSystem { subvendor : 7695u16 , subdevice : 1u16 , name : "Generic NVMe CM6" }] } , Device { vendor_id : 7695u16 , id : 8u16 , name : "RD500/Exceria Plus/Exceria Plus G2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 9u16 , name : "NVMe SSD" , subsystems : & [SubSystem { subvendor : 7695u16 , subdevice : 1u16 , name : "Toshiba RC500 Series NVMe SSD" } , SubSystem { subvendor : 7695u16 , subdevice : 50u16 , name : "KIOXIA EXCERIA RC10 Series NVMe SSD" }] } , Device { vendor_id : 7695u16 , id : 12u16 , name : "NVMe SSD Controller BG5 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 13u16 , name : "NVMe SSD Controller XG7" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 16u16 , name : "NVMe SSD Controller XG8" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 17u16 , name : "NVMe SSD Controller CD7" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8585u16 , name : "DC NVMe SED CD7 RI 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8586u16 , name : "DC NVMe CD7 RI 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8587u16 , name : "DC NVMe SED CD7 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8588u16 , name : "DC NVMe CD7 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8589u16 , name : "DC NVMe SED CD7 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8590u16 , name : "DC NVMe CD7 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8591u16 , name : "DC NVMe SED CD7 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8592u16 , name : "DC NVMe CD7 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8593u16 , name : "DC NVMe SED CD7 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8594u16 , name : "DC NVMe CD7 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8595u16 , name : "NVMe CD7 E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8596u16 , name : "NVMe CD7 E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8597u16 , name : "NVMe CD7 E3.S 7.68TB" }] } , Device { vendor_id : 7695u16 , id : 19u16 , name : "NVMe SSD Controller CM7" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8749u16 , name : "Ent NVMe CM7 FIPS U.2 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8750u16 , name : "Ent NVMe CM7 FIPS U.2 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8751u16 , name : "Ent NVMe CM7 FIPS U.2 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8752u16 , name : "Ent NVMe CM7 FIPS U.2 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8753u16 , name : "Ent NVMe CM7 FIPS U.2 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8754u16 , name : "Ent NVMe CM7 U.2 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8755u16 , name : "Ent NVMe CM7 U.2 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8756u16 , name : "Ent NVMe CM7 U.2 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8757u16 , name : "Ent NVMe CM7 U.2 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8758u16 , name : "Ent NVMe CM7 U.2 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8759u16 , name : "Ent NVMe CM7 FIPS U.2 MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8760u16 , name : "Ent NVMe CM7 FIPS U.2 MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8761u16 , name : "Ent NVMe CM7 FIPS U.2 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8762u16 , name : "Ent NVMe CM7 FIPS U.2 MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8763u16 , name : "Ent NVMe CM7 U.2 MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8764u16 , name : "Ent NVMe CM7 U.2 MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8765u16 , name : "Ent NVMe CM7 U.2 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8766u16 , name : "Ent NVMe CM7 U.2 MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8844u16 , name : "Ent NVMe CM7 U.2 MU 6.4TB" }] } , Device { vendor_id : 7695u16 , id : 20u16 , name : "NVMe SSD Controller CM7 EDSFF" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8767u16 , name : "Ent NVMe CM7 FIPS E3.S RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8768u16 , name : "Ent NVMe CM7 FIPS E3.S RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8769u16 , name : "Ent NVMe CM7 FIPS E3.S RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8770u16 , name : "Ent NVMe CM7 E3.S RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8771u16 , name : "Ent NVMe CM7 E3.S RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8772u16 , name : "Ent NVMe CM7 E3.S RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8773u16 , name : "Ent NVMe CM7 E3.S RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8774u16 , name : "Ent NVMe CM7 E3.S MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8775u16 , name : "Ent NVMe CM7 E3.S MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8776u16 , name : "Ent NVMe CM7 E3.S MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8777u16 , name : "Ent NVMe CM7 E3.S MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8887u16 , name : "Ent NVMe CM7 FIPS E3.S MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8888u16 , name : "Ent NVMe CM7 FIPS E3.S MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8889u16 , name : "Ent NVMe CM7 FIPS E3.S MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8890u16 , name : "Ent NVMe CM7 FIPS E3.S MU 12.8TB" }] } , Device { vendor_id : 7695u16 , id : 24u16 , name : "Exceria Pro NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 26u16 , name : "NVMe SSD Controller BG6 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 27u16 , name : "NVMe SSD Controller EG6 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 30u16 , name : "NVMe SSD Controller LD2-L" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 31u16 , name : "NVMe SSD Controller CD8" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8739u16 , name : "DC NVMe CD8 U.2 SED 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8740u16 , name : "DC NVMe CD8 U.2 SED 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8741u16 , name : "DC NVMe CD8 U.2 SED 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8742u16 , name : "DC NVMe CD8 U.2 SED 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8743u16 , name : "DC NVMe CD8 U.2 SED 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8744u16 , name : "DC NVMe CD8 U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8745u16 , name : "DC NVMe CD8 U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8746u16 , name : "DC NVMe CD8 U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8747u16 , name : "DC NVMe CD8 U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8748u16 , name : "DC NVMe CD8 U.2 960GB" }] } , Device { vendor_id : 7695u16 , id : 37u16 , name : "NVMe SSD Controller CM7 2.5\"" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8749u16 , name : "Ent NVMe CM7 FIPS U.2 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8750u16 , name : "Ent NVMe CM7 FIPS U.2 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8751u16 , name : "Ent NVMe CM7 FIPS U.2 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8752u16 , name : "Ent NVMe CM7 FIPS U.2 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8753u16 , name : "Ent NVMe CM7 FIPS U.2 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8754u16 , name : "Ent NVMe CM7 U.2 RI 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8755u16 , name : "Ent NVMe CM7 U.2 RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8756u16 , name : "Ent NVMe CM7 U.2 RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8757u16 , name : "Ent NVMe CM7 U.2 RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8758u16 , name : "Ent NVMe CM7 U.2 RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8759u16 , name : "Ent NVMe CM7 FIPS U.2 MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8760u16 , name : "Ent NVMe CM7 FIPS U.2 MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8761u16 , name : "Ent NVMe CM7 FIPS U.2 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8762u16 , name : "Ent NVMe CM7 FIPS U.2 MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8763u16 , name : "Ent NVMe CM7 U.2 MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8764u16 , name : "Ent NVMe CM7 U.2 MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8765u16 , name : "Ent NVMe CM7 U.2 MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8766u16 , name : "Ent NVMe CM7 U.2 MU 1.6TB" }] } , Device { vendor_id : 7695u16 , id : 42u16 , name : "Exceria Plus G3 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 43u16 , name : "NVMe SSD Controller CD8P" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 44u16 , name : "NVMe SSD Controller CD8P EDSFF" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8895u16 , name : "DC NVMe CD8P E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8896u16 , name : "DC NVMe CD8P E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8897u16 , name : "DC NVMe CD8P E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8898u16 , name : "DC NVMe CD8P E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8903u16 , name : "DC NVMe CD8P E3.S MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8904u16 , name : "DC NVMe CD8P E3.S MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8905u16 , name : "DC NVMe CD8P E3.S MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8906u16 , name : "DC NVMe CD8P E3.S MU 1.6TB" }] } , Device { vendor_id : 7695u16 , id : 46u16 , name : "NVMe SSD Controller XD8" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 47u16 , name : "NVMe SSD Controller BG7 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 48u16 , name : "NVMe SSD Controller EG7" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 49u16 , name : "PCIe 5.0 NVMe SSD Controller XG10" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 50u16 , name : "PCIe 4.0 NVMe SSD Controller XG10d" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 51u16 , name : "Exceria Plus G4 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 52u16 , name : "CM9-based E3.S NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9231u16 , name : "KCM9FRJE1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9232u16 , name : "KCM9FRJE3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9233u16 , name : "KCM9FRJE7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9234u16 , name : "KCM9FRJE15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9235u16 , name : "KCM9FRJE30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9236u16 , name : "KCM9XRJE1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9237u16 , name : "KCM9XRJE3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9238u16 , name : "KCM9XRJE7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9239u16 , name : "KCM9XRJE15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9240u16 , name : "KCM9XRJE30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9241u16 , name : "KCM9FVJE1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9242u16 , name : "KCM9FVJE3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9243u16 , name : "KCM9FVJE6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9244u16 , name : "KCM9FVJE12T8" } , SubSystem { subvendor : 4136u16 , subdevice : 9245u16 , name : "KCM9XVJE1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9246u16 , name : "KCM9XVJE3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9247u16 , name : "KCM9XVJE6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9248u16 , name : "KCM9XVJE12T8" }] } , Device { vendor_id : 7695u16 , id : 53u16 , name : "CM9-based U3 NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9211u16 , name : "KCM9FRUL1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9212u16 , name : "KCM9FRUL3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9213u16 , name : "KCM9FRUL7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9214u16 , name : "KCM9FRUL15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9215u16 , name : "KCM9FRUL30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9216u16 , name : "KCM9FRUL61T4" } , SubSystem { subvendor : 4136u16 , subdevice : 9217u16 , name : "KCM9XRUL1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9218u16 , name : "KCM9XRUL3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9219u16 , name : "KCM9XRUL7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9220u16 , name : "KCM9XRUL15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9221u16 , name : "KCM9XRUL30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9222u16 , name : "KCM9XRUL61T4" } , SubSystem { subvendor : 4136u16 , subdevice : 9223u16 , name : "KCM9FVUL1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9224u16 , name : "KCM9FVUL3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9225u16 , name : "KCM9FVUL6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9226u16 , name : "KCM9FVUL12T8" } , SubSystem { subvendor : 4136u16 , subdevice : 9227u16 , name : "KCM9XVUL1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9228u16 , name : "KCM9XVUL3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9229u16 , name : "KCM9XVUL6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9230u16 , name : "KCM9XVUL12T8" }] } , Device { vendor_id : 7695u16 , id : 54u16 , name : "NVMe SSD Controller CD9P U.2" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9173u16 , name : "KCD9DPUG1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9174u16 , name : "KCD9DPUG3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9175u16 , name : "KCD9DPUG7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9176u16 , name : "KCD9DPUG15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9177u16 , name : "KCD9DPUG30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9178u16 , name : "KCD9DPUG61T4" } , SubSystem { subvendor : 4136u16 , subdevice : 9179u16 , name : "KCD9XPUG1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9180u16 , name : "KCD9XPUG3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9181u16 , name : "KCD9XPUG7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9182u16 , name : "KCD9XPUG15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9183u16 , name : "KCD9XPUG30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9184u16 , name : "KCD9XPUG61T4" } , SubSystem { subvendor : 4136u16 , subdevice : 9185u16 , name : "KCD9DPUG1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9186u16 , name : "KCD9DPUG3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9187u16 , name : "KCD9DPUG6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9188u16 , name : "KCD9DPUG12T8" } , SubSystem { subvendor : 4136u16 , subdevice : 9189u16 , name : "KCD9XPUG1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9190u16 , name : "KCD9XPUG3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9191u16 , name : "KCD9XPUG6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9192u16 , name : "KCD9XPUG12T8" }] } , Device { vendor_id : 7695u16 , id : 55u16 , name : "NVMe SSD Controller CD9P E3.S" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9193u16 , name : "KCD9DPJE1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9194u16 , name : "KCD9DPJE3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9195u16 , name : "KCD9DPJE7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9196u16 , name : "KCD9DPJE15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9197u16 , name : "KCD9DPJE30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9198u16 , name : "KCD9XPJE1T92" } , SubSystem { subvendor : 4136u16 , subdevice : 9199u16 , name : "KCD9XPJE3T84" } , SubSystem { subvendor : 4136u16 , subdevice : 9200u16 , name : "KCD9XPJE7T68" } , SubSystem { subvendor : 4136u16 , subdevice : 9201u16 , name : "KCD9XPJE15T3" } , SubSystem { subvendor : 4136u16 , subdevice : 9202u16 , name : "KCD9XPJE30T7" } , SubSystem { subvendor : 4136u16 , subdevice : 9203u16 , name : "KCD9DPJE1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9204u16 , name : "KCD9DPJE3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9205u16 , name : "KCD9DPJE6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9206u16 , name : "KCD9DPJE12T8" } , SubSystem { subvendor : 4136u16 , subdevice : 9207u16 , name : "KCD9XPJE1T60" } , SubSystem { subvendor : 4136u16 , subdevice : 9208u16 , name : "KCD9XPJE3T20" } , SubSystem { subvendor : 4136u16 , subdevice : 9209u16 , name : "KCD9XPJE6T40" } , SubSystem { subvendor : 4136u16 , subdevice : 9210u16 , name : "KCD9XPJE12T8" }] } , Device { vendor_id : 7695u16 , id : 58u16 , name : "NVMe SSD Controller BG8 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 59u16 , name : "Exceria Basic NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 61u16 , name : "LC9 E3.L NVMe SSD" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9295u16 , name : "RLC9GZV245T" } , SubSystem { subvendor : 4136u16 , subdevice : 9296u16 , name : "RLC9CZV245T" }] } , Device { vendor_id : 7695u16 , id : 62u16 , name : "LC9 E3.S NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 63u16 , name : "LC9 U.2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7695u16 , id : 66u16 , name : "NVMe SSD Controller LD4" , subsystems : & [] }] }),
        (7410, Vendor { id : 7410u16 , name : "ZTE Corp." , devices : & [Device { vendor_id : 7410u16 , id : 32822u16 , name : "DH Switch" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32832u16 , name : "NEO X510 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32833u16 , name : "NEO X510 SRIOV VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32834u16 , name : "NX I512 VDPA VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32836u16 , name : "NEO X510 MPF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32837u16 , name : "NX I512 BOND PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32838u16 , name : "NX I512 OVS PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32839u16 , name : "NEO X510 BOND PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32840u16 , name : "NEO X510 OVS PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32841u16 , name : "NX E312-02R00 Ethernet Controller [RDMA]" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32842u16 , name : "NEO X512 NOF PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32843u16 , name : "NEO X512 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32844u16 , name : "NEO X512 INITIATOR1 PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32845u16 , name : "NEO X512 INITIATOR2 PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32846u16 , name : "NX I512 UPF PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32847u16 , name : "NX I512 UPF VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32864u16 , name : "NX E312-02R00 Ethernet Controller Virtual Function [RDMA]" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32865u16 , name : "NX E310-01N00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32866u16 , name : "NX E310-01N00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32867u16 , name : "NX I510 BOND PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32868u16 , name : "NX I510 OVS PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32869u16 , name : "NX I510 VDPA VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32870u16 , name : "NX I511 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32871u16 , name : "NX I511 OVS PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32872u16 , name : "NX I511 VDPA VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32875u16 , name : "NEO X512 RDMA PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32876u16 , name : "NEO X512 RDMA VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32877u16 , name : "NX I512 RDMA PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32878u16 , name : "NX I512 RDMA VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32879u16 , name : "NX I512 UPF BOND PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32893u16 , name : "NX E312S-01R00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32894u16 , name : "NX E316-01R00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32895u16 , name : "NX E316-01R00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32896u16 , name : "NX E311-01N00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32897u16 , name : "NX E311-01N00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32898u16 , name : "NX I511 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32899u16 , name : "NX I511 SRIOV VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32900u16 , name : "NX E310-02R00 Ethernet Controller [RDMA]" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32901u16 , name : "NX E310-02R00 Ethernet Controller Virtual Function [RDMA]" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32902u16 , name : "NX I510 SRIOV SEC PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32903u16 , name : "NX I510 SRIOV SEC VF Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32904u16 , name : "NX E312S-01R00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32905u16 , name : "NEO X512 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32906u16 , name : "NEO X512 SRIOV PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32928u16 , name : "NX E312-01N00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32929u16 , name : "NX E312-01N00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32930u16 , name : "NX E312SD-01R00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32931u16 , name : "NX E312SD-01R00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32932u16 , name : "NX I512 OFFLOAD PF Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32939u16 , name : "NX E318-01R00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32940u16 , name : "NX E318-01R00 Ethernet Controller Virtual Function" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32950u16 , name : "NX E310S-01N00 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 7410u16 , id : 32951u16 , name : "NX E310S-01N00 Ethernet Controller Virtual Function" , subsystems : & [] }] }),
        (21580, Vendor { id : 21580u16 , name : "Teralogic Inc" , devices : & [Device { vendor_id : 21580u16 , id : 848u16 , name : "TL880-based HDTV/ATSC tuner" , subsystems : & [] }] }),
        (4986, Vendor { id : 4986u16 , name : "Mark of the Unicorn Inc" , devices : & [Device { vendor_id : 4986u16 , id : 1u16 , name : "PCI-324 Audiowire Interface" , subsystems : & [] }] }),
        (5059, Vendor { id : 5059u16 , name : "Janz Computer AG" , devices : & [] }),
        (32749, Vendor { id : 32749u16 , name : "PowerTV" , devices : & [] }),
        (736, Vendor { id : 736u16 , name : "XFX Pine Group Inc. (Wrong ID)" , devices : & [] }),
        (6282, Vendor { id : 6282u16 , name : "Ample Communications, Inc" , devices : & [] }),
        (5260, Vendor { id : 5260u16 , name : "Tul Corporation / PowerColor" , devices : & [] }),
        (49406, Vendor { id : 49406u16 , name : "Motion Engineering, Inc." , devices : & [] }),
        (65534, Vendor { id : 65534u16 , name : "VMWare Inc (temporary ID)" , devices : & [Device { vendor_id : 65534u16 , id : 1808u16 , name : "Virtual SVGA" , subsystems : & [] }] }),
        (16730, Vendor { id : 16730u16 , name : "Auzentech, Inc." , devices : & [] }),
        (5462, Vendor { id : 5462u16 , name : "PLDA" , devices : & [Device { vendor_id : 5462u16 , id : 4352u16 , name : "PCI Express Core Reference Design" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4367u16 , name : "PCI Express Core Reference Design Virtual Function" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4368u16 , name : "XpressRich Reference Design" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4369u16 , name : "XpressRich-AXI Ref Design" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4370u16 , name : "QuickPCIe" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4371u16 , name : "XpressSwitch" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4372u16 , name : "Inspector" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4373u16 , name : "XpressLINK Ref Design" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 4374u16 , name : "XpressLINK-SOC Ref Design" , subsystems : & [] } , Device { vendor_id : 5462u16 , id : 48640u16 , name : "PCI Express Bridge" , subsystems : & [] }] }),
        (26896, Vendor { id : 26896u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (5528, Vendor { id : 5528u16 , name : "Granite Microsystems" , devices : & [] }),
        (8373, Vendor { id : 8373u16 , name : "Shanghai StarFive Technology Co., Ltd." , devices : & [] }),
        (60888, Vendor { id : 60888u16 , name : "ARK Logic Inc" , devices : & [Device { vendor_id : 60888u16 , id : 41105u16 , name : "1000PV [Stingray]" , subsystems : & [] } , Device { vendor_id : 60888u16 , id : 41113u16 , name : "2000PV [Stingray]" , subsystems : & [] } , Device { vendor_id : 60888u16 , id : 41121u16 , name : "2000MT" , subsystems : & [] } , Device { vendor_id : 60888u16 , id : 41129u16 , name : "2000MI" , subsystems : & [] }] }),
        (5196, Vendor { id : 5196u16 , name : "Catalina Research Inc" , devices : & [] }),
        (7897, Vendor { id : 7897u16 , name : "Myrtle.ai" , devices : & [] }),
        (7396, Vendor { id : 7396u16 , name : "Exablaze" , devices : & [Device { vendor_id : 7396u16 , id : 1u16 , name : "ExaNIC X4" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 2u16 , name : "ExaNIC X2" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 3u16 , name : "ExaNIC X10" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 4u16 , name : "ExaNIC X10-GM" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 5u16 , name : "ExaNIC X40" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 6u16 , name : "ExaNIC X10-HPT" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 7u16 , name : "ExaNIC X40" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 8u16 , name : "ExaNIC V5P" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 9u16 , name : "ExaNIC X25" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 10u16 , name : "ExaNIC X100" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 11u16 , name : "ExaNIC V9P" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 12u16 , name : "ExaNIC V9P-3" , subsystems : & [] } , Device { vendor_id : 7396u16 , id : 256u16 , name : "ExaDISK FX1" , subsystems : & [] }] }),
        (42791, Vendor { id : 42791u16 , name : "3Com Corporation" , devices : & [Device { vendor_id : 42791u16 , id : 19u16 , name : "3CRPAG175 Wireless PC Card" , subsystems : & [] } , Device { vendor_id : 42791u16 , id : 26627u16 , name : "3CRDAG675B Wireless 11a/b/g Adapter" , subsystems : & [] }] }),
        (7636, Vendor { id : 7636u16 , name : "Swissbit AG" , devices : & [Device { vendor_id : 7636u16 , id : 16u16 , name : "N-10m2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 22u16 , name : "N-16" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 32u16 , name : "EN-20 BGA NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 48u16 , name : "N-30m2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 54u16 , name : "N-36m2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 4608u16 , name : "D1200 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 12288u16 , name : "N3000 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 12290u16 , name : "N3002 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 13826u16 , name : "N3602 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 16689u16 , name : "A1x NVMe SSD Series" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 17458u16 , name : "D2x NVMe SSD Series" , subsystems : & [] } , Device { vendor_id : 7636u16 , id : 20992u16 , name : "N5200 NVMe SSD" , subsystems : & [] }] }),
        (8032, Vendor { id : 8032u16 , name : "Accelecom" , devices : & [Device { vendor_id : 8032u16 , id : 1u16 , name : "XELE-NIC 25K5" , subsystems : & [] } , Device { vendor_id : 8032u16 , id : 84u16 , name : "XELE-NIC 25K5" , subsystems : & [] }] }),
        (6997, Vendor { id : 6997u16 , name : "NetUP Inc." , devices : & [Device { vendor_id : 6997u16 , id : 6390u16 , name : "Dual DVB Universal CI card" , subsystems : & [] } , Device { vendor_id : 6997u16 , id : 6391u16 , name : "Dual DVB Universal CI card rev 1.4" , subsystems : & [] } , Device { vendor_id : 6997u16 , id : 10796u16 , name : "Dual DVB-S2-CI card" , subsystems : & [] } , Device { vendor_id : 6997u16 , id : 58084u16 , name : "Dual DVB-T/C-CI RF card" , subsystems : & [] } , Device { vendor_id : 6997u16 , id : 58868u16 , name : "MPEG2 and H264 Encoder-Transcoder" , subsystems : & [] } , Device { vendor_id : 6997u16 , id : 61892u16 , name : "Dual ASI-RX/TX-CI card" , subsystems : & [] }] }),
        (5431, Vendor { id : 5431u16 , name : "DATALEX COMMUNCATIONS" , devices : & [] }),
        (149, Vendor { id : 149u16 , name : "Silicon Image, Inc. (Wrong ID)" , devices : & [Device { vendor_id : 149u16 , id : 1664u16 , name : "Ultra ATA/133 IDE RAID CONTROLLER CARD" , subsystems : & [] }] }),
        (6225, Vendor { id : 6225u16 , name : "Microtune, Inc." , devices : & [] }),
        (4614, Vendor { id : 4614u16 , name : "Amdahl Corporation" , devices : & [] }),
        (5056, Vendor { id : 5056u16 , name : "Microgate Corporation" , devices : & [Device { vendor_id : 5056u16 , id : 16u16 , name : "SyncLink Adapter v1" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 32u16 , name : "SyncLink SCC Adapter" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 48u16 , name : "SyncLink Multiport Adapter" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 112u16 , name : "SyncLink GT Adapter" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 128u16 , name : "SyncLink GT4 Adapter" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 160u16 , name : "SyncLink GT2 Adapter" , subsystems : & [] } , Device { vendor_id : 5056u16 , id : 528u16 , name : "SyncLink Adapter v2" , subsystems : & [] }] }),
        (6305, Vendor { id : 6305u16 , name : "Astute Networks Inc." , devices : & [] }),
        (5211, Vendor { id : 5211u16 , name : "PRAIM SRL" , devices : & [] }),
        (65213, Vendor { id : 65213u16 , name : "Ultraview Corp." , devices : & [] }),
        (5418, Vendor { id : 5418u16 , name : "QUICKTURN DESIGN Systems" , devices : & [] }),
        (4515, Vendor { id : 4515u16 , name : "Deuretzbacher GmbH & Co. Eng. KG" , devices : & [] }),
        (5812, Vendor { id : 5812u16 , name : "Aspex Semiconductor Ltd" , devices : & [] }),
        (4236, Vendor { id : 4236u16 , name : "Oakleigh Systems Inc." , devices : & [] }),
        (4727, Vendor { id : 4727u16 , name : "Comstream" , devices : & [] }),
        (4964, Vendor { id : 4964u16 , name : "ATM Communications Inc" , devices : & [] }),
        (6356, Vendor { id : 6356u16 , name : "Celestica" , devices : & [Device { vendor_id : 6356u16 , id : 4103u16 , name : "Time Card" , subsystems : & [] }] }),
        (6345, Vendor { id : 6345u16 , name : "ARVOO Engineering BV" , devices : & [] }),
        (5152, Vendor { id : 5152u16 , name : "Psion Dacom plc" , devices : & [Device { vendor_id : 5152u16 , id : 32770u16 , name : "Gold Card NetGlobal 56k+10/100Mb CardBus (Ethernet part)" , subsystems : & [] } , Device { vendor_id : 5152u16 , id : 32771u16 , name : "Gold Card NetGlobal 56k+10/100Mb CardBus (Modem part)" , subsystems : & [] }] }),
        (4449, Vendor { id : 4449u16 , name : "PFU Limited" , devices : & [] }),
        (4767, Vendor { id : 4767u16 , name : "OEC Medical Systems, Inc." , devices : & [] }),
        (14402, Vendor { id : 14402u16 , name : "EVGA Corporation" , devices : & [] }),
        (4737, Vendor { id : 4737u16 , name : "Yokogawa Electric Corporation" , devices : & [] }),
        (5311, Vendor { id : 5311u16 , name : "SPIDER Communications Inc." , devices : & [] }),
        (6252, Vendor { id : 6252u16 , name : "Humusoft, s.r.o." , devices : & [Device { vendor_id : 6252u16 , id : 1554u16 , name : "AD612 Data Acquisition Device" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1556u16 , name : "MF614 Multifunction I/O Card" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1570u16 , name : "AD622 Data Acquisition Device" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1572u16 , name : "MF624 Multifunction I/O PCI Card" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1573u16 , name : "MF625 3-phase Motor Driver" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1588u16 , name : "MF634 Multifunction I/O PCIe Card" , subsystems : & [] } , Device { vendor_id : 6252u16 , id : 1604u16 , name : "MF644 Multifunction I/O Thb Card" , subsystems : & [] }] }),
        (5169, Vendor { id : 5169u16 , name : "Gilat Satellite Networks" , devices : & [] }),
        (4337, Vendor { id : 4337u16 , name : "Tyan Computer" , devices : & [Device { vendor_id : 4337u16 , id : 10341u16 , name : "Tyan Thunder K8E S2865" , subsystems : & [] } , Device { vendor_id : 4337u16 , id : 21248u16 , name : "Tyan S5380 Mainboard" , subsystems : & [] }] }),
        (4908, Vendor { id : 4908u16 , name : "Micrel Inc" , devices : & [] }),
        (41563, Vendor { id : 41563u16 , name : "Hewlett Packard GmbH PL24-MKT" , devices : & [] }),
        (7023, Vendor { id : 7023u16 , name : "Etron Technology, Inc." , devices : & [Device { vendor_id : 7023u16 , id : 28707u16 , name : "EJ168 USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 20487u16 , name : "GA-880GMA-USB3" }] } , Device { vendor_id : 7023u16 , id : 28754u16 , name : "EJ188/EJ198 USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 28754u16 , name : "QC5000-ITX/PH" }] }] }),
        (1653, Vendor { id : 1653u16 , name : "Dynalink" , devices : & [Device { vendor_id : 1653u16 , id : 5888u16 , name : "IS64PH ISDN Adapter" , subsystems : & [] } , Device { vendor_id : 1653u16 , id : 5890u16 , name : "IS64PH ISDN Adapter" , subsystems : & [] } , Device { vendor_id : 1653u16 , id : 5891u16 , name : "ISDN Adapter (PCI Bus, DV, W)" , subsystems : & [] } , Device { vendor_id : 1653u16 , id : 5892u16 , name : "ISDN Adapter (PCI Bus, D, C)" , subsystems : & [] }] }),
        (37274, Vendor { id : 37274u16 , name : "Gigapixel Corp" , devices : & [] }),
        (4389, Vendor { id : 4389u16 , name : "Eurocore" , devices : & [] }),
        (5985, Vendor { id : 5985u16 , name : "Pickering Interfaces Ltd" , devices : & [Device { vendor_id : 5985u16 , id : 17425u16 , name : "Pickering Devices with FPGA Based Bus Communication" , subsystems : & [SubSystem { subvendor : 5985u16 , subdevice : 2095u16 , name : "40-737-901" } , SubSystem { subvendor : 5985u16 , subdevice : 2156u16 , name : "40-576-001" } , SubSystem { subvendor : 5985u16 , subdevice : 2177u16 , name : "40-584-001" } , SubSystem { subvendor : 5985u16 , subdevice : 12570u16 , name : "42-738-001" } , SubSystem { subvendor : 5985u16 , subdevice : 12688u16 , name : "41-765-004" } , SubSystem { subvendor : 5985u16 , subdevice : 12715u16 , name : "41-670-003" } , SubSystem { subvendor : 5985u16 , subdevice : 13087u16 , name : "50-297A-014" } , SubSystem { subvendor : 5985u16 , subdevice : 13088u16 , name : "50-297A-050" } , SubSystem { subvendor : 5985u16 , subdevice : 13089u16 , name : "50-297A-056" } , SubSystem { subvendor : 5985u16 , subdevice : 13158u16 , name : "41-625-004" } , SubSystem { subvendor : 5985u16 , subdevice : 13160u16 , name : "50-297A-130" } , SubSystem { subvendor : 5985u16 , subdevice : 13170u16 , name : "50-297A-122" } , SubSystem { subvendor : 5985u16 , subdevice : 13180u16 , name : "50-297A-144" } , SubSystem { subvendor : 5985u16 , subdevice : 13207u16 , name : "50-297A-030" } , SubSystem { subvendor : 5985u16 , subdevice : 13217u16 , name : "40-419-004" } , SubSystem { subvendor : 5985u16 , subdevice : 13219u16 , name : "41-770-002" } , SubSystem { subvendor : 5985u16 , subdevice : 13236u16 , name : "50-297A-034" } , SubSystem { subvendor : 5985u16 , subdevice : 13265u16 , name : "50-297A-151" } , SubSystem { subvendor : 5985u16 , subdevice : 14100u16 , name : "42-297A-050" }] }] }),
        (4110, Vendor { id : 4110u16 , name : "Weitek" , devices : & [Device { vendor_id : 4110u16 , id : 36864u16 , name : "P9000 Viper" , subsystems : & [] } , Device { vendor_id : 4110u16 , id : 36865u16 , name : "P9000 Viper" , subsystems : & [] } , Device { vendor_id : 4110u16 , id : 36866u16 , name : "P9000 Viper" , subsystems : & [] } , Device { vendor_id : 4110u16 , id : 37120u16 , name : "P9100 Viper Pro/SE" , subsystems : & [] }] }),
        (4124, Vendor { id : 4124u16 , name : "Western Digital" , devices : & [Device { vendor_id : 4124u16 , id : 403u16 , name : "33C193A" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 406u16 , name : "33C196A" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 407u16 , name : "33C197A" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 662u16 , name : "33C296A" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 12691u16 , name : "7193" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 12695u16 , name : "7197" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 12950u16 , name : "33C296A" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 17046u16 , name : "34C296" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 38672u16 , name : "Pipeline 9710" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 38674u16 , name : "Pipeline 9712" , subsystems : & [] } , Device { vendor_id : 4124u16 , id : 49738u16 , name : "90C" , subsystems : & [] }] }),
        (5475, Vendor { id : 5475u16 , name : "A-Trend Technology Co Ltd" , devices : & [] }),
        (5527, Vendor { id : 5527u16 , name : "Memec Design Services" , devices : & [] }),
        (4531, Vendor { id : 4531u16 , name : "Barr Systems Inc." , devices : & [] }),
        (4895, Vendor { id : 4895u16 , name : "Siig Inc" , devices : & [Device { vendor_id : 4895u16 , id : 4096u16 , name : "CyberSerial (1-port) 16550" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4097u16 , name : "CyberSerial (1-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4098u16 , name : "CyberSerial (1-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4112u16 , name : "Duet 1S(16550)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4113u16 , name : "Duet 1S(16650)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4114u16 , name : "Duet 1S(16850)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4128u16 , name : "CyberParallel (1-port)" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4129u16 , name : "CyberParallel (2-port)" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4144u16 , name : "CyberSerial (2-port) 16550" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4145u16 , name : "CyberSerial (2-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4146u16 , name : "CyberSerial (2-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4148u16 , name : "Trio 2S(16550)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4149u16 , name : "Trio 2S(16650)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4150u16 , name : "Trio 2S(16850)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4176u16 , name : "CyberSerial (4-port) 16550" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4177u16 , name : "CyberSerial (4-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 4178u16 , name : "CyberSerial (4-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8192u16 , name : "CyberSerial (1-port) 16550" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8193u16 , name : "CyberSerial (1-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8194u16 , name : "CyberSerial (1-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8208u16 , name : "Duet 1S(16550)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8209u16 , name : "Duet 1S(16650)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8210u16 , name : "Duet 1S(16850)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8224u16 , name : "CyberParallel (1-port)" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8225u16 , name : "CyberParallel (2-port)" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8240u16 , name : "CyberSerial (2-port) 16550" , subsystems : & [SubSystem { subvendor : 4895u16 , subdevice : 8240u16 , name : "PCI Serial Card" }] } , Device { vendor_id : 4895u16 , id : 8241u16 , name : "CyberSerial (2-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8242u16 , name : "CyberSerial (2-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8256u16 , name : "Trio 1S(16550)+2P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8257u16 , name : "Trio 1S(16650)+2P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8258u16 , name : "Trio 1S(16850)+2P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8272u16 , name : "CyberSerial (4-port) 16550" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8273u16 , name : "CyberSerial (4-port) 16650" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8274u16 , name : "CyberSerial (4-port) 16850" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8288u16 , name : "Trio 2S(16550)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8289u16 , name : "Trio 2S(16650)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8290u16 , name : "Trio 2S(16850)+1P" , subsystems : & [] } , Device { vendor_id : 4895u16 , id : 8321u16 , name : "CyberSerial (8-port) ST16654" , subsystems : & [] }] }),
        (4561, Vendor { id : 4561u16 , name : "Auravision" , devices : & [Device { vendor_id : 4561u16 , id : 503u16 , name : "VxP524" , subsystems : & [] } , Device { vendor_id : 4561u16 , id : 505u16 , name : "VxP951" , subsystems : & [] }] }),
        (7561, Vendor { id : 7561u16 , name : "YEESTOR Microelectronics Co., Ltd" , devices : & [Device { vendor_id : 7561u16 , id : 640u16 , name : "PCIe NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7561u16 , id : 61221u16 , name : "GDRAMARS GMT1000 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5482, Vendor { id : 5482u16 , name : "Avtec Systems" , devices : & [] }),
        (51771, Vendor { id : 51771u16 , name : "Cambrionix Ltd." , devices : & [] }),
        (5452, Vendor { id : 5452u16 , name : "VISUAL Technology Inc" , devices : & [] }),
        (4140, Vendor { id : 4140u16 , name : "Chips and Technologies" , devices : & [Device { vendor_id : 4140u16 , id : 184u16 , name : "F64310" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 192u16 , name : "F69000 HiQVideo" , subsystems : & [SubSystem { subvendor : 4140u16 , subdevice : 192u16 , name : "F69000 HiQVideo" } , SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4112u16 , name : "CP5/CR6 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4128u16 , name : "VR6 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4144u16 , name : "PC5 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" }] } , Device { vendor_id : 4140u16 , id : 208u16 , name : "F65545" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 216u16 , name : "F65545" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 220u16 , name : "F65548" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 224u16 , name : "F65550" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 228u16 , name : "F65554" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 229u16 , name : "F65555 HiQVPro" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45129u16 , name : "Armada 1700 Laptop Display Controller" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Satellite Pro/Satellite" }] } , Device { vendor_id : 4140u16 , id : 240u16 , name : "F68554" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 244u16 , name : "F68554 HiQVision" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 245u16 , name : "F68555" , subsystems : & [] } , Device { vendor_id : 4140u16 , id : 3120u16 , name : "F69030" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4176u16 , name : "CT7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4177u16 , name : "CE7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" }] }] }),
        (5215, Vendor { id : 5215u16 , name : "Baldor Electric Company" , devices : & [Device { vendor_id : 5215u16 , id : 1u16 , name : "NextMove PCI" , subsystems : & [] }] }),
        (4711, Vendor { id : 4711u16 , name : "S. A. Telecommunications" , devices : & [Device { vendor_id : 4711u16 , id : 21330u16 , name : "PCR2101" , subsystems : & [] } , Device { vendor_id : 4711u16 , id : 23115u16 , name : "Telsat Turbo" , subsystems : & [] }] }),
        (4527, Vendor { id : 4527u16 , name : "Avid Technology Inc." , devices : & [Device { vendor_id : 4527u16 , id : 1u16 , name : "Cinema" , subsystems : & [] } , Device { vendor_id : 4527u16 , id : 60961u16 , name : "Digidesign DSP Farm" , subsystems : & [] } , Device { vendor_id : 4527u16 , id : 60992u16 , name : "Digidesign Audiomedia III" , subsystems : & [] } , Device { vendor_id : 4527u16 , id : 61024u16 , name : "Digidesign SampleCell II / II Plus" , subsystems : & [] }] }),
        (5407, Vendor { id : 5407u16 , name : "TOPIC SEMICONDUCTOR Corp" , devices : & [Device { vendor_id : 5407u16 , id : 0u16 , name : "TP560 Data/Fax/Voice 56k modem" , subsystems : & [] }] }),
        (4950, Vendor { id : 4950u16 , name : "The Logical Co" , devices : & [] }),
        (8180, Vendor { id : 8180u16 , name : "DEEPX Co., Ltd." , devices : & [Device { vendor_id : 8180u16 , id : 256u16 , name : "M1 [Series]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 257u16 , name : "M1 [H1]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 258u16 , name : "M1 [H1 V-NPU]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 272u16 , name : "M1M [Series]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 273u16 , name : "M1M [H1M]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 274u16 , name : "M1M [H1M V-NPU]" , subsystems : & [] } , Device { vendor_id : 8180u16 , id : 8193u16 , name : "VPU [H1/H1M V-NPU]" , subsystems : & [] }] }),
        (5188, Vendor { id : 5188u16 , name : "TRW" , devices : & [] }),
        (60544, Vendor { id : 60544u16 , name : "Belkin Corporation" , devices : & [Device { vendor_id : 60544u16 , id : 60416u16 , name : "F5D6000" , subsystems : & [] }] }),
        (4332, Vendor { id : 4332u16 , name : "Realtek Semiconductor Co., Ltd." , devices : & [Device { vendor_id : 4332u16 , id : 313u16 , name : "RTL-8139/8139C/8139C+ Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 9728u16 , name : "Killer E2600 GbE Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 12288u16 , name : "Killer E3000 2.5GbE Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 17185u16 , name : "RTL8852BE 802.11ax PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 20480u16 , name : "Killer E5000 5GbE Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21000u16 , name : "RTS5208 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21001u16 , name : "RTS5209 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21031u16 , name : "RTS5227 PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8718u16 , name : "ThinkPad T440p" } , SubSystem { subvendor : 6058u16 , subdevice : 8724u16 , name : "ThinkPad X240" }] } , Device { vendor_id : 4332u16 , id : 21032u16 , name : "RTS5288 PCIe SD UHS-I Card Reader controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21033u16 , name : "RTS5229 PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 2067u16 , name : "Aspire R7-571" } , SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 6058u16 , subdevice : 14386u16 , name : "Yoga 520" }] } , Device { vendor_id : 4332u16 , id : 21034u16 , name : "RTS522A PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 32889u16 , name : "EliteBook 840 G3" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" }] } , Device { vendor_id : 4332u16 , id : 21065u16 , name : "RTS5249 PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6409u16 , name : "ZBook 15" }] } , Device { vendor_id : 4332u16 , id : 21066u16 , name : "RTS524A PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21072u16 , name : "RTS5250 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21082u16 , name : "RTS525A PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1750u16 , name : "Latitude 7275 tablet" } , SubSystem { subvendor : 4136u16 , subdevice : 1756u16 , name : "Latitude E7470" } , SubSystem { subvendor : 4136u16 , subdevice : 1764u16 , name : "XPS 15 9550" } , SubSystem { subvendor : 4136u16 , subdevice : 1766u16 , name : "Latitude 11 5175 2-in-1" } , SubSystem { subvendor : 4136u16 , subdevice : 1796u16 , name : "XPS 13 9350" } , SubSystem { subvendor : 4136u16 , subdevice : 2494u16 , name : "Latitude 7410" } , SubSystem { subvendor : 4136u16 , subdevice : 2832u16 , name : "Precision 3571" } , SubSystem { subvendor : 4136u16 , subdevice : 3078u16 , name : "Precision 3580" } , SubSystem { subvendor : 5218u16 , subdevice : 5295u16 , name : "MSI Claw A8" } , SubSystem { subvendor : 6058u16 , subdevice : 8783u16 , name : "ThinkPad X1 Carbon 5th Gen" }] } , Device { vendor_id : 4332u16 , id : 21088u16 , name : "RTS5260 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21089u16 , name : "RTS5261 PCIe SD Express Card Reader controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21092u16 , name : "RTS5264 PCIe SD UHS-II & SD Express Card Reader controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21126u16 , name : "RTS5286 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21127u16 , name : "RTL8411B PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4244u16 , name : "Acer Aspire E5-575G" }] } , Device { vendor_id : 4332u16 , id : 21128u16 , name : "RTS5288 PCI Express Card Reader" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 21129u16 , name : "RTL8411 PCI Express Card Reader" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5207u16 , name : "K55A Laptop" }] } , Device { vendor_id : 4332u16 , id : 22368u16 , name : "RTS5763DL x2 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22370u16 , name : "RTS5762 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22371u16 , name : "RTS5763DL NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22373u16 , name : "RTS5765DL NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22384u16 , name : "RTS5770DL NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22385u16 , name : "RTS5771 NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 22386u16 , name : "RTS5772DL NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 32809u16 , name : "RTL-8029(AS)" , subsystems : & [SubSystem { subvendor : 4280u16 , subdevice : 8209u16 , name : "EZ-Card (SMC1208)" } , SubSystem { subvendor : 4332u16 , subdevice : 32809u16 , name : "RTL-8029(AS)" } , SubSystem { subvendor : 4371u16 , subdevice : 4616u16 , name : "EN1208" } , SubSystem { subvendor : 4486u16 , subdevice : 768u16 , name : "DE-528" } , SubSystem { subvendor : 4697u16 , subdevice : 9216u16 , name : "AT-2400" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 4332u16 , id : 33061u16 , name : "RTL8125 2.5GbE Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 8226u16 , name : "LRES2022PT Single-port 2.5Gb Ethernet Network Adapter" }] } , Device { vendor_id : 4332u16 , id : 33062u16 , name : "RTL8126 5GbE Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33063u16 , name : "RTL8127 10GbE Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33065u16 , name : "RTL-8129" , subsystems : & [SubSystem { subvendor : 4332u16 , subdevice : 33065u16 , name : "RT8129 Fast Ethernet Adapter" } , SubSystem { subvendor : 4588u16 , subdevice : 33065u16 , name : "RTL8111/8168 PCIe Gigabit Ethernet (misconfigured)" }] } , Device { vendor_id : 4332u16 , id : 33078u16 , name : "RTL810xE PCI Express Fast Ethernet controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "RTL8106E on Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4156u16 , subdevice : 10892u16 , name : "Compaq 500B Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 10929u16 , name : "Pavilion p6774" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4473u16 , subdevice : 65380u16 , name : "RTL8102E PCI-E Fast Ethernet NIC" } , SubSystem { subvendor : 6080u16 , subdevice : 4179u16 , name : "RTL8101e Medion WIM 2210 Notebook PC [MD96850]" }] } , Device { vendor_id : 4332u16 , id : 33079u16 , name : "RTL8104E PCIe Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33080u16 , name : "RT8139 (B/C) Cardbus Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4332u16 , subdevice : 33080u16 , name : "RT8139 (B/C) Fast Ethernet Adapter" }] } , Device { vendor_id : 4332u16 , id : 33081u16 , name : "RTL-8100/8101L/8139 PCI Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 855u16 , subdevice : 10u16 , name : "TTP-Monitoring Card V2.0" } , SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 35104u16 , name : "ALN-325" } , SubSystem { subvendor : 4133u16 , subdevice : 35105u16 , name : "ALN-325" } , SubSystem { subvendor : 4156u16 , subdevice : 106u16 , name : "NX9500" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12505u16 , name : "Presario C700" } , SubSystem { subvendor : 4163u16 , subdevice : 4165u16 , name : "L8400B, L3C/S, X58LE notebook" } , SubSystem { subvendor : 4163u16 , subdevice : 33033u16 , name : "P5P800-MX Mainboard" } , SubSystem { subvendor : 4209u16 , subdevice : 33120u16 , name : "MIM2000" } , SubSystem { subvendor : 4285u16 , subdevice : 800u16 , name : "EP-320X-R" } , SubSystem { subvendor : 4332u16 , subdevice : 33081u16 , name : "RTL-8100/8101L/8139 PCI Fast Ethernet Adapter" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 4371u16 , subdevice : 60417u16 , name : "LevelOne FNC-0107TX/FNC-0109TX" } , SubSystem { subvendor : 4486u16 , subdevice : 4356u16 , name : "DFE-520TX Fast Ethernet PCI Adapter (rev. D1)" } , SubSystem { subvendor : 4486u16 , subdevice : 4864u16 , name : "DFE-538TX" } , SubSystem { subvendor : 4486u16 , subdevice : 4896u16 , name : "SN5200" } , SubSystem { subvendor : 4486u16 , subdevice : 33081u16 , name : "DRN-32TX" } , SubSystem { subvendor : 4598u16 , subdevice : 33081u16 , name : "FN22-3(A) LinxPRO Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 9472u16 , name : "AT-2500TX" } , SubSystem { subvendor : 4697u16 , subdevice : 9475u16 , name : "AT-2500TX/ACPI" } , SubSystem { subvendor : 4997u16 , subdevice : 62237u16 , name : "FA311 v2" } , SubSystem { subvendor : 5013u16 , subdevice : 8448u16 , name : "AMB2100" } , SubSystem { subvendor : 5161u16 , subdevice : 53264u16 , name : "ND010/ND012" } , SubSystem { subvendor : 5170u16 , subdevice : 37168u16 , name : "EN-9130TX" } , SubSystem { subvendor : 5174u16 , subdevice : 33081u16 , name : "RT8139" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "GA-7VM400M/7VT600 Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 8572u16 , name : "Aspire L250" } , SubSystem { subvendor : 5218u16 , subdevice : 30860u16 , name : "865PE Neo2-V Mainboard" } , SubSystem { subvendor : 5228u16 , subdevice : 5177u16 , name : "FE-1439TX" } , SubSystem { subvendor : 5257u16 , subdevice : 24577u16 , name : "GF100TXRII" } , SubSystem { subvendor : 5257u16 , subdevice : 24578u16 , name : "GF100TXRA" } , SubSystem { subvendor : 5276u16 , subdevice : 5018u16 , name : "LFE-8139ATX" } , SubSystem { subvendor : 5276u16 , subdevice : 33081u16 , name : "LFE-8139TX" } , SubSystem { subvendor : 5323u16 , subdevice : 512u16 , name : "LNR-100 Family 10/100 Base-TX Ethernet" } , SubSystem { subvendor : 5477u16 , subdevice : 8960u16 , name : "P4TSV Onboard LAN (RTL8100B)" } , SubSystem { subvendor : 5681u16 , subdevice : 28675u16 , name : "Onboard RTL8111 on GA-8SIML Rev1.0 Mainboard" } , SubSystem { subvendor : 5781u16 , subdevice : 36865u16 , name : "Onboard RTL8101L 10/100 MBit" } , SubSystem { subvendor : 5868u16 , subdevice : 255u16 , name : "USR997900A" } , SubSystem { subvendor : 6041u16 , subdevice : 20480u16 , name : "F5D5000 PCI Card/Desktop Network PCI Card" } , SubSystem { subvendor : 6041u16 , subdevice : 20496u16 , name : "F5D5010 CardBus Notebook Network Card" } , SubSystem { subvendor : 6270u16 , subdevice : 13059u16 , name : "FN312" } , SubSystem { subvendor : 6404u16 , subdevice : 33081u16 , name : "RTL8139D Fast Ethernet Adapter" } , SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" } , SubSystem { subvendor : 9798u16 , subdevice : 1u16 , name : "KNE120TX" } , SubSystem { subvendor : 36398u16 , subdevice : 28672u16 , name : "KF-230TX" } , SubSystem { subvendor : 36398u16 , subdevice : 28928u16 , name : "KF-230TX/2" } , SubSystem { subvendor : 41120u16 , subdevice : 7u16 , name : "ALN-325C" }] } , Device { vendor_id : 4332u16 , id : 33121u16 , name : "RTL8111/8168/8411 PCI Express Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4332u16 , subdevice : 33128u16 , name : "TP-Link TG-3468 v4.0 Gigabit PCI Express Network Adapter" }] } , Device { vendor_id : 4332u16 , id : 33127u16 , name : "RTL-8110SC/8169SC Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3600u16 , name : "RTL-8110SC-GR on a N15235/A74MX mainboard" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "GA-MA69G-S3H Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 9052u16 , name : "P965 Neo MS-7235 mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 9068u16 , name : "945P Neo3-F motherboard" }] } , Device { vendor_id : 4332u16 , id : 33128u16 , name : "RTL8111/8168/8211/8411 PCI Express Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 33128u16 , name : "RTL8111/8168 PCI Express Gigabit Ethernet controller" } , SubSystem { subvendor : 4133u16 , subdevice : 4244u16 , name : "Acer Aspire E5-575G" } , SubSystem { subvendor : 4136u16 , subdevice : 643u16 , name : "Vostro 220" } , SubSystem { subvendor : 4136u16 , subdevice : 1202u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1242u16 , name : "Vostro 3750" } , SubSystem { subvendor : 4136u16 , subdevice : 1495u16 , name : "Alienware X51 R2" } , SubSystem { subvendor : 4136u16 , subdevice : 1778u16 , name : "Latitude 3470" } , SubSystem { subvendor : 4136u16 , subdevice : 1779u16 , name : "Latitude 3570" } , SubSystem { subvendor : 4136u16 , subdevice : 2153u16 , name : "Vostro 3470" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4156u16 , subdevice : 6480u16 , name : "ProBook 450/455" } , SubSystem { subvendor : 4156u16 , subdevice : 10863u16 , name : "Asus IPIBL-LB Motherboard" } , SubSystem { subvendor : 4156u16 , subdevice : 33371u16 , name : "OMEN-17-w001nv" } , SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4156u16 , subdevice : 34516u16 , name : "Pavilion Laptop 15-ec0xxx" } , SubSystem { subvendor : 4156u16 , subdevice : 34946u16 , name : "HP ProDesk 405 G8 Desktop Mini PC" } , SubSystem { subvendor : 4156u16 , subdevice : 35607u16 , name : "ProBook 445 G9/455 G9 [RTL8111HSH-CG GbE Controller]" } , SubSystem { subvendor : 4163u16 , subdevice : 4597u16 , name : "Notebook motherboard (one of many models)" } , SubSystem { subvendor : 4163u16 , subdevice : 5845u16 , name : "U6V/U31J laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 33194u16 , name : "P5B" } , SubSystem { subvendor : 4163u16 , subdevice : 33478u16 , name : "M3A78 Series Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33699u16 , name : "M4A785/P7P55/AT3IONT-I Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33842u16 , name : "P8P67 and other motherboards" } , SubSystem { subvendor : 4163u16 , subdevice : 34053u16 , name : "P8 series motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 34132u16 , name : "H81M-C Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 34206u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 34423u16 , name : "Onboard RTL8111H Ethernet" } , SubSystem { subvendor : 4187u16 , subdevice : 3452u16 , name : "D270S/D250S Motherboard" } , SubSystem { subvendor : 4332u16 , subdevice : 33128u16 , name : "RTL8111/8168 PCI Express Gigabit Ethernet controller" } , SubSystem { subvendor : 5197u16 , subdevice : 50770u16 , name : "RTL8168 on a NP300E5C series laptop" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "Onboard Ethernet" } , SubSystem { subvendor : 5218u16 , subdevice : 9100u16 , name : "Onboard RTL8111b on MSI P965 Platinum Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 13404u16 , name : "RTL8111B on MS-7345 Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 13964u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 5218u16 , subdevice : 16768u16 , name : "Wind PC MS-7418" } , SubSystem { subvendor : 5218u16 , subdevice : 29986u16 , name : "X58 Pro-E" } , SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" } , SubSystem { subvendor : 5940u16 , subdevice : 4544u16 , name : "RTL8211DN on Esprimo P510 D3171 motherboard" } , SubSystem { subvendor : 6005u16 , subdevice : 4556u16 , name : "CC11/CL11" } , SubSystem { subvendor : 6058u16 , subdevice : 12440u16 , name : "ThinkCentre E73" } , SubSystem { subvendor : 6058u16 , subdevice : 14356u16 , name : "Z50-75" } , SubSystem { subvendor : 6058u16 , subdevice : 14371u16 , name : "Lenovo V130-15IGM Laptop - Type 81HL" } , SubSystem { subvendor : 6058u16 , subdevice : 20584u16 , name : "Thinkpad E480/E580" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 6217u16 , subdevice : 33128u16 , name : "Motherboard (one of many)" } , SubSystem { subvendor : 29808u16 , subdevice : 13416u16 , name : "TG-3468 Gigabit PCI Express Network Adapter" } , SubSystem { subvendor : 32902u16 , subdevice : 8277u16 , name : "NUC Kit DN2820FYKH" } , SubSystem { subvendor : 32902u16 , subdevice : 54805u16 , name : "Desktop Board D510MO/D525MW" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4332u16 , id : 33129u16 , name : "RTL8169 PCI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 121u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4285u16 , subdevice : 12802u16 , name : "EP-320G-TX1 32-bit PCI Gigabit Ethernet Adapter" } , SubSystem { subvendor : 4332u16 , subdevice : 33129u16 , name : "RTL8169/8110 Family PCI Gigabit Ethernet NIC" } , SubSystem { subvendor : 4697u16 , subdevice : 49415u16 , name : "CG-LAPCIGT" } , SubSystem { subvendor : 4977u16 , subdevice : 17230u16 , name : "ProG-2000L" } , SubSystem { subvendor : 4997u16 , subdevice : 12570u16 , name : "GA311" } , SubSystem { subvendor : 4997u16 , subdevice : 20992u16 , name : "GA511 Gigabit PC Card" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "GA-8I915ME-G Mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 780u16 , name : "K8N Neo-FSR v2.0 mainboard" } , SubSystem { subvendor : 5218u16 , subdevice : 1628u16 , name : "Hetis 865GV-E (MS-7065)" } , SubSystem { subvendor : 5218u16 , subdevice : 28716u16 , name : "K8T NEO 2 motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 28820u16 , name : "K8T Neo2-F V2.0" } , SubSystem { subvendor : 5868u16 , subdevice : 287u16 , name : "USR997903" } , SubSystem { subvendor : 5940u16 , subdevice : 4241u16 , name : "D2030-A1" } , SubSystem { subvendor : 41120u16 , subdevice : 1097u16 , name : "AK86-L motherboard" }] } , Device { vendor_id : 4332u16 , id : 33130u16 , name : "RTL8111xP UART #1" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4332u16 , id : 33131u16 , name : "RTL8111xP UART #2" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4332u16 , id : 33132u16 , name : "RTL8111xP IPMI interface" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4332u16 , id : 33133u16 , name : "RTL811x EHCI host controller" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4332u16 , id : 33134u16 , name : "RealManage BMC" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33137u16 , name : "RTL8191SEvA Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33138u16 , name : "RTL8191SEvB Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33139u16 , name : "RTL8192SE Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33140u16 , name : "RTL8192SE Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33142u16 , name : "RTL8188CE 802.11b/g/n WiFi Adapter" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33973u16 , name : "PCE-N10" } , SubSystem { subvendor : 6715u16 , subdevice : 4409u16 , name : "AW-NE139H Half-size Mini PCIe Card" }] } , Device { vendor_id : 4332u16 , id : 33143u16 , name : "RTL8191CE PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33144u16 , name : "RTL8192CE PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33145u16 , name : "RTL8188EE Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6525u16 , name : "RTL8188EE mini-PCIe card" }] } , Device { vendor_id : 4332u16 , id : 33152u16 , name : "RTL8180L 802.11b MAC" , subsystems : & [SubSystem { subvendor : 4997u16 , subdevice : 18176u16 , name : "MA521 802.11b Wireless PC Card" } , SubSystem { subvendor : 5943u16 , subdevice : 25u16 , name : "WPC11v4 802.11b Wireless-B Notebook Adapter" }] } , Device { vendor_id : 4332u16 , id : 33157u16 , name : "RTL-8185 IEEE 802.11a/b/g Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 6270u16 , subdevice : 33317u16 , name : "ZyAIR G-302 v3" }] } , Device { vendor_id : 4332u16 , id : 33163u16 , name : "RTL8192EE PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33168u16 , name : "RTL8190 802.11n PCI Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33169u16 , name : "RTL8192CE PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33170u16 , name : "RTL8192E/RTL8192SE Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33171u16 , name : "RTL8192DE Wireless LAN Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33174u16 , name : "RTL8196 Integrated PCI-e Bridge" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33175u16 , name : "SmartLAN56 56K Modem" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 33177u16 , name : "RTL8187SE Wireless LAN Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 26772u16 , name : "MN54G2 / MS-6894 Wireless Mini PCIe Card" }] } , Device { vendor_id : 4332u16 , id : 34595u16 , name : "RTL8723AE PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 34834u16 , name : "RTL8812AE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 34835u16 , name : "RTL8813AE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 34849u16 , name : "RTL8821AE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 34898u16 , name : "RTL8852AE 802.11ax PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 35106u16 , name : "RTL8922AE 802.11be PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 43098u16 , name : "RTL8852AE WiFi 6 802.11ax PCIe Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 46368u16 , name : "RTL8852BE-VT PCIe 802.11ax Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 46883u16 , name : "RTL8723BE PCIe Wireless Network Adapter" , subsystems : & [SubSystem { subvendor : 4332u16 , subdevice : 34617u16 , name : "Dell Wireless 1801" } , SubSystem { subvendor : 6058u16 , subdevice : 46902u16 , name : "Z50-75" }] } , Device { vendor_id : 4332u16 , id : 47137u16 , name : "RTL8821CE PCIe 802.11ac Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 47138u16 , name : "RTL8822BE 802.11a/b/g/n/ac WiFi adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33563u16 , name : "Realtek RTL8822BE 802.11ac 2x2 Wi-Fi + Bluetooth 4.2 Combo Adapter (MU-MIMO supported)" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 6058u16 , subdevice : 45091u16 , name : "ThinkPad E595" }] } , Device { vendor_id : 4332u16 , id : 47185u16 , name : "RTL8851BE PCIe 802.11ax Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 47186u16 , name : "RTL8852BE PCIe 802.11ax Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 47195u16 , name : "RTL8852BE PCIe 802.11ax Wireless Network Controller [1T1R]" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 51233u16 , name : "RTL8821CE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 51234u16 , name : "RTL8822CE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 51247u16 , name : "RTL8822CE 802.11ac PCIe Wireless Network Adapter" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 51282u16 , name : "RTL8852CE PCIe 802.11ax Wireless Network Controller" , subsystems : & [] } , Device { vendor_id : 4332u16 , id : 55075u16 , name : "RTL8723DE 802.11b/g/n PCIe Adapter" , subsystems : & [] }] }),
        (7218, Vendor { id : 7218u16 , name : "Highland Technology, Inc." , devices : & [] }),
        (5681, Vendor { id : 5681u16 , name : "Packard Bell B.V." , devices : & [] }),
        (1161, Vendor { id : 1161u16 , name : "SIFIVE" , devices : & [] }),
        (4544, Vendor { id : 4544u16 , name : "Hewlett Packard" , devices : & [] }),
        (4417, Vendor { id : 4417u16 , name : "Crest Microsystem Inc" , devices : & [] }),
        (4808, Vendor { id : 4808u16 , name : "G Force Co, Ltd" , devices : & [] }),
        (5670, Vendor { id : 5670u16 , name : "TDK Semiconductor Corp." , devices : & [Device { vendor_id : 5670u16 , id : 33808u16 , name : "RTL81xx Fast Ethernet" , subsystems : & [] }] }),
        (5591, Vendor { id : 5591u16 , name : "Rockwell-Collins Inc" , devices : & [] }),
        (4451, Vendor { id : 4451u16 , name : "Rendition" , devices : & [Device { vendor_id : 4451u16 , id : 1u16 , name : "Verite 1000" , subsystems : & [] } , Device { vendor_id : 4451u16 , id : 8192u16 , name : "Verite V2000/V2100/V2200" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 8192u16 , name : "Stealth II S220" }] }] }),
        (19267, Vendor { id : 19267u16 , name : "KonteX Inc." , devices : & [] }),
        (30912, Vendor { id : 30912u16 , name : "Herrick Technology Laboratories, Inc. [HTL]" , devices : & [Device { vendor_id : 30912u16 , id : 0u16 , name : "HTLv-1 / HTLv-2 / HTLv-13 / HTLv-23" , subsystems : & [] } , Device { vendor_id : 30912u16 , id : 17u16 , name : "HTLv-53" , subsystems : & [] }] }),
        (5106, Vendor { id : 5106u16 , name : "Ford Microelectronics Inc" , devices : & [] }),
        (5153, Vendor { id : 5153u16 , name : "Ads Technologies Inc" , devices : & [] }),
        (855, Vendor { id : 855u16 , name : "TTTech Computertechnik AG (Wrong ID)" , devices : & [Device { vendor_id : 855u16 , id : 10u16 , name : "TTP-Monitoring Card V2.0" , subsystems : & [] }] }),
        (7516, Vendor { id : 7516u16 , name : "Fantasia Trading LLC" , devices : & [] }),
        (5234, Vendor { id : 5234u16 , name : "DAIKIN Industries, Ltd" , devices : & [] }),
        (7066, Vendor { id : 7066u16 , name : "XAVi Technologies Corp." , devices : & [] }),
        (5224, Vendor { id : 5224u16 , name : "AMBIT Microsystem Corp." , devices : & [] }),
        (53601, Vendor { id : 53601u16 , name : "Digium, Inc." , devices : & [Device { vendor_id : 53601u16 , id : 288u16 , name : "Wildcard TE120P single-span T1/E1/J1 card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 517u16 , name : "Wildcard TE205P/TE207P dual-span T1/E1/J1 card 5.0V" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 528u16 , name : "Wildcard TE210P/TE212P dual-span T1/E1/J1 card 3.3V" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 544u16 , name : "Wildcard TE220 dual-span T1/E1/J1 card 3.3V (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 1029u16 , name : "Wildcard TE405P/TE407P quad-span T1/E1/J1 card 5.0V" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 1040u16 , name : "Wildcard TE410P/TE412P quad-span T1/E1/J1 card 3.3V" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 1056u16 , name : "Wildcard TE420P quad-span T1/E1/J1 card 3.3V (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 2048u16 , name : "Wildcard TDM800P 8-port analog card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 4613u16 , name : "Wildcard TE205P/TE207P dual-span T1/E1/J1 card 5.0V (u1)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 4640u16 , name : "Wildcard TE220 dual-span T1/E1/J1 card 3.3V (PCI-Express) (5th gen)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 5125u16 , name : "Wildcard TE405P/TE407P quad-span T1/E1/J1 card 5.0V (u1)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 5136u16 , name : "Wildcard TE410P quad-span T1/E1/J1 card 3.3V (5th Gen)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 5152u16 , name : "Wildcard TE420 quad-span T1/E1/J1 card 3.3V (PCI-Express) (5th gen)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 6176u16 , name : "Wildcard TE820 octal-span T1/E1/J1 card 3.3V (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 9216u16 , name : "Wildcard TDM2400P 24-port analog card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 13312u16 , name : "Wildcard TC400P transcoder base card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32768u16 , name : "Wildcard TE121 single-span T1/E1/J1 card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32769u16 , name : "Wildcard TE122 single-span T1/E1/J1 card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32770u16 , name : "Wildcard AEX800 8-port analog card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32771u16 , name : "Wildcard AEX2400 24-port analog card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32772u16 , name : "Wildcard TCE400P transcoder base card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32773u16 , name : "Wildcard TDM410 4-port analog card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32774u16 , name : "Wildcard AEX410 4-port analog card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32775u16 , name : "Hx8 Series 8-port Base Card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32776u16 , name : "Hx8 Series 8-port Base Card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32778u16 , name : "Wildcard TE133 single-span T1/E1/J1 card (PCI Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32779u16 , name : "Wildcard TE134 single-span T1/E1/J1 card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32780u16 , name : "Wildcard A8A 8-port analog card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32781u16 , name : "Wildcard A8B 8-port analog card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32782u16 , name : "Wildcard TE235/TE435 quad-span T1/E1/J1 card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32783u16 , name : "Wildcard A4A 4-port analog card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32784u16 , name : "Wildcard A4B 4-port analog card (PCI-Express)" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 32787u16 , name : "Wildcard TE236/TE436 quad-span T1/E1/J1 card" , subsystems : & [] } , Device { vendor_id : 53601u16 , id : 46096u16 , name : "Wildcard B410 quad-BRI card" , subsystems : & [] }] }),
        (19539, Vendor { id : 19539u16 , name : "SBS Technologies" , devices : & [Device { vendor_id : 19539u16 , id : 0u16 , name : "PLUSTEST device" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 12288u16 , name : "PLUSTEST card (PC104+)" } , SubSystem { subvendor : 19539u16 , subdevice : 12289u16 , name : "PLUSTEST card (PMC)" }] } , Device { vendor_id : 19539u16 , id : 1u16 , name : "PLUSTEST-MM device" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 12290u16 , name : "PLUSTEST-MM card (PMC)" }] }] }),
        (5061, Vendor { id : 5061u16 , name : "Alphi Technology Corp" , devices : & [] }),
        (4657, Vendor { id : 4657u16 , name : "Woodward McCoach, Inc." , devices : & [Device { vendor_id : 4657u16 , id : 1249u16 , name : "Desktop PCI Telephony 4" , subsystems : & [] } , Device { vendor_id : 4657u16 , id : 1505u16 , name : "Desktop PCI Telephony 5/6" , subsystems : & [] } , Device { vendor_id : 4657u16 , id : 3328u16 , name : "LightParser" , subsystems : & [] } , Device { vendor_id : 4657u16 , id : 3330u16 , name : "LightParser 2" , subsystems : & [] } , Device { vendor_id : 4657u16 , id : 3347u16 , name : "Desktop PCI L1/L3 Telephony" , subsystems : & [] }] }),
        (4580, Vendor { id : 4580u16 , name : "Second Wave Inc" , devices : & [] }),
        (5520, Vendor { id : 5520u16 , name : "Hewlett Packard Enterprise" , devices : & [Device { vendor_id : 5520u16 , id : 1u16 , name : "Eagle Cluster Manager" , subsystems : & [] } , Device { vendor_id : 5520u16 , id : 2u16 , name : "Osprey Cluster Manager" , subsystems : & [] } , Device { vendor_id : 5520u16 , id : 3u16 , name : "Harrier Cluster Manager" , subsystems : & [] } , Device { vendor_id : 5520u16 , id : 881u16 , name : "Cassini 2 [Slingshot 400Gb]" , subsystems : & [] } , Device { vendor_id : 5520u16 , id : 882u16 , name : "Cassini 2 [Slingshot 400Gb] SR-IOV VF" , subsystems : & [] } , Device { vendor_id : 5520u16 , id : 40989u16 , name : "FC044X Fibre Channel HBA" , subsystems : & [] }] }),
        (5465, Vendor { id : 5465u16 , name : "SI LOGIC Ltd" , devices : & [] }),
        (5104, Vendor { id : 5104u16 , name : "Sundance Technology Inc / IC Plus Corp" , devices : & [Device { vendor_id : 5104u16 , id : 512u16 , name : "IC Plus IP100A Integrated 10/100 Ethernet MAC + PHY" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33299u16 , name : "NX1001" }] } , Device { vendor_id : 5104u16 , id : 513u16 , name : "ST201 Sundance Ethernet" , subsystems : & [] } , Device { vendor_id : 5104u16 , id : 4129u16 , name : "TC902x Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 5104u16 , id : 4131u16 , name : "IP1000 Family Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33152u16 , name : "NX1101" }] }] }),
        (4143, Vendor { id : 4143u16 , name : "Toshiba America" , devices : & [Device { vendor_id : 4143u16 , id : 9u16 , name : "r4x00" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 10u16 , name : "TX3927 MIPS RISC PCI Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 32u16 , name : "ATM Meteor 155" , subsystems : & [SubSystem { subvendor : 4143u16 , subdevice : 248u16 , name : "ATM Meteor 155" }] } , Device { vendor_id : 4143u16 , id : 48u16 , name : "TC35815CF PCI 10/100 Mbit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 49u16 , name : "TC35815CF PCI 10/100 Mbit Ethernet Controller with WOL" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 50u16 , name : "TC35815CF PCI 10/100 Mbit Ethernet Controller on TX4939" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 261u16 , name : "TC86C001 [goku-s] IDE" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 262u16 , name : "TC86C001 [goku-s] USB 1.1 Host" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 263u16 , name : "TC86C001 [goku-s] USB Device Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 264u16 , name : "TC86C001 [goku-s] I2C/SIO/GPIO Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 384u16 , name : "TX4927/38 MIPS RISC PCI Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 385u16 , name : "TX4925 MIPS RISC PCI Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 386u16 , name : "TX4937 MIPS RISC PCI Controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 436u16 , name : "Celleb platform IDE interface" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 437u16 , name : "SCC USB 2.0 EHCI controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 438u16 , name : "SCC USB 1.1 OHCI controller" , subsystems : & [] } , Device { vendor_id : 4143u16 , id : 442u16 , name : "SpursEngine SE1000 accelerator" , subsystems : & [] }] }),
        (6830, Vendor { id : 6830u16 , name : "Global Velocity, Inc." , devices : & [] }),
        (5007, Vendor { id : 5007u16 , name : "Patapsco Designs Inc" , devices : & [] }),
        (8330, Vendor { id : 8330u16 , name : "MICIUS Laboratory" , devices : & [] }),
        (26889, Vendor { id : 26889u16 , name : "TBS Technologies (wrong ID)" , devices : & [] }),
        (4422, Vendor { id : 4422u16 , name : "Force Computers" , devices : & [] }),
        (4836, Vendor { id : 4836u16 , name : "Brooktrout Technology Inc" , devices : & [] }),
        (4765, Vendor { id : 4765u16 , name : "Compcore Multimedia, Inc." , devices : & [] }),
        (4308, Vendor { id : 4308u16 , name : "Hualon Microelectronics" , devices : & [] }),
        (4319, Vendor { id : 4319u16 , name : "Emulex Corporation" , devices : & [Device { vendor_id : 4319u16 , id : 1824u16 , name : "OneConnect NIC (Skyhawk)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6452u16 , name : "FlexFabric 20Gb 2-port 650M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6453u16 , name : "FlexFabric 20Gb 2-port 650FLB Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8660u16 , name : "StoreFabric CN1200E 10Gb Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8714u16 , name : "FlexFabric 10Gb 2-port 556FLR-SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 32831u16 , name : "Ethernet 10Gb 2-port 557SFP+ Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 33092u16 , name : "FlexFabric 10GB 2-port 556FLR-T Adapter" } , SubSystem { subvendor : 6058u16 , subdevice : 4182u16 , name : "ThinkServer OCm14102-UX-L AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 4183u16 , name : "ThinkServer OCm14104-UX-L AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 4185u16 , name : "ThinkServer OCm14104-UT-L AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 16404u16 , name : "ThinkServer OCm14102-NX-L AnyFabric" }] } , Device { vendor_id : 4319u16 , id : 1826u16 , name : "OneConnect iSCSI Initiator (Skyhawk)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1827u16 , name : "OneConnect iSCSI Initiator + Target (Skyhawk)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1828u16 , name : "OneConnect FCoE Initiator (Skyhawk)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1832u16 , name : "OneConnect NIC (Skyhawk-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1834u16 , name : "OneConnect iSCSI Initiator (Skyhawk-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1835u16 , name : "OneConnect iSCSI Initiator + Target (Skyhawk-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 1836u16 , name : "OneConnect FCoE Initiator (Skyhawk-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 6885u16 , name : "LP6000 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 54016u16 , name : "LPE41100/LPE42100 Series 64G/128G Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57600u16 , name : "Proteus-X: LightPulse IOV Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57649u16 , name : "LightPulse 8Gb/s PCIe Shared I/O Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57728u16 , name : "Proteus-X: LightPulse IOV Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57856u16 , name : "LPe15000/LPe16000 Series 8Gb/16Gb Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1009u16 , name : "PCIe2 2-Port 16Gb Fibre Channel Adapter for POWER (FC EL5B; CCIN 577F)" } , SubSystem { subvendor : 4116u16 , subdevice : 1251u16 , name : "PCIe3 4-Port 10GbE SR Adapter for POWER (FC EN15/EN16; CCIN 2CE3)" } , SubSystem { subvendor : 4116u16 , subdevice : 1252u16 , name : "PCIe3 4-Port 10GbE SFP+ Adapter for POWER (FC EN18; CCIN 2CE4)" } , SubSystem { subvendor : 4319u16 , subdevice : 57984u16 , name : "LPe16002B-M6 2-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 57985u16 , name : "LPe16000B-M6 1-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 57986u16 , name : "Flex System FC5054 4-port 16Gb FC Adapter" }] } , Device { vendor_id : 4319u16 , id : 57864u16 , name : "LightPulse 16Gb Fibre Channel Host Adapter (Lancer-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57888u16 , name : "OneConnect NIC (Lancer)" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 4180u16 , name : "ThinkServer LPm16002B-M6-L AnyFabric" } , SubSystem { subvendor : 6058u16 , subdevice : 4181u16 , name : "ThinkServer LPm16004B-M8-L AnyFabric" }] } , Device { vendor_id : 4319u16 , id : 57920u16 , name : "OneConnect iSCSI Initiator (Lancer)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57952u16 , name : "OneConnect FCoE Initiator (Lancer)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 57960u16 , name : "OneConnect 10Gb FCoE Converged Network Adapter (Lancer-VF)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 58112u16 , name : "LPe31000/LPe32000 Series 16Gb/32Gb Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1556u16 , name : "PCIe3 4-Port 16Gb Fibre Channel Adapter for POWER (FC EN1C/EN1D; CCIN 578E)" } , SubSystem { subvendor : 4116u16 , subdevice : 1557u16 , name : "PCIe3 2-Port 32Gb Fibre Channel Adapter for POWER (FC EN1A/EN1B; CCIN 578F)" } , SubSystem { subvendor : 4116u16 , subdevice : 1696u16 , name : "PCIe3 2-Port 16Gb Fibre Channel Adapter for POWER (FC EN1L/EN1M; CCIN 2CFC)" } , SubSystem { subvendor : 4319u16 , subdevice : 58112u16 , name : "LPe32002-M2 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58113u16 , name : "LPe32000-M2 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58128u16 , name : "LPe31002-M6 2-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58129u16 , name : "LPe31000-M6 1-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58130u16 , name : "LPe31004-M6 4-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58144u16 , name : "LPe32002-M2-D 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58145u16 , name : "LPe32000-M2-D 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58146u16 , name : "LPe31002-M6-D 2-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58147u16 , name : "LPe31000-M6-D 1-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58148u16 , name : "LPm32002-D 2-Port 32Gb Fibre Channel Mezz Card" } , SubSystem { subvendor : 4319u16 , subdevice : 58149u16 , name : "LPm31002-D 2-Port 16Gb Fibre Channel Mezz Card" } , SubSystem { subvendor : 4319u16 , subdevice : 58160u16 , name : "LPe32002-M2-L 2-Port 32Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58161u16 , name : "LPe32000-M2-L 1-Port 32Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58162u16 , name : "LPe31002-M6-L 2-Port 16Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 58163u16 , name : "LPe31000-M6-L 1-Port 16Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 513u16 , name : "StoreFabric SN1600E 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 514u16 , name : "StoreFabric SN1600E 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 531u16 , name : "StoreFabric SN1200E 1-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 532u16 , name : "StoreFabric SN1200E 2-Port 16Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 558u16 , name : "Synergy 5330C 2-Port 32Gb Fibre Channel Mezz Card" } , SubSystem { subvendor : 6461u16 , subdevice : 4192u16 , name : "NIC-FC730i-Mb-2P" }] } , Device { vendor_id : 4319u16 , id : 61457u16 , name : "Saturn: LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61461u16 , name : "Saturn: LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61573u16 , name : "LP850 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61589u16 , name : "LP952 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61592u16 , name : "LP982 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61601u16 , name : "Thor LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61605u16 , name : "Thor LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61621u16 , name : "Viper LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61649u16 , name : "Helios LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61653u16 , name : "Helios LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61665u16 , name : "Zephyr LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61669u16 , name : "Zephyr LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61685u16 , name : "Neptune LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61696u16 , name : "LPe12000 Series 8Gb Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 906u16 , name : "8Gb PCI Express Dual Port FC Adapter for POWER" } , SubSystem { subvendor : 4156u16 , subdevice : 12930u16 , name : "8Gb Dual-port PCI-e FC HBA" } , SubSystem { subvendor : 4319u16 , subdevice : 61760u16 , name : "LPe12000-M8-L 1-Port 8Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 61761u16 , name : "LPe12002-M8-L 2-Port 8Gb PCIe Fibre Channel Adapter" }] } , Device { vendor_id : 4319u16 , id : 61713u16 , name : "Saturn-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61714u16 , name : "Saturn-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 61824u16 , name : "LPSe12002 EmulexSecure Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 62464u16 , name : "LPe35000/LPe36000 Series 32Gb/64Gb Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4319u16 , subdevice : 62465u16 , name : "LPe35000-M2 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62466u16 , name : "LPe35002-M2 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62467u16 , name : "LPe36000-M64 1-Port 64Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62468u16 , name : "LPe36002-M64 2-Port 64Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62469u16 , name : "LPe35004-M2 4-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62470u16 , name : "LPe35004-X6 4-Port Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62480u16 , name : "LPe35002-M2-D 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62481u16 , name : "LPe35000-M2-D 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62488u16 , name : "LPe35000-M2-L 1-Port 32Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62489u16 , name : "LPe35002-M2-L 2-Port 32Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62497u16 , name : "LPe36002-M2-L 2-Port 64Gb PCIe Fibre Channel Adapter" } , SubSystem { subvendor : 4319u16 , subdevice : 62498u16 , name : "LPe36002-M64-D 2-Port 64Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 725u16 , name : "StoreFabric SN1610E 1-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 726u16 , name : "StoreFabric SN1610E 2-Port 32Gb Fibre Channel Adapter" }] } , Device { vendor_id : 4319u16 , id : 62720u16 , name : "LPe37000/LPe38000 Series 32Gb/64Gb Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1729u16 , name : "PCIe4 4-Port 32Gb Fibre Channel Adapter for POWER (FC EN1L/EN1M; CCIN 2CFC)" } , SubSystem { subvendor : 4116u16 , subdevice : 1730u16 , name : "PCIe4 2-Port 64Gb Fibre Channel Adapter for POWER (FC EN1N/EN1P; CCIN 2CFD)" } , SubSystem { subvendor : 5520u16 , subdevice : 1070u16 , name : "Synergy 7330C 64Gb Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 1105u16 , name : "SN1620E 2-Port 32Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 1106u16 , name : "SN1720E 2-Port 64Gb Fibre Channel Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 1108u16 , name : "Synergy 5331C 32Gb Fibre Channel Host Bus Adapter" }] } , Device { vendor_id : 4319u16 , id : 62976u16 , name : "LPe37100S/LPe38100S Series 32Gb/64Gb Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63232u16 , name : "LP7000 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63233u16 , name : "LP7000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63488u16 , name : "LP8000 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63489u16 , name : "LP8000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63744u16 , name : "LP9000 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63745u16 , name : "LP9000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63872u16 , name : "LP9802 Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63873u16 , name : "LP9802 Fibre Channel Host Adapter Alternate ID" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 63874u16 , name : "LP9802 Fibre Channel Host Adapter Alternate ID" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64000u16 , name : "Thor-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64256u16 , name : "Viper LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64512u16 , name : "Thor-X LightPulse Fibre Channel Host Adapter" , subsystems : & [SubSystem { subvendor : 4319u16 , subdevice : 64512u16 , name : "LP10000 LightPulse Fibre Channel Host Adapter" }] } , Device { vendor_id : 4319u16 , id : 64528u16 , name : "Helios-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64544u16 , name : "Zephyr-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64576u16 , name : "Saturn-X: LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64592u16 , name : "Proteus-X: LightPulse IOV Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64768u16 , name : "Helios-X LightPulse Fibre Channel Host Adapter" , subsystems : & [SubSystem { subvendor : 4319u16 , subdevice : 64770u16 , name : "LightPulse LP11002 Dual-port 4Gigabit PCI Fibre Channel Adapter" }] } , Device { vendor_id : 4319u16 , id : 64785u16 , name : "Helios-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 64786u16 , name : "Helios-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 65024u16 , name : "Zephyr-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 65029u16 , name : "Zephyr-X: LightPulse FCoE Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 65041u16 , name : "Zephyr-X LightPulse Fibre Channel Host Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 65042u16 , name : "Zephyr-X LightPulse FCoE Adapter" , subsystems : & [] } , Device { vendor_id : 4319u16 , id : 65280u16 , name : "Neptune LightPulse Fibre Channel Host Adapter" , subsystems : & [] }] }),
        (7827, Vendor { id : 7827u16 , name : "Beijing Bytedance Network Technology Co., Ltd." , devices : & [] }),
        (5175, Vendor { id : 5175u16 , name : "Nissin Inc Co" , devices : & [] }),
        (4818, Vendor { id : 4818u16 , name : "NVidia / SGS Thomson (Joint Venture)" , devices : & [Device { vendor_id : 4818u16 , id : 8u16 , name : "NV1" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 9u16 , name : "DAC64" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 24u16 , name : "Riva128" , subsystems : & [SubSystem { subvendor : 4168u16 , subdevice : 3088u16 , name : "VICTORY Erazor" } , SubSystem { subvendor : 4168u16 , subdevice : 3093u16 , name : "VICTORY Erazor LT-8" } , SubSystem { subvendor : 4219u16 , subdevice : 32816u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4242u16 , subdevice : 848u16 , name : "Viper V330" } , SubSystem { subvendor : 4242u16 , subdevice : 4242u16 , name : "Viper V330" } , SubSystem { subvendor : 4276u16 , subdevice : 6939u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4276u16 , subdevice : 6941u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4276u16 , subdevice : 6942u16 , name : "STB Velocity 128, PAL TV-Out" } , SubSystem { subvendor : 4276u16 , subdevice : 6944u16 , name : "STB Velocity 128 Sapphire" } , SubSystem { subvendor : 4276u16 , subdevice : 6945u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4276u16 , subdevice : 6946u16 , name : "STB Velocity 128 AGP, NTSC TV-Out" } , SubSystem { subvendor : 4276u16 , subdevice : 6947u16 , name : "STB Velocity 128 AGP, PAL TV-Out" } , SubSystem { subvendor : 4276u16 , subdevice : 6951u16 , name : "STB Velocity 128 DVD" } , SubSystem { subvendor : 4276u16 , subdevice : 7048u16 , name : "MVP Pro 128" } , SubSystem { subvendor : 4276u16 , subdevice : 8746u16 , name : "STB Velocity 128 AGP" } , SubSystem { subvendor : 4276u16 , subdevice : 8752u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4276u16 , subdevice : 8754u16 , name : "STB Velocity 128" } , SubSystem { subvendor : 4276u16 , subdevice : 8757u16 , name : "STB Velocity 128 AGP" } , SubSystem { subvendor : 10773u16 , subdevice : 21667u16 , name : "3DVision-SAGP / 3DexPlorer 3000" }] } , Device { vendor_id : 4818u16 , id : 25u16 , name : "Riva128ZX" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 32u16 , name : "TNT" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 40u16 , name : "TNT2" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 41u16 , name : "UTNT2" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 44u16 , name : "VTNT2" , subsystems : & [] } , Device { vendor_id : 4818u16 , id : 160u16 , name : "ITNT2" , subsystems : & [] }] }),
        (4293, Vendor { id : 4293u16 , name : "Xerox Corporation" , devices : & [] }),
        (5036, Vendor { id : 5036u16 , name : "Motion Media Technology Ltd" , devices : & [] }),
        (4402, Vendor { id : 4402u16 , name : "Mitel Corp." , devices : & [] }),
        (7882, Vendor { id : 7882u16 , name : "Lightmatter" , devices : & [Device { vendor_id : 7882u16 , id : 0u16 , name : "Envise-B" , subsystems : & [] }] }),
        (7635, Vendor { id : 7635u16 , name : "Sage Microelectronics Corp." , devices : & [] }),
        (4520, Vendor { id : 4520u16 , name : "Systech Corp." , devices : & [] }),
        (4228, Vendor { id : 4228u16 , name : "Parador" , devices : & [] }),
        (4672, Vendor { id : 4672u16 , name : "Marathon Technologies Corp." , devices : & [] }),
        (6247, Vendor { id : 6247u16 , name : "Topspin Communications" , devices : & [Device { vendor_id : 6247u16 , id : 23108u16 , name : "MT23108 InfiniHost HCA" , subsystems : & [] } , Device { vendor_id : 6247u16 , id : 23109u16 , name : "MT23108 InfiniHost HCA flash recovery" , subsystems : & [] } , Device { vendor_id : 6247u16 , id : 23110u16 , name : "MT23108 InfiniHost HCA bridge" , subsystems : & [] } , Device { vendor_id : 6247u16 , id : 25208u16 , name : "MT25208 InfiniHost III Ex (Tavor compatibility mode)" , subsystems : & [] } , Device { vendor_id : 6247u16 , id : 25218u16 , name : "MT25208 InfiniHost III Ex" , subsystems : & [] }] }),
        (5327, Vendor { id : 5327u16 , name : "TEK Microsystems Inc." , devices : & [] }),
        (4340, Vendor { id : 4340u16 , name : "S-MOS Systems, Inc." , devices : & [] }),
        (5350, Vendor { id : 5350u16 , name : "SHINING Technology Inc" , devices : & [] }),
        (5115, Vendor { id : 5115u16 , name : "Aydin Corp" , devices : & [] }),
        (4258, Vendor { id : 4258u16 , name : "Quantum Corporation" , devices : & [] }),
        (32893, Vendor { id : 32893u16 , name : "Asustek Computer, Inc." , devices : & [] }),
        (5316, Vendor { id : 5316u16 , name : "IWASAKI Information Systems Co Ltd" , devices : & [] }),
        (4327, Vendor { id : 4327u16 , name : "Vadem" , devices : & [] }),
        (5193, Vendor { id : 5193u16 , name : "TUT Systems Inc" , devices : & [] }),
        (38672, Vendor { id : 38672u16 , name : "MosChip Semiconductor Technology Ltd." , devices : & [Device { vendor_id : 38672u16 , id : 37456u16 , name : "PCI-to-PCI bridge [MCS9250]" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 38917u16 , name : "PCI 1 port parallel adapter" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 38933u16 , name : "PCI 9815 Multi-I/O Controller" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 32u16 , name : "2P0S (2 port parallel adaptor)" }] } , Device { vendor_id : 38672u16 , id : 38944u16 , name : "PCI 9820 Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 38965u16 , name : "PCI 9835 Multi-I/O Controller" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 2u16 , name : "2S (16C550 UART)" } , SubSystem { subvendor : 4096u16 , subdevice : 18u16 , name : "1P2S" }] } , Device { vendor_id : 38672u16 , id : 38981u16 , name : "PCI 9845 Multi-I/O Controller" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 4u16 , name : "0P4S (4 port 16550A serial card)" } , SubSystem { subvendor : 4096u16 , subdevice : 6u16 , name : "0P6S (6 port 16550a serial card)" } , SubSystem { subvendor : 4096u16 , subdevice : 20u16 , name : "1P4S (1 Parallel / 4 16550A Serial Port Adapter)" }] } , Device { vendor_id : 38672u16 , id : 38997u16 , name : "PCI 9855 Multi-I/O Controller" , subsystems : & [SubSystem { subvendor : 4096u16 , subdevice : 20u16 , name : "1P4S" } , SubSystem { subvendor : 4096u16 , subdevice : 34u16 , name : "2P2S (2 Parallel / 2 16550A Serial Port Adapter)" }] } , Device { vendor_id : 38672u16 , id : 39013u16 , name : "PCI 9865 Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39168u16 , name : "MCS9900 Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39169u16 , name : "PCIe 9901 Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39172u16 , name : "4-Port PCIe Serial Adapter" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39186u16 , name : "PCIe 9912 Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39202u16 , name : "MCS9922 PCIe Multi-I/O Controller" , subsystems : & [] } , Device { vendor_id : 38672u16 , id : 39312u16 , name : "MCS9990 PCIe to 4-Port USB 2.0 Host Controller" , subsystems : & [] }] }),
        (4268, Vendor { id : 4268u16 , name : "Honeywell IAC" , devices : & [] }),
        (4639, Vendor { id : 4639u16 , name : "Arcus Technology, Inc." , devices : & [] }),
        (5271, Vendor { id : 5271u16 , name : "SMA Regelsysteme GmBH" , devices : & [Device { vendor_id : 5271u16 , id : 5271u16 , name : "SMA Technologie AG" , subsystems : & [] }] }),
        (5101, Vendor { id : 5101u16 , name : "Raytheion E-Systems" , devices : & [] }),
        (5381, Vendor { id : 5381u16 , name : "ITA INGENIEURBURO FUR TESTAUFGABEN GmbH" , devices : & [] }),
        (7734, Vendor { id : 7734u16 , name : "Shanghai Enflame Technology Co. Ltd" , devices : & [Device { vendor_id : 7734u16 , id : 1u16 , name : "T10 [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 2u16 , name : "T11 [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 3u16 , name : "T10(QSFP-DD) [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 33u16 , name : "T20(32GB) [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 34u16 , name : "T20(64GB) [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 35u16 , name : "T21(32GB) [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 36u16 , name : "T21(64GB) [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 32769u16 , name : "I20 [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 32785u16 , name : "I10 [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 32786u16 , name : "I10L [CloudBlazer]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 32817u16 , name : "S6 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 32818u16 , name : "S6 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49201u16 , name : "S30 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49202u16 , name : "S90 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49203u16 , name : "S60G [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49205u16 , name : "S60 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49217u16 , name : "L300 [Enflame]" , subsystems : & [] } , Device { vendor_id : 7734u16 , id : 49218u16 , name : "L600 [Enflame]" , subsystems : & [] }] }),
        (29316, Vendor { id : 29316u16 , name : "HT OMEGA Inc." , devices : & [] }),
        (6670, Vendor { id : 6670u16 , name : "DekTec Digital Video B.V." , devices : & [Device { vendor_id : 6670u16 , id : 2111u16 , name : "DTA-2111 VHF/UHF Modulator" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2144u16 , name : "DTA-2144(B) Quad ASI/SDI in+out" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2145u16 , name : "DTA-2145 ASI/SDI in+out with bypass relay" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2172u16 , name : "DTA-2172 Dual 3G-SDI/ASI ports" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2174u16 , name : "DTA-2174 Quad 3G-SDI/ASI in+out" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2175u16 , name : "DTA-2175 3G-SDI/ASI input+output with bypass relay" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 2178u16 , name : "DTA-2178 Octal 12G-SDI/ASI ports with genlock" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 43138u16 , name : "DTA-2178-ASI Octal ASI Ports" , subsystems : & [] } , Device { vendor_id : 6670u16 , id : 47230u16 , name : "DTA-2174B Quad 3G-SDI/ASI ports (1x12G) with genlock" , subsystems : & [] }] }),
        (5280, Vendor { id : 5280u16 , name : "SOFTING GmBH" , devices : & [] }),
        (7776, Vendor { id : 7776u16 , name : "Hailo Technologies Ltd." , devices : & [Device { vendor_id : 7776u16 , id : 10340u16 , name : "Hailo-8 AI Processor" , subsystems : & [] } , Device { vendor_id : 7776u16 , id : 17860u16 , name : "Hailo-10H AI Processor" , subsystems : & [] }] }),
        (5405, Vendor { id : 5405u16 , name : "Fujitsu Computer Products Of America" , devices : & [] }),
        (6798, Vendor { id : 6798u16 , name : "DRS Technologies" , devices : & [Device { vendor_id : 6798u16 , id : 8336u16 , name : "Model 2090 PCI Express" , subsystems : & [] }] }),
        (6307, Vendor { id : 6307u16 , name : "AT&T" , devices : & [] }),
        (4243, Vendor { id : 4243u16 , name : "National Instruments" , devices : & [Device { vendor_id : 4243u16 , id : 352u16 , name : "PCI-DIO-96" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 354u16 , name : "PCI-MIO-16XE-50" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4065u16 , name : "PXI-8320" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4432u16 , name : "PCI-6533 (PCI-DIO-32HS)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4464u16 , name : "PCI-MIO-16XE-10" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4480u16 , name : "PCI-MIO-16E-1" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4496u16 , name : "PCI-MIO-16E-4" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4528u16 , name : "PXI-6070E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4544u16 , name : "PXI-6040E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4560u16 , name : "PXI-6030E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4720u16 , name : "PCI-6032E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4752u16 , name : "PCI-6704" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4784u16 , name : "PCI-6534" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4880u16 , name : "PCI-6602" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4896u16 , name : "PXI-6533" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4912u16 , name : "PCI-6031E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4928u16 , name : "PCI-6033E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4944u16 , name : "PCI-6071E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 4960u16 , name : "PXI-6602" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5056u16 , name : "PXI-6508" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5264u16 , name : "PXI-6534" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5344u16 , name : "PCI-6110" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5360u16 , name : "PCI-6111" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5504u16 , name : "PXI-6031E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5552u16 , name : "PXI-6071E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 5904u16 , name : "PXI-6509" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6080u16 , name : "PXI-5690" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6096u16 , name : "PCI-6503" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6256u16 , name : "PCI-6713" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6272u16 , name : "PCI-6711" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6320u16 , name : "PCI-6052E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6336u16 , name : "PXI-6052E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6432u16 , name : "PXI-6704" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6448u16 , name : "PCI-6040E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6592u16 , name : "PCI-4472" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6816u16 , name : "PXI-4110" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6864u16 , name : "PCI-6133" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 6880u16 , name : "PXI-6133" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 7728u16 , name : "PCI-6624" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 7744u16 , name : "PXI-6624" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 7760u16 , name : "PXI-5404" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9232u16 , name : "PCI-6733" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9248u16 , name : "PXI-6733" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9264u16 , name : "PCI-6731" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9328u16 , name : "PCI-4474" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9376u16 , name : "PCI-4065" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9392u16 , name : "PXI-4200" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9456u16 , name : "PXI-4472" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9488u16 , name : "PCI-4472" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 9504u16 , name : "PCI-4474" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10144u16 , name : "PCI-6123" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10160u16 , name : "PXI-6123" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10368u16 , name : "DAQCard-6601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10384u16 , name : "PCI-6036E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10400u16 , name : "PXI-4461" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10416u16 , name : "PCI-6013" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10432u16 , name : "PCI-6014" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10448u16 , name : "PCI-5122" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10464u16 , name : "PXI-5122" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10736u16 , name : "PXI-7334" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10752u16 , name : "PXI-7344" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10848u16 , name : "PCI-6023E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10864u16 , name : "PCI-6024E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10880u16 , name : "PCI-6025E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 10928u16 , name : "PXI-6025E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11024u16 , name : "PXI-6527" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11040u16 , name : "PCI-6527" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11136u16 , name : "PXI-6713" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11152u16 , name : "PXI-6711" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11360u16 , name : "PCI-6601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11376u16 , name : "PXI-6601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11392u16 , name : "PCI-6035E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11408u16 , name : "PCI-6703" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11424u16 , name : "PCI-6034E" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11440u16 , name : "PCI-7344" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11456u16 , name : "PXI-6608" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11552u16 , name : "PXI-5600" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11696u16 , name : "PCI-6608" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11712u16 , name : "PCI-4070" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11728u16 , name : "PXI-4070" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11952u16 , name : "PXI-4472" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11968u16 , name : "PXI-6115" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 11984u16 , name : "PCI-6115" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12000u16 , name : "PXI-6120" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12016u16 , name : "PCI-6120" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12241u16 , name : "PCI-7334" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12242u16 , name : "PCI-7350" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12243u16 , name : "PCI-7342" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12245u16 , name : "PXI-7350" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 12246u16 , name : "PXI-7342" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28675u16 , name : "PCI-6551" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28676u16 , name : "PXI-6551" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28683u16 , name : "PXI-5421" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28684u16 , name : "PCI-5421" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28698u16 , name : "VXIpc-87xB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28699u16 , name : "VXIpc-770" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28707u16 , name : "PXI-2593" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28711u16 , name : "PCI-MXI-2 Universal" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28716u16 , name : "PXI-7831R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28717u16 , name : "PCI-7831R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28718u16 , name : "PXI-7811R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28719u16 , name : "PCI-7811R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28720u16 , name : "PCI-CAN (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28721u16 , name : "PCI-CAN/2 (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28722u16 , name : "PCI-CAN/LS (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28723u16 , name : "PCI-CAN/LS2 (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28724u16 , name : "PCI-CAN/DS (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28725u16 , name : "PXI-8460 (Series 2, 1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28726u16 , name : "PXI-8460 (Series 2, 2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28727u16 , name : "PXI-8461 (Series 2, 1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28728u16 , name : "PXI-8461 (Series 2, 2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28729u16 , name : "PXI-8462 (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28735u16 , name : "PXI-2566" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28736u16 , name : "PXI-2567" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28740u16 , name : "MXI-4 Connection Monitor" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28743u16 , name : "PXI-6653" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28748u16 , name : "PXI-2530" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28751u16 , name : "PXI-4220" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28752u16 , name : "PXI-4204" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28757u16 , name : "PXI-7830R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28758u16 , name : "PCI-7830R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28762u16 , name : "PCI-CAN/XS (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28763u16 , name : "PCI-CAN/XS2 (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28764u16 , name : "PXI-8464 (Series 2, 1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28765u16 , name : "PXI-8464 (Series 2, 2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28766u16 , name : "cRIO-9102" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28768u16 , name : "PXI-5610" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28772u16 , name : "PXI-1045 Trigger Routing Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28773u16 , name : "PXI-6652" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28774u16 , name : "PXI-6651" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28775u16 , name : "PXI-2529" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28776u16 , name : "PCI-CAN/SW (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28777u16 , name : "PCI-CAN/SW2 (Series 2)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28778u16 , name : "PXI-8463 (Series 2, 1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28779u16 , name : "PXI-8463 (Series 2, 2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28787u16 , name : "PCI-6723" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28788u16 , name : "PXI-7833R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28789u16 , name : "PXI-6552" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28790u16 , name : "PCI-6552" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28796u16 , name : "PXI-1428" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28798u16 , name : "PXI-4462" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28800u16 , name : "PXI-8430/2 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28801u16 , name : "PXI-8431/2 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28803u16 , name : "PCI-7833R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28805u16 , name : "PCI-6509" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28806u16 , name : "PXI-6528" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28807u16 , name : "PCI-6515" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28808u16 , name : "PCI-6514" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28812u16 , name : "PXI-2568" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28813u16 , name : "PXI-2569" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28841u16 , name : "PCI-6528" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28842u16 , name : "PCI-6229" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28843u16 , name : "PCI-6259" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28844u16 , name : "PCI-6289" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28845u16 , name : "PXI-6251" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28846u16 , name : "PXI-6220" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28847u16 , name : "PCI-6221" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28848u16 , name : "PCI-6220" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28849u16 , name : "PXI-6229" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28850u16 , name : "PXI-6259" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28851u16 , name : "PXI-6289" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28852u16 , name : "PCI-6250" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28853u16 , name : "PXI-6221" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28854u16 , name : "PCI-6280" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28855u16 , name : "PCI-6254" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28856u16 , name : "PCI-6251" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28857u16 , name : "PXI-6250" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28858u16 , name : "PXI-6254" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28859u16 , name : "PXI-6280" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28860u16 , name : "PCI-6284" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28861u16 , name : "PCI-6281" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28862u16 , name : "PXI-6284" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28863u16 , name : "PXI-6281" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28864u16 , name : "PCI-6143" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28867u16 , name : "PCI-6511" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28868u16 , name : "PXI-7330" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28869u16 , name : "PXI-7340" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28870u16 , name : "PCI-7330" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28871u16 , name : "PCI-7340" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28872u16 , name : "PCI-6513" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28873u16 , name : "PXI-6515" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28874u16 , name : "PCI-1405" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28876u16 , name : "PCI-6512" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28877u16 , name : "PXI-6514" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28878u16 , name : "PXI-1405" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28879u16 , name : "PCIe-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28880u16 , name : "PXI-2570" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28881u16 , name : "PXI-6513" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28882u16 , name : "PXI-6512" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28883u16 , name : "PXI-6511" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28884u16 , name : "PCI-6722" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28886u16 , name : "PXI-4072" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28887u16 , name : "PXI-6541" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28888u16 , name : "PXI-6542" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28889u16 , name : "PCI-6541" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28890u16 , name : "PCI-6542" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28891u16 , name : "PCI-8430/2 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28892u16 , name : "PCI-8431/2 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28893u16 , name : "PXI-8430/4 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28894u16 , name : "PXI-8431/4 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28895u16 , name : "PCI-8430/4 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28896u16 , name : "PCI-8431/4 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28897u16 , name : "PXI-2532" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28898u16 , name : "PXI-8430/8 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28899u16 , name : "PXI-8431/8 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28900u16 , name : "PCI-8430/8 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28901u16 , name : "PCI-8431/8 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28902u16 , name : "PXI-8430/16 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28903u16 , name : "PCI-8430/16 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28904u16 , name : "PXI-8432/2 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28905u16 , name : "PXI-8433/2 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28906u16 , name : "PCI-8432/2 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28907u16 , name : "PCI-8433/2 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28908u16 , name : "PXI-8432/4 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28909u16 , name : "PXI-8433/4 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28910u16 , name : "PCI-8432/4 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28911u16 , name : "PCI-8433/4 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28912u16 , name : "PXI-5922" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28913u16 , name : "PCI-5922" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28914u16 , name : "PCI-6224" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28915u16 , name : "PXI-6224" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28918u16 , name : "cRIO-9101" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28919u16 , name : "cRIO-9103" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28920u16 , name : "cRIO-9104" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28927u16 , name : "PXI-6723" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28928u16 , name : "PXI-6722" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28932u16 , name : "PCIx-1429" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28933u16 , name : "PCIe-1429" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28938u16 , name : "PXI-4071" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28941u16 , name : "PXI-6143" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28942u16 , name : "PCIe-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28943u16 , name : "PXI-5422" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28944u16 , name : "PCI-5422" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28945u16 , name : "PXI-5441" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28953u16 , name : "PXI-6561" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28954u16 , name : "PXI-6562" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28955u16 , name : "PCI-6561" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28956u16 , name : "PCI-6562" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28960u16 , name : "PCI-7390" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28961u16 , name : "PXI-5122EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28962u16 , name : "PCI-5122EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28963u16 , name : "PXIe-5653" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28964u16 , name : "PCI-6510" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28965u16 , name : "PCI-6516" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28966u16 , name : "PCI-6517" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28967u16 , name : "PCI-6518" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28968u16 , name : "PCI-6519" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28983u16 , name : "PXI-2575" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28988u16 , name : "PXI-2585" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28989u16 , name : "PXI-2586" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28994u16 , name : "PXI-4224" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28996u16 , name : "PXI-5124" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28997u16 , name : "PCI-5124" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28998u16 , name : "PCI-6132" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 28999u16 , name : "PXI-6132" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29000u16 , name : "PCI-6122" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29001u16 , name : "PXI-6122" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29004u16 , name : "PXI-5114" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29005u16 , name : "PCI-5114" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29008u16 , name : "PXI-2564" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29010u16 , name : "PCI-5640R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29014u16 , name : "PXI-1044 Trigger Routing Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29021u16 , name : "PCI-1426" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29031u16 , name : "PXI-5412" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29032u16 , name : "PCI-5412" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29035u16 , name : "PCI-6230" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29036u16 , name : "PCI-6225" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29037u16 , name : "PXI-6225" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29039u16 , name : "PCI-4461" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29040u16 , name : "PCI-4462" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29041u16 , name : "PCI-6010" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29044u16 , name : "PXI-8360" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29047u16 , name : "PXI-6230" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29053u16 , name : "PCIe-6251" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29055u16 , name : "PCIe-6259" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29063u16 , name : "PCI-1410" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29067u16 , name : "PCI-6521" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29068u16 , name : "PXI-6521" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29073u16 , name : "PCI-6154" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29075u16 , name : "PXI-7813R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29076u16 , name : "PCI-7813R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29077u16 , name : "PCI-8254R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29079u16 , name : "PXI-5402" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29080u16 , name : "PCI-5402" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29087u16 , name : "PCIe-6535" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29088u16 , name : "PCIe-6536" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29091u16 , name : "PXI-5650" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29092u16 , name : "PXI-5652" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29093u16 , name : "PXI-2594" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29095u16 , name : "PXI-2595" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29097u16 , name : "PXI-2596" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29098u16 , name : "PXI-2597" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29099u16 , name : "PXI-2598" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29100u16 , name : "PXI-2599" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29101u16 , name : "PCI-GPIB+" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29102u16 , name : "PCIe-1430" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29111u16 , name : "PXI-1056 Trigger Routing Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29112u16 , name : "PXI-1045 Trigger Routing Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29113u16 , name : "PXI-1044 Trigger Routing Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29115u16 , name : "PXI-2584" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29116u16 , name : "PCI-6221 (37-pin)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29119u16 , name : "PCIe-1427" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29125u16 , name : "PCI-6520" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29126u16 , name : "PXI-2576" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29127u16 , name : "cRIO-9072" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29148u16 , name : "PCI-1588" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29152u16 , name : "PCI-6255" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29153u16 , name : "PXI-6255" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29154u16 , name : "PXI-5406" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29155u16 , name : "PCI-5406" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29180u16 , name : "PXI-4022" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29193u16 , name : "PCI-6233" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29194u16 , name : "PXI-6233" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29195u16 , name : "PCI-6238" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29196u16 , name : "PXI-6238" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29280u16 , name : "PXI-5142" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29281u16 , name : "PCI-5142" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29293u16 , name : "PXI-5651" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29299u16 , name : "PXI-4461" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29300u16 , name : "PXI-4462" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29305u16 , name : "PCI-6232" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29306u16 , name : "PXI-6232" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29307u16 , name : "PCI-6239" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29308u16 , name : "PXI-6239" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29310u16 , name : "SMBus Controller" , subsystems : & [SubSystem { subvendor : 4243u16 , subdevice : 30124u16 , name : "PXIe-8388" } , SubSystem { subvendor : 4243u16 , subdevice : 30125u16 , name : "PXIe-8389" } , SubSystem { subvendor : 4243u16 , subdevice : 30288u16 , name : "PXIe-8381" } , SubSystem { subvendor : 4243u16 , subdevice : 33632u16 , name : "PXIe-8360" } , SubSystem { subvendor : 4243u16 , subdevice : 33648u16 , name : "PXIe-8370" } , SubSystem { subvendor : 4243u16 , subdevice : 33653u16 , name : "PXIe-8375" }] } , Device { vendor_id : 4243u16 , id : 29313u16 , name : "PCI-6236" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29314u16 , name : "PXI-6236" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29315u16 , name : "PXI-2554" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29320u16 , name : "PXIe-5611" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29331u16 , name : "PCIe-8255R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29341u16 , name : "cRIO-9074" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29348u16 , name : "PCIe-4065" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29351u16 , name : "PCIe-6537" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29352u16 , name : "PXI-5152" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29353u16 , name : "PCI-5152" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29354u16 , name : "PXI-5105" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29355u16 , name : "PCI-5105" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29368u16 , name : "PXI-6682" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29392u16 , name : "PXI-2545" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29393u16 , name : "PXI-2546" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29394u16 , name : "PXI-2547" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29395u16 , name : "PXI-2548" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29396u16 , name : "PXI-2549" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29397u16 , name : "PXI-2555" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29398u16 , name : "PXI-2556" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29399u16 , name : "PXI-2557" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29400u16 , name : "PXI-2558" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29401u16 , name : "PXI-2559" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29416u16 , name : "PXIe-6251" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29417u16 , name : "PXIe-6259" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29423u16 , name : "PXI-4498" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29424u16 , name : "PXI-4496" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29435u16 , name : "PXIe-6672" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29454u16 , name : "PXI-4130" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29455u16 , name : "PXI-5922EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29456u16 , name : "PCI-5922EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29468u16 , name : "PXI-2535" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29469u16 , name : "PXI-2536" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29474u16 , name : "PXIe-6124" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29479u16 , name : "PXI-6529" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29484u16 , name : "VXI-8360T" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29489u16 , name : "PXIe-5602" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29490u16 , name : "PXIe-5601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29491u16 , name : "PXI-5900" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29493u16 , name : "PXI-2533" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29494u16 , name : "PXI-2534" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29506u16 , name : "PXI-4461" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29513u16 , name : "PXI-5154" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29514u16 , name : "PCI-5154" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29527u16 , name : "PXI-4065" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29529u16 , name : "PXI-4495" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29552u16 , name : "PXI-4461" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29555u16 , name : "sbRIO-9601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29556u16 , name : "IOtech-9601" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29557u16 , name : "sbRIO-9602" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29560u16 , name : "sbRIO-9641" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29565u16 , name : "PXI-5124EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29572u16 , name : "PXI-7851R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29573u16 , name : "PXI-7852R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29574u16 , name : "PCIe-7851R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29575u16 , name : "PCIe-7852R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29584u16 , name : "PXI-7841R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29585u16 , name : "PXI-7842R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29586u16 , name : "PXI-7853R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29587u16 , name : "PCIe-7841R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29588u16 , name : "PCIe-7842R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29591u16 , name : "sbRIO-9611" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29592u16 , name : "sbRIO-9612" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29593u16 , name : "sbRIO-9631" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29594u16 , name : "sbRIO-9632" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29595u16 , name : "sbRIO-9642" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29601u16 , name : "PXIe-4498" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29602u16 , name : "PXIe-4496" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29605u16 , name : "PXIe-5641R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29607u16 , name : "PXI-8250 Chassis Monitor Module" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29608u16 , name : "PXI-8511 CAN/LS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29609u16 , name : "PXI-8511 CAN/LS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29610u16 , name : "PXI-8512 CAN/HS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29611u16 , name : "PXI-8512 CAN/HS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29612u16 , name : "PXI-8513 CAN/XS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29613u16 , name : "PXI-8513 CAN/XS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29615u16 , name : "PXI-8516 LIN" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29617u16 , name : "PXI-8517 FlexRay" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29618u16 , name : "PXI-8531 CANopen" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29619u16 , name : "PXI-8531 CANopen" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29620u16 , name : "PXI-8532 DeviceNet" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29621u16 , name : "PXI-8532 DeviceNet" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29622u16 , name : "PCI-8511 CAN/LS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29623u16 , name : "PCI-8511 CAN/LS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29624u16 , name : "PCI-8512 CAN/HS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29625u16 , name : "PCI-8512 CAN/HS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29626u16 , name : "PCI-8513 CAN/XS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29627u16 , name : "PCI-8513 CAN/XS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29629u16 , name : "PCI-8516 LIN" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29631u16 , name : "PCI-8517 FlexRay" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29632u16 , name : "PCI-8531 CANopen" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29633u16 , name : "PCI-8531 CANopen" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29634u16 , name : "PCI-8532 DeviceNet" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29635u16 , name : "PCI-8532 DeviceNet" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29637u16 , name : "PXIe-2527" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29638u16 , name : "PXIe-2529" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29640u16 , name : "PXIe-2530" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29641u16 , name : "PXIe-2532" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29642u16 , name : "PXIe-2569" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29643u16 , name : "PXIe-2575" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29644u16 , name : "PXIe-2593" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29653u16 , name : "PXI-7951R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29654u16 , name : "PXI-7952R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29655u16 , name : "PXI-7953R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29665u16 , name : "PXI-7854R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29676u16 , name : "PXI-7954R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29677u16 , name : "cRIO-9073" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29680u16 , name : "PXI-5153" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29681u16 , name : "PCI-5153" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29684u16 , name : "PXI-2515" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29686u16 , name : "cRIO-9111" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29687u16 , name : "cRIO-9112" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29688u16 , name : "cRIO-9113" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29689u16 , name : "cRIO-9114" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29690u16 , name : "cRIO-9116" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29691u16 , name : "cRIO-9118" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29700u16 , name : "PXI-4132" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29701u16 , name : "PXIe-6674T" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29702u16 , name : "PXIe-6674" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29710u16 , name : "PCIe-8430/16 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29711u16 , name : "PCIe-8430/8 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29712u16 , name : "PCIe-8431/16 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29713u16 , name : "PCIe-8431/8 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29716u16 , name : "PCIe-GPIB+" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29724u16 , name : "PXI-5691" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29725u16 , name : "PXI-5695" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29756u16 , name : "CSC-3059" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29768u16 , name : "PXI-2510" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29780u16 , name : "PXI-2512" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29781u16 , name : "PXI-2514" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29782u16 , name : "PXIe-2512" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29783u16 , name : "PXIe-2514" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29786u16 , name : "PXI-6682H" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29790u16 , name : "PXI-5153EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29791u16 , name : "PCI-5153EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29792u16 , name : "PXI-5154EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29793u16 , name : "PCI-5154EX" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29805u16 , name : "PXIe-5650" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29806u16 , name : "PXIe-5651" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29807u16 , name : "PXIe-5652" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29810u16 , name : "PXI-2800" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29845u16 , name : "PXIe-5603" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29847u16 , name : "PXIe-5605" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29870u16 , name : "PXIe-2515" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29876u16 , name : "PXI-2531" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29877u16 , name : "PXIe-2531" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29889u16 , name : "PXIe-8430/16 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29890u16 , name : "PXIe-8430/8 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29891u16 , name : "PXIe-8431/16 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29892u16 , name : "PXIe-8431/8 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29909u16 , name : "PXIe-5630" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29913u16 , name : "PCIe-8432/2 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29914u16 , name : "PCIe-8433/2 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29915u16 , name : "PCIe-8432/4 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29916u16 , name : "PCIe-8433/4 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29928u16 , name : "NI 9148" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29973u16 , name : "PCIe-8430/2 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29974u16 , name : "PCIe-8430/4 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29975u16 , name : "PCIe-8431/2 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29976u16 , name : "PCIe-8431/4 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29979u16 , name : "cRIO-9081" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29980u16 , name : "cRIO-9082" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29992u16 , name : "PXIe-4497" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29993u16 , name : "PXIe-4499" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 29994u16 , name : "PXIe-4492" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30009u16 , name : "NI 9157" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30010u16 , name : "NI 9159" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30104u16 , name : "PXI-2571" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30116u16 , name : "PXI-4131A" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30129u16 , name : "PCIe-7854R" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30138u16 , name : "PXI-2543" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30139u16 , name : "PXIe-2543" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30181u16 , name : "PXI-6683" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30182u16 , name : "PXI-6683H" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30191u16 , name : "PXIe-5632" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30236u16 , name : "VXI-8360LT" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30239u16 , name : "PXI-2540" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30240u16 , name : "PXIe-2540" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30241u16 , name : "PXI-2541" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30242u16 , name : "PXIe-2541" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30246u16 , name : "NI 9154" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30247u16 , name : "NI 9155" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30264u16 , name : "PXI-2720" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30265u16 , name : "PXI-2722" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30266u16 , name : "PXIe-2725" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30267u16 , name : "PXIe-2727" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30268u16 , name : "PXI-4465" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30283u16 , name : "PXIe-2790" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30284u16 , name : "PXI-2520" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30285u16 , name : "PXI-2521" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30286u16 , name : "PXI-2522" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30287u16 , name : "PXI-2523" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30290u16 , name : "PXIe-4080" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30292u16 , name : "PXI-2796" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30293u16 , name : "PXI-2797" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30294u16 , name : "PXI-2798" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30295u16 , name : "PXI-2799" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30301u16 , name : "PXI-2542" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30302u16 , name : "PXIe-2542" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30303u16 , name : "PXI-2544" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30304u16 , name : "PXIe-2544" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30317u16 , name : "PCIe-6535B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30318u16 , name : "PCIe-6536B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30319u16 , name : "PCIe-6537B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30371u16 , name : "PXIe-6535B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30372u16 , name : "PXIe-6536B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30373u16 , name : "PXIe-6537B" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30424u16 , name : "PXIe-4081" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30425u16 , name : "PXIe-4082" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30632u16 , name : "PXIe-6375" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30782u16 , name : "PXI-8368" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30850u16 , name : "PXIe-6376" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 30851u16 , name : "PXIe-6378" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 31134u16 , name : "PXIe-6386" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 31135u16 , name : "PXIe-6396" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36896u16 , name : "PXI-2501" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36912u16 , name : "PXI-2503" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36928u16 , name : "PXI-2527" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36944u16 , name : "PXI-2565" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36960u16 , name : "PXI-2590" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36976u16 , name : "PXI-2591" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 36992u16 , name : "PXI-2580" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 37008u16 , name : "PCI-4021" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 37024u16 , name : "PXI-4021" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 40961u16 , name : "PCI-MXI-2" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45057u16 , name : "PCI-1408" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45073u16 , name : "PXI-1408" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45089u16 , name : "PCI-1424" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45090u16 , name : "PXI-1424" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45105u16 , name : "PCI-1413" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45121u16 , name : "PCI-1407" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45137u16 , name : "PXI-1407" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45153u16 , name : "PCI-1411" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45169u16 , name : "PCI-1422" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45185u16 , name : "PXI-1422" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45201u16 , name : "PXI-1411" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45233u16 , name : "PCI-1409" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45249u16 , name : "PXI-1409" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 45281u16 , name : "PCI-1428" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 50372u16 , name : "PXIe/PCIe Device" , subsystems : & [SubSystem { subvendor : 4243u16 , subdevice : 29322u16 , name : "PXIe-5421" } , SubSystem { subvendor : 4243u16 , subdevice : 29323u16 , name : "PXIe-5442" } , SubSystem { subvendor : 4243u16 , subdevice : 29325u16 , name : "PXIe-5451" } , SubSystem { subvendor : 4243u16 , subdevice : 29346u16 , name : "PXIe-5122" } , SubSystem { subvendor : 4243u16 , subdevice : 29402u16 , name : "PXIe-5422" } , SubSystem { subvendor : 4243u16 , subdevice : 29431u16 , name : "PXIe-6535" } , SubSystem { subvendor : 4243u16 , subdevice : 29432u16 , name : "PXIe-6536" } , SubSystem { subvendor : 4243u16 , subdevice : 29433u16 , name : "PXIe-6537" } , SubSystem { subvendor : 4243u16 , subdevice : 29450u16 , name : "PXIe-5142" } , SubSystem { subvendor : 4243u16 , subdevice : 29478u16 , name : "PCIe-6509" } , SubSystem { subvendor : 4243u16 , subdevice : 29548u16 , name : "PXIe-4140" } , SubSystem { subvendor : 4243u16 , subdevice : 29579u16 , name : "PXIe-5622" } , SubSystem { subvendor : 4243u16 , subdevice : 29636u16 , name : "PXIe-5450" } , SubSystem { subvendor : 4243u16 , subdevice : 29639u16 , name : "PXIe-6545" } , SubSystem { subvendor : 4243u16 , subdevice : 29652u16 , name : "PXIe-6544" } , SubSystem { subvendor : 4243u16 , subdevice : 29733u16 , name : "PCIe-6320" } , SubSystem { subvendor : 4243u16 , subdevice : 29735u16 , name : "PCIe-6321" } , SubSystem { subvendor : 4243u16 , subdevice : 29736u16 , name : "PXIe-6323" } , SubSystem { subvendor : 4243u16 , subdevice : 29737u16 , name : "PCIe-6323" } , SubSystem { subvendor : 4243u16 , subdevice : 29738u16 , name : "PXIe-6341" } , SubSystem { subvendor : 4243u16 , subdevice : 29739u16 , name : "PCIe-6341" } , SubSystem { subvendor : 4243u16 , subdevice : 29740u16 , name : "PXIe-6343" } , SubSystem { subvendor : 4243u16 , subdevice : 29741u16 , name : "PCIe-6343" } , SubSystem { subvendor : 4243u16 , subdevice : 29743u16 , name : "PCIe-6351" } , SubSystem { subvendor : 4243u16 , subdevice : 29745u16 , name : "PCIe-6353" } , SubSystem { subvendor : 4243u16 , subdevice : 29746u16 , name : "PXIe-6361" } , SubSystem { subvendor : 4243u16 , subdevice : 29747u16 , name : "PCIe-6361" } , SubSystem { subvendor : 4243u16 , subdevice : 29748u16 , name : "PXIe-6363" } , SubSystem { subvendor : 4243u16 , subdevice : 29749u16 , name : "PCIe-6363" } , SubSystem { subvendor : 4243u16 , subdevice : 29750u16 , name : "PXIe-6356" } , SubSystem { subvendor : 4243u16 , subdevice : 29751u16 , name : "PXIe-6358" } , SubSystem { subvendor : 4243u16 , subdevice : 29752u16 , name : "PXIe-6366" } , SubSystem { subvendor : 4243u16 , subdevice : 29753u16 , name : "PXIe-6368" } , SubSystem { subvendor : 4243u16 , subdevice : 29800u16 , name : "PXIe-5185" } , SubSystem { subvendor : 4243u16 , subdevice : 29801u16 , name : "PXIe-5186" } , SubSystem { subvendor : 4243u16 , subdevice : 29842u16 , name : "PXIe-4300" } , SubSystem { subvendor : 4243u16 , subdevice : 29848u16 , name : "PXIe-6548" } , SubSystem { subvendor : 4243u16 , subdevice : 29849u16 , name : "PXIe-6547" } , SubSystem { subvendor : 4243u16 , subdevice : 29864u16 , name : "PXIe-4330" } , SubSystem { subvendor : 4243u16 , subdevice : 29865u16 , name : "PXIe-4331" } , SubSystem { subvendor : 4243u16 , subdevice : 29873u16 , name : "PXIe-4154" } , SubSystem { subvendor : 4243u16 , subdevice : 29874u16 , name : "PXIe-4353" } , SubSystem { subvendor : 4243u16 , subdevice : 29878u16 , name : "PCIe-1433" } , SubSystem { subvendor : 4243u16 , subdevice : 29901u16 , name : "PXIe-5643R" } , SubSystem { subvendor : 4243u16 , subdevice : 29904u16 , name : "PXIe-7961R" } , SubSystem { subvendor : 4243u16 , subdevice : 29917u16 , name : "PXIe-6376" } , SubSystem { subvendor : 4243u16 , subdevice : 29918u16 , name : "PXIe-6378" } , SubSystem { subvendor : 4243u16 , subdevice : 29922u16 , name : "PXIe-7962R" } , SubSystem { subvendor : 4243u16 , subdevice : 29923u16 , name : "PXIe-7965R" } , SubSystem { subvendor : 4243u16 , subdevice : 29925u16 , name : "PXIe-4844" } , SubSystem { subvendor : 4243u16 , subdevice : 29939u16 , name : "PCIe-5140" } , SubSystem { subvendor : 4243u16 , subdevice : 30012u16 , name : "PXIe-1435" } , SubSystem { subvendor : 4243u16 , subdevice : 30024u16 , name : "PXIe-5622 (25MHz DDC)" } , SubSystem { subvendor : 4243u16 , subdevice : 30029u16 , name : "PCIe-5155" } , SubSystem { subvendor : 4243u16 , subdevice : 30033u16 , name : "PXIe-6556" } , SubSystem { subvendor : 4243u16 , subdevice : 30035u16 , name : "PCIe-1473R" } , SubSystem { subvendor : 4243u16 , subdevice : 30064u16 , name : "PCIe-1474R" } , SubSystem { subvendor : 4243u16 , subdevice : 30065u16 , name : "PXIe-1475R" } , SubSystem { subvendor : 4243u16 , subdevice : 30066u16 , name : "PXIe-1476R" } , SubSystem { subvendor : 4243u16 , subdevice : 30114u16 , name : "PXIe-5693" } , SubSystem { subvendor : 4243u16 , subdevice : 30115u16 , name : "PXIe-5694" } , SubSystem { subvendor : 4243u16 , subdevice : 30117u16 , name : "PXIe-4141" } , SubSystem { subvendor : 4243u16 , subdevice : 30158u16 , name : "PXIe-7966R" } , SubSystem { subvendor : 4243u16 , subdevice : 30159u16 , name : "PXIe-4357" } , SubSystem { subvendor : 4243u16 , subdevice : 30162u16 , name : "PXIe-RevB-5643R" } , SubSystem { subvendor : 4243u16 , subdevice : 30163u16 , name : "PXIe-5644R" } , SubSystem { subvendor : 4243u16 , subdevice : 30190u16 , name : "PXIe-5645R" } , SubSystem { subvendor : 4243u16 , subdevice : 30227u16 , name : "PXIe-6555" } , SubSystem { subvendor : 4243u16 , subdevice : 30233u16 , name : "PXIe-5185" } , SubSystem { subvendor : 4243u16 , subdevice : 30234u16 , name : "PXIe-5186" } , SubSystem { subvendor : 4243u16 , subdevice : 30249u16 , name : "PXIe-4142" } , SubSystem { subvendor : 4243u16 , subdevice : 30250u16 , name : "PXIe-4143" } , SubSystem { subvendor : 4243u16 , subdevice : 30251u16 , name : "PXIe-4138" } , SubSystem { subvendor : 4243u16 , subdevice : 30252u16 , name : "PXIe-4144" } , SubSystem { subvendor : 4243u16 , subdevice : 30253u16 , name : "PXIe-4145" } , SubSystem { subvendor : 4243u16 , subdevice : 30254u16 , name : "PXIe-5606" } , SubSystem { subvendor : 4243u16 , subdevice : 30276u16 , name : "PXIe-4841" } , SubSystem { subvendor : 4243u16 , subdevice : 30282u16 , name : "PCIe-8237R-S" } , SubSystem { subvendor : 4243u16 , subdevice : 30290u16 , name : "PXIe-4080" } , SubSystem { subvendor : 4243u16 , subdevice : 30296u16 , name : "PXIe-5162 (4CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30379u16 , name : "PXIe-4322" } , SubSystem { subvendor : 4243u16 , subdevice : 30381u16 , name : "PXIe-4112" } , SubSystem { subvendor : 4243u16 , subdevice : 30382u16 , name : "PXIe-4113" } , SubSystem { subvendor : 4243u16 , subdevice : 30389u16 , name : "PXIe-7971R" } , SubSystem { subvendor : 4243u16 , subdevice : 30390u16 , name : "PXIe-7972R" } , SubSystem { subvendor : 4243u16 , subdevice : 30391u16 , name : "PXIe-7975R" } , SubSystem { subvendor : 4243u16 , subdevice : 30392u16 , name : "PXIe-5696" } , SubSystem { subvendor : 4243u16 , subdevice : 30393u16 , name : "PXIe-5654" } , SubSystem { subvendor : 4243u16 , subdevice : 30408u16 , name : "PXIe-6614" } , SubSystem { subvendor : 4243u16 , subdevice : 30409u16 , name : "PXIe-6612" } , SubSystem { subvendor : 4243u16 , subdevice : 30411u16 , name : "PXIe-5646R" } , SubSystem { subvendor : 4243u16 , subdevice : 30412u16 , name : "PXIe-5162 (2CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30414u16 , name : "CVS-1459" } , SubSystem { subvendor : 4243u16 , subdevice : 30416u16 , name : "PXIe-5160 (2CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30417u16 , name : "PXIe-5160 (4CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30424u16 , name : "PXIe-4081" } , SubSystem { subvendor : 4243u16 , subdevice : 30425u16 , name : "PXIe-4082" } , SubSystem { subvendor : 4243u16 , subdevice : 30428u16 , name : "PXIe-4610" } , SubSystem { subvendor : 4243u16 , subdevice : 30444u16 , name : "PXIe-2524" } , SubSystem { subvendor : 4243u16 , subdevice : 30445u16 , name : "PXIe-2525" } , SubSystem { subvendor : 4243u16 , subdevice : 30446u16 , name : "PXIe-2526" } , SubSystem { subvendor : 4243u16 , subdevice : 30447u16 , name : "PXIe-2737" } , SubSystem { subvendor : 4243u16 , subdevice : 30448u16 , name : "PXIe-2738" } , SubSystem { subvendor : 4243u16 , subdevice : 30449u16 , name : "PXIe-2739" } , SubSystem { subvendor : 4243u16 , subdevice : 30459u16 , name : "PCIe-1473R-LX110" } , SubSystem { subvendor : 4243u16 , subdevice : 30460u16 , name : "PXIe-5105" } , SubSystem { subvendor : 4243u16 , subdevice : 30461u16 , name : "PXIe-5114" } , SubSystem { subvendor : 4243u16 , subdevice : 30462u16 , name : "PXIe-5644R" } , SubSystem { subvendor : 4243u16 , subdevice : 30463u16 , name : "PXIe-5644R" } , SubSystem { subvendor : 4243u16 , subdevice : 30464u16 , name : "PXIe-5644R" } , SubSystem { subvendor : 4243u16 , subdevice : 30465u16 , name : "PXIe-5645R" } , SubSystem { subvendor : 4243u16 , subdevice : 30466u16 , name : "PXIe-5645R" } , SubSystem { subvendor : 4243u16 , subdevice : 30467u16 , name : "PXIe-5645R" } , SubSystem { subvendor : 4243u16 , subdevice : 30476u16 , name : "PXIe-4139" } , SubSystem { subvendor : 4243u16 , subdevice : 30481u16 , name : "PXIe-4464" } , SubSystem { subvendor : 4243u16 , subdevice : 30482u16 , name : "PXIe-4463" } , SubSystem { subvendor : 4243u16 , subdevice : 30486u16 , name : "PCIe-6612" } , SubSystem { subvendor : 4243u16 , subdevice : 30493u16 , name : "Unconfigured CA4 Switch" } , SubSystem { subvendor : 4243u16 , subdevice : 30494u16 , name : "PXIe-4339" } , SubSystem { subvendor : 4243u16 , subdevice : 30517u16 , name : "cRIO-9033" } , SubSystem { subvendor : 4243u16 , subdevice : 30526u16 , name : "PXIe-5624R" } , SubSystem { subvendor : 4243u16 , subdevice : 30539u16 , name : "cRIO-9031" } , SubSystem { subvendor : 4243u16 , subdevice : 30541u16 , name : "cRIO-9034" } , SubSystem { subvendor : 4243u16 , subdevice : 30549u16 , name : "cRIO-9030" } , SubSystem { subvendor : 4243u16 , subdevice : 30568u16 , name : "PXIe-2747" } , SubSystem { subvendor : 4243u16 , subdevice : 30569u16 , name : "PXIe-2748" } , SubSystem { subvendor : 4243u16 , subdevice : 30570u16 , name : "PXIe-2746" } , SubSystem { subvendor : 4243u16 , subdevice : 30583u16 , name : "PXIe-7976R" } , SubSystem { subvendor : 4243u16 , subdevice : 30594u16 , name : "PXIe-5646R" } , SubSystem { subvendor : 4243u16 , subdevice : 30595u16 , name : "PXIe-5646R" } , SubSystem { subvendor : 4243u16 , subdevice : 30596u16 , name : "PXIe-5646R" } , SubSystem { subvendor : 4243u16 , subdevice : 30608u16 , name : "PXIe-5170R (4CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30609u16 , name : "PXIe-5170R (8CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30611u16 , name : "PXIe-5171R (8CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30612u16 , name : "PXIe-5172 (4CH - 325T)" } , SubSystem { subvendor : 4243u16 , subdevice : 30613u16 , name : "PXIe-5172 (8CH - 410T)" } , SubSystem { subvendor : 4243u16 , subdevice : 30629u16 , name : "PXIe-6345" } , SubSystem { subvendor : 4243u16 , subdevice : 30630u16 , name : "PXIe-6355" } , SubSystem { subvendor : 4243u16 , subdevice : 30631u16 , name : "PXIe-6365" } , SubSystem { subvendor : 4243u16 , subdevice : 30632u16 , name : "PXIe-6375" } , SubSystem { subvendor : 4243u16 , subdevice : 30634u16 , name : "CVS-1458" } , SubSystem { subvendor : 4243u16 , subdevice : 30637u16 , name : "IC-3173" } , SubSystem { subvendor : 4243u16 , subdevice : 30644u16 , name : "PXIe-7820R" } , SubSystem { subvendor : 4243u16 , subdevice : 30645u16 , name : "PXIe-7821R" } , SubSystem { subvendor : 4243u16 , subdevice : 30646u16 , name : "PXIe-7822R" } , SubSystem { subvendor : 4243u16 , subdevice : 30649u16 , name : "cRIO-9038" } , SubSystem { subvendor : 4243u16 , subdevice : 30650u16 , name : "PXIe-4136" } , SubSystem { subvendor : 4243u16 , subdevice : 30651u16 , name : "PXIe-4137" } , SubSystem { subvendor : 4243u16 , subdevice : 30656u16 , name : "PXIe-5624R" } , SubSystem { subvendor : 4243u16 , subdevice : 30657u16 , name : "PXIe-5624R" } , SubSystem { subvendor : 4243u16 , subdevice : 30658u16 , name : "PXIe-5624R" } , SubSystem { subvendor : 4243u16 , subdevice : 30666u16 , name : "PXIe-6738" } , SubSystem { subvendor : 4243u16 , subdevice : 30667u16 , name : "PXIe-6739" } , SubSystem { subvendor : 4243u16 , subdevice : 30683u16 , name : "cRIO-9035" } , SubSystem { subvendor : 4243u16 , subdevice : 30684u16 , name : "cRIO-9036" } , SubSystem { subvendor : 4243u16 , subdevice : 30685u16 , name : "cRIO-9039" } , SubSystem { subvendor : 4243u16 , subdevice : 30722u16 , name : "PXIe-4302" } , SubSystem { subvendor : 4243u16 , subdevice : 30723u16 , name : "PXIe-4303" } , SubSystem { subvendor : 4243u16 , subdevice : 30725u16 , name : "PXIe-4305" } , SubSystem { subvendor : 4243u16 , subdevice : 30750u16 , name : "PXIe-4135" } , SubSystem { subvendor : 4243u16 , subdevice : 30752u16 , name : "PXIe-5164" } , SubSystem { subvendor : 4243u16 , subdevice : 30781u16 , name : "PXIe-6570" } , SubSystem { subvendor : 4243u16 , subdevice : 30801u16 , name : "PXIe-5172 (8CH - 325T)" } , SubSystem { subvendor : 4243u16 , subdevice : 30831u16 , name : "PXIe-4163" } , SubSystem { subvendor : 4243u16 , subdevice : 30849u16 , name : "PXIe-5163" } , SubSystem { subvendor : 4243u16 , subdevice : 30862u16 , name : "PXIe-4304" } , SubSystem { subvendor : 4243u16 , subdevice : 30933u16 , name : "PXIe-5413 (1CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30934u16 , name : "PXIe-5413 (2CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30935u16 , name : "PXIe-5423 (1CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30936u16 , name : "PXIe-5423 (2CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30937u16 , name : "PXIe-5433 (1CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30938u16 , name : "NI PXIe-5433 (2CH)" } , SubSystem { subvendor : 4243u16 , subdevice : 30968u16 , name : "NI FlexRIO Module (KU035)" } , SubSystem { subvendor : 4243u16 , subdevice : 30969u16 , name : "NI FlexRIO Module (KU040)" } , SubSystem { subvendor : 4243u16 , subdevice : 30970u16 , name : "NI FlexRIO Module (KU060)" } , SubSystem { subvendor : 4243u16 , subdevice : 30975u16 , name : "PXIe-4162" } , SubSystem { subvendor : 4243u16 , subdevice : 31023u16 , name : "PXIe-4190" } , SubSystem { subvendor : 4243u16 , subdevice : 31029u16 , name : "PXIe-5111" } , SubSystem { subvendor : 4243u16 , subdevice : 31030u16 , name : "PXIe-5110" } , SubSystem { subvendor : 4243u16 , subdevice : 31125u16 , name : "PXIe-7911R" } , SubSystem { subvendor : 4243u16 , subdevice : 31126u16 , name : "PXIe-7912R" } , SubSystem { subvendor : 4243u16 , subdevice : 31127u16 , name : "PXIe-7915R" } , SubSystem { subvendor : 4243u16 , subdevice : 31181u16 , name : "PXIe-5113" } , SubSystem { subvendor : 4243u16 , subdevice : 31187u16 , name : "NI FlexRIO PCIe Module (KU035)" } , SubSystem { subvendor : 4243u16 , subdevice : 31188u16 , name : "NI FlexRIO PCIe Module (KU040)" } , SubSystem { subvendor : 4243u16 , subdevice : 31189u16 , name : "NI FlexRIO PCIe Module (KU060)" } , SubSystem { subvendor : 4243u16 , subdevice : 31224u16 , name : "PXIe-6571" } , SubSystem { subvendor : 4243u16 , subdevice : 31254u16 , name : "PXIe-4147" } , SubSystem { subvendor : 4243u16 , subdevice : 31386u16 , name : "PXIe-4137 (40W)" } , SubSystem { subvendor : 4243u16 , subdevice : 31396u16 , name : "PXIe-4135 (40W)" } , SubSystem { subvendor : 4243u16 , subdevice : 31434u16 , name : "PXIe-4051" } , SubSystem { subvendor : 4243u16 , subdevice : 31435u16 , name : "PXIe-4150" } , SubSystem { subvendor : 4243u16 , subdevice : 31436u16 , name : "PXIe-4151" } , SubSystem { subvendor : 4243u16 , subdevice : 31456u16 , name : "PXIe-4163 (10 pA)" } , SubSystem { subvendor : 4243u16 , subdevice : 31457u16 , name : "PXIe-4162 (10 pA)" } , SubSystem { subvendor : 4243u16 , subdevice : 31471u16 , name : "PXIe-4190 (500 kHz)" } , SubSystem { subvendor : 4243u16 , subdevice : 31519u16 , name : "PXIe-6571 (8CH)" }] } , Device { vendor_id : 4243u16 , id : 51201u16 , name : "PCI-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 51217u16 , name : "PCI-GPIB+" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 51233u16 , name : "PXI-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 51249u16 , name : "PMC-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 51264u16 , name : "PCI-GPIB" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53552u16 , name : "PCI-232/2 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53568u16 , name : "PCI-232/4 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53584u16 , name : "PCI-232/8 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53600u16 , name : "PCI-485/2 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53616u16 , name : "PCI-485/4 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53648u16 , name : "PXI-8422/2 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53664u16 , name : "PXI-8422/4 (Isolated RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53680u16 , name : "PXI-8423/2 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53696u16 , name : "PXI-8423/4 (Isolated RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53712u16 , name : "PXI-8420/2 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53728u16 , name : "PXI-8420/4 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53744u16 , name : "PXI-8420/8 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53745u16 , name : "PXI-8420/16 (RS-232) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53808u16 , name : "PXI-8421/2 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53824u16 , name : "PXI-8421/4 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53840u16 , name : "PCI-232/2 (Isolated) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53856u16 , name : "PCI-485/2 (Isolated) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53872u16 , name : "PCI-232/4 (Isolated) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53888u16 , name : "PCI-485/4 (Isolated) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53904u16 , name : "PCI-485/8 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53920u16 , name : "PXI-8421/8 (RS-485) Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 53936u16 , name : "PCI-232/16 Interface" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57617u16 , name : "PCI-CAN" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57649u16 , name : "PXI-8461 (1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57665u16 , name : "PCI-CAN/LS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57681u16 , name : "PXI-8460 (1 port)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57873u16 , name : "PCI-CAN/2" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57905u16 , name : "PXI-8461 (2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57921u16 , name : "PCI-CAN/LS2" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57937u16 , name : "PXI-8460 (2 ports)" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57953u16 , name : "PCI-CAN/DS" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 57969u16 , name : "PXI-8462" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 61712u16 , name : "VMEpc-650" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 61728u16 , name : "VXIpc-650" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 65024u16 , name : "VXIpc-87x" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 65089u16 , name : "VXIpc-860" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 65105u16 , name : "VXIpc-74x" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 65121u16 , name : "VXIpc-850" , subsystems : & [] } , Device { vendor_id : 4243u16 , id : 65136u16 , name : "VXIpc-880" , subsystems : & [] }] }),
        (5499, Vendor { id : 5499u16 , name : "Star Multimedia Corp" , devices : & [] }),
        (5055, Vendor { id : 5055u16 , name : "Sharewave Inc" , devices : & [] }),
        (5079, Vendor { id : 5079u16 , name : "Toshiba Engineering Corporation" , devices : & [] }),
        (7294, Vendor { id : 7294u16 , name : "TTTech Computertechnik AG" , devices : & [Device { vendor_id : 7294u16 , id : 512u16 , name : "zFAS Debug Port" , subsystems : & [] }] }),
        (4184, Vendor { id : 4184u16 , name : "Electronics & Telecommunications RSH" , devices : & [] }),
        (5184, Vendor { id : 5184u16 , name : "ALGOL Corp." , devices : & [] }),
        (5446, Vendor { id : 5446u16 , name : "IOI Technology Corp" , devices : & [] }),
        (65242, Vendor { id : 65242u16 , name : "Broadcom Inc" , devices : & [Device { vendor_id : 65242u16 , id : 41210u16 , name : "BCM4210 iLine10 HomePNA 2.0" , subsystems : & [] } , Device { vendor_id : 65242u16 , id : 41230u16 , name : "BCM4230 iLine10 HomePNA 2.0" , subsystems : & [] }] }),
        (4476, Vendor { id : 4476u16 , name : "ATTO Technology, Inc." , devices : & [Device { vendor_id : 4476u16 , id : 44u16 , name : "ExpressSAS R380" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 45u16 , name : "ExpressSAS R348" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 48u16 , name : "Ultra320 SCSI Host Adapter" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 32787u16 , name : "ExpressPCI UL4D" } , SubSystem { subvendor : 4476u16 , subdevice : 32788u16 , name : "ExpressPCI UL4S" } , SubSystem { subvendor : 4476u16 , subdevice : 32807u16 , name : "ExpressPCI UL5D" } , SubSystem { subvendor : 4476u16 , subdevice : 32815u16 , name : "ExpressPCI UL5D Low Profile" }] } , Device { vendor_id : 4476u16 , id : 51u16 , name : "SAS Adapter" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 65u16 , name : "ExpressSAS R30F" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 66u16 , name : "ExpressSAS 6Gb/s SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 66u16 , name : "ExpressSAS H680" } , SubSystem { subvendor : 4476u16 , subdevice : 67u16 , name : "ExpressSAS H608" } , SubSystem { subvendor : 4476u16 , subdevice : 68u16 , name : "ExpressSAS H60F" } , SubSystem { subvendor : 4476u16 , subdevice : 69u16 , name : "ExpressSAS H6F0" } , SubSystem { subvendor : 4476u16 , subdevice : 70u16 , name : "ExpressSAS H644" } , SubSystem { subvendor : 4476u16 , subdevice : 79u16 , name : "ExpressSAS M608" } , SubSystem { subvendor : 4476u16 , subdevice : 87u16 , name : "ExpressSAS M680" } , SubSystem { subvendor : 4476u16 , subdevice : 88u16 , name : "ExpressSAS M644" } , SubSystem { subvendor : 4476u16 , subdevice : 89u16 , name : "ExpressSAS W608" } , SubSystem { subvendor : 4476u16 , subdevice : 90u16 , name : "ExpressSAS W680" } , SubSystem { subvendor : 4476u16 , subdevice : 91u16 , name : "ExpressSAS W644" }] } , Device { vendor_id : 4476u16 , id : 73u16 , name : "ExpressSAS 6Gb SAS/SATA RAID Adapter" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 73u16 , name : "ExpressSAS R680" } , SubSystem { subvendor : 4476u16 , subdevice : 74u16 , name : "ExpressSAS R608" } , SubSystem { subvendor : 4476u16 , subdevice : 75u16 , name : "ExpressSAS R60F" } , SubSystem { subvendor : 4476u16 , subdevice : 76u16 , name : "ExpressSAS R6F0" } , SubSystem { subvendor : 4476u16 , subdevice : 77u16 , name : "ExpressSAS R644" } , SubSystem { subvendor : 4476u16 , subdevice : 78u16 , name : "ExpressSAS R648" }] } , Device { vendor_id : 4476u16 , id : 100u16 , name : "Celerity FC 16Gb/s Gen 5 Fibre Channel HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 99u16 , name : "Celerity FC-161E" } , SubSystem { subvendor : 4476u16 , subdevice : 100u16 , name : "Celerity FC-162E" } , SubSystem { subvendor : 4476u16 , subdevice : 101u16 , name : "Celerity FC-164E" }] } , Device { vendor_id : 4476u16 , id : 148u16 , name : "Celerity FC 16/32Gb/s Gen 6 Fibre Channel HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 148u16 , name : "Celerity FC-162P" } , SubSystem { subvendor : 4476u16 , subdevice : 160u16 , name : "Celerity FC-161P" } , SubSystem { subvendor : 4476u16 , subdevice : 161u16 , name : "Celerity FC-164P" } , SubSystem { subvendor : 4476u16 , subdevice : 162u16 , name : "Celerity FC-321E" } , SubSystem { subvendor : 4476u16 , subdevice : 163u16 , name : "Celerity FC-322E" } , SubSystem { subvendor : 4476u16 , subdevice : 172u16 , name : "Celerity FC-324E" } , SubSystem { subvendor : 4476u16 , subdevice : 16550u16 , name : "ThunderLink TLFC-3162" } , SubSystem { subvendor : 4476u16 , subdevice : 16551u16 , name : "ThunderLink TLFC-3322" }] } , Device { vendor_id : 4476u16 , id : 165u16 , name : "ExpressSAS 24Gb/s SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 208u16 , name : "ExpressSAS H24F0" } , SubSystem { subvendor : 4476u16 , subdevice : 209u16 , name : "ExpressSAS H240F" } , SubSystem { subvendor : 4476u16 , subdevice : 210u16 , name : "ExpressSAS H240N" }] } , Device { vendor_id : 4476u16 , id : 187u16 , name : "Celerity FC 32/64Gb/s Gen 7 Fibre Channel HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 188u16 , name : "Celerity FC-321P" } , SubSystem { subvendor : 4476u16 , subdevice : 189u16 , name : "Celerity FC-322P" } , SubSystem { subvendor : 4476u16 , subdevice : 190u16 , name : "Celerity FC-324P" } , SubSystem { subvendor : 4476u16 , subdevice : 201u16 , name : "Celerity FC-641E" } , SubSystem { subvendor : 4476u16 , subdevice : 202u16 , name : "Celerity FC-642E" } , SubSystem { subvendor : 4476u16 , subdevice : 212u16 , name : "Celerity FC-644E" } , SubSystem { subvendor : 4476u16 , subdevice : 16600u16 , name : "ThunderLink FC 5642" } , SubSystem { subvendor : 4476u16 , subdevice : 16601u16 , name : "ThunderLink FC 5322" }] } , Device { vendor_id : 4476u16 , id : 197u16 , name : "ExpressNVM PCIe Gen4 Switch" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 198u16 , name : "ExpressNVM S48F PCIe Gen4" } , SubSystem { subvendor : 4476u16 , subdevice : 203u16 , name : "ExpressNVM S4FF PCIe Gen4" }] } , Device { vendor_id : 4476u16 , id : 230u16 , name : "ExpressSAS GT 12Gb/s SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 192u16 , name : "ExpressSAS H1280 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 193u16 , name : "ExpressSAS H1208 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 194u16 , name : "ExpressSAS H1244 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 195u16 , name : "ExpressSAS H12F0 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 196u16 , name : "ExpressSAS H120F GT" } , SubSystem { subvendor : 4476u16 , subdevice : 225u16 , name : "ExpressSAS H1280 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 226u16 , name : "ExpressSAS H1208 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 227u16 , name : "ExpressSAS H1244 GT" } , SubSystem { subvendor : 4476u16 , subdevice : 16608u16 , name : "ThunderLink SH 5128" } , SubSystem { subvendor : 4476u16 , subdevice : 16612u16 , name : "ThunderLink SH 5128" }] } , Device { vendor_id : 4476u16 , id : 32787u16 , name : "ExpressPCI UL4D" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 32788u16 , name : "ExpressPCI UL4S" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 32807u16 , name : "ExpressPCI UL5D" , subsystems : & [] } , Device { vendor_id : 4476u16 , id : 32880u16 , name : "ExpressSAS 12Gb/s SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 112u16 , name : "ExpressSAS H1280" } , SubSystem { subvendor : 4476u16 , subdevice : 113u16 , name : "ExpressSAS H1208" } , SubSystem { subvendor : 4476u16 , subdevice : 128u16 , name : "ExpressSAS H1244" } , SubSystem { subvendor : 4476u16 , subdevice : 16558u16 , name : "ThunderLink SH 3128" }] } , Device { vendor_id : 4476u16 , id : 32882u16 , name : "ExpressSAS 12Gb/s SAS/SATA HBA" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 114u16 , name : "ExpressSAS H12F0" } , SubSystem { subvendor : 4476u16 , subdevice : 115u16 , name : "ExpressSAS H120F" } , SubSystem { subvendor : 4476u16 , subdevice : 129u16 , name : "ExpressSAS H1248" } , SubSystem { subvendor : 4476u16 , subdevice : 130u16 , name : "ExpressSAS H1288" }] }] }),
        (5355, Vendor { id : 5355u16 , name : "SEIKO EPSON Corp" , devices : & [] }),
        (16713, Vendor { id : 16713u16 , name : "AIMOTIVE Kft." , devices : & [Device { vendor_id : 16713u16 , id : 53265u16 , name : "HIL camera injector board" , subsystems : & [] }] }),
        (4104, Vendor { id : 4104u16 , name : "Epson" , devices : & [] }),
        (4503, Vendor { id : 4503u16 , name : "Gage Applied Sciences, Inc." , devices : & [Device { vendor_id : 4503u16 , id : 268u16 , name : "CompuScope 82G 8bit 2GS/s Analog Input Card" , subsystems : & [] }] }),
        (5631, Vendor { id : 5631u16 , name : "Fastpoint Technologies Inc" , devices : & [] }),
        (16389, Vendor { id : 16389u16 , name : "Avance Logic Inc." , devices : & [Device { vendor_id : 16389u16 , id : 768u16 , name : "ALS300 PCI Audio Device" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 776u16 , name : "ALS300+ PCI Audio Device" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 777u16 , name : "PCI Input Controller" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 4196u16 , name : "ALG-2064" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 8292u16 , name : "ALG-2064i" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 8488u16 , name : "ALG-2364A GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 8961u16 , name : "ALG-2301" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 8962u16 , name : "ALG-2302" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 8963u16 , name : "AVG-2302 GUI Accelerator" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 9060u16 , name : "ALG-2364A" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 9316u16 , name : "ALG-2464" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 9473u16 , name : "ALG-2564A/25128A" , subsystems : & [] } , Device { vendor_id : 16389u16 , id : 16384u16 , name : "ALS4000 Audio Chipset" , subsystems : & [SubSystem { subvendor : 16389u16 , subdevice : 16384u16 , name : "ALS4000 Audio Chipset" }] } , Device { vendor_id : 16389u16 , id : 18192u16 , name : "ALC200/200P" , subsystems : & [] }] }),
        (7349, Vendor { id : 7349u16 , name : "Focusrite Audio Engineering Ltd" , devices : & [Device { vendor_id : 7349u16 , id : 2u16 , name : "Clarett" , subsystems : & [] }] }),
        (4674, Vendor { id : 4674u16 , name : "JNI Corporation" , devices : & [Device { vendor_id : 4674u16 , id : 5472u16 , name : "JNIC-1560 PCI-X Fibre Channel Controller" , subsystems : & [SubSystem { subvendor : 4674u16 , subdevice : 25954u16 , name : "FCX2-6562 Dual Channel PCI-X Fibre Channel Adapter" } , SubSystem { subvendor : 4674u16 , subdevice : 25962u16 , name : "FCX-6562 PCI-X Fibre Channel Adapter" }] } , Device { vendor_id : 4674u16 , id : 17987u16 , name : "FCI-1063 Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4674u16 , id : 25954u16 , name : "FCX2-6562 Dual Channel PCI-X Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4674u16 , id : 25962u16 , name : "FCX-6562 PCI-X Fibre Channel Adapter" , subsystems : & [] }] }),
        (5030, Vendor { id : 5030u16 , name : "Videonics Inc" , devices : & [] }),
        (18767, Vendor { id : 18767u16 , name : "ACCES I/O Products, Inc." , devices : & [Device { vendor_id : 18767u16 , id : 1288u16 , name : "PCI-IDO-16A FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 1304u16 , name : "PCI-IDO-32A FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 1312u16 , name : "PCI-IDO-48 FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 1313u16 , name : "PCI-IDO-48A FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 1795u16 , name : "PCIe-RO-4 Electromechanical Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 2000u16 , name : "PCIe-IDO-24 FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 2336u16 , name : "PCI-IDI-48 Isolated Digital Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3024u16 , name : "PCIe-IDI-24 Isolated Digital Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3152u16 , name : "PCI-DIO-24H 1x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3153u16 , name : "PCI-DIO-24D 1x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3154u16 , name : "PCIe-DIO-24 1x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3155u16 , name : "PCIe-DIO-24H 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3159u16 , name : "mPCIe-DIO-24 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3168u16 , name : "PCI-DIO-48H 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3169u16 , name : "PCIe-DIO-48 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3170u16 , name : "P104-DIO-48 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3176u16 , name : "PCI-DIO-72 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3177u16 , name : "P104-DIO-96 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3184u16 , name : "PCI-DIO-96 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3192u16 , name : "PCI-DIO-120 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3528u16 , name : "PCI-IDIO-16 Isolated Digital Input / FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3664u16 , name : "PCI-DIO-24S 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3665u16 , name : "PCI-DIO-24H(C) 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3666u16 , name : "PCI-DIO-24D(C) 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3667u16 , name : "PCIe-DIO-24S 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3668u16 , name : "PCIe-DIO-24HS 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3669u16 , name : "PCIe-DIO-24DC 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3670u16 , name : "PCIe-DIO-24DCS 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3671u16 , name : "mPCIe-DIO-24S 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3680u16 , name : "PCI-DIO-48S 2x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3681u16 , name : "PCIe-DIO-48S 2x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3682u16 , name : "P104-DIO-48S 2x 8255 Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3840u16 , name : "PCI-IIRO-8 Isolated Digital / Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3841u16 , name : "LPCI-IIRO-8 Isolated Digital / Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3842u16 , name : "PCIe-IIRO-8 Isolated Digital / Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3848u16 , name : "PCI-IIRO-16 Isolated Digital / Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 3849u16 , name : "PCIe-IIRO-16 Isolated Digital / Relay Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4032u16 , name : "PCIe-IDIO-12 Isolated Digital Input / FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4033u16 , name : "PCIe-IDI-12 Isolated Digital Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4034u16 , name : "PCIe-IDO-12 FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4048u16 , name : "PCIe-IDIO-24 Isolated Digital Input / FET Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4176u16 , name : "PCI-422/485-2 2x RS422/RS484 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4177u16 , name : "PCIe-COM-2SRJ 2x RS422/RS484 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4178u16 , name : "104I-COM-2S 2x RS422/RS484 PCI/104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4179u16 , name : "mPCIe-COM-2S 2x RS422/RS484 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4184u16 , name : "PCI-COM422/4 4x RS422 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4185u16 , name : "PCI-COM485/4 4x RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4186u16 , name : "PCIe-COM422-4 4x RS422 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4187u16 , name : "PCIe-COM485-4 4x RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4188u16 , name : "PCIe-COM-4SRJ 4x RS422/RS485 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4189u16 , name : "104I-COM-4S 4x RS422/RS484 PCI/104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4190u16 , name : "mPCIe-COM-4S 4x RS422/RS484 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4200u16 , name : "PCI-COM422/8 8x RS422 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4201u16 , name : "PCI-COM485/8 8x RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4202u16 , name : "PCIe-COM422-8 8x RS422 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4203u16 , name : "PCIe-COM485-8 8x RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4204u16 , name : "104I-COM-8S 8x RS422/RS485 PCI/104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4232u16 , name : "PCI-COM232/1 1x RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4240u16 , name : "PCI-COM232/2 2x RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4241u16 , name : "PCIe-COM232-2RJ 2x RS232 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4243u16 , name : "mPCIe-COM232-2 2x RS232 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4248u16 , name : "PCIe-COM232-4 4x RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4249u16 , name : "PCIe-COM232-4RJ 4x RS232 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4251u16 , name : "mPCIe-COM232-4 4x RS232 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4264u16 , name : "P104-COM232-8 8x RS232 PC-104+ Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4265u16 , name : "PCIe-COM232-8 8x RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4297u16 , name : "PCI-COM-1S 1x RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4304u16 , name : "PCI-COM2S 2x RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4305u16 , name : "PCIe-COM-2SMRJ 2x RS232/RS422/RS485 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4306u16 , name : "104I-COM-2SM 2x RS232/RS422/RS485 PCI/104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4307u16 , name : "mPCIe-COM-2SM 2x RS232/RS422/RS485 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4312u16 , name : "PCI-COM-4SM 4x RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4313u16 , name : "PCIe-COM-4SM 4x RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4314u16 , name : "PCIe-COM-4SMRJ 4x RS232/RS422/RS485 Card w/RJ45 Connectors" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4315u16 , name : "104I-COM-4SM 4x RS232/RS422/RS485 PCI/104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4316u16 , name : "mPCIe-COM-4SM 4x RS232/RS422/RS485 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4328u16 , name : "PCI-COM-8SM 8x RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4329u16 , name : "PCIe-COM-8SM 8x RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4330u16 , name : "104I-COM-8SM 8x RS232/RS422/RS485 PCI-104 Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4360u16 , name : "mPCIe-ICM485-1 1x Isolated RS485 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4368u16 , name : "mPCIe-ICM422-2 2x Isolated RS422 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4369u16 , name : "mPCIe-ICM485-2 2x Isolated RS485 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4376u16 , name : "mPCIe-ICM422-4 4x Isolated RS422 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4377u16 , name : "mPCIe-ICM485-4 4x Isolated RS485 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4424u16 , name : "PCI-ICM-1S 1x Isolated RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4432u16 , name : "PCI-ICM-2S 2x Isolated RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4434u16 , name : "PCIe-ICM-2S 2x Isolated RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4440u16 , name : "PCI-ICM422/4 4x Isolated RS422 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4441u16 , name : "PCI-ICM485/4 4x Isolated RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4442u16 , name : "PCIe-ICM-4S 4x Isolated RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4496u16 , name : "PCIe-ICM232-2 2x Isolated RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4497u16 , name : "mPCIe-ICM232-2 2x Isolated RS232 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4504u16 , name : "PCIe-ICM232-4 4x Isolated RS232 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4505u16 , name : "mPCIe-ICM232-4 4x Isolated RS422 PCI Express Mini Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4560u16 , name : "PCIe-ICM-2SM 2x Isolated RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4568u16 , name : "PCIe-ICM-4SM 4x Isolated RS232/RS422/RS485 Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4688u16 , name : "PCI-WDG-2S Watchdog and 2x Serial Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 4816u16 , name : "PCI-WDG-IMPAC" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 8752u16 , name : "PCI-QUAD-8 8x Quadrature Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 8753u16 , name : "PCI-QUAD-4 4x Quadrature Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 8896u16 , name : "PCI-WDG-CSM Watchdog Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 9664u16 , name : "P104-WDG-E Watchdog PC/104+ Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 11344u16 , name : "PCI-DIO-96CT 96x Digital Input / Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 11352u16 , name : "PCI-DIO-96C3 96x Digital Input / Output Card w/3x 8254 Counter Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 12000u16 , name : "PCIe-DIO24S-CTR12 24x Digital Input / Output Card w/4x 8254 Counter Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 12224u16 , name : "P104-WDG-CSM Watchdog PC/104+ Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 12225u16 , name : "P104-WDG-CSMA Advanced Watchdog PC/104+ Board" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 24272u16 , name : "PCI-DAC" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27792u16 , name : "PCI-DA12-2 2x 12-bit Analog Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27800u16 , name : "PCI-DA12-4 4x 12-bit Analog Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27808u16 , name : "PCI-DA12-6 6x 12-bit Analog Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27816u16 , name : "PCI-DA12-8 8x 12-bit Analog Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27817u16 , name : "PCI-DA12-8V" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27824u16 , name : "PCI-DA12-16 16x 12-bit Analog Output Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 27825u16 , name : "PCI-DA12-16V" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 36592u16 , name : "P104-FAS16-16" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 44200u16 , name : "PCI-AI12-16 12-bit 100kHz Analog Input Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 44201u16 , name : "PCI-AI12-16A 12-bit 100kHz Analog Input w/FIFO Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 60584u16 , name : "PCI-AIO12-16 12-bit 100kHz Analog Input w/2x Analog Output and FIFO Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 60586u16 , name : "PCI-A12-16A 12-bit 100kHz Analog Input w/2x Analog Output and FIFO Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 60648u16 , name : "LPCI-A16-16A 16-bit 500kHz Analog Input low-profile Card" , subsystems : & [] } , Device { vendor_id : 18767u16 , id : 60649u16 , name : "LPCI-AIO16A 16-bit 500kHz Analog Input low-profile Card" , subsystems : & [] }] }),
        (43521, Vendor { id : 43521u16 , name : "iTuner" , devices : & [] }),
        (4522, Vendor { id : 4522u16 , name : "Actel" , devices : & [] }),
        (5080, Vendor { id : 5080u16 , name : "Phobos corporation" , devices : & [] }),
        (6849, Vendor { id : 6849u16 , name : "Global Unichip Corp." , devices : & [Device { vendor_id : 6849u16 , id : 2202u16 , name : "Coral Edge TPU" , subsystems : & [] }] }),
        (16720, Vendor { id : 16720u16 , name : "ONA Electroerosion" , devices : & [Device { vendor_id : 16720u16 , id : 1u16 , name : "PCI32TLITE FILSTRUP1 PCI to VME Bridge Controller" , subsystems : & [] } , Device { vendor_id : 16720u16 , id : 6u16 , name : "PCI32TLITE UART 16550 Opencores" , subsystems : & [] } , Device { vendor_id : 16720u16 , id : 7u16 , name : "PCI32TLITE CAN Controller Opencores" , subsystems : & [] }] }),
        (4963, Vendor { id : 4963u16 , name : "Phoenix Technology Ltd" , devices : & [] }),
        (4523, Vendor { id : 4523u16 , name : "Marvell Technology Group Ltd." , devices : & [Device { vendor_id : 4523u16 , id : 256u16 , name : "88F3700 [Armada 3700 Family] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 272u16 , name : "88F60x0/88F70x0/88F80x0/CN913x ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 326u16 , name : "GT-64010/64010A System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 3923u16 , name : "88E6318 Link Street network controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 4523u16 , name : "MV88SE614x SATA II PCI-E controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 5007u16 , name : "W8300 802.11 Adapter (rev 07)" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 8102u16 , name : "Marvell W8300 802.11 Adapter" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 15112u16 , name : "AirPlus G DWL-G630 Wireless Cardbus Adapter (rev.A1)" } , SubSystem { subvendor : 4486u16 , subdevice : 15113u16 , name : "AirPlus G DWL-G510 Wireless G PCI Card" }] } , Device { vendor_id : 4523u16 , id : 8103u16 , name : "88W8310 and 88W8000G [Libertas] 802.11g client chipset" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 8106u16 , name : "88w8335 [Libertas] 802.11b/g Wireless" , subsystems : & [SubSystem { subvendor : 4486u16 , subdevice : 15137u16 , name : "EH101 Wireless G Notebook Adapter" } , SubSystem { subvendor : 4486u16 , subdevice : 15138u16 , name : "EH102 Wireless G Desktop Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 19968u16 , name : "WG511v2 54 Mbps Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 27392u16 , name : "WG311v3 802.11g Wireless PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 64u16 , name : "WPC54G v5 802.11g Wireless-G Notebook Adapter" } , SubSystem { subvendor : 42791u16 , subdevice : 26626u16 , name : "3CRGPC10075 OfficeConnect Wireless 54Mbps 11g PC Card" }] } , Device { vendor_id : 4523u16 , id : 8721u16 , name : "88SB2211 PCI Express to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10753u16 , name : "88W8335 [Libertas] 802.11b/g Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10754u16 , name : "88W8361 [TopDog] 802.11n Wireless" , subsystems : & [SubSystem { subvendor : 2001u16 , subdevice : 15106u16 , name : "DIR-615 rev. A1 Mini PCI Wireless Module" } , SubSystem { subvendor : 4997u16 , subdevice : 31744u16 , name : "WN511T RangeMax Next 300 Mbps Wireless PC Card" } , SubSystem { subvendor : 4997u16 , subdevice : 31745u16 , name : "WN511T RangeMax Next 300 Mbps Wireless Notebook Adapter" } , SubSystem { subvendor : 4997u16 , subdevice : 32256u16 , name : "WN311T RangeMax Next 300 Mbps Wireless PCI Adapter" } , SubSystem { subvendor : 5943u16 , subdevice : 101u16 , name : "WPC4400N Wireless-N Business Notebook Adapter" } , SubSystem { subvendor : 6041u16 , subdevice : 32795u16 , name : "F5D8011 v2 802.11n N1 Wireless Notebook Card" }] } , Device { vendor_id : 4523u16 , id : 10760u16 , name : "88W8362e [TopDog] 802.11a/b/g/n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10762u16 , name : "88W8363 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10763u16 , name : "88W8363 [TopDog] 802.11n Wireless" , subsystems : & [SubSystem { subvendor : 4436u16 , subdevice : 854u16 , name : "WLI-CB-AMG144N" } , SubSystem { subvendor : 4436u16 , subdevice : 855u16 , name : "WLI-CB-AG300N Nfiniti 802.11a/g/b + Draft 802.11n CardBus Card" } , SubSystem { subvendor : 4436u16 , subdevice : 860u16 , name : "WLI-CB-AMG300N" }] } , Device { vendor_id : 4523u16 , id : 10764u16 , name : "88W8363 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10788u16 , name : "88W8363 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10795u16 , name : "88W8687 [TopDog] 802.11b/g Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10800u16 , name : "88W8687 [TopDog] 802.11b/g Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10816u16 , name : "88W8366 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10817u16 , name : "88W8366 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10818u16 , name : "88W8366 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10819u16 , name : "88W8366 [TopDog] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 10837u16 , name : "88W8864 [Avastar] 802.11ac Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 11062u16 , name : "88W8764 [Avastar] 802.11n Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 11064u16 , name : "88W8897 [AVASTAR] 802.11ac Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 11072u16 , name : "88W8964 [Avastar] 802.11ac Wireless" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 16641u16 , name : "OLPC Cafe Controller Secure Digital Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 16710u16 , name : "GT-64111 System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17184u16 , name : "88E8001 Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 3896u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (ECS)" } , SubSystem { subvendor : 4121u16 , subdevice : 32769u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (ECS)" } , SubSystem { subvendor : 4163u16 , subdevice : 5948u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Asus)" } , SubSystem { subvendor : 4163u16 , subdevice : 33050u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Asus)" } , SubSystem { subvendor : 4187u16 , subdevice : 3097u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Foxconn)" } , SubSystem { subvendor : 4280u16 , subdevice : 46162u16 , name : "EZ Card 1000 (SMC9452TXV.2)" } , SubSystem { subvendor : 4523u16 , subdevice : 289u16 , name : "Marvell RDK-8001" } , SubSystem { subvendor : 4523u16 , subdevice : 801u16 , name : "Marvell RDK-8003" } , SubSystem { subvendor : 4523u16 , subdevice : 4129u16 , name : "Marvell RDK-8010" } , SubSystem { subvendor : 4523u16 , subdevice : 17184u16 , name : "Marvell Yukon Gigabit Ethernet 10/100/1000Baset-T Constroller (Asus)" } , SubSystem { subvendor : 4523u16 , subdevice : 20513u16 , name : "Marvell Yukon Gigabit Ethernet 10/100/1000Base-T Controller (64 bit)" } , SubSystem { subvendor : 4523u16 , subdevice : 38177u16 , name : "Marvell Yukon Gigabit Ethernet 10/100/1000Base-T Controller (32 bit)" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Gigabyte)" } , SubSystem { subvendor : 5243u16 , subdevice : 5126u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Abit)" } , SubSystem { subvendor : 5588u16 , subdevice : 71u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Iwill)" } , SubSystem { subvendor : 5781u16 , subdevice : 36901u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Epox)" } , SubSystem { subvendor : 6130u16 , subdevice : 7171u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Albatron)" } , SubSystem { subvendor : 9999u16 , subdevice : 10243u16 , name : "Marvell 88E8001 Gigabit Ethernet Controller (Chaintech)" }] } , Device { vendor_id : 4523u16 , id : 17216u16 , name : "88E8021 PCI-X IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17217u16 , name : "88E8022 PCI-X IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17218u16 , name : "88E8061 PCI-E IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17219u16 , name : "88E8062 PCI-E IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17220u16 , name : "88E8021 PCI-X IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17221u16 , name : "88E8022 PCI-X IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17222u16 , name : "88E8061 PCI-E IPMI Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17223u16 , name : "88E8062 PCI-E IPMI Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4304u16 , name : "Telum ASLP10 PrAMC Gigabit Ethernet" }] } , Device { vendor_id : 4523u16 , id : 17232u16 , name : "88E8035 PCI-E Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Marvell 88E8035 Fast Ethernet Controller (Toshiba)" } , SubSystem { subvendor : 4523u16 , subdevice : 13601u16 , name : "Marvell RDK-8035" } , SubSystem { subvendor : 6228u16 , subdevice : 13u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 14u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 15u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 17u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 18u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 22u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 23u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 24u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 25u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 28u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 30u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 32u16 , name : "Marvell 88E8035 Fast Ethernet Controller (LGE)" }] } , Device { vendor_id : 4523u16 , id : 17233u16 , name : "88E8036 PCI-E Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4219u16 , subdevice : 16393u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Wistron)" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Panasonic)" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Toshiba)" } , SubSystem { subvendor : 4473u16 , subdevice : 65280u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Compal)" } , SubSystem { subvendor : 4473u16 , subdevice : 65296u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Inventec)" } , SubSystem { subvendor : 4523u16 , subdevice : 13857u16 , name : "Marvell RDK-8036" } , SubSystem { subvendor : 5073u16 , subdevice : 44050u16 , name : "Abocom EFE3K - 10/100 Ethernet Expresscard" } , SubSystem { subvendor : 5663u16 , subdevice : 8253u16 , name : "Marvell 88E8036 Fast Ethernet Controller (Arima)" } , SubSystem { subvendor : 6228u16 , subdevice : 13u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 14u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 15u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 17u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 18u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 22u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 23u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 24u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 25u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 28u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 30u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 32u16 , name : "Marvell 88E8036 Fast Ethernet Controller (LGE)" }] } , Device { vendor_id : 4523u16 , id : 17234u16 , name : "88E8038 PCI-E Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17235u16 , name : "88E8039 PCI-E Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4173u16 , subdevice : 36909u16 , name : "VAIO VGN-NR120E" }] } , Device { vendor_id : 4523u16 , id : 17236u16 , name : "88E8040 PCI-E Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 5197u16 , subdevice : 49258u16 , name : "R730 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49266u16 , name : "Notebook N150P" }] } , Device { vendor_id : 4523u16 , id : 17237u16 , name : "88E8040T PCI-E Fast Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" }] } , Device { vendor_id : 4523u16 , id : 17238u16 , name : "88EC033 PCI-E Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17239u16 , name : "88E8042 PCI-E Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17242u16 , name : "88E8048 PCI-E Fast Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17248u16 , name : "88E8052 PCI-E ASF Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33076u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (Asus)" } , SubSystem { subvendor : 4219u16 , subdevice : 16393u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (Wistron)" } , SubSystem { subvendor : 4523u16 , subdevice : 21025u16 , name : "Marvell RDK-8052" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (Gigabyte)" } , SubSystem { subvendor : 5218u16 , subdevice : 1324u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (MSI)" } , SubSystem { subvendor : 6217u16 , subdevice : 32850u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (ASRock)" } , SubSystem { subvendor : 41120u16 , subdevice : 1289u16 , name : "Marvell 88E8052 Gigabit Ethernet Controller (Aopen)" }] } , Device { vendor_id : 4523u16 , id : 17249u16 , name : "88E8050 PCI-E ASF Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4219u16 , subdevice : 12309u16 , name : "Marvell 88E8050 Gigabit Ethernet Controller (Gateway)" } , SubSystem { subvendor : 4523u16 , subdevice : 20513u16 , name : "Marvell 88E8050 Gigabit Ethernet Controller (Intel)" } , SubSystem { subvendor : 32902u16 , subdevice : 12387u16 , name : "D925XCVLK mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13369u16 , name : "Marvell 88E8050 Gigabit Ethernet Controller (Intel)" }] } , Device { vendor_id : 4523u16 , id : 17250u16 , name : "88E8053 PCI-E Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10765u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Asus)" } , SubSystem { subvendor : 4163u16 , subdevice : 33090u16 , name : "Marvell 88E8053 Gigabit Ethernet controller PCIe (Asus)" } , SubSystem { subvendor : 4255u16 , subdevice : 12695u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Trigem)" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Panasonic)" } , SubSystem { subvendor : 4349u16 , subdevice : 42032u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (SOYO)" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Toshiba)" } , SubSystem { subvendor : 4473u16 , subdevice : 65280u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Compal)" } , SubSystem { subvendor : 4473u16 , subdevice : 65296u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Inventec)" } , SubSystem { subvendor : 4523u16 , subdevice : 21281u16 , name : "Marvell RDK-8053" } , SubSystem { subvendor : 4759u16 , subdevice : 49728u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)" } , SubSystem { subvendor : 4759u16 , subdevice : 49729u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)" } , SubSystem { subvendor : 4759u16 , subdevice : 49730u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)" } , SubSystem { subvendor : 4759u16 , subdevice : 49731u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)" } , SubSystem { subvendor : 4759u16 , subdevice : 49732u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)" } , SubSystem { subvendor : 5073u16 , subdevice : 44049u16 , name : "EGE5K - Giga Ethernet Expresscard" } , SubSystem { subvendor : 5208u16 , subdevice : 57344u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Gigabyte)" } , SubSystem { subvendor : 5218u16 , subdevice : 1420u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (MSI)" } , SubSystem { subvendor : 5312u16 , subdevice : 18u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Compal)" } , SubSystem { subvendor : 5464u16 , subdevice : 1184u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Clevo)" } , SubSystem { subvendor : 5565u16 , subdevice : 4099u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (DFI)" } , SubSystem { subvendor : 5663u16 , subdevice : 8252u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Arima)" } , SubSystem { subvendor : 5663u16 , subdevice : 8253u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Arima)" } , SubSystem { subvendor : 5781u16 , subdevice : 36905u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Epox)" } , SubSystem { subvendor : 6130u16 , subdevice : 11272u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Albatron)" } , SubSystem { subvendor : 6143u16 , subdevice : 1413u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Quanta)" } , SubSystem { subvendor : 6217u16 , subdevice : 32851u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (ASRock)" } , SubSystem { subvendor : 6228u16 , subdevice : 11u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 12u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 16u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 19u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 20u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 21u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 26u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 27u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 29u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 31u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 33u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 6228u16 , subdevice : 34u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (LGE)" } , SubSystem { subvendor : 9999u16 , subdevice : 10241u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Chaintech)" } , SubSystem { subvendor : 41120u16 , subdevice : 1286u16 , name : "Marvell 88E8053 Gigabit Ethernet Controller (Aopen)" }] } , Device { vendor_id : 4523u16 , id : 17251u16 , name : "88E8055 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17252u16 , name : "88E8056 PCI-E Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33272u16 , name : "Motherboard" } , SubSystem { subvendor : 4538u16 , subdevice : 186u16 , name : "8056 Gigabit Ethernet Controller" }] } , Device { vendor_id : 4523u16 , id : 17253u16 , name : "88E8070 based Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17254u16 , name : "88EC036 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17255u16 , name : "88EC032 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17256u16 , name : "88EC034 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17257u16 , name : "88EC042 Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17258u16 , name : "88E8058 PCI-E Gigabit Ethernet Controller" , subsystems : & [SubSystem { subvendor : 4523u16 , subdevice : 186u16 , name : "Imac 8,1 Wired Ethernet Adapter" }] } , Device { vendor_id : 4523u16 , id : 17259u16 , name : "88E8071 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17260u16 , name : "88E8072 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17261u16 , name : "88E8055 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17264u16 , name : "88E8075 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17280u16 , name : "88E8057 PCI-E Gigabit Ethernet Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17281u16 , name : "Yukon Optima 88E8059 [PCIe Gigabit Ethernet Controller with AVB]" , subsystems : & [SubSystem { subvendor : 4697u16 , subdevice : 10243u16 , name : "AT-2814FX" } , SubSystem { subvendor : 4697u16 , subdevice : 10244u16 , name : "AT-2874xx" }] } , Device { vendor_id : 4523u16 , id : 17937u16 , name : "GT-64115 System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 17952u16 , name : "GT-64120/64120A/64121A System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 18433u16 , name : "GT-48001" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20485u16 , name : "Belkin F5D5005 Gigabit Desktop Network PCI Card" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20544u16 , name : "MV88SX5040 4-port SATA I PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20545u16 , name : "MV88SX5041 4-port SATA I PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20608u16 , name : "MV88SX5080 8-port SATA I PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20609u16 , name : "MV88SX5081 8-port SATA I PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20865u16 , name : "88f5181 [Orion-1] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 20866u16 , name : "88f5182 [Orion-NAS] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 21121u16 , name : "88f5281 [Orion-2] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24641u16 , name : "MV88SX6041 4-port SATA II PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24642u16 , name : "88SX6042 PCI-X 4-Port SATA-II" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24705u16 , name : "MV88SX6081 8-port SATA II PCI-X Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24833u16 , name : "88SE6101/6102 single-port PATA133 interface" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33504u16 , name : "P5K PRO Motherboard" }] } , Device { vendor_id : 4523u16 , id : 24865u16 , name : "88SE6111/6121 SATA II / PATA Controller" , subsystems : & [SubSystem { subvendor : 4523u16 , subdevice : 24865u16 , name : "88SE6111/6121 1/2 port SATA II + 1 port PATA Controller" }] } , Device { vendor_id : 4523u16 , id : 24897u16 , name : "88SE614x SATA II PCI-E controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24901u16 , name : "88SE6145 SATA II PCI-E controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24960u16 , name : "88F6180 [Kirkwood] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 24978u16 , name : "88F6190/6192 [Kirkwood] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25217u16 , name : "88F6281 [Kirkwood] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25218u16 , name : "88F6282 [Armada 300] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25473u16 , name : "MV78xx0 [Discovery Innovation] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25664u16 , name : "88SE63x0 x1, 88SE6440 x4 PCIe SAS/SATA 3Gb/s RAID controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25680u16 , name : "64560 System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25696u16 , name : "MV64360/64361/64362 System Controller" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25728u16 , name : "MV64460/64461/64462 System Controller" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 49664u16 , name : "C2K CompactPCI single board computer" }] } , Device { vendor_id : 4523u16 , id : 25733u16 , name : "MV64460/64461/64462 System Controller, Revision B" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 25952u16 , name : "88F6560 [Avanta] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 26384u16 , name : "88F6710 [Armada 370] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 26656u16 , name : "88F6820 [Armada 385] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 26664u16 , name : "88F6828 [Armada 388] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 26912u16 , name : "88F6920 [Armada 390] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 28738u16 , name : "88SX7042 PCIe 4-port SATA-II controller" , subsystems : & [SubSystem { subvendor : 5816u16 , subdevice : 17227u16 , name : "Tempo SATA E4P" }] } , Device { vendor_id : 4523u16 , id : 30736u16 , name : "MV78100 [Discovery Innovation] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 30752u16 , name : "MV78200 [Discovery Innovation] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 30755u16 , name : "MV78230 [Armada XP] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 30790u16 , name : "MV78460 [Armada XP] ARM SoC" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 54287u16 , name : "Bobcat3 Ethernet Switch" , subsystems : & [] } , Device { vendor_id : 4523u16 , id : 61443u16 , name : "GT-64010 Primary Image Piranha Image Generator" , subsystems : & [] }] }),
        (7672, Vendor { id : 7672u16 , name : "V&G Information System Co.,Ltd" , devices : & [Device { vendor_id : 7672u16 , id : 12288u16 , name : "PC NVMe SSD" , subsystems : & [SubSystem { subvendor : 7672u16 , subdevice : 12544u16 , name : "M.2 NVMe Gen3*4 SSD" }] } , Device { vendor_id : 7672u16 , id : 16384u16 , name : "M.2 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7672u16 , subdevice : 16385u16 , name : "M.2 NVMe Gen4*4 SSD" } , SubSystem { subvendor : 7672u16 , subdevice : 16386u16 , name : "M.2 NVMe Gen4*4 SSD" }] } , Device { vendor_id : 7672u16 , id : 49152u16 , name : "DC NVMe SSD" , subsystems : & [SubSystem { subvendor : 7672u16 , subdevice : 50688u16 , name : "Enterprise U.2 NVMe SSD" }] } , Device { vendor_id : 7672u16 , id : 53248u16 , name : "PC NVMe SSD" , subsystems : & [SubSystem { subvendor : 7672u16 , subdevice : 53504u16 , name : "M.2 NVMe SSD" } , SubSystem { subvendor : 7672u16 , subdevice : 53761u16 , name : "M.2 NVMe SSD" } , SubSystem { subvendor : 7672u16 , subdevice : 54784u16 , name : "M.2 NVMe SSD" }] }] }),
        (4339, Vendor { id : 4339u16 , name : "Alaris, Inc." , devices : & [] }),
        (7208, Vendor { id : 7208u16 , name : "Lite-On IT Corp. / Plextor" , devices : & [Device { vendor_id : 7208u16 , id : 290u16 , name : "M6e PCI Express SSD [Marvell 88SS9183]" , subsystems : & [] }] }),
        (6940, Vendor { id : 6940u16 , name : "Corsair" , devices : & [] }),
        (57005, Vendor { id : 57005u16 , name : "Indigita Corporation" , devices : & [] }),
        (4689, Vendor { id : 4689u16 , name : "VLSI Solutions Oy" , devices : & [] }),
        (17224, Vendor { id : 17224u16 , name : "WCH.CN" , devices : & [Device { vendor_id : 17224u16 , id : 8819u16 , name : "CH351 PCI Dual Serial Port Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 12883u16 , name : "CH352 PCI Dual Serial Port Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 13395u16 , name : "CH353 PCI Quad Serial Port Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 20563u16 , name : "CH352 PCI Serial and Parallel Port Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 28755u16 , name : "CH353 PCI Dual Serial and Parallel Ports Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 28787u16 , name : "CH356 PCI Quad Serial and Parallel Ports Controller" , subsystems : & [] } , Device { vendor_id : 17224u16 , id : 29043u16 , name : "CH355 PCI Quad Serial Port Controller" , subsystems : & [] }] }),
        (5622, Vendor { id : 5622u16 , name : "Extreme Packet Device Inc" , devices : & [] }),
        (9032, Vendor { id : 9032u16 , name : "Racore" , devices : & [Device { vendor_id : 9032u16 , id : 8208u16 , name : "8142 100VG/AnyLAN" , subsystems : & [] }] }),
        (6379, Vendor { id : 6379u16 , name : "Advance Multimedia Internet Technology, Inc." , devices : & [] }),
        (4182, Vendor { id : 4182u16 , name : "ICL" , devices : & [] }),
        (4490, Vendor { id : 4490u16 , name : "Hilevel Technology" , devices : & [] }),
        (7905, Vendor { id : 7905u16 , name : "Suzhou Kuhan Information Technologies" , devices : & [Device { vendor_id : 7905u16 , id : 80u16 , name : "Aurora NVMe SSD Controller" , subsystems : & [SubSystem { subvendor : 7905u16 , subdevice : 9u16 , name : "Airglow A430 NVMe SSD U.2 1.6TB" } , SubSystem { subvendor : 7905u16 , subdevice : 10u16 , name : "Airglow A430 NVMe SSD U.2 3.2TB" } , SubSystem { subvendor : 7905u16 , subdevice : 11u16 , name : "Airglow A430 NVMe SSD U.2 4.8TB" } , SubSystem { subvendor : 7905u16 , subdevice : 18u16 , name : "Airglow Z400 NVMe ZNS SSD U.2 5.76TB" }] }] }),
        (5460, Vendor { id : 5460u16 , name : "PROLINK Microsystems Corp" , devices : & [] }),
        (8347, Vendor { id : 8347u16 , name : "BitIntelligence Technology" , devices : & [Device { vendor_id : 8347u16 , id : 4096u16 , name : "TCU Family - TCU-1" , subsystems : & [] } , Device { vendor_id : 8347u16 , id : 4097u16 , name : "TCU Family - TCU-1 Virtual Function" , subsystems : & [] } , Device { vendor_id : 8347u16 , id : 4098u16 , name : "TCU Family - TCU-1 PCIe Switch" , subsystems : & [] }] }),
        (1, Vendor { id : 1u16 , name : "SafeNet (wrong ID)" , devices : & [] }),
        (5073, Vendor { id : 5073u16 , name : "Abocom Systems Inc" , devices : & [Device { vendor_id : 5073u16 , id : 43778u16 , name : "ADMtek Centaur-C rev 17 [D-Link DFE-680TX] CardBus Fast Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 5073u16 , id : 43779u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] } , Device { vendor_id : 5073u16 , id : 43782u16 , name : "RTL8139 [FE2000VX] CardBus Fast Ethernet Attached Port Adapter" , subsystems : & [] } , Device { vendor_id : 5073u16 , id : 43784u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] }] }),
        (7570, Vendor { id : 7570u16 , name : "Abaco Systems Inc." , devices : & [] }),
        (394, Vendor { id : 394u16 , name : "LevelOne" , devices : & [Device { vendor_id : 394u16 , id : 262u16 , name : "FPC-0106TX misprogrammed [RTL81xx]" , subsystems : & [] }] }),
        (41732, Vendor { id : 41732u16 , name : "Sony" , devices : & [] }),
        (5592, Vendor { id : 5592u16 , name : "Cybernetics Technology Co Ltd" , devices : & [] }),
        (7165, Vendor { id : 7165u16 , name : "EeeTOP" , devices : & [] }),
        (3938, Vendor { id : 3938u16 , name : "Acrox Technologies Co., Ltd." , devices : & [] }),
        (4816, Vendor { id : 4816u16 , name : "GDE Systems, Inc." , devices : & [] }),
        (4098, Vendor { id : 4098u16 , name : "Advanced Micro Devices, Inc. [AMD/ATI]" , devices : & [Device { vendor_id : 4098u16 , id : 4372u16 , name : "Krackan [Radeon 840M / 860M Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4868u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4869u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4870u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4871u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4872u16 , name : "Kaveri HDMI/DP Audio Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" }] } , Device { vendor_id : 4098u16 , id : 4873u16 , name : "Kaveri [Radeon R6/R7 Graphics]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14384u16 , name : "Z50-75" }] } , Device { vendor_id : 4098u16 , id : 4874u16 , name : "Kaveri [Radeon R6 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4875u16 , name : "Kaveri [Radeon R4 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4876u16 , name : "Kaveri [Radeon R7 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4877u16 , name : "Kaveri [Radeon R6 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4878u16 , name : "Kaveri [Radeon R5 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4879u16 , name : "Kaveri [Radeon R7 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4880u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4881u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4882u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4883u16 , name : "Kaveri [Radeon R7 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4884u16 , name : "Wrestler HDMI Audio" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 4885u16 , name : "Kaveri [Radeon R5 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4886u16 , name : "Kaveri [Radeon R5 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4887u16 , name : "Kaveri" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4888u16 , name : "Kaveri [Radeon R5 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4891u16 , name : "Kaveri [Radeon R4 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4892u16 , name : "Kaveri [Radeon R7 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 4893u16 , name : "Kaveri [Radeon R6 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5056u16 , name : "Granite Ridge [Radeon Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5083u16 , name : "Cyan Skillfish [PlayStation 5 APU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5097u16 , name : "Ariel/Navi10Lite" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5098u16 , name : "Cyan Skillfish HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5113u16 , name : "Oberon/Navi12Lite" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5115u16 , name : "Cyan Skillfish [PlayStation 5 APU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5118u16 , name : "Cyan Skillfish [BC-250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5173u16 , name : "Sephiroth [AMD Custom GPU 0405]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5210u16 , name : "Dummy Function (absent graphics controller)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5240u16 , name : "Navi 10 XL Upstream Port of PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5241u16 , name : "Navi 10 XL Downstream Port of PCI Express Switch" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5382u16 , name : "Mendocino [Radeon 610M]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 14350u16 , name : "IdeaPad 1 15AMN7" }] } , Device { vendor_id : 4098u16 , id : 5390u16 , name : "Strix [Radeon 880M / 890M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5452u16 , name : "Kryptos [Radeon RX 350]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31784u16 , name : "MS-7C28 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 5454u16 , name : "Garfield" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5457u16 , name : "Arlene" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5458u16 , name : "Pooky" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5473u16 , name : "Anubis" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5510u16 , name : "Strix Halo [Radeon Graphics / Radeon 8050S Graphics / Radeon 8060S Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5555u16 , name : "Stoney HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5567u16 , name : "Phoenix1" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5576u16 , name : "Phoenix2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5592u16 , name : "Picasso/Raven 2 [Radeon Vega Series / Radeon Vega Mobile Series]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 6058u16 , subdevice : 12673u16 , name : "ThinkCentre M75n IoT" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 7618u16 , subdevice : 8713u16 , name : "Avita Pura 14 Notebook" } , SubSystem { subvendor : 59984u16 , subdevice : 52240u16 , name : "RXi2-BP" }] } , Device { vendor_id : 4098u16 , id : 5597u16 , name : "Raven Ridge [Radeon Vega Series / Radeon Vega Mobile Series]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33734u16 , name : "Radeon Vega 8 Mobile" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "Radeon RX Vega 11" } , SubSystem { subvendor : 59984u16 , subdevice : 52240u16 , name : "RXi2-BP" }] } , Device { vendor_id : 4098u16 , id : 5598u16 , name : "Raven/Raven2/Fenghuang HDMI/DP Audio Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME B450M-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52240u16 , name : "RXi2-BP" }] } , Device { vendor_id : 4098u16 , id : 5599u16 , name : "Raven/Raven2/Fenghuang/Renoir Cryptographic Coprocessor" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4098u16 , id : 5607u16 , name : "Barcelo" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 35607u16 , name : "ProBook 445 G9/455 G9 [Ryzen 7 Integrated Radeon GPU]" }] } , Device { vendor_id : 4098u16 , id : 5631u16 , name : "Fenghuang [Zhongshan Subor Z+]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5639u16 , name : "Arden" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5686u16 , name : "Renoir [Radeon Vega Series / Radeon Vega Mobile Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5687u16 , name : "Renoir/Cezanne HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5688u16 , name : "Cezanne [Radeon Vega Series / Radeon Vega Mobile Series]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 5826u16 , name : "Radeon Vega 8" }] } , Device { vendor_id : 4098u16 , id : 5695u16 , name : "VanGogh [AMD Custom GPU 0405]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5696u16 , name : "Radeon High Definition Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5708u16 , name : "Lucienne" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5709u16 , name : "Rembrandt" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5710u16 , name : "Raphael" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5711u16 , name : "Phoenix" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5761u16 , name : "Rembrandt [Radeon 680M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 5908u16 , name : "BeaverCreek HDMI Audio [Radeon HD 6500D and 6400G-6600G series]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5771u16 , name : "ProBook 4535s" }] } , Device { vendor_id : 4098u16 , id : 6400u16 , name : "HawkPoint1" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 6401u16 , name : "HawkPoint2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 6402u16 , name : "Krackan2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 12624u16 , name : "RV380/M24 [Mobility Radeon X600]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 2356u16 , name : "nx8220" }] } , Device { vendor_id : 4098u16 , id : 12625u16 , name : "RV380 GL [FireMV 2400]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 12626u16 , name : "RV370/M22 [Mobility Radeon X300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 12628u16 , name : "RV380/M24 GL [Mobility FireGL V3200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 12629u16 , name : "RV380 GL [FireMV 2400]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 12657u16 , name : "RV380 GL [FireMV 2400] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 15952u16 , name : "RV380 [Radeon X550/X600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 15956u16 , name : "RV380 GL [FireGL V3200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 15984u16 , name : "RV380 [Radeon X550/X600] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16694u16 , name : "RS100 [Mobility IGP 320M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16695u16 , name : "RS200 [Radeon IGP 340]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16708u16 , name : "R300 [Radeon 9500]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16710u16 , name : "R300 [Radeon 9700 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16711u16 , name : "R300 GL [FireGL Z1]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16712u16 , name : "R350 [Radeon 9800/9800 SE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16720u16 , name : "RV350 [Radeon 9550/9600/X1050 Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2u16 , name : "R9600 Pro primary (Asus OEM for HP)" } , SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "R9600 Pro secondary (Asus OEM for HP)" } , SubSystem { subvendor : 4098u16 , subdevice : 18210u16 , name : "All-in-Wonder 2006 AGP Edition" } , SubSystem { subvendor : 5208u16 , subdevice : 16420u16 , name : "GV-R96128D" } , SubSystem { subvendor : 5260u16 , subdevice : 8292u16 , name : "R96A-C3N" } , SubSystem { subvendor : 5260u16 , subdevice : 8294u16 , name : "R96A-C3N" } , SubSystem { subvendor : 5963u16 , subdevice : 31769u16 , name : "Atlantis Radeon 9600 Pro" } , SubSystem { subvendor : 5963u16 , subdevice : 31785u16 , name : "GC-R9600PRO" } , SubSystem { subvendor : 6126u16 , subdevice : 8194u16 , name : "Radeon 9600 256Mb Primary" } , SubSystem { subvendor : 6332u16 , subdevice : 257u16 , name : "GC-R9600PRO (Primary)" }] } , Device { vendor_id : 4098u16 , id : 16721u16 , name : "RV350 [Radeon 9600 Series]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 49156u16 , name : "A9600SE" } , SubSystem { subvendor : 5963u16 , subdevice : 31799u16 , name : "Radeon 9600SE 128M DDR V/D/VO" }] } , Device { vendor_id : 4098u16 , id : 16722u16 , name : "RV360 [Radeon 9600/X1050 Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2u16 , name : "Radeon 9600XT" } , SubSystem { subvendor : 4098u16 , subdevice : 18290u16 , name : "All-in-Wonder 9600 XT" } , SubSystem { subvendor : 4163u16 , subdevice : 49154u16 , name : "Radeon 9600 XT TVD" } , SubSystem { subvendor : 4163u16 , subdevice : 49178u16 , name : "A9600XT/TD" } , SubSystem { subvendor : 5218u16 , subdevice : 38160u16 , name : "RX9600XT (MS-8951)" } , SubSystem { subvendor : 5963u16 , subdevice : 31785u16 , name : "Radeon 9600XT" } , SubSystem { subvendor : 6023u16 , subdevice : 16386u16 , name : "Radeon 9600 XT" }] } , Device { vendor_id : 4098u16 , id : 16723u16 , name : "RV350 [Radeon 9550]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 268u16 , name : "A9550GE/TD" } , SubSystem { subvendor : 5218u16 , subdevice : 37676u16 , name : "RX9550SE-TD128 (MS-8932)" }] } , Device { vendor_id : 4098u16 , id : 16724u16 , name : "RV350 GL [FireGL T2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16725u16 , name : "RV350 [Radeon 9600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16727u16 , name : "RV350 GL [FireGL T2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16728u16 , name : "68800AX [Graphics Ultra Pro PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16740u16 , name : "R300 [Radeon 9500 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16741u16 , name : "R300 [Radeon 9700 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16742u16 , name : "R300 [Radeon 9700 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16744u16 , name : "RV350 [Radeon 9800 SE] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 16752u16 , name : "RV350 [Radeon 9550/9600/X1050 Series] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "R9600 Pro secondary (Asus OEM for HP)" } , SubSystem { subvendor : 4098u16 , subdevice : 18211u16 , name : "All-in-Wonder 2006 AGP Edition (Secondary)" } , SubSystem { subvendor : 5208u16 , subdevice : 16421u16 , name : "GV-R96128D (Secondary)" } , SubSystem { subvendor : 5260u16 , subdevice : 8295u16 , name : "R96A-C3N (Secondary)" } , SubSystem { subvendor : 5963u16 , subdevice : 31784u16 , name : "GC-R9600PRO (Secondary)" } , SubSystem { subvendor : 6126u16 , subdevice : 8195u16 , name : "Radeon 9600 256Mb (Secondary)" } , SubSystem { subvendor : 6332u16 , subdevice : 256u16 , name : "GC-R9600PRO (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 16753u16 , name : "RV350 [Radeon 9600] (Secondary)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 49157u16 , name : "A9600SE (Secondary)" } , SubSystem { subvendor : 5963u16 , subdevice : 31798u16 , name : "Radeon 9600SE 128M DDR V/D/VO (secondary)" }] } , Device { vendor_id : 4098u16 , id : 16754u16 , name : "RV350 [Radeon 9600/X1050 Series] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "Radeon 9600XT (Secondary)" } , SubSystem { subvendor : 4098u16 , subdevice : 18291u16 , name : "All-in-Wonder 9600 XT (Secondary)" } , SubSystem { subvendor : 4163u16 , subdevice : 49155u16 , name : "A9600XT (Secondary)" } , SubSystem { subvendor : 4163u16 , subdevice : 49179u16 , name : "A9600XT/TD (Secondary)" } , SubSystem { subvendor : 5963u16 , subdevice : 31784u16 , name : "Radeon 9600XT (Secondary)" } , SubSystem { subvendor : 6023u16 , subdevice : 16387u16 , name : "Radeon 9600 XT (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 16755u16 , name : "RV350 [Radeon 9550] (Secondary)" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 269u16 , name : "A9550GE/TD (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 16962u16 , name : "R200 [All-In-Wonder Radeon 8500 DV]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 682u16 , name : "Radeon 8500 AIW DV Edition" }] } , Device { vendor_id : 4098u16 , id : 16963u16 , name : "R200 PCI Bridge [All-in-Wonder Radeon 8500DV]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17206u16 , name : "RS100 [Radeon IGP 320M]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 17206u16 , name : "Pavilion ze4300 ATI Radeon Mobility U1 (IGP 320 M)" } , SubSystem { subvendor : 4156u16 , subdevice : 36u16 , name : "Pavilion ze4400 builtin Video" } , SubSystem { subvendor : 5663u16 , subdevice : 8233u16 , name : "eMachines M5312 builtin Video" }] } , Device { vendor_id : 4098u16 , id : 17207u16 , name : "RS200M [Radeon IGP 330M/340M/345M/350M]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1338u16 , name : "ThinkPad R40e" } , SubSystem { subvendor : 4156u16 , subdevice : 2128u16 , name : "Radeon IGP 345M" }] } , Device { vendor_id : 4098u16 , id : 17217u16 , name : "SB200 AC97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17218u16 , name : "SB200 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17221u16 , name : "SB200 EHCI USB Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17222u16 , name : "Crayola 6 [XENOS Parent Die (XBOX 360)]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17223u16 , name : "SB200 OHCI USB Controller #1" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17224u16 , name : "SB200 OHCI USB Controller #2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17225u16 , name : "SB200 IDE Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17228u16 , name : "SB200 PCI to LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17229u16 , name : "SB200 AC97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17235u16 , name : "SB200 SMBus Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17236u16 , name : "215CT [Mach64 CT PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17240u16 , name : "Mach64 CX [Graphics Xpression]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17249u16 , name : "SB300 AC'97 Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17250u16 , name : "SB300 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17251u16 , name : "SB300 SMBus Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17253u16 , name : "SB300 USB Controller (EHCI)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17255u16 , name : "SB300 USB Controller (EHCI)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17256u16 , name : "SB300 USB Controller (EHCI)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17257u16 , name : "SB300 IDE Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17260u16 , name : "SB300 PCI to LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17261u16 , name : "SB300 AC97 Modem Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17262u16 , name : "SB300 Serial ATA Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17264u16 , name : "IXP SB400 AC'97 Audio Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 121u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4133u16 , subdevice : 145u16 , name : "Aspire 5032WXMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10757u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 4187u16 , subdevice : 3201u16 , name : "Realtek ALC 653" } , SubSystem { subvendor : 4219u16 , subdevice : 768u16 , name : "MX6421" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" }] } , Device { vendor_id : 4098u16 , id : 17265u16 , name : "IXP SB4x0 PCI-PCI Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17266u16 , name : "IXP SB4x0 SMBus Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17267u16 , name : "IXP SB4x0 USB2 Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17268u16 , name : "IXP SB4x0 USB Host Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17269u16 , name : "IXP SB4x0 USB Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17270u16 , name : "IXP SB4x0 IDE Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17271u16 , name : "IXP SB4x0 PCI-ISA Bridge" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17272u16 , name : "IXP SB400 AC'97 Modem Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" }] } , Device { vendor_id : 4098u16 , id : 17273u16 , name : "IXP SB4x0 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 28993u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 17274u16 , name : "IXP SB400 Serial ATA Controller" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 17273u16 , name : "4379 Serial ATA Controller" } , SubSystem { subvendor : 4098u16 , subdevice : 17274u16 , name : "437A Serial ATA Controller" } , SubSystem { subvendor : 5218u16 , subdevice : 28993u16 , name : "Aspire L250" } , SubSystem { subvendor : 5361u16 , subdevice : 34816u16 , name : "Leadtek WinFast TV2000XP Expert" }] } , Device { vendor_id : 4098u16 , id : 17275u16 , name : "IXP SB4x0 High Definition Audio Controller" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 17275u16 , name : "IXP SB4x0 High Definition Audio Controller" } , SubSystem { subvendor : 4303u16 , subdevice : 4902u16 , name : "Fujitsu Lifebook A3040" } , SubSystem { subvendor : 5940u16 , subdevice : 4280u16 , name : "Realtek High Definition Audio" }] } , Device { vendor_id : 4098u16 , id : 17280u16 , name : "SB600 Non-Raid-5 SATA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10259u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5208u16 , subdevice : 45059u16 , name : "GA-MA790FX-DS5 (rev. 1.0)" } , SubSystem { subvendor : 5208u16 , subdevice : 45061u16 , name : "Gigabyte GA-MA69G-S3H Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29479u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 22937u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17281u16 , name : "SB600 SATA Controller (RAID 5 mode)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17282u16 , name : "SB600 AC97 Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17283u16 , name : "SBx00 Azalia (Intel HDA)" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion dm1z-3000" } , SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4163u16 , subdevice : 33328u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33644u16 , name : "M4A785TD Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33808u16 , name : "M4A89GTD PRO/USB3 Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33819u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4163u16 , subdevice : 33861u16 , name : "M5A78L LE" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 4473u16 , subdevice : 65310u16 , name : "Satellite C660D-113" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5208u16 , subdevice : 40994u16 , name : "GA-770/78-series motherboard" } , SubSystem { subvendor : 5208u16 , subdevice : 41218u16 , name : "GA-880GMA-USB3" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17284u16 , name : "SBx00 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17285u16 , name : "SBx00 SMBus Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33673u16 , name : "M4A785TD Motherboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5208u16 , subdevice : 17285u16 , name : "GA-770/78-series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17286u16 , name : "SB600 USB Controller (EHCI)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17287u16 , name : "SB600 USB (OHCI0)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17288u16 , name : "SB600 USB (OHCI1)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17289u16 , name : "SB600 USB (OHCI2)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17290u16 , name : "SB600 USB (OHCI3)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17291u16 , name : "SB600 USB (OHCI4)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17292u16 , name : "SB600 IDE" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5208u16 , subdevice : 20482u16 , name : "Gigabyte GA-MA69G-S3H Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17293u16 , name : "SB600 PCI to LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 5218u16 , subdevice : 29544u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17294u16 , name : "SB600 AC97 Modem" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17296u16 , name : "SB7x0/SB8x0/SB9x0 SATA Controller [IDE mode]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33673u16 , name : "M4A785TD Motherboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 45058u16 , name : "GA-MA770-DS3rev2.0 Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 17296u16 , name : "Motherboard (one of many)" }] } , Device { vendor_id : 4098u16 , id : 17297u16 , name : "SB7x0/SB8x0/SB9x0 SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33859u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4163u16 , subdevice : 34013u16 , name : "M5A99X EVO (R1.0) SB950" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 45058u16 , name : "GA-78/880-series motherboard" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 17298u16 , name : "SB7x0/SB8x0/SB9x0 SATA Controller [Non-RAID5 mode]" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" }] } , Device { vendor_id : 4098u16 , id : 17299u16 , name : "SB7x0/SB8x0/SB9x0 SATA Controller [RAID5 mode]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17300u16 , name : "SB7x0/SB8x0/SB9x0 SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17301u16 , name : "SB8x0/SB9x0 SATA Controller [Storage mode]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17302u16 , name : "SB7x0/SB8x0/SB9x0 USB EHCI Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33859u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-78/880-series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 17303u16 , name : "SB7x0/SB8x0/SB9x0 USB OHCI0 Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33859u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-78/880-series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 17304u16 , name : "SB7x0 USB OHCI1 Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-MA78GM-S2H motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" }] } , Device { vendor_id : 4098u16 , id : 17305u16 , name : "SB7x0/SB8x0/SB9x0 USB OHCI2 Controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33859u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 20484u16 , name : "GA-78/880-series motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 17308u16 , name : "SB7x0/SB8x0/SB9x0 IDE Controller" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 17298u16 , name : "MSI MS-7713 motherboard" } , SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5208u16 , subdevice : 20482u16 , name : "GA-MA78GM-S2H motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" }] } , Device { vendor_id : 4098u16 , id : 17309u16 , name : "SB7x0/SB8x0/SB9x0 LPC host controller" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4156u16 , subdevice : 5649u16 , name : "Pavilion DM1Z-3000" } , SubSystem { subvendor : 4163u16 , subdevice : 33519u16 , name : "M3A78-EH Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33859u16 , name : "M5A88-V EVO" } , SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard / AMD SB700" } , SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" } , SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 17312u16 , name : "SB700/SB800/SB900 PCI to PCI bridge (PCIE port 0)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17313u16 , name : "SB700/SB800/SB900 PCI to PCI bridge (PCIE port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17314u16 , name : "SB900 PCI to PCI bridge (PCIE port 2)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17315u16 , name : "SB900 PCI to PCI bridge (PCIE port 3)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17463u16 , name : "RS250 [Mobility Radeon 7000 IGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17748u16 , name : "210888ET [Mach64 ET]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17968u16 , name : "XENOS Parent Die (XBOX 360)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 17969u16 , name : "XENOS Daughter Die (XBOX 360)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18004u16 , name : "Mach64 VT" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18242u16 , name : "Rage 3 [3D Rage PRO AGP 2X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 64u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 68u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 97u16 , name : "Rage Pro AIW AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 98u16 , name : "Rage Pro AIW AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 99u16 , name : "Rage Pro AIW AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 128u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 132u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 18242u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 32769u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4136u16 , subdevice : 130u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4136u16 , subdevice : 16514u16 , name : "Optiplex GX1 Onboard Display Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 32898u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 4136u16 , subdevice : 49282u16 , name : "Rage Pro Turbo AGP 2X" } , SubSystem { subvendor : 32902u16 , subdevice : 16722u16 , name : "Xpert 98D AGP 2X" } , SubSystem { subvendor : 32902u16 , subdevice : 17994u16 , name : "Rage Pro Turbo AGP 2X" }] } , Device { vendor_id : 4098u16 , id : 18244u16 , name : "Rage 3 [3D Rage PRO AGP 1X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 18244u16 , name : "Rage Pro Turbo AGP" } , SubSystem { subvendor : 32902u16 , subdevice : 19797u16 , name : "Rage 3D Pro AGP 1X [Intel MU440EX]" }] } , Device { vendor_id : 4098u16 , id : 18249u16 , name : "3D Rage PRO PCI" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 97u16 , name : "Rage Pro AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 98u16 , name : "Rage Pro AIW" }] } , Device { vendor_id : 4098u16 , id : 18253u16 , name : "Rage XL AGP 2X" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 4u16 , name : "Xpert 98 RXL AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Xpert 98 RXL AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 128u16 , name : "Rage XL AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 132u16 , name : "Xpert 98 AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 18253u16 , name : "Rage XL AGP" } , SubSystem { subvendor : 4147u16 , subdevice : 32874u16 , name : "Rage XL AGP" }] } , Device { vendor_id : 4098u16 , id : 18254u16 , name : "Rage XC AGP" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 18254u16 , name : "Rage XC AGP" }] } , Device { vendor_id : 4098u16 , id : 18255u16 , name : "Rage XL" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Rage XL" } , SubSystem { subvendor : 4098u16 , subdevice : 18255u16 , name : "Rage XL" }] } , Device { vendor_id : 4098u16 , id : 18256u16 , name : "3D Rage Pro PCI" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 64u16 , name : "Rage Pro Turbo" } , SubSystem { subvendor : 4098u16 , subdevice : 68u16 , name : "Rage Pro Turbo" } , SubSystem { subvendor : 4098u16 , subdevice : 128u16 , name : "Rage Pro Turbo" } , SubSystem { subvendor : 4098u16 , subdevice : 132u16 , name : "Rage Pro Turbo" } , SubSystem { subvendor : 4098u16 , subdevice : 18256u16 , name : "Rage Pro Turbo" }] } , Device { vendor_id : 4098u16 , id : 18258u16 , name : "Rage 3 [Rage XL PCI]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 30u16 , name : "Proliant Rage XL" } , SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Rage XL" } , SubSystem { subvendor : 4098u16 , subdevice : 18258u16 , name : "Proliant Rage XL" } , SubSystem { subvendor : 4098u16 , subdevice : 32776u16 , name : "Rage XL" } , SubSystem { subvendor : 4116u16 , subdevice : 576u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 206u16 , name : "PowerEdge 1400" } , SubSystem { subvendor : 4136u16 , subdevice : 209u16 , name : "PowerEdge 2550" } , SubSystem { subvendor : 4136u16 , subdevice : 217u16 , name : "PowerEdge 2500" } , SubSystem { subvendor : 4136u16 , subdevice : 308u16 , name : "PowerEdge 600SC" } , SubSystem { subvendor : 4136u16 , subdevice : 330u16 , name : "PowerEdge 1750" } , SubSystem { subvendor : 4136u16 , subdevice : 357u16 , name : "PowerEdge 750" } , SubSystem { subvendor : 4156u16 , subdevice : 4321u16 , name : "NetServer Rage XL" } , SubSystem { subvendor : 4156u16 , subdevice : 12808u16 , name : "ProLiant DL140 G2" } , SubSystem { subvendor : 4219u16 , subdevice : 25600u16 , name : "6400 Server" } , SubSystem { subvendor : 5940u16 , subdevice : 122u16 , name : "PRIMERGY RX/TX series onboard VGA" } , SubSystem { subvendor : 5940u16 , subdevice : 4211u16 , name : "Primergy Econel 200 D2020 mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13329u16 , name : "SDS2 Mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13351u16 , name : "S875WP1-E mainboard" } , SubSystem { subvendor : 32902u16 , subdevice : 22340u16 , name : "S845WD1-E mainboard" }] } , Device { vendor_id : 4098u16 , id : 18259u16 , name : "Rage XC" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 18259u16 , name : "Rage XC" }] } , Device { vendor_id : 4098u16 , id : 18260u16 , name : "Mach64 GT/GT-B [3D Rage I/II]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18261u16 , name : "Mach64 GT-B [3D Rage II+ DVD]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18262u16 , name : "Rage 2 [3D Rage IIC PCI]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 18262u16 , name : "Rage IIC" }] } , Device { vendor_id : 4098u16 , id : 18263u16 , name : "Rage 2 [3D Rage IIC AGP]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 18263u16 , name : "Rage IIC AGP" } , SubSystem { subvendor : 4136u16 , subdevice : 137u16 , name : "Rage 3D IIC" } , SubSystem { subvendor : 4136u16 , subdevice : 142u16 , name : "PowerEdge 1300 onboard video" } , SubSystem { subvendor : 4136u16 , subdevice : 16514u16 , name : "Rage 3D IIC" } , SubSystem { subvendor : 4136u16 , subdevice : 32898u16 , name : "Rage 3D IIC" } , SubSystem { subvendor : 4136u16 , subdevice : 49282u16 , name : "Rage 3D IIC" }] } , Device { vendor_id : 4098u16 , id : 18264u16 , name : "Mach64 GX [WinTurbo]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18265u16 , name : "Rage 3 [3D Rage IIC PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18266u16 , name : "3D Rage IIC AGP" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 132u16 , name : "Rage 3D Pro AGP 2x XPERT 98" } , SubSystem { subvendor : 4098u16 , subdevice : 135u16 , name : "Rage 3D IIC" } , SubSystem { subvendor : 4098u16 , subdevice : 18266u16 , name : "Rage IIC AGP" }] } , Device { vendor_id : 4098u16 , id : 18501u16 , name : "Xilleon 220 HBIU for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18502u16 , name : "Xilleon 220 IDE for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18503u16 , name : "Xilleon 220 USB for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18504u16 , name : "Xilleon 220 DAIO-0 for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18505u16 , name : "Xilleon 220 DAIO-1 for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18506u16 , name : "Xilleon 220 LPC for HDTV2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18512u16 , name : "Xilleon 215 HBIU for X215" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18513u16 , name : "Xilleon 215 IDE for X215" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18514u16 , name : "Xilleon 215 USB for X215" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18515u16 , name : "Xilleon 215 DAIO-0 for X215" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18516u16 , name : "Xilleon 215 DAIO-1 for X215" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18517u16 , name : "Xilleon 225 HBIU for X225" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18518u16 , name : "Xilleon 225 IDE for X225" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18519u16 , name : "Xilleon 225 USB for X225" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18520u16 , name : "Xilleon 225 DAIO-0 for X225" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18521u16 , name : "Xilleon 225 DAIO-1 for X225" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18528u16 , name : "Xilleon 210 HBIU for X210" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18529u16 , name : "Xilleon 210 IDE for X210" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18530u16 , name : "Xilleon 210 USB for X210" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18531u16 , name : "Xilleon 210 DAIO-0 for X210" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18532u16 , name : "Xilleon 210 DAIO-1 for X210" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18533u16 , name : "Xilleon 226 HBIU for X226" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18534u16 , name : "Xilleon 226 IDE for X226" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18535u16 , name : "Xilleon 226 USB for X226" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18536u16 , name : "Xilleon 226 DAIO-0 for X226" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18537u16 , name : "Xilleon 226 DAIO-1 for X226" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18538u16 , name : "Xilleon 240S HBIU for X240S" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18539u16 , name : "Xilleon 240H HBIU for X240H" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18540u16 , name : "Xilleon 240S USB for X240S" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18541u16 , name : "Xilleon 240H USB for X240H" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18542u16 , name : "Xilleon 250 USB 1.1 for X250" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18543u16 , name : "Xilleon 260 USB 1.1 for X260" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18544u16 , name : "Xilleon 250 HBIU for X250" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18545u16 , name : "Xilleon 250 IDE for X250" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18546u16 , name : "Xilleon 234/235 HBIU for X234/X235" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18547u16 , name : "Xilleon 244/245 HBIU for X244/X245" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18548u16 , name : "Xilleon 234/235 USB 1.1 for X234/X235" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18549u16 , name : "Xilleon 260 HBIU for X260" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18550u16 , name : "Xilleon 260 IDE for X260" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18551u16 , name : "Xilleon 244/245 USB 1.1 for X244/X245" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18552u16 , name : "Xilleon 270 HBIU for X270" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18555u16 , name : "Xilleon 242 HBIU for X242" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18557u16 , name : "Xilleon 242 USB 1.1 for X242" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18560u16 , name : "Xilleon 254 HBIU for X254" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18561u16 , name : "Xilleon 254 USB 1.1 for X254" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18562u16 , name : "Xilleon 255 HBIU for X255" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18563u16 , name : "Xilleon 255 USB 1.1 for X255" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18564u16 , name : "Xilleon 243 HBIU for X243" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18565u16 , name : "Xilleon 243 USB 1.1 for X243" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18566u16 , name : "Xilleon 233 HBIU for X233" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18567u16 , name : "Xilleon 233 USB 1.1 for X233" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18568u16 , name : "Xilleon 143 HBIU for X143" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18569u16 , name : "Xilleon 143 HBIU for X143L" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18570u16 , name : "Xilleon 143 HBIU for X143S" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 18790u16 , name : "RV250 [Radeon 9000 Series]" , subsystems : & [SubSystem { subvendor : 4337u16 , subdevice : 2u16 , name : "RV250 If [Tachyon G9000 PRO]" } , SubSystem { subvendor : 5260u16 , subdevice : 8249u16 , name : "RV250 If [Radeon 9000 Pro \"Evil Commando\"]" } , SubSystem { subvendor : 5385u16 , subdevice : 39424u16 , name : "RV250 If [Radeon 9000 \"AT009\"]" } , SubSystem { subvendor : 5761u16 , subdevice : 64u16 , name : "RV250 If [3D prophet 9000]" } , SubSystem { subvendor : 5963u16 , subdevice : 29046u16 , name : "Radeon 9000 Pro" } , SubSystem { subvendor : 5963u16 , subdevice : 29074u16 , name : "RV250 If [Radeon 9000 \"Atlantis\"]" } , SubSystem { subvendor : 6063u16 , subdevice : 8197u16 , name : "RV250 If [Excalibur Radeon 9000 Pro]" } , SubSystem { subvendor : 6063u16 , subdevice : 8198u16 , name : "RV250 If [Excalibur Radeon 9000]" }] } , Device { vendor_id : 4098u16 , id : 18798u16 , name : "RV250 [Radeon 9000] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19017u16 , name : "R420 [Radeon X800 PRO/GTO AGP]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 9760u16 , name : "R420 [Radeon X800 GTO AGP]" }] } , Device { vendor_id : 4098u16 , id : 19018u16 , name : "R420 [Radeon X800 GT AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19019u16 , name : "R420 [Radeon X800 AGP Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19021u16 , name : "R420 GL [FireGL X3-256]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19022u16 , name : "RV420/M18 [Mobility Radeon 9800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19023u16 , name : "R420 [Radeon X850 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19024u16 , name : "R420 [Radeon X800 XT Platinum Edition AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19028u16 , name : "R420 [Radeon X800 VE AGP]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 17442u16 , name : "All-In-Wonder X800 VE AGP" }] } , Device { vendor_id : 4098u16 , id : 19049u16 , name : "R420 [Radeon X800 PRO/GTO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19050u16 , name : "R420 [Radeon X800] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19051u16 , name : "R420 [Radeon X800 XT AGP] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19056u16 , name : "R420 [Radeon X800 XT Platinum Edition AGP] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19060u16 , name : "R420 [Radeon X800 VE] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19273u16 , name : "R481 [Radeon X850 XT AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19275u16 , name : "R481 [Radeon X850 PRO AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19276u16 , name : "R481 [Radeon X850 XT Platinum Edition AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19305u16 , name : "R481 [Radeon X850 XT AGP] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19307u16 , name : "R481 [Radeon X850 PRO AGP] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19308u16 , name : "R481 [Radeon X850 XT Platinum Edition AGP] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19522u16 , name : "Mach64 LT [3D Rage LT PRO AGP]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45287u16 , name : "Rage LT Pro (Compaq Presario 5240)" } , SubSystem { subvendor : 3601u16 , subdevice : 45288u16 , name : "Rage 3D LT Pro" } , SubSystem { subvendor : 3601u16 , subdevice : 45326u16 , name : "3D Rage LT Pro (Compaq Armada 1750)" } , SubSystem { subvendor : 4098u16 , subdevice : 64u16 , name : "Rage LT Pro AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 68u16 , name : "Rage LT Pro AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 19522u16 , name : "Rage LT Pro AGP 2X" } , SubSystem { subvendor : 4098u16 , subdevice : 32769u16 , name : "Rage LT Pro AGP 2X" } , SubSystem { subvendor : 4136u16 , subdevice : 133u16 , name : "Rage 3D LT Pro" }] } , Device { vendor_id : 4098u16 , id : 19526u16 , name : "Rage Mobility 128 AGP 2X/Mobility M3" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 341u16 , name : "IBM Thinkpad A22p" } , SubSystem { subvendor : 4116u16 , subdevice : 341u16 , name : "Thinkpad A22p" } , SubSystem { subvendor : 4136u16 , subdevice : 177u16 , name : "Latitude C600" }] } , Device { vendor_id : 4098u16 , id : 19527u16 , name : "3D Rage IIC PCI / Mobility Radeon 7500/7500C" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19529u16 , name : "3D Rage LT PRO PCI" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 4u16 , name : "Rage LT Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 64u16 , name : "Rage LT Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 68u16 , name : "Rage LT Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 19529u16 , name : "Rage LT Pro" }] } , Device { vendor_id : 4098u16 , id : 19533u16 , name : "Rage Mobility AGP 2x Series" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45329u16 , name : "Armada M700" } , SubSystem { subvendor : 3601u16 , subdevice : 45408u16 , name : "Armada E500" } , SubSystem { subvendor : 4098u16 , subdevice : 132u16 , name : "Xpert 98 AGP 2X (Mobility)" } , SubSystem { subvendor : 4116u16 , subdevice : 340u16 , name : "ThinkPad A20m/A21m" } , SubSystem { subvendor : 4136u16 , subdevice : 170u16 , name : "Latitude CPt" } , SubSystem { subvendor : 4136u16 , subdevice : 187u16 , name : "Latitude CPx" } , SubSystem { subvendor : 4473u16 , subdevice : 65280u16 , name : "Satellite 1715XCDS laptop" } , SubSystem { subvendor : 5053u16 , subdevice : 4121u16 , name : "PC-AR10" }] } , Device { vendor_id : 4098u16 , id : 19534u16 , name : "Rage Mobility L AGP 2x" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19536u16 , name : "Rage 3 LT [3D Rage LT PRO PCI]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 19536u16 , name : "Rage LT Pro" }] } , Device { vendor_id : 4098u16 , id : 19538u16 , name : "M1 [Rage Mobility-M1 PCI]" , subsystems : & [SubSystem { subvendor : 4147u16 , subdevice : 33042u16 , name : "Versa Note VXi" }] } , Device { vendor_id : 4098u16 , id : 19540u16 , name : "264LT [Mach64 LT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19543u16 , name : "RV200/M7 [Mobility Radeon 7500]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 183u16 , name : "Evo N610c" } , SubSystem { subvendor : 4116u16 , subdevice : 1303u16 , name : "ThinkPad T30" } , SubSystem { subvendor : 4116u16 , subdevice : 1328u16 , name : "ThinkPad T4x Series" } , SubSystem { subvendor : 4136u16 , subdevice : 230u16 , name : "Radeon Mobility M7 LW (Dell Inspiron 8100)" } , SubSystem { subvendor : 4136u16 , subdevice : 298u16 , name : "Latitude C640" } , SubSystem { subvendor : 4163u16 , subdevice : 5666u16 , name : "Mobility Radeon M7 (L3C/S)" } , SubSystem { subvendor : 5197u16 , subdevice : 49158u16 , name : "Radeon Mobility M7 LW in vpr Matrix 170B4" }] } , Device { vendor_id : 4098u16 , id : 19544u16 , name : "RV200/M7 GL [Mobility FireGL 7800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19545u16 , name : "RV100/M6 [Rage/Radeon Mobility Series]" , subsystems : & [SubSystem { subvendor : 3601u16 , subdevice : 45329u16 , name : "Evo N600c" } , SubSystem { subvendor : 4116u16 , subdevice : 565u16 , name : "ThinkPad A30/A30p (2652/2653)" } , SubSystem { subvendor : 4116u16 , subdevice : 569u16 , name : "ThinkPad X22/X23/X24" } , SubSystem { subvendor : 4156u16 , subdevice : 37u16 , name : "XE4500 Notebook" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 4173u16 , subdevice : 33088u16 , name : "PCG-Z1SP laptop" } , SubSystem { subvendor : 5385u16 , subdevice : 6448u16 , name : "Medion MD9703" }] } , Device { vendor_id : 4098u16 , id : 19558u16 , name : "RV250/M9 GL [Mobility FireGL 9000/Radeon 9000]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1357u16 , name : "ThinkPad T41" }] } , Device { vendor_id : 4098u16 , id : 19566u16 , name : "RV250/M9 [Mobility Radeon 9000] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19782u16 , name : "Rage Mobility 128 AGP 4X/Mobility M4" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19794u16 , name : "Theater 550 PRO PCI [ATI TV Wonder 550]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 19795u16 , name : "Theater 550 PRO PCIe" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20036u16 , name : "R300 [Radeon 9700/9700 PRO]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 20830u16 , name : "Radeon ES1000" } , SubSystem { subvendor : 4098u16 , subdevice : 22885u16 , name : "Radeon ES1000" }] } , Device { vendor_id : 4098u16 , id : 20037u16 , name : "R300 [Radeon 9500 PRO/9700]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2u16 , name : "Radeon R300 NE [Radeon 9500 Pro]" } , SubSystem { subvendor : 5761u16 , subdevice : 2u16 , name : "Hercules 3D Prophet 9500 PRO [Radeon 9500 Pro]" }] } , Device { vendor_id : 4098u16 , id : 20038u16 , name : "R300 [Radeon 9600 TX]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20039u16 , name : "R300 GL [FireGL X1]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20040u16 , name : "R350 [Radeon 9800 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20041u16 , name : "R350 [Radeon 9800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20042u16 , name : "R360 [Radeon 9800 XXL/XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 20042u16 , name : "R360 [Radeon 9800 XT]" }] } , Device { vendor_id : 4098u16 , id : 20043u16 , name : "R350 GL [FireGL X2 AGP Pro]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20048u16 , name : "RV350/M10 / RV360/M11 [Mobility Radeon 9600 (PRO) / 9700]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 90u16 , name : "TravelMate 290" } , SubSystem { subvendor : 4133u16 , subdevice : 100u16 , name : "Extensa 3000 series laptop: ATI RV360/M11 [Mobility Radeon 9700]" } , SubSystem { subvendor : 4156u16 , subdevice : 2188u16 , name : "NC8000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 2192u16 , name : "NC6000 laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P35 notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 785u16 , name : "MSI M510A" } , SubSystem { subvendor : 5940u16 , subdevice : 4181u16 , name : "Amilo M1420W" }] } , Device { vendor_id : 4098u16 , id : 20049u16 , name : "RV350 [Radeon 9550/9600/X1050 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20050u16 , name : "RV350/M10 [Mobility Radeon 9500/9700 SE]" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P35 notebook" }] } , Device { vendor_id : 4098u16 , id : 20052u16 , name : "RV350/M10 GL [Mobility FireGL T2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20054u16 , name : "RV360/M12 [Mobility Radeon 9550]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20068u16 , name : "R300 [Radeon 9700 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20069u16 , name : "R300 [Radeon 9500 PRO] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3u16 , name : "Radeon R300 NE [Radeon 9500 Pro]" } , SubSystem { subvendor : 5761u16 , subdevice : 3u16 , name : "Hercules 3D Prophet 9500 PRO [Radeon 9500 Pro] (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 20070u16 , name : "RV350 [Radeon 9600] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20071u16 , name : "R300 GL [FireGL X1] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20072u16 , name : "R350 [Radeon 9800 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20073u16 , name : "R350 [Radeon 9800] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20074u16 , name : "RV350 [Radeon 9800 XT] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 20074u16 , name : "R360 [Radeon 9800 XT] (Secondary)" } , SubSystem { subvendor : 4098u16 , subdevice : 20081u16 , name : "M10 NQ [Radeon Mobility 9600]" }] } , Device { vendor_id : 4098u16 , id : 20081u16 , name : "RV350/M10 [Mobility Radeon 9600] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20338u16 , name : "RV250 [Radeon 9000 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20339u16 , name : "RV250 [Radeon 9000 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20548u16 , name : "All-In-Wonder 128 PCI" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Rage 128 AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 41u16 , name : "Rage 128 AIW" }] } , Device { vendor_id : 4098u16 , id : 20550u16 , name : "Rage 4 [Rage 128 PRO AGP 4X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 4u16 , name : "Rage Fury Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Rage Fury Pro/Xpert 2000 Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 20u16 , name : "Rage Fury Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 24u16 , name : "Rage Fury Pro/Xpert 2000 Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Rage 128 Pro AIW AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 42u16 , name : "Rage 128 Pro AIW AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 72u16 , name : "Rage Fury Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 8192u16 , name : "Rage Fury MAXX AGP 4x (TMDS) (VGA device)" } , SubSystem { subvendor : 4098u16 , subdevice : 8193u16 , name : "Rage Fury MAXX AGP 4x (TMDS) (Extra device?!)" }] } , Device { vendor_id : 4098u16 , id : 20560u16 , name : "Rage 4 [Rage 128 PRO PCI / Xpert 128 PCI]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Xpert 128" }] } , Device { vendor_id : 4098u16 , id : 20562u16 , name : "Rage 4 [Rage 128 PRO AGP 4X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20804u16 , name : "R100 [Radeon 7200 / All-In-Wonder Radeon]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 9u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 10u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 26u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 41u16 , name : "Radeon AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 56u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 57u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 138u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 186u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 313u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 650u16 , name : "Radeon 7000/Radeon" } , SubSystem { subvendor : 4098u16 , subdevice : 682u16 , name : "Radeon AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 1338u16 , name : "Radeon 7000/Radeon" }] } , Device { vendor_id : 4098u16 , id : 20808u16 , name : "R200 GL [FireGL 8800]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 266u16 , name : "FireGL 8800 64Mb" } , SubSystem { subvendor : 4098u16 , subdevice : 338u16 , name : "FireGL 8800 128Mb" } , SubSystem { subvendor : 4098u16 , subdevice : 354u16 , name : "FireGL 8700 32Mb" } , SubSystem { subvendor : 4098u16 , subdevice : 370u16 , name : "FireGL 8700 64Mb" }] } , Device { vendor_id : 4098u16 , id : 20812u16 , name : "R200 [Radeon 8500/8500 LE]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 58u16 , name : "Radeon R200 QL [Radeon 8500 LE]" } , SubSystem { subvendor : 4098u16 , subdevice : 314u16 , name : "Radeon 8500" } , SubSystem { subvendor : 5260u16 , subdevice : 8230u16 , name : "R200 QL [Radeon 8500 Evil Master II Multi Display Edition]" } , SubSystem { subvendor : 5761u16 , subdevice : 16u16 , name : "Radeon 8500 [3D Prophet 8500 128Mb]" } , SubSystem { subvendor : 5963u16 , subdevice : 29001u16 , name : "Radeon 8500 LE" } , SubSystem { subvendor : 6023u16 , subdevice : 3848u16 , name : "Radeon R200 QL [PowerMagic Radeon 8500]" }] } , Device { vendor_id : 4098u16 , id : 20813u16 , name : "R200 [Radeon 9100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 20823u16 , name : "RV200 [Radeon 7500/7500 LE]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 314u16 , name : "Radeon 7500" } , SubSystem { subvendor : 4098u16 , subdevice : 3883u16 , name : "ALL-IN-WONDER VE PCI" } , SubSystem { subvendor : 4098u16 , subdevice : 4154u16 , name : "Dell Optiplex GX260" } , SubSystem { subvendor : 5208u16 , subdevice : 16384u16 , name : "RV200 QW [RADEON 7500 PRO MAYA AR]" } , SubSystem { subvendor : 5260u16 , subdevice : 8228u16 , name : "RV200 QW [Radeon 7500LE Dual Display]" } , SubSystem { subvendor : 5260u16 , subdevice : 8229u16 , name : "RV200 QW [Radeon 7500 Evil Master Multi Display Edition]" } , SubSystem { subvendor : 5260u16 , subdevice : 8246u16 , name : "RV200 QW [Radeon 7500 PCI Dual Display]" } , SubSystem { subvendor : 5963u16 , subdevice : 28998u16 , name : "RV200 QW [Radeon 7500 LE]" } , SubSystem { subvendor : 5963u16 , subdevice : 28999u16 , name : "Radeon 7500 LE" } , SubSystem { subvendor : 5963u16 , subdevice : 29025u16 , name : "Radeon RV200 QW [Radeon 7500 LE]" } , SubSystem { subvendor : 6063u16 , subdevice : 514u16 , name : "RV200 QW [Excalibur Radeon 7500LE]" }] } , Device { vendor_id : 4098u16 , id : 20825u16 , name : "RV100 [Radeon 7000 / Radeon VE]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 10u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 11u16 , name : "Radeon 7000" } , SubSystem { subvendor : 4098u16 , subdevice : 56u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 58u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 186u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 314u16 , name : "Radeon 7000/Radeon VE" } , SubSystem { subvendor : 4098u16 , subdevice : 2312u16 , name : "XVR-100 (supplied by Sun)" } , SubSystem { subvendor : 4116u16 , subdevice : 666u16 , name : "Remote Supervisor Adapter II (RSA2)" } , SubSystem { subvendor : 4116u16 , subdevice : 712u16 , name : "eServer xSeries server mainboard" } , SubSystem { subvendor : 4136u16 , subdevice : 364u16 , name : "PowerEdge 1850 Embedded Radeon 7000/VE" } , SubSystem { subvendor : 4136u16 , subdevice : 365u16 , name : "PowerEdge 2850 Embedded Radeon 7000-M" } , SubSystem { subvendor : 4136u16 , subdevice : 368u16 , name : "PowerEdge 6850 Embedded Radeon 7000/VE" } , SubSystem { subvendor : 4136u16 , subdevice : 410u16 , name : "PowerEdge SC1425" } , SubSystem { subvendor : 4156u16 , subdevice : 4754u16 , name : "Radeon 7000" } , SubSystem { subvendor : 4163u16 , subdevice : 49162u16 , name : "A7000/T/64M" } , SubSystem { subvendor : 5208u16 , subdevice : 16386u16 , name : "RV100 QY [RADEON 7000 PRO MAYA AV Series]" } , SubSystem { subvendor : 5260u16 , subdevice : 8195u16 , name : "RV100 QY [Radeon 7000 Multi-Display Edition]" } , SubSystem { subvendor : 5260u16 , subdevice : 8227u16 , name : "RV100 QY [Radeon 7000 Evil Master Multi-Display]" } , SubSystem { subvendor : 5260u16 , subdevice : 8321u16 , name : "RV6DE" } , SubSystem { subvendor : 5963u16 , subdevice : 640u16 , name : "Radeon RV100 QY [Radeon 7000/VE]" } , SubSystem { subvendor : 5963u16 , subdevice : 28946u16 , name : "Radeon VE 7000" } , SubSystem { subvendor : 5963u16 , subdevice : 31784u16 , name : "Radeon VE 7000 DDR" } , SubSystem { subvendor : 6023u16 , subdevice : 514u16 , name : "RV100 QY [Excalibur Radeon 7000]" } , SubSystem { subvendor : 6126u16 , subdevice : 4097u16 , name : "Radeon 7000 64MB DDR + DVI" }] } , Device { vendor_id : 4098u16 , id : 20830u16 , name : "ES1000" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 443u16 , name : "PowerEdge 1955 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 479u16 , name : "PowerEdge SC440" } , SubSystem { subvendor : 4136u16 , subdevice : 486u16 , name : "PowerEdge 860" } , SubSystem { subvendor : 4136u16 , subdevice : 496u16 , name : "PowerEdge R900 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "PowerEdge T605 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 527u16 , name : "PowerEdge R300 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 528u16 , name : "PowerEdge T300 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 545u16 , name : "PowerEdge R805 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 547u16 , name : "PowerEdge R905 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 549u16 , name : "PowerEdge T105 Embedded ATI ES1000" } , SubSystem { subvendor : 4136u16 , subdevice : 572u16 , name : "PowerEdge R200 Embedded ATI ES1000" } , SubSystem { subvendor : 4156u16 , subdevice : 4868u16 , name : "Integrity iLO2 Advanced KVM VGA [AD307A]" } , SubSystem { subvendor : 4156u16 , subdevice : 12795u16 , name : "ProLiant DL360 G5" } , SubSystem { subvendor : 5593u16 , subdevice : 34432u16 , name : "X7DVL-E-O motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 38528u16 , name : "X7DBN Motherboard" } , SubSystem { subvendor : 32902u16 , subdevice : 13430u16 , name : "S5000PSLSATA Server Board" }] } , Device { vendor_id : 4098u16 , id : 21061u16 , name : "Rage 128 GL PCI" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Xpert 128" } , SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Rage 128 AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 41u16 , name : "Rage 128 AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 104u16 , name : "Rage 128 AIW" }] } , Device { vendor_id : 4098u16 , id : 21062u16 , name : "Rage 128 (Rage 4) series" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 4u16 , name : "Magnum/Xpert 128/Xpert 99" } , SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Rage 128 AGP 2x" } , SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Rage 128 AIW AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 68u16 , name : "Rage Fury/Xpert 128/Xpert 2000" } , SubSystem { subvendor : 4098u16 , subdevice : 104u16 , name : "Rage 128 AIW AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 1096u16 , name : "Rage Fury" }] } , Device { vendor_id : 4098u16 , id : 21067u16 , name : "Rage 128 VR PCI" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21068u16 , name : "Rage 128 VR AGP" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Xpert 99/Xpert 2000" } , SubSystem { subvendor : 4098u16 , subdevice : 136u16 , name : "Xpert 99" }] } , Device { vendor_id : 4098u16 , id : 21325u16 , name : "Rage 128 4X AGP 4x" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Xpert 99/Xpert 2000" } , SubSystem { subvendor : 4098u16 , subdevice : 24u16 , name : "Xpert 2000" }] } , Device { vendor_id : 4098u16 , id : 21332u16 , name : "Mach 64 VT" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 22100u16 , name : "Mach 64 reference" }] } , Device { vendor_id : 4098u16 , id : 21574u16 , name : "Rage 128 PRO Ultra AGP 4x" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 4u16 , name : "Rage Fury Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 8u16 , name : "Rage Fury Pro/Xpert 2000 Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 24u16 , name : "Rage Fury Pro/Xpert 2000 Pro" } , SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Rage 128 AIW Pro AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 41u16 , name : "Rage 128 AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 42u16 , name : "Rage 128 AIW Pro AGP" } , SubSystem { subvendor : 4098u16 , subdevice : 43u16 , name : "Rage 128 AIW" } , SubSystem { subvendor : 4098u16 , subdevice : 72u16 , name : "Xpert 2000 Pro" }] } , Device { vendor_id : 4098u16 , id : 21586u16 , name : "Rage 128 PRO Ultra4XL VR-R AGP" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 28u16 , name : "Rage 128 Pro 4XL" } , SubSystem { subvendor : 4156u16 , subdevice : 4729u16 , name : "Rage 128 Pro 4XL" }] } , Device { vendor_id : 4098u16 , id : 21600u16 , name : "RV370/M22 [Mobility Radeon X300]" , subsystems : & [SubSystem { subvendor : 6005u16 , subdevice : 4352u16 , name : "CR11/VR11 Single Board Computer" }] } , Device { vendor_id : 4098u16 , id : 21601u16 , name : "RV370/M22 [Mobility Radeon X300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21602u16 , name : "RV380/M24C [Mobility Radeon X600 SE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21604u16 , name : "RV370/M22 GL [Mobility FireGL V3100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21833u16 , name : "R423 [Radeon X800 GTO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21834u16 , name : "R423 [Radeon X800 XT Platinum Edition]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21835u16 , name : "R423 [Radeon X800 GT/SE]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 770u16 , name : "Radeon X800 SE" }] } , Device { vendor_id : 4098u16 , id : 21837u16 , name : "R480 [Radeon X800 GTO2/XL]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 802u16 , name : "All-In-Wonder X800 XL" } , SubSystem { subvendor : 5208u16 , subdevice : 8484u16 , name : "GV-R80L256V-B (AGP)" }] } , Device { vendor_id : 4098u16 , id : 21838u16 , name : "R430 [All-In-Wonder X800 GT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21839u16 , name : "R430 [Radeon X800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21840u16 , name : "R423 GL [FireGL V7100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21841u16 , name : "R423 GL [FireGL V5100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21865u16 , name : "R423 [Radeon X800 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21867u16 , name : "R423 [Radeon X800 GT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21869u16 , name : "R480 [Radeon X800 GTO2/XL] (Secondary)" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 8485u16 , name : "GV-R80L256V-B (AGP)" }] } , Device { vendor_id : 4098u16 , id : 21871u16 , name : "R430 [Radeon X800] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 21873u16 , name : "R423 GL [FireGL V5100] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22091u16 , name : "RV410/M26 GL [Mobility FireGL V5000]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22095u16 , name : "RV410/M26 [Mobility Radeon X700 XL]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22098u16 , name : "RV410/M26 [Mobility Radeon X700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22099u16 , name : "RV410/M26 [Mobility Radeon X700]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMi" } , SubSystem { subvendor : 4156u16 , subdevice : 2368u16 , name : "Compaq NW8240 Mobile Workstation" }] } , Device { vendor_id : 4098u16 , id : 22100u16 , name : "Mach64 VT [Video Xpression]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 22100u16 , name : "Mach64VT Reference" }] } , Device { vendor_id : 4098u16 , id : 22101u16 , name : "264VT3 [Mach64 VT3]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22102u16 , name : "Mach64 VT4 [Video Xpression+]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22103u16 , name : "RV410 [Radeon X550 XTX / X700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22576u16 , name : "RS300 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22577u16 , name : "RS300 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22578u16 , name : "RS300 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22579u16 , name : "RS300 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22580u16 , name : "RS300 [Radeon 9100 IGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22581u16 , name : "RS300M [Mobility Radeon 9100 IGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22584u16 , name : "RS300 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22612u16 , name : "RS480 [Radeon Xpress 200 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22644u16 , name : "RS480 [Radeon Xpress 1150] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22848u16 , name : "RV280 [Radeon 9200 PRO] (Secondary)" , subsystems : & [SubSystem { subvendor : 6063u16 , subdevice : 8225u16 , name : "Excalibur Radeon 9250 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 22849u16 , name : "RV280 [Radeon 9200] (Secondary)" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 16409u16 , name : "Radeon 9200" } , SubSystem { subvendor : 5963u16 , subdevice : 31762u16 , name : "Radeon 9200" } , SubSystem { subvendor : 6063u16 , subdevice : 8205u16 , name : "Excalibur Radeon 9200" } , SubSystem { subvendor : 6332u16 , subdevice : 80u16 , name : "GC-R9200-C3 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 22852u16 , name : "RV280 [Radeon 9200 SE PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22864u16 , name : "RS480/RS482/RS485 Host Bridge" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 128u16 , name : "Aspire 5024WLMMi" } , SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 4156u16 , subdevice : 10784u16 , name : "Pavilion t3030.de Desktop PC" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" } , SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 22865u16 , name : "RX480/RX482 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22866u16 , name : "RD580 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22868u16 , name : "RS480 [Radeon Xpress 200 Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 22868u16 , name : "RV370 [Radeon Xpress 200G Series]" }] } , Device { vendor_id : 4098u16 , id : 22869u16 , name : "RS480M [Mobility Radeon Xpress 200]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 22869u16 , name : "RS480 0x5955 [Radeon XPRESS 200M 5955 (PCIE)]" } , SubSystem { subvendor : 4156u16 , subdevice : 12427u16 , name : "MX6125" } , SubSystem { subvendor : 5218u16 , subdevice : 305u16 , name : "MS-1013 Notebook" }] } , Device { vendor_id : 4098u16 , id : 22870u16 , name : "RD790 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22871u16 , name : "RX780/RX790 Host Bridge" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 22871u16 , name : "A770CrossFire Motherboard" }] } , Device { vendor_id : 4098u16 , id : 22872u16 , name : "RD780 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22880u16 , name : "RV280 [Radeon 9200 PRO / 9250]" , subsystems : & [SubSystem { subvendor : 6063u16 , subdevice : 8224u16 , name : "Excalibur Radeon 9250" }] } , Device { vendor_id : 4098u16 , id : 22881u16 , name : "RV280 [Radeon 9200]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 12146u16 , name : "All-in-Wonder 9200 Series" } , SubSystem { subvendor : 4121u16 , subdevice : 19504u16 , name : "Radeon 9200 VIVO" } , SubSystem { subvendor : 4779u16 , subdevice : 22881u16 , name : "YUAN SMARTVGA Radeon 9200" } , SubSystem { subvendor : 5208u16 , subdevice : 16408u16 , name : "Radeon 9200" } , SubSystem { subvendor : 5963u16 , subdevice : 31763u16 , name : "Radeon 9200" } , SubSystem { subvendor : 6063u16 , subdevice : 8204u16 , name : "Excalibur Radeon 9200" } , SubSystem { subvendor : 6332u16 , subdevice : 80u16 , name : "Radeon 9200 Game Buster" } , SubSystem { subvendor : 6332u16 , subdevice : 81u16 , name : "GC-R9200-C3" } , SubSystem { subvendor : 6332u16 , subdevice : 83u16 , name : "Radeon 9200 Game Buster VIVO" }] } , Device { vendor_id : 4098u16 , id : 22882u16 , name : "RV280 [Radeon 9200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22884u16 , name : "RV280 [Radeon 9200 SE]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 22884u16 , name : "Radeon 9200 SE, 64-bit 128MB DDR, 200/166MHz" } , SubSystem { subvendor : 4163u16 , subdevice : 49158u16 , name : "Radeon 9200 SE / TD / 128M" } , SubSystem { subvendor : 5208u16 , subdevice : 16408u16 , name : "R92S128T [Radeon 9200 SE 128MB]" } , SubSystem { subvendor : 5208u16 , subdevice : 16434u16 , name : "Radeon 9200 SE 128MB" } , SubSystem { subvendor : 5243u16 , subdevice : 24977u16 , name : "R9200SE-DT" } , SubSystem { subvendor : 5260u16 , subdevice : 8307u16 , name : "CN-AG92E" } , SubSystem { subvendor : 5963u16 , subdevice : 31763u16 , name : "Radeon 9200 SE" } , SubSystem { subvendor : 6023u16 , subdevice : 22884u16 , name : "Excalibur 9200SE VIVO 128M" } , SubSystem { subvendor : 6063u16 , subdevice : 8210u16 , name : "Radeon 9200 SE Excalibur" } , SubSystem { subvendor : 6332u16 , subdevice : 368u16 , name : "Sapphire Radeon 9200 SE 128MB Game Buster" } , SubSystem { subvendor : 6332u16 , subdevice : 371u16 , name : "GC-R9200L(SE)-C3H [Radeon 9200 Game Buster]" }] } , Device { vendor_id : 4098u16 , id : 22885u16 , name : "RV280 GL [FireMV 2200 PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22900u16 , name : "RS482/RS485 [Radeon Xpress 1100/1150]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 10250u16 , name : "DC5750 Microtower" } , SubSystem { subvendor : 5218u16 , subdevice : 28993u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 22901u16 , name : "RS482M [Mobility Radeon Xpress 200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22904u16 , name : "RX780/RD790 PCI to PCI bridge (external gfx0 port A)" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 22871u16 , name : "A770CrossFire Motherboard" }] } , Device { vendor_id : 4098u16 , id : 22905u16 , name : "RD790 PCI to PCI bridge (external gfx0 port B)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22906u16 , name : "RD790 PCI to PCI bridge (PCI express gpp port A)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22907u16 , name : "RX780/RD790 PCI to PCI bridge (PCI express gpp port B)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22908u16 , name : "RD790 PCI to PCI bridge (PCI express gpp port C)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22909u16 , name : "RX780/RD790 PCI to PCI bridge (PCI express gpp port D)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22910u16 , name : "RD790 PCI to PCI bridge (PCI express gpp port E)" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 22871u16 , name : "A770CrossFire Motherboard" }] } , Device { vendor_id : 4098u16 , id : 22911u16 , name : "RD790 PCI to PCI bridge (PCI express gpp port F)" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 22871u16 , name : "A770CrossFire Motherboard" }] } , Device { vendor_id : 4098u16 , id : 22912u16 , name : "RD790 PCI to PCI bridge (external gfx1 port A)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22913u16 , name : "RD790 PCI to PCI bridge (external gfx1 port B)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 22914u16 , name : "RD790 PCI to PCI bridge (NB-SB link)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23056u16 , name : "RD890 Northbridge only dual slot (2x16) PCI-e GFX Hydra part" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23057u16 , name : "RD890 Northbridge only single slot PCI-e GFX Hydra part" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23058u16 , name : "RD890 Northbridge only dual slot (2x8) PCI-e GFX Hydra part" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" }] } , Device { vendor_id : 4098u16 , id : 23059u16 , name : "RD890S/SR5650 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23060u16 , name : "RD9x0/RX980 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23061u16 , name : "RD890 PCI to PCI bridge (PCI express gpp port A)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23062u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GFX port 0)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23063u16 , name : "RD890/RD9x0 PCI to PCI bridge (PCI Express GFX port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23064u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 0)" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" }] } , Device { vendor_id : 4098u16 , id : 23065u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23066u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 2)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23067u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 3)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23068u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 4)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23069u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP Port 5)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23070u16 , name : "RD890/RD9x0/RX980 PCI to PCI bridge (PCI Express GPP2 Port 0)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23071u16 , name : "RD890/RD990 PCI to PCI bridge (PCI Express GFX2 port 0)" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 43025u16 , name : "H8DGU" }] } , Device { vendor_id : 4098u16 , id : 23072u16 , name : "RD890/RD990 PCI to PCI bridge (PCI Express GFX2 port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23075u16 , name : "RD890S/RD990 I/O Memory Management Unit (IOMMU)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23089u16 , name : "RC410 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23091u16 , name : "RS400 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23092u16 , name : "RS4xx PCI Express Port [ext gfx]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23094u16 , name : "RC4xx/RS4xx PCI Express Port 1" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23095u16 , name : "RC4xx/RS4xx PCI Express Port 2" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23096u16 , name : "RC4xx/RS4xx PCI Express Port 3" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23097u16 , name : "RC4xx/RS4xx PCI Express Port 4" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23103u16 , name : "RC4xx/RS4xx PCI Bridge [int gfx]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29207u16 , name : "Aspire L250" }] } , Device { vendor_id : 4098u16 , id : 23105u16 , name : "RS400 [Radeon Xpress 200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23106u16 , name : "RS400M [Radeon Xpress 200M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23137u16 , name : "RC410 [Radeon Xpress 200/1100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23138u16 , name : "RC410M [Mobility Radeon Xpress 200M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23392u16 , name : "RV370 [Radeon X300]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 42u16 , name : "Extreme AX300SE-X" } , SubSystem { subvendor : 4163u16 , subdevice : 814u16 , name : "Extreme AX300/TD" } , SubSystem { subvendor : 5208u16 , subdevice : 8450u16 , name : "GV-RX30S128D (X300SE)" } , SubSystem { subvendor : 5218u16 , subdevice : 1024u16 , name : "RX300SE-TD128E (MS-8940 REV:200)" } , SubSystem { subvendor : 5218u16 , subdevice : 1026u16 , name : "RX300SE-TD128E (MS-8940)" } , SubSystem { subvendor : 5963u16 , subdevice : 1280u16 , name : "Radeon X300 (PCIE)" } , SubSystem { subvendor : 6509u16 , subdevice : 4230u16 , name : "X300SE HM" }] } , Device { vendor_id : 4098u16 , id : 23394u16 , name : "RV370 [Radeon X600/X600 SE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23395u16 , name : "RV370 [Radeon X300/X550/X1050 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23396u16 , name : "RV370 GL [FireGL V3100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23397u16 , name : "RV370 GL [FireMV 2200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23398u16 , name : "RV370X" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23408u16 , name : "RV370 [Radeon X300 SE]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 1027u16 , name : "Radeon X300 SE 128MB DDR" } , SubSystem { subvendor : 5963u16 , subdevice : 1281u16 , name : "Radeon X300 SE" } , SubSystem { subvendor : 6509u16 , subdevice : 4231u16 , name : "Radeon X300 SE HyperMemory" }] } , Device { vendor_id : 4098u16 , id : 23410u16 , name : "RV380 [Radeon X300/X550/X1050 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23411u16 , name : "RV370 [Radeon X300/X550/X1050 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23412u16 , name : "RV370 GL [FireGL V3100] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23413u16 , name : "RV370 GL [FireMV 2200] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23649u16 , name : "RV280/M9+ [Mobility Radeon 9200 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23651u16 , name : "RV280/M9+ [Mobility Radeon 9200 AGP]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 23651u16 , name : "Apple iBook G4 2004" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30 notebook" }] } , Device { vendor_id : 4098u16 , id : 23876u16 , name : "RV280 [Radeon 9200 SE] (Secondary)" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 16409u16 , name : "R92S128T (Radeon 9200 SE 128MB Secondary)" } , SubSystem { subvendor : 5208u16 , subdevice : 16434u16 , name : "Radeon 9200 SE 128MB" } , SubSystem { subvendor : 5243u16 , subdevice : 24976u16 , name : "R9200SE-DT (Secondary)" } , SubSystem { subvendor : 5963u16 , subdevice : 31762u16 , name : "Radeon 9200 SE (Secondary)" } , SubSystem { subvendor : 6023u16 , subdevice : 22885u16 , name : "Excalibur 9200SE VIVO 128M (Secondary)" } , SubSystem { subvendor : 6063u16 , subdevice : 8211u16 , name : "Radeon 9200 SE Excalibur (Secondary)" } , SubSystem { subvendor : 6332u16 , subdevice : 369u16 , name : "Radeon 9200 SE 128MB Game Buster (Secondary)" } , SubSystem { subvendor : 6332u16 , subdevice : 370u16 , name : "GC-R9200L(SE)-C3H [Radeon 9200 Game Buster]" }] } , Device { vendor_id : 4098u16 , id : 23877u16 , name : "RV280 GL [FireMV 2200 PCI] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23880u16 , name : "R423/M28 [Mobility Radeon X800 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23881u16 , name : "R423/M28 GL [Mobility FireGL V5100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23882u16 , name : "R423/M28 [Mobility Radeon X800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23885u16 , name : "R480 [Radeon X850 XT Platinum Edition]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23886u16 , name : "R480 [Radeon X850 SE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23887u16 , name : "R480 [Radeon X800 GTO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23888u16 , name : "R480 GL [FireGL V7200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23890u16 , name : "R480 [Radeon X850 XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2834u16 , name : "PowerColor X850XT PCIe (Primary)" }] } , Device { vendor_id : 4098u16 , id : 23895u16 , name : "R423 [Radeon X800 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23917u16 , name : "R480 [Radeon X850 XT Platinum Edition] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23919u16 , name : "R480 [Radeon X800 GTO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 23922u16 , name : "R480 [Radeon X850 XT] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2835u16 , name : "PowerColor X850XT PCIe (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 23927u16 , name : "R423 [Radeon X800 XT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24136u16 , name : "RV410 GL [FireGL V5000]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24137u16 , name : "RV410 [Radeon X700 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24138u16 , name : "RV410 [Radeon X700 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24139u16 , name : "RV410 [Radeon X700 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24140u16 , name : "RV410 [Radeon X700 SE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24141u16 , name : "RV410 [Radeon X700]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 8470u16 , name : "Bravo X700" }] } , Device { vendor_id : 4098u16 , id : 24143u16 , name : "RV410 [Radeon X700]" , subsystems : & [SubSystem { subvendor : 5481u16 , subdevice : 7759u16 , name : "Radeon X550 XT" }] } , Device { vendor_id : 4098u16 , id : 24171u16 , name : "RV410 [Radeon X700 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 24173u16 , name : "RV410 [Radeon X700] (Secondary)" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 8471u16 , name : "Bravo X700 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 24407u16 , name : "R423 [Radeon X800 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26112u16 , name : "Mars [Radeon HD 8670A/8670M/8750M / R7 M370]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6482u16 , name : "ProBook 455 G1" }] } , Device { vendor_id : 4098u16 , id : 26113u16 , name : "Mars [Radeon HD 8730M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 8448u16 , name : "FirePro M4100" }] } , Device { vendor_id : 4098u16 , id : 26116u16 , name : "Opal XT [Radeon R7 M265/M365X/M465]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 1910u16 , name : "Aspire V5 Radeon R7 M265" } , SubSystem { subvendor : 4156u16 , subdevice : 32774u16 , name : "FirePro M4170" } , SubSystem { subvendor : 4156u16 , subdevice : 33103u16 , name : "Litho XT [Radeon R7 M365X]" } , SubSystem { subvendor : 4156u16 , subdevice : 33450u16 , name : "Litho XT [Radeon R7 M465]" } , SubSystem { subvendor : 6058u16 , subdevice : 13891u16 , name : "Radeon R7 A360" }] } , Device { vendor_id : 4098u16 , id : 26117u16 , name : "Opal PRO [Radeon R7 M260X]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 8793u16 , name : "FirePro M4150" }] } , Device { vendor_id : 4098u16 , id : 26118u16 , name : "Mars XTX [Radeon HD 8790M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1668u16 , name : "FirePro W4170M" }] } , Device { vendor_id : 4098u16 , id : 26119u16 , name : "Mars LE [Radeon HD 8530M / R5 M240]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26120u16 , name : "Oland GL [FirePro W2100]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15656u16 , name : "MXRT-2600" }] } , Device { vendor_id : 4098u16 , id : 26121u16 , name : "Oland GL [FirePro W2100 / Barco MXRT 2600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26128u16 , name : "Oland XT [Radeon HD 8670 / R5 340X OEM / R7 250/350/350X OEM]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 48u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 4136u16 , subdevice : 129u16 , name : "Radeon R7 350X OEM" } , SubSystem { subvendor : 4136u16 , subdevice : 131u16 , name : "Radeon R5 340X OEM" } , SubSystem { subvendor : 4136u16 , subdevice : 8480u16 , name : "Radeon R7 250" } , SubSystem { subvendor : 4136u16 , subdevice : 8994u16 , name : "Radeon R7 250" } , SubSystem { subvendor : 5218u16 , subdevice : 10512u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 5218u16 , subdevice : 10513u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 5260u16 , subdevice : 29520u16 , name : "Radeon R7 350" } , SubSystem { subvendor : 5698u16 , subdevice : 15489u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 5698u16 , subdevice : 15505u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 5698u16 , subdevice : 16137u16 , name : "Radeon R7 350" }] } , Device { vendor_id : 4098u16 , id : 26129u16 , name : "Oland [Radeon HD 8570 / R5 430 OEM / R7 240/340/430 / Radeon 520 OEM]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 4097u16 , name : "Radeon R5 430 OEM (1024 MByte)" } , SubSystem { subvendor : 4136u16 , subdevice : 4098u16 , name : "Radeon R5 430 OEM (2048 MByte)" } , SubSystem { subvendor : 4136u16 , subdevice : 5905u16 , name : "R5 430 OEM (2048 MByte)" } , SubSystem { subvendor : 4136u16 , subdevice : 8459u16 , name : "Radeon R5 240 OEM" } , SubSystem { subvendor : 4136u16 , subdevice : 8481u16 , name : "Radeon HD 8570 OEM" } , SubSystem { subvendor : 4156u16 , subdevice : 13173u16 , name : "Radeon R7 430 OEM" } , SubSystem { subvendor : 4303u16 , subdevice : 6281u16 , name : "Radeon HD 8570 OEM" } , SubSystem { subvendor : 5698u16 , subdevice : 6249u16 , name : "Radeon 520 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 16968u16 , name : "Radeon R7 240 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 41536u16 , name : "Radeon R7 240 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 54080u16 , name : "Radeon R7 340 OEM" } , SubSystem { subvendor : 6922u16 , subdevice : 37075u16 , name : "Radeon R7 240 OEM" }] } , Device { vendor_id : 4098u16 , id : 26131u16 , name : "Oland PRO [Radeon R7 240/340 / Radeon 520]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 29504u16 , name : "Radeon R7 340" } , SubSystem { subvendor : 5762u16 , subdevice : 29248u16 , name : "R7 240 2048 MB" } , SubSystem { subvendor : 7631u16 , subdevice : 12288u16 , name : "Oland PRO [Radeon R7 240/340 / Radeon 520]" }] } , Device { vendor_id : 4098u16 , id : 26135u16 , name : "Oland LE [Radeon R7 240]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26161u16 , name : "Oland" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26176u16 , name : "Saturn XT [FirePro M6100]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 331u16 , name : "Tropo XT [Radeon R9 M380 Mac Edition]" }] } , Device { vendor_id : 4098u16 , id : 26177u16 , name : "Saturn PRO [Radeon HD 8930M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26182u16 , name : "Bonaire XT [Radeon R9 M280X / FirePro W6150M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26183u16 , name : "Saturn PRO/XT [Radeon R9 M270X/M280X]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 8765u16 , name : "N551ZU laptop Radeon R9 M280X" }] } , Device { vendor_id : 4098u16 , id : 26185u16 , name : "Bonaire [FirePro W5100]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2828u16 , name : "FirePro W4300" } , SubSystem { subvendor : 4156u16 , subdevice : 2828u16 , name : "Bonaire [FirePro W4300]" } , SubSystem { subvendor : 4156u16 , subdevice : 8972u16 , name : "FirePro W5100" } , SubSystem { subvendor : 5068u16 , subdevice : 15658u16 , name : "MXRT-5600" }] } , Device { vendor_id : 4098u16 , id : 26189u16 , name : "Bonaire [FirePro W5100 / Barco MXRT-5600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26192u16 , name : "Bonaire" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26193u16 , name : "Bonaire" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26200u16 , name : "Bonaire XTX [Radeon R7 260X/360]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1167u16 , name : "R7260X-DC2OC-2GD5" } , SubSystem { subvendor : 4163u16 , subdevice : 1235u16 , name : "AMD Radeon R7 260X" } , SubSystem { subvendor : 5208u16 , subdevice : 8827u16 , name : "Radeon R7 260X" } , SubSystem { subvendor : 5260u16 , subdevice : 2311u16 , name : "Radeon R7 360" } , SubSystem { subvendor : 5762u16 , subdevice : 2311u16 , name : "Radeon R7 360" } , SubSystem { subvendor : 5762u16 , subdevice : 29536u16 , name : "Radeon R7 360" }] } , Device { vendor_id : 4098u16 , id : 26204u16 , name : "Bonaire XT [Radeon HD 7790/8770 / R7 360 / R9 260/360 OEM]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1106u16 , name : "Radeon HD 7790 DirectCU II OC" } , SubSystem { subvendor : 5218u16 , subdevice : 10544u16 , name : "Radeon HD 7790 OC" } , SubSystem { subvendor : 5218u16 , subdevice : 10546u16 , name : "Radeon HD 8770" } , SubSystem { subvendor : 5218u16 , subdevice : 10548u16 , name : "Radeon R9 260 OEM" } , SubSystem { subvendor : 5218u16 , subdevice : 10552u16 , name : "Radeon R9 360 OEM" } , SubSystem { subvendor : 5260u16 , subdevice : 2311u16 , name : "Radeon R7 360" } , SubSystem { subvendor : 5260u16 , subdevice : 37472u16 , name : "Radeon R9 260 OEM" } , SubSystem { subvendor : 5260u16 , subdevice : 37728u16 , name : "Radeon R9 360 OEM" } , SubSystem { subvendor : 5762u16 , subdevice : 2311u16 , name : "Radeon R7 360" } , SubSystem { subvendor : 5762u16 , subdevice : 13072u16 , name : "Radeon HD 7790 Black Edition 2 GB" } , SubSystem { subvendor : 5963u16 , subdevice : 57939u16 , name : "Radeon HD 7790 Dual-X OC" } , SubSystem { subvendor : 6023u16 , subdevice : 9001u16 , name : "Radeon HD 7790 TurboDuo" }] } , Device { vendor_id : 4098u16 , id : 26205u16 , name : "Bonaire [Radeon R7 200 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26207u16 , name : "Tobago PRO [Radeon R7 360 / R9 360 OEM]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2820u16 , name : "Radeon R9 360 OEM" } , SubSystem { subvendor : 5218u16 , subdevice : 10552u16 , name : "Radeon R9 360 OEM" } , SubSystem { subvendor : 5218u16 , subdevice : 12913u16 , name : "Radeon R9 360 OEM" } , SubSystem { subvendor : 5762u16 , subdevice : 29536u16 , name : "Radeon R7 360" }] } , Device { vendor_id : 4098u16 , id : 26208u16 , name : "Sun XT [Radeon HD 8670A/8670M/8690M / R5 M330 / M430 / Radeon 520 Mobile]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1514u16 , name : "Radeon HD 8670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1727u16 , name : "Radeon R5 M335" } , SubSystem { subvendor : 4156u16 , subdevice : 6512u16 , name : "Radeon HD 8670M" } , SubSystem { subvendor : 4156u16 , subdevice : 32958u16 , name : "Radeon R5 M330" } , SubSystem { subvendor : 4156u16 , subdevice : 33078u16 , name : "Radeon R5 M330" } , SubSystem { subvendor : 4156u16 , subdevice : 33577u16 , name : "Radeon R7 M520" } , SubSystem { subvendor : 6058u16 , subdevice : 13875u16 , name : "Radeon R5 A330" } , SubSystem { subvendor : 6058u16 , subdevice : 14340u16 , name : "Radeon R5 M330" } , SubSystem { subvendor : 6058u16 , subdevice : 14345u16 , name : "Radeon R5 M330" } , SubSystem { subvendor : 6058u16 , subdevice : 14362u16 , name : "Radeon R5 M430" } , SubSystem { subvendor : 6058u16 , subdevice : 14604u16 , name : "Radeon R5 M330" }] } , Device { vendor_id : 4098u16 , id : 26211u16 , name : "Sun PRO [Radeon HD 8570A/8570M]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 2118u16 , name : "Radeon HD 8570A" } , SubSystem { subvendor : 6058u16 , subdevice : 14341u16 , name : "Radeon HD 8570M" }] } , Device { vendor_id : 4098u16 , id : 26212u16 , name : "Jet XT [Radeon R5 M240]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26213u16 , name : "Jet PRO [Radeon R5 M230 / R7 M260DX / Radeon 520/610 Mobile]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 4873u16 , name : "Z50-75 Radeon R7 M260DX" } , SubSystem { subvendor : 6058u16 , subdevice : 13967u16 , name : "Radeon R5 A230" }] } , Device { vendor_id : 4098u16 , id : 26215u16 , name : "Jet ULT [Radeon R5 M230]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26223u16 , name : "Sun LE [Radeon HD 8550M / R5 M230]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26272u16 , name : "Vega 20 [Radeon Pro/Radeon Instinct]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26273u16 , name : "Vega 20 [Radeon Pro VII/Radeon Instinct MI50]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26274u16 , name : "Vega 20" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26275u16 , name : "Vega 20 [Radeon Pro Vega II/Radeon Pro Vega II Duo]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26279u16 , name : "Vega 20 [Radeon Pro Vega 20]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26287u16 , name : "Vega 20 [Radeon VII]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26372u16 , name : "Cayman PRO GL [FirePro V7900]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26375u16 , name : "Cayman LE GL [FirePro V5900]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26392u16 , name : "Cayman XT [Radeon HD 6970]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26393u16 , name : "Cayman PRO [Radeon HD 6950]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26396u16 , name : "Antilles [Radeon HD 6990]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26397u16 , name : "Antilles [Radeon HD 6990]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26399u16 , name : "Cayman CE [Radeon HD 6930]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26400u16 , name : "Blackcomb [Radeon HD 6970M/6990M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1167u16 , name : "Radeon HD 6990M" } , SubSystem { subvendor : 4136u16 , subdevice : 1168u16 , name : "Alienware M17x R3 Radeon HD 6970M" } , SubSystem { subvendor : 4136u16 , subdevice : 1188u16 , name : "FirePro M8900" } , SubSystem { subvendor : 4136u16 , subdevice : 1210u16 , name : "Radeon HD 6990M" } , SubSystem { subvendor : 4136u16 , subdevice : 1343u16 , name : "FirePro M8900" } , SubSystem { subvendor : 4203u16 , subdevice : 2816u16 , name : "Radeon HD 6970M" } , SubSystem { subvendor : 5464u16 , subdevice : 20738u16 , name : "Radeon HD 6970M" } , SubSystem { subvendor : 5464u16 , subdevice : 20740u16 , name : "Radeon HD 6990M" } , SubSystem { subvendor : 5464u16 , subdevice : 29185u16 , name : "Radeon HD 6990M" } , SubSystem { subvendor : 5963u16 , subdevice : 57736u16 , name : "Radeon HD 6970M" }] } , Device { vendor_id : 4098u16 , id : 26424u16 , name : "Barts XT [Radeon HD 6870]" , subsystems : & [SubSystem { subvendor : 5762u16 , subdevice : 12547u16 , name : "Radeon HD 8670" } , SubSystem { subvendor : 6023u16 , subdevice : 8218u16 , name : "Barts XT [Radeon HD 6870 X2]" } , SubSystem { subvendor : 6023u16 , subdevice : 8219u16 , name : "Barts XT [Radeon HD 6870 X2]" }] } , Device { vendor_id : 4098u16 , id : 26425u16 , name : "Barts PRO [Radeon HD 6850]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 948u16 , name : "EAH6850 [Radeon HD 6850]" }] } , Device { vendor_id : 4098u16 , id : 26430u16 , name : "Barts LE [Radeon HD 6790]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 30496u16 , name : "Radeon HD 7720 OEM" }] } , Device { vendor_id : 4098u16 , id : 26432u16 , name : "Whistler [Radeon HD 6730M/6770M/7690M XT]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 9100u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4121u16 , subdevice : 9102u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4121u16 , subdevice : 9105u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4121u16 , subdevice : 9106u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" } , SubSystem { subvendor : 4136u16 , subdevice : 1342u16 , name : "FirePro M5950" } , SubSystem { subvendor : 4156u16 , subdevice : 5680u16 , name : "FirePro M5950" } , SubSystem { subvendor : 4156u16 , subdevice : 5681u16 , name : "FirePro M5950" } , SubSystem { subvendor : 4156u16 , subdevice : 5707u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4156u16 , subdevice : 5710u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4156u16 , subdevice : 5719u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5720u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5722u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5723u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5768u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5769u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 5770u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 6238u16 , name : "Radeon HD 7690M XT" } , SubSystem { subvendor : 4156u16 , subdevice : 13192u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 13193u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 13698u16 , name : "Radeon HD 6770M" } , SubSystem { subvendor : 4156u16 , subdevice : 13932u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4163u16 , subdevice : 7426u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4163u16 , subdevice : 7442u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4173u16 , subdevice : 36996u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 4173u16 , subdevice : 36997u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 5197u16 , subdevice : 45172u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 5197u16 , subdevice : 45175u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 5197u16 , subdevice : 45188u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 5197u16 , subdevice : 45192u16 , name : "Radeon HD 6730M" } , SubSystem { subvendor : 6058u16 , subdevice : 14722u16 , name : "Radeon HD 6730M" }] } , Device { vendor_id : 4098u16 , id : 26433u16 , name : "Whistler [Radeon HD 6630M/6650M/6750M/7670M/7690M]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 9102u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4121u16 , subdevice : 9103u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 889u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 891u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 894u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 898u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 900u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 901u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 902u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 903u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 904u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1090u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1105u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1161u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1163u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1164u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1290u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1291u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1292u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1294u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1295u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1299u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1300u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1301u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1302u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1310u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1311u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1312u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1313u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1322u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1365u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1366u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1373u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1374u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1389u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1434u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1435u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1438u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1439u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1536u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1541u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1542u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1561u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4136u16 , subdevice : 1217u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1221u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1229u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1239u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1241u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4136u16 , subdevice : 1325u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4156u16 , subdevice : 5655u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4156u16 , subdevice : 5702u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 5703u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4156u16 , subdevice : 5707u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4156u16 , subdevice : 5710u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4156u16 , subdevice : 5768u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 5769u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 5770u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 6240u16 , name : "Radeon HD 7690M" } , SubSystem { subvendor : 4156u16 , subdevice : 13189u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4156u16 , subdevice : 13664u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 13709u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 13712u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 13715u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 4156u16 , subdevice : 13932u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4163u16 , subdevice : 7378u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8481u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8482u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8483u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8485u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8487u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4173u16 , subdevice : 36987u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4173u16 , subdevice : 36992u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4173u16 , subdevice : 36993u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4203u16 , subdevice : 226u16 , name : "MacBookPro8,2 [Core i7, 15\", Late 2011]" } , SubSystem { subvendor : 4473u16 , subdevice : 64867u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 4473u16 , subdevice : 64869u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 5197u16 , subdevice : 49299u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 5197u16 , subdevice : 49324u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 5197u16 , subdevice : 49331u16 , name : "Radeon HD 6750M" } , SubSystem { subvendor : 5197u16 , subdevice : 50489u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 5197u16 , subdevice : 50697u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 5421u16 , subdevice : 2324u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 6058u16 , subdevice : 8673u16 , name : "Radeon HD 6630M" } , SubSystem { subvendor : 6058u16 , subdevice : 14704u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 6058u16 , subdevice : 14710u16 , name : "Radeon HD 6650M" } , SubSystem { subvendor : 6228u16 , subdevice : 2311u16 , name : "Radeon HD 6650M" }] } , Device { vendor_id : 4098u16 , id : 26434u16 , name : "Whistler LE [Radeon HD 6610M/7610M]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 4121u16 , subdevice : 9107u16 , name : "Radeon HD 6610M" } , SubSystem { subvendor : 4163u16 , subdevice : 7554u16 , name : "K53SK Laptop Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64290u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64291u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64295u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64298u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64300u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64304u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64305u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64306u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64312u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64313u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64314u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64315u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64320u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64321u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64327u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64328u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64329u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64337u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64338u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64339u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64342u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64385u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64386u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64387u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64598u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64724u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 4473u16 , subdevice : 64750u16 , name : "Radeon HD 7610M" } , SubSystem { subvendor : 5208u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 5218u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 5260u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 5762u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 5963u16 , subdevice : 21872u16 , name : "Turks [Radeon HD 5570]" } , SubSystem { subvendor : 5963u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 5963u16 , subdevice : 30064u16 , name : "Turks [Radeon HD 7570]" } , SubSystem { subvendor : 5963u16 , subdevice : 34064u16 , name : "Turks [Radeon HD 8510]" } , SubSystem { subvendor : 5963u16 , subdevice : 34160u16 , name : "Turks [Radeon HD 8570]" } , SubSystem { subvendor : 6023u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 6063u16 , subdevice : 25968u16 , name : "Turks [Radeon HD 6570]" } , SubSystem { subvendor : 32902u16 , subdevice : 8465u16 , name : "Radeon HD 6625M" }] } , Device { vendor_id : 4098u16 , id : 26435u16 , name : "Whistler [Radeon E6760]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26441u16 , name : "Turks GL [FirePro V4900]" , subsystems : & [SubSystem { subvendor : 5571u16 , subdevice : 11014u16 , name : "MED-X4900 (EIZO)" }] } , Device { vendor_id : 4098u16 , id : 26442u16 , name : "Turks GL [FirePro V3900]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15650u16 , name : "MXRT-2500" } , SubSystem { subvendor : 5571u16 , subdevice : 262u16 , name : "MED-X3900" }] } , Device { vendor_id : 4098u16 , id : 26448u16 , name : "Onega [Radeon HD 6650A/7650A]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 9840u16 , name : "Radeon HD 6670A" } , SubSystem { subvendor : 6058u16 , subdevice : 12409u16 , name : "Radeon HD 7650A" } , SubSystem { subvendor : 6058u16 , subdevice : 12410u16 , name : "Radeon HD 6650A" } , SubSystem { subvendor : 6058u16 , subdevice : 12423u16 , name : "Radeon HD 7650A" } , SubSystem { subvendor : 6058u16 , subdevice : 13848u16 , name : "Radeon HD 6650A" } , SubSystem { subvendor : 6058u16 , subdevice : 13859u16 , name : "Radeon HD 6650A" } , SubSystem { subvendor : 6058u16 , subdevice : 13863u16 , name : "Radeon HD 6650A" }] } , Device { vendor_id : 4098u16 , id : 26449u16 , name : "Turks [Radeon HD 7650A/7670A]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1352u16 , name : "Radeon HD 7650A" } , SubSystem { subvendor : 5218u16 , subdevice : 9841u16 , name : "Radeon HD 7670A" } , SubSystem { subvendor : 5218u16 , subdevice : 9842u16 , name : "Radeon HD 7670A" } , SubSystem { subvendor : 5218u16 , subdevice : 9856u16 , name : "Radeon HD 7650A" } , SubSystem { subvendor : 5218u16 , subdevice : 9857u16 , name : "Radeon HD 7650A" } , SubSystem { subvendor : 6058u16 , subdevice : 12423u16 , name : "Radeon HD 7650A" }] } , Device { vendor_id : 4098u16 , id : 26456u16 , name : "Turks XT [Radeon HD 6670/7670]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2830u16 , name : "Radeon HD 6670" } , SubSystem { subvendor : 4156u16 , subdevice : 26754u16 , name : "Radeon HD 6670" } , SubSystem { subvendor : 5218u16 , subdevice : 9482u16 , name : "Radeon HD 7670" } , SubSystem { subvendor : 5260u16 , subdevice : 30320u16 , name : "Radeon HD 7670" } , SubSystem { subvendor : 5445u16 , subdevice : 30320u16 , name : "Radeon HD 7670" } , SubSystem { subvendor : 5762u16 , subdevice : 13056u16 , name : "Radeon HD 7670" } , SubSystem { subvendor : 5963u16 , subdevice : 30320u16 , name : "Radeon HD 7670" } , SubSystem { subvendor : 5963u16 , subdevice : 57729u16 , name : "Radeon HD 6670" } , SubSystem { subvendor : 6023u16 , subdevice : 8969u16 , name : "Radeon HD 6670" }] } , Device { vendor_id : 4098u16 , id : 26457u16 , name : "Turks PRO [Radeon HD 6570/7570/8550 / R5 230]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12592u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 4163u16 , subdevice : 1027u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 5218u16 , subdevice : 9472u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 5218u16 , subdevice : 9481u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5260u16 , subdevice : 30064u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5698u16 , subdevice : 14951u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 5762u16 , subdevice : 12928u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5762u16 , subdevice : 13616u16 , name : "Radeon HD 8550" } , SubSystem { subvendor : 5762u16 , subdevice : 21040u16 , name : "Radeon R5 230 series" } , SubSystem { subvendor : 5762u16 , subdevice : 25680u16 , name : "Radeon HD 6450 series" } , SubSystem { subvendor : 5963u16 , subdevice : 30064u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5963u16 , subdevice : 34128u16 , name : "Radeon HD8550 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 34160u16 , name : "Radeon HD8550 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 57666u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 5963u16 , subdevice : 57729u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6023u16 , subdevice : 41520u16 , name : "Radeon R5 230 series" } , SubSystem { subvendor : 6023u16 , subdevice : 42064u16 , name : "Radeon HD 6450 series" } , SubSystem { subvendor : 6922u16 , subdevice : 37007u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6922u16 , subdevice : 37008u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6922u16 , subdevice : 37009u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6922u16 , subdevice : 37010u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6922u16 , subdevice : 37022u16 , name : "Radeon HD 6570" } , SubSystem { subvendor : 6922u16 , subdevice : 37045u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 6922u16 , subdevice : 37046u16 , name : "Radeon HD 7570" }] } , Device { vendor_id : 4098u16 , id : 26459u16 , name : "Turks [Radeon HD 7600 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26461u16 , name : "Turks PRO [Radeon HD 7570]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26463u16 , name : "Turks LE [Radeon HD 5570/6510/7510/8510]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 25872u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 5260u16 , subdevice : 25904u16 , name : "Radeon HD 6530" } , SubSystem { subvendor : 5260u16 , subdevice : 29968u16 , name : "Radeon HD 7510" } , SubSystem { subvendor : 5445u16 , subdevice : 30064u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5963u16 , subdevice : 25872u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 5963u16 , subdevice : 29968u16 , name : "Radeon HD 7510" } , SubSystem { subvendor : 5963u16 , subdevice : 34064u16 , name : "Radeon HD 8510" } , SubSystem { subvendor : 6023u16 , subdevice : 8210u16 , name : "Radeon HD 5570 2GB GDDR3" } , SubSystem { subvendor : 6023u16 , subdevice : 8980u16 , name : "Radeon HD 5570 1GB DDR2/GDDR3" }] } , Device { vendor_id : 4098u16 , id : 26464u16 , name : "Seymour [Radeon HD 6400M/7400M Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 292u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4098u16 , subdevice : 308u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4121u16 , subdevice : 9099u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4121u16 , subdevice : 9102u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4121u16 , subdevice : 9104u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4121u16 , subdevice : 39301u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1217u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1219u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1226u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1227u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1228u16 , name : "Vostro 3350" } , SubSystem { subvendor : 4136u16 , subdevice : 1233u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1235u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1239u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1282u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1283u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1286u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1287u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1300u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4136u16 , subdevice : 1308u16 , name : "Radeon HD 6450M / 7430M" } , SubSystem { subvendor : 4136u16 , subdevice : 1309u16 , name : "Radeon HD 6450M / 7430M" } , SubSystem { subvendor : 4156u16 , subdevice : 5658u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5659u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5662u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5663u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5666u16 , name : "Radeon HD 6450M" } , SubSystem { subvendor : 4156u16 , subdevice : 5667u16 , name : "Radeon HD 6450M" } , SubSystem { subvendor : 4156u16 , subdevice : 5706u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5709u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5713u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5718u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5720u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5721u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5723u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5725u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5727u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5729u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5731u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5733u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5735u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5737u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5739u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5740u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5742u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5744u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5746u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5754u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5755u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5757u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5759u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 5772u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5775u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5780u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5782u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5784u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5786u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 5788u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 6229u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 4156u16 , subdevice : 6233u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 4156u16 , subdevice : 6236u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 4156u16 , subdevice : 6237u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4156u16 , subdevice : 6239u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4156u16 , subdevice : 6243u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 4156u16 , subdevice : 13660u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13663u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13667u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13669u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13671u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13673u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13697u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13700u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13708u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13711u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13714u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13718u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 4156u16 , subdevice : 13931u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4156u16 , subdevice : 13937u16 , name : "FirePro M3900" } , SubSystem { subvendor : 4156u16 , subdevice : 13939u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4106u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4108u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4123u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4124u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4138u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 4163u16 , subdevice : 4140u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4171u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4189u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4203u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4205u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 4221u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 7346u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 7458u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 7474u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 8193u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 8194u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 8455u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 8456u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 8457u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 33952u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 34025u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4163u16 , subdevice : 34069u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 34071u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4163u16 , subdevice : 34138u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4173u16 , subdevice : 36987u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4173u16 , subdevice : 36993u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4173u16 , subdevice : 36996u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4173u16 , subdevice : 36997u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Radeon HD 6450M" } , SubSystem { subvendor : 4473u16 , subdevice : 3u16 , name : "Radeon HD 6450M" } , SubSystem { subvendor : 4473u16 , subdevice : 4u16 , name : "Radeon HD 6450M" } , SubSystem { subvendor : 4473u16 , subdevice : 64290u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64291u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64300u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64305u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64306u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64307u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64312u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64313u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64314u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64320u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64321u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64322u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64327u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64328u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64337u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64338u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64339u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64385u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64386u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64387u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64593u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64594u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64598u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64723u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64724u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 64750u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 4473u16 , subdevice : 65006u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 5197u16 , subdevice : 45172u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 45188u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 49301u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 49331u16 , name : "Radeon HD 6490M" } , SubSystem { subvendor : 5197u16 , subdevice : 50488u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 50561u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 50569u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5197u16 , subdevice : 50697u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 5197u16 , subdevice : 50725u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 5197u16 , subdevice : 50742u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 5218u16 , subdevice : 4268u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 5421u16 , subdevice : 2326u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 8677u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 14592u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6058u16 , subdevice : 14594u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6058u16 , subdevice : 14697u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 14704u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6058u16 , subdevice : 14710u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 14715u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 14717u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6058u16 , subdevice : 20737u16 , name : "Radeon HD 7470M" } , SubSystem { subvendor : 6058u16 , subdevice : 20738u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6058u16 , subdevice : 20739u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6058u16 , subdevice : 20742u16 , name : "Radeon HD 7450M" } , SubSystem { subvendor : 6228u16 , subdevice : 2199u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6228u16 , subdevice : 2304u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6228u16 , subdevice : 2312u16 , name : "Radeon HD 6470M" } , SubSystem { subvendor : 6228u16 , subdevice : 8213u16 , name : "Radeon HD 6470M" }] } , Device { vendor_id : 4098u16 , id : 26465u16 , name : "Seymour LP [Radeon HD 6430M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26467u16 , name : "Seymour [Radeon E6460]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26468u16 , name : "Seymour [Radeon HD 6400M Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26469u16 , name : "Seymour [Radeon HD 6400M Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26470u16 , name : "Caicos" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26471u16 , name : "Caicos" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26472u16 , name : "Caicos" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26480u16 , name : "Caicos [Radeon HD 6450A/7450A]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 12429u16 , name : "Radeon HD 7450A" } , SubSystem { subvendor : 6058u16 , subdevice : 13859u16 , name : "Radeon HD 6450A" } , SubSystem { subvendor : 6058u16 , subdevice : 13863u16 , name : "Radeon HD 6450A" } , SubSystem { subvendor : 6058u16 , subdevice : 13865u16 , name : "Radeon HD 6450A" } , SubSystem { subvendor : 6058u16 , subdevice : 13884u16 , name : "Radeon HD 6450A" } , SubSystem { subvendor : 6058u16 , subdevice : 13912u16 , name : "Radeon HD 7470A" }] } , Device { vendor_id : 4098u16 , id : 26481u16 , name : "Caicos XTX [Radeon HD 8490 / R5 235X OEM]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26482u16 , name : "Caicos [Radeon HD 7450A]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26488u16 , name : "Caicos XT [Radeon HD 7470/8470 / R5 235/310 OEM]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 36u16 , name : "Radeon HD 7470" } , SubSystem { subvendor : 4121u16 , subdevice : 39u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 4136u16 , subdevice : 8480u16 , name : "Radeon HD 7470" } , SubSystem { subvendor : 5218u16 , subdevice : 46225u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 5218u16 , subdevice : 46226u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 5218u16 , subdevice : 46227u16 , name : "Radeon HD 8470 OEM" } , SubSystem { subvendor : 5218u16 , subdevice : 46233u16 , name : "Radeon R5 235 OEM" } , SubSystem { subvendor : 5698u16 , subdevice : 15461u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 5698u16 , subdevice : 15477u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 5963u16 , subdevice : 33093u16 , name : "Radeon HD 8470" } , SubSystem { subvendor : 5963u16 , subdevice : 53573u16 , name : "Radeon R5 235 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 54069u16 , name : "Radeon R5 310 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 57669u16 , name : "Radeon HD 7470" } , SubSystem { subvendor : 6058u16 , subdevice : 13972u16 , name : "Radeon R5 A220" }] } , Device { vendor_id : 4098u16 , id : 26489u16 , name : "Caicos [Radeon HD 6450/7450/8450 / R5 230 OEM]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 22u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 4121u16 , subdevice : 23u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 4121u16 , subdevice : 24u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 4136u16 , subdevice : 8480u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 4156u16 , subdevice : 8488u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 4156u16 , subdevice : 10990u16 , name : "Radeon HD 7450A" } , SubSystem { subvendor : 4163u16 , subdevice : 1147u16 , name : "EAH6450 SILENT/DI/1GD3(LP)" } , SubSystem { subvendor : 4242u16 , subdevice : 25680u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5218u16 , subdevice : 8485u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5218u16 , subdevice : 9030u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5218u16 , subdevice : 9360u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5218u16 , subdevice : 9364u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5218u16 , subdevice : 9366u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5260u16 , subdevice : 29776u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5260u16 , subdevice : 33872u16 , name : "Radeon HD 8450 OEM" } , SubSystem { subvendor : 5445u16 , subdevice : 29808u16 , name : "Radeon HD 7470" } , SubSystem { subvendor : 5698u16 , subdevice : 14949u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5698u16 , subdevice : 14950u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5698u16 , subdevice : 14965u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5698u16 , subdevice : 14966u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5762u16 , subdevice : 12800u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5963u16 , subdevice : 29776u16 , name : "Radeon HD 7450" } , SubSystem { subvendor : 5963u16 , subdevice : 57639u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5963u16 , subdevice : 57683u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5963u16 , subdevice : 57700u16 , name : "Radeon HD 6450 1 GB DDR3" } , SubSystem { subvendor : 5963u16 , subdevice : 57728u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 5963u16 , subdevice : 57857u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 6023u16 , subdevice : 8977u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 6063u16 , subdevice : 33872u16 , name : "Radeon HD 8450 OEM" } , SubSystem { subvendor : 6922u16 , subdevice : 37014u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 6922u16 , subdevice : 37015u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 6922u16 , subdevice : 37032u16 , name : "Radeon HD 6450A" } , SubSystem { subvendor : 6922u16 , subdevice : 37041u16 , name : "Radeon HD 6450" } , SubSystem { subvendor : 6922u16 , subdevice : 37043u16 , name : "Radeon HD 7450A" } , SubSystem { subvendor : 6922u16 , subdevice : 37051u16 , name : "Radeon HD 7450A" }] } , Device { vendor_id : 4098u16 , id : 26491u16 , name : "Caicos PRO [Radeon HD 7450]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26496u16 , name : "Tahiti XT GL [FirePro W9000]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26500u16 , name : "Tahiti [FirePro Series Graphics Adapter]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26504u16 , name : "Tahiti [FirePro Series Graphics Adapter]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26506u16 , name : "Tahiti PRO GL [FirePro Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 780u16 , name : "FirePro W8000" } , SubSystem { subvendor : 4098u16 , subdevice : 784u16 , name : "FirePro S9000" } , SubSystem { subvendor : 4098u16 , subdevice : 1056u16 , name : "Radeon Sky 700" } , SubSystem { subvendor : 4098u16 , subdevice : 1058u16 , name : "Radeon Sky 900" } , SubSystem { subvendor : 4098u16 , subdevice : 1808u16 , name : "FirePro S9050" } , SubSystem { subvendor : 4098u16 , subdevice : 2830u16 , name : "FirePro S10000 Passive" } , SubSystem { subvendor : 4098u16 , subdevice : 2858u16 , name : "FirePro S10000" } , SubSystem { subvendor : 4136u16 , subdevice : 780u16 , name : "FirePro W8000" } , SubSystem { subvendor : 4136u16 , subdevice : 1808u16 , name : "FirePro S9000" }] } , Device { vendor_id : 4098u16 , id : 26520u16 , name : "Tahiti XT [Radeon HD 7970/8970 OEM / R9 280X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 12288u16 , name : "Tahiti XT2 [Radeon HD 7970 GHz Edition]" } , SubSystem { subvendor : 4098u16 , subdevice : 12289u16 , name : "Tahiti XTL [Radeon R9 280X]" } , SubSystem { subvendor : 4098u16 , subdevice : 16384u16 , name : "Radeon HD 8970 OEM" } , SubSystem { subvendor : 4163u16 , subdevice : 1052u16 , name : "HD 7970 DirectCU II" } , SubSystem { subvendor : 4163u16 , subdevice : 1056u16 , name : "HD 7970 DirectCU II TOP" } , SubSystem { subvendor : 4163u16 , subdevice : 1092u16 , name : "HD 7970 DirectCU II TOP" } , SubSystem { subvendor : 4163u16 , subdevice : 1096u16 , name : "HD 7970 DirectCU II TOP" } , SubSystem { subvendor : 4163u16 , subdevice : 1098u16 , name : "Tahiti XT2 [Matrix HD 7970]" } , SubSystem { subvendor : 4163u16 , subdevice : 1100u16 , name : "Tahiti XT2 [Matrix HD 7970 Platinum]" } , SubSystem { subvendor : 4163u16 , subdevice : 12289u16 , name : "Tahiti XTL [ROG Matrix R9 280X]" } , SubSystem { subvendor : 4163u16 , subdevice : 12294u16 , name : "Tahiti XTL [Radeon R9 280X DirectCU II TOP]" } , SubSystem { subvendor : 4163u16 , subdevice : 39321u16 , name : "ARES II" } , SubSystem { subvendor : 4203u16 , subdevice : 295u16 , name : "FirePro D700" } , SubSystem { subvendor : 4203u16 , subdevice : 296u16 , name : "FirePro D700" } , SubSystem { subvendor : 4242u16 , subdevice : 12288u16 , name : "Tahiti XT2 [Radeon HD 7970 GHz Edition]" } , SubSystem { subvendor : 5208u16 , subdevice : 8801u16 , name : "Tahiti XT2 [Radeon HD 7970 GHz Edition OC]" } , SubSystem { subvendor : 5208u16 , subdevice : 12289u16 , name : "Tahiti XTL [Radeon R9 280X OC]" } , SubSystem { subvendor : 5218u16 , subdevice : 10100u16 , name : "HD 7970 TwinFrozr III Boost Edition OC" } , SubSystem { subvendor : 5762u16 , subdevice : 12289u16 , name : "Tahiti XTL [Radeon R9 280X]" } , SubSystem { subvendor : 5762u16 , subdevice : 12817u16 , name : "Double D HD 7970 Black Edition" } , SubSystem { subvendor : 5762u16 , subdevice : 12819u16 , name : "HD 7970 Black Edition" } , SubSystem { subvendor : 5762u16 , subdevice : 12820u16 , name : "Double D HD 7970" } , SubSystem { subvendor : 6023u16 , subdevice : 8220u16 , name : "HD 7970 IceQ X²" } , SubSystem { subvendor : 6023u16 , subdevice : 8983u16 , name : "Radeon HD 7990" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Tahiti XT2 [Radeon HD 7970 GHz Edition]" }] } , Device { vendor_id : 4098u16 , id : 26522u16 , name : "Tahiti PRO [Radeon HD 7950/8950 OEM / R9 280]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2817u16 , name : "Radeon HD 8950 OEM" } , SubSystem { subvendor : 4098u16 , subdevice : 12288u16 , name : "Tahiti PRO2 [Radeon HD 7950 Boost]" } , SubSystem { subvendor : 5218u16 , subdevice : 12288u16 , name : "Radeon HD 8950 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 40963u16 , name : "Radeon R9 280" }] } , Device { vendor_id : 4098u16 , id : 26523u16 , name : "Malta [Radeon HD 7990/8990 OEM]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2856u16 , name : "Radeon HD 8990 OEM" } , SubSystem { subvendor : 4098u16 , subdevice : 2858u16 , name : "Radeon HD 7990" } , SubSystem { subvendor : 5218u16 , subdevice : 32822u16 , name : "Radeon HD 8990 OEM" } , SubSystem { subvendor : 5260u16 , subdevice : 35216u16 , name : "Radeon HD 8990 OEM" }] } , Device { vendor_id : 4098u16 , id : 26526u16 , name : "Tahiti LE [Radeon HD 7870 XT]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 293u16 , name : "FirePro D500" } , SubSystem { subvendor : 4203u16 , subdevice : 294u16 , name : "FirePro D500" } , SubSystem { subvendor : 6023u16 , subdevice : 9000u16 , name : "Radeon HD 7870 Black Edition 2 GB GDDR5 [2GBD5-2DHV3E]" }] } , Device { vendor_id : 4098u16 , id : 26527u16 , name : "Tahiti" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26528u16 , name : "Hawaii XT GL [FirePro W9100]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 821u16 , name : "FirePro S9150" } , SubSystem { subvendor : 4098u16 , subdevice : 1845u16 , name : "FirePro S9170" } , SubSystem { subvendor : 4136u16 , subdevice : 799u16 , name : "FirePro W9100" } , SubSystem { subvendor : 4136u16 , subdevice : 821u16 , name : "FirePro S9150" }] } , Device { vendor_id : 4098u16 , id : 26529u16 , name : "Hawaii PRO GL [FirePro W8100]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 821u16 , name : "FirePro S9100" } , SubSystem { subvendor : 4136u16 , subdevice : 821u16 , name : "FirePro S9100" }] } , Device { vendor_id : 4098u16 , id : 26530u16 , name : "Hawaii GL" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26536u16 , name : "Hawaii" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26537u16 , name : "Hawaii" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26538u16 , name : "Hawaii" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26544u16 , name : "Hawaii XT / Grenada XT [Radeon R9 290X/390X]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2816u16 , name : "Grenada XT [Radeon R9 390X]" } , SubSystem { subvendor : 4156u16 , subdevice : 25958u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 4163u16 , subdevice : 1130u16 , name : "R9 290X DirectCU II" } , SubSystem { subvendor : 4163u16 , subdevice : 1132u16 , name : "R9 290X DirectCU II OC" } , SubSystem { subvendor : 4163u16 , subdevice : 1140u16 , name : "Matrix R9 290X Platinum" } , SubSystem { subvendor : 4163u16 , subdevice : 1142u16 , name : "ARES III" } , SubSystem { subvendor : 4163u16 , subdevice : 1239u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 4163u16 , subdevice : 1243u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 4163u16 , subdevice : 1247u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 4163u16 , subdevice : 1257u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 5208u16 , subdevice : 8828u16 , name : "R9 290X WindForce 3X OC" } , SubSystem { subvendor : 5208u16 , subdevice : 8833u16 , name : "R9 290X WindForce 3X OC" } , SubSystem { subvendor : 5208u16 , subdevice : 8844u16 , name : "R9 290X WindForce 3X" } , SubSystem { subvendor : 5208u16 , subdevice : 8845u16 , name : "R9 290X WindForce 3X OC" } , SubSystem { subvendor : 5208u16 , subdevice : 8848u16 , name : "R9 290X WindForce 3X" } , SubSystem { subvendor : 5208u16 , subdevice : 8892u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 5208u16 , subdevice : 8897u16 , name : "Grenada PRO [Radeon R9 390]" } , SubSystem { subvendor : 5218u16 , subdevice : 8213u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 5218u16 , subdevice : 12400u16 , name : "R9 290X Lightning" } , SubSystem { subvendor : 5218u16 , subdevice : 12401u16 , name : "R9 290X Lightning" } , SubSystem { subvendor : 5218u16 , subdevice : 12402u16 , name : "R9 290X Lightning LE" } , SubSystem { subvendor : 5218u16 , subdevice : 12416u16 , name : "R9 290X Gaming" } , SubSystem { subvendor : 5218u16 , subdevice : 12418u16 , name : "R9 290X Gaming OC" } , SubSystem { subvendor : 5260u16 , subdevice : 9031u16 , name : "Devil 13 Dual Core R9 290X" } , SubSystem { subvendor : 5260u16 , subdevice : 9047u16 , name : "Grenada XT [Radeon R9 390X]" } , SubSystem { subvendor : 5762u16 , subdevice : 37520u16 , name : "Double Dissipation R9 290X" } , SubSystem { subvendor : 5762u16 , subdevice : 37781u16 , name : "Grenada XT [Radeon R9 390X]" } , SubSystem { subvendor : 5963u16 , subdevice : 3636u16 , name : "Radeon R9 390X" } , SubSystem { subvendor : 5963u16 , subdevice : 57986u16 , name : "Vapor-X R9 290X Tri-X OC" } , SubSystem { subvendor : 5963u16 , subdevice : 57989u16 , name : "R9 290X Tri-X OC" } , SubSystem { subvendor : 5963u16 , subdevice : 58148u16 , name : "Grenada XT2 [Radeon R9 390X]" } , SubSystem { subvendor : 6023u16 , subdevice : 8224u16 , name : "R9 290X IceQ X² Turbo" } , SubSystem { subvendor : 6023u16 , subdevice : 9047u16 , name : "Grenada XT [Radeon R9 390X]" }] } , Device { vendor_id : 4098u16 , id : 26545u16 , name : "Hawaii PRO [Radeon R9 290/390]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1245u16 , name : "STRIX R9 390" } , SubSystem { subvendor : 5260u16 , subdevice : 9048u16 , name : "Radeon R9 390" } , SubSystem { subvendor : 5963u16 , subdevice : 58148u16 , name : "Sapphire Nitro R9 390" }] } , Device { vendor_id : 4098u16 , id : 26552u16 , name : "Hawaii XT [Radeon R9 290X Engineering Sample]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26553u16 , name : "Vesuvius [Radeon R9 295X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26558u16 , name : "Hawaii LE" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26560u16 , name : "Ellesmere [Radeon Pro WX 7100 Mobile]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26562u16 , name : "Ellesmere [Radeon Pro V7300X / V7350x2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26564u16 , name : "Ellesmere [Radeon Pro WX 7100]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 822u16 , name : "Radeon Pro Duo" } , SubSystem { subvendor : 4098u16 , subdevice : 4918u16 , name : "Radeon Pro Duo" }] } , Device { vendor_id : 4098u16 , id : 26567u16 , name : "Ellesmere [Radeon Pro WX 5100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26570u16 , name : "Ellesmere [Polaris10]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26572u16 , name : "Ellesmere [Polaris10]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26575u16 , name : "Ellesmere [Polaris10]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26576u16 , name : "Ellesmere [Radeon Pro V7300X / V7350x2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26580u16 , name : "Ellesmere [Radeon Pro WX 7100 / Barco MXRT-8700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26583u16 , name : "Ellesmere [Radeon Pro WX 5100 / Barco MXRT-6700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26591u16 , name : "Ellesmere [Radeon RX 470/480/570/570X/580/580X/590]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2871u16 , name : "Radeon RX 480" } , SubSystem { subvendor : 4136u16 , subdevice : 5922u16 , name : "Radeon RX 570X" } , SubSystem { subvendor : 4136u16 , subdevice : 5923u16 , name : "Radeon RX 580X" } , SubSystem { subvendor : 4156u16 , subdevice : 33806u16 , name : "Radeon RX 580 4GB" } , SubSystem { subvendor : 4163u16 , subdevice : 1192u16 , name : "Radeon RX 480" } , SubSystem { subvendor : 4163u16 , subdevice : 1200u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 4163u16 , subdevice : 1275u16 , name : "Radeon RX 480" } , SubSystem { subvendor : 4163u16 , subdevice : 1277u16 , name : "Radeon RX 480 8GB" } , SubSystem { subvendor : 4163u16 , subdevice : 1386u16 , name : "Radeon RX 590" } , SubSystem { subvendor : 4203u16 , subdevice : 353u16 , name : "Radeon Pro 580" } , SubSystem { subvendor : 4203u16 , subdevice : 354u16 , name : "Radeon Pro 575" } , SubSystem { subvendor : 4203u16 , subdevice : 355u16 , name : "Radeon Pro 570" } , SubSystem { subvendor : 5208u16 , subdevice : 8944u16 , name : "Radeon RX 570" } , SubSystem { subvendor : 5208u16 , subdevice : 8951u16 , name : "Radeon RX 570 Gaming 4G" } , SubSystem { subvendor : 5218u16 , subdevice : 13329u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 5218u16 , subdevice : 13331u16 , name : "Radeon RX 480 Gaming X 8GB" } , SubSystem { subvendor : 5218u16 , subdevice : 13334u16 , name : "Radeon RX 570" } , SubSystem { subvendor : 5218u16 , subdevice : 13336u16 , name : "Radeon RX 580 ARMOR 8G OC" } , SubSystem { subvendor : 5218u16 , subdevice : 13339u16 , name : "Radeon RX 570 Armor 8G OC" } , SubSystem { subvendor : 5218u16 , subdevice : 13342u16 , name : "Radeon RX 570 Armor 4G OC" } , SubSystem { subvendor : 5218u16 , subdevice : 32926u16 , name : "Radeon RX 480 4GB" } , SubSystem { subvendor : 5218u16 , subdevice : 35474u16 , name : "Radeon RX 580" } , SubSystem { subvendor : 5260u16 , subdevice : 9074u16 , name : "Radeon RX 480 [Red Dragon]" } , SubSystem { subvendor : 5260u16 , subdevice : 9075u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 5260u16 , subdevice : 9079u16 , name : "Red Devil RX 580 8G Golden" } , SubSystem { subvendor : 5260u16 , subdevice : 9080u16 , name : "Radeon RX 580" } , SubSystem { subvendor : 5260u16 , subdevice : 9081u16 , name : "Radeon RX 570 4G [Red Dragon]" } , SubSystem { subvendor : 5260u16 , subdevice : 9105u16 , name : "Radeon RX 590 [Red Devil]" } , SubSystem { subvendor : 5762u16 , subdevice : 38000u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 5762u16 , subdevice : 38016u16 , name : "Radeon RX 480" } , SubSystem { subvendor : 5762u16 , subdevice : 38279u16 , name : "Radeon RX 590 FATBOY 8GB" } , SubSystem { subvendor : 5762u16 , subdevice : 38280u16 , name : "Radeon RX 580 XTR" } , SubSystem { subvendor : 5762u16 , subdevice : 50544u16 , name : "Radeon RX 570" } , SubSystem { subvendor : 5762u16 , subdevice : 50560u16 , name : "Radeon RX 580" } , SubSystem { subvendor : 5963u16 , subdevice : 58183u16 , name : "Radeon RX 470/480" } , SubSystem { subvendor : 5963u16 , subdevice : 58185u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 6023u16 , subdevice : 42096u16 , name : "Radeon RX 470" } , SubSystem { subvendor : 6023u16 , subdevice : 42112u16 , name : "Radeon RX 480" } , SubSystem { subvendor : 6217u16 , subdevice : 20481u16 , name : "Phantom Gaming X RX 580 OC" } , SubSystem { subvendor : 6217u16 , subdevice : 20528u16 , name : "Phantom Gaming D Radeon RX580 8G OC" } , SubSystem { subvendor : 7586u16 , subdevice : 58179u16 , name : "Radeon RX 570 Pulse ITX 4GB" } , SubSystem { subvendor : 7586u16 , subdevice : 58195u16 , name : "Radeon RX 570 Pulse 4GB" } , SubSystem { subvendor : 7586u16 , subdevice : 58214u16 , name : "Nitro+ Radeon RX 570/580/590" } , SubSystem { subvendor : 7586u16 , subdevice : 58247u16 , name : "Radeon RX 580 Pulse 4GB" }] } , Device { vendor_id : 4098u16 , id : 26592u16 , name : "Baffin [Radeon Pro WX 4170]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33392u16 , name : "Radeon Pro WX 4170" } , SubSystem { subvendor : 4156u16 , subdevice : 33394u16 , name : "Radeon Pro WX 4170" }] } , Device { vendor_id : 4098u16 , id : 26593u16 , name : "Baffin [Polaris11]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26595u16 , name : "Baffin [Radeon Pro WX 4100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26600u16 , name : "Baffin [Radeon Pro WX 4130/4150]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1885u16 , name : "Radeon Pro WX 4150" } , SubSystem { subvendor : 4136u16 , subdevice : 1968u16 , name : "Radeon Pro WX 4130/4150" } , SubSystem { subvendor : 4136u16 , subdevice : 1969u16 , name : "Radeon Pro WX 4130" } , SubSystem { subvendor : 4136u16 , subdevice : 5981u16 , name : "Radeon Pro WX 4150" } , SubSystem { subvendor : 4136u16 , subdevice : 6064u16 , name : "Radeon Pro WX 4130/4150" } , SubSystem { subvendor : 4136u16 , subdevice : 6065u16 , name : "Radeon Pro WX 4130" } , SubSystem { subvendor : 4156u16 , subdevice : 33397u16 , name : "Radeon Pro WX 4150" } , SubSystem { subvendor : 4156u16 , subdevice : 33399u16 , name : "Radeon Pro WX 4150" }] } , Device { vendor_id : 4098u16 , id : 26601u16 , name : "Baffin [Polaris11]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26603u16 , name : "Baffin [Radeon Pro V5300X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26607u16 , name : "Baffin [Radeon RX 460/560D / Pro 450/455/460/555/555X/560/560X]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4967u16 , name : "RX560X 4GB" } , SubSystem { subvendor : 4136u16 , subdevice : 5891u16 , name : "RX 560D OEM OC 2 GB" } , SubSystem { subvendor : 4156u16 , subdevice : 13345u16 , name : "Radeon RX 460" } , SubSystem { subvendor : 4163u16 , subdevice : 1377u16 , name : "AREZ Radeon RX 560" } , SubSystem { subvendor : 4203u16 , subdevice : 352u16 , name : "Radeon Pro 460" } , SubSystem { subvendor : 4203u16 , subdevice : 358u16 , name : "Radeon Pro 455" } , SubSystem { subvendor : 4203u16 , subdevice : 359u16 , name : "Radeon Pro 450" } , SubSystem { subvendor : 4203u16 , subdevice : 377u16 , name : "Radeon Pro 560" } , SubSystem { subvendor : 4203u16 , subdevice : 378u16 , name : "Radeon Pro 555" } , SubSystem { subvendor : 4203u16 , subdevice : 399u16 , name : "Radeon Pro 560X" } , SubSystem { subvendor : 4203u16 , subdevice : 400u16 , name : "Radeon Pro 555X" } , SubSystem { subvendor : 5698u16 , subdevice : 5927u16 , name : "Polaris 21 XL [Radeon RX 560D]" } , SubSystem { subvendor : 5762u16 , subdevice : 38253u16 , name : "Polaris 21 XL [Radeon RX 560D]" }] } , Device { vendor_id : 4098u16 , id : 26623u16 , name : "Baffin [Radeon RX 550 640SP / RX 560/560X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2820u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 4136u16 , subdevice : 5921u16 , name : "Radeon RX 560X" } , SubSystem { subvendor : 4136u16 , subdevice : 5926u16 , name : "Radeon RX 560DX" } , SubSystem { subvendor : 4156u16 , subdevice : 33913u16 , name : "Radeon RX 560X Mobile" } , SubSystem { subvendor : 4163u16 , subdevice : 1212u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 4163u16 , subdevice : 1327u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 5208u16 , subdevice : 8941u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 5260u16 , subdevice : 9089u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 5762u16 , subdevice : 38240u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 6217u16 , subdevice : 20535u16 , name : "PHANTOM G R RX550 2G/M/ASRK" } , SubSystem { subvendor : 7586u16 , subdevice : 58184u16 , name : "Radeon RX 560" } , SubSystem { subvendor : 7586u16 , subdevice : 58215u16 , name : "Radeon RX 550 640SP" }] } , Device { vendor_id : 4098u16 , id : 26624u16 , name : "Wimbledon XT [Radeon HD 7970M]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 292u16 , name : "Radeon HD 7970M" } , SubSystem { subvendor : 32902u16 , subdevice : 8464u16 , name : "Radeon HD 7970M" } , SubSystem { subvendor : 32902u16 , subdevice : 8465u16 , name : "Radeon HD 7970M" }] } , Device { vendor_id : 4098u16 , id : 26625u16 , name : "Neptune XT [Radeon HD 8970M]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 292u16 , name : "Radeon HD 8970M" } , SubSystem { subvendor : 5218u16 , subdevice : 4375u16 , name : "Radeon R9 M290X" } , SubSystem { subvendor : 32902u16 , subdevice : 8464u16 , name : "Radeon HD 8970M" } , SubSystem { subvendor : 32902u16 , subdevice : 8465u16 , name : "Radeon HD 8970M" }] } , Device { vendor_id : 4098u16 , id : 26626u16 , name : "Wimbledon" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26630u16 , name : "Neptune" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26632u16 , name : "Pitcairn XT GL [FirePro W7000]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 784u16 , name : "FirePro S7000" } , SubSystem { subvendor : 4098u16 , subdevice : 1056u16 , name : "Radeon Sky 500" } , SubSystem { subvendor : 4156u16 , subdevice : 780u16 , name : "MED-X7000" } , SubSystem { subvendor : 5068u16 , subdevice : 15653u16 , name : "MXRT-7500" } , SubSystem { subvendor : 5571u16 , subdevice : 780u16 , name : "MED-X7000" }] } , Device { vendor_id : 4098u16 , id : 26633u16 , name : "Pitcairn LE GL [FirePro W5000]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15651u16 , name : "MXRT-5500" } , SubSystem { subvendor : 5068u16 , subdevice : 15652u16 , name : "MXRT-5550" } , SubSystem { subvendor : 5571u16 , subdevice : 2822u16 , name : "MED-X5000" }] } , Device { vendor_id : 4098u16 , id : 26640u16 , name : "Curacao XT / Trinidad XT [Radeon R7 370 / R9 270X/370X]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 298u16 , name : "FirePro D300" } , SubSystem { subvendor : 4203u16 , subdevice : 299u16 , name : "FirePro D300" } , SubSystem { subvendor : 5260u16 , subdevice : 2312u16 , name : "Radeon R9 370 OEM" } , SubSystem { subvendor : 5762u16 , subdevice : 29552u16 , name : "Radeon R7 370" }] } , Device { vendor_id : 4098u16 , id : 26641u16 , name : "Curacao PRO [Radeon R7 370 / R9 270/370 OEM]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 2816u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 4163u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 5208u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 5218u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 5218u16 , subdevice : 12368u16 , name : "R9 270 Gaming OC" } , SubSystem { subvendor : 5260u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 5762u16 , subdevice : 8213u16 , name : "Trinidad PRO [Radeon R7 370]" } , SubSystem { subvendor : 5963u16 , subdevice : 8213u16 , name : "NITRO Radeon R7 370" } , SubSystem { subvendor : 5963u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" } , SubSystem { subvendor : 6023u16 , subdevice : 8214u16 , name : "Trinidad PRO [Radeon R9 370 OEM]" }] } , Device { vendor_id : 4098u16 , id : 26646u16 , name : "Pitcairn" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26647u16 , name : "Pitcairn" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26648u16 , name : "Pitcairn XT [Radeon HD 7870 GHz Edition]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2821u16 , name : "Radeon HD 8870 OEM" } , SubSystem { subvendor : 5963u16 , subdevice : 35588u16 , name : "Radeon HD 8860" }] } , Device { vendor_id : 4098u16 , id : 26649u16 , name : "Pitcairn PRO [Radeon HD 7850 / R7 265 / R9 270 1024SP]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1068u16 , name : "Radeon HD 7850" } , SubSystem { subvendor : 5762u16 , subdevice : 29289u16 , name : "Radeon R9 270 1024SP" } , SubSystem { subvendor : 5762u16 , subdevice : 37496u16 , name : "Radeon R9 270 1024SP" } , SubSystem { subvendor : 5963u16 , subdevice : 40968u16 , name : "Radeon R9 270 1024SP" } , SubSystem { subvendor : 5963u16 , subdevice : 57889u16 , name : "Radeon HD 7850 2GB GDDR5 DVI-I/DVI-D/HDMI/DP" }] } , Device { vendor_id : 4098u16 , id : 26656u16 , name : "Venus XTX [Radeon HD 8890M / R9 M275X/M375X]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6225u16 , name : "Radeon HD 7750M" } , SubSystem { subvendor : 6058u16 , subdevice : 13891u16 , name : "Radeon R9 A375" } , SubSystem { subvendor : 6058u16 , subdevice : 14337u16 , name : "Radeon R9 M275" } , SubSystem { subvendor : 6058u16 , subdevice : 14372u16 , name : "Radeon R9 M375" } , SubSystem { subvendor : 7586u16 , subdevice : 57962u16 , name : "Radeon R7 250" }] } , Device { vendor_id : 4098u16 , id : 26657u16 , name : "Venus XT [Radeon HD 8870M / R9 M270X/M370X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 798u16 , name : "FirePro SX4000" } , SubSystem { subvendor : 4136u16 , subdevice : 1484u16 , name : "FirePro M5100" } , SubSystem { subvendor : 4136u16 , subdevice : 5580u16 , name : "FirePro M5100" } , SubSystem { subvendor : 4203u16 , subdevice : 329u16 , name : "Radeon R9 M370X Mac Edition" }] } , Device { vendor_id : 4098u16 , id : 26658u16 , name : "Venus PRO [Radeon E8860]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26659u16 , name : "Venus PRO [Radeon HD 8850M / R9 M265X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26661u16 , name : "Heathrow XT [Radeon HD 7870M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1343u16 , name : "FirePro M6000" } , SubSystem { subvendor : 4136u16 , subdevice : 1485u16 , name : "FirePro M6000" } , SubSystem { subvendor : 4136u16 , subdevice : 5581u16 , name : "FirePro M6000" } , SubSystem { subvendor : 4156u16 , subdevice : 5996u16 , name : "FirePro M6000" } , SubSystem { subvendor : 32902u16 , subdevice : 8465u16 , name : "Chelsea PRO" }] } , Device { vendor_id : 4098u16 , id : 26662u16 , name : "Chelsea LP [Radeon HD 7700M Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26663u16 , name : "Heathrow PRO [Radeon HD 7850M/8850M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26664u16 , name : "Cape Verde PRO [FirePro W600]" , subsystems : & [SubSystem { subvendor : 5571u16 , subdevice : 11038u16 , name : "MED-X6000" }] } , Device { vendor_id : 4098u16 , id : 26665u16 , name : "Cape Verde" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26666u16 , name : "Venus PRO" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26667u16 , name : "Cape Verde PRO / Venus LE / Tropo PRO-L [Radeon HD 8830M / R7 250 / R7 M465X]" , subsystems : & [SubSystem { subvendor : 296u16 , subdevice : 1948u16 , name : "Radeon R7 465X" } , SubSystem { subvendor : 5218u16 , subdevice : 12306u16 , name : "Radeon R7 250" }] } , Device { vendor_id : 4098u16 , id : 26668u16 , name : "Cape Verde GL [FirePro W4100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26669u16 , name : "Chelsea XT GL [FirePro M4000]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26671u16 , name : "Chelsea LP [Radeon HD 7730M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6225u16 , name : "Radeon HD 7750M" }] } , Device { vendor_id : 4098u16 , id : 26677u16 , name : "Cape Verde PRX [Radeon R9 255 OEM]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26679u16 , name : "Cape Verde LE [Radeon HD 7730/8730]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 10134u16 , name : "Radeon HD 8730" } , SubSystem { subvendor : 5218u16 , subdevice : 32914u16 , name : "Radeon HD 8730" } , SubSystem { subvendor : 5260u16 , subdevice : 34608u16 , name : "Radeon HD 8730" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 6570" }] } , Device { vendor_id : 4098u16 , id : 26685u16 , name : "Cape Verde XT [Radeon HD 7770/8760 / R7 250X]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 48u16 , name : "Radeon HD 8760 OEM" } , SubSystem { subvendor : 4121u16 , subdevice : 48u16 , name : "Radeon HD 8760 OEM" } , SubSystem { subvendor : 4156u16 , subdevice : 26768u16 , name : "Radeon HD 8760 OEM" } , SubSystem { subvendor : 4163u16 , subdevice : 34656u16 , name : "Radeon HD 8760 OEM" } , SubSystem { subvendor : 5218u16 , subdevice : 10000u16 , name : "R7770-PMD1GD5" } , SubSystem { subvendor : 5963u16 , subdevice : 33540u16 , name : "Radeon HD 8760 OEM" }] } , Device { vendor_id : 4098u16 , id : 26687u16 , name : "Cape Verde PRO [Radeon HD 7750/8740 / R7 250E]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 10128u16 , name : "Radeon HD 8740" } , SubSystem { subvendor : 5218u16 , subdevice : 10129u16 , name : "Radeon HD 8740" } , SubSystem { subvendor : 5698u16 , subdevice : 15255u16 , name : "Radeon HD 8740" }] } , Device { vendor_id : 4098u16 , id : 26688u16 , name : "Thames [Radeon HD 7500M/7600M Series]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 1294u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1295u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1299u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1300u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1389u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1434u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1435u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1438u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1536u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1542u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4133u16 , subdevice : 1686u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1687u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1688u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1689u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4133u16 , subdevice : 1879u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1386u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1390u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1432u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1437u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1443u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1465u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4136u16 , subdevice : 1467u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6025u16 , name : "FirePro M2000" } , SubSystem { subvendor : 4156u16 , subdevice : 6129u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4156u16 , subdevice : 6132u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6163u16 , name : "Radeon HD 7590M" } , SubSystem { subvendor : 4156u16 , subdevice : 6191u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6192u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6197u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6202u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6204u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6206u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6208u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6210u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6212u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6216u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6218u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6220u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6293u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6295u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6309u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6311u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4156u16 , subdevice : 6388u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 4106u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 4171u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 4316u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8481u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8482u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8483u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8485u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4163u16 , subdevice : 8487u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64273u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64290u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64291u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64300u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64305u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64306u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64312u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64313u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64314u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64320u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64321u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64327u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64328u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64337u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64338u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64339u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64385u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64386u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64387u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64598u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64724u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 4473u16 , subdevice : 64750u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 5197u16 , subdevice : 49349u16 , name : "Radeon HD 7690M" } , SubSystem { subvendor : 5197u16 , subdevice : 49358u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 5197u16 , subdevice : 49370u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 6058u16 , subdevice : 14704u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 6058u16 , subdevice : 14715u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 6058u16 , subdevice : 20737u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 6058u16 , subdevice : 20738u16 , name : "Radeon HD 7670M" } , SubSystem { subvendor : 6058u16 , subdevice : 20739u16 , name : "Radeon HD 7670M" }] } , Device { vendor_id : 4098u16 , id : 26689u16 , name : "Thames [Radeon HD 7550M/7570M/7650M]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1377u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4136u16 , subdevice : 1388u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4136u16 , subdevice : 1407u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4156u16 , subdevice : 6129u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4156u16 , subdevice : 6132u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6163u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4156u16 , subdevice : 6202u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6204u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6206u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6208u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6210u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4156u16 , subdevice : 6212u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4163u16 , subdevice : 4106u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4163u16 , subdevice : 4171u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4163u16 , subdevice : 4316u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4163u16 , subdevice : 8500u16 , name : "Radeon HD 7650M" } , SubSystem { subvendor : 4473u16 , subdevice : 1u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4473u16 , subdevice : 2u16 , name : "Radeon HD 7570M" } , SubSystem { subvendor : 4473u16 , subdevice : 64323u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 4473u16 , subdevice : 64401u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 4473u16 , subdevice : 64402u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 4473u16 , subdevice : 64403u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 4473u16 , subdevice : 64418u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 4473u16 , subdevice : 64419u16 , name : "Radeon HD 7550M" } , SubSystem { subvendor : 5197u16 , subdevice : 49351u16 , name : "Radeon HD 7550M" }] } , Device { vendor_id : 4098u16 , id : 26690u16 , name : "Thames LE [Radeon HD 7000M Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26691u16 , name : "Thames [Radeon HD 7670M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26720u16 , name : "Vega 10 [Instinct MI25/MI25x2/V340/V320]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3125u16 , name : "Radeon PRO V320" } , SubSystem { subvendor : 4098u16 , subdevice : 27765u16 , name : "Radeon PRO V320" } , SubSystem { subvendor : 4203u16 , subdevice : 380u16 , name : "Radeon Pro Vega 64" }] } , Device { vendor_id : 4098u16 , id : 26721u16 , name : "Vega 10 XT [Radeon PRO WX 9100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26722u16 , name : "Vega 10 XT [Radeon PRO SSG]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26723u16 , name : "Vega 10 XTX [Radeon Vega Frontier Edition]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26724u16 , name : "Vega 10 [Radeon Pro V340/Instinct MI25x2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26727u16 , name : "Vega 10 XL [Radeon Pro Vega 56]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26728u16 , name : "Vega 10 [Radeon PRO WX 8100/8200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26729u16 , name : "Vega 10 XGA [Radeon Pro Vega 48]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26730u16 , name : "Vega 10 LEA" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26731u16 , name : "Vega 10 XTXA [Radeon Pro Vega 64X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26732u16 , name : "Vega 10 [Instinct MI25 MxGPU/MI25x2 MxGPU/V340 MxGPU/V340L MxGPU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26733u16 , name : "Vega 10 GLXTA" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26734u16 , name : "Vega 10 GLXLA" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26751u16 , name : "Vega 10 XL/XT [Radeon RX Vega 56/64]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2870u16 , name : "RX Vega64" } , SubSystem { subvendor : 4098u16 , subdevice : 27510u16 , name : "AMD Radeon RX Vega 56 8GB" } , SubSystem { subvendor : 4163u16 , subdevice : 1220u16 , name : "Radeon RX Vega 64" } , SubSystem { subvendor : 4163u16 , subdevice : 1365u16 , name : "ROG STRIX RX Vega 56 8GB" } , SubSystem { subvendor : 5208u16 , subdevice : 8972u16 , name : "Radeon RX VEGA 56 GAMING OC 8G" } , SubSystem { subvendor : 7586u16 , subdevice : 58230u16 , name : "Radeon RX VEGA 56 Pulse 8GB OC HBM2" }] } , Device { vendor_id : 4098u16 , id : 26752u16 , name : "Lexington [Radeon HD 6550M]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5692u16 , name : "Pavilion dv6 Radeon HD 6550M" }] } , Device { vendor_id : 4098u16 , id : 26760u16 , name : "Cypress XT [FirePro V8800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26761u16 , name : "Cypress PRO [FirePro V7800]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 769u16 , name : "FirePro V7800P" } , SubSystem { subvendor : 5068u16 , subdevice : 15647u16 , name : "MXRT-7400" }] } , Device { vendor_id : 4098u16 , id : 26762u16 , name : "Cypress XT [FirePro V9800]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 780u16 , name : "FirePro V9800P" }] } , Device { vendor_id : 4098u16 , id : 26764u16 , name : "Cypress XT GL [FireStream 9370]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26765u16 , name : "Cypress PRO GL [FireStream 9350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26776u16 , name : "Cypress XT [Radeon HD 5870]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2816u16 , name : "Radeon HD 5870 Eyefinity Edition" } , SubSystem { subvendor : 4203u16 , subdevice : 208u16 , name : "Radeon HD 5870 Mac Edition" } , SubSystem { subvendor : 5218u16 , subdevice : 32818u16 , name : "Radeon HD 5870 1 GB GDDR5" } , SubSystem { subvendor : 5963u16 , subdevice : 26736u16 , name : "Radeon HD 6870 1600SP Edition" }] } , Device { vendor_id : 4098u16 , id : 26777u16 , name : "Cypress PRO [Radeon HD 5850]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 816u16 , name : "Radeon HD 5850" } , SubSystem { subvendor : 5963u16 , subdevice : 9083u16 , name : "Radeon HD 5850 X2" } , SubSystem { subvendor : 5963u16 , subdevice : 26704u16 , name : "Radeon HD 6850 1440SP Edition" }] } , Device { vendor_id : 4098u16 , id : 26779u16 , name : "Cypress PRO [Radeon HD 6800 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26780u16 , name : "Hemlock [Radeon HD 5970]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 850u16 , name : "ARES" }] } , Device { vendor_id : 4098u16 , id : 26781u16 , name : "Hemlock [Radeon HD 5970]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26782u16 , name : "Cypress LE [Radeon HD 5830]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26784u16 , name : "Broadway XT [Mobility Radeon HD 5870]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 4847u16 , name : "FirePro M7820" } , SubSystem { subvendor : 4156u16 , subdevice : 5408u16 , name : "FirePro M7820" }] } , Device { vendor_id : 4098u16 , id : 26785u16 , name : "Broadway PRO [Mobility Radeon HD 5850]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 204u16 , name : "iMac MC511 Mobility Radeon HD 5850 MXM Module" }] } , Device { vendor_id : 4098u16 , id : 26792u16 , name : "Granville [Radeon HD 6850M/6870M]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 1090u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1105u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1290u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1291u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1292u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1294u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1295u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1299u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1300u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1301u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1302u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1317u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1318u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4133u16 , subdevice : 1389u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 4136u16 , subdevice : 1167u16 , name : "Radeon HD 6870M" } , SubSystem { subvendor : 4136u16 , subdevice : 1168u16 , name : "Radeon HD 6870M" } , SubSystem { subvendor : 4136u16 , subdevice : 1209u16 , name : "Radeon HD 6870M" } , SubSystem { subvendor : 4136u16 , subdevice : 1210u16 , name : "Radeon HD 6870M" } , SubSystem { subvendor : 4156u16 , subdevice : 5531u16 , name : "Radeon HD 6850M" } , SubSystem { subvendor : 5197u16 , subdevice : 49325u16 , name : "Radeon HD 6850M" }] } , Device { vendor_id : 4098u16 , id : 26793u16 , name : "Juniper XT [FirePro V5800]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15646u16 , name : "MXRT-5400" } , SubSystem { subvendor : 5068u16 , subdevice : 15648u16 , name : "MXRT-5450" }] } , Device { vendor_id : 4098u16 , id : 26808u16 , name : "Juniper XT [Radeon HD 5770]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 207u16 , name : "MacPro5,1 [Mac Pro 2.8GHz DDR3]" }] } , Device { vendor_id : 4098u16 , id : 26809u16 , name : "Juniper LE [Radeon HD 5670 640SP Edition]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26810u16 , name : "Juniper XT [Radeon HD 6770]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26814u16 , name : "Juniper PRO [Radeon HD 5750]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Radeon HD 6750" }] } , Device { vendor_id : 4098u16 , id : 26815u16 , name : "Juniper PRO [Radeon HD 6750]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 26448u16 , name : "Radeon HD 6750" }] } , Device { vendor_id : 4098u16 , id : 26816u16 , name : "Madison [Mobility Radeon HD 5730 / 6570M]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 9091u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 4136u16 , subdevice : 674u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 4136u16 , subdevice : 766u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 4136u16 , subdevice : 1049u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 4156u16 , subdevice : 5245u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 4156u16 , subdevice : 5409u16 , name : "Madison XT [FirePro M5800]" } , SubSystem { subvendor : 4156u16 , subdevice : 5523u16 , name : "Mobility Radeon HD 6570" } , SubSystem { subvendor : 4156u16 , subdevice : 5526u16 , name : "Mobility Radeon HD 6570" } , SubSystem { subvendor : 4156u16 , subdevice : 5529u16 , name : "Mobility Radeon HD 6570" } , SubSystem { subvendor : 4163u16 , subdevice : 7202u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 6058u16 , subdevice : 14631u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 6058u16 , subdevice : 14674u16 , name : "Mobility Radeon HD 5730" } , SubSystem { subvendor : 6058u16 , subdevice : 14712u16 , name : "Radeon HD 6570M" }] } , Device { vendor_id : 4098u16 , id : 26817u16 , name : "Madison [Mobility Radeon HD 5650/5750 / 6530M/6550M]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 517u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 659u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 660u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 662u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 776u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 778u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 785u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 786u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 796u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 797u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 829u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 830u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 831u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 838u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" } , SubSystem { subvendor : 4133u16 , subdevice : 840u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 854u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 855u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 856u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 857u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 858u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 859u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 860u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 861u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 862u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 864u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 866u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 868u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 869u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 870u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 871u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 872u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 876u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 877u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 878u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 879u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 882u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 883u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 887u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 888u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 889u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 890u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 891u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 894u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 895u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 898u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 899u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 900u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 901u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 902u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 903u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 904u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 907u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 908u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 922u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1041u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1042u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1048u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1049u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1056u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1057u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1061u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1066u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1070u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1071u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1074u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1075u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1090u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1100u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1102u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1105u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1108u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1109u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1141u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1142u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1159u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1161u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1176u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 1303u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1306u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1307u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1308u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1309u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1317u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1318u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1323u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1324u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1340u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1341u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1342u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1343u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4133u16 , subdevice : 1543u16 , name : "Radeon HD 6550M" } , SubSystem { subvendor : 4136u16 , subdevice : 1051u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4136u16 , subdevice : 1095u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4136u16 , subdevice : 1096u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4136u16 , subdevice : 1110u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4136u16 , subdevice : 1111u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5174u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5175u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5184u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5192u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5193u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5194u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5195u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5243u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5276u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5278u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4156u16 , subdevice : 5409u16 , name : "Madison Pro [FirePro M5800]" } , SubSystem { subvendor : 4163u16 , subdevice : 7106u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4173u16 , subdevice : 36977u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4173u16 , subdevice : 36983u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4173u16 , subdevice : 36993u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64768u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64786u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64794u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64816u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64817u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64848u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64850u16 , name : "Radeon HD 6530M" } , SubSystem { subvendor : 4473u16 , subdevice : 64867u16 , name : "Radeon HD 6530M" } , SubSystem { subvendor : 4473u16 , subdevice : 64869u16 , name : "Radeon HD 6530M" } , SubSystem { subvendor : 4473u16 , subdevice : 64976u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4473u16 , subdevice : 64978u16 , name : "Radeon HD 6530M" } , SubSystem { subvendor : 5197u16 , subdevice : 49278u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 5197u16 , subdevice : 49285u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 5312u16 , subdevice : 67u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 5312u16 , subdevice : 77u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 6058u16 , subdevice : 14632u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 6058u16 , subdevice : 14673u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 6058u16 , subdevice : 14711u16 , name : "Radeon HD 6550M" }] } , Device { vendor_id : 4098u16 , id : 26823u16 , name : "Pinewood [Mobility Radeon HD 5570/6550A]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 8769u16 , name : "Mobility Radeon HD 5570" } , SubSystem { subvendor : 5218u16 , subdevice : 8771u16 , name : "Mobility Radeon HD 5570" } , SubSystem { subvendor : 5218u16 , subdevice : 8772u16 , name : "Mobility Radeon HD 5570" } , SubSystem { subvendor : 5218u16 , subdevice : 8773u16 , name : "Radeon HD 6550A" } , SubSystem { subvendor : 5218u16 , subdevice : 8774u16 , name : "Radeon HD 6550A" }] } , Device { vendor_id : 4098u16 , id : 26824u16 , name : "Redwood XT GL [FirePro V4800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26825u16 , name : "Redwood PRO GL [FirePro V3800]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15645u16 , name : "MXRT-2400" }] } , Device { vendor_id : 4098u16 , id : 26840u16 , name : "Redwood XT [Radeon HD 5670/5690/5730]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 26848u16 , name : "Radeon HD 5670" } , SubSystem { subvendor : 5963u16 , subdevice : 22160u16 , name : "Radeon HD 5690" } , SubSystem { subvendor : 5963u16 , subdevice : 22320u16 , name : "Radeon HD 5730" } , SubSystem { subvendor : 5963u16 , subdevice : 57681u16 , name : "Radeon HD 5670" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 5730" } , SubSystem { subvendor : 6063u16 , subdevice : 12304u16 , name : "Radeon HD 5730" } , SubSystem { subvendor : 6063u16 , subdevice : 12305u16 , name : "Radeon HD 5690" }] } , Device { vendor_id : 4098u16 , id : 26841u16 , name : "Redwood PRO [Radeon HD 5550/5570/5630/6510/6610/7570]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 26736u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 4156u16 , subdevice : 26738u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 4163u16 , subdevice : 974u16 , name : "Radeon HD 5550" } , SubSystem { subvendor : 5218u16 , subdevice : 8529u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 5218u16 , subdevice : 8768u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 5260u16 , subdevice : 12289u16 , name : "Radeon HD 6610" } , SubSystem { subvendor : 5445u16 , subdevice : 21840u16 , name : "Radeon HD 5550" } , SubSystem { subvendor : 5445u16 , subdevice : 30064u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5698u16 , subdevice : 14725u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 5698u16 , subdevice : 14742u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 5963u16 , subdevice : 25872u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 5963u16 , subdevice : 26128u16 , name : "Radeon HD 6610" } , SubSystem { subvendor : 5963u16 , subdevice : 57666u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 6063u16 , subdevice : 12288u16 , name : "Radeon HD 6510" } , SubSystem { subvendor : 6063u16 , subdevice : 12304u16 , name : "Radeon HD 5630" }] } , Device { vendor_id : 4098u16 , id : 26842u16 , name : "Redwood LE [Radeon HD 5550/5570/5630/6390/6490/7570]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 32881u16 , name : "VR5550-MD1G (Radeon HD 5550)" } , SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Radeon HD 6390" } , SubSystem { subvendor : 5260u16 , subdevice : 12289u16 , name : "Radeon HD 6490" } , SubSystem { subvendor : 5445u16 , subdevice : 30064u16 , name : "Radeon HD 7570" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Radeon HD 6390" } , SubSystem { subvendor : 5963u16 , subdevice : 21872u16 , name : "Radeon HD 5570" } , SubSystem { subvendor : 5963u16 , subdevice : 22064u16 , name : "Radeon HD 5630" } , SubSystem { subvendor : 5963u16 , subdevice : 25744u16 , name : "Radeon HD 6490" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 5630" } , SubSystem { subvendor : 6063u16 , subdevice : 12288u16 , name : "Radeon HD 6390" } , SubSystem { subvendor : 6063u16 , subdevice : 12304u16 , name : "Radeon HD 5630" }] } , Device { vendor_id : 4098u16 , id : 26846u16 , name : "Redwood" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26848u16 , name : "Park [Mobility Radeon HD 5430/5450/5470]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1028u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4136u16 , subdevice : 1044u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4136u16 , subdevice : 1076u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5171u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5172u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5225u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5227u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5254u16 , name : "TouchSmart tm2-2050er discrete GPU (Mobility Radeon HD 5450)" } , SubSystem { subvendor : 4156u16 , subdevice : 5666u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 5667u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 61166u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 4173u16 , subdevice : 36982u16 , name : "Mobility Radeon HD 5450" } , SubSystem { subvendor : 5762u16 , subdevice : 12366u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5762u16 , subdevice : 24576u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 6058u16 , subdevice : 40530u16 , name : "FirePro M3800" } , SubSystem { subvendor : 6058u16 , subdevice : 40531u16 , name : "FirePro M3800" }] } , Device { vendor_id : 4098u16 , id : 26849u16 , name : "Park [Mobility Radeon HD 5430]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1055u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 4163u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5260u16 , subdevice : 12289u16 , name : "Caicos [Radeon HD 6230]" } , SubSystem { subvendor : 5260u16 , subdevice : 12290u16 , name : "Caicos [Radeon HD 6250]" } , SubSystem { subvendor : 5260u16 , subdevice : 12291u16 , name : "Caicos [Radeon HD 6350]" } , SubSystem { subvendor : 5260u16 , subdevice : 29520u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 5260u16 , subdevice : 33616u16 , name : "Caicos [Radeon HD 8350]" } , SubSystem { subvendor : 5445u16 , subdevice : 21584u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5445u16 , subdevice : 29520u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 5762u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5762u16 , subdevice : 24576u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5762u16 , subdevice : 29520u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5963u16 , subdevice : 21616u16 , name : "Caicos [Radeon HD 5470]" } , SubSystem { subvendor : 5963u16 , subdevice : 24576u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 5963u16 , subdevice : 25136u16 , name : "Caicos [Radeon HD 6230]" } , SubSystem { subvendor : 5963u16 , subdevice : 25424u16 , name : "Caicos [Radeon HD 6350]" } , SubSystem { subvendor : 5963u16 , subdevice : 29520u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 6063u16 , subdevice : 12288u16 , name : "Caicos [Radeon HD 5450]" } , SubSystem { subvendor : 6063u16 , subdevice : 12289u16 , name : "Caicos [Radeon HD 6230]" } , SubSystem { subvendor : 6063u16 , subdevice : 12308u16 , name : "Caicos [Radeon HD 6350]" } , SubSystem { subvendor : 6063u16 , subdevice : 12309u16 , name : "Caicos [Radeon HD 7350]" } , SubSystem { subvendor : 6063u16 , subdevice : 33616u16 , name : "Caicos [Radeon HD 8350 OEM]" }] } , Device { vendor_id : 4098u16 , id : 26852u16 , name : "Robson CE [Radeon HD 6370M/7370M]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 9094u16 , name : "Radeon HD 6350M" } , SubSystem { subvendor : 4121u16 , subdevice : 9095u16 , name : "Radeon HD 6350M" } , SubSystem { subvendor : 4121u16 , subdevice : 9101u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4121u16 , subdevice : 9102u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4133u16 , subdevice : 898u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4133u16 , subdevice : 1161u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4133u16 , subdevice : 1162u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4133u16 , subdevice : 1163u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4133u16 , subdevice : 1164u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4136u16 , subdevice : 1217u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4136u16 , subdevice : 1226u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4136u16 , subdevice : 1228u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4136u16 , subdevice : 1229u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4136u16 , subdevice : 1239u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5137u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5153u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5158u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5160u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5162u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5163u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5178u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5180u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5189u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5676u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5677u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5678u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5679u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5689u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5690u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5691u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5692u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5693u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5694u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5695u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5697u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 5699u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 13688u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 13690u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 13939u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4156u16 , subdevice : 13941u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4163u16 , subdevice : 7314u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4163u16 , subdevice : 33953u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4163u16 , subdevice : 33965u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 4173u16 , subdevice : 36993u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 5445u16 , subdevice : 29520u16 , name : "Cedar [Radeon HD 7350]" } , SubSystem { subvendor : 5464u16 , subdevice : 17680u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 5464u16 , subdevice : 21765u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 5963u16 , subdevice : 21584u16 , name : "Cedar [Radeon HD 5450]" } , SubSystem { subvendor : 6058u16 , subdevice : 8669u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 6058u16 , subdevice : 8681u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 6058u16 , subdevice : 14705u16 , name : "Radeon HD 6370M" } , SubSystem { subvendor : 6058u16 , subdevice : 14706u16 , name : "Radeon HD 7370M" } , SubSystem { subvendor : 6058u16 , subdevice : 14714u16 , name : "Radeon HD 6370M/7370M" } , SubSystem { subvendor : 6058u16 , subdevice : 14715u16 , name : "Radeon HD 6370M/7370M" } , SubSystem { subvendor : 6058u16 , subdevice : 14719u16 , name : "Radeon HD 7370M" }] } , Device { vendor_id : 4098u16 , id : 26853u16 , name : "Robson LE [Radeon HD 6330M]" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 64828u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64848u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64850u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64867u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64869u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64883u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64885u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64976u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 64978u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 65002u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 4473u16 , subdevice : 65016u16 , name : "Radeon HD 6330M" } , SubSystem { subvendor : 5260u16 , subdevice : 21584u16 , name : "Cedar [Radeon HD 5450]" } , SubSystem { subvendor : 5260u16 , subdevice : 25424u16 , name : "Cedar [Radeon HD 6350]" } , SubSystem { subvendor : 5260u16 , subdevice : 29520u16 , name : "Cedar [Radeon HD 7350]" } , SubSystem { subvendor : 5260u16 , subdevice : 33616u16 , name : "Cedar [Radeon HD 8350]" } , SubSystem { subvendor : 5445u16 , subdevice : 29520u16 , name : "Cedar [Radeon HD 7350]" }] } , Device { vendor_id : 4098u16 , id : 26856u16 , name : "Cedar" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26857u16 , name : "Cedar [ATI FirePro (FireGL) Graphics Adapter]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26865u16 , name : "Cedar GL [FirePro 2460]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26866u16 , name : "Cedar GL [FirePro 2270]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26872u16 , name : "Cedar [Radeon HD 7300 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26873u16 , name : "Cedar [Radeon HD 5000/6000/7350/8350 Series]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 1u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4121u16 , subdevice : 2u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4121u16 , subdevice : 25u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 4133u16 , subdevice : 1304u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4133u16 , subdevice : 1305u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4136u16 , subdevice : 270u16 , name : "XPS 8300" } , SubSystem { subvendor : 4136u16 , subdevice : 8486u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 4156u16 , subdevice : 8486u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 4156u16 , subdevice : 10924u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 10926u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4156u16 , subdevice : 13696u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4163u16 , subdevice : 902u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 4163u16 , subdevice : 962u16 , name : "EAH5450 SILENT/DI/512MD2 (LP)" } , SubSystem { subvendor : 5208u16 , subdevice : 8664u16 , name : "Radeon HD 5450 1GB DDR3 Silent" } , SubSystem { subvendor : 5218u16 , subdevice : 8496u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 8497u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 8499u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5218u16 , subdevice : 8576u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 8577u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 8578u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5218u16 , subdevice : 8579u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5218u16 , subdevice : 8752u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 8753u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5218u16 , subdevice : 9365u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5260u16 , subdevice : 12289u16 , name : "Radeon HD 5530/6250" } , SubSystem { subvendor : 5260u16 , subdevice : 12290u16 , name : "Radeon HD 6290" } , SubSystem { subvendor : 5260u16 , subdevice : 12291u16 , name : "Radeon HD 6230" } , SubSystem { subvendor : 5260u16 , subdevice : 12292u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5260u16 , subdevice : 29520u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5260u16 , subdevice : 33616u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 5445u16 , subdevice : 29520u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5698u16 , subdevice : 14723u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5698u16 , subdevice : 14724u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5698u16 , subdevice : 14727u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5698u16 , subdevice : 14743u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5698u16 , subdevice : 14853u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5698u16 , subdevice : 15153u16 , name : "Radeon HD 6350A" } , SubSystem { subvendor : 5762u16 , subdevice : 12912u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Radeon HD 6230" } , SubSystem { subvendor : 5963u16 , subdevice : 14727u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5963u16 , subdevice : 21616u16 , name : "Radeon HD 5470" } , SubSystem { subvendor : 5963u16 , subdevice : 21648u16 , name : "Radeon HD 5490" } , SubSystem { subvendor : 5963u16 , subdevice : 21808u16 , name : "Radeon HD 5530" } , SubSystem { subvendor : 5963u16 , subdevice : 25136u16 , name : "Radeon HD 6230" } , SubSystem { subvendor : 5963u16 , subdevice : 25168u16 , name : "Radeon HD 6250" } , SubSystem { subvendor : 5963u16 , subdevice : 25232u16 , name : "Radeon HD 6290" } , SubSystem { subvendor : 5963u16 , subdevice : 25424u16 , name : "Radeon HD 6350" } , SubSystem { subvendor : 5963u16 , subdevice : 29520u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 33616u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 5963u16 , subdevice : 57639u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5963u16 , subdevice : 57669u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 5963u16 , subdevice : 57683u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 5470" } , SubSystem { subvendor : 6023u16 , subdevice : 12289u16 , name : "Radeon HD 5530" } , SubSystem { subvendor : 6023u16 , subdevice : 12290u16 , name : "Radeon HD 5490" } , SubSystem { subvendor : 6058u16 , subdevice : 13826u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 6058u16 , subdevice : 13827u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 6058u16 , subdevice : 13839u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 6058u16 , subdevice : 13849u16 , name : "Radeon HD 5450" } , SubSystem { subvendor : 6063u16 , subdevice : 12288u16 , name : "Radeon HD 6250" } , SubSystem { subvendor : 6063u16 , subdevice : 12289u16 , name : "Radeon HD 6230" } , SubSystem { subvendor : 6063u16 , subdevice : 12290u16 , name : "Radeon HD 6290" } , SubSystem { subvendor : 6063u16 , subdevice : 12305u16 , name : "Radeon HD 5470" } , SubSystem { subvendor : 6063u16 , subdevice : 12306u16 , name : "Radeon HD 5490" } , SubSystem { subvendor : 6063u16 , subdevice : 12307u16 , name : "Radeon HD 5470" } , SubSystem { subvendor : 6063u16 , subdevice : 12308u16 , name : "Radeon HD 6350" }] } , Device { vendor_id : 4098u16 , id : 26874u16 , name : "Cedar [Radeon HD 7350/8350 / R5 220]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 25u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 4121u16 , subdevice : 33u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 4121u16 , subdevice : 34u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 4121u16 , subdevice : 38u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 4156u16 , subdevice : 10975u16 , name : "Radeon HD 7350A" } , SubSystem { subvendor : 4156u16 , subdevice : 10984u16 , name : "Radeon HD 7350A" } , SubSystem { subvendor : 4163u16 , subdevice : 33616u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 5218u16 , subdevice : 8488u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5218u16 , subdevice : 8580u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5218u16 , subdevice : 8582u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5218u16 , subdevice : 9365u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5218u16 , subdevice : 46224u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5698u16 , subdevice : 14725u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 13584u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 5963u16 , subdevice : 13601u16 , name : "Radeon R5 220" } , SubSystem { subvendor : 5963u16 , subdevice : 13602u16 , name : "Radeon R5 220" } , SubSystem { subvendor : 5963u16 , subdevice : 29520u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 33107u16 , name : "Radeon HD 8350" } , SubSystem { subvendor : 5963u16 , subdevice : 57639u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 57683u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 5963u16 , subdevice : 57728u16 , name : "Radeon HD 7350" } , SubSystem { subvendor : 6063u16 , subdevice : 12309u16 , name : "Radeon HD 7350" }] } , Device { vendor_id : 4098u16 , id : 26878u16 , name : "Cedar LE" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26880u16 , name : "Topaz XT [Radeon R7 M260/M265 / M340/M360 / M440/M445 / 530/535 / 620/625 Mobile]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 4182u16 , name : "Radeon R7 M360 / R8 M365DX" } , SubSystem { subvendor : 4136u16 , subdevice : 1600u16 , name : "Radeon R7 M260/M265" } , SubSystem { subvendor : 4136u16 , subdevice : 1603u16 , name : "Radeon R7 M260/M265" } , SubSystem { subvendor : 4136u16 , subdevice : 1663u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4136u16 , subdevice : 1895u16 , name : "Radeon R7 M445" } , SubSystem { subvendor : 4136u16 , subdevice : 2064u16 , name : "Radeon 530" } , SubSystem { subvendor : 4136u16 , subdevice : 4874u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 8803u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 8809u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 8902u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 8904u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 11077u16 , name : "Radeon R7 A360" } , SubSystem { subvendor : 4156u16 , subdevice : 32908u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4156u16 , subdevice : 32921u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 4156u16 , subdevice : 32949u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 4156u16 , subdevice : 32953u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 4156u16 , subdevice : 33052u16 , name : "Radeon R7 M340" } , SubSystem { subvendor : 4156u16 , subdevice : 33318u16 , name : "Radeon R7 M440" } , SubSystem { subvendor : 4303u16 , subdevice : 6406u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4464u16 , subdevice : 39289u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 4473u16 , subdevice : 63747u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4473u16 , subdevice : 63778u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4473u16 , subdevice : 63779u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 4473u16 , subdevice : 63796u16 , name : "Radeon R7 M260" } , SubSystem { subvendor : 6058u16 , subdevice : 14370u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 6058u16 , subdevice : 14372u16 , name : "Radeon R7 M360" } , SubSystem { subvendor : 6058u16 , subdevice : 20513u16 , name : "Radeon R7 M260" }] } , Device { vendor_id : 4098u16 , id : 26881u16 , name : "Topaz PRO [Radeon R5 M255]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4888u16 , name : "Radeon R6 M255DX" }] } , Device { vendor_id : 4098u16 , id : 26887u16 , name : "Meso XT [Radeon R5 M315]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26912u16 , name : "Amethyst [Radeon R9 M395/ M395X Mac Edition]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26913u16 , name : "Amethyst XT [Radeon R9 M295X / M390X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26921u16 , name : "Tonga XT GL [FirePro S7150]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26923u16 , name : "Tonga PRO GL [FirePro W7100]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15659u16 , name : "MXRT-7600" }] } , Device { vendor_id : 4098u16 , id : 26927u16 , name : "Tonga XTV GL [FirePro S7150V]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26928u16 , name : "Tonga PRO [Radeon R9 380 4GB]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26936u16 , name : "Tonga XT / Amethyst XT [Radeon R9 380X / R9 M295X]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1269u16 , name : "Radeon R9 380X" } , SubSystem { subvendor : 4163u16 , subdevice : 1271u16 , name : "Radeon R9 380X" } , SubSystem { subvendor : 4203u16 , subdevice : 314u16 , name : "Radeon R9 M295X Mac Edition" } , SubSystem { subvendor : 5208u16 , subdevice : 8904u16 , name : "Radeon R9 380X" } , SubSystem { subvendor : 5260u16 , subdevice : 9040u16 , name : "Radeon R9 380X" } , SubSystem { subvendor : 5762u16 , subdevice : 37765u16 , name : "Radeon R9 380X" } , SubSystem { subvendor : 5963u16 , subdevice : 58120u16 , name : "Radeon R9 380X Nitro 4G D5" } , SubSystem { subvendor : 6063u16 , subdevice : 8198u16 , name : "Radeon R9 380X" }] } , Device { vendor_id : 4098u16 , id : 26937u16 , name : "Tonga PRO [Radeon R9 285/380]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 8213u16 , name : "Radeon R9 380 Gaming 4G" } , SubSystem { subvendor : 5260u16 , subdevice : 37760u16 , name : "Radeon R9 380" } , SubSystem { subvendor : 5963u16 , subdevice : 58120u16 , name : "Radeon R9 380 Nitro 4G D5" } , SubSystem { subvendor : 5963u16 , subdevice : 58133u16 , name : "Radeon R9 285" }] } , Device { vendor_id : 4098u16 , id : 26939u16 , name : "Tonga PRO GL [FirePro W7100 / Barco MXRT-7600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26956u16 , name : "Polaris 22 XT [Radeon RX Vega M GH]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26958u16 , name : "Polaris 22 XL [Radeon RX Vega M GL]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 26959u16 , name : "Polaris 22 MGL XL [Radeon Pro WX Vega M GL]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27008u16 , name : "Polaris12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27009u16 , name : "Lexa XT [Radeon PRO WX 3200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27013u16 , name : "Lexa XT [Radeon PRO WX 3100]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 33717u16 , name : "Radeon PRO WX 3100" }] } , Device { vendor_id : 4098u16 , id : 27014u16 , name : "Polaris12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27015u16 , name : "Lexa [Radeon 540X/550X/630 / RX 640 / E9171 MCM]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27023u16 , name : "Lexa XT [Radeon PRO WX 3100 / Barco MXRT 4700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27029u16 , name : "Lexa XT [Radeon PRO WX 2100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27039u16 , name : "Lexa PRO [Radeon 540/540X/550/550X / RX 540X/550/550X]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 5920u16 , name : "Radeon RX 550X" } , SubSystem { subvendor : 5208u16 , subdevice : 8947u16 , name : "Radeon RX 550 2GB" } , SubSystem { subvendor : 5260u16 , subdevice : 9088u16 , name : "Lexa XL [Radeon RX 550]" } , SubSystem { subvendor : 6058u16 , subdevice : 20585u16 , name : "Thinkpad E480/E580" } , SubSystem { subvendor : 7586u16 , subdevice : 58215u16 , name : "Lexa PRO [Radeon RX 550]" }] } , Device { vendor_id : 4098u16 , id : 27040u16 , name : "Vega 12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27041u16 , name : "Vega 12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27042u16 , name : "Vega 12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27043u16 , name : "Vega 12" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 27055u16 , name : "Vega 12 [Radeon Pro Vega 20]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28639u16 , name : "Polaris 20 XL [Radeon RX 580 2048SP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28687u16 , name : "RS100 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28688u16 , name : "RS200/RS250 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28928u16 , name : "R520 [Radeon X1800 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28929u16 , name : "R520/M58 [Mobility Radeon X1800 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28930u16 , name : "R520/M58 [Mobility Radeon X1800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28932u16 , name : "R520 GL [FireGL V7200 / Barco MXTR-5100]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15626u16 , name : "MXRT-5100" }] } , Device { vendor_id : 4098u16 , id : 28937u16 , name : "R520 [Radeon X1800 XL]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 802u16 , name : "All-in-Wonder X1800XL" } , SubSystem { subvendor : 4098u16 , subdevice : 3330u16 , name : "Radeon X1800 CrossFire Edition" }] } , Device { vendor_id : 4098u16 , id : 28938u16 , name : "R520 [Radeon X1800 GTO]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2834u16 , name : "Radeon X1800 GTO²" }] } , Device { vendor_id : 4098u16 , id : 28939u16 , name : "R520 [Radeon X1800 GTO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28942u16 , name : "R520 GL [FireGL V7300]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15628u16 , name : "MXRT-5150" }] } , Device { vendor_id : 4098u16 , id : 28943u16 , name : "R520 GL [FireGL V7350]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15630u16 , name : "MXRT-7100" }] } , Device { vendor_id : 4098u16 , id : 28960u16 , name : "R520 [Radeon X1800] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28964u16 , name : "R520 GL [FireGL V7200] (Secondary)" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15627u16 , name : "MXRT-5100 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 28969u16 , name : "R520 [Radeon X1800] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 803u16 , name : "All-In-Wonder X1800 XL (Secondary)" } , SubSystem { subvendor : 4098u16 , subdevice : 3331u16 , name : "Radeon X1800 CrossFire Edition (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 28974u16 , name : "R520 GL [FireGL V7300] (Secondary)" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15629u16 , name : "MXRT-5150 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 28975u16 , name : "R520 GL [FireGL V7350] (Secondary)" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15631u16 , name : "MXRT-7100 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 28992u16 , name : "RV515 [Radeon X1300/X1550/X1600 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28994u16 , name : "RV515 PRO [Radeon X1300/X1550 Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 802u16 , name : "All-in-Wonder 2006 PCI-E Edition" } , SubSystem { subvendor : 4163u16 , subdevice : 322u16 , name : "EAX1300PRO/TD/256M" }] } , Device { vendor_id : 4098u16 , id : 28995u16 , name : "RV505 [Radeon X1300/X1550 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 28997u16 , name : "RV515/M54 [Mobility Radeon X1400]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8198u16 , name : "Thinkpad T60 model 2007" }] } , Device { vendor_id : 4098u16 , id : 28998u16 , name : "RV515 [Radeon X1300/X1550]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 802u16 , name : "All-in-Wonder 2006 PCI-E Edition" } , SubSystem { subvendor : 5445u16 , subdevice : 6550u16 , name : "Radeon X1300 512MB PCI-e" }] } , Device { vendor_id : 4098u16 , id : 28999u16 , name : "RV505 [Radeon X1550 64-bit]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29001u16 , name : "RV515/M52 [Mobility Radeon X1300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29002u16 , name : "RV515/M52 [Mobility Radeon X1300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29010u16 , name : "RV515 GL [FireGL V3300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29011u16 , name : "RV515 GL [FireGL V3350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29023u16 , name : "RV505 CE [Radeon X1550 64-bit]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29026u16 , name : "RV515 PRO [Radeon X1300/X1550 Series] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 803u16 , name : "All-in-Wonder 2006 PCI-E Edition (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29027u16 , name : "RV505 [Radeon X1550 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29030u16 , name : "RV515 [Radeon X1300/X1550 Series] (Secondary)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 803u16 , name : "All-in-Wonder 2006 PCI-E Edition (Secondary)" } , SubSystem { subvendor : 5445u16 , subdevice : 6551u16 , name : "Radeon X1300 512MB PCI-e (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29031u16 , name : "RV515 [Radeon X1550 64-bit] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29042u16 , name : "RV515 GL [FireGL V3300] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29043u16 , name : "RV515 GL [FireGL V3350] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29057u16 , name : "RV516 [Radeon X1600/X1650 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29059u16 , name : "RV516 [Radeon X1300/X1550 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29062u16 , name : "RV516/M64 [Mobility Radeon X1450]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29063u16 , name : "RV516 [Radeon X1300/X1550 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29064u16 , name : "RV516/M64-S [Mobility Radeon X2300]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "6910p" }] } , Device { vendor_id : 4098u16 , id : 29066u16 , name : "RV516/M64 [Mobility Radeon X2300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29067u16 , name : "RV516/M62 [Mobility Radeon X1350]" , subsystems : & [SubSystem { subvendor : 4209u16 , subdevice : 33289u16 , name : "Medion MIM 2240 Notebook PC [MD98100]" }] } , Device { vendor_id : 4098u16 , id : 29068u16 , name : "RV516/M62-CSP64 [Mobility Radeon X1350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29069u16 , name : "RV516/M64-CSP128 [Mobility Radeon X1450]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29075u16 , name : "RV516 [Radeon X1550 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29078u16 , name : "RV516/M62-S [Mobility Radeon X1350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29083u16 , name : "RV516 GL [FireMV 2250]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15634u16 , name : "MXRT-1150" } , SubSystem { subvendor : 5068u16 , subdevice : 15636u16 , name : "MXRT-2150" }] } , Device { vendor_id : 4098u16 , id : 29087u16 , name : "RV516 [Radeon X1550 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29088u16 , name : "RV516 [Radeon X1300/X1550 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29089u16 , name : "RV516 [Radeon X1600/X1650 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29091u16 , name : "RV516 [Radeon X1300/X1550 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29095u16 , name : "RV516 [Radeon X1300/X1550 Series] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29115u16 , name : "RV516 GL [FireMV 2250] (Secondary)" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15635u16 , name : "MXRT-1150 (Secondary)" } , SubSystem { subvendor : 5068u16 , subdevice : 15637u16 , name : "MXRT-2150 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29120u16 , name : "RV530 [Radeon X1600 XT/X1650 GTO]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 57696u16 , name : "Radeon X1650 GTO" } , SubSystem { subvendor : 5963u16 , subdevice : 57696u16 , name : "Radeon X1650 GTO" }] } , Device { vendor_id : 4098u16 , id : 29121u16 , name : "RV535 [Radeon X1650 PRO]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 2176u16 , name : "Radeon X1700 FSC" }] } , Device { vendor_id : 4098u16 , id : 29122u16 , name : "RV530 [Radeon X1600 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29123u16 , name : "RV530 [Radeon X1600 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29124u16 , name : "RV530/M56 GL [Mobility FireGL V5200]" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8199u16 , name : "ThinkPad T60p" }] } , Device { vendor_id : 4098u16 , id : 29125u16 , name : "RV530/M56-P [Mobility Radeon X1600]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 12447u16 , name : "Compaq nx9420 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 12451u16 , name : "Compaq NW8440" } , SubSystem { subvendor : 4163u16 , subdevice : 4274u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4203u16 , subdevice : 128u16 , name : "MacBook Pro" }] } , Device { vendor_id : 4098u16 , id : 29126u16 , name : "RV530LE [Radeon X1600/X1650 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29127u16 , name : "RV535 [Radeon X1650 PRO]" , subsystems : & [SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "PowerColor X1650 PRO AGP" }] } , Device { vendor_id : 4098u16 , id : 29134u16 , name : "RV530 [Radeon X1300 XT/X1600 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29138u16 , name : "RV530 GL [FireGL V3400]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15624u16 , name : "MXRT-2100" }] } , Device { vendor_id : 4098u16 , id : 29140u16 , name : "RV530/M66 GL [Mobility FireGL V5250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29141u16 , name : "RV530/M66-P [Mobility Radeon X1700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29142u16 , name : "RV530/M66-XT [Mobility Radeon X1700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29150u16 , name : "RV530/M66 [Mobility Radeon X1700/X2500]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29152u16 , name : "RV530 [Radeon X1600] (Secondary)" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 57697u16 , name : "Radeon X1600 GTO (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29153u16 , name : "RV535 [Radeon X1650 PRO] (Secondary)" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 2177u16 , name : "Radeon X1700 FSC (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29154u16 , name : "RV530 [Radeon X1600] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29158u16 , name : "RV530 [Radeon X1650] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29159u16 , name : "RV535 [Radeon X1650 PRO] (Secondary)" , subsystems : & [SubSystem { subvendor : 6023u16 , subdevice : 12289u16 , name : "Radeon X1650 PRO AGP" }] } , Device { vendor_id : 4098u16 , id : 29170u16 , name : "RV530 GL [FireGL V3400] (Secondary)" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15625u16 , name : "MXRT-2100 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29200u16 , name : "RV550/M71 [Mobility Radeon HD 2300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29201u16 , name : "RV550/M71 [Mobility Radeon X2300 HD]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29248u16 , name : "R580+ [Radeon X1950 XTX]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3330u16 , name : "Radeon X1950 CrossFire Edition" }] } , Device { vendor_id : 4098u16 , id : 29252u16 , name : "R580+ [Radeon X1950 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29256u16 , name : "R580 [Radeon X1950]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29257u16 , name : "R580 [Radeon X1900 XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 1042u16 , name : "All-In-Wonder X1900" } , SubSystem { subvendor : 4098u16 , subdevice : 2834u16 , name : "Radeon X1900 XT/XTX" } , SubSystem { subvendor : 4098u16 , subdevice : 3330u16 , name : "Radeon X1900 CrossFire Edition" } , SubSystem { subvendor : 4163u16 , subdevice : 352u16 , name : "Radeon X1900 XTX 512 MB GDDR3" }] } , Device { vendor_id : 4098u16 , id : 29259u16 , name : "R580 [Radeon X1900 GT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2834u16 , name : "Radeon X1900 (Primary)" } , SubSystem { subvendor : 4098u16 , subdevice : 2835u16 , name : "Radeon X1900 (Secondary)" }] } , Device { vendor_id : 4098u16 , id : 29262u16 , name : "R580 GL [FireGL V7350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29289u16 , name : "R580 [Radeon X1900 XT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29291u16 , name : "R580 [Radeon X1900 GT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29294u16 , name : "R580 [AMD Stream Processor] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29312u16 , name : "RV570 [Radeon X1950 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29320u16 , name : "RV570 [Radeon X1950 GT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29329u16 , name : "RV560 [Radeon X1650 XT]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 2064u16 , name : "Radeon X1700 SE" }] } , Device { vendor_id : 4098u16 , id : 29331u16 , name : "RV560 [Radeon X1650 GT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29344u16 , name : "RV570 [Radeon X1950 PRO] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29352u16 , name : "RV570 [Radeon X1950 GT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29361u16 , name : "RV560 [Radeon X1650 XT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29363u16 , name : "RV560 [Radeon X1650 GT] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29440u16 , name : "Fiji [Radeon R9 FURY / NANO Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2869u16 , name : "FirePro S9300 X2" } , SubSystem { subvendor : 4098u16 , subdevice : 2870u16 , name : "Radeon R9 FURY X / NANO" } , SubSystem { subvendor : 4098u16 , subdevice : 6966u16 , name : "Radeon Pro Duo" } , SubSystem { subvendor : 4163u16 , subdevice : 1182u16 , name : "Radeon R9 FURY" } , SubSystem { subvendor : 4163u16 , subdevice : 1184u16 , name : "Radeon R9 FURY X" } , SubSystem { subvendor : 5963u16 , subdevice : 58153u16 , name : "Radeon R9 FURY" }] } , Device { vendor_id : 4098u16 , id : 29456u16 , name : "Navi 10 [Radeon Pro W5700X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29458u16 , name : "Navi 10 [Radeon Pro W5700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29460u16 , name : "Navi 10 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29465u16 , name : "Navi 10 [Radeon Pro 5700 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29467u16 , name : "Navi 10 [Radeon Pro 5700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29470u16 , name : "TDC-150" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29471u16 , name : "Navi 10 [Radeon RX 5600 OEM/5600 XT / 5700/5700 XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 2870u16 , name : "Reference RX 5700 XT" } , SubSystem { subvendor : 5208u16 , subdevice : 8979u16 , name : "Radeon RX 5700 XT Gaming OC" } , SubSystem { subvendor : 5208u16 , subdevice : 8989u16 , name : "Radeon RX 5600 XT/REV 2.0 [Windforce 6GB OC]" } , SubSystem { subvendor : 5218u16 , subdevice : 14366u16 , name : "RX 5600 XT MECH OC" } , SubSystem { subvendor : 5260u16 , subdevice : 9112u16 , name : "AXRX 5700 XT 8GBD6-3DHE/OC [PowerColor Red Devil Radeon RX 5700 XT]" } , SubSystem { subvendor : 5762u16 , subdevice : 22273u16 , name : "RX 5700 XT RAW II" } , SubSystem { subvendor : 6217u16 , subdevice : 20738u16 , name : "RX5700 CLD 8GO [ASRock Challenger D RX 5700 OC]" } , SubSystem { subvendor : 6217u16 , subdevice : 20768u16 , name : "Radeon RX 5600 XT" } , SubSystem { subvendor : 7586u16 , subdevice : 58377u16 , name : "Sapphire Technology Limited Navi 10 [Radeon RX 5600 OEM/5600 XT / 5700/5700 XT]" } , SubSystem { subvendor : 7586u16 , subdevice : 58384u16 , name : "Sapphire NITRO+ RX 5700 XT" } , SubSystem { subvendor : 7586u16 , subdevice : 58385u16 , name : "Navi 10 [Radeon RX 5600 OEM/5600 XT / 5700/5700 XT]Navi 10 [Radeon RX 5600 OEM/5600 XT / 5700/5700 XT]" }] } , Device { vendor_id : 4098u16 , id : 29504u16 , name : "Navi 14 [Radeon RX 5500/5500M / Pro 5300/5300M/5500M]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 528u16 , name : "MacBookPro16,1 (16\", 2019) [Radeon Pro 5300M]" } , SubSystem { subvendor : 4203u16 , subdevice : 537u16 , name : "iMac (Retina 5K, 27-inch, 2020) [Radeon Pro 5300]" }] } , Device { vendor_id : 4098u16 , id : 29505u16 , name : "Navi 14 [Radeon Pro W5500]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29511u16 , name : "Navi 14 [Radeon Pro W5500M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29519u16 , name : "Navi 14 [Radeon Pro W5300M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29536u16 , name : "Navi 12 [Radeon Pro 5600M/V520/BC-160]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29538u16 , name : "Navi 12 [Radeon Pro V520/V540]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29576u16 , name : "Arcturus GL-XL" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29580u16 , name : "Arcturus GL-XL [Instinct MI100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29582u16 , name : "Arcturus GL-XL [Instinct MI100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29601u16 , name : "Navi 21 [Radeon Pro V620]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29602u16 , name : "Navi 21 Pro-XTA [Radeon Pro W6900X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29603u16 , name : "Navi 21 GL-XL [Radeon PRO W6800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29604u16 , name : "Navi 21 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29605u16 , name : "Navi 21 [Radeon RX 6950 XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3642u16 , name : "Radeon RX 6950 XT" } , SubSystem { subvendor : 6217u16 , subdevice : 21040u16 , name : "Navi 21 [ASRock OC Forumla Radeon RX 6950XT]" } , SubSystem { subvendor : 6217u16 , subdevice : 21048u16 , name : "Navi 21 [ASRock Radeon RX 6950 XT Phantom Gaming OC]" } , SubSystem { subvendor : 7586u16 , subdevice : 17437u16 , name : "Navi 21 [Sapphire Nitro+ Radeon RX 6950 XT]" } , SubSystem { subvendor : 7854u16 , subdevice : 26960u16 , name : "Navi 21 [XFX Speedster MERC319 Radeon RX 6950 XT]" }] } , Device { vendor_id : 4098u16 , id : 29611u16 , name : "Navi 21 Pro-XLA [Radeon Pro W6800X/Radeon Pro W6800X Duo]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29614u16 , name : "Navi 21 [Radeon Pro V620 MxGPU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29615u16 , name : "Navi 21 [Radeon RX 6900 XT]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 9236u16 , name : "Navi 21 XTXH [PowerColor Red Devil RX 6900 XT Ultimate]" }] } , Device { vendor_id : 4098u16 , id : 29631u16 , name : "Navi 21 [Radeon RX 6800/6800 XT / 6900 XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3642u16 , name : "Radeon RX 6900 XT" } , SubSystem { subvendor : 5260u16 , subdevice : 9224u16 , name : "Red Devil AMD Radeon RX 6900 XT" } , SubSystem { subvendor : 7586u16 , subdevice : 17294u16 , name : "NITRO+ Radeon RX 6800 XT" } , SubSystem { subvendor : 7586u16 , subdevice : 17423u16 , name : "TOXIC RX 6900 XT" } , SubSystem { subvendor : 7586u16 , subdevice : 58423u16 , name : "Pulse Radeon RX 6800" } , SubSystem { subvendor : 7854u16 , subdevice : 26369u16 , name : "Speedster MERC 319 AMD Radeon RX 6800 XT Black" } , SubSystem { subvendor : 7854u16 , subdevice : 26881u16 , name : "Speedster MERC 319 AMD Radeon RX 6900 XT Black" }] } , Device { vendor_id : 4098u16 , id : 29635u16 , name : "Navi 22" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29636u16 , name : "Navi 22 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29646u16 , name : "Navi 22-XL SRIOV MxGPU" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29663u16 , name : "Navi 22 [Radeon RX 6700/6700 XT/6750 XT / 6800M/6850M XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3638u16 , name : "Radeon RX 6700 XT 12G" } , SubSystem { subvendor : 4163u16 , subdevice : 5826u16 , name : "Radeon RX 6800M" } , SubSystem { subvendor : 5208u16 , subdevice : 9224u16 , name : "Radeon RX 6750 XT GAMING OC 12G" } , SubSystem { subvendor : 5218u16 , subdevice : 14720u16 , name : "Radeon RX 6700 XT Mech 2X 12G [MSI]" } , SubSystem { subvendor : 5260u16 , subdevice : 9225u16 , name : "Red Devil RX 6700 XT" } , SubSystem { subvendor : 6217u16 , subdevice : 21008u16 , name : "Radeon RX 6700 XT Challenger D" } , SubSystem { subvendor : 6217u16 , subdevice : 21017u16 , name : "Radeon RX 6700 XT Challenger D" } , SubSystem { subvendor : 6217u16 , subdevice : 21026u16 , name : "RX 6700 XT Challenger D OC" } , SubSystem { subvendor : 7586u16 , subdevice : 17502u16 , name : "Radeon RX 6700 XT GAMING OC 12G [Sapphire PULSE]" } , SubSystem { subvendor : 7586u16 , subdevice : 18014u16 , name : "Radeon RX 6750 XT PULSE OC" } , SubSystem { subvendor : 7586u16 , subdevice : 58437u16 , name : "Sapphire Radeon RX 6700" } , SubSystem { subvendor : 7854u16 , subdevice : 26113u16 , name : "Speedster QICK 319 RX 6700 XT" } , SubSystem { subvendor : 7854u16 , subdevice : 26138u16 , name : "Radeon RX 6700 [SPEEDSTER SWFT 309]" }] } , Device { vendor_id : 4098u16 , id : 29664u16 , name : "Navi 23" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29665u16 , name : "Navi 23 WKS-XM [Radeon PRO W6600M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29667u16 , name : "Navi 23 WKS-XL [Radeon PRO W6600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29668u16 , name : "Navi 23 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29679u16 , name : "Navi 23 [Radeon RX 6650 XT / 6700S / 6800S]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 9221u16 , name : "Navi 23 [Radeon RX 6650 XT]" } , SubSystem { subvendor : 5218u16 , subdevice : 20519u16 , name : "RX 6650XT MECH 2X" } , SubSystem { subvendor : 6217u16 , subdevice : 21046u16 , name : "RX 6650 XT Challenger D OC" }] } , Device { vendor_id : 4098u16 , id : 29680u16 , name : "Navi 33 [Radeon RX 7600M XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29695u16 , name : "Navi 23 [Radeon RX 6600/6600 XT/6600M]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 20513u16 , name : "MSI RX 6600XT MECH 2X" } , SubSystem { subvendor : 5218u16 , subdevice : 20514u16 , name : "RX 6600 MECH 2X" } , SubSystem { subvendor : 5260u16 , subdevice : 9234u16 , name : "PowerColor Red Devil RX 6600 XT" } , SubSystem { subvendor : 6217u16 , subdevice : 21016u16 , name : "Radeon RX 6600 Challenger ITX 8GB" } , SubSystem { subvendor : 7586u16 , subdevice : 17550u16 , name : "Radeon RX 6600 XT Pulse" }] } , Device { vendor_id : 4098u16 , id : 29704u16 , name : "Aldebaran/MI200 [Instinct MI250X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29708u16 , name : "Aldebaran/MI200 [Instinct MI250X/MI250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29711u16 , name : "Aldebaran/MI200 [Instinct MI210]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29729u16 , name : "Navi 24 [Radeon PRO W6500M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29730u16 , name : "Navi 24 [Radeon PRO W6400]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29731u16 , name : "Navi 24 [Radeon PRO W6300/W6300M]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29732u16 , name : "Navi 24 [Radeon RX 6300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29759u16 , name : "Navi 24 [Radeon RX 6400/6500 XT/6500M]" , subsystems : & [SubSystem { subvendor : 7586u16 , subdevice : 58455u16 , name : "PULSE AMD Radeon RX 6500 XT" }] } , Device { vendor_id : 4098u16 , id : 29766u16 , name : "Navi 31 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29768u16 , name : "Navi 31 [Radeon Pro W7900]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29769u16 , name : "Navi 31 [Radeon Pro W7800 48GB]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29770u16 , name : "Navi 31 [Radeon Pro W7900 Dual Slot]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29771u16 , name : "Navi 31 [Radeon Pro W7900D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29772u16 , name : "Navi 31 [Radeon RX 7900 XT/7900 XTX/7900 GRE/7900M]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 3643u16 , name : "RX 7900 XTX / RX 7900 GRE [XFX]" } , SubSystem { subvendor : 4163u16 , subdevice : 1286u16 , name : "TUF Gaming Radeon RX 7900 XTX OC" } , SubSystem { subvendor : 5260u16 , subdevice : 9253u16 , name : "HELLHOUND RX 7900 GRE" } , SubSystem { subvendor : 6217u16 , subdevice : 21252u16 , name : "Radeon RX 7900 XTX" } , SubSystem { subvendor : 7586u16 , subdevice : 18206u16 , name : "PULSE RX 7900 XTX" } , SubSystem { subvendor : 7586u16 , subdevice : 18270u16 , name : "PULSE RX 7900 GRE" } , SubSystem { subvendor : 7586u16 , subdevice : 58481u16 , name : "NITRO+ RX 7900 XTX Vapor-X" } , SubSystem { subvendor : 7854u16 , subdevice : 30977u16 , name : "RX-79XMERCB9 [SPEEDSTER MERC 310 RX 7900 XTX]" } , SubSystem { subvendor : 7854u16 , subdevice : 30986u16 , name : "RX-79GMERCBR [XFX RX 7900 GRE]" }] } , Device { vendor_id : 4098u16 , id : 29790u16 , name : "Navi 31 [Radeon Pro W7800]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29792u16 , name : "Navi 32 GL-XL [AMD Radeon PRO V710]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29793u16 , name : "Navi 32 [AMD Radeon PRO V710]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29808u16 , name : "Navi 32 [Radeon PRO W7700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29822u16 , name : "Navi 32 [Radeon RX 7700 XT / 7800 XT]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 32294u16 , name : "Radeon RX 7800 XT [Nitro+]" } , SubSystem { subvendor : 5260u16 , subdevice : 9255u16 , name : "RX 7800 XT [Hellhound / Red Devil]" }] } , Device { vendor_id : 4098u16 , id : 29824u16 , name : "Navi 33 [Radeon RX 7600/7600 XT/7600M XT/7600S/7700S / PRO W7600]" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 21267u16 , name : "RX 7600 Challenger OC" }] } , Device { vendor_id : 4098u16 , id : 29825u16 , name : "Navi 33 [Steam Machine]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29827u16 , name : "Navi 33 [Radeon RX 7600M/7600M XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29831u16 , name : "Navi 33 [Radeon Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29833u16 , name : "Navi 33 [Radeon Pro W7500]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29835u16 , name : "Navi 33 [Radeon Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29849u16 , name : "Navi 33 [Radeon RX 7400/7300/Pro W7400]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29855u16 , name : "Navi 33 [Radeon RX 7500]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29856u16 , name : "Aqua Vanjaram [Instinct MI300A]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29857u16 , name : "Aqua Vanjaram [Instinct MI300X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29858u16 , name : "Aqua Vanjaram [Instinct MI308X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29861u16 , name : "Aqua Vanjaram [Instinct MI325X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29865u16 , name : "Aqua Vanjaram [Instinct MI300X HF]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29877u16 , name : "Aqua Vanjaram [Instinct MI300X VF]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29881u16 , name : "Aqua Vanjaram [Instinct MI325X VF]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 29885u16 , name : "Aqua Vanjaram [Instinct MI300X HF]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30032u16 , name : "Navi 48 [Radeon RX 9070/9070 XT/9070 GRE]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 9271u16 , name : "Navi 48 XTX [Radeon RX 9070 XT Gaming OC ICE 16G]" } , SubSystem { subvendor : 5260u16 , subdevice : 9269u16 , name : "Radeon RX 9070 XT 16GB" } , SubSystem { subvendor : 6217u16 , subdevice : 21507u16 , name : "Navi 48 XTX [Steel Legend Radeon RX 9070 XT]" } , SubSystem { subvendor : 7586u16 , subdevice : 58512u16 , name : "Navi 48 XTX [Sapphire Pulse Radeon RX 9070 XT]" }] } , Device { vendor_id : 4098u16 , id : 30033u16 , name : "Navi 48 [Radeon AI PRO R9700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30096u16 , name : "Navi 44 [Radeon RX 9050 / 9060 XT]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 1593u16 , name : "Navi 44 [Radeon RX 9060 XT]" } , SubSystem { subvendor : 5208u16 , subdevice : 9257u16 , name : "GV-R9060XTGAMING OC-16GD [Radeon RX 9060 XT GAMING OC 16G]" } , SubSystem { subvendor : 7854u16 , subdevice : 34305u16 , name : "RX-96TS316W7 [SWIFT RX 9060 XT OC White Triple Fan Gaming Edition 16GB]" }] } , Device { vendor_id : 4098u16 , id : 30112u16 , name : "Aqua Vanjaram [Instinct MI350X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30115u16 , name : "Aqua Vanjaram [Instinct MI355X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30120u16 , name : "Aqua Vanjaram [Instinct MI350P]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30771u16 , name : "RS350 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30772u16 , name : "RS350 [Radeon 9100 PRO/XT IGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30773u16 , name : "RS350M [Mobility Radeon 9000 IGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30776u16 , name : "RS350 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30992u16 , name : "RS690 Host Bridge" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 30993u16 , name : "RS690/RS740 Host Bridge" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 30992u16 , name : "RS690/RS740 Host Bridge" }] } , Device { vendor_id : 4098u16 , id : 30994u16 , name : "RS690/RS740 PCI to PCI Bridge (Internal gfx)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30995u16 , name : "RS690 PCI to PCI Bridge (PCI Express Graphics Port 0)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30997u16 , name : "RS690 PCI to PCI Bridge (PCI Express Port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30998u16 , name : "RS690 PCI to PCI Bridge (PCI Express Port 2)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 30999u16 , name : "RS690 PCI to PCI Bridge (PCI Express Port 3)" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 30992u16 , name : "RS690 PCI to PCI Bridge" }] } , Device { vendor_id : 4098u16 , id : 31001u16 , name : "RS690 HDMI Audio [Radeon Xpress 1200 Series]" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 31001u16 , name : "Satellite P305D-S8995E" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 31006u16 , name : "RS690 [Radeon X1200]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29479u16 , name : "K9AG Neo2" } , SubSystem { subvendor : 6130u16 , subdevice : 20480u16 , name : "KI690-AM2 Motherboard" }] } , Device { vendor_id : 4098u16 , id : 31007u16 , name : "RS690M [Radeon Xpress 1200/1250/1270]" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 65360u16 , name : "Satellite P305D-S8995E" }] } , Device { vendor_id : 4098u16 , id : 31024u16 , name : "RS600 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31026u16 , name : "RS600 PCI to PCI Bridge (Internal gfx)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31027u16 , name : "RS600 PCI to PCI Bridge (PCI Express Graphics Port 0)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31029u16 , name : "RS600 PCI to PCI Bridge (PCI Express Port 1)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31030u16 , name : "RS600 PCI to PCI Bridge (PCI Express Port 2)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31031u16 , name : "RS690 PCI to PCI Bridge (PCI Express Port 3)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31035u16 , name : "RS600 HDMI Audio [Radeon Xpress 1250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31039u16 , name : "RS690M [Radeon Xpress 1200/1250/1270] (Secondary)" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31041u16 , name : "RS600 [Radeon Xpress 1250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31042u16 , name : "RS600M [Radeon Xpress 1250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 31086u16 , name : "RS740 [Radeon 2100]" , subsystems : & [SubSystem { subvendor : 4187u16 , subdevice : 3603u16 , name : "N15235/A74MX mainboard" }] } , Device { vendor_id : 4098u16 , id : 37888u16 , name : "R600 [Radeon HD 2900 PRO/XT]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 9554u16 , name : "Radeon HD 2900 XT" } , SubSystem { subvendor : 4098u16 , subdevice : 12288u16 , name : "Radeon HD 2900 PRO" } , SubSystem { subvendor : 4098u16 , subdevice : 12610u16 , name : "HIS Radeon HD 2900XT 512MB GDDR3 VIVO PCIe" }] } , Device { vendor_id : 4098u16 , id : 37889u16 , name : "R600 [Radeon HD 2900 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37891u16 , name : "R600 [Radeon HD 2900 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37893u16 , name : "R600 [Radeon HD 2900 GT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37898u16 , name : "R600 GL [FireGL V8650]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15638u16 , name : "MXRT-7200" }] } , Device { vendor_id : 4098u16 , id : 37899u16 , name : "R600 GL [FireGL V8600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37903u16 , name : "R600 GL [FireGL V7600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37952u16 , name : "RV770 [Radeon HD 4870]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37953u16 , name : "R700 [Radeon HD 4870 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37954u16 , name : "RV770 [Radeon HD 4850]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 1282u16 , name : "MSI Radeon HD 4850 512MB GDDR3" } , SubSystem { subvendor : 5963u16 , subdevice : 59408u16 , name : "Radeon HD 4850 512MB GDDR3" }] } , Device { vendor_id : 4098u16 , id : 37955u16 , name : "R700 [Radeon HD 4850 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37956u16 , name : "RV770 GL [FirePro V8750]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37958u16 , name : "RV770 GL [FirePro V7760]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37962u16 , name : "RV770/M98L [Mobility Radeon HD 4850]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37963u16 , name : "RV770/M98 [Mobility Radeon HD 4850 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37964u16 , name : "RV770 LE [Radeon HD 4830]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37966u16 , name : "RV770 CE [Radeon HD 4710]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 12897u16 , name : "Radeon HD 4810" }] } , Device { vendor_id : 4098u16 , id : 37968u16 , name : "RV770 GL [FireStream 9270]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37970u16 , name : "RV770 GL [FireStream 9250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37974u16 , name : "RV770 GL [FirePro V8700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37978u16 , name : "RV770/M98-XT [Mobility Radeon HD 4870]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37984u16 , name : "RV790 [Radeon HD 4890]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37986u16 , name : "RV790 [Radeon HD 4860]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 37994u16 , name : "RV770 GL [FirePro M7750]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38016u16 , name : "RV730/M96 [Mobility Radeon HD 4650/5165]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "Mobility Radeon HD 4650 [dv6-1190en]" }] } , Device { vendor_id : 4098u16 , id : 38024u16 , name : "RV730/M96-XT [Mobility Radeon HD 4670]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38025u16 , name : "RV730/M96 GL [Mobility FireGL V5725]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38032u16 , name : "RV730 XT [Radeon HD 4670]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 59520u16 , name : "Radeon HD 4670 512MB GDDR3 Dual DVI-I/TVO" }] } , Device { vendor_id : 4098u16 , id : 38033u16 , name : "RV730/M96-CSP [Radeon E4690]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38037u16 , name : "RV730 [Radeon HD 4600 AGP Series]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 40u16 , name : "Radeon HD 4650/4670 AGP" } , SubSystem { subvendor : 4242u16 , subdevice : 40u16 , name : "Radeon HD 4670 AGP 512MB DDR2" } , SubSystem { subvendor : 5208u16 , subdevice : 40u16 , name : "Radeon HD 4650 AGP" } , SubSystem { subvendor : 5762u16 , subdevice : 40u16 , name : "Radeon HD 4650 AGP" } , SubSystem { subvendor : 5963u16 , subdevice : 40u16 , name : "Radeon HD 4650 AGP DDR2" }] } , Device { vendor_id : 4098u16 , id : 38040u16 , name : "RV730 PRO [Radeon HD 4650]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38044u16 , name : "RV730 GL [FirePro V7750]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15643u16 , name : "MXRT-7300" }] } , Device { vendor_id : 4098u16 , id : 38046u16 , name : "RV730 GL [FirePro V5700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38047u16 , name : "RV730 GL [FirePro V3750]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38048u16 , name : "RV740/M97 [Mobility Radeon HD 4830]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38049u16 , name : "RV740/M97-XT [Mobility Radeon HD 4860]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38051u16 , name : "RV740/M97 GL [FirePro M7740]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38067u16 , name : "RV740 PRO [Radeon HD 4770]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38068u16 , name : "RV740 PRO [Radeon HD 4750]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38081u16 , name : "RV610 [Radeon HD 2400 PRO/XT]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 529u16 , name : "Optiplex 755" } , SubSystem { subvendor : 4136u16 , subdevice : 3330u16 , name : "Optiplex 755" }] } , Device { vendor_id : 4098u16 , id : 38083u16 , name : "RV610 [Radeon HD 2400 PRO]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 770u16 , name : "Radeon HD 2400 Pro" } , SubSystem { subvendor : 5963u16 , subdevice : 58368u16 , name : "Radeon HD 2400 PRO" } , SubSystem { subvendor : 6332u16 , subdevice : 13648u16 , name : "Radeon HD 2400 PRO" }] } , Device { vendor_id : 4098u16 , id : 38084u16 , name : "RV610 LE [Radeon HD 2400 PRO AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38085u16 , name : "RV610 [Radeon HD 2400 LE]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38087u16 , name : "RV610 [Radeon HD 2350]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38088u16 , name : "RV610/M74 [Mobility Radeon HD 2400 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38089u16 , name : "RV610/M72-S [Mobility Radeon HD 2400]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 38089u16 , name : "Radeon HD2400" }] } , Device { vendor_id : 4098u16 , id : 38091u16 , name : "RV610 [Radeon E2400]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38092u16 , name : "RV610 LE [Radeon HD 2400 PRO PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38144u16 , name : "RV670 [Radeon HD 3850 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38145u16 , name : "RV670 [Radeon HD 3870]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 58912u16 , name : "Radeon HD 3870" } , SubSystem { subvendor : 6023u16 , subdevice : 8772u16 , name : "Radeon HD 3870" }] } , Device { vendor_id : 4098u16 , id : 38148u16 , name : "RV670/M88 [Mobility Radeon HD 3850]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38149u16 , name : "RV670 [Radeon HD 3690/3850]" , subsystems : & [SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Radeon HD 3850" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Radeon HD 3690/3850" } , SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 3690" }] } , Device { vendor_id : 4098u16 , id : 38150u16 , name : "RV670/M88 [Mobility Radeon HD 3850 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38151u16 , name : "RV670 [Radeon HD 3830]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38152u16 , name : "RV670/M88-XT [Mobility Radeon HD 3870]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38153u16 , name : "RV670/M88 [Mobility Radeon HD 3870 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38159u16 , name : "R680 [Radeon HD 3870 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38161u16 , name : "RV670 GL [FireGL V7700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38163u16 , name : "RV670 [Radeon HD 3850 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38165u16 , name : "RV670 PRO [Radeon HD 3850 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38169u16 , name : "RV670 GL [FireStream 9170]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38208u16 , name : "RV710 [Radeon HD 4550]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38223u16 , name : "RV710 [Radeon HD 4350/4550]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 5656u16 , name : "R4350 MD512H (MS-V161)" }] } , Device { vendor_id : 4098u16 , id : 38226u16 , name : "RV710/M92 [Mobility Radeon HD 4330/4350/4550]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 4355u16 , name : "M92 [Mobility Radeon HD 4330]" } , SubSystem { subvendor : 5208u16 , subdevice : 8620u16 , name : "Radeon HD 4350" } , SubSystem { subvendor : 5208u16 , subdevice : 8685u16 , name : "Radeon HD 4550" } , SubSystem { subvendor : 5260u16 , subdevice : 12288u16 , name : "Radeon HD 4350 Go! Green 512MB GDDR3" } , SubSystem { subvendor : 5963u16 , subdevice : 12288u16 , name : "Radeon HD 4350/4550 HyperMemory DDR2" }] } , Device { vendor_id : 4098u16 , id : 38227u16 , name : "RV710/M92 [Mobility Radeon HD 4530/4570/5145/530v/540v/545v]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 350u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4133u16 , subdevice : 381u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4133u16 , subdevice : 517u16 , name : "Mobility Radeon HD 4570 / 545v" } , SubSystem { subvendor : 4133u16 , subdevice : 518u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4133u16 , subdevice : 567u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4136u16 , subdevice : 702u16 , name : "Mobility Radeon HD 4570 / 545v" } , SubSystem { subvendor : 4136u16 , subdevice : 744u16 , name : "Mobility Radeon HD 4530" } , SubSystem { subvendor : 4156u16 , subdevice : 5180u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 4156u16 , subdevice : 5190u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 4156u16 , subdevice : 13860u16 , name : "Mobility Radeon HD 4530" } , SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "Mobility Radeon HD 4530" } , SubSystem { subvendor : 4156u16 , subdevice : 13878u16 , name : "Mobility Radeon HD 4530" } , SubSystem { subvendor : 4163u16 , subdevice : 6962u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4163u16 , subdevice : 6978u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4173u16 , subdevice : 36950u16 , name : "Mobility Radeon HD 4570" } , SubSystem { subvendor : 4473u16 , subdevice : 65410u16 , name : "Satellite L505-13T GPU (Mobility Radeon HD 5145)" } , SubSystem { subvendor : 5197u16 , subdevice : 49279u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 5197u16 , subdevice : 50545u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 5218u16 , subdevice : 4102u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 6058u16 , subdevice : 8489u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 6058u16 , subdevice : 8539u16 , name : "Mobility Radeon HD 545v" } , SubSystem { subvendor : 6058u16 , subdevice : 8635u16 , name : "Mobility Radeon HD 545v" }] } , Device { vendor_id : 4098u16 , id : 38229u16 , name : "RV711/M93 [Mobility Radeon HD 4350/4550/530v/540v/545v / FirePro RG220]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5137u16 , name : "ProBook 4720s GPU (Mobility Radeon HD 4350)" }] } , Device { vendor_id : 4098u16 , id : 38231u16 , name : "RV711/M93 GL [FirePro RG220]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38239u16 , name : "RV710/M92 [Mobility Radeon HD 4330]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38272u16 , name : "RV630 [Radeon HD 2600 PRO]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38273u16 , name : "RV630/M76 [Mobility Radeon HD 2600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38275u16 , name : "RV630/M76 [Mobility Radeon HD 2600 XT/2700]" , subsystems : & [SubSystem { subvendor : 4203u16 , subdevice : 131u16 , name : "iMac 7,1" } , SubSystem { subvendor : 5940u16 , subdevice : 4359u16 , name : "Mobility Radeon HD 2700" }] } , Device { vendor_id : 4098u16 , id : 38278u16 , name : "RV630 XT [Radeon HD 2600 XT AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38279u16 , name : "RV630 PRO [Radeon HD 2600 PRO AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38280u16 , name : "RV630 XT [Radeon HD 2600 XT]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 8556u16 , name : "Radeon HD 2600 XT, 256MB GDDR3, 2x DVI, TV-out, PCIe (GV-RX26T256H)" }] } , Device { vendor_id : 4098u16 , id : 38281u16 , name : "RV630 PRO [Radeon HD 2600 PRO]" , subsystems : & [SubSystem { subvendor : 6023u16 , subdevice : 12288u16 , name : "Radeon HD 3610" }] } , Device { vendor_id : 4098u16 , id : 38282u16 , name : "RV630 [Radeon HD 2600 X2]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38283u16 , name : "RV630/M76 [Mobility Radeon HD 2600 XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38284u16 , name : "RV630 GL [FireGL V5600]" , subsystems : & [SubSystem { subvendor : 5068u16 , subdevice : 15640u16 , name : "MXRT-5200" }] } , Device { vendor_id : 4098u16 , id : 38285u16 , name : "RV630 GL [FireGL V3600]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38289u16 , name : "RV635/M86 [Mobility Radeon HD 3650]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 38289u16 , name : "Mobility Radeon HD 3650" }] } , Device { vendor_id : 4098u16 , id : 38291u16 , name : "RV635/M86 [Mobility Radeon HD 3670]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38293u16 , name : "RV635/M86 GL [Mobility FireGL V5700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38294u16 , name : "RV635 PRO [Radeon HD 3650 AGP]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 40u16 , name : "EAH3650 SILENT/HTDI/512M/A" }] } , Device { vendor_id : 4098u16 , id : 38295u16 , name : "RV635 PRO [Radeon HD 3650 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38296u16 , name : "RV635 [Radeon HD 3650/3750/4570/4580]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 38296u16 , name : "Mobility Radeon HD 3600" } , SubSystem { subvendor : 4163u16 , subdevice : 470u16 , name : "EAH3650 Silent" } , SubSystem { subvendor : 4163u16 , subdevice : 12289u16 , name : "Radeon HD 4570" } , SubSystem { subvendor : 5963u16 , subdevice : 12289u16 , name : "Radeon HD 3750" } , SubSystem { subvendor : 5963u16 , subdevice : 17792u16 , name : "RV635 PRO [Radeon HD 4580]" } , SubSystem { subvendor : 6063u16 , subdevice : 12305u16 , name : "RV635 PRO [Radeon HD 4580]" }] } , Device { vendor_id : 4098u16 , id : 38297u16 , name : "RV635 PRO [Radeon HD 3650 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38336u16 , name : "RV620 PRO [Radeon HD 3470]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 38336u16 , name : "Mobility Radeon HD 3470" } , SubSystem { subvendor : 4136u16 , subdevice : 12867u16 , name : "C120D" }] } , Device { vendor_id : 4098u16 , id : 38338u16 , name : "RV620/M82 [Mobility Radeon HD 3410/3430]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38340u16 , name : "RV620/M82 [Mobility Radeon HD 3450/3470]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 38340u16 , name : "Mobility Radeon HD 3400" }] } , Device { vendor_id : 4098u16 , id : 38341u16 , name : "RV620 LE [Radeon HD 3450]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 834u16 , name : "Radeon HD 3450 DMS-59" }] } , Device { vendor_id : 4098u16 , id : 38342u16 , name : "RV620 LE [Radeon HD 3450 AGP]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38345u16 , name : "RV620 LE [Radeon HD 3450 PCI]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38348u16 , name : "RV620 GL [FirePro V3700]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38349u16 , name : "RV620 GL [FirePro 2450]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38351u16 , name : "RV620 GL [FirePro 2260]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38415u16 , name : "RS780 HDMI Audio [Radeon 3000/3100 / HD 3200/3300]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 30102u16 , name : "760GM-E51(MS-7596) Motherboard" }] } , Device { vendor_id : 4098u16 , id : 38416u16 , name : "RS780 [Radeon HD 3200]" , subsystems : & [SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "GA-MA78GM-S2H Motherboard" }] } , Device { vendor_id : 4098u16 , id : 38417u16 , name : "RS780C [Radeon 3100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38418u16 , name : "RS780M [Mobility Radeon HD 3200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38419u16 , name : "RS780MC [Mobility Radeon HD 3100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38420u16 , name : "RS780D [Radeon HD 3300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38421u16 , name : "RS780E [Radeon HD 3200]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38422u16 , name : "RS780L [Radeon 3000]" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 29953u16 , name : "760GM-E51(MS-7596) Motherboard" }] } , Device { vendor_id : 4098u16 , id : 38464u16 , name : "Sumo [Radeon HD 6550D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38465u16 , name : "Sumo [Radeon HD 6620G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38466u16 , name : "SuperSumo [Radeon HD 6370D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38467u16 , name : "SuperSumo [Radeon HD 6380G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38468u16 , name : "SuperSumo [Radeon HD 6410D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38469u16 , name : "SuperSumo [Radeon HD 6410D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38471u16 , name : "Sumo [Radeon HD 6520G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38472u16 , name : "Sumo [Radeon HD 6480G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38473u16 , name : "SuperSumo [Radeon HD 6480G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38474u16 , name : "Sumo [Radeon HD 6530D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38475u16 , name : "Sumo" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38476u16 , name : "Sumo" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38478u16 , name : "Sumo" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38479u16 , name : "Sumo" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38671u16 , name : "RS880 HDMI Audio [Radeon HD 4200 Series]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4163u16 , subdevice : 33698u16 , name : "M4A785TD Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33854u16 , name : "M5A88-V EVO" }] } , Device { vendor_id : 4098u16 , id : 38672u16 , name : "RS880 [Radeon HD 4200]" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4163u16 , subdevice : 33698u16 , name : "M4A785TD Motherboard" }] } , Device { vendor_id : 4098u16 , id : 38674u16 , name : "RS880M [Mobility Radeon HD 4225/4250]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" }] } , Device { vendor_id : 4098u16 , id : 38675u16 , name : "RS880M [Mobility Radeon HD 4100]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38676u16 , name : "RS880 [Radeon HD 4290]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38677u16 , name : "RS880 [Radeon HD 4250]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33854u16 , name : "M5A88-V EVO" }] } , Device { vendor_id : 4098u16 , id : 38914u16 , name : "Wrestler [Radeon HD 6310]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4098u16 , id : 38915u16 , name : "Wrestler [Radeon HD 6310]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38916u16 , name : "Wrestler [Radeon HD 6250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38917u16 , name : "Wrestler [Radeon HD 6250]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38918u16 , name : "Wrestler [Radeon HD 6320]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38919u16 , name : "Wrestler [Radeon HD 6290]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38920u16 , name : "Wrestler [Radeon HD 7340]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38921u16 , name : "Wrestler [Radeon HD 7310]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38922u16 , name : "Wrestler [Radeon HD 7290]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38960u16 , name : "Kabini [Radeon HD 8400 / R3 Series]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" }] } , Device { vendor_id : 4098u16 , id : 38961u16 , name : "Kabini [Radeon HD 8400E]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38962u16 , name : "Kabini [Radeon HD 8330]" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 38962u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4098u16 , id : 38963u16 , name : "Kabini [Radeon HD 8330E]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38964u16 , name : "Kabini [Radeon HD 8210]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38965u16 , name : "Kabini [Radeon HD 8310E]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38966u16 , name : "Kabini [Radeon HD 8280 / R3 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38967u16 , name : "Kabini [Radeon HD 8280E]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38968u16 , name : "Kabini [Radeon HD 8240 / R3 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38969u16 , name : "Kabini [Radeon HD 8180]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38973u16 , name : "Temash [Radeon HD 8250/8280G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38976u16 , name : "Kabini HDMI/DP Audio" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 38976u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4098u16 , id : 38992u16 , name : "Mullins [Radeon R2/R3 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38993u16 , name : "Mullins [Radeon R4/R5 Graphics]" , subsystems : & [SubSystem { subvendor : 4473u16 , subdevice : 63784u16 , name : "Beema [Radeon R5 Graphics]" }] } , Device { vendor_id : 4098u16 , id : 38994u16 , name : "Mullins [Radeon R2 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38995u16 , name : "Mullins [Radeon R2 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38996u16 , name : "Mullins [Radeon R3E Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38997u16 , name : "Mullins [Radeon R6 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38998u16 , name : "Mullins [Radeon R1E/R2E Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 38999u16 , name : "Mullins [Radeon APU XX-2200M with R2 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39000u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39001u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39002u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39003u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39004u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39005u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39006u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39007u16 , name : "Mullins" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39028u16 , name : "Wani [Radeon R5/R6/R7 Graphics]" , subsystems : & [SubSystem { subvendor : 4098u16 , subdevice : 6257u16 , name : "Radeon R5 Graphics" } , SubSystem { subvendor : 4098u16 , subdevice : 7712u16 , name : "Radeon R7 Graphics" } , SubSystem { subvendor : 4136u16 , subdevice : 1725u16 , name : "Radeon R6 Graphics" } , SubSystem { subvendor : 4156u16 , subdevice : 11076u16 , name : "Radeon R6 Graphics" } , SubSystem { subvendor : 4156u16 , subdevice : 33313u16 , name : "Radeon R5 Graphics" } , SubSystem { subvendor : 4156u16 , subdevice : 33315u16 , name : "Radeon R5 Graphics" } , SubSystem { subvendor : 4156u16 , subdevice : 33336u16 , name : "Radeon R7 Graphics" } , SubSystem { subvendor : 4156u16 , subdevice : 33619u16 , name : "Radeon R7 Graphics" } , SubSystem { subvendor : 5208u16 , subdevice : 53248u16 , name : "Radeon R7 Graphics" } , SubSystem { subvendor : 6058u16 , subdevice : 20755u16 , name : "Radeon R6 Graphics" } , SubSystem { subvendor : 6058u16 , subdevice : 20758u16 , name : "Radeon R6 Graphics" } , SubSystem { subvendor : 6058u16 , subdevice : 20760u16 , name : "Radeon R5 Graphics" }] } , Device { vendor_id : 4098u16 , id : 39056u16 , name : "Amur" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39104u16 , name : "Nolan" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39140u16 , name : "Stoney [Radeon R2/R3/R4/R5 Graphics]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39168u16 , name : "Trinity [Radeon HD 7660G]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" }] } , Device { vendor_id : 4098u16 , id : 39169u16 , name : "Trinity [Radeon HD 7660D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39170u16 , name : "Trinity HDMI Audio Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" }] } , Device { vendor_id : 4098u16 , id : 39171u16 , name : "Trinity [Radeon HD 7640G]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6482u16 , name : "ProBook 455 G1 Notebook" }] } , Device { vendor_id : 4098u16 , id : 39172u16 , name : "Trinity [Radeon HD 7560D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39173u16 , name : "Trinity GL [FirePro A300]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39174u16 , name : "Trinity GL [FirePro A320]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39175u16 , name : "Trinity [Radeon HD 7620G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39176u16 , name : "Trinity [Radeon HD 7600G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39177u16 , name : "Trinity [Radeon HD 7500G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39178u16 , name : "Trinity [Radeon HD 7500G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39179u16 , name : "Richland [Radeon HD 8650G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39180u16 , name : "Richland [Radeon HD 8670D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39181u16 , name : "Richland [Radeon HD 8550G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39182u16 , name : "Richland [Radeon HD 8570D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39183u16 , name : "Richland [Radeon HD 8610G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39184u16 , name : "Trinity [Radeon HD 7660G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39187u16 , name : "Trinity [Radeon HD 7640G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39191u16 , name : "Trinity [Radeon HD 7620G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39192u16 , name : "Trinity [Radeon HD 7600G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39193u16 , name : "Trinity [Radeon HD 7500G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39198u16 , name : "Bishop [Xbox One S APU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39200u16 , name : "Liverpool [Playstation 4 APU]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39201u16 , name : "Liverpool HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39202u16 , name : "Starshp" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39203u16 , name : "Starsha2 [Kingston/Clayton]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39204u16 , name : "Gladius" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39205u16 , name : "Kingston/Clayton/Jupiter/Gladius/Montego HDMI Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39206u16 , name : "Jupiter" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39312u16 , name : "Trinity 2 [Radeon HD 7520G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39313u16 , name : "Trinity 2 [Radeon HD 7540D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39314u16 , name : "Trinity 2 [Radeon HD 7420G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39315u16 , name : "Trinity 2 [Radeon HD 7480D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39316u16 , name : "Trinity 2 [Radeon HD 7400G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39317u16 , name : "Richland [Radeon HD 8450G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39318u16 , name : "Richland [Radeon HD 8470D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39319u16 , name : "Richland [Radeon HD 8350G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39320u16 , name : "Richland [Radeon HD 8370D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39321u16 , name : "Richland [Radeon HD 8510G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39322u16 , name : "Richland [Radeon HD 8410G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39323u16 , name : "Richland [Radeon HD 8310G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39324u16 , name : "Richland [Radeon HD 8650D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39325u16 , name : "Richland [Radeon HD 8550D]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39328u16 , name : "Trinity 2 [Radeon HD 7520G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39330u16 , name : "Trinity 2 [Radeon HD 7420G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 39332u16 , name : "Trinity 2 [Radeon HD 7400G]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43520u16 , name : "R600 HDMI Audio [Radeon HD 2900 GT/PRO/XT]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43528u16 , name : "RV630 HDMI Audio [Radeon HD 2600 PRO/XT / HD 3610]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43536u16 , name : "RV610 HDMI Audio [Radeon HD 2350 PRO / 2400 PRO/XT / HD 3410]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 43536u16 , name : "Radeon HD 2400 PRO" } , SubSystem { subvendor : 6332u16 , subdevice : 43536u16 , name : "Radeon HD 2400 PRO" }] } , Device { vendor_id : 4098u16 , id : 43544u16 , name : "RV670/680 HDMI Audio [Radeon HD 3690/3800 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43552u16 , name : "RV635 HDMI Audio [Radeon HD 3650/3730/3750]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43560u16 , name : "RV620 HDMI Audio [Radeon HD 3450/3470/3550/3570]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43568u16 , name : "RV770 HDMI Audio [Radeon HD 4850/4870]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 43568u16 , name : "Radeon HD 4850 512MB GDDR3 PCI-E Dual Slot Fansink" }] } , Device { vendor_id : 4098u16 , id : 43576u16 , name : "RV710/730 HDMI Audio [Radeon HD 4000 series]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13864u16 , name : "dv6-1190en" }] } , Device { vendor_id : 4098u16 , id : 43600u16 , name : "Cypress HDMI Audio [Radeon HD 5830/5850/5870 / 6850/6870 Rebrand]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43608u16 , name : "Juniper HDMI Audio [Radeon HD 5700 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43616u16 , name : "Redwood HDMI Audio [Radeon HD 5000 Series]" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 829u16 , name : "Mobility Radeon HD 5650" } , SubSystem { subvendor : 4133u16 , subdevice : 839u16 , name : "Aspire 7740G" }] } , Device { vendor_id : 4098u16 , id : 43624u16 , name : "Cedar HDMI Audio [Radeon HD 5400/6300/7300 Series]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 43624u16 , name : "XPS 8300" }] } , Device { vendor_id : 4098u16 , id : 43648u16 , name : "Cayman/Antilles HDMI Audio [Radeon HD 6930/6950/6970/6990]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43656u16 , name : "Barts HDMI Audio [Radeon HD 6790/6850/6870 / 7720 OEM]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43664u16 , name : "Turks HDMI Audio [Radeon HD 6500/6600 / 6700M Series]" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1187u16 , name : "Precision M4600" }] } , Device { vendor_id : 4098u16 , id : 43672u16 , name : "Caicos HDMI Audio [Radeon HD 6450 / 7450/8450/8490 OEM / R5 230/235/235X OEM]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 43672u16 , name : "EAH6450 SILENT/DI/1GD3(LP)" } , SubSystem { subvendor : 5963u16 , subdevice : 43672u16 , name : "Radeon HD 6450 1GB DDR3" }] } , Device { vendor_id : 4098u16 , id : 43680u16 , name : "Tahiti HDMI Audio [Radeon HD 7870 XT / 7950/7970]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43696u16 , name : "Oland/Hainan/Cape Verde/Pitcairn HDMI Audio [Radeon HD 7000 Series]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43704u16 , name : "Tiran HDMI Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43712u16 , name : "Tobago HDMI Audio [Radeon R7 360 / R9 360 OEM]" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 43712u16 , name : "R7260X-DC2OC-2GD5" }] } , Device { vendor_id : 4098u16 , id : 43720u16 , name : "Hawaii HDMI Audio [Radeon R9 290/290X / 390/390X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43736u16 , name : "Tonga HDMI Audio [Radeon R9 285/380]" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 43736u16 , name : "Radeon R9 285/380 HDMI Audio" }] } , Device { vendor_id : 4098u16 , id : 43744u16 , name : "Baffin HDMI/DP Audio [Radeon RX 550 640SP / RX 560/560X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43752u16 , name : "Fiji HDMI/DP Audio [Radeon R9 Nano / FURY/FURY X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43760u16 , name : "Ellesmere HDMI Audio [Radeon RX 470/480 / 570/580/590]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43768u16 , name : "Vega 10 HDMI Audio [Radeon Vega 56/64]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43776u16 , name : "Baffin HDMI/DP Audio [Radeon RX 550 640SP / RX 560/560X]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43784u16 , name : "Polaris 22 HDMI Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43792u16 , name : "Lexa HDMI Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43800u16 , name : "Vega 12 HDMI Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43808u16 , name : "Vega 20 HDMI Audio [Radeon VII]" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43816u16 , name : "Navi 21/23 HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43824u16 , name : "Navi 31 HDMI/DP Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43832u16 , name : "Navi 10 HDMI Audio" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 43840u16 , name : "Navi 48 HDMI/DP Audio Controller" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44032u16 , name : "Theater 506 World-Wide Analog Decoder" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44033u16 , name : "Theater 506 World-Wide Analog Decoder" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44034u16 , name : "TV Wonder HD 600 PCIe" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44035u16 , name : "Theater 506 PCIe" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44036u16 , name : "Theater 506 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44037u16 , name : "Theater 506 USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44038u16 , name : "Theater 506 External USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44039u16 , name : "Theater 506 External USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44040u16 , name : "Theater 506A World-Wide Analog Decoder + Demodulator" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44041u16 , name : "Theater 506A World-Wide Analog Decoder + Demodulator" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44042u16 , name : "Theater 506A PCIe" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44043u16 , name : "Theater 506A PCIe" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44044u16 , name : "Theater 506A USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44045u16 , name : "Theater 506A USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44046u16 , name : "Theater 506A External USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44047u16 , name : "Theater 506A External USB" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 44050u16 , name : "Theater HD T507 (DVB-T) TV tuner/capture device" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 51888u16 , name : "RS100 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 51890u16 , name : "RS200 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 51891u16 , name : "RS250 Host Bridge" , subsystems : & [] } , Device { vendor_id : 4098u16 , id : 52146u16 , name : "RS200 Host Bridge" , subsystems : & [] }] }),
        (5284, Vendor { id : 5284u16 , name : "Lite-On Technology Corporation" , devices : & [Device { vendor_id : 5284u16 , id : 8448u16 , name : "CA1-8D128 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 8689u16 , name : "Plextor NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 8704u16 , name : "CX2-8B256, CX2-8B512 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 8864u16 , name : "EP2-KB960 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 8945u16 , name : "M8Pe Series NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 8960u16 , name : "CA3-8D256, CA3-8D512 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 9201u16 , name : "M9PeG, M9PeGN, M9PeY NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 12032u16 , name : "CAZ-82512 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 13568u16 , name : "CA5-8D512 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 13809u16 , name : "PLEXTOR PX-256M9PGN+ NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 17176u16 , name : "Broadcom BCM4318 [AirForce One 54g] 802.11g WLAN Controller" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 20736u16 , name : "CB1-SD256, CB1-SD512 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5284u16 , id : 37120u16 , name : "CL1-3D256, CL1-8D512 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (30212, Vendor { id : 30212u16 , name : "O.N. Electronic Co Ltd." , devices : & [] }),
        (5396, Vendor { id : 5396u16 , name : "TFL LAN Inc" , devices : & [] }),
        (7982, Vendor { id : 7982u16 , name : "AzurEngine Technologies" , devices : & [Device { vendor_id : 7982u16 , id : 161u16 , name : "[AE] AzureBlade Accelerator" , subsystems : & [] } , Device { vendor_id : 7982u16 , id : 177u16 , name : "[AE] AzureBlade Accelerator" , subsystems : & [] } , Device { vendor_id : 7982u16 , id : 193u16 , name : "[AE] AzureBlade Accelerator" , subsystems : & [] }] }),
        (4226, Vendor { id : 4226u16 , name : "EFA Corporation of America" , devices : & [] }),
        (5555, Vendor { id : 5555u16 , name : "Mellanox Technologies" , devices : & [Device { vendor_id : 5555u16 , id : 401u16 , name : "MT25408 [ConnectX IB Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 502u16 , name : "MT27500 Family [ConnectX-3 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 504u16 , name : "MT27520 Family [ConnectX-3 Pro Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 511u16 , name : "MT27600 Family [Connect-IB Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 521u16 , name : "MT27700 Family [ConnectX-4 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 523u16 , name : "MT27710 Family [ConnectX-4 Lx Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 525u16 , name : "MT28800 Family [ConnectX-5 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 527u16 , name : "MT28908A0 Family [ConnectX-6 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 528u16 , name : "MT28908A0 Family [ConnectX-6 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 529u16 , name : "MT416842 Family [BlueField SoC Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 530u16 , name : "MT2892 Family [ConnectX-6 Dx Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 531u16 , name : "MT2892 Family [ConnectX-6 Dx Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 532u16 , name : "MT42822 Family [BlueField-2 SoC Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 533u16 , name : "MT42822 Family [BlueField-2 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 534u16 , name : "MT2894 Family [ConnectX-6 Lx Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 535u16 , name : "MT2894 Family [ConnectX-6 Lx Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 536u16 , name : "MT2910 Family [ConnectX-7 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 537u16 , name : "MT2910 Family [ConnectX-7 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 538u16 , name : "MT43162 Family [BlueField-3 Lx SoC Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 539u16 , name : "MT43162 Family [BlueField-3 Lx Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 540u16 , name : "MT43244 Family [BlueField-3 SoC Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 541u16 , name : "MT43244 Family [BlueField-3 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 542u16 , name : "CX8 Family [ConnectX-8 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 543u16 , name : "CX8 Family [ConnectX-8 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 544u16 , name : "BF4 Family Flash Recovery [BlueField-4 SoC Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 545u16 , name : "BF4 Family Secure Flash Recovery [BlueField-4 Secure Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 546u16 , name : "CX8 PCIe Switch Family [ConnectX-8 PCIe Switch Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 547u16 , name : "CX8 PCIe Switch Family [ConnectX-8 PCIe Switch Secure Flash Recovery-RMA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 548u16 , name : "CX9 Family [ConnectX-9 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 549u16 , name : "CX9 Family [ConnectX-9 Secure Flash Recovery-RMA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 550u16 , name : "CX10 Family [ConnectX-10 Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 551u16 , name : "CX10 Family [ConnectX-10 RMA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 552u16 , name : "CX9 PCIe Switch Family [ConnectX-9 PCIe Switch Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 553u16 , name : "CX9 PCIe Switch Family [ConnectX-9 PCIe Switch Secure Flash Recovery-RMA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 590u16 , name : "MT53100 [Spectrum-2, Flash recovery mode]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 591u16 , name : "MT53100 [Spectrum-2, Secure Flash recovery mode]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 592u16 , name : "Spectrum-3, Flash recovery mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 593u16 , name : "Spectrum-3, Secure Flash recovery mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 594u16 , name : "Amos chiplet" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 595u16 , name : "Amos GearBox Manager" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 596u16 , name : "Spectrum-4, Flash recovery mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 597u16 , name : "Spectrum-4 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 598u16 , name : "Abir GearBox" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 599u16 , name : "Quantum-2 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 600u16 , name : "Quantum-2 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 601u16 , name : "Abir Chiplet" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 603u16 , name : "Quantum-3 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 604u16 , name : "Quantum-3 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 605u16 , name : "Quantum-3CPO in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 606u16 , name : "Quantum-3CPO RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 610u16 , name : "MT27710 [ConnectX-4 Lx Programmable] EN" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 611u16 , name : "MT27710 [ConnectX-4 Lx Programmable Virtual Function] EN" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 612u16 , name : "Innova-2 Flex Burn image" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 624u16 , name : "Spectrum-5 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 625u16 , name : "Spectrum-5 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 628u16 , name : "Spectrum-6 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 629u16 , name : "Spectrum-6 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 631u16 , name : "Spectrum-6 Tile" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 632u16 , name : "NVLink-6 Switch in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 633u16 , name : "NVLink-6 Switch RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 634u16 , name : "Eros Chiplet" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 636u16 , name : "NVLink-7 Switch in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 637u16 , name : "NVLink-7 Switch RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 638u16 , name : "Spectrum-7 Tile" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 639u16 , name : "Spectrum-8 tile" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 641u16 , name : "NPS-600 Flash Recovery" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 642u16 , name : "ArcusE Flash recovery" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 643u16 , name : "ArcusE RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 644u16 , name : "Sagitta" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 645u16 , name : "Sagitta RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 646u16 , name : "LibraE Flash Recovery" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 647u16 , name : "LibraE RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 648u16 , name : "Arcus2 Flash Recovery" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 649u16 , name : "Arcus2 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 656u16 , name : "SagittaZ" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 658u16 , name : "Arcus3 Flash Recovery" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 659u16 , name : "Arcus3 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 660u16 , name : "OPHY2.1 [SagittaZ]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 662u16 , name : "OPHY2.6 [Sagitta]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 664u16 , name : "OPHY3.0 [Sagitta]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 666u16 , name : "OPHY3.1 [Sagitta]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 668u16 , name : "OPHY3.5 [Sagitta]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 672u16 , name : "NVLink-8 Switch in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 673u16 , name : "NVLink-8 Switch RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 674u16 , name : "Spectrum-7 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 675u16 , name : "Spectrum-7 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 678u16 , name : "OrionR" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 679u16 , name : "OrionR RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 680u16 , name : "Spectrum-8 in Flash Recovery Mode" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 681u16 , name : "Spectrum-8 RMA" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4098u16 , name : "MT25400 Family [ConnectX-2 Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4099u16 , name : "MT27500 Family [ConnectX-3]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1205u16 , name : "PCIe3 40GbE RoCE Converged Host Bus Adapter for Power" } , SubSystem { subvendor : 4156u16 , subdevice : 6007u16 , name : "InfiniBand FDR/EN 10/40Gb Dual Port 544FLR-QSFP Adapter (Rev Cx)" } , SubSystem { subvendor : 4156u16 , subdevice : 6089u16 , name : "Infiniband QDR/Ethernet 10Gb 2-port 544i Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6350u16 , name : "InfiniBand QDR/EN 10Gb Dual Port 544M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6351u16 , name : "InfiniBand FDR/EN 10/40Gb Dual Port 544M Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6358u16 , name : "InfiniBand FDR/EN 10/40Gb Dual Port 544QSFP Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 37u16 , name : "ConnectX-3 IB QDR Dual Port Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 38u16 , name : "ConnectX-3 IB FDR Dual Port Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 40u16 , name : "ConnectX-3 VPI Dual QSFP+ Port QDR Infiniband 40Gb/s or 10Gb Ethernet" } , SubSystem { subvendor : 5555u16 , subdevice : 85u16 , name : "ConnectX-3 10 GbE Single Port SFP+ Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 89u16 , name : "ConnectX-3 VPI IB FDR/40 GbE Single Port QSFP+ Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 100u16 , name : "ConnectX-3 EN 10/40 GbE Single Port QSFP+ Adapter (MCX313A-BCBT)" } , SubSystem { subvendor : 5555u16 , subdevice : 101u16 , name : "ConnectX-3 VPI IB FDR/40 GbE Dual Port QSFP+ Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 102u16 , name : "ConnectX-3 IB FDR10 Dual Port Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 103u16 , name : "ConnectX-3 VPI IB FDR/40 GbE Single Port QSFP+ Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 113u16 , name : "ConnectX-3 VPI IB FDR/40 GbE Dual Port QSFP+ Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 120u16 , name : "ConnectX-3 10 GbE Dual Port KR Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 121u16 , name : "ConnectX-3 40 GbE Dual Port QSFP+ Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 128u16 , name : "ConnectX-3 10 GbE Dual Port SFP+ Adapter" }] } , Device { vendor_id : 5555u16 , id : 4100u16 , name : "MT27500/MT27520 Family [ConnectX-3/ConnectX-3 Pro Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4101u16 , name : "MT27510 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4102u16 , name : "MT27511 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4103u16 , name : "MT27520 Family [ConnectX-3 Pro]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1259u16 , name : "2-Port 10GbE NIC and RoCE SR PCIe3" } , SubSystem { subvendor : 4156u16 , subdevice : 8947u16 , name : "InfiniBand FDR/Ethernet 10Gb/40Gb 2-port 544+QSFP Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 8948u16 , name : "InfiniBand FDR/Ethernet 10Gb/40Gb 2-port 544+FLR-QSFP Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 32799u16 , name : "Ethernet 10G 2-port 546SFP+ Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 144u16 , name : "FastFrame NQ41" } , SubSystem { subvendor : 4476u16 , subdevice : 145u16 , name : "FastFrame NQ42" } , SubSystem { subvendor : 4476u16 , subdevice : 146u16 , name : "FastFrame NQ11" } , SubSystem { subvendor : 4476u16 , subdevice : 147u16 , name : "FastFrame NQ12" } , SubSystem { subvendor : 5555u16 , subdevice : 6u16 , name : "Mellanox Technologies ConnectX-3 Pro Stand-up dual-port 40GbE MCX314A-BCCT" } , SubSystem { subvendor : 5555u16 , subdevice : 12u16 , name : "ConnectX-3 Pro 10 GbE Dual Port SFP+ Adapter revision A1" } , SubSystem { subvendor : 5555u16 , subdevice : 120u16 , name : "ConnectX-3 Pro 10 GbE Dual Port KR Mezzanine Card" } , SubSystem { subvendor : 5555u16 , subdevice : 121u16 , name : "ConnectX-3 Pro 40 GbE Dual Port QSFP+ Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 128u16 , name : "ConnectX-3 Pro 10 GbE Dual Port SFP+ Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4098u16 , name : "520F-B" }] } , Device { vendor_id : 5555u16 , id : 4105u16 , name : "MT27530 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4106u16 , name : "MT27531 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4107u16 , name : "MT27540 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4108u16 , name : "MT27541 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4109u16 , name : "MT27550 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4110u16 , name : "MT27551 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4111u16 , name : "MT27560 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4112u16 , name : "MT27561 Family" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4113u16 , name : "MT27600 [Connect-IB]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4114u16 , name : "MT27600 Family [Connect-IB Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4115u16 , name : "MT27700 Family [ConnectX-4]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1271u16 , name : "PCIe3 2-port 100 GbE (NIC and RoCE) QSFP28 Adapter for Power" } , SubSystem { subvendor : 4476u16 , subdevice : 175u16 , name : "FastFrame N351 Single-port 50Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 176u16 , name : "FastFrame N352 Dual-port 50Gb Ethernet Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 3u16 , name : "Mellanox Technologies ConnectX-4 Stand-up single-port 40GbE MCX413A-BCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 5u16 , name : "Mellanox Technologies ConnectX-4 Stand-up single-port 40GbE MCX415A-BCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 6u16 , name : "MCX416A-BCAT, ConnectX-4 EN, 40/56GbE 2P, PCIe3.0 x16" } , SubSystem { subvendor : 5555u16 , subdevice : 7u16 , name : "ConnectX-4 EN network interface card, 40/56GbE dual-port QSFP28, PCIe3.0 x16, tall bracket" } , SubSystem { subvendor : 5555u16 , subdevice : 8u16 , name : "ConnectX-4 Stand-up dual-port 100GbE MCX416A-CCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 51u16 , name : "ConnectX-4 VPI IB EDR/100 GbE Single Port QSFP28 Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 52u16 , name : "ConnectX-4 VPI IB EDR/100 GbE Dual Port QSFP28 Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 80u16 , name : "ConnectX-4 100 GbE Dual Port QSFP28 Adapter" }] } , Device { vendor_id : 5555u16 , id : 4116u16 , name : "MT27700 Family [ConnectX-4 Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4117u16 , name : "MT27710 Family [ConnectX-4 Lx]" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 180u16 , name : "FastFrame N322 Dual-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16567u16 , name : "ThunderLink TLN3-3252 Dual-port 25Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 16568u16 , name : "ThunderLink TLN3-3102 Dual-port 10Gb Ethernet Adapter" } , SubSystem { subvendor : 5208u16 , subdevice : 640u16 , name : "CLN4C44" } , SubSystem { subvendor : 5208u16 , subdevice : 641u16 , name : "CLN4C44 with NCSI" } , SubSystem { subvendor : 5555u16 , subdevice : 1u16 , name : "ConnectX-4 Lx EN network interface card, 25GbE single-port SFP28, PCIe3.0 x8, tall bracket, ROHS R6" } , SubSystem { subvendor : 5555u16 , subdevice : 3u16 , name : "Stand-up ConnectX-4 Lx EN, 25GbE dual-port SFP28, PCIe3.0 x8, MCX4121A-ACAT" } , SubSystem { subvendor : 5555u16 , subdevice : 4u16 , name : "ConnectX-4 Lx Stand-up dual-port 10GbE MCX4121A-XCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 5u16 , name : "Mellanox Technologies ConnectX-4 Lx Stand-up single-port 40GbE MCX4131A-BCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 32u16 , name : "MCX4411A-ACQN, ConnectX-4 Lx EN OCP, 1x25Gb" } , SubSystem { subvendor : 5555u16 , subdevice : 33u16 , name : "MCX4421A-ACQN ConnectX-4 Lx EN OCP,2x25G" } , SubSystem { subvendor : 5555u16 , subdevice : 37u16 , name : "ConnectX-4 Lx 25 GbE Dual Port SFP28 rNDC" } , SubSystem { subvendor : 6461u16 , subdevice : 4106u16 , name : "620F-B" } , SubSystem { subvendor : 6461u16 , subdevice : 4131u16 , name : "NIC-ETH540F-LP-2P" } , SubSystem { subvendor : 6461u16 , subdevice : 4145u16 , name : "NIC-ETH640i-Mb-2x25G" } , SubSystem { subvendor : 6461u16 , subdevice : 4227u16 , name : "NIC-ETH640F-3S-2P" } , SubSystem { subvendor : 6461u16 , subdevice : 4228u16 , name : "NIC-ETH540F-3S-2P" } , SubSystem { subvendor : 7809u16 , subdevice : 3088u16 , name : "25GbE dual-port SFP28, PCIe3.0 x8 [3SC10]" } , SubSystem { subvendor : 7999u16 , subdevice : 3088u16 , name : "25GbE dual-port SFP28, PCIe3.0 x8, 3SC10" }] } , Device { vendor_id : 5555u16 , id : 4118u16 , name : "MT27710 Family [ConnectX-4 Lx Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4119u16 , name : "MT27800 Family [ConnectX-5]" , subsystems : & [SubSystem { subvendor : 4476u16 , subdevice : 177u16 , name : "FastFrame N311 Single-port 10Gb Ethernet Adapter" } , SubSystem { subvendor : 4476u16 , subdevice : 178u16 , name : "FastFrame N312 Dual-port 10Gb Ethernet Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 6u16 , name : "ConnectX-5 EN network interface card, 100GbE single-port QSFP28, PCIe3.0 x16, tall bracket; MCX515A-CCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 7u16 , name : "Mellanox ConnectX-5 MCX516A-CCAT" } , SubSystem { subvendor : 5555u16 , subdevice : 32u16 , name : "ConnectX-5 EN network interface card, 10/25GbE dual-port SFP28, PCIe3.0 x8, tall bracket ; MCX512A-ACAT" } , SubSystem { subvendor : 5555u16 , subdevice : 104u16 , name : "ConnectX-5 EN network interface card for OCP2.0, Type 1, with host management, 25GbE dual-port SFP28, PCIe3.0 x8, no bracket Halogen free ; MCX542B-ACAN" } , SubSystem { subvendor : 6461u16 , subdevice : 4177u16 , name : "NIC-IB1040i-Mb-2P" }] } , Device { vendor_id : 5555u16 , id : 4120u16 , name : "MT27800 Family [ConnectX-5 Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4121u16 , name : "MT28800 Family [ConnectX-5 Ex]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1559u16 , name : "PCIe4 x16 2-Port EDR IB-only ConnectX-5 CAPI Capable Adapter [IBM EC64]" } , SubSystem { subvendor : 4116u16 , subdevice : 1589u16 , name : "PCIe4 2-port 100 GbE RoCE x16 adapter [IBM EC66]" } , SubSystem { subvendor : 5555u16 , subdevice : 8u16 , name : "ConnectX-5 Ex EN network interface card, 100GbE dual-port QSFP28, PCIe4.0 x16, tall bracket; MCX516A-CDAT" } , SubSystem { subvendor : 5555u16 , subdevice : 293u16 , name : "Tencent ConnectX-5 EN Ex network interface card for OCP 3.0, with host management, 50GbE Dual-port QSFP28, PCIe4.0 x16, Thumbscrew (pull-tab) bracket" } , SubSystem { subvendor : 5555u16 , subdevice : 294u16 , name : "PCIe4 x16 2-port EDR 100GbE ConnectX-5 CAPI Capable adapter [IBM AJP1]" }] } , Device { vendor_id : 5555u16 , id : 4122u16 , name : "MT28800 Family [ConnectX-5 Ex Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4123u16 , name : "MT28908 Family [ConnectX-6]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4124u16 , name : "MT28908 Family [ConnectX-6 Virtual Function]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4125u16 , name : "MT2892 Family [ConnectX-6 Dx]" , subsystems : & [SubSystem { subvendor : 6461u16 , subdevice : 4181u16 , name : "NIC-ETH1040F-LP-2P QSFP56 2x100GbE PCIe Network Adapter" }] } , Device { vendor_id : 5555u16 , id : 4126u16 , name : "ConnectX Family mlx5Gen Virtual Function" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4127u16 , name : "MT2894 Family [ConnectX-6 Lx]" , subsystems : & [SubSystem { subvendor : 6461u16 , subdevice : 4149u16 , name : "NIC-ETH641F-LP-2P SFP28 2x25GbE PCIe Network Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 172u16 , name : "O252MCX6Lx" } , SubSystem { subvendor : 7124u16 , subdevice : 174u16 , name : "S252MCX6Lx" } , SubSystem { subvendor : 7999u16 , subdevice : 3089u16 , name : "25GbE dual-port SFP28, PCIe4.0 x8, 3SC1125GbE dual-port SFP28, PCIe4.0 x8, 3SC11" } , SubSystem { subvendor : 8185u16 , subdevice : 173u16 , name : "ENFM6251-SP2" } , SubSystem { subvendor : 8185u16 , subdevice : 175u16 , name : "ENPM6251-SP2" }] } , Device { vendor_id : 5555u16 , id : 4128u16 , name : "MT28860" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4129u16 , name : "MT2910 Family [ConnectX-7]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4131u16 , name : "CX8 Family [ConnectX-8]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4132u16 , name : "CX8 PCIe Switch Family [ConnectX-8 PCIe Switch]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4133u16 , name : "CX9 Family [ConnectX-9]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4135u16 , name : "CX10 Family [ConnectX-10]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 4136u16 , name : "CX10 Family [ConnectX-10 Trusted Network Control Memory]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6516u16 , name : "MT28800 Family [ConnectX-5 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6517u16 , name : "MT416842 Family [BlueField SoC PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6518u16 , name : "MT28908 Family [ConnectX-6 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6519u16 , name : "MT2892 Family [ConnectX-6 Dx PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6520u16 , name : "MT42822 Family [BlueField-2 SoC PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6521u16 , name : "MT2910 Family [ConnectX-7 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6522u16 , name : "MT43162 Family [BlueField-3 Lx SoC PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6523u16 , name : "MT43244 Family [BlueField-3 SoC PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6524u16 , name : "ConnectX/BlueField Family mlx5Gen PCIe Bridge [PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6525u16 , name : "CX8 Family [ConnectX-8 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6526u16 , name : "CX9 Family [ConnectX-9 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 6527u16 , name : "CX10 Family [ConnectX-10 PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8224u16 , name : "MT2892 Family [ConnectX-6 Dx Emulated PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8225u16 , name : "MT42822 Family [BlueField-2 SoC Emulated PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8227u16 , name : "MT2910 Family [ConnectX-7 Emulated PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8228u16 , name : "MT43244 Family [BlueField-3 SoC Emulated PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8229u16 , name : "ConnectX/BlueField Family mlx5Gen Emulated PCIe Bridge [Emulated PCIe Bridge]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8448u16 , name : "CX8 Family [CX8 Data Direct Interface]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8449u16 , name : "CX10 Family [ConnectX-10 NVLink-C2C]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 8960u16 , name : "ConnectX/BlueField Family Hardware Performance Monitors [HWPM]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 16663u16 , name : "MT27712A0-FDCF-AE" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 57u16 , name : "SN10XMP2P25" } , SubSystem { subvendor : 7124u16 , subdevice : 58u16 , name : "25G SFP28 SP EO251FM9 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 77u16 , name : "SN10XMP2P25,YZPC-01191-101" }] } , Device { vendor_id : 5555u16 , id : 21108u16 , name : "MT21108 InfiniBridge" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 23108u16 , name : "MT23108 InfiniHost" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 23109u16 , name : "MT23108 [Infinihost HCA Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 23110u16 , name : "MT23108 PCI Bridge" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 24204u16 , name : "MT24204 [InfiniHost III Lx HCA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 24205u16 , name : "MT25204 [InfiniHost III Lx HCA Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 24577u16 , name : "NVMe SNAP Controller" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25204u16 , name : "MT25204 [InfiniHost III Lx HCA]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25208u16 , name : "MT25208 InfiniHost III Ex (Tavor compatibility mode)" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25209u16 , name : "MT25208 [InfiniHost III Ex HCA Flash Recovery]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25218u16 , name : "MT25208 [InfiniHost III Ex]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25408u16 , name : "MT25408A0-FCC-SI ConnectX, Dual Port 10Gb/s InfiniBand / 10GigE Adapter IC with PCIe 2.0 x8 2.5GT/s Interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25418u16 , name : "MT25408A0-FCC-DI ConnectX, Dual Port 20Gb/s InfiniBand / 10GigE Adapter IC with PCIe 2.0 x8 2.5GT/s Interface" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 4116u16 , name : "4X InfiniBand DDR Expansion Card (CFFh) for IBM BladeCenter" }] } , Device { vendor_id : 5555u16 , id : 25448u16 , name : "MT25448 [ConnectX EN 10GigE, PCIe 2.0 2.5GT/s]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 25458u16 , name : "MT25458 ConnectX EN 10GBASE-T PCIe 2.5 GT/s" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 26418u16 , name : "MT25408A0-FCC-GI ConnectX, Dual Port 20Gb/s InfiniBand / 10GigE Adapter IC with PCIe 2.0 x8 5.0GT/s Interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 26428u16 , name : "MT25408A0-FCC-QI ConnectX, Dual Port 40Gb/s InfiniBand / 10GigE Adapter IC with PCIe 2.0 x8 5.0GT/s Interface" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1045u16 , name : "PCIe2 2-port 4X InfiniBand QDR Adapter for Power" } , SubSystem { subvendor : 4116u16 , subdevice : 1159u16 , name : "GX++ 1-port 4X IB QDR Adapter for Power 795" } , SubSystem { subvendor : 4156u16 , subdevice : 6018u16 , name : "4X QDR InfiniBand Mezzanine HCA for c-Class BladeSystem" } , SubSystem { subvendor : 5555u16 , subdevice : 33u16 , name : "HP InfiniBand 4X QDR CX-2 PCI-e G2 Dual Port HCA" }] } , Device { vendor_id : 5555u16 , id : 26438u16 , name : "MT26438 [ConnectX VPI PCIe 2.0 5GT/s - IB QDR / 10GigE Virtualization+]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6017u16 , name : "NC543i 1-port 4x QDR IB/Flex-10 10Gb Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 13129u16 , name : "NC543i 2-port 4xQDR IB/10Gb Adapter" }] } , Device { vendor_id : 5555u16 , id : 26448u16 , name : "MT26448 [ConnectX EN 10GigE, PCIe 2.0 5GT/s]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1046u16 , name : "PCIe2 2-Port 10GbE RoCE SFP+ Adapter" } , SubSystem { subvendor : 4116u16 , subdevice : 1121u16 , name : "PCIe2 2-Port 10GbE RoCE SR Adapter" } , SubSystem { subvendor : 5555u16 , subdevice : 24u16 , name : "HP 10 GbE PCI-e G2 Dual-Port NIC (rev C1)" } , SubSystem { subvendor : 5555u16 , subdevice : 25970u16 , name : "IBM Flex System EN4132 2-port 10Gb RoCE Adapter" }] } , Device { vendor_id : 5555u16 , id : 26458u16 , name : "MT26458 ConnectX EN 10GBASE-T PCIe Gen2 5.0 GT/s" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 26468u16 , name : "MT26468 [ConnectX EN 10GigE, PCIe 2.0 5GT/s Virtualization+]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13075u16 , name : "NC542m Dual Port Flex-10 10GbE BLc Adapter" }] } , Device { vendor_id : 5555u16 , id : 26478u16 , name : "MT26478 [ConnectX EN 40GigE, PCIe 2.0 5GT/s]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 26488u16 , name : "MT26488 [ConnectX VPI PCIe 2.0 5GT/s - IB DDR / 10GigE Virtualization+]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28929u16 , name : "NPS-400 configuration and management interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28930u16 , name : "NPS-400 network interface PF" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28931u16 , name : "NPS-400 network interface VF" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28961u16 , name : "NPS-600 configuration and management interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28962u16 , name : "NPS-600 network interface PF" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 28963u16 , name : "NPS-600 network interface VF" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 33280u16 , name : "Innova-2 Flex Shell Logic" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41680u16 , name : "MT416842 BlueField SoC Crypto enabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41681u16 , name : "MT416842 BlueField SoC Crypto disabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41682u16 , name : "MT416842 BlueField integrated ConnectX-5 network controller" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41683u16 , name : "MT416842 BlueField multicore SoC family VF" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41684u16 , name : "MT42822 BlueField-2 SoC Crypto enabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41685u16 , name : "MT42822 BlueField-2 SoC Crypto disabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41686u16 , name : "MT42822 BlueField-2 integrated ConnectX-6 Dx network controller" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41687u16 , name : "MT43162 BlueField-3 Lx SoC Crypto enabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41688u16 , name : "MT43162 BlueField-3 Lx SoC Crypto disabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41689u16 , name : "MT43162 BlueField-3 Lx integrated ConnectX-7 network controller" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41690u16 , name : "MT43244 BlueField-3 SoC Crypto enabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41691u16 , name : "MT43244 BlueField-3 SoC Crypto disabled" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41692u16 , name : "MT43244 BlueField-3 integrated ConnectX-7 network controller" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41693u16 , name : "BF4 Family Crypto enabled [BlueField-4 SoC Crypto enabled]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41694u16 , name : "BF4 Family Crypto disabled [BlueField-4 SoC Crypto disabled]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 41695u16 , name : "BF4 Family integrated network controller [BlueField-4 integrated network controller]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 45568u16 , name : "ArcusE" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 45569u16 , name : "LibraE" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 45570u16 , name : "Arcus2" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 45571u16 , name : "Arcus3" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 45572u16 , name : "OrionR" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49873u16 , name : "BlueField DPU Family Auxiliary Communication Channel [BlueField Family]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49874u16 , name : "MT416842 BlueField SoC management interfac" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49875u16 , name : "MT42822 BlueField-2 SoC Management Interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49876u16 , name : "MT43162 BlueField-3 Lx SoC Management Interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49877u16 , name : "MT43244 BlueField-3 SoC Management Interface" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 49878u16 , name : "BF4 Family Management Interface [BlueField-4 SoC Management Interface]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51000u16 , name : "MT51136" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51001u16 , name : "MT51136 GW" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51080u16 , name : "Spectrum-8" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51256u16 , name : "MT52236" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51257u16 , name : "MT52236 router" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 51953u16 , name : "ConnectX-4 CAPI Function" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 52100u16 , name : "MT52100" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53000u16 , name : "Switch-IB2" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53001u16 , name : "Quantum Aggregation Node" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53100u16 , name : "MT53100 [Spectrum-2]" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53104u16 , name : "Spectrum-3" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53120u16 , name : "Spectrum-4" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53122u16 , name : "Spectrum-5" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53124u16 , name : "Spectrum-6" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 53126u16 , name : "Spectrum-7" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54000u16 , name : "Quantum HDR (200Gbps) switch" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54002u16 , name : "Quantum-2 NDR (400Gbps) switch" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54004u16 , name : "Quantum-3" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54006u16 , name : "Quantum-3CPO" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54008u16 , name : "NVLink-6 Switch" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54010u16 , name : "NVLink-7 Switch" , subsystems : & [] } , Device { vendor_id : 5555u16 , id : 54012u16 , name : "NVLink-8 Switch" , subsystems : & [] }] }),
        (8423, Vendor { id : 8423u16 , name : "TOPSSD" , devices : & [] }),
        (7614, Vendor { id : 7614u16 , name : "INNOGRIT Corporation" , devices : & [Device { vendor_id : 7614u16 , id : 21000u16 , name : "NVMe SSD Controller IG5208 [Shasta] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7614u16 , id : 21014u16 , name : "NVMe SSD Controller IG5216 [Shasta+] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7614u16 , id : 21024u16 , name : "NVMe SSD Controller IG5220 [RainierQX] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7614u16 , id : 21026u16 , name : "NVMe SSD Controller IG5222 [RainierX] (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7614u16 , id : 21046u16 , name : "NVMe SSD Controller IG5236 [RainierPC]" , subsystems : & [SubSystem { subvendor : 7614u16 , subdevice : 16385u16 , name : "Dongting-B1 DC SSD M.2 480GB" } , SubSystem { subvendor : 7614u16 , subdevice : 16386u16 , name : "Dongting-B1 DC SSD M.2 960GB" }] } , Device { vendor_id : 7614u16 , id : 22070u16 , name : "NVMe DC SSD IG5636" , subsystems : & [SubSystem { subvendor : 7614u16 , subdevice : 1u16 , name : "Dongting-N1 DC SSD U.2 1600GB" } , SubSystem { subvendor : 7614u16 , subdevice : 2u16 , name : "Dongting-N1 DC SSD U.2 1920GB" } , SubSystem { subvendor : 7614u16 , subdevice : 3u16 , name : "Dongting-N1 DC SSD U.2 3200GB" } , SubSystem { subvendor : 7614u16 , subdevice : 4u16 , name : "Dongting-N1 DC SSD U.2 3840GB" } , SubSystem { subvendor : 7614u16 , subdevice : 5u16 , name : "Dongting-N1 DC SSD U.2 6400GB" } , SubSystem { subvendor : 7614u16 , subdevice : 6u16 , name : "Dongting-N1 DC SSD U.2 7680GB" } , SubSystem { subvendor : 7614u16 , subdevice : 4097u16 , name : "Donghu-Z1 DC ZNS SSD U.2 4000GB" } , SubSystem { subvendor : 7614u16 , subdevice : 4098u16 , name : "Donghu-Z1 DC ZNS SSD U.2 8000GB" }] } , Device { vendor_id : 7614u16 , id : 22072u16 , name : "NVMe DC SSD IG5638" , subsystems : & [SubSystem { subvendor : 7614u16 , subdevice : 8193u16 , name : "Dongting-N2 DC SSD U.2 1600GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8194u16 , name : "Dongting-N2 DC SSD U.2 1920GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8195u16 , name : "Dongting-N2 DC SSD U.2 3200GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8196u16 , name : "Dongting-N2 DC SSD U.2 3840GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8197u16 , name : "Dongting-N2 DC SSD U.2 6400GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8198u16 , name : "Dongting-N2 DC SSD U.2 7680GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8472u16 , name : "Dongting-N2Q DC QLC SSD U.2 15360GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8592u16 , name : "Dongting-N2Q DC QLC SSD U.2 30720GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8705u16 , name : "Dongting-N2 2000 DC SSD U.2 1600GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8706u16 , name : "Dongting-N2 2000 DC SSD U.2 1920GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8707u16 , name : "Dongting-N2 2000 DC SSD U.2 3200GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8708u16 , name : "Dongting-N2 2000 DC SSD U.2 3840GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8709u16 , name : "Dongting-N2 2000 DC SSD U.2 6400GB" } , SubSystem { subvendor : 7614u16 , subdevice : 8710u16 , name : "Dongting-N2 2000 DC SSD U.2 7680GB" } , SubSystem { subvendor : 7614u16 , subdevice : 12289u16 , name : "Donghu-Z2 DC ZNS SSD U.2 4000GB" } , SubSystem { subvendor : 7614u16 , subdevice : 12290u16 , name : "Donghu-Z2 DC ZNS SSD U.2 8000GB" }] } , Device { vendor_id : 7614u16 , id : 22118u16 , name : "NVMe SSD Controller IG5666" , subsystems : & [] } , Device { vendor_id : 7614u16 , id : 22120u16 , name : "NVMe PCIe 5.0 DC SSD" , subsystems : & [SubSystem { subvendor : 7614u16 , subdevice : 20483u16 , name : "Dongting-N3 DC SSD U.2 3200GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20484u16 , name : "Dongting-N3 DC SSD U.2 3840GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20485u16 , name : "Dongting-N3 DC SSD U.2 6400GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20486u16 , name : "Dongting-N3 DC SSD U.2 7680GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20487u16 , name : "Dongting-N3 DC SSD U.2 12800GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20488u16 , name : "Dongting-N3 DC SSD U.2 15360GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20489u16 , name : "Dongting-N3 DC SSD U.2 25600GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20608u16 , name : "Dongting-N3 DC SSD U.2 30720GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20760u16 , name : "Dongting-N3Q DC QLC SSD U.2 15360GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20880u16 , name : "Dongting-N3Q DC QLC SSD U.2 30720GB" } , SubSystem { subvendor : 7614u16 , subdevice : 20882u16 , name : "Dongting-N3Q DC QLC SSD U.2 61440GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22531u16 , name : "Dongting-N3 DC SSD E3.S 3200GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22532u16 , name : "Dongting-N3 DC SSD E3.S 3840GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22533u16 , name : "Dongting-N3 DC SSD E3.S 6400GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22534u16 , name : "Dongting-N3 DC SSD E3.S 7680GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22535u16 , name : "Dongting-N3 DC SSD E3.S 12800GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22536u16 , name : "Dongting-N3 DC SSD E3.S 15360GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22537u16 , name : "Dongting-N3 DC SSD E3.S 25600GB" } , SubSystem { subvendor : 7614u16 , subdevice : 22656u16 , name : "Dongting-N3 DC SSD E3.S 30720GB" } , SubSystem { subvendor : 7614u16 , subdevice : 24854u16 , name : "Dongting-Z3Q DC ZNS SSD U.2 7680GB" }] } , Device { vendor_id : 7614u16 , id : 22121u16 , name : "NVMe SSD Controller IG5669 [Tacoma]" , subsystems : & [] }] }),
        (5122, Vendor { id : 5122u16 , name : "Meilhaus Electronic GmbH" , devices : & [Device { vendor_id : 5122u16 , id : 1584u16 , name : "ME-630" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 2368u16 , name : "ME-94" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 2384u16 , name : "ME-95" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 2400u16 , name : "ME-96" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 4096u16 , name : "ME-1000" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 4106u16 , name : "ME-1000" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 4107u16 , name : "ME-1000" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5120u16 , name : "ME-1400" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5130u16 , name : "ME-1400A" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5131u16 , name : "ME-1400B" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5132u16 , name : "ME-1400C" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5133u16 , name : "ME-1400D" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5134u16 , name : "ME-1400E" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5354u16 , name : "ME-1400EA" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5355u16 , name : "ME-1400EB" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5636u16 , name : "ME-1600/4U" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5640u16 , name : "ME-1600/8U" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5644u16 , name : "ME-1600/12U" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5647u16 , name : "ME-1600/16U" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 5775u16 , name : "ME-1600/16U8I" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 17936u16 , name : "ME-4610" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18000u16 , name : "ME-4650" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18016u16 , name : "ME-4660" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18017u16 , name : "ME-4660I" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18018u16 , name : "ME-4660" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18019u16 , name : "ME-4660I" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18032u16 , name : "ME-4670" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18033u16 , name : "ME-4670I" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18034u16 , name : "ME-4670S" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18035u16 , name : "ME-4670IS" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18048u16 , name : "ME-4680" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18049u16 , name : "ME-4680I" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18050u16 , name : "ME-4680S" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 18051u16 , name : "ME-4680IS" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24580u16 , name : "ME-6000/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24584u16 , name : "ME-6000/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24591u16 , name : "ME-6000/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24596u16 , name : "ME-6000I/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24600u16 , name : "ME-6000I/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24607u16 , name : "ME-6000I/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24628u16 , name : "ME-6000ISLE/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24632u16 , name : "ME-6000ISLE/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24639u16 , name : "ME-6000ISLE/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24644u16 , name : "ME-6000/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24648u16 , name : "ME-6000/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24655u16 , name : "ME-6000/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24660u16 , name : "ME-6000I/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24664u16 , name : "ME-6000I/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24671u16 , name : "ME-6000I/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24692u16 , name : "ME-6000ISLE/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24696u16 , name : "ME-6000ISLE/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24703u16 , name : "ME-6000ISLE/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24836u16 , name : "ME-6100/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24840u16 , name : "ME-6100/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24847u16 , name : "ME-6100/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24852u16 , name : "ME-6100I/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24856u16 , name : "ME-6100I/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24863u16 , name : "ME-6100I/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24884u16 , name : "ME-6100ISLE/4" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24888u16 , name : "ME-6100ISLE/8" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24895u16 , name : "ME-6100ISLE/16" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24900u16 , name : "ME-6100/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24904u16 , name : "ME-6100/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24911u16 , name : "ME-6100/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24916u16 , name : "ME-6100I/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24920u16 , name : "ME-6100I/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24927u16 , name : "ME-6100I/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24948u16 , name : "ME-6100ISLE/4/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24952u16 , name : "ME-6100ISLE/8/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 24959u16 , name : "ME-6100ISLE/16/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 25177u16 , name : "ME-6200I/9/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 25433u16 , name : "ME-6300I/9/DIO" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 33034u16 , name : "ME-8100A" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 33035u16 , name : "ME-8100B" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 33290u16 , name : "ME-8200A" , subsystems : & [] } , Device { vendor_id : 5122u16 , id : 33291u16 , name : "ME-8200B" , subsystems : & [] }] }),
        (7935, Vendor { id : 7935u16 , name : "Rebellions Inc." , devices : & [Device { vendor_id : 7935u16 , id : 4128u16 , name : "RBLN-CA2 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4129u16 , name : "RBLN-CA2 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4368u16 , name : "RBLN-CA11 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4369u16 , name : "RBLN-CA11 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4384u16 , name : "RBLN-CA12 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4385u16 , name : "RBLN-CA12 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4624u16 , name : "RBLN-CA21 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4625u16 , name : "RBLN-CA21 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4640u16 , name : "RBLN-CA22 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4641u16 , name : "RBLN-CA22 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4688u16 , name : "RBLN-CA25 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 4689u16 , name : "RBLN-CA25 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8240u16 , name : "RBLN-CR03 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8241u16 , name : "RBLN-CR03 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8496u16 , name : "RBLN-CR13 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8497u16 , name : "RBLN-CR13 (VF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8752u16 , name : "RBLN-CR23 (PF)" , subsystems : & [] } , Device { vendor_id : 7935u16 , id : 8753u16 , name : "RBLN-CR23 (VF)" , subsystems : & [] }] }),
        (4984, Vendor { id : 4984u16 , name : "Telemann Co. Ltd" , devices : & [] }),
        (5352, Vendor { id : 5352u16 , name : "RAYCER Inc" , devices : & [] }),
        (5133, Vendor { id : 5133u16 , name : "Matsushita Electric Works Ltd" , devices : & [] }),
        (5023, Vendor { id : 5023u16 , name : "Aethra s.r.l." , devices : & [] }),
        (45312, Vendor { id : 45312u16 , name : "OpenVox Communication Co. Ltd." , devices : & [] }),
        (5143, Vendor { id : 5143u16 , name : "Convergenet Technologies Inc" , devices : & [] }),
        (8192, Vendor { id : 8192u16 , name : "Smart Link Ltd." , devices : & [Device { vendor_id : 8192u16 , id : 10240u16 , name : "SmartPCI2800 V.92 PCI Soft DFT" , subsystems : & [] }] }),
        (4466, Vendor { id : 4466u16 , name : "Altera Corporation" , devices : & [Device { vendor_id : 4466u16 , id : 4u16 , name : "PF5102 board" , subsystems : & [] } , Device { vendor_id : 4466u16 , id : 167u16 , name : "Stratix V" , subsystems : & [] } , Device { vendor_id : 4466u16 , id : 1328u16 , name : "Stratix IV" , subsystems : & [] } , Device { vendor_id : 4466u16 , id : 25708u16 , name : "KT-500/KT-521 board" , subsystems : & [] }] }),
        (5266, Vendor { id : 5266u16 , name : "Time Logic Corporation" , devices : & [] }),
        (5354, Vendor { id : 5354u16 , name : "Planex Communications, Inc" , devices : & [Device { vendor_id : 5354u16 , id : 43782u16 , name : "FNW-3603-TX CardBus Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 5354u16 , id : 43783u16 , name : "RTL81xx RealTek Ethernet" , subsystems : & [] } , Device { vendor_id : 5354u16 , id : 43784u16 , name : "FNW-3602-TX CardBus Fast Ethernet" , subsystems : & [] }] }),
        (4935, Vendor { id : 4935u16 , name : "Odetics" , devices : & [] }),
        (4360, Vendor { id : 4360u16 , name : "Proteon, Inc." , devices : & [Device { vendor_id : 4360u16 , id : 256u16 , name : "p1690plus_AA" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 257u16 , name : "p1690plus_AB" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 261u16 , name : "P1690Plus" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 264u16 , name : "P1690Plus" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 312u16 , name : "P1690Plus" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 313u16 , name : "P1690Plus" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 316u16 , name : "P1690Plus" , subsystems : & [] } , Device { vendor_id : 4360u16 , id : 317u16 , name : "P1690Plus" , subsystems : & [] }] }),
        (4375, Vendor { id : 4375u16 , name : "Datacube, Inc" , devices : & [Device { vendor_id : 4375u16 , id : 38144u16 , name : "Max-1C SVGA card" , subsystems : & [] } , Device { vendor_id : 4375u16 , id : 38145u16 , name : "Max-1C image processing" , subsystems : & [] }] }),
        (4735, Vendor { id : 4735u16 , name : "Fujifilm" , devices : & [] }),
        (4324, Vendor { id : 4324u16 , name : "Tandem Computers" , devices : & [Device { vendor_id : 4324u16 , id : 32809u16 , name : "Realtek 8029 Network Card" , subsystems : & [] }] }),
        (5251, Vendor { id : 5251u16 , name : "LABWAY Corporation" , devices : & [] }),
        (4225, Vendor { id : 4225u16 , name : "Supermac Technology" , devices : & [Device { vendor_id : 4225u16 , id : 3399u16 , name : "Radius PCI to NuBUS Bridge" , subsystems : & [] }] }),
        (4334, Vendor { id : 4334u16 , name : "Xilinx Corporation" , devices : & [Device { vendor_id : 4334u16 , id : 1u16 , name : "EUROCOM for PCI (ECOMP)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 2u16 , name : "Octal E1/T1 for PCI ETP Card" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 7u16 , name : "Default PCIe endpoint ID" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 517u16 , name : "Wildcard TE205P" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 528u16 , name : "Wildcard TE210P" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 768u16 , name : "Spartan 3 Designs (Xilinx IP)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 788u16 , name : "Wildcard TE405P/TE410P (1st Gen)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 1029u16 , name : "Wildcard TE405P (2nd Gen)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 1040u16 , name : "Wildcard TE410P (2nd Gen)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 1536u16 , name : "Xilinx 6 Designs (Xilinx IP)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16320u16 , name : "RME Digi96" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16321u16 , name : "RME Digi96/8" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16322u16 , name : "RME Digi96/8 Pro" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16323u16 , name : "RME Digi96/8 Pad" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16324u16 , name : "RME Digi9652 (Hammerfall)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16325u16 , name : "RME Hammerfall DSP" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 16326u16 , name : "RME Hammerfall DSP MADI" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 20480u16 , name : "Alveo U200 XDMA Platform" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 20484u16 , name : "Alveo U250 XDMA Platform" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 20485u16 , name : "Alveo U250" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 20492u16 , name : "Alveo U280 XDMA Platform" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 20512u16 , name : "Alveo U50 XMDA Platform" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 20572u16 , name : "Alveo U55C" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 20596u16 , name : "Alveo X3522, Quad Port, 10/25GbE Adaptable Accelerator Card" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 20612u16 , name : "Alveo X3522, Quad Port, 10/25GbE Low Latency Network Adapter" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 27015u16 , name : "SmartSSD" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 27016u16 , name : "SmartSSD" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 28689u16 , name : "7-Series FPGA Hard PCIe block (AXI/debug)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 28728u16 , name : "FPGA Card XC7VX690T" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 16431u16 , name : "FPGA XC7VX690T-3FFG1157E" }] } , Device { vendor_id : 4334u16 , id : 32793u16 , name : "Memory controller" , subsystems : & [SubSystem { subvendor : 7916u16 , subdevice : 513u16 , name : "VSEC10232X Dual-port 100Gb/s Etherent PCIe" }] } , Device { vendor_id : 4334u16 , id : 33664u16 , name : "Ellips ProfiXpress Profibus Master" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 33665u16 , name : "Ellips Santos Frame Grabber" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 37172u16 , name : "SmartSSD" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 37428u16 , name : "SmartSSD" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 37940u16 , name : "SmartSSD" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 53248u16 , name : "Alveo U200 Golden Image" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 53252u16 , name : "Alveo U250 Golden Image" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 53260u16 , name : "Alveo U280 Golden Image" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 53280u16 , name : "Alveo U50 Golden Image" , subsystems : & [SubSystem { subvendor : 4334u16 , subdevice : 14u16 , name : "Alveo card" }] } , Device { vendor_id : 4334u16 , id : 53588u16 , name : "Copley Controls CAN card (PCI-CAN-02)" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 60400u16 , name : "SED Systems Modulator/Demodulator" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 60401u16 , name : "SED Systems Audio Interface Card" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 60402u16 , name : "SED Systems Common PCI Interface" , subsystems : & [] } , Device { vendor_id : 4334u16 , id : 60403u16 , name : "SED Systems PCIe-AXI Bridge" , subsystems : & [] }] }),
        (4977, Vendor { id : 4977u16 , name : "CNet Technology Inc" , devices : & [Device { vendor_id : 4977u16 , id : 17230u16 , name : "GigaCard Network Adapter" , subsystems : & [SubSystem { subvendor : 4977u16 , subdevice : 17230u16 , name : "N-Way PCI-Bus Giga-Card 1000/100/10Mbps(L)" }] }] }),
        (5980, Vendor { id : 5980u16 , name : "AudioScience Inc" , devices : & [] }),
        (4371, Vendor { id : 4371u16 , name : "Accton Technology Corporation" , devices : & [Device { vendor_id : 4371u16 , id : 4625u16 , name : "SMC2-1211TX" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4615u16 , name : "EN-1207D Fast Ethernet Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 4625u16 , name : "EN-1207D Fast Ethernet Adapter" }] } , Device { vendor_id : 4371u16 , id : 4630u16 , name : "EN-1216 Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4371u16 , subdevice : 4630u16 , name : "EN1207F series PCI Fast Ethernet Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 8736u16 , name : "EN2220A Cardbus Fast Ethernet Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 8770u16 , name : "EN2242 10/100 Ethernet Mini-PCI Card" } , SubSystem { subvendor : 4378u16 , subdevice : 4128u16 , name : "SpeedStream 1020 PCI 10/100 Ethernet Adaptor [EN-1207F-TX ?]" }] } , Device { vendor_id : 4371u16 , id : 4631u16 , name : "EN-1217 Ethernet Adapter" , subsystems : & [] } , Device { vendor_id : 4371u16 , id : 20741u16 , name : "10Mbps Network card" , subsystems : & [] } , Device { vendor_id : 4371u16 , id : 37393u16 , name : "EN-1207D Fast Ethernet Adapter" , subsystems : & [SubSystem { subvendor : 4371u16 , subdevice : 37393u16 , name : "EN-1207D Fast Ethernet Adapter" }] } , Device { vendor_id : 4371u16 , id : 38161u16 , name : "21x4x DEC-Tulip compatible Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 4371u16 , id : 54017u16 , name : "CPWNA100 (Philips wireless PCMCIA)" , subsystems : & [] } , Device { vendor_id : 4371u16 , id : 60418u16 , name : "SMC 1244TX v3" , subsystems : & [] } , Device { vendor_id : 4371u16 , id : 60963u16 , name : "SMCWPCIT-G 108Mbps Wireless PCI adapter" , subsystems : & [] }] }),
        (4912, Vendor { id : 4912u16 , name : "MMC Networks" , devices : & [] }),
        (4221, Vendor { id : 4221u16 , name : "LeadTek Research Inc." , devices : & [Device { vendor_id : 4221u16 , id : 0u16 , name : "P86C850" , subsystems : & [] }] }),
        (5503, Vendor { id : 5503u16 , name : "PX Instruments Technology Ltd" , devices : & [] }),
        (4538, Vendor { id : 4538u16 , name : "Videotron Corp" , devices : & [] }),
        (1801, Vendor { id : 1801u16 , name : "LJMicro Co., Ltd" , devices : & [Device { vendor_id : 1801u16 , id : 257u16 , name : "GP101" , subsystems : & [] } , Device { vendor_id : 1801u16 , id : 258u16 , name : "GP102" , subsystems : & [] } , Device { vendor_id : 1801u16 , id : 513u16 , name : "GP201" , subsystems : & [] } , Device { vendor_id : 1801u16 , id : 514u16 , name : "GP202" , subsystems : & [] }] }),
        (4755, Vendor { id : 4755u16 , name : "Media Reality Technology" , devices : & [] }),
        (5033, Vendor { id : 5033u16 , name : "Siemens Medical Systems, Ultrasound Group" , devices : & [] }),
        (7880, Vendor { id : 7880u16 , name : "Beijing Fantasy Technology Co., Ltd." , devices : & [Device { vendor_id : 7880u16 , id : 34816u16 , name : "Fantasy I" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy I Device" }] } , Device { vendor_id : 7880u16 , id : 34832u16 , name : "Fantasy I" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy I Device" }] } , Device { vendor_id : 7880u16 , id : 34833u16 , name : "Fantasy I-V" , subsystems : & [] } , Device { vendor_id : 7880u16 , id : 34848u16 , name : "Fantasy III" , subsystems : & [] } , Device { vendor_id : 7880u16 , id : 35072u16 , name : "GR308" , subsystems : & [] } , Device { vendor_id : 7880u16 , id : 35074u16 , name : "GR3 Audio" , subsystems : & [] } , Device { vendor_id : 7880u16 , id : 35076u16 , name : "GR308" , subsystems : & [] } , Device { vendor_id : 7880u16 , id : 38912u16 , name : "Fantasy II" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy II Device" }] } , Device { vendor_id : 7880u16 , id : 38914u16 , name : "Fantasy II" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy II Device" }] } , Device { vendor_id : 7880u16 , id : 38916u16 , name : "Fantasy II" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy II Device" }] } , Device { vendor_id : 7880u16 , id : 38928u16 , name : "Fantasy II-M" , subsystems : & [SubSystem { subvendor : 7880u16 , subdevice : 4770u16 , name : "Fantasy II Device" }] } , Device { vendor_id : 7880u16 , id : 39424u16 , name : "GR316" , subsystems : & [] }] }),
        (4623, Vendor { id : 4623u16 , name : "Essential Communications" , devices : & [Device { vendor_id : 4623u16 , id : 1u16 , name : "Roadrunner serial HIPPI" , subsystems : & [] }] }),
        (4470, Vendor { id : 4470u16 , name : "SBE Incorporated" , devices : & [] }),
        (6956, Vendor { id : 6956u16 , name : "Opal-RT Technologies Inc." , devices : & [] }),
        (25929, Vendor { id : 25929u16 , name : "Teradici Corp." , devices : & [Device { vendor_id : 25929u16 , id : 4608u16 , name : "TERA1200 PC-over-IP Host" , subsystems : & [] }] }),
        (4604, Vendor { id : 4604u16 , name : "Silicon Magic" , devices : & [] }),
        (5046, Vendor { id : 5046u16 , name : "Dlog GmbH" , devices : & [] }),
        (43981, Vendor { id : 43981u16 , name : "Vadatech Inc." , devices : & [] }),
        (5965, Vendor { id : 5965u16 , name : "WellX Telecom SA" , devices : & [] }),
        (5259, Vendor { id : 5259u16 , name : "INNOMEDIALOGIC Inc." , devices : & [] }),
        (4876, Vendor { id : 4876u16 , name : "Ambex Technologies, Inc" , devices : & [] }),
        (59984, Vendor { id : 59984u16 , name : "Emerson Automation Solutions" , devices : & [] }),
        (4253, Vendor { id : 4253u16 , name : "Zida Technologies Ltd." , devices : & [] }),
        (4931, Vendor { id : 4931u16 , name : "Phylon Communications Inc" , devices : & [] }),
        (4247, Vendor { id : 4247u16 , name : "Appian Technology" , devices : & [] }),
        (512, Vendor { id : 512u16 , name : "Dell (wrong ID)" , devices : & [] }),
        (5524, Vendor { id : 5524u16 , name : "Netgame Ltd" , devices : & [] }),
        (6920, Vendor { id : 6920u16 , name : "Tria Technologies GmbH" , devices : & [] }),
        (4348, Vendor { id : 4348u16 , name : "I-O Data Device, Inc." , devices : & [Device { vendor_id : 4348u16 , id : 3u16 , name : "Cardbus IDE Controller" , subsystems : & [] } , Device { vendor_id : 4348u16 , id : 5u16 , name : "Cardbus SCSI CBSC II" , subsystems : & [] }] }),
        (296, Vendor { id : 296u16 , name : "Dell (wrong ID)" , devices : & [] }),
        (7367, Vendor { id : 7367u16 , name : "Radian Memory Systems Inc." , devices : & [Device { vendor_id : 7367u16 , id : 512u16 , name : "RMS-200 PCIe NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7367u16 , id : 592u16 , name : "RMS-250 U.2 NVMe SSD" , subsystems : & [] }] }),
        (4917, Vendor { id : 4917u16 , name : "Videomail, Inc" , devices : & [] }),
        (5026, Vendor { id : 5026u16 , name : "Ositech Communications Inc" , devices : & [] }),
        (4240, Vendor { id : 4240u16 , name : "Compro Computer Services, Inc." , devices : & [Device { vendor_id : 4240u16 , id : 17936u16 , name : "PCI RTOM" , subsystems : & [] } , Device { vendor_id : 4240u16 , id : 17952u16 , name : "GPIO HSD" , subsystems : & [] }] }),
        (5900, Vendor { id : 5900u16 , name : "YottaYotta Inc." , devices : & [] }),
        (4423, Vendor { id : 4423u16 , name : "Interface Corp" , devices : & [] }),
        (4890, Vendor { id : 4890u16 , name : "Finisar Corp." , devices : & [] }),
        (4511, Vendor { id : 4511u16 , name : "Bull HN Information Systems" , devices : & [Device { vendor_id : 4511u16 , id : 4225u16 , name : "BXI Host Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4511u16 , id : 4353u16 , name : "BXI Host Channel Adapter v1.2" , subsystems : & [] } , Device { vendor_id : 4511u16 , id : 4385u16 , name : "BXI Host Channel Adapter v1.3" , subsystems : & [] }] }),
        (5443, Vendor { id : 5443u16 , name : "SILICON Laboratories" , devices : & [Device { vendor_id : 5443u16 , id : 12370u16 , name : "Intel 537 [Winmodem]" , subsystems : & [] } , Device { vendor_id : 5443u16 , id : 19490u16 , name : "Si3036 MC'97 DAA" , subsystems : & [] }] }),
        (8251, Vendor { id : 8251u16 , name : "XTX Markets Technologies Ltd." , devices : & [] }),
        (771, Vendor { id : 771u16 , name : "Hewlett-Packard Company (Wrong ID)" , devices : & [] }),
        (33668, Vendor { id : 33668u16 , name : "SigmaTel" , devices : & [] }),
        (4270, Vendor { id : 4270u16 , name : "Cornerstone Technology" , devices : & [] }),
        (5028, Vendor { id : 5028u16 , name : "Rascom Inc" , devices : & [] }),
        (7736, Vendor { id : 7736u16 , name : "Blaize, Inc" , devices : & [Device { vendor_id : 7736u16 , id : 258u16 , name : "Xplorer X1600" , subsystems : & [] }] }),
        (6126, Vendor { id : 6126u16 , name : "Connect Components Ltd" , devices : & [] }),
        (8525, Vendor { id : 8525u16 , name : "Equalizer Technologies" , devices : & [Device { vendor_id : 8525u16 , id : 1u16 , name : "Xeetah Base Firmware" , subsystems : & [] } , Device { vendor_id : 8525u16 , id : 2u16 , name : "Xeetah Development Platform" , subsystems : & [] }] }),
        (4653, Vendor { id : 4653u16 , name : "Aztech System Ltd" , devices : & [Device { vendor_id : 4653u16 , id : 4614u16 , name : "368DSP" , subsystems : & [] } , Device { vendor_id : 4653u16 , id : 5120u16 , name : "Trident PCI288-Q3DII (NX)" , subsystems : & [] } , Device { vendor_id : 4653u16 , id : 20700u16 , name : "3328 Audio" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 1u16 , name : "3328 Audio" }] } , Device { vendor_id : 4653u16 , id : 32986u16 , name : "3328 Audio" , subsystems : & [SubSystem { subvendor : 4653u16 , subdevice : 1u16 , name : "3328 Audio" }] }] }),
        (7869, Vendor { id : 7869u16 , name : "EMERGETECH Company Ltd." , devices : & [Device { vendor_id : 7869u16 , id : 257u16 , name : "Seirios 2063 Video Codec" , subsystems : & [] }] }),
        (20056, Vendor { id : 20056u16 , name : "Nutanix, Inc." , devices : & [Device { vendor_id : 20056u16 , id : 1u16 , name : "Virtual NVMe Controller" , subsystems : & [] }] }),
        (31710, Vendor { id : 31710u16 , name : "MIDAC Corporation" , devices : & [] }),
        (4878, Vendor { id : 4878u16 , name : "Yamatake-Honeywell Co. Ltd" , devices : & [] }),
        (4745, Vendor { id : 4745u16 , name : "AVC Technology, Inc." , devices : & [] }),
        (6080, Vendor { id : 6080u16 , name : "Wistron Corp." , devices : & [] }),
        (4112, Vendor { id : 4112u16 , name : "Video Logic, Ltd." , devices : & [] }),
        (5454, Vendor { id : 5454u16 , name : "SERVOTEST Ltd" , devices : & [] }),
        (6552, Vendor { id : 6552u16 , name : "Toyou Feiji Electronics Co., Ltd." , devices : & [Device { vendor_id : 6552u16 , id : 1u16 , name : "TOBOLT1 51987 NVMe SSD" , subsystems : & [SubSystem { subvendor : 6552u16 , subdevice : 402u16 , name : "TOBOLT1 45967 1920G M.2 NVMe SSD" } , SubSystem { subvendor : 6552u16 , subdevice : 900u16 , name : "TOBOLT1 51987 3840G 2.5\" U.2 NVMe SSD" } , SubSystem { subvendor : 6552u16 , subdevice : 1896u16 , name : "TOBOLT1 51987 7680G 2.5\" U.2 NVMe SSD" } , SubSystem { subvendor : 6552u16 , subdevice : 8210u16 , name : "TOBOLT1 51987 3840G 2.5\" U.2 NVMe SSD" } , SubSystem { subvendor : 6552u16 , subdevice : 24900u16 , name : "TOBOLT1 51995 6144G 2.5\" U.2 NVMe SSD" }] }] }),
        (5838, Vendor { id : 5838u16 , name : "Roland Corp." , devices : & [] }),
        (22136, Vendor { id : 22136u16 , name : "Dawicontrol Computersysteme GmbH" , devices : & [] }),
        (9580, Vendor { id : 9580u16 , name : "Graphics Technology (HK) Co., Ltd." , devices : & [Device { vendor_id : 9580u16 , id : 109u16 , name : "HS610" , subsystems : & [] }] }),
        (4766, Vendor { id : 4766u16 , name : "Victor Company of Japan, Ltd." , devices : & [] }),
        (2341, Vendor { id : 2341u16 , name : "VIA Technologies, Inc. (Wrong ID)" , devices : & [] }),
        (4461, Vendor { id : 4461u16 , name : "Martin-Marietta" , devices : & [] }),
        (8628, Vendor { id : 8628u16 , name : "Hunan Goke Microelectronics Co., Ltd" , devices : & [] }),
        (11224, Vendor { id : 11224u16 , name : "ROPEX Industrie-Elektronik GmbH" , devices : & [] }),
        (5173, Vendor { id : 5173u16 , name : "RTD Embedded Technologies, Inc." , devices : & [Device { vendor_id : 5173u16 , id : 17696u16 , name : "PCI4520" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 24608u16 , name : "SPM6020" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 24624u16 , name : "SPM6030" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 25632u16 , name : "SPM186420" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 25648u16 , name : "SPM176430" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 25649u16 , name : "SPM176431" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 29984u16 , name : "DM7520" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 30016u16 , name : "SDM7540" , subsystems : & [] } , Device { vendor_id : 5173u16 , id : 30752u16 , name : "DM7820" , subsystems : & [] }] }),
        (7559, Vendor { id : 7559u16 , name : "Rockchip Electronics Co., Ltd" , devices : & [Device { vendor_id : 7559u16 , id : 256u16 , name : "RK3399 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 6152u16 , name : "RK1808 Neural Network Processor Card" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 6186u16 , name : "RK182x Edge AI Computing Coprocessor" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 13608u16 , name : "RK3528 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 13666u16 , name : "RK3562 PCI Express Root Port" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 13670u16 , name : "RK3568 Remote Signal Processor" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 13686u16 , name : "RK3576" , subsystems : & [] } , Device { vendor_id : 7559u16 , id : 13704u16 , name : "RK3588" , subsystems : & [] }] }),
        (4883, Vendor { id : 4883u16 , name : "Yaskawa Electric Co." , devices : & [] }),
        (5147, Vendor { id : 5147u16 , name : "Zoom Telephonics Inc" , devices : & [] }),
        (5456, Vendor { id : 5456u16 , name : "OPEN NETWORK Co Ltd" , devices : & [] }),
        (5359, Vendor { id : 5359u16 , name : "CARRY Computer ENG. CO Ltd" , devices : & [] }),
        (4954, Vendor { id : 4954u16 , name : "Brainboxes Ltd" , devices : & [Device { vendor_id : 4954u16 , id : 2113u16 , name : "UC-268 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2145u16 , name : "UC-257 2 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2146u16 , name : "UC-257 2 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2147u16 , name : "UC-257 2 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2177u16 , name : "UC-279 8 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2209u16 , name : "UC-313 2 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2210u16 , name : "UC-313 2 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2211u16 , name : "UC-313 2 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2241u16 , name : "UC-310 2 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2273u16 , name : "UC-302 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2274u16 , name : "UC-302 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2275u16 , name : "UC-302 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2305u16 , name : "UC-431 3 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2337u16 , name : "UC-420 3 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2433u16 , name : "UC-475 1 + 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2434u16 , name : "UC-475 1 + 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2465u16 , name : "UC-607 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2466u16 , name : "UC-607 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2467u16 , name : "UC-607 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2657u16 , name : "UC-324 1 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2689u16 , name : "UC-357 1 port RS-232 + 1 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2690u16 , name : "UC-357 1 port RS-232 + 1 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2691u16 , name : "UC-357 1 port RS-232 + 1 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2721u16 , name : "UC-246 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2722u16 , name : "UC-246 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2753u16 , name : "UP-189 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2754u16 , name : "UP-189 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2755u16 , name : "UP-189 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2817u16 , name : "UC-346 4 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2818u16 , name : "UC-346 4 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2849u16 , name : "UP-200 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2850u16 , name : "UP-200 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2851u16 , name : "UP-200 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2977u16 , name : "UC-101 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2978u16 , name : "UC-101 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 2979u16 , name : "UC-101 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3009u16 , name : "UC-203 1 + 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3010u16 , name : "UC-203 1 + 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3041u16 , name : "UC-146 LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3042u16 , name : "UC-146 LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3073u16 , name : "UP-869 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3074u16 , name : "UP-869 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3075u16 , name : "UP-869 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3105u16 , name : "UP-880 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3106u16 , name : "UP-880 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3107u16 , name : "UP-880 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3137u16 , name : "UC-368 4 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3138u16 , name : "UC-368 4 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3139u16 , name : "UC-368 4 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3233u16 , name : "UC-253 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3361u16 , name : "UC-260 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3393u16 , name : "UC-836 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3424u16 , name : "IS-100 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3456u16 , name : "IS-200 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3488u16 , name : "IS-300 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3520u16 , name : "IS-400 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3552u16 , name : "IS-500 LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3649u16 , name : "PX-279 8 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 3681u16 , name : "UC-414 3 + 1 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16384u16 , name : "PX-420 3 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16385u16 , name : "PX-431 3 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16386u16 , name : "PX-820 Powered 3 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16387u16 , name : "PX-831 Powered 3 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16388u16 , name : "PX-235 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16389u16 , name : "PX-101 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16390u16 , name : "PX-257 1 + 1 port RS-232 + LPT card (Serial port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16391u16 , name : "PX-257 1 + 1 port RS-232 + LPT card (LPT port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16392u16 , name : "PX-835 Powered 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16393u16 , name : "PX-857 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16394u16 , name : "PX-260 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16395u16 , name : "PX-320 1 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16396u16 , name : "PX-313 2 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16398u16 , name : "PX-310 2 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16399u16 , name : "PX-346 4 port RS-422/485 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16400u16 , name : "PX-368 4 port RS-422/485 Opto Isolated card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16401u16 , name : "PX-420 3 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16402u16 , name : "PX-431 3 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16403u16 , name : "PX-820 Powered 3 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16404u16 , name : "PX-831 Powered 3 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16405u16 , name : "PX-257 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16406u16 , name : "PX-235 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16407u16 , name : "PX-835 Powered 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16408u16 , name : "PX-857 Powered 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16409u16 , name : "PX-101 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16412u16 , name : "PX-146 LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16413u16 , name : "PX-475 1 port RS-232 + LPT card (Serial port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16414u16 , name : "PX-803 Powered 1 + 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16415u16 , name : "PX-475 1 port RS-232 + LPT card (LPT port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16416u16 , name : "XC-157 LPT ExpressCard" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16417u16 , name : "XC-475 1 port RS-232 + LPT ExpressCard (Serial port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16418u16 , name : "XC-475 1 port RS-232 + LPT ExpressCard (LPT port)" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16422u16 , name : "XC-235 1 port RS-232 ExpressCard" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16423u16 , name : "IX-100 1 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16424u16 , name : "IX-200 2 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16425u16 , name : "IX-400 4 port RS-232 card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16426u16 , name : "IX-500 LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16428u16 , name : "PX-263 4 port RS-232 + LPT card" , subsystems : & [] } , Device { vendor_id : 4954u16 , id : 16640u16 , name : "PX-272 4 + 1 port RS-232 + LPT card" , subsystems : & [] }] }),
        (53347, Vendor { id : 53347u16 , name : "SolidRun" , devices : & [Device { vendor_id : 53347u16 , id : 4096u16 , name : "SolidNET DPU" , subsystems : & [] } , Device { vendor_id : 53347u16 , id : 5376u16 , name : "Emulated NVMe" , subsystems : & [] }] }),
        (4773, Vendor { id : 4773u16 , name : "Vision Dynamics Ltd." , devices : & [] }),
        (5116, Vendor { id : 5116u16 , name : "Computer Peripherals International" , devices : & [] }),
        (6418, Vendor { id : 6418u16 , name : "Renesas Electronics Corp." , devices : & [Device { vendor_id : 6418u16 , id : 2u16 , name : "SH7780 PCI Controller (PCIC)" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 4u16 , name : "SH7763 PCI Controller (PCIC)" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 17u16 , name : "SH7757 PCIe End-Point [PBI]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 18u16 , name : "SH7757 PCIe-PCI Bridge [PPB]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 19u16 , name : "SH7757 PCIe Switch [PS]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 20u16 , name : "uPD720201 USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 21u16 , name : "uPD720202 USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 39538u16 , name : "LRSU9A72 2-Port USB 3.0 Exchange Adapter" }] } , Device { vendor_id : 6418u16 , id : 26u16 , name : "SH7758 PCIe-PCI Bridge [PPB]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 27u16 , name : "SH7758 PCIe End-Point [PBI]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 29u16 , name : "SH7758 PCIe Switch [PS]" , subsystems : & [] } , Device { vendor_id : 6418u16 , id : 51u16 , name : "RZ/G3S PCIe 2.0 Controller" , subsystems : & [] }] }),
        (4480, Vendor { id : 4480u16 , name : "Ricoh Co Ltd" , devices : & [Device { vendor_id : 4480u16 , id : 1125u16 , name : "RL5c465" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1126u16 , name : "RL5c466" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1141u16 , name : "RL5c475" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49158u16 , name : "vpr Matrix 170B4 CardBus bridge" }] } , Device { vendor_id : 4480u16 , id : 1142u16 , name : "RL5c476 II" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 389u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4116u16 , subdevice : 1365u16 , name : "ThinkPad X41" } , SubSystem { subvendor : 4116u16 , subdevice : 1388u16 , name : "ThinkPad Z60t" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300 laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 392u16 , name : "Inspiron 6000 laptop" } , SubSystem { subvendor : 4156u16 , subdevice : 12480u16 , name : "Compaq 6710b" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 6503u16 , name : "V6800V" } , SubSystem { subvendor : 4163u16 , subdevice : 6535u16 , name : "A4K and Z81K notebooks, possibly others ( mid-2005 machines )" } , SubSystem { subvendor : 4173u16 , subdevice : 32991u16 , name : "Vaio PCG-FX403" } , SubSystem { subvendor : 4173u16 , subdevice : 32999u16 , name : "VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP" } , SubSystem { subvendor : 4173u16 , subdevice : 33102u16 , name : "VAIO GRZ390Z" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 5359u16 , subdevice : 544u16 , name : "PCD-RP-220S" } , SubSystem { subvendor : 6058u16 , subdevice : 8220u16 , name : "ThinkPad X60/X60s" } , SubSystem { subvendor : 6058u16 , subdevice : 8388u16 , name : "ThinkPad T61/R61" } , SubSystem { subvendor : 6058u16 , subdevice : 8390u16 , name : "ThinkPad R61/T400" }] } , Device { vendor_id : 4480u16 , id : 1143u16 , name : "RL5c477" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1144u16 , name : "RL5c478" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 388u16 , name : "ThinkPad A30p" }] } , Device { vendor_id : 4480u16 , id : 1297u16 , name : "R5C511" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1314u16 , name : "R5C522 IEEE 1394 Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 463u16 , name : "ThinkPad A30p" } , SubSystem { subvendor : 4163u16 , subdevice : 6503u16 , name : "V6800V" }] } , Device { vendor_id : 4480u16 , id : 1361u16 , name : "R5C551 IEEE 1394 Controller" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 49158u16 , name : "vpr Matrix 170B4" }] } , Device { vendor_id : 4480u16 , id : 1362u16 , name : "R5C552 IEEE 1394 Controller" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1297u16 , name : "ThinkPad A/T/X Series" } , SubSystem { subvendor : 4136u16 , subdevice : 335u16 , name : "Latitude X300 laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 392u16 , name : "Inspiron 6000 laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 5975u16 , name : "M2400N laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49157u16 , name : "X10 Laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49164u16 , name : "P30/P35 notebook" } , SubSystem { subvendor : 6058u16 , subdevice : 8222u16 , name : "ThinkPad X60/X60s" }] } , Device { vendor_id : 4480u16 , id : 1364u16 , name : "R5C554" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1397u16 , name : "R5C575 SD Bus Host Adapter" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1398u16 , name : "R5C576 SD Bus Host Adapter" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 1426u16 , name : "R5C592 Memory Stick Bus Host Adapter" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 5409u16 , name : "HP EliteBook 8540w" } , SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv95xx/96xx/97xx/98xx series" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 6503u16 , name : "V6800V" } , SubSystem { subvendor : 4173u16 , subdevice : 36917u16 , name : "VAIO VGN-FW11ZRU" } , SubSystem { subvendor : 5197u16 , subdevice : 49176u16 , name : "X20 IV" } , SubSystem { subvendor : 6058u16 , subdevice : 8394u16 , name : "ThinkPad T61/T400" }] } , Device { vendor_id : 4480u16 , id : 2065u16 , name : "R5C811" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 2082u16 , name : "R5C822 SD/SDIO/MMC/MS/MSPro Host Adapter" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1366u16 , name : "ThinkPad X40 / X41 / X60s / Z60t" } , SubSystem { subvendor : 4116u16 , subdevice : 1432u16 , name : "ThinkPad Z60m" } , SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 392u16 , name : "Inspiron 6000 laptop" } , SubSystem { subvendor : 4136u16 , subdevice : 418u16 , name : "Inspiron 9200" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 949u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4163u16 , subdevice : 4663u16 , name : "A6J-Q008" } , SubSystem { subvendor : 4163u16 , subdevice : 6503u16 , name : "V6800V" } , SubSystem { subvendor : 4343u16 , subdevice : 33592u16 , name : "Panasonic CF-Y5 laptop" } , SubSystem { subvendor : 5197u16 , subdevice : 49176u16 , name : "X20 IV" } , SubSystem { subvendor : 6058u16 , subdevice : 8221u16 , name : "ThinkPad X60/X60s" } , SubSystem { subvendor : 6058u16 , subdevice : 8391u16 , name : "ThinkPad T61" } , SubSystem { subvendor : 6058u16 , subdevice : 8392u16 , name : "ThinkPad T400/W500" }] } , Device { vendor_id : 4480u16 , id : 2098u16 , name : "R5C832 IEEE 1394 Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4136u16 , subdevice : 589u16 , name : "Latitude E4300" } , SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4156u16 , subdevice : 12481u16 , name : "Compaq 6910p" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" } , SubSystem { subvendor : 6058u16 , subdevice : 8389u16 , name : "ThinkPad R61" } , SubSystem { subvendor : 6058u16 , subdevice : 8391u16 , name : "ThinkPad R61/T400" }] } , Device { vendor_id : 4480u16 , id : 2113u16 , name : "R5C841 CardBus/SD/SDIO/MMC/MS/MSPro/xD/IEEE1394" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 2115u16 , name : "R5C843 MMC Host Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 471u16 , name : "XPS M1210" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 501u16 , name : "Inspiron 1501" } , SubSystem { subvendor : 4136u16 , subdevice : 591u16 , name : "Latitude e6500" } , SubSystem { subvendor : 4156u16 , subdevice : 949u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 5409u16 , name : "HP EliteBook 8540w" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9500/9600/9700 series" } , SubSystem { subvendor : 4163u16 , subdevice : 4119u16 , name : "X58LE" } , SubSystem { subvendor : 4483u16 , subdevice : 2115u16 , name : "Alienware Aurora m9700" }] } , Device { vendor_id : 4480u16 , id : 2130u16 , name : "xD-Picture Card Controller" , subsystems : & [SubSystem { subvendor : 4133u16 , subdevice : 289u16 , name : "Aspire 5920G" } , SubSystem { subvendor : 4136u16 , subdevice : 499u16 , name : "Inspiron 1420" } , SubSystem { subvendor : 4136u16 , subdevice : 559u16 , name : "Inspiron 1525" } , SubSystem { subvendor : 4156u16 , subdevice : 12469u16 , name : "Presario V3242AU" } , SubSystem { subvendor : 4156u16 , subdevice : 12471u16 , name : "Presario V6133CL" } , SubSystem { subvendor : 4156u16 , subdevice : 12492u16 , name : "Pavilion dv6700" } , SubSystem { subvendor : 4156u16 , subdevice : 12495u16 , name : "Pavilion dv9668eg Laptop" } , SubSystem { subvendor : 4163u16 , subdevice : 6503u16 , name : "V6800V" } , SubSystem { subvendor : 4480u16 , subdevice : 2130u16 , name : "Pavilion 2410us" } , SubSystem { subvendor : 4900u16 , subdevice : 4303u16 , name : "P7120" } , SubSystem { subvendor : 6058u16 , subdevice : 8395u16 , name : "ThinkPad T400" }] } , Device { vendor_id : 4480u16 , id : 57904u16 , name : "R5U2xx (R5U230 / R5U231 / R5U241) [Memory Stick Host Controller]" , subsystems : & [] } , Device { vendor_id : 4480u16 , id : 58486u16 , name : "R5U242 PCIe to CardBus Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" }] } , Device { vendor_id : 4480u16 , id : 59426u16 , name : "MMC/SD Host Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 4480u16 , id : 59427u16 , name : "PCIe SDXC/MMC Host Controller" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 4480u16 , id : 59442u16 , name : "R5C832 PCIe IEEE 1394 Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1034u16 , name : "Latitude E6410" } , SubSystem { subvendor : 4136u16 , subdevice : 1035u16 , name : "Latitude E6510" } , SubSystem { subvendor : 6058u16 , subdevice : 8655u16 , name : "ThinkPad T520" }] } , Device { vendor_id : 4480u16 , id : 59474u16 , name : "PCIe xD-Picture Card Controller" , subsystems : & [] }] }),
        (6189, Vendor { id : 6189u16 , name : "SiteCom Europe BV" , devices : & [Device { vendor_id : 6189u16 , id : 12393u16 , name : "ISDN PCI DC-105V2" , subsystems : & [] } , Device { vendor_id : 6189u16 , id : 38800u16 , name : "WL-121 Wireless Network Adapter 100g+ [Ver.3]" , subsystems : & [] }] }),
        (7532, Vendor { id : 7532u16 , name : "Atomic Rules LLC" , devices : & [Device { vendor_id : 7532u16 , id : 4097u16 , name : "A5PL-E1" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4098u16 , name : "A5PL-E7" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4099u16 , name : "S5PEDS-AB" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4100u16 , name : "KC705-K325" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4101u16 , name : "ZC706-Z045" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4102u16 , name : "KCU105-KU040" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4103u16 , name : "XUSP3S-VU095 [Jasper]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4104u16 , name : "XUSPL4-VU065 [Mustang UltraScale]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4105u16 , name : "XUSPL4-VU3P [Mustang UltraScale+]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4106u16 , name : "A10PL4-A10GX115" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4107u16 , name : "K35-2SFP" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4108u16 , name : "K35-4SFP" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4109u16 , name : "AR-ARKA-FX0 [Arkville 32B DPDK Data Mover]" , subsystems : & [SubSystem { subvendor : 7532u16 , subdevice : 8193u16 , name : "DPDK-Aware Virtual Function [Arkville VF]" }] } , Device { vendor_id : 7532u16 , id : 4110u16 , name : "AR-ARKA-FX1 [Arkville 64B DPDK Data Mover]" , subsystems : & [SubSystem { subvendor : 7532u16 , subdevice : 8193u16 , name : "DPDK-Aware Virtual Function [Arkville VF]" }] } , Device { vendor_id : 7532u16 , id : 4111u16 , name : "AR-ARKA-FX1 [Arkville 64B DPDK Data Mover for Versal]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4112u16 , name : "AR-ARKA-FX1 [Arkville 64B DPDK Data Mover for Agilex]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4113u16 , name : "AR-MAN-U50 [Manitou Class Accelerator for U50]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4114u16 , name : "AR-MAN-U200 [Manitou Class Accelerator for U200]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4115u16 , name : "AR-MAN-U250 [Manitou Class Accelerator for U250]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4116u16 , name : "AR-MAN-U280 [Manitou Class Accelerator for U280]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4117u16 , name : "AR-ARK-BBDEV-FX0 [Arkville 32B DPDK Baseband Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4118u16 , name : "AR-ARK-BBDEV-FX1 [Arkville 64B DPDK Baseband Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4119u16 , name : "AR-ARK-FX1 [Arkville 64B Multi-Homed Primary Endpoint]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4120u16 , name : "AR-ARK-FX1 [Arkville 64B Multi-Homed Secondary Endpoint]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4121u16 , name : "AR-ARK-FX1 [Arkville 64B Multi-Homed Tertiary Endpoint]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4122u16 , name : "AR-ARK-SRIOV-FX0 [Arkville 32B Primary Physical Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4123u16 , name : "AR-ARK-SRIOV-FX1 [Arkville 64B Primary Physical Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4124u16 , name : "AR-ARK-SRIOV-VF [Arkville Virtual Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4125u16 , name : "AR-ARK-NIC [Arkville ArkNIC Kernel Path Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4126u16 , name : "AR-ARKA-FX1 [Arkville 64B DPDK Data Mover for Agilex R-Tile]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4127u16 , name : "AR-TK242 [2x100GbE Packet Capture Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4128u16 , name : "AR-GP-PF [General-Purpose Physical Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4129u16 , name : "AR-MGMT-PF [Board-Generic Management Physical Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4130u16 , name : "AR-ARKA-FX2 [Arkville 128B DPDK Data Mover for Agilex]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4131u16 , name : "AR-BD-FX1 [BufferDirect Zero-Bounce Data Motion Endpoint]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4132u16 , name : "AR-TK242 [2x10GbE Packet Capture Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4133u16 , name : "AR-TK242-FX2 [2x100GbE Gen5 Packet Capture Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4134u16 , name : "AR-TK242-FX2 [1x200GbE Gen5 Packet Capture Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4135u16 , name : "AR-P2P-DBG [P2P Debug Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4136u16 , name : "AR-P2P-ATR [P2P Actor Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4137u16 , name : "AR-P2P-UTL [P2P Utility Function]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4138u16 , name : "AR-TK242-FX2 [4x100GbE Gen5 Packet Capture-Replay Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4139u16 , name : "AR-ARKV-FX1 [Arkville 128B DPDK Data Mover for Versal/CPM5]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4140u16 , name : "AR-TK242-V80 [Gen5 PCAP Processor]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4141u16 , name : "AR-TK242-FX2 [8x10GbE Gen5 Packet Capture-Replay Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4142u16 , name : "AR-TK242-FX2 [8x25GbE Gen5 Packet Capture-Replay Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4143u16 , name : "AR-TK242-FX2 [1x400GbE Gen5 Packet Capture-Replay Device]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 4144u16 , name : "AR-ARKSTREAM [Arkville Streaming DMA]" , subsystems : & [] } , Device { vendor_id : 7532u16 , id : 16896u16 , name : "A5PL-E1-10GETI [10 GbE Ethernet Traffic Instrument]" , subsystems : & [] }] }),
        (25460, Vendor { id : 25460u16 , name : "c't Magazin fuer Computertechnik" , devices : & [Device { vendor_id : 25460u16 , id : 26483u16 , name : "GPPCI" , subsystems : & [] }] }),
        (4592, Vendor { id : 4592u16 , name : "Compu-Shack" , devices : & [Device { vendor_id : 4592u16 , id : 16945u16 , name : "FDDI" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 16946u16 , name : "FASTline UTP Quattro" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 16947u16 , name : "FASTline FO" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 16948u16 , name : "FASTline UTP" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 16949u16 , name : "FASTline-II UTP" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 16950u16 , name : "FASTline-II FO" , subsystems : & [] } , Device { vendor_id : 4592u16 , id : 18225u16 , name : "GIGAline" , subsystems : & [] }] }),
        (5858, Vendor { id : 5858u16 , name : "Marvin Test Solutions" , devices : & [Device { vendor_id : 5858u16 , id : 4148u16 , name : "GX1034 Certification and Verification Module PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 4352u16 , name : "GX1100 Arbitrary Waveform and Function Generator PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 4384u16 , name : "GX1120 Arbitrary Waveform and Function Generator PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 4452u16 , name : "GX1164 Multi-Channel Programmable Resistor PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 5682u16 , name : "GX1632e 32-Channel Arbitrary Analog Output PXIe Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 5704u16 , name : "GX1648 64-Channel Analog Output PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 5705u16 , name : "GX1649 64-Channel Arbitrary Analog Output PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 6200u16 , name : "GX1838 Precision DC Source PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 8293u16 , name : "GX2065 Digital Multimeter PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 8704u16 , name : "GTX2200 High Resolution Universal Time Interval Counter PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 8705u16 , name : "GC2200 High Resolution Universal Time Interval Counter PCI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 12548u16 , name : "GX3104 4-Channel 20V, 250mA Per Channel SMU PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 12552u16 , name : "GX3108 4-Channel Source Measure Unit PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 12566u16 , name : "GX3116e 16-Channel DUT Power Supply PXIe Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 13128u16 , name : "GX3348 Multi-Channel DC Source and Switch Matrix Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 13568u16 , name : "GX3500 Digital I/O FLEX FPGA Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 14080u16 , name : "GX3700e Digital I/O FLEX FPGA PXIe Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 14081u16 , name : "GX3700 Digital I/O FLEX FPGA PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 14336u16 , name : "GX3800e Digital I/O FLEX FPGA PXIe Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 18755u16 , name : "GX4943 HellFire/Longbow Serial Communication Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20560u16 , name : "GC5050 Dynamic Digital I/O with Algorithmic Sequencer PCI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20565u16 , name : "GX5055 Dynamic Digital I/O with Pin Electronics PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20611u16 , name : "GX5083 Dynamic Digital I/O with Algorithmic Sequencer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20741u16 , name : "GX5105 High-Speed Digital I/O Timing (Master) Module PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20742u16 , name : "GX5106 High-Speed Digital I/O Cycle Timing (Master) Module PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20817u16 , name : "GX5150/1 High Speed Digital I/O (50/100MHz) PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 20818u16 , name : "GX5152/3 DSR High Speed Digital Stimulus/Response PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21072u16 , name : "GX5250 Digital I/O Streaming PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21120u16 , name : "GX5280 High-Speed Digital I/O PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21136u16 , name : "GX5290 High-Speed Dynamic Digital I/O PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21140u16 , name : "GX5294 High-Speed Digital I/O PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21141u16 , name : "GX5295 Digital I/O with Pin Electronics and PMU PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21142u16 , name : "GX5296 Digital I/O with Pin Electronics and PMU PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 21840u16 , name : "GX5050 Dynamic Digital I/O with Algorithmic Sequencer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 22081u16 , name : "GX5641/2 Bi-directional Differential-TTL I/O PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 22321u16 , name : "GX5731 Digital I/O with 128 TTL Channels and 3x32 Customizable Channels PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 22322u16 , name : "GX5732 224-Channel Static Digital I/O PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 22323u16 , name : "GX5733 Digital I/O with 96 TTL Channels and 32 Customizable Channels PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 22880u16 , name : "GX5961/4 Digital I/O Timing/Sync with Pin Electronics and PMU PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24609u16 , name : "GX6021 20-Channel RF Multiplexer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24674u16 , name : "GX6062 60-Channel RF Multiplexer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24853u16 , name : "GX6115 15-Channel High Current SPDT Form-C Relay PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24869u16 , name : "GX6125 25-Channel High-Density SPDT Form-C Relay PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24888u16 , name : "GX6138 38-Channel SPST Form-A Relay PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24968u16 , name : "GX6188 104-Channel 8x8 Multiplexer/Matrix PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24978u16 , name : "GX6192 16, 16x2 High Frequency, High Density Multiplexer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 24982u16 , name : "GX6196 96-Channel DPST Form-A Relay Carrier PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25174u16 , name : "GX6256 16, 16x2 Low Frequency, High Density Multiplexer PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25188u16 , name : "GX6264 128-Channel Scanner/Multiplexer Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25365u16 , name : "GX6315 45-Channel High Current SPDT Form-C Relay Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25381u16 , name : "GX6325 75-Channel High-Density SPDT Form-C Relay Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25400u16 , name : "GX6338 114-Channel SPST Form-A Relay PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25463u16 , name : "GX6377 Multifunction Switching PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 25476u16 , name : "GX6384 Configurable High-Density Switch Matrix PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 26134u16 , name : "GX6616 6x2:16 Switch Matrix PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 26724u16 , name : "GX6864 Video Switching PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 29696u16 , name : "GX7400 Dual-Output Programmable Power Supply PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 29700u16 , name : "GX7404 Prototyping and Power Interface PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 30583u16 , name : "GX7777 Generic Interface PXI Board" , subsystems : & [] } , Device { vendor_id : 5858u16 , id : 30585u16 , name : "GX7779 Backplane Tester PXI Board" , subsystems : & [] }] }),
        (5281, Vendor { id : 5281u16 , name : "Systembase Co Ltd" , devices : & [Device { vendor_id : 5281u16 , id : 19714u16 , name : "SB16C1052PCI [Multi-2/PCIe RS232]" , subsystems : & [] }] }),
        (5488, Vendor { id : 5488u16 , name : "Teledyne Lecroy" , devices : & [] }),
        (6058, Vendor { id : 6058u16 , name : "Lenovo" , devices : & [Device { vendor_id : 6058u16 , id : 3u16 , name : "LENSE20256GMSP34MEAT2TA" , subsystems : & [] } , Device { vendor_id : 6058u16 , id : 4u16 , name : "LENSE20512GMSP34MEAT2TA" , subsystems : & [] } , Device { vendor_id : 6058u16 , id : 5u16 , name : "LENSE30256GMSP34MEAT3TA" , subsystems : & [] } , Device { vendor_id : 6058u16 , id : 6u16 , name : "LENSE30512GMSP34MEAT3TA" , subsystems : & [] } , Device { vendor_id : 6058u16 , id : 16427u16 , name : "Intel 82599ES 10Gb 2-port Server Adapter X520-2" , subsystems : & [] } , Device { vendor_id : 6058u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] }] }),
        (7809, Vendor { id : 7809u16 , name : "Ramaxel Technology(Shenzhen) Limited" , devices : & [Device { vendor_id : 7809u16 , id : 4611u16 , name : "NVMe SSD Controller UHXXXa series" , subsystems : & [SubSystem { subvendor : 7809u16 , subdevice : 41249u16 , name : "NVMe SSD UHXXXa series U.2 960GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41250u16 , name : "NVMe SSD UHXXXa series U.2 1920GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41251u16 , name : "NVMe SSD UHXXXa series U.2 3840GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41252u16 , name : "NVMe SSD UHXXXa series U.2 7680GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41253u16 , name : "NVMe SSD UHXXXa series U.2 15360GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41489u16 , name : "NVMe SSD UHXXXa series U.2 800GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41490u16 , name : "NVMe SSD UHXXXa series U.2 1600GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41491u16 , name : "NVMe SSD UHXXXa series U.2 3200GB" } , SubSystem { subvendor : 7809u16 , subdevice : 41492u16 , name : "NVMe SSD UHXXXa series U.2 6400GB" } , SubSystem { subvendor : 7809u16 , subdevice : 61731u16 , name : "NVMe SSD TP6500 series U.2 3840GB" }] } , Device { vendor_id : 7809u16 , id : 25094u16 , name : "AM620 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7809u16 , id : 27138u16 , name : "AM6A0 NVMe SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7809u16 , id : 27668u16 , name : "RPEYJ1T24 NVMe SSD (DRAM-less)" , subsystems : & [] }] }),
        (5357, Vendor { id : 5357u16 , name : "DATAKINETICS Ltd" , devices : & [] }),
        (4697, Vendor { id : 4697u16 , name : "Allied Telesis" , devices : & [Device { vendor_id : 4697u16 , id : 9568u16 , name : "AT-2560 Fast Ethernet Adapter (i82557B)" , subsystems : & [] } , Device { vendor_id : 4697u16 , id : 10241u16 , name : "AT-2801FX (RTL-8139)" , subsystems : & [] } , Device { vendor_id : 4697u16 , id : 41239u16 , name : "RTL81xx Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 4697u16 , id : 41246u16 , name : "RTL81xx Fast Ethernet" , subsystems : & [] } , Device { vendor_id : 4697u16 , id : 41248u16 , name : "21x4x DEC-Tulip compatible 10/100 Ethernet" , subsystems : & [] }] }),
        (6938, Vendor { id : 6938u16 , name : "K&F Computing Research Co." , devices : & [Device { vendor_id : 6938u16 , id : 3696u16 , name : "GRAPE" , subsystems : & [] }] }),
        (7671, Vendor { id : 7671u16 , name : "opencpi.org" , devices : & [Device { vendor_id : 7671u16 , id : 1u16 , name : "ml605" , subsystems : & [] } , Device { vendor_id : 7671u16 , id : 2u16 , name : "alst4" , subsystems : & [] } , Device { vendor_id : 7671u16 , id : 3u16 , name : "alst4x" , subsystems : & [] }] }),
        (10776, Vendor { id : 10776u16 , name : "Video Transcode Controller" , devices : & [Device { vendor_id : 10776u16 , id : 10786u16 , name : "Video Transcode Controller" , subsystems : & [] }] }),
        (5181, Vendor { id : 5181u16 , name : "Tamarack Microelectronics Inc" , devices : & [] }),
        (17915, Vendor { id : 17915u16 , name : "Baldor Electric Company" , devices : & [] }),
        (4175, Vendor { id : 4175u16 , name : "Co-time Computer Ltd" , devices : & [] }),
        (4346, Vendor { id : 4346u16 , name : "Truevision" , devices : & [Device { vendor_id : 4346u16 , id : 12u16 , name : "TARGA 1000" , subsystems : & [] }] }),
        (5378, Vendor { id : 5378u16 , name : "MITSUBISHI ELECTRIC LOGISTICS SUPPORT Co Ltd" , devices : & [] }),
        (61790, Vendor { id : 61790u16 , name : "SiFive, Inc." , devices : & [Device { vendor_id : 61790u16 , id : 0u16 , name : "FU740-C000 RISC-V SoC PCI Express x8 to AXI4 Bridge" , subsystems : & [] }] }),
        (4266, Vendor { id : 4266u16 , name : "ACC Microelectronics" , devices : & [Device { vendor_id : 4266u16 , id : 0u16 , name : "ACCM 2188" , subsystems : & [] } , Device { vendor_id : 4266u16 , id : 8273u16 , name : "2051 CPU bridge" , subsystems : & [] } , Device { vendor_id : 4266u16 , id : 22594u16 , name : "2051 ISA bridge" , subsystems : & [] }] }),
        (4130, Vendor { id : 4130u16 , name : "Advanced Micro Devices, Inc. [AMD]" , devices : & [Device { vendor_id : 4130u16 , id : 4352u16 , name : "K8 [Athlon64/Opteron] HyperTransport Technology Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4353u16 , name : "K8 [Athlon64/Opteron] Address Map" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4354u16 , name : "K8 [Athlon64/Opteron] DRAM Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4355u16 , name : "K8 [Athlon64/Opteron] Miscellaneous Control" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4368u16 , name : "Krackan Internal PCIe GPP Bridge 0 to Bus A" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4369u16 , name : "Krackan Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4370u16 , name : "Krackan Internal PCIe GPP Bridge 0 to Bus C" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4371u16 , name : "Krackan PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4374u16 , name : "Krackan PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4375u16 , name : "Krackan PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4386u16 , name : "Krackan Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4387u16 , name : "Krackan IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4608u16 , name : "Family 10h Processor HyperTransport Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4609u16 , name : "Family 10h Processor Address Map" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4610u16 , name : "Family 10h Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4611u16 , name : "Family 10h Processor Miscellaneous Control" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4612u16 , name : "Family 10h Processor Link Control" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4680u16 , name : "Krackan Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4681u16 , name : "Krackan Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4682u16 , name : "Krackan Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4683u16 , name : "Krackan Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4684u16 , name : "Krackan Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4685u16 , name : "Krackan Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4686u16 , name : "Krackan Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4687u16 , name : "Krackan Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4792u16 , name : "Strix Halo Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4793u16 , name : "Strix Halo Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4794u16 , name : "Strix Halo Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4795u16 , name : "Strix Halo Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4796u16 , name : "Strix Halo Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4797u16 , name : "Strix Halo Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4798u16 , name : "Strix Halo Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4799u16 , name : "Strix Halo Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4800u16 , name : "Turin Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4801u16 , name : "Turin Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4802u16 , name : "Turin Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4803u16 , name : "Turin Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4804u16 , name : "Turin Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4805u16 , name : "Turin Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4806u16 , name : "Turin Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4807u16 , name : "Turin Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4864u16 , name : "Family 11h Processor HyperTransport Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4865u16 , name : "Family 11h Processor Address Map" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4866u16 , name : "Family 11h Processor DRAM Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4867u16 , name : "Family 11h Processor Miscellaneous Control" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4868u16 , name : "Family 11h Processor Link Control" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4869u16 , name : "Griffin Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4870u16 , name : "Griffin Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4871u16 , name : "Griffin Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4872u16 , name : "Kaveri Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 4884u16 , name : "Wrestler/Bheem/Ontario/Krishna Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5088u16 , name : "Ariel Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5089u16 , name : "Ariel IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5090u16 , name : "Ariel PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5091u16 , name : "Ariel PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5092u16 , name : "Ariel PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5093u16 , name : "Ariel Internal PCIe GPP Bridge 0 to Bus A" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5094u16 , name : "Ariel Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5095u16 , name : "Ariel SMBus Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5096u16 , name : "Ariel LPC Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5097u16 , name : "Ariel Internal GPU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5098u16 , name : "Ariel HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5099u16 , name : "Ariel HD Audio Coprocessor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5100u16 , name : "Ariel Cryptographic Coprocessor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5101u16 , name : "Ariel USB 3.1 Type C: Gen2 x 1port + DP Alt Mode" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5102u16 , name : "Ariel USB 3.1 Type A: Gen2 x 2 ports" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5103u16 , name : "Ariel ZCN/MP4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5104u16 , name : "Ariel Device 24: Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5105u16 , name : "Ariel Device 24: Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5106u16 , name : "Ariel Device 24: Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5107u16 , name : "Ariel Device 24: Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5108u16 , name : "Ariel Device 24: Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5109u16 , name : "Ariel Device 24: Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5110u16 , name : "Ariel Device 24: Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5111u16 , name : "Ariel Device 24: Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5120u16 , name : "Family 15h (Models 10h-1fh) Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5121u16 , name : "Family 15h (Models 10h-1fh) Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5122u16 , name : "Family 15h (Models 10h-1fh) Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5123u16 , name : "Family 15h (Models 10h-1fh) Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5124u16 , name : "Family 15h (Models 10h-1fh) Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5125u16 , name : "Family 15h (Models 10h-1fh) Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5136u16 , name : "Family 15h (Models 10h-1fh) Processor Root Complex" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" }] } , Device { vendor_id : 4130u16 , id : 5138u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [SubSystem { subvendor : 4130u16 , subdevice : 4660u16 , name : "Trinity A-series APU" }] } , Device { vendor_id : 4130u16 , id : 5139u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5140u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [SubSystem { subvendor : 4130u16 , subdevice : 4660u16 , name : "Trinity A-series APU" }] } , Device { vendor_id : 4130u16 , id : 5141u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5142u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5143u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5144u16 , name : "Family 15h (Models 10h-1fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5145u16 , name : "Family 15h (Models 10h-1fh) I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5146u16 , name : "Family 15h (Models 30h-3fh) Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5147u16 , name : "Family 15h (Models 30h-3fh) Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5148u16 , name : "Family 15h (Models 30h-3fh) Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5149u16 , name : "Family 15h (Models 30h-3fh) Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5150u16 , name : "Family 15h (Models 30h-3fh) Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5151u16 , name : "Family 15h (Models 30h-3fh) Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5154u16 , name : "Family 15h (Models 30h-3fh) Processor Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5155u16 , name : "Family 15h (Models 30h-3fh) I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5156u16 , name : "Family 15h (Models 30h-3fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5157u16 , name : "Kaveri P2P Bridge for GFX PCIe Port [1:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5158u16 , name : "Family 15h (Models 30h-3fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5166u16 , name : "Liverpool Processor HT configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5167u16 , name : "Liverpool Processor Address Maps" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5168u16 , name : "Liverpool Processor DRAM configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5169u16 , name : "Liverpool Processor Misc configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5170u16 , name : "Liverpool Processor PM configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5171u16 , name : "Liverpool Processor NB Performance Monitor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5172u16 , name : "Liverpool Processor SPLL Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5174u16 , name : "Liverpool Processor Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5175u16 , name : "Liverpool I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5176u16 , name : "Liverpool UMI PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5177u16 , name : "Family 16h Processor Functions 5:1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5178u16 , name : "Kingston/Clayton/Gladius/Montego Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5179u16 , name : "Kingston/Clayton/Gladius/Montego P2P Bridge for UMI Link" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5184u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5185u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5186u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5187u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5188u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5189u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5190u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5191u16 , name : "Matisse/Vermeer Data Fabric: Device 18h; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5192u16 , name : "Renoir Device 24: Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5193u16 , name : "Renoir Device 24: Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5194u16 , name : "Renoir Device 24: Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5195u16 , name : "Renoir Device 24: Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5196u16 , name : "Renoir Device 24: Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5197u16 , name : "Renoir Device 24: Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5198u16 , name : "Renoir Device 24: Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5199u16 , name : "Renoir Device 24: Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5200u16 , name : "Family 17h (Models 00h-0fh) Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5201u16 , name : "Family 17h (Models 00h-0fh) I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5202u16 , name : "Family 17h (Models 00h-1fh) PCIe Dummy Host Bridge" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5203u16 , name : "Family 17h (Models 00h-0fh) PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5204u16 , name : "Family 17h (Models 00h-0fh) Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5205u16 , name : "Zeppelin/Renoir PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5206u16 , name : "Family 17h (Models 00h-0fh) Platform Security Processor (PSP) 3.0 Device" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5207u16 , name : "Family 17h (Models 00h-0fh) HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5208u16 , name : "XGMAC 10GbE Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5209u16 , name : "XGMAC 10GbE Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5210u16 , name : "Zeppelin/Raven/Raven2 PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5211u16 , name : "Zeppelin Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5212u16 , name : "Family 17h (Models 00h-0fh) USB 3.0 Host Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5213u16 , name : "Zeppelin Switch Upstream (PCIE SW.US)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5214u16 , name : "Zeppelin Switch Downstream (PCIE SW.DS)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5215u16 , name : "Zeppelin USB 3.0 xHCI Compliant Host Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5216u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5217u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5218u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5219u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5220u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5221u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5222u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5223u16 , name : "Family 17h (Models 00h-0fh) Data Fabric: Device 18h; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5224u16 , name : "Zeppelin Cryptographic Coprocessor NTBCCP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5232u16 , name : "Vega 10 PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5233u16 , name : "Vega 10 PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5248u16 , name : "Starship/Matisse Root Complex" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 7061u16 , name : "H12SSL-i" } , SubSystem { subvendor : 6217u16 , subdevice : 5248u16 , name : "ROME2D32LM3" }] } , Device { vendor_id : 4130u16 , id : 5249u16 , name : "Starship/Matisse IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5250u16 , name : "Starship/Matisse PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5251u16 , name : "Starship/Matisse GPP Bridge" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65529u16 , name : "Gimlet Baseboard" }] } , Device { vendor_id : 4130u16 , id : 5252u16 , name : "Starship/Matisse Internal PCIe GPP Bridge 0 to bus[E:B]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5253u16 , name : "Starship/Matisse Reserved SPP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5254u16 , name : "Starship/Matisse Cryptographic Coprocessor PSPCPP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5255u16 , name : "Starship/Matisse HD Audio Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 39991u16 , name : "X570-A PRO motherboard" }] } , Device { vendor_id : 4130u16 , id : 5256u16 , name : "Starship Reserved SSP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5257u16 , name : "Starship Reserved SSP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5258u16 , name : "Starship/Matisse PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5259u16 , name : "Starship/Matisse Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5260u16 , name : "Starship USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 5593u16 , subdevice : 5212u16 , name : "H12SSL-i" }] } , Device { vendor_id : 4130u16 , id : 5261u16 , name : "Starship/Matisse Switch Upstream (PCIE SW.US)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5262u16 , name : "Starship/Matisse Switch Downstream (PCIE SW.DS)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5263u16 , name : "Starship Reserved SSP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5264u16 , name : "Starship Device 24; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5265u16 , name : "Starship Device 24; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5266u16 , name : "Starship Device 24; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5267u16 , name : "Starship Device 24; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5268u16 , name : "Starship Device 24; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5269u16 , name : "Starship Device 24; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5270u16 , name : "Starship Device 24; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5271u16 , name : "Starship Device 24; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5272u16 , name : "Starship/Matisse PTDMA" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5273u16 , name : "Starship/Matisse NVMe" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5274u16 , name : "Starship PCIe GPP Bridge [1:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5275u16 , name : "Starship Reserved SSP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5276u16 , name : "Matisse USB 3.0 Host Controller" , subsystems : & [SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" }] } , Device { vendor_id : 4130u16 , id : 5277u16 , name : "Vangogh CVIP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5278u16 , name : "Genoa/Bergamo IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5279u16 , name : "Genoa/Bergamo Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5284u16 , name : "Genoa/Bergamo Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5285u16 , name : "Genoa/Bergamo GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5286u16 , name : "Genoa/Bergamo RCEC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5287u16 , name : "Genoa/Bergamo Internal PCIe GPP Bridge to Bus [D:B]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5290u16 , name : "Genoa/Bergamo GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5291u16 , name : "Genoa/Bergamo GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5292u16 , name : "Genoa/Bergamo Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5293u16 , name : "Genoa/Bergamo Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5294u16 , name : "Genoa/Bergamo Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5295u16 , name : "Genoa/Bergamo Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5296u16 , name : "Genoa/Bergamo Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5297u16 , name : "Genoa/Bergamo Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5298u16 , name : "Genoa/Bergamo Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5299u16 , name : "Genoa/Bergamo Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5300u16 , name : "Genoa/Bergamo Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5301u16 , name : "Family 17h-19h PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5302u16 , name : "Family 17h-19h IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5303u16 , name : "Family 17h-19h PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5304u16 , name : "Family 17h-19h PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5305u16 , name : "Family 17h-19h Internal PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5306u16 , name : "Family 17h-19h PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5313u16 , name : "Secondary vNTB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5322u16 , name : "Genoa CCP/PSP 4.0 Device" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5325u16 , name : "Family 19h USB4/Thunderbolt PCIe tunnel" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5336u16 , name : "Raphael/Granite Ridge Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5337u16 , name : "Raphael/Granite Ridge IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5338u16 , name : "Raphael/Granite Ridge Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5339u16 , name : "Raphael/Granite Ridge GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5340u16 , name : "SDXI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5341u16 , name : "Raphael/Granite Ridge Internal GPP Bridge to Bus [C:A]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5342u16 , name : "Raphael/Granite Ridge PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5344u16 , name : "Raphael/Granite Ridge Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5345u16 , name : "Raphael/Granite Ridge Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5346u16 , name : "Raphael/Granite Ridge Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5347u16 , name : "Raphael/Granite Ridge Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5348u16 , name : "Raphael/Granite Ridge Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5349u16 , name : "Raphael/Granite Ridge Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5350u16 , name : "Raphael/Granite Ridge Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5351u16 , name : "Raphael/Granite Ridge Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5352u16 , name : "Phoenix Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5353u16 , name : "Phoenix IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5354u16 , name : "Phoenix Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5355u16 , name : "Phoenix Internal GPP Bridge to Bus [C:A]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5356u16 , name : "Phoenix Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5357u16 , name : "Phoenix GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5358u16 , name : "Phoenix GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5359u16 , name : "Family 19h USB4/Thunderbolt PCIe tunnel" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5360u16 , name : "Phoenix Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5361u16 , name : "Phoenix Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5362u16 , name : "Phoenix Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5363u16 , name : "Phoenix Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5364u16 , name : "Phoenix Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5365u16 , name : "Phoenix Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5366u16 , name : "Phoenix Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5367u16 , name : "Phoenix Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5378u16 , name : "AMD IPU Device" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5383u16 , name : "Strix/Strix Halo Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5384u16 , name : "Strix/Strix Halo IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5385u16 , name : "Strix/Strix Halo Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5386u16 , name : "Strix/Strix Halo PCIe USB4 Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5387u16 , name : "Strix/Strix Halo GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5388u16 , name : "Strix/Strix Halo Internal GPP Bridge to Bus [C:A]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5389u16 , name : "Strix/Strix Halo PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5392u16 , name : "Family 14h Processor Root Complex" , subsystems : & [SubSystem { subvendor : 5963u16 , subdevice : 4097u16 , name : "PURE Fusion Mini" }] } , Device { vendor_id : 4130u16 , id : 5394u16 , name : "Family 14h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5395u16 , name : "Family 14h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5396u16 , name : "Family 14h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5397u16 , name : "Family 14h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5398u16 , name : "Family 14h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5404u16 , name : "USB4 Router 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5405u16 , name : "USB4 Router 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5424u16 , name : "Family 16h Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5425u16 , name : "Family 16h Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5426u16 , name : "Family 16h Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5427u16 , name : "Family 16h Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5428u16 , name : "Family 16h Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5429u16 , name : "Family 16h Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5430u16 , name : "Family 16h Processor Root Complex" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6217u16 , subdevice : 5430u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 5431u16 , name : "Kabini/Mullins PSP-Platform Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5432u16 , name : "Family 16h Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5433u16 , name : "Kabini P2P Bridge for PCIe Ports[4:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5434u16 , name : "Turin Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5435u16 , name : "Turin IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5436u16 , name : "Turin RCEC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5437u16 , name : "Turin PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5438u16 , name : "Turin GPP Bridge" , subsystems : & [SubSystem { subvendor : 478u16 , subdevice : 65528u16 , name : "Cosmo Baseboard" }] } , Device { vendor_id : 4130u16 , id : 5440u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky HT Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5441u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky Address Maps" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5442u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky DRAM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5443u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky Miscellaneous Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5444u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky PM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5445u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky NB Performance Monitor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5446u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5447u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5448u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky UMI PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5449u16 , name : "Kryptos/Cato/Garfield/Garfield+ P2P Bridge for PCIe Port [3:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5450u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky Audio Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5451u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5453u16 , name : "Kryptos/Cato/Garfield/Garfield+/Arlene/Pooky/Anubis HDMI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5455u16 , name : "Anubis Audio Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5456u16 , name : "Garfield+/Arlene/Pooky/Anubis SPLL Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5459u16 , name : "Arlene/Pooky P2P Bridge for PCIE (3:0)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5460u16 , name : "Turin GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5461u16 , name : "Turin Internal PCIe GPP Bridge to Bus [D:C]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5462u16 , name : "Turin PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5463u16 , name : "Turin USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5465u16 , name : "Turin Secondary PCIe Dummy Function" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5467u16 , name : "Anubis Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5468u16 , name : "Anubis IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5469u16 , name : "Anubis UMI PCIe Dummy Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5470u16 , name : "Anubis P2P Bridge for PCIe Ports [4:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5472u16 , name : "Anubis Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5478u16 , name : "Family 16h (Models 30h-3fh) Processor Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5479u16 , name : "Mullins IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5483u16 , name : "Family 16h (Models 30h-3fh) Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5486u16 , name : "Turin CCP/ASP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5488u16 , name : "Family 15h (Models 60h-6fh) Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5489u16 , name : "Family 15h (Models 60h-6fh) Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5490u16 , name : "Family 15h (Models 60h-6fh) Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5491u16 , name : "Family 15h (Models 60h-6fh) Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5492u16 , name : "Family 15h (Models 60h-6fh) Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5493u16 , name : "Family 15h (Models 60h-6fh) Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5494u16 , name : "Family 15h (Models 60h-6fh) Processor Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5495u16 , name : "Family 15h (Models 60h-6fh) I/O Memory Management Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5496u16 , name : "Carrizo Platform Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5497u16 , name : "Carrizo Audio Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5498u16 , name : "Family 15h (Models 60h-6fh) Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5499u16 , name : "Family 15h (Models 60h-6fh) Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5500u16 , name : "Family 15h (Models 60h-6fh) Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5501u16 , name : "Carrizo Audio Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5502u16 , name : "Carrizo Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5504u16 , name : "Family 16h (Models 30h-3fh) Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5505u16 , name : "Family 16h (Models 30h-3fh) Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5506u16 , name : "Family 16h (Models 30h-3fh) Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5507u16 , name : "Family 16h (Models 30h-3fh) Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5508u16 , name : "Family 16h (Models 30h-3fh) Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5509u16 , name : "Family 16h (Models 30h-3fh) Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5511u16 , name : "Strix Halo USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5512u16 , name : "Strix Halo USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5513u16 , name : "Strix Halo USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5515u16 , name : "Strix Halo USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5517u16 , name : "Strix Halo USB4 Host Router" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5518u16 , name : "Strix Halo USB4 Host Router" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5520u16 , name : "Amur/Nolan HT Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5521u16 , name : "Amur/Nolan Address Maps" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5522u16 , name : "Amur/Nolan DRAM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5523u16 , name : "Amur/Nolan Miscellaneous Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5524u16 , name : "Amur/Nolan PM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5525u16 , name : "Amur/Nolan NB Performance Monitor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5526u16 , name : "Amur/Nolan Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5527u16 , name : "Amur/Nolan IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5528u16 , name : "Amur/Nolan Platform Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5529u16 , name : "Amur/Nolan PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5533u16 , name : "Amur Function 6: Gasket" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5552u16 , name : "Stoney HT Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5553u16 , name : "Stoney Address Maps" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5554u16 , name : "Stoney DRAM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5555u16 , name : "Stoney Miscellaneous Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5556u16 , name : "Stoney PM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5557u16 , name : "Stoney NB Performance Monitor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5558u16 , name : "Raphael/Granite Ridge USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5559u16 , name : "Raphael/Granite Ridge USB 3.1 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5560u16 , name : "Raphael/Granite Ridge USB 2.0 xHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5564u16 , name : "Stoney PCIe [GFX,GPP] Bridge [4:0]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5566u16 , name : "Stoney Audio Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5572u16 , name : "Phoenix USB4/Thunderbolt NHI controller #1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5573u16 , name : "Phoenix USB4/Thunderbolt NHI controller #2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5575u16 , name : "Phoenix CCP/PSP 3.0 Device" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5584u16 , name : "Raven/Raven2 Root Complex" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME B450M-A Motherboard" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5585u16 , name : "Raven/Raven2 IOMMU" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME B450M-A Motherboard" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5586u16 , name : "Raven/Raven2 PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5587u16 , name : "Raven/Raven2 PCIe GPP Bridge [6:0]" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5588u16 , name : "FireFlight USB 3.1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5589u16 , name : "FireFlight USB 3.1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5590u16 , name : "Rembrandt USB4 XHCI controller #5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5591u16 , name : "Rembrandt USB4 XHCI controller #6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5594u16 , name : "Raven/Raven2 PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5595u16 , name : "Raven/Raven2 Internal PCIe GPP Bridge 0 to Bus A" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5596u16 , name : "Raven/Raven2 Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5598u16 , name : "Raven/Raven2/FireFlight HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5599u16 , name : "Raven/Raven2/FireFlight/Renoir/Cezanne Platform Security Processor" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5600u16 , name : "Raven USB 3.1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5601u16 , name : "Raven USB 3.1" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5602u16 , name : "Audio Coprocessor" , subsystems : & [SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5603u16 , name : "Ryzen HD Audio Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4156u16 , subdevice : 35607u16 , name : "ProBook 445 G9/455 G9" } , SubSystem { subvendor : 4163u16 , subdevice : 34503u16 , name : "PRIME B450M-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" }] } , Device { vendor_id : 4130u16 , id : 5604u16 , name : "Sensor Fusion Hub" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5605u16 , name : "Raven2 USB 3.1" , subsystems : & [SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5606u16 , name : "Raven/Raven2/Renoir Non-Sensor Fusion Hub KMDF driver" , subsystems : & [SubSystem { subvendor : 4130u16 , subdevice : 5604u16 , name : "Raven/Raven2/Renoir Sensor Fusion Hub" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 5608u16 , name : "Raven/Raven2 Device 24: Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5609u16 , name : "Raven/Raven2 Device 24: Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5610u16 , name : "Raven/Raven2 Device 24: Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5611u16 , name : "Raven/Raven2 Device 24: Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5612u16 , name : "Raven/Raven2 Device 24: Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5613u16 , name : "Raven/Raven2 Device 24: Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5614u16 , name : "Raven/Raven2 Device 24: Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5615u16 , name : "Raven/Raven2 Device 24: Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5616u16 , name : "FireFlight Device 24: Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5617u16 , name : "FireFlight Device 24: Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5618u16 , name : "FireFlight Device 24: Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5619u16 , name : "FireFlight Device 24: Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5620u16 , name : "FireFlight Device 24: Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5621u16 , name : "FireFlight Device 24: Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5622u16 , name : "FireFlight Device 24: Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5623u16 , name : "FireFlight Device 24: Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5624u16 , name : "FireFlight Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5625u16 , name : "FireFlight IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5626u16 , name : "FireFlight PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5627u16 , name : "FireFlight PCIe GPP Bride 3:0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5628u16 , name : "FireFlight PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5629u16 , name : "FireFlight Internal PCIe GPP Bridge 0 to Bus A" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5630u16 , name : "FireFlight Internal PCIe GPP Bridge 0 to Bus B" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5631u16 , name : "FireFlight Bus A; Device 0: Function 0: Internal GPU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5632u16 , name : "Family 15h Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5633u16 , name : "Family 15h Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5634u16 , name : "Family 15h Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5635u16 , name : "Family 15h Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5636u16 , name : "Family 15h Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5637u16 , name : "Family 15h Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5638u16 , name : "Arden Security Processor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5640u16 , name : "Arden Device 18h: Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5641u16 , name : "Arden Device 18h: Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5642u16 , name : "Arden Device 18h: Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5643u16 , name : "Arden Device 18h: Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5644u16 , name : "Arden Device 18h: Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5645u16 , name : "Arden Device 18h: Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5646u16 , name : "Arden Device 18h: Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5647u16 , name : "Arden Device 18h: Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5658u16 , name : "Rembrandt USB4 XHCI controller #1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5659u16 , name : "Rembrandt USB4 XHCI controller #2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5660u16 , name : "Rembrandt USB4 XHCI controller #7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5661u16 , name : "Rembrandt USB4 XHCI controller #3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5662u16 , name : "Rembrandt USB4 XHCI controller #4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5663u16 , name : "Rembrandt USB4 XHCI controller #8" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5664u16 , name : "Anubis HT Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5665u16 , name : "Anubis Address Maps" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5666u16 , name : "Anubis DRAM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5667u16 , name : "Anubis Miscellaneous Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5668u16 , name : "Anubis PM Configuration" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5669u16 , name : "Anubis NB Performance Monitor" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5670u16 , name : "Arden Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5671u16 , name : "Arden IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5672u16 , name : "Arden PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5673u16 , name : "Arden PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5674u16 , name : "Arden Internal PCIe GPP Bridge 0 to bus X" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5675u16 , name : "Arden PCIe Non-Transparent Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5676u16 , name : "VanGogh USB2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5678u16 , name : "Rembrandt USB4/Thunderbolt NHI controller #1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5679u16 , name : "Rembrandt USB4/Thunderbolt NHI controller #2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5680u16 , name : "Renoir/Cezanne Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5681u16 , name : "Renoir/Cezanne IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5682u16 , name : "Renoir PCIe Dummy Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5683u16 , name : "Renoir PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5684u16 , name : "Renoir/Cezanne PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5685u16 , name : "Renoir Internal PCIe GPP Bridge to Bus" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5687u16 , name : "Renoir HD Audio Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5689u16 , name : "Renoir/Cezanne USB 3.1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5690u16 , name : "VanGogh USB0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5691u16 , name : "VanGogh USB1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5692u16 , name : "VanGogh SecUSB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5693u16 , name : "VanGogh SecureFunction" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5694u16 , name : "VanGogh HSP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5697u16 , name : "Renoir 10GbE Controller Port 0 (XGBE0/1)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5698u16 , name : "Renoir WLAN" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5699u16 , name : "Renoir BT" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5700u16 , name : "Renoir I2S" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5701u16 , name : "VanGogh Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5702u16 , name : "VanGogh IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5703u16 , name : "VanGogh PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5704u16 , name : "VanGogh Internal PCIe GPP Bridge to Bus" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5705u16 , name : "Family 19h PSP/CCP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5706u16 , name : "Sensor Fusion Hub" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5707u16 , name : "Non-Sensor Fusion Hub" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5711u16 , name : "Milan IOMMU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5712u16 , name : "Milan Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5713u16 , name : "Milan Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5714u16 , name : "Milan Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5715u16 , name : "Milan Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5716u16 , name : "Milan Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5717u16 , name : "Milan Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5718u16 , name : "Milan Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5719u16 , name : "Milan Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5728u16 , name : "VanGogh Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5729u16 , name : "VanGogh Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5730u16 , name : "VanGogh Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5731u16 , name : "VanGogh Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5732u16 , name : "VanGogh Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5733u16 , name : "VanGogh Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5734u16 , name : "VanGogh Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5735u16 , name : "VanGogh Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5736u16 , name : "Pink Sardine USB4/Thunderbolt NHI controller #1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5737u16 , name : "Pink Sardine USB4/Thunderbolt NHI controller #2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5738u16 , name : "Cezanne Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5739u16 , name : "Cezanne Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5740u16 , name : "Cezanne Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5741u16 , name : "Cezanne Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5742u16 , name : "Cezanne Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5743u16 , name : "Cezanne Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5744u16 , name : "Cezanne Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5745u16 , name : "Cezanne Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5753u16 , name : "Rembrandt Data Fabric: Device 18h; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5754u16 , name : "Rembrandt Data Fabric: Device 18h; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5755u16 , name : "Rembrandt Data Fabric: Device 18h; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5756u16 , name : "Rembrandt Data Fabric: Device 18h; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5757u16 , name : "Rembrandt Data Fabric: Device 18h; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5758u16 , name : "Rembrandt Data Fabric: Device 18h; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5759u16 , name : "Rembrandt Data Fabric: Device 18h; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5760u16 , name : "Rembrandt Data Fabric: Device 18h; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5880u16 , name : "Strix Data Fabric; Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5881u16 , name : "Strix Data Fabric; Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5882u16 , name : "Strix Data Fabric; Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5883u16 , name : "Strix Data Fabric; Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5884u16 , name : "Strix Data Fabric; Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5885u16 , name : "Strix Data Fabric; Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5886u16 , name : "Strix Data Fabric; Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5887u16 , name : "Strix Data Fabric; Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5888u16 , name : "Family 12h/14h Processor Function 0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5889u16 , name : "Family 12h/14h Processor Function 1" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5890u16 , name : "Family 12h/14h Processor Function 2" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5891u16 , name : "Family 12h/14h Processor Function 3" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5892u16 , name : "Family 12h/14h Processor Function 4" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5893u16 , name : "Family 12h Processor Root Complex" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5894u16 , name : "Llano P2P Bridge to external GPU" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5895u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5896u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5897u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5898u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5899u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5900u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5901u16 , name : "Family 12h Processor Root Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5910u16 , name : "Family 12h/14h Processor Function 5" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5912u16 , name : "Family 12h/14h Processor Function 6" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 5913u16 , name : "Family 12h/14h Processor Function 7" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 6112u16 , name : "Strix/Krackan/Strix Halo CCP/ASP" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 6128u16 , name : "Strix/Krackan/Strix Halo Neural Processing Unit" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8192u16 , name : "79C97x [PCnet32 LANCE]" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 8192u16 , name : "NetFinity 10/100 Fast Ethernet" } , SubSystem { subvendor : 4130u16 , subdevice : 8192u16 , name : "PCnet - Fast 79C971" } , SubSystem { subvendor : 4156u16 , subdevice : 4172u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4196u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4197u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4204u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4206u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 4330u16 , name : "Ethernet with LAN remote power Adapter" } , SubSystem { subvendor : 4371u16 , subdevice : 4640u16 , name : "EN1220 10/100 Fast Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9296u16 , name : "AT-2450 10/100 Fast Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9300u16 , name : "AT-2450v4 10Mb Ethernet Adapter" } , SubSystem { subvendor : 4697u16 , subdevice : 9984u16 , name : "AT-2700TX 10/100 Fast Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9985u16 , name : "AT-2700FX 100Mb Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9986u16 , name : "AT-2700FTX (AM79C976KD [PCnet-PRO] chipset) 10/100 Mb Fiber/Copper Fast Ethernet" } , SubSystem { subvendor : 4697u16 , subdevice : 9987u16 , name : "AT-2701FX" } , SubSystem { subvendor : 4697u16 , subdevice : 9988u16 , name : "AT-2701FTX 10/100 Mb Fiber/Copper Fast Ethernet" } , SubSystem { subvendor : 19539u16 , subdevice : 4096u16 , name : "CC7/CR7/CP7/VC7/VP7/VR7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4112u16 , name : "CP5/CR6 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4128u16 , name : "VR6 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4144u16 , name : "PC5 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4160u16 , name : "CL7 mainboard" } , SubSystem { subvendor : 19539u16 , subdevice : 4192u16 , name : "PC7 mainboard" }] } , Device { vendor_id : 4130u16 , id : 8193u16 , name : "Am79C978 PCnet Home (HomePNA) 1/10 PCI Ethernet Adapter [Am79C971 PHY]" , subsystems : & [SubSystem { subvendor : 4242u16 , subdevice : 2680u16 , name : "Multimedia Home Network Adapter" } , SubSystem { subvendor : 5736u16 , subdevice : 665u16 , name : "ActionLink Home Network Adapter" }] } , Device { vendor_id : 4130u16 , id : 8195u16 , name : "Am 1771 MBW [Alchemy]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8224u16 , name : "AM53/79C974 [PC-SCSI]" , subsystems : & [SubSystem { subvendor : 6900u16 , subdevice : 4352u16 , name : "QEMU Virtual Machine" }] } , Device { vendor_id : 4130u16 , id : 8256u16 , name : "79c974" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8320u16 , name : "CS5536 [Geode companion] Host Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8321u16 , name : "Geode LX Video" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8322u16 , name : "Geode LX AES Security Block" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8335u16 , name : "CS5536 GeodeLink PCI South Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8336u16 , name : "CS5536 [Geode companion] ISA" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8337u16 , name : "CS5536 [Geode companion] FLASH" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8339u16 , name : "CS5536 [Geode companion] Audio" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8340u16 , name : "CS5536 [Geode companion] OHC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8341u16 , name : "CS5536 [Geode companion] EHC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8342u16 , name : "CS5536 [Geode companion] UDC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8343u16 , name : "CS5536 [Geode companion] UOC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 8346u16 , name : "CS5536 [Geode companion] IDE" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 9765u16 , name : "Am79C973 [Lance/PCI PCNet/32]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 9767u16 , name : "Am79C975 [Lance/PCI PCNet/32]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 12288u16 , name : "ELanSC520 Microcontroller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17312u16 , name : "Hudson PCI to PCI bridge (PCIE port 0)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17313u16 , name : "Hudson PCI to PCI bridge (PCIE port 1)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17314u16 , name : "Hudson PCI to PCI bridge (PCIE port 2)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17315u16 , name : "Hudson PCI to PCI bridge (PCIE port 3)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17328u16 , name : "X370 Series Chipset PCIe Upstream Port" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 17350u16 , name : "Fatal1ty X370 Professional Gaming" }] } , Device { vendor_id : 4130u16 , id : 17329u16 , name : "X399 Series Chipset PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17332u16 , name : "300 Series Chipset PCIe Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17333u16 , name : "X370 Series Chipset SATA Controller" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 17352u16 , name : "Fatal1ty X370 Professional Gaming" }] } , Device { vendor_id : 4130u16 , id : 17334u16 , name : "X399 Series Chipset SATA Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17335u16 , name : "300 Series Chipset SATA Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17336u16 , name : "A320 Chipset SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17337u16 , name : "X370 Series Chipset USB 3.1 xHCI Controller" , subsystems : & [SubSystem { subvendor : 6217u16 , subdevice : 17360u16 , name : "Fatal1ty X370 Professional Gaming" }] } , Device { vendor_id : 4130u16 , id : 17338u16 , name : "X399 Series Chipset USB 3.1 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17339u16 , name : "300 Series Chipset USB 3.1 xHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17340u16 , name : "A320 USB 3.1 XHCI Host Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17350u16 , name : "400 Series Chipset PCIe Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17351u16 , name : "400 Series Chipset PCIe Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17352u16 , name : "400 Series Chipset SATA Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17365u16 , name : "400 Series Chipset USB 3.1 xHCI Compliant Host Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17385u16 , name : "500 Series Chipset Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17386u16 , name : "500 Series Chipset Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17387u16 , name : "500 Series Chipset SATA Controller" , subsystems : & [SubSystem { subvendor : 6945u16 , subdevice : 4194u16 , name : "ASM1062 Serial ATA Controller" }] } , Device { vendor_id : 4130u16 , id : 17388u16 , name : "A520 Series Chipset USB 3.1 XHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17390u16 , name : "500 Series Chipset USB 3.1 XHCI Controller" , subsystems : & [SubSystem { subvendor : 6945u16 , subdevice : 4418u16 , name : "ASM1042A USB 3.0 Host Controller" }] } , Device { vendor_id : 4130u16 , id : 17396u16 , name : "600 Series Chipset PCIe Switch Upstream Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17397u16 , name : "600 Series Chipset PCIe Switch Downstream Port" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17398u16 , name : "600 Series Chipset SATA Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17399u16 , name : "600 Series Chipset USB 3.2 Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17401u16 , name : "B665 Series Chipset USB 3.1 Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17402u16 , name : "A620 Series Chipset USB 3.1 Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17404u16 , name : "800 Series Chipset USB 3.x XHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 17405u16 , name : "800 Series Chipset USB 3.x XHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 22435u16 , name : "Matisse PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 22436u16 , name : "Matisse PCIe GPP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 22445u16 , name : "Matisse Switch Upstream" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28678u16 , name : "AMD-751 [Irongate] System Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28679u16 , name : "AMD-751 [Irongate] AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28682u16 , name : "AMD-IGR4 AGP Host to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28683u16 , name : "AMD-IGR4 PCI to PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28684u16 , name : "AMD-760 MP [IGD4-2P] System Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28685u16 , name : "AMD-760 MP [IGD4-2P] AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28686u16 , name : "AMD-760 [IGD4-1P] System Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 28687u16 , name : "AMD-760 [IGD4-1P] AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29696u16 , name : "AMD-755 [Cobra] ISA" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29697u16 , name : "AMD-755 [Cobra] IDE" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29699u16 , name : "AMD-755 [Cobra] ACPI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29700u16 , name : "AMD-755 [Cobra] USB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29704u16 , name : "AMD-756 [Viper] ISA" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29705u16 , name : "AMD-756 [Viper] IDE" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29707u16 , name : "AMD-756 [Viper] ACPI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29708u16 , name : "AMD-756 [Viper] USB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29712u16 , name : "AMD-766 [ViperPlus] ISA" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29713u16 , name : "AMD-766 [ViperPlus] IDE" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29715u16 , name : "AMD-766 [ViperPlus] ACPI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29716u16 , name : "AMD-766 [ViperPlus] USB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29760u16 , name : "AMD-768 [Opus] ISA" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32836u16 , name : "A7M-D Mainboard" }] } , Device { vendor_id : 4130u16 , id : 29761u16 , name : "AMD-768 [Opus] IDE" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29763u16 , name : "AMD-768 [Opus] ACPI" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 32836u16 , name : "A7M-D Mainboard" }] } , Device { vendor_id : 4130u16 , id : 29765u16 , name : "AMD-768 [Opus] Audio" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29766u16 , name : "AMD-768 [Opus] MC97 Modem" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29768u16 , name : "AMD-768 [Opus] PCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29769u16 , name : "AMD-768 [Opus] USB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29776u16 , name : "AMD-8131 PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29777u16 , name : "AMD-8131 PCI-X IOAPIC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29780u16 , name : "AMD-8151 System Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29781u16 , name : "AMD-8151 AGP Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29784u16 , name : "AMD-8132 PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29785u16 , name : "AMD-8132 PCI-X IOAPIC" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29792u16 , name : "AMD-8111 PCI" , subsystems : & [SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29793u16 , name : "AMD-8111 USB" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29794u16 , name : "AMD-8111 Ethernet" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29795u16 , name : "AMD-8111 USB EHCI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29796u16 , name : "AMD-8111 USB OHCI" , subsystems : & [SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29800u16 , name : "AMD-8111 LPC" , subsystems : & [SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29801u16 , name : "AMD-8111 IDE" , subsystems : & [SubSystem { subvendor : 4130u16 , subdevice : 11136u16 , name : "AMD-8111 IDE [Quartet]" } , SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29802u16 , name : "AMD-8111 SMBus 2.0" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 29803u16 , name : "AMD-8111 ACPI" , subsystems : & [SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29805u16 , name : "AMD-8111 AC97 Audio" , subsystems : & [SubSystem { subvendor : 5663u16 , subdevice : 12311u16 , name : "HDAMB" }] } , Device { vendor_id : 4130u16 , id : 29806u16 , name : "AMD-8111 MC97 Modem" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30059u16 , name : "AMD-8111 ACPI" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30720u16 , name : "FCH SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30721u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5771u16 , name : "ProBook 4535s Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30721u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30722u16 , name : "FCH SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30723u16 , name : "FCH SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30724u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" }] } , Device { vendor_id : 4130u16 , id : 30725u16 , name : "FCH SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30726u16 , name : "FCH SD Flash Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30727u16 , name : "FCH USB OHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30727u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30728u16 , name : "FCH USB EHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30728u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30729u16 , name : "FCH USB OHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" }] } , Device { vendor_id : 4130u16 , id : 30730u16 , name : "Kabini/Mullins SATA Raid/AHCI Mode (DotHill driver)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30731u16 , name : "FCH SMBus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30731u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30732u16 , name : "FCH IDE Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30733u16 , name : "FCH Azalia Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 33860u16 , name : "F2A85-M Series" } , SubSystem { subvendor : 4163u16 , subdevice : 34166u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 34962u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30734u16 , name : "FCH LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30734u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30735u16 , name : "FCH PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30738u16 , name : "FCH USB XHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30739u16 , name : "FCH SD Flash Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30740u16 , name : "FCH USB XHCI Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6478u16 , name : "ProBook 455 G1 Notebook" } , SubSystem { subvendor : 4156u16 , subdevice : 6533u16 , name : "Pavilion 17-e163sg Notebook PC" } , SubSystem { subvendor : 4163u16 , subdevice : 34339u16 , name : "AM1I-A Motherboard" } , SubSystem { subvendor : 6058u16 , subdevice : 14728u16 , name : "Z50-75" } , SubSystem { subvendor : 6217u16 , subdevice : 30740u16 , name : "QC5000-ITX/PH" }] } , Device { vendor_id : 4130u16 , id : 30976u16 , name : "FCH SATA Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30977u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 30977u16 , name : "H12SSL-i" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 30978u16 , name : "FCH SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30979u16 , name : "FCH SATA Controller [RAID mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30980u16 , name : "FCH SATA Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30982u16 , name : "FCH SD Flash Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30984u16 , name : "FCH USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30987u16 , name : "FCH SMBus Controller" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4156u16 , subdevice : 35607u16 , name : "ProBook 445 G9/455 G9" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 30987u16 , name : "H12SSL-i" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 30990u16 , name : "FCH LPC Bridge" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 34325u16 , name : "Pavilion Laptop 15-cw1xxx" } , SubSystem { subvendor : 4163u16 , subdevice : 34667u16 , name : "PRIME B450M-A Motherboard" } , SubSystem { subvendor : 5218u16 , subdevice : 31799u16 , name : "X570-A PRO motherboard" } , SubSystem { subvendor : 5593u16 , subdevice : 30990u16 , name : "H12SSL-i" } , SubSystem { subvendor : 6058u16 , subdevice : 20772u16 , name : "ThinkPad E595" } , SubSystem { subvendor : 59984u16 , subdevice : 52761u16 , name : "mCOM10-L1900" }] } , Device { vendor_id : 4130u16 , id : 30991u16 , name : "FCH PCI Bridge" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 30996u16 , name : "FCH USB XHCI Controller" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38400u16 , name : "RS780 Host Bridge" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33518u16 , name : "M378A-CM Motherboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33521u16 , name : "M3A78-EH Motherboard" }] } , Device { vendor_id : 4130u16 , id : 38401u16 , name : "RS880 Host Bridge" , subsystems : & [SubSystem { subvendor : 4121u16 , subdevice : 8480u16 , name : "A785GM-M" } , SubSystem { subvendor : 4136u16 , subdevice : 1075u16 , name : "Optiplex 580" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" } , SubSystem { subvendor : 4163u16 , subdevice : 33698u16 , name : "M4A785-M Mainboard" } , SubSystem { subvendor : 4163u16 , subdevice : 33854u16 , name : "M5A88-V EVO" }] } , Device { vendor_id : 4130u16 , id : 38402u16 , name : "RS780/RS880 PCI to PCI bridge (int gfx)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38403u16 , name : "RS780 PCI to PCI bridge (ext gfx port 0)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1075u16 , name : "Optiplex 580" } , SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" }] } , Device { vendor_id : 4130u16 , id : 38404u16 , name : "RS780/RS880 PCI to PCI bridge (PCIE port 0)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1075u16 , name : "OptiPlex 580" }] } , Device { vendor_id : 4130u16 , id : 38405u16 , name : "RS780/RS880 PCI to PCI bridge (PCIE port 1)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38406u16 , name : "RS780 PCI to PCI bridge (PCIE port 2)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 5641u16 , name : "ProLiant MicroServer N36L" }] } , Device { vendor_id : 4130u16 , id : 38407u16 , name : "RS780/RS880 PCI to PCI bridge (PCIE port 3)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38408u16 , name : "RS780/RS880 PCI to PCI bridge (PCIE port 4)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 1075u16 , name : "OptiPlex 580" }] } , Device { vendor_id : 4130u16 , id : 38409u16 , name : "RS780/RS880 PCI to PCI bridge (PCIE port 5)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38410u16 , name : "RS780 PCI to PCI bridge (NB-SB link)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 38411u16 , name : "RS780 PCI to PCI bridge (ext gfx port 1)" , subsystems : & [] } , Device { vendor_id : 4130u16 , id : 45056u16 , name : "RAID Bottom Device" , subsystems : & [] }] }),
        (4880, Vendor { id : 4880u16 , name : "Gespac" , devices : & [] }),
        (4713, Vendor { id : 4713u16 , name : "Thales" , devices : & [Device { vendor_id : 4713u16 , id : 179u16 , name : "5G Data Card [Cinterion MV31-W]" , subsystems : & [] }] }),
        (4215, Vendor { id : 4215u16 , name : "QLogic Corp." , devices : & [Device { vendor_id : 4215u16 , id : 4118u16 , name : "ISP10160 Single Channel Ultra3 SCSI Processor" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 4128u16 , name : "ISP1020/1040 Fast-wide SCSI" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 4130u16 , name : "ISP1022 Fast-wide SCSI" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 4224u16 , name : "ISP1080 SCSI Host Adapter" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "QLA1080" }] } , Device { vendor_id : 4215u16 , id : 4630u16 , name : "ISP12160 Dual Channel Ultra3 SCSI Processor" , subsystems : & [SubSystem { subvendor : 4126u16 , subdevice : 33905u16 , name : "QLA12160 on AMI MegaRAID" } , SubSystem { subvendor : 4126u16 , subdevice : 33939u16 , name : "QLA12160 on AMI MegaRAID" }] } , Device { vendor_id : 4215u16 , id : 4672u16 , name : "ISP1240 SCSI Host Adapter" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 4736u16 , name : "ISP1280 SCSI Host Adapter" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 5684u16 , name : "FastLinQ QL45000 Series 40GbE Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 58609u16 , name : "FastLinQ QL45212H 40GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58610u16 , name : "FastLinQ QL45211H 40GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58611u16 , name : "FastLinQ QL45412H 40GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58612u16 , name : "FastLinQ QL45411H 40GbE Adapter" }] } , Device { vendor_id : 4215u16 , id : 5700u16 , name : "FastLinQ QL45000 Series 100GbE Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 58616u16 , name : "FastLinQ QL45611H 100GbE Adapter" }] } , Device { vendor_id : 4215u16 , id : 5716u16 , name : "FastLinQ QL45000 Series 50GbE Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 50u16 , name : "QL45212 Flex 50Gb 2-port Ethernet Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 547u16 , name : "Synergy 6810C 25/50Gb Ethernet Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 647u16 , name : "Synergy 6820C 25/50Gb CNA" }] } , Device { vendor_id : 4215u16 , id : 5718u16 , name : "FastLinQ QL45000 Series 25GbE Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 51u16 , name : "QL45214 Flex 25Gb 4-port Ethernet Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 679u16 , name : "QL45212-DE 25GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58614u16 , name : "FastLinQ QL45211H 25GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58615u16 , name : "FastLinQ QL45212H 25GbE Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 581u16 , name : "10/20/25GbE 2P 4820c CNA" }] } , Device { vendor_id : 4215u16 , id : 5724u16 , name : "FastLinQ QL45000 Series 10/25/40/50GbE Controller (FCoE)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 52u16 , name : "QL45262 Flex 50Gb 2-port Ethernet Adapter w/ iSCSI/FCoE" } , SubSystem { subvendor : 4215u16 , subdevice : 58609u16 , name : "FastLinQ QL45462H 40GbE FCoE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58610u16 , name : "FastLinQ QL45461H 40GbE FCoE Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 581u16 , name : "10/20/25GbE 2P 4820c CNA FCoE" }] } , Device { vendor_id : 4215u16 , id : 5726u16 , name : "FastLinQ QL45000 Series 10/25/40/50GbE Controller (iSCSI)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 52u16 , name : "QL45262 Flex 50Gb 2-port Ethernet Adapter w/ iSCSI/FCoE" } , SubSystem { subvendor : 4215u16 , subdevice : 58609u16 , name : "FastLinQ QL45462H 40GbE iSCSI Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 58610u16 , name : "FastLinQ QL45461H 40GbE iSCSI Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 581u16 , name : "10/20/25GbE 2P 4820c CNA iSCSI" }] } , Device { vendor_id : 4215u16 , id : 5732u16 , name : "FastLinQ QL45000 Series Gigabit Ethernet Controller (SR-IOV VF)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 58609u16 , name : "FastLinQ QL45462H 40GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58610u16 , name : "FastLinQ QL45461H 40GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58611u16 , name : "FastLinQ QL45412H 40GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58612u16 , name : "FastLinQ QL45411H 40GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58614u16 , name : "FastLinQ QL45211H 25GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58615u16 , name : "FastLinQ QL45212H 25GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 58616u16 , name : "FastLinQ QL45611H 100GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 5520u16 , subdevice : 581u16 , name : "10/20/25GbE 2P 4820c CNA SRIOV" }] } , Device { vendor_id : 4215u16 , id : 5793u16 , name : "BCM57840 NetXtreme II 10 Gigabit Ethernet" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 236u16 , name : "FlexFabric 10Gb 4-port 536FLR-T Adapter" }] } , Device { vendor_id : 4215u16 , id : 5803u16 , name : "NetXtreme II BCM57840 10/20 Gigabit Ethernet Multi Function" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 5805u16 , name : "NetXtreme II BCM57840 10/20 Gigabit Ethernet Virtual Function" , subsystems : & [SubSystem { subvendor : 5520u16 , subdevice : 236u16 , name : "FlexFabric 10Gb 4-port 536FLR-T Adapter" }] } , Device { vendor_id : 4215u16 , id : 8224u16 , name : "ISP2020A Fast!SCSI Basic Adapter" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 8241u16 , name : "ISP8324-based 16Gb Fibre Channel to PCI Express Adapter" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 6119u16 , name : "SN1000Q 16Gb Single Port Fibre Channel Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6120u16 , name : "SN1000Q 16Gb Dual Port Fibre Channel Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 6457u16 , name : "QMH2672 16Gb Dual Port Fibre Channel Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 32770u16 , name : "3830C 16G Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 577u16 , name : "QLE2670 16Gb Single Port Fibre Channel Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 585u16 , name : "QLE2672 16Gb Dual Port Fibre Channel Adapter" }] } , Device { vendor_id : 4215u16 , id : 8305u16 , name : "ISP2714-based 16/32Gb Fibre Channel to PCIe Adapter" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 643u16 , name : "QLE2764 Quad Port 32Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 670u16 , name : "QLE2694 Quad Port 16Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 674u16 , name : "QLE2694L Quad Port 16Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 685u16 , name : "QLE2694U Quad Port 16/32Gb Fibre Channel to PCIe Adapter" }] } , Device { vendor_id : 4215u16 , id : 8321u16 , name : "ISP2814-based 64/32G Fibre Channel to PCIe Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 737u16 , name : "QLE2874 Quad Port 64GFC PCIe Gen4 x16 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 739u16 , name : "QLE2774 Quad Port 32GFC PCIe Gen4 x16 Adapter" }] } , Device { vendor_id : 4215u16 , id : 8329u16 , name : "ISP2854-based 64/32G Fibre Channel to PCIe Controller with StorCryption" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 744u16 , name : "QLE2884 Quad Port 64GFC PCIe Gen4 x16 Adapter with StorCryption" } , SubSystem { subvendor : 4215u16 , subdevice : 746u16 , name : "QLE2784 Quad Port 32GFC PCIe Gen4 x16 Adapter with StorCryption" }] } , Device { vendor_id : 4215u16 , id : 8448u16 , name : "QLA2100 64-bit Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "QLA2100 64-bit Fibre Channel Adapter" }] } , Device { vendor_id : 4215u16 , id : 8704u16 , name : "QLA2200 64-bit Fibre Channel Adapter" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 2u16 , name : "QLA2200" }] } , Device { vendor_id : 4215u16 , id : 8801u16 , name : "ISP2722-based 16/32Gb Fibre Channel to PCIe Adapter" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 665u16 , name : "QLE2740 Single Port 32Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 666u16 , name : "QLE2742 Dual Port 32Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 667u16 , name : "QLE2690 Single Port 16Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 668u16 , name : "QLE2692 Dual Port 16Gb Fibre Channel to PCIe Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 679u16 , name : "QLE2690 Single Port 16Gb FC to PCIe Gen3 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 680u16 , name : "QLE2692 Dual Port 16Gb FC to PCIe Gen3 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 683u16 , name : "QLE2740 Single Port 32Gb FC to PCIe Gen3 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 684u16 , name : "QLE2742 Dual Port 32Gb FC to PCIe Gen3 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 696u16 , name : "2x16Gb QME2692 FC HBA" } , SubSystem { subvendor : 4215u16 , subdevice : 697u16 , name : "2x32Gb QME2742 FC HBA" } , SubSystem { subvendor : 5520u16 , subdevice : 249u16 , name : "StoreFabric SN1100Q 16Gb Single Port Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 250u16 , name : "StoreFabric SN1100Q 16Gb Dual Port Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 515u16 , name : "StoreFabric SN1600Q 32Gb Single Port Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 516u16 , name : "StoreFabric SN1600Q 32Gb Dual Port Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 557u16 , name : "5830C 32Gb Dual Port Fibre Channel Adapter" } , SubSystem { subvendor : 6461u16 , subdevice : 4109u16 , name : "NIC-FC680i-Mb-2x16G" }] } , Device { vendor_id : 4215u16 , id : 8833u16 , name : "ISP2812-based 64/32G Fibre Channel to PCIe Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 738u16 , name : "QLE2872 Dual Port 64GFC PCIe Gen4 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 740u16 , name : "QLE2772 Dual Port 32GFC PCIe Gen4 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 750u16 , name : "QLE2870 Single Port 64GFC PCIe Gen4 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 752u16 , name : "QLE2770 Single Port 32GFC PCIe Gen4 x8 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 754u16 , name : "QLogic 1x32Gb QLE2770 FC HBA" } , SubSystem { subvendor : 4215u16 , subdevice : 755u16 , name : "QLogic 2x32Gb QLE2772 FC HBA" } , SubSystem { subvendor : 5520u16 , subdevice : 723u16 , name : "SN1610Q - 1P Enhanced 32GFC Single Port Fibre Channel Host Bus Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 724u16 , name : "SN1610Q - 2P Enhanced 32GFC Dual Port Fibre Channel Host Bus Adapter" }] } , Device { vendor_id : 4215u16 , id : 8841u16 , name : "ISP2852-based 64/32G Fibre Channel to PCIe Controller with StorCryption" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 745u16 , name : "QLE2882 Dual Port 64GFC PCIe Gen4 x8 Adapter with StorCryption" } , SubSystem { subvendor : 4215u16 , subdevice : 747u16 , name : "QLE2782 Dual Port 32GFC PCIe Gen4 x8 Adapter with StorCryption" } , SubSystem { subvendor : 4215u16 , subdevice : 751u16 , name : "QLE2880 Single Port 64GFC PCIe Gen4 x8 Adapter with StorCryption" } , SubSystem { subvendor : 4215u16 , subdevice : 753u16 , name : "QLE2780 Single Port 32GFC PCIe Gen4 x8 Adapter with StorCryption" }] } , Device { vendor_id : 4215u16 , id : 8857u16 , name : "SC2962I 128/64G Fibre Channel to PCIe Controller" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 8960u16 , name : "QLA2300 64-bit Fibre Channel Adapter" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 8978u16 , name : "ISP2312-based 2Gb Fibre Channel to PCI-X HBA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 305u16 , name : "2Gb Fibre Channel - Single port [A7538A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4794u16 , name : "2Gb Fibre Channel - Dual port [A6826A]" }] } , Device { vendor_id : 4215u16 , id : 8994u16 , name : "ISP2322-based 2Gb Fibre Channel to PCI-X HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 9250u16 , name : "ISP2422-based 4Gb Fibre Channel to PCI-X HBA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 4823u16 , name : "4Gb Fibre Channel [AB379A]" } , SubSystem { subvendor : 4156u16 , subdevice : 4829u16 , name : "4Gb Fibre Channel [AB429A]" }] } , Device { vendor_id : 4215u16 , id : 9266u16 , name : "ISP2432-based 4Gb Fibre Channel to PCI Express HBA" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 28736u16 , name : "FC1142SR 4Gb 1-port PCIe Fibre Channel Host Bus Adapter [HPAE311A]" } , SubSystem { subvendor : 4215u16 , subdevice : 311u16 , name : "QLE2460 Single-Port 4Gbps FC-to-PCI-X/PCIe Host Bus Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 312u16 , name : "QLE2462 Dual-Port 4Gbps FC-to-PCI-X/PCIe Host Bus Adapter" }] } , Device { vendor_id : 4215u16 , id : 9522u16 , name : "ISP2532-based 8Gb Fibre Channel to PCI Express HBA" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1054u16 , name : "FC EN0Y/EN12 PCIe2 LP 8 Gb 4-port Fibre Channel Adapter for POWER" } , SubSystem { subvendor : 4156u16 , subdevice : 12898u16 , name : "StorageWorks 81Q" } , SubSystem { subvendor : 4156u16 , subdevice : 12899u16 , name : "StorageWorks 82Q" } , SubSystem { subvendor : 4215u16 , subdevice : 348u16 , name : "QLE2560 PCI Express to 8Gb FC Single Channel" } , SubSystem { subvendor : 4215u16 , subdevice : 349u16 , name : "QLE2562 PCI Express to 8Gb FC Dual Channel" } , SubSystem { subvendor : 4215u16 , subdevice : 350u16 , name : "QLE2564 PCI Express to 8Gb FC Quad Channel" } , SubSystem { subvendor : 4215u16 , subdevice : 359u16 , name : "QME2572 Dual Port FC8 HBA Mezzanine" } , SubSystem { subvendor : 5520u16 , subdevice : 252u16 , name : "StoreFabric 84Q 8Gb Quad Port Fibre Channel Host Bus Adapter" }] } , Device { vendor_id : 4215u16 , id : 10609u16 , name : "ISP2684" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 10625u16 , name : "32G SFP28 EP2744 Apatter" , subsystems : & [SubSystem { subvendor : 7124u16 , subdevice : 176u16 , name : "32G SFP28 EP2744 Apatter" }] } , Device { vendor_id : 4215u16 , id : 12322u16 , name : "ISP4022-based Ethernet NIC" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 12338u16 , name : "ISP4032-based Ethernet IPv6 NIC" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 16400u16 , name : "ISP4010-based iSCSI TOE HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 16418u16 , name : "ISP4022-based iSCSI TOE HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 16434u16 , name : "ISP4032-based iSCSI TOE IPv6 HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 21554u16 , name : "SP232-based 4Gb Fibre Channel to PCI Express HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 25362u16 , name : "SP202-based 2Gb Fibre Channel to PCI-X HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 25378u16 , name : "SP212-based 2Gb Fibre Channel to PCI-X HBA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 29216u16 , name : "IBA7220 InfiniBand HCA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 29474u16 , name : "IBA7322 QDR InfiniBand HCA" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 32768u16 , name : "10GbE Converged Network Adapter (TCP/IP Networking)" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 32769u16 , name : "10GbE Converged Network Adapter (FCoE)" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 943u16 , name : "FC 5708/5270 10 Gb FCoE PCIe Dual Port Adapter for POWER" }] } , Device { vendor_id : 4215u16 , id : 32800u16 , name : "cLOM8214 1/10GbE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8036u16 , name : "QMD8262-k 10G DP bNDC KR" } , SubSystem { subvendor : 4156u16 , subdevice : 13126u16 , name : "CN1000Q Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4156u16 , subdevice : 14131u16 , name : "NC523SFP 10Gb 2-port Server Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 515u16 , name : "8200 Series Single Port 10GbE Converged Network Adapter (TCP/IP Networking)" } , SubSystem { subvendor : 4215u16 , subdevice : 519u16 , name : "8200 Series Dual Port 10GbE Converged Network Adapter (TCP/IP Networking)" } , SubSystem { subvendor : 4215u16 , subdevice : 523u16 , name : "3200 Series Dual Port 10Gb Intelligent Ethernet Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 524u16 , name : "3200 Series Quad Port 1Gb Intelligent Ethernet Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 527u16 , name : "3200 Series Single Port 10Gb Intelligent Ethernet Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 528u16 , name : "QME8242-k 10GbE Dual Port Mezzanine Card" } , SubSystem { subvendor : 4215u16 , subdevice : 563u16 , name : "QME8262-k 10GbE Dual Port Mezzanine Card" }] } , Device { vendor_id : 4215u16 , id : 32801u16 , name : "8200 Series 10GbE Converged Network Adapter (FCoE)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13128u16 , name : "CN1000Q Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 529u16 , name : "QME8242-k 10GbE Dual Port Mezzanine Card, FCoE" }] } , Device { vendor_id : 4215u16 , id : 32802u16 , name : "8200 Series 10GbE Converged Network Adapter (iSCSI)" , subsystems : & [SubSystem { subvendor : 4156u16 , subdevice : 13127u16 , name : "CN1000Q Dual Port Converged Network Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 530u16 , name : "QME8242-k 10GbE Dual Port Mezzanine Card, iSCSI" }] } , Device { vendor_id : 4215u16 , id : 32816u16 , name : "ISP8324 1/10GbE Converged Network Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 579u16 , name : "8300 Series Single Port 10GbE Converged Network Adapter (TCP/IP Networking)" } , SubSystem { subvendor : 4215u16 , subdevice : 582u16 , name : "8300 Series Dual Port 10GbE Converged Network Adapter (TCP/IP Networking)" }] } , Device { vendor_id : 4215u16 , id : 32817u16 , name : "8300 Series 10GbE Converged Network Adapter (FCoE)" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 32818u16 , name : "8300 Series 10GbE Converged Network Adapter (iSCSI)" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 32880u16 , name : "FastLinQ QL41000 Series 10/25/40/50GbE Controller" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "10GE 2P QL41162HxRJ-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 2u16 , name : "10GE 2P QL41112HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 4u16 , name : "4x10GE QL41164HFCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 5u16 , name : "QLogic 4x10GE QL41164HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 6u16 , name : "QLogic 4x10GE QL41164HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 7u16 , name : "QLogic 2x1GE+2x10GE QL41264HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 9u16 , name : "QLogic 2x1GE+2x10GE QL41162HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 11u16 , name : "25GE 2P QL41262HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 15u16 , name : "2x25GE QL41262HMKR CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 16u16 , name : "2x25GE QL41232HMKR NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 17u16 , name : "FastLinQ QL41212HLCU 25GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 18u16 , name : "FastLinQ QL41112H 10GbE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 25u16 , name : "QL41232HOCU - Dual Port 25/10GbE SFP28 OCP Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 57u16 , name : "QLogic QL41262 PCIe 25Gb 2-Port SFP28 Ethernet Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 83u16 , name : "QLogic 2x25GE QL41232HQCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 84u16 , name : "2x10GE QL41132HQRJ NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 85u16 , name : "QLogic 2x10GE QL41132HQCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 86u16 , name : "2x10GE QL41132HxRJ NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 87u16 , name : "2x25GE QL41232HxCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 101u16 , name : "QLogic 4x10GE QL41154HQRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 102u16 , name : "QLogic 4x10GE QL41154HQCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 104u16 , name : "10GbE 2p SFP+ QL41132HLCU-HC Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 105u16 , name : "10GbE 2p BASE-T QL41132HQRJ-HC OCP3 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 112u16 , name : "10GbE 2p BASE-T QL41132HLRJ-HC Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 113u16 , name : "10GbE 2p SFP+ QL41132HQCU-HC OCP3 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 114u16 , name : "10GbE 4p SFP+ QL41134HLCU-HC Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 115u16 , name : "10/25GbE 2p SFP28 QL41232HQCU-HC OCP3 Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 116u16 , name : "10/25GbE 2p SFP28 QL41232HLCU-HC Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 538u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 539u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 541u16 , name : "10/25GbE 2P QL41222HLCU-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 542u16 , name : "10/25GbE 2P QL41162HMRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 543u16 , name : "10/25GbE 2P QL41262HMCU-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 544u16 , name : "10/25GbE 2P QL41122HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 701u16 , name : "10Gb 2P 524SFP+ NIC" } , SubSystem { subvendor : 6461u16 , subdevice : 4144u16 , name : "NIC-ETH681i-Mb-2x25G" } , SubSystem { subvendor : 6461u16 , subdevice : 4146u16 , name : "NIC-ETH682i-Mb-2x25G" }] } , Device { vendor_id : 4215u16 , id : 32896u16 , name : "FastLinQ QL41000 Series 10/25/40/50GbE Controller (FCoE)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "10GE 2P QL41162HxRJ-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 2u16 , name : "10GE 2P QL41112HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 4u16 , name : "4x10GE QL41164HFCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 5u16 , name : "QLogic 4x10GE QL41164HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 6u16 , name : "QLogic 4x10GE QL41164HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 7u16 , name : "QLogic 2x1GE+2x10GE QL41264HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 9u16 , name : "QLogic 2x1GE+2x10GE QL41162HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 11u16 , name : "25GE 2P QL41262HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 12u16 , name : "QLogic 2x25GE QL41262HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 13u16 , name : "FastLinQ QL41262H 25GbE FCoE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 14u16 , name : "FastLinQ QL41162H 10GbE FCoE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 15u16 , name : "2x25GE QL41262HMKR CNA" } , SubSystem { subvendor : 5520u16 , subdevice : 538u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 539u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" }] } , Device { vendor_id : 4215u16 , id : 32900u16 , name : "FastLinQ QL41000 Series 10/25/40/50GbE Controller (iSCSI)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "10GE 2P QL41162HxRJ-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 2u16 , name : "10GE 2P QL41112HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 3u16 , name : "4x10GE QL41164HxRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 4u16 , name : "4x10GE QL41164HFCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 5u16 , name : "QLogic 4x10GE QL41164HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 6u16 , name : "QLogic 4x10GE QL41164HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 7u16 , name : "QLogic 2x25GE QL41262HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 9u16 , name : "QLogic 2x1GE+2x10GE QL41162HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 11u16 , name : "25GE 2P QL41262HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 12u16 , name : "QLogic 2x25GE QL41262HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 13u16 , name : "FastLinQ QL41262H 25GbE iSCSI Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 14u16 , name : "FastLinQ QL41162H 10GbE iSCSI Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 15u16 , name : "2x25GE QL41262HMKR CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 101u16 , name : "QLogic 4x10GE QL41154HQRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 102u16 , name : "QLogic 4x10GE QL41154HQCU CNA" } , SubSystem { subvendor : 5520u16 , subdevice : 538u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 539u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" }] } , Device { vendor_id : 4215u16 , id : 32912u16 , name : "FastLinQ QL41000 Series Gigabit Ethernet Controller (SR-IOV VF)" , subsystems : & [SubSystem { subvendor : 4215u16 , subdevice : 1u16 , name : "25GE 2P QL41262HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 2u16 , name : "10GE 2P QL41112HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 3u16 , name : "4x10GE QL41164HxRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 4u16 , name : "4x10GE QL41164HFCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 5u16 , name : "QLogic 4x10GE QL41164HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 6u16 , name : "QLogic 4x10GE QL41164HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 7u16 , name : "QLogic 2x1GE+2x10GE QL41264HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 9u16 , name : "QLogic 2x1GE+2x10GE QL41162HMRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 11u16 , name : "25GE 2P QL41262HxCU-DE Adapter" } , SubSystem { subvendor : 4215u16 , subdevice : 12u16 , name : "QLogic 2x25GE QL41262HMCU CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 13u16 , name : "FastLinQ QL41262H 25GbE FCoE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 14u16 , name : "FastLinQ QL41162H 10GbE iSCSI Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 15u16 , name : "2x25GE QL41262HMKR CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 16u16 , name : "2x25GE QL41232HMKR NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 17u16 , name : "FastLinQ QL41212H 25GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 18u16 , name : "FastLinQ QL41112H 10GbE Adapter (SR-IOV VF)" } , SubSystem { subvendor : 4215u16 , subdevice : 83u16 , name : "QLogic 2x25GE QL41232HQCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 84u16 , name : "QLogic 2x10GE QL41132HQRJ NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 85u16 , name : "QLogic 2x10GE QL41132HQCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 86u16 , name : "2x10GE QL41132HxRJ NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 87u16 , name : "2x25GE QL41232HxCU NIC" } , SubSystem { subvendor : 4215u16 , subdevice : 101u16 , name : "QLogic 4x10GE QL41154HQRJ CNA" } , SubSystem { subvendor : 4215u16 , subdevice : 102u16 , name : "QLogic 4x10GE QL41154HQCU CNA" } , SubSystem { subvendor : 5520u16 , subdevice : 538u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 539u16 , name : "10GbE 2P QL41162HLRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 542u16 , name : "10/25GbE 2P QL41162HMRJ-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 543u16 , name : "10/25GbE 2P QL41262HMCU-HP Adapter" } , SubSystem { subvendor : 5520u16 , subdevice : 701u16 , name : "10Gb 2P 524SFP+ NIC" }] } , Device { vendor_id : 4215u16 , id : 33840u16 , name : "ISP8324 1/10GbE Converged Network Controller (NIC VF)" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 33841u16 , name : "8300 Series 10GbE Converged Network Adapter (FCoE VF)" , subsystems : & [] } , Device { vendor_id : 4215u16 , id : 33842u16 , name : "ISP2432M-based 10GbE Converged Network Adapter (CNA)" , subsystems : & [] }] }),
        (6568, Vendor { id : 6568u16 , name : "DAQDATA GmbH" , devices : & [] }),
        (5197, Vendor { id : 5197u16 , name : "Samsung Electronics Co Ltd" , devices : & [Device { vendor_id : 5197u16 , id : 5632u16 , name : "S4LN053X01 AHCI SSD Controller(Apple slot)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 42308u16 , name : "Exynos 8890 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 42357u16 , name : "Exynos 7420 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 42467u16 , name : "Exynos 5433 PCIe Root Complex" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43008u16 , name : "XP941 PCIe SSD" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43009u16 , name : "S4LN058A01[SSUBX] AHCI SSD Controller (Apple slot)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43010u16 , name : "NVMe SSD Controller SM951/PM951" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 43009u16 , name : "PM963 2.5\" NVMe PCIe SSD" }] } , Device { vendor_id : 5197u16 , id : 43012u16 , name : "NVMe SSD Controller SM961/PM961/SM963" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 43009u16 , name : "SM963 2.5\" NVMe PCIe SSD" }] } , Device { vendor_id : 5197u16 , id : 43014u16 , name : "PM971 BGA PCIe x2 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43016u16 , name : "NVMe SSD Controller SM981/PM981/PM983" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 43009u16 , name : "SSD 970 EVO/PRO" } , SubSystem { subvendor : 7497u16 , subdevice : 16443u16 , name : "Thinksystem U.2 PM983 NVMe SSD" }] } , Device { vendor_id : 5197u16 , id : 43017u16 , name : "NVMe SSD Controller 980 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43018u16 , name : "NVMe SSD Controller PM9A1/PM9A3/980PRO" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8538u16 , name : "DC NVMe PM9A3 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8539u16 , name : "DC NVMe PM9A3 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8540u16 , name : "DC NVMe PM9A3 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8541u16 , name : "Dell DC NVMe PM9A3 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8550u16 , name : "DC NVMe PM9A3 RI 110M.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8551u16 , name : "DC NVMe PM9A3 RI 110M.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8552u16 , name : "DC NVMe PM9A3 RI 80M.2 480GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8553u16 , name : "DC NVMe PM9A3 RI 80M.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8822u16 , name : "DC NVMe PM9A3 RI 110M.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8823u16 , name : "DC NVMe PM9A3 RI 110M.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 20781u16 , name : "DC NVMe PM9A3 RI U.2 7.68TB" } , SubSystem { subvendor : 5197u16 , subdevice : 43009u16 , name : "SSD 980 PRO" } , SubSystem { subvendor : 5197u16 , subdevice : 43027u16 , name : "General DC NVMe PM9A3" }] } , Device { vendor_id : 5197u16 , id : 43019u16 , name : "NVMe SSD Controller PM9B1 (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43020u16 , name : "NVMe SSD Controller S4LV008[Pascal]" , subsystems : & [SubSystem { subvendor : 5197u16 , subdevice : 43009u16 , name : "SSD 990 PRO" }] } , Device { vendor_id : 5197u16 , id : 43021u16 , name : "NVMe SSD Controller PM9C1a (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43022u16 , name : "NVMe SSD Controller PM9D3a" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43023u16 , name : "BM9C1 QLC NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43024u16 , name : "NVMe SSD 9100 PRO [PM9E1]" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43025u16 , name : "BM9H1 NVME SSD (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 5197u16 , id : 43040u16 , name : "NVMe SSD Controller 171X" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8085u16 , name : "Express Flash NVMe XS1715 SSD 400GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8086u16 , name : "Express Flash NVMe XS1715 SSD 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8087u16 , name : "Express Flash NVMe XS1715 SSD 1600GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8100u16 , name : "Express Flash NVMe SM1715 3.2TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8102u16 , name : "Express Flash NVMe SM1715 3.2TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8122u16 , name : "Express Flash NVMe SM1715 800GB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8123u16 , name : "Express Flash NVMe SM1715 1.6TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8124u16 , name : "Express Flash NVMe SM1715 1.6TB AIC" }] } , Device { vendor_id : 5197u16 , id : 43041u16 , name : "NVMe SSD Controller 172X" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8119u16 , name : "Express Flash NVMe PM1725 3.2TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8120u16 , name : "Express Flash NVMe PM1725 3.2TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8121u16 , name : "Express Flash NVMe PM1725 6.4TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8129u16 , name : "Express Flash NVMe PM1725 800GB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8130u16 , name : "Express Flash NVMe PM1725 1.6TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8132u16 , name : "Express Flash NVMe PM1725 1.6TB AIC" }] } , Device { vendor_id : 5197u16 , id : 43042u16 , name : "NVMe SSD Controller 172Xa/172Xb" , subsystems : & [SubSystem { subvendor : 4116u16 , subdevice : 1569u16 , name : "PCIe3 1.6TB NVMe Flash Adapter II x8" } , SubSystem { subvendor : 4116u16 , subdevice : 1570u16 , name : "PCIe3 3.2TB NVMe Flash Adapter II x8" } , SubSystem { subvendor : 4116u16 , subdevice : 1577u16 , name : "PCIe3 6.4TB NVMe Flash Adapter II x8" } , SubSystem { subvendor : 4116u16 , subdevice : 1610u16 , name : "PCIe3 1.6TB NVMe Flash Adapter III x8" } , SubSystem { subvendor : 4116u16 , subdevice : 1611u16 , name : "PCIe3 3.2TB NVMe Flash Adapter III x8" } , SubSystem { subvendor : 4116u16 , subdevice : 1612u16 , name : "PCIe3 6.4TB NVMe Flash Adapter III x8" } , SubSystem { subvendor : 4136u16 , subdevice : 8153u16 , name : "Express Flash PM1725a 800GB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8154u16 , name : "Express Flash PM1725a 1.6TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8155u16 , name : "Express Flash PM1725a 3.2TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8156u16 , name : "Express Flash PM1725a 6.4TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8157u16 , name : "Express Flash PM1725a 1.6TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8158u16 , name : "Express Flash PM1725a 3.2TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8159u16 , name : "Express Flash PM1725a 6.4TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8179u16 , name : "Express Flash PM1725b 1.6TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8180u16 , name : "Express Flash PM1725b 3.2TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8181u16 , name : "Express Flash PM1725b 6.4TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8182u16 , name : "Express Flash PM1725b 12.8TB SFF" } , SubSystem { subvendor : 4136u16 , subdevice : 8183u16 , name : "Express Flash PM1725b 1.6TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8184u16 , name : "Express Flash PM1725b 3.2TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8185u16 , name : "Express Flash PM1725b 6.4TB AIC" } , SubSystem { subvendor : 4136u16 , subdevice : 8186u16 , name : "Express Flash PM1725b 12.8TB AIC" }] } , Device { vendor_id : 5197u16 , id : 43044u16 , name : "NVMe SSD Controller PM173X" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8256u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU U.2 Gen4 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8257u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU U.2 Gen4 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8258u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU U.2 Gen4 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8259u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU U.2 Gen4 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8260u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU AIC Gen4 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8261u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU AIC Gen4 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8262u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN MU AIC Gen4 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8304u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI U.2 Gen4 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8305u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI U.2 Gen4 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8306u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI U.2 Gen4 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8307u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI U.2 Gen4 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8308u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI AIC Gen4 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8309u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI AIC Gen4 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8310u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN RI AIC Gen4 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8336u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU U.2 Gen4 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8337u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU U.2 Gen4 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8338u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU U.2 Gen4 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8339u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU U.2 Gen4 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8340u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU AIC Gen4 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8341u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU AIC Gen4 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8342u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED MU AIC Gen4 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8343u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED RI U.2 Gen4 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8344u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED RI U.2 Gen4 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8345u16 , name : "EMC PowerEdge Express Flash Ent NVMe AGN SED RI U.2 Gen4 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8472u16 , name : "Ent NVMe v2 AGN FIPS MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8473u16 , name : "Ent NVMe v2 AGN MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8480u16 , name : "Ent NVMe v2 AGN FIPS MU U.2 3.2T" } , SubSystem { subvendor : 4136u16 , subdevice : 8481u16 , name : "Ent NVMe v2 AGN MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8482u16 , name : "Ent NVMe v2 AGN FIPS MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8483u16 , name : "Ent NVMe v2 AGN MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8484u16 , name : "Ent NVMe v2 AGN FIPS MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8485u16 , name : "Ent NVMe v2 AGN MU U.2 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8486u16 , name : "Ent NVMe v2 AGN FIPS RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8487u16 , name : "Ent NVMe v2 AGN RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8488u16 , name : "Ent NVMe v2 AGN FIPS RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8489u16 , name : "Ent NVMe v2 AGN RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8496u16 , name : "Ent NVMe v2 AGN FIPS RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8497u16 , name : "Ent NVMe v2 AGN RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8498u16 , name : "Ent NVMe v2 AGN FIPS RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8499u16 , name : "Ent NVMe v2 AGN RI U.2 15.36TB" }] } , Device { vendor_id : 5197u16 , id : 43045u16 , name : "NVMe SSD Controller PM173Xa" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8568u16 , name : "Ent NVMe FIPS PM1733a RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8569u16 , name : "Ent NVMe PM1733a RI 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8570u16 , name : "Ent NVMe FIPS PM1733a RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8571u16 , name : "Ent NVMe PM1733a RI 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8572u16 , name : "Ent NVMe FIPS PM1733a RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8573u16 , name : "Ent NVMe PM1733a RI 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8574u16 , name : "Ent NVMe FIPS PM1733a RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8575u16 , name : "Ent NVMe PM1733a RI 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8576u16 , name : "Ent NVMe FIPS PM1735a MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8577u16 , name : "Ent NVMe PM1735a MU 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8578u16 , name : "Ent NVMe FIPS PM1735a MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8579u16 , name : "Ent NVMe PM1735a MU 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8580u16 , name : "Ent NVMe FIPS PM1735a MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8581u16 , name : "Ent NVMe PM1735a MU 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8582u16 , name : "Ent NVMe FIPS PM1735a MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8583u16 , name : "Ent NVMe PM1735a MU 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8810u16 , name : "Ent NVMe FIPS PM1733a RI 30.72TB" }] } , Device { vendor_id : 5197u16 , id : 43046u16 , name : "NVMe SSD Controller PM174X" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8724u16 , name : "NVMe PM1743 RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8726u16 , name : "NVMe PM1743 RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8728u16 , name : "NVMe PM1743 RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8730u16 , name : "NVMe PM1743 RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8732u16 , name : "NVMe PM1745 MU E3.S 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8734u16 , name : "NVMe PM1745 MU E3.S 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8736u16 , name : "NVMe PM1745 MU E3.S 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8738u16 , name : "NVMe PM1745 MU E3.S 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8782u16 , name : "NVMe FIPS PM1743 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8783u16 , name : "NVMe PM1743 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8784u16 , name : "NVMe FIPS PM1743 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8785u16 , name : "NVMe PM1743 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8786u16 , name : "NVMe FIPS PM1743 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8787u16 , name : "NVMe PM1743 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8788u16 , name : "NVMe FIPS PM1743 RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8789u16 , name : "NVMe PM1743 RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8792u16 , name : "NVMe FIPS PM1745 MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8793u16 , name : "NVMe PM1745 MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8794u16 , name : "NVMe FIPS PM1745 MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8795u16 , name : "NVMe PM1745 MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8796u16 , name : "NVMe FIPS PM1745 MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8797u16 , name : "NVMe PM1745 MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8798u16 , name : "NVMe FIPS PM1745 MU U.2 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8799u16 , name : "NVMe PM1745 MU U.2 12.8TB" }] } , Device { vendor_id : 5197u16 , id : 43264u16 , name : "NVMe SSD Controller PM9DXa" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8975u16 , name : "DC NVMe PM9D3a RI 80M.2 480GB ISE" } , SubSystem { subvendor : 4136u16 , subdevice : 8976u16 , name : "DC NVMe PM9D3a RI 80M.2 960GB ISE" } , SubSystem { subvendor : 4136u16 , subdevice : 8977u16 , name : "DC NVMe PM9D3a RI 80M.2 1.92TB ISE" } , SubSystem { subvendor : 4136u16 , subdevice : 9025u16 , name : "DC NVMe PM9D3a RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 9026u16 , name : "DC NVMe PM9D3a RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9027u16 , name : "DC NVMe PM9D3a RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9028u16 , name : "DC NVMe PM9D3a RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9029u16 , name : "DC NVMe PM9D3a RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9030u16 , name : "DC NVMe FIPS PM9D3a RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 9031u16 , name : "DC NVMe FIPS PM9D3a RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9032u16 , name : "DC NVMe FIPS PM9D3a RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9033u16 , name : "DC NVMe FIPS PM9D3a RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9034u16 , name : "DC NVMe FIPS PM9D3a RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9037u16 , name : "DC NVMe PM9D3a RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9038u16 , name : "DC NVMe PM9D3a RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9039u16 , name : "DC NVMe PM9D3a RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9040u16 , name : "DC NVMe PM9D3a RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9041u16 , name : "DC NVMe FIPS PM9D3a RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9042u16 , name : "DC NVMe FIPS PM9D3a RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9043u16 , name : "DC NVMe FIPS PM9D3a RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9044u16 , name : "DC NVMe FIPS PM9D3a RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9045u16 , name : "DC NVMe PM9D5a MU U.2 800GB" } , SubSystem { subvendor : 4136u16 , subdevice : 9046u16 , name : "DC NVMe PM9D5a MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9047u16 , name : "DC NVMe PM9D5a MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9048u16 , name : "DC NVMe PM9D5a MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9049u16 , name : "DC NVMe PM9D5a MU E3.S 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9050u16 , name : "DC NVMe PM9D5a MU E3.S 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9051u16 , name : "DC NVMe PM9D5a MU E3.S 6.4TB" }] } , Device { vendor_id : 5197u16 , id : 43520u16 , name : "NVMe SSD Controller BM1743" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8978u16 , name : "NVMe FIPS BM1743 QLC U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8979u16 , name : "NVMe FIPS BM1743 QLC U.2 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8980u16 , name : "NVMe FIPS BM1743 QLC U.2 61.44TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8981u16 , name : "NVMe BM1743 QLC U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8982u16 , name : "NVMe BM1743 QLC U.2 30.72TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8983u16 , name : "NVMe BM1743 QLC U.2 61.44TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9060u16 , name : "NVMe BM1743 QLC U.2 122.88TB" } , SubSystem { subvendor : 4136u16 , subdevice : 9062u16 , name : "MZ3MO15THCLCAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9063u16 , name : "MZ3MO30THCLFAD3" }] } , Device { vendor_id : 5197u16 , id : 44032u16 , name : "NVMe SSD Controller PM175x" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 9091u16 , name : "MZ3L91T9HFJAAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9092u16 , name : "MZ3L91T9HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9093u16 , name : "MZ3L93T8HFJAAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9094u16 , name : "MZ3L93T8HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9095u16 , name : "MZ3L97T6HFLTAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9096u16 , name : "MZ3L97T6HFLTAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9097u16 , name : "MZ3L915THBLCAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9098u16 , name : "MZ3L915THBLCAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9099u16 , name : "MZ3L930THBLFAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9100u16 , name : "MZ3L930THBLFAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9101u16 , name : "MZ3L91T6HFJAAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9102u16 , name : "MZ3L91T6HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9103u16 , name : "MZ3L93T2HFJAAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9104u16 , name : "MZ3L93T2HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9105u16 , name : "MZ3L96T4HFLTAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9106u16 , name : "MZ3L96T4HFLTAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9108u16 , name : "MZWL91T9HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9110u16 , name : "MZWL93T8HFLTAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9112u16 , name : "MZWL97T6HFLAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9114u16 , name : "MZWL915THBLFAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9115u16 , name : "MZWL930THBLFAD9" } , SubSystem { subvendor : 4136u16 , subdevice : 9116u16 , name : "MZWL930THBLFAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9118u16 , name : "MZWL91T6HFJAAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9119u16 , name : "MZWL93T2HFLTAD3" } , SubSystem { subvendor : 4136u16 , subdevice : 9120u16 , name : "MZWL96T4HFLAAD3" }] } , Device { vendor_id : 5197u16 , id : 60652u16 , name : "Exynos 8895 PCIe Root Complex" , subsystems : & [] }] }),
        (4845, Vendor { id : 4845u16 , name : "Optivision Inc." , devices : & [] }),
        (7788, Vendor { id : 7788u16 , name : "Xsight Labs Ltd." , devices : & [Device { vendor_id : 7788u16 , id : 1u16 , name : "XLX1A256A Switch ASIC" , subsystems : & [] } , Device { vendor_id : 7788u16 , id : 2u16 , name : "XLX2B128A Switch ASIC" , subsystems : & [] }] }),
        (43520, Vendor { id : 43520u16 , name : "iTuner" , devices : & [] }),
        (6135, Vendor { id : 6135u16 , name : "Topdek Semiconductor Inc." , devices : & [] }),
        (5128, Vendor { id : 5128u16 , name : "Aloka Co. Ltd" , devices : & [] }),
        (4202, Vendor { id : 4202u16 , name : "Aten Research Inc" , devices : & [] }),
        (4302, Vendor { id : 4302u16 , name : "Radius" , devices : & [] }),
        (5640, Vendor { id : 5640u16 , name : "Automated Wagering International" , devices : & [] }),
        (7308, Vendor { id : 7308u16 , name : "Mobiveil, Inc." , devices : & [] }),
        (4438, Vendor { id : 4438u16 , name : "Periscope Engineering" , devices : & [] }),
        (5087, Vendor { id : 5087u16 , name : "E-Tech Inc" , devices : & [Device { vendor_id : 5087u16 , id : 1u16 , name : "PCI56RVP Modem" , subsystems : & [SubSystem { subvendor : 5087u16 , subdevice : 1u16 , name : "PCI56RVP Modem" }] }] }),
        (5139, Vendor { id : 5139u16 , name : "Addonics" , devices : & [] }),
        (6218, Vendor { id : 6218u16 , name : "Thales Computers" , devices : & [Device { vendor_id : 6218u16 , id : 4352u16 , name : "MAX II cPLD" , subsystems : & [] }] }),
        (5961, Vendor { id : 5961u16 , name : "RLX Technologies" , devices : & [] }),
        (7931, Vendor { id : 7931u16 , name : "Flexxon Pte Ltd" , devices : & [] }),
        (4218, Vendor { id : 4218u16 , name : "NetWorth" , devices : & [] }),
        (5057, Vendor { id : 5057u16 , name : "3ware Inc" , devices : & [Device { vendor_id : 5057u16 , id : 4096u16 , name : "5xxx/6xxx-series PATA-RAID" , subsystems : & [] } , Device { vendor_id : 5057u16 , id : 4097u16 , name : "7xxx/8xxx-series PATA/SATA-RAID" , subsystems : & [SubSystem { subvendor : 5057u16 , subdevice : 4097u16 , name : "7xxx/8xxx-series PATA/SATA-RAID" }] } , Device { vendor_id : 5057u16 , id : 4098u16 , name : "9xxx-series SATA-RAID" , subsystems : & [] } , Device { vendor_id : 5057u16 , id : 4099u16 , name : "9550SX SATA-II RAID PCI-X" , subsystems : & [] } , Device { vendor_id : 5057u16 , id : 4100u16 , name : "9650SE SATA-II RAID PCIe" , subsystems : & [] } , Device { vendor_id : 5057u16 , id : 4101u16 , name : "9690SA SAS/SATA-II RAID PCIe" , subsystems : & [] } , Device { vendor_id : 5057u16 , id : 4112u16 , name : "9750 SAS2/SATA-II RAID PCIe" , subsystems : & [] }] }),
        (5637, Vendor { id : 5637u16 , name : "Pairgain Technologies" , devices : & [] }),
        (4996, Vendor { id : 4996u16 , name : "Reality Simulation Systems Inc" , devices : & [] }),
        (5411, Vendor { id : 5411u16 , name : "MUSIC Semiconductors" , devices : & [] }),
        (5328, Vendor { id : 5328u16 , name : "Ericsson Axe R & D" , devices : & [] }),
        (7854, Vendor { id : 7854u16 , name : "XFX Limited" , devices : & [] }),
        (29559, Vendor { id : 29559u16 , name : "Shenzhen Colorful Yugong Technology and Development Co." , devices : & [] }),
        (8139, Vendor { id : 8139u16 , name : "Varex Imaging Deutschland AG" , devices : & [Device { vendor_id : 8139u16 , id : 1u16 , name : "XRD_FG (PCGR-200) [95510214H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 2u16 , name : "XRD_FGx Opto (PCGR-300) [95510215H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 3u16 , name : "XRD_FGe Opto (PCGR-400 / PCGR-410) [95510216H / 95510218H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 4u16 , name : "XRD_FGe (PCGR-210) [95510217H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 5u16 , name : "XRD_FGe Opto V3 (PCGR-415) [95510220H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 6u16 , name : "XRD_FGe Opto V4 (PCGR-500) [95510221H]" , subsystems : & [] } , Device { vendor_id : 8139u16 , id : 7u16 , name : "XRD_FGe Opto V5 (PCGR-510) [95510222H]" , subsystems : & [] }] }),
        (8178, Vendor { id : 8178u16 , name : "Linkdata" , devices : & [Device { vendor_id : 8178u16 , id : 4257u16 , name : "NIC1160 Ethernet Controller Family" , subsystems : & [SubSystem { subvendor : 8178u16 , subdevice : 3089u16 , name : "10GE Ethernet Adapter 1160-2X" }] } , Device { vendor_id : 8178u16 , id : 4258u16 , name : "NIC1160 Ethernet Controller Virtual Function Family" , subsystems : & [] } , Device { vendor_id : 8178u16 , id : 4273u16 , name : "NIC 1260 Ethernet Controller Family" , subsystems : & [] } , Device { vendor_id : 8178u16 , id : 4274u16 , name : "NIC 1260 Ethernet Controller Virtual Function Family" , subsystems : & [] } , Device { vendor_id : 8178u16 , id : 4275u16 , name : "NIC 1260C Ethernet Controller Family" , subsystems : & [] } , Device { vendor_id : 8178u16 , id : 4276u16 , name : "NIC 1260C Ethernet Controller Virtual Function Family" , subsystems : & [] } , Device { vendor_id : 8178u16 , id : 8353u16 , name : "IOC2110 Storage Controller" , subsystems : & [SubSystem { subvendor : 8178u16 , subdevice : 2577u16 , name : "2120-16i SATA3/SAS3 HBA Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2578u16 , name : "2120-8i SATA3/SAS3 HBA Adapter" }] } , Device { vendor_id : 8178u16 , id : 8354u16 , name : "IOC2250 Storage Controller" , subsystems : & [SubSystem { subvendor : 8178u16 , subdevice : 2593u16 , name : "2230-18i Tri-mode HBA Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2594u16 , name : "2230-10i Tri-mode HBA Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2595u16 , name : "2230-16i Tri-mode HBA Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2596u16 , name : "2230-8i Tri-mode HBA Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2600u16 , name : "2233-16i Tri-mode HBA Adapter" }] } , Device { vendor_id : 8178u16 , id : 12450u16 , name : "ROC3250 Storage Controller" , subsystems : & [SubSystem { subvendor : 8178u16 , subdevice : 2849u16 , name : "3260-18i Tri-mode RAID Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2850u16 , name : "3260-10i Tri-mode RAID Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2851u16 , name : "3260-16i Tri-mode RAID Adapter" } , SubSystem { subvendor : 8178u16 , subdevice : 2852u16 , name : "3260-8i Tri-mode RAID Adapter" }] }] }),
        (4309, Vendor { id : 4309u16 , name : "Autologic Inc." , devices : & [] }),
        (4882, Vendor { id : 4882u16 , name : "Acuity Imaging, Inc" , devices : & [] }),
        (5663, Vendor { id : 5663u16 , name : "Rioworks" , devices : & [] }),
        (7408, Vendor { id : 7408u16 , name : "Akitio" , devices : & [] }),
        (5442, Vendor { id : 5442u16 , name : "Concurrent Real-Time" , devices : & [Device { vendor_id : 5442u16 , id : 37472u16 , name : "RCIM-II Real-Time Clock & Interrupt Module" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37489u16 , name : "RCIM-III Real-Time Clock & Interrupt Module (PCIe)" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37490u16 , name : "Pulse Width Modulator Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37491u16 , name : "RCIM-IV Real-Time Clock & Interrupt Module (PCIe)" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37495u16 , name : "5 Volt Delta Sigma Converter Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37496u16 , name : "10 Volt Delta Sigma Converter Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37511u16 , name : "Analog Output Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37520u16 , name : "FPGA Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37632u16 , name : "Universal Exhaust Gas Oxygen Sensor Simulator" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37648u16 , name : "Digital Programmable Resistor" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37664u16 , name : "Arria 10 FPGA Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37680u16 , name : "Cyclone V FPGA Card" , subsystems : & [] } , Device { vendor_id : 5442u16 , id : 37712u16 , name : "Analog Input Card" , subsystems : & [] }] }),
        (4261, Vendor { id : 4261u16 , name : "Smart Link Ltd." , devices : & [Device { vendor_id : 4261u16 , id : 12370u16 , name : "SmartPCI562 56K Modem" , subsystems : & [] } , Device { vendor_id : 4261u16 , id : 21577u16 , name : "SmartPCI561 modem" , subsystems : & [] }] }),
        (5109, Vendor { id : 5109u16 , name : "Kansai Electric Co. Ltd" , devices : & [] }),
        (5306, Vendor { id : 5306u16 , name : "INTERNIX Inc." , devices : & [Device { vendor_id : 5306u16 , id : 1536u16 , name : "ARC-PCI/22" , subsystems : & [] }] }),
        (256, Vendor { id : 256u16 , name : "nCipher Security" , devices : & [] }),
        (5041, Vendor { id : 5041u16 , name : "Tamura corporation" , devices : & [] }),
        (4362, Vendor { id : 4362u16 , name : "Siemens AG" , devices : & [Device { vendor_id : 4362u16 , id : 2u16 , name : "Pirahna 2-port" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 5u16 , name : "Tulip controller, power management, switch extender" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 6u16 , name : "FSC PINC (I/O-APIC)" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 21u16 , name : "FSC Multiprocessor Interrupt Controller" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 29u16 , name : "FSC Copernicus Management Controller" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 123u16 , name : "FSC Remote Service Controller, mailbox device" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 124u16 , name : "FSC Remote Service Controller, shared memory device" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 125u16 , name : "FSC Remote Service Controller, SMIC device" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 8449u16 , name : "HST SAPHIR V Primary PCI (ISDN/PMx)" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 8450u16 , name : "DSCC4 PEB/PEF 20534 DMA Supported Serial Communication Controller with 4 Channels" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 8452u16 , name : "Eicon Diva 2.02 compatible passive ISDN card" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 12545u16 , name : "HiPath 4000 PCI card" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 12609u16 , name : "SIMATIC NET CP 5611 / 5621" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 12610u16 , name : "SIMATIC NET CP 5613 / 5614" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 12611u16 , name : "SIMATIC NET CP 1613" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16417u16 , name : "SIMATIC NET CP 5512 (Profibus and MPI Cardbus Adapter)" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16425u16 , name : "SIMATIC NET CP 5613 A2" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 16425u16 , name : "SIMATIC NET CP 5613 A2" } , SubSystem { subvendor : 4362u16 , subdevice : 49193u16 , name : "SIMATIC NET CP 5614 A2" }] } , Device { vendor_id : 4362u16 , id : 16437u16 , name : "SIMATIC NET CP 1613 A2" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16438u16 , name : "SIMATIC NET CP 1616" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16440u16 , name : "SIMATIC NET CP 1604" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16489u16 , name : "SIMATIC NET CP 5623" , subsystems : & [SubSystem { subvendor : 4362u16 , subdevice : 16489u16 , name : "SIMATIC NET CP 5623" } , SubSystem { subvendor : 4362u16 , subdevice : 49257u16 , name : "SIMATIC NET CP 5624" }] } , Device { vendor_id : 4362u16 , id : 16508u16 , name : "SIMATIC NET CP 5612" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16509u16 , name : "SIMATIC NET CP 5613 A3" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16510u16 , name : "SIMATIC NET CP 5622" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16515u16 , name : "SIMATIC NET CP 5614 A3" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 16516u16 , name : "SIMATIC NET CP 1626" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 18754u16 , name : "FPGA I-Bus Tracer for MBD" , subsystems : & [] } , Device { vendor_id : 4362u16 , id : 24864u16 , name : "SZB6120" , subsystems : & [] }] }),
        (4559, Vendor { id : 4559u16 , name : "Pioneer Electronic Corporation" , devices : & [] }),
        (4750, Vendor { id : 4750u16 , name : "Hoontech Corporation/Samho Multi Tech Ltd." , devices : & [Device { vendor_id : 4750u16 , id : 8u16 , name : "ST128 WSS/SB" , subsystems : & [] } , Device { vendor_id : 4750u16 , id : 9u16 , name : "ST128 SAM9407" , subsystems : & [] } , Device { vendor_id : 4750u16 , id : 10u16 , name : "ST128 Game Port" , subsystems : & [] } , Device { vendor_id : 4750u16 , id : 11u16 , name : "ST128 MPU Port" , subsystems : & [] } , Device { vendor_id : 4750u16 , id : 12u16 , name : "ST128 Ctrl Port" , subsystems : & [] }] }),
        (4803, Vendor { id : 4803u16 , name : "Holtek Microelectronics Inc" , devices : & [Device { vendor_id : 4803u16 , id : 88u16 , name : "PCI NE2K Ethernet" , subsystems : & [] } , Device { vendor_id : 4803u16 , id : 21912u16 , name : "PCI NE2K Ethernet" , subsystems : & [] }] }),
        (5550, Vendor { id : 5550u16 , name : "Amersham Pharmacia Biotech" , devices : & [] }),
        (6987, Vendor { id : 6987u16 , name : "Marvell Technology Group Ltd." , devices : & [Device { vendor_id : 6987u16 , id : 256u16 , name : "88F3700 [Armada 3700 Family] ARM SoC" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 1600u16 , name : "88SE9128 SATA III 6Gb/s RAID Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 4242u16 , name : "88SS1092 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 4448u16 , name : "88NV1160 PCIe x2 NVMe SSD Controller (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 4897u16 , name : "88SS1321 NVMe SSD Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 8769u16 , name : "88NR2241 Non-Volatile memory controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8466u16 , name : "BOSS-N1 Monolithic" } , SubSystem { subvendor : 4136u16 , subdevice : 8467u16 , name : "BOSS-N1 Modular" } , SubSystem { subvendor : 4136u16 , subdevice : 8529u16 , name : "BOSS-N1 Modular ET" } , SubSystem { subvendor : 4136u16 , subdevice : 8598u16 , name : "ROR-N1" } , SubSystem { subvendor : 4136u16 , subdevice : 8838u16 , name : "BOSS-N1 DC-MHS" } , SubSystem { subvendor : 4136u16 , subdevice : 8839u16 , name : "BOSS-N1 Modular DC-MHS" } , SubSystem { subvendor : 4136u16 , subdevice : 9136u16 , name : "eBOSS-N1 DC-MHS" } , SubSystem { subvendor : 5520u16 , subdevice : 758u16 , name : "NS204i-p Gen10+ Boot Controller" } , SubSystem { subvendor : 6987u16 , subdevice : 8769u16 , name : "Santa Cruz NVMe Host Adapter" } , SubSystem { subvendor : 7062u16 , subdevice : 16384u16 , name : "WD_BLACK AN1500 NVMe SSD" } , SubSystem { subvendor : 7497u16 , subdevice : 774u16 , name : "ThinkSystem M.2 NVMe 2-Bay RAID Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 775u16 , name : "ThinkSystem 7mm NVMe 2-Bay Rear RAID Enablement Kit" } , SubSystem { subvendor : 8317u16 , subdevice : 2048u16 , name : "TrustRAID B310n" } , SubSystem { subvendor : 19538u16 , subdevice : 38209u16 , name : "LRNV9541 2-port M.2 NVMe Raid Adapter" }] } , Device { vendor_id : 6987u16 , id : 11074u16 , name : "88W8997 2.4/5 GHz Dual-Band 2x2 Wi-Fi® 5 (802.11ac) + Bluetooth® 5.3 Solution" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 11075u16 , name : "NXP 88W9098 Wi-Fi 6 (ax) MAC #1" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 11076u16 , name : "NXP 88W9098 Wi-Fi 6 (ax) MAC #2" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 11077u16 , name : "NXP 88W9098 Bluetooth 5.3" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 11094u16 , name : "NXP IW620 Wi-Fi 6 (ax) + Bluetooth 5.3 Combo" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 25166u16 , name : "DAWICONTROL DC-624e RAID" , subsystems : & [SubSystem { subvendor : 56467u16 , subdevice : 25166u16 , name : "DC-624e RAID" }] } , Device { vendor_id : 6987u16 , id : 37152u16 , name : "88SE9120 SATA 6Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37154u16 , name : "88SE912x SATA 6Gb/s Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37155u16 , name : "88SE9123 PCIe SATA 6.0 Gb/s controller" , subsystems : & [SubSystem { subvendor : 56467u16 , subdevice : 24590u16 , name : "DC-6xxe series SATA 6G controller" }] } , Device { vendor_id : 6987u16 , id : 37157u16 , name : "88SE9125 PCIe SATA 6.0 Gb/s controller" , subsystems : & [SubSystem { subvendor : 19538u16 , subdevice : 38421u16 , name : "LRST9615 4-port SATA3(6Gb/s) Exchange Adapter" }] } , Device { vendor_id : 6987u16 , id : 37160u16 , name : "88SE9128 PCIe SATA 6 Gb/s RAID controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37168u16 , name : "88SE9128 PCIe SATA 6 Gb/s RAID controller with HyperDuo" , subsystems : & [SubSystem { subvendor : 4163u16 , subdevice : 33848u16 , name : "P8P67 Deluxe Motherboard" }] } , Device { vendor_id : 6987u16 , id : 37232u16 , name : "88SE9170 PCIe 2.0 x1 2-port SATA 6 Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37233u16 , name : "88SE9171 PCIe 2.0 x1 1-port SATA 6 Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37234u16 , name : "88SE9172 SATA 6Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37240u16 , name : "88SE9170 PCIe SATA 6Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37242u16 , name : "88SE9172 SATA III 6Gb/s Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37250u16 , name : "88SE9182 PCIe 2.0 x2 2-port SATA 6 Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37251u16 , name : "88SS9183 PCIe SSD Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37254u16 , name : "88SE9186 6Gb/s SATA 6Gb/s Controller [AHCI mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37258u16 , name : "88SE9182 PCIe SATA 6 Gb/s controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37266u16 , name : "88SE9172 SATA III 6Gb/s RAID Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37280u16 , name : "88SE912x SATA 6Gb/s Controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37283u16 , name : "88SE9128 PCIe SATA 6 Gb/s controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37284u16 , name : "88SE912x IDE Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37296u16 , name : "88SE9172 SATA 6 Gb/s controller [IDE mode]" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37397u16 , name : "88SE9215 PCIe 2.0 x1 4-port SATA 6 Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37408u16 , name : "88SE9220 PCIe 2.0 x2 2-port SATA 6 Gb/s RAID Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37424u16 , name : "88SE9230 PCIe 2.0 x2 4-port SATA 6 Gb/s RAID Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8150u16 , name : "BOSS-S1 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8159u16 , name : "BOSS-S1 Modular" } , SubSystem { subvendor : 4136u16 , subdevice : 8162u16 , name : "BOSS-S1 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8208u16 , name : "BOSS-S2 Adapter" } , SubSystem { subvendor : 4136u16 , subdevice : 8800u16 , name : "BOSS-S1 Modular" } , SubSystem { subvendor : 4136u16 , subdevice : 8801u16 , name : "BOSS-S2 Adapter" } , SubSystem { subvendor : 7124u16 , subdevice : 115u16 , name : "RS0200L6R2iM2" } , SubSystem { subvendor : 7497u16 , subdevice : 768u16 , name : "ThinkSystem M.2 with Mirroring Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 769u16 , name : "ThinkSystem SR630 x16 PCIE with 4 SATA ports Riser" } , SubSystem { subvendor : 7497u16 , subdevice : 770u16 , name : "ThinkSystem SE350 M.2 SATA 4-Bay Data RAID Mirroring Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 771u16 , name : "ThinkSystem SE350 M.2 SATA 4-Bay Data RAID Mirroring Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 772u16 , name : "ThinkSystem M.2 SATA 2-Bay RAID Enablement Kit" } , SubSystem { subvendor : 7497u16 , subdevice : 773u16 , name : "ThinkSystem 7mm SATA 2-Bay Rear RAID Enablement Kit" } , SubSystem { subvendor : 19538u16 , subdevice : 38448u16 , name : "LRST9630 4-port SATA3(6Gb/s) Raid Adapter" } , SubSystem { subvendor : 56467u16 , subdevice : 25166u16 , name : "DC-624e RAID" }] } , Device { vendor_id : 6987u16 , id : 37429u16 , name : "88SE9235 PCIe 2.0 x2 4-port SATA 6 Gb/s Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 37957u16 , name : "88SE9445 PCIe 2.0 x4 4-Port SAS/SATA 6 Gbps RAID Controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 38016u16 , name : "88SE9480 SAS/SATA 6Gb/s RAID controller" , subsystems : & [] } , Device { vendor_id : 6987u16 , id : 38021u16 , name : "88SE9485 SAS/SATA 6Gb/s controller" , subsystems : & [] }] }),
        (5472, Vendor { id : 5472u16 , name : "Terayon Communications Systems" , devices : & [] }),
        (16708, Vendor { id : 16708u16 , name : "Alpha Data" , devices : & [Device { vendor_id : 16708u16 , id : 68u16 , name : "ADM-XRCIIPro" , subsystems : & [] }] }),
        (6001, Vendor { id : 6001u16 , name : "InnoVISION Multimedia Ltd." , devices : & [] }),
        (7157, Vendor { id : 7157u16 , name : "Greenliant" , devices : & [Device { vendor_id : 7157u16 , id : 4096u16 , name : "G7200 series U.2 NVMe SSD" , subsystems : & [] }] }),
        (4533, Vendor { id : 4533u16 , name : "Radstone Technology Plc" , devices : & [] }),
        (4454, Vendor { id : 4454u16 , name : "Broadcom" , devices : & [Device { vendor_id : 4454u16 , id : 0u16 , name : "CMIC-LE" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 5u16 , name : "CNB20-LE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 6u16 , name : "CNB20HE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 7u16 , name : "CNB20-LE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 8u16 , name : "CNB20HE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 9u16 , name : "CNB20LE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 16u16 , name : "CIOB30" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 17u16 , name : "CMIC-HE" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 18u16 , name : "CMIC-WS Host Bridge (GC-LE chipset)" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 19u16 , name : "CNB20-HE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 20u16 , name : "CMIC-LE Host Bridge (GC-LE chipset)" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 21u16 , name : "CMIC-GC Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 22u16 , name : "CMIC-GC Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 23u16 , name : "GCNB-LE Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 49u16 , name : "HT1100 HPX0 HT Host Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 54u16 , name : "BCM5785 [HT1000] PCI/PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 257u16 , name : "CIOB-X2 PCI-X I/O Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 259u16 , name : "EPB PCI-Express to PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 260u16 , name : "BCM5785 [HT1000] PCI/PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 272u16 , name : "CIOB-E I/O Bridge with Gigabit Ethernet" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 304u16 , name : "BCM5780 [HT2000] PCI-X bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 306u16 , name : "BCM5780 [HT2000] PCI-Express Bridge" , subsystems : & [SubSystem { subvendor : 4454u16 , subdevice : 306u16 , name : "HT2000 PCI-Express bridge" }] } , Device { vendor_id : 4454u16 , id : 320u16 , name : "HT2100 PCI-Express Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 321u16 , name : "HT2100 PCI-Express Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 322u16 , name : "HT2100 PCI-Express Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 324u16 , name : "HT2100 PCI-Express Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 512u16 , name : "OSB4 South Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 513u16 , name : "CSB5 South Bridge" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" }] } , Device { vendor_id : 4454u16 , id : 515u16 , name : "CSB6 South Bridge" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4114u16 , name : "PRIMERGY RX/TX series" }] } , Device { vendor_id : 4454u16 , id : 517u16 , name : "BCM5785 [HT1000] Legacy South Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 529u16 , name : "OSB4 IDE Controller" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 530u16 , name : "CSB5 IDE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 330u16 , name : "PowerEdge 1750" } , SubSystem { subvendor : 4136u16 , subdevice : 33035u16 , name : "PowerEdge 1650/2550" } , SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" }] } , Device { vendor_id : 4454u16 , id : 531u16 , name : "CSB6 RAID/IDE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 16692u16 , name : "PowerEdge 600SC" } , SubSystem { subvendor : 4136u16 , subdevice : 49460u16 , name : "Poweredge SC600" } , SubSystem { subvendor : 5940u16 , subdevice : 4114u16 , name : "PRIMERGY RX/TX series onboard IDE" }] } , Device { vendor_id : 4454u16 , id : 532u16 , name : "BCM5785 [HT1000] IDE" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 HT1000 IDE" }] } , Device { vendor_id : 4454u16 , id : 535u16 , name : "CSB6 IDE Controller" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 16692u16 , name : "Poweredge SC600" }] } , Device { vendor_id : 4454u16 , id : 539u16 , name : "HT1100 HD Audio" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 544u16 , name : "OSB4/CSB5 OHCI USB Controller" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" }] } , Device { vendor_id : 4454u16 , id : 545u16 , name : "CSB6 OHCI USB Controller" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4114u16 , name : "PRIMERGY RX/TX series onboard OHCI" }] } , Device { vendor_id : 4454u16 , id : 547u16 , name : "BCM5785 [HT1000] USB" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 HT1000 USB Controller" } , SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "PowerEdge T605 HT1000 USB Controller" }] } , Device { vendor_id : 4454u16 , id : 549u16 , name : "CSB5 LPC bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 551u16 , name : "GCLE-2 Host Bridge" , subsystems : & [SubSystem { subvendor : 5940u16 , subdevice : 4114u16 , name : "PRIMERGY RX/TX series" }] } , Device { vendor_id : 4454u16 , id : 560u16 , name : "CSB5 LPC bridge" , subsystems : & [SubSystem { subvendor : 19539u16 , subdevice : 4224u16 , name : "CT8 mainboard" }] } , Device { vendor_id : 4454u16 , id : 564u16 , name : "BCM5785 [HT1000] LPC" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 HT1000 LPC" } , SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "PowerEdge T605 HT1000 LPC" }] } , Device { vendor_id : 4454u16 , id : 565u16 , name : "BCM5785 [HT1000] XIOAPIC0-2" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 568u16 , name : "BCM5785 [HT1000] WDTimer" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 576u16 , name : "K2 SATA" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 577u16 , name : "RAIDCore RC4000" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 578u16 , name : "RAIDCore BC4000" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 586u16 , name : "BCM5785 [HT1000] SATA (Native SATA Mode)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 523u16 , name : "PowerEdge T605 onboard SATA Controller" }] } , Device { vendor_id : 4454u16 , id : 587u16 , name : "BCM5785 [HT1000] SATA (PATA/IDE Mode)" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 517u16 , name : "PowerEdge 2970 HT1000 SATA controller" }] } , Device { vendor_id : 4454u16 , id : 1030u16 , name : "HT1100 PCI-X Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1032u16 , name : "HT1100 Legacy Device" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1034u16 , name : "HT1100 ISA-LPC Bridge" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 547u16 , name : "PowerEdge R905 HT1100 ISA-LPC Bridge" }] } , Device { vendor_id : 4454u16 , id : 1040u16 , name : "HT1100 SATA Controller (Native SATA Mode)" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1041u16 , name : "HT1100 SATA Controller (PATA / IDE Mode)" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1042u16 , name : "HT1100 USB OHCI Controller" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1044u16 , name : "HT1100 USB EHCI Controller" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1046u16 , name : "HT1100 USB EHCI Controller (with Debug Port)" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1056u16 , name : "HT1100 PCI-Express Bridge" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1057u16 , name : "HT1100 SAS/SATA Controller" , subsystems : & [] } , Device { vendor_id : 4454u16 , id : 1058u16 , name : "HT1100 PCI-Express Bridge" , subsystems : & [] }] }),
        (4127, Vendor { id : 4127u16 , name : "PictureTel" , devices : & [] }),
        (5212, Vendor { id : 5212u16 , name : "Cryptek" , devices : & [] }),
        (7260, Vendor { id : 7260u16 , name : "SK hynix" , devices : & [Device { vendor_id : 7260u16 , id : 4201u16 , name : "PCB01 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4738u16 , name : "PC300 NVMe Solid State Drive 128GB" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4739u16 , name : "PC300 NVMe Solid State Drive 256GB" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4740u16 , name : "PC300 NVMe Solid State Drive 512GB" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4741u16 , name : "PC300 NVMe Solid State Drive 1TB" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4903u16 , name : "BC501 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 4921u16 , name : "BC511 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5380u16 , name : "PC400 NVMe SSD" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5415u16 , name : "PC401 NVMe Solid State Drive 256GB" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5671u16 , name : "PC601 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5689u16 , name : "PC611 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5945u16 , name : "BC701 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 5962u16 , name : "Gold P31/BC711/PC711 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 6489u16 , name : "Platinum P41/PC801 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 6505u16 , name : "PC811 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 7513u16 , name : "BC901 NVMe Solid State Drive (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 8041u16 , name : "PVC10 NVMe Solid State Drive (DRAM-less)" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 8708u16 , name : "960GB TLC PCIe Gen3 x4 NVMe M.2 22110" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 9255u16 , name : "PE6010 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 9257u16 , name : "PE6011 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 9275u16 , name : "PE6110 NVMe Solid State Drive" , subsystems : & [SubSystem { subvendor : 7260u16 , subdevice : 256u16 , name : "PE6110 NVMe Solid State Drive" }] } , Device { vendor_id : 7260u16 , id : 9511u16 , name : "PE4010 NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 10297u16 , name : "PE8000 Series NVMe Solid State Drive" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8515u16 , name : "DC NVMe SED PE8010 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8516u16 , name : "DC NVMe PE8010 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8517u16 , name : "DC NVMe SED PE8010 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8518u16 , name : "DC NVMe PE8010 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8519u16 , name : "DC NVMe SED PE8010 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8520u16 , name : "DC NVMe PE8010 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8521u16 , name : "DC NVMe SED PE8010 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8522u16 , name : "DC NVMe PE8010 RI U.2 7.68TB" } , SubSystem { subvendor : 7260u16 , subdevice : 256u16 , name : "PE8000 Series NVMe Solid State Drive" }] } , Device { vendor_id : 7260u16 , id : 10313u16 , name : "PE81x0 U.2/3 NVMe Solid State Drive" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8802u16 , name : "DC NVMe OPAL PE8110 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8803u16 , name : "DC NVMe OPAL PE8110 RI U.2 1920GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8804u16 , name : "DC NVMe OPAL PE8110 RI U.2 3840GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8805u16 , name : "DC NVMe OPAL PE8110 RI U.2 7680GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8806u16 , name : "DC NVMe ISE PE8110 RI U.2 960GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8807u16 , name : "DC NVMe ISE PE8110 RI U.2 1920GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8808u16 , name : "DC NVMe ISE PE8110 RI U.2 3840GB" } , SubSystem { subvendor : 4136u16 , subdevice : 8809u16 , name : "DC NVMe ISE PE8110 RI U.2 7680GB" } , SubSystem { subvendor : 7260u16 , subdevice : 257u16 , name : "PE81x0 U.2/3 NVMe Solid State Drive" }] } , Device { vendor_id : 7260u16 , id : 10314u16 , name : "PE8110 Series NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 10825u16 , name : "PE9110 Series NVMe Solid State Drive" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 10841u16 , name : "PE9010 Series NVMe Solid State Drives" , subsystems : & [] } , Device { vendor_id : 7260u16 , id : 11097u16 , name : "Px10x0 Series NVMe Solid State Drives" , subsystems : & [SubSystem { subvendor : 4136u16 , subdevice : 8853u16 , name : "NVMe ISE PS1010 RI U.2 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8854u16 , name : "NVMe ISE PS1010 RI U.2 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8855u16 , name : "NVMe ISE PS1010 RI U.2 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8856u16 , name : "NVMe ISE PS1010 RI U.2 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8857u16 , name : "NVMe ISE PS1030 MU U.2 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8858u16 , name : "NVMe ISE PS1030 MU U.2 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8859u16 , name : "NVMe ISE PS1030 MU U.2 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8860u16 , name : "NVMe ISE PS1030 MU U.2 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8871u16 , name : "NVMe ISE PS1010 RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8872u16 , name : "NVMe ISE PS1010 RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8873u16 , name : "NVMe ISE PS1010 RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8874u16 , name : "NVMe ISE PS1010 RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8875u16 , name : "NVMe ISE PS1030 MU E3.S 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8876u16 , name : "NVMe ISE PS1030 MU E3.S 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8877u16 , name : "NVMe ISE PS1030 MU E3.S 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8878u16 , name : "NVMe ISE PS1030 MU E3.S 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8924u16 , name : "NVMe FIPS PS1010 RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8925u16 , name : "NVMe FIPS PS1010 RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8926u16 , name : "NVMe FIPS PS1010 RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8927u16 , name : "NVMe FIPS PS1010 RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8928u16 , name : "NVMe FIPS PS1030 MU E3.S 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8929u16 , name : "NVMe FIPS PS1030 MU E3.S 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8930u16 , name : "NVMe FIPS PS1030 MU E3.S 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8931u16 , name : "NVMe FIPS PS1030 MU E3.S 12.8TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8952u16 , name : "NVMe ISE PE1010 RI E3.S 1.92TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8953u16 , name : "NVMe ISE PE1010 RI E3.S 3.84TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8954u16 , name : "NVMe ISE PE1010 RI E3.S 7.68TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8955u16 , name : "NVMe ISE PE1010 RI E3.S 15.36TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8956u16 , name : "NVMe ISE PE1030 MU E3.S 1.6TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8957u16 , name : "NVMe ISE PE1030 MU E3.S 3.2TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8958u16 , name : "NVMe ISE PE1030 MU E3.S 6.4TB" } , SubSystem { subvendor : 4136u16 , subdevice : 8959u16 , name : "NVMe ISE PE1030 MU E3.S 12.8TB" }] }] }),
        (5011, Vendor { id : 5011u16 , name : "Moxa Technologies Co Ltd" , devices : & [Device { vendor_id : 5011u16 , id : 1u16 , name : "UC7000 Serial" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4128u16 , name : "CP-102 (2-port RS-232 PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4129u16 , name : "CP-102UL (2-port RS-232 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4130u16 , name : "CP-102U (2-port RS-232 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4131u16 , name : "CP-102UF" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4132u16 , name : "CP-102E (2-port RS-232 Smart PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4133u16 , name : "CP-102EL (2-port RS-232 Smart PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4160u16 , name : "Smartio C104H/PCI" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4161u16 , name : "CP104U (4-port RS-232 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4162u16 , name : "CP104JU (4-port RS-232 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4163u16 , name : "CP104EL (4-port RS-232 Smart PCI Express)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4164u16 , name : "POS104UL (4-port RS-232 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4165u16 , name : "CP-104EL-A (4-port RS-232 PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4224u16 , name : "CB108 (8-port RS-232 PC/104-plus Module)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4416u16 , name : "CT-114 series" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4417u16 , name : "Industrio CP-114" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4418u16 , name : "CB114 (4-port RS-232/422/485 PC/104-plus Module)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4419u16 , name : "CP-114UL (4-port RS-232/422/485 Smart Universal PCI Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4420u16 , name : "CP-114EL (4-port RS-232/422/485 Smart PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4480u16 , name : "CP118U (8-port RS-232/422/485 Smart Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4481u16 , name : "CP118EL (8-port RS-232/422/485 Smart PCI Express)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4482u16 , name : "CP-118EL-A (8-port RS-232/422/485 PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4896u16 , name : "CP132 (2-port RS-422/485 PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4897u16 , name : "CP132U (2-Port RS-422/485 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4898u16 , name : "CP-132EL (2-port RS-422/485 Smart PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4928u16 , name : "CP134U (4-Port RS-422/485 Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4929u16 , name : "CB134I (4-port RS-422/485 PC/104-plus Module)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 4992u16 , name : "CP138U (8-port RS-232/422/485 Smart Universal PCI)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 5760u16 , name : "Smartio C168H/PCI" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 5761u16 , name : "CP-168U V2 Smart Serial Board (8-port RS-232)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 5762u16 , name : "CP-168EL (8-port RS-232 Smart PCI Express)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 5763u16 , name : "CP-168EL-A (8-port RS-232 PCI Express Serial Board)" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 8256u16 , name : "Intellio CP-204J" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 8576u16 , name : "Intellio C218 Turbo PCI" , subsystems : & [] } , Device { vendor_id : 5011u16 , id : 12800u16 , name : "Intellio C320 Turbo PCI" , subsystems : & [] }] }),
        (5241, Vendor { id : 5241u16 , name : "TORAY Industries Inc" , devices : & [] }),
        (7972, Vendor { id : 7972u16 , name : "xFusion Digital Technologies Co., Ltd." , devices : & [Device { vendor_id : 7972u16 , id : 4184u16 , name : "EP500/EP600 NVMe SSD" , subsystems : & [SubSystem { subvendor : 7972u16 , subdevice : 4372u16 , name : "EP500 NVMe SSD(RI)" } , SubSystem { subvendor : 7972u16 , subdevice : 4373u16 , name : "EP600 NVMe SSD(MU)" }] }] }),
        (43527, Vendor { id : 43527u16 , name : "iTuner" , devices : & [] }),
        (4233, Vendor { id : 4233u16 , name : "Data General Corporation" , devices : & [] }),
        (8158, Vendor { id : 8158u16 , name : "Kratos Defense & Security Solutions, Inc." , devices : & [Device { vendor_id : 8158u16 , id : 4389u16 , name : "OpenEdge 1125P" , subsystems : & [] } , Device { vendor_id : 8158u16 , id : 9472u16 , name : "OpenEdge 2500P" , subsystems : & [] }] }),
        (5180, Vendor { id : 5180u16 , name : "Amlogic Inc" , devices : & [] }),
        (6344, Vendor { id : 6344u16 , name : "Cray Inc" , devices : & [] }),
        (32896, Vendor { id : 32896u16 , name : "Chengdu Storeswift Technology Co., Ltd." , devices : & [Device { vendor_id : 32896u16 , id : 16406u16 , name : "CX4016A NVMe SSD Controller" , subsystems : & [] }] }),
        (4439, Vendor { id : 4439u16 , name : "Avsys Corporation" , devices : & [] }),
        (2001, Vendor { id : 2001u16 , name : "D-Link System Inc" , devices : & [] }),
        (5363, Vendor { id : 5363u16 , name : "BroadLogic" , devices : & [Device { vendor_id : 5363u16 , id : 8240u16 , name : "2030 DVB-S Satellite Receiver" , subsystems : & [] } , Device { vendor_id : 5363u16 , id : 8245u16 , name : "2035 DVB-S Satellite Receiver" , subsystems : & [] } , Device { vendor_id : 5363u16 , id : 8272u16 , name : "2050 DVB-T Terrestrial (Cable) Receiver" , subsystems : & [] } , Device { vendor_id : 5363u16 , id : 8288u16 , name : "2060 ATSC Terrestrial (Cable) Receiver" , subsystems : & [] }] }),
        (5206, Vendor { id : 5206u16 , name : "Advanced Hardware Architectures" , devices : & [] }),
        (6041, Vendor { id : 6041u16 , name : "Belkin" , devices : & [Device { vendor_id : 6041u16 , id : 24577u16 , name : "F5D6001 Wireless PCI Card [Realtek RTL8180]" , subsystems : & [] } , Device { vendor_id : 6041u16 , id : 24608u16 , name : "F5D6020 v3000 Wireless PCMCIA Card [Realtek RTL8180]" , subsystems : & [] } , Device { vendor_id : 6041u16 , id : 24672u16 , name : "F5D6060 Wireless PDA Card" , subsystems : & [] } , Device { vendor_id : 6041u16 , id : 28687u16 , name : "F5D7000 v7000 Wireless G Desktop Card [Realtek RTL8185]" , subsystems : & [] } , Device { vendor_id : 6041u16 , id : 28703u16 , name : "F5D7010 v7000 Wireless G Notebook Card [Realtek RTL8185]" , subsystems : & [] }] }),
        (5603, Vendor { id : 5603u16 , name : "Networth Technologies Inc" , devices : & [] }),
        (4427, Vendor { id : 4427u16 , name : "Canopus Co., Ltd" , devices : & [] }),
        (4956, Vendor { id : 4956u16 , name : "Quatech Inc" , devices : & [Device { vendor_id : 4956u16 , id : 16u16 , name : "QSC-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 32u16 , name : "DSC-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 48u16 , name : "DSC-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 64u16 , name : "QSC-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 80u16 , name : "ESC-100D" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 96u16 , name : "ESC-100M" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 240u16 , name : "MPAC-100 Synchronous Serial Card (Zilog 85230)" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 288u16 , name : "QSCP-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 304u16 , name : "DSCP-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 320u16 , name : "QSCP-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 336u16 , name : "DSCP-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 368u16 , name : "QSCLP-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 384u16 , name : "DSCLP-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 385u16 , name : "DSC-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 400u16 , name : "SSCLP-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 416u16 , name : "QSCLP-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 432u16 , name : "DSCLP-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 433u16 , name : "DSC-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 448u16 , name : "SSCLP-200/300" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 480u16 , name : "ESC(LP)-100" , subsystems : & [] } , Device { vendor_id : 4956u16 , id : 600u16 , name : "DSPSX-200/300" , subsystems : & [] }] }),
        (4292, Vendor { id : 4292u16 , name : "Award Software International Inc." , devices : & [] }),
        (5410, Vendor { id : 5410u16 , name : "MainPine Ltd" , devices : & [Device { vendor_id : 5410u16 , id : 256u16 , name : "PCI <-> IOBus Bridge" , subsystems : & [SubSystem { subvendor : 5410u16 , subdevice : 512u16 , name : "RockForceDUO 2 Port V.92/V.44 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 768u16 , name : "RockForceQUATRO 4 Port V.92/V.44 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 1024u16 , name : "RockForceDUO+ 2 Port V.92/V.44 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 1280u16 , name : "RockForceQUATRO+ 4 Port V.92/V.44 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 1536u16 , name : "RockForce+ 2 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 1792u16 , name : "RockForce+ 4 Port V.90 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 2048u16 , name : "RockForceOCTO+ 8 Port V.92/V.44 Data/Fax/Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 3072u16 , name : "RockForceDUO+ 2 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 3328u16 , name : "RockForceQUATRO+ 4 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 7424u16 , name : "RockForceOCTO+ 8 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 8192u16 , name : "RockForceD1 1 Port V.90 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 8448u16 , name : "RockForceF1 1 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 8704u16 , name : "RockForceD2 2 Port V.90 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 8960u16 , name : "RockForceF2 2 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 9216u16 , name : "RockForceD4 4 Port V.90 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 9472u16 , name : "RockForceF4 4 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 9728u16 , name : "RockForceD8 8 Port V.90 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 9984u16 , name : "RockForceF8 8 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 12288u16 , name : "IQ Express D1 - 1 Port V.92 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 12544u16 , name : "IQ Express F1 - 1 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 12800u16 , name : "IQ Express D2 - 2 Port V.92 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 13056u16 , name : "IQ Express F2 - 2 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 13312u16 , name : "IQ Express D4 - 4 Port V.92 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 13568u16 , name : "IQ Express F4 - 4 Port V.34 Super-G3 Fax Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 15360u16 , name : "IQ Express D8 - 8 Port V.92 Data Modem" } , SubSystem { subvendor : 5410u16 , subdevice : 15616u16 , name : "IQ Express F8 - 8 Port V.34 Super-G3 Fax Modem" }] } , Device { vendor_id : 5410u16 , id : 16384u16 , name : "PCI Express UART" , subsystems : & [SubSystem { subvendor : 5410u16 , subdevice : 16385u16 , name : "IQ Express 1-port V.34 Super-G3 Fax" } , SubSystem { subvendor : 5410u16 , subdevice : 16386u16 , name : "IQ Express 2-port V.34 Super-G3 Fax" } , SubSystem { subvendor : 5410u16 , subdevice : 16388u16 , name : "IQ Express 4-port V.34 Super-G3 Fax" } , SubSystem { subvendor : 5410u16 , subdevice : 16392u16 , name : "IQ Express 8-port V.34 Super-G3 Fax" } , SubSystem { subvendor : 5410u16 , subdevice : 16640u16 , name : "IQ Express SideBand" }] }] }),
        (8468, Vendor { id : 8468u16 , name : "EigenQ, Inc." , devices : & [Device { vendor_id : 8468u16 , id : 7u16 , name : "QMA Board M.2 Gen2" , subsystems : & [] } , Device { vendor_id : 8468u16 , id : 10u16 , name : "QMA Board PCIe Gen2" , subsystems : & [] } , Device { vendor_id : 8468u16 , id : 26u16 , name : "QMA Board PCIe Gen3" , subsystems : & [] }] }),
        (5504, Vendor { id : 5504u16 , name : "Primex Aerospace Co" , devices : & [] }),
        (6871, Vendor { id : 6871u16 , name : "Spectracom Corporation" , devices : & [Device { vendor_id : 6871u16 , id : 32768u16 , name : "TSync-PCIe Time Code Processor" , subsystems : & [] } , Device { vendor_id : 6871u16 , id : 37120u16 , name : "TPRO-PCI-66U Timecode Reader/Generator" , subsystems : & [] } , Device { vendor_id : 6871u16 , id : 40960u16 , name : "OCP-TAP [ARTCard]" , subsystems : & [] }] }),
        (5416, Vendor { id : 5416u16 , name : "ACKSYS" , devices : & [] }),
        (4800, Vendor { id : 4800u16 , name : "Infimed" , devices : & [] }),
        (4539, Vendor { id : 4539u16 , name : "Pyramid Technology" , devices : & [] }),
        (7093, Vendor { id : 7093u16 , name : "Quantenna Communications, Inc." , devices : & [] }),
        (4619, Vendor { id : 4619u16 , name : "Adaptive Solutions" , devices : & [] }),
        (5171, Vendor { id : 5171u16 , name : "Eltec Elektronik GmbH" , devices : & [Device { vendor_id : 5171u16 , id : 2u16 , name : "XRD_FG (PCGR-200) [95510214H]" , subsystems : & [] } , Device { vendor_id : 5171u16 , id : 6u16 , name : "XRD_FGx Opto (PCGR-300) [95510215H]" , subsystems : & [] }] }),
        (7913, Vendor { id : 7913u16 , name : "SUSE LLC" , devices : & [] }),
        (4806, Vendor { id : 4806u16 , name : "Mitani Corporation" , devices : & [] }),
        (14654, Vendor { id : 14654u16 , name : "Lenovo (wrong ID)" , devices : & [] }),
    ],
};
static CLASSES: phf::Map<u8, Class> = ::phf::Map {
    key: 12913932095322966823,
    disps: &[
        (3, 16),
        (9, 1),
        (2, 5),
        (5, 0),
        (0, 5),
    ],
    entries: &[
        (12, Class { id : 12u8 , name : "Serial bus controller" , subclasses : & [Subclass { class_id : 12u8 , id : 0u8 , name : "FireWire (IEEE 1394)" , prog_ifs : & [ProgIf { id : 0u8 , name : "Generic" } , ProgIf { id : 16u8 , name : "OHCI" }] } , Subclass { class_id : 12u8 , id : 1u8 , name : "ACCESS Bus" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 2u8 , name : "SSA" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 3u8 , name : "USB controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "UHCI" } , ProgIf { id : 16u8 , name : "OHCI" } , ProgIf { id : 32u8 , name : "EHCI" } , ProgIf { id : 48u8 , name : "XHCI" } , ProgIf { id : 64u8 , name : "USB4 Host Interface" } , ProgIf { id : 128u8 , name : "Unspecified" } , ProgIf { id : 254u8 , name : "USB Device" }] } , Subclass { class_id : 12u8 , id : 4u8 , name : "Fibre Channel" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 5u8 , name : "SMBus" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 6u8 , name : "InfiniBand" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 7u8 , name : "IPMI Interface" , prog_ifs : & [ProgIf { id : 0u8 , name : "SMIC" } , ProgIf { id : 1u8 , name : "KCS" } , ProgIf { id : 2u8 , name : "BT (Block Transfer)" }] } , Subclass { class_id : 12u8 , id : 8u8 , name : "SERCOS interface" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 9u8 , name : "CANBUS" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 10u8 , name : "MIPI I3C" , prog_ifs : & [] } , Subclass { class_id : 12u8 , id : 128u8 , name : "Serial bus controller" , prog_ifs : & [] }] }),
        (10, Class { id : 10u8 , name : "Docking station" , subclasses : & [Subclass { class_id : 10u8 , id : 0u8 , name : "Generic Docking Station" , prog_ifs : & [] } , Subclass { class_id : 10u8 , id : 128u8 , name : "Docking Station" , prog_ifs : & [] }] }),
        (6, Class { id : 6u8 , name : "Bridge" , subclasses : & [Subclass { class_id : 6u8 , id : 0u8 , name : "Host bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 1u8 , name : "ISA bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 2u8 , name : "EISA bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 3u8 , name : "MicroChannel bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 4u8 , name : "PCI bridge" , prog_ifs : & [ProgIf { id : 0u8 , name : "Normal decode" } , ProgIf { id : 1u8 , name : "Subtractive decode" }] } , Subclass { class_id : 6u8 , id : 5u8 , name : "PCMCIA bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 6u8 , name : "NuBus bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 7u8 , name : "CardBus bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 8u8 , name : "RACEway bridge" , prog_ifs : & [ProgIf { id : 0u8 , name : "Transparent mode" } , ProgIf { id : 1u8 , name : "Endpoint mode" }] } , Subclass { class_id : 6u8 , id : 9u8 , name : "Semi-transparent PCI-to-PCI bridge" , prog_ifs : & [ProgIf { id : 64u8 , name : "Primary bus towards host CPU" } , ProgIf { id : 128u8 , name : "Secondary bus towards host CPU" }] } , Subclass { class_id : 6u8 , id : 10u8 , name : "InfiniBand to PCI host bridge" , prog_ifs : & [] } , Subclass { class_id : 6u8 , id : 128u8 , name : "Bridge" , prog_ifs : & [] }] }),
        (0, Class { id : 0u8 , name : "Unclassified device" , subclasses : & [Subclass { class_id : 0u8 , id : 0u8 , name : "Non-VGA unclassified device" , prog_ifs : & [] } , Subclass { class_id : 0u8 , id : 1u8 , name : "VGA compatible unclassified device" , prog_ifs : & [] } , Subclass { class_id : 0u8 , id : 5u8 , name : "Image coprocessor" , prog_ifs : & [] }] }),
        (19, Class { id : 19u8 , name : "Non-Essential Instrumentation" , subclasses : & [] }),
        (255, Class { id : 255u8 , name : "Unassigned class" , subclasses : & [] }),
        (3, Class { id : 3u8 , name : "Display controller" , subclasses : & [Subclass { class_id : 3u8 , id : 0u8 , name : "VGA compatible controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "VGA controller" } , ProgIf { id : 1u8 , name : "8514 controller" }] } , Subclass { class_id : 3u8 , id : 1u8 , name : "XGA compatible controller" , prog_ifs : & [] } , Subclass { class_id : 3u8 , id : 2u8 , name : "3D controller" , prog_ifs : & [] } , Subclass { class_id : 3u8 , id : 128u8 , name : "Display controller" , prog_ifs : & [] }] }),
        (5, Class { id : 5u8 , name : "Memory controller" , subclasses : & [Subclass { class_id : 5u8 , id : 0u8 , name : "RAM memory" , prog_ifs : & [] } , Subclass { class_id : 5u8 , id : 1u8 , name : "FLASH memory" , prog_ifs : & [] } , Subclass { class_id : 5u8 , id : 2u8 , name : "CXL" , prog_ifs : & [ProgIf { id : 0u8 , name : "CXL Memory Device - vendor specific" } , ProgIf { id : 16u8 , name : "CXL Memory Device (CXL 2.0 or later)" }] } , Subclass { class_id : 5u8 , id : 128u8 , name : "Memory controller" , prog_ifs : & [] }] }),
        (1, Class { id : 1u8 , name : "Mass storage controller" , subclasses : & [Subclass { class_id : 1u8 , id : 0u8 , name : "SCSI storage controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "Vendor specific" } , ProgIf { id : 17u8 , name : "SCSI storage device (SOP target port using PQI)" } , ProgIf { id : 18u8 , name : "SCSI controller (SOP target port using PQI)" } , ProgIf { id : 19u8 , name : "SCSI storage device & controller (SOP target port using PQI)" } , ProgIf { id : 33u8 , name : "SCSI storage device (SOP target port using NVMe)" }] } , Subclass { class_id : 1u8 , id : 1u8 , name : "IDE interface" , prog_ifs : & [ProgIf { id : 0u8 , name : "ISA Compatibility mode-only controller" } , ProgIf { id : 5u8 , name : "PCI native mode-only controller" } , ProgIf { id : 10u8 , name : "ISA Compatibility mode controller, supports both channels switched to PCI native mode" } , ProgIf { id : 15u8 , name : "PCI native mode controller, supports both channels switched to ISA compatibility mode" } , ProgIf { id : 128u8 , name : "ISA Compatibility mode-only controller, supports bus mastering" } , ProgIf { id : 133u8 , name : "PCI native mode-only controller, supports bus mastering" } , ProgIf { id : 138u8 , name : "ISA Compatibility mode controller, supports both channels switched to PCI native mode, supports bus mastering" } , ProgIf { id : 143u8 , name : "PCI native mode controller, supports both channels switched to ISA compatibility mode, supports bus mastering" }] } , Subclass { class_id : 1u8 , id : 2u8 , name : "Floppy disk controller" , prog_ifs : & [] } , Subclass { class_id : 1u8 , id : 3u8 , name : "IPI bus controller" , prog_ifs : & [] } , Subclass { class_id : 1u8 , id : 4u8 , name : "RAID bus controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "Vendor specific" }] } , Subclass { class_id : 1u8 , id : 5u8 , name : "ATA controller" , prog_ifs : & [ProgIf { id : 32u8 , name : "ADMA single stepping" } , ProgIf { id : 48u8 , name : "ADMA continuous operation" }] } , Subclass { class_id : 1u8 , id : 6u8 , name : "SATA controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "Vendor specific" } , ProgIf { id : 1u8 , name : "AHCI 1.0" } , ProgIf { id : 2u8 , name : "Serial Storage Bus" }] } , Subclass { class_id : 1u8 , id : 7u8 , name : "Serial Attached SCSI controller" , prog_ifs : & [ProgIf { id : 1u8 , name : "Serial Storage Bus" }] } , Subclass { class_id : 1u8 , id : 8u8 , name : "Non-Volatile memory controller" , prog_ifs : & [ProgIf { id : 1u8 , name : "NVMHCI" } , ProgIf { id : 2u8 , name : "NVM Express" } , ProgIf { id : 3u8 , name : "NVM Express administrative controller" }] } , Subclass { class_id : 1u8 , id : 9u8 , name : "Universal Flash Storage controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "Vendor specific" } , ProgIf { id : 1u8 , name : "UFSHCI" }] } , Subclass { class_id : 1u8 , id : 128u8 , name : "Mass storage controller" , prog_ifs : & [] }] }),
        (13, Class { id : 13u8 , name : "Wireless controller" , subclasses : & [Subclass { class_id : 13u8 , id : 0u8 , name : "IRDA controller" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 1u8 , name : "Consumer IR controller" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 16u8 , name : "RF controller" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 17u8 , name : "Bluetooth" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 18u8 , name : "Broadband" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 32u8 , name : "802.11a 5 GHz controller" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 33u8 , name : "802.11b 2.4 GHz controller" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 64u8 , name : "Cellular controller/modem" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 65u8 , name : "Cellular controller/modem plus Ethernet (802.11)" , prog_ifs : & [] } , Subclass { class_id : 13u8 , id : 128u8 , name : "Wireless controller" , prog_ifs : & [] }] }),
        (15, Class { id : 15u8 , name : "Satellite communications controller" , subclasses : & [Subclass { class_id : 15u8 , id : 1u8 , name : "Satellite TV controller" , prog_ifs : & [] } , Subclass { class_id : 15u8 , id : 2u8 , name : "Satellite audio communication controller" , prog_ifs : & [] } , Subclass { class_id : 15u8 , id : 3u8 , name : "Satellite voice communication controller" , prog_ifs : & [] } , Subclass { class_id : 15u8 , id : 4u8 , name : "Satellite data communication controller" , prog_ifs : & [] }] }),
        (14, Class { id : 14u8 , name : "Intelligent controller" , subclasses : & [Subclass { class_id : 14u8 , id : 0u8 , name : "I2O" , prog_ifs : & [] }] }),
        (16, Class { id : 16u8 , name : "Encryption controller" , subclasses : & [Subclass { class_id : 16u8 , id : 0u8 , name : "Network and computing encryption device" , prog_ifs : & [] } , Subclass { class_id : 16u8 , id : 16u8 , name : "Entertainment encryption device" , prog_ifs : & [] } , Subclass { class_id : 16u8 , id : 128u8 , name : "Encryption controller" , prog_ifs : & [] }] }),
        (64, Class { id : 64u8 , name : "Coprocessor" , subclasses : & [] }),
        (4, Class { id : 4u8 , name : "Multimedia controller" , subclasses : & [Subclass { class_id : 4u8 , id : 0u8 , name : "Multimedia video controller" , prog_ifs : & [] } , Subclass { class_id : 4u8 , id : 1u8 , name : "Multimedia audio controller" , prog_ifs : & [] } , Subclass { class_id : 4u8 , id : 2u8 , name : "Computer telephony device" , prog_ifs : & [] } , Subclass { class_id : 4u8 , id : 3u8 , name : "Audio device" , prog_ifs : & [ProgIf { id : 0u8 , name : "HDA compatible" } , ProgIf { id : 128u8 , name : "HDA compatible with vendor specific extensions" }] } , Subclass { class_id : 4u8 , id : 128u8 , name : "Multimedia controller" , prog_ifs : & [] }] }),
        (2, Class { id : 2u8 , name : "Network controller" , subclasses : & [Subclass { class_id : 2u8 , id : 0u8 , name : "Ethernet controller" , prog_ifs : & [ProgIf { id : 1u8 , name : "Ethernet Controller with IDPF Compliant Interface" }] } , Subclass { class_id : 2u8 , id : 1u8 , name : "Token ring network controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 2u8 , name : "FDDI network controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 3u8 , name : "ATM network controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 4u8 , name : "ISDN controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 5u8 , name : "WorldFip controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 6u8 , name : "PICMG controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 7u8 , name : "Infiniband controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 8u8 , name : "Fabric controller" , prog_ifs : & [] } , Subclass { class_id : 2u8 , id : 128u8 , name : "Network controller" , prog_ifs : & [] }] }),
        (7, Class { id : 7u8 , name : "Communication controller" , subclasses : & [Subclass { class_id : 7u8 , id : 0u8 , name : "Serial controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "8250" } , ProgIf { id : 1u8 , name : "16450" } , ProgIf { id : 2u8 , name : "16550" } , ProgIf { id : 3u8 , name : "16650" } , ProgIf { id : 4u8 , name : "16750" } , ProgIf { id : 5u8 , name : "16850" } , ProgIf { id : 6u8 , name : "16950" }] } , Subclass { class_id : 7u8 , id : 1u8 , name : "Parallel controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "SPP" } , ProgIf { id : 1u8 , name : "BiDir" } , ProgIf { id : 2u8 , name : "ECP" } , ProgIf { id : 3u8 , name : "IEEE1284" } , ProgIf { id : 254u8 , name : "IEEE1284 Target" }] } , Subclass { class_id : 7u8 , id : 2u8 , name : "Multiport serial controller" , prog_ifs : & [] } , Subclass { class_id : 7u8 , id : 3u8 , name : "Modem" , prog_ifs : & [ProgIf { id : 0u8 , name : "Generic" } , ProgIf { id : 1u8 , name : "Hayes/16450" } , ProgIf { id : 2u8 , name : "Hayes/16550" } , ProgIf { id : 3u8 , name : "Hayes/16650" } , ProgIf { id : 4u8 , name : "Hayes/16750" }] } , Subclass { class_id : 7u8 , id : 4u8 , name : "GPIB controller" , prog_ifs : & [] } , Subclass { class_id : 7u8 , id : 5u8 , name : "Smard Card controller" , prog_ifs : & [] } , Subclass { class_id : 7u8 , id : 128u8 , name : "Communication controller" , prog_ifs : & [] }] }),
        (11, Class { id : 11u8 , name : "Processor" , subclasses : & [Subclass { class_id : 11u8 , id : 0u8 , name : "386" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 1u8 , name : "486" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 2u8 , name : "Pentium" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 16u8 , name : "Alpha" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 32u8 , name : "Power PC" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 48u8 , name : "MIPS" , prog_ifs : & [] } , Subclass { class_id : 11u8 , id : 64u8 , name : "Co-processor" , prog_ifs : & [] }] }),
        (17, Class { id : 17u8 , name : "Signal processing controller" , subclasses : & [Subclass { class_id : 17u8 , id : 0u8 , name : "DPIO module" , prog_ifs : & [] } , Subclass { class_id : 17u8 , id : 1u8 , name : "Performance counters" , prog_ifs : & [] } , Subclass { class_id : 17u8 , id : 16u8 , name : "Communication synchronizer" , prog_ifs : & [] } , Subclass { class_id : 17u8 , id : 32u8 , name : "Signal processing management" , prog_ifs : & [] } , Subclass { class_id : 17u8 , id : 128u8 , name : "Signal processing controller" , prog_ifs : & [] }] }),
        (18, Class { id : 18u8 , name : "Processing accelerators" , subclasses : & [Subclass { class_id : 18u8 , id : 0u8 , name : "Processing accelerators" , prog_ifs : & [] } , Subclass { class_id : 18u8 , id : 1u8 , name : "SNIA Smart Data Accelerator Interface (SDXI) controller" , prog_ifs : & [] }] }),
        (9, Class { id : 9u8 , name : "Input device controller" , subclasses : & [Subclass { class_id : 9u8 , id : 0u8 , name : "Keyboard controller" , prog_ifs : & [] } , Subclass { class_id : 9u8 , id : 1u8 , name : "Digitizer Pen" , prog_ifs : & [] } , Subclass { class_id : 9u8 , id : 2u8 , name : "Mouse controller" , prog_ifs : & [] } , Subclass { class_id : 9u8 , id : 3u8 , name : "Scanner controller" , prog_ifs : & [] } , Subclass { class_id : 9u8 , id : 4u8 , name : "Gameport controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "Generic" } , ProgIf { id : 16u8 , name : "Extended" }] } , Subclass { class_id : 9u8 , id : 128u8 , name : "Input device controller" , prog_ifs : & [] }] }),
        (8, Class { id : 8u8 , name : "Generic system peripheral" , subclasses : & [Subclass { class_id : 8u8 , id : 0u8 , name : "PIC" , prog_ifs : & [ProgIf { id : 0u8 , name : "8259" } , ProgIf { id : 1u8 , name : "ISA PIC" } , ProgIf { id : 2u8 , name : "EISA PIC" } , ProgIf { id : 16u8 , name : "IO-APIC" } , ProgIf { id : 32u8 , name : "IO(X)-APIC" }] } , Subclass { class_id : 8u8 , id : 1u8 , name : "DMA controller" , prog_ifs : & [ProgIf { id : 0u8 , name : "8237" } , ProgIf { id : 1u8 , name : "ISA DMA" } , ProgIf { id : 2u8 , name : "EISA DMA" }] } , Subclass { class_id : 8u8 , id : 2u8 , name : "Timer" , prog_ifs : & [ProgIf { id : 0u8 , name : "8254" } , ProgIf { id : 1u8 , name : "ISA Timer" } , ProgIf { id : 2u8 , name : "EISA Timers" } , ProgIf { id : 3u8 , name : "HPET" }] } , Subclass { class_id : 8u8 , id : 3u8 , name : "RTC" , prog_ifs : & [ProgIf { id : 0u8 , name : "Generic" } , ProgIf { id : 1u8 , name : "ISA RTC" }] } , Subclass { class_id : 8u8 , id : 4u8 , name : "PCI Hot-plug controller" , prog_ifs : & [] } , Subclass { class_id : 8u8 , id : 5u8 , name : "SD Host controller" , prog_ifs : & [] } , Subclass { class_id : 8u8 , id : 6u8 , name : "IOMMU" , prog_ifs : & [] } , Subclass { class_id : 8u8 , id : 7u8 , name : "Root Complex Event Collector" , prog_ifs : & [] } , Subclass { class_id : 8u8 , id : 128u8 , name : "System peripheral" , prog_ifs : & [] } , Subclass { class_id : 8u8 , id : 153u8 , name : "Timing Card" , prog_ifs : & [ProgIf { id : 1u8 , name : "TAP Timing Card" }] }] }),
    ],
};
